\ MscDrv.fth - USB Mass Storage Class operations with DMA option.

((
Copyright (c) 2008
Synics AG

and

Copyright (c) 2008, 2009, 2010, 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20170606 BJC012 In ?MSCDRV shows console output when card inserted or removed.
                this output is now conditional on dbgMSC?
20130909 SFP011 More updates for USBCV20 compliance.
		The code now passes the command verifier.
20130904 SFP010 Updated for USBCV20 compliance.
20101115 FVD009 Added USBdisk? to enable access.
20101101 SFP008	Refactored start up code.
20090105 SFP007 Renamed INITCACHE to INITSECCACHE in FATfiler.
		Added InitSecCache to drive removal detection.
20081122 MVD006 Interlocked CSW in DMA driver - needed by Mac.
20081120 SFP005 Improved card insertion recovery.
20081119 SFP004 Card detect and write protection installed.
20081114 SFP003 Flushed FAT file system cache before read/write.
20081112 SFP002 DMA version.
20080731 SFP001 Initial release.
))

only forth definitions
decimal

\ =========
\ *! mscdrv
\ *T Mass Storage Class (MSC) driver
\ =========
\ *P The Mass Storage Class (MSC) driver uses the USB Mass Storage
\ ** Class Bulk Only Transport (BOT) specification. This, in turn,
\ ** simulates a SCSI disc controller and so many SCSI terms
\ ** are used here.

\ *P The driver relies on user-supplied
\ ** sector read/write routines. In order to reduce the time spent
\ ** in the USB interrupt handler, the sector read/write routines
\ ** are accessed by a separate task. A consequence of this is that
\ ** if the USB and the device's application share access to the
\ ** drive, e.g. through the FAT file system, semaphores *\b{must}
\ ** be used for disk access.

\ *P Note that the file *\i{%FATfiler%\SDspi.fth} of 19 November 2008
\ ** or later is required for use with SD/MMC cards that use the
\ ** generic SPI driver.

\ *P The file system needs these words to access the mass storage:
\ *D SecRead  ( addr sector dev -- ) Reads a sector from the specified device.
\ **          *\fo{THROW}s on error. Use *\i{dev=0}.
\ *D SecWrite ( addr sector dev -- ) Writes a sector to the specified device.
\ **          *\fo{THROW}s on error. Use *\i{dev=0}.
\ *D MassInit ( -- ) Initializes mass storage access.
\ *D MassTerm ( -- ) Terminates mass storage access.
\ *D CDin?    ( -- flag ) Return true if card/device ready.
\ *D WPin?    ( -- flag ) Return true if card/device write protected.

\ *P To simplify the code, the raw disc read/write interface treats all
\ ** read/write errors as fatal, and *\fo{THROW}s on error. Retries
\ ** should be accommodated within the sector read/write code.

\ *P The code is sensitive to the condition of the equate
\ ** *\fo{usbDMA?}. We *\b{strongly} recommend that this is set
\ ** non-zero if your hardware supports DMA for USB transfers.


\ ********************
\ *S MSC state machine
\ ********************
\ *P The Mass Storage Class (MSC) bulk endpoint actions are
\ ** controlled by a state machine.

interpreter
: [sm		\ -- 0
\ *G Starts the definition of a state machine's states.
  0  ;
: smState	\ n -- n+1
\ *G Defines the next state as an *\fo{EQU} and increments the
\ ** state number.
  dup equ  1+  ;
: sm]		\ n --
\ *G Finishes the state machine and defines an equate of the
\ ** number of states.
  equ  ;
target

\ *P The MSC bulk endpoint state machine.
\ *[
[sm
  smState smCBW		\ wait for a CBW from host, idle state
  smState smDOUT	\ receive data blocks (sectors) from host
  smState smDIN		\ transmit data blocks (sectors) to host
  smState smCSW		\ send CSW to host
  smstate smCIN		\ send command response to host
  smState smStalled	\ both bulk endpoints stalled
sm] #MSCsm	\ -- n
\ *]

smCBW value MSCsm#	\ -- state
\ *G MSC Bulk endpoint state.

dbgMSC? [if]
: .MSCstate	\ --
\ *G Display the current state of the MSC state machine.
  case  MSCsm#
    smCBW	of  ." smCBW "  endof
    smDOUT	of  ." smDOUT "  endof
    smDIN	of  ." smDIN "  endof
    smCSW	of  ." smCSW "  endof
    smCIN       of  ." smCIN "  endof
    smStalled	of  ." smStalled "  endof
      ." ??? "
  endcase
;
[then]


\ ***********
\ *S MSC data
\ ***********

struct /DataPtr	\ -- len
\ *G describes the layout of length/addr data pointers.
  int .len	\ holds length
  int .addr	\ holds address
end-struct

2variable mscBulkData	\ -- addr
\ *G The data pointer for the remaining data of the bulk sector transfer.
\ ** Note that this may be only part of the transfer, especially
\ ** for multi-sector disk reads and writes.
variable mscBulkLen	\ -- addr
\ *G Holds the remaining length of the complete transfer. Before
\ ** the transfer starts, the data corresponds to the Dx value
\ ** in the "13 conditions" tests.
variable mscHostLen	\ -- addr
\ *G Holds the transfer size read from the CBW data block. This
\ ** corresponds to the Hx value in the "13 conditions" tests.
: initBulk	\ addr len --
\ *G Set the data pointers for a bulk transfer.
  dup mscBulkLen !  mscBulkData 2!
;

struct /CBW	\ -- len
\ *G SCSI Command Block Wrapper (CBW). The CBW fields are little
\ ** endian for USB, but the CDB fields are big-endian SCSI data.
  int  cbw.Sig			\ signature
  int  cbw.Tag			\ tag
  int  cbw.DataLen		\ length of data to follow
  byte cbw.bFlags		\ flags, bit 7 set=IN
  byte cbw.bLun			\ LUN, always 0 for us
  byte cbw.bCDBlen		\ number of bytes in CDB
  #16 field cbw.CDB		\ command data block, first byte = command
end-struct

struct /CSW	\ -- len
\ *G SCSI Command Status Wrapper (CSW). Little-endian data for USB.
  int  csw.Sig			\ signature
  int  csw.Tag			\ tag, copy of that in CBW
  int  csw.Residue		\ remaining data length expected, 0=done
  byte csw.bStatus		\ status code, 0=ok, 1=failed, 2=phase
end-struct


\ *P According to the specs (see the MSC overview), there should
\ ** only be one CBW outstanding until the corresponding CSW has
\ ** been sent. Consequently, we do not need to queue commands,
\ ** and only need one set of buffers.

[sections URAMP  uAlign
usbDMA? [if]
MscBoutEP BulkDD:
\ *G Bulk OUT DMA descriptor.
MscBinEP BulkDD:
\ *G Bulk IN DMA descriptor.
[then]
UAlign  /MaxPacket buffer: CBWbuff	\ -- addr
\ *G Holds a packet expected to be the next CBW. Includes the CDB.
UAlign  /MaxPacket buffer: MSCbuff	\ -- addr
\ *G Holds miscellaneous data in or out that is not sector data.
UAlign  /CSW buffer: CSWbuff	\ -- addr
\ *G Holds the CSW being assembled.
UAlign  #512 buffer: USBSecBuff	\ -- addr
\ *G Single sector buffer.
sections]

CBWbuff cbw.CDB constant CDBbuff	\ -- addr
\ *G The base address of the CDB in *\fo{CBWbuff}. The first byte
\ ** contains the operation code.

0x43425355 equ CBWsig#		\ 55 is first byte, "USBC"
0x53425355 equ CSWsig#		\ 55 is first byte, "USBS"

\ CSW status codes
$00 equ CSWok
$01 equ CSWfailed
$02 equ CSWphaseErr

0 value SenseCode	\ -- 00kkccqq
\ *G The REQUEST SENSE command returns data in the KEY, ASC and
\ ** ASQ fields. These are merged into the low 24 bits of
\ ** *\fo{SenseCode}.
\ *[
0x030C00 equ SenseWriteError	\ Card fail
0x031100 equ SenseReadError	\ Card fail
0x052000 equ SenseInvalidOp	\ bad command
0x052400 equ SenseInvalidCDB	\ command format bad
0x023A00 equ SenseNotReady	\ not ready, no card
0x0B0C00 equ SenseWriteFail	\ USB fail timeout
0x0B1100 equ SenseReadFail	\ USB fail timeout
\ *]

\ SCSI commands - M=Mandatory in SCSI command set,
\                 U=mandatory for USB
\                 O=Optional for USB
\                 W=Required by Windows
\                 X=Required by Mac/Linux
\ See MSC compliance tests for more details.
0x00 equ cmdTestUnitReady		\ M U
0x03 equ cmdRequestSense		\ M U
0x04 equ cmdFormatUnit			\ M
0x08 equ cmdRead6			\ M	(deprecated)
0x0A equ cmdWrite6			\   O	(deprecated)
0x12 equ cmdInquiry			\ M U
0x15 equ cmdModeSelect6			\   O
0x1A equ cmdModeSense6			\   O/X
0x1B equ cmdStartStopUnit		\   U/O
0x1D equ cmdSendDiagnostic		\ M
0x1E equ cmdMediaRemoval		\   O	(Prevent Allow Medium Removal)
0x23 equ cmdReadFormatCapacities	\   O	CD/DVD option
0x25 equ cmdReadCapacity10		\ M U
0x28 equ cmdRead10			\ M U
0x2A equ cmdWrite10			\   U
0x2F equ cmdVerify10			\   W
0x55 equ cmdModeSelect10		\   O
0x5A equ cmdModeSense10			\   O
0xA0 equ cmdReportLuns			\ M
0xA8 equ cmdRead12			\   O
0xAA equ cmdWrite12			\   O


\ ********
\ *S Tools
\ ********

: StallMSCin	\ --
\ *G Stall the bulk IN endpoint.
[ dbgMSC? ] [if] ." StallMSCin " [then]
  MscBinEP USBSetStallEP
;

: StallMSCout	\ --
\ *G Stall the bulk OUT endpoint.
[ dbgMSC? ] [if] ." StallMSCout " .MSCstate [then]
  MscBoutEP USBSetStallEP
;

: BOTStall	\ --
\ *G Stall the transaction. Which endpoint to stall is determined
\ ** by looking at the transfer direction intended by the host.
[ dbgMSC? ] [if] ." BOT " [then]
  CBWbuff dup cbw.bFlags c@ $80 and	\ -- addr dir
  swap cbw.DataLen lel@ 0= or
  if  StallMSCin  else  StallMSCout  endif
;
\ Should we actually stall an in/out request with length=0?

usbDMA? 0= [if]
: readMSCOut	\ caddr -- len
\ *G Read the bulk OUT endpoint into a buffer of at least
\ ** *\fo{/MaxPacket} bytes and return the number read.
  MscBoutEP swap USBReadEP
;

: writeMSCIn	\ caddr len --
\ *G Write the buffer to the bulk IN endpoint.
  MscBinEP -rot USBWriteEP drop
;
[then]


\ ***********************
\ *S Disk Read/Write task
\ ***********************
\ *P Disk read and write is controlled by a task. This is done to
\ ** prevent the USB interrupt taking too long.
\ *D SecRead  ( addr sector dev -- ) Reads a sector from the specified device.
\ **          *\fo{THROW}s on error.
\ *D SecWrite ( addr sector dev -- ) Writes a sector to the specified device.
\ **          *\fo{THROW}s on error.

-1 value USBdisk?	\ -- flag ; FVD009
\ *G Returns true if the USB drive should be available to the USB host.
0 value DriveBusy?	\ -- x
\ *G Returns the current drive command and direction.
$100 DirIn or equ SecReadCmd	\ -- x
\ *G Indicates that sectors are being read.
$100 DirOut or equ SecWriteCmd	\ --
\ *G Indicates that sectors are being written.
0 value DriveComplete	\ -- x
\ *G Set non-zero when the USB system has finished a command.
0 value FirstSector	\ -- u
\ *G First sector to be read/written.
0 value NumSectors	\ -- u
\ *G Number of sectors to be read/written.
0 value DriveStatus	\ -- x
\ *G Holds the drive action completion status. This is read
\ ** by the USB interrupt handler to decide how to terminate
\ ** the action and is cleared when termination has been handled.
\ ** Negative values indicate errors, 0 is success
0 equ DrvOK	\ -- x
\ *G Drive status good.
-128 equ DrvWriteFail	\ --
\ *G Fatal write error.
-129 equ DrvReadFail
\ *G Fatal read error.
-130 equ DrvBadCmd
\ *G Invalid drive command.
-131 equ DrvAborted.
\ *G Command aborted by USB
-132 equ DrvUSBto
\ *G Command aborted by USB timeout.

compiler
: setDriveCmd	\ x --
\ *G Set the drive command and clear the status.
  DrvOK -> DriveStatus  -> DriveBusy?
;
target

: ResetUSBdisk	\ --
  0 -> FirstSector  0 -> NumSectors
  0 setDriveCmd
  -1 -> DriveComplete
;

: Sector<>USB	\ -- ior ; 0=success, nz=timeout
\ *G Send or receive a 512 byte sector data to/from *\fo{USBSecBuff}.
\ ." s"
  1000 later >r
  USBSecBuff #512 mscBulkData 2!		\ Give it somewhere to go
  begin
    r@ timedout? if
\ ." T"
       r> drop  -1 exit
    endif
    mscBulkData .len @ 0 >
   while
    pause  DriveComplete
  until then
  begin
    r@ timedout? if
\ ." t"
       r> drop  -1 exit
    endif
    mscBulkData .addr @
   while
    pause  DriveComplete
  until then
\ ." e " out @ 66 > if cr endif
  r> drop  0
;

: -BulkData	\ --
\ *G Acknowledge a sector transfer by zeroing the transfer address
\ ** and length.
  0 0 mscBulkData 2!
;

: Disk>USB	\ --
\ *G Read the given sectors from the disk.
\ cr cr ." RS " FirstSector u. NumSectors u.
  FirstSector NumSectors bounds ?do	\ -- status
\ [ dbgMSC? ] [if]  cr ." R" i .  [then]
\ ." R" i .
    usbLED2on
    USBSecBuff i 0 ['] SecRead catch if
      3drop				\ clean up after THROW
      SenseReadError -> SenseCode	\ for SCSI responses
      DrvReadFail -> DriveStatus	\ to abandon disk transfer
      usbLED2off
      unloop exit
    endif
    usbLED2off
    Sector<>USB if			\ wait for transfer to complete
      SenseReadFail -> SenseCode	\ for SCSI responses
      DrvReadFail -> DriveStatus	\ to abandon disk transfer
    endif
    DriveStatus DriveComplete or	\ get out if aborted
    if  unloop exit  endif
\ ." r"
  loop
  DrvOK -> DriveStatus
;

: USB>Disk	\ --
\ *G Write the given sectors to the disk.
\ cr cr ." WS " FirstSector . NumSectors .
  FirstSector NumSectors bounds ?do	\ -- status
[ dbgMSC? ] [if] ." W" i . [then]
\ ." W" i .
    Sector<>USB if			\ wait for transfer to complete
      SenseWriteFail -> SenseCode	\ for SCSI responses
      DrvWriteFail -> DriveStatus	\ to abandon disk transfer
    else
      usbLED2on
      USBSecBuff i 0 ['] SecWrite catch if
        3drop				\ clean up after THROW
        SenseWriteError -> SenseCode	\ for SCSI responses
        DrvWriteFail -> DriveStatus	\ to abandon disk transfer
      endif
      usbLED2off
    endif
    DriveStatus DriveComplete or	\ get out if aborted
    if  unloop exit  endif
\ ." w"
  loop
  DrvOK -> DriveStatus
;

: DriveIn?	\ -- flag
\ *G Return true if the drive is in.
  50 ms CDin? if
    50 ms  CDin? if
      50 ms  CDin? exit
    endif
  endif
  0
;

: ?MSCdrive	\ --
\ *G Wait for drive insertion after removal.
  CDin? 0= if				\ card detect
[ dbgMSC? ] [if]
      cr ." Card removed"
[endif]
    initSecCache			\ clean sector cache ; SFP007
    begin  DriveIn?  until
[ dbgMSC? ] [if]
      cr ." Card inserted"
[endif]
[defined] initFATfs [if]		\ check for file system
    initFATfs drop
[else]					\ or just storage unit
    MassInit
[then]
    ResetUSBdisk			\ SFP005
  endif
;

: WaitSecCmd	\ --
\ *G Wait for a sector command
  begin  ?MSCdrive DriveBusy? 0= while  pause  repeat
;

: WaitDrvComplete	\ --
\ *G Wait for the drive action on the USB side to complete.
  begin  DriveComplete 0= while  pause  repeat
  0 -> DriveComplete
;

: doSectorRW	\ --
  consoleIO decimal
[ dbgMSC? ] [if]  cr ." USB disk transfer task running"  [then]
  ResetUSBdisk
  begin
    ?StackEmpty
    WaitSecCmd  0 -> DriveComplete
\ cr ." SecCmd=" DriveBusy? .word space FirstSector .  NumSectors .
[defined] SyncDrive [if] SyncDrive [then]	\ Flush FAT sector cache
    case DriveBusy?
      SecReadCmd of  Disk>USB  endof
      SecWriteCmd of  USB>Disk  endof
        SenseInvalidOp -> SenseCode
        DrvBadCmd -> DriveStatus
    endcase
\ cr ." SecStat=" DriveStatus .
    WaitDrvComplete
  again
;

Task USBdiskTask	\ -- addr
\ *G The task control block for the MSC sector transfer task.

0 value USBdiskTask?	\ -- x
\ *G Returns non-zero when the task is running.

: +USBdiskTask	\ --
\ *G Start the USB disk task if not already running.
  USBdiskTask? 0= if
    ['] doSectorRW  USBdiskTask initiate
    -1 -> USBdiskTask?
  endif
;

: -USBdiskTask	\ --
\ *G Stop the USB disk task if not already stopped.
  USBdiskTask? if
     USBdiskTask terminate  0 -> USBdiskTask?
  endif
;


\ *****************
\ *S CSW operations
\ *****************

: NoResidue	\ --
\ *G Zero the *\fo{CSWbuff csw.Residue} field.
  0 CSWbuff csw.Residue !		\ this field is aligned.
;

: initCSW	\ --
\ *G Set up the CSW from the CBW. The residue field is set to the
\ ** CBW data length field. For actions that complete immediately,
\ ** this should be cleared by *\fo{NoResidue} above.
  CBWbuff CSWbuff			\ -- src dest
  over cbw.DataLen lel@ mscHostLen !	\ host data length ; SFP011
  CSWsig# over csw.Sig lel!
  over cbw.tag @ over csw.Tag !		\ endian independent
  swap cbw.DataLen @ over csw.Residue !	\ endian independent
  CSWok swap csw.bStatus c!
;

: sendCSW	\ --
\ *G Send the CSW as is.
  smCSW -> MSCsm#			\ move the state machine
;
compiler
: sendCSW	\ --
  smCSW -> MSCsm#			\ move the state machine
;
target

: sendCSWfail	\ --
  CSWfailed CSWbuff csw.bStatus c!  SendCSW
;

: sendCSWphase	\ --
  CSWphaseErr CSWbuff csw.bStatus c!  SendCSW
;

: badRWphase	\ --
\ *G Perform the bad R/W setup actions, returning a phase error
\ ** status (0x02).
  SenseInvalidCDB -> SenseCode
  BOTstall sendCSWphase
;

: badRWfail	\ --
\ *G Perform the bad R/W setup actions, returning a fail status
\ ** (0x01).
  SenseInvalidCDB -> SenseCode
  BOTstall sendCSWfail
;


\ ***************
\ *S The 13 cases
\ ***************
\ *P The USB specifications define the host and device transactions
\ ** in terms of 13 cases. The file *\i{usbmassbulk_10.pdf}
\ ** contains the gory details and can be downloaded from
\ ** *\f{www.usb.org}. You will probably use that document
\ ** in conjunction with *\i{INF-8070.pdf} to be found at
\ ** *\f{ftp://ftp.seagate.com/sff/INF-8070.PDF}.

\ SFP011...

: mscHostDir@	\ -- $00/$80
\ *G Return the Host direction, where $80=IN to host.
  CBWbuff cbw.bFlags c@ $80 and
;

compiler
: mscHostLen@	\ -- len
\ *G Return the host length from the CBW.
  mscHostLen @
;
target

: NoMscDevData?	\ -- ior ; 0=success
\ *G Checks cases 1, 4, and 9. On an error (Hx>0), a pipe is
\ ** stalled and a failure CSW sent. On success, no action is
\ ** taken.
  mscHostLen@				\ cases 4, 9
  if  badRWfail 1 exit  endif
  0					\ case 1, Hn=Dn=0
;

: MscDevDataIN-1?	\ caddr len -- caddr len' 0 | ior ; 0=success
\ *G Checks cases 2, 5, 6, 7, and 10. On error, appropriate action
\ ** is performed. On success, no action is taken. This version is
\ ** used for short responses less than *\fo{/MaxPacket}, when
\ ** data is transmitted for case 5.
\ case 2: Hn<Di - stall, 02h status
\ case 5: Hi>Di - stall, 01h status
\ case 6: Hi=Di - success
\ case 7: Hi<Di - stall, 02h status
\ case 10: Ho<>Di (dir error) - stall, 02h status
  mscHostDir@ $080 <>			\ case 10
  if  2drop 1  badRWphase exit  endif
  mscHostLen@ over < 			\ case 2, 7
  if  2drop 1  badRWphase exit  endif
  mscHostLen@ over > if
    mscHostLen@ over - 			\ new residue
    CSWbuff csw.Residue lel!
    drop mscHostLen@			\ case 5
  else
    NoResidue
  endif
  0					\ case 6
;

: MscDevDataIN-2?	\ len -- ior ; 0=success
\ *G Checks cases 2, 5, 6, 7, and 10. On error, appropriate action
\ ** is performed. On success, no action is taken. This version
\ ** stalls and sends no data for case 5.
\ case 2: Hn<Di - stall, 02h status
\ case 5: Hi>Di - stall, 01h status
\ case 6: Hi=Di - success
\ case 7: Hi<Di - stall, 02h status
\ case 10: Ho<>Di (dir error) - stall, 02h status
  mscHostDir@ $080 <>			\ case 10
  if  drop 1  badRWphase exit  endif
  mscHostLen@ over < 			\ case 2, 7
  if  drop 1  badRWphase exit  endif
  mscHostLen@ swap >
  if  1 badRWfail exit  endif		\ case 5
  0					\ case 6
;

: MscDevDataOUT?	\ len -- ior ; 0=success
\ *G Checks cases 3, 8, 11, 12, and 13. On error, appropriate action
\ ** is performed. On success, no action is taken.
\ case 3: Hn<Do - stall, 02h status
\ case 8: Hi<>Do (dir error) - stall, 02h status
\ case 11: Ho>Do - stall, 01h status
\ case 12: Ho=Do - success
\ case 13: Ho<Do - stall, 02h status
  mscHostDir@ $00 <>			\ case 8
  if  drop 1  badRWphase exit  endif
  mscHostLen@ over <			\ case 3, 13
  if  drop 1  badRWphase exit  endif
  mscHostLen@ swap >			\ case 11
  if  1 badRWfail exit  endif
  0					\ case 12
;

\ ...SFP011


\ *****************
\ *S CBW operations
\ *****************

usbDMA? [if]
2variable mscRespCIN	\ -- addr
\ *G Contains a len/addr pair for command responses,

: -RespCIN	\ --
\ *G Mark no command response data, and switch to a CSW send.
  0 0 mscRespCIN 2!  NoResidue SendCSW
;

: +RespCIN	\ caddr len --
\ *G Trigger a command response phase.
  mscHostLen@ min			\ MvD workaround for Mac! ; SFP011
\  CBWbuff cbw.DataLen lel@ min		\ MvD workaround for Mac!
  /MaxPacket min			\ SFP011
  MscDevDataIN-1? 0= if			\ -- caddr len 0 | ior ; SFP011
    mscRespCIN 2!  smCIN -> MSCsm#
  endif
;
[else]
: +RespCIN	\ caddr len --
  mscHostLen@ min			\ MvD workaround for Mac! ; SFP011
\  CBWbuff cbw.DataLen lel@ min		\ MvD workaround!
  /MaxPacket min			\ SFP011
  MscDevDataIN-1? 0= if			\ -- caddr len ; good ; SFP011
    writeMSCin				\ write buffer to IN endpoint
    SendCSW				\ later send CSW
  endif
;
[then]

((
The TEST UNIT READY command provides a
means to check if the logical unit is ready. This
is not a request for self-check. If the logical unit
is able to accept an appropriate medium access
command, without returning a Check Condition
status, this command will return a Good status.
Otherwise, the command is terminated with the
Check Condition status and the SENSE KEY is
set to reflect the error condition.
))
: doTUR		\ --
\ *G Perform the TEST UNIT READY actions.
  USBdisk? CDin? and 0=			\ FVD009
  if  SenseNotReady -> SenseCode  endif
  NoMscDevData? if  exit  endif		\ SFP011
  SenseCode
  if  sendCSWfail  exit  endif
  sendCSW
;

#18 equ /SenseData	\ -- len
\ *G Size of the response to the REQUEST SENSE command.
create SenseTemplate	\ -- addr
\ *G Holds the 18 byte response to the REQUEST SENSE command.
\ ** The $FF bytes are filled in with data from *\fo{SenseCode}.
  $70 c,  0 c,  $FF c,  0 c,		\ third byte is KEY
  0 c,  0 c,  0 c,  $0A c,		\ last is additional length
  0 c,  0 c,  0 c,  0 c,
  $FF c,  $FF c,  0 c,  0 c,		\ first two are ASC and ASQ
  0 c,  0 c,

: doRS		\ --
\ *G Respond to the REQUEST SENSE command.
  SenseTemplate MSCbuff /SenseData cmove
  SenseCode MscBuff			\ -- code buff
  over #16 rshift  over 2 + c!		\ Key
  over #8 rshift  over #12 + c!		\ ASC
  tuck #13 + c!				\ ASQ
  0 -> SenseCode			\ clear
  /SenseData +RespCIN			\ write buffer to IN endpoint
;

#36 equ /InqData	\ -- len
\ *G Size of returned INQUIRY data.

create InqData	\ -- addr
\ *G The returned INQUIRY data.
  0 c,					\ PDT = direct access device
  $80 c,				\ removable media, e.g. SD card
  0 c,					\ no standard claim
  $02 c,				\ standard response

  #32 c,				\ 32 additional bytes inc. this.
  0 c,  0 c,  0 c,			\ not a storage array
  #28 allot&erase
  s" MPE     " InqData 8 + swap cmove	\ 8 byte Vendor string
  s" Mass Storage Dev" InqData #16 + swap cmove	\ 16 byte Product string
  s" 1.00" InqData #32 + swap cmove	\ 4 byte Version string

: doINQ		\ --
\ *G Return the INQUIRY data.
  InqData MSCbuff /InqData cmove	\ needed for LPC2xxx DMA
  MSCbuff /InqData +RespCIN		\ Send INQUIRY data
;

: doRC10	\ --
\ *G Performs the READ CAPACITY (10) command operations.
[ dbgMSC? ] [if] ." RC10 " [then]
  #Sectors 1- MSCbuff bel!		\ last sector (0..n-1)
  #512 MSCbuff 4 + bel!			\ sector size
  MSCbuff 8 +RespCIN			\ respond
;

: initMSCrw	\ --
\ *G Initialise the read/write data from the CDBbuff
  0 -> DriveStatus			\ not finished
  CDBbuff 2+ bel@ -> FirstSector	\ first sector
  CDBbuff 7 + bew@ dup -> NumSectors	\ number of sectors
  #512 * mscBulkLen !			\ total number of bytes in transfer
;

: +DiskRD	\ --
\ *G Start the disk task read.
\ [ dbgMSC? ] [if] cr ." +DiskRD " FirstSector . NumSectors . [then]
   SecReadCmd -> DriveBusy?  smDIN -> MSCsm#
;

: +DiskWR	\ --
\ *G Start the disk task write.
\ [ dbgMSC? ] [if] cr ." +DiskWR " FirstSector . NumSectors . [then]
   SecWriteCmd -> DriveBusy?  smDOUT -> MSCsm#
;

: doRD10	\ --
\ *G Performs the READ (10) command operations.
  USBdisk? CDin? and 0= if		\ FVD009
\    SenseNotReady -> SenseCode
    StallMSCin  SendCSWfail  exit
  endif
  DriveBusy? if
    StallMSCin  SendCSWfail  exit
  endif
  -BulkData initMSCrw
  mscBulkLen @ MscDevDataIN-2?		\ 13 condition check ; SFP011
  if  exit  endif
  NumSectors				\ check number of sectors to send
  if  +DiskRD  else  sendCSW  endif	\ 0 sectors is valid
;

: doWR10	\ --
\ *G Performs the WRITE (10) command operations.
  USBdisk? CDin? and 0= if		\ FVD009
\    SenseNotReady -> SenseCode
    StallMSCout  SendCSWfail  exit
  endif
  DriveBusy? if
    StallMSCout  SendCSWfail  exit
  endif
  -BulkData initMSCrw
  mscBulkLen @ MscDevDataOUT?		\ 13 condition check ; SFP011
  if  exit  endif
  NumSectors				\ check number of sectors to send
  if  +DiskWR  else  sendCSW  endif	\ 0 sectors is valid
;

: doVRFY10	\ --
\ *G Performs the VERIFY (10) command operations.
\ ** For the moment this is a dummy that always succeeds unless
\ ** no drive is present.
\ cr ." VRFY " CBWbuff /CBW dump
  CDin? 0= if
\    SenseNotReady -> SenseCode
    SendCSWfail  exit
  endif
  NoResidue sendCSW
;

((
MS6 response
============
Header
======
0	byte	length in bytes after this
1  	byte	medium type		See SPC3r23, 3.1.18
2	byte	device specific
3	byte	block length (8)
Block
=====
0	byte	density code		See SPC3r23, 3.1.18
1	3 bytes	number of blocks/sectors
4       byte	reserved
5	3 bytes	block length
))

create MS6resp	\ -- addr
\ *G The MODE SENSE (6) response.
\ header
  $0B c,  0 c,  0 c,  8 c,
\ block
  0 ,					\ num sectors, filled in later
  0 c,  0 c,  #512 lew,

: doMS6		\ --
\ *G Performs the MODE SENSE (6) command operations.
\ ** Taken from other code - I have no idea!
\ ** This command is not needed by Windows, but is required by
\ ** Linux.
  MS6resp MSCbuff tuck #12 cmove	\ copy template
  WPin? if				\ check for write protected
    $80 MSCbuff 2 + c!
  endif
  #Sectors over 4 + bel!
  #12 +RespCIN
;

((
: doRFC		\ --
\ *G Performs the READ FORMAT CAPACITY command operations.
\ ** This command is supposed to be obsolete or vendor specific,
\ ** but Windows may want it. See *\i{INF-8070.PDF} for the details.
\ ** NO LONGER REQUIRED.
  MSCbuff dup #12 erase
  8 over 3 + c!				\ num bytes that follow
  #Sectors over 4 + bel!		\ sectors on dist
  #512 over 8 + bel!			\ sector size
  2 over 4 + c!				\ descriptor code - formatted media
  #12 +RespCIN
;
))

: ProcessCBW	\ --
  case  CDBbuff c@			\ inspect command code
\ [ dbgMSC? ] [if]
\ dup if  cr ." PCBW_" dup .byte space  then	\ ignore TUR
\ [then]
    cmdTestUnitReady of  doTUR  endof
    cmdRequestSense of  doRS  endof
    cmdInquiry of  doINQ  endof
    cmdFormatUnit of  sendCSW  endof	\ acknowledge but do nothing
    cmdStartStopUnit of  sendCSW  endof	\ acknowledge but do nothing
    cmdReadCapacity10 of  doRC10  endof
\    cmdReadFormatCapacities of  doRFC  endof	\ cos of Windows
    cmdRead10 of  doRD10  endof
    cmdWrite10 of  doWR10  endof
    cmdVerify10 of  doVRFY10  endof
    cmdModeSense6 of  doMS6  endof
[ dbgMSC? ] [if] cr ." PCBW fail" [then]
      SenseInvalidOp -> SenseCode
      BOTstall  sendCSWfail
  endcase
;


\ **********************
\ *S Endpoint despatcher
\ **********************
\ *P This section contains the handlers for the MSC bulk IN and
\ ** OUT handlers, plus the actions required for set up operations
\ ** performed on endpoint 0.

: +DiskData	\ u --
\ *G Add *\i{u} bytes to the bulk transfer control variables.
  dup mscBulkData .addr +!
  negate dup mscBulkData .len +!
  mscBulkLen +!
;

: ?AckSector	\ --
\ *G If a full sector has been transferred, acknowledge it.
  mscBulkData .len @ 1 <
  if  -BulkData  endif
;


\ **************
\ *N Without DMA
\ **************

usbDMA? 0= [if]

: ReadCBW	\ --
\ *G Read a CBW from the bulk OUT endpoint and process it.
  CBWbuff ReadMSCout /CBW = if		\ must be 31 bytes
\ [ dbgMSC? ] [if]  CBWbuff /CBW dump  [then]
    InitCSW				\ set up good CSW response
    CBWbuff cbw.Sig lel@ CBWsig# = if	\ sig must be good
      CBWbuff cbw.bLun c@ 0= if		\ we only support one LUN
        CBWbuff cbw.bCDBlen c@		\ CDB length
        1 17 within if			\ must be 1..16
          ProcessCBW exit		\ now it's good
        endif
      endif
    endif
    SenseInvalidCDB -> SenseCode
    BOTstall
    SendCSWfail  exit			\ bad command format
  endif
[ dbgMSC? ] [if]  ." bad_CBW "  [then]
  0 CBWbuff !				\ wipe first four bytes
  SenseInvalidCDB -> SenseCode
  StallMSCin  StallMSCout		\ stall both bulk endpoints
  smStalled -> MSCsm#			\ mark as stalled
;

: TransmitCSW	\ --
\ *G Transmit the CSW on the bulk IN endpoint, and switch
\ ** the state machine to wait for the next CBW.
[ dbgMSC? ] [if] ." TxCSW " [then]
  CSWBuff /CSW WriteMSCin
  0 0 initBulk  0 setDriveCmd  -1 -> DriveComplete
  smCBW -> MSCsm#
;

: readMSCdata	\ --
\ *G Read the bulk OUT endpoint for a transfer operation
  DriveBusy? SecWriteCmd <>
  if  exit  endif
  SenseCode
  if   StallMSCout sendCSWfail  exit  endif
  mscBulkData .addr @ if		\ if there's somewhere to put it
    mscBulkData .addr @ readMSCout
    +DiskData  ?AckSector
  endif
  DriveStatus mscBulkLen @ 1 < or if
    mscBulkLen @ CSWbuff csw.Residue lel!
    DriveStatus DrvOK = if
      sendCSW
    else
      stallMSCout
      SenseWriteError -> SenseCode
      sendCSWfail
    endif
  endif
;

: writeMSCdata	\ --
\ *G Write the bulk IN endpoint for a transfer operation. Note
\ ** that this may receive NAK interrupts as it is a Bulk IN endpoint.
  DriveBusy? SecReadCmd <>
  if  exit  endif
  SenseCode
  if   StallMSCin sendCSWfail  exit  endif
  mscBulkData .addr @ if		\ if there's some thing to send
    mscBulkData 2@ /MaxPacket min dup >r
    writeMSCin
    r> +DiskData  ?AckSector
  endif
  DriveStatus mscBulkLen @ 1 < or if
    mscBulkLen @ CSWbuff csw.Residue lel!
    DriveStatus DrvOK = mscBulkLen @ 0= or if
      sendCSW
    else
      stallMSCin
      SenseReadError -> SenseCode
      sendCSWfail
    endif
  endif
;

: DiscardMSCout	\ --
\ *G Read and discard data received on the MSC Bulk OUT endpoint.
  MSCbuff readMSCout drop
;

: MSCBulkOut	\ --
\ *G Handle bulk OUT actions to the device. Note that this includes NAK
\ ** interrupts.
  UsbEPhalt MscBoutEPMask and 0= if	\ if the endpoint is not stalled
    case  MSCsm#
      smCBW of  ReadCBW  endof		\ read and process expected CBW
      smDOUT of  readMSCdata  endof	\ read expected data
      smDIN of  DiscardMSCout  endof
      smCSW of  DiscardMSCout  endof
      smStalled of  StallMSCout  endof	\ stall it again until reset
\        ." badBO " .MSCstate
        StallMSCout sendCSWphase
    endcase
  endif
;

: MSCBulkIn	\ --
\ *G Handle bulk IN actions to the host. Note that this includes NAK
\ ** interrupts.
  UsbEPhalt MscBinEPMask and 0= if	\ if the endpoint is not stalled
    case  MSCsm#
      smDIN of  WriteMSCdata  endof	\ write or finish data transmission
      smCSW of  TransmitCSW  endof	\ transmit CSW
      smCBW of  endof			\ ignore
      smDOUT of  endof			\ ignore
      smStalled of  StallMSCin  endof	\ stall it again until reset
\        ." badBI " .MSCstate
        StallMSCin sendCSWphase
    endcase
  endif
;

: doEP2		\ event --
\ *G The action of endpoint 2.
  case
    USB_EVT_OUT of  MSCBulkOut  endof
    USB_EVT_IN  of  MSCBulkIn  endof
  endcase
;
' doEP2 2 SetEPhandler

((
: MscFrameCheck	\ --
\ *G Executed in the frame interrupt.
\ ** No longer needed as it is done by default for non-DMA drivers.
  +BulkNAKout
;
' MscFrameCheck AtUSBframe
))

: DevMSCreset	\ --
\ *G Reset the MSC driver. Performed at start up and as an EP0 action.
  smCBW -> MSCsm#  0 -> SenseCode	\ state and error status
  0 0 initBulk				\ no pending data transfers
  resetUSBdisk				\ reset disk task
;
' DevMSCreset AtUSBreset
Synonym resetMSCdata DevMSCreset

[then]


\ ***********
\ *N With DMA
\ ***********

usbDMA? [if]

: startReadMSCdata	\ --
\ *G Prepare the bulk OUT endpoint for a transfer operation
  DriveBusy? SecWriteCmd = if
    mscBulkData .addr @ if		\ if there's somewhere to put it
      mscBulkData 2@ MscBoutEP setEPdma	\ start if off
    endif
  endif
;

: endReadMSCdata	\ --
\ *G Handle the bulk OUT endpoint data just received.
  DriveBusy? SecWriteCmd = if
    mscBulkData .addr @ if		\ if there's somewhere to put it
     MscBoutEP #EPtrans +DiskData	\ update pointers
     ?AckSector				\ check completion
    endif
    DriveStatus mscBulkLen @ 1 < or if
      mscBulkLen @ CSWbuff csw.Residue lel!
      DriveStatus DrvOK = if
        sendCSW
      else
        stallMSCout
        SenseWriteError -> SenseCode
        sendCSWfail
      endif
    endif
  endif
;

: startWriteMSCdata	\ --
\ *G Write the bulk IN endpoint for a transfer operation.
  DriveBusy? SecReadCmd = if
    mscBulkData .addr @ if		\ if there's some thing to send
      mscBulkData 2@ MscBinEP setEPdma	\ start transfer
    endif
  endif
;

: endWriteMSCdata	\ --
\ *G Update after a bulk IN endpoint data transfer operation.
  DriveBusy? SecReadCmd = if
    mscBulkData .addr @ if		\ if there was some thing to send
      MscBinEP #EPtrans +DiskData	\ count it
     ?AckSector
    endif
    DriveStatus mscBulkLen @ 1 < or if
      mscBulkLen @ CSWbuff csw.Residue lel!
      DriveStatus DrvOK = mscBulkLen @ 0= or if
        sendCSW
      else
        stallMSCin
        SenseReadError -> SenseCode
        sendCSWfail
      endif
    endif
  endif
;

\ MVD006...
0 value txCSWdone?	\ -- flag
\ *G An interlock used to prevent multiple transmissions of
\ ** the CSW.

: startTransmitCSW	\ --
\ *G Transmit the CSW on the bulk IN endpoint.
[ dbgMSC? ] [if] ." sTxCSW " [then]
  CSWBuff /CSW MscBinEP setEPdma
  -1 to txCSWdone?
;

: endTransmitCSW	\ --
\ *G CSW transmit done, switch the state machine to wait for the next CBW.
[ dbgMSC? ] [if] ." eTxCSW " [then]
  txCSWdone? if
    0 0 initBulk  0 setDriveCmd  -1 -> DriveComplete
    smCBW -> MSCsm#  0 to txCSWdone?
  then
;
\ ...MVD006

: startRespCIN	\ --
\ *G Start the command response transfer.
  mscRespCIN .len @ if
    mscRespCIN 2@ MscBinEP setEPdma
    mscRespCIN .len off
  endif
;

: endRespCIN	\ --
\ *G Mark the end of the command response stage.
  mscRespCIN .len @ if
    mscRespCIN 2@ MscBinEP setEPdma
    mscRespCIN .len off
  else
    -RespCIN
  endif
;

: startReadCBW	\ --
\ *G prepare for a CBW read.
  CBWbuff /MaxPacket MscBoutEP setEPDMA
;

: endReadCBW	\ --
\ *G Read a CBW from the bulk OUT endpoint and process it.
  MscBoutEP #EPtrans /CBW = if		\ must be 31 bytes
\ [ dbgMSC? ] [if]  CBWbuff /CBW dump  [then]
    InitCSW				\ set up good CSW response
    CBWbuff cbw.Sig lel@ CBWsig# = if	\ sig must be good
      CBWbuff cbw.bLun c@ 0= if		\ we only support one LUN
        CBWbuff cbw.bCDBlen c@		\ CDB length
        1 17 within if			\ must be 1..16
          ProcessCBW exit		\ now it's good
        endif
      endif
    endif
    SenseInvalidCDB -> SenseCode
    BOTstall SendCSWfail  exit		\ bad command format
  endif
[ dbgMSC? ] [if]  ." bad_CBW "  [then]
  0 CBWbuff !				\ wipe first four bytes
  SenseInvalidCDB -> SenseCode
  StallMSCin  StallMSCout		\ stall both bulk endpoints
  smStalled -> MSCsm#			\ mark as stalled
;

: MSCBulkOutNdr	\ --
\ *G Handle bulk OUT requests (data to the device).
  MscBoutEP enEPdma			\ it was disabled by the hardware
  UsbEPhalt MscBoutEPMask and 0= if	\ if the endpoint is not stalled
    case  MSCsm#
      smCBW of  startReadCBW  endof	\ start read of expected CBW
      smDOUT of  startReadMSCdata  endof \ read expected data
      smDIN of  endof
      smCSW of  endof
      smStalled of  StallMSCout  endof	\ stall it again until reset
\        ." badBO " .MSCstate
        StallMSCout sendCSWphase
    endcase
  endif
;

: MSCBulkOutEot	\ --
\ *G Handle bulk OUT completions (data to the device).
  UsbEPhalt MscBoutEPMask and 0= if	\ if the endpoint is not stalled
    case  MSCsm#
      smCBW of  endReadCBW  endof	\ read and process expected CBW
      smDOUT of  endReadMSCdata  endof	\ read expected data
      smDIN of  endof
      smCSW of  endof
      smStalled of  StallMSCout  endof	\ stall it again until reset
\        ." badBO " .MSCstate
        StallMSCout sendCSWphase
    endcase
  endif
;

: MSCBulkInNdr	\ --
\ *G Handle requests for bulk IN actions (data to the host).
  MscBinEP enEPdma			\ it was disabled by the hardware
  UsbEPhalt MscBinEPMask and 0= if	\ if the endpoint is not stalled
    case  MSCsm#
      smDIN of startWriteMSCdata endof	\ write or finish data transmission
      smCSW of startTransmitCSW endof	\ transmit CSW
\      smCBW of  StallMSCin sendCSWfail endof	\ ignore ; SFP011
\      smDOUT of  StallMSCin  endof	\ ignore ; SFP011
      smCBW of  endof			\ ignore
      smDOUT of  endof			\ ignore
      smCIN of  startRespCIN  endof	\ command response start
      smStalled of StallMSCin endof	\ stall it again until reset
\        ." badBI " .MSCstate
        StallMSCin sendCSWphase
    endcase
  endif
;

: MSCBulkInEOT	\ --
\ *G Handle completion of bulk IN actions (data to the host).
  UsbEPhalt MscBinEPMask and 0= if	\ if the endpoint is not stalled
    case  MSCsm#
      smDIN of  endWriteMSCdata  endof	\ write or finish data transmission
      smCSW of  endTransmitCSW  endof	\ transmit CSW
      smCBW of  endof			\ ignore
      smDOUT of  endof			\ ignore
      smCIN of  endRespCIN  endof	\ command response end
      smStalled of  StallMSCin  endof	\ stall it again until reset
\        ." badBI " .MSCstate
        StallMSCin sendCSWphase
    endcase
  endif
;

: doEP2		\ event --
\ *G The action of endpoint 2.
  case
    USB_EVT_OUT_DMA_NDR of  MSCBulkOutNdr  endof
    USB_EVT_OUT_DMA_EOT of  MSCBulkOutEot  endof
    USB_EVT_IN_DMA_NDR  of  MSCBulkInNdr  endof
    USB_EVT_IN_DMA_EOT  of  MSCBulkInEot  endof
  endcase
;
' doEP2 2 SetEPhandler

: MscFrameCheck	\ --
\ *G Executed in the frame interrupt. Enables both Bulk endpoints,
\ ** and seems to be required. We do not yet know why!
  MscBinEP enEPdma  MscBoutEP enEPdma
;
' MscFrameCheck AtUSBframe

: resetMSCdata	\ --
\ *G Reset the MSC driver data. Performed at start up and as
\ ** an EP0 action.
  smCBW -> MSCsm#  0 -> SenseCode	\ state and error status
  0 0 initBulk				\ no pending data transfers
  0 0 mscRespCIN 2!			\ no pending command response
  resetUSBdisk				\ reset disk task
;

: DevMSCreset	\ --
\ *G Reset the MSC driver. Performed as an EP0 action.
  resetMSCdata
  CBWbuff /MaxPacket MscBoutEP setEPdma
;
' DevMSCreset AtUSBreset

[then]


\ ********************
\ *S EP0 SETUP actions
\ ********************

: MscReset	\ -- flag ; 0=failed
\ *G Reset the MSC driver. EP0 action.
  SetupPacket usp.wValue uaLEw@ 0<>	\ Validate command ; SFP011
  SetupPacket usp.wIndex uaLEw@ MscIf# <> or
  SetupPacket usp.wLength uaLEw@ 0<> or
  if  0 exit  endif
  DevMSCreset 1
;

: MSCGetMaxLUN	\ -- flag ; 0=failed
\ *G This device does not distinguish between LUNs. EP0 action.
  SetupPacket usp.wValue uaLEw@ 0<>	\ Validate command ; SFP011
  SetupPacket usp.wIndex uaLEw@ MscIf# <> or
  SetupPacket usp.wLength uaLEw@ 1 <> or
  if  0 exit  endif
  0 EP0Buf c!  1
;

: MSCdoEP0ClassRTI	\ --
\ *G Handle the MSC class REQUEST_TO_INTERFACE operations.
  case  SetupPacket usp.bRequest c@
    MSC_REQUEST_RESET of
      MscReset
      if  USBStatusInStage  else  stall_i  endif
    endof
    MSC_REQUEST_GET_MAX_LUN of
      MscGetMaxLUN if
        EP0Buf EP0Data ep.*Data !  USBDataInStage  0 EP0Data ep.Count !
      else
        stall_i
      endif
    endof
      stall_i
  endcase
;


\ ****************************
\ *S Diagnostics and Test code
\ ****************************

\ SFP008...
: initMSCdata	\ --
\ *G Initialise the MSC data and task.
  initFATfs drop
  resetMSCdata				\ initialise data
  -USBdiskTask +USBdiskTask		\ start disk task
;
\ ...SFP008


\ ======
\ *> ###
\ ======

decimal
