\ UsbEP0.fth - USB Endpoint 0 operations, hardware independent

((
Copyright (c) 2008, 2013, 2017
Synics AG
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20171201 SFP002 Changes for USBSetAddress functionality.
20130905 SFP001 Removed SetInterface operation.
))

only forth definitions
decimal

\ =========
\ *! usbep0
\ *T USB EndPoint 0
\ =========
\ *P The file *\i{UsbBase.fth} contains an (almost) hardware
\ ** independent USB code base. The hardware layer files,
\ ** *\i{UsbHwxxxx.fth} provide a standard interface that isolates
\ ** hardware from the base layer. Class drivers such as *\i{MscDrv.fth}
\ ** should be compiled after the core layer and can reference the
\ ** hardware layer. After the class files have been compiled,
\ ** compile *\fo{UsbEP0.fth}, which contains the USB control and
\ ** initialisation code using endpoint 0.

UsbMin? [if]  internal  [then]

\ ===================
\ *S Setup operations
\ ===================

compiler
: RTD?		\ -- flag
\ *G Is this a request to the device? If so return true.
  ReqType ReqRecipMask and REQUEST_TO_DEVICE =
;
: RTI?		\ -- flag
\ *G Is this a request to the device? If so return true.
  ReqType ReqRecipMask and REQUEST_TO_INTERFACE =
;
target

: doEP0SuStd	\ --
\ *G Handle the EP0 REQUEST_TYPE_STANDARD setup operations.
[ dbgCore? ] [if] ." EP0SuStd " [then]
  case  SetupPacket usp.bRequest c@
    USB_REQUEST_GET_STATUS of
      USBGetStatus 0= if  stall_i  endif
    endof
    USB_REQUEST_CLEAR_FEATURE of
      0 USBSetClrFeature 0=
      if  stall_i  else  USBStatusInStage  endif
    endof
    USB_REQUEST_SET_FEATURE of
      1 USBSetClrFeature 0=
      if  stall_i  else  USBStatusInStage  endif
    endof
    USB_REQUEST_SET_ADDRESS of
      RTD? if
        SetupPacket usp.wValue c@
        USBStatusInStage
        USBSetAddress			\ must be after USBStatusInStage
      else
        stall_i
      endif
    endof
    USB_REQUEST_GET_DESCRIPTOR of
      USBGetDescriptor 0= if  stall_i  endif
    endof
    USB_REQUEST_SET_DESCRIPTOR of
      $00 USBSetStallEP  EP0Data ep.Count off
    endof
    USB_REQUEST_GET_CONFIGURATION of
      RTD? if
        addr UsbConfig EP0Data ep.*Data !
        USBDataInStage
      else
        stall_i
      endif
    endof
    USB_REQUEST_SET_CONFIGURATION of
      RTD? if
        USBSetConfig 0=
        if  stall_i  else  USBStatusInStage  endif
      else
        stall_i
      endif
    endof
    USB_REQUEST_GET_INTERFACE of
      RTI? if
        UsbConfig  SetupPacket usp.wIndex c@ Usb#IfSet < and if
         UsbAltSets SetupPacket usp.wIndex c@ Usb#IfSet +
         EP0Data ep.*Data !
         USBDataInStage
        else
          stall_i
        endif
      else
        stall_i
      endif
    endof
    USB_REQUEST_SET_INTERFACE of
      stall_i				\ SFP001
((
      RTI? if
        USBSetIface 0=
        if  stall_i  else  USBStatusInStage  endif
      else
        stall_i
      endif
))
    endof
      stall_i
  endcase
;

: doEP0ClassRTI	\ --
\ *G Handle the class REQUEST_TO_INTERFACE operations.
[ MSC? ] [if]
  SetupPacket usp.wIndex c@ MscIf# = if	\ Mass Storage Class request?
    MSCdoEP0ClassRTI  exit
  endif
[then]
[ CDC? ] [if]
  SetupPacket usp.wIndex c@		\ interface number
  dup CdcCIf# = swap CdcDIf# = or if	\ CDC class request?
    CDCdoEP0ClassRTI  exit
  endif
[then]
[ HID? ] [if]  ...  [then]
[ Audio? ] [if]  ...  [then]
  stall_i
;

: doEP0SuClass	\ --
\ *G Handle the EP0 REQUEST_TYPE_CLASS setup operations.
  case  ReqRecipient
    REQUEST_TO_INTERFACE of doEP0ClassRTI  endof
      stall_i
  endcase
;

\ ===========
\ *S Despatch
\ ===========

: doEP0setup	\ --
\ *G Perform the EP0 SETUP operations.
[ dbgCore? ] [if] ." EP0Setup " [then]
  USBSetupStage
  ReqDir USBDirCtrlEP
  SetupPacket usp.wLength lew@ EP0Data ep.Count !
  case  ReqTypeBits
    ReqTypeStd   of  doEP0SuStd  endof
    ReqTypeClass of  doEP0SuClass  endof
      stall_i
  endcase
;

: doEP0out	\ --
\ *G Perform the EP0 OUT (to device) operations.
  ReqDir
  if  USBStatusOutStage exit  endif
  EP0Addr if				\ address set?
    /EP0Data if				\ more to do?
      USBDataOutStage
      /EP0Data 0= if			\ all data transferred
        NoEP0Data  USBStatusInStage
        exit
      endif
      /EP0Data 0> if			\ data left to do is ok
        exit
      endif
    endif
  endif
  stall_o				\ only get here on error
;
((
\ !!Expand here!! for other classes - code taken from from a C example
\ on success, status in stage, otherwise stall
      case ReqTypeBits
        REQUEST_CLASS of
          case ReqRecipient
            REQUEST_TO_INTERFACE of
[ HID? ] [if]  ...  [then]
[ AUDIO? ] [if]  ...  [then]
              stall_i
            endof
            REQUEST_TO_ENDPOINT of
[ AUDIO? ] [if]  ...  [then]
              stall_i
            endof
              stall_i
          endcase
        endof
          stall_i
      endcase
))

: doEP0in	\ --
\ *G Perform the EP0 IN (to host) operations.
  ReqDir if				\ sanity check
    USBDataInStage
  endif
;

: doEP0		\ event --
\ *G The despatcher for the actions of endpoint 0.
[ dbgCore? ] [if] ." EP0 "  [then]
  case
    USB_EVT_SETUP of  doEP0setup  endof
    USB_EVT_IN    of  doEP0in  endof
    USB_EVT_OUT   of  doEP0out  endof
    USB_EVT_IN_STALL  of  $80 USBClrStallEP  endof
    USB_EVT_OUT_STALL of  $00 USBClrStallEP  endof
  endcase
;
' doEP0 0 SetEPhandler

UsbMin? [if]  external  [then]


\ ****************************
\ *S USB system initialisation
\ ****************************
\ *P As of November 2010, start up of the USB code is performed by
\ ** *\fo{startUSBsys ( -- )}, which can be found in *\i{UsbEP0.fth}.
\ ** *\fo{startUSBsys} uses conditional compilation to produce a
\ ** start up word that initialises all the required classes.
\ ** Please use *\fo{startUSBsys} as your USB start up word from
\ ** now on.

((
CDC? MSC? 0= and [if]
: StartCDC	\ --
\ *G Start the USB and CDC systems. Used for CDC only devices.
\ ** NOW OBSOLETE - see *\fo{startUSBsys} below.
  resetCDCdata startUSB  ;		\ initialise CDC, start USB
[then]

CDC? 0= MSC? and [if]
: StartMSC	\ --
\ *G Initialise MSC and start USB.
\ ** NOW OBSOLETE - see *\fo{startUSBsys} below.
  initMSCdata startUSB  ;		\ initialise USB and connect
[then]

CDC? MSC? and [if]
: StartCompo	\ --
\ *G Start the USB CDC and MSC systems. Used for composite devices.
\ ** NOW OBSOLETE - see *\fo{startUSBsys} below.
  resetCDCdata initMSCdata startUSB  ;
[then]
))

: startUSBsys	\ --
\ *G Initialise all the selected USB components and then
\ ** start the USB connection.
[ CDC? ] [if]  resetCDCdata  [then]
[ MSC? ] [if]  initMSCdata  [then]
  startUSB
;


\ ======
\ *> ###
\ ======

decimal
