\ UsbHwLPC2xxx.fth - LPC2xxx USB hardware driver with DMA option.

((
Copyright (c) 2008, 2010, 2011
Synics AG, CDS Advanced Technology bv
  and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20171201 SFP008 Changes for STN32F0x2 compatibility.
20110817 SFP007 Updated PORTL> and >PORTL.
20101101 SFP006 Updated start up code
20100903 FVD005 Updated for systems that can control VBus.
20081117 MPE004 Isolated LPC2xxx differences.
20081111 MPE003 DMA changes.
20081104 MPE002 Recoded PORTL> and >PORTL.
20081028 CDS001 Converted to LPC23xx and ISP1301.
))

only forth definitions
decimal

\ ===============
\ *! usbhwlpc2xxx
\ *T USB hardware layer for LPC2xxx
\ ===============
\ *P The file *\i{UsbHwLPC2xxx.fth} contains the USB hardware layer for
\ ** NXP LPC2xxx devices.

\ *P DMA operation is supported if the equate *\fo{usbDMA?} is
\ ** set non-zero in the USB configuration file, e.g. *\i{MscConfig.fth}.
\ ** Use of DMA is *\b{strongly} recommended.
\ ** For most real-time applications the benefit of DMA operation
\ ** is the considerable reduction of interrupt overhead at the
\ ** expense of code size. If even this is unacceptable, consider
\ ** the following solutions.
\ *(
\ *B Permit nested interrupts and reduce the priority of the
\ ** USB interrupt handler.
\ *B Do not use a USB interrupt handler at all. Define a USB task
\ ** and poll the handler.
\ *)


\ ****************
\ *S Configuration
\ ****************

IsLeaf
: EPaddr	\ ep# -- physaddr
\ *G Translate the endpoint number to a physical address. Bit 7
\ ** of the endpoint number *\i{ep#} (0..15) is set if the write
\ ** buffer is required.
  dup $0F and 1 lshift  swap bit7 and
  if  1+ endif
;
interpreter
: EPaddr	\ ep# -- physaddr
  dup $0F and 1 lshift  swap bit7 and
  if  1+ endif
;
target
\ A COMPILER version costs 300 bytes or so

[undefined] usbDMA? [if]
1 equ usbDMA?	\ -- flag
\ *G Set non-zero if the USB hardware is to use DMA for bulk and
\ ** isochronous transfers. The definition in this file is only
\ ** used if not previously defined.
\ ** Use of DMA is *\b{strongly} recommended.
[then]


\ *****************
\ *S DMA operations
\ *****************
\ *P The USB hardware contains its own DMA controller. Whether
\ ** this is used is set by the equate *\fo{usbDMA?} in the USB
\ ** configuration file. Each endpoint that uses DMA *\b{must}
\ ** define at least one DMA descriptor and its type.

usbDMA? [if]

[undefined] UDCA [if]
[sections URAMP
32 cells buffer: UDCA	\ -- addr
\ *G USB DMA communication area. *\fo{UDCA} is only defined here
\ ** if it has not already been defined.
sections]
[then]

UDCA #127 and [if]
  cr ." UDCA must be on a 128 byte aligned boundary in USB RAM" abort
[then]

create UDCAcfg	\ -- addr
\ *G A table of device descriptors used to initialise the
\ ** *\fo{UDCA}.
  32 cells allot&erase

create EpDmaCfg	\ -- addr
\ *G A table containing a byte per endpoint. The byte indicates
\ ** how the endpoint is used with DMA:
\ *D 0    no DMA
\ *D 16   Bulk or interrupt endpoint
\ *D 20   Isochronous endpoint.
  32 allot

0 equ NonDmaEP
16 equ BulkDmaEP
16 equ IntDmaEP
20 equ IsoDmaEP


\ *P DMA use is set at compile time. The following *\fo{INTERPRETER}
\ ** words are used.

interpreter
: BulkDD:	\ ep# -- ; e.g. $0x or $8x
\ *G Creates a bulk DMA Descriptor in the *\fo{URAMP}
\ ** section, and adds it to the *\fo{UDCAcfg} template.
  EPaddr  BulkDmaEP over EpDmaCfg + !  >r
  [sections URAMP
    here r> cells UDCAcfg + !		\ fill slot in template
    4 cells allot
  sections]
;

: IntDD:	\ ep# -- ; e.g. $0x or $8x
\ *G Creates an interrupt DMA Descriptor in the *\fo{URAMP}
\ ** section, and adds it to the *\fo{UDCAcfg} template.
  EPaddr  IntDmaEP over EpDmaCfg + !  >r
  [sections URAMP
    here r> cells UDCAcfg + !		\ fill slot in template
    4 cells allot
  sections]
;

: IsoDD:	\ ep# -- ; e.g. $0x or $8x
\ *G Creates an Isochronous DMA Descriptor in the *\fo{URAMP} section,
\ ** and adds it to the *\fo{UDCAcfg} template.
  EPaddr  IsoDmaEP over EpDmaCfg + !  >r
  [sections URAMP
    here r> cells UDCAcfg + !		\ fill slot in template
    5 cells allot
  sections]
;
target

: resetDD	\ physaddr dd --
\ *G Reset a device descriptor to No_Packet state. The DMA
\ ** buffer address, length and max packet size are all zero.
\ ** See LPC23xx User Manual, Chapter 13, 15.5.6.
  swap EpDmaCfg + c@ 2dup erase		\ -- dd type/len
  IsoDmaEP = if				\ -- dd
    bit4 swap cell + or!
  else
    drop
  endif
;

: initEPdma	\ physaddr --
\ *G Set up the DMA for the given endpoint (0..31) and apply
\ ** *\fo{resetDD}.
  dup EpDmaCfg + c@ if			\ DMA endpoint
    dup cells UDCAcfg + @		\ -- physaddr dd
    2dup resetDD  over cells UDCA + !	\ -- physaddr ; reset DD and set in table
    bitn _USBdev USBDevEpIntEn + bic!	\ disable ordinary device interrupt
  else
    dup cells UDCA + off		\ no DMA descriptor
    bitn _USBdev USBDevEPIntEn + or!	\ enable ordinary device interrupt
  endif
;

: initUSBdma	\ --
\ *G Initialise the USB DMA.
  UDCAcfg UDCA 32 cells cmove		\ copy template to RAM
  32 2
  do  i initEPdma  loop
  _USBdev
  UDCA over USBDevCaHead + !		\ set UDCA head pointer
\  $07 swap USBDevDmaIntEn + !		\ enable DMA interrupts with SysErrs
  $03 swap USBDevDmaIntEn + !		\ enable DMA interrupts without SysErrs
;

: setEPdma	\ caddr len ep# --
\ *G Set the next transfer for an endpoint ($0x/$8x).
  EPaddr dup >r  cells UDCA + @ tuck	\ -- caddr dd len dd ; DD for this endpoint
  swap #16 lshift /MaxPacket 5 lshift or \ set cell 1 in descriptor
  $FFFF:FFC0 rot cell + setMask		\   length and packet size
  tuck 2 cells + !			\ buffer address
  0 swap 3 cells + !			\ status
  _USBdev
  1 r> lshift over USBDevEpDmaEn + !	\ enable DMA operation/interrupt for this EP
  0 swap USBDevEpDmaDis + !		\ clear DMA PROCEED flag
;

: #EPtrans	\ ep# -- len
\ *G Return the number of bytes transferred on this endpoint.
\ ** The return value is only valid after a transfer is complete.
  EPaddr cells UDCA + @			\ get descriptor
  3 cells + @ #16 rshift		\ Present_DMA_count
;

: enEPdma	\ ep# --
\ *G Enable endpoint DMA.
  _USBdev  swap EPaddr bitn		\ -- addr mask
  over USBDevEpDmaEn + !		\ enable DMA operation/interrupt
  0 swap USBDevEpDmaDis + !		\ clear DMA PROCEED flag
;

: disEPdma	\ ep# --
\ *G Disable endpoint DMA.
  EPaddr bitn  _USBdev USBDevEpDmaDis + !
;

[then]


\ *****************
\ *S Data and tools
\ *****************

\ Device Interrupt Status Register bits
bit9 equ EPR_INT
bit8 equ EP_RLZED
bit7 equ TxEndPkt
bit6 equ RxEndPkt
bit5 equ CDFULL
bit4 equ CCEMTY
bit3 equ DEV_STAT
bit2 equ EP_SLOW
bit1 equ EP_FAST
bit0 equ FRAME

\ Device Status Register Definitions
#define DEV_CON             0x01
#define DEV_CON_CH          0x02
#define DEV_SUS             0x04
#define DEV_SUS_CH          0x08
#define DEV_RST             0x10

\ Device Address Register Definitions
#define DEV_ADDR_MASK       0x7F
#define DEV_EN              0x80

\ Endpoint Select Register Definitions
#define EP_SEL_F            0x01
#define EP_SEL_ST           0x02
#define EP_SEL_STP          0x04
#define EP_SEL_PO           0x08
#define EP_SEL_EPN          0x10
#define EP_SEL_B_1_FULL     0x20
#define EP_SEL_B_2_FULL     0x40

\ USB Endpoint Callback Events
#define USB_EVT_SETUP       1   /* Setup Packet */
#define USB_EVT_OUT         2   /* OUT Packet */
#define USB_EVT_IN          3   /*  IN Packet */
#define USB_EVT_OUT_NAK     4   /* OUT Packet - Not Acknowledged */
#define USB_EVT_IN_NAK      5   /*  IN Packet - Not Acknowledged */
#define USB_EVT_OUT_STALL   6   /* OUT Packet - Stalled */
#define USB_EVT_IN_STALL    7   /*  IN Packet - Stalled */
#define USB_EVT_OUT_DMA_EOT 8   /* DMA OUT EP - End of Transfer */
#define USB_EVT_IN_DMA_EOT  9   /* DMA  IN EP - End of Transfer */
#define USB_EVT_OUT_DMA_NDR 10  /* DMA OUT EP - New Descriptor Request */
#define USB_EVT_IN_DMA_NDR  11  /* DMA  IN EP - New Descriptor Request */
#define USB_EVT_OUT_DMA_ERR 12  /* DMA OUT EP - Error */
#define USB_EVT_IN_DMA_ERR  13  /* DMA  IN EP - Error */

\ Rx & Tx Packet Length Definitions
#define PKT_LNGTH_MASK      0x000003FF
#define PKT_DV              0x00000400
#define PKT_RDY             0x00000800

create ^USBdev	\ -- addr
\ *G Holds the address of the USB device peripheral.
  _USBdev ,

compiler
: EpIntSt@	\ -- x
\ *G Read the endpoint interrupt status register.
  ^USBdev @ USBdevEpIntSt + @
;

: EpIntClr!	\ x --
\ *G Write the endpoint interrupt clear register.
  ^USBdev @ USBdevEpIntClr + !
;

: DevIntSt@	\ -- x
\ *G Read the device interrupt status register.
  ^USBDev @ USBDevIntSt + @
;

: CmdData@	\ -- x
\ *G Read the command data register.
  ^USBDev @ USBDevCmdData + @
;

: EpRe!		\ x --
\ *G Write the EpRe register.
  ^USBDev @ USBDevEpRe + !
;
target


\ ***************************
\ *S Protocol Engine commands
\ ***************************
\ *P The USB Protocol Engine is triggered by writing to the
\ ** *\fo{UsbDevCmdCode} register. Commands are in the form:
\ *C   00ppcc00
\ *P where *\i{pp} is the command phase and *\i{cc} is the
\ ** command code.
\ ** For more details, read the source code.

\ Command Codes
#define CMD_SET_ADDR        0x00D00500
#define CMD_CFG_DEV         0x00D80500
#define CMD_SET_MODE        0x00F30500
#define CMD_RD_FRAME        0x00F50500
#define DAT_RD_FRAME        0x00F50200
#define CMD_RD_TEST         0x00FD0500
#define DAT_RD_TEST         0x00FD0200
#define CMD_SET_DEV_STAT    0x00FE0500
#define CMD_GET_DEV_STAT    0x00FE0500
#define DAT_GET_DEV_STAT    0x00FE0200
#define CMD_GET_ERR_CODE    0x00FF0500
#define DAT_GET_ERR_CODE    0x00FF0200
#define CMD_RD_ERR_STAT     0x00FB0500
#define DAT_RD_ERR_STAT     0x00FB0200
#define CMD_CLR_BUF         0x00F20500
#define DAT_CLR_BUF         0x00F20200
#define CMD_VALID_BUF       0x00FA0500

\ USB Control Definitions
#define CTRL_RD_EN          0x00000001
#define CTRL_WR_EN          0x00000002

\ Device Configure Register Definitions
#define CONF_DVICE          0x01

\ These masks contain 16 bits, set if the endpoint matches the feature.
#define EP_MSK_CTRL 0x0001      \ Control Endpoint Logical Address Mask
#define EP_MSK_BULK 0xC924      \ Bulk Endpoint Logical Address Mask
#define EP_MSK_INT  0x4492      \ Interrupt Endpoint Logical Address Mask
#define EP_MSK_ISO  0x1248	\ Isochronous Endpoint Logical Address Mask

6 equ #MaxBulkEPs	\ max number of bulk endpoints
4 equ #MaxIsoEPs	\ max number of isochronous emdpoints
5 equ #MaxIntEPs	\ max number of interrupt endpoints

\ SIE Endpoint Status Register Definitions
#define EP_STAT_ST          0x01
#define EP_STAT_DA          0x20
#define EP_STAT_RF_MO       0x40
#define EP_STAT_CND_ST      0x80

compiler
: DAT_WR_BYTE(x)	\ x -- x'
  #16 lshift  0x00000100 or  ;
: CMD_SEL_EP(x)		\ x -- x'
  #16 lshift  0x00000500 or  ;
: DAT_SEL_EP(x)		\ x -- x'
  #16 lshift  0x00000200 or  ;
: CMD_SEL_EP_CLRI(x)	\ x -- x'
  #16 lshift  0x00400500 or  ;
: DAT_SEL_EP_CLRI(x)	\ x -- x'
  #16 lshift  0x00400200 or  ;
: CMD_SET_EP_STAT(x)	\ x -- x'
  #16 lshift  0x00400500 or  ;
target


\ *P Commands for the LPC2xxx USB controller are 32 bit items
\ ** of the form $aabb:0500. When data is also written the 8 bit
\ ** data is written as a 32 bit item of the form $00bb:0100.
\ ** When a data byte is to be read, the read is performed by
\ ** reissuing the command in the form $aabb:0200, and waiting
\ ** for the CDFULL interrupt rather than the CCEMTY interrupt.

IsLeaf
 : WrCmd	\ cmd --
\ *G Write *\i{cmd} to the CmdCode register.
  ^USBdev @				\ -- cmd dev
  CCEMTY over USBDevIntClr + !		\ clear interrupt
  tuck USBDevCmdCode + !		\ write command ; -- dev
  begin					\ wait for completion
    dup USBDevIntSt + @ CCEMTY and
  until
  drop
;

: WrCmdDat	\ cmd val --
\ *G Write *\i{cmd} followed by *\i{val} to the CmdCode register.
  swap WrCmd  #16 lshift $0100 or WrCmd
;

IsLeaf
: RdCmdDat	\ cmd -- val
\ *G Write a command and read the data.
  ^USBdev @				\ -- cmd dev
  CCEMTY CDFULL or over USBDevIntClr + !  \ clear interrupts
  2dup USBDevCmdCode + !		\ write command
  begin					\ wait for completion ; -- cmd dev
    dup USBDevIntSt + @ CCEMTY and
  until
  CCEMTY over USBDevIntClr + !		\ clear interrupt
  swap $FFFF0000 and $0200 or		\ read command
  over USBDevCmdCode + !		\ set command ; -- dev
  begin					\ wait for completion
    dup USBDevIntSt + @ CDFULL and
  until
  CDFULL over USBDevIntClr + !		\ clear interrupt
  USBDevCmdData + @ $FF and
;


\ *********************
\ *S Handling endpoints
\ *********************
\ *P This section contains an array of endpoint handlers and default
\ ** actions for other operations which are hardware dependent.

create EPxts	\ -- addr
\ *G Holds the xts of the 16 endpoint handler words. The action
\ ** of each handler must have the stack effect below:
\ *C   event --
  ' drop ,  ' drop ,  ' drop ,  ' drop ,
  ' drop ,  ' drop ,  ' drop ,  ' drop ,
  ' drop ,  ' drop ,  ' drop ,  ' drop ,
  ' drop ,  ' drop ,  ' drop ,  ' drop ,

: EPhandler	\ event u --
\ *G Run the endpoint handler associated with endpoint *\i{u},
\ ** where endpoints are numbered 0..15. Each endpoint handler
\ ** is passed an event type which allows it to determine what
\ ** it has to do. The action of each handler must have the stack
\ ** effect below.
\ *C   event --
  $0F and cells EPxts + @ execute
;

interpreter
: SetEPhandler	\ xt ep# --
\ *G Install the endpoint handler for endpoint *\i{ep#}.
\ ** Used during interpretation.
  $0F and cells EPxts + !
;
target

0 value UsbDevSt	\ -- x
\ *G USB device Status.

: USBsuspend	\ --
\ *G Suspend operations are performed by the hardware. This
\ ** is a hook for installing actions such as toggling an LED.
;

: USBresume	\ --
\ *G Resume operations are performed by the hardware. This
\ ** is a hook for installing actions such as toggling an LED.
;

: USBconnect	\ --
\ *G Connect to the USB bus.
  [defined] USBhwconnect [if] USBhwconnect [then]	\ FVD005
  CMD_SET_DEV_STAT DEV_CON WrCmdDat
;

: USBdisconnect	\ --
\ *G Disconnect from the USB bus.
  CMD_SET_DEV_STAT 0 WrCmdDat
  [defined] USBhwdisconnect [if] USBhwdisconnect [then]	\ FVD005
  0 -> UsbConfig			\ FVD005
;

: USBWakeUp	\ --
\ *G Called automatically on USB Remote Wakeup.
  UsbDevSt USB_GETSTATUS_REMOTE_WAKEUP and if
    CMD_SET_DEV_STAT DEV_CON WrCmdDat
  endif
;

: USBSetAddress	\ addr --
\ *G Set the USB assigned address in the protocol engine.
  DEV_EN or  CMD_SET_ADDR swap WrCmdDat	\ takes effect after status in
;

IsLeaf
: WaitEPrlzed	\ --
\ *G Wait for the EP_RLZED interrupt status and clear it.
  ^USBdev @
  begin
    dup USBDevIntSt + @ EP_RLZED and
  until
  EP_RLZED swap UsbDevIntClr + !
;

: USBConfigure	\ cfg -- ; 0=unconfigure
\ *G Set the SIE configuration state.
  0<> CONF_DVICE and  CMD_CFG_DEV swap WrCmdDat
  $0000:0003 ^USBdev @ USBDevEpRe + !
  WaitEPrlzed
;

((
: USBConfigEP	\ desc --
\ *G Configure USB Endpoint according to Descriptor.
  dup epd.Addr c@ EPaddr
  dup bitn _USBdev USBDevEpRe + or!	\ realise using mask
  _USBdev USBDevEpIndx + !		\ set endpoint
  epd./MaxPacket lew@
  _USBdev USBDevEpSize + !		\ set packet size
  WaitEPrlzed
;
))

: USBConfig2EP	\ desc ep# -- desc
\ *G Configure USB Endpoint according to Descriptor and EndPoint.
  over
\ USBConfigEP
  dup epd.Addr c@ EPaddr
  dup bitn _USBdev USBDevEpRe + or!	\ realise using mask
  _USBdev USBDevEpIndx + !		\ set endpoint
  epd./MaxPacket lew@
  _USBdev USBDevEpSize + !		\ set packet size
  WaitEPrlzed				\ -- desc
\ dup USBenableEP
  dup EPaddr CMD_SET_EP_STAT(x) 0 WrCmdDat
\ USBResetEP
  EPaddr CMD_SET_EP_STAT(x) 0 WrCmdDat
;

compiler
: USBDirCtrlEP	\ dir -- ; 0=out
\ *G This word is a dummy for this hardware.
  drop
;
target

((
: USBEnableEP	\ ep# -- ; bit7=dir, bits3:0=num
\ *G Enable the endpoint.
  EPaddr CMD_SET_EP_STAT(x) 0 WrCmdDat
;
))

: USBDisableEP	\ ep# -- ; bit7=dir, bits3:0=num
\ *G Disable the endpoint.
  EPaddr CMD_SET_EP_STAT(x) EP_STAT_DA WrCmdDat
;

((
: USBResetEP	\ ep# -- ; bit7=dir, bits3:0=num
\ *G Reset the endpoint.
  EPaddr CMD_SET_EP_STAT(x) 0 WrCmdDat
;
))

variable SieModeMask	\ -- addr
\ *G A shadow variable that holds the last mode mask set.

: setSIEmode	\ mask --
\ *G Set the USB device mode register in the SIE. A shadow
\ ** variable is used to hold the current state.
  dup SieModeMask !  CMD_SET_MODE swap WrCmdDat
;

((
  : EnBulkNAKin	\ --
  \ *G Enable NAK interrupts on bulk IN endpoints.
  \ ** All other bits are cleared.
    bit5 setSIEmode
  ;

  : EnBulkNAKout	\ --
  \ *G Enable NAK interrupts on bulk OUT endpoints.
  \ ** All other bits are cleared.
    bit6 setSIEmode
  ;

  : EnBulkNAKs	\ --
  \ *G Enable NAK interrupts on bulk IN and OUT endpoints.
  \ ** All other bits are cleared.
    bit5 bit6 or setSIEmode
  ;
))

: +BulkNAKin	\ --
\ *G Enable NAK interrupts on bulk IN endpoints.
\ ** All other SIE mode bits are left alone.
  SieModeMask @ bit5 or setSIEmode
;

: -BulkNAKin	\ --
\ *G Disable NAK interrupts on bulk IN endpoints.
\ ** All other SIE mode bits are left alone.
  SieModeMask @ bit5 invert and setSIEmode
;

: +BulkNAKout	\ --
\ *G Enable NAK interrupts on bulk OUT endpoints.
\ ** All other SIE mode bits are left alone.
  SieModeMask @ bit6 or setSIEmode
;

: -BulkNAKout	\ --
\ *G Disable NAK interrupts on bulk OUT endpoints.
\ ** All other SIE mode bits are left alone.
  SieModeMask @ bit6 invert and setSIEmode
;

: USBSetStallEP	\ ep# -- ; bit7=dir, bits3:0=num
\ *G Stall the endpoint. This word *\b{must} clear the relevant
\ ** bit in *\fo{UsbEpHalt}.
  dup EPaddr CMD_SET_EP_STAT(x) EP_STAT_ST WrCmdDat
  EP>mask addr UsbEpHalt or!
;

: USBClrStallEP	\ ep# -- ; bit7=dir, bits3:0=num
\ *G Clear the endpoint stall. This word *\b{must} clear the relevant
\ ** bit in *\fo{UsbEpHalt}, and should ensure that bulk IN/OUT NAKs cause
\ ** interrupts.
\ cr ." CES " dup .byte space .MSCstate
  dup EPaddr CMD_SET_EP_STAT(x) 0 WrCmdDat	\ clear stall
  dup EP>mask addr UsbEpHalt bic!	\ clear halt status bit
  $080 and if				\ IN endpoint?
    +BulkNAKin
[ usbDMA? 0= ] [if]			\ only enable BO ints for non-DMA
  else
    +BulkNAKout
[then]
  endif
;

((
: USBReadEP	\ ep# addr -- #read ; bit7=dir, bits3:0=num
 \ *G Read endpoint data from endpoint *\i{ep#} to *\i{addr}.
 \ ** It will be faster if *\i{addr} is aligned. Return the
 \ ** number of bytes read.
 \ ** Note that data is transferred in units of four bytes, so
 \ ** the buffer must be large enough to accommodate up to three
 \ ** extra bytes.
  swap dup >r				\ -- addr ep# ; R: -- ep#
  $0F and  2 lshift CTRL_RD_EN or
  ^USBdev @ tuck UsbDevCtrl + !		\ -- addr usb ; R: -- ep#
  0 begin				\ wait for packet ready
    drop  dup USBDevRxLen + @
    dup PKT_RDY and
  until
  PKT_LNGTH_MASK and			\ -- addr usb #read
  tuck swap 2swap			\ -- #read usb addr #read
  over 3 and 0= if			\ aligned address, fast
    3 + -4 and bounds ?do		\ -- #read usb ; read data to memory
      dup USBDevRxData + @ i !		\ write as a CELL
    cell +loop
  else					\ unaligned address, slow
    3 + -4 and bounds ?do		\ -- #read usb ; read data to memory
      dup USBDevRxData + @ i lel!	\ write as four bytes
    cell +loop
  endif
  0 swap UsbDevCtrl + !			\ -- #read
  EP_MSK_ISO r@ $0F and rshift		\ 16 bits, set if isochronous EP.
  1 and 0= if				\ not an isochronous endpoint
    r@ EPaddr CMD_SEL_EP(x) WrCmd
    CMD_CLR_BUF WrCmd
  endif
  r> drop
;
))

code portl>	\ caddr len port --
\ *G Read the port into the the memory block a cell at a time.
\ ** If *\i{caddr} is aligned, the data is stored cell by cell,
\ ** otherwise it is assembled a byte at a time before being sent
\ ** to the port.
  ldr      r0, [ psp ], # 4		\ r0=len
  ldr      r1, [ psp ], # 4		\ r1=caddr
  add      r0, r0, # 3			\ align length
  bic .s   r0, r0, # 3
  ne, if,				\ if something to do
\ r0=len, r1=caddr, tos=port, r3..r8=scratch
    and .s   r2, r1, # 3		\ address aligned?
    eq, if,				\ yes
      cmp      r0, # 16
      ge, if,
        begin,
          sub      r0, r0, # 16
          ldr      r4, [ tos ]
          ldr      r5, [ tos ]
          ldr      r6, [ tos ]
          ldr      r7, [ tos ]
          cmp      r0, # 16
          stmia    r1 ! { r4-r7 }
        lt, until,
      endif,
      cmp      r0, # 0
      ne, if,
        begin,
          ldr      r3, [ tos ]		\ read port
          sub .s   r0, r0, # 4
          str      r3, [ r1 ], # 4	\ store cell
        eq, until,
      endif,
    else,				\ unaligned
      begin,
        ldr      r3, [ tos ]		\ read port
        sub .s   r0, r0, # 4
        strb     r3, [ r1 ], # 1	\ store byte by byte
        mov      r3, r3 lsl # 8
        strb     r3, [ r1 ], # 1
        mov      r3, r3 lsl # 8
        strb     r3, [ r1 ], # 1
        mov      r3, r3 lsl # 8
        strb     r3, [ r1 ], # 1
      eq, until,
    endif,
  endif,
  ldr      tos,  [ psp ], # 4		\ restore TOS
  mov      pc, link
end-code


: USBReadEP	\ ep# addr -- #read ; bit7=dir, bits3:0=num
\ *G Read endpoint data from endpoint *\i{ep#} to *\i{addr}.
\ ** It will be faster if *\i{addr} is aligned. Return the
\ ** number of bytes read.
\ ** Note that data is transferred in units of four bytes, so
\ ** the buffer must be large enough to accommodate up to three
\ ** extra bytes.
  swap dup >r				\ -- addr ep# ; R: -- ep#
  $0F and  2 lshift CTRL_RD_EN or
  ^USBdev @ tuck UsbDevCtrl + !		\ -- addr usb ; R: -- ep#
  0 begin				\ wait for packet ready
    drop  dup USBDevRxLen + @
    dup PKT_RDY and
  until
  PKT_LNGTH_MASK and			\ -- addr usb #read
  swap 2dup 2>r	USBDevRxData + portl> 2r>	\ -- #read usb
  0 swap UsbDevCtrl + !			\ -- #read
  EP_MSK_ISO r@ $0F and rshift		\ 16 bits, set if isochronous EP.
  1 and 0= if				\ not an isochronous endpoint
    r@ EPaddr CMD_SEL_EP(x) WrCmd
    CMD_CLR_BUF WrCmd
  endif
  r> drop
;


((
: USBWriteEP	\ ep# addr len -- cnt
 \ *G Write data *\i{addr/len} to host from endpoint *\i{ep#}.
 \ ** Return the number of bytes written.
  dup >r  rot dup >r
  $0F and  2 lshift CTRL_WR_EN or	\ write ep# to control
  ^USBdev @ tuck UsbDevCtrl + !		\ -- addr len usb ; R: -- len ep#
  2dup UsbDevTxLen + !			\ write length
  -rot					\ -- usb addr len ; R: -- len ep#
  over 3 and 0= if			\ if the address is aligned, fast
    3 + -4 and bounds ?do		\ -- usb ; R: -- len ep#
      i @ over UsbDevTxData + !		\ fetch by CELL
    cell +loop
  else					\ address not aligned, slow
    3 + -4 and bounds ?do		\ -- usb ; R: -- len ep#
      i lel@ over UsbDevTxData + !	\ fetch four bytes
    cell +loop
  endif
  0 swap UsbDevCtrl + !			\ essential for stall handling
  r> EPaddr CMD_SEL_EP(x) WrCmd
  CMD_VALID_BUF WrCmd
  r>
;
))

code >portl	\ caddr len port --
\ *G Transfer the memory block to the port a cell at a time.
\ ** If *\i{caddr} is aligned, the data is fetched cell by cell,
\ ** otherwise it is assembled a byte at a time before being sent
\ ** to the port.
  ldr      r0, [ psp ], # 4		\ r0=len
  ldr      r1, [ psp ], # 4		\ r1=caddr
  add      r0, r0, # 3			\ align length
  bic .s   r0, r0, # 3
  ne, if,				\ if something to do
    and .s   r2, r1, # 3		\ address aligned?
    eq, if,				\ yes
\ r0=len, r1=caddr, tos=port, r3..r8=scratch
      cmp      r0, # 16
      ge, if,
        begin,
          sub      r0, r0, # 16		\ count length down
          ldmia    r1 ! { r4-r7 }	\ read 16 bytes
          cmp      r0, # 16
          str      r4, [ tos ]		\ store 16 bytes to port
          str      r5, [ tos ]
          str      r6, [ tos ]
          str      r7, [ tos ]
        lt, until,
      endif,
      cmp      r0, # 0
      ne, if,
        begin,
          ldr      r3, [ r1 ], # 4	\ fetch cell
          sub .s   r0, r0, # 4
          str      r3, [ tos ]		\ transfer to port
        eq, until,
      endif,
    else,				\ unaligned
      begin,
        ldrb     r3, [ r1 ], # 1	\ fetch byte by byte
        ldrb     r4, [ r1 ], # 1
        orr      r3, r3, r4 .lsl # 8
        ldrb     r4, [ r1 ], # 1
        orr      r3, r3, r4 .lsl # #16
        ldrb     r4, [ r1 ], # 1
        orr      r3, r3, r4 .lsl # #24
        sub .s   r0, r0, # 4
        str      r3, [ tos ]		\ transfer to port
      eq, until,
    endif,
  endif,
  ldr      tos,  [ psp ], # 4		\ restore TOS
  mov      pc, link
end-code

: USBWriteEP	\ ep# addr len -- cnt
\ *G Write data *\i{addr/len} to host from endpoint *\i{ep#}.
\ ** Return the number of bytes written.
  dup >r  rot dup >r
  $0F and  2 lshift CTRL_WR_EN or	\ write ep# to control
  ^USBdev @ tuck UsbDevCtrl + !		\ -- addr len usb ; R: -- len ep#
  2dup UsbDevTxLen + !			\ write length
  dup >r UsbDevTxData + >portl r>	\ -- usb ; R: -- len ep#
  0 swap UsbDevCtrl + !			\ essential for stall handling
  r> EPaddr CMD_SEL_EP(x) WrCmd
  CMD_VALID_BUF WrCmd
  r>
;


\ *****************
\ *S Initialisation
\ *****************

: hwUSBreset	\ --
\ *G Reset the USB device hardware after access has been enabled.
  ^USBdev @				\ -- addr
\ Set up the control endpoint, 0=read, 1=write
  0 over USBDevEpIndx + !
  /MaxPacket0 over USBDevEpSize + !
  1 over USBDevEpIndx + !
  /MaxPacket0 over USBDevEpSize + !
  waitEPrlzed
  $FFFFFFFF over USBDevEpIntClr + !	\ clear all EP ints
  0         over USBDevEpIntPri + !	\ all SLOW
  $FFFFFFFF over USBDevEpIntEn + !	\ enable all EP ints, no DMA
  $FFFFFFFF over USBDevIntClr + !	\ clear device ints
  DEV_STAT EP_SLOW or FRAME or
            over USBdevIntEn + !	\ enable some device ints

  0         over USBDevDmaIntEn + !	\ no DMA interrupts
  0         over USBDevCaHead + !	\ no DMA yet
  $FFFFFFFF over USBDevEpDmaDis + !	\ disable all DMA endpoints
  $FFFFFFFF over USBDevDmaReqClr + !	\ clear pending DMA interrupts
  $FFFFFFFF over USBDevEotIntClr + !	\ clear pending EOT interrupts
  $FFFFFFFF over USBDevDdrIntClr + !	\ clear pending DDreq interrupts
  $FFFFFFFF over USBDevSysErrIntClr + !	\ clear pending SysErr interrupts
  drop
;

: USBswReset	\ --
\ *G The action performed for a software reset of the USB, either
\ ** at reset or in response to a USB reset.
  hwUSBreset				\ primary reset
[ usbDMA? ] [if]
  initUSBdma
[then]
  UsbResetChain ExecChain		\ reset core layer and class drivers
  0 USBSetAddress			\ bug in some LPC variants
;


\ ******************************
\ *S Endpoint interrupt handling
\ ******************************

IsLeaf
: waitCmdData	\ -- x
\ *G Wait for CDFULL and read the data
  ^USBdev @
  begin  dup USBDevIntSt + @ CDFULL and  until
  USBDevCmdData + @
;

: doEPslow	\ --
\ *G Process an EP_SLOW interrupt.

\ Process EP0 OUT to device
  EpIntSt@ bit0 and if
    bit0 EpIntClr!
    WaitCmdData EP_SEL_STP and if	\ SETUP request?
[ dbgUHW? ] [if]  ." Setup "  [then]
      USB_EVT_SETUP 0 EPhandler
    else
      USB_EVT_OUT 0 EPhandler
    endif
  endif

\ Process EP0 IN to host
  EpIntSt@ bit1 and if
    bit1 EpIntClr!
    waitCmdData drop
    USB_EVT_IN 0 EPhandler
  endif

\ Process other endpoints
  $0000:0004  #EPs 2 do		\ -- mask
    EpIntSt@ ?dup 0= ?leave	\ get out if nothing to do
    over and if
      dup EpIntClr!
      waitCmdData drop
      i 1 and 0= if
        USB_EVT_OUT i 1 rshift EPhandler
      else
        USB_EVT_IN i 1 rshift EPhandler
      endif
    endif
    1 lshift
  loop
  drop
;

usbDMA? [if]

: doDmaEOT	\ --
\ *G Handle DMA End Of Transfer interrupt.
  _USBdev USBDevEotIntSt + @ dup -4 and	\ remove phys 0/1
  #EPs 2 do				\ -- st st'
    dup 1 i lshift and if
      i bitn invert and			\ clear bit in st'
      i 1 and 0= if
        USB_EVT_OUT_DMA_EOT i 1 rshift EPhandler
      else
        USB_EVT_IN_DMA_EOT i 1 rshift EPhandler
      endif
    endif
    dup 0= ?leave			\ get out if nothing left
  loop
  drop  _USBdev USBDevEotIntClr + !
;

: doDmaNDDReq	\ --
\ *G Handle DMA New DD request interrupt.
  _USBdev USBDevDdrIntSt + @ dup -4 and	\ remove phys 0/1
  #EPs 2 do				\ -- st st'
    dup i bitn and if
      i bitn invert and			\ clear bit in st'
      i 1 and 0= if
        USB_EVT_OUT_DMA_NDR i 1 rshift EPhandler
      else
        USB_EVT_IN_DMA_NDR i 1 rshift EPhandler
      endif
    endif
    dup 0= ?leave			\ get out if nothing left
  loop
  drop  _USBdev USBDevDdrIntClr + !
;

\ *P USB Ssytem Errors are AHB bus errors. As yet, the AHB bus
\ ** configuration registers are undocumented. Consequently
\ ** there is no point in trying to handle these errors, and the
\ ** code framework to do so is commented out.
((
: doDmaSysErr	\ --
\ *G Handle DMA System Error interrupt.
  _USBdev USBDevSysErrIntSt + @ dup -4 and	\ remove phys 0/1
  #EPs 2 do				\ -- st st'
    dup 1 i lshift and if
      i bitn invert and			\ clear bit in st'
      i 1 and 0= if
        USB_EVT_OUT_DMA_ERR i 1 rshift EPhandler
      else
        USB_EVT_IN_DMA_ERR i 1 rshift EPhandler
      endif
    endif
    dup 0= ?leave			\ get out if nothing left
  loop
  drop  _USBdev USBDevSysErrIntClr + !
;
))

[then]

dbgUHW? [if]  variable #USBints  [then]

: doUSBint	\ disr -- disr'
\ *G The primary word called from *\fo{USBinterrupt} below.
\ ** Processes DEV_STAT, EP_SLOW and FRAME interrupt sources,
\ ** ignoring and clearing all others. If DMA is enabled, DMA
\ ** interrupts are also processed.
  console dup opvec ! ipvec !  decimal	\ for debugging
[ dbgUHW? ] [if]  cr ." USBstat = " dup .dword space  1 #USBints +!  [then]

  dup DEV_STAT and if
    DEV_STAT ^USBDev @ USBDevIntClr + !	\ See LPC2148 User Manual 25 July 2006
    CMD_GET_DEV_STAT WrCmd
    DAT_GET_DEV_STAT RdCmdDat		\ -- disr stat
    dup DEV_RST and if			\ Reset
      drop USBswReset USBconnect exit	\ FVD005
    endif
    dup DEV_CON_CH and if		\ Connect Change
      drop exit
    endif
    dup DEV_SUS_CH and if
      DEV_SUS and
      if  USBsuspend  else  USBresume  endif
      exit
    endif
    drop
  endif

  dup EP_SLOW and if
[ dbgUHW? ] [if] ." EPslow " [then]
    doEPslow
  endif

  dup FRAME and if
[ dbgUHW? ] [if] ." FRAME " [then]
    +BulkNAKin
[ usbDMA? 0= ] [if] +BulkNAKout [then]	\ non DMA only
    UsbFrameChain ExecChain
  endif

[ usbDMA? ] [if]
  _USBdev USBDevDmaIntSt + @ bit0 and if
    doDmaEOT
  endif
  _USBdev USBDevDmaIntSt + @ bit1 and if
    doDmaNDDReq
  endif
((
  _USBdev USBDevDmaIntSt + @ bit2 and if
    doDmaSysErr
  endif
))
[then]
;

: USBinterrupt	\ --
\ *G The USB interrupt handler. It calls *\fo{doUSBint} above to
\ ** permit better factoring.
  usbLED1on
  ^USBDev @ USBDevIntSt + @
  doUSBint
  ^USBDev @ USBDevIntClr + !
  usbLED1off
;
' USBinterrupt IRQ: USB_ISR	\ -- addr
\ *G The entry point for the USB interrupt.

: USBintInit	\ --
\ *G Initialise USB hardware interrupt handler.
  USB_ISR USBint# USBpri# SetIrqIsr	\ install interrupt handler
  DEV_STAT ^USBdev @ USBDevIntEn + !	\ enable USB interrupt device
;

: InitUSB	\ --
\ *G Initialise the USB hardware.
  USBhwInit USBintInit USBswReset	\ all primary initialisation
  0 USBSetAddress			\ LPC2148 Manual section 9.1
;

\ SFP006...
: startUSB	\ --
\ *G Start the USB system. This word gives more control to devices
\ ** that have VBUS control. If such devices are rebooted without
\ ** power down, they may/will need to perform a *\fo{USBdisconnect}
\ ** operation as part of the start up sequence. A disconnect,
\ ** delay, connect sequence is part of *\fo{startUSB}.
  initUSB  USBdisconnect  #1000 ms  USBconnect  ;
\ ...SFP006

dbgUHW? [if]
_USB constant _USB
_USBRAM constant _USBRAM
[then]


\ **************************
\ *S Development and Gotchas
\ **************************
\ *P Under some conditions, normally after testing a faulty USB
\ ** device, the Windows USB system can become faulty. During
\ ** device development, rebooting your PC will be a common
\ ** occurrence.

\ *P When using console debug messages, note that you must not use
\ ** serial drivers with queued output. These use *\fo{PAUSE}
\ ** which *\b{must not} be called from an interrupt handler.


\ ======
\ *> ###
\ ======

decimal
