\ devRDB4078.fth - hardware set up for Code Red RDB4078 board

((
Copyright (c) 2008, 2011, 2014, 2015, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20160208 SFP004 More conversion work.
20150702 SFP003 Converted to Code Red RDB4078 device port.
20140402 BJC002 Conversion to CCJ-SP05 PCB using LPC1788 chip.
		Should be easily portable to other boards with similar chips.
		NOTE: Main differences between this file and LPC17xx is pin
		configuration & clock register bits.  Otherwise driver same
		as LPC17xx.
20110817 SFP001 Conversion to device connection of Olimex
                LPC-1766STK board.
))

only forth definitions
decimal

\ ===========
\ *! devRBD4078.fth
\ *T USB device set up for Code Red RBD4078
\ ===========
\ *P The file *\i{devRBD4078.fth} provides USB device mode
\ ** initialisation and LED drivers for Code Red RBD4078
\ ** boards. It will require minor changes for other LPC devices
\ ** and boards, in particular in the USB Connect function and
\ ** LEDs.


\ *******
\ *S LEDs
\ *******
\ *P This board uses the hardware driven 'up' LED, so these words are stubs.

((
: initLEDs	\ --
\ *G Initialise the two LEDs to off.
  LED1val LED1mask  _PINSEL3 setMask	\ configure LED1
  LED2val LED2mask  _PINSEL0 setMask	\ configure LED2
  LED1bit _FIO1
  2dup FIODIR + or!			\ set DIR bits
  FIOSET + !				\ set high
  LED2bit _FIO0
  2dup FIODIR + or!			\ set DIR bits
  FIOSET + !				\ set high
;
' initLEDs AtCold
))

compiler
: usbLED1on	\ --
\ *G Turn LED1 on.
;
: usbLED1off	\ --
\ *G Turn LED1 off.
;
: usbLED2on	\ --
\ *G Turn LED2 on.
;
: usbLED2off	\ --
\ *G Turn LED2 off.
;
target

\ ******************************
\ *S USB clock PLL configuration
\ ******************************

interpreter
[-short-branches
: gen-PLL1	\ -- Mval Pval
\ *G INTERPRETER: Generate the data for the PLL1CFG register.
\ ** The PLL input, Fin, must lie in the range 10 to 25 MHz.
\ ** The PLL osc. output, Fcco, must lie in the range 155 and 320 MHz.
\ ** Fcco = (Fin * M * 2 * P)
\ ** The final PLL output is (Fin * M)
\ ** Mval and Pval are the register values to use Mval = M-1
  { | P bestP bestM -- Mval Pval }
  usb-speed xtal-speed u< if
    cr ." USB speed less than xtal speed" abort
  endif
  xtal-speed #10 MHZ u< if
    cr ." Xtal frequency too low to use for PLL input!" abort
  endif
  1	to P				\ initial P
  -1 to bestP				\ best P is invalid
  -1 to bestM				\ best M invalid
  -1					\ maximum for best error difference
  4 0 do				\ scan possible Pvalues (1, 2, 4, 8)
    33 1 do				\ scan possible M values (1 .. 32)
      P 2* i * xtal-speed *		\ calc Fcco; 2 * P * M * Fin
      dup #156 MHz #320 MHz within? if	\ .. within PLL range
        P 2* /				\ calc PLL output frequency
        usb-speed
      	2dup u<				\ absolute difference unsigned
        if  swap  endif
	-
      	dup 2 pick u< if		\ lower error than last guess
      	  nip				\ leave new error
      	  i to bestM
      	  P to bestP
      	else
          drop
      	endif
      else
      	drop
      endif
    loop
    dup 0= ?leave			\ stop if found a solution with low P value
    P 2* to P				\ .. next P
  loop
  bestM 0< if
    cr ." Acceptable Mval not found!"  abort
  endif
  cr ." Alt PLL: best match found; P = " bestP 1 u.r
  ." , M = " bestM 1 u.r ."  (Fcco = " bestM xtal-speed * bestP 2* * 1 MHZ / 1 u.r
  ." MHz), error = " 1 u.r ." Hz"
  bestM 1-						\ Mval
  case bestP 						\ Pval
    1 of 0 endof
    2 of 1 endof
    4 of 2 endof
    8 of 3 endof
  endcase
;
short-branches]
target

usb-speed xtal-speed <> [if]
  gen-PLL1 equ P1val  equ M1val
  P1val #5 lshift  M1val or  equ PLL1CFGval
[then]

usb-speed 48 MHz / equ USBCLKSELval	\ -- 1/2/3
\ *G Divider value.
USBCLKSELval 0=  USBCLKSELval 3 >  or [if]
  cr ." Invalid USBCLKSEL divider value!"  abort
[endif]


\ **********************
\ *S USB hardware set up
\ **********************
\ *P The primary USB setup is performed in *\i{UsbHwLPC1xxx.fth}.
\ ** Because LPC17xx/LPC40xx variants and hardware differ, the
\ ** actual hardware initialisation is performed here.
\ ** These port assignments are for the USB2 port.
\ *E USB_UP_LED2   P0.13        I/O type A
\ ** USB_CONNECT2  P0.14        I/O type D
\ ** USB_VBUS2     P1.30        I/O type A
\ ** USB_D+2       P0.31        I/O type U
\ ** USB_D-2       USB_D2       dedicated pin

~USBCLK_DEV_CLK ~USBCLK_PORTSEL_CLK or ~USBCLK_AHB_CLK or
  equ		~USB_ALL_CLOCKS

~USBCLK_DEV_CLK ~USBCLK_AHB_CLK or
  equ   ~USB_RUN_CLOCKS

: doPLL1Feed	\ -- ;
\ *G Send the PLL feed sequence to PLL1.
  [i
  _SYSCON
  $AA over PLL1FEED + !			\ feed sequence to load PLL
  $55 swap PLL1FEED + !
  i]
;

: ConfigUSBclock	\ --
\ *G Configure the USB clock using PLL1.
  [ cr ." Configuring PLL1 with PLL1CFG val $" PLL1CFGval .hex ]
  _SYSCON
  0 over USBCLKSEL + !			\ disable clock
  0 over PLL1CON + !  doPLL1Feed	\ disable PLL
  begin					\ wait for lock
    pause
    dup PLL1STAT + @ bit8 and 0=	\ test PLLE status
  until
  PLL1CFGval over PLL1CFG + !  doPLL1Feed  \ set PLL divider
  bit0 over PLL1CON + !  doPLL1Feed	\ enable PLL
  begin					\ wait for lock
    pause
    dup PLL1STAT + @ bit10 and		\ test PLOCK
  until
  USBCLKSELval over USBCLKSEL + !	\ set up  clock divider, still using SYSCLK source
  $0200 USBCLKSELval or  over USBCLKSEL + !  \ switch USB core clock to use PLL1.
  drop
;

: USBhwInit	\ --
\ *G Enable access to the USB Device peripheral.
\ ** The main clock PLL is set in the start up code.
  ~pcusb _PCONP  or!			\ USB power on

  ConfigUSBclock			\ Configure USB clocking

  _USB
  $01A over USBClkCtrl + !		\ enable AHB, Portsel, USB device clocks
  begin					\ turn on clocks
    pause
    dup USBClkSt + @ $01A and $01A =	\ check status ; SFP004
  until
  $03 over USBPortSel + !		\ use USB2 port for device
  $012 over USBClkCtrl + !		\ just AHB & USB device clocks (portsel now hidden)
  drop

  \ configure IO pins
\ *E USB_UP_LED2   P0.13        I/O type A
\ ** USB_CONNECT2  P0.14        I/O type D
\ ** USB_VBUS2     P1.30        I/O type A
\ ** USB_D+2       P0.31        I/O type U
\ ** USB_D-2       USB_D-2      dedicated pin
  _IOCON
  AF1  over IOCON_P0_31 + !		\ USB D+2, USB D-2 is dedicated
  AF3  over IOCON_P0_14 + !		\ USB Connect2, Type D
  AF1 bit7 or  over IOCON_P0_13 + !	\ 'up' LED indicator, Type A, digital mode
  AF2 bit7 or  over IOCON_P1_30 + !	\ USB Vbus sense, Type A, digital mode
  drop
;

_PCONP constant _PCONP
_USB constant _USB

\ not needed if software connect works right or is available
\ comment these defs if soft connect is used.
((
: USBhwconnect	\ --
 \ *G Connect to the USB bus.
  BIT19 _FIO1 FIOCLR + !
;
: USBhwdisconnect	\ --
 \ *G Disconnect from the USB bus.
  BIT19 _FIO1 FIOSET + !
;
))


\ ======
\ *> ###
\ ======

decimal

