\ EB55_DATAFLASH.FTH - AT91EB55 SPI DataFlash Driver

((
Copyright (c) 2002
MicroProcessor Engineering and Electricity Data Services

MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20020924 CFB001 First release
))

only forth definitions  decimal


\ *****************
\ *! EB55_DataFlash
\ *T SPI Data Flash driver for AT91EB55 Evaluation Kit
\ *****************
\ *P This code drives the DataFlash devices connected to
\ ** SPI port on the EB55 board.

\ *******
\ Equates
\ *******
  $68 equ CONT_ARRAY_READ
  $52 equ PAGE_READ
  $53 equ PAGE_TO_BUFFER1
  $55 equ PAGE_TO_BUFFER2
  $57 equ STATUS_READ
  $84 equ BUFFER1_WRITE
  $87 equ BUFFER2_WRITE
  $83 equ BUFFER1_TO_PAGE_WITH_ERASE
  $86 equ BUFFER2_TO_PAGE_WITH_ERASE

  #528 equ DB321PAGESIZE
  #512 equ PAGESIZE
  $400000 equ DB321SIZE


\ =============
\ *S Primitives
\ =============

: InitSPI	\ --
\ *G Initialise the SPI Chip Select and Mode Registers.
\ ** The AT45DB321 DataFlash ICs are on Chip Select 0 to 3.
\ ** The AT25256W (not fitted) IC is on Chip Select 4.
  SP_CSR_CPOL			\ -- Clock polarity (inactive is high)
  SP_CSR_BITS_8 OR		\ -- OR 8 bits per transfer
  $00000500 OR			\ -- OR SPI baud rate = SPI MC / 2 * 5
  $00010000 OR			\ -- OR nCS -> SPCK delay = SPI MC * 1 
  $20000000 OR			\ -- OR 32 * 48 SPI Clks between cons. xfers
  _SPI_BASE SP_CSR0 + !		\ --
  \ Mode Register
  SP_MR_MSTR			\ --		; Set SPI mode to MASTER
  SP_MR_PS_VARIABLE OR		\ --		; Variable peripheral select
  SP_MR_PCSDEC OR		\ --		; Chip selects are decoded
\ SP_MR_DIV32 OR		\ --		; Slow SPI master clock
  $0A000000 OR			\ --		; 10 SPI Clks between CSs
  _SPI_BASE SP_MR + !
;

: OpenSPI	\ --
\ *G Open the SPI.
  APMC_PCER_SPI _APMC_BASE APMC_PCER + ! \ --	; Enable the SPI clock
  $3F800000 _PIOA_BASE PIO_PDR + ! \ --		; Disable PIO control of SPI pins
  $3F800000 _PIOA_BASE PIO_IFDR + ! \ --	; Disable PIO Glitch Filters
  $3F800000 _PIOA_BASE PIO_MDDR + ! \ --	; Disable Open Drain Config on pins
  SP_CR_SWRST _SPI_BASE SP_CR + ! \ --		; Reset the SPI
  SP_CR_SPIEN _SPI_BASE SP_CR + ! \ --		; Enable the SPI
  InitSPI
;

: CloseSPI	\ --
\ *G Close the SPI.
  SP_CR_SPIDIS _SPI_BASE SP_CR + ! \ --		; Disable the SPI
  $3F800000 PIO_PER _PIOA_BASE + ! \ --		; Enable PIO control of SPI pins
  APMC_PCDR_SPI _APMC_BASE APMC_PCDR + ! \ --	; Disable the SPI clock
;

: XferSPI			\ Tx -- Rx
\ *G Transmit and Receive via the SPI.
  _SPI_BASE TUCK		\ -- SPIBASE Tx SPIBASE
  SP_TDR + !			\ -- SPIBASE
  BEGIN
    DUP SP_SR + @ SP_SR_RDRF AND \ -- SPIBASE Rxed? ; Wait for byte rxed
  UNTIL
  SP_RDR + @			\ -- rx
  16 LSHIFT 16 RSHIFT		\ -- rx&$FFFF	; Mask out top 16 bits
;

: CSHigh			\ --
\ *G Wait till SPI Transfer is finished - At least 1 SPI DLYBCT period.
  512 0 DO			\ --		; 3 cycles per iteration
  LOOP				\ --		; Wait at least one DLYBCT
;

: DF_Xfer			\ pcs TxByte -- RxByte
\ *G Transfer a byte to/from the AT45DB321 DataFlash bank
\ ** - PCS needs to be orred into bits 19..16 before transmission
  OR XferSPI			\ -- RxData	; Add PCS then Xfer
  $FF AND			\ -- RxByte	; Return a byte
;

: DF_PCS			\ addr -- pcs
\ *G Calculate the Peripheral Chip Select (PCS) value given a DataFlash Address.
  6 RSHIFT $000F0000 AND	\ -- pcs
;

: DF_PageOffset			\ addr --
\ *G Set the AT45DB321 Page and Offset.
  DUP DF_PCS >R			\ -- addr	; Save PCS
  R@ OVER 15 RSHIFT $7F AND	\ -- addr pcs page>>6
  DF_Xfer DROP			\ -- addr	; Send first 8 bits
  DUP 7 RSHIFT $FC AND		\ -- addr page<<2
  OVER 8 RSHIFT $01 AND OR	\ -- addr page<<2|offset>>8
  R@ SWAP DF_Xfer DROP		\ -- addr	; Send second 8 bits
  $FF AND R> SWAP		\ -- pcs offset&FF
  DF_Xfer DROP			\ --		; Send last 8 bits
;

: DF_GetStatus			\ pcs -- status
\ *G Read the status of an AT45DB321 DataFlash chip.
  DUP STATUS_READ DF_Xfer	\ -- pcs DC	; Command
  DF_Xfer			\ -- status	; Get the status
  CSHigh			\ -- status	; Stop Xfer
;

: DF_SetupPageRead		\ src --
\ *G Setup an AT45DB321 DataFlash for a Page Read.
  DUP DF_PCS SWAP		\ -- pcs src	; Get PCS
  OVER PAGE_READ DF_Xfer DROP	\ -- pcs src ; Send Command
  DF_PageOffset			\ -- pcs	; Send Page & Offset
  DUP $FF DF_Xfer DROP		\ -- pcs	; Send 4 "Don't Care" bytes
  DUP $FF DF_Xfer DROP		\ -- pcs
  DUP $FF DF_Xfer DROP		\ -- pcs
  DUP $FF DF_Xfer DROP		\ -- pcs
  DROP				\ --
;

: DF_PageToBuffer1		\ dest --
\ *G Copy a page in an AT45DB321 DataFlash to it's SRAM Buffer 1.
  DUP DF_PCS SWAP		\ -- pcs dest	; Get PCS
  OVER PAGE_TO_BUFFER1 DF_Xfer DROP \ -- pcs dest ; Send Command
  DF_PageOffset			\ -- pcs	; Send Page
  CSHigh			\ -- pcs	; End SPI Transfer
  BEGIN
    DUP DF_GetStatus		\ -- pcs status	; Get AT45DB321's Status
    $80 AND    
  UNTIL				\ -- pcs	; Wait for DataFlash ready
  DROP				\ --
;

: DF_SetupBuffer1Write		\ dest --
\ *G Setup an AT45DB321 DataFlash for a buffer 1 write.
  DUP DF_PCS SWAP		\ -- pcs dest	; Get PCS
  OVER BUFFER1_WRITE DF_Xfer DROP \ -- pcs dest	; Send Command
  DF_PageOffset			\ -- pcs	; Send Offset
  DROP				\ --
;

: DF_Buffer1ToPage		\ dest --
\ *G Copy the SRAM buffer 1 in an AT45DB321 DataFlash to a page.
  DUP DF_PCS SWAP		\ -- pcs dest	; Get PCS
  OVER BUFFER1_TO_PAGE_WITH_ERASE DF_Xfer DROP \ -- pcs dest ; Send Command
  DF_PageOffset			\ -- pcs	; Send Page
  CSHigh			\ -- pcs	; End SPI Transfer
  BEGIN
    DUP DF_GetStatus		\ -- pcs status	; Get AT45DB321's Status
    $80 AND    
  UNTIL				\ -- pcs	; Wait for DataFlash ready
  DROP				\ --
;

: DF_FetchPage			\ dest src len -- dest' src' len'
\ *G Get a page (or partial page) of data from "src" and store it in "dest",
\ ** also update "src", "len" & "dest" accordingly.
  \ Setup the relevant AT45DB321 DataFlash chip for a Page Read
  OVER DF_SetupPageRead
  \ Calculate the next Block size (Smaller of "len" and "Bytes to End of Page")
  OVER $1FF AND PAGESIZE SWAP -	\ -- dest src len BytesToEndOfPage
  OVER MIN			\ -- dest src len BlockSize
  >R				\ -- dest src len ; Save it
  -ROT SWAP R@			\ -- len src dest BS
  \ Read the Bytes!!
  0 DO				\ -- len src dest ; For each byte
    OVER DF_PCS $FF DF_Xfer	\ -- len src dest DataByte ; Get the byte
    OVER C!			\ -- len src dest ; Store it
    1+				\ -- len src dest+1 ; Inc dest address
  LOOP
  CSHigh			\ -- len src dest ; Stop SPI transfer
  \ Update the pointers & counter
  SWAP R@ +			\ -- len dest' src' ; "src" += BlockSize
  ROT				\ -- dest' src' len
  R> -				\ -- dest' src' len' ; "len" -= BlockSize
;

: DF_StorePage			\ dest src len -- dest' src' len'
\ *G Store a page (or partial page) of data from "src" and store it in "dest",
\ ** also update "src", "len" & "dest" accordingly.
  \ Copy the page that contains dest into an SRAM buffer
  ROT DUP DF_PageToBuffer1	\ -- src len dest ; Copy DataFlash Page
  SWAP				\ -- src dest len
  \ Setup the AT45DB321 DataFlash chip for a Buffer write
  OVER DF_SetupBuffer1Write	\ -- src dest len
  \ Calculate the next Block size (Smaller of "len" and "Bytes to End of Page")
  OVER $1FF AND PAGESIZE SWAP -	\ -- src dest len BytesToEndOfPage
  OVER MIN			\ -- src dest len BlockSize
  >R				\ -- src dest len ; Save it
  -ROT SWAP R@			\ -- len dest src BS
  \ Write the Bytes!!
  0 DO				\ -- len dest src ; For each byte
    OVER DF_PCS OVER C@		\ -- len dest src pcs byte
    DF_Xfer DROP		\ -- len dest src ; Store the byte
    1+				\ -- len dest src+1 ; Inc src address
  LOOP
  CSHigh			\ -- len dest src ; End SPI Transfer
  \ Write the SRAM buffer back into the page
  OVER DF_Buffer1ToPage		\ -- len dest src ; Write the page
  \ Update the pointers & counter
  SWAP R@ +			\ -- len src' dest' ; "dest" += BlockSize
  SWAP ROT			\ -- dest' src' len
  R> -				\ -- dest' src' len' ; "len" -= BlockSize
;


\ ====================
\ *S Application words
\ ====================

: DF_Capacity			\ -- capacity
\ *G Sample the first four SPI selects to find out how much is there.
  0				\ -- capacity
  4 0 DO
    I 16 LSHIFT DF_GetStatus	\ -- capacity status
    3 RSHIFT $7 AND
    $6 = IF			\ -- capacity	; If a AT45DB321
      DB321SIZE +		\ -- capacity	; Capacity += AT45DB321 size
    ENDIF
  LOOP
;

: DATAFLASH@	\ sram_addr length dataflash_addr -- t/f(good)
\ *G Get data from the DataFlash, returns Success flag (0 = Success).
  SWAP				\ -- dest src len
  dup 0=
  if drop drop drop 0 exit endif
  OpenSPI			\ -- dest src len ; Open the SPI
  BEGIN 			\ -- dest src len ; While (len > 0)
    DF_FetchPage		\ -- dest' src' len'
    dup 0=
  UNTIL				\ -- dest src len
  CSHigh			\ -- dest src len ; End SPI Transfer
  DROP DROP DROP
  CloseSPI			\ --		; Close the SPI
  FALSE				\ -- FALSE	; Return "success"
;

: DATAFLASH!	\ sram_addr length dataflash_addr -- t/f(good)
\ *G Store data in the DataFlash, returning Success flag (0 = Success).
  over 0=
  if  drop drop drop  0  exit  endif
  OpenSPI
  -rot
  BEGIN				\ -- dest src len ; While (len > 0)
    DF_StorePage		\ -- dest' src' len'
    dup 0=
  UNTIL				\ -- dest src len
  DROP DROP DROP
  CloseSPI			\ --		; Close the SPI
  FALSE				\ -- FALSE	; Return "success"
;


\ ************
\ *S Test Code
\ ************

((
: Str!		\ $addr DFaddr -- T/F(Good) ; counted string
\ *G Write a counted string to the Flash.
  over c@ 1+ swap		\ -- src len dest
  DATAFLASH!			\ -- T/F(Good)
;

: Str@		\ $addr len DFaddr -- T/F(Good)
\ *G Use this to fetch a counted string stored in DataFlash
\ ** into a buffer of size len and display it.
  2 pick >r
  DATAFLASH@			\ -- T/F(Good)
  r> count type
;
))

decimal

\ ======
\ *> ###
\ ======

