\ FlashProg.fth - Program SAM7x Flash

((
Copyright (c) 2006
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
At end of programming, check for a partial block to program.

Change history
==============
20061101 MPE001 First release
))

only forth definitions  decimal


\ ===========
\ *! flashsam7
\ *T SAM7x Flash tools
\ ===========
\ *P These tools are provided for applications which reserve
\ ** part of the Flash for data.

\ *P N.B. All writes to Flash must be from RAM as the Flash
\ ** is unavailable at the time any part is being written.


\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the SAM7 series, they are designed to be easily expanded
\ ** for future SAM7 parts with as yet unknown features.


\ ***************
\ *S Code for RAM
\ ***************

Proc WriteSAM	\ src dest -- ior
\ *G This procedure is copied into RAM and contains the
\ ** Flash erase/program tools. It is passed TOS=dest, NOS=
\ ** src, and returns a zero/one status, where zero=success.
\ ** 256 bytes are programmed. *\i{Dest} must be 256 byte
\ ** aligned.

\ Uses R8=src, R7=dest, R6=count, R), R1, R2
  mov     r7, TOS		\ preserve dest for page number
  ldr     r8, [ psp ], # 4	\ R8 := src
  mov     r6, # #256		\ 256 bytes per page
  begin,
    ldr     r0, [ r8 ], # 4	\ read data from RAM
    str     r0, [ TOS ], # 4	\ write to page buffer
    sub .s  r6, r6, # 4
  eq, until,
  mov     r0, r7		\ page base address
  mvn     r2, # $FF		\ r2 := FFFF:FF00 = _MC
  begin,			\ wait for MC_FSR0 FRDY bit to be set 
    ldr     r1, [ r2, # MC_FSR0 ]
    and .s  r1, r1, # 1
  ne, until,
  \ form the command
  bic     r0, r0, # $FF000000	\ clear key field
  bic     r0, r0, # $00F00000	\ clear top of 1Mb range
  bic     r0, r0, # $000000FF	\ clear command field
  orr     r0, r0, # $5A000000	\ set key field
  orr     r0, r0, # $00000001	\ set command field
  str     r0, [ r2, # MC_FCR0 ]	\ write command
  begin,			\ wait for MC_FSR0 FRDY bit to be set 
    ldr     TOS, [ r2, # MC_FSR0 ]
    and .s  r1, TOS, # 1
  ne, until,
  and .s  TOS, TOS, # $0C	\ check status bits
  mov .ne TOS, # 1		\ if either set, return 1
  mov   pc, link		\ return
EndProc
here WriteSAM - constant LenSAM

LenSAM buffer: RAMprog


\ ***************
\ *S Flash driver
\ ***************
\ *P The 256k flash is divided into 16 regions of 16kb which can be locked.
\ ** The Flash is programmed in pages of 256 bytes. The Flash controller
\ ** includes a 256 byte buffer. A page is written by writing 32 bit
\ ** words to the address to be programmed until the buffer contains
\ ** 256 bytes and then issuing a page write command. Pages are 256
\ ** byte aligned.

cell buffer: FlDest	\ -- addr
\ *G Holds next destination address

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count

/MCK #1000000 / 1+ equ #FMCN1	\ -- n
\ *G Number of clocks in 1us.
#FMCN1 3 2 */ 1+ equ #FMCN1.5	\ -- n
\ *G Number of clocks in 1.5us.

: SetFMCN	\ nmcn --
\ Write the value into the FMCN field of MC_FMR0.
 $FF and #16 lshift  #FWS 8 lshift or	\ FMR0 field construction
 _MC MC_FMR0 + !
  begin  _MC MC_FSR0 + @ 1 and  until	\ wait for Flash ready
;

: Prog256	\ src dest -- ior
\ *G Program a 256 byte block into RAM. The Flash write routine
\ ** must already be in RAM and the EFC unit must have been
\ ** initialised.
  [i dfi  RamProg execute  i]
;

: (EraseFlash)	\ dest dlen --
\ *G Erase the flash for the given range. A dummy placeholder.
  2drop
;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at dest.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Src} is forced to a 4 byte boundary, *\i{dest}
\ ** is forced to a 256 byte boundary and *\i{len} is rounded
\ ** up to the next 256 byte unit. 
  rot -4 and -rot			\ force src
  swap #-256 and swap			\ force dest
  #255 + #-256 and			\ force len
  bounds ?do
    dup i Prog256 #PrgErrs +!  #256 +	\ program 256 byte page
  #256 +loop
  drop
;

: .src/dest	\ src dest -- src dest
\ *G Display the addresses and contents of the source
\ ** and destination.
  ."  source: " over .lword  over c@ ." (" .byte ." )"
  ."  dest: " dup .lword  dup c@ ." (" .byte ." )"
;

: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over c@ over c@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    1 + swap  1 + swap
  1 +loop
  2drop
;

: ProgramFlash	\ src dest len --
\ *G Using the RAM memory buffer at src, program the Flash at dest
\ ** with len bytes. The relevant Flash pages are erased, the
\ ** Flash is programmed, and the result verified.
  #255 + #-256 and			\ round to 256 byte length
  WriteSAM RAMprog LenSAM cmove		\ copy code to RAM
  #FMCN1.5 SetFMCN			\ prep Flash
  0 #PrgErrs !
  cr ." Programming "  3dup (ProgFlash)
  space #PrgErrs ? ." errors"
  cr ." Verifying "  (VerifyFlash)
;

\ Test code

decimal

1024 buffer: foo

: t
  foo 1024 $AA fill
  foo $120000 1024 ProgramFlash
;

\ ======
\ *> ###
\ ======

