\ Boot loader application code for MPE StrongBox
\ The required ethernet MAC Address and IPv4 addresses are configured
\ by the two CREATEs in the Ethernet driver

((
Copyright (c) 2000, 2002
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20020831 SFP002 Faster, smaller checksum code.
20020830 SFP002 Modified for Samsung board.
		Made the network order fetch/store operations
		usable on 16 bit alignments rather than 32 bit
		alignments as these depend on the ARM CPU bus
		interface permissions. Some ARMs with 16 bit
		bus interfaces permit 16 bit aligned 32 bit
		operations, but 32 bit bus systems may/will fail.
		Improved high level checksum code for ARM.
20020627 SFP001 Updated and tested on StrongBox rev 1.
20020617 SFP000 First release
))


\ **********
\ *! Netboot
\ *T Network boot code and TFTP
\ **********
\ *P The file NETBOOT.FTH provides facilities for file transfer into Flash
\ ** using Ethernet and TFTP. Menu options allow various facilities for 
\ ** boot management, replacement of user applications, replacement of
\ ** boot code, and Ethernet MAC and IP address configuration.

decimal


\ ****************
\ *S Configuration
\ ****************

true equ ping?	\ -- flag
\ *G Set this equate non-zero to include the PING code.


\ ****************************
\ *S Handling the Flash memory
\ ****************************

cell buffer: DestPtr	\ -- addr
\ *G Holds the next address in Flash to be written to.

#512 buffer: FlWriteBuff
\ *G Scratch buffer to hold TFTP data when the source address
\ ** is not cell aligned. This is usually the case because of
\ ** the header layout.

: WriteFlashBlock       \ addr len --
\ *G Write the buffer addr/len to Flash at the address
\ ** in DESTPTR and then add len to DESTPTR.
  over 3 and if				\ If unaligned
    >r
    FlWriteBuff r@ cmove		\ copy to aligned buffer
    FlWriteBuff r>
  endif
  tuck
  DestPtr @ swap (ProgFlash)
  DestPtr +!
;

: run-user	\ --
\ *G Copy flash into ram at address APPRAM, and execute at address APPRAM.
\ ** If the data at the start of the application Flash area is $FFFFFFFF,
\ ** which is the signature of unprogrammed flash, RUN-USER issues an
\ ** error message and returns.
  UserFlash @ $FFFFFFFF =		\ Check for application
  if  cr ." No application present" exit  then

  stop-timers  di			\ disable all interrupts
  UserFlash AppRam UserFlashLen cmove	\ copy Flash to RAM
  AppRam execute			\ go do it
;


\ ****************************************
\ *S Network order (big-endian) operations
\ ****************************************
\ *P Note that these functions have to be capable of fetching 32 bit cells
\ ** from 16-bit aligned addresses, not necessarily from 32 bit aligned
\ ** addresses.

little-endian? [if]
: w@(n)         \ addr -- u16
\ *G Network order 16 bit fetch.
  dup c@ 8 lshift swap 1+ c@ or
;

: w!(n)         \ u16 addr --
\ *G Network order 16 bit store.
  over 255 and over 1+ c!
  swap 8 rshift swap c!
;

: @(n)          \ addr -- u32
\ *G Network order 32 bit fetch.
  dup w@(n) 16 lshift swap 2 + w@(n) or
;

: !(n)		\ u32 addr --
\ *G Network order 32 bit store.
  over 65535 and over 2 + w!(n)
  swap 16 rshift swap w!(n)
;
[else]
compiler
: w@(n)		w@  ;
: w!(n)		w!  ;
target
interpreter
: w@(n)		w@  ;
: w!(n)		w!  ;
target
: w@(n)		w@  ;
: w!(n)		w!  ;

: @(n)          \ addr -- u32
  dup w@(n) #16 lshift swap 2 + w@(n) or
;

: !(n)		\ u32 addr --
  over $FFFF and over 2 + w!(n)
  swap #16 rshift swap w!(n)
;
[then]


\ ****************
\ *S Miscellaneous
\ ****************

: memswap       \ a1 a2 len --
\ *G Exchange the contents of two buffers.
  bounds ?do
    i c@                \ a1 c2 --
    over c@             \ a1 c2 c1 --
    i c!                \ a1 c2 --
    over c!             \ a1 --
    1+
  loop
  drop
;

: .z$           \ addr --
\ *G Display a zero terminated string at addr.
  begin
    dup c@ dup
  while
    emit 1+
  repeat
  2drop
;


\ ********************
\ *S Internet checksum
\ ********************

code cksum	\ addr len -- u16
\ *G ARM IP Checksum routine in 16 instructions.
\ ** Routine is safe against alignment issues as the network order word fetch
\ ** is done as two byte fetches.
        ldmfd    r12 ! { r0 }           \ r0 = address
        add      r1, r10, r0            \ r1 = end-address-1
        sub      r1, r1, # 1
        mov      r10, # 0               \ r10 = checksum-acc
        mvn      r3, # 0                \ r3 = $FFFF
        mov      r3, r3, lsr # 16
l: ckl2
        cmp      r0, r1
        ldrb .le r2, [ r0 ], # 1
        add  .le r10, r10, r2 lsl # 8
        ldrb .lt r2, [ r0 ], # 1
        add  .lt r10, r10, r2
        add  .le r10, r10, r10 lsr # 16
        and  .le r10, r10, r3
        b    .lt ckl2
        eor      r10, r10, r3
        next,
end-code

[undefined] cksum [if]
: cksum		\ addr len -- cksum ; computed as 16 bit ints
\ *G Compute the internet checksum of a memory area.
  0 -rot				\ initial value of checksum
  swap 2- swap				\ start address 2- so we can inc it more efficiently
  begin
    dup 1 >				\ check if more than 1 byte left
  while
    2- >r				\ dec & shift the length out of the way
    2+ tuck w@(n)			\ get 16 bits of data
    +					\ and add it into cksum
    dup $80000000 and if		\ check top bit
      dup $0FFFF and			\ get low 16 bits
      swap #16 rshift +			\ and add in the high 16 bits
    endif
    swap 				\ re-order values
    r>					\ retrieve length
  repeat
  if					\ odd byte left
    2+ c@ #8 lshift +
    dup $80000000 and if		\ check top bit
      dup $0FFFF and			\ get low 16 bits
      swap #16 rshift +			\ and add in the high 16 bits
    endif
  else					\ all done
    drop
  endif
  begin
    dup $FFFF0000 and
  while
    dup $0FFFF and			\ get low 16 bits
    swap #16 rshift +			\ and add in the high 16 bits
  repeat
  $FFFF xor
;
[then]


\ *****************************
\ *S Misc. IP and MAC functions
\ *****************************

create EtherAddressBC	\ -- addr ; for checking for broadcast addresses
  255 c, 255 c, 255 c, 255 c, 255 c, 255 c,

: MyIp?         \ *ip -- flag ; true for match
  IpAddress 4 s=
;

: BroadCastMac? \ addr -- flag
  EtherAddressBC 6 s=
;


\ *******************************************
\ *S Ethernet and IP structures and constants
\ *******************************************

128 equ T_IGNOREPACKET	\ -- n ; error codes
129 equ T_IP_BADSIZE
130 equ T_IP_BADVERSION
131 equ T_IP_OFFSET
132 equ T_IP_NOTFORME
133 equ T_IP_UNHANDLED
134 equ T_ICMP_BADSIZE
135 equ T_ICMP_BADTYPE
136 equ T_UNKNOWN_DOMAIN

struct EHDR
  6 field EHDR.DstMAC
  6 field EHDR.SrcMAC
  2 field EHDR.type
end-struct

$0800 equ ET_IP
$0806 equ ET_ARP

struct ARP
  EHDR field      _ehdr1
  2 field AHDR.htype
  2 field AHDR.ptype
  1 field AHDR.hlen
  1 field AHDR.plen
  2 field AHDR.op
  6 field ARP.senderMAC
  4 field ARP.senderIP
  6 field ARP.targetMAC
  4 field ARP.targetIP
end-struct

$0001	equ ARPHARD_ETHER
$0800	equ ARPPROTO_IP

$0001	equ ARPOP_REQUEST
$0002	equ ARPOP_REPLY

struct  IPHDR
  EHDR field      _ehdr2
  1 field IPHDR.version
  1 field IPHDR.tos
  2 field IPHDR.len
  2 field IPHDR.id
  2 field IPHDR.off
  1 field IPHDR.ttl
  1 field IPHDR.p
  2 field IPHDR.cksum
  4 field IPHDR.src
  4 field IPHDR.dst
end-struct

1	equ IPPROTO_ICMP	\ control message protocol
17	equ IPPROTO_UDP		\ user datagram protocol

struct  ICMP
  IPHDR field     _iphdr1
  1 field ICMP.type
  1 field ICMP.code
  2 field ICMP.cksum
  2 field ICMP.id
  2 field ICMP.seq
  0 field ICMP.data
end-struct

$0000 equ ICMP_ECHOREPLY
$0008 equ ICMP_ECHOREQUEST

struct  UDP
  IPHDR field     _iphdr2
  2 field UDP.sport
  2 field UDP.dport
  2 field UDP.ulen
  2 field UDP.usum
  0 field UDP.data
end-struct


\ ******************************
\ *S Buffers and sending packets
\ ******************************

cell buffer: pktInSize
#1536 buffer: pktIn
cell buffer: udp-data
cell buffer: udp-datalen

: SendEtherPacket	\ -- 
\ *G Send a raw Ethernet packet from the buffer PKTIN whose
\ ** size is contained in the variable PKTINSIZE.
  PktIn PktInSize @ send_ether_pkt
;

: SendPacket	\ --
\ *G Send an IP packet over Ethernet from the buffer PKTIN whose
\ ** size is contained in the variable PKTINSIZE. The IP checksum
\ ** is calculated and inserted into the IP header before
\ ** transmission.
  pktinsize @ EHDR - pktin IPHDR.len w!(n)      \ new ip header length
  PktIn EHDR +  IPHDR EHDR -  cksum             \ .. and recalculate for header
  PktIn IPHDR.cksum w!(n)
  SendEtherPacket
;


\ *******
\ *S TFTP
\ *******

$001 equ TFTP_RRQ                \ Read ReQuest
$002 equ TFTP_WRQ                \ Write ReQuest
$003 equ TFTP_DATA               \ DATA packet
$004 equ TFTP_ACK                \ ACKnowledge
$005 equ TFTP_ERROR              \ ERROR condition

cell buffer: Uploading?
cell buffer: blockid
cell buffer: UploadComplete?

: ResetTFTP	\ --
  false Uploading? !
  0 UpLoadComplete? !
  0 blockid !
;

: smERROR       \ --
  cr ." TFTP Error: "
  udp-data @ 4 + .z$			\ display error text from server
  ResetTFTP  T_IGNOREPACKET throw
;

: smDone        \ --
  cr ." Upload Complete"
  ResetTFTP  UpLoadComplete? on
;

: AckData       \ --
  TFTP_ACK udp-data @ w!(n)
  blockid @ udp-data @ 2+ w!(n)
  4 udp-datalen !
  1 blockid +!
;

: smBegin       \ --
  UpLoading? @ if
    cr ." TFTP Error: Upload begun before previous complete."
    ResetTFTP  T_IGNOREPACKET throw
  then

  cr ." TFTP Receiving file: " udp-data @ 2+ .z$ space
  true Uploading? !
  AckData
;

: smData        \ --
  udp-data @ cell+ udp-datalen @ cell -
  WriteFlashBlock
  udp-datalen @ #516 <> >r
  AckData
  r> if smDone then
;

: process-ip-udp-tftp   \ --
  udp-data @ w@(n)
  case
    TFTP_WRQ of         smBegin         endof
    TFTP_ERROR of       smError         endof
    TFTP_DATA of        smData          endof
      cr ." TFTP Error: Unexpected packet"
      ResetTFTP  T_IGNOREPACKET throw
  endcase
;

\ ********************
\ *S IP, UDP, and ICMP
\ ********************

20 buffer: temp-ip

: process-ip-udp        \ --
  pktInSize @ UDP                       <  if T_IGNOREPACKET throw then

  pktin udp.sport dup w@(n)     	\ *s s --
  pktin udp.dport dup w@(n)     	\ *s s *d d --
  dup >r                        	\ *s s *d d -- ; d --
  -rot w!(n)                    	\ swap port numbers for reply
  swap w!(n)
  r>                            	\ dest-port --

  pktin UDP.data udp-data !             \ split buffer to header and data
  pktinsize @ UDP - udp-datalen !       \ sections
  UDP pktinsize !

  case
    #69 of      process-ip-udp-tftp     endof
    drop T_IGNOREPACKET throw
  end-case

  \ recombine udp data with original header (assumes data in same place)
  udp-datalen @ pktinsize +!

  \ Update UDP data length ( data length+sizeofUDP header )
  udp-datalen @ 8 + pktin UDP.ulen w!(n)

  \ Update UDP checksum
  0 pktin UDP.usum w!(n)

  pktin IPHDR.src @(n) temp-ip !(n)
  pktin IPHDR.dst @(n) temp-ip 1 cells + !(n)
  IPPROTO_UDP temp-ip 2 cells + w!(n)
  udp-datalen @ 8 + temp-ip 10 + w!(n)          \ size of udp data + udp header
  pktin IPHDR + temp-ip 12 + 8 cmove

  temp-ip 20 cksum $FFFF xor
  udp-data @ udp-datalen @ cksum $FFFF xor +
  dup $FFFF > if 1+ then $FFFF and $FFFF xor

  pktin UDP.usum w!(n)

  \ .. and send
  SendPacket
;

: process-ip-icmp	\ --
  pktInSize @ ICMP                      <  if T_ICMP_BADSIZE throw then
  PktIn ICMP.type c@ ICMP_ECHOREQUEST   <> if T_ICMP_BADTYPE throw then

  ICMP_ECHOREPLY PktIn ICMP.type c!     \ switch type to reply
  0 PktIn ICMP.cksum w!(n)              \ reset checksum

  PktIn IPHDR +                         \ pointer to icmp header
  PktInSize @ IPHDR -                   \ size of ICMP block
  cksum PktIn ICMP.cksum w!(n)          \ recalc checksum

  SendPacket
;

: process-ip	\ --
  pktInSize @ IPHDR                     <  if T_IP_BADSIZE    throw then
  pktIn IPHDR.version c@ $45            <> if T_IP_BADVERSION throw then
  pktIn IPHDR.off w@(n)                    if T_IP_OFFSET     throw then
  pktIn IPHDR.dst MyIp?                 0= if T_IP_NOTFORME   throw then


  PktIn IPHDR.src                       \ swap addresses ready for reply
  PktIn IPHDR.dst 4 memswap

  0 PktIn IPHDR.cksum w!(n)             \ clear checksum field

  pktIn IPHDR.p c@
  case
    IPPROTO_ICMP of     process-ip-icmp         endof
    IPPROTO_UDP  of     process-ip-udp          endof
    drop T_IP_UNHANDLED throw
  end-case
;


\ ******
\ *S ARP
\ ******

: process-arp-request   \ --
  pktIn AHDR.htype w@(n) ARPHARD_ETHER <> if T_IGNOREPACKET throw then
  pktIn AHDR.ptype w@(n) ARPPROTO_IP   <> if T_IGNOREPACKET throw then
  pktIn AHDR.hlen c@ 6                 <> if T_IGNOREPACKET throw then
  pktIn AHDR.plen c@ 4                 <> if T_IGNOREPACKET throw then
  pktIn ARP.targetIP MyIp?             0= if T_IGNOREPACKET throw then
  pktIn ARP.senderMAC			\ copy sender MAC and IP to target
  pktIn ARP.targetMAC #10 cmove
  EtherAddress PktIn ARP.senderMAC 6 cmove       \ put in my details
  IpAddress    PktIn ARP.senderIP  4 cmove

  ARPOP_REPLY PktIn AHDR.op w!(n)               \ set ARPOP_REPLY

  pktIn ARP.targetMAC pktIn EHDR.DstMac 6 cmove  \ set ethernet packet addresses
  pktIn ARP.senderMAC pktIn EHDR.SrcMac 6 cmove
  SendEtherPacket
;

: process-arp   \ --
  pktInSize @ ARP < if T_IGNOREPACKET throw then
  ARP PktInSize !
  pktIn AHDR.op w@(n) ARPOP_REQUEST = if
    process-arp-request
  else
    T_IGNOREPACKET throw
  then
;


\ *********************
\ *S Tieing it together
\ *********************

: boot-process	\ --

  pktInSize @ EHDR < if T_IGNOREPACKET throw then       \ too short

  pktIn EHDR.DstMac BroadCastMAC? 0= if         \ swap if not broadcast
    PktIn EHDR.DstMac
    Pktin EHDR.SrcMac
    6 memswap
  then

  case  pktIn EHDR.type w@(n)
    ET_IP  of  process-ip   endof
    ET_ARP of  process-arp  endof
      T_UNKNOWN_DOMAIN throw
  endcase
;



\ *************************************
\ Serial and Ethernet Flash programming
\ *************************************

((
: nextSerUnit	{ *flash | buff[ 4 ] -- flag }
\ *G Program the next serial data into a Flash unit,
\ ** assuming that the data is being sent as a memory
\ ** image. The returned flag is 0 for success.
   buff[ 4 bounds
   do  key i c!  loop
   buff[ @ *flash 4 write32
;

: serial-download	{ *flash len -- }
  *flash len (eraseFlash)
  cr ." Start serial transfer now"
  cr ." Reset board when transfer is complete"
  cr
  *flash len bounds ?do
    nextSerUnit if
      cr ." Write failed at address: " i .lword
      unloop exit
    endif
    i $FF and 0= if [char] . emit then
  4 +loop				\ must match size of nextSerUnit
;
))

: ether-download	{ *flash len -- }
  *flash 0= if  [I  endif		\ protect if overwriting interrupt vectors

  *flash DestPtr !
  *flash len (eraseFlash)

  ResetTFTP  InitEther  CurrentIP
  cr ." Start Ethernet TFTP transfer now"
  cr ." Press <ESC> to abort transfer"
  cr

  begin
    isrx? if	\ check for ethernet download
      pktIn #1536 get_ether_pkt pktInSize !
      [I  ['] boot-process catch  I]
      ?dup if
        case
\          T_IGNOREPACKET   of   cr ." >> Ignored"         endof
\          T_UNKNOWN_DOMAIN of   cr ." >> Unknown Domain"  endof
          T_IP_BADSIZE     of   cr ." >> Bad IP Size"     endof
          T_IP_BADVERSION  of   cr ." >> Bad IP Version"  endof
          T_IP_OFFSET      of   cr ." >> Bad IP Offset"   endof
\          T_IP_NOTFORME    of   cr ." >> IP Not for me"   endof
          T_IP_UNHANDLED   of   cr ." >> Unhandled IP"    endof
          T_ICMP_BADSIZE   of   cr ." >> Bad ICMP Size"   endof
          T_ICMP_BADTYPE   of   cr ." >> Bad ICMP Type"   endof
          drop
        end-case
      then
    then
    *flash if
      key? if
        key $1B =
        if  UpLoadComplete? on  then
      then
    then
    UpLoadComplete? @
  until
  base @ decimal
  cr  DestPtr @ *flash - . ." bytes programmed"
  base !

  *flash 0= if  I]  then
;

: download      \ *flash len --
\ *G Download a file by Ethernet/TFTP
\ or by serial line
\ ** and program it into the Flash Block at *flash/len,
\ ** which will be erased before the file transfer starts.
  ether-download
;
((
: download      \ *flash len --
  cr ." Select"
  cr ."  E - Ethernet and TFTP"
  cr ."  S - Serial 8 bit binary"
  cr ."  other - no action"
  cr ." S/E " key upc dup emit
  case
    [char] S of  serial-download  endof
    [char] E of  ether-download   endof
      nip nip
  endcase
;
))


\ ***********************************************
\ *S Displaying and Entering IP and MAC addresses
\ ***********************************************

: getByte	\ addr char -- addr'
\ *G Collect an integer delimited by char and store it
\ ** at addr, returning addr+1.
  word integer? 1 <> throw		\ collect and convert
  over c!  1+				\ stash data and update address
;

: getData	\ addr char n --
\ *G Collect n integers delimited by char, and store the bytes
\ ** sequentially at addr.
  1- 0 max 0 ?do
    tuck getByte swap
  loop
  drop  bl getByte drop
;

: CurrentIP	\ -- ; display IP address
\ *G Display the current IP address.
  base @ >r  decimal
  cr ." Current IP address: "
  IpAddress
  dup c@     0 .r ." ."
  dup 1 + c@ 0 .r ." ."
  dup 2 + c@ 0 .r ." ."
      3 + c@ 0 .r
  r> base !
;

: GetIp		\ -- ; input new IP address
  base @ >r  decimal
  begin
    cr ." Please Enter New IP Address: xxx.xxx.xxx.xxx "
    cr Query  #tib @ if
      IPaddress [char] . 4 ['] getData catch dup
      if  nip nip nip  then
    else
      0
    endif
   while
     2drop drop
     cr ." Invalid IP address"
     CurrentIP
  repeat
  CurrentIP
  r> base !
;

: CurrentMAC	\ --
\ *G Display current MAC address.
  base @ >r  hex
  cr ." Current MAC Address: " 
  EtherAddress
  dup     c@ 0 .r [char] - emit
  dup 1 + c@ 0 .r [char] - emit
  dup 2 + c@ 0 .r [char] - emit
  dup 3 + c@ 0 .r [char] - emit
  dup 4 + c@ 0 .r [char] - emit
  5 + c@ 0 .r 
  r> base !
;

: GetMac	\ -- ; input new MAC address
\ *G Get a new MAC address from the user console.
  base @ >r  hex
  begin
    cr ." Please Enter New MAC Address: aa-bb-cc-dd-ee-ff (base is hex) "
    cr Query  #tib @ if
      EtherAddress [char] - 6 ['] getData catch dup
      if  nip nip nip  then
    else
      0
    endif
   while
    2drop drop
    cr ." Invalid MAC address"
    CurrentMAC
  repeat
  CurrentMAC
  r> base !
;

: RAM>DataFlash	{ | buff[ #10 ] -- }
\ *G When data such as the Ethernet IP or MAC address is changed, the
\ ** data sector must be reprogrammed. Note that the word DATAFLASH>RAM
\ ** must be used to reload the data at powerup.
  IpAddress buff[ 4 cmove
  EtherAddress buff[ 4 + 6 cmove
  buff[ DataFlash #12 ProgramFlash	\ force integer number of cells
;

: DataFlash>RAM	\ --
\ *G Set up the IP and Ethernet addresses from the data flash area.
  DataFlash @ $FFFFFFFF <> if
    DataFlash @ IpAddress !
  then
  DataFlash 4 + @ $FFFFFFFF <> if
    DataFlash 4 + EtherAddress 6 cmove
  then
;

: SetupIP	\ --
\ *G Get a new Ethernet IP address from the user and reprogram the Flash
\ ** boot sector.
  CurrentIP GetIP RAM>DataFlash
;

: SetupMAC
\ *G Get a new Ethernet MAC address from the user and reprogram the Flash
\ ** boot sector.
  CurrentMAC GetMAC RAM>DataFlash
;


\ ************************
\ *S User maintenance menu
\ ************************
\ *P The maintenance system provides the following features,
\ ** accessed from the menu.
\ *B I - setup the Ethernet IP address
\ *B M - setup the Ethernet MAC address
\ *B F - run the Forth monitor
\ *B R - run the application
\ *B E - erase the user Flash area
\ *B D - download and Flash a new user program
\ *B B - download and Flash a new boot program
\ *B Q - quit maintenace system


decimal

((
: nowReset	\ -- 
  cr ." Please Reset board: "
  begin again
;
))

: .maintenance-menu	\ --
\ *G Display the maintenance menu.
  cr ." Command keys:"
  cr ." I - setup the Ethernet IP address"
  cr ." M - setup the Ethernet MAC address"
  cr ." F - run the Forth monitor"
  cr ." R - run the application"
  cr ." E - erase the user Flash area"
  cr ." D - download and Flash a new user program"
  cr ." B - download and Flash a new boot program"
  cr ." Q - quit maintenance system"
;

: (maintenance)	\ mode -- flag ; flag true if Q pressed
\ *G Enter a single command and process it. If mode is non-zero,
\ ** an invalid key runs an application program if it exists.
\ ** Flag is returned non-zero if Q was pressed.
  >r
  0					\ Q not pressed yet
  case  cr ." Enter command: " key upc dup emit
    [char] I of  
      cr ." Setup Ethernet IP address" SetupIP  ( nowReset )
    endof
    [char] M of
      cr ." Setup MAC address" SetupMAC  ( nowReset )
    endof
    [char] F of
      cr ." Forth Monitor"  cr abort
    endof
    [char] R of
      cr ." Run user application" run-user
    endof
    [char] D of
      cr ." Download new user program"
      userflash userflashlen download  ( nowReset )
    endof
    [char] B of
      cr ." Download new boot program:"
      cr ."   Are you sure? Y/N: " key upc  dup emit  [char] Y = if
\        Stop-Timers
        BootFlash BootLen download  ( nowReset )
\        Start-Timers
      endif
    endof
    [char] E of
      cr ." Erasing user Flash area"
      userflash userflashlen (EraseFlash)  ( nowReset )
    endof
    [char] Q of  cr  drop true  endof
      cr ." Invalid command"
      r@ if
        ."  - Running User program"
        drop run-user
      endif
  endcase
  r> drop
;

: Maintenance 	\ mode -- flag ; true if Q pressed
\ *G Enter the maintenance system. If mode is non-zero,
\ ** the system will time out after 3 seconds and run
\ ** an application program if it exists. Flag is
\ ** returned non-zero if Q was pressed.
  CurrentIP CurrentMAC
  .maintenance-menu
  Start-Timers
  if					\ mode is non-zero
    cr ." You have 3 seconds ..."
    0
    begin
      #10 ms
      1+ dup #300 =
      key? or
    until
    #300 < if
      1 (maintenance)
    else
      cr ." no key pressed - Running User program"
      run-user
    then
  else
    0 (maintenance)
  endif
  stop-timers
;

: maintain	\ --
\ *G Run the maintenance system until Q is pressed.
  DataFlash>RAM
  begin  0 maintenance  until
;

: bootstrap	\ --
  (init) init-ser			\ initialise Forth and serial line
  console ipvec !  console opvec !	\ select default I/O device
  cr
  cr ." MPE S3C4510 BootStrap v1.310, 30 August 2002"
  cr ." Copyright (c) 1999-2002 MicroProcessor Engineering Ltd."
  DataFlash>RAM
  begin  1 maintenance drop  again
;


\ *******
\ *S PING
\ *******
\ *P The PING example is used to test Ethernet transmission and
\ ** reception. Before testing transmission, test reception by
\ ** using the maintenance system to download a user application
\ ** and then pinging the target board instead of TFTPing to it.
\ *E Example ARP request:  $42/#2A bytes
\ ** C020.0600  FF FF FF FF FF FF 00 C0  F0 0D 9D 7E 08 06 00 01
\ ** C020.0610  08 00 06 04 00 01 00 C0  F0 0D 9D 7E C0 A8 01 7F
\ ** C020.0620  00 00 00 00 00 00 C0 A8  FF FE 11 0A 75 9C C0 A8

\ *P Note that the equate PING? must be set true to compile
\ ** the PING code.

ping? [if]

create requested-ip	\ -- addr
\ *G Holds IP address to be pinged.
  #192 c,  #168 c,  0 c,  1 c,
create zero-mac
\ *G Holds MAC address 00-00-00-00-00-00
  0 c,  0 c,  0 c,  0 c,  0 c,  0 c,

: SendArpReq	\ -- ; send ARP request
\ *G Send an ARP request.
\ ** All initialisation must have been performed before
\ ** using this word.
  pktIn >r

  EtherAddressBC r@ EHDR.DstMac 6 cmove
  EtherAddress   r@ EHDR.SrcMac 6 cmove
  ET_ARP         r@ EHDR.type w!(n)

  ARPHARD_ETHER  r@ AHDR.htype w!(n)
  ARPPROTO_IP    r@ AHDR.ptype w!(n)
  6              r@ AHDR.hlen c!
  4              r@ AHDR.plen c!
  ARPOP_REQUEST  r@ AHDR.op w!(n)

  EtherAddress   r@ ARP.SenderMAC 6 cmove
  IpAddress      r@ ARP.SenderIP 4 cmove
  zero-mac       r@ ARP.TargetMAC 6 cmove
  requested-ip   r@ ARP.TargetIP 4 cmove

  r> ARP send_ether_pkt
;

: pingonce	\ -- ; an ARP request
\ *G Send an ARP request and process the response once.
\ ** All initialisation must have been performed before
\ ** using this word.
  cr  ticks				\ starting time
  SendArpReq

  #1000 later
  begin
    isrx? if
      pktIn #1536 get_ether_pkt pktInSize !
      pktIn EHDR.type w@(n) ET_ARP =
      pktIn AHDR.op w@(n) ARPOP_REPLY =  and
      pktIn ARP.targetMAC EtherAddress 6 s=  and
    else
      dup expired dup
      if  ." Timed out "  endif
    endif
  until
  drop
  ticks swap - . ." ms"			\ elapsed time
;

: pingit	\ -- 
\ *G Ping the IP address at REQUESTED-IP until a key is pressed.
  Start-Timers  InitEther
  begin
    pingonce  #100 ms
    key?
  until
  Stop-Timers
;

[then]


\ ******
\ *> ###
\ ******

