\ StampApps.fth - Application save, restore and execute routines

((
Copyright (c) 2004
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
))

decimal

\ ============
\ *! stampapps
\ *T Creating turnkey applications
\ ============

\ ===============
\ *S Introduction
\ ===============
\ *P Applications compiled on the Stamp boards are compiled into
\ ** RAM. This area of RAM can be saved in the serial EEPROM
\ ** (USB Stamp, 16k or 64k bytes) or CPU Flash (e.g. TiniARM)
\ ** and reloaded at power up. Optionally a word can be executed
\ ** at power up, so making a turnkey application.

\ *P An application is saved by the word *\fo{TURNKEY ( xt|0 -- )}
\ ** which saves a header and the compiled image, including previously
\ ** compiled code and data, into the serial EEPROM or Flash. If you do
\ ** not want a word to run at power up, a value of 0 is used
\ ** instead of an XT.
\ *E ' MyApp turnkey    \ will execute MyApp at power up
\ ** 0 turnkey          \ will just restore the image 

\ *P At power up, the image is copied into RAM. If the internal
\ ** CRC check fails, the interactive Forth is started
\ ** but may eventually crash if the image is badly corrupted.
\ ** If a turnkey word has been set, it will be executed. If the
\ ** turnkey word returns, the interactive Forth is started.

\ *P If you want to remove a previously compiled image, do this
\ ** with *\fo{EMPTY ( -- )}, which resets the header block

(( in control file
    here equ CommitXt
      0 , 0 , 0 , 0 ,				\ len, xt, inverted xt, CRCapp
      -1 , -1 , -1 , -1 ,			\ kernel CRC, 3 by RFU
    4 cells equ /Commit				\ size of used header
    8 cells equ /Header				\ total size of header
    /Commit equ MinAppSize			\ minimum application size
    sec-end CommitXt - equ MaxAppSize		\ maximum application size
))

SerStamp? [if]
[undefined] AppFlash [if]
$00010000 equ AppFlash	\ -- addr
\ *G Serial Stamp, e.g. TiniARM: Base address of the Flash area
\ ** where RAM programs are saved. The address here is only used
\ ** if it has not already been defined. Note that this address
\ ** should be at the start of a Flash sector.
[then]
[then]


\ **********************
\ *S Saving applications
\ **********************

USBStamp? [if]
: Empty		\ --
\ *G USB Stamp: Erase the header information in the serial EEPROM. This does
\ ** not modify the software in RAM. In order to run without the
\ ** previously loaded software use *\fo{EMPTY} and then *\fo{REBOOT}.
  here /Commit 2dup erase  0 SEPROM!
;
[then]

SerStamp? [if]
: Empty		\ --
\ *G Serial Stamp, e.g. TiniARM: Erase the header information in
\ ** the LPC2106 Flash at *\fo{AppFlash}. This does not modify the software
\ ** in RAM. In order to run without the previously loaded software
\ ** use *\fo{EMPTY} and then *\fo{REBOOT}.
  AppFlash /Commit (EraseFlash)
;
[then]

: Proglen	\ -- len
\ *G Gives the length of the compiled application.
  here Commitxt -
;

CommitXt constant ProgStart	\ -- addr
\ *G Returns the start address of the RAM program

: +CRC16	\ crc b -- crc'
\ *G Update CRC16 for one byte
  8 0 do
    2dup xor  1 and if
      swap 1 rshift $A001 xor swap	\ XOR in polynomial to crc
    Else
      swap 1 rshift swap		\ just shift crc
    Endif
    1 rshift				\ shift b
  Loop
  drop					\ discard b
;

: genCRC	\ crc addr len -- crc'
\ *G Generate a CRC for a block of memory, given an initial CRC
\ ** value.
  [i
  ticks >r  0 -> ticks			\ prevents ticker problems
  ack_i2c? >r  1 -> ack_i2c?		\ prevents I2C problems in save
  bounds				\ -- crc addr+len addr
  ?do  i c@ +crc16  loop
  $FFFF and
  r> -> ack_i2c?
  r> -> ticks
  i]
;

: AppCRC	\ -- crc
\ *G Generate the application CRC.
  0 ProgStart ProgLen genCRC
;

: KernelCRC	\ -- crc
\ *G Generate the kernel CRC for the first 64k bytes of Flash.
  0 0 $10000 GenCRC
;

: Commit	\ xt|0 --
\ *G Prepares the RAM header for saving.
  here 4 erase  align			\ requires four byte boundary
  CommitXt dup /Commit erase		\ -- xt header ; wipe program info
  ProgLen over !			\ -- xt header ; set program length
  tuck  over if				\ -- header xt header
    2dup cell + !			\ set xt
    swap invert swap  2 cells + !	\ set inverted xt
  else
    2dup cell + !			\ xt=0
    2 cells + !				\ inverted xt=0
  endif					\ -- header
  KernelCRC over 4 cells + !		\ fill in the kernel CRC
  AppCRC swap 3 cells + !		\ fill in the application CRC
;

USBStamp? [if]  
: SaveApp	\ --
\ *G USB Stamp: Write the RAM image to the serial EEPROM.
  CommitXt ProgLen 0 SEPROM!
  cr ProgLen . ." bytes saved"
;
[then]

SerStamp? [if]  
: SaveApp	\ --
\ *G Serial Stamp, e.g. TiniARM: Write the RAM image to the
\ ** Flash at 0001:0000h. The ticker interrupt is halted 
\ ** during this operation.
  ticks >r  stop-clock  0 -> ticks
  CommitXt AppFlash ProgLen ProgramFlash
  cr ProgLen . ." bytes saved"
  start-clock  r> -> ticks
;
[then]

cell buffer: NoReload	\ -- addr
\ *G Set this location to $5555AAAA to prevent application reload
\ ** at *\fo{COLD} or reboot. This location is *\b{not} preserved
\ ** across power-down.

: TurnKey	\ xt|0 --
\ *G Initialise the program header and save the application to the
\ ** serial EEPROM. An XT of 0 will cause the image to be reloaded
\ ** without any start up action.
  Commit SaveApp  NoReload off
;


\ **************************************
\ *S Reloading and starting applications
\ **************************************

USBStamp? [if]
: Reload	\ --
\ *G USB Stamp: Reload an application from the serial EEPROM.
  CommitXt /Commit erase		\ wipe RAM header
  0 eel@ MinAppSize MaxAppSize within if \ check program size
    CommitXt  0 eel@ 0
    ?do  i eeb@ over c!  1+  loop
    drop
  endif
;
[then]

SerStamp? [if]
: Reload	\ --
\ Serial Stamp, e.g. TiniARM: Reload an application from the internal Flash.
  CommitXt /Commit erase		\ wipe RAM header
  AppFlash @ MinAppSize MaxAppSize within if \ check program size
    AppFlash CommitXt over @ cmove
  endif
;
[then]

: CheckApp	\ -- flag ; true if ok
\ *G Returns non-zero if the CRC of the application in RAM is good.
  CommitXt 3 cells + dup @off		\ -- addr crc
  dup AppCRC = -rot			\ -- flag addr crc
  swap !				\ -- flag
;

: CheckKernel	\ -- flag ; true if ok
\ *G Returns non-zero if the CRC of the kernel matches the kernel
\ ** CRC saved in the application.
  KernelCRC  CommitXt 4 cells + @ =
;

: ?Restart	\ --
\ *G Reload the application, check the CRC, and execute a correct
\ ** application if required. This word is executed at power up.
\ ** If the application cannot be installed, a warning message 
\ ** is issued and further reloads are inhibited.
\ ** The Forth text interpreter then starts.
  NoReload @ $5555AAAA =		\ check for reload inhibited
  if  exit  endif
  Reload
  CommitXt @ 0= if  exit  then		\ check for 0 length
  CheckKernel CheckApp and if
    CommitXt
    dup cell + @			\ -- header xt
    dup 0=				\ don't use 0 xt
    if  2drop  exit  endif
    swap 2 cells + @			\ -- xt xt'
    over xor -1 <>			\ don't use bad xt
    if  drop  exit  endif
    execute				\ execute good xt
  else
    cr cr ." Application install failed, restarting"
    $5555AAAA NoReload !  Cold
  endif
;
' ?Restart AtCold


\ *******************************
\ *S Cross Compiler Compatibility
\ *******************************

((
: equ		\ x -- ; -- x
\ *G A synonym for *\fo{CONSTANT}, useful when interactively
\ ** compiling code that will later be cross compiled.
  constant
;

: buffer:	\ size -- ; -- addr
\ *G Create a buffer of the given size. At run-time the address
\ ** is returned.
  create allot
;
))

: or!		\ mask addr --
\ *G OR the *\i{mask} with the data at *\i{addr}
  or!
;

: and!		\ mask addr --
\ *G AND the *\i{mask} with the data at *\i{addr}
  and!
;

: bic!		\ mask addr --
\ *G AND the inverted *\i{mask} with the data at *\i{addr}, so
\ ** that any '1' bits in mask are cleared at *\i{addr}.
  bic!
;


\ ==========
\ *S Gotchas
\ ==========
\ *P Once a turnkey application has started, there is no way
\ ** back to the interactive Forth for debugging unless you
\ ** provide it. It is common practice to provide a backdoor
\ ** for engineering and maintenance access.

\ *P Because of supply problems, the first ten engineering sample
\ ** boards only have 16k of serial EEPROM. Do not try to save more
\ ** than 16k in these systems.

\ *P Because the cold chain mechanism is executed before applications
\ ** are reloaded into RAM, application programs may not use the
\ ** word *\fo{ATCOLD}. Applications must explicitly perform any
\ ** initialisation they require.

\ *P The saved image is a snapshot of RAM, including all system variables.
\ ** If your application requires different initial conditions, it must
\ ** explicitly set these variables.


\ ======================
\ *S Application License
\ ======================
\ *P This software is provided for use with boards manufactured
\ ** by MicroProcessor Engineering Ltd (MPE) and New Micros Inc (NMI) only.
\ ** If you want to run this code on other hardware, you MUST
\ ** obtain a license from MicroProcessor Engineering Ltd.

\ *P Your application may provide access to the open Forth for what
\ ** we term "engineering and maintenance access" only. If you
\ ** need your clients to be able to do more than this, e.g. to
\ ** create new words themselves, you *\b{must} obtain a license from
\ ** MicroProcessor Engineering Ltd. In many low volume cases, this
\ ** just consists of purchasing development systems rather than
\ ** production boards. The advantage of this is that your clients
\ ** receive all the latest tools and documentation. You can also
\ ** choose to purchase an MPE cross compiler which enables you
\ ** to add to and modify the contents of the Flash image, as well
\ ** as producing smaller and much faster code.

\ *P Note that none of files on the development distribution may be
\ ** redistributed without permission. If you have any questions
\ ** or concerns about the licenses, please contact MPE directly.


\ ======
\ *> ###
\ ======

(( \ test code
: DumpImg	\ -- ; dump
  here proglen 0 seprom@
  here proglen dump
;

: DumpApp	\ --
  CommitXt proglen dump
;

: CompImg	\ --
  Proglen 0 ?do
    CommitXt i + c@
    i eeb@  <> if
      cr i .word space
      Commitxt c@ .byte space
      i eeb@ .byte space
    endif
  loop
;
))


