\ BootLoader.fth - Bootloader for CDS.

((
Copyright (c) 2009
CDS Advanced Technology bv
  and
Copyright (c) 2009
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *! bootloader
\ *T CDS bootloader
\ ===========
\ *P The bootloader allows applications to be reloaded from
\ ** the FAT file system as required. Applications are placed
\ ** in Flash at $0001:0000..$0007:CFFF, with the Flash sector
\ ** at $0007:D000 being reserved for unit configuration data.

\ *P Application files must be binary image files, with the
\ ** following header:
\ *E  len  contents
\ **   4   CRC16 checksum from next cell to end
\ **   4   length of image including checksum
\ **   4   jump to entry point
\ **       ...

\ *P Note that because the bootloader occupies the Flash at
\ ** address zero, the initialisation code must remap the
\ ** interrupt vectors to RAM, usually at $4000:0000.
\ ** Because of this, an additional 64 bytes (8 entries and
\ ** vectors) must be reserved at the start of RAM.



\ **************
\ *S Error codes
\ **************

interpreter
: NextErr	dup equ  1+  ;
target

200
  NextErr err_bopen		\ file open failed
  NextErr err_bclose		\ file close failed
  NextErr err_bsize		\ file size failed
  NextErr err_bread		\ file read failed
  NextErr err_btoobig		\ file too big
  NextErr err_bprog		\ programming failed
  Nexterr err_bcrc		\ CRC check failed
drop


\ ******************
\ *S File operations
\ ******************
\ *P All fatal errors cause a *\fo{THROW}.

$0001:0000 constant BootImg	\ -- addr
\ *G Address where boot image is loaded.
$0006:D000 constant /MaxBootImg	\ -- len
\ *G Maximum size of the boot image.

-1 value hBootFile	\ -- handle
\ *G Handle for BootFile
0 value /BootFile	\ -- len
\ *G Size of boot file.
0 value ProgNext	\ -- addr
\ *G Address to be programmed next.
[sections URAME
#512 buffer: BootBuff	\ -- addr
\ *G Buffer for the file to be programmed.
sections]

: OpenBootFile	\ --
\ *G Open the boot file *\i{\LCP-SYS.DAT} on the SD card
  s" LCP-SYS.DAT" r/w open-file err_bopen ?throw
  dup to hBootFile
  file-size err_bsize ?throw  drop
  dup to /BootFile
  /MaxBootImg u> err_btoobig ?throw
;

: CloseBootFile	\ --
\ *G Close the boot file
  hBootFile close-file
  -1 to hBootFile  0 to /BootFile
   err_bclose ?throw
;

: ReadBootFile	\ caddr len --
\ *G Read a block from the file.
  hBootFile read-file err_bread ?throw drop
;


\ ******************
\ *S CRC16 and CRC32
\ *****************

((
: +CRC16	\ crc b -- crc'
 \ *G Update CRC16 for one byte
  8 0 do
    2dup xor  1 and if
      swap 1 rshift $A001 xor swap	\ XOR in polynomial to crc
    Else
      swap 1 rshift swap		\ just shift crc
    Endif
    1 rshift				\ shift b
  Loop
  drop					\ discard b
;

: genCRC16	\ crc addr len -- crc'
 \ *G Generate a 16 bit CRC for a block of memory, given an initial CRC.
 \ ** value.
  bounds				\ -- crc addr+len addr
  ?do  i c@ +crc16  loop
  $FFFF and
;
))

: +CRC32	\ crc b -- crc'
\ *G Update CRC32 with one byte.
  8 0 do
    2dup xor  1 and if
      swap 1 rshift $EDB8:8320 xor swap	\ XOR in polynomial to crc
    Else
      swap 1 rshift swap		\ just shift crc
    Endif
    1 rshift				\ shift b
  Loop
  drop					\ discard b
;

: genCRC32	\ crc addr len -- crc'
\ *G Generate a 32 bit CRC for a block of memory, given an initial CRC.
\ ** value.
  bounds				\ -- crc addr+len addr
  ?do  i c@ +crc32  loop
;

: ValidApp?	\ -- flag
\ *G Returns true if the programmed app meets the requirements.
  BootImg 4 + @ /MaxBootImg u> 0= if
    0 					\ initial CRC
    BootImg cell + dup @ cell -		\ -- crc caddr len
    genCRC32  BootImg @ =
    exit
  endif
  0
;


\ ************************
\ *S Read and program file
\ ************************

: EraseApp	\ --
\ *G Erase the application area.
  0 #PrgErrs !  BootImg /MaxBootImg (EraseFlash)
;

: ProgBlock	\ --
\ *G Programming is performed in blocks of 512 bytes.
  BootBuff ProgNext Prog512
  #PrgErrs @ err_bprog ?throw
  #512 addr ProgNext +!
;

: ProgBootFile	\ --
\ *G Read and program the bootfile to Flash.
  BootImg to Prognext			\ start address
  0 #PrgErrs !
  /BootFile begin
     dup
    while
     BootBuff #512 ReadBootFile
     ProgBlock
     #512 - 0 max
  repeat
  drop
;

: (ProgImgFile)	\ --
\ *G Open, read/program and close the image file. All errors
\ ** are fatal and cause a *\fo{THROW}. The application area
\ ** is erased after opening the file.
  OpenBootFile
  EraseApp
  ['] ProgBootFile catch
  CloseBootFile
  throw
;

: ProgImgFile	\ -- ior
\ *G Open, read/program and close the image file. The result
\ ** code is returned 0 on success.
  initFATfs dup 0= if
    drop
    ['] (ProgImgFile) catch
    termFATfs
  endif
;


\ *****************************
\ *S Direct download of new app
\ *****************************

: ProgXmodem1024	\ --
\ *G Program 1024 bytes from *\fo{X-BUFFER} into the memory
\ ** pointed to by *\fo{CUR-ADDRESS}.
  /Xblk #1024 <>
  if  #PrgErrs incr  exit  endif
  X-buffer CUR-ADDRESS @ (Prog512)			\ first 512 bytes
  X-buffer #512 + CUR-ADDRESS @ #512 + (Prog512)	\ second 512 bytes
  #1024 cur-address +!
;

: ReflashApp	\ --
\ *G Download a new application image by 1024/CRC Xmodem and
\ ** program it into Flash at $0001:0000.
  cr ." Erasing " EraseApp
  assign ProgXmodem1024 to-do From-Buffer
  Xmodem-1k  0 #PrgErrs !
  BootImg /MaxBootImg RecvXmodem 0= if
    cr ." Success, " . ." bytes programmed"
  else
    cr ." Failed" drop
  endif
;


\ **********
\ *S Booting
\ **********

: PreBoot	\ --
\ *G Preliminary boot. If the user presses the <Esc> key within
\ ** five seconds, the boot sequence is aborted, and the Forth
\ ** monitor is entered.
  msg" Boot"
  500 0 do
    key? if
      key $01B = if
        unloop
        cr ." Forth monitor entry" cr  quit
        exit
      endif
    endif
    10 ms
  loop
;

: ExecApp	\ --
\ *G Check for a valid app and run it if valid.
  ValidApp? if
    msg" Exec"
    dfi di				\ disable interrupts
    BootImg 2 cells + execute		\ execute app
    msg" Retn"
  else
    msg" NoAp"
  endif
  200 ms
;

: FlashApp	\ --
\ *G Try to Flash a new app from the card and execute it.
  msg" Flsh"
  ProgImgFile 0=
  if  ExecApp  endif
  200 ms
;

: BootSequence	\ --
\ *G Wait for a key press or timeout, check for a valid app,
\ ** Flash a new one, and try again. This is an endless loop.
  begin
    PreBoot  ExecApp  FlashApp
  again
;
\ ' BootSequence AtCold


\ CRC testing

((
Variable Bytewise       \ true for bytewise CCITT, SDLC/HDLC
Variable Feedback       \ generating polynomial
Variable CRC

: Odd   ( char -- flag ; is char odd? )
  1 And
;

: Lo    ( num -- num.lo ; return low byte of num )
  $ff And
;

\ Process CRC for a single byte
: UpdateCRC     \ char -- ; update CRC for a byte
  { B -- }
  Bytewise @ If
    \ Bytewise operation for CCITT, SDLC/HDLC
    B  CRC @ Lo  Xor  Lo -> B
    B Dup  4 LShift  Xor Lo -> B

    CRC @ 8 RShift
    B Dup 8 LShift
    Over  3 LShift
    Rot   4 RShift
    Xor Xor Xor
    CRC !
  Else
    \ Bitwise operation for CRC-16, LRCC-16
    8 0 Do
      B Odd  CRC @ Odd  Xor If
        CRC @  1 RShift  Feedback @ Xor  CRC !
      Else
        CRC @  1 RShift  CRC !
      Endif
      B  1 RShift  -> B
    Loop
  Endif
;

: SendString    \ addr1 addr2 -- ; calculate BCC for range (addr1,addr2)
  1+ swap
  Do
    i c@ UpdateCRC
  Loop
;

: CRCtest	\ addr1 addr2 -- crc ; 16bit CRC-16
  0 CRC !               \ initialise BCC to zero
  $A001 Feedback !
  0 Bytewise !
  SendString            \ addr3 addr1 addr2 addr3 -- ; calculate BCC
  CRC @
;

: crcCheck	\ caddr len --
  2dup 0 -rot gencrc16
  cr ." genCRC:  " .dword
  over + 1- CRCtest
  cr ." CRCtest: " .dword
;
))


\ ======
\ *> ###
\ ======

decimal
