\ FlashProg.fth - Flash programming for STR91x CPUs

((
Copyright (c) 2007
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal


\ ============
\ *! flashstr9
\ *T Programming STR91x Flash
\ ============
\ *P The code in the file ARM\Hardware\STR91x\FlashProg.fth is
\ ** derived from C example code. The STR91x CPUs have two Flash
\ ** banks. The programming code here can only be used for the bank
\ ** that does *\b{not} contain this code.

\ Flash Commands
#define CMD_PROT1CFM    0x01            \ Protect Level 1 Confirm
#define CMD_SERS        0x20            \ Sector Erase Set-up
#define CMD_PRGS        0x40            \ Program Set-up
#define CMD_CLRSTAT     0x50            \ Clear Status Register
#define CMD_PROT1S      0x60            \ Protect Level 1 Set-up
#define CMD_RDSTAT      0x70            \ Read Status Register
#define CMD_BNKERS      0x80            \ Bank Erase Set-up
#define CMD_RSIG        0x90            \ Read Electronic Signature
#define CMD_RDOTP       0x98            \ Read OTP Sector
#define CMD_CFM         0xD0            \ Prog/Ers Resume, Ers Confirm
                                        \ Level 1 Unprotect Confirm
#define CMD_RDARR       0xFF            \ Read Array

\ Status register bits
#define PECS            0x80            \ Prog/Ers Controller Status
#define ESS             0x40            \ Erase Suspend Status
#define ES              0x20            \ Erase Status
#define PS              0x10            \ Program Status
#define PSS             0x04            \ Program Suspend Status
#define SP              0x02            \ Sector Protection Status


: FlDelay	\ --
\ *G Short delay to allow Flash to settle or operation to finish.
  #10 begin 1- dup 0= until drop
;

: FlConfirm	\ cmd addr --
\ *G Issue the command *\i{cmd} at *\i{addr}, a confirm, and
\ ** return to Read Array mode.
\ ** The address must be 32 bit aligned.
  tuck w!  0xD0 over w!  0xFF swap w!  FlDelay
;

: Unprotect	\ --
\ *G remove write protection from both Flash blocks.
  1 _SCU SCU_SCR0 + bic!		\ disable PFQ/BC
  FlashBase BootFlashKb #1024 * bounds
  ?do  0x60 i FlConfirm  $08000 +loop
  NBFlashBase NonBootFlashKb #1024 * bounds
  ?do  0x60 i FlConfirm  $2000 +loop
;


\ From jwSTR91x.fth

[undefined] #PrgErrs [if]
variable #PrgErrs
[then]

: FlWaitDone	\ addr -- status
\ *G Wait for last operation with Flash to finish and return status.
\ ** The input *\i{addr} is that of the Flash area to poll. The Flash
\ ** is returned to Read Array mode.
\ ** The address must be 32 bit aligned.
  begin  dup w@ $80 and  until		\ wait until done
  dup w@ swap				\ final status read
  0x50 over w!				\ clear status reg
  0xFF swap w!				\ back to Read Array mode
  0xFF and				\ mask status bits
;

: Unlock	\ addr --
\ *G Unlock the sector containg addr.
  [i  -4 and  $60 over w!  $D0 swap w!  i]
;

: ReLock	\ addr --
\ *G relock the sector containg addr.
  [i  -4 and  $60 over w!  $01 swap w!  i]
;

: SecErase	\ addr -- ior ; 0=success
\ *G Erase the sector containing the address.
  [i
  $20 over w!  $D0 over w!
  -4 and FlWaitDone $22 and		\ return ES and SP bits
  i]
;

: (EraseFlash)	\ addr len --
\ *G Erase the given region. 
  bounds ?do
    i dup Unlock SecErase
    if  #PrgErrs incr  endif
    i NBFlashBase u<			\ step depends on which Flash region
    if  $8000  else  $2000  endif
  +loop
;

: Prog2		\ w addr -- ior ; 0=success
\ *G Program 2 bytes at a half word address.
  [i
  $40 over -4 and w!			\ Program Set-up
  tuck w!				\ write 16 bit halfword
  -4 and FlWaitDone $12 and		\ return PS and SP bits
  i]
;

: (ProgFlash)	\ src dest len --
\ *G Program *\i{len} bytes from memory at *\i{src} to Flash at
\ ** *\i{dest}. Note that the Flash is assumed to be erased, and
\ ** that no verification is performed except when each half-word
\ ** is programmed.
  bounds ?do				\ -- src
    dup w@ i Prog2			\ program 16 bit word
    if  #PrgErrs incr  endif		\ check for error
    2+
  2 +loop
  drop
;

: .src/dest	\ src dest -- src dest
\ *G Display the addresses and contents of the source
\ ** and destination.
  ."  source: " over .lword  over c@ ." (" .byte ." )"
  ."  dest: " dup .lword  dup c@ ." (" .byte ." )"
;

: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over c@ over c@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    1 + swap  1 + swap
  1 +loop
  2drop
;

: ?prgExit		\ --
\ *G Check the programming status. On error, issue a message and
\ ** exit the *\b{caller}.
  #PrgErrs @ if
    ."  failed, aborted"  r> drop
  endif
;

: ProgramFlash	\ src dest len --
\ *G Using the RAM memory buffer at src, program the Flash at dest
\ ** with len bytes. The relevant Flash pages are erased, the
\ ** Flash is programmed, and the result verified.
\ ** *\i{Src} and *\i{Dest} are forced to 2 byte boundaries, and
\ ** *\i{len} is rounded up to the next 128 byte unit. 
  rot -2 and -rot			\ src to 2 byte alignment
  swap -2 and swap			\ dest to 2 byte alignment
  #127 + #-128 and 			\ len to 128 byte boundary
  0 #PrgErrs !
  cr ." Erasing " 2dup (EraseFlash) ?prgExit
  cr ." Programming "  3dup (ProgFlash) ?PrgExit
  cr ." Verifying "  (VerifyFlash) ?prgExit
  cr ." successful"
;


\ ======
\ *> ###
\ ======

decimal

