\ ReprogApp.fth - reprogram STR91x Flash

((
Copyright (c) 2004, 2007
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
Change max program size to depend on CPU.

Change history
==============
20070815 Converted for STR91x.
20060701 SFP003 Added CopyFlash.
20050627 MPE002 Now uses sector tables.
20040212 MPE001 First release
))

only forth definitions  decimal


\ ****************
\ *S Flash handler
\ ****************

cell buffer: FlDest	\ -- addr
\ *G Holds next destination address

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count

: FlWaitDone	\ addr -- status
\ *G Wait for last operation with Flash to finish and return status.
\ ** The input *\i{addr} is that of the Flash area to poll. The Flash
\ ** is returned to Read Array mode.
\ ** The address must be 32 bit aligned.
  begin  dup w@ $80 and  until		\ wait until done
  dup w@ swap				\ final status read
  0x50 over w!				\ clear status reg
  0xFF swap w!				\ back to Read Array mode
  0xFF and				\ mask status bits
;

: Unlock	\ addr --
\ *G Unlock the sector containg addr.
  -4 and  $60 over w!  $D0 swap w!
;

: SecErase	\ addr -- ior ; 0=success
\ *G Erase the sector containing the address.
  $20 over w!  $D0 over w!
  -4 and FlWaitDone $22 and		\ return ES and SP bits
;

: (EraseFlash)	\ addr len --
\ *G Erase the given region of the boot Flash. 
  bounds ?do
    i dup Unlock SecErase
    if  #PrgErrs incr  endif
  $8000 +loop
;

: EraseFlash	\ addr len --
\ *G Erase all off the boot Flash. 
  FlashBase BootFlashKb #1024 * (EraseFlash)
;

: Prog2		\ w addr -- ior ; 0=success
\ *G Program 2 bytes at a half word address.
  $40 over -4 and w!			\ Program Set-up
  tuck w!				\ write 16 bit halfword
  -4 and FlWaitDone $12 and		\ return PS and SP bits
;

: ProgBuff	\ src dest len --
\ *G Program *\i{len} bytes from memory at *\i{src} to Flash at
\ ** *\i{dest}. Note that the Flash is assumed to be erased, and
\ ** that no verification is performed except when each half-word
\ ** is programmed.
  bounds ?do				\ -- src
    dup w@ i Prog2			\ program 16 bit word
    if  #PrgErrs incr  endif		\ check for error
    2+
  2 +loop
  drop
;

: InitXBuff	\ --
\ *G Initialise the Flash buffer
  Xbase ^FlBuffer !
;

: toFlash	\ src dest --
\ *G Handle 128 bytes at src to go to dest in Flash.
  swap #128 +				\ update RAM pointer
  dup Xtop = if				\ -- dest' src ; got complete unit?
    2drop
    Xbase FlDest @ Xlen ProgBuff	\ program unit
    InitXbuff  Xlen FlDest +!		\ update pointers
  else
    ^FlBuffer !  drop
  endif
;

: LastToFlash	\ -- 
\ *G Program the partial buffer.
  Xbase ^FlBuffer @ over - dup if	\ -- src len
    FlDest @ swap ProgBuff
  else
    2drop
  endif
;


\ *****************
\ *S Numeric output
\ *****************

: .nibble	\ n --
\ *G Convert a nibble to a hex ASCII digit and display it.
  $0F and  dup 9 >
  if  7 +  then
  $30 + emit
;

: .WORD		\ w --
\ *G Display w as a 4 digit unsigned hexadecimal number.
  dup #12 rshift .nibble
  dup 8 rshift .nibble
  dup 4 rshift .nibble
  .nibble
;

: .dword	\ x --
\ *G Display x as an 8 digit unsigned hexadecimal number.
  dup #16 rshift .word .word
;


\ ***************
\ *S Reprogrammer
\ ***************

l: m1
  ", Erasing "
l: m2
  ",  done "
l: m3
  ", Start Xmodem TX "
[defined] REBOOT [if]
l: m4
  ", Press R to reboot, other to reprogram "
[else]
l: m4
  ", Reset or press key to reprogram"
[then]
l: m5
  ", Errors"
l: m6
  ", No "

: $.	\ c-addr --
\ *G Display counted string.
  count type
;

: .PrgErrs
\ *G Display number of errors.
  cr  #PrgErrs @ .word  bl emit
  m5 $.
;

: ReprogFlash	\ --
\ *G Erase and program the Flash.
  init-ser  #20 ms
  begin
    InitXbuff  #PrgErrs off
    cr m1 $.  #20 ms
    EraseFlash
    m2 $.  .PrgErrs  #PrgErrs off
    cr m3 $.  $01A emit			\ send message and trigger
    0 dup FlDest !  $00400000 Bin-Up drop
    LastToFlash
    .PrgErrs
    #100 ms
    cr m4 $.
[defined] REBOOT [if]
    key $DF and dup emit  [char] R =
  until
  #100 ms
  reboot
[else]
    key drop
  again
[then]
;

\ SFP002...
l: msg7
  ", Copy "

: (CopyFlash)	\ src dest len --
\ *G Copy the flash. The parameters are as for *\fo{CMOVE}.
  init-ser  #100 ms
  cr m1 $.  #20 ms			\ "Erase" message
  #PrgErrs off
  2dup (EraseFlash) .PrgErrs  #10 ms	\ -- src dest len
  #PrgErrs @
  if  3drop exit  endif
  cr msg7 $.				\ "Copy" message
  ProgBuff
  .PrgErrs  #20 ms			\ error count
;

: CopyFlash	\ src dest len --
\ *G Copy the flash. The parameters are as for *\fo{CMOVE}.
\ ** The process is repeated until there are no errors, and
\ ** the system is then rebooted.
  begin
    3dup (CopyFlash)  #PrgErrs @ 0=
  until
  #100 ms  reboot
;
\ ...SFP002


\ ======
\ *> ###
\ ======

