\ ARM exception (interrupt) handlers

((
Copyright (c) 2000-2003
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20040127 MPE003 Updated SWI, PAbort and DAbort handlers.
		Fixed MSRs to SPSR to set Cand F fields.
20021009 SFP002 Initialised S0 and R0.
20020430 SFP001 Simplified entry/exit code.
		Added better nesting support.
))

\ =========
\ *! intarm2
\ *T Interrupt handlers
\ =========
\ *P The file ARM\INTARM2.FTH contains generic ARM interrupt
\ ** handlers in the v6.2 style. You are strongly recommended tp
\ ** to use INTARM2 for new code. The file INTARM.FTH is retained
\ ** but will not be supported in future releases.

\ *P Separate files are provided for CPUs with vectored
\ ** interrupt controllers, e.g. INTAT91.FTH, and these may have
\ ** a different word set for initialising interrupts.

\ *P Each interrupt type requires a predefined stack frame on which
\ ** the interrupt handler builds the Forth stacks and USER area.
\ ** The system initialisation code must preset the banked R13 registers
\ ** to the top of this area.

\ *P The IRQ handlers all share a common "stack of stacks". On entry to
\ ** the interrupt handler, the next stack is allocated. It is assumed
\ ** that an interrupt handler may not itself be interrupted by an interrupt
\ ** of the same level. If this will happen, the return stack pointers
\ ** of each banked interrupt must be pre-allocated and set up by the
\ ** interrupt initialisation routine INIT-INT-USER which is called
\ ** by the start-up code. This will require each interrupt routine
\ ** to have its own stack area, and for the user areas to be allocated
\ ** dynamically.

\ *P The number of stacks allocated is defined by the equate #IRQs
\ ** in the control file.

\ *P Note that each interrupt has its own USER area, but that
\ ** NO USER VARIABLES ARE INITIALISED.

\ *P Note that it is assumed that the banked stack pointers have
\ ** already been set up by the hardware initialisation code.

\ *P Note that it is implicit throughout the code that the three
\ ** system registers UP, PSP, RSP are always set so that:
\ *C   UP > PSP > RSP
\ *P Because of this layout, data stack underflows may corrupt
\ ** the first part of the USER area. YOu have been warned.

1 equ test-isr?	\ -- n ; non-zero to compile test code
\ *G Non-zero to compile test code.

only forth definitions
decimal

: init-int-user \ -- ; initialise interrupt user area
\ *G Initialise interrupt user area, usually a noop.
;


\ ********************
\ *S Interrupt management
\ ********************

[undefined] EI [if]
code EI		\ -- ; enable interrupts
\ *G Global enable interrupts, both FIQ and IRQ.
  mrs  r0, cpsr
  bic  r0, r0, # $C0			\ clear I and F bits
  msr  cpsr _c r0			\ modify control bits
  next,
end-code

code DI         \ -- ; disable interrupts
\ *G Global disable interrupts, both FIQ and IRQ.
  mrs  r0, cpsr
  orr  r0, r0, # $C0			\ set I and F bits
  msr  cpsr _c r0			\ modify control bits
  next,
end-code

code [I		\ R: -- x ; preserve I/F status on return stack, disable ints
\ *G Preserve I/F status on return stack, disable interrupts.
\ ** The state is restored by I].
  mrs   r0, cpsr			\ get CPSR
  stmfd rsp ! { r0 }			\ save
  orr   r0, r0, # $C0			\ set I and F bits
  msr   cpsr _c r0			\ modify control bits
  next,
end-code

code I]		\ R: x -- ; restore int. status from r. stack
\ *G Restore interrupt status saved by [I from the return stack.
  ldmfd rsp ! { r0 }
  msr   cpsr _c r0
  next,
end-code
[then]

code SAVE-INT	\ -- x ; return interrupt status and disable interrupts
\ *G Geturn interrupt status and then disable interrupts.
\ ** Use [I and I] for new code.
  stmfd psp ! { tos }			\ preserve TOS
  mrs   r0, cpsr			\ get status register
  mov   tos, r0
  orr   r0, r0, # $C0			\ set I and F bits
  msr   cpsr _c r0			\ modify control bits
  next,
end-code

code RESTORE-INT	\ x -- ; restore state returned by SAVE-INT
\ *G Restore state returned by SAVE-INT.
\ ** Use [I and I] for new code.
  msr   cpsr _c  tos
  ldmfd psp ! { tos }
  next,
end-code


\ **************
\ *S SWI handler
\ **************

\ *P This SWI handler assumes that we may already be in supervisor mode,
\ ** and that the RSP must be preserved. A new RSP stack is allocated, and
\ ** the original RSP, R4..R12 and LINK are saved on the new RSP stack.
\ ** New data stack and USER area are then allocated, parameters R0..R2
\ ** and the SWI# are put on the new data stack, and the handler is executed.
\ ** On return, registers RSP and R4..12 are restored. R2 and R3 will be
\ ** destroyed. Return data (if any) may be placed in R0 and R1.

\ *P By default, on entry to SWI_HANDLER, TOS contains the SWI#, and the
\ ** next three items contain R0..2. This is done so that the standard
\ ** ARM SWI calls can be emulated. Although it is not strictly necessary
\ ** to remove this data, a canonical handler will have the stack effect:
\ *C   SWI_ACTION	\ r2 r1 r0 swi# -- r1' r0'

defer SWI_handler	\ r2 r1 r0 swi# -- r1' r0' ; default action
\ *G The default action is 2DROP. Assign your own action
\ ** to this word.
\ *C   assign mySwi to-do SWI_handler
  assign 2drop to-do SWI_handler	\ set to do nothing

PROC SWI_exception	\ --
\ *G The actual SWI handler which calls SWI_handler above.
\ ** It assumes RSP=R13, R0-2 are parameters, R2-3 will be destroyed,
\ ** Return parameters are in R0-R1, R3..R13 will be preserved
  stmfd rsp ! { r3-r12, link }		\ save unbanked registers + LINK (11 regs)
  mrs   r3, SPSR			\ save SPSR for nesting
  stmfd rsp ! { r3 }

  sub   up, rsp, # up-size		\ USER area at top
  mov   psp, up				\ with DATA stack next
  sub   rsp, up, # sp-size		\ reserve space for data stack
  					\ leaving the return stack lowest
  ldr tos, [ link, # -4 ]		\ Get SWI instruction
  bic tos, tos, # $FF000000		\ mask out instruction bits to get number
  stmfd psp ! { r0-r2 }			\ tos=SWI#, r0..2=parameters
  bl    ' SWI_handler			\ run required word
  mov   r0, tos				\ get return data into registers
  ldmfd psp ! { r1 }

  add   rsp, rsp, # sp-size up-size +	\ restore rsp - does not affect flags
  ldmfd rsp ! { r3 }			\ restore SPSR
  msr   SPSR, _c _f r3
  ldmfd rsp ! { r3-r12, link }		\ restore registers

  mov .s  pc, link			\ return from exception
end-code

SWI_exception SWIV !


\ -------------
\ SWI test code
\ -------------

test-isr? [if]
: testswi	\ r2 r1 r0 -- r1' r0' ; executes swi0
\ *G An example SWI handler.
  decimal  console opvec !  console ipvec !
  hex
  cr ." In SWI: "
  ." RSP = " rp@ u.  ."  PSP = " sp@ u.  ."  UP = " up@ u.
  cr ." SWI exception"
  cr ." SWI number = " .
  cr ." args = " . . .
  cr
  10 11
;
assign testswi to-do SWI_handler

code (run-swi0)	\ r2 r1 r0 -- ; executes swi0
\ *G The primitive that calls SWI 0.
  ldmfd psp ! { r1-r2 }		\ massage parameters
  mov   r0, tos
  stmfd rsp ! { link }		\ nest return address
  swi   # 0
  mov   tos, r0			\ set return data
  stmfd psp ! { r1 }
  ldmfd rsp ! { pc }		\ unnest
end-code

: run-swi0	\ r2 r1 r0 -- ; executes swi0
\ *G Test execution of SWI 0.
  base @ >r
  hex
  cr ." Before: "
  ." RSP = " rp@ u.  ."  PSP = " sp@ u.  ."  UP = " up@ u.
  (run-swi0)
  decimal
  cr ." Returns: " . .
  hex
  cr ." After : "
  ." RSP = " rp@ u.  ."  PSP = " sp@ u.  ."  UP = " up@ u.
  r> base !
;
[then]


\ *****************************
\ *S Undefined instruction handler
\ *****************************

defer Undef_handler	\ ^ins --
\ *G The place holder for the Undefined Instruction handler has
\ ** a default action of DROP.
  assign drop to-do Undef_handler

PROC Undef_exception
  stmfd rsp ! { r0-r12, link }		\ save unbanked registers and LINK
  mrs   r3, SPSR			\ save SPSR for nesting
  stmfd rsp ! { r3 }

  sub   up, rsp, # up-size		\ USER area at top
  mov   psp, up				\ with DATA stack next
  sub   rsp, up, # sp-size		\ reserve space for data stack
  					\ leaving the return stack lowest
  sub   psp, psp, # 4			\ indicate depth=1
  sub   tos, link, # 4			\ point to undefined instruction
  bl    ' Undef_handler

  add   rsp, rsp, # sp-size up-size +	\ restore rsp - does not affect flags
  ldmfd rsp ! { r3 }			\ restore SPSR
  msr   SPSR, _c _f r3
  ldmfd rsp ! { r0-r12, link }		\ restore unbanked registers and LINK
  mov .s pc, link			\ return from exception
end-code

Undef_exception UndefV !


\ ---------------
\ Undef test code
\ ---------------

test-isr? [if]
: testundef	\ ^ins -- ; handle undefined instruction
\ *G Test handler for Undefined Instructions.
  hex  console opvec !  console ipvec !
  cr ." Undef exception at " dup .lword
  cr ." Instruction        " @ .lword
;
assign testundef to-do Undef_handler

: run-undef	\ --
\ *G Causes an Undefined Instruction exception.
  cr ." before"
  [asm
    mrc  7 0 r0, cr0, cr0, 0		\ CP7 doesn't exist, so should fault
  asm]
  cr ." after"
;
[then]


\ **********************
\ *S Prefetch Abort handler
\ **********************

defer PAbort_handler	\ ^ins -- fixed?
\ *G The place holder for the PABORT handler has
\ ** a default action of NOOP.
  assign noop to-do PAbort_handler

PROC PAbort_exception	\ --
\ The PABORT exception code calls PAbort_handler above.
  stmfd rsp ! { r0-r12, link }		\ save unbanked registers and LINK
  mrs   r3, SPSR			\ save SPSR for nesting
  stmfd rsp ! { r3 }

  sub   up, rsp, # up-size		\ USER area at top
  mov   psp, up				\ with DATA stack next
  sub   rsp, up, # sp-size		\ reserve space for data stack
  					\ leaving the return stack lowest
  sub   psp, psp, # 4			\ indicate depth=1
  sub   tos, link, # 4			\ point to bad instruction
  bl    ' PAbort_handler

  cmp   tos, # 0
  add   rsp, rsp, # sp-size up-size +	\ restore rsp - does not affect flags
  ldmfd rsp ! { r3 }			\ restore SPSR
  msr   SPSR, _c _f r3
  ldmfd rsp ! { r0-r12, link }		\ restore unbanked registers and LINK
  sub .ne .s  pc, link, # 4		\ return from exception, rerun instruction
  mov .s      pc, link			\ return, skip instruction
end-code

PAbort_exception PAbortV !

test-isr? [if]
: testpabort	\ --
\ *G Test handler for PABORT exceptions.
  decimal  console opvec !  console ipvec !
  cr ." PAbort exception"
;
assign testpabort to-do PAbort_handler
[then]


\ ******************
\ *S Data Abort handler
\ ******************

defer DAbort_handler	\ ^ins -- fixed? ; true if instruction fixed up
\ *G The place holder for the DABORT handler has
\ ** a default action of NOOP.
  assign noop to-do DAbort_handler

PROC DAbort_exception
  stmfd rsp ! { r0-r12, link }		\ save unbanked registers and LINK
  mrs   r3, SPSR			\ save SPSR for nesting
  stmfd rsp ! { r3 }

  sub   up, rsp, # up-size		\ USER area at top
  mov   psp, up				\ with DATA stack next
  sub   rsp, up, # sp-size		\ reserve space for data stack
  					\ leaving the return stack lowest
  sub   psp, psp, # 4			\ indicate depth=1
  sub   tos, link, # 8			\ point to bad instruction
  bl    ' DAbort_handler

  cmp   tos, # 0
  add   rsp, rsp, # sp-size up-size +	\ restore rsp - does not affect flags
  ldmfd rsp ! { r3 }			\ restore SPSR
  msr   SPSR, _c _f r3
  ldmfd rsp ! { r0-r12, link }		\ restore unbanked registers and LINK
  sub .s .ne  pc, link, # 8		\ return, rerun instruction
  sub .s      pc, link, # 4		\ return, skip instruction
end-code

DAbort_exception DAbortV !

test-isr? [if]
: testdabort
\ *G Test DABORT handler.
  decimal  console opvec !  console ipvec !
  cr ." DAbort exception"
;
assign testdabort to-do DAbort_handler
[then]


\ *******************************************
\ *S Reserved (26 bit address exception) handler
\ *******************************************

defer Unused_handler	\ ^ins -- fixed?
\ *G The place holder for the UNUSED handler has
\ ** a default action of NOOP.
  assign noop to-do Unused_handler

PROC Unused_exception	\ --
\ *G The UNUSED exception handler calls UNUSED_handler above.
  stmfd rsp ! { r0-r12, link }		\ save unbanked registers and LINK
  mrs   r3, SPSR			\ save SPSR for nesting
  stmfd rsp ! { r3 }

  sub   up, rsp, # up-size		\ USER area at top
  mov   psp, up				\ with DATA stack next
  sub   rsp, up, # sp-size		\ reserve space for data stack
  					\ leaving the return stack lowest
  sub   psp, psp, # 4			\ indicate depth=1
  sub   tos, link, # 4			\ point to bad instruction
  bl    ' Unused_handler

  cmp   tos, # 0			\ check return result
  add   rsp, rsp, # sp-size up-size +	\ restore rsp - does not affect flags
  ldmfd rsp ! { r3 }			\ restore SPSR
  msr   SPSR, _c _f r3
  ldmfd rsp ! { r0-r12, link }		\ restore unbanked registers and LINK
  sub .ne .s  pc, link, # 4		\ return from exception, rerun instruction
  mov .s      pc, link			\ return, skip instruction
end-code

Unused_exception UnusedV !

test-isr? [if]
: testunused	\ --
\ *G Test code for the UNUSED exception.
  decimal  console opvec !  console ipvec !
  cr ." Unused exception"
;
assign testunused to-do Unused_handler
[then]


\ **************
\ *S IRQ handler
\ **************
\ *P The ARM architecture does not define a vectored
\ ** interrupt controller, although several CPUs provide
\ ** one. If the CPU has a vectored controller, it should
\ ** be used. This implementation provides a linked list
\ ** of routines to be called, each of which tests for its own
\ ** interrupt and then handles it if required. The last item
\ ** in the chain is the return from IRQ routine.

\ *P To support nested interrupts, the variable NEXT-IRQ
\ ** points to the first ISR to execute.

variable next-irq	\ -- addr
\ *G Holds pointer to next ISR to run.
\ ** ISRs are added to this chain.
l: ^next-irq
  next-irq ,

PROC IRQ_exception	\ -- ; IRQ entry code
\ *G The IRQ handler entry point which executes the chain.
\ ** The USER variables S0 and R0 are initialised.
  stmfd rsp ! { r0-r12, link }		\ save unbanked registers and LINK
  mrs   r3, SPSR			\ save SPSR for nesting
  stmfd rsp ! { r3 }

  sub   up, rsp, # up-size		\ USER area at top
  mov   psp, up				\ with DATA stack next
  sub	psp, psp, # sp-guard cells	\ generate new PSP ; SFP002
  sub	r0, psp, # tos-cached? cells	\ generate new S0 ; SFP002
  str   r0, [ up, # s0-offset ]		\ that is compatible with SP! ; SFP002
  sub   rsp, up, # sp-size		\ reserve space for data stack
  					\ leaving the return stack lowest
  str   rsp, [ up, # R0-offset ]	\ generate new R0 ; SFP002
  ldr   r0, ^next-irq			\ get address of variable NEXT-ISR
  ldr	pc, [ r0 ]			\ branch to contents
end-code

proc IRQ_reti		\ -- ; IRQ exit code
\ *G This code cleans up after the IRQ and is the first item added
\ ** to the IRQ chain.
  add   rsp, rsp, # sp-size up-size +	\ restore rsp - does not affect flags
  ldmfd rsp ! { r3 }			\ restore SPSR
  msr   SPSR, _c _f r3
  ldmfd rsp ! { r0-r12, link }		\ restore unbanked registers and LINK
  sub .s  pc, link, # 4			\ return from exception
end-code
IRQ_reti next-irq !			\ no handlers defined yet
IRQ_exception IRQV !

interpreter
: br24,		\ xt opcode --
  swap
  chere 8 + -				\ subtract HERE (pc in effect)
  4/ $00FFFFFF and                      \ shift addr and mask
  or ,					\ make B/BL and lay down
;

: add-isr	\ xt chain -- ; ' <name> <chain_var> ADD-ISR adds name to the ISR handler list
  cdata align
  dup @ >r  chere swap !		\ get current chain end, and update
  $EB000000 br24,			\ BL xt
  r> $EA000000 br24,			\ B previous
;

: add-irq	\ xt -- ; ' <name> ADD-IRQ adds name to the IRQ handler list
  next-irq add-isr
;
target
heads? [if]
: br24,		\ xt opcode --
\ *G Given an opcode (B or BL) and an execution address,
\ ** compiles a branch or call to it.
  swap
  here 8 + -				\ subtract HERE (pc in effect)
  4/ $00FFFFFF and                      \ shift addr and mask
  or ,					\ make B/BL and lay down
;

: add-isr	\ xt chain -- ; ' <name> <chain_var> ADD-ISR adds name to the ISR handler list
\ *G An action is added to the IRQ or FIQ chain by a phrase of the form:
\ *C   ' <name> <chain_var> ADD-ISR
  align					\ force cell boundary
  dup @ >r  here swap !			\ get current chain end, and update
  $EB000000 br24,			\ BL xt
  r> $EA000000 br24,			\ B previous
;

: add-irq	\ xt -- ; ' <name> ADD-IRQ adds name to the IRQ handler list
\ *G An action is added to the IRQ chain by a phrase of the form:
\ *C   ' <name> ADD-IRQ
  next-irq add-isr
;
[then]


\ **************
\ *S FIQ handler
\ **************

variable next-fiq	\ -- addr
\ *G Holds pointer to next ISR to run. FIQ ISRs are
\ ** added to this chain
l: ^next-fiq
  next-fiq ,

PROC FIQ_exception	\ --
\ *G FIQ entry code.
  stmfd rsp ! { r0-r7, link }		\ save unbanked registers and LINK
  mrs   r3, SPSR			\ save SPSR for nesting
  stmfd rsp ! { r3 }

  sub   up, rsp, # up-size		\ USER area at top
  mov   psp, up				\ with DATA stack next
  sub   rsp, up, # sp-size		\ reserve space for data stack
  					\ leaving the return stack lowest
  ldr   r0, ^next-fiq			\ get address of variable NEXT-FIQ
  ldr	pc, [ r0 ]			\ branch to contents
end-code

proc FIQ_reti		\ --
\ *G FIQ exit code.
  add   rsp, rsp, # sp-size up-size +	\ restore rsp - does not affect flags
  ldmfd rsp ! { r3 }			\ restore SPSR
  msr   SPSR, _c _f r3
  ldmfd rsp ! { r0-r7, link }		\ restore unbanked registers and LINK
  sub .s  pc, link, # 4			\ return from exception
end-code
FIQ_reti next-fiq !			\ no handlers defined yet
FIQ_exception FIQV !			\ set vector

interpreter
: add-fiq	\ xt -- ; ' <name> ADD-FIQ adds name to the FIQ handler list
  next-fiq add-isr
;
target

heads? [if]
: add-fiq	\ xt -- ; ' <name> ADD-FIQ adds name to the FIQ handler list
\ *G An action is added to the FIQ chain by a phrase of the form:
\ *C   ' <name> ADD-FIQ
  next-fiq add-isr
;
[then]


\ ======
\ *> ###
\ ======

