\ LPC-GPIO.fth - easy access to LPC2xxx GPIO pins.

((
Copyright (c) 2010, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20120321 SFP001 Set GPIO mode for ISINPUT and ISOUTPUT.
		Corrected ISPINMODE
))

only forth definitions
decimal

\ ===========
\ *! lpcgpio
\ *T LPC2xxx GPIO utilities
\ ===========
\ *P The code in *\i{ARM\LPC-GPIO.fth} provides utility words for
\ ** accessing the GPIO/FIO pins on a bit by bit basis. The code
\ ** is written for ease of use rather than performance.


\ ****************
\ *S Configuration
\ ****************
\ *P The equates in this block are defaults used if the equates
\ ** have not been defined before this file is compiled.

[undefined] useFIOs? [if]
1 equ useFIOs?	\ -- flag
\ *G Use the FIO access reather than the legacy GPIO access.
\ ** Note that on the Cortex-M3 devices, legacy access is not
\ ** supported. FIO (Fast I/O) is *\b{much} faster. Unless you
\ ** are using LPC21xx/22xx devices before the /O1 releases,
\ ** leave *\fo{useFIOs} set to 1.
[then]

interpreter
: within?  within?  ;
target

useFIOs? [if]
LPCtype #100 / 21 22 within? [if]
: initFIOs	\ --
\ *G To use the FIOs on LPC21xx and LPC22xx parts, SCS bits 0 and
\ ** 1 must be set. This is done at power up.
  $0003 _SCS or!  ;
' initFIOs AtCold
[then]
LPCtype #100 / 23 24 within? [if]
: initFIOs	\ --
\ *G To use the FIOs on LPC23xx and LPC24xx parts, SCS bit 0 must
\ ** be set. This is done at power up.
  $0001 _SCS or!  ;
' initFIOs AtCold
[then]
[then]

useFIOs? [if]
_FIO0 equ PortBase	\ -- addr
\ *G Base address of port definitions.
$20 equ /PortBlock	\ -- len
\ *G Number of bytes per port for peripheral registers.
5 equ #PortShift	\ -- u
\ *G Number of bits to shift to offset by *\fo{/PortBlock}.
FIOSET equ xIOSET	\ -- u
\ *G Offset to port set register
FIOCLR equ xIOCLR	\ -- u
\ *G Offset to port clear register
FIOPIN equ xIOPIN	\ -- u
\ *G Offset to port pin register
FIODIR equ xIODIR	\ -- u
\ *G Offset to port pin register
[else]
_GPIO0 equ PortBase	\ -- addr
$10 equ /PortBlock	\ -- len
4 equ #PortShift	\ -- u
IOSET equ xIOSET	\ -- u
IOCLR equ xIOCLR	\ -- u
IOPIN equ xIOPIN	\ -- u
IODIR equ xIODIR	\ -- u
[then]


\ ********************
\ *S Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number. Bits 0..31 form
\ ** P0.0 to P0.31, bits 32..63 form P1.0 to P1.31 and so on.

target-only
: PIO:		\ port# bit# -- ; -- iobit#
\ *G Define a port I/O bit by name. For example
\ *C   2 11 PIO: P2.11
\ ** defines an I/O bit on port 2, bit 11. At run time, the I/O
\ ** bit number is returned.
  swap 5 lshift or constant  ;
host&target
interpreter
: PIO:		\ port# bit# -- ; -- iobit#
  swap 5 lshift or constant  ;
target


\ ****************
\ *S IO pin access
\ ****************

compiler
: io>port	\ iobit# -- mask offset
  dup $1F and 1 swap lshift		\ -- iobit# mask
  swap 5 rshift #Portshift lshift	\ -- mask offset
;
target

: setPin	\ iobit# --
\ *G Set the pin high.
  io>port xIOSET + PortBase + !  ;

: clrPin	\ iobit# --
\ *G Set the pin low.
  io>port xIOCLR + PortBase + !  ;

: getPin	\ iobit# -- 0/1
\ *G Read the pin state.
  dup $1F and				\ -- iobit# pin#
  swap 5 rshift #Portshift lshift	\ -- pin# offset
  xIOPIN + PortBase + @
  swap rshift 1 and
;

: ioport	\ iobit# -- mask offset
  io>port  ;


\ ***********************
\ *S IO pin configuration
\ ***********************

: mode>mask	\ mode -- modemask
  3 and
  dup 2 lshift or			\ 4 bits
  dup 4 lshift or			\ 8 bits
  dup 8 lshift or			\ 16 bits
  dup #16 lshift or			\ 32 bits
;

: io>mode	\ mode iobit# -- val mask offset
  swap mode>mask swap			\ -- modemask iobit#
  dup $0F and 2* $03 swap lshift	\ -- modemask iobit# bitmask
  swap -rot				\ -- iobit# modemask bitmask
  tuck and swap rot			\ -- val mask iobit#
  4 rshift 2 lshift			\ -- val mask offset
;

: isPinsel	\ sel# iobit# --
\ *G Set the pin configuration 0..3. By default, nearly all pins
\ ** are in I/O pin mode after reset.
  io>mode _PINSEL0 + setMask  ;

[defined] _PINMODE0 [if]
: isPinmode	\ mode# iobit# --
\ *G Set the pin mode 0..3. Not available for CPUs that do not
\ ** have *\fo{PINMODE} registers.
  io>mode _PINMODE0 + setMask  ;	\ SFP001
0 constant PulledUp	\ -- mode#
\ *G Used as the mode# with *\fo{isPinMode} above to define how
\ ** the pin operates as an input.
2 constant NotPulled	\ -- mode#
\ *G Used as the mode# with *\fo{isPinMode} above to define how
\ ** the pin operates as an input.
3 constant PulledDown	\ -- mode#
\ *G Used as the mode# with *\fo{isPinMode} above to define how
\ ** the pin operates as an input.
[then]

: isInput	\ iobit# --
\ *G Set the pin as an input. Selects GPIO mode.
  0 over isPinsel			\ GPIO mode ; SFP001
  io>port xIODIR + Portbase + bic!
;

: isOutput	\ iobit# --
\ *G Set the pin as an input. Selects GPIO mode.
  0 over isPinsel			\ GPIO mode ; SFP001
  io>port xIODIR + Portbase + or!
;



\ ==========================
\ *S Test code for LPC-E2468
\ ==========================

((
\ *[
decimal

4 16 pio: led2
4 17 pio: led1

: tled1		\ --
  led1 isOutput
  begin
    led1 clrPin  200 ms
    led1 setPin  200 ms
  key? until
;

: tled2		\ --
  led2 isOutput
  begin
    led2 clrPin  200 ms
    led2 setPin  200 ms
  key? until
;
\ *]
))

\ ==========================
\ *S Test code for LPC-P2148
\ ==========================

((
\ *[
decimal

0 15 pio: but1
0 16 pio: but2

\ but1 getPin .
\ but2 getPin .
\ *]
))

\ ======
\ *> ###
\ ======

decimal
