\ MinARM.fth - ARM Umbilical Forth minimal code definitions

((
Released for use with the MPE Forth Cross Compiler by:

Copyright (c) 2000, 2008
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

PLEASE NOTE OUR PHONE NUMBER CHANGE


Change history
==============
20081128 SFP010 Added S=.
20000620 SFP009 Matched USER to updated compiler [build 561]
20000608 SFP008 Converted to v6.1 generic I/O model


********************
Resource definitions
********************

On the ARM the following register usage is assumed
  r15         pc      program counter
  r14         link    link register
  r13         rsp     return stack pointer
  r12         psp     data stack pointer
  r11         up      user area pointer
  r10         tos     cached top of stack
  r9          lp      locals pointer
  r8          fsp     float stack pointer
  r0-r7       scratch
))

\ =========
\ *! minarm
\ *T Minimal Umbilical code definitions
\ =========
\ *P The file ARM\MINARM.FTH contains the minimum code definitions
\ ** required to support Umbilical Forth. If additional definitions
\ ** are required, they may be copied to a new file from ARM\CODEARM.FTH
\ ** or COMMON\KERNEL62.FTH.


decimal

\ ***************
\ *S Flow of control
\ ***************

(( have code generators
CODE LEAVE 	\ --
\ *G Remove the current DO..LOOP parameters and jump to the
\ ** end of the DO..LOOP structure.
  add rsp, rsp, # 8                     \ remove limit, index
  ldmfd rsp ! { pc }                    \ jump to exit address
END-CODE

CODE ?LEAVE 	\ flag --
\ *G If flag is non-zero, remove the current DO..LOOP parameters
\ ** and jump to the end of the DO..LOOP structure.
  mov .s tos, tos                       \ set flags
  ldmfd psp ! { tos }                   \ restore TOS
  mov .eq pc, link                      \ flag false so continue
  add rsp, rsp, # 8                     \ flag true so remove limit, index  - if old TOS<>0
  ldmfd rsp ! { pc }                    \ flag true so jump to exit address - if old TOS<>0
END-CODE
))

CODE (DO)	\ limit index --
\ *G The run time action of DO compiled on the target.
  ldmfd psp ! { r1 }                    \ get limit
L: PDO
  ldr r2, [ link ], # 4                 \ get LEAVE address and incr BRANCH addr
  add r1, r1, # $80000000
  sub r0, tos, r1
  stmfd rsp ! { r0-r2 }                 \ push LEAVE then limit, then index on ret. stack
  ldmfd psp ! { tos }                   \ update tos
  next,
END-CODE

CODE (?DO) 	\ limit index --
\ *G The run time action of ?DO compiled on the target.
  ldmfd psp ! { r1 }                    \ get limit
  cmp r1, tos                           \ check not equal
  b .ne pdo                             \ take DO ?
  ldmfd psp ! { tos }                   \ update tos
  ldr pc, [ link ]                      \ and take leave branch
END-CODE

CODE EXECUTE 	\ xt --
\ *G Execute the code described by the XT. This is a Forth equivalent
\ ** to an assembler indirect JSR/CALL instruction.
  mov r0, tos			        \ move CFA
  ldmfd psp ! { tos }		        \ get new TOS
  mov pc, r0                            \ execute CFA - link contains ret addr of execute
END-CODE


\ **************************
\ *S Stack operations and maths
\ **************************

CODE NOOP	\ --
\ *G A NOOP, null instruction.                                          )
  next,
END-CODE

CODE DROP 	\ x -- ; needed by default interrupt handler
\ *G Lose the top data stack item and promote NOS to TOS.
  ldmfd psp ! { tos }
  next,
END-CODE

(( have code generators
CODE MIN	\ n1 n2 -- n1|n2
\ *G Given two data stack items preserve only the smaller.
  ldmfd psp ! { r0 }
  cmp tos, r0
  mov .gt tos, r0
  next,
END-CODE

CODE MAX	\ n1 n2 -- n1|n2
\ *G Given two data stack items preserve only the larger.
  ldmfd psp ! { r0 }
  cmp tos, r0
  mov .lt tos, r0
  next,
END-CODE
))

CODE WITHIN? 	\ n1 n2 n3 -- flag
\ *G Return TRUE if N1 is within the range N2..N3.
\ ** This word uses signed arithmetic.
  ldmfd psp ! { r0, r1 }
  mov r2, # 0
  mov r3, # 0
  cmp r1, r0
  mov .ge r2, # 1
  cmp r1, tos
  mov .le r3, # 1
  tst r2, r3
  mvn .ne tos, # 0
  mov .eq tos, # 0
  next,
END-CODE

CODE WITHIN 	\ n1|u1 n2|u2 n3|u3 -- flag
\ *G The ANS version of WITHIN?.
\ ** Return TRUE if N1 is within the range N2..N3-1.
\ ** This word uses unsigned arithmetic, so that signed compares are
\ ** treated as existing on a number circle.
  ldmfd psp ! { r0, r1 }                \ tos = n3, r0 = n2, r1 = n1
  sub tos, tos, r0
  sub r1, r1, r0
  cmp r1, tos
  mvn .lo tos, # 0
  mov .hs tos, # 0
  next,
END-CODE

(( have code generators
CODE LSHIFT 	\ x1 u -- x2
\ *G Logically shift X1 by U bits left.
  ldmfd psp ! { r0 }
  mov tos, r0, lsl tos
  next,
END-CODE

CODE RSHIFT 	\ x1 u -- x2
\ *G Logically shift X1 by U bits right.
  ldmfd psp ! { r0 }
  mov tos, r0, lsr tos
  next,
END-CODE

CODE S>D 	\ n -- d
\ *G Convert a single number to a double one.
  stmfd psp ! { tos }                   \ push n
  mov .s tos, tos                       \ test for negative n
  mvn .mi tos, # 0                      \ make tos=-1
  mov .pl tos, # 0
  next,
END-CODE
))

CODE UM*	\ u1 u2 -- ud
\ *G Perform unsigned-multiply between two numbers and return double
\ ** result.
  ldmfd psp ! { r0 }
  umull r1, tos, r0, tos                \ hi:lo = u1 * u2
  stmfd psp ! { r1 }                    \ use r1 because of umull register restrictions
  next,
END-CODE

CODE *		\ n1 n2 -- n3
\ *G Standard signed multiply. N3 = n1 * n2.
  ldmfd psp ! { r0 }
  mul tos, r0, tos
  next,
END-CODE

CODE m*		\ n1 n2 -- d
\ *G Signed multiply yielding double result.
  ldmfd psp ! { r0 }
  smull r1, tos, r0, tos                \ hi:lo = n1 * n2
  stmfd psp ! { r1 }
  next,
END-CODE

\ division subroutine - unrolled for a bit extra speed
asmcode
l: udiv64/32	\ r0:tos / tos ; 64/32 unsigned divide -> tos=quot, r0=rem
  rsb .s tos, tos, # 0                  \ negate divisor
  add .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 1
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 2
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 3
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 4
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 5
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 6
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 7
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 8
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 9
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 10
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 11
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 12
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 13
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 14
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 15
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 16
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 17
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 18
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 19
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 20
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 21
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 22
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 23
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 24
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 25
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 26
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 27
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 28
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 29
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 30
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 31
  sub .cc r0, r0, tos
  adc .s r1, r1, r1
  adc .s r0, tos, r0, lsl # 1   \ 32
  sub .cc r0, r0, tos
  adc tos, r1, r1
  next,
end-code

CODE UM/MOD	\ ud un -- urem uquot
\ *G Perform unsigned division of double number UD by single number U
\ ** and return remainder and quotient.
  ldmfd psp ! { r0, r1 }                \ HI:LO=r0:r1=dividend, tos=divisor
  mov r3, link                          \ save link - assumes r3 not corrupted
  bl udiv64/32
  stmfd psp ! { r0 }                    \ store remainder, quotient in tos
  mov pc, r3
END-CODE

: MU/MOD 	\ d n -- rem d#quot
\ *G Perform an unsigned divide of a double by a single, returning
\ ** a single remainder and a double quotient.
  >r  0  r@  um/mod  r>  swap  >r  um/mod  r>
;

CODE FM/MOD 	\  d1 n2 -- rem quot ; symmetric division
\ *G Perform a signed division of double number D1 by single number N2
\ ** and return remainder and quotient using floored division. See the
\ ** ANS Forth specification for more details of floored division.
  ldmfd psp ! { r0, r1 }                \ r0:r1=dividend
l: fm/mod1
  mov .s r3, r0                         \ r3=sign of dividend
  b .pl fm/mod2                         \ form |dividend|
  rsb .s r1, r1, # 0
  rsc r0, r0, # 0
l: fm/mod2
  mov .s r2, tos                        \ r2=sign of divisor
  rsb .mi tos, tos, # 0                 \ form |divisor|
  mov r6, link                          \ store link - r6 is free for use here
  bl udiv64/32
  teq r0, # 0                           \ remainder = 0?
  b .eq fm/mod3                         \ NO
  cmp r2, # 0
  rsb .mi r0, r0, # 0                   \   remainder takes sign of divisor
  eor .s r3, r3, r2                     \   r3=sign of quotient
  rsb .mi tos, tos, # 0                 \   form quotient if neg.
  sub .mi r0, r2, r0                    \   form remainder if neg.
  sub .mi tos, tos, # 1                 \   form quotient if neg.
  stmfd psp ! { r0 }                    \   push rem
  mov pc, r6                            \   exit
l: fm/mod3                              \ YES
  eor .s r3, r3, r2                     \   r3=sign of quotient
  rsb .mi tos, tos, # 0                 \   form quotient
  stmfd psp ! { r0 }                    \   push rem
  mov pc, r6                            \   exit
END-CODE

CODE SM/REM 	\ d1 n2 -- rem quot ; symmetric division
\ *G Perform a signed division of double number D1 by single number N2
\ ** and return remainder and quotient using symmetric (normal) division.
  ldmfd psp ! { r0, r1 }		\ r0:r1 = d1
l: sm/rem1
  mov .s r3, r0
  b .pl sm/rem2
  rsb .s r1, r1, # 0
  rsc r0, r0, # 0
l: sm/rem2
  mov .s r2, tos
  rsb .mi tos, tos, # 0
  mov r6, link
  bl udiv64/32
  teq r0, # 0
  b .eq sm/rem3
  eor .s r3, r3, r2
  rsb .mi tos, tos, # 0
  rsb .mi r0, r0, # 0
  cmp r2, # 0
  rsb .mi r0, r0, # 0
  stmfd psp ! { r0 }
  mov pc, r6
l: sm/rem3
  eor .s r3, r3, r2
  rsb .mi tos, tos, # 0
  stmfd psp ! { r0 }
  mov pc, r6
END-CODE

CODE /MOD 	\ n1 n2 -- rem quot
\ *G Signed division of N1 by N2 single-precision yielding remainder
\ ** and quotient.
  ldmfd psp ! { r1 }                    \ get n1=dividend
  mov .s r1, r1                         \ sign extend
  mvn .mi r0, # 0
  mov .pl r0, # 0
  b sm/rem1				\ join common code
END-CODE

: / 		\ n1 n2  -- n3
\ *G Standard signed division operator. n3 = n1/n2.
  /mod nip
;

: MOD 		\ n1 n2 -- n3
\ *G Return remainder of division of N1 by N2. n3 = n1 mod n2.
  /mod drop
;

: */MOD 	\ n1 n2 n3 -- n4 n4
\ *G Multiply n1 by n2 to give a double precision result, and then
\ ** divide it by n3 returning the remainder and quotient. The point
\ ** of this operation is to avoid loss of precision.
  >r m* r> sm/rem
;

: */ 		\ n1 n2 n3 -- n4
\ *G Multiply n1 by n2 to give a double precision result, and then
\ ** divide it by n3 returning the quotient. The point
\ ** of this operation is to avoid loss of precision.
  */mod nip
;

: M/ 		\ d n1 -- n2
\ *G Signed divide of a double by a single integer.
  sm/rem nip
;

CODE D+ 	\ d1 d2 -- d3
\ *G Add two double precision integers.
  ldmfd psp ! { r0-r2 }                 \ r0 = d2 low, r1 = d1 high, r2 = d1 low
  add .s r0, r0, r2                     \ d3l = d2l + d1l
  adc tos, tos, r1                      \ d3h = d2h + d1h
  stmfd psp ! { r0 }                    \ push d3l
  next,
END-CODE

CODE D- 	\ d1 d2 -- d3
\ *G Subtract two double precision integers. D3=D1-D2.
  ldmfd psp ! { r0-r2 }                 \ r0 = d2 low, r1 = d1 high, r2 = d1 low
  sub .s r0, r2, r0                     \ d3l = d2l - d1l
  sbc tos, r1, tos                      \ d3h = d2h - d1h
  stmfd psp ! { r0 }                    \ push d3l
  next,
END-CODE

CODE DNEGATE 	\ d1 -- d2
\ *G Negate a double number.
L: DNEG1
  ldmfd psp ! { r0 }                    \ r0 = d1l, tos = d1h
  rsb .s r0, r0, # 0                    \ negate low
  rsc tos, tos, # 0                     \ negate high carrying through
  stmfd psp ! { r0 }
  next,
END-CODE

CODE ?NEGATE 	\ n1 flag -- n1|n2
\ *G If flag is negative, then negate n1.
  mov .s tos, tos                       \ set processor flags
  ldmfd psp ! { tos }                   \ discard flag
  rsb .mi tos, tos, # 0                 \ negate n1 if flag=true
  next,
END-CODE

CODE ?DNEGATE 	\ d1 flag -- d1|d2
\ *G If flag is negative, then negate d1.
  mov .s tos, tos                       \ set processor flags
  ldmfd psp ! { tos }                   \ discard flag
  mov r1, link                          \ preserve link
  bl .mi DNEG1                          \ negate d1 if flag=true
  mov pc, r1                            \ restore link and exit
END-CODE

CODE ABS	\ n -- u
\ *G If n is negative, return its positive equivalent (absolute value).
  teq tos, # 0
  rsb .mi tos, tos, # 0
  next,
END-CODE

CODE DABS 	\ d -- ud
\ *G If d is negative, return its positive equivalent (absolute value).
  mov .s tos, tos                       \ set processor flags
  mov r1, link                          \ preserve link
  bl .mi DNEG1                          \ negate d if flag=true
  mov pc, r1                            \ restore link and exit
END-CODE

CODE ROLL 	\ xu xu-1 .. x0 u -- xu-1 .. x0 xu
\ *G Rotate the order of the top N stack items by one place such that
\ ** the current top of stack becomes the second item and the Nth item
\ ** becomes TOS. See also ROT.
  mov r0, tos, asl # 2			\ r0=position of xu in bytes
  cmp r0, # 0				\ u is valid?
  ldmfd .le psp ! { tos }		\   no - get new tos
  mov .le pc, link			\   no - exit
  ldr tos, [ psp, ++ r0 ]		\ put xu in tos
L: ROLL1
  sub r1, r0, # 4
  ldr r2, [ psp, ++ r1 ]
  str r2, [ psp, ++ r0 ]
  mov .s r0, r1
  b .ne ROLL1
  add psp, psp, # 4
  next,
END-CODE


\ *******
\ *S Strings
\ *******

CODE CMOVE	\ c-addr1 c-addr2 u --
\ *G Copy U bytes of memory forwards from C-ADDR1 to C-ADDR2.
  ldmfd psp ! { r0, r1 }                \ r0=adr2, r1=adr1
  mov .s tos, tos
  b .eq CMOV2                           \ u=0
L: CMOV1
  ldrb r2, [ r1 ], # 1
  strb r2, [ r0 ], # 1
  sub .s tos, tos, # 1
  b .ne CMOV1
L: CMOV2
  ldmfd psp ! { tos }
  next,
END-CODE

CODE CMOVE>	\ c-addr1 c-addr2 u --
\ *G As CMOVE but working in the opposite direction,
\ ** copying the last character in the string first.
  ldmfd psp ! { r0, r1 }                \ r0=adr2, r1=adr1
  mov .s tos, tos
  b .eq CMV>2                           \ u=0
  add r0, r0, tos
  add r1, r1, tos
L: CMV>1
  ldrb r2, [ r1, # -1 ] !
  strb r2, [ r0, # -1 ] !
  sub .s tos, tos, # 1
  b .ne CMV>1
L: CMV>2
  ldmfd psp ! { tos }
  next,
END-CODE

CODE FILL	\ c-addr u char -- ; used by multitasker
\ *G Fill U bytes of memory starting at C-ADDR with the byte information
\ ** specified as CHAR.
  ldmfd psp ! { r0, r1 }                \ tos=char, r0=count, r1=addr
L: FIL1
  sub .s r0, r0, # 1
  strb .pl tos, [ r1 ], # 1
  b .pl FIL1
  ldmfd psp ! { tos }
  next,
END-CODE

: erase		\ c-addr u -- ; wipe memory
\ *G Set U bytes of memory starting at C-ADDR with zeros.
  0 fill
;

CODE S=		\ c-addr1 c-addr2 u -- flag
\ *G Compare two same-length strings/memory blocks, returning TRUE if
\ ** they are identical.
  ldmfd psp ! { r0, r1 }                \ r0=adr2, r1=adr1
  mov .s tos, tos
  b .eq S=X                             \ TRUE as len=0
L: S=1
  ldrb r2, [ r0 ], # 1
  ldrb r3, [ r1 ], # 1
  cmp r2, r3
  b .ne S=X                             \ FALSE as mismatched chars
  sub .s tos, tos, # 1
  b .ne S=1                             \ if count=0 then TRUE
L: S=X
  mvn .eq tos, # 0                      \ TRUE
  mov .ne tos, # 0                      \ FALSE
  next,
END-CODE

CODE (") 	\ -- a-addr ; return address of string, skip over it
\ *G Return the address of a counted string that is inline after the
\ ** CALLING word, and adjust the CALLING word's return address to
\ ** step over the inline string. See the definition of (.") for
\ ** an example.
  stmfd psp ! { tos }                   \ save tos
  ldr tos, [ rsp ]                      \ get a-addr
  ldrb r0, [ tos ]                      \ get count byte
  add r0, r0, # 4                       \ 1 for count, 3 to make aligned
  mvn r1, # 3                           \ mov -4 into r1
  and r0, r0, r1                        \ force 4 byte even boundary
  add r0, r0, tos                       \ add to a-addr
  str r0, [ rsp ]                       \ skip over
  next,
END-CODE

: (C")		\ -- c-addr
\ *G The run time action compiled by C".
  (")
;

: (S")		\ -- c-addr u
\ *G The run time action compiled by S".
  (") count
;


\ ***************
\ Runtime support
\ ***************

\ ************************************
\ *S Umbilical versions of defining words
\ ************************************

asmcode
here is-action-of constant
\ *G The runtime code for a CONSTANT.
  stmfd psp ! { tos }                   \ save old TOS
  ldr tos, [ link ]                     \ get const val from PFA
  ldmfd rsp ! { pc }
  next,
END-CODE

asmcode
here is-action-of variable
\ *G The runtime action for a VARIABLE.
  stmfd psp ! { tos }                   \ save old tos
  ldr tos, [ link ]                     \ get var addr from PFA
  ldmfd rsp ! { pc }
  next,
END-CODE

asmcode
here is-action-of value
\ *G The runtime action of a VALUE.
  stmfd psp ! { tos }                   \ save old tos
  ldr tos, [ link ]                     \ get var addr from PFA
  ldr tos, [ tos ]			\ get data
  ldmfd rsp ! { pc }
  next,
END-CODE

asmcode
here is-action-of user
\ *G The runtime action of a USER variable.
  stmfd psp ! { tos }                   \ save old tos
  ldr r0 [ link ]                       \ get ptr to user val from PFA
  add tos, up, r0                       \ add user pointer to user value
  ldmfd rsp ! { pc }
  next,
end-code

interpreter
: u#		\ "<name>"-- u
\ *G An *\fo{INTERPRETER} word that returns the index of the
\ ** *\fo{USER} variable whose name follows, e.g.
\ *C   u# S0
  ' >body
;
target

: CRASH		\ --  ; used as action of DEFER
\ *G The default action of a DEFERed word. A NOOP.
;

asmcode
here is-action-of DEFER	\ Comp: "<spaces>name" -- ; Run: i*x -- j*x
\ *G The runtime action of a DEFERred word.
  ldr r0, [ link ]
  ldmfd rsp ! { link }                  \ allows return to 'top-level' from word
  ldr pc, [ r0 ]                        \ jump to word
end-code


\ ****************
\ *S Display words
\ ****************

: SPACE         \ --
\ *G Output a blank space (ASCII 32) character.
  bl emit
;

: SPACES        \ n --
\ *G Output 'n' spaces, where 'n' > 0.
\ ** If 'n' < 0, no action is taken.
  0 max 0 ?do  space  loop
;

: .nibble	\ n --
\ *G Convert a nibble to a hex ASCII digit and display it.
  $0F and  dup 9 >
  if  7 +  then
  $30 + emit
;

: .BYTE         \ b --
\ *G Display the byte *\i{b} as a 2 digit hex number.
  dup 4 rshift .nibble  .nibble
;

: .WORD		\ w --
\ *G Display w as a 4 digit unsigned hexadecimal number.
  dup #12 rshift .nibble
  dup 8 rshift .nibble
  dup 4 rshift .nibble
  .nibble
;

: .dword	\ x --
\ *G Display x as an 8 digit unsigned hexadecimal number.
  dup #16 rshift .word .word
;

: .lword	\ x --
\ *G A synonym for *\fo{.DWORD} above.
  .dword
;


\ ======
\ *> ###
\ ======

