\ RTC214x.fth - LPC214x Real-Time Clock driver

((
Copyright (c) 2005
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   htpp://www.mpeforth.com

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

\ ==========
\ *! rtc214x
\ *T LPC214x Real Time Clock
\ ==========

\ *************
\ *S RTC access
\ *************

: init_RTC	\ --
\ *G Initialise RTC - does not affect current state or contents.
\ ** Uses the 32kHz crystal.
  $11 _RTC CCR + !		\ clear reset, enable, use 32 kHz crystal
;
' init_RTC AtCold		\ performed at power up.

: term_RTC	\ --
\ *G Stop the RTC, but leave it connected to the 32 kHz crystal.
\ ** Used when setting the clock.
  $10 _RTC CCR + !
;

: getRTC	\ -- date time
\ *G Return date and time from consolidated registers.
  _RTC
  dup CTIME1 + @ $0FFF0F1F and
  swap CTIME0 + @ $071F3F3F and
;

0 value dow	\ -- dow ; 0=Sunday
\ *G Returns the local day of the week, starting at 0=Sunday.
\ ** This value is updated when *\fo{TIME&DATE} below is called.

: RTC>TD	\ date time -- seconds mins hours day month year
\ *G Convert the consolidated date/time values to Forth standard form.
  dup $3F and -rot			\ seconds
  dup 8 rshift $3F and -rot		\ minutes
  dup #16 rshift $1F and -rot		\ hours
  #24 rshift $07 and to dow		\ day of week
  dup $1F and swap			\ day of month
  dup 8 rshift $0F and swap		\ month
  #16 rshift $FFF and			\ year
;

: time&date     \ -- seconds mins hours day month year
\ *G Return the operating system local time and date, and
\ ** set *\fo{DOW} as a side effect.
  getRTC
  begin					\ rollover check
[ tasking? ] [if]  pause  [then]
    getRTC 2swap			\ -- dnew dold
    2over d=
  until					\ -- dnew
  RTC>TD
;

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set the RTC.
\ ** The current value of *\fo{DOW} is also set.
  term_RTC
  dup #100 <				\ change 0..99 to 2000+
  if  #2000 +  endif
  _RTC					\ clock base address
  tuck rtcYEAR + !  tuck rtcMONTH + !  tuck rtcDOM + !
  tuck rtcHOUR + !  tuck rtcMIN + !  tuck rtcSEC + !
  dow swap rtcDOW + !
  init_RTC
;


\ ***********************
\ *S Time and date output
\ ***********************

((
[undefined] dow [if]
2 constant dow	\ -- 2
[then]

[undefined] time&date [if]
: time&date	\ -- s m h dd mm yy
  0 55 11  22 4 2003
;
[then]

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow		\ --
\ *G Display day of week.
  dow 3 * days$ 1+ + 3 type
;

create months	\ -- addr
\ *G String containing 3 character text for the months.
  z", JanFebMarAprMayJunJulAugSepOctNovDec"

: .2r		\ n --
\ *G Display n as a two digit number with leading zeros.
  s>d <# # # #> type
;

: .4r		\ n --
\ *G Display n as a four digit number with leading zeros.
  s>d <# # # # # #> type
;

: .Time&Date	\ --
\ *G Display the system time
\ ** The format is:
\ *C   hh:mm:ss dd Mmm yyyy
  base @ decimal
  time&date				\ -- s m h dd mm yy
  dup #100 <				\ correct two digit year clocks
  if  #2000 +  endif
  rot .2r space				\ day - generate date
  swap 1- 3 * months + 3 type space	\ month
  .4r space				\ year
  .2r ." :"			 	\ hours - generate time
  .2r ." :"				\ minutes
  .2r					\ seconds
  base !
;
))
: .2d		\ u --
\ *G Display u as two digits with leading zero.
  0 <# # # #> type
;

: (.time)	\ secs mins hours --
\ *G Display given time.
  .2d  ." :" .2d  ." :" .2d
;

create months$	\ -- addr
\ *G String containing 3 character text for the months.
  ", JanFebMarAprMayJunJulAugSepOctNovDec"

: (.date)	\ day month year --
\ *G Display given date.
  rot .2d space				\ day
  swap 1- 3 * months$ 1+ + 3 type space	\ month
  4 .r					\ year
;

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow		\ --
\ *G Display day of week from *\fo{DOW}.
  dow 3 * days$ 1+ + 3 type
;

: .time&date	\ --
\ *G Display the day of week, date and time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  .dow space  (.date) space  (.time)
  r> base !
;

: .time		\ --
\ *G Read and display time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  drop drop drop  (.time)
  r> base !
;

: .date		\ --
\ *G Read and display date.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  (.date)  drop drop drop
  r> base !
;

: .zone		\ zone --
  if  3 spaces  else  ." GMT"  endif
;

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  .dow ." , "  (.date) space  (.time)
  space .zone space
  r> base !
;


\ ======
\ *> ###
\ ======
