\ RTC23xx.fth - LPC23xx/24xx Real-Time Clock driver

((
Copyright (c) 2005, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   htpp://www.mpeforth.com

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20101119 SFP001 Made interrupt configurable.
		Refactored to ease use with UnixTime.fth.
))

\ ==========
\ *! rtc23xx
\ *T LPC23xx Real Time Clock
\ ==========
\ *P The RTC for the LPC23xx/24xx devices is very similar to the
\ ** RTC in the LPC214x devices, but is not identical, especially
\ ** with regard to clocking.

[undefined] RTCint? [if]
1 equ RTCint?	\ -- flag
\ *G Set this equate true to compile the RTC one second interrupt
\ ** handler. This version is compiled if *\fo{RTCint?} has not
\ ** been defined yet.
[then]


\ *************
\ *S RTC access
\ *************

: init_RTC	\ --
\ *G Initialise RTC - does not affect current state or contents.
\ ** Uses the 32kHz crystal.
  ~pcRTC _PCONP or!		\ power it up (may be the first time)
  $11 _RTC CCR + !		\ clear reset, enable, use 32 kHz crystal
;
' init_RTC AtCold		\ performed at power up.

: term_RTC	\ --
\ *G Stop the RTC, but leave it connected to the 32 kHz crystal.
\ ** Used when setting the clock.
  $10 _RTC CCR + !
;

: getRTC	\ -- date time
\ *G Return date and time from consolidated registers.
  _RTC
  dup CTIME1 + @ $0FFF0F1F and
  swap CTIME0 + @ $071F3F3F and
;

0 value dow	\ -- dow ; 0=Sunday
\ *G Returns the local day of the week, starting at 0=Sunday.
\ ** This value is updated when *\fo{TIME&DATE} below is called.

: RTC>TD	\ date time -- seconds mins hours day month year
\ *G Convert the consolidated date/time values to Forth standard form.
  dup $3F and -rot			\ seconds
  dup 8 rshift $3F and -rot		\ minutes
  dup #16 rshift $1F and -rot		\ hours
  #24 rshift $07 and to dow		\ day of week
  dup $1F and swap			\ day of month
  dup 8 rshift $0F and swap		\ month
  #16 rshift $FFF and			\ year
;

: time&date     \ -- seconds mins hours day month year
\ *G Return the operating system local time and date, and
\ ** set *\fo{DOW} as a side effect.
  getRTC
  begin					\ rollover check
[ tasking? ] [if]  pause  [then]
    getRTC 2swap			\ -- dnew dold
    2over d=
  until					\ -- dnew
  RTC>TD
;

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set the RTC.
\ ** The current value of *\fo{DOW} is also used.
\ ** You can set the clock from AIDE's PowerTerm window by pressing
\ ** Control-D which will send the current day-of-week, date and
\ ** time, e.g.
\ *C   5 to dow  0 56 14 10 10 2008 SetTime&Date
  term_RTC
  dup #100 <				\ change 0..99 to 2000+
  if  #2000 +  endif
  _RTC					\ clock base address
  tuck rtcYEAR + !  tuck rtcMONTH + !  tuck rtcDOM + !
  tuck rtcHOUR + !  tuck rtcMIN + !  tuck rtcSEC + !
  dow swap rtcDOW + !
  init_RTC
;


\ ***********************
\ *S Time and date output
\ ***********************

: .2d		\ u --
\ *G Display u as two digits with leading zero.
  0 <# # # #> type
;

: (.time)	\ secs mins hours --
\ *G Display given time.
  .2d  ." :" .2d  ." :" .2d
;

create months$	\ -- addr
\ *G String containing 3 character text for the months.
  ", JanFebMarAprMayJunJulAugSepOctNovDec"

: (.date)	\ day month year --
\ *G Display given date.
  rot .2d space				\ day
  swap 1- 3 * months$ 1+ + 3 type space	\ month
  4 .r					\ year
;

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow#		\ dow --
\ *G Display day of week by name.
  3 * days$ 1+ + 3 type  ;

: .time&date	\ --
\ *G Display the day of week, date and time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  dow .dow# space  (.date) space  (.time)
  r> base !
;

: .time		\ --
\ *G Read and display time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  drop drop drop  (.time)
  r> base !
;

: .date		\ --
\ *G Read and display date.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  (.date)  drop drop drop
  r> base !
;

: .zone		\ zone --
  if  3 spaces  else  ." GMT"  endif
;

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  dow .dow# ." , "  (.date) space  (.time)
  space .zone space
  r> base !
;


\ ************************
\ *S RTC interrupt handler
\ ************************

RTCint? [if]
\ *G The code here is compiled if *\fo{RTCint?} is non-zero.

#15 equ RTCSlot#	\ -- n
\ *G Select the VIC slot/priority for the ticker interrupt.
\ ** This is usually the lowest priority interrupt (15).

[undefined] lseconds [if]
variable lseconds
[then]

defer RTCticker	\ --
\ *G By default executes *\fo{NOOP}, assign your own action as
\ ** required. Called in *\fo{RTC-isr} below.
  assign noop to-do RTCticker

: RTC-isr	\ --
\ *G The high level interrupt handler for the one second ticker.
  1 _RTC ILR + !			\ reset RTC interrupt
  1 lseconds +!				\ count seconds
  RTCticker				\ user action
;
' RTC-isr IRQ: RTC-irq	\ -- addr
\ *G Creates the IRQ handler for the ticker.

: +RTCint	\ -- ; start RTC interrupt
\ *G Start the RTC 1 second interrupt. Run at start up.
  di
  _RTC
  0 over rtcCISS + !			\ no subsecond interrupts
  $0FF over AMR + !			\ no alarm interrupts
  1 over CIIR + !			\ interrupt 1 sec
  7 swap ILR + !			\ reset any pending RTC interrupts
  RTC-irq RTCint# RTCSlot# SetIrqIsr 	\ set up and enable interrupt
  ei
;
' +RTCint AtCold

: -RTCint	\ -- ; stop clock interrupt
\ *G Stop the RTC periodic interrupt.
  RTCInt# DisInt
;

[then]


\ ***********************
\ *S LSECONDS conversions
\ ***********************
\ *P If you have LSECONDS, you will need conversion between Unix
\ ** seconds and the time/data functions supported by the RTC.
\ ** This code is taken from *\i{UnixTime.fth} and is compiled if
\ ** the equate *\fo{ConvertSecs?} is true.

#24 #60 * #60 * equ secs/day	\ -- 86400
\ *G Seconds per day.
#60 #60 * equ secs/hr		\ -- 3600
\ *G Seconds per hour
#60 equ secs/min		\ -- 60
\ *G Seconds per minute
#60 equ mins/hr			\ -- 60
\ *G Minutes per hour
#24 equ hrs/day			\ -- 24
\ *G Hours per day
#34698 equ uday0		\ -- 34698
\ *G The Julian day number for Unix day 0 at 1 Mar 1995.

compiler
: j>uday	\ jday -- uday
\ *G Convert a Julian day based at 0 for 1 March 1900 to a Unix
\ ** day number derived from *\fo{LSECONDS}.
  uday0 -
;

: u>jday	\ jday -- uday
\ *G Convert a Unix day number as used by *\fo{LSECONDS} to a Julian
\ ** day based at 0 for 1 March 1900.
  uday0 +
;
target

\ works with symmetric division
: >date		\ jday -- day month year
\ *G Convert a Julian day number to day month year format.
  1461 /mod				\ number of leap-year periods
   4 * >r				\ convert to number of years
   365 /mod				\ number of years and days
   dup 4 = if 2drop 365 3 then		\ correct if Feb 29th
   r> + >r				\ -- days | years
   123 + dup 5 153 */			\ compute month
   dup >r 153 5 */ - dup 0=		\ if 31 day month
   if  31 +  -1 else 0 then r> +	\ compute day
   1- dup 12 > if 12 - r> 1+ >r then	\ correct month and year
   r> 1900 +				\ -- day month year
;

: date>         \ day month year -- jday ; see month codes
\ *G Given day month and year generate the day number.
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if 12 + r> 1- >r then    	\ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

[undefined] /umod [if]
: /umod		\ u1 u2 -- urem uquot
\ *G Single precision division of unsigned U1 by U2 returning
\ ** both remainder and quotient.
  0 swap um/mod
;
[then]

: secs>td	\ lsecs -- sec min hr day mon yr
\ *G Convert a Unix seconds value to time and date.
  secs/min /umod
  mins/hr /umod
  hrs/day /umod				\ -- sec min hr day#
  u>jday >date				\ -- sec min hr day mon yr
;

: td>secs	\ sec min hr day mon yr -- secs
\ *G Use the given values for time and date, and return the Unix
\ ** *\fo{LSECONDS} value.
  date> j>uday secs/day *		\ -- sec min hr lsecs
  swap secs/hr * +
  swap secs/min * +
  +
;


\ ======
\ *> ###
\ ======
