\ SMC91C111 Ethernet driver

((
Copyright (c) 1998-2005
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20050204 SFO004 Conversion to LAN91C111
20030630 SFP003 Fixed Odd/Even problem in GET_ETHER_PKT
20020627 SFP002 Updated documentation.
20020618 SFP001 Revamped for tidiness and PowerNet v3+.
))


\ =======================================================================
\ *! Smc91C9111
\ *T SMC LAN91C111 Ethernet Driver Code
( **                                                                    )
\ *S Introduction
\ *P This is the hardware driver layer for the Standard MicroSystems 
\ ** LAN91C111 Ethernet controller chips as implemented on the
\ ** LogicPD boards.
\ ** The data sheets for these devices may be found at www.smsc.com and
\ ** the part number is LAN91C111.

( **                                                                    )
( *P The controller has 64 8 bit registers organised as four banks of  )
( ** 16.                                                                )
( **                                                                    )
\ *P The source code for the driver is in the file SMC91C111.FTH. The
\ ** code is written for portability rather than speed and uses byte
\ ** accesses only. This permits the code to be used without change on
\ ** big and little endian CPUs regardless of the bus interface width
\ ** 8, 16 or 32 bits.

\ *S Hardware gotchas
\ *P This code assumes a CPU with byte addressing. A cell 
\ ** addressed machine, e.g. most DSPs, will require considerable
\ ** changes to the source code for memory buffer transfers.

\ *P When operating with fast CPUs note the following:
\ *B The cycle time of the chip is 185ns (min. back to back access),
\ ** but the read access time is only 40ns. This is detailed in the
\ ** data sheet.
\ *B At least 400ns is required between a data register access and
\ ** a pointer register access. At least 400ns is required after
\ ** writing the pointer register and accessing the data register.
\ ** See the data sheet.
\ *B The words EC@ and EC! and the optional EW@ and EW! 16 bit words
\ ** by default assume that the SMC chip is memory mapped and that
\ ** that the SMC memory area is contiguous. If you have a 16 or 32
\ ** bit bus and one or more low address lines are ignored you must
\ ** adapt these words yourself to suit your hardware.


\ ****************
\ *S Configuration
\ ****************
\ *P The following must be defined before the file DRIVERS\SMC91C111.FTH
\ ** is compiled.

interpreter
: const equ ;	\ n -- ; -- n
\ *G If you are using a standalone target with heads and you want
\ ** interactive access to all the registers and bit masks, define
\ ** CONST as CONSTANT, otherwise by default CONST is defined as EQU.
target

[undefined] SMC16? [if]
0 equ SMC16?	\ -- flag
\ *G If SMC16 is non-zero, the driver will use 16 bit register
\ ** accesses where possible, otherwise it will use two 8 bit
\ ** accesses. SMC16 should only be set non-zero for little
\ ** endian CPUs.
[then]

[undefined] fastCPU? [if]
1 equ fastCPU?	\ -- n
\ *G Set this value false if no software intervention is
\ ** required to meet the 400ns timing requirement before and after
\ ** changing the pointer register.
[then]

[undefined] Etherbase [if]
$4C000000 const EtherBase	\ -- addr
\ *G Define the base address of the Ethernet controller.
\ ** The default is for the LogicPD 75401-11 board.
[then]

[undefined] smcDiags? [if]
0 equ smcDiags?	\ -- flag
\ *G Set this equate true to compile diagnostic code for
\ ** register dumping and so on. False by default.
[then]

[undefined] eeprom? [if]
0 equ eeprom?	\ -- flag
\ *G Set this equate true if the LAN91C9x has an attached EEPROM
\ ** for configuration data storage. False by default.
[then]

[undefined] sniff? [if]
1 equ sniff?	\ -- flag
\ *G Set this equate true to compile the packet sniffer code,
\ ** which can be used to test Ethernet reception.
[then]

[undefined] EtherAddress [if]
create EtherAddress	\ -- addr
\ *G Holds the Ethernet MAC address (six bytes). Note that you
\ ** must obtain these from the IEEE (www.ieee.org) or from other
\ ** sources.
  $00 c, $10 c, $8B c, $F1 c, $44 c, $20 c,	\ six bytes for 48 bit MAC address
					\ you must obtain these yourself!
[then]

[undefined] IpAddress [if]
create IpAddress	\ -- addr
\ *G Holds the Ethernet IP address (four bytes).
  192 c,  168 c,  1 c,  251 c,		\ assign these as required
[then]


\ ************
\ *S Constants
\ ************

( *D BankSelect         Offset of bank select register                  )
( **                                                                    )
( *H Bank0 Registers                                                    )
( *D TCRL               Transmit Control Register - lo byte             )
( *D TCRH               Transmit Control Register - hi byte             )
( *D TXSTATUS           Transmit Status Register                        )
( *D RCR                Receive Control Register                        )
( *D MCR                Memory Control Register                         )
( **                                                                    )
( *H Bank1 Registers                                                    )
( *D Configl            Configuration Register - lo byte                )
( *D Configh            Configuration Register - hi byte                )
( *D IA0                Hardware Address, Ethernet MSB                  )
( *D IA1                Hardware addr                                   )
( *D IA2                Hardware addr                                   )
( *D IA3                Hardware addr                                   )
( *D IA4                Hardware addr                                   )
( *D IA5                Hardware addr                                   )
( *D Ctrh               Control Register - hi byte                      )
( **                                                                    )
( *H Bank2 Registers                                                    )
( *D MMU                Memory management unit cmd reg                  )
( *D PNR                Packet number register                          )
( *D ARR                Allocation result register.                     )
( *D Pointerl           Memory pointer register                         )
( *D Pointerh           Memory pointer register                         )
( *D Datal              Reg to send packets too...                      )
( *D Datah              Reg to send packets too...                      )
( *D IntStatus                                                          )
( *D IntMask                                                            )
( **                                                                    )
( *H Misc. Constants                                                    )
( *D RelRx              MMU Command to release memory from an rx        )
( *D RxRd               SMC Command to read received packet             )
( *D TxWr               SMC Command to write to tx area of ram          )
( *D AllocIntMask       Memory allocated mask                           )
( *D MaxMsgSize         Max 802.3 Ether frame size in bytes             )
( *D AllocTx                                                            )
( *D bit0               Bitmask                                         )
( *D bit1               Bitmask                                         )
( *D bit2               Bitmask                                         )
( *D bit3               Bitmask                                         )
( *D bit4               Bitmask                                         )
( *D bit5               Bitmask                                         )
( *D bit6               Bitmask                                         )
( *D bit7               Bitmask                                         )
( *D EvenTx             Control byte for tx of even bytes               )
( *D OddTx              Control byte for tx of odd bytes                )
( *D RxTask#            Multi Task ID                                   )
( **                                                                    )

\ =======================================================================

$0E const BankSelect
$00 const Bank0
$01 const Bank1
$02 const Bank2
$03 const Bank3

\ Bank 0 registers
$00 const TCRL
$01 const TCRH
$02 const TxStatus
$05 const RCR
$0A const MCR

\ Bank 1 registers
$00 const Configl
$01 const Configh
$04 const IA0
$05 const IA1
$06 const IA2
$07 const IA3
$08 const IA4
$09 const IA5
$0A const gprl
$0B const gprh
$0C const Ctrl
$0D const Ctrh

\ Bank 2 registers
$00 const MMU
$02 const PNR
$03 const ARR
$06 const Pointerl
$07 const Pointerh
$08 const Datal
$09 const Datah
$0C const IntStatus
$0D const IntMask

\ Various bits and bobs...
$80 const RelRx
$E0 const RxRd
$40 const TxWr
$08 const AllocIntMask
#1530 const MaxMsgSize
$20 const AllocTx

[undefined] bit0 [if]
$01 const bit0
$02 const bit1
$04 const bit2
$08 const bit3
$10 const bit4
$20 const bit5
$40 const bit6
$80 const bit7
[then]

$00 const EvenTx
$20 const OddTx
$20 const OddRx			\ SFP003
$01 const RxTask#


\ =======================================================================
( **                                                                    )
( *S Hardware Interface Layer                                           )
( **                                                                    )
\ =======================================================================

: ec!           \ b offset --
\ *G Set an 8 bit register contents to b, in the selected bank.
  EtherBase + c!
;

: ec@           \ offset -- val
\ *G Read an 8 bit register in the selected bank.
  EtherBase + c@
;

((
SMC16? [if]
: ew!           \ w offset --
\ *G For LITTLE-ENDIAN CPUs only.
\ ** Set a 16 bit register contents to w, in the selected bank.
  EtherBase + w!
;

: ew@           \ offset -- val
\ *G For LITTLE-ENDIAN CPUs only.
\ ** Read a 16 bit register in the selected bank.
  EtherBase + w@
;
[then]
))

optimised? [if]
compiler
: ec!           \ val offset --
\ *G VFX optimising compilers will probably produce shorter
\ ** and faster code by using a compiler macro for EC!.
  EtherBase + c!
;

: ec@           \ offset -- val
\ *G VFX optimising compilers will probably produce shorter
\ ** and faster code by using a compiler macro for EC@.
  EtherBase + c@
;

((
SMC16? [if]
: ew!           \ val offset --
\ *G VFX optimising compilers will probably produce shorter
\ ** and faster code by using a compiler macro for EW!.
  EtherBase + w!
;

: ew@		\ offset -- val
\ *G VFX optimising compilers will probably produce shorter
\ ** and faster code by using a compiler macro for EW@.
  EtherBase + w@
;
[then]
))

target
[then]

fastCPU? [if]
: 400ns		\ --
\ *G For fast CPUs, use this hardware dependent word to ensure
\ ** at least 400ns between back to back Ethernet chip accesses.
\ ** This is required for correct operation of the pointer register.
\ ** The default version here assumes 4 instructions per iteration
\ ** at 5ns per instruction, and the call/return and set up overheads
\ ** are ignored. See FASTCPU? above.
  #20 0 do loop
;
[else]
compiler
: 400ns		\ --
\ *G Use this version if your CPU takes at least 400 ns between back
\ ** to back Ethernet chip accesses. See FASTCPU? above.
;
target
[then]

: SetBank	\ bank -- ; bank = 0 to 3
\ *G Select the active register bank.
  $03 and  BankSelect EtherBase + c!
;


\ =======================================================================
( **                                                                    )
( *S Diagnostics                                                        )
( **                                                                    )
\ =======================================================================

smcDiags? [if]

: .bank         \ --
( *G Display register contents for currently selected SMC bank.         )
  $10 0
  do  i ec@ .byte space  loop
  cr
;

: .reg          \ --
( *G Display contents of all SMC Ethernet Controller's registers        )
  base @  hex
  cr ." Offset  " $10 0 do i .byte space loop cr cr
  #4 0 do
    i setbank  ." Bank " i . space  .bank
  loop
  base !
;

[then]


\ =======================================================================
( **                                                                    )
( *S Driver Layer                                                       )
( **                                                                    )
\ =======================================================================

: AllocateMemory	\ n --
\ *G Allocate n bytes of memory to hold a packet for transmission.
\ ** This word will block until sufficient memory is available.
  bank2 setbank
  $10 +                         \ because we will add to the msg length later
  #8 rshift			\ because we allocate in 256 byte chuncks
  AllocTx or                    \ add command and length together
  MMU ec!                       \ send the rconstantest to ether

  begin
    bank2 setbank
    IntStatus ec@               \ get interrupt status register
    AllocIntMask and 0=         \ loop until bit set
   while
[ tasking? ] [if]  pause  [then]
  repeat

  ARR ec@ $1F and                \ extract allocated packet number
  PNR ec!                        \ store at packet number register
;

: QueueTx       \ --
( *G Enqueue last written packet for transmission.                      )
  $C0 MMU ec!                    \ enqueue packet command
;

: WritePacket   \ addr len --
( *G Write an Ethernet frame to chip for transmission                   )
  >r                            \ store len for later
  $00 Datah ec!                 \ write status word
  $00 Datah ec!
  r@ 6 +                        \ len+6,=status,length and control word
  $FFFE and                     \ ensure an even number
  dup 8 rshift  swap $FF and	\ split len to two 8 bit units - faster on ARM
  Datah ec! Datah ec!           \ store low then high count byte
  r@ bounds                     \ addr len bounds for loop
  do  i c@ Datah ec!  loop	\ get buffer entry, store to chip
  r> 1 and                      \ was length odd ?
  if                            \ y:
    OddTx Datah ec!             \ packet control byte
  else                          \ n:
    $00 Datah ec!               \ dummy write
    EvenTx Datah ec!            \ packet control byte
  endif
;

: init-SMC	\ --
\ *G Initialise SMC chip.
\  cr ." Configure SMC Ether Device"
\  cr ."   Reset MMU"
  Bank2 SetBank  $40 MMU ec!		\ Reset MMU
\  cr ."   Interrupt Source (unused)"
  bank1 SetBank  $00 Configl ec!	\ Active HI ints
\  cr ."   Twisted Pair media"
  Bank1 SetBank  $00 Configh ec!	\ select 10BaseT transmission media
\  cr ."   Reserve TX frame space"
  Bank0 SetBank  $06 MCR ec!		\ reserves 6*256=1536 bytes for TX
;

: init-EtherTx  \ --
\ *G Enable Ethernet TX module.
\  cr ." SMC Ether Device, Enable TX"
  Bank0 SetBank
  Tcrl ec@ 1 or Tcrl ec!
  Bank1 SetBank  $08 Ctrh ec!    \ set auto release of transmit memory
;

: init-EtherRx  \ --
\ *G Enable Ethernet RX module.
\  cr ." SMC Ether Device, Enable RX"
  Bank0 SetBank
  $03 Rcr ec!                    \ enable receiver and strip rx crc
  Bank2 SetBank
  bit0 IntMask ec!               \ enable rx ints on ether chip
;

: etheradd>CS   \ addr --
\ *G Store 6 byte Ethernet address into SMC chip from memory buffer.
  cr ." SMC Ether Device, Hardware Address "
  Bank1 SetBank
  dup c@ dup .byte IA0 ec! 1+ [char] - emit
  dup c@ dup .byte IA1 ec! 1+ [char] - emit
  dup c@ dup .byte IA2 ec! 1+ [char] - emit
  dup c@ dup .byte IA3 ec! 1+ [char] - emit
  dup c@ dup .byte IA4 ec! 1+ [char] - emit
      c@ dup .byte IA5 ec!
;

: InitEther	\ --
\ *G Perform a full initialisation of the chip, enabling
\ ** Rx and Tx, and setting up the Ethernet MAC address
\ ** from ETHERADDRESS.
  init-SMC  EtherAddress etheradd>CS
  init-EtherRx init-EtherTx
;

: GetAddrs	\ -- ipaddr macaddr 0
\ *G For Generic I/O IPDGetAddr function
  IpAddress EtherAddress 0
;

: SetAddrs	\ ipaddr|0 macaddr|0 mode --
\ *G Set up the Ethernet MAC and the IP addresses.
\ ** At present mode is always 0, but will be used in
\ ** future releases to indicate data formats. If
\ ** ipaddr or macaddr are zero, the stored data
\ ** will not be changed.
  drop					\ flags
  ?dup if				\ macaddr
    dup EtherAddress 6 cmove
    etheradd>CS
  endif
  ?dup if				\ ipaddr
    IpAddress 4 cmove
  endif
;

: IsRx?         \ -- t|f
\ *G Check if incoming data is present, returning true if a
\ ** packet is available for *\fo{get_ether_pkt} below.
  Bank2 SetBank
  IntStatus ec@ 1 and
;

: get_ether_pkt \ *dest maxlen -- len
\ *G Get pending receive packet to a buffer.
\ ** Note that data is valid only when *\fo{ISRX?} returns true,
\ ** so that you must poll with *\fo{ISRX?} before using
\ ** *\fo{GET_ETHER_PKT}.
  Bank2 SetBank
  $00 Pointerl ec!              \ reset pointer value
  RxRd Pointerh ec!             \ read rx ram area, auto inc low pntr
  400ns				\ See SMC91C9x data sheet
  Datah ec@ drop                \ discard low status
  Datah ec@ drop                \ discard high status
  Datah ec@ Datah ec@ 8 lshift or \ 2 * 8 bit -> 16 bit number ; includes control info
  $07FF and                     \ byte count only uses 11 bits - always even
  $06 -                         \ already read 4 bytes, ignore last control bytes
  min				\ -- addr len ; prevent buffer overrun
(( del SFP003
  tuck bounds ?do		\ loop for rest of message
    datah ec@  i c!		\ copy byte from ether chip ram to buffer
  loop
))
\ SFP003...
  2dup bounds ?do		\ loop for even bytes of message
    datah ec@  i c!		\ copy byte from ether chip ram to buffer
  loop
  tuck				\ -- len addr len
  datah ec@			\ potential last odd byte
  datah ec@ OddRx and if	\ control byte, check if odd count
    -rot + c!			\ stash last byte
    1+				\ increment count
  else
    drop 2drop
  endif
\ ...SFP003
    
  RelRx MMU ec!                 \ remove and release memory from rx
;

: DiscardRX     \ --
\ *G Throw away pending input packet due to lack of memory.
  Bank2 SetBank  RelRx MMU ec!	\ remove and release memory from rx
;

: IsTx?		\ n -- flag
\ *G Return true if a packet of length n can be sent.
  drop true
;

: send_ether_pkt        \ addr len --
\ *G Send packet from supplied buffer.
\ ** The word will block until sufficient packet memory is available.
\ ** Note that you should read incoming packets regularly, otherwise
\ ** it is possible to get into a situation in which SEND_ETHER_PKT
\ ** blocks for ever because there is not enough space in the
\ ** device for the transmission packet. Theoretically, this should
\ ** not happen because 1536 bytes are reserved for transmission,
\ ** but ...
  $40 max			\ ensure packet is at least 64 bytes
  Bank2 SetBank
  dup AllocateMemory		\ get memory for transmission
  $00 Pointerl ec!		\ reset offset value
  TxWr Pointerh ec!		\ set to write Tx ram area
  400ns				\ See SMC91C9x data sheet
  WritePacket			\ send packet to ether chip
  QueueTx
;

((
: set-ether     \ e1 e2 e3 e4 e5 e6 --
( *G Set device 48 bit Ethernet address. )
  Bank1 SetBank
  IA5 ec!
  IA4 ec!
  IA3 ec!
  IA2 ec!
  IA1 ec!
  IA0 ec!
;
))


\ ******************
\ *S Attached EEPROM
\ ******************

eeprom? [if]

: ee!           \ word offset --
\ *G Write a 16 bit word into the EEPROM at EEPROM address
\ ** offset. The EEPROM may not be present in all SMC91C9x
\ ** implementations.
  Bank2 SetBank  pointerl ec!
  Bank1 SetBank  dup 255 and gprl ec!
  8 rshift gprh ec!
  ctrl ec@  $F8 and $5 or  ctrl ec!
  begin  ctrl ec@ 1 and  0= until
  ctrl ec@  $F8 and  ctrl ec!
;

: ee@           \ offset -- word
\ *G Read a 16 bit word from the EEPROM at EEPROM address
\ ** offset. The EEPROM may not be present in all SMC91C9x
\ ** implementations.
  Bank2 SetBank pointerl ec!
  Bank1 SetBank
  ctrl ec@  $F8 and $6 or  ctrl ec!
  begin  ctrl ec@ 2 and  0= until
  ctrl ec@  $F8 and  ctrl ec!
  gprh ec@ 8 lshift  gprl ec@ or
;

[then]


\ ****************************************
\ *S Generic I/O for PowerNet v3 and above
\ ****************************************

\ *P The layout and usage of this structure is defined in the
\ ** file COMMON\ETHERCOM.FTH.
\ *E create IPdevice
\ **   ' MyInit ,       \ 0: initialisation
\ **   ' MyTerm ,	\ 1: shutdown
\ **   ' MyRx? ,	\ 2: receive test
\ **   ' MyRx		\ 3: receive packet
\ **   ' MyTx? ,	\ 4: transmit test
\ **   ' MyTx ,		\ 5: transmit packet
\ **   ' MyGetAddr ,    \ 6: Get device addresses
\ **   ' MySetAddr ,    \ 7: Set device addresses
\ **   ' MySave ,	\ 8: Save IP device state

create SMCvector	\ -- addr
\ *G The device vector needed by PowerNet v3+ and
\ ** COMMON\ETHERCOM.FTH.
  ' InitEther ,         \ initialisation      
  ' noop ,	        \ shutdown            
  ' IsRx? ,	        \ receive test        
  ' get_ether_pkt ,     \ receive packet      
  ' IsTx? ,	        \ transmit test       
  ' send_ether_pkt ,    \ transmit packet     
  ' GetAddrs ,		\ Get device addresses
  ' SetAddrs ,	        \ Set device addresses
  ' noop ,	        \ Save IP device state

SMCvector constant IPDevice	\ -- addr
\ *G IPDevice is the default device name required by ETHERCOM.FTH.
\ ** If you have multiple ports, only one should be called IPDevice.

  
\ **************
\ *S System test
\ **************

sniff? [if]

1536 buffer: pbuff	\ -- addr
\ *G Temporary buffer for SNIFF.

: sniff		\ --
\ *G Listens to the network and displays all the traffic that
\ ** has a broadcast destination or is for this device.
\ ** SNIFF can be used to test reception.
  InitEther
  begin
    isrx? if
      pbuff 1536 erase
      pbuff 1536 get_ether_pkt
      pbuff swap 1 max dump
    then
[ tasking? ] [if] pause [then]
    key?
  until
;

[then]


\ ******
\ *> ###
\ ******


