\ i2cLPC23xx.fth - I2C driver for LPC23xx/LPC24xx

((
Copyright (c) 2008, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20151015 BJC002 Added optional diagnostic output
		Added Timeouts for I2C wait loops in (i2cwait) and  (stop_i2c)
		Corrected conditional compilation for error display in two places.
		Corrected clearing of interrupt flag before issuing STOP - hardware gets
		confused if done the other way around.
		Improved error checking for final bytes in transfers - code used to go through
		error path on NACK'd bytes always.
		Commented debug reporting of NACK when #SHO-I2C-ERRS is set, as this generates
		false reports due to polling for EEPROM ready. 
))
DECIMAL

#102   EQU   I2CLPC23xx_FILE_VER    \ ADC file revision   

only forth definitions decimal

#100   EQU   #I2C-TIMEOUT    \ timeout for hardware, in ms

[DEFINED] #SHO-I2C-ERRS [IF]
  cr ." Compiling error reports for I2C words"
[ENDIF]

\ ============
\ *! i2lpc23xx
\ *T I2C primitives for LPC23xx/24xx
\ ============
\ *P The code in *\i{ARM\drivers\i2cLPC23xx} provides the following
\ ** words for constructing drivers for I2C devices:
\ *D START_I2C Generate a START condition
\ *D STOP_I2C  Generate a STOP condition
\ *D READ_BYTE_I2C Read a byte from the I2C lines
\ *D WRITE_BYTE_I2C Write a byte to the I2C lines
\ *D +ACK_I2C Generate and test the ACK bit
\ *D -ACK_I2C Do not generate and test the ACK bit.

\ *P The code provides the same interface as that provided by
\ ** *\i{Examples\i2c\i2cBase.fth}.

\ Bits in the I2CONSET and I2CONCLR registers
bit6 equ ~I2EN
bit5 equ ~STA
bit4 equ ~STO
bit3 equ ~SI
bit2 equ ~AA

#98  equ err_I2C_Timeout \ -- n 
\ *G Error code returned on timeout.
#99  equ err_I2C_Gen	\ -- n
\ *G Error code returned on general failure.
#100 equ err_I2C_ACK	\ -- n
\ *G Error code returned on ACK failure.

: (INITI2C)	\ high/low base --
\ *G After power up and pin selection, initialise the I2C channel
\ ** at *\i{base} in master mode with clocking defined by
\ ** *\i{high/low} where the top 16 bits are the high value).
  $6C over I2CONCLR + !			\ reset
  over $FFFF and over I2SCLL + !	\ low clock count
  swap #16 rshift over I2SCLH + !	\ high clock count
  $40 swap I2CONSET + !			\ enable
;

: (I2CWAIT)	\ base --
\ *G Wait for an I2C operation to complete.
  #I2C-TIMEOUT later swap	\ leave expiry time under addr
  begin
    over expired if		\ NOTE: EXPIRED has PAUSE, so must re-check to avoid false timeouts
      dup I2CONSET + @ ~SI and 0= if
[ [DEFINED] #SHO-I2C-ERRS ] [IF]
        [io >ser cr ." (I2CWAIT): Timeout!" io]
[ENDIF]
        init_I2Cx		\ attempt to reset hardware
        err_I2C_Timeout throw
      endif
    endif
    dup I2CONSET + @ ~SI and
  until
  2drop
;

: (START_I2C)	\ base --
\ *G Issue a START on the device.
  ~SI over I2CONCLR + !			\ clear pending interrupt
  ~STA over I2CONSET + !
  dup (i2cWait)
  ~STA swap I2CONCLR + !
;


: (STOP_I2C)	\ base --
\ *G Issue a STOP on the device
  ~STO over I2CONSET + !			\ DO THIS FIRST
  ~SI over I2CONCLR + !			        \ THEN, clear pending interrupt
  #I2C-TIMEOUT later swap			\ leave expiry time under addr
  begin
    over expired if				\ pause called via expired
[ [DEFINED] #SHO-I2C-ERRS ] [IF]
      [io >ser cr ." (STOP_I2C): Timeout!" io]
[ENDIF]
      init_I2Cx      \ attempt to reset hardware
      err_I2C_Timeout throw
    endif
    dup I2CONSET + @ ~STO and 0=
  until
  2drop
;

cell +user <ack_i2c?>	\ -- addr
\ *G Holds -1 to acknowledge ACK bits. Note that this is a *\fo{USER} variable
\ ** and hence only applies to the current task.

: (i2cAA)	\ base --
  <ack_i2c?> @
  if  ~AA swap I2CONSET + !  else  ~AA swap I2CONCLR + !  then
;

: (i2c@)	\ base -- 8b
\ *G Receive a byte from the given channel, checking the status
\ ** if required.
\ NOTE:  Does not clear the interrupt flag if status is not ACK'd data.  This
\        is because a STOP will follow, and the interrupt flag should be set
\        AFTER stop is sent.
  dup (i2cAA)				\ set/clear AA bit
  ~SI over I2CONCLR + !			\ clear pending interrupt
  dup (i2cWait)				\ wait until read
  dup I2STAT + @ $F8 and		\ get the status value
  dup $50 <>				\ not ACK'd data byte
\  dup 0= if				\ .. an ack'd byte   | base stat flag
\    ~SI 3 pick I2CONCLR + !		\ clear pending interrupt
\  endif
  swap $58 <> and			\ and not NAK'd data byte
  err_I2C_Gen ?throw
  I2DAT + @ $FF and			\ return data
;

: (i2c!)	\ 8b base --
\ *G Transmit a byte to the given channel, checking the status
\ ** if required.
  dup (i2cAA)				\ set/clear AA bit
  tuck I2DAT + !			\ data byte
  ~SI over I2CONCLR + !			\ clear pending interrupt
  dup (i2cwait)				\ wait for completion
  I2STAT + @ $F8 and		\ get the status value
  dup $020 =
  over $048 = or        \ .. byte was NACK'd
  <ack_i2c?> @ and if   \ .. ACK is expected
((
[ [DEFINED] #SHO-I2C-ERRS ] [IF]
    [io >ser cr ." (I2C!) NACK " .byte io]
[ENDIF]
))
    err_I2C_ACK throw
  endif
  dup  $028 <>			    \ good data byte
  over $018 <> and			\ good SLA+W
  over $040 <> and			\ good SLA+R
  if
[ [DEFINED] #SHO-I2C-ERRS ] [IF]
    [io >ser cr ." (I2C!) +ack: bad status " .byte io]
[ENDIF]
    err_I2C_Gen throw    \ some other error
  else
    drop
  endif
;


\ ************************
\ *S Standard MPE word set
\ ************************

_I2C0 equ _I2Cx	\ -- base
\ *G Defines the channel to which the standard word set applies.

: START_I2C   	\ --
\ *G Define a start condition, bring SDA high-low while SCL is high
  _I2Cx (start_i2c)
;

: STOP_I2C  	\ --
\ *G Define a stop condition, bring SDA low-high while SCL is high
  _I2Cx (stop_i2c)
;

: ACK_I2C?	( -- n )  <ack_i2c?> @  ;
\ *G If a non-zero value is used, the system will
\ ** generate the ACK bit on SDA for reads, otherwise SDA
\ ** is left high. For a non-zero value on writes, the
\ ** system will check the ACK bit and *\fo{THROW} if it is non-zero.
compiler
: ACK_I2C?	( -- n )  <ack_i2c?> @  ;
target

: +ACK_I2C	\ --
\ *G check ACK cycles.
  <ack_i2c?> on
;

: -ACK_I2C	\ --
\ *G Do not check ACK cycles.
  <ack_i2c?> off
;

: READ_BYTE_I2C		\ -- byte
\ *G Read a byte, MSB first. If *\fo{I2C_ACK?} is
\ ** non-zero, pull SDA low during the ACK bit, otherwise
\ ** leave it high
  _I2Cx (i2c@)
;

: WRITE_BYTE_I2C	\ byte --
\ *G Write a byte by shifting, MSB first. If *\fo{ACK_I2C?} is
\ ** set, read ack bit on SDA, and *\fo{THROW} if the ACK bit is
\ ** high.
  _I2Cx (i2c!)
;

#100000 equ I2C0-SPEED	\ -- hz
\ *G Required I2C clock speed.
4 equ pcdI2C0	\ -- n
\ *G Peripheral clock division ratio for I2C0.
pcdI2C0 PCLKdiv #14 lshift equ /pcdI2C0	\ -- mask
\ *G Bit mask for UART0 peripheral clock divider.
1 [ps 27 +ps 28 +ps ps] equ I2C0psval equ I2C0psmask
\ *G I2C0 pinselct value and mask for PINSEL1
system-speed pcdI2C0 / I2C0-speed / equ I2C0div	\ -- u
\ *G Divider for I2C SCLL+SCLH registers
I2C0div 2/ dup #16 lshift or equ I2C0divHL	\ --
\ *G Merged divsion values assuming a 50% duty cycle clock.

: INIT_I2Cx	\ --
\ *G Initialise default I2C channel (I2C0).
  ~pcI2C0 _PCONP or!			\ power up
\  /pcdI2C0 $C000 _PCLKSEL0 setMask	\ clock speed - causes crash!
  I2C0psval I2C0psmask _PINSEL1 setMask	\ pin selection
  I2C0divHL _I2Cx (initI2C)
  +ack_i2c  stop_i2c  #10 ms
;


\ ************
\ *S Test Code
\ ************
\ *G The tests were performed using an SAA1064 display controller
\ ** at I2C address $70.

0 [if]
\ *G Use *\fo{0 [if]} not to compile the test code.

((
_PCONP constant _PCONP
_PINSEL0 constant _PINSEL0
_PINSEL1 constant _PINSEL1
_PCLKSEL0 constant _PCLKSEL0
_PCLKSEL1 constant _PCLKSEL1
_I2Cx constant _I2Cx
))

: .i2cstat	\ --
\ *G Display I2C status byte.
  _I2Cx I2STAT + @ .byte
;

: >disp		\ nn..n1 n reg --
\ *G Send *\i{n} bytes starting at register *\i{reg}.
  +ack_i2c				\ all bytes acknowledged
  start_i2c  $70 write_byte_i2c		\ SAA1064 address +w
  $07 and write_byte_i2c		\ first reg
  0 ?do  write_byte_i2c  loop		\ n bytes
  stop_i2c
;

: .Boot		\ --
\ *G Put a "boot" message on the display.
  $78 $5c $5c $7c $37 5 0 >Disp
;

: DispStat	\ -- b
\ *G Read the display status, $80 if enabled.
  +ack_i2c				\ all bytes acknowledged
  start_i2c  $71 write_byte_i2c		\ SAA1064 address +r
  read_byte_i2c  stop_i2c
;

: AutoDisp	\ --
\ *G Initialise and display a boot message at start up.
  init_i2cx .Boot
;
' AutoDisp AtCold

[then]


\ ======
\ *> ###
\ ======

