\ SA1110 RTC driver

((
Copyright (c) 2003
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
fax: +44 (0)2380 339691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com
))

decimal

\ *! rtc1110
\ *T SA1110 RTC driver
\ ===============
\ *S Introduction
\ ===============
\ *P The file DRIVERS\RTC1110.FTH provides a seconds timer
\ ** using the on-chip RTC driven from 32.768kHz crystal.
\ ** Support is provided for the ANS word TIME&DATE. The
\ ** day of week format uses 0=Sunday which is compatible
\ ** with the file EXAMPLES\CALENDAR.FTH for date handling.

\ *P The code for this RTC was tested on a Hiden StrongBox.


\ ****************
\ *S Date handling
\ ****************
\ *P These routines are taken from the Julian calendar in
\ ** EXAMPLES\CALENDAR.FTH. They handle dates by converting
\ ** then to a single precision Julian day number. Day 0 is
\ ** 1 March 1900.

\ *P Other routines in EXAMPLES\CALENDAR.FTH can handle a
\ ** wider range of dates including historical ones.

\ *P Time and dates are displayed and entered as follows:
\ *D weekday: 0=Sun, 1=Mon ... 6=Sat
\ *D seconds: 0..59
\ *D minutes: 0..60
\ *D hours 0..23
\ *D days: 1..31
\ *D months: 1=Jan, 2=Feb ... 12=Dec
\ *D years: 1995..


#24 #60 * #60 * equ secs/day	\ -- 86400
\ *G Seconds per day.
#60 #60 * equ secs/hr		\ -- 3600
\ *G Seconds per hour
#60 equ secs/min		\ -- 60
\ *G Seconds per minute
#60 equ mins/hr			\ -- 60
\ *G Minutes per hour
#24 equ hrs/day			\ -- 24
\ *G Hours per day
#34698 equ uday0		\ -- 34698
\ *G The Julian day number for Unix day 0 at 1 Mar 1995.

compiler
: j>uday	\ jday -- uday
\ *G Convert a Julian day based at 0 for 1 March 1900 to a Unix
\ ** day number derived from LSECONDS.
  uday0 -
;

: u>jday	\ jday -- uday
\ *G Convert a Unix day number as used by LSECONDS to a Julian
\ ** day based at 0 for 1 March 1900.
  uday0 +
;
target

: date>         \ day month year -- jday ; see month codes
\ Given day month and year generate the day number. 
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if 12 + r> 1- >r then    	\ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

: >date         \ jday -- day month year
\ *G Convert a Julian day number to day month year format.
  0 1461 um/mod                         \ no leap-year periods
  4 * >r                                \ convert to no years
  365 /mod                              \ no years and days
  dup 4 = if 2drop 365 3 then           \ correct if Febr.29th
  r> + >r                               \ -- days | years
  123 + dup 5 153 */                    \ compute month
  dup >r 153 5 */ - r>                  \ compute day
  1- dup 12 > if 12 - r> 1+ >r then     \ correct month and year
  r> 1900 +                        	\ -- day month year
;

: weekday       \ jday -- n ; 0=Sunday
\ *G Given a Julian day return the day of week number where
\ ** 0 is Sunday, 1 is Monday and so on.
  4 + 0 7 um/mod drop
;

: >time		\ lsecs -- sec min hr
  secs/day mod				\ -- n ; strip date
  secs/min /mod  mins/hr /mod		\ -- secs min hr
;

: time>		\ sec min hr -- lsecs
  secs/hr *  swap secs/min * + +
;

[undefined] u/ [if]
: u/		\ u1 u2 --
  0 swap um/mod nip
;
[then]


\ ************************
\ *S Application interface
\ ************************

_rtc rcnr + constant lseconds	\ -- addr
\ *G Incremented once per second by the CPU hardware.

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set the RTC.
  date> j>uday secs/day *		\ -- sec min hr lsecs
  swap secs/hr * +
  swap secs/min * +
  +
  lseconds !
;

: time&date	\ -- sec min hr day mon yr ; 10.6.2.2292 
\ *G The ANS Forth word. Reads the RTC directly, and extracts
\ ** the time and date.
  lseconds @ dup >r  >time
  r> secs/day u/ u>jday >date
;

$7FFF equ init_rttr	\ -- n
\ *G The trim register initial value. See INIT_RTC. For
\ ** details of the trim register values see the Intel
\ ** SA1110 developer manual section 9.3.5.
\ ** The trim register value is by default
\ ** set to $7FFF by this equate which assumes a
\ ** perfect 32768Hz crystal. The exact value required
\ ** will need to be defined for each crystal. 

: init_RTC	\ --
\ *G Set up The RTC. INIT_RTC should only be executed after
\ ** power up (hardware) reset.
  _rtc
  3 over rtsr + !			\ clear AL and HZ, disable interrupts
  init_rttr swap rttr + !		\ set trim register to 32768Hz -1
;

: time		\ -- sec min hr
\ *G Return the last read time. Sec is the second
\ ** {0...59}, min is the minute {0...59}, hr is the hour.
  lseconds @ >time
;

: date		\ -- day mon yr
\ *G Return the last read date. Day is the day {1...31},
\ ** mon is the month {1...12}, and yr is the year (e.g., 1991).
  lseconds @ secs/day u/ u>jday >date
;

: dow		\ -- dow ; 0=Sunday
\ *G Return the day-of-week value last read from the RTC.
  lseconds @ secs/day u/ u>jday weekday
;


\ ****************
\ *S Time and date
\ ****************
\ *P These functions rely on the ANS Forth word TIME&DATE
\ ** ( -- s m h dd mm yyyy ) and the non-standard DOW
\ ** ( -- dow, 0=Sun) to get the day of the week.

\ *P This section of code is commented out by default.

((
create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow		\ --
\ *G Display day of week.
  dow 3 * days$ 1+ + 3 type
;

create months	\ -- addr
\ *G String containing 3 character text for the months.
  z", JanFebMarAprMayJunJulAugSepOctNovDec"

: .2r		\ n --
\ *G Display n as a two digit number with leading zeros.
  s>d <# # # #> type
;

: .4r		\ n --
\ *G Display n as a four digit number with leading zeros.
  s>d <# # # # # #> type
;

: .Time&Date	\ --
\ *G Display the system time
\ ** The format is:
\ *C   hh:mm:ss dd Mmm yyyy
  base @ decimal
  time&date				\ -- s m h dd mm yy
  dup #100 <				\ correct two digit year clocks
  if  #2000 +  endif
  rot .2r space				\ day - generate date
  swap 1- 3 * months + 3 type space	\ month
  .4r space				\ year
  .2r ." :"			 	\ hours - generate time 
  .2r ." :"				\ minutes
  .2r					\ seconds
  base !
;

: .zone		\ zone --
  if  3 spaces  else  ." GMT"  endif
;

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  .dow ." , " .time&date space .zone space
;
))


\ ======
\ *> ###
\ ======
