\ serBitBang.fth - Bit banged serial driver

((
Copyright (c) 2005
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20050710 GS_001 Changed RX start bit test.
))

only forth definitions

\ =============
\ *! serbitbang
\ =============
\ *T Bit-banged serial driver
\ *P The driver provides bit banged serial ports using
\ ** a timer running at four times the required baud rate.
\ ** All the real work is done in the timer interrupt routine.
\ ** Each serial port requires an input and an output line
\ ** on a GPIO port. Inverting buffers are required for RS232.
\ ** The serial format is no parity, 8 data bits 1 stop bit.

\ *P The default configuration uses the UART1 pins on an
\ ** LPC21xx system and Timer 1 for the timer interrupt.

\ *P The driver is based around a state machine
\ *E   State   function
\ **     0       idle
\ **     1       start bit 
\ **     2       bit 0
\ **     3       bit 1
\ **     ...
\ **     9       bit 7
\ **     10      stop bit
\ **     11      TX clean up or RX start bit error recovery

\ *P The states are defined so that the state counter is
\ ** incremented by every timer interrupt. If *\fo{S/C} below
\ ** is set to four for a 32 bit system, serial bit samples
\ ** are performed every four interrupts and the state number
\ ** serves as an index into separate receive and transmit
\ ** state tables. State 0 is the idle state, and the routine
\ ** for this is performed on every interrupt. All bit-banged
\ ** serial ports operate at the same baud rate.


4 equ s/c	\ -- n ; samples per serial clock bit.
\ *G Number of timer interrupts per bit time, must be four for
\ ** 32 bit systems. 

\ *P The following state numbers are defined. If two stop bits
\ ** are required, add a second stop bit state and insert handlers
\ ** in the receive and transmit despatch tables *\fo{RXexec} and
\ ** *\fo{TXexec} below.
 0 s/c * equ idling
 1 s/c * equ starting
 2 s/c * equ doBit0
 3 s/c * equ doBit1
 4 s/c * equ doBit2
 5 s/c * equ doBit3
 6 s/c * equ doBit4
 7 s/c * equ doBit5
 8 s/c * equ doBit6
 9 s/c * equ doBit7
10 s/c * equ stopbit
11 s/c * equ recovering

\ *P The serial port is defined by tables in Flash and RAM.
\ *E RAM
\ **     byte    TX state number*4
\ **     byte    RX state number*4
\ **     byte    TX data to transmit
\ **     byte    RX last data byte
\ **     cell    current TX data
\ **     cell    current RX data
\ **     byte    TX status
\ **     byte    RX status
\ **     byte    Reserved 1
\ **     byte    Reserved 2
\ ** 
\ ** Flash
\ **     cell    TX/RX port address
\ **     cell    TX bit mask - identifies TX bit in port
\ **     cell    RX bit mask - identifies RX bit in port
\ **     cell    address of RAM table

\ *P The testing was performed on an LPC2138 using the UART1 pins
\ ** as GPIO pins.

interpreter
: byte		1 field  ;
\ *G Define a one byte field in a data structure
: hword		2 field  ;
\ *G Define two byte field in a data structure
target

struct /SBBflash	\ -- len
\ *G The Fash data structure for a bit-banged serial port.
  int sbf.ioport	\ base address of I/O port peripheral
  int sbf.txmask	\ bit mask for transmit line
  int sbf.rxmask	\ bit mask for receive line
  int sbf.RAM		\ address of RAM data
end-struct

struct /SBBram	\ -- len
\ *G The RAM data structure for a bit-banged serial port.
  byte sbr.TXstate	\ TX state number*4
  byte sbr.RXstate	\ RX state number*4
  byte sbr.TXchar	\ TX data to transmit
  byte sbr.RXchar	\ RX last data byte
  int sbr.TXacc		\ current TX data, bit 0 = 1 = TX in progress
  int sbr.RXacc		\ current RX data, bit 0 = 1 = char available
  byte sbr.TXstatus	\ TX status
  byte sbr.Rxstatus	\ RX status
  aligned
end-struct

/SBBram buffer: SBBram0	\ -- addr
\ *G RAM buffer for first bit banged port.

create SBBport0		\ -- addr
\ *G Flash data for first bit banged port.
\ ** Modify this for your hardware.
  _GPIO ,		\ base address of GPIO unit 0
  bit8 ,		\ TxD1=P0.8
  bit9 ,		\ RxD1=P0.9
  SBBram0 ,		\ pointer to RAM table


\ *********************
\ *S SBB state machines
\ *********************

\ ===========
\ *N Receiver
\ ===========
\ *P In the idle state, the line is sampled on every timer
\ ** interrupt. As soon as the line goes low, the current
\ ** state is set so that the start bit is checked one half-bit
\ ** time later. If the start bit is still low, the data bits
\ ** and stop bit are sampled every full bit time later.
\ ** If the start bit is high, a noise spike is assumed and
\ ** idling is resumed. If the stop bit is low, the current
\ ** character is discarded, the receiver waits until the line
\ ** goes high and then resumes idling.

\ *P Because the receive line is only sampled four times per bit
\ ** in this system, rather than 16 times per bit as is common
\ ** for hardware UARTs, the bit-banged receivers are more
\ ** sensitive to baud rate inaccuracies than hardware UARTs.
\ ** You can sometimes obtain better results by adjusting the
\ ** start bit sampling point using the equate *\fo{preRXstart}
\ ** below.

s/c 2/ equ #preRX	\ -- n
\ *G Defines the number of interrupt ticks between detecting RX
\ ** low and checking the start bit again.

compiler

: getRXstate	\ flash -- state
\ *G Macro to get receive state.
  sbf.RAM @ sbr.RXstate c@
;

: setRXstate	\ state flash --
\ *G Macro to set receive state.
  sbf.RAM @ sbr.RXstate c!
;

: readRX	\ flash -- mask
\ *G Macro to read RX bit, giving non-zero if high.
\ ** Modify this for your hardware
  dup sbf.ioport @ IO0PIN + @		\ port address
  swap sbf.RXmask @ and			\ read it
;

target

IsLeaf : doRXIdle	\ flash -- flash
\ *G Idle state handler.
  dup readRX				\ read RX bit
  if  exit  endif			\ if high, still in idle
  starting #preRX - over setRXstate	\ start sampling at half bit point
;

IsLeaf : doRXstart	\ flash -- flash
\ *G Start bit handler.
  dup readRX 0= if			\ start bit must be low
    0 over sbf.RAM @ sbr.RXacc c!	\ RX accumulator = 0
    exit
  then
  idling over setRXstate		\ high, back to idle
;

IsLeaf : doRXBitN	\ flash -- flash
\ *G Data bit handler.
  dup sbf.RAM @ tuck sbr.RXacc c@	\ -- ram flash char
  1 rshift				\ shift previous value
  over readRX 0<> bit7 and or		\ -- ram flash char ; merge current bit
\ over readRX if  bit7 or  then		\ -- ram flash char ; merge current bit
  rot sbr.RXacc c!			\ save new value
;

IsLeaf : doRXstop	\ flash -- flash
\ *G Stop bit handler.
  dup readRX if
    dup sbf.RAM @
    dup sbr.RXacc c@			\ save character
    over sbr.RXchar c!
    bit0 over sbr.RXstatus c!		\ flag character available
    idling swap sbr.RXstate c!		\ wait for next character
    exit
  then
  recovering s/c - over setRXstate	\ wait for real stop bit
;

IsLeaf : doRXrecover	\ flash -- flash
\ *G Error recovery handler.
  dup readRX if				\ line idle if high
    idling over setRXstate
    exit
  then
  recovering s/c - over setRXstate	\ wait for real stop bit
;

create RXexec	\ -- addr
\ *G Holds a table of pointers to words for the RX state machine.
  ' doRXidle ,
  ' doRXstart ,
  ' doRXbitN ,
  ' doRXbitN ,
  ' doRXbitN ,
  ' doRXbitN ,
  ' doRXbitN ,
  ' doRXbitN ,
  ' doRXbitN ,
  ' doRXbitN ,
  ' doRXstop ,
  ' doRXrecover ,

: cycleRX	\ flash -- flash
\ *G During idle, the receive line is sampled on every tick.
\ ** When a start bit is detected, the start bit is sampled again
\ ** at the half bit time and the data and stop bits at subsequent
\ ** full bit times.
  dup getRXstate dup if			\ -- flash state ; not idling?
    1+ 2dup swap setRXstate		\ -- flash state' ; update state
    dup s/c 1- and 0= if
      RXexec + @ execute  exit
    endif
    drop  exit
  endif
  drop doRXidle
;


\ ==============
\ *N Transmitter
\ ==============
\ *P During idle, the transmit status bit is monitored on every
\ ** timer interrupt. If set the state machine copies the character
\ ** into a working buffer, clears the status bit and clocks out
\ ** data stream.

compiler
: getTXstate	\ flash -- state
\ *G Macro to get the transmit state.
  sbf.RAM @ sbr.TXstate c@
;

: setTXstate	\ state flash --
\ *G Macro to set the transmit state.
  sbf.RAM @ sbr.TXstate c!
;

: setTX		\ flash --
\ *G Macro to set the TX line high.
\ ** Modify this for your hardware.
  dup sbf.TXmask @			\ -- flash mask
  swap sbf.ioport @ IO0SET + !
;

: clrTX		\ flash --
\ *G Macro to set the TX line low.
\ ** Modify this for your hardware.
  dup sbf.TXmask @			\ -- flash mask
  swap sbf.ioport @ IO0CLR + !
;

target

IsLeaf : doTXidle	\ flash -- flash
\ *G Idle handler. Check for character to send, and start
\ ** the TX system if required.
  dup sbf.RAM @
  dup sbr.TXstatus c@ 1 and if		\ -- flash ram ; char to send
    dup sbr.TXchar c@ over sbr.TXacc c!	\ copy character
    0 over sbr.TXstatus c!		\ mark character accepted
    starting s/c 2/ - swap sbr.TXstate c! \ start output at half bit point
    exit				\ -- flash
  then
  drop					\ -- flash
;

IsLeaf : doTXstart	\ flash -- flash
\ *G Set start bit low.
  dup clrTX				\ start bit
;

IsLeaf : doTXbitN	\ flash -- flash
\ *G Issue the next data bit. The code is not pretty,
\ ** but minimises memory traffic.
  dup sbf.RAM @ sbr.TXacc c@ 1 and if	\ data bit
    dup setTX
  else
    dup clrTX
  endif
  dup sbf.RAM @
  dup sbr.TXacc c@			\ shift for next bit
  1 rshift
  swap sbr.TXacc c!
;

IsLeaf : doTXstop	\ flash -- flash
\ *G Issue a stop bit.
  dup setTX				\ stop bit
;

IsLeaf : doTXrecover	\ flash -- flash
\ *G A place-holder for future error recovery routines.
  idling over setTXstate		\ all done
;

create TXexec	\ -- addr
\ *G Holds a table of pointers to words for the TX state machine.
  ' doTXidle ,
  ' doTXstart ,
  ' doTXbitN ,
  ' doTXbitN ,
  ' doTXbitN ,
  ' doTXbitN ,
  ' doTXbitN ,
  ' doTXbitN ,
  ' doTXbitN ,
  ' doTXbitN ,
  ' doTXstop ,
  ' doTXrecover ,

: cycleTX	\ flash -- flash
\ *G During idle, the transmit status bit is monitored on every
\ ** timer interrupt. If set the state machine copies the character
\ ** into a working buffer, clears the status bit and clocks out
\ ** data stream.
  dup getTXstate dup if			\ -- flash state ; not idling?
    1+ 2dup swap setTXstate		\ -- flash state' ; update state
    dup s/c 1- and 0= if
      TXexec + @ execute  exit
    endif
    2drop  exit
  endif
  drop doTXidle
;


\ ****************************
\ *S Interrupt service routine
\ ****************************
\ *P This code is CPU specific and is for LPC21xx CPUs.

\ *P For the LPC210x, the default is Timer 1, Match
\ ** Register 0 being used to reset the count and generate an
\ ** interrupt. Note that if the PWM outputs are not being used,
\ ** the PWM unit can be used as it includes the same register
\ ** set as Timer 0 and Timer 1. The interrupt uses slot 4
\ ** in the Vectored Interrupt Controller.

\ *P This code requires the interrupt facilities in
\ ** *\i{ARM\INTARM3FTH}. If you need higher baud rates than 9600
\ ** baud, or need more than one port, or have other long interrupt
\ ** routines (such as with *\i{TimeBase.fth}) you should consider
\ ** moving the interrupt handler on to the FIQ line to reduce
\ ** interrupt latency and jitter.

_timer1 equ _sbbTimer	\ -- addr
\ *G Select the timer/counter unit to use for the ticker.

BIT2 equ sbbBit	\ -- mask
\ *G The bit to be set in PCONP to power the counter/timer.

t1int# equ sbbInt#	\ -- src
\ *G select the source number of the ticker interrupt.

#4 equ sbbSlot#	\ -- n
\ *G Select the VIC slot/priority for the ticker interrupt.
\ ** This is usually a high priority interrupt to avoid phase
\ ** jitter.

1 equ /preSbb	\ -- n
\ *G Select the prescaler division ratio.
/preSbb 1- equ init_preDivSbb
\ *G The initial value of the prescaler divider.

decimal

/vpb /preSbb * s/c * equ sbbDiv	\ -- n
\ *G A scale factor used for setting the timer.

interpreter
: gen-sbb	\ bps-rate clock -- divisor16
\ *G An *\fo{INTERPRETER} definition that takes the required
\ ** serial comms rate (in BPS) and the clock speed in herz
\ ** returning the required timer match register initialisation
\ ** value.
  swap /				\ ratio
  sbbDiv /mod				\ scaled
  swap sbbDiv 2/ >=
  if  1+  endif
;
target

[undefined] sbbBaud0 [if]
  #9600 equ sbbBaud0
\ *G If not previously defined, 9600 baud will be used.
[then]

sbbBaud0 system-speed gen-sbb equ #MRsbb0
\ *G The initial value of the match register calculated from the
\ ** required baud rate and the system clock speed *\fo{SYSTEM-SPEED}
\ ** in herz.

: SBBisr	\ --
\ *G The high level interrupt handler for bit-banged serial lines.
  1 _sbbTimer TnIR + !			\ reset MR0 interrupt
  SBBport0 cycleRX cycleTX drop		\ cycle state machines for SBB0
;

' SBBisr IRQ: SBB-irq	\ -- addr
\ *G Creates the IRQ handler for the serial timer.


\ *****************
\ *S Initialisation
\ *****************
\ *P This code is CPU specific and is for LPC21xx CPUs.


: +SBBclock	\ -- ; start clock interrupt
\ *G Start the serial ticker periodic interrupt.
\ ** Modify this for your hardware
  di
  sbbBit _PCONP or!			\ power up timer block
  _sbbTimer				\ base address of Timer
  0 over TnTCR + !			\ disable counter
  init_preDivSbb over TnPC + !		\ set prescaler
  #MRsbb0 over TnMR0 + !		\ set match register
  3 over TnMCR + !			\ interrupt and reset on MR0
  SBB-irq sbbInt# sbbSlot# SetIrqIsr	\ set up and enable interrupt
  1 swap TnTCR + !			\ enable counter
  ei
;

: -SBBclock	\ -- ; stop clock interrupt
\ *G Stop the ticker periodic interrupt.
\ ** Modify this for your hardware
  sbbInt# DisInt
;

: initSbb0	\ --
\ *G Initialise first bit-banged serial port.
\ ** Modify this for your hardware
  di
  _PINSEL0 dup @			\ enable GPIO pins
    $000F0000 invert and		\ clear bits 16..19 for GPIO use  
  swap !
  SBBport0				\ set up IO ports
     dup sbf.ioport @ tuck IO0DIR + @	\ -- ioport flash IODIR
     over sbf.TXmask @ or		\ set TX bit for o/p
     over sbf.RXmask @ invert and	\ clear RX bit for input
     rot IO0DIR + !			\ set new direction mask
  sbf.RAM @ /SBBram erase		\ wipe buffer
  ei
;

: initSBBs	\ --
\ *G Initialise bit-banged serial port(s). This word is
\ ** run at cold start.
  initSbb0  +SBBclock
;
' initSBBs AtCold


\ ********************
\ *S Serial primitives
\ ********************

: (sbbEmit)	\ char flash --
\ *G Transmit a character on UART.
  sbf.RAM @
  begin					\ wait for TX ready
    dup sbr.TXstatus c@ 1 and 0=
  until
  tuck sbr.TXchar c!			\ set character
  1 swap sbr.TXstatus c!  
;

: (sbbkey?)	\ flash -- t/f
\ *G Return true if the UART has a character available.
  sbf.RAM @ sbr.RXstatus c@ 1 and 0<>
;

tasking? [if]
: (sbbKey)	\ flash -- char
\ *G Wait for a character on the UART and return it.
  sbf.RAM @ 
  begin
    dup sbr.RXstatus c@ 1 and 0=
   while
    pause
  repeat
  dup sbr.RXchar c@
  0 rot sbr.RXstatus c!
;
[else]
: (sbbKey)	\ flash -- char
  sbf.RAM @ 
  begin
    dup sbr.RXstatus c@ 1 and
  until
  dup sbr.RXchar c@
  0 rot sbr.RXstatus c!
;
[then]

\ external

: (sbbType)	\ c-addr len flash --
\ *G Type a string to the UART.
  -rot  bounds
  ?do  i c@ over (sbbEmit)  loop
  drop
;

: (sbbCR)	\ flash --
\ *G Issue a CR/LF pair to the UART.
  $0D over (sbbEmit)  $0A swap (sbbEmit)
;

: sbbKey?0	( -- flag )	SBBport0 (sbbKey?)  ;
\ *\fo{KEY?} for bit-banged port 0.
: sbbKey0	( -- char )	SBBport0 (sbbKey)  ;
\ *\fo{KEY} for bit-banged port 0.
: sbbEmit0	( char -- )	SBBport0 (sbbEmit)  ;
\ *\fo{EMIT} for bit-banged port 0.
: sbbType0	( addr n -- )	SBBport0 (sbbType)  ;
\ *\fo{TYPE} for bit-banged port 0.
: sbbCR0	( -- )		SBBport0 (sbbCR)  ;
\ *\fo{CR} for bit-banged port 0.


\ **************************
\ *S Generic i/o assignments
\ **************************

cdata
create Sbb0Console	\ -- addr ;
\ *G Generic I/O device for the first bit-banged UART. *\fo{OUT}
\ ** is managed by the upper level driver in KERNEL62.FTH.
  ' sbbKey0 ,		\ -- char ; receive char
  ' sbbKey?0 ,		\ -- flag ; check receive char
  ' sbbEmit0 ,		\ -- char ; display char
  ' sbbType0 ,		\ caddr len -- ; display string
  ' sbbCr0 ,		\ -- ; display new line

console-port 10 = [if]
  Sbb0Console constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for
\ ** interaction. It may be changed by a phrase of the form:
\ *C   <device>  dup opvec !  ipvec !
[then]


\ ======
\ *> ###
\ ======

decimal
