\ SerLPC23xxqi.fth - LPC23xx queued interrupt RX and TX UART driver

((
Copyright (c) 2008
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20081010 SFP002 Corrected and refactored initialisations.
20081007 MPE001 First release
))

only forth definitions


\ =============
\ *! serlpc23xxqi
\ *T LPC23xx serial driver
\ =============
\ *P This driver provides queued interrupt-driven serial comms for LPC23xx CPUs
\ ** using the internal UARTs. Because there are alternate pin
\ ** assignments for some UARTs, you *\b{must} check that the
\ ** supplied code pin setup is correct for your board.


\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUART0? [if]
1 equ useUART0?	\ -- n
\ *G Compile UART0 code if non-zero.
[then]

[undefined] useUART1? [if]
0 equ useUART1?	\ -- n
\ *G Compile UART1 code if non-zero.
[then]

[undefined] useUART2? [if]
0 equ useUART2?	\ -- n
\ *G Compile UART2 code if non-zero.
[then]

[undefined] useUART3? [if]
0 equ useUART3?	\ -- n
\ *G Compile UART3 code if non-zero.
[then]


\ ********
\ *S Tools
\ ********

[undefined] setMask [if]
: setMask	\ value mask addr --
\ *G Clear the *\i{mask} bits at *\i{addr} and set (or) the
\ ** bits defined by *\i{value}.
  tuck @				\ -- value addr mask x
  swap invert and			\ -- value addr x'
  rot or				\ -- addr x''
  swap !
;
[then]


\ *P *\fo{GenBaudRate} uses the system clock rate (in Hz), the
\ ** perpheral clock divider ration, and the required serial speed
\ ** (in BPS) and returns the 16 bit divider value for the baud
\ ** rate generator
\ *E              system-speed
\ ** divisor = -------------------
\ **           16 * pcd * baudrate

interpreter
: GenBaudRate	\ bps-rate pcd -- divisor16
\ *G Given a baud rate and the peripheral clock divider
\ ** (1, 2, 4, 8), calculate the 16 bit baud rate divisor.
\ **
  system-speed swap / swap / #16 /mod
  swap 8 >=
  if  1+  endif
;
target

internal
: >RxQ		\ char queue --
\ *G Put character from UART into queue.
\ ** INTERNAL.
  dup (cqFull?)				\ if queue full
  if  2drop  exit  endif		\  discard character
  (>cqueue)				\  queue character
;

: FIFO>RxQ      \ base queue --
\ *G Given a UART base address and a character queue,
\ ** empty the UART Rx FIFO into the queue.
\ ** INTERNAL.
  begin
    over UnLSR + @ $01 and
   while
    over UnRBR + @ $FF and over >RxQ
  repeat
  2drop
;

: TxQ>FIFO	\ base queue
\ *G Given a UART base address and a character queue,
\ ** put transmit characters into the UART Tx FIFO.
  dup cqEmpty? if
    drop
    0 swap UnSCR + !			\ mark empty for EMIT routine
    exit
  endif
  (cqueue>) swap UnTHR + !		\ write character
;
external


\ ********************
\ *S Serial primitives
\ ********************

internal

0 value FaultConsole?	\ -- flag
\ *G When set non-zero, transmit is in polled mode. This is
\ ** required when fault exceptions provide debug information.

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
  -1 to FaultConsole?
  [ tasking? ] [if]  single  [then]
;

: (qseremit)	\ char base queue --
\ *G Send a character on the UART whose base address and queue are given.
  FaultConsole? 0= if
    begin				\ wait for non-full queue
      dup cqfull?
     while
      pause
    repeat
    [i
    over UnSCR + @ $FF and if		\ if TX in progress
      nip (>cqueue)  i]			\ just stash the character
      exit
    then
    drop				\ -- char base
    tuck UnTHR + !			\ write character
    $FF swap UnSCR + !			\ mark TX int started
    i]  exit
  endif
  drop
  begin
    dup UnLSR + c@ %00100000 and
  until
  UnTHR + c!
;

: (qsertype)	\ caddr len base queue --
\ *G Send a string on the UART whose base address and queue are given.
  2swap bounds				\ -- base queue
  ?do  2dup i c@ -rot (qseremit)  loop
  2drop
;

: (qsercr)	\ base queue --
\ *G Send a CR/LF on the UART whose base address and queue are given.
  2dup $0D -rot (qseremit)  $0A -rot (qseremit)
;

external

useUART0? [if]
1 equ pcdU0	\ -- n
\ *G Peripheral clock division ratio for UART0.
console0-speed pcdU0 GenBaudRate equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.
pcdu0 PCLKdiv 6 lshift equ /pcdu0	\ -- mask
\ *G Bit mask for UART0 peripheral clock divider.

#128 cqueue: SerInpQ0	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ0	\ -- addr
\ *G Transmitter circular byte queue.
: ser0-isrh	\ --
\ *G UART0 high level ISR.
\ ** INTERNAL.
  case  _UART0 dup UnIIR + @ 7 and
\   0 of  UnMSR + @ drop  endof		\ No MSR for UART0
    2 of  SerOutQ0 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ0 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser0-isrh IRQ: ser0-isr
\ *G UART0 interrupt service routine entry point.

l: uart0_tab	\ -- addr ; initialisation table
  _UART0 UnFCR + ,	$81 ,			\ FIFO enable (8 char delay)
  _UART0 UnLCR + ,	$03 $80 or ,        	\ divisor access via DLAB
  u0-divisor dup				\ set divisor
  _UART0 UnDLM + ,	8 rshift $FF and ,	\ high
  _UART0 UnDLL + ,	$FF and ,		\ low
  _UART0 UnLCR + ,	$03 ,			\ disable DLAB access
  _UART0 UnIER + ,	$03 ,			\ enable Tx and RX interrupts
  _UART0 UnSCR + ,	$00 ,			\ SCR used as interrupt flag
  0 ,

: init-ser0	\ --
\ *G Initialise serial port 0
  di
  SerInpQ0 init-cqueue  SerOutQ0 init-cqueue	\ initialise character queues
  bit3 _PCONP or!				\ enable clock & pwr
  /pcdU0 $00C0 _SYSCON PCLKSEL0 + setMask 	\ set peripheral clock speed
  $0050 $00F0 _PINSEL0 setMask			\ PINSEL0 bits 7:4
  uart0_tab init-io				\ use table
  ser0-isr U0Int# #14 SetIrqIsr			\ set up interrupt
  ei
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 SerOutQ0 (qserEmit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 SerOutQ0 (qsertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 SerOutQ0 (qsercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  SerInpQ0 cqnotempty?  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  SerInpQ0 cqueue>  ;
create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]
[then]

useUART1? [if]
1 equ pcdU1	\ -- n
\ *G Peripheral clock division ratio for UART.
console1-speed pcdU1 GenBaudRate equ u1-divisor	\ -- n
\ *G UART baud rate divisor.
pcdu1 PCLKdiv 8 lshift equ /pcdu1	\ -- mask
\ *G Bit mask for UART peripheral clock divider.

#128 cqueue: SerInpQ1	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ1	\ -- addr
\ *G Transmitter circular byte queue.
: ser1-isrh	\ --
\ *G UART1 high level ISR.
\ ** INTERNAL.
  case  _UART1 dup UnIIR + @ 7 and
    0 of  U1MSR + @ drop  endof		\ Modem Status interrupt
    2 of  SerOutQ1 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ1 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser1-isrh IRQ: ser1-isr
\ *G UART1 interrupt service routine entry point.

l: uart1_tab	\ -- addr ; initialisation table
  _UART1 UnFCR + ,	$81 ,			\ FIFO enable (8 char delay)
  _UART1 UnLCR + ,	$03 $80 or ,        	\ divisor access via DLAB
  u1-divisor dup				\ set divisor
  _UART1 UnDLM + ,	8 rshift $FF and ,	\ high
  _UART1 UnDLL + ,	$FF and ,		\ low
  _UART1 UnLCR + ,	$03 ,			\ disable DLAB access
  _UART1 UnIER + ,	$03 ,			\ enable Tx and RX interrupts
  _UART1 UnSCR + ,	$00 ,			\ SCR used as interrupt flag
  0 ,

: init-ser1	\ --
\ *G Initialise serial port 1.
  di
  SerInpQ1 init-cqueue  SerOutQ1 init-cqueue	\ initialise character queues
  ~pcUART1 _PCONP or!				\ enable clock & pwr
  /pcdU1 $0300 _SYSCON PCLKSEL0 + setMask 	\ set peripheral clock speed
  $000A $000F _PINSEL4 setMask			\ RXD1, TXD1, mode 2
  uart1_tab init-io				\ use table
  ser1-isr U1Int# #14 SetIrqIsr			\ set up interrupt
  ei
;
: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 SerOutQ1 (qserEmit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 SerOutQ1 (qsertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 SerOutQ1 (qsercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  SerInpQ1 cqnotempty?  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  SerInpQ1 cqueue>  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]
[then]

useUART2? [if]
1 equ pcdU2	\ -- n
\ *G Peripheral clock division ratio for UART.
console2-speed pcdU2 GenBaudRate equ u2-divisor	\ -- n
\ *G UART baud rate divisor.
pcdu2 PCLKdiv #16 lshift equ /pcdu2	\ -- mask
\ *G Bit mask for UART peripheral clock divider.

#128 cqueue: SerInpQ2	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ2	\ -- addr
\ *G Transmitter circular byte queue.
: ser2-isrh	\ --
\ *G UART2 high level ISR.
\ ** INTERNAL.
  case  _UART2 dup UnIIR + @ 7 and
\    0 of  U1MSR + @ drop  endof		\ Modem Status interrupt
    2 of  SerOutQ2 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ2 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser2-isrh IRQ: ser2-isr
\ *G UART2 interrupt service routine entry point.

l: uart2_tab	\ -- addr ; initialisation table
  _UART2 UnFCR + ,	$81 ,			\ FIFO enable (8 char delay)
  _UART2 UnLCR + ,	$03 $80 or ,        	\ divisor access via DLAB
  u2-divisor dup				\ set divisor
  _UART2 UnDLM + ,	8 rshift $FF and ,	\ high
  _UART2 UnDLL + ,	$FF and ,		\ low
  _UART2 UnLCR + ,	$03 ,			\ disable DLAB access
  _UART2 UnIER + ,	$03 ,			\ enable Tx and RX interrupts
  _UART2 UnSCR + ,	$00 ,			\ SCR used as interrupt flag
  0 ,

: init-ser2	\ --
\ *G Initialise serial port 2.
  di
  SerInpQ2 init-cqueue  SerOutQ2 init-cqueue	\ initialise character queues
  ~pcUART2 _PCONP or!				\ enable clock & pwr
  /pcdU2 $30000 _SYSCON PCLKSEL1 + setMask	\ set peripheral clock speed
  $0050:0000 $00F0:0000 _PINSEL0 setMask	\ PINSEL0 bits 23:20
  uart2_tab init-io				\ use table
  ser2-isr U2Int# #14 SetIrqIsr			\ set up interrupt
  ei
;
: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 SerOutQ2 (qserEmit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 SerOutQ2 (qsertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 SerOutQ2 (qsercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  SerInpQ2 cqnotempty?  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  SerInpQ2 cqueue>  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]
[then]

useUART3? [if]
1 equ pcdU3	\ -- n
\ *G Peripheral clock division ratio for UART.
console3-speed pcdU3 GenBaudRate equ u3-divisor	\ -- n
\ *G UART baud rate divisor.
pcdu3 PCLKdiv #18 lshift equ /pcdu3	\ -- mask
\ *G Bit mask for UART peripheral clock divider.

#128 cqueue: SerInpQ3	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ3	\ -- addr
\ *G Transmitter circular byte queue.
: ser3-isrh	\ --
\ *G UART3 high level ISR.
\ ** INTERNAL.
  case  _UART3 dup UnIIR + @ 7 and
\    0 of  U1MSR + @ drop  endof		\ Modem Status interrupt
    2 of  SerOutQ3 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ3 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser3-isrh IRQ: ser3-isr
\ *G UART3 interrupt service routine entry point.

l: uart3_tab	\ -- addr ; initialisation table
  _UART3 UnFCR + ,	$81 ,			\ FIFO enable (8 char delay)
  _UART3 UnLCR + ,	$03 $80 or ,        	\ divisor access via DLAB
  u3-divisor dup				\ set divisor
  _UART3 UnDLM + ,	8 rshift $FF and ,	\ high
  _UART3 UnDLL + ,	$FF and ,		\ low
  _UART3 UnLCR + ,	$03 ,			\ disable DLAB access
  _UART3 UnIER + ,	$03 ,			\ enable Tx and RX interrupts
  _UART3 UnSCR + ,	$00 ,			\ SCR used as interrupt flag
  0 ,

: init-ser3	\ --
\ *G Initialise serial port 3.
  di
  SerInpQ3 init-cqueue  SerOutQ3 init-cqueue	\ initialise character queues
  ~pcUART3 _PCONP or!				\ enable clock & pwr
  /pcdU3 $C0000 _SYSCON PCLKSEL1 + setMask	\ set peripheral clock speed
  $003C:0000 $003C:0000 _PINSEL1 setMask	\ PINSEL1 bits 21:18, mode 3
  uart3_tab init-io				\ use table
  ser3-isr U3Int# #14 SetIrqIsr			\ set up interrupt
  ei
;
: seremit3	\ char --
\ *G Transmit a character on UART3.
  _UART3 SerOutQ3 (qserEmit)  ;
: sertype3	\ c-addr len --
\ *G Transmit a string on UART3.
  _UART3 SerOutQ3 (qsertype)  ;
: sercr3	\ --
\ *G Issue a CR/LF pair to UART3.
  _UART3 SerOutQ3 (qsercr)  ;
: serkey?3	\ -- flag
\ *G Return true if UART3 has a character available.
  SerInpQ3 cqnotempty?  ;
: serkey3	\ -- char
\ *G Wait for a character on UART3.
  SerInpQ3 cqueue>  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART3.
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line

console-port 3 = [if]
  console3 constant console
[then]
[then]


: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
  [ useUART3? ] [if]  init-ser3  [then]
;



\ *********
\ Test code
\ *********

((
: txu1		\ --
  500 0 do  bl seremit1  loop  ;

: txu2		\ --
  500 0 do  bl seremit2  loop  ;

: txu3		\ --
  500 0 do  bl seremit3  loop  ;
))

\ *******
\ Perhaps
\ *******

((
: SetBaud	\ bps-rate uart --
 \ *G Takes the required serial comms rate (in BPS) and the UART
 \ ** base and sets the baud rate.
  swap system-speed /vpb / swap / #16 /mod
  swap 8 >=
  if  1+  endif
  dup 8 rshift $FF and  swap $FF and swap	\ -- uart divlo divhi
  [i
    rot						\ -- divlo divhi uart
    $83 over UnLCR + !				\ enable DLAB access
    tuck UnDLM + !				\ -- divlo uart ; high divisor ; MPE003
    tuck UnDLL + !				\ -- uart ; MPE003
    $03 swap UnLCR + !
  i]
;
))


\ ======
\ *> ###
\ ======

decimal
