\ MPE Character Queue Support

((
Released for use with the MPE Forth Cross Compiler by:
and copyright (c) 2001

MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com

To do
=====


Change history
==============
20160316 BJC004 Added alignment to CQUEUE:.
20090811 SFP003 Added extended functions.
20041109 SFP003 Restored compiler directives.
20030305 SFP002 Added INIT-HCQUEUE for dynamically allocated
		queues.
20021009 SFP001 Refactored to enhance interrupt locking in
		interrupt driven drivers.
))


\ ==========
\ *! cqueues
\ *T Character Queues
\ ==========
\ *P The file *\i{Common\Cqueues.fth} provides circular character
\ ** (byte) queues. If the equate *\fo{TASKING?} is non-zero, the
\ ** blocking routines will use *\fo{PAUSE}. Interrupts are
\ ** disabled for the queue empty/full checks.


\ ***********************
\ *S Queue data structure
\ ***********************

struct /cqueue	\ -- size ; character queue structure in idata, buffer in udata
\ *G Circular queue data structure.
\ *[
  int >qhead		\ Offset of head, where characters are taken from
  int >qtail		\ Offset of tail, where characters are put
  int >qchars		\ Number of characters in the queue
  int >qmask		\ Mask to apply to pointers
  ptr >qbuffer		\ Base address of character buffer
end-struct
\ *]

interpreter
: cqueue:	\ size -- ; -- cqueue ; size CQUEUE: <name>
\ *G An interpreter definition to build a character queue of the
\ ** specified size. The queue data structure is built in the current
\ ** *\fo{IDATA} space, and the buffer itself is in the current
\ ** *\fo{UDATA} space. Executing *\fo{<name>} returns the address
\ ** of the queue data structure.
\ ** N.B. The size of a queue must be a power of two, e.g. 32, 64 ...
  ialign ualign
  ihere  /cqueue iallot		\ -- size cqueue
  dup >qhead off
  dup >qtail off
  dup >qchars off
  over 1- over >qmask !
  uhere over >qbuffer !
  constant uallot
;
target

: init-cqueue	\ cqueue -- ; initialise queue
\ *G Initialise the specified queue created by *\fo{CQUEUE:}.
  dup >qhead off
  dup >qtail off
  >qchars off
;

sizeofheap [if]
: init-hcqueue	\ size cqueue -- ; SFP002
\ *G Initialise the specified queue created by *\fo{ALLOCATE}.
\ ** When a queue is allocated from the heap by a phrase
\ ** of the form *\fo{/CQUEUE <size> + ALLOCATE}, this word
\ ** must be used.
  swap 1- over >qmask !			\ set queue size
  dup  dup /cqueue + swap >qbuffer !	\ set buffer pointer
  init-cqueue				\ perform normal initialisation
;
[then]


\ *******************
\ *S Queue primitives
\ *******************

: (>cqueue)	\ char cqueue -- ; put character on cqueue
\ *G Put char into the queue with no checks.
  dup >qbuffer @ over >qtail @ +	\ -- char queue q*
  rot swap c! 				\ -- queue
  dup >qchars incr
  dup >qtail @ 1+
  over >qmask @ and
  swap >qtail !
;

: (cqueue>)	\ cqueue -- char ; get next character from queue
\ *G Get the next character from the queue with no checks.
  dup >qbuffer @ over >qhead @ +
  c@ swap				\ -- char queue
  dup >qchars decr
  dup >qhead @ 1+
  over >qmask @ and
  swap >qhead !
;

: (cqfull?)	\ queue -- flag ; TRUE if queue full
\ *G Return true if the queue is full. No interrupt
\ ** protection is provided.
  dup >qchars @  swap >qmask @ u>
;
compiler
: (cqfull?)	\ queue -- flag ; TRUE if queue full
  dup >qchars @  swap >qmask @ u>
;
target

: cqfull?	\ queue -- flag ; TRUE if queue full
\ *G Return true if the queue is full. Interrupt
\ ** protection is provided.
  [i  dup >qchars @  swap >qmask @ u>  i]
;

: cqchars	\ queue -- n
\ *G Return the number of characters in the queue.
  >qchars @
;

: cqempty?	\ queue -- flag ; TRUE if queue empty
\ *G Return true if the queue is empty.
  >qchars @ 0=
;

: cqnotempty?	\ queue -- flag ; TRUE if queue not empty
\ *G Return true if the queue is not empty, i.e. if it contains any
\ ** characters.
  >qchars @ 0<>
;

: cqRoom	\ queue -- u
\ *G Return the number of free bytes in the queue.
  dup >qmask @ 1+  swap >qchars @ -
;

: cqroom?	\ +n queue -- flag
\ *G Return true if there is enough room in the queue for +n more
\ ** characters.
  dup >qmask @ 1+  swap >qchars @ -  <=
;

compiler
: cqchars	\ queue -- n
  >qchars @
;

: cqempty?	\ queue -- flag ; TRUE if queue empty
  >qchars @ 0=
;

: cqnotempty?	\ queue -- flag ; TRUE if queue not empty
  >qchars @ 0<>
;
target

: >cqueue	\ char cqueue -- ; spins if full
\ *G Put a character into the queue. If the queue is full, the
\ ** system waits (blocks) until there is enough space.
  begin
    [I dup (cqfull?)
   while
    I]
[ Tasking? ] [IF] pause [THEN]
  repeat
  (>cqueue) I]
;

: cqueue>	\ queue -- char ; spins while queue empty
\ *G Remove the next character, waiting if the queue is empty.
  begin
    [I dup cqempty?
   while
    I]
[ Tasking? ] [IF] pause [THEN]
  repeat
  (cqueue>) I]
;


\ ****************************
\ *S Extended queue operations
\ ****************************
\ *P When queues are filled and emptied in blocks rather than
\ ** character by character, faster operation can be achieved
\ ** using these words. You may see additional benefit using the
\ ** fast version of *\fo{CMOVE}.

\ *P The extended operations are compiled if the equate
\ ** *\fo{extCQ?} is set non-zero before *\i{Cqueues.fth}
\ ** is compiled.

[undefined] extCQ? [if]
0 equ extCQ?	\ -- flag
\ *G If set non-zero here or before *\i{Cqueues.fth} is compiled,
\ ** the extended wordset below will be compiled.
[then]

extCQ? [if]


\ =============
\ *N Primitives
\ =============

: /tailw		\ queue -- len
\ *G Number of bytes from the tail to the end.
  dup >qmask @ 1+ swap >qtail @ -
;

: /gapw		\ queue -- len
\ *G Number of bytes from the tail to the head, assuming no wrap.
  dup cqchars if
    dup >qhead @ swap >qtail @ -
  else
    >qmask @ 1+
  endif
;

: +tail		\ n queue --
\ *G Add n characters to the tail pointer and count.
  [i
  2dup >qchars +!
  tuck >qtail @ +  over >qmask @ and  swap >qtail !
  i]
;

: $>cq		\ caddr len n queue -- caddr' len' queue
\ *G Copy *\i{n} bytes of the the source block into the buffer.
  >r  2 pick				\ src
  r@ dup >qbuffer @ swap >qtail @ +	\ dest
  2 pick cmove				\ copy n bytes
  dup r@ +tail				\ update queue
  /string				\ update string
  r>
;

: $>tail	\ caddr len queue -- caddr' len' queue
\ *G Add block in space between tail and end
  over
  if  2dup /tailw min swap $>cq  endif
;

: $>gap		\ caddr len queue -- caddr' len' queue
\ *G Add block in space between tail and head.
  over
  if  2dup /gapw min swap $>cq  endif
;

compiler
: /headr	\ queue -- len
\ *G Number of bytes from the head to the end.
  dup >qmask @ 1+ swap >qhead @ -
;

: /gapr		\ queue -- len
\ *G Number of bytes from the head to the tail, assuming no wrap.
\  dup >qtail @ swap >qhead @ -
  >qchars @
;
target

: -head		\ n queue --
\ *G Remove n characters from the head pointer and count.
  [i
  over negate over >qchars +!
  tuck >qhead @ +  over >qmask @ and  swap >qhead !
  i]
;

: cq>$		\ caddr len n queue -- caddr' len' queue
\ *G Copy *\i{n} bytes from the queue into the buffer
\ ** *\i{caddr/len}.
  dup >r
  dup >qbuffer @ swap >qhead @ +	\ src
  3 pick				\ dest
  2 pick cmove				\ copy n bytes
  dup r@ -head				\ update queue
  /string				\ update string
  r>
;

: head>$	\ caddr len queue -- caddr' len' queue
\ *G Copy queue block in space between head and end
  over
  if  2dup /headr min swap cq>$  endif
;

: gap>$		\ caddr len queue -- caddr' len' queue
\ *G Add block in space between head and tail.
  over
  if  2dup /gapr min swap cq>$  endif
;


\ =============
\ *N User words
\ =============

: cqWrite	\ caddr len queue --
\ *G Write the given string to the queue.
\ ** This word blocks until the operation is complete.
  over if
    begin
      dup >qtail @ over >qhead @ u< not
      if  $>tail  then
      $>gap  over
     while
      pause
    repeat
  endif
  3drop
;

: cqStuff	\ caddr len queue -- n
\ *G Write as much data from the given string to the queue as
\ ** there is room for. Return the number of bytes written.
  tuck cqRoom min swap cqWrite
;

: cqRead	\ caddr len queue --
\ *G Read the given string from the queue.
\ ** This word blocks until the operation is complete.
  over if
    begin
      dup >qtail @ over >qhead @ u> not
      if  head>$  then
      gap>$  over
     while
      pause
    repeat
  endif
  3drop
;

: cqExtract	\ caddr len queue -- n
\ *G Read data from the queue to the buffer. The amount read
\ ** is limited by the size of the buffer and the number of
\ ** bytes in the queue. The number of bytes read is returned.
  tuck cqchars min swap cqRead
;

0 [if]		\ test code

create $src
  ", abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"

64 cqueue: testme

68 buffer: $dest

: wipedest	\ --
  $dest 68 erase  ;

: .testme	\ --
  cr testme cqchars . ." in TESTME"
  ." , head=" testme >qhead ?
  ." tail=" testme >qtail ?
  ." chars=" testme >qchars ?
  testme >qbuffer @ 64 dump
;

: testit	\ --
  wipedest
  $src count testme cqWrite		\ 52 chars in testme
  .testme
  $dest 30 testme cqRead		\ 22 chars in testme, 30 in dest
  .testme  $dest 30 dump
  $src 1+ 42 testme cqWrite		\ 64 characters in testme
  .testme
  $dest 64 testme cqRead		\ 0 in testme, 64 in dest
  .testme $dest 64 dump
  testme cqchars abort" TESTME not empty"
;

[then]		\ test code

[then]		\ extCQ?


\ ======
\ *> ###
\ ======

