\ FracMath.fth - Fractional arithmetic.

((
Copyright (c) 2008
Gary Bergstrom
  and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *! fracmath
\ *T Fractional Arithmetic
\ ===========
\ *P Fractional arithmetic uses integer arithmetic to represent
\ ** a range of numbers. In this case, we use 2 bits before the
\ ** fractional point and 14/30 after it. This is called 2.14 format.
\ ** Allowing for signs, this allows us to represent numbers in
\ ** in the range -1.999..+1.999, which is ideal for trigonometric
\ ** functions without having to use floating point.

\ *P By Forth tradition, the operators that handle fractional
\ ** arithmetic are named with a '.' at the end.

\ *P If your CPU has no divide instruction or division is slow,
\ ** performance of the multiply operation can usually be improved
\ ** by using shifts for the final scaling.

32bit? [if] 32 [else] 16 [then] equ #bits	\ -- u
\ *G Number of bits in a cell.
#2 equ #before			\ -- u
\ *G Number of bits before the fractional point.
#bits #before - equ #after	\ -- u
\ *G Number of bits after the fractional point.
1 #after lshift equ +1.		\ -- $4000
\ *G The fractional arithmetic equivalent of 1.
#after #301 #1000 */ equ #digs	\ -- u
\ *G Number of decimal digits after the point. The ratio 1000:301
\ ** is an approximation for ln2(10).

interpreter
: 10^n		\ n -- 10^n
  1  swap 0 ?do  #10 * loop
;
target
#digs 10^n equ frscale	\ --
\ *G Scale factor used for display.

+1. constant +1.	\ -- fr=1
\ Constant for debugging.
: *.		\ fr1 fr2 -- fr3
\ *G Fractional multiply.
  +1. */  ;				\ scaled
: /.		\ fr1 fr2 -- fr3
\ *G Fractional divide.
  +1. swap */  ;
: +.		\ fr1 fr2 -- fr3
\ *G Fractional add. A synonym for *\fo{+}.
  +  ;
: -.		\ fr1 fr2 -- fr3
\ *G Fractional subtract. A synonym for *\fo{-}.
  -  ;
compiler
: +.		\ fr1 fr2 -- fr3
  +  ;
: -.		\ fr1 fr2 -- fr3
  -  ;
target

: frac#s	\ d1 -- d2
\ *G Create the number of digits specified by *\fo{#digs}.
  #digs 0 do  #  loop  ;
: (.f)		\ fr -- caddr len
\ *G Convert *\i{fr} to an ASCII decimal string with four digits after
\ ** the decimal point.
  base @ swap decimal
  frscale *.  DUP ABS 0 <#		\ number is scaled
    frac#s [CHAR] . HOLD #S ROT SIGN
  #> rot base !
;
: .F		\ fr --
\ *G Convert a fractional number and display it in decimal.
  (.f) TYPE SPACE  ;
: D>FR		\ d -- fr
\ *G Convert a double number to a fraction.
  drop frscale /.  ;


\ ****************************
\ *S A and B address registers
\ ****************************
\ *P It has been proposed, principally by Chuck Moore, that the
\ ** Forth virtual machine could usefully be extended by two
\ ** scratch/index registers. Such registers have been used in
\ ** various silicon stack machines, including the IntellaSys C18
\ ** cores.

\ *P Using the A and B registers improves many DSP operations.
\ ** It improves code readability and reduces the amount of stack
\ ** shuffling required to get at the three operands often
\ ** required, e.g. source address, destination address and
\ ** length.

\ *P This section provides a high level simulation of these
\ ** registers using global variables. For high performance,
\ ** these routines can be coded or, on CPUs with enough real
\ ** registers, replaced by code generators.

variable Areg	\ -- addr
\ *G Simulates the A register.
variable Breg	\ -- addr
\ *G Simulates the B register.

: >A		\ x --
\ *G Writes to the A register.
  Areg !  ;
: >B		\ x --
\ *G Writes to the B register.
  Breg !  ;
: A>		\ -- x
\ *G Reads the A register.
  Areg @  ;
: B>		\ -- x
\ *G Reads the B register.
  Breg @  ;
: A@		\ -- x
\ *G Read the memory pointed to by the A register.
  Areg @ @  ;
: A!		\ x --
\ *G Write the memory pointed to by the A register
  Areg @ !  ;
: B@		\ -- x
\ *G Read the memory pointed to by the B register.
  Breg @ @  ;
: B!		\ x --
\ *G Write the memory pointed to by the B register
  Breg @ !  ;
: A@+		\ -- x
\ *G Read memory pointed to by A, increment A by one cell.
\ ** A post-incremented read.
  Areg dup @ dup @			\ -- addr A (A)
  -rot  cell + swap !  ;
: B@+		\ -- x
\ *G Read memory pointed to by B, increment B by one cell.
\ ** A post-incremented read.
  Breg dup @ dup @			\ -- addr B (B)
  -rot  cell + swap !  ;
: A@-		\ -- x
\ *G Read memory pointed to by A, decrement A by one cell.
\ ** A post-decremented read.
  Areg dup @ dup @			\ -- addr A (A)
  -rot  cell - swap !  ;
: B@-		\ -- x
\ *G Read memory pointed to by B, decrement B by one cell.
\ ** A post-decremented read.
  Breg dup @ dup @			\ -- addr B (B)
  -rot  cell - swap !  ;
: A!+		\ x --
\ *G Write to the memory pointed to by A, and update A.
  Areg dup @ dup			\ -- x reg A A
  -rot 2swap !				\ -- reg A
  cell + swap !
;
: B!+		\ x --
\ *G Write to the memory pointed to by B, and update B.
  Breg dup @ dup			\ -- x reg A A
  -rot 2swap !				\ -- reg A
  cell + swap !
;

compiler
: >A		\ x --
  Areg !  ;
: >B		\ x --
  Breg !  ;
: A>		\ -- x
  Areg @  ;
: B>		\ -- x
  Breg @  ;
: A@		\ -- x
  A> @  ;
: A!		\ x --
  A> !  ;
: B@		\ -- x
  B> @  ;
: B!		\ x --
  B> !  ;
: A@+		\ -- x
  Areg dup @ dup @  -rot  cell + swap !  ;
: B@+		\ -- x
  Breg dup @ dup @  -rot  cell + swap !  ;
: A@-		\ -- x
  Areg dup @ dup @  -rot  cell - swap !  ;
: B@-		\ -- x
  Breg dup @ dup @  -rot  cell - swap !  ;
: A!+		\ x --
  Areg dup @ dup  -rot 2swap !  cell + swap !  ;
: B!+		\ x --
  Breg dup @ dup  -rot 2swap !  cell + swap !  ;
target


\ *********************
\ *S IIR filter example
\ *********************
\ *P This example uses:
\ *(
\ *B Fractional arithmetic,
\ *B The A and B register mechanism,
\ *B A address register for data pointing,
\ *B B address register for coefficient pointing.
\ *)
\ *P Note that the following routines all leave A and B pointing
\ ** to the next appropriate place. These routines change and do
\ ** not save or restore A and B.

\ ================
\ *N Biquad filter
\ ================

: 1STEP+	\ sum -- sum'
\ *G Perform a multply/accumulate step, incrementing both
\ ** pointers.
  B@+ A@+ *. +  ;
: 1STEP-	\ sum -- sum'
\ *G Perform a multply/accumulate step, incrementing the
\ ** coefficient pointer and decrementing the data pointer.
  B@+ A@- *. +  ;
: SHIFT2	\ fr --
\ *G The last step of the filter. The current data item is shifted
\ ** into the next data slot and replaced by *\i{fr}.
  A@ SWAP A!+ A!+ ;

: (BIQUAD)	\ frx -- fry
\ *G The core of the biquad filter operation.
  DUP >R
  B@+ *.			( initial sum = B0*input )
  1STEP+ 1STEP-  R> SHIFT2
  1STEP+ 1STEP-  ;

: BIQUAD	\ fx addr-filt addr-coef -- fry
\ *G A single order biquad filter.
  >B >A  (BIQUAD) DUP SHIFT2  ;
: 2xBIQUAD	\ fx addr-filt addr-coef -- fry
\ *G A second order biquad filter.
  >B >A (BIQUAD) (BIQUAD) DUP SHIFT2  ;
\ higher order filters just add more (BIQUAD) stages

: BIQUADinit	\ addr --
\ *G Initialise a first order biquad data array.
  4 CELLS 0 FILL  ;
: 2xBIQUADinit	\ addr --
\ *G Initialise a second order biquad data array.
  6 CELLS 0 FILL  ;


\ =================
\ *N Biquad example
\ =================

4 cells buffer: 'Filter1	\ -- addr
\ *G A first order BIQUAD array.

Create 'Filter1_coef	\ -- addr
\ *G Coefficients for the first order operation.
  131  ,	\ B0    0.007995605
  -255  ,	\ B1   -0.015563965
  131  ,	\ B2    0.007995605
  32427  ,	\ A1    1.979187012
  -16050  ,	\ A2   -0.979614258

: Filter1	\ fr1 -- fr2
\ *G Perform the first order biquad.
  'Filter1 'Filter1_coef BIQUAD  ;

6 cells buffer: 'Filter2	\ -- addr
\ *G A second order BIQUAD array.

Create 'Filter2_coef	\ -- addr
\ *G Coefficients for the second order operation.
  144  ,		\ B0    first section
  -279  ,		\ B1
  144  ,		\ B2
  32155  ,		\ A1
  -15779  ,		\ A2

  1465  ,		\ B0    2nd section
  -2914  ,		\ B1
  1465  ,		\ B2
  32517  ,		\ A1
  -16148  ,		\ A2

: Filter2	\ fr1 -- fr2
\ *G Perform the second order biquad.
  'Filter2 'Filter2_coef 2xBIQUAD  ;


\ ======
\ *> ###
\ ======

decimal
