\ Timebase - high level code

((
Copyright (c) 1998, 1999, 2000, 2005
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 80 631441
fax: +44 (0)23 80 339691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

20050225 MPE004 Moved delay words into DELAYS.FTH.
		Reformatted DocGen comments.
20000511 SFP003 Removed ACTUALPERIOD from embedded systems
20000314 SFP002 Reused for embedded systems
19991201 SFP001	Converted to use Timer IDs rather than structure addresses.
		Added ACTUALPERIOD value that can be used by timer actions.
))


\ ===========
\ *! timebase
\ *T Periodic Timers
\ ===========
\ ***************
\ *S Introduction
\ ***************
\ *P This code provides a timer system that allows many timers
\ to be defined, all slaved from a single periodic interrupt.
\ ** The Forth words in the user accessible group documented
\ ** below are compatible with the code supplied with MPE's
\ ** embedded targets, and with VFX Forth. This code assumes the
\ ** presence of a global value *\fo{TICKS} which holds a time
\ ** value incremented in milliseconds. The timebase is
\ ** approximate. Granularity and jitter are affected by the timer
\ ** ISR and the time taken by your own code to execute. By
\ ** default, the ticker is set to run every 1..100 ms, usually
\ ** defined by the *\fo{EQU}ate *\fo{TICK-MS}. The source code
\ ** is in the the file *\i{TIMEBASE.FTH}.

\ *P The file *\i{DELAYS.FTH} should be compiled after
\ ** *\i{TIMEBASE.FTH}. The code to start and stop the timebase
\ ** system is part of the ticker interrupt system, which is
\ ** compiled after *\i{DELAYS.FTH}. If you need to write a new
\ ** ticker interrupt handler, there will be examples to start
\ ** from in the *\i{<CPU>\DRIVERS} folder. The required
\ ** compilation order is this:
\ *E multitasker (optional)
\ ** TIMEBASE.FTH (optional)
\ ** DELAYS.FTH
\ ** Ticker driver

\ *P The timer chain is built using a buffer area, and two chain
\ ** pointers. Each timer is linked into either the free timer chain,
\ ** or into the active timer chain.

\ *P All time periods are in milliseconds. Note that on a 32 bit
\ ** system, these time periods must be less than 2^31-1 milliseconds,
\ ** say 596 hours or 24 days, whereas if the code is on a 16 bit
\ ** system, time periods must be less than 2^15-1 milliseconds,
\ ** say 32 seconds.

\ ***********************
\ *S The basics of timers
\ ***********************
\ *P These basic words are defined for applications to use the timer system.
\ ** Other words are detailed elswhere in this chapter.

\ *E START-TIMERS       \ -- ; must do this first
\ ** STOP-TIMERS        \ -- ; closes timers
\ ** AFTER              \ xt period -- timerid/0 ; runs xt once after period ms
\ ** EVERY              \ xt period -- timerid/0 ; runs xt every period ms
\ ** TSTOP              \ timerid -- ; stops the timer
\ ** MS                 \ period -- ; wait for period ms

\ *P After the timers have been started, actions can be added.
\ ** The example below starts a timer which puts a character on
\ ** the debug console every two seconds. Note that when using
\ ** generic I/O, the output and input devices MUST be specified.

\ *E start-timers
\ ** : t        \ -- ; will run every 2 seconds
\ **   console opvec !
\ **   [char] * emit
\ ** ;
\ ** ' t 2 seconds every  \ returns timer id, use TSTOP to stop it

\ *P The item on stack is a timer handle, use *\fo{TSTOP} to halt
\ ** this timer.

\ *P *\fo{AFTER} is very useful for creating timeouts, such as
\ ** required to determine if something has happened in time.
\ ** *\fo{AFTER} returns a timerid. If the action you are
\ ** protecting happens in time, just use *\fo{TSTOP} when the
\ ** action happens, and the timer will never trigger. If the
\ ** action does not happen, the timer event will be triggered.

\ ***********************************
\ *S Considerations when using timers
\ ***********************************
\ *P All timers are executed within a single interrupt, and so
\ ** all timer action words share a common user area. This has
\ ** some impact on timer action words. Since you do not know in
\ ** which order timer action words are executed, you must set up
\ ** any *\fo{USER} variables such as *\fo{BASE} that you use,
\ ** either directly or indirectly.

\ *P The interrupt that handles all the timers does not set
\ ** *\fo{IPVEC} and *\fo{OPVEC} to a default value. If you use
\ ** I/O words such as *\fo{EMIT} and *\fo{TYPE} within a timer
\ ** action, you MUST set *\fo{IPVEC} and *\fo{OPVEC} before
\ ** using the I/O. For the sake of other timer action routines
\ ** that may still be using default I/O, it is polite to save
\ ** and restore *\fo{IPVEC} and *\fo{OPVEC} in your timer action
\ ** words.

\ *P Do not worry about calling *\fo{TSTOP} with a timerid that
\ ** has already been executed and removed from the active timer
\ ** chain; if *\fo{TSTOP} cannot find the timer, it will ignore
\ ** the request.

\ *P Under some conditions, the execution time of all the timer
\ ** routines may be longer than the requested period of the
\ ** timer. In addition, the timer interrupt may be subject to
\ ** jitter.

\ ************************
\ *S Implementation issues
\ ************************
\ *P The following discussion is relevant if you want to modify
\ ** this code. Functionally equivalent code is provided with
\ ** MPE's VFX Forth systems. In the Windows environment, timer
\ ** interrupts are implemented by callbacks and critical sections.

\ *P By default, the word *\fo{DO-TIMERS} is run from within the
\ ** periodic timer interrupt. If interrupts are not re-enabled
\ ** after resetting the timer interrupt, you may have latency
\ ** issues if a number of timers is used, or if one of the timer
\ ** routines takes a considerable time, or if you want to use
\ *** *\fo{KEY} or *\fo{EMIT} in the timer action. In this case,
\ ** it would be better to set up the timer routine to *\fo{RESTART}
\ ** a task which calls *\fo{DO-TIMERS},e.g.

\ *E : TIMER-TASK    \ --
\ **   <initialise>
\ **   BEGIN
\ **     DO-TIMERS STOP
\ **   AGAIN
\ ** ;

\ *P Such a strategy also permits you to use a fast interrupt,
\ ** say 1 ms, for the clock, and to trigger the *\fo{TIMER-TASK}
\ ** every say 32 ms. Another strategy is to leave *\fo{DO-TIMERS}
\ ** in the interrupt routine and for each timer action be to put
\ ** a different character into a circular buffer (and to restart
\ ** a task).

\ *E 16 cqueue: timerQ
\ **
\ ** : TimerAction    \ --
\ **   <initialise>
\ **   begin
\ **     case  timerQ cqueue>
\ **       [char] A of  ...  endof
\ **       [char] B of  ...  endof
\ **            ...
\ **     endcase
\ **   again
\ ** ;

\ *P Providing that the queue is big enough, no events
\ ** are lost and event order is preserved. However, an
\ ** action that takes 20 seconds may/will delay all subsequent
\ ** actions.

\ *P The best handler structure depends heavily on your application
\ ** design. You need to consider things like
\ *(
\ *B Required interrupt latency
\ *B Required handler latency
\ *B Must handlers use comms?
\ *B Have I enough RAM for a task?
\ *)
\ *P As with all software design, it's usually a trade-off. The
\ ** default "all in the interrupt" design is good enough for
\ ** many users. For all the rest, adding a task is a good first
\ ** step.


\ ********************
\ *S Timebase glossary
\ ********************

\ ***********************************
\ Structures, variables and constants
\ ***********************************

[undefined] ticks [if]
0 value ticks	\ -- addr ; holds timer count
\ *G Get current clock value in milliseconds.
[then]

#8 constant #timers	\ -- n ; maximum number of timers
\ *G A constant used at compile time to set the maximum number of
\ ** timers required. Each timer requires RAM as defined by the
\ ** *\fo{ITIMER} structure.

struct itimer   \ -- size ; interval timer
  cell field itlink                     \ link to next timer ; MUST be first
  cell field itTimerId			\ timer ID
  cell field itinterval                 \ period of timer in MS
  cell field ittimeout                  \ next timeout
  cell field itmode                     \ mode/flags, 0=periodic, 1=one shot
  cell field itxt                       \ word to execute
end-struct

#timers itimer * constant /itimers	\ size of required buffer
/itimers buffer: timer-chain            \ for all timer chains

variable active-timers                  \ pointer to head of active timers chain
variable free-timers                    \ pointer to head of free timers chain
variable nextTimerID			\ next timer ID to be allocated

#8192                                   \ timer errors
  dup constant err_no_timers  1+
  dup constant err_bad_timer  1+
drop

0 constant periodic                     \ timer modes
1 constant single-shot


\ *******************
\ Timer ID management
\ *******************

: FormID	\ struct/0 -- ID/0
  dup if
    NextTimerID @ 0=			\ 0 is bad!
    if  1 NextTimerID +!  then
    NextTimerID @ dup rot itTimerID !
    1 NextTimerID +!
  endif
;

: FindID	\ id -- struct/0
  active-timers @
  begin
    dup
   while				\ get out at end of chain
    2dup itTimerID @ =			\ if this timer is the one
    if  nip  exit  endif		\   get out returning structure
    itlink @
  repeat
  nip					\ discard ID, return 0
;


\ **********************
\ Timer chain management
\ **********************

: init-timers   \ -- ; initialises timer system
  timer-chain /itimers erase            \ wipe buffer area
  0 active-timers !                     \ no active timers
  0                                     \ no free timers yet
  timer-chain /itimers bounds do
    i itlink !                          \ set old value
    i                                   \ this becomes new value
  itimer +loop
  free-timers !
;

: get-timer     \ -- struct ; throw if no free timer
  free-timers @ 0= err_no_timers ?throw
  di					\ make this safe!
  free-timers @                         \ new structure
  dup @ free-timers !                   \ unhook from free chain
  active-timers @ over itlink !         \ add to active timer chain
  dup active-timers !
  ei
;

: release-timer \ struct -- ; exit quietly if timer can't be found
  active-timers
  begin                                 \ -- struct s2
    2dup itlink @ <>                    \ itlink MUST be first
   while
     itlink @ dup 0=			\ if at end of chain
     if  2drop  exit  endif		\   get out quietly
  repeat
\ struct prev -- ; take the next link, and place it in the previous link
  over itlink @  swap itlink !
  free-timers @ over itlink !           \ add to free chain
  free-timers !
;

: (set-timer)   \ xt period mode -- struct ; set up timer
  get-timer
  tuck itmode !                         \ operating mode
  tuck  2dup itinterval !               \ period
  swap ticks +  swap ittimeout !	\ period+now = trigger
  tuck itxt !                           \ word to execute
;

: set-timer     \ xt period mode -- struct/0
  ['] (set-timer) catch
  if  2drop drop  0  endif
;

decimal

-short-branches
: do-timers     \ --
\ *G Process all the timers in the chain
  active-timers @
  begin
    dup
   while                                \ -- struct
    dup itlink @ swap                   \ get next in chain now - might be released
    dup ittimeout @  ticks - 0< if	\ if the interval has expired
      dup itxt @ execute                \ run the word
      dup itmode @ single-shot = if	\ if a single shot timer
        release-timer                   \   dispose of the timer
      else
        dup itinterval @                \   update the timeout point
        swap ittimeout +!
      endif
    else
      drop
    endif
  repeat
  drop                                  \ discard the end of chain
;
+short-branches


\ ****************
\ Diagnostic tools
\ ****************

((
: .t-chain      \ chain -- ; display a chain
  begin
    @ ?dup
   while
    dup .dword space
  repeat
;

: .t-active     \ -- ; display active timer chain
  cr ." Active timers:"
  cr
  active-timers .t-chain
;

: .t-free       \ -- ; display free timer chain
  cr ." Free timers:"
  cr
  free-timers .t-chain
;
))


\ **********************
\ User level timer words
\ **********************

: after		\ xt period -- timerid/0 ; xt is executed once,
\ *G Starts a timer that executes once after the given period.
\ ** A timer ID is returned if the timer could be started,
\ ** otherwise 0 is returned.
  single-shot set-timer formID
;

: every         \ xt period -- timerid/0 ; periodically
\ *G Starts a timer that executes every given period. A timer
\ ** ID is returned if the timer could be started, otherwise 0
\ ** is returned. The returned timerID can be used by *\fo{TSTOP}
\ ** to stop the timer.
  periodic set-timer formID
;

: tstop         \ timerid --
\ *G Removes the given timer from the active list.
  findID ?dup
  if  release-timer  endif
;

\ ***********************
\ Clock hardware handling
\ ***********************


(( Moved to CLOCKINT.FTH in CPU specific directory
#32 equ clock-update            \ adjust this for your own hardware
                                \ it should be the number of milliseconds
                                \ of the clock interrupt period
proc clockisr
  push.l  er0

\ first process the clock interrupt
  mov.l   @ clock er0           \ inc clock
  add.l   # clock-update er0
  mov.l   er0     @ clock
  bclr    # 2     @ _TSR0       \ and clear interrupt (ack)

\ run the clock handler word
  push.l  er1                   \ ER0 already saved.
  push.l  er2
  push.l  er3
  push.l  er4
  push.l  er5
  mov.l   # int-init-u0 up      \ interrupt user area
  jsr     @ ' do-timers         \ run timer handler (use NOOP for testing)
  pop.l   er5
  pop.l   er4
  pop.l   er3
  pop.l   er2
  pop.l   er1

\ Restore
  pop.l   er0
  rte
end-code
clockisr ovi0-vector !                  \ set vector

: start-timers  \ -- ; Start internal time clock
  init-timers                           \ initialise time base system
  $8B _TCR0 c!                          \ TCNT not cleared, Falling Edge, IClk/8
  $FC _TIER0 c!                         \ Overflow interrupts
  _TSTR c@ $001 or _TSTR c!             \ start timer 0
  ei
;

: Stop-timers   \ -- ; disable interrupts
  _TSTR c@ $0FE and _TSTR c!            \ stop timer 0
;
))


\ *********
\ Test code
\ *********

((
Start-timers

: t	\ -- ; will run every 2 seconds
  console opvec !		\ MUST define output device
  [char] * emit
;

' t 2 seconds every		\ returns a timer id, use TSTOP to stop it

cr
cr ." Item on stack is a timer handle, use TSTOP to halt this timer"
cr
))


\ ======
\ *> ###
\ ======

