\ XmodemTxRx.fth - XMODEM file upload/download tools

((
Copyright (c) 2001, 2002, 2004, 2005, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20110107 SFP006 Added key timeouts.
20051004 MPE005 Added 1k blocks and CRC check options.
		Revamped block send and receive.
20040504 MPE004 Simplified by rounding lengths to 128 byte units.
20040112 SFP003 Added ^Y equate. Reordered some code.
20021014 SFP002 Modified for COMMON directory.
20010717 EJB000 First MPE release
))


\ ==================
\ *! xmodemtxrx
\ *T XMODEM Receiver and Transmitter
\ ==================
\ *S Introduction
\ *P The file *\i{Common\XmodemTxRx.fth} implements the XMODEM
\ ** 128 and 1024 byte protocols in both directions. Use with AIDE
\ ** requires AIDE release 3.00 upwards. A very simplified version
\ ** of the 128 byte checksum receive code may be found in
\ ** *\b{Common\MinXmodemRx.fth} and is ideal for Flash reprogramming.

\ *P The original shorter code that just handles the 128 byte
\ ** protocol is available as *\i{Common\XmodemTxRx128.fth}.

\ *P No test code is provided for this file as the system has
\ ** been tested by comparison of transferred binary files.

\ ==========================
\ *S Words in XmodemTxRx.fth
\ ==========================

decimal


\ ================
\ *N Configuration
\ ================

[undefined] XmodemTx? [if]
1 equ XmodemTx?	\ -- n
\ *G Non-zero to compile transmit code
[then]

[undefined] XmodemRx? [if]
1 equ XmodemRx?	\ -- n
\ *G Non-zero to compile receive code.
[then]


\ ==========================
\ *N Constants and variables
\ ==========================

\ Characters
$001 equ SOH				\ for Xodem-128
$002 equ STX				\ for XModem-1k
$004 equ EOT				\ End of Transmission
$006 equ ACK				\ Acknowledge
$015 equ NAK				\ Negative Acknowledge
$018 equ CAN				\ Cancel
$019 equ ^Y				\ AIDE: Xmodem to host
$01A equ ^Z				\ AIDE: Xmodem/128/cksum from host
$01C equ ^\				\ AIDE: Xmodem/1024/crc from host

\ Error codes
$0101 equ blkerror
\ *G A block number error has occurred.
$0103 equ noreply
\ *G There was no reply within one second.
$0104 equ crcerror
\ *G Bad CRC or checksum.
$0105 equ overflow
\ *G Too many blocks were sent.

$010 equ maxerrs	\ -- n
\ *G Maximum number of errors before transfer is aborted.

\ Variables
variable BLK#                           \ Current Block No.
variable CUR-ADDRESS                    \ Current host address
variable LAST-BLOCK                     \ last block we can receive
variable rxstatus			\ receiver status
variable rxerrcnt			\ counter for rx errors
#1024 Buffer: x-buffer	\ -- addr
\ *G Holds a 128 or 1024 byte Xmodem data block.
#128 value /Xblk	\ -- n
\ *G Holds Xmodem block size, 1024 or 128.
0 value Xmode		\ -- n
\ *G Holds 0 for checksum mode, nz for CRC-16.


\ ==============
\ *N Common code
\ ==============

: init-blks		\ addr #bytes -- #blks
\ *G Given the size of an image, return the number of complete
\ ** 128 byte blocks, set the variable CUR-ADDRESS to ADDR and
\ ** set the variable BLK# to 1.
  swap cur-address !			\ set starting data address
  /Xblk 1- +  /Xblk /			\ calculate number of blocks
  1 blk# !				\ 1 is first block
;

[defined] CPU=ARM [if]	\ use ARM coded version
code +Xcrc	\ crc char -- crc'
\ tos=crc, r3=scratch, r4=polynomial
  ldr      r1, [ psp ], # 4		\ r1 := crc
  eor      tos, r1, tos lsl # 8
  mov      r4, # $1000
  orr      r4, r4, # $0021
    and .s   r3, tos, # $8000		\ 1
    mov      tos, tos lsl # 1
    eor .ne  tos, tos, r4
    and .s   r3, tos, # $8000		\ 2
    mov      tos, tos lsl # 1
    eor .ne  tos, tos, r4
    and .s   r3, tos, # $8000		\ 3
    mov      tos, tos lsl # 1
    eor .ne  tos, tos, r4
    and .s   r3, tos, # $8000		\ 4
    mov      tos, tos lsl # 1
    eor .ne  tos, tos, r4
    and .s   r3, tos, # $8000		\ 5
    mov      tos, tos lsl # 1
    eor .ne  tos, tos, r4
    and .s   r3, tos, # $8000		\ 6
    mov      tos, tos lsl # 1
    eor .ne  tos, tos, r4
    and .s   r3, tos, # $8000		\ 7
    mov      tos, tos lsl # 1
    eor .ne  tos, tos, r4
    and .s   r3, tos, # $8000		\ 8
    mov      tos, tos lsl # 1
    eor .ne  tos, tos, r4
  bic      tos, tos, # $FF000000	\ zero top 16 bits
  bic      tos, tos, # $00FF0000
  mov      pc, link
end-code
[then]

[undefined] +Xcrc [if]		\ high level version
: +Xcrc		\ crc char -- crc'
\ *G Update the XMODEM CRC with the given character. The initial
\ ** value should be zero.
  8 lshift xor
  8 0 do
    dup $8000 and if
      1 lshift  $1021 xor
    else
      1 lshift
    then
  loop
  $FFFF and
;
[then]

[undefined] .decimal [if]
: .decimal  base @ >r  decimal .  r> base !  ;
[then]

: .StartX	\ --
\ Display Xmodem start message.
  cr ." Start host's Xmodem-" /Xblk .decimal
;


\ ======================
\ *N XMODEM transmission
\ ======================

XmodemTx? [if]

: (To-Buffer)	\ --
\ *G Move 128/1024 bytes from the memory pointed to by
\ ** *\fo{CUR-ADDRESS} into *\fo{X-BUFFER}. This is the default action
\ ** of *\fo{TO-BUFFER}. The variable *\fo{CUR-ADDRESS} is set by
\ ** *\fo{BIN-DOWN} and friends.
  CUR-ADDRESS @ X-buffer /Xblk cmove    \ Move 128/1024 bytes from
  /Xblk cur-address +!  ;		\ memory pointed to by
                                        \ CUR-ADDRESS into buff

Defer To-Buffer	\ --
\ *G Copy the next 128/1024 bytes to transmit to *\fo{X-BUFFER}.
\ ** They are then transmitted from *\fo{X-BUFFER}. You can change
\ ** this action as required by your application.
\ ** The default action is *\fo{(TO-BUFFER)}.
  assign (to-buffer) to-do to-buffer

: ?Ack          \ -- ; wait for char, abort if not ACK
\ *G Wait for a character and terminate the transfer and abort
\ ** if the character is not an ACK.
  ser-key ACK <>                        \ wait for response
  IF  EOT ser-emit                      \ flag end of XMODEM
      1 abort" File error"		\ MUST be ACK
  ENDIF  ;

: Send-Block	\ Blk# -- ; transmit a block
\ *G Transmit the 128/1024 byte contents of *\fo{X-BUFFER} to the host.
  /Xblk #128 =
  if  SOH  else  STX  then  ser-emit    \ Start of header
  dup ser-emit                          \ Block Number
  invert ser-emit			\ 1's complement Block#
  Xmode if
    0  X-buffer /Xblk bounds DO		\ -- crc; Send 128/1024 byte Block
      i c@ dup ser-emit  +Xcrc
    LOOP
    dup 8 rshift ser-emit  ser-emit	\ send CRC
  else
    0  X-buffer /Xblk bounds DO		\ -- crc; Send 128/1024 byte Block
      i c@ dup ser-emit  +
    LOOP
    ser-emit				\ send checksum
  endif
  ?Ack					\ wait response
;

: Bin-Down	\ addr #bytes -- ; transfer memory to host
\ *G Download (transmit) the given block of memory to the host
\ ** using the XMODEM 128/1024 byte block protocol.
\ ** On entry, the variable *\fo{CUR-ADDRESS} is set to *\i{addr}
\ ** on entry and *\i{#bytes} is rounded up to a 128/1024 byte
\ ** unit.
  .StartX ." receiver " cr
  ^Y ser-emit                           \ trigger for AIDE/PowerTerm
  init-blks				\ -- #blks
  begin					\ Wait for receiver ...
    case  ser-key
      NAK      of 0 to Xmode  1  endof	\ simple checksum
      [char] C of 1 dup to Xmode endof	\ crc-16
        0
    endcase
  until
  1+ 1                                  \ block numbers from 1
  ?DO  to-buffer                        \ Transfer mem to buffer
       i send-block                     \ Send this block
       1 BLK# +!                        \ Count blocks sent
  LOOP
  EOT ser-emit  ?Ack                    \ terminate XMODEM
  cr ." File transfer successful" cr
;

[then]


\ ===================
\ *N XMODEM reception
\ ===================

XmodemRX? [if]

[undefined] ser-flush [if]
: ser-flush	\ -- ; flush the link input
\ *G Flush all input characters from the host/target link.
  begin
    ser-key?
  while
    ser-key drop
  repeat
;
[then]

: send-ack	\ -- ; send ACK
\ *G Transmit an ACK character.
  ACK ser-emit
;

: send-nak	\ -- ; Transmit a NAK character
\ *G Transmit a NAK character, usually to trigger a retransmit.
  NAK ser-emit
;

: send-can	\ -- ; send CAN character
\ *G Transmit a CAN character.
  CAN ser-emit
;

: toomanyerrs?	\ -- T|F ; true if too many errors
\ *G Return true if too many comms errors have occurred.
  1 rxerrcnt +!
  rxerrcnt @ maxerrs >
;

: (From-Buffer)	\ --
\ *G Move 128/1024 bytes from *\fo{X-BUFFER} into the memory
\ ** pointed to by *\fo{CUR-ADDRESS}.
  X-buffer CUR-ADDRESS @ /Xblk cmove	\ Move 128/1024 bytes
  /Xblk cur-address +!
;

Defer From-Buffer	\ --
\ *G Move 128/1024 bytes from *\fo{X-BUFFER} into the memory
\ ** pointed to by *\fo{CUR-ADDRESS}. *\fo{CUR-ADDRESS} is set
\ ** up by *\fo{BIN-UP} and friends. The default action is
\ ** *\fo{(FROM-BUFFER)}. You can modify the action to suit your
\ ** own application.
  assign (from-buffer) to-do from-buffer

\ variable xnum
\ SFP006...
: TimedKey	\ -- char|-1
\ *G As *\fo{KEY} but returns -1 on timeout of 100 ms
  ticks #100 + begin			\ -- ms
    ser-key?
    if  drop  ser-key exit  endif	\ valid key
[ tasking? ] [if] pause [then]
    dup ticks - 0<			\ timer expired?
  until
  drop -1				\ timeout value
;
\ ...SFP006

: RecvXData	\ -- flag
\ Receive an Xmodem data block and return true if the checksum or CRC matches.
  Xmode if
    0					\ initial CRC
    X-buffer /Xblk bounds do		\ Get 128/1024 byte Block
      TimedKey dup -1 =			\ SFP006
      if  drop leave endif
      dup i c!				\ save byte to buffer
      +Xcrc				\ update CRC
    loop
    TimedKey 8 lshift  TimedKey or  =	\ test CRC ; SFP006
  else
    0					\ initial checksum
    X-buffer /Xblk bounds do		\ get 128/1024 byte Block
      TimedKey dup -1 =			\ SFP006
      if  drop leave endif
      dup i c!				\ save byte to buffer
      +					\ add to Checksum
    loop
    $FF and  TimedKey =			\ test checksum ; SFP006
  endif
;

: Get-Block	\ --
\ *G Receive an XMODEM 128/1024 byte data block from the host,
\ ** processing the header and checksum data.
  TimedKey TimedKey			\ get block and inverse ; SFP006
  over + $0FF <> if
    toomanyerrs? if
      send-can send-can
      blkerror rxstatus !  ( 1 xnum ! )
    else
      send-nak ( 2 xnum ! )
    endif
    send-can
    blkerror rxstatus !  ( 3 xnum ! )
    drop				\ lose incoming block number
  else
    Blk# @ $00FF and <> if		\ correct block ??
      toomanyerrs? if
        send-can send-can
        blkerror rxstatus !  ( 4 xnum ! )
      else
        send-nak  ( 5 xnum ! )
      endif
    else
\      Get-BlockData GetCheck = if	\ get cksum or CRC
      RecvXData if			\ if good block
        Blk# @ last-block @ > if
          send-can send-can		\ 2 CAN's aborts transfer
          overflow rxstatus !		\ no space left in buffer
	  ( 6 xnum ! )
        else
          From-Buffer			\ copy data to memory
          send-ack			\ all OK send ACK
          1 Blk# +!			\ bump for next block
          rxerrcnt off			\ clear error count again
        endif
      else
        toomanyerrs? if
          send-can send-can		\ 2 CAN's aborts transfer
          crcerror rxstatus !  ( 7 xnum ! )
        else
          send-nak  ( 8 xnum ! )
        endif
      endif
    endif
  endif
;

: (WaitResponse)	\ --
\ *G Wait for up to one second for a character.
  ticks #1000 +
  begin
[ tasking? ] [if] pause [then]
    dup ticks - 0<			\ timer expired?
    ser-key? or				\ or a character received
  until
  drop
;

: SendReq	\ --
\ *G If Xmode is set, send a C character, otherwise send a NAK.
  Xmode
  if  [char] C ser-emit  else  NAK ser-emit  endif
;

0 value /RXms	\ --
\ *G Holds the transfer time in milliseconds.

: Bin-Up	\ addr len -- status ; status 0 = GOOD
\ *G Upload (receive) a block of data of the given size into memory
\ ** using the XMODEM 128/1024 byte block protocol. An error status
\ ** is returned, 0 indicating success.
\ ** On entry, the variable *\fo{CUR-ADDRESS} is set to *\i{addr}
\ ** on entry and *\fo{len} is rounded up to a 128/1024 byte unit.
\ ** Note that an error return of $0105 indicates the the file
\ ** being sent is larger than the *\i{len} input parameter.
  .StartX ." transmitter" cr
  /Xblk #128 =
  if  ^Z  else  ^\  endif  ser-emit	\ Trigger for AIDE/PowerTerm
  rxstatus off				\ start with good status
  rxerrcnt off				\ and no errors
  ser-flush				\ flush input

  init-blks last-block !		\ find last block

  SendReq				\ NAK for checksum, C for CRC
  #40 0 do				\ wait for remote startup
    (WaitResponse) ser-key? ?leave	\ done if a reply
    i #30 >
    if  0 to Xmode  endif	        \ cancel CRC req after 30 secs.
    SendReq
  loop

  ticks >r				\ start ms

  begin					\ outer loop handles blocks
    (WaitResponse) ser-key? 0=
    if  noreply rxstatus !  endif	\ flag no response
    rxstatus @ 0= if
      case  ser-key				\ get response
        EOT of  EOT rxstatus !  send-ack  endof	\ all finished
        CAN of  CAN rxstatus !  send-ack  endof	\ remote cancelled
        SOH of  #128 to /Xblk  Get-block  endof	\ start of 128 byte data block
        STX of  #1024 to /Xblk  Get-block endof	\ start of 1024 byte data block
      endcase					\ ignore noise characters
    endif
    rxstatus @				\ loop until all done
  until

  ticks r> - abs to /RXms		\ save receive time

  rxstatus @ dup EOT = if
    drop 0				\ returns 0 for good transfer
  else
    #150 ms ser-flush
\    cr ." Upload failed, code = "
\    dup .dword				\ else a fail code
  endif
;

: RecvXmodem	\ addr len -- len' status ; status=0=good
\ *G Upload (receive) a block of data of the given maximum size into memory
\ ** using the XMODEM 128/1024 byte block protocol. The number of bytes
\ ** correctly received and an error status are returned, 0
\ ** indicating success. See *\fo{BIN-UP} above for more details.
  over >r
  ['] bin-up catch
  if  2drop  rxstatus @  endif
  cur-address @ r> - swap
;

[then]


\ ===========
\ *N Defaults
\ ===========
\ *P If you have changed the operation of the buffer handling
\ ** routines, you can restore them.

: Xmodem-1k	\ --
\ *G Default to 1k byte Xmodem blocks with CRCs. This usually
\ ** gives the fastest transfers and best error checking.
  1 to Xmode  #1024 to /Xblk
;

: Xmodem-128	\ --
\ *G Default to 128 byte Xmodem blocks with checksums.
\ ** This works with virtually all terminal emulators.
  0 to Xmode  #128 to /Xblk
;

: XmodemDefaults	\ --
\ *G Set the XModem transfer routines to their default
\ ** host copy operations.
[ XmodemTx? ] [if]
  assign (to-buffer) to-do to-buffer
[then]
[ XmodemRx? ] [if]
  assign (from-buffer) to-do from-buffer
[then]
;


\ ************
\ *S Test code
\ ************

0 [if]

: (NullFrom)	\ --
\ *G Dummy for testing receive.
  /Xblk cur-address +!
;

: +NullRecv	\ --
\ *G Use the dummy receiver.
  assign (NullFrom) to-do from-buffer
;

[then]


\ ======
\ *> ###
\ ======

