\ ANS memory management extensions for 16 bit targets
\ Use HEAP32.FTH for the documentation.

((
Copyright (c) 1998, 1999, 2000, 2003, 2009
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Change History
==============
20090530 SFP005 Made HEAP32.FTH the documentation master.
20000529 SFP004 Converted for 16 bit targets in which the control information
                uses two cells. See below for further documentation.
20000321 SFP003 Fixed more signed/unsigned issues. These are all benign
                in this code, as the maximum allocation is 0x00FFFFFF,
                but changing it makes the code better to read and maintain.
19990315 SJC001 Fixed sign bug in size. Must treat the requested size as
                unsigned in ALLOCATE. 
19980803 MSD001 Correct the returned address when RESIZE fails.  The ANS spec
         MSD001 says that in this situation, RESIZE should return a
         MSD001 (non-zero) return code and the *original* block pointer
         MSD001 address.  This code used to return the block pointer of the
         MSD001 failed ALLOCATE when the block was being increased in size.
         MSD001 Since ALLOCATE returns an undefined value, RESIZE *cannot*
         MSD001 use it as the block size to return when trying to increase
         MSD001 the block size and ALLOCATE fails.
))


\ =======
\ *! heap
\ *T Heap Memory Allocation
\ =======

( *S Heap definition )
( *P The heap is allocated from a predefined section of memory. Facilities are )
( ** provided for user expansion of the heap to mass storage, although the current )
( ** code makes no provision for page management. When the heap is initialised, )
( ** a free block and an end block are created. The end block is of zero size, and )
( ** is used only as a marker. The address returned by ALLOCATE and RESIZE is the )
( ** address of the first data byte, as is the address consumed by FREE. )

( *P The heap MUST be initialised before use by calling INIT-HEAP. )
( ** Heap access words return status=0 for success, and status<>0 for error. )

( *P Two equates are required during compilation to allocate a contiguous block )
( ** of RAM for the heap. )
( *E   STARTOFHEAP is the start address of the heap )
( **   SIZEOFHEAP is the size of the RAM for the heap )

\ *P There are two versions of this code provided. HEAP32.FTH is provided for
\ ** 32 bit targets and is optimised for the VFX code generator. HEAP16.FTH
\ ** is for 16 bit targets, and is optimised for code density.

\ *N 16 bit targets - HEAP16.FTH
( *P The heap is controlled using two cells per block. )
( ** This information is used in three parts: )
( *E cell = #bytes, number of bytes in this block )
( ** cell = flag, split between a four bit and a 12 bit field )
( *P The top four bits of the flag are used to indicate the block type, where )
( ** $E = End, $F = Free, $A = Allocated. Others may be added later for type )
( ** management. )
( *P The bottom 12 bits of the flag are currently unused, and should be set to )
( ** zero. )

\ *N 32 bit targets - HEAP32.FTH
\ *P The heap is controlled using a single cell per block.
\ ** This information is used in two parts:
\ *E bits 31..24: $EE - End, $FF - Free, $AA - Allocated
\ ** bits 23..0: 24 bits for number of data bytes in block.
\ *P A consequence of this is that the maximum block size that
\ ** can be allocated is 16Mb-1 bytes.

\ *P If you use a pre-emptive scheduler or need to use the heap
\ ** routines inside interrupt routines, you must define suitable
\ ** heap lock and unlock routines and set the equate LOCKHEAP? to
\ ** non-zero.
\ *D LockHeap=0 no heap locking
\ *D LockHeap=1 heap locking by turning off interrupts
\ *D LockHeap=2 heap locking by semaphore.


\ *S Gotchas
\ *P The heap routines must be protected if they are to be used both
\ ** in normal code and in interrupts. In this case the code must be
\ ** modified to be interrupt safe, but this may have a significant
\ ** impact on interrupt latency. Examples may be found in
\ ** HEAP32.FTH.

( *S Glossary )
( *P The glossary does not include all the factors used in the )
( ** code. If you are interested in the implementation, please read the )
( ** sources. )

only forth definitions
decimal

[defined] environment [if]
  also environment definitions
  TRUE constant MEMORY-ALLOC
  previous definitions
[then]


$F000 equ MASKTYPE	\ mask to extract type information
$E000 equ LASTBLOCK	\ marker for end of heap
$F000 equ FREEBLOCK	\ marker for free block on heap
$A000 equ ALLOCBLOCK	\ marker for allocated block

internal		\ hide the implementation factors

optimised? [if]
struct /HEAPINFO	\ size of the management info in a block
  int blocksize		\ size of data in the block
  int blockinfo		\ state of the block
end-struct
[else]
2 cells equ /HEAPINFO
compiler
: blocksize  	;		\ immediate NOOP
: blockinfo	cell +  ;
target
[then]


\ The following constant is the minimum *total* size occupied by a block. This
\ size includes overheads generated by alignment retrictions as well as those
\ incurred by the fact that the size and status of a block have to be stored
\ somewhere.
1 aligned /HEAPINFO + equ /MINLEFT	\ minimumum left to split block
sizeofheap /HEAPINFO 2* - equ /initFREE	\ calculate initial free block size

sizeofheap buffer: STARTOFHEAP		\ allocate heap space
startofheap sizeofheap + /HEAPINFO -	\ point to last info block in heap
  equ LASTINFO


\ ***************
\ Heap Primitives
\ ***************

: AllocBlock?	\ ^heap -- flag ; true if allocated
  blockinfo @ MASKTYPE and ALLOCBLOCK =
;

: FreeBlock?	\ ^heap -- flag ; true if free
  blockinfo @ MASKTYPE and FREEBLOCK =
;

: LastBlock?	\ ^heap -- flag ; true if last
  blockinfo @ MASKTYPE and LASTBLOCK =
;

: SkipBlock	\ ^heap -- ^heap' ; skip a block
  dup blocksize @ + /HEAPINFO +		\ skip to next block
;

: SkipAlloc	\ ^heap -- ^heap' ; skip allocated blocks
  begin
    dup AllocBlock?
   while
    SkipBlock
  repeat
;


\ ******************
\ GARBAGE-COLLECTION
\ ******************

: compact	\ -- ; merge free blocks
  startofheap
  begin
    SkipAlloc				\ skip over allocated blocks
    dup freeBlock?			\ if not allocated, its the last one
   while
    dup					\ -- current next
    begin				\ skip over further free blocks
      SkipBlock dup FreeBlock? 0=
    until				\ -- current next'
    tuck				\ next will be new current
    over - /HEAPINFO - swap blocksize !	\ set existing current to merged size
  repeat
  drop
;

: SplitBlock	\ addr size -- ; split block, mark top as free
		\ N.B. calls COMPACT, must mark block as allocated FIRST
  tuck /MINLEFT +  over blocksize @ u< if	\ if size+min < free, split block
  						\ -- size addr
    2dup + /HEAPINFO + >r			\ start of new block
    FREEBLOCK r@ blockinfo !			\ mark free
    2dup blocksize @ swap - /HEAPINFO -		\ mark size of remaining free data
    r> blocksize !
    blocksize !					\ mark new size of original block
    compact					\ try to merge space
  else
    2drop
  endif
;


\ ********************
\ ALLOCATE FREE RESIZE - the ANS words
\ ********************

external

: allocate	\ #bytes -- addr status
( *G Attempt to allocate some memory from the heap. )
( ** Walk the heap looking for a single big enough block. )
( ** If the block is larger than than required split it into two blocks. )
( ** Allocate part or all of the free block. Status=0 for success. )
\ STEP#1: Find a big enough free block
  aligned  startofheap			\ -- size current
  begin
    SkipAlloc dup LastBlock? if		\ no space if last block
      2drop  0 -1  exit
    endif
    2dup blocksize @ u>			\ check for size <= current.size
   while
    SkipBlock
  repeat				\ -- size current'
\ STEP#2: mark allocated, split block if necessary
  ALLOCBLOCK over blockinfo !		\ mark it as allocated
  dup rot SplitBlock			\ split the block
  /HEAPINFO +  0			\ return data address and success
;

: free		\ address -- status
\ *G Attempt to free a heap block. Status=0 for success.
\ ** If *\i{addr} is zero, no action is taken and zero is returned.
  dup if
  /HEAPINFO -  dup AllocBlock? if	\ check validity
    FREEBLOCK swap blockinfo !		\ mark it as free
    compact				\ merge free blocks
    0					\ return success
  else
    drop  -1				\ error if invalid
  endif
  endif
;

\ In the following
\ comments, "new size" refers to the new size requested by the user *after*
\ alignment requirements have been taken into consideration.)
\
\ There are three different cases that we have to consider:
\
\       The new size is less than or equal to the old size minus the minimum
\       size taken by a heap entry.  In this case, we shrink the block size
\       making the free space into a free block.  We return the *existing*
\       block pointer because the user's data hasn't been moved.
\
\       The new size is *greater* than the old block size.  Since the user
\       has requested more space than is available in the current block, we
\       have to allocate a *new* block, copy the user's data into it and
\       then free the *old* block.  We return the pointer to the
\       newly-allocated block, of course.
\
\       If neither of the above cases apply, then we just return the
\       existing pointer because there is not enough unused space in the
\       block to create a new (free) heap block.
\
: resize	\ addr1 size -- addr2 ior
( *G Try to resize an allocated block to a new size, allowing for alignment. )
( ** If the existing memory block is not big enough, the data will be copied )
( ** to a new block, and the returned addr2 will not be the same as addr1. )
( ** Status=0 for success. )
  aligned swap				\ align size
  dup >r				\ preserve addr1
  /HEAPINFO -				\ -- size block1 ; step to management block
  dup AllocBlock? 0= if			\ consistency check
    2drop  r> -1  exit
  endif
  2dup blocksize @ u< if 		\ downsizing?
    swap SplitBlock			\  try to split it
    r> 0  exit				\  return addr and success
  then
  over swap blocksize @ u> if		\ if new is larger than current
    dup allocate if			\ -- size addr2 ; failed?
      2drop  r> -1			\ return original address and error
    else
      tuck				\ -- addr2 size addr2
      r@ swap rot cmove			\ -- addr2 ; copy data
      r> free				\ -- addr2 ior
    endif
    exit
  then
  drop  r> 0				\ otherwise, recycle current block
;


\ **************
\ Initialisation
\ **************

: init-heap	\ -- ; initialise the heap structures
( *G The heap is initialised by creating 2 blocks. )
( ** Block 1 starts at the beginning and is marked as a free block. )
( ** Block 2 Is a null marker at the end of heap space. )
  startofheap sizeofheap erase		\ erase heap memory

  /initFREE StartOfHeap blocksize !	\ set size of free block
  FREEBLOCK StartOfHeaP blockInfo !	\ and mark it as free

  LASTBLOCK LASTINFO blockInfo !	\ store end of heap marker
;


\ **************
\ *S Diagnostics
\ **************

heap-diags? [if]
: size		\ addr -- currsize | -1
( *G Return the size of an allocated block or -1 if there's an error. )
  /HEAPINFO -  dup AllocBlock? if
    blocksize @
  else
    drop -1
  then
;

: .heap		\ -- ; display heap info
( *G Walk the heap displaying block information. )
  0 0 0 0  startofheap			\ -- freecnt freesz usedcnt usedsz curr
  base @ >r  hex
  cr ." Heap is " sizeofheap u. ." bytes, at address " startofheap u.
  cr
  cr ." Addr  Size  Type"
  begin
    cr  dup u.  dup blocksize @ u.		\ display address and size
    >r						\ save current block
    r@ FreeBlock? if
      ."  FREE"
      2swap
      r@ blocksize @ +  swap 1+ swap		\ update free space info
      2swap
    endif
    r@ AllocBlock? if
      ."  USED"
      r@ blocksize @ +  swap 1+ swap		\ update used space info
    endif
    r@ LastBlock? if
      ."  LAST"
    endif
    r> dup LastBlock? 0=
   while
    SkipBlock
  repeat
  drop
  cr
  cr u. ." bytes used in " u. ." blocks"
  cr u. ." bytes free in " u. ." blocks"
  r> base !
;

: heapok?	\ -- t/f ; check heap
( *G Walk the heap and return TRUE if the heap is "well". )
  startofheap
  begin					\ -- current
    dup LASTINFO u> if
      cr ." Past end of heap at " u.
      false  exit
    then
    dup LastBlock?
    over LASTINFO <>
    and if
      cr ." Misplaced end-of-heap marker at " u.
      false  exit
    then
    dup FreeBlock? 0=
    over AllocBlock? 0= and
    over LastBlock? 0= and if
      cr ." Bad heap marker at " u.
      false  exit
    then
    dup LastBlock? 0=
   while
    SkipBlock
  repeat
  drop  true
;
[then]


\ ******
\ *> ###
\ ******


