\ ANS memory management extensions for 32 bit targets
\ Use HEAP32.FTH for the documentation.

((
Copyright (c) 1998, 1999, 2000, 2003, 2009, 2011, 2018
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
De-local and Forthify.

Change History
==============
20190509 BJC010 Fixed HEAPOK
20180702 SFP009 Recoded .HEAP to force short branches.
20110719 SFP008 Added equate ENDOFHEAP.
20090530 SFP007 Made HEAP32.FTH the documentation master.
20030331 SFP006 Added semaphore locking option.
20030315 SFP005 Updated ALLOCATE and took advantage of signed
		size arithmetic again for better code on some CPUs.
20030312 SFP004 Revamped .HEAP to ease leak checking.
20000321 SFP003 Fixed more signed/unsigned issues. These are all benign
                in this code, as the maximum allocation is 0x00FFFFFF,
                but changing it makes the code better to read and maintain.
19990315 SJC001 Fixed sign bug in size. Must treat the requested size as
                unsigned in ALLOCATE.
19980803 MSD001 Correct the returned address when RESIZE fails.  The ANS spec
         MSD001 says that in this situation, RESIZE should return a
         MSD001 (non-zero) return code and the *original* block pointer
         MSD001 address.  This code used to return the block pointer of the
         MSD001 failed ALLOCATE when the block was being increased in size.
         MSD001 Since ALLOCATE returns an undefined value, RESIZE *cannot*
         MSD001 use it as the block size to return when trying to increase
         MSD001 the block size and ALLOCATE fails.

we use 2 equates during compilation
to allocate a contiguous block of RAM for the heap
 STARTOFHEAP is the start address of the heap
 SIZEOFHEAP is the max size of RAM available for the heap

the heap is controlled using a single cell (32 bits) per block
this cell is used as two parts:

 bits 31-24 = flag
 bits 23-0 = #bytes

the flag indicates the status of the block:
      EE = End
      FF = Free
      AA = Allocated
      others may be added later for type management

the heap MUST be initialised before use by calling init-heap

status returned is 0 for OK, <> 0 is a fail code
))


\ =======
\ *! heap
\ *T Heap Memory Allocation
\ =======

\ ******************
\ *S Heap definition
\ ******************
\ *P The heap is allocated from a predefined section of memory.
\ ** Facilities are  provided for user expansion of the heap to
\ ** mass storage, although the current code makes no provision
\ ** for page management. When the heap is initialised, a free
\ ** block and an end block are created. The end block is of zero
\ ** size, and is used only as a marker. The address returned by
\ ** *\fo{ALLOCATE} and *\fo{RESIZE} is the address of the first
\ ** data byte, as is the address consumed by *\fo{FREE}.

\ *P The heap *\b{must} be initialised before use by calling
\ ** *\fo{INIT-HEAP}. Heap access words return status=0 for
\ ** success, and status<>0 for error.

\ *P Two equates are required during compilation to allocate a
\ ** contiguous block of RAM for the heap.
\ *E   STARTOFHEAP is the start address of the heap
\ **   SIZEOFHEAP is the size of the RAM for the heap

\ *P There are two versions of this code provided. *\i{Heap32.fth}
\ ** is provided for 32-bit targets and is optimised for the VFX
\ ** code generator. *\i{Heap16.fth} is for 16 bit targets, and
\ ** is optimised for code density.

\ ==============================
\ *N 32 bit targets - HEAP32.FTH
\ ==============================
\ *P The heap is controlled using a single cell per block.
\ ** This information is used in two parts:
\ *E bits 31..24: $EE - End, $FF - Free, $AA - Allocated
\ ** bits 23..0: 24 bits for number of data bytes in block.
\ *P A consequence of this is that the maximum block size that
\ ** can be allocated is 16Mb-1 bytes.

\ *P If you use a pre-emptive scheduler or need to use the heap
\ ** words inside interrupt routines, you must define suitable
\ ** heap lock and unlock words and set the equate *\fo{LOCKHEAP?} to
\ ** non-zero.
\ *D LockHeap=0 no heap locking
\ *D LockHeap=1 heap locking by turning off interrupts
\ *D LockHeap=2 heap locking by semaphore.

\ ==============================
\ *N 16 bit targets - HEAP16.FTH
\ ==============================
( *P The heap is controlled using two cells per block. )
( ** This information is used in three parts: )
( *E cell = #bytes, number of bytes in this block )
( ** cell = flag, split between a four bit and a 12 bit field )
( *P The top four bits of the flag are used to indicate the block type, where )
( ** $E = End, $F = Free, $A = Allocated. Others may be added later for type )
( ** management. )
( *P The bottom 12 bits of the flag are currently unused, and should be set to )
( ** zero. )


\ **********
\ *S Gotchas
\ **********
\ *P The heap routines must be protected if they are to be used both
\ ** in normal code and in interrupts. In this case the code must be
\ ** modified to be interrupt safe, but this may have a significant
\ ** impact on interrupt latency. Examples may be found in
\ ** *\i{heap32.fth}.

\ *P If you use a pre-emptive scheduler, you must also define
\ ** suitable heap lock and unlock routines.

\ *P The equate *\fo{LOCKHEAP?} controls the usage:
\ *D LockHeap=0 no heap locking
\ *D LockHeap=1 heap locking by turning off interrupts
\ *D LockHeap=2 heap locking by semaphore.


\ ***********
\ *S Glossary
\ ***********
( *P The glossary does not include all the factors used in the )
( ** code. If you are interested in the implementation, please read the )
( ** sources. )

only forth definitions
decimal

[defined] environment [if]
  also environment definitions
  TRUE constant MEMORY-ALLOC
  previous definitions
[then]


$00FFFFFF equ MASKSIZE
$FF000000 equ MASKTYPE

$EE000000 equ LASTBLOCK			\ marker for end of heap
$FF000000 equ FREEBLOCK                 \ marker for free block on heap
$AA000000 equ ALLOCBLOCK                \ marker for allocated block



0 equ LockHeap?	\ -- n

LockHeap? 0 = [if]	\ no heap locking
compiler
: InitHeapLock	;	\ --
: LockHeap	;	\ --
: UnlockHeap	;	\ --
target
[then]

LockHeap? 1 = [if]	\ interrupt locking
compiler
: InitHeapLock	;	\ --
: LockHeap	[i  ;	\ -- ; R: -- x
: UnlockHeap	i]  ;	\ -- ; R: x --
target
[then]

LockHeap? 2 = [if]	\ semaphore locking
Semaphore Heap-Access
compiler
: InitHeapLock	0 1 Heap-Access 2!  ;	\ --
: LockHeap	Heap-Access request  ;	\ --
: UnlockHeap	Heap-Access signal  ;	\ --
target
[then]


\ The following constant is the minimum *total* size occupied by a block. This
\ size includes overheads generated by alignment restrictions as well as those
\ incurred by the fact that the size and status of a block have to be stored
\ somewhere.
1 cells equ /HEAPINFO				\ size of heap information
1 aligned /HEAPINFO + equ MINIMUMHEAPENTRYSIZE	\ minimum total block size


sizeofheap buffer: STARTOFHEAP	\ -- addr
\ *G The heap memory is defined at compile time.
STARTOFHEAP sizeofheap + equ ENDOFHEAP	\ -- addr
\ *G The end+1 of the heap.

: init-heap	\ --
( *G The heap is initialised by creating 2 blocks. )
( ** Block 1 starts at the beginning and is marked as a free block. )
( ** Block 2 Is a null marker at the end of heap space. )
  InitHeapLock

  startofheap sizeofheap erase          \ erase heap memory

  sizeofheap /HEAPINFO 2* -		\ calculate free block size
  MASKSIZE and  FREEBLOCK or		\ mask in FREEBLOCK indicator
  startofheap !                         \ write header

  LASTBLOCK				\ end of heap marker
  startofheap sizeofheap + /HEAPINFO - ! \ to last block in heap
;


\ ******************
\ GARBAGE-COLLECTION
\ ******************

-short-branches

: compact	\ --
  startofheap
  begin
    \ skip past any allocated blocks
    begin				\ -- *curr
      dup @ MASKTYPE and ALLOCBLOCK =
     while
      dup @ MASKSIZE and + /HEAPINFO +	\ step to next block
    repeat
    dup @ MASKTYPE and FREEBLOCK =	\ are we on a free block?
   while
    \ scan for a new non-free block
    dup
    begin				\ -- *curr *next
      dup @ MASKSIZE and + /HEAPINFO +	\ move on to next block
      dup @ MASKTYPE and FREEBLOCK <>	\ until not FREE
    until				\ -- *curr *next'
    \ *curr = current free block
    \ *next = next non free block
    tuck				\ preserve *next for next iteration
    over - /HEAPINFO -			\ -- *next *curr sizefree
    MASKSIZE and FREEBLOCK or		\ mark as free block
    swap !
  repeat
  drop
;

+short-branches


[-short-branches
: SplitBlk	\ size *blk -- ; split block, mark top as free
  over 2 cells +			\ split above this size
  over @ MASKSIZE and			\ current block size
  u< if					\ split?
     \ Size of new free block (top section of *blk) is current
     \ freesize -CELL since we now need a new control CELL
     2dup @ MASKSIZE and swap cell+ -	\ new free block size
     FREEBLOCK or			\ -- size *blk sizef ; mask in FREEBLOCK indicator
     2 pick 2 pick + /HEAPINFO + !	\ -- size *blk ; set second block control info
     swap ALLOCBLOCK or  swap !		\ set new control info
  else
     nip				\ discard size
     dup @ MASKSIZE and			\ use existing size
     ALLOCBLOCK or  swap !		\ mark allocated
  then
;
short-branches]


\ ********************
\ ALLOCATE FREE RESIZE - the ANS words
\ ********************

[-short-branches
: allocate	\ #bytes -- addr status
( *G Attempt to allocate some memory from the heap. )
( ** Walk the heap looking for a single big enough block. )
( ** If the block is larger than than required split it into two blocks. )
( ** Allocate part or all of the free block. Status=0 for success. )
  aligned dup MASKSIZE u>		\ check #bytes, we can use
  if  -1  exit  endif			\ signed arithmetic on #bytes
  \ -- #bytes
\ Find a big enough free block, fail at LASTBLOCK
  LockHeap				\ SFP006
  startofheap				\ -- #bytes start
  begin
    dup @ MASKTYPE and LASTBLOCK <>	\ continue while not LASTBLOCK
   while				\ -- #bytes *blk
    dup @ MASKTYPE and FREEBLOCK = if	\ if block is free
      2dup @ MASKSIZE and <= if		\ and big enough
        tuck SplitBlk			\ split the block to alloc+free
	/HEAPINFO +  0 			\ return pointer and 0
	UnlockHeap  exit		\ SFP006
      then
    then
    dup @ MASKSIZE and + cell+		\ step to next block
  repeat
  2drop  0 -1
  UnlockHeap				\ SFP006
;
short-branches]


: free		\ addr -- status
\ *G Attempt to free a heap block. Status=0 for success.
\ ** If *\i{addr} is zero, no action is taken and zero is returned.
  dup if
    LockHeap				\ SFP006
    /HEAPINFO -				\ step back to control block
    dup @ MASKTYPE and ALLOCBLOCK = if	\ if block was allocated
      dup @
      MASKSIZE and  FREEBLOCK or	\ reset type to FREE
      swap !
      0					\ GOOD
      compact
    else
      drop  -1				\ failed
    then
    UnlockHeap				\ SFP006
  endif
;

\ RESIZE
\
\ Try and re-size an allocated block to a new size.  In the following
\ comments, "new size" refers to the new size requested by the user *after*
\ alignment requirements have been taken into consideration.
\
\ There are three different cases that we have to consider:
\
\       The new size is less than or equal to the old size minus the minimum
\       size taken by a heap entry.  In this case, we shrink the block size
\       making the free space into a free block.  We return the *existing*
\       block pointer because the user's data hasn't been moved.
\
\       The new size is *greater* than the old block size.  Since the user
\       has requested more space than is available in the current block, we
\       have to allocate a *new* block, copy the user's data into it and
\       then free the *old* block.  We return the pointer to the
\       newly-allocated block, of course.
\
\       If neither of the above cases apply, then we just return the
\       existing pointer because there is not enough un-used space in the
\       block to create a new (free) heap block.

[-short-branches
\ works, ARM=396 bytes
: resize	{ *ptr newsize | currsize -- *newptr ior }
( *G Try to resize an allocated block to a new size, allowing for alignment. )
( ** If the existing memory block is not big enough, the data will be copied )
( ** to a new block, and the returned addr2 will not be the same as addr1. )
( ** Status=0 for success. )
  newsize aligned -> newsize            \ align new size
  *ptr cell- @ MASKSIZE and -> currsize \ get current block size
  newsize currsize MINIMUMHEAPENTRYSIZE - u> 0=  \ Will we have a block's worth
  if                                    \ of free space after resizing?
    LockHeap				\ SFP006
    newsize MASKSIZE and                \ mask newsize
    ALLOCBLOCK or *ptr cell- !          \ and store as size of curblk
    currsize newsize - cell -           \ calc new free size for 2nd
    MASKSIZE and FREEBLOCK or           \ mask it up
    *ptr newsize +                      \ ptr to new block
    !                                   \ store it
    *ptr 0                              \ rc's
    UnlockHeap  exit			\ SFP006
  endif
  newsize currsize u> if		\ Must we allocate bigger block
    newsize allocate 0= if              \ get a new block
      dup >r                            \ preserve new ptr
      *ptr swap currsize move           \ move existing data to new blk
      *ptr free drop                    \ free old block now
      r> 0                              \ flag as ok
    else
      drop *ptr -1                      \ failed to alloc new big block \ MSD001
    then
  else                                  \ No space to generate a new
    *ptr 0                              \ free block - use current block.
  then
;
(( In progress
: resize	\ *ptr newsize -- *newptr ior
  aligned dup MASKSIZE u> if		\ align and check size
    drop -1  exit
  endif
  LockHeap				\ SFP006
  swap /HEAPINFO -			\ -- newsize *blk
  dup @ MASKSIZE and >r			\ get current size
  over r@ u< if				\ new < current?
    tuck SplitBlk
    /HEAPINFO + 0
    r> drop  unlockHeap  exit		\ SFP006
  then
  over r@ = if				\ new = current?
    nip  /HEAPINFO + 0
    r> drop  UnlockHeap  exit		\ SFP006
  then					\ -- newsize *blk
  \ larger
  swap allocate if			\ -- *blk addr
    drop
    /HEAPINFO +  -1			\ failed
    r> drop  UnlockHeap  exit		\ SFP006
  then
  swap /HEAPINFO + 2dup			\ -- addr *data addr *data
  swap r> move				\ -- addr *data ; copy data
  free drop				\ free old block
  0
  UnlockHeap
;
))
short-branches]

: size		\ *ptr -- currsize|-1
\ *G Return the size of an allocated block or -1 if there's an error.
  /HEAPINFO -				\ control block
  @ dup  MASKTYPE and  ALLOCBLOCK = if	\ allocated?
    MASKSIZE and			\ return size
  else
    drop -1
  then
;


\ **************
\ *S Diagnostics
\ **************

heap-diags? [if]

[-short-branches
\ SFP009...
: .lastBlk	\ size type start -- size type
  CELL+ .dword  ."       LAST     " over .dword
;
: .freeBlk	\ size type start -- size type
  CELL+ .dword  ."       FREE     " over .dword
;
: .usedBlk	\ size type start -- size type
  CELL+ .dword  ."       USED     " over .dword
;
: .badBlk	\ size type start -- size type
  CELL+ .dword  ."       LAST     " over .dword
;
\ ...SFP009
: .heap		\ --
\ *G Walk the heap displaying block information.
  { | *curr freecnt freesz usedcnt usedsz -- ; SFP004 }
  cr
  ." Heap size is " sizeofheap . ." bytes at " startofheap .dword
  cr
  startofheap -> *curr
  0 -> freecnt
  0 -> freesz
  0 -> usedcnt
  0 -> usedsz

  cr ." BlockAddr   BlockType   BlockSize "
  begin
    *curr @ MASKSIZE and                \ get block size and type
    *curr @ MASKTYPE and		\ -- size type
    case  dup
      LASTBLOCK of
        over CELL + usedsz + -> usedsz
        cr  *curr .lastBlk		\ SFP009
      endof
      FREEBLOCK of
        over CELL + freesz + -> freesz
        freecnt 1+ -> freecnt
        cr  *curr .freeBlk		\ SFP009
      endof
      ALLOCBLOCK of
        over CELL + usedsz + -> usedsz
        usedcnt 1+ -> usedcnt
        cr  *curr .usedBlk		\ SFP009
      endof
        cr  *curr .badBlk		\ SFP009
    endcase
    swap *curr + cell+ -> *curr         \ point *curr to next block
\    LASTBLOCK =			\ finished when hit LASTBLOCK
\ [ .ws .sequence ]
\  until
    LASTBLOCK <>			\ finished when hit LASTBLOCK
   while
  repeat

  cr
  cr freesz . ." bytes free in " freecnt . ." blocks"
  cr usedsz . ." bytes used in " usedcnt . ." blocks"
  cr freesz usedsz + . ." bytes total"
  cr sizeofheap . ." is SIZEOFHEAP"
;
short-branches]


: heapok?	\ -- t/f
\ *G Walk the heap and return TRUE if the heap is "well".
  startofheap
  begin
    dup endofheap cell- u> if		\ SFP008
      cr ." Past end of heap at " .dword
      false exit
    then
    dup @ LASTBLOCK = if
      endofheap cell- =			\ SFP008 ; Last block must be at *end* of heap.
      dup not if
        cr ." Misplaced end-of-heap marker"
      endif
      exit
    then
    dup @ MASKTYPE and
    dup FREEBLOCK =
    swap ALLOCBLOCK =
    or not if
      cr ." Bad heap marker at " .dword
      false exit			\ BJC010
    then
    dup @ MASKSIZE and + cell+          \ Point to next block.
  again
;

[then]


\ ******
\ *> ###
\ ******


