\ SVDconv.fth - Convert SVD files to Forth

((
Copyright (c) 2016, 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)

Please note that MPE files have hard tabs set to 8. Hence, tabs
are at columns 1, 9, 17, 25 ...

To do
=====

Change history
==============
20230524 SFP003 Updated description output for RP2040.
20230520 SFP002 Enlarged buffers and mede the word-counted.
20190228 SFP001 Updated for STM32 oddity.
))

only forth definitions
decimal

\ ===========
\ *! svdconv
\ *T SVD file converter
\ ===========
\ *P The code here converts ARM CMSIS System View Description
\ ** (SVD) files to a set of Forth equates for a cross compiler.
\ ** These equates are in the same format as the existing "Special
\ ** Function Register" files usually named *\i{sfr<cpu>.fth} in
\ ** the *\i{Cortex} target directory.
\ ** Processing SVD files automatically to produce Forth
\ ** source significantly reduces the effort involved in porting
\ ** Forth to a CPU. This code requires the use of the file
\ ** *\i{XML.fth} supplied with VFX Forth. In the cross compiler,
\ ** the files are in the *\i{ArmCortex/SVDparser} or
\ ** *\i{compiler/SVDparser} directory, depending on the edition
\ ** of the compiler.
\ *E   include XML.fth
\ **   include SVDconv.fth

\ *P The converted SVD (XML) to Forth is output using *\fo{TYPE}
\ ** and *\fo{EMIT} and friends. This can be copied to an editor
\ ** or the output redirected to a file. See *\fo{SVDconv} at the
\ ** end of this chapter if you want the output sent to a file.
\ ** Otherwise just *\fo{include} the SVD file.
\ *C   include ATSAM4SD32C.svd
\ *C   svdconv ATSAM4SD32C.svd  sfrATSAM4SD32C.raw.fth

\ *P Because of lack of structure namespaces in Forth, most
\ ** register names will need a prefix to identify the peripheral
\ ** the register applies to. In the SVD specification, this is
\ ** achieved using the *\i{prependToName} tag, e.g.
\ *C   <prependToName>HSMCI_</prependToName>
\ *P for an MMC/SD card peripheral. MPE standard practice is
\ ** to use a short prefix such as "mmc.". To use these, add a
\ ** tag line such as
\ *C   <mpePrepends>mmc.</mpePrepends>
\ *P to each peripheral section before the first register definition.
\ ** Search for <peripheral> when editing the SVD file. You can select
\ ** whether or not text is added to register names using the
\ ** directives *\fo{-prepends}, *\fo{+prependC} and *\fo{+prependMPE}
\ ** below. The default is *\fo{+prependMPE}.

\ *P The resulting output may/will require a little hand editing.
\ ** Many SVD files repeat the register definitions for identical
\ ** or similar peripherals such as UARTs. By using the same prefices,
\ ** you can remove the repetitions. Providing that the values are
\ ** the same, redefinitions of *\fo{EQU}ates are harmless, albeit
\ ** noisy and inelegant. Personally, I prefer to remove the
\ ** duplicates as it reduces the size of the file.


\ ****************
\ *S Time and date
\ ****************
\ *P These functions rely on the ANS Forth word
\ ** *\fo{TIME&DATE ( -- s m h dd mm yyyy )} and the non-standard
\ ** *\fo{DOW ( -- dow, 0=Sun)} to get the day of the week.

[undefined] .Time&Date [if]

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

create months	\ -- addr
\ *G String containing 3 character text for the months.
  z", JanFebMarAprMayJunJulAugSepOctNovDec"

[-sin

: .dow		\ dow --
\ *G Display day of week.
  3 * days$ 1+ + 3 type
;

: .2r		\ n --
\ *G Display n as a two digit number with leading zeros.
  s>d <# # # #> type
;

: .4r		\ n --
\ *G Display n as a four digit number with leading zeros.
  s>d <# # # # # #> type
;

: .Time&Date	\ s m h dd mm yy --
\ *G Display the system time
\ ** The format is:
\ *C   hh:mm:ss dd Mmm yyyy
  base @ >r  decimal
  dup #100 <				\ correct two digit year clocks
  if  #2000 +  endif
  rot .2r space				\ day - generate date
  swap 1- 3 * months + 3 type space	\ month
  .4r space				\ year
  .2r ." :"			 	\ hours - generate time
  .2r ." :"				\ minutes
  .2r					\ seconds
  r> base !
; doNotSin

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT (system time) is displayed, otherwise local time
\ ** is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  if
    time&date				\ -- s m h dd mm yy
    dow .dow ." , " .time&date
  else
    systime&date			\ -- s m h dd mm yy
    sysdow .dow ." , " .time&date ."  GMT "
  endif
;

sin]

[then]


\ ***************
\ *S Data buffers and variables
\ ***************
\ *P SVD files are fairly simple as XML files go, so data is
\ ** extracted from tags and saved in global or dynamic buffers.

1024 buffer: opfile$	\ -- addr
\ *G If set, the name of the output file. 16 bit Word counted string.
32 kb buffer: description$	\ -- addr
\ *G Contents of the <description> tags. 16 bit Word counted string.
1024 buffer: name$	\ -- addr
\ *G Contents of the <name> tags. 16 bit Word counted string.
1024 buffer: revision$	\ -- addr
\ *G Contents of the <revision> tags. 16 bit Word counted string.
1024 buffer: endian$	\ -- addr
\ *G Contents of the <endian> tags. 16 bit Word counted string.
1024 buffer: mpupresent$	\ -- addr
\ *G Contents of the <mpupresent> tags. 16 bit Word counted string.
1024 buffer: fpupresent$	\ -- addr
\ *G Contents of the <fpupresent> tags. 16 bit Word counted string.
1024 buffer: nvicPrioBits$	\ -- addr
\ *G Contents of the <nvicPrioBits> tags. 16 bit Word counted string.
1024 buffer: vendorSystickConfig$	\ -- addr
\ *G Contents of the <vendorSystickConfig> tags. 16 bit Word counted string.
1024 buffer: prepend$	\ -- addr
\ *G Holds the text prepended to each register name. This text
\ ** is cleared by each *\i{peripheral} tag. 16 bit Word counted string.
1024 buffer: clusName$	\ -- addr
\ *G Holds <name> from last <cluster>. 16 bit Word counted string.

0 value valueData	\ -- int
\ *G Integer from the contents of the <value> tag.
0 value dimData	\ -- int
\ *G Integer from the contents of the <dim> tag.
0 value inPeriph?	\ -- flag
\ *G True between <peripheral> and </peripheral>.
0 value PeriphHead?	\ -- flag
\ *G True when the peripheral header has been written.
0 value BaseAddrValue	\ -- addr
\ *G Peripheral base address from the <baseAddress> tag.
0 value AddrOffset	\ -- offset
\ *G Offset from the <AddressOffset> tag.
0 value ClusterOffset	\ -- addr
\ *G Offset from the <AddressOffset> tag inside a cluster.
0 value InReg?		\ -- flag
\ *G True between <register> and </register>.
0 value RegWritten?	\ -- flag
\ *G True when the register data has been written.
0 value InCluster?	\ -- flag
\ *G True between <cluster> and </cluster> tags.
0 value ClusterWritten?	\ -- flag
\ *G True when the cluster header data has been written.
0 value prependC?	\ -- flag
\ *G True to prepend C text to register names. This text is
\ ** provided by lines in the SVD file such as
\ *C   <prependToName>HSMCI_</prependToName>
0 value prependMPE?	\ -- flag
\ *G True to prepend MPE text to register names. This text is
\ ** provided by lines (that you must add) in the SVD file such as
\ *C   <mpePrepends>mmc.</mpePrepends>

[-sin
: -contents	\ --
\ *G Discard the contents of the last tag.
  contents dup sdrop snew
;

: saveTagData	\ spad dest --
\ *G Save the tag data (as given by the spad), to a given buffer.
  swap stos  ( cr 2dup type )  -white $FFFE umin rot wplace	\ SFP002
;

: saveContents	\ dest --
\ *G Save the last tag's contents to a buffer.
  contents swap saveTagData  -contents
;

[undefined] w$. [if]
: w$.	( wbuff -- ) wcount type  ;
[then]

: contents>val	\ -- int
\ *G Convert the last tag's contents to an integer.
  contents stos -white			\ -- caddr len
  isInteger? case
    2 of  drop  endof
    1 of        endof
      1 abort" Cannot convert number"
  endcase
  -contents
;

: \n?		\ caddr len -- caddr len flag
\ *G Return true if the string starts with '\n'.
  over c@ '\' =  2 pick 1+ c@ 'n' = and  ;

: ?ignLineFeed	\ caddr len -- caddr' len
\ *G Step over \n, ignore following LF or CRLF.
  over c@  case
    $0A of  1 /string  endof
    $0D of  2 /string  endof
  endcase
;

: .periphChar	\ caddr len -- caddr' len'
  \n? if
    2 /string  0 max  ?ignLineFeed	\ step over \n, ignore LF or CRLF
    cr ." \ " skip-white
  else
    over c@ emit  1 /string
  then
;

: .periphDesc	\ buffer len --
\ *G Output a string translating \n pairs.
  begin  dup while  .periphChar  repeat 2drop  ;

: ?.PeriphHead	\ -- ; SFP002
\ *G Output the peripheral header if not already done.
  InPeriph? if
    PeriphHead? 0= if
      cr cr
      cr ." \ ====================================== "
      cr ." \ "name$ w$. ."  - " description$ wcount .periphDesc
      cr ." \ ====================================== "
      cr
      cr ." $" BaseAddrValue .dword ."  equ _" name$ w$.
      -1 to PeriphHead?
    endif
  endif
;

: ?.%s		\ --
\ *G Output the value of the <dim> tag.
  dimData if
    base @ >r decimal  ."  %s=" dimData .  r> base !
  endif
;

: .regChar	\ caddr len -- caddr' len'
  \n? if
    2 /string  0 max  ?ignLineFeed	\ step over \n, ignore LF or CRLF
    cr 33 >pos  ." \ " skip-white
  else
    over c@ emit  1 /string
  then
;

: .regDesc	\ buffer len --
\ *G Output a string translating \n pairs.
  begin  dup while  .regChar  repeat 2drop  ;

: ?.Register	\ -- ; SFP002
\ *G Output the register line if not already done.
\ *C  $xx equ regname   \ description
  InReg? if
    RegWritten? 0= if
      cr ."   $" CLusterOffset AddrOffset + .word ."  equ "
      prepend$ w$. name$ w$. space
      33 >pos ." \ " ?.%s description$ wcount .regDesc
      -1 to RegWritten?
    endif
  endif
;

: ?.Cluster	\ -- ; SFP002
\ *G Output the cluster line if not already done.
\ *C  \ Cluster: <name> - <description>
  InCluster? if
    ClusterWritten? 0= if
      AddrOffset -> ClusterOffset
      cr ."   \ <Cluster> " name$ w$.  ?.%s  ."  - " description$ w$.
      -1 to ClusterWritten?
    endif
  endif
;

: .MPEhead	\ --
\ *G Display an MPE header.
  cr ." MicroProcessor Engineering"
  cr ." 133 Hill Lane"
  cr ." Southampton SO15 5AF"
  cr ." England"
  cr
  cr ." tel:   +44 (0)7803 903 612"
  cr ."        +34 649 662 974"
  cr ." email: mpe@mpeforth.com"
  cr ."        tech-support@vfxforth.com"
  cr ." web:   http://vfxforth.com"
  cr ."        http://www.mpeforth.com"
  cr ." Skype: mpe_sfp"
  cr cr
  cr ." This file produced by the MPE SVDconv utility"
  cr ." on " 1 .AnsiDate ." ."
;

sin]


\ ***************
\ *S Tag handling
\ ***************
\ *P Tags that are handled when encountered are words of the same
\ ** name in the *\fo{inputTags} vocabulary. There are separate
\ ** entries for the opening and closing tags. This can be useful
\ ** and is also much quieter when debugging. The tag handlers
\ ** have no overall stack effect. Many closing tags just save
\ ** the contents, others display data.

\ *[
also inputTags definitions
: description	( -- )  ;
: /description	\ --
  description$ saveContents
;
: cpu		\ --
  ." \ " opfile$ w@
  if  opfile$ w$. ." - "  then
  description$ w$.
;
: /cpu		\ --
  cr cr ." (("
  .MPEhead
  cr cr
  cr description$ w$.
  cr ." CPU: " name$ w$. ." , " revision$ w$.
  cr ."   Endian: " endian$ w$.
  cr ."   MPU: " mpupresent$ w$.
  cr ."   FPU: " fpupresent$ w$.
  cr ."   NVIC priority bits: " nvicPrioBits$ w$.
  cr ."   Vendor Systick Config: " vendorSystickConfig$ w$.
  cr ." ))"
  contents dup sdrop snew
;
: value		( -- )  ;
: /value	( -- )  contents>val to ValueData  ;
: dim		( -- )  ;
: /dim		( -- )  contents>val to dimData  ;
: name		( -- )  ;
: /name		( -- )  name$ saveContents  ;
: revision	( -- )  ;
: /revision	( -- )  revision$ saveContents  ;
: endian	( -- )  ;
: /endian	( -- )  endian$ saveContents  ;
: mpuPresent	( -- )  ;
: /mpuPresent	( -- )  mpuPresent$ saveContents  ;
: fpuPresent	( -- )  ;
: /fpuPresent	( -- )  fpuPresent$ saveContents  ;
: nvicPrioBits	( -- )  ;
: /nvicPrioBits	( -- )  nvicPrioBits$ saveContents  ;
: vendorSystickConfig	( -- )  ;
: /vendorSystickConfig	( -- )  vendorSystickConfig$ saveContents  ;
: interrupt	( -- )  ?.PeriphHead  ;
: /interrupt	\ --
  base @ >r decimal
  cr  ." #" ValueData #16 + . ." equ " name$ w$. s" _IRQn" type
  r> base !  -contents
;
: peripheral	( -- )
  1 to InPeriph?  0 to PeriphHead?  0 prepend$ c!
;
: /peripheral	( -- )  ?.PeriphHead  0 to InPeriph?  0 to PeriphHead?  -contents  ;	\ SFP001
: prependToName	( -- )  ;
: /prependToName ( -- )
  prependC? if  prepend$ saveContents  else  -contents  then
;
: mpePrepends	( -- )  ;
: /mpePrepends	( -- )
  prependMPE? if  prepend$ saveContents  else  -contents  then
;
: baseaddress	( -- )  ;
: /baseaddress	( -- )  contents>val to BaseAddrValue  ;
: addressOffset	( -- )  ;
: /addressOffset ( -- )  contents>val to AddrOffset  ;
: registers	( -- )  ?.PeriphHead  ;
: /registers	( -- )  ( key drop )  -contents  ;
: register	( -- )  ?.Cluster  1 to inReg?  0 to RegWritten?  0 -> dimData  ;
: /register	( -- )  ?.Register  0 to inReg?  0 to RegWritten?  -contents  ;
: cluster	( -- )
  InCluster? abort" Nested clusters not supported"
  1 -> InCluster?  0 -> ClusterWritten?
  0 -> AddrOffset  0 -> ClusterOffset
;
: /cluster	( -- )
  cr ."   \ </cluster>"
  0 -> InCluster?  0 -> ClusterWritten?  0 -> ClusterOffset
  -contents
;

: fields	( -- )  s" /fields>" skipPast  ;
((
: fields	( -- )  ?.Cluster ?.Register  ;
: /fields	( -- )  -contents  ;
))
previous definitions
\ *]


\ ***************
\ *S File version
\ ***************
\ *P The word *\fo{SVDconv} below redirects output to a named
\ ** file.

: -prepends	\ --
\ *G Do not prepend text to register names.
  0 prepend$ w!  0 -> prependC?  0 -> prependMPE?
;
: +prependC	\ --
\ *G The C names are prepended to register names. These are found
\ ** in the SVD file in the form:
\ *C   <prependToName>HSMCI_</prependToName>
  1 -> prependC?  0 -> prependMPE?
;
: +prependMPE	\ --
\ *G The MPE names are prepended to register names. These should be
\ ** in the SVD file in the form:
\ *C   <mpePrepends>mmc.</mpePrepends>
\ *P Note that you will have to add these lines to the SVD file
\ ** yourself.
  0 -> prependC?  1 -> prependMPE?
;
+prependMPE

filebuff: opdev	\ -- sid

also system
: SVDconv	\ -- ; SVDconv <ifile> <ofile>
\ *G Includes an SVD file and writes the output to a named file.
\ ** Use in the form:
\ *C   SVDconv input.svd  output.fth
  {: | ipfile[ 256 ] opfile[ 256 ] -- :}
  op-handle @ >r			\ preserve current output
  bVerboseInclude? >r			\ turn noise off
  0 -> bVerboseInclude?
  getPathSpec ipfile[ place
  getPathSpec opfile[ place
  opfile[ count delete-file drop	\ delete previous
  opfile[ count r/w opdev open-gio abort" can't create output file" drop
  opdev op-handle !			\ set new output
  ipfile[ count ['] included catch
  opdev close-gio drop
  r> -> bVerboseInclude?		\ restore noise level
  r> op-handle !			\  restore output
  throw
;
previous

0 to .UnknownXML?			\ ignore unknown tags


\ ======
\ *> ###
\ ======

decimal
