\ Bin2Hex.fth - miscellaneous output file formats

((
Copyright (c) 2005, 2006, 2010, 2023
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   htpp://www.mpeforth.com

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20230530 SFP004 Added UF2 output format
20210604 SFP003 Added Verilog ReadMemH format
20100105 SFP002 Added ELF output.
20060123 NSL001 Fixed SEND-S1
))

\ *P Output format selectors are:
\ *D NoHex	Do not generate hex files (default).
\ *D HEX-I16	Intel Hex for 64k bytes maximum as used for 8-bit CPUs
\ *D HEX-I32	Intel Hex for 32 bit linear addresses
\ *D HEX-ISEG	Intel segmented format for 80x86 real mode - not supported
\ *D HEX-S19	Motorola S19 format
\ *D HEX-S28	Motorola S28 format
\ *D HEX-S37	Motorola S37 format
\ *D ReadMemH   Verilog ReadMemH format in 32 bit units.
\ *D UF2-format UF2 format as used for Raspberry Pi Pico and RP2040.

\ *P The initial execution address can be set for S28, S37 and
\ ** ELF formats by:
\ *C  <addr> SetBoot


\ ***********************
\ +S Additional VFX tools
\ ***********************

[undefined] buffer: [if]
: buffer:	\ n -- ; -- addr
  create  allot&erase
;
[then]


vocabulary HexFiles.voc
also sections.voc  also HexFiles.voc definitions


\ **************
\ +S Global data
\ **************

((
defer send-blocks	\ -- ; write out the data

variable host-address	\ -- addr ; holds image address in host
variable host-length	\ -- addr ; holds image length
variable target-address	\ -- addr ; holds target image load address
variable exec-address	\ -- addr ; holds target start address

#16 buffer: $ext-name	\ -- addr ; holds output file name extension WITH '.'
))

variable checksum	\ -- addr ; holds checksum so far
variable length		\ -- addr ; holds number of bytes/data to send
variable record-type	\ -- addr ; holds record type

variable record#	\ -- addr ; holds record number counter

$010 constant b/line	\ -- n ; no of data bytes output in a line


\ *******************
\ +S Common interface
\ *******************

: SetBoot	\ addr --
  exec-address !
;


\ ****************
\ Hex output tools
\ ****************

: ImageRange	\ -- end+1 start ; host memory for target image
  host-address @ host-length @ bounds
;

: bin-ascii	\ nibble -- char ; convert nibble to ascii char
  $0F and  dup 9 >
  if  $037 +  else  $030 +  endif	\ a..f or 0..9
;

: byte-ascii	\ byte -- lsc msc ; convert byte to ascii
  dup bin-ascii  swap $010 /  bin-ascii
;

: sum           \ data -- ; advance checksum
  checksum +!
;

: byte-target	\ byte -- ; send 8 bit
  $0FF and  dup sum  byte-ascii emit emit
;

: word-target	\ word -- ; send 16 bit data, msb first
  $FFFF and 0 $0100 um/mod  byte-target  byte-target
;

: dword-target	\ dword -- ; send 32 bit data
  dup #16 rshift word-target word-target
;

: 24b-target	\ 24b -- ; send 24 bit data
  dup #16 rshift byte-target word-target
;

: data		\ -- ; send data
  host-address @ length @ bounds
  ?do  i c@ byte-target  loop		\ send each byte in turn
;

: send-Icheck	\ -- ; send intel checksum
  checksum @ negate byte-target
  checksum off
;


\ *******************
\ *S Intel HEX format
\ *******************

((
 The records are:
  Data Record (8-, 16-, or 32-bit formats)
  End of File Record (8-, 16-, or 32-bit formats)
  Extended Segment Address Record (16- or 32-bit formats)
  Start Segment Address Record (16- or 32-bit formats)
  Extended Linear Address Record (32-bit format only)
  Start Linear Address Record (32-bit format only)

General Record Format
     | RECORD  |     LOAD     |         |         |  INFO   |         |
     |  MARK   |    RECLEN    | OFFSET  | RECTYP  |   or    | CHKSUM  |
     |  ':'    |              |         |         |  DATA   |         |
Data:  1-byte       1-byte      2-bytes   1-byte    n-bytes   1-byte
Ascii: 1            2           4         2         2n        2

Record types are:
  '00' Data Record
  '01' End of File Record
  '02' Extended Segment Address Record
  '03' Start Segment Address Record
  '04' Extended Linear Address Record
  '05' Start Linear Address Record


Data Record (8-, 16-, or 32-bit formats)
| RECORD  |     LOAD     |         |         |         |         |
|  MARK   |    RECLEN    | OFFSET  | RECTYP  |  DATA   | CHKSUM  |
|  ':'    |              |         |  '00'   |         |         |
  1-byte       1-byte      2-bytes   1-byte    n-bytes   1-byte

End of File Record (8-, 16-, or 32-bit formats)
| RECORD  |     LOAD     |         |         |         |
|  MARK   |    RECLEN    | OFFSET  | RECTYP  | CHKSUM  |
|  ':'    |    '00'      | '0000'  |  '01'   |  'FF'   |
  1-byte       1-byte      2-bytes   1-byte    1-byte

Extended Segment Address Record (16- or 32-bit formats)
| RECORD  |     LOAD     |         |         |         |         |
|  MARK   |    RECLEN    | OFFSET  | RECTYP  |  USBA   | CHKSUM  |
|  ':'    |    '02'      | '0000'  |  '02'   |         |         |
  1-byte       1-byte      2-bytes   1-byte    2-bytes   1-byte

Start Segment Address Record (16- or 32-bit formats)
| RECORD  |     LOAD     |         |         |         |         |
|  MARK   |    RECLEN    | OFFSET  | RECTYP  |  CS/IP  | CHKSUM  |
|  ':'    |    '04'      | '0000'  |  '03'   |         |         |
  1-byte       1-byte      2-bytes   1-byte    4-bytes   1-byte

Extended Linear Address Record (32-bit format only)
| RECORD  |     LOAD     |         |         |         |         |
|  MARK   |    RECLEN    | OFFSET  | RECTYP  |  ULBA   | CHKSUM  |
|  ':'    |    '02'      | '0000'  |  '04'   |         |         |
  1-byte       1-byte      2-bytes   1-byte    2-bytes   1-byte

Start Linear Address Record (32-bit format only)
| RECORD  |     LOAD     |         |         |         |         |
|  MARK   |    RECLEN    | OFFSET  | RECTYP  |  EIP    | CHKSUM  |
|  ':'    |    '04'      | '0000'  |  '05'   |         |         |
  1-byte       1-byte      2-bytes   1-byte    4-bytes   1-byte
))

: send-header	\ -- ; send line header
  [char] : emit  checksum off
;

: record-length	\ -- ; send record length
  length @ byte-target
;

: start-address	\ -- ; send start address
  target-address @ word-target
;

: type-of-record	\ -- ; send type (line/end)
  record-type @ byte-target
;

: send-record	\ -- ; send one record
  send-header                   \ the header
  record-length                 \ the length
  start-address                 \ the address
  type-of-record                \ the type
  data                          \ the data
  send-Icheck                    \ the checksum
  cr                            \ ready for next line
;

: send-last-record	\ -- ; send final record
  length off  target-address off  $01 record-type !
  send-header
  record-length
  start-address
  type-of-record
  send-Icheck
  cr
  $00 record-type !		\ data record type
  b/line length !		\ data record length
;

: send-I16	\ -- ; send the data in Intel 16 bit format
\ *G The data to send is in a buffer
  record# off                           \ start at record 0
  b/line length !
  ImageRange ?do
    1 record# +!			\ increment count
    send-record				\ send the record
    b/line host-address +!		\ increment the address
    b/line target-address +!		\ and this address
  b/line +loop
  send-last-record			\ send last record
;
assign send-i16 to-do send-blocks

: send-ulba	\ -- ; send new upper 16 bits
  send-header
  $02 byte-target  $0000 word-target  $04 byte-target
  target-address @ #16 rshift word-target
  send-Icheck  cr
;

variable last-ulba	\ -- ; holds last upper target address

: send-I32	\ -- ; send the data in Intel 16 bit format
\ *G The data to send is in a buffer
  b/line length !
  -1 last-ulba !			\ force new ULBA
  ImageRange ?do
    target-address @ last-ulba @ xor	\ target-address changed
    $FFFF0000 and if			\ in upper 16 bits
      send-ulba				\ send new ULBA
      target-address @ last-ulba !
    endif
    send-record				\ send the record
    b/line host-address +!		\ increment the address
    b/line target-address +!		\ and this address
  b/line +loop
  send-last-record			\ send last record
;

also forth definitions
: NoHex		\ -- ; select no hex output format
  ['] noop to-do send-blocks
  c" .img" $ext-name $move
;
NoHex

: HEX-I16	\ -- ; select Intel 16 bit format
  ['] send-I16 to-do send-blocks
  c" .hex" $ext-name $move
;

: HEX-I32	\ -- ; select Intel 16 bit format
  ['] send-I32 to-do send-blocks
  c" .hex" $ext-name $move
;
previous definitions


\ ********************
\ Motorola Sxx formats
\ ********************

((  from: 'man 4 srec' on some Unix box
          The general format of an S-record follows:

          +-------------------//------------------//-----------------------+
          | type | count | address  |            data           | checksum |
          +-------------------//------------------//-----------------------+

          type      A char[2] field.  These characters describe the
                    type of record (S0, S1, S2, S3, S5, S7, S8, or
                    S9).
          count     A char[2] field.  These characters when paired and
                    interpreted as a hexadecimal value, display the
                    count of remaining character pairs in the record.
          address   A char[4,6, or 8] field.  These characters grouped
                    and interpreted as a hexadecimal value, display
                    the address at which the data field is to be
                    loaded into memory.  The length of the field
                    depends on the number of bytes necessary to hold
                    the address.  A 2-byte address uses 4 characters,
                    a 3-byte address uses 6 characters, and a 4-byte
                    address uses 8 characters.
          data      A char [0-64] field.  These characters when paired
                    and interpreted as hexadecimal values represent
                    the memory loadable data or descriptive
                    information.
          checksum  A char[2] field.  These characters when paired and
                    interpreted as a hexadecimal value display the
                    least significant byte of the ones complement of
                    the sum of the byte values represented by the
                    pairs of characters making up the count, the
                    address, and the data fields.


          S0 Record  The type of record is 'S0' (0x5330).  The address
                     field is unused and will be filled with zeros
                     (0x0000).  The header information within the data
                     field is divided into the following subfields.

                                  mname      is char[20] and is the
                                             module name.
                                  ver        is char[2] and is the
                                             version number.
                                  rev        is char[2] and is the
                                             revision number.
                                  description is char[0-36] and is a
                                             text comment.

                     Each of the subfields is composed of ASCII bytes
                     whose associated characters, when paired,
                     represent one byte hexadecimal values in the case
                     of the version and revision numbers, or represent
                     the hexadecimal values of the ASCII characters
                     comprising the module name and description.
          S1 Record  The type of record field is 'S1' (0x5331).  The
                     address field is intrepreted as a 2-byte address.
                     The data field is composed of memory loadable
                     data.
          S2 Record  The type of record field is 'S2' (0x5332).  The
                     address field is intrepreted as a 3-byte address.
                     The data field is composed of memory loadable
                     data.
          S3 Record  The type of record field is 'S3' (0x5333).  The
                     address field is intrepreted as a 4-byte address.
                     The data field is composed of memory loadable
                     data.
          S5 Record  The type of record field is 'S5' (0x5335).  The
                     address field is intrepreted as a 2-byte value
                     and contains the count of S1, S2, and S3 records
                     previously transmitted.  There is no data field.
          S7 Record  The type of record field is 'S7' (0x5337).  The
                     address field contains the starting execution
                     address and is intrepreted as  4-byte address.
                     There is no data field.
          S8 Record  The type of record field is 'S8' (0x5338).  The
                     address field contains the starting execution
                     address and is intrepreted as  3-byte address.
                     There is no data field.
          S9 Record  The type of record field is 'S9' (0x5339).  The
                     address field contains the starting execution
                     address and is intrepreted as  2-byte address.
                     There is no data field.
))

: send-Mcheck	\ -- ; send the checksum
  checksum @ $0FF xor byte-target  0 checksum !
;

: send-s1	\ -- ; send the next record
  ." S1"  0 checksum !
  length @ 3 + byte-target		\ #data-bytes + 1 c/s + 2 addr
  start-address				\ send start addr ; NSL001
  data send-Mcheck cr			\ send data, checksum and cr
;

: send-s0	\ --
  ." S00B00004C4C4C2E4D4F5420D2" cr	\ LLL.MOT_
\  ." S00600004844521B" cr		\ HDR
;

: send-S9	\ -- ; send the terminator record
  ." S9030000FC" cr
;

also forth definitions
: send-S19	\ -- ; send the data in Motorola S19 format
\ *G The data to send is in a buffer
  b/line length !
  send-s0
  ImageRange ?do
    send-s1				\ send data record
    b/line host-address +!		\ increment the address
    b/line target-address +!		\ and this address
  b/line +loop
  send-S9				\ send last record
;

: HEX-S19	\ -- ; select Motorola S19 output format
  ['] send-s19 to-do send-blocks
  c" .s19" $ext-name $move
;
previous definitions


\ *******************
\ Motorola S28 format
\ *******************

: send-s2	\ -- ; send the next record
  ." S2"  0 checksum !			\ send header
  length @ 4 + byte-target		\ #data-bytes + 1 len + 3 addr
  target-address @ 24b-target		\ send start addr
  data  send-Mcheck  cr
;

: send-s8	\ send terminator
  ." S8"  0 checksum !			\ send header
  $04 byte-target			\ length
  exec-address @ 24b-target		\ execution adress
  send-Mcheck cr			\ no data
;

also forth definitions
: send-S28	\ -- ; send the data in Motorola S29 format
\ The data to send is in a buffer
  b/line length !
  send-s0
  host-address @ host-length @ bounds
  ?do
    send-s2				\ send data record
    b/line host-address +!		\ increment the address
    b/line target-address +!		\ and this address
  b/line +loop
  send-s8				\ send last record
;

: HEX-S28	\ -- ; select Motorola S29 output format
  ['] send-s28 to-do send-blocks
  c" .s28" $ext-name $move
;
previous definitions


\ *******************
\ Motorola S37 format
\ *******************

: send-s3	\ -- ; send the next record
  ." S3"  0 checksum !			\ header
  length @ 5 + byte-target		\ #data-bytes + 1 len + 4 addr
  target-address @ dword-target		\ 4 byte address
  data  send-Mcheck  cr			\ send data, checksum and cr
;

: send-s7	\ -- ; send the terminator record
  ." S7"  0 checksum !			\ send header
  $05 byte-target			\ length
  exec-address @ dword-target		\ execution adress
  send-Mcheck cr			\ no data
;

also forth definitions
: send-S37	\ -- ; send the data in Motorola S39 format
\ *G The data to send is in a buffer
  b/line length !
  send-s0
  host-address @ host-length @ bounds
  ?do
    send-s3				\ send data record
    b/line host-address +!		\ increment the address
    b/line target-address +!		\ and this address
  b/line +loop
  send-s7				\ send last record
;

: HEX-S37	\ --
\ *G Select Motorola S37 output format
  ['] send-s37 to-do send-blocks
  c" .s37" $ext-name $move
;
previous definitions


\ **********
\ ELF format
\ **********

$34 CONSTANT /ELFhdr	\ -- len
\ Length of ELF header

$80 constant /ELF+PHT	\ -- len
\ *G Size of the combined ELF header and PHT.

create ELFhdr	\ -- addr
\ ELF file header with PHT containing one segment
  \ Offset $00
  $7f c,
  char E c, char L c, char F c,	\  db      0x7F, "ELF", 1, 1, 1	; e_ident
  1 c, 1 c, 1 c,
  0 c, 0 c, 0 c, 0 c, 0 c, 	\ times   9 db 0
  0 c, 0 c, 0 c, 0 c,
  \ Offset $10
  2 w,				\  dw 2 ; e_type     2=executable, 3=shared lib
  3 w,				\  dw 3 ; e_machine  3=i32/386
  1 ,				\  dd 1 ; e_version  1=current
  0 ,				\  dd entry ; e_entry    absolute entry point
  $40 ,				\  dd PHT-$$ ; e_phoff    file offset to PHT, 0=none
  \ Offset $20
  0 ,				\  dd 0 ; e_shoff    file offset to SHT, 0=none
  0 ,				\  dd 0 ; e_flags    0 for i386
  /ELFhdr w,			\  dw ehdrsize ; e_ehsize   ELF header size
  #32 w,			\  dw phdrsize ; e_phentsize PHT entry size
  1 w,				\  dw 1 ; e_phnum    #entries in PHT
  $28 w,			\  dw 0 ; e_shentsize SHT entry size
  \ Offset $30
  0 w,				\  dw 0 ; e_shnum    #entries in SHT
  0 w,				\  dw 0 ; e_shstrndx file offset to string table, 0=none
  \ Extend the size to $040 bytes
  0 , 0 , 0 ,

\ Here starts the PHT with one entry
\ PHtext: Main code (.text) and fixed data
  1 ,			\ dd 1     ; p_type     1=load into memory
  /ELF+PHT ,		\ dd 0     ; p_offset   file offset to start of segment
  0 ,			\ dd $$    ; p_vaddr    virt. address where loaded
  0 ,			\ dd $$    ; p_paddr    abs. address where loaded
  0 ,			\ dd filesize   ; p_filesz   how much loaded from file
  0 ,			\ dd filesize   ; p_memsz    segment size in meory
  7 ,			\ dd 5 ; p_flags permissions RWX, 4=read,2=write,1=exec
  $04 ,			\ dd 0x1000     ; p_align    alignment required
  \ Extend the PHT to a 64 byte boundary
  0 , 0 , 0 , 0 ,  0 , 0 , 0 , 0 ,

ELFhdr $12 + constant ELFmachine	\ -- addr
\ Host address of the ELF machine code (16 bits).
ELFhdr $18 + constant ELFentry		\ -- addr
\ Host address of the absolute entry point (32 bits).
ELFhdr $1C + constant ELFphtOff		\ -- addr
\ Host address of the file offset to the PHT (32 bits).
ELFhdr $24 + constant ELFflags		\ -- addr
\ Host address of the ELF header flags (32 bits).
ELFhdr $2A + constant ELFphentsize	\ -- addr
\ Host address of the size of an entry in the PHT (16 bits).
ELFhdr $2C + constant ELFphnum		\ -- addr
\ Host address of the number of entries in the PHT (16 bits).
ELFhdr $40 + constant PHThdr

PHThdr $08 + constant PHTvirtaddr	\ -- addr
\ Host address of the section virtual address (32 bits).
PHThdr $0C + constant PHTphysaddr	\ -- addr
\ Host address of the physical address (32 bits).
PHThdr $10 + constant PHTfilesz		\ -- addr
\ Host address of the section size in the file (32 bits).
PHThdr $14 + constant PHTmemsz		\ -- addr
\ Host address of the memory size (32 bits).

also forth definitions
: send-elf	\ -- ; send the data in ELF format
\ The data to send is in a buffer defined by variables
  exec-address @ ELFentry !		\ set execution address
  target-address @ dup			\ target memory address
    PHTvirtaddr ! PHTphysaddr !
  host-length @ aligned dup		\ target size
    PHTfilesz ! PHTmemsz !
  op-handle @ gen-handle @ >r
  ELFhdr /ELF+PHT r@ write-file drop	\ write header
  host-address @ host-length @ r> write-file drop
;

: ELF-format	\ machine flags -- ; select ELF output format
  ELFflags !  ELFmachine w!
  ['] send-elf to-do send-blocks
  c" .elf" $ext-name $move
;

: ELF-ARM	\ --
  #40 $0500:0002 ELF-Format  ;

: ELF-386	\ --
  3 0 ELF-format  ;
\  4 0 Elf-format		\ 68k ELF
\ #52 0 Elf-format		\ Coldfire ELF

previous definitions


\ ***********************
\ Verilog ReadMemH format
\ ***********************

: .hex32	\ dw --
\ Display the 32 bit long word 'dw' as an 8 digit hex number.
  base @ hex  swap
  0 <# # # # # # # # # #> type
  base !
;

: sendRMH	\ -- ; send data in ReadMemH format
  ." @" target-address @ .hex32 cr
  Target-address @ Host-length @ bounds ?do
    #16 0 do  j i + l@c(t) .hex32 space  4 +loop
    cr
  #16 +loop
;

: ReadMemH	\ -- ; select ReadMemH format
  ['] sendRMH to-do send-blocks
  c" .dat" $ext-name $move
;


\ ==========
\ UF2 format
\ ==========

variable UF2dest	\ -- addr(t)
\ Holds address in target at which the UF2 file is placed. This is
\ not always the address at which the UF2 code is executed, e.g. 
\ when the data is copied from Flash to RAM. A value of zero 
\ indicates thet the section address is the start address.
 
((
defer send-blocks	\ -- ; write out the data

variable host-address	\ -- addr ; holds image address in host
variable host-length	\ -- addr ; holds image length
variable target-address	\ -- addr ; holds target image load address
variable exec-address	\ -- addr ; holds target start address

#16 buffer: $ext-name	\ -- addr ; holds output file name extension WITH '.'
))

struct /UF2blk	\ -- len
  4 field uf2.MagicStart0	\ 0x0A324655 = "UF2\n"
  4 field uf2.MagicStart1	\ 0x9E5D5157 =
  4 field uf2.flags
  4 field uf2.TargetAddress	\ address in target
  4 field uf2.payloadsize	\ usually 256
  4 field uf2.blockno		\ 0..n-1
  4 field uf2.numBlocks		\ number of blocks in file
  4 field uf2.fileSize		\ file size or family ID or zero
  #476 field uf2.data		\ zero padded data
  4 field uf2.MagicEnd		\ 0x0AB16F30
end-struct

/UF2blk buffer: UF2block	\ -- addr
variable NumBlocks		\ -- addr ; total number of block
variable UF2family		\ -- addr ; UF2 family or 0
variable UF2flags		\ -- addr ; flags or 0
variable currblock		\ -- addr ; current block number


: initUF2block	\ hostaddr targetaddr --
  UF2block dup #512 erase  >r
  0x0A324655 r@ uf2.MagicStart0 l!
  0x9E5D5157 r@ uf2.MagicStart1 l!
  UF2flags @ r@ uf2.flags l!
             r@ uf2.TargetAddress l!
  #256       r@ uf2.payloadsize l!
  currblock @ r@ uf2.blockno l!
  Numblocks @ r@ uf2.numBlocks l!
  UF2family @ r@ uf2.filesize l!
             r@ uf2.data #256 cmove
  0x0AB16F30 r> uf2.MagicEnd l!
;

: sendUF2block	\ host target --
  initUF2block
  UF2block #512 op-handle @ gen-handle @ write-file drop
;

also forth definitions
: send-UF2	\ -- ; send data in UF2 format
  target-address @ >r			\ target address start
  UF2dest @ if				\ actual flash address ?
    UF2dest @ target-address !
  then
  host-address @			\ Host start
  Host-Length @ #255 + #-256 and 	\ size 256 byte aligned
  dup 8 rshift numBlocks !
  0 currblock !
  bounds ?do
    i target-address @ sendUF2block
    #256 target-address +!  currblock incr
  #256 +loop
  r> target-address !
;

: UF2-format	\ -- ; select UF2 format
  ['] send-UF2 to-do send-blocks
  c" .uf2" $ext-name $move
;

: set-UF2dest	\ addr(t) --
  UF2dest !
;

: set-UF2family	\ family --
  dup UF2family ! 
  if  $2000 UF2flags or!
  else  $2000 UF2flags bic!
  then
;

: set-UF2-RP2040	\ --
  $E48BFF56 set-UF2family  ;

previous definitions


\ ***********************
\ Directives for the user
\ ***********************

[defined] cc/i [if]
also cc/i

: NoHex		cc/i  NoHex       0 -> SaveAllSecData?  ;
: Hex-I16	cc/i  Hex-I16     0 -> SaveAllSecData?  ;
: Hex-I32	cc/i  Hex-I32     0 -> SaveAllSecData?  ;
: Hex-S19	cc/i  Hex-S19     0 -> SaveAllSecData?  ;
: Hex-S28	cc/i  Hex-S28     0 -> SaveAllSecData?  ;
: Hex-S37	cc/i  Hex-S37     0 -> SaveAllSecData?  ;
: ELF-format	cc/i  ELF-format  0 -> SaveAllSecData?  ;
: ELF-ARM	cc/i  ELF-ARM     0 -> SaveAllSecData?  ;
: ELF-386	cc/i  ELF-386     0 -> SaveAllSecData?  ;
: ReadMemH	cc/i  ReadMemH    1 -> SaveAllSecData?  ;
: UF2-format	cc/i  UF2-format  0 -> SaveAllSecData?  ;

: SetBoot	cc/i  SetBoot  ;
: set-UF2dest	cc/i  set-UF2dest  ;
: set-UF2family cc/i  set-UF2family  ;
: set-UF2-RP2040 cc/i  set-UF2-RP2040  ;

previous definitions
[then]

previous previous definitions

