\ CHECKSUM.FTH - Checksum generation

((
	Copyright (c) 1997, 2008, 2010, 2011, 2023
	MicroProcessor Engineering
	133 Hill Lane
	Southampton SO15 5AF
	England

	tel: (+44) 2380 631441
	email: mpe@mpeforth.com

Changes
-------
20230710 SFP006 Add Rp2040 CRC32 from Matthias Koch.
20110415 FVD005 Added two forms of XModem CRC.
20101130 SFP004 Added checksum string generation.
20101015 SFP003 Added -Checksums to disable previous checksums.
20080605 SFP002 DocGenned
         JL001  Amended SIMPLE32 to use LxC(T) for 32bit target access.
))

\ =========
\ *> interp
\ *S Checksum generation
\ =========
\ *P Checksums of various types can be generated for your code.
\ ** More than one checksum can be generated in different memory
\ ** areas. To avoid problems with forward references, checksums
\ ** are not generated until *\fo{FINIS} has finished patching
\ ** various labels and has laid the initialised RAM table.

only forth also c-c definitions
decimal


((
-------------------------------------------------------------------------------
        checksum details are kept in checksum structures, all
        the structures are held in a linked list
-------------------------------------------------------------------------------
))

list-make constant CHECKSUM-LIST

struct CHECKSUM-STRUCT
  cell field CHECKSUM.START                     \ start address
  cell field CHECKSUM.END                       \ end address
  cell field CHECKSUM.LOCATION                  \ address where to store checksum
  cell field CHECKSUM.TYPE                      \ type of checksum required
  cell field CHECKSUM.SECTION                   \ section to produce checksum for
end-struct

((
-------------------------------------------------------------------------------
        the checksum generation functions
-------------------------------------------------------------------------------
))

\ ----------------------------------------------------------------------------
\ Checksum types
  1 Constant Simple8
  2 Constant Simple16
  3 Constant Simple32
  4 Constant CCITT
  5 Constant CRC16
  6 Constant LRCC16
  7 Constant SDLC
  8 constant CRC32
  9 constant CRCxModem16
#10 constant CRCxModem16-0
#11 constant CRC32RP2040

\ ----------------------------------------------------------------------------
\ General checksum words
: Aligned?      \ addr1 addr2 #bytes -- flag
  -rot
  1+ swap -
  swap
  /mod
  drop
  0=
;

\ ----------------------------------------------------------------------------
\ Specific Checksum words
: Checksum8     \ addr1 addr2 addr3 -- ;
  { startaddr endaddr result -- }
  0 result c!c(t)
  startaddr c@c(t) result c!c(t)        \ store first byte

  endaddr 1+ startaddr 1+
  Do
    i c@c(t)                     \ get next byte
    result c@c(t)                \ get intermediate result
    +                           \ 16b result...
    result c!c(t)                \ ...but only store lower 8b
  Loop
\  result c@c(t) 2 formCRC$
;

: Checksum16    \ addr1 addr2 addr3 -- ; buffer must be a multiple of 2 bytes
  { startaddr endaddr result -- }
  startaddr endaddr 2 Aligned?  \ Is the buffer len a multiple of 2 bytes?
  If
    0 result w!c(t)
    startaddr w@c(t) result w!c(t)      \ store first word

    endaddr 1+ startaddr 2+
    Do
      i w@c(t)                   \ get next word
      result w@c(t)              \ get intermediate result
      +                         \ 32b result...
      result w!c(t)              \ ... but only store lower 16b
    2 +Loop
\    result w@c(t) 4 formCRC$
  Else
    ." Error! Buffer must be a multiple of 2 bytes!" cr
  Endif
;

\ JL001...
: Checksum32    \ addr1 addr2 addr3 -- ; buffer must be a multiple of 4 bytes
  { startaddr endaddr result -- }
  startaddr endaddr 4 Aligned?  \ Is the buffer len a multiple of 4 bytes?
  If
    0 result l!c(t)
    startaddr l@c(t) result l!c(t)      \ store first word

    endaddr 1+ startaddr cell+
    Do
      i l@c(t)                          \ get next word
      result l@c(t)                     \ get intermediate result
      +                                 \ 64b result...
      result l!c(t)                     \ ... but only store 32b
    cell +Loop
\    result l@c(t) 8 formCRC$
  Else
    ." Error! Buffer must be a multiple of 4 bytes!" cr
  Endif
;
\ ...JL001

\ ----------------------------------------------------------------------------
\ General CRC Words
Variable Bytewise       \ true for bytewise CCITT, SDLC/HDLC
Variable Feedback       \ generating polynomial
Variable CRC

: Odd   ( char -- flag ; is char odd? )
  1 And
;

: Lo    ( num -- num.lo ; return low byte of num )
  $ff And
;

\ Process CRC for a single byte
: UpdateCRC     \ char -- ; update CRC for a byte
  { B | -- }
  Bytewise @ If
    \ Bytewise operation for CCITT, SDLC/HDLC
    B  CRC @ Lo  Xor  Lo -> B
    B Dup  4 LShift  Xor Lo -> B

    CRC @ 8 RShift
    B Dup 8 LShift
    Over  3 LShift
    Rot   4 RShift
    Xor Xor Xor
    CRC !
  Else
    \ Bitwise operation for CRC-16, LRCC-16
    8 0 Do
      B Odd  CRC @ Odd  Xor If
        CRC @  1 RShift  Feedback @ Xor  CRC !
      Else
        CRC @  1 RShift  CRC !
      Endif
      B  1 RShift  -> B
    Loop
  Endif
;

\ Send a message string
: SendString            \ addr1 addr2 -- ; calculate BCC for range (addr1,addr2)
  1+ swap
  Do
    i c@c(t) UpdateCRC
  Loop
;

\ ----------------------------------------------------------------------------
\ Specific CRC Words
: CRC-CCITT     \ addr1 addr2 addr3 -- ; 16bit CCITT
  0 CRC !               \ initialise BCC to zero
  $8408 Feedback !
  1 Bytewise !
  -Rot
  SendString            \ addr3 addr1 addr2 -- addr3 ; calculate BCC

  \ The variable CRC contains the computed BCC
  CRC @  ( dup 4 formCRC$ )  swap w!c(t)
;

: CRC-16        \ addr1 addr2 addr3 -- ; 16bit CRC-16
  0 CRC !               \ initialise BCC to zero
  $A001 Feedback !
  0 Bytewise !
  -Rot
  SendString            \ addr3 addr1 addr2 addr3 -- ; calculate BCC

  \ The variable CRC contains the computed BCC
  CRC @  ( dup 4 formCRC$ )  swap w!c(t)
;

: CRC-LRCC16    \ addr1 addr2 addr3 -- ; 16bit LRCcc-16
  0 CRC !               \ initialise BCC to zero
  $8000 Feedback !
  0 Bytewise !
  -Rot
  SendString            \ addr3 addr1 addr2 -- ; calculate BCC

  \ The variable CRC contains the computed BCC
  CRC @  ( dup 4 formCRC$ )  swap w!c(t)
;

: CRC-SDLC      \ addr1 addr2 addr3 -- ; 16bit SDLC/HDLC
  $ffff CRC !           \ initialise BCC to $ffff
  -Rot
  1 Bytewise !
  $F088 Feedback !
  -rot SendString			\ addr3 addr1 addr2 -- ; calculate BCC

  \ The variable CRC contains the computed BCC
  CRC @  ( dup 4 formCRC$ )  swap w!c(t)
;

: CRC-32        \ addr1 addr2 addr3 -- ; 32 bit CRC
  0 CRC !               \ initialise BCC to zero
  $EDB8:8320 Feedback !
  0 Bytewise !
  -rot SendString			\ addr3 addr1 addr2 -- ; calculate BCC
  CRC @  ( dup 8 formCRC$ )  swap l!c(t)
;


\ 16 bit XModem

: +X16crc	\ crc char -- crc'
  8 lshift xor
  8 0 do
    dup $8000 and if
      1 lshift  $1021 xor
    else
      1 lshift
    then
  loop
  $FFFF and
;

: crc-xmodem	\ start-addr end-addr crc-addr --
  >r  0 -rot  1+ swap
  ?do  i c@c(t) +X16crc loop
  r> w!c(t)
;

: crc-xmodem0	\ start-addr end-addr crc-addr --
  >r  0 -rot  1+ swap
  ?do  i c@c(t) +X16crc loop
  r> over 8 rshift over c!c(t) 1+ c!c(t)
;

: +crcrp2040	\ crc byte -- crc'
\ Update non-reflected $104c11db7 polynom CRC32 checksum with next data byte
  #24 lshift xor
  8 0 do
    dup 1 lshift  swap $80000000 and
    if  $104c11db7 xor  then
  loop
;

	
: crc-32-RP2040	\ start end crc-addr --
  >r  $FFFF:FFFF -rot			\ -- crc-addr crc start end 
  1+ swap ?do 				\ -- crc-addr crc
    i c@c(t) +crcrp2040			\ -- crc-addr crc'
  loop
  r> l!c(t)
;
    

\ ----------------------------------------------------------------------------
\ Checksum type table
Create type_table	\ -- addr
        ' Checksum8     ,       (  8bit checksum )
        ' Checksum16    ,       ( 16bit checksum )
        ' Checksum32    ,       ( 32bit checksum )
        ' CRC-CCITT     ,       ( 16bit CRC      )
        ' CRC-16        ,       ( 16bit CRC      )
        ' CRC-LRCC16    ,       ( 16bit CRC      )
        ' CRC-SDLC      ,       ( 16bit CRC      )
        ' CRC-32        ,	( 32 bit CRC     )
        ' crc-xmodem    ,	\ 16 bit XModem CRC
        ' crc-xmodem0   ,	\ 16 bit XModem CRC - final result 0
        ' crc-32-RP2040 ,	\ 32 bit CRC for RP240 boot block
        

\ ----------------------------------------------------------------------------
\ Checksum word definition
: (Checksum)      \ addr1 addr2 addr3 type -- ;
  \ Generate checksum for memory block [addr1..addr2] and store at addr3
  1- cells type_table + @ execute	\ addr1 addr2 addr3 --
;


((
-------------------------------------------------------------------------------
        checksum list management
-------------------------------------------------------------------------------
))
: ADD-CHECKSUM  (  start end loc type  --  )
  here  >r  checksum-struct allot               \ use dictionary for structure
  r@ checksum.type !                            \ save the parameters in struct...
  r@ checksum.location !
  r@ checksum.end !
  r@ checksum.start !
  r> checksum-list list-add                     \ add structure to list
;

: DO-CHECKSUM  (  ^checksum-struct  --  )
  >r
  r@ checksum.start @                           \ retrieve all parameters
  r@ checksum.end @
  r@ checksum.location @
  r@ checksum.type @
  (checksum)                                    \ generate the checksum
  r> drop
;


((
-------------------------------------------------------------------------------
       public functions
-------------------------------------------------------------------------------
))

only forth also c-c also cc/i

: CHECKSUM	cc/i	\ start end addr type --
\ *G Set a region to be checksummed when *\fo{FINIS} executes.
\ ** More than one region can be checksummed.
\ *D Start	Start address.
\ *D End	Last address to be included in the checksum. For
\ **            16 and 32 bit operations, this address must be
\ **            aligned as required. Note that CRCs are byte
\ **            operations.
\ *D addr	Where the checksum is placed.
\ *D type	Type of checksum from the words below.
  add-checksum
;

\ checksum types
: SIMPLE8	cc/i	\ -- n
\ *G Generate an 8 bit checksum by adding bytes.
  simple8   ;
: SIMPLE16	cc/i	\ -- n
\ *G Generate an 16 bit checksum by adding byte pairs.
  simple16  ;
: SIMPLE32	cc/i	\ -- n
\ *G Generate an 32 bit checksum by adding four-byte units.
  simple32  ;
: CCITT		cc/i	\ -- n
\ *G Generate a CCITT checksum.
  ccitt     ;
: CRC16		cc/i	\ -- n
\ *G Generate a CRC16 checksum.
  crc16     ;
: LRCC16	cc/i	\ -- n
\ *G Generate an LRCC16 checksum.
  lrcc16    ;
: SDLC		cc/i	\ -- n
\ *G Generate an SDLC checksum.
  sdlc      ;
: CRC32		cc/i	\ -- n
\ *G Generate a 32 bit CRC.
  crc32  ;
: CRCxModem16	cc/i	\ -- n
\ *G Generate a 16 bit XModem CRC. The result is stored in native
\ ** order.
  CRCxModem16  ;
: CRCxModem16-0	cc/i	\ -- n
\ *G Generate a 16 bit XModem CRC. The result is stored in
\ ** big-endian order.
\ *P *\fo{CRCxModem16-0} is usually used when
\ ** you want to force the checksum of a block to zero, e.g.
\ *C   <start> <end-2> <end-1> CRCxModem16-0 checksum
\ *P where <start> is the first address of the block and <end> is
\ ** the last address in the block. If you later perform an XModem
\ ** checksum from <start> to <end> the CRC will be zero.
  CRCxModem16-0  ;
: CRC32RP2040  cc/i	\ -- n
\ *G Used to generate a 32 bit little endian CRC32 as used for 
\ ** Raspberry Pi Pico and RP2040 boot blocks.
\ ** The last four bytes of the image loaded from flash (which
\ ** we hope is a valid flash second stage) are a CRC32 checksum
\ ** of the first 252 bytes. The parameters of the checksum are:
\ *(
\ *B  Polynomial: 0x04c11db7
\ *B  Input reflection: no
\ *B  Output reflection: no
\ *B  Initial value: 0xffffffff
\ *B  Final XOR: 0x00000000
\ *B  Checksum value appears as little-endian integer at end of image
\ *)
\ *P Note that for the calculation of the RP2040 boot2 loader,
\ **the end address is start+251 bytes. The address range is
\ ** start:start+251 and the CRC is stored at start+252.
  CRC32RP2040  ;


only forth also c-c definitions

: ?GENERATE-CHECKSUMS	\ --
  ['] do-checksum checksum-list list-traverse  ;

: -checksums	\ -- ; disable checksums
  checksum-list list-clear  ;


\ ======
\ *> ###
\ ======

