\ Literal pool handling for VFX optimiser

((
Copyright (c) 2002, 2004, 2023
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441


To do
=====

Change history
==============
20230713 SFP003 Added ASMLITLOADINS, and C_ASM-LITPOOL
20110605 SFP002 Aligned pool.
20041110 SFP001 Added merge for duplicates
))

decimal

\ ======
\ *! litpool
\ *T Literal pool
\ ======

(( Simple one
The literal pool consists of a list of references which are resolved
after an unconditional return or branch. Since the pool is flushed on
a regular basis, it is not necessary for it to be large, and the pool
is currently implemented as a fixed size table of up to 256 items.

The format of a literal pool entry is:
  cell   lp-value        literal value
  cell   lp-addrins      target address of load instruction
  cell   lp-prev         pointer to previous /LP-ENTRY with same LP-VALUE
  cell   lp-addrconst    pointer to where constant was laid for duplicates

A literal that is to be handled by the literal pool system is compiled by
the word C_LITPOOL (n -- ).
))

[undefined] buffer: [if]
: buffer:       \ size -- ; -- addr ; build a buffer of the required size
  create  allot&erase  ;
[then]

variable lpoffset	\ -- addr
\ *G Offset of next entry in table

struct /lp-entry	\ -- len
  cell field lp-value		\ value of literal
  cell field lp-addrins		\ address of load instruction
  cell field lp-prev		\ 0 for first entry, pointer to previous /LP-ENTRY
  cell field lp-addrconst	\ 0 or address of laid constant for duplicates
end-struct

#256 /lp-entry *
  constant lpsize		\ size of pool

lpsize buffer: LiteralPool

: initlitpool    \ -- ; initialise pool
( *G Initialise the compiler's literal pool. )
  lpoffset off			\ mark unused
;
initlitpool

defer LitLoadIns,	\ dest-res -- ^ins ; lay literal load instruction
\ *G A DEFERed word that loads a literal from the pool into a destination. 
\ ** resource. Used as a factor of C_LITPOOL. It is deferred because it is 
\ ** CPU dependent. Assignment of the action is usually performed by the 
\ ** code generator during the cross compiler build. 

defer asmLitLoadIns,	\ reg# -- ^ins ;  lay literal load instruction
\ *G A DEFERed word that loads a literal from the pool into a destination. 
\ ** register. Used as a factor of C_ASM-LITPOOL. It is deferred because it is 
\ ** CPU dependent. Assignment of the action is usually performed by the 
\ ** code generator during the cross compiler build. 

defer ResLoadIns	\ ^const ^ins -- ; resolve load offset
( *G A DEFERed word that resolves a load instruction laid by LITLOADINS, )
( ** with an unkown offset. Used as a factor of FLUSHLITPOOL. It is deferred  )
( ** because it is CPU dependent. Assignment of the action is usually )
( ** performed by the code generator during the cross compiler build. )

: inLitpool?	\ n -- addr|0
\ *G Checks whether n is already in the pool, returning the
\ ** /LP-ENTRY address if already there, or zero if not found.
\ ** The pool must be scanned from the start.
  0
  LiteralPool lpoffset @ bounds ?do	\ -- n flag
    over i lp-value @ =			\ if n already exists
    if  drop i  leave  endif		\   flag := record
  /lp-entry +loop
  nip					\ -- 0|addr
;

: c_litpool      \ n dest-res -- ; add n to pool
( *G Lays a literal load instruction from the literal pool that )
( ** will be resolved when FLUSHLITPOOL is next called. )
  lpoffset @ lpsize >= abort" literal pool overflow"
  literalpool lpoffset @ + >r		\ slot in pool
  LitLoadIns, r@ lp-addrins !		\ lay load instruction, stash ins. addr.
  dup r@ lp-value !			\ stash literal value to load
  inLitpool? r@ lp-prev !		\ check for duplicate entry
  r> lp-addrconst off			\ unknown at present
  /lp-entry lpoffset +!			\ step to next slot
;

: (c_asm-litpool)	\ n reg# -- ; add n to pool
\ *G Lays a literal load instruction from the literal pool that 
\ ** will be resolved when FLUSHLITPOOL is next called. 
  lpoffset @ lpsize >= abort" literal pool overflow"
  literalpool lpoffset @ + >r		\ slot in pool
  asmLitLoadIns, r@ lp-addrins !	\ lay load instruction, stash ins. addr.
  dup r@ lp-value !			\ stash literal value to load
  inLitpool? r@ lp-prev !		\ check for duplicate entry
  r> lp-addrconst off			\ unknown at present
  /lp-entry lpoffset +!			\ step to next slot
;
assign (c_asm-litpool) to-do c_asm-litpool

: flushlitpool   \ --
( *G Flushes the literal pool at the current CDATA location, and fixes )
( ** up all outstanding load instructions referencing the literal pool. )
  lpoffset @ if
    calign(t)
    LiteralPool lpoffset @ bounds ?do
      i lp-prev @ ?dup 0= if		\ -- ; if not a duplicate
        chere(t) dup i lp-addrconst !	\ -- ^const ; fill in address of const
        i lp-value @ l,c(t)		\ lay const
      else				\ -- prev
        lp-addrconst @			\ -- ^const
      endif
      i lp-addrins @ resloadins		\ resolve load ins.
    /lp-entry +loop
    lpoffset off
  endif
;

: #lp-entries	\ -- n
\ *G Return the number of entries in the literal pool, including
\ ** duplicates.
  lpoffset @ /lp-entry /
;


\ ======
\ *> ###
\ ======

decimal


