\ Umbilical Forth host side

((
Copyright (c) 1988, 1989, 1992, 1997, 1998, 2000, 2001, 2013, 2022
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: (+44) 2380 631441
net: mpe@mpeforth.com

Change History
==============
V2 changes
20220323 SFP020 removed ACK control from data cells
V1 changes
20131108 SFP014 Improved portability of FROM-TARG-EMIT.
20131031 SFP013 Added host break to KEY operation.
20010711 SFP012 Added LINK-FLUSH to SEND-MESSAGE
20000424 SFP011 Allowed for Harvard targets to access CDATA over the link
19980304 MSD010 Inhibit EMITs of carriage-returns when the target sends them
         MSD010 over the umbilical serial link.  This keeps the generated
         MSD010 output more in line with what you see when running standalone.
19980304 MSD009 Cosmetic change to the error message generated by ?SYNCH-BYTE.
         MSD009 (A space was added before the "Expecting ACK!" display.)
19980222 SFP008 Converted C@ to use byte data
                Converted RECEIVE-CELL to give no ACK!s
19980221 SFP007 Removed PUSH-HOST-MESS(R) and replaced it with
                TYPE-HOST-MESS(R). Converted EMIT.. to send byte
19980216 MSD006 Removed "ACK! SEND-BYTE" line from before the "BOUNDS"
         MSD006 in COPY-HOST-MESS(S) (the "send buffer" command).
19970219 SFP005 Added THROW! handling.
19970126 SFP004 Added LINK-FLUSH.
19970114 SFP003 Added download for non-paged targets.
19970112 SFP002 I/O words deferred for allow different links.
19970110 SFP001 Converted to PFW host.


))


\ start it off

cr cr ." Loading HOST message system handler"

only forth also c-c definitions

\ Defer the link primitives
\ link-open link-close query-byte wait-byte send-byte link-reset .link-state
\ SFP002
Defer link-open         \ -- ; open the link if not already done
Defer link-close        \ -- ; close the link if not already done
Defer query-byte        \ -- t/f ; Like KEY?
Defer wait-byte         \ -- b ; like KEY
Defer send-byte         \ b -- ; like EMIT
Defer link-reset        \ -- ; resets target
Defer .link-state       \ -- ; show link details
Defer link-flush        \ -- ; flush any input characters       SFP004

only forth also c-c also Umbilical definitions

\ query-byte wait-byte byte Send-Byte ?synch-byte

(( Sample implementations from PF386 implementation SFP002
: query-byte    \ -- flag
  ser-key?
;

: Wait-byte     \ -- b
  begin  ser-key? 0=            \ From serial port
    while  key?
           if  case  key
                 2 of  1 abort" ^B - Serial break"  endof
                 3 of  1 abort" ^C - Serial break"  endof
               endcase
           endif
  repeat
  ser-key
;

: Send-byte     \ b --
  ser-emit                      \ To serial port
;
))


\ send-cell *** DEPENDENT *** ........ J L   06/03/88

\ Cells are transmitted and received in big-endian format, high byte first,
\ regardless of the processor intrinsic data format.
\ Send each byte and wait for ACK! to be returned.

hex

: ?synch-byte   \ b -- ( expected )
  dup ACK! <>                           \ test for ACK
  if  pcr ." Value: " .byte             \ show value if failed
      1 abort"  Expecting ACK!"         \ report error                          \ MSD009
  else  drop
  endif
;

: send&ack      \ b --
  send-byte  wait-byte ?synch-byte  ;   \ transmit and acknowledge

: send-ack      \ --
  ACK! send-byte  ;

: recv&ack      \ -- b
  wait-byte  send-ack  ;


16BIT? [if]
(( \ V1
: send-cell     \ n --
  0 0100 um/mod                         \ -- rem highbyte
  send&ack                              \ transmit and acknowledge
  send&ack                              \ transmit and acknowledge
;
: receive-cell  \ -- n
  wait-byte ( send-ack )                \ -- bh
  0100 *
  wait-byte ( send-ack )                \ -- bl
  +                                     \ Convert to a word -- n
;
))
\ V2
: send-cell     \ n --
  dup 8 rshift send-byte		\ -- rem highbyte
  send-byte
;
: receive-cell  \ -- n
  wait-byte 8 lshift  wait-byte  or		\ bml
;

[then]

32BIT? [if]
(( \ V1
: send-cell     \ n --
  0 01000000 um/mod send&ack            \ -- rem highbyte
  0 010000 um/mod   send&ack            \ -- rem highbyte
  0 0100 um/mod     send&ack            \ -- rem highbyte
                    send&ack            \ transmit and acknowledge
;
: receive-cell  \ -- n
  wait-byte 01000000 *                   \ bh
  wait-byte 010000 *  +                  \ bmh
  wait-byte 0100 *  +                    \ bml
  wait-byte       +                      \ bl
;
))
\ V2
: send-cell     \ n --
  dup #24 rshift send-byte	\ bh
  dup #16 rshift send-byte	\ bmh
  dup 8   rshift send-byte	\ bml
                 send-byte	\ bl
;
: receive-cell  \ -- n
  wait-byte #24 lshift		\ bh
  wait-byte #16 lshift  or	\ bmh
  wait-byte 8 lshift  or	\ bml
  wait-byte  or 		\ bl
;

[then]

64BIT? [if]
(( \ V1
: send-cell     \ x64 --
  8 0 do				\ -- x64'
    8 rol dup $FF and send&ack
  loop
  drop
;
: receive-cell  \ -- x64 ; receive big-endian number
  0  8 0 do				\ sum
    8 lshift wait-byte or
  loop
;
))
\ V2
: send-cell     \ x64 --
  8 0 do			\ -- x64'
    8 rol dup  $FF and send-byte
  loop
  drop
;
: receive-cell  \ -- x64 ; receive big-endian number
  0  8 0 do			\ sum
    8 lshift wait-byte or
  loop
;
))
[then]


\ send-stack receive-stack                         sfp 28/06/90

: send-stack    \ sys -- ; send stack to target
  depth dup send-cell  ?dup             \ say what's to come
  if  0 swap 1-                         \ send bottom first
      do  i roll send-cell  -1 +loop
  endif  ;

: receive-stack \ -- sys ; receive stack from target
  receive-cell 0                        \ how many words ?
  ?do  receive-cell  loop  ;            \ receive them


\ .-host-mess(r) ............................... J sfp 28/06/90

((
These messages have names ending in (R) because these messages are
*R*eceived by the cross compiler from the target
))
: .-host-mess(r)        \ -- ; From serial
  send-ack  receive-cell .  ;

: s>t-host-mess(r)      \ sys -- ; send host stack to target
  send-ack  send-stack  ;

: s>h-host-mess(r)      \ -- sys ; receive stack from target
  send-ack  receive-stack  ;


\ type-host-mess(r) Key?-host-mess(r)

: from-targ-emit	\ char -- ; SFP014
  case
    $0A of  endof			\ ignore
    $0D of  cr  endof			\ line terminator
      dup emit
  endcase
;
((
: from-targ-emit        ( char -- )     \ MSD010
  dup $0A <> if emit else drop then     \ MSD010
;                                       \ MSD010
))

: type-host-mess(r)     \ -- ; From serial
  send-ack
  wait-byte 0                   \ get length
  ?do  wait-byte from-targ-emit  loop  ;                                        \ MSD010

: Key?-host-mess(r)     \ -- ; ( To serial )
  send-ack  key? send-cell  ;   \ Send flag to serial


\ Key-host-mess(r) emit-host-mess(r) .. ........ J sfp 10/09/92

: Key-host-mess(r)      \ -- ( To serial )
  send-ack  key dup
  case					\ SFP013
    2 of  1 abort" ^B - Host break"  endof
    3 of  1 abort" ^C - Host break"  endof
  endcase
  send-cell                         \ Send char to serial
;

: Emit-host-mess(r)     \ -- ( From serial )
  send-ack
  wait-byte from-targ-emit  ;           \ Emit char to serial                   \ MSD010


\ Host-message(r) .............................. J sfp 10/09/92

: Host-message(r)       \ mess# --
  CASE
    key-cmnd>h  OF  key-host-mess(r)  ENDOF
    emit-cmnd>h OF  emit-host-mess(r) ENDOF
    type-cmnd>h OF  type-host-mess(r) ENDOF
    key?-cmnd>h OF  key?-host-mess(r) ENDOF
    .-cmnd>h    OF  .-host-mess(r)    ENDOF
\    s>t-cmnd>h  OF  s>t-host-mess(r)  ENDOF
\    s>h-cmnd>h  OF  s>h-host-mess(r)  ENDOF
                    pcr ." Value: " .byte
                    1 abort"  Illegal message from target"
  ENDCASE
;


\ receive-message

variable last-status    \ -- addr ; last-status code from target execute

: Receive-message       \ --
  BEGIN
    wait-byte                           \ Wait for TARGET MESSAGE or STATUS CODE
    dup status>h-start status>h-end     \ -- status lower upper
    within? 0=                          \ if not a status code
  WHILE
    host-message(r)                     \ assume a message from the target
  REPEAT
  dup last-status !                     \ SFP005
  case
    OK!     of   endof                                  \ successful completion
    ACK!    of  1 abort" Unexpected ACK!"  endof        \ ??
    ERROR!  of  send-ack 1 abort" Bad command to target "  endof        \ ??
    SUFL!   of  send-ack 1 abort" Target stack underflow"  endof        \ stack underflow
    THROW!  of   endof
                1 abort" Unknown status code"           \ all others
  endcase
\  send-ack                                     \ acknowledge status
;


\ Pop-host-message(s) .......................... J sfp 06/07/90

((
These messages have names ending in (S) because these messages are
*S*ent by the cross compiler to the target
))

: Pop-host-mess(s)      \ -- n
  wait-byte ?synch-byte
  receive-cell  ;


\ exec-host-mess(s) push-host-mess(s) .......... J sfp 27/10/91

: ?target-threw         \ -- ; message if target threw  SFP005
  last-status @off  THROW! =          \ THROW code is special
  if  cr ."  Target THROW code " .  endif
;

: Exec-host-mess(s)     \ n --
  wait-byte ?synch-byte               \ Error if not ACK
  send-stack                          \ Send stack over
  receive-message                     \ wait for target message
  receive-stack                       \ get stack back again
  ?target-threw                       \ display message on THROW  SFP005
;

: Push-host-mess(s)     \ n --
  wait-byte                           \ Wait for ACK! code
  ?synch-byte                         \ Error if not
  send-cell
;


\ s>t-host-mess(s) s>h-host-mess(s)                sfp 06/07/90

: s>t-host-mess(s)      \ sys -- ; send host stack to target
  wait-byte ?synch-byte                 \ wait for acknowledge
  send-stack  ;

: s>h-host-mess(s)      \ -- sys ; receive target stack
  wait-byte ?synch-byte                 \ wait for acknowledge
  receive-stack  ;


\ @-host-mess(s) !-host-mess(s)                    sfp 06/07/90

: C@-host-mess(s)       \ addr -- n
  wait-byte ?synch-byte                 \ wait for acknowledge
\  send-cell  receive-cell  ;
  send-cell  wait-byte  ;

: @-host-mess(s)        \ addr -- n
  wait-byte ?synch-byte                 \ wait for acknowledge
  send-cell  receive-cell  ;

: !-host-mess(s)        \ n addr --
  wait-byte ?synch-byte                 \ wait for acknowledge
  swap send-cell send-cell  ;


\ copy-host-mess(s)

: copy-host-mess(s)   \ source dest-addr len --
  wait-byte ?synch-byte         \ wait for acknowledgement
  swap send-cell                \ send dest-addr
  dup send-cell                 \ send length
  bounds
  ?Do
    i c@ send&ack
  Loop
;


\ Host-message                                     sfp 06/07/90

variable ackdelay  #1 ackdelay !

: Host-message(s)       \ mess# --
  CASE
    execute-cmnd>t OF  exec-host-mess(s)  ENDOF \ EXECUTE
    push-cmnd>t    OF  push-host-mess(s)  ENDOF \ *** unused ***
    pop-cmnd>t     OF  pop-host-mess(s)   ENDOF \ *** unused ***
    s>t-cmnd>t     OF  s>t-host-mess(s)   ENDOF \ *** unused *** stack > target
    s>h-cmnd>t     OF  s>h-host-mess(s)   ENDOF \ *** unused *** stack > host
    W@-cmnd>t      OF  @-host-mess(s)     ENDOF \ W@(T)
    W!-cmnd>t      OF  !-host-mess(s)     ENDOF \ W!(T)
    C@-cmnd>t      OF  C@-host-mess(s)    ENDOF \ C@(T)
    C!-cmnd>t      OF  !-host-mess(s)     ENDOF \ C!(T)
[ 32bit? ] [if]
    L@-cmnd>t      OF  @-host-mess(s)     ENDOF \ L@(T)
    L!-cmnd>t      OF  !-host-mess(s)     ENDOF \ L!(T)
[then]
    copy-cmnd>t    OF  copy-host-mess(s)  ENDOF \ Copy bytes to target
[ harvard? ] [if]	\ SFP011
    W@C-cmnd>t     OF  @-host-mess(s)     ENDOF \ W@C(T)
    W!C-cmnd>t     OF  !-host-mess(s)     ENDOF \ W!C(T)
    C@C-cmnd>t     OF  C@-host-mess(s)    ENDOF \ C@C(T)
    C!C-cmnd>t     OF  !-host-mess(s)     ENDOF \ C!C(T)
[ 32bit? ] [if]
    L@C-cmnd>t     OF  @-host-mess(s)     ENDOF \ L@C(T)
    L!C-cmnd>t     OF  !-host-mess(s)     ENDOF \ L!C(T)
[then]
[then]
                       pcr ." Value: " .byte
                       1 abort"  Illegal Message to target"
  ENDCASE
  ackdelay @ 0 ?do loop
  send-ack  ;


\ Send-message >Target ......................... J sfp 10/09/92

: Send-message  \ mess# --
  link-flush			\ flush junk from input queue ; SFP012
  dup send-byte			\ Send MESSAGE byte
  host-message(s)		\ Execute host part
;

: >Target       \ pars... mess# --
  send-message
  receive-message
;


\ >Target target-exec

: ?Token        \ addr -- targ-addr
  @(s) ?dup ?undef
;

: target-exec   \ symbol --
\ ."  TARGET-EXEC "
  ?token                              \ Is token defined ?
  execute-cmnd>t >target
;


\ all done

cr ." HOST message system handler loaded" cr

decimal


