\ EtherLMrom.fth - Ethernet driver for Luminary ROM.

((
Copyright (c) 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20100405 SFP001 First release.
))

only forth definitions
decimal

\ ===============
\ *! etherlmrom
\ *T Luminary/Stellaris Ethernet driver
\ ===============
\ *P The code in *\i{Cortex\Drivers\EtherLMrom.fth} mostly uses
\ ** the ROM calls in Luminary devices to provide a standard MPE
\ ** Ethernet layer for PowerNet. Note that some calls are
\ ** unimplemented in some versions of some devices. Because the
\ ** ROM API does not provide transmit and receive routines usable
\ ** by PowerNet, these have been provided in Forth. Note also
\ ** that these routines rely on the ability of Cortex-M3 cores
\ ** to perform unaligned memory accesses.


\ ****************
\ *S Configuration
\ ****************

interpreter
: const equ ;	\ n -- ; -- n
\ *G If you are using a standalone target with heads and you want
\ ** interactive access to all the registers and bit masks, define
\ ** *\fo{CONST} as *\fo{CONSTANT}, otherwise by default *\fo{CONST}
\ ** is defined as *\fo{EQU}.
target

[undefined] EtherCom? [if]
0 equ EtherCom?	\ -- flag
\ *G Only needed if the EtherCom generic I/O device is being used.
\ ** This only happens for systems with multiple IP ports and
\ ** PowerNet v3.x upwards.
[then]

[undefined] EtherDiags? [if]
0 equ EtherDiags?	\ -- flag
\ *G Set this equate true to compile diagnostic code for
\ ** register dumping and so on. False by default.
[then]

[undefined] sniff? [if]
0 equ sniff?	\ -- flag
\ *G Set this equate true to compile the packet sniffer code,
\ ** which can be used to test Ethernet reception.
[then]

[undefined] pingit? [if]
0 equ pingit?	\ -- flag
\ *G Set this equate true to compile the PING code,
\ ** which tests Ethernet transmission and reception.
[then]

[undefined] EtherAddress [if]
create EtherAddress	\ -- addr
\ *G Holds the Ethernet MAC address (six bytes). Note that you
\ ** must obtain these from the IEEE (www.ieee.org) or from other
\ ** sources.
  $00 c, $10 c, $8B c, $F1 c, $44 c, $51 c,	\ six bytes for 48 bit MAC address
\  $00 c, $08 c, $EF c, $00 c, $5E c, $C4 c,	\ six bytes for 48 bit MAC address
[then]						\ you must obtain these yourself!

[undefined] IpAddress [if]
create IpAddress	\ -- addr
\ *G Holds the Ethernet IP address (four bytes).
\  10 c,  102 c,  0 c,  11 c,		\ assign these as required
  192 c,  168 c,  0 c,  251 c,		\ assign these as required
[then]


\ ************************
\ *S ROM calls and equates
\ ************************

ETH_CFG_RX_BADCRCDIS
ETH_CFG_TX_DPLXEN or
ETH_CFG_TX_CRCEN or
ETH_CFG_TX_PADEN or
  equ /EthCfg

: InitEmac	\ -- ior
\ *G Initialise the Ethernet system, returning zero on success.
  SYSCTL_PERIPH_ETH PeriphEnable	\ enable Ethernet
  SYSCTL_PERIPH_ETH PeriphReset		\ reset Ethernet
  SYSCTL_PERIPH_GPIOF PeriphEnable	\ enable GPIOF
(( not on B1 devices
  GPIO_PF2_LED1 GPIOpinCfg
  GPIO_PF3_LED0 GPIOpinCfg
))
  $1100 $FF00 _GPIOF gpioPCTL + setMask	\ function 1 for bits 2/3
  _GPIOF bit2 bit3 or 2 GPIOdirModeSet	\ 0=in, 1=out, 2=hw
  _GPIOF bit2 bit3 or 1 8 GPIOpadCfgSet	\ 1=2ma, 8=STD
  _EMAC system-speed EtherInitExpClk
  _EMAC /EthCfg EtherConfigSet
  _EMAC EtherAddress EtherMACset
  _EMAC EtherEnable
  0
;

: InitEther	\ --
\ *G Perform a full initialisation of the chip, enabling
\ ** Rx and Tx, and setting up the Ethernet MAC address
\ ** from *\fo{EtherAddress}.
  cr ." Initialising Ethernet ... "
  InitEMAC
  if ." un" then ." successful"
;

: ReadPhy	\ reg -- w
\ *G Read a PHY register. Note that these are 16 bit values.
  _EMAC swap EtherPHYread $FFFF and  ;

: WritePhy	\ w reg --
\ *G Write a PHY register. Note that these are 16 bit values.
  swap _EMAC -rot EtherPHYwrite  ;

: OrPhy		\ mask reg --
\ *G Or mask into reg in the PHY. Set bits in mask are set in the
\ ** register.
  tuck ReadPhy  or  swap WritePhy  ;

: EtherLink?	\ -- flag
\ *G Return true if the Ethernet link is established. For most
\ ** PHYs, this is performed by testing bit 2 of register 1.
  1 ReadPhy bit2 and 0<>  ;


\ ********************
\ *S Packet read/write
\ ********************

: IsRx?         \ -- flag ; nz = has data
\ *G Check if incoming data is present, returning true if a
\ ** packet is available for *\fo{get_ether_pkt} below.
  _EMAC macNP + @ $3F and  ;


\ We cannot use the ROM API for reading a packet because it
\ transfers the length information.

\ Based on the following table of how the receive frame is stored in the
\ receive FIFO, this function will extract a packet from the FIFO and store
\ it in the packet buffer that was passed in.
\
\ Format of the data in the RX FIFO is as follows:
\
\ \verbatim
\ +---------+----------+----------+----------+----------+
\ |         | 31:24    | 23:16    | 15:8     | 7:0      |
\ +---------+----------+----------+----------+----------+
\ | Word 0  | DA 2     | DA 1     | FL MSB   | FL LSB   |
\ +---------+----------+----------+----------+----------+
\ | Word 1  | DA 6     | DA 5     | DA 4     | DA 3     |
\ +---------+----------+----------+----------+----------+
\ | Word 2  | SA 4     | SA 3     | SA 2     | SA 1     |
\ +---------+----------+----------+----------+----------+
\ | Word 3  | FT LSB   | FT MSB   | SA 6     | SA 5     |
\ +---------+----------+----------+----------+----------+
\ | Word 4  | DATA 4   | DATA 3   | DATA 2   | DATA 1   |
\ +---------+----------+----------+----------+----------+
\ | Word 5  | DATA 8   | DATA 7   | DATA 6   | DATA 5   |
\ +---------+----------+----------+----------+----------+
\ | Word 6  | DATA 12  | DATA 11  | DATA 10  | DATA 9   |
\ +---------+----------+----------+----------+----------+
\ | ...     |          |          |          |          |
\ +---------+----------+----------+----------+----------+
\ | Word X  | DATA n   | DATA n-1 | DATA n-2 | DATA n-3 |
\ +---------+----------+----------+----------+----------+
\ | Word Y  | FCS 4    | FCS 3    | FCS 2    | FCS 1    |
\ +---------+----------+----------+----------+----------+
\ \endverbatim
\
\ Where FL is Frame Length, (FL + DA + SA + FT + DATA + FCS) Bytes.
\ Where DA is Destination (MAC) Address.
\ Where SA is Source (MAC) Address.
\ Where FT is Frame Type (or Frame Length for Ethernet).
\ Where DATA is Payload Data for the Ethernet Frame.
\ Where FCS is the Frame Check Sequence.

: discardFrame	\ len -- ; len=framelen-6
  7 + -4 and 0 ?do
    _EMAC macDATA + @ drop
  4 + loop
;

: get_ether_pkt \ *dest maxlen -- len
\ *G Get a pending receive packet to a buffer of size *\b{maxlen).
\ ** Return the number of bytes received, or zero for a bad packet.
\ ** Note that data is valid only when *\fo{ISRX?} returns true,
\ ** so that you must poll with *\fo{ISRX?} before using
\ ** *\fo{GET_ETHER_PKT}. The received size of the packet
\ ** (not clipped to the buffer size) is returned.
  _EMAC macDATA + @
  dup #16 rshift >r			\ 16 bit data
  $FFFF and 6 - 0 max			\ frame-FL-FCS
  \ -- dest dlen flen ; R: -- wx
  tuck < if				\ -- dest flen ; R: -- wx
    r> drop nip discardFrame 0 exit	\ too long
  endif
  dup #60 < if
    r> drop nip discardFrame 0 exit	\ runt
  endif
  tuck					\ -- flen dest flen ; we return this length
  over r> swap w!  2 /string		\ save first two bytes
  3 + -4 and bounds ?do			\ -- align remaining length
    _EMAC macDATA + @ i !		\ relies on CM3 unaligned access
  4 +loop
  _EMAC macDATA + @ drop		\ read and discard FCS
;

: DiscardRX     \ --
\ *G Throw away a pending input packet due to lack of host buffer
\ ** memory.
  _EMAC macDATA + @ $FFFF and 6 - 0 max	\ frame-FL
  discardFrame
;

\ Format of the data in the TX FIFO is as follows:
\
\ +---------+----------+----------+----------+----------+
\ |         | 31:24    | 23:16    | 15:8     | 7:0      |
\ +---------+----------+----------+----------+----------+
\ | Word 0  | DA 2     | DA 1     | PL MSB   | PL LSB   |
\ +---------+----------+----------+----------+----------+
\ | Word 1  | DA 6     | DA 5     | DA 4     | DA 3     |
\ +---------+----------+----------+----------+----------+
\ | Word 2  | SA 4     | SA 3     | SA 2     | SA 1     |
\ +---------+----------+----------+----------+----------+
\ | Word 3  | FT LSB   | FT MSB   | SA 6     | SA 5     |
\ +---------+----------+----------+----------+----------+
\ | Word 4  | DATA 4   | DATA 3   | DATA 2   | DATA 1   |
\ +---------+----------+----------+----------+----------+
\ | Word 5  | DATA 8   | DATA 7   | DATA 6   | DATA 5   |
\ +---------+----------+----------+----------+----------+
\ | Word 6  | DATA 12  | DATA 11  | DATA 10  | DATA 9   |
\ +---------+----------+----------+----------+----------+
\ | ...     |          |          |          |          |
\ +---------+----------+----------+----------+----------+
\ | Word X  | DATA n   | DATA n-1 | DATA n-2 | DATA n-3 |
\ +---------+----------+----------+----------+----------+
\
\ Where PL is Payload Length, (DATA) only
\ Where DA is Destination (MAC) Address
\ Where SA is Source (MAC) Address
\ Where FT is Frame Type (or Frame Length for Ethernet)
\ Where DATA is Payload Data for the Ethernet Frame

: IsTx?		\ len -- flag
\ *G Return true if a packet of length *\i{len} can be sent.
  drop  _EMAC macTR + @ bit0 and 0=  ;

: waitTX	\ len --
\ *G Wait until a packet of length *\i{len} can be sent.
  begin  dup IsTx? 0= while  pause  repeat
  drop
;

: send_ether_pkt	\ caddr len --
\ *G Send packet from supplied buffer.
\ Perform sanity checks
  dup #14 #1520 within 0=		\ range check length
  if  2drop exit  endif
  #64 max dup waitTX			\ force length to min bytes
  over w@ #16 lshift  over #14 - or	\ Word 0
  _EMAC macDATA + !
  2 /string 3 + -4 and			\ round to number of complete words
  bounds ?do
    i @ _EMAC macDATA + !		\ relies on CM3 unaligned addressing
  4 +loop
  bit0 _EMAC macTR + or!		\ start TX
;


\ ****************************************
\ *S Generic I/O for PowerNet v3 and above
\ ****************************************

EtherCom? [if]

: GetAddrs	\ -- ipaddr macaddr 0
\ *G For Generic I/O IPDGetAddr function
  IpAddress EtherAddress 0
;

: SetAddrs	\ ipaddr|0 macaddr|0 mode --
\ *G Set up the Ethernet MAC and the IP addresses.
\ ** At present mode is always 0, but will be used in
\ ** future releases to indicate data formats. If
\ ** ipaddr or macaddr are zero, the corresponding stored data
\ ** will not be changed.
  drop					\ mode flags
  ?dup if				\ macaddr
    dup EtherAddress 6 cmove		\ save
    SetMACaddress			\ put in EMAC
  endif
  ?dup if				\ ipaddr
    IpAddress 4 cmove
  endif
;

\ *P The layout and usage of the Ethernet common I/O structure is
\ ** defined in the file *\i{COMMON\ETHERCOM.FTH}.
\ *E create IPdevice
\ **   ' MyInit ,       \ 0: initialisation
\ **   ' MyTerm ,	\ 1: shutdown
\ **   ' MyRx? ,	\ 2: receive test
\ **   ' MyRx		\ 3: receive packet
\ **   ' MyTx? ,	\ 4: transmit test
\ **   ' MyTx ,		\ 5: transmit packet
\ **   ' MyGetAddr ,    \ 6: Get device addresses
\ **   ' MySetAddr ,    \ 7: Set device addresses
\ **   ' MySave ,	\ 8: Save IP device state

create E23xxvector	\ -- addr
\ *G The device vector needed by PowerNet v3+ and
\ ** *\i{COMMON\ETHERCOM.FTH}.
  ' InitEther ,         \ initialisation
  ' noop ,	        \ shutdown
  ' IsRx? ,	        \ receive test
  ' get_ether_pkt ,     \ receive packet
  ' IsTx? ,	        \ transmit test
  ' send_ether_pkt ,    \ transmit packet
  ' GetAddrs ,		\ Get device addresses
  ' SetAddrs ,	        \ Set device addresses
  ' noop ,	        \ Save IP device state

E23xxvector constant IPDevice	\ -- addr
\ *G *\fo{IPDevice} is the default device name required by
\ ** *\{ETHERCOM.FTH}. If you have multiple ports, only one
\ ** should be called *\fo{IPDevice}.

[then]


\ **************
\ *S System test
\ **************
\ *P The packet sniffer code is only compiled if the equate
\ ** *\fo{SNIFF?} is non-zero.

EtherDiags? [if]

_EMAC constant _EMAC
: .em		\ --
  _EMAC $20 pdump  ;
: .ed		\ --
  _EMAC $100 + $30 pdump  ;
: .td		\ --
  ETXCtlDescs $40 pdump  ;

[then]

sniff? [if]

1536 buffer: pbuff	\ -- addr
\ *G Temporary buffer for *\fo{SNIFF}.

: NextRxPkt	\ --
\ *G Collect and display packet
  pbuff 1536 erase
  pbuff 1536 get_ether_pkt dup .word
  pbuff swap 1 max dump
;

: sniff		\ --
\ *G Listens to the network and displays all the traffic that
\ ** has a broadcast destination or is for this device.
\ ** *\fo{SNIFF} can be used to test reception.
  begin
    isrx?
    if  NextRxPkt  endif
    #10 ms
[ tasking? ] [if] pause [then]
    key?
  until
;

[then]

\ *P The PING example is used to test Ethernet transmission and
\ ** reception. Before testing transmission, test reception by
\ ** using the maintenance system to download a user application
\ ** and then pinging the target board instead of TFTPing to it.
\ *E Example ARP request:  #42/$2A bytes
\ ** C020.0600  FF FF FF FF FF FF 00 C0  F0 0D 9D 7E 08 06 00 01
\ ** C020.0610  08 00 06 04 00 01 00 C0  F0 0D 9D 7E C0 A8 01 7F
\ ** C020.0620  00 00 00 00 00 00 C0 A8  FF FE 11 0A 75 9C C0 A8

\ *P Note that the equates *\fo{SNIFF?} and *\fo{PINGIT?} must both
\ ** be set non-zero to compile the PING code.

pingit? [if]

create requested-ip	\ -- addr
\ *G Holds IP address to be pinged.
  #192 c,  #168 c,  0 c,  50 c,

create zero-mac		\ -- addr
\ *G Holds MAC address 00-00-00-00-00-00
  0 c,  0 c,  0 c,  0 c,  0 c,  0 c,

create EtherBC		\ -- addr
\ *G Holds MAC address FF-FF-FF-FF-FF-FF
  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,

struct EHDR
  6 field EHDR.DstMAC
  6 field EHDR.SrcMAC
  2 field EHDR.type
end-struct

struct ARP
  EHDR field      _ehdr1
  2 field AHDR.htype
  2 field AHDR.ptype
  1 field AHDR.hlen
  1 field AHDR.plen
  2 field AHDR.op
  6 field ARP.senderMAC
  4 field ARP.senderIP
  6 field ARP.targetMAC
  4 field ARP.targetIP
end-struct

$0800 equ ET_IP
$0806 equ ET_ARP

$0001 equ ARPHARD_ETHER
$0800 equ ARPPROTO_IP

$0001 equ ARPOP_REQUEST
$0002 equ ARPOP_REPLY

: SendArpReq	\ -- ; send ARP request
\ *G Send an ARP request.
\ ** All initialisation must have been performed before
\ ** using this word.
  pbuff >r

  EtherBC r@ EHDR.DstMac 6 cmove
  EtherAddress   r@ EHDR.SrcMac 6 cmove
  ET_ARP         r@ EHDR.type w!(n)

  ARPHARD_ETHER  r@ AHDR.htype w!(n)
  ARPPROTO_IP    r@ AHDR.ptype w!(n)
  6              r@ AHDR.hlen c!
  4              r@ AHDR.plen c!
  ARPOP_REQUEST  r@ AHDR.op w!(n)

  EtherAddress   r@ ARP.SenderMAC 6 cmove
  IpAddress      r@ ARP.SenderIP 4 cmove
  zero-mac       r@ ARP.TargetMAC 6 cmove
  requested-ip   r@ ARP.TargetIP 4 cmove

  r> ARP send_ether_pkt
;

: pingonce	\ -- ; an ARP request
\ *G Send an ARP request and process the response once.
\ ** All initialisation must have been performed before
\ ** using this word.
  cr  ticks				\ starting time
  SendArpReq
  #1000 later begin
    isrx? if
      pbuff #1536 get_ether_pkt .
      pbuff EHDR.type w@(n) ET_ARP =
      pbuff AHDR.op w@(n) ARPOP_REPLY =  and
      pbuff ARP.targetMAC EtherAddress 6 s=  and
    else
      dup expired dup
      if  ." Timed out "  endif
    endif
  until
  drop
  ticks swap - . ." ms"			\ elapsed time
;

: pingit	\ --
\ *G Ping the IP address at *\fo{REQUESTED-IP} until a key is pressed.
  InitEther
  begin
    pingonce  #500 ms  key?
  until
  ( Stop-Timers )
;

[then]


\ ******
\ *> ###
\ ******

decimal

