\ EtherLPC17xx.fth - Ethernet driver for LPC17xx devices

((
Copyright (c) 2008, 2010, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20110728 SFP003 Converted to LPC17xx.
20101118 FVD003 Added EtherVerbose? control.
20081124 SFP002 Refactored Ethernet and PHY initialisation.
20081016 SFP001 First release.
))

only forth definitions
decimal

\ ===============
\ *! etherlpc17xx
\ *T LPC17xx Ethernet driver
\ ===============
\ *P This code requires modification for use with Revision '-'
\ ** devices. The PHY is initialised in RMII mode.


\ ****************
\ *S Configuration
\ ****************

interpreter
: const equ ;	\ n -- ; -- n
\ *G If you are using a standalone target with heads and you want
\ ** interactive access to all the registers and bit masks, define
\ ** *\fo{CONST} as *\fo{CONSTANT}, otherwise by default *\fo{CONST}
\ ** is defined as *\fo{EQU}.
target

[undefined] EtherVerbose? [if]
1 equ EtherVerbose?	\ -- flag
\ *G If non-zero, Ethernet initialisation produces messages.
[then]

[undefined] EtherCom? [if]
0 equ EtherCom?	\ -- flag
\ *G Only needed if the EtherCom generic I/O device is being used.
\ ** This only happens for systems with multiple IP ports and
\ ** PowerNet v3.x upwards.
[then]

[undefined] MultiPhy? [if]
0 equ MultiPhy?	\ -- flag
\ *G Set this flag true if you are using multiple PHYs or do not
\ ** know the address of the PHY.
[then]

[undefined] EtherDiags? [if]
0 equ EtherDiags?	\ -- flag
\ *G Set this equate true to compile diagnostic code for
\ ** register dumping and so on. False by default.
[then]

[undefined] sniff? [if]
0 equ sniff?	\ -- flag
\ *G Set this equate true to compile the packet sniffer code,
\ ** which can be used to test Ethernet reception.
[then]

[undefined] pingit? [if]
0 equ pingit?	\ -- flag
\ *G Set this equate true to compile the PING code,
\ ** which tests Ethernet transmission and reception.
[then]

[undefined] EtherAddress [if]
create EtherAddress	\ -- addr
\ *G Holds the Ethernet MAC address (six bytes). Note that you
\ ** must obtain these from the IEEE (www.ieee.org) or from other
\ ** sources.
  $00 c, $10 c, $8B c, $F1 c, $44 c, $51 c,	\ six bytes for 48 bit MAC address
[then]						\ you must obtain these yourself!

[undefined] IpAddress [if]
create IpAddress	\ -- addr
\ *G Holds the Ethernet IP address (four bytes).
  192 c,  168 c,  0 c,  251 c,		\ assign these as required
[then]


\ ********
\ *S Tools
\ ********

: emWait	\ --
\ *G A short wait because the example code does it.
  #512 0 do loop
;


\ ******************
\ *S RAM allocations
\ ******************
\ *P The Ethernet unit has its own 16kb RAM area. This is the
\ ** *\fo{ERAM} section in the control file.

7 equ #ERX		\ -- u
\ *G Number of Ethernet receive buffers
3 equ #ETX		\ -- u
\ *G Number of Ethernet transmit buffers
#1536 equ /edata	\ -- len
\ *G Size of an Ethernet data buffer.
8 equ /edesc		\ -- len
\ *G Size of an Ethernet descriptor.
4 equ /etxStatDesc	\ -- len
\ *G Size of an Ethernet transmit status descriptor.

/edesc #ERX * equ /ERXCtlDescs
/edesc #ERX * equ /ERXStatDescs
/edesc #ETX * equ /ETXCtlDescs
/etxStatDesc #ETX * equ /ETXStatDescs

[sections ERAM cdata
/edata #ERX * buffer: ERXbuffs		\ -- addr
\ *G Receive data buffers.
/edata #ETX * buffer: ETXbuffs		\ -- addr
\ *G Transmit data buffers.
/ERXCtlDescs buffer: ERXCtlDescs	\ -- addr
\ *G Receive control descriptors. The control and status
\ ** descriptors must be contiguous.
/ERXStatDescs buffer: ERXStatDescs	\ -- addr
\ *G Receive status descriptors.
/ETXCtlDescs buffer: ETXCtlDescs	\ -- addr
\ *G Transmit control descriptors.
/ETXStatDescs buffer: ETXStatDescs	\ -- addr
\ *G Transmit status descriptors.
ERXCtlDescs equ DescStart	\ -- addr
\ *G Start address of descriptors.
uhere ERXCtlDescs - equ /Descs	\ -- len
\ *G Length of descriptors.
sections]


\ **********************
\ *S Descriptor handling
\ **********************

struct /CtlDesc	\ -- len
\ *G Ethernet control descriptor.
  int ecd.addr
  int ecd.ctrl
end-struct

: initDescs	\ --
\ *G Initialise transmit and receive descriptors.
  DescStart /Descs erase
\ Receive Descriptors - status descriptors already erased
  _EMAC
  ERXCtlDescs over emRxDesc + !		\ control descriptors
  ERXStatDescs over emRxSt + !		\ status descriptors
  #ERX 1- over emRxDesc# + !		\ number of descriptors
  #ERX 0 do
    ERXbuffs /edata i * +		\ data buffer address
    ERXCtlDescs /edesc i * + 		\ descriptor address
    tuck ecd.addr !			\ buffer address
    /edata 1- swap ecd.ctrl !		\ buffer len, no interrupts
  loop
  0 over emRxConsInd + !		\ consumer index
\ Transmit descriptors - status descriptors already erased
  ETXCtlDescs over emTxDesc + !		\ control descriptors
  ETXStatDescs over emTxSt + !		\ status descriptors
  #ETX 1- over emTxDesc# + !		\ number of descriptors
  #ETX 0 do
    ETXbuffs /edata i * +		\ data buffer address
    ETXCtlDescs /edesc i * +		\ descriptor address
    tuck ecd.addr !			\ buffer address
    /edata 1- swap ecd.ctrl !		\ buffer len, no interrupts
  loop
  0 over emTxProdInd + !		\ producer index
  drop
;


\ *************
\ *S PHY access
\ *************

\ PHY registers
0x00 const PHY_CNTL_REG
0x01 const PHY_STATUS_REG
0x02 const PHY_ID_REG1
0x03 const PHY_ID_REG2
0x04 const PHY_ANA_REG
0x05 const PHY_ANLPAR_REG
0x06 const PHY_ANE_REG

\ PHY control register bits
bit15 const RESET_PHY
bit14 const ENABLE_LOOPBACK
bit13 const DR_100MB
bit12 const ENABLE_AN
bit11 const PHY_POWERDOWN
bit10 const PHY_MAC_ISOLATE
bit9  const RESTART_AN
bit8  const PHY_FULLDUPLEX
bit7  const PHY_COL_TEST

\ PHY Status register bits
bit5 const AN_COMPLETE

\ PHY Auto-negotiation Advertisement register.
\ Also applies to ANLPAR register.
bit8 const DR100_TX_FULL
bit7 const DR100_TX_HALF
bit6 const DR10_TX_FULL
bit5 const DR10_TX_HALF
DR100_TX_FULL DR100_TX_HALF or DR10_TX_FULL or
  const ANE_ALL	\ request everything

[undefined] Phy# [if]
1 equ Phy#	\ -- n
\ *G Single PHY: The number of the PHY in use: 0..31
\ ** This equate is only compiled if not already set.
\ ** Some boards, e.g. Olimex LPC1766-STK, require this to be
\ ** set to 1.
[then]
Phy# #8 lshift equ ~Phy#	\ -- mask
\ *G Single PHY: Mask used for PHY command generation.

((
1 value Phy#
: ~Phy#		\ -- mask
  Phy# #8 lshift
;
))

: ReadPhy	\ reg -- w
\ *G Read a PHY register. Note that these are 16 bit values.
  _EMAC
  $0001 over emMCMD + !			\ read command
  swap $1F and ~Phy# or over emMADR + !	\ register address
  begin
    dup emMIND + @ $0005 and
   while
    pause
  repeat
  $0000 over emMCMD + !			\ cancel command
  emMRDD + @ $FFFF and
;

: WritePhy	\ w reg --
\ *G Write a PHY register. Note that these are 16 bit values.
  _EMAC
  $0000 over emMCMD + !			\ write command
  swap $1F and ~Phy# or over emMADR + !	\ register address
  tuck emMWTD + !			\ data
  begin  dup emMIND + @ bit0 and while  pause  repeat
  drop
;

: OrPhy		\ mask reg --
\ *G Or mask into reg in the PHY. Set bits in mask are set in the
\ ** register.
  tuck ReadPhy  or  swap WritePhy
;

defer UserResetPhy	\ --
\ *G If you want to change the behaviour of the PHY, this word is
\ ** executed at the end of *\fo{resetPhy} below. The default
\ ** action is *\fo{noop}.
  assign noop to-do UserResetPhy

interpreter
decimal
: genPhyClk	\ -- div
  system-speed 1000000 /
  dup 10 <=
  if  drop 0 exit  endif
  dup 15 <=
  if  drop 2 exit  endif
  dup 20 <=
  if  drop 3 exit  endif
  dup 25 <=
  if  drop 4 exit  endif
  dup 35 <=
  if  drop 5 exit  endif
  dup 50 <=
  if  drop 6 exit  endif
  dup 70 <=
  if  drop 7 exit  endif
  dup 80 <=
  if  drop 8 exit  endif
  dup 90 <=
  if  drop 9 exit  endif
  dup 100 <
  if  drop 10 exit  endif
  dup 120 <
  if  drop 11 exit  endif
  dup 130 <
  if  drop 12 exit  endif
  dup 140 <=
  if  drop 13 exit  endif
  150 <
  if  14 exit  endif
  15
;
target
genPhyClk 2 lshift equ initMCFG
\ *G Divider field for MCFG register.

: configPhy	\ --
\ configure EMAC to run PHY
  _EMAC
  $8000 initMCFG + over emMCFG + !	\ reset PHY, slowest clock
  10 ms
  initMCFG over emMCFG + !		\ unreset PHY, slowest clock
  $0000 over emCommand + !
  $0200 over emCommand + !		\ RMII mode
  $0100 over emSUPP + !			\ 100 Mbps
  drop  emWait
;

: resetPhy	\ --
\ *G Start up the PHY with autonegotiation enabled, but do not
\ ** perform autonegotiation.

  configPhy

  RESET_PHY PHY_CNTL_REG OrPhy  emWait	\ software reset the chip
  begin					\ wait until reset complete
    pause
    PHY_CNTL_REG ReadPhy RESET_PHY and 0=
  until
  ANE_ALL PHY_ANA_REG OrPhy		\ request 10/100BaseT features
  ENABLE_AN PHY_CNTL_REG OrPhy		\ enable autonegotiation

  UserResetPhy				\ for CDS
;

: autoNegotiate	\ -- ior ; 0=success
\ *G Perform the autonegotiation sequence. If no link can be
\ ** established, return ior=non-zero. On success, return zero.
\ ** For most PHYs this is done by a timeout, say 5 seconds.
  ENABLE_AN PHY_CNTL_REG OrPhy		\ enable autonegotiation
\ cr ." Starting AutoNegotiation"
  RESTART_AN PHY_CNTL_REG OrPhy  emWait	\ start autonegotiation
  5000 later				\ limit to 5 seconds
  begin
    pause				\ wait until done
    dup timedout? 0=
   while
    PHY_STATUS_REG ReadPhy AN_COMPLETE and
  until then
  timedout?
  if  -1  exit  endif
\ cr ." AutoNegotiation done"
\ set up control reg using the results
  PHY_ANLPAR_REG ReadPhy
  0					\ -- partner cntlset
  over DR100_TX_FULL and
  if  DR_100MB PHY_FULLDUPLEX or or  endif
  over DR100_TX_HALF and
  if  DR_100MB or  endif
  over DR10_TX_FULL and
  if  PHY_FULLDUPLEX or  endif
  PHY_CNTL_REG WritePhy
  drop 0
;


\ *****************
\ *S Initialisation
\ *****************
\ *P The Ethernet I/O is on Port 1 and operates in RMII mode.
\ *E   P1.0      ENET_TXD0
\ **   P1.1      ENET_TXD1
\ **   P1.4      ENET_TXEN
\ **   P1.8      ENET_CRS
\ **   P1.9      ENET_RXD0
\ **   P1.10     ENET_RXD1
\ **   P1.14     ENET_RXER
\ **   P1.15     ENET_REF_CLK
\ **   P1.16     ENET_MDC
\ **   P1.17     ENET_MDIO

decimal

\ Rev '-' chips only
1 [ps
  0 +ps  1 +ps  4 +ps  6 +ps  8 +ps  9 +ps
  10 +ps  14 +ps  15  +ps
ps] equ enetPS2val- equ enetPS2mask-
\ *G Pin select value and mask for PINSEL2 - rev '-' chips only.
2 [ps
  0 +ps  1 +ps  4 +ps  6 +ps  8 +ps  9 +ps
  10 +ps  14 +ps  15  +ps
ps] equ enetPM2val- equ enetPM2mask-
\ *G Pin mode value and mask for PINMODE2 - rev '-' chips only.

\ Rev A chips onwards
1 [ps
  0 +ps  1 +ps  4 +ps  8 +ps  9 +ps
  10 +ps  14 +ps  15  +ps
ps] equ enetPS2valA equ enetPS2maskA
\ *G Pin select value and mask for PINSEL2 - rev A chips onwards.
2 [ps
  0 +ps  1 +ps  4 +ps  8 +ps  9 +ps
  10 +ps  14 +ps  15  +ps
ps] equ enetPM2valA equ enetPM2maskA
\ *G Pin mode value and mask for PINMODE2 - rev A chips onwards.

1 [ps
  16  +ps  17 +ps
ps] equ enetPS3val  equ enetPS3mask
\ *G Pin select value and mask for PINSEL3.
2 [ps
  16  +ps  17 +ps
ps] equ enetPM3val  equ enetPM3mask
\ *G Pin select value and mask for PINMODE3.

: setMACAddress	\ addr --
\ *G Set the MAC address in the EMAC unit. For a MAC address of
\ ** the form *\f{aa-bb-cc-dd-ee-ff}, the *\f{aa} byte must appear
\ ** on the wire first. In the NXP documentation, this is octet 6
\ ** of the station address. In MPE code, the six bytes of a MAC
\ ** address are stored sequentially in memory starting with the
\ ** *\f{aa} byte.
  _EMAC
  over 4 + w@ over emSA0 + !
  over 2 + w@ over emSA1 + !
  swap 0 + w@ swap emSA2 + !
;

: resetMAC	\ --
\ *G Reset and initialise the MAC unit.
  _EMAC
  $CF00 over emMAC1 + !			\ reset all MAC modules
  $0038 over emCommand + !  emWait	\ reset all datapaths
  $0000 over emMAC1 + !			\ clear resets
  $0000 over emCommand + !		\ stop TX and RX paths
  $0000 over emMAC2 + !			\ set after PHY initialisation
  $0C12 over emIPGR + !			\ recommended non back to back inter-packet gap
  $370F over emCLRT + !			\ recommended collision window / retry
  $0600 over emMAXF + !			\ max 1536 byte frame length
  drop
;

: configMAC	\ --
\ *G Configure the MAC after autonegotiation.
  _EMAC  PHY_CNTL_REG ReadPhy $2100 and	\ bit13=speed (1=100Mb), bit8=duplex (1=full)
  case					\ -- emac mask
    $0000 of	\ 10Mb, half duplex
       $0030 over emMAC2 + !		\ CRC/PAD
       bit8 over emSUPP + bic!		\ PHY 10Mb
       $0240 over emCommand + !		\ RMII and RUNT
       $0012 over emIPGT + !		\ half duplex IPG
    endof
    $0100 of	\ 10Mb, full duplex
       $0031 over emMAC2 + !		\ CRC/PAD, full duplex
       bit8 over emSUPP + bic!		\ PHY 10Mb
       $0640 over emCommand + !		\ FD, RMII and RUNT
       $0015 over emIPGT + !		\ full duplex IPG
    endof
    $2000 of	\ 100Mb, half duplex
       $0030 over emMAC2 + !		\ CRC/PAD
       bit8 over emSUPP + or!		\ PHY 100Mb
       $0240 over emCommand + !		\ RMII and RUNT
       $0012 over emIPGT + !		\ half duplex IPG
    endof
    $2100 of	\ 100Mb, full duplex
       $0031 over emMAC2 + !		\ CRC/PAD, full duplex
       bit8 over emSUPP + or!		\ PHY 100Mb
       $0640 over emCommand + !		\ FD, RMII and RUNT
       $0015 over emIPGT + !		\ full duplex IPG
    endof
  endcase
  drop
;

: configFilter	\ --
\ *G Set up the receive filtering.
  $0022 _EMAC tuck emRxFiltCTrl + !	\ station and broadcast only
  -1 over emRxFiltWolClr + !		\ no Wakeup-On-Lan (WOL)
  0 over emHashFiltL + !		\ no Hash filter
  0 over emHashFiltH + !
  drop
;

: EnableRxTx	\ --
\ *G Enable RX and TX.
  $0003 _EMAC tuck emCommand + or!	\ command reg
  $0001 swap emMAC1 + or!		\ and MAC receive
;

$3902:2000 equ EmacModuleId

: InitEHW	\ --
\ *G Power up and select the pins.
  ~pcENET _PCONP or!  emWait		\ power up
  0 _EMAC emPwrDn + !			\ remove power down lock
  _EMAC emModId + @ EmacModuleId <> if	\ Rev '-' device
    enetPS2val- enetPS2mask- _PINSEL2 setMask	\ configure pins
    enetPM2val- enetPM2mask- _PINMODE2 setMask
  else
    enetPS2valA enetPS2maskA _PINSEL2 setMask	\ configure pins
    enetPM2valA enetPM2maskA _PINMODE2 setMask
  endif
  enetPS3val enetPS3mask _PINSEL3 setMask	\ configure pins
  enetPM3val enetPM3mask _PINMODE3 setMask
;

: initEMAC	\ -- ior
\ *G Initialise the Ethernet system, returning zero on success.
  InitEHW  resetMAC			\ set up pins, Reset the MAC
  EtherAddress setMACaddress		\ set MAC address
  resetPhy  autoNegotiate		\ start PHY, get link
  if  -1  exit  endif
  configMAC  initDescs			\ set up MAC and descriptors
  configFilter  EnableRxTx		\ set up filters, enable RX and TX
  0
;

EtherVerbose? [if]
: InitEther	\ --
\ *G Perform a full initialisation of the chip, enabling
\ ** Rx and Tx, and setting up the Ethernet MAC address
\ ** from *\fo{EtherAddress}.
  cr ." Initialising Ethernet ... "
  InitEMAC
  if ." un" then ." successful"
;
[else]
: InitEther	\ --
  InitEMAC drop  ;
[then]

: EtherLink?	\ -- flag
\ *G Return true if the Ethernet link is established. For most
\ ** PHYs, this is performed by testing bit 2 of register 1.
  1 ReadPhy bit2 and 0<>
;


\ ********************
\ *S Packet read/write
\ ********************

: +RxConsInd	\ --
\ *G Step on to the next RX descriptor.
  _EMAC emRxConsInd + @
  1+  dup #ERX 1- >
  if  drop 0  endif
  _EMAC emRxConsInd + !
;

: IsRx?         \ -- flag ; nz = has data
\ *G Check if incoming data is present, returning true if a
\ ** packet is available for *\fo{get_ether_pkt} below.
  _EMAC dup emRxProdInd + @
  swap emRxConsInd + @ =		\ if prod=cons
  if  0  exit  endif			\ no data is available
  _EMAC emRxConsInd + @ /edesc *	\ offset into RX status descriptor
  ERXStatDescs + @			\ first cell is status
  dup $3B9C:0000 and 0=			\ error check, all zero
  swap bit30 and 0<>  and		\ and LastFlag set
  dup 0=				\ if bad, discard it
  if  +RxConsInd  endif
;

: get_ether_pkt \ *dest maxlen -- len
\ *G Get a pending receive packet to a buffer of size *\b{maxlen).
\ ** Return the number of bytes received, or zero for a bad packet.
\ ** Note that data is valid only when *\fo{ISRX?} returns true,
\ ** so that you must poll with *\fo{ISRX?} before using
\ ** *\fo{GET_ETHER_PKT}. The received size of the packet
\ ** (not clipped to the buffer size) is returned.
  _EMAC emRxConsInd + @ /edesc *	\ offset into RX status/control descriptors
  dup ERXStatDescs + @ $7FF and 1+ >r	\ first cell is status, 11 bit length
  ERXCtlDescs + ecd.addr @		\ -- *dest maxlen *src ; R: -- len
  -rot r@ min cmove  r>			\ -- len
  +RxConsInd
;

: DiscardRX     \ --
\ *G Throw away a pending input packet due to lack of host buffer
\ ** memory.
  +RxConsInd
;

: +TxProdInd	\ --
\ *G Step on to the next TX descriptor.
  _EMAC emTxProdInd + @
  1+  dup #ETX 1- >
  if  drop 0  endif
  _EMAC emTxProdInd + !
;

: (TxFull?)	\ -- flag
\ *G Return true if the TX system is full.
  _EMAC emTxConsInd + @ dup 0=		\ make Cons-1 positive/zero
  if  drop #ETX  endif
  1-  _EMAC emTxProdInd + @ =		\ (Cons-1 = Prod) means full
;

: TxFull?	\ -- flag
\ *G Return true if the TX system is full. Check whether the
\ ** transmitter has locked up by trying again after 3ms. If it has
\ ** locked up, reset the transmitter and try again.
  (TxFull?) if
    3 ms  (TxFull?) if			\ max frame time = 1.5ms @ 10Mbps
      bit1 _EMAC emCommand + bic!  1 ms	\ stop TX
      bit1 _EMAC emCommand + or!	\ restart TX, sets Cons=Prod
      3 ms  (TxFull?) exit
    endif
  endif
  0
;

: IsTx?		\ n -- flag
\ *G Return true if a packet of length n can be sent.
  drop TxFull? 0=
;

: send_ether_pkt	\ caddr len --
\ *G Send packet from supplied buffer. Note that we copy the
\ ** data to a private area so that the buffer can be released
\ ** immediately.

\ Perform sanity checks
  dup #14 #1520 within 0=		\ range check length
  if  2drop exit  endif
  TxFull? if				\ no buffer space?
cr ." Ether TX locked up"
    2drop exit
  endif
  #64 max				\ force length to min bytes

\ Copy packet to Ethernet DMA memory
  _EMAC emTxProdInd + @ /edesc * ETXCtlDescs +	\ -- caddr len desc
  over 1- bit30 or  over ecd.ctrl !	\ last bit and len-1
  ecd.addr @ swap cmove			\ copy to DMA buffer
  +TxProdInd				\ next TX buffer
\ At this point, the TX DMA will pick the packet up & send it if TX is running.
;


\ ****************************************
\ *S Generic I/O for PowerNet v3 and above
\ ****************************************

EtherCom? [if]

: GetAddrs	\ -- ipaddr macaddr 0
\ *G For Generic I/O IPDGetAddr function
  IpAddress EtherAddress 0
;

: SetAddrs	\ ipaddr|0 macaddr|0 mode --
\ *G Set up the Ethernet MAC and the IP addresses.
\ ** At present mode is always 0, but will be used in
\ ** future releases to indicate data formats. If
\ ** ipaddr or macaddr are zero, the corresponding stored data
\ ** will not be changed.
  drop					\ mode flags
  ?dup if				\ macaddr
    dup EtherAddress 6 cmove		\ save
    SetMACaddress			\ put in EMAC
  endif
  ?dup if				\ ipaddr
    IpAddress 4 cmove
  endif
;

\ *P The layout and usage of the Ethernet common I/O structure is
\ ** defined in the file *\i{COMMON\ETHERCOM.FTH}.
\ *E create IPdevice
\ **   ' MyInit ,       \ 0: initialisation
\ **   ' MyTerm ,	\ 1: shutdown
\ **   ' MyRx? ,	\ 2: receive test
\ **   ' MyRx		\ 3: receive packet
\ **   ' MyTx? ,	\ 4: transmit test
\ **   ' MyTx ,		\ 5: transmit packet
\ **   ' MyGetAddr ,    \ 6: Get device addresses
\ **   ' MySetAddr ,    \ 7: Set device addresses
\ **   ' MySave ,	\ 8: Save IP device state

create E23xxvector	\ -- addr
\ *G The device vector needed by PowerNet v3+ and
\ ** *\i{COMMON\ETHERCOM.FTH}.
  ' InitEther ,         \ initialisation
  ' noop ,	        \ shutdown
  ' IsRx? ,	        \ receive test
  ' get_ether_pkt ,     \ receive packet
  ' IsTx? ,	        \ transmit test
  ' send_ether_pkt ,    \ transmit packet
  ' GetAddrs ,		\ Get device addresses
  ' SetAddrs ,	        \ Set device addresses
  ' noop ,	        \ Save IP device state

E23xxvector constant IPDevice	\ -- addr
\ *G *\fo{IPDevice} is the default device name required by
\ ** *\{ETHERCOM.FTH}. If you have multiple ports, only one
\ ** should be called *\fo{IPDevice}.

[then]


\ **************
\ *S System test
\ **************
\ *P The packet sniffer code is only compiled if the equate
\ ** *\fo{SNIFF?} is non-zero.

EtherDiags? [if]

_EMAC constant _EMAC
: .em		\ --
  _EMAC $20 pdump  ;
: .ed		\ --
  _EMAC $100 + $30 pdump  ;
: .td		\ --
  ETXCtlDescs $40 pdump  ;

: .phy		\ --
\ *G Display contents of the useful PHY registers.
  cr ."  0 Control: " 0 readphy .word
  cr ."  1  Status: " 1 readphy .word
  cr ."  2 PhyID#1: " 2 readphy .word
  cr ."  3 PhyID#2: " 3 readphy .word
  cr ."  4 AutoNeg: " 4 readphy .word
  cr ."  5 Ability: " 5 readphy .word
  cr ."  6 Autoexp: " 6 readphy .word
  cr ." 17 Conf/St: " #17 readphy .word
  cr ." 18 TCIE:    " #18 readphy .word
;

((
: .phys		\ --
\ *G If *\fo{Phy#} is a *\fo{VALUE}, then this word can be
\ ** uncommented  and used to find the PHY on the board.
  32 0 do
    i to Phy#  cr ." Phy# " i .  .phy
  loop
;
))
[then]

sniff? [if]

1536 buffer: pbuff	\ -- addr
\ *G Temporary buffer for *\fo{SNIFF}.

: NextRxPkt	\ --
\ *G Collect and display packet
  pbuff 1536 erase
  pbuff 1536 get_ether_pkt dup .word
  pbuff swap 1 max dump
;

: sniff		\ --
\ *G Listens to the network and displays all the traffic that
\ ** has a broadcast destination or is for this device.
\ ** *\fo{SNIFF} can be used to test reception.
  begin
    isrx?
    if  NextRxPkt  endif
    #10 ms
[ tasking? ] [if] pause [then]
    key?
  until
;

[then]

\ *P The PING example is used to test Ethernet transmission and
\ ** reception. Before testing transmission, test reception by
\ ** using the maintenance system to download a user application
\ ** and then pinging the target board instead of TFTPing to it.
\ *E Example ARP request:  #42/$2A bytes
\ ** C020.0600  FF FF FF FF FF FF 00 C0  F0 0D 9D 7E 08 06 00 01
\ ** C020.0610  08 00 06 04 00 01 00 C0  F0 0D 9D 7E C0 A8 01 7F
\ ** C020.0620  00 00 00 00 00 00 C0 A8  FF FE 11 0A 75 9C C0 A8

\ *P Note that the equates *\fo{SNIFF?} and *\fo{PINGIT?} must both
\ ** be set non-zero to compile the PING code.

pingit? [if]

create requested-ip	\ -- addr
\ *G Holds IP address to be pinged.
  #192 c,  #168 c,  0 c,  50 c,

create zero-mac		\ -- addr
\ *G Holds MAC address 00-00-00-00-00-00
  0 c,  0 c,  0 c,  0 c,  0 c,  0 c,

create EtherBC		\ -- addr
\ *G Holds MAC address FF-FF-FF-FF-FF-FF
  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,  $FF c,

struct EHDR
  6 field EHDR.DstMAC
  6 field EHDR.SrcMAC
  2 field EHDR.type
end-struct

struct ARP
  EHDR field      _ehdr1
  2 field AHDR.htype
  2 field AHDR.ptype
  1 field AHDR.hlen
  1 field AHDR.plen
  2 field AHDR.op
  6 field ARP.senderMAC
  4 field ARP.senderIP
  6 field ARP.targetMAC
  4 field ARP.targetIP
end-struct

$0800 equ ET_IP
$0806 equ ET_ARP

$0001 equ ARPHARD_ETHER
$0800 equ ARPPROTO_IP

$0001 equ ARPOP_REQUEST
$0002 equ ARPOP_REPLY

: SendArpReq	\ -- ; send ARP request
\ *G Send an ARP request.
\ ** All initialisation must have been performed before
\ ** using this word.
  pbuff >r

  EtherBC r@ EHDR.DstMac 6 cmove
  EtherAddress   r@ EHDR.SrcMac 6 cmove
  ET_ARP         r@ EHDR.type w!(n)

  ARPHARD_ETHER  r@ AHDR.htype w!(n)
  ARPPROTO_IP    r@ AHDR.ptype w!(n)
  6              r@ AHDR.hlen c!
  4              r@ AHDR.plen c!
  ARPOP_REQUEST  r@ AHDR.op w!(n)

  EtherAddress   r@ ARP.SenderMAC 6 cmove
  IpAddress      r@ ARP.SenderIP 4 cmove
  zero-mac       r@ ARP.TargetMAC 6 cmove
  requested-ip   r@ ARP.TargetIP 4 cmove

  r> ARP send_ether_pkt
;

: pingonce	\ -- ; an ARP request
\ *G Send an ARP request and process the response once.
\ ** All initialisation must have been performed before
\ ** using this word.
  cr  ticks				\ starting time
  SendArpReq
  #1000 later begin
    isrx? if
      pbuff #1536 get_ether_pkt .
      pbuff EHDR.type w@(n) ET_ARP =
      pbuff AHDR.op w@(n) ARPOP_REPLY =  and
      pbuff ARP.targetMAC EtherAddress 6 s=  and
    else
      dup expired dup
      if  ." Timed out "  endif
    endif
  until
  drop
  ticks swap - . ." ms"			\ elapsed time
;

: pingit	\ --
\ *G Ping the IP address at *\fo{REQUESTED-IP} until a key is pressed.
  InitEther
  begin
    pingonce  #500 ms  key?
  until
  ( Stop-Timers )
;

[then]


\ ******
\ *> ###
\ ******

decimal

