\ dmaLPC40xx.fth - LPC40xx GPDMA driver

((
Copyright (c) 2008, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20150720 SFP002 LPC40xx conversion from LPC23xx
))


\ =============
\ *! dmalpc40xx
\ *T DMA driver for LPC40xx
\ =============

\ *****************
\ *S DMA primitives
\ *****************
\ *P The DMA engine has problems transferring data to the primary
\ ** internal RAM block. We recommend that you use the upper portion
\ ** of the USB block, which has to be powered up.

\ *P By default, MPE code for LPC40xx devices uses DMA
\ ** channel 0 for the MCI driver and DMA channel 1 for the USB
\ ** driver.

: initDMA	\ --
\ *G Initialise the GPDMA controller. The USB block is also
\ ** powered up so that the USB RAM can be used.
  ~pcUSB ~pcGPDMA or _PCONP or!		\ power up USB and GPDMA
  _GPDMA
  $FF over dmaIntTCClr + !		\ clear interrupts 0..7
  $FF over dmaIntErrClr + !
  $01 over dmaConfig + !		\ DMA is little-endian, enabled
  drop  20 0 do loop
;

$02  12 lshift equ /SBsize	\ source burst = 8
$02  15 lshift equ /DBsize	\ destination burst = 8
$02  18 lshift equ /SBwidth	\ source data = 32 bits
$02  21 lshift equ /DBwidth	\ dest data = 32 bits
1    26 lshift equ SIm2p	\ increment source after transfer
0    27 lshift equ DIm2p	\ do not increment dest after transfer
%101 ( 0 ) 28 lshift equ ProtM2p	\ cacheable, non-buff, privileged
				\ unused in LPC40xx
1    31 lshift equ TCIntEn	\ enable termal count interrupt

0    26 lshift equ SIp2m	\ do not increment source after transfer
1    27 lshift equ DIp2m	\ increment dest after transfer
%011 28 lshift equ ProtP2M	\ non-cacheable, buff, privileged
				\ unused in LPC40xx

TCIntEn  ProtM2p or  DIm2p or SIm2p or
/DBwidth or  /SBwidth or  /DBsize or  /SBsize or
  equ CtrlValM2P	\ -- mask
\ *G Channel control register value for M2P transfers.

TCIntEn  ProtP2m or  DIp2m or SIp2m or
/DBwidth or  /SBwidth or  /DBsize or  /SBsize or
  equ CtrlValP2M	\ -- mask ; B849:2000
\ *G Channel control register value for P2M transfers.

$020 equ /DmaCh		\ -- bytes
\ *G Number of bytes for each DMA channel
    $000 equ dmaChSrcAddr
    $004 equ dmaChDestAddr
    $008 equ dmaChLLI
    $00C equ dmaChCtrl
    $010 equ dmaChConfig

compiler
: n>dmaCh	\ n -- addr
\ *G Get the base address of DMA channel n.
  /DmaCh * _GPDMA DmaChan0 + +
;
target

: clrDmaChn	\ n --
\ *G Clear DMA channel *\i{n} (0..7).
  1 swap lshift  _GPDMA
  2dup dmaIntTCClr + !			\ clear indicators
  dmaIntErrClr + !
;

: +m2pn		\ src dest cfg n --
\ *G Initialise DMA channel *\i{n} for a memory to peripheral transfer.
\ ** The peripheral is the flow controller, so a length is not needed.
\ ** The *\i{cfg} value is set in the *\fo{dmaChConfig} register.
  2swap rot n>DmaCh			\ -- cfg src dest dmachn
  tuck dmaChDestAddr + !		\ set dest (periph) address
  tuck dmaChSrcAddr + !			\ set source (memory) address
  0 over dmaChLLI + !			\ no linked list
  CtrlValM2P over dmaChCtrl + !		\ set control reg
  dmaChConfig + !			\ must be last
;

: +p2mn		\ src dest cfg n --
\ *G Initialise DMA channel *\i{n} for a peripheral to memory transfer.
\ ** The peripheral is the flow controller, so a length is not needed.
\ ** The *\i{cfg} value is set in the *\fo{dmaChConfig} register.
  dup clrDmaChn
  2swap swap rot n>DmaCh		\ -- cfg dest src dmachn
  tuck dmaChSrcAddr + !			\ set source (memory) address
  tuck dmaChDestAddr + !		\ set dest (periph) address
  0 over dmaChLLI + !			\ no linked list
  CtrlValP2M over dmaChCtrl + !		\ set control reg
  dmaChConfig + !			\ must be last
;

: -dman		\ n --
\ *G Finish the transfer on DMA channel n.
  n>DmaCh  $00 swap dmaChConfig + !	\ turn everything off
;

: dmanErr?	\ n -- flag
\ *G Return non-zero if there is an error on DMA channel *\i{n}.
  1 swap lshift _GPDMA dmaRawIntErrSt + @ and
;


\ *************************
\ *S Channel specific words
\ *************************

: +m2p0		\ src dest cfg --
\ *G Initialise DMA channel 0 for a memory to peripheral transfer.
\ ** The peripheral is the flow controller, so a length is not needed.
\ ** The *\i{cfg} value is set in the *\fo{dma0Config} register.
  0 +m2pn
;

: +p2m0		\ src dest cfg --
\ *G Initialise DMA channel 0 for a peripheral to memory transfer.
\ ** The peripheral is the flow controller, so a length is not needed.
\ ** The *\i{cfg} value is set in the *\fo{dma0Config} register.
  0 +p2mn
;

: -dma0		\ --
\ *G Finish the transfer on DMA channel 0.
  0 -dman
;

: dma0Err?	\ -- flag ; true on error
\ *G Return non-zero if there is an error on DMA channel 0.
  0 dmanErr?
;


\ **************
\ *S DMA for MCI
\ **************

\ 4 equ mciDma#	\ LPC23xx/24xx
1 equ mciDma#	\ LPC40xx
$1E801 mciDMA# 6 lshift or equ M2PvalMCI	\ -- x
\ *G DMA channel configuration register value for an MCI M2P
\ ** transfer, flow control by the peripheral, i.e. a write to
\ ** the card.
$1F001 mciDMA# 1 lshift or equ P2MvalMCI	\ -- x
\ *G DMA channel configuration register value for an MCI P2M
\ ** transfer, flow control by the peripheral, i.e. a read from
\ ** the card.


\ ======
\ *> ###
\ ======

decimal
