\ EEPROM_LPC1788.fth - on-chip EEPROM driver
((
Copyright (c) 2014
Culver Consultancy Ltd.
www.culverconsultancy.co.uk
Written by Barry Culver.

Destributed for use with MPE Compiler Products by:
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)

Driver for the on-chip EEPROM memory

This driver is for the LPC17xx family devices that have EEPROM
The size of the EEPROM block will vary.

We don't use interrupts in this code - wait while reads
or writes complete.  If using the multi-tasker resource locks are
used to prevent tasks simultaneously accessing EEPROM, and PAUSE is
used while waiting for write to complete.

To do
=====

Change history
==============
20140402 BJC001	First release.
))

only forth definitions
decimal		\ use $ prefix for hex

tasking? [if]
SEMAPHORE EE_Res			\ resource lock for EERPOM writes
  1 EE_Res !				\ single device instance
[endif]

\ This word should not be called unless the EEPROM block is idle
\ Note: EEPROM is powered up following a reset
: EEPROM-DIS		\ -- ; disable the EEPROM block to save power
  bit0 _EEPWRDWN !			\ turn off power
;

\ EEPROM wait and clock calculations

\ Calc min 15ns delay as system clocks
#10000000
system-speed #1000 /  /			\ clock period in 0.1 ns units
1 max					\ ensure non-zero
#15					\ desired period, ns
#10 * swap /mod
swap 0<> abs +				\ round up
1- equ EE_PHASE3_VAL	\ adjust for 1 based register setting

\ repeat for 55ns phase 2
#10000000
system-speed #1000 /  /			\ clock period in 0.1 ns units
1 max					\ ensure non-zero
#55					\ desired period, ns
#10 * swap /mod
swap 0<> abs +				\ round up
1- equ EE_PHASE2_VAL	\ adjust for 1 based register setting

\ finally phase 1; 35ns
#10000000
system-speed #1000 /  /			\ clock period in 0.1 ns units
1 max					\ ensure non-zero
#35					\ desired period, ns
#10 * swap /mod
swap 0<> abs +				\ round up
1- equ EE_PHASE1_VAL	\ adjust for 1 based register setting

\ merge into register value
EE_PHASE3_VAL
EE_PHASE2_VAL 8 lshift or
EE_PHASE1_VAL 16 lshift or
  equ EEWSTATE_VAL

\ Calculate division factor for EEPROM clock
  system-speed 375600 /			\ required frequency is 375.6 kHz
  1- equ EECLKDIV_VAL 			\ adjust for 1 based register setting


: EEPROM-INIT		\ --; initialise the on-chip EEPROM
  _EEPROM
  0 over EEPWRDWN + !			\ power up
  EEWSTATE_VAL over EEWSTATE + !	\ set wait states
  EECLKDIV_VAL over EECLKDIV + !	\ set clock divider
  bit26 bit28 or over EEENCLR + !	\ clear interrupt enables
  bit26 bit28 or swap EESTATCLR + !	\ clear interrupt status flags
  1 ms															\ allow EEPROM to start up (actually needs 100us)
;
' eeprom-init AtCold

((		EEPROM ACCESS WORDS
			~~~~~~~~~~~~~~~~~~~
	EEPROM addresses are not part of the general address space.  The words
	below use an address range of 0 through to the highest implemented address
	(2047 or 4095 for current devices - see data sheet).

	Be aware that writing is slow - circa 3ms in current devices.  The WRITE
	words will block (using pauses if multi-tasker in use).

	Read access is somewhat faster, but still orders of magnitude slower than
	for RAM or FLASH.  Read words will stall the bus while awaiting data.

	The implementation here is geared for simplicity rather than speed, so no
	interrupts.

	IMPORTANT: Reads and writes of word (16 bit) and doubleword (32 bit) values
	MUST USE ALIGNED ADDRESSES.  I.E. word read writes to EVEN addresses,
	doublewords must use addresses with bits 0,1 = 0.

	The EEPROM control block has no error flags - if it is unhappy with anything
	either the operation just does not work, or you get a bus error.
))

\ Reads are straightforward, though they may cause brief bus stalls.
\ Note that a 32 bit read will cause a longer stall.

: EE-RD-BYTE	\ addr -- byte ; read a byte value
  _EEPROM
  swap over EEADDR + !			\ select the address
  EE_CMD_8bREAD over EECMD + !		\ request a read
  EERDATA + @				\ get the data
;

: EE-RD-16B	\ addr -- word ; read a 16 bit value from an EVEN address
  _EEPROM
  swap over EEADDR + !			\ select the address
  EE_CMD_16bREAD over EECMD + !		\ request a read
  EERDATA + @				\ get the data
;

\   Note that a 32 bit read will cause a longer bus stall as the hardware
\		treats this as two successive 16b transfers.
: EE-RD-32B	\ addr -- lword; read a 32 bit value from an ALIGNED address
  _EEPROM
  swap over EEADDR + !			\ select the address
  EE_CMD_32bREAD over EECMD + !		\ request a read
  EERDATA + @				\ get the data
;

(( WRITING TO EEPROM
EEPROM Write is a two stage process; first the byte / word / dword is
written to the page register, then a erase, write operation loads the
page register data back into the EEPROM array.  The erase, write takes
3 ms to complete.  This implementation uses polling.

We use three primitives to do a write:
(EE-WR-xxxx)		sets a 8b/16b/32b value in the page register
[repeat if required for other data in the same 64 bit page]
(EE-PROG)		erases, writes page to EEPROM, non-blocking
(EE-WAIT-COMPLETE)	spins until the page programming completes
))

: (EE-WR-BYTE)	\ byte addr -- ; write a byte value into the page register
  _EEPROM
  swap over EEADDR + !			\ select the address - only bits 0:5 are used (page offset)
  EE_CMD_8bWRITE over EECMD + !		\ request a page register write
  EEWDATA + !				\ write data to page register
;

: (EE-WR-16B)	\ word addr -- ; write a 16 bit value into the page register
  _EEPROM
  swap over EEADDR + !			\ select the address - only bits 0:5 are used (page offset)
  EE_CMD_16bWRITE over EECMD + !	\ request a page register write
  EEWDATA + !				\ write data to page register
;

: (EE-WR-32B)	\ lword addr --; write a 32 bit value into the page register
  _EEPROM
  swap over EEADDR + !			\ select the address - only bits 0:5 are used (page offset)
  EE_CMD_32bWRITE over EECMD + !	\ request a page register write
  EEWDATA + !									\ write data to page register
;

\ This word leaves the EEPROM controller in a busy state for typ. 3ms
: EE-PROG	\ addr --; erase and program the page register into EEPROM
  _EEPROM
  ~EESTAT_PROG over EESTATCLR + !	\ ensure flag is clear
  swap over EEADDR + !			\ set page address (ls 6 bits ignored)
  EE_CMD_EP_PAGE swap EECMD + !		\ request erase and program
;

: EE-WAIT-RDY		\ --; await erase, program completion
  [ tasking? ] [if]  EE_Res request  [endif]
  _EEPROM  begin
    [ tasking? ] [if]  pause  [then]
    dup EESTAT + @
    ~EESTAT_PROG and
   until
  ~EESTAT_PROG swap EESTATCLR + !	\ ensure completed flag is clear
  [ tasking? ] [if]  EE_Res signal  [endif]
;

\ High level words to write EEPROM values, awaiting completion
\ These words will pause, so take care if other tasks may use EEPROM
\  - risk of 'deadly embrace'!

: EE-WR-BYTE	\ byte addr --
  dup >r
  (ee-wr-byte)			\ write to page register
  r> ee-prog
  ee-wait-rdy
;

: EE-WR-16B	\ 16b addr --
  dup >r
  (ee-wr-16b)			\ write to page register
  r> ee-prog
  ee-wait-rdy
;

: EE-WR-32B	\ 32b addr --
  dup >r
  (ee-wr-32b)			\ write to page register
  r> ee-prog
  ee-wait-rdy
;

true [if]		\ test code

_EEPROM EECMD + constant _EECMD
_EEPROM EEADDR + constant _EEADDR
_EEPROM EESTAT + constant _EESTAT
_EEPROM EEWDATA + constant _EEWDATA
_EEPROM EERDATA + CONSTANT _EERDATA

: EE-SHO-REGS		\ --
  _EEPROM
  cr ." Status   " _EESTAT @ .lword
  cr ." Address  " _EEADDR @ .lword
  cr ." Cmd reg. " _EECMD @ .lword
  cr ." Rd data  " _EERDATA @ .lword
  drop
;

: EE-CLRSTAT		\ --
  bit26 bit28 or _EEPROM EESTATCLR + !
;

: EE-SHO-STAT
  cr
	_EEPROM EESTAT + @
	dup bit26 and if ." RDWR " endif
	bit28 and if ." PROG" endif
;

[endif]
