\ gpioK60.fth - easy access to Freescale K60 GPIO pins.

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20111115 SFP001 Adapted from Nuvoton version.
))

only forth definitions
decimal

\ ===========
\ *! gpiok60
\ *T Freescale K60 GPIO utilities
\ ===========
\ *P The code in *\i{Cortex\Drivers\gpioK60.fth} provides
\ ** utility words for accessing GPIO pins on a bit by bit basis.
\ ** The code is written for ease of use rather than performance.


\ ********************
\ *S Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number. Bits 0..31 form
\ ** P0.0 to P0.31, bits 32..63 form P1.0 to P1.31 and so on.
\ ** Although Freescale name their ports from A, here we number
\ ** them from 0.

create PortTable	\ -- addr
\ *G Holds data about each port.
  _GPIOA ,  _PORTA portPCR + ,		\ pin, control blocks
  _GPIOB ,  _PORTB portPCR + ,
  _GPIOC ,  _PORTC portPCR + ,
  _GPIOD ,  _PORTD portPCR + ,
  _GPIOE ,  _PORTE portPCR + ,
2 cells equ /PortItem

\ *P We can make life easier with symbolic names for the GPIO ports.
\ *[
0 constant PTA
1 constant PTB
2 constant PTC
3 constant PTD
4 constant PTE
\ *]

target-only
: PIO:		\ port# bit# -- ; -- iobit#
\ *G Define a port I/O bit by name. For example
\ *C   PTC 11 PIO: PTC.11
\ ** defines an I/O bit on port 2, bit 11. At run time, the I/O
\ ** bit number is returned.
  create
    over /PortItem * PortTable + @ ,	\ pins base
    1 over lshift ,			\ bit mask
    swap /PortItem * PortTable + cell + @ \ PORT address
    swap 2 lshift or ,			\ PCRx address
;
host&target
interpreter
: PIO:		\ port# bit# -- ; -- iobit#
  create
    over /PortItem * PortTable + @ ,	\ pins base
    1 over lshift ,			\ bit mask
    swap /PortItem * PortTable + cell + @ \ PORT address
    swap 2 lshift or ,			\ PCRx address
;
target


\ ******************
\ *S GPIO pin access
\ ******************

: setPin	\ struct --
\ *G Set the pin high.
  2@ ptPSOR + !  ;

: clrPin	\ struct --
\ *G Set the pin low.
  2@ ptPCOR + !  ;

: getPin	\ struct -- 0/1
\ *G Read the pin state.
  2@ ptPDIR + @ and 0<> 1 and
;


\ ***********************
\ *S IO pin configuration
\ ***********************

: setPinCtrl	\ value mask struct --
\ *G Apply value and mask to the pin's control register.
  2 cells + @ setMask
;

: isPinMode	\ mode struct --
\ *G Sets the GPIO mode for a pin. Use with the constants below.
  swap 8 lshift  $0000:0700		\ MUX value and mask
  rot setPinCtrl			\ apply to pin control reg
;

: pinBit	\ 0/nz bit# struct --
\ *G Set/clear the given bit in the control regster.
  >r
  swap 0<> 1 and over lshift		\ -- bit# value ; R: -- struct
  swap 1 lshift				\ -- value mask ; R: -- struct
  r> setPinCtrl				\ --
;

\ *P Pins can be in one of eight modes, 0..7. In the majority of
\ ** cases, mode 0 means disabled and mode 1 means GPIO mode.
\ ** Most multi-function pins are disabled after reset.
\ ** Consult the Chip Configuration chapter of the chip's Reference
\ ** Manual for the full details.

0 constant UnusedMode	\ -- mode#
\ *G The pin is unused or analog.
1 constant GPIOmode	\ -- mode#
\ *G Specifies GPIO mode.

: isInput	\ struct --
\ *G Set the pin to GPIO mode and input.
  GPIOmode over isPinMode
  2@ ptPDDR + bic!
;

: isOutput	\ struct --
\ *G Set the pin to GPIO mode and output.
  GPIOmode over isPinMode
  2@ ptPDDR + or!
;

: isPinStrength	\ 0/1 struct --
\ *G Set the pin driver to low (0) or high (nz) drive.
  6 swap pinBit  ;

: isPinOD	\ 0/1 struct --
\ *G Enable (0) or disable (nz) the open drain driver.
  5 swap pinBit  ;

: NotPulled	\ struct --
\ *G No pull up or pull down.
  0 1 rot pinBit  ;

: PulledUp	\ struct --
\ *G Enable the pin pull up resistor.
  >r  1 0 r@ pinBit  1 1 r> pinBit
;

: PulledDown	\ struct --
\ *G Enable the pin pull down resistor.
  >r  0 0 r@ pinBit  1 1 r> pinBit
;


\ ======================================
\ *S Test code for Freescale TWR-K60N512
\ ======================================
\ *[
0 [if]
decimal

0 11 pio: led1		\ GPA.11
0 28 pio: led2		\ GPA.28

: tled1		\ --
  led1 isOutput
  begin
    led1 clrPin  200 ms
    led1 setPin  200 ms
  key? until
;

: tled2		\ --
  led2 isOutput
  begin
    led2 clrPin  200 ms
    led2 setPin  200 ms
  key? until
;
[then]
\ *]


\ ======
\ *> ###
\ ======

decimal
