\ gpioLPC40xx.fth - easy access to LPC40xx family GPIO pins.

((
Copyright (c) 2010, 2012, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20150622 SFP004 Enhanced for LPC40xx.
20140107 BJC003 Converted IO configuration for LPC1788 family.
20120514 PAK002 Fixed isPinMode.
20101027 SFP001 Adapted from ARM version.
))

only forth definitions
decimal

\ ===========
\ *! gpiolpc1788
\ *T LPC1788 GPIO utilities
\ ===========
\ *P The code in *\i{Cortex\Drivers\gpioLPC1788.fth} provides
\ ** tools to access the GPIO/FIO pins on a bit by bit basis.
\ ** The code is written for ease of use rather than performance.

\ *P There are similar files in the *\i{Cortex\Drivers} directory
\ ** for different CPU families.


\ ****************
\ *S Configuration
\ ****************
\ *P The equates in this block are defaults used if the equates
\ ** have not been defined before this file is compiled.

interpreter
: within?  within?  ;
target

_IOCON equ IoconBase	\ -- addr
\ *G Base address of Pin I/O control registers. One four-byte
\ ** register per pin.
_FIO0 equ PortBase	\ -- addr
\ *G Base address of GPIO port usage registers. Each port
\ ** has up to 8 registers in 32 byte blocks.
$20 equ /PortBlock	\ -- len
\ *G Number of bytes per port for peripheral registers.
5 equ #PortShift	\ -- u
\ *G Number of bits to shift to offset by *\fo{/PortBlock}.


\ ********************
\ *S Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number. Bits 0..31 form
\ ** P0.0 to P0.31, bits 32..63 form P1.0 to P1.31 and so on.

target-only
: PIO:		\ port# bit# -- ; -- iobit#
\ *G Define a port I/O bit by name. For example
\ *C   2 11 PIO: P2.11
\ ** defines an I/O bit on port 2, bit 11. At run time, the I/O
\ ** bit number is returned.
  swap 5 lshift or constant  ;
host&target
interpreter
: PIO:		\ port# bit# -- ; -- iobit#
  swap 5 lshift or constant  ;
target

\ *P Define the ports as constants.
\ *[
0 constant P0
1 constant P1
2 constant P2
3 constant P3
4 constant P4
5 constant P5
\ *]


\ ******************************
\ *S Pin configuration functions
\ ******************************
\ *P These words perform operations on the IOCON registers.

: changeIOCON	\ data mask iobit# --
\ *G Apply the *\i{data} to the *\i{mask} bits of the IOCON
\ ** register for the given *\i{iobit#}.
  cells IoconBase + setMask
;

: setPinFunc	\ func# iobit# --
\ *G Set the alternate function of the I/O pin. Zero corresponds
\ ** to GPIO usage for nearly all pins.
  $0007 swap  changeIOCON
;

\ *P Define the alternate functions for readability.
0 constant AF0
1 constant AF1
2 constant AF2
3 constant AF3
4 constant AF4
5 constant AF5
6 constant AF6
7 constant AF7

: isPinPuPd	\ mode iobit# --
\ *G Sets the pull up/down for a pin.
  swap 3 lshift swap  $0018 swap changeIOCON
;

: NotPulled	\ iobit# --
\ *G No pull up or pull down.
  0 swap isPinPuPd  ;

: PulledUp	\ iobit# --
\ *G Enable the pin pull up resistor.
  2 swap isPinPuPd  ;

: PulledDown	\ iobit# --
\ *G Enable the pin pull down resistor.
  1 swap isPinPuPd  ;

: Repeater	\ iobit# --
\ *G Set the pin into repeater mode.
  3 swap isPinPuPd  ;

: isPinSpeed	\ mode iobit# --
\ *G Sets the output speed for a pin. 0=standard, 1=fast. Fast
\ ** mode is recommended for the EMC, LCD and SPIFI interfaces.
  swap 9 lshift swap  bit9 swap  changeIOCON
;
: standardPin	\ iobit# --
\ *G Set pin speed to standard.
  0 swap isPinSpeed  ;
: fastPin	\ iobit# --
\ *G Set pin speed to fast.
  1 swap isPinSpeed  ;

: isPinOD	\ mode iobit# --
\ *G Sets the open-drain mode for a pin. 0=push-pull, 1=open-drain.
  swap #10 lshift swap  bit10 swap  changeIOCON
;
: ODpin	\ iobit# --
\ *G Set pin to open-drain.
  1 swap isPinOD  ;
: PPpin	\ iobit# --
\ *G Set pin to push-pull.
  0 swap isPinOD  ;


\ ****************
\ *S IO pin access
\ ****************

compiler
: io>port	\ iobit# -- mask offset
  dup $1F and 1 swap lshift		\ -- iobit# mask
  swap 5 rshift #Portshift lshift	\ -- mask offset
;
target

: setPin	\ iobit# --
\ *G Set the pin high.
  io>port FIOSET + PortBase + !  ;

: clrPin	\ iobit# --
\ *G Set the pin low.
  io>port FIOCLR + PortBase + !  ;

: getPin	\ iobit# -- 0/1
\ *G Read the pin state.
  dup $1F and				\ -- iobit# pin#
  swap 5 rshift #Portshift lshift	\ -- pin# offset
  FIOPIN + PortBase + @
  swap rshift 1 and
;

: isInput	\ iobit# --
\ *G Set the pin as an input.
  AF0 over setPinFunc
  io>port FIODIR + Portbase + bic!  ;

: isOutput	\ iobit# --
\ *G Set the pin as an output.
  AF0 over setPinFunc
  io>port FIODIR + Portbase + or!  ;



\ ========================
\ *S Test code for RDB4078
\ ========================

1 [if]

\ *[
decimal

P1 24 pio: uled1	\ all active high
P1 25 pio: uled2
P1 28 pio: uled3
P1 29 pio: uled4

: tled1		\ --
  uled1 isOutput
  begin
    uled1 setPin  500 ms
    uled1 clrPin  500 ms
  key? until
;

: tled2		\ --
  uled2 isOutput
  begin
    uled2 setPin  500 ms
    uled2 clrPin  500 ms
  key? until
;
\ *]

[then]


\ ======
\ *> ###
\ ======

decimal
