\ gpioNRF51xxx.fth - easy access to nRF51 GPIO pins.

((
Copyright (c) 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20161031 SFP001 First release.
))

only forth definitions
decimal

\ ===========
\ *! gpionrf51
\ *T NRF51 GPIO utilities
\ ===========
\ *P The code in *\i{Cortex\Drivers\gpioNRF51xxx.fth} provides
\ ** utility words for accessing GPIO pins on a bit by bit basis
\ ** for the nRF51 CPUs.
\ ** The code is written for ease of use rather than performance.


\ ********************
\ *S Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number, 0..31 which
\ ** select a bit in the single GPIO port.

target-only
: PIO:		\ bit# -- ; -- struct
\ *G Define a port I/O bit by name. For example
\ *C   12 pio: LED1   \ P0.12
\ ** defines an I/O bit on GPIO bit 12 as *\fo{LED1}. At run
\ ** time, the I/O bit structure is returned.
  create
    1 over lshift ,			\ bit mask
    ,					\ I/O bit number
;
host&target
interpreter
: PIO:		\ bit# -- ; -- struct
  create
    1 over lshift ,			\ bit mask
    ,					\ I/O bit number
;
target


\ ******************
\ *S GPIO pin access
\ ******************

: setPin	\ struct --
\ *G Set the pin high.
  @ _GPIO gpio.OUTSET + !  ;

: clrPin	\ struct --
\ *G Set the pin low.
  @ _GPIO gpio.OUTCLR + !  ;

: getPin	\ struct -- 0/1
\ *G Read the pin state.
  _GPIO gpio.IN @			\ read pins
  swap cell + @ rshift  1 and		\ shift right and mask
;


\ ***********************
\ *S I/O pin configuration
\ ***********************
\ *P Pins can be in one of several modes. You can define additional
\ ** constants for yourself. The masks below are for the
\ ** CNF (bits 3:2) and MODE bits (bits 1:0) in GPIOx_CRl/H.

%0000 constant InputMode	\ -- mode
\ *G The pin is a general purpose floating input.
%0001 constant OutputMode	\ -- mode
\ *G The pin is an output.

\ *P The following modifiers affect the pin configuration.
\ ** Standard output drive is 0.5 ma, high output drive is 5 ma.
\ ** An example is:
\ *C   OutputMode +H0D1 LED5 iSConfig
\ *P This configures the I/O pin *\fo{LED5} as an output with
\ ** high current drive for a zero and no drive for a one;
\ ** a simulation of a high-current (5 ma) open-drain driver.

: +PD		( mode -- mode' )  %0100 or  ;
\ *G Selects a pull-down resistor.
: +PD		( mode -- mode' )  %1100 or  ;
\ *G Selects a pull-up resistor.
: +S0S1		( mode -- mode' )  $000 or  ;
\ *G Standard drive for a '0', standard drive for a '1'. Default.
: +H0S1		( mode -- mode' )  $100 or  ;
\ *G High drive for a '0', standard drive for a '1'.
: +S0H1		( mode -- mode' )  $200 or  ;
\ *G Standard drive for a '0', high drive for a '1'.
: +H0H1		( mode -- mode' )  $300 or  ;
\ *G High drive for a '0', high drive for a '1'.
: +D0S1		( mode -- mode' )  $400 or  ;
\ *G Disconnect '0', standard drive for a '1'.
: +D0H1		( mode -- mode' )  $500 or  ;
\ *G Disconnect '0', standard drive for a '1'.
: +S0D1		( mode -- mode' )  $600 or  ;
\ *G Standard for a '0', disconnect for a '1'. Low current open-drain.
: +H0D1		( mode -- mode' )  $700 or  ;
\ *G Standard for a '0', disconnect for a '1'. High current open-drain.

: isConfig	\ mode struct --
\ *G Set the pin's configuration bits.
  _GPIO gpio.PIN_CNF[%s] +  swap cell + @ cells +  !
;

: isInput	\ struct --
\ *G Set the pin to GPIO mode and input.
  InputMode swap isConfig
;

: isOutput	\ struct --
\ *G Set the pin as an output.
  OutputMode swap isConfig
;

: isDriver	\ struct --
\ *G Set the pin as an output with high drive.
  OutputMode +H0H1 swap isConfig
;


\ ===============================
\ *S Test code for nRF51 DK board
\ ===============================
\ *P The user LEDs on the STM32-P107 board are set as follows:
\ *E  STAT1  green   PC6
\ **  STAT2  orange  PC7
\ *P Note that the LEDs are driven active high/on.
\ *[

1 [if]
decimal

21 pio: led1
22 pio: led2
23 pio: led3
24 pio: led4

: ledDelay	\ --
  500 ms
;

: tled1		\ --
  led1 isDriver
  begin
    led1 clrPin  ledDelay
    led1 setPin  ledDelay
  key? until
;

: tleds		\ --
  led1 isDriver  led2 isDriver
  begin
    led1 setPin  ledDelay  led1 clrPin
    led2 setPin  ledDelay  led2 clrPin
  key? until
  led1 setPin  led2 setPin
;
[then]
\ *]


\ ======
\ *> ###
\ ======

decimal
