\ gpioSTM32.fth - easy access to STM32 GPIO pins.

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20120326 SFP004 Converion for generic STM32.
20120210 SFP003 Corrected ISFUNCTION.
20120130 SFP002 Refactored, added pin speed.
20111201 SFP001 Adapted from Kinetis version.
))

only forth definitions
decimal

\ ===========
\ *! gpiostm32
\ *T STM32 GPIO utilities
\ ===========
\ *P The code in *\i{Cortex\Drivers\gpioSTM32.fth} provides
\ ** utility words for accessing GPIO pins on a bit by bit basis
\ ** for the STM32 CPUs.
\ ** The code is written for ease of use rather than performance.


\ ********************
\ *S Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number. Bits 0..31 form
\ ** P0.0 to P0.31, bits 32..63 form P1.0 to P1.31 and so on.
\ ** Although ST name their ports from A, here we number
\ ** them from 0. Note also that ST usually only provide 16
\ ** I/O bits per port.

$0400 equ /Port		\ -- u
\ *G Separation in memory between GPIO port base addresses.
#10 equ PortShift	\ -- u
\ *\fo{/Port} as a shift.

target-only
: PIO:		\ port# bit# -- ; -- struct
\ *G Define a port I/O bit by name. For example
\ *C   3 12 pio: LED1   \ PD.12
\ ** defines an I/O bit on GPIOD, bit 12 as *\fo{LED1}. At run
\ ** time, the I/O bit structure is returned.
  create
    over PortShift lshift _GPIOA + ,	\ pins base
    1 over lshift ,			\ bit mask
    swap 5 lshift or ,			\ I/O bit number
;
host&target
interpreter
: PIO:		\ port# bit# -- ; -- struct
  create
    over PortShift lshift _GPIOA + ,	\ pins base
    1 over lshift ,			\ bit mask
    swap 5 lshift or ,			\ I/O bit number
;
target

\ *P Define the ports as constants. So that we can use names
\ ** rather than numbers, e.g.
\ *C   PD 12 pio: LED1   \ PD.12
\ *[
0 constant PA
1 constant PB
2 constant PC
3 constant PD
4 constant PE
5 constant PF
6 constant PG
7 constant PH
8 constant PI
9 constant PJ
\ *]


\ ******************
\ *S GPIO pin access
\ ******************

: setPin	\ struct --
\ *G Set the pin high.
  2@ gp.BSRR + !  ;

: clrPin	\ struct --
\ *G Set the pin low.
  2@ swap #16 lshift swap gp.BSRR + !  ;

: getPin	\ struct -- 0/1
\ *G Read the pin state.
  2@ gp.IDR + @ and 0<> 1 and
;


\ ***********************
\ *S IO pin configuration
\ ***********************

: af#>mask	\ mode -- afbits
\ Repeats a four bit mask 8 times.
  $0F and
  dup 4 lshift or			\ 8 bits
  dup 8 lshift or			\ 16 bits
  dup #16 lshift or			\ 32 bits
;

: >4bitmask/off	\ bit# -- bitmask offset
\ Converts a bit number 0..15 to a 4 bit mask and an offset (0/4)
\ for a 32 bit word,
  dup 7 and 4* $0F swap lshift		\ -- bit# mask
  swap $08 and 1 rshift			\ -- mask offset
;

: af#>valmask	\ af# bit# -- val mask offset
\ Convert a 4 bit alternate function number and bit number
\ (0..15) into a value and bitmask to use with setMask.
  swap af#>mask 			\ -- bit# afmask
  swap >4bitmask/off 			\ -- afmask bitmask offset
  >r  tuck and swap  r>			\ -- val bitmask offset
;

\ *P Pins can be in one of several modes. You can define additional
\ ** constants for yourself. The four-bit masks below are for the
\ ** CNF (bits 3:2) and MODE bits (bits 1:0) in GPIOx_CRl/H.

%0100 constant InputMode	\ -- mode#
\ *G The pin is a general purpose floating input.
%0010 constant Output2MHz	\ -- mode#
\ *G The pin is a 2MHz output.
%0001 constant Output10MHz	\ -- mode#
\ *G The pin is a 10MHz output.
%0011 constant Output50MHz	\ -- mode#
\ *G The pin is a 50MHz output.
%1011 constant AFOutput50MHz	\ -- mode#
\ *G The pin is an alternate function push-pull 50MHz output.

: isFunction	\ af# struct --
\ *G Set the pin's MODE and CNF bits in GPIOx_CRl/H.
  dup >r
  2 cells + @ $0F and af#>valmask	\ -- val mask offset ; -- R: struct
  r> @ gp.CRL + + setMask		\ set bits in CRL or CRH
;

: isInput	\ struct --
\ *G Set the pin to GPIO mode and input.
  InputMode swap isFunction
;

: isOutput	\ struct --
\ *G Set the pin as an output.
  Output2MHz swap isFunction
;

: isAnalog	\ struct --
\ *G Set the pin as an analog pin.
  %0000 swap isFunction
;

: initGPIO	\ --
\ *G Enable clocks to all GPIO ports and take them out of reset.
\ ** performed in the cold chain.
  $007C _RCC  2dup rccAPB2en + or!  rccAPB2rst + bic!
;
' initGPIO atCold


\ ========================================
\ *S Test code for Olimex STM32-P107 board
\ ========================================
\ *P The user LEDs on the STM32-P107 board are set as follows:
\ *E  STAT1  green   PC6
\ **  STAT2  orange  PC7
\ *P Note that the LEDs are driven active high/on.
\ *[

0 [if]
decimal

2 6 pio: led1		\ PC.6
2 7 pio: led2		\ PC.7

: ledDelay	\ --
  200 ms
;

: tled1		\ --
  led1 isOutput
  begin
    led1 clrPin  ledDelay
    led1 setPin  ledDelay
  key? until
;

: tleds		\ --
  led1 isOutput  led2 isOutput
  begin
    led1 setPin  ledDelay  led1 clrPin
    led2 setPin  ledDelay  led2 clrPin
  key? until
;
[then]
\ *]


\ ========================================
\ *S Test code for Blue Pill STM32103 board
\ ========================================
\ *P The user LEDs on the Blue Pill board are set as follows:
\ *E  LED2  blue   PC6
\ *P Note that the LEDs are driven active high/on.
\ *[

1 [if]
decimal

PC 13 pio: uLED		\ PC.6

: ledDelay	\ --
  200 ms
;

: tuled		\ --
  uled isOutput
  begin
    uled clrPin  ledDelay
    uled setPin  ledDelay
  key? until
;
[then]
\ *]


\ ======
\ *> ###
\ ======

decimal
