\ gpioSTM32F30x.fth - easy access to STM32F30x GPIO pins.

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)

Adapted to STM32F30x by
Synics AG

To do
=====

Change history
==============
20140620 K_H004 Added a link list of PIOs usefull for experimental project with not centralised PIO settings
20120210 SFP003 Corrected ISFUNCTION.
20120130 SFP002 Refactored, added pin speed.
20111201 SFP001 Adapted from Kinetis version.
))

only forth definitions
decimal

\ ===========
\ *! gpiostm32f30x
\ *T STM32F30x GPIO utilities
\ ===========
\ *P The code in *\i{Cortex\Drivers\gpioSTM32F30x.fth} provides
\ ** utility words for accessing GPIO pins on a bit by bit basis
\ ** for the STM32F3 CPUs.
\ ** The code is written for ease of use rather than performance.


\ ********************
\ *S Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number. Bits 0..31 form
\ ** P0.0 to P0.31, bits 32..63 form P1.0 to P1.31 and so on.
\ ** Although ST name their ports from A, here we number
\ ** them from 0. Note also that ST usually only provide 16
\ ** I/O bits per port.

$0400 equ /Port		\ -- u
\ *G Separation in memory between GPIO port base addresses.
#10 equ PortShift	\ -- u
\ *\fo{/Port} as a shift.

target-only
: PIO:		\ port# bit# -- ; -- struct
\ *G Define a port I/O bit by name. For example
\ *C   3 12 pio: LED1   \ PD.12
\ ** defines an I/O bit on GPIOD, bit 12 as *\fo{LED1}. At run
\ ** time, the I/O bit structure is returned.
  create
    over PortShift lshift _GPIOA + ,	\ pins base
    1 over lshift ,			\ bit mask
    swap 5 lshift or ,			\ I/O bit number
;
host&target

align here 0 , constant piolink  \ link list pointer for timers, first val is 0

interpreter
: PIO:		\ port# bit# -- ; -- struct
  create
    here >r
    over PortShift lshift _GPIOA + ,	\ pins base
    1 over lshift ,			\ bit mask
    swap 5 lshift or ,			\ I/O bit number
    r> piolink dup @ ,                  \ back link is set
    !                                   \ store link value
;
target

\ ----- PIO debugging -----

3 equ PIOSTRUCTSIZE                     \ magic number of cells used by PIO definition

: .pios \ --;
\ *G Print the list the declared PIOs in interactive mode
  piolink @
  100 0 do                              \ fictive limit of number of PIOs
     cr dup @ _GPIOA - PortShift rshift
     ." P" [char] A + emit
     dup 2 cells + @
     1 5 lshift 1 - and
     dup . base @ < if                  \ only 1 or chars output -> simplfied justifying
       space
     then
     dup 2 cells -                      \ move from XT to CFA, a kind of reverse >BODY (see offset in the def. of >BODY)
     >name space .name
     PIOSTRUCTSIZE cells + @
     dup 0= if leave then
  loop
  drop cr
;

\ ******************
\ *S GPIO pin access
\ ******************

: setPin	\ struct --
\ *G Set the pin high.
  2@ gpio.BSRRL + w!  ;

: clrPin	\ struct --
\ *G Set the pin low.
  2@ gpio.BSRRH + w!  ;

: getPin	\ struct -- 0/1
\ *G Read the pin state.
  2@ gpio.IDR + @ and 0<> 1 and
;


\ ***********************
\ *S IO pin configuration
\ ***********************

: mode>mask	\ mode -- modemask
\ Repeats the two bit mask 16 times.
  3 and
  dup 2 lshift or			\ 4 bits
  dup 4 lshift or			\ 8 bits
  dup 8 lshift or			\ 16 bits
  dup #16 lshift or			\ 32 bits
;

: >2bitmask	\ bit# -- bitmask
\ Converts a bit number 0..15 to a 2 bit mask for a 32 bit word,
  $0F and 2* $03 swap lshift  ;

: mode>valmask	\ mode bit# -- val bitmask
\ Convert a two bit mode and bit number (0..15) into a value and
\ bitmask to use with setMask.
  swap mode>mask 			\ -- bit# modemask
  swap >2bitmask			\ -- modemask bitmask
  tuck and swap 			\ -- val bitmask
;

: ms>vba	\ mode struct -- val bitmask *gpio
  tuck  2 cells + @ $0F and		\ -- struct mode bit#
  mode>valmask				\ -- struct val bitmask
  rot @					\ -- val bitmask *gpio
;

: isPinMode	\ mode struct --
\ *G Sets the GPIO mode for a pin. Use with the constants below.
  ms>vba gpio.MODER + setMask		\ set pin's bits in MODER
;

: af#>mask	\ mode -- afbits
\ Repeats a four bit mask 8 times.
  $0F and
  dup 4 lshift or			\ 8 bits
  dup 8 lshift or			\ 16 bits
  dup #16 lshift or			\ 32 bits
;

: >4bitmask/off	\ bit# -- bitmask offset
\ Converts a bit number 0..15 to a 4 bit mask and offset for a
\ 32 bit word,
  dup 7 and 4* $0F swap lshift		\ -- bit# mask
  swap $08 and 1 rshift			\ -- mask offset
;

: af#>valmask	\ af# bit# -- val mask offset
\ Convert a 4 bit alternate function number and bit number
\ (0..15) into a value and bitmask to use with setMask.
  swap af#>mask 			\ -- bit# afmask
  swap >4bitmask/off 			\ -- afmask bitmask offset
  >r  tuck and swap  r>			\ -- val bitmask offset
;

\ *P Pins can be in one of four modes, 0..3.

0 constant InputMode	\ -- mode#
\ *G The pin is an input.
1 constant OutputMode	\ -- mode#
\ *G The pin is an output.
2 constant AFmode	\ -- mode#
\ *G The pin is for one of the alternate function modes.
3 constant AnalogMode	\ -- mode#
\ *G The pin is an analogue pin.

: isInput	\ struct --
\ *G Set the pin to GPIO mode and input.
  InputMode swap isPinMode
;

: isOutput	\ struct --
\ *G Set the pin as an output.
  OutputMode swap isPinMode
;

: isFunction	\ af# struct --
\ * Set the pin to be used with one of the alternate functions.
  >r
  AFmode r@ isPinMode			\ uses an alternate function.
  r@ 2 cells + @ $0F and af#>valmask	\ -- val mask offset ; -- R: struct
  r> @ gpio.AFRL + + setMask		\ set bits in AFRL or AFRH
;

: isAnalog	\ struct --
\ *G Set the pin as an analog pin.
  AnalogMode swap isPinMode
;

: isPinOD	\ 0/1 struct --
\ *G Enable (nz) or disable (0) the open drain driver.
  swap if
    2@ gpio.OTYPER + or!
  else
    2@ gpio.OTYPER + bic!
  endif
;

: isPinPuPd	\ mode struct --
\ *G Sets the pull up/down for a pin.
  ms>vba gpio.PUPDR + setMask		\ set pin's bits in PUPDR
;

: NotPulled	\ struct --
\ *G No pull up or pull down.
  0 swap isPinPuPd  ;

: PulledUp	\ struct --
\ *G Enable the pin pull up resistor.
  1 swap isPinPuPd  ;

: PulledDown	\ struct --
\ *G Enable the pin pull down resistor.
  2 swap isPinPuPd  ;

: isPinSpeed	\ mode struct --
\ *G Sets the output speed for a pin according to the constants
\ ** below.
  ms>vba gpio.OSPEEDR + setMask		\ set pin's bits in PUPDR
;

0 constant LowSpeed	\ -- mask
\ *G The default is 2MHz, low speed.
1 constant MediumSpeed	\ -- mask
\ *G 10 MHz, medium speed.
3 constant HighSpeed	\ -- mask
\ *G 50 MHz, fast speed.

: initGPIO	\ --
\ *G Enable clocks to all GPIO ports and take them out of reset.
\ ** performed in the cold chain.
  $00FF:0000 _RCC 2dup rcc.AHBENR + or!  rcc.AHBRSTR + bic!
;
' initGPIO atCold


\ ========================================
\ *S Test code for STM32F3 Nucleo 32 board
\ ========================================

1 [if]
decimal

0 8 pio: PA8	\ -- struct

: initMCO	\ --
\ *G Display SYCLK/128 on PA8 as MCO pin
  PA8					\ configure PA8 for MCO
  0 over isFunction  0 over isPinOD
  dup notPulled  HighSpeed swap isPinSpeed
  $7400:0000 $7700:0000 _RCC rcc.CFGR + setMask \ divide by 128 from SYSCLK
;
\ ' initMCO AtCold
[then]


\ ======
\ *> ###
\ ======

decimal
