\ gpioTM4C12xx.fth - easy access to TI TM4C12xx GPIO pins.

((
Copyright (c) 2011, 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20190223 SFP002 Adapted from Luminary LM3S9B9x version.
20111011 SFP001 Adapted from Nuvoton version.
))

only forth definitions
decimal

\ ===========
\ *! gpiotm4c12xx
\ *T TI TM4C12cc GPIO utilities
\ ===========
\ *P The code in *\i{Cortex\Drivers\gpioTM4C12xx.fth} provides
\ ** utility words for accessing GPIO pins on a bit by bit basis.
\ ** The code is written for ease of use rather than performance,
\ ** and assumes that the system start up code has enabled and
\ ** clocked the port.

\ *P The directory *\i{Cortex\Drivers} contains a range of files
\ ** caled *\i{gpio*.fth} which provide similar facilities for
\ ** different CPUs.


\ ********************
\ *S Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number. Bits 0..31 form
\ ** P0.0 to P0.31, bits 32..63 form P1.0 to P1.31 and so on.
\ ** Although TI/Luminary name their ports from A, here we number
\ ** them from 0 and provide constants PA .. PQ

gpio.DATA 0 <> [if]
  cr ." This code relies on GPIODATA = $0000." abort
[then]

create GPIOtab	\ -- addr
  _GPIOA ,  _GPIOB ,  _GPIOC ,  _GPIOD ,
  _GPIOE ,  _GPIOF ,  _GPIOG ,  _GPIOH ,
  _GPIOJ ,  _GPIOK ,  _GPIOL ,  _GPIOM ,
  _GPION ,  _GPIOP ,  _GPIOQ ,

\ *P Ports by name.
\ *[
0 constant PA
1 constant PB
2 constant PC
3 constant PD
4 constant PE
5 constant PF
6 constant PG
7 constant PH
8 constant PJ
9 constant PK
#10 constant PL
#11 constant PM
#12 constant PN
#13 constant PP
#14 constant PQ
\ *]

target-only
: PIO:		\ port# bit# -- ; -- struct
\ *G Define a port I/O bit by name. For example
\ *C   PC 6 PIO: PC.6
\ ** defines an I/O bit on port 2 (GPIOC), bit 6. At run time, the
\ ** address of a structure is returned. The first cell contains
\ ** the I/O address used to set/clear the bit. The second cell
\ ** contains the bit number (0..7).
  create
    tuck swap 2 lshift GPIOtab + @	\ port base address
    1 rot 2 + lshift or ,		\ form address bits 9:2
    ,
  does>
;
host&target
interpreter
: PIO:		\ port# bit# -- ; -- struct
  create
    tuck swap 2 lshift GPIOtab + @	\ port base address
    1 rot 2 + lshift or ,		\ form address bits 9:2
    ,
  does>
;
target

struct /pin	\ -- len
\ *G Structure created  by PIO:
\ *[
  int pin.ioa		\ io address, bit mask in bits 9:2
  int pin.bit		\ bit number
end-struct
\ *]


\ ****************
\ *S IO pin access
\ ****************

: setPin	\ struct --
\ *G Set the pin high.
  pin.ioa @  $FF swap c!  ;

: clrPin	\ struct --
\ *G Set the pin low.
  pin.ioa @  $00 swap c!  ;

: getPin	\ struct -- 0/1
\ *G Read the pin state.
  pin.ioa @ @ 0<> 1 and  ;


\ ***********************
\ *S IO pin configuration
\ ***********************

: pin>mask/base	\ struct -- mask base
\ *G From the structure address generate the GPIO base address
\ ** and the bit mask for the pin in bits 7:0.
  dup pin.bit @ 1 swap lshift		\ -- struct mask
  swap pin.ioa @ $FFFF:FC00 and		\ mask base
;

: initPin	\ struct --
\ *G Perform default initialisation of a pin.
  pin>mask/base
  2dup gpio.DR2R + or!		\ pin has 2mA drive
  2dup gpio.SLR + bic!		\ pin has no slew control
  2dup gpio.DEN + or!		\ enable pin
  2dup gpio.ODR + bic!		\ no open drain
  2dup gpio.PUR + bic!		\ no pull ups
  2dup gpio.PDR + bic!		\ no pull downs
  2drop
;

: isInput	\ struct --
\ *G Set the pin as an input.
  dup  pin>mask/base gpio.DIR + bic!  initPin
;

: isOutput	\ struct --
\ *G Set the pin as an input.
  dup  pin>mask/base gpio.DIR + or!  initPin
;

: setAF		\ func struct --
\ *G Set the pin configuration. Function=0 is for GPIO usage,
\ ** otherwise the alternate function. By default, nearly all pins
\ ** are in I/O pin mode after reset.
  over 0= if
    dup pin>mask/base gpio.AFSEL + bic!	\ AF no
    2drop exit
  then
  dup pin>mask/base gpio.AFSEL + or!	\ AF yes
  swap  over pin.bit @ 4 * lshift  swap	\ -- 000f0000 struct
  $0F over pin.bit @ 4 * lshift	 swap	\ -- 000f0000 000F0000 struct
  gpio.PCTL + setmask
;


: isPinFunc	\ func struct --
\ *G Set the pin configuration. Function=0 is for GPIO usage,
\ ** otherwise the alternate function. By default, nearly all pins
\ ** are in I/O pin mode after reset. The pin is initialised.
  tuck setAF  initpin
;


\ ============
\ *S Test code
\ ============

\ *[
1 [if]
decimal

PN 1 pio: led1		\ GPB.2
PN 0 pio: led2		\ GPB.3

: tled1		\ --
  led1 isOutput
  begin
    led1 setPin  200 ms
    led1 clrPin  200 ms
  key? until
;

: tled2		\ --
  led2 isOutput
  begin
    led2 setPin  200 ms
    led2 clrPin  200 ms
  key? until
;
[then]
\ *]


\ ======
\ *> ###
\ ======

decimal
