\ i2cLPC40xx.fth - I2C hardware driver for LPC40xx/178x

((
Copyright (c) 2008-2014, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20150716 SFP002 Changed user channel selection to be more like the UART
		device selection.
20140402 BJC001	Modified for LPC1788 family devices.  These use a rather
		different pin assignment scheme from older LPC1xxx and
		LPC2xxx devices. The interface follows MPE standard.
))

only forth definitions decimal


\ ============
\ *! i2clpc40xx
\ *T I2C primitives for LPC40xx, using hardware I2C device
\ ============
\ *P NOTE: this code defaults to using I2C0, unless equate I2C_DEV is defined
\ ** previously to select another port.
\ *P The code in *\i{Cortex\Drivers\i2cLPC40xx.fth} provides the following
\ ** words for constructing drivers for I2C devices:
\ *D START_I2C Generate a START condition
\ *D STOP_I2C  Generate a STOP condition
\ *D READ_BYTE_I2C Read a byte from the I2C lines
\ *D WRITE_BYTE_I2C Write a byte to the I2C lines
\ *D +ACK_I2C Generate and test the ACK bit
\ *D -ACK_I2C Do not generate and test the ACK bit.

\ *P The code provides the same interface as that provided by
\ ** *\i{Examples\i2c\i2cBase.fth}.


\ *****************
\ *S I2C primitives
\ *****************
\ *P Each task that controls I2C uses two *\fo{USER} variables
\ ** to hold the channel (base address) and I2C ACK conditions.

cell +user <i2c_dev>	\ -- addr
\ *G Holds the base address of the I2C peripheral.
cell +user <ack_i2c?>	\ -- addr
\ *G Holds non-zero to acknowledge ACK bits.


: (i2cwait)	\ base --
\ *G Wait for an I2C operation to complete.
  begin
    [ tasking? ] [if] pause [then]
    dup I2CONSET + @ ~SI and
  until
  drop
;

: (start_i2c)	\ base --
\ *G Issue a START on the device.
  ~SI over I2CONCLR + !			\ clear pending interrupt
  ~STA over I2CONSET + !
  dup (i2cWait)
  ~STA swap I2CONCLR + !
;

: (stop_i2c)	\ base --
\ *G Issue a STOP on the device.
  ~SI over I2CONCLR + !			\ clear pending interrupt
  ~STO over I2CONSET + !
  begin
    [ tasking? ] [if] pause [then]
    dup I2CONSET + @ ~STO and 0=
  until
  drop
;

: (i2cAA)	\ base --
  <ack_i2c?> @
  if  ~AA swap I2CONSET + !  else  ~AA swap I2CONCLR + !  then
;

: (i2c@)	\ base -- 8b
\ *G Receive a byte from the given channel, checking the status
\ ** if required.
  dup (i2cAA)				\ set/clear AA bit
  ~SI over I2CONCLR + !			\ clear pending interrupt
  dup (i2cWait)				\ wait until read
  <ack_i2c?> @ if
    dup I2STAT + @ $F8 and		\ status value
    dup $50 <>				\ good data byte
    over $018 <> and			\ good SLA+W
    swap $040 <> and			\ good SLA+R
    err_I2C_ACK ?throw
  endif
  I2DAT + @ $FF and			\ return data
;

: (i2c!)	\ 8b base --
\ *G Transmit a byte to the given channel, checking the status
\ ** if required.
  dup (i2cAA)				\ set/clear AA bit
  tuck I2DAT + !			\ character
  ~SI over I2CONCLR + !			\ clear pending interrupt
  dup (i2cwait)				\ wait for completion
  <ack_i2c?> @ if
    I2STAT + @ $F8 and			\ status value
    dup  $028 <>			\ good data byte
    over $018 <> and			\ good SLA+W
    swap $040 <> and			\ good SLA+R
    err_I2C_ACK ?throw
  else
    drop
  endif
;

: (initI2c)	\ high/low base --
\ *G After power up and pin selection, initialise the I2C channel
\ ** at *\i{base} in master mode with clocking defined by
\ ** *\i{high/low} where the top 16 bits are the high value).
  $6C over I2CONCLR + !			\ reset
  over $FFFF and over I2SCLL + !	\ low clock count
  swap #16 rshift over I2SCLH + !	\ high clock count
  $40 over I2CONSET + !			\ enable
  (stop_I2C)
;


\ ************************
\ *S I2C Channel specifics
\ ************************
\ *G To use an I2C device you must set the equate *\fo{useI2Cx?}
\ ** where x=0, 1 or 2 to non-zero. You must also set the equate
\ ** *\fo{I2Cx-speed} to the bit rate in Hz. You can set these
\ ** before the file *\i{Cortex\Drivers\i2cLPC40xx.fth} is
\ ** compiled. There are no defaults.

\ *P You must also provide a pin selection and set up routine
\ ** *\fo{selio-I2Cx}. For these, defaults are provided.

useI2C0? [if]
periph-speed I2C0-speed / 1+ 2/		\ total period/count; clock high/low separate
  equ I2CSCL0				\ assume 50% duty cycle
\ *G Initial value of the I2C clock registers.

[undefined] selio-I2C0 [if]
: selio-I2C0	\ --
\ *G Turn on device, connect the IO pins
  ~pcI2C0 _PCONP or!			\ enable power
  _IOCON
  AF5 over IOCON_P5_02 + !		\ SDA on P5[2]
  AF5 swap IOCON_P5_03 + !		\ SCL on P5[3]
;
[then]
[then]

useI2C1? [if]
periph-speed I2C1-speed / 1+ 2/		\ total period/count; clock high/low separate
  equ I2CSCL1				\ assume 50% duty cycle
\ *G Initial value of the I2C clock registers.

[undefined] selio-I2C1 [if]
: selio-I2C1	\ --
\ *G Turn on device, connect the IO pins
  ~pcI2C1 _PCONP or!			\ enable power
  _IOCON
  AF3 bit10 or over IOCON_P0_00 + !	\ SDA on P0[0], Open Drain
  AF3 bit10 or swap IOCON_P0_01 + !	\ SCL on P0[1], Open Drain
;
[then]
[then]

useI2C2? [if]
periph-speed I2C2-speed / 1+ 2/		\ total period/count; clock high/low separate
  equ I2CSCL2				\ assume 50% duty cycle
\ *G Initial value of the I2C clock registers.

[undefined] selio-I2C2 [if]
: selio-I2C2	\ -- ; turn on device, connect the IO pins
\ *G Turn on device, connect the IO pins
  ~pcI2C2 _PCONP or!			\ enable power
  _IOCON
  AF2 bit10 or over IOCON_P0_10 + !	\ SDA on P0[10], Open Drain
  AF2 bit10 or swap IOCON_P0_11 + !	\ SCL on P0[11], Open Drain
;
[then]
[then]

: init-I2C		\ --
[ useI2C0? ] [if]
  selio-I2C0  I2CSCL0 _I2C0 (initi2c)
[then]
[ useI2C1? ] [if]
  selio-I2C1  I2CSCL1 _I2C1 (initi2c)
[then]
[ useI2C2? ] [if]
  selio-I2C2  I2CSCL2 _I2C2 (initi2c)
[then]
  +ack_i2c  #10 ms
;
' init-I2C AtCold


\ ************************
\ *S Standard MPE word set
\ ************************
\ *P The user variable *\fo{<i2c_dev>} holds the base address of
\ ** the I2C peripheral.

: start_i2c   	\ --
\ *G Define a start condition, bring SDA high-low while SCL is high
  <i2c_dev> @ (start_i2c)
;

: stop_i2c  	\ --
\ *G Define a stop condition, bring SDA low-high while SCL is high
  <i2c_dev> @ (stop_i2c)
;

: ack_i2c?	( -- n )  <ack_i2c?> @  ;
\ *G If a non-zero value is used, the system will
\ ** generate the ACK bit on SDA for reads, otherwise SDA
\ ** is left high. For a non-zero value on writes, the
\ ** system will check the ACK bit and *\fo{THROW} if it is non-zero.
compiler
: ack_i2c?	( -- n )  <ack_i2c?> @  ;
target

: +ack_i2c	\ --
\ *G check ACK cycles.
  <ack_i2c?> on
;

: -ack_i2c	\ --
\ *G Do not check ACK cycles.
  <ack_i2c?> off
;

: read_byte_i2c		\ -- byte
\ *G Read a byte, MSB first. If *\fo{ack_i2c?} is
\ ** non-zero, pull SDA low during the ACK bit, otherwise
\ ** leave it high
  <i2c_dev> @ (i2c@)
;

: write_byte_i2c	\ byte --
\ *G Write a byte by shifting, MSB first. If *\fo{ack_i2c?} is
\ ** set, read ack bit on SDA, and *\fo{THROW} if the ACK bit is
\ ** high.
  <i2c_dev> @ (i2c!)
;


\ ************
\ *S Test Code
\ ************

0 [if]
\ *G Use *\fo{0 [if]} not to compile the test code.

: .i2cstat	\ --
\ *G Display I2C status byte.
  <i2c_dev> @ I2STAT + @ .byte
;

: >disp		\ nn..n1 n reg --
\ *G Send *\i{n} bytes starting at register *\i{reg}.
  +ack_i2c				\ all bytes acknowledged
  start_i2c  $70 write_byte_i2c		\ SAA1064 address +w
  $07 and write_byte_i2c		\ first reg
  0 ?do  write_byte_i2c  loop		\ n bytes
  stop_i2c
;

: .Boot		\ --
\ *G Put a "boot" message on the display.
  $78 $5c $5c $7c $37 5 0 >Disp
;

: DispStat	\ -- b
\ *G Read the display status, $80 if enabled.
  +ack_i2c				\ all bytes acknowledged
  start_i2c  $71 write_byte_i2c		\ SAA1064 address +r
  read_byte_i2c  stop_i2c
;

: AutoDisp	\ --
\ *G Initialise and display a boot message at start up.
  init_i2cx .Boot
;
' AutoDisp AtCold

[then]


\ ======
\ *> ###
\ ======

