\ I2C bit-bang driver for LPC2148

((
Copyright (c) 1999, 2004, 2006
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
fax: +44 (0)2380 339691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com
))

\ *! i2clpcfiobb
\ *T LPC Fast I/O software I2C driver
\ ===============
\ *S Introduction
\ ===============
\ *P The I2C interface is a software bit-banging system using
\ ** GPIO bits 2, 3 and 4 in Fast I/O mode. The initialisation
\ ** code sets up these bits as open drain outputs.

\ *P The driver code in *\i{i2cLpcFIObb.fth} must be compiled
\ ** before the generic drive code in I2CBASE.FTH and the
\ ** device specific files (e.g. AT24C512.FTH) are
\ ** compiled.

1 equ TestI2C?	\ -- n
\ *G Set this equate non-zero to include the test code
\ ** at the end of the file. This equate also controls
\ ** the compilation of I2C test code in other files.

bit2 equ SCLmask	\ -- mask ; for SCL bit
2 equ #SCLshift		\ -- n; right shifts for SCL to bit0
bit3 equ SDAmask	\ -- mask ; for SDA bit
3 equ #SDAshift		\ -- n; right shifts for SDA to bit0
SDAmask SCLmask or equ I2Cmask	\ -- mask ; for SDA and SCL bits
$00F0 equ i2cPS0mask	\ -- mask ; SCL and SDA as GPIO bits

l: ^FIO0	\ -- addr
  _FIO0 ,

: initI2C	\ --
\ *G Initialise the I2C hardware. Performed after reset.
  bit0 _SYSCON SCS + or!		\ enable fast I/O for GPIO0
  i2cPS0mask _PINSEL0 bic!		\ use bits for GPIO
  ^FIO0 @
  I2Cmask over FIODIR + bic!		\ set I/O direction, 1=o/p
  I2Cmask swap FIOCLR + !		\ SDA and SCL potentially low
\  start-timers				\ already done?
; ' initI2C atCold


\ =========
\ *S Timing
\ =========

decimal

#10 equ I2CPeriod	\ -- us
\ *G Define the I2C bit time in microseconds.

I2Cperiod system-speed 1000000 */ equ I2Cclocks	\ -- n
\ *G The number of CPU clocks in a bit time.

1 equ clocks/ins	\ -- n
\ *G The number of CPU clocks per instruction. Tune this to
\ ** your hardware set up (memory width and speed). On most ARMs,
\ ** a store takes 2 cycles, a load takes 3, and branch takes 3 and
\ ** all others take 1. Look at the specific core documentation on
\ ** the ARM technical reference CD for clock/instruction details.
\ ** Look at your board schematics and setup code for details of the
\ ** memory bandwidth. For example a 16 bit bus with two wait states
\ ** and no cache may lead to six or more clocks per instruction.

#20 equ clocks/DO
\ *G the number of clocks needed to execute DO at the
\ ** start of DO..LOOP including parameter passing and the call
\ ** to (DO).
\ ** This number may be sensitive to the compiler version and
\ ** switch settings.

9 equ clocks/LOOP
\ *G the number of clocks needed to execute LOOP at the
\ ** end of DO..LOOP.
\ ** This number may be sensitive to the compiler version and
\ ** switch settings.

I2Cclocks 4 / clocks/DO - clocks/LOOP /
  equ /I2Cqbit	\ -- n
\ *G The number n for "n 0 DO LOOP" to generate an I2C
\ ** quarter bit time.

/I2Cqbit 1 < [if]
  cr ." *** I2C requested rate too fast for CPU ***" abort
[then]

: I2Cdelay	\ --
\ *G Wait one quarter of an I2C bit. Tune the equates
\ ** above for a one quarter bit time.
  /I2Cqbit 0 do loop
;


\ ====================
\ *S I2C bit functions
\ ====================

: read_scl	\ -- bit
\ *G Read SCL bit and return in l.s. bit. No delays.
   ^FIO0 @ FIOPIN + @ SCLmask and #SCLshift rshift
;

IsLeaf
: write_scl	\ b -- ; output clock bit
\ *G Write to SCL with no delays.
  ^FIO0 @ swap if			\ high
    SCLmask swap FIODIR + bic!		\ i/p, so R pulls high
  else					\ low
    SCLmask swap FIODIR + or!		\ o/p, preset low
  endif
;

: SCL_low	\ --
\ *G Set SCL low. A quarter cycle delay is performed
\ ** before and after the transition.
  I2Cdelay
  SCLmask ^FIO0 @ FIODIR + or!		\ o/p, preset low
  I2Cdelay
;

: SCL_high	\ --
\ *G Set SCL high. A quarter cycle delay is performed
\ ** before and after the transition.
  I2Cdelay
  SCLmask ^FIO0 @ FIODIR + bic!		\ i/p, R pulls high
  I2Cdelay
;

: read_sda	\ -- bit
\ *G Read SCL bit and return in l.s. bit.
  ^FIO0 @ FIOPIN + @ SDAmask and #SDAshift rshift
;

IsLeaf
: write_sda	\ b -- ; output to SDA
\ *G Write the SDA bit.
  ^FIO0 @ swap if			\ high
    SDAmask swap FIODIR + bic!		\ i/p, so R pulls high
  else					\ low
    SDAmask swap FIODIR + or!		\ o/p, preset low
  endif
;

: sda_low	\ --
\ *G Set SDA low.
  SDAmask ^FIO0 @ FIODIR + or!		\ o/p, preset low
;

: sda_high	\ --
\ *G Set SDA high.
  SDAmask ^FIO0 @ FIODIR + bic!		\ i/p, R pulls high
;


\ *********
\ test code
\ *********

TestI2C? [if]
: tcl		\ -- ; rattle the SCL line
  begin  scl_low scl_high  key? until
;

: tda		\ -- ; rattle the SCL line
  begin  sda_low sda_high  key? until
;
[then]


\ ======
\ *> ###
\ ======

