\ i2cNUC1xx.fth - I2C driver 

((
Copyright (c) 2008, 2011, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

>From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20120119 MPE001 Documentation changes
20111213 WIC001 NUC1xx version
))

only forth definitions decimal


\ ============
\ *! i2cnuc1xx
\ *T I2C primitives for NUC1xx
\ ============
\ *P The code in *\i{Cortex\Drivers\i2cNUC1xx} provides the following
\ ** words for constructing drivers for I2C devices:
\ *D START_I2C Generate a START condition
\ *D STOP_I2C  Generate a STOP condition
\ *D READ_BYTE_I2C Read a byte from the I2C lines
\ *D WRITE_BYTE_I2C Write a byte to the I2C lines
\ *D +ACK_I2C Generate and test the ACK bit
\ *D -ACK_I2C Do not generate and test the ACK bit.

\ *P The code provides the same interface as that provided by
\ ** *\i{Examples\i2c\i2cBase.fth}.

\ Bits in the I2CCON register
bit7 equ I2CIEN		\ interrupt enable
bit6 equ I2EN		\ I2C enable
bit5 equ STA		\ start control
bit4 equ STO		\ stop control
bit3 equ SI		\ interrupt flag
bit2 equ AA~		\ assert acknowledge

\ bits peripheral reset control reg 2
bit8 equ I2C0_RST

0 8 pio: i2csda  \ pa.8
\ *G SDA pin.
0 9 pio: i2cscl  \ pa.9
\ *G SCL pin

#100 equ err_I2C_ACK	\ -- n
\ *G Error code returned on ACK failure.

: (initI2c)	\ clk base --
\ *G After power up and pin selection, initialise the I2C channel
\ ** at *\i{base} in master mode with clocking defined by
\ ** *\i{clk}.
  $08  over i2ccon + !			\ disable, reset int
  dup -rot I2CCLK + !			\ clock count
  I2EN  over i2ccon + or!		\ enable
  drop
;

: (i2cwait)	\ base --
\ *G Wait for an I2C operation to complete.
  begin
    [ tasking? ] [if] pause [then]
    dup I2CCON + @ SI and
  until
  drop
;

: (start_i2c)	\ base --
\ *G Issue a START on the device.
  SI over I2CCON + or!			\ clear pending interrupt
  STA over I2CCON + or!
  dup (i2cWait)
  STA swap I2CCON + bic!
;

: (stop_i2c)	\ base --
\ *G Issue a STOP on the device.
  SI over I2CCON + or!			\ clear pending interrupt
  STO over I2CCON + or!
  STO over I2CCON + bic!  \ seems stupid but works ??
  begin
   [ tasking? ] [if] pause [then]
    dup I2CCON + @ STO and  0=
  until
  drop
;

variable <ack_i2c?>	\ changed from user var
\ *G Holds -1 to acknowledge ACK bits.

: (i2cAA)	\ base --
\ *G Set AA bit in I2CCON as required.
  <ack_i2c?> @
  if  AA~ swap I2CCON + or!  else  AA~ swap I2CCON + bic!  then
;

: (i2c@)	\ base -- 8b
\ *G Receive a byte from the given channel, checking the status
\ ** if required.
  dup (i2cAA)				\ set/clear AA~ bit
  SI over I2CCON + or!			\ clear pending interrupt
  dup (i2cWait)				\ wait until read
  <ack_i2c?> @ if
    dup I2CSTATUS + @ $F8 and		\ status value
    dup $50 <>				\ good data byte
    over $018 <> and			\ good SLA+W
    swap $040 <> and			\ good SLA+R
    err_I2C_ACK ?throw
  endif
  I2CDATA + @ $FF and			\ return data
;

: (i2c!)	\ 8b base --
\ *G Transmit a byte to the given channel, checking the status
\ ** if required.
  dup (i2cAA)				\ set/clear AA~ bit
  tuck I2CDATA + !			\ character
  SI over I2CCON + or!			\ clear pending interrupt
  dup (i2cwait)				\ wait for completion
  <ack_i2c?> @ if
    I2CSTATUS + @ $F8 and		\ status value
    dup  $028 <>			\ good data byte
    over $018 <> and			\ good SLA+W
    swap $040 <> and			\ good SLA+R
    err_I2C_ACK ?throw
  else
    drop
  endif
;


\ ************************
\ *S Standard MPE word set
\ ************************

_I2C0 equ _I2Cx	\ -- base
\ *G Defines the channel to which the standard word set applies.

: start_i2c   	\ --
\ *G Define a start condition, bring SDA high-low while SCL is high
  _I2Cx (start_i2c)
;

: stop_i2c  	\ --
\ *G Define a stop condition, bring SDA low-high while SCL is high
  _I2Cx (stop_i2c)
;

: ack_i2c?	( -- n )  <ack_i2c?> @  ;
\ *G If a non-zero value is used, the system will
\ ** generate the ACK bit on SDA for reads, otherwise SDA
\ ** is left high. For a non-zero value on writes, the
\ ** system will check the ACK bit and *\fo{THROW} if it is non-zero.
compiler
: ack_i2c?	( -- n )  <ack_i2c?> @  ;
target

: +ack_i2c	\ --
\ *G check ACK cycles.
  <ack_i2c?> on
;

: -ack_i2c	\ --
\ *G Do not check ACK cycles.
  <ack_i2c?> off
;

: read_byte_i2c		\ -- byte
\ *G Read a byte, MSB first. If *\fo{I2C_ACK?} is
\ ** non-zero, pull SDA low during the ACK bit, otherwise
\ ** leave it high
  _I2Cx (i2c@)
;

: write_byte_i2c	\ byte --
\ *G Write a byte by shifting, MSB first. If *\fo{ACK_I2C?} is
\ ** set, read ack bit on SDA, and *\fo{THROW} if the ACK bit is
\ ** high.
  _I2Cx (i2c!)
;

#100000 equ I2C0-speed	\ -- hz
\ *G Required I2C clock speed.
system-speed 4 I2C0-speed * 1+ /  equ I2C0div	\ -- u
\ *G Divider for I2C clk register

: init_I2Cx	\ --
\ *G Initialise default I2C channel (I2C0).
  $100 _clk apbclk +  or!		\ enable i2c0 clk
  1 i2csda isPinFunc			\ pin selection
  1 i2cscl isPinFunc
  I2C0div _I2Cx (initI2C)		\ set clk speed and start
  +ack_i2c  
  stop_i2c  
  #10 ms
;


\ ************
\ *S Test Code
\ ************
\ *G The tests were performed using an SAA1064 display controller
\ ** at I2C address $70.

0 [if]
\ *G Use *\fo{0 [if]} not to compile the test code.

: .i2cstat	\ --
\ *G Display I2C status byte.
  _I2Cx I2CSTATUS + @ .byte
;

: >disp		\ nn..n1 n reg --
\ *G Send *\i{n} bytes starting at register *\i{reg}.
  +ack_i2c				\ all bytes acknowledged
  start_i2c  $70 write_byte_i2c		\ SAA1064 address +w
  $07 and write_byte_i2c		\ first reg
  0 ?do  write_byte_i2c  loop		\ n bytes
  stop_i2c
;

: .Boot		\ --
\ *G Put a "boot" message on the display.
  $78 $5c $5c $7c $37 5 0 >Disp
;

: DispStat	\ -- b
\ *G Read the display status, $80 if enabled.
  +ack_i2c				\ all bytes acknowledged
  start_i2c  $71 write_byte_i2c		\ SAA1064 address +r
  read_byte_i2c  stop_i2c
;

: AutoDisp	\ --
\ *G Initialise and display a boot message at start up.
  init_i2cx .Boot
;
' AutoDisp AtCold

[then]


\ ======
\ *> ###
\ ======


