\ I2C bit-bang driver for STM32

((
Copyright (c) 1999
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
fax: +44 (0)2380 339691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com
))

\ *! i2cstm32bb
\ *T STM32 software I2C driver
\ ===============
\ *S Introduction
\ ===============
\ *P The I2C interface is a software bit-banging system using
\ ** GPIO bits PB6 (SCL) and PB9 (SDA). The initialisation code
\ ** sets up these bits as open drain outputs.

\ *P The driver code in *\i{Cortex/Drivers/i2cSTM32F4xxbb.fth}
\ ** must be compiled before the generic drive code in
\ ** *\i{Examples/i2c/i2cBase.fth} and the device specific files
\ ** (e.g. AT24C64.FTH) are compiled.

[undefined] TestI2C? [if]
0 equ TestI2C?	\ -- n
\ *G Set this equate non-zero to include the test code at the end
\ ** of the file. This equate also controls the compilation of
\ ** I2C test code in other files. The version in the file
\ ** *\i{i2cSTM32F4xxbb.fth} is used only if the equate has not
\ ** already been defined.
[then]


\ ===============================
\ *S I/O Configuration and Timing
\ ===============================

decimal

1 6 pio: SCLio	\ --
\ *G The I/O bit for SCL.
1 9 pio: SDAio	\ --
\ *G The I/O bit for SDA.

#100 kHz equ i2c-speed	\ -- hz
\ *G Define the I2C clock speed in Herz.

system-speed i2c-speed / equ I2Cclocks	\ -- n
\ *G The number of CPU clocks in a bit time.

#10 equ clocksOver
\ *G the number of clocks needed over and above the loop itself.
\ ** This number may be sensitive to the compiler version and
\ ** switch settings.

4 equ clocksPer
\ *G the number of clocks needed for each iteration of the loop
\ ** in *\fo{i2cDelay} below. This number may be sensitive to the
\ ** compiler version and switch settings. On most Cortex CPUs,
\ ** a store takes 2 cycles, a load takes 3, a branch takes 3
\ ** and all others take 1. Look at the core documentation from
\ ** ARM for clock/instruction details.
\ ** Look at your board schematics and setup code for details of the
\ ** memory bandwidth. For example a 16 bit bus with two wait states
\ ** and no cache may lead to six or more clocks per instruction.

I2Cclocks 4 / clocksOver - clocksPer / equ /I2Cqbit	\ -- n
\ *G The number n for "begin 1- dup 0= until" to generate an
\ ** I2C quarter bit time.

/I2Cqbit 1 < [if]
  cr ." *** I2C requested rate too fast for CPU ***" abort
[then]

: I2Cdelay	\ --
\ *G Wait one quarter of an I2C bit. Tune the equates
\ ** above for a one quarter bit time.
  /I2Cqbit begin  1- dup 0=  until  drop
;

: initI2C	\ --
\ *G Initialise the I2C hardware. Performed after reset. This
\ ** code needs *\fo{Cortex/Drivers/gpioSTM32F4xx.fth}.
  SCLio dup isOutput  1 over isPinOD  PulledUp
  SDAio dup isOutput  1 over isPinOD  PulledUp
;  ' initI2C atCold


\ ====================
\ *S I2C bit functions
\ ====================

: read_scl	\ -- bit
\ *G Read SCL bit and return in l.s. bit. No delays.
  SCLio getPin
;

: write_scl	\ b -- ; output clock bit
\ *G Write to SCL with no delays.
  SCLio swap
  if  setPin  else  clrPin  endif
;

: SCL_low	\ --
\ *G Set SCL low. A quarter cycle delay is performed
\ ** before and after the transition.
  I2Cdelay  SCLio clrPin  I2Cdelay
;

: SCL_high	\ --
\ *G Set SCL high. A quarter cycle delay is performed
\ ** before and after the transition.
  I2Cdelay  SCLio setPin  I2Cdelay
;

: read_sda	\ -- bit
\ *G Read SCL bit and return in l.s. bit. No delays.
  SDAio getPin
;

: write_sda	\ b -- ; output to SDA
\ *G Write the SDA bit.
  SDAio swap
  if  setPin  else  clrPin  endif
;

: sda_low	\ --
\ *G Set SDA low.
  SDAio clrPin
;

: sda_high	\ --
\ *G Set SDA high.
  SDAio setPin
;


\ *********
\ test code
\ *********

TestI2C? [if]
: tcl		\ -- ; rattle the SCL line
  begin  scl_low scl_high  key? until
;

: tda		\ -- ; rattle the SCL line
  begin  sda_low sda_high  key? until
;

: tSCL		\ -- ; rattle the SCL line
  begin  scl_low scl_high  key? until
;

: tSDA		\ -- ; rattle the SCL line
  begin  i2cDelay sda_low i2cDelay  i2cDelay sda_high i2cDelay  key? until
;
[then]


\ ======
\ *> ###
\ ======

