\ rtcK60.fth - K60 Real-Time Clock driver

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   htpp://www.mpeforth.com

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20111124 MPE001 First release
))

only forth definitions  decimal

\ ==========
\ *! rtck60
\ *T Kinetis K60 Real Time Clock
\ ==========


\ *************
\ *S RTC access
\ *************

: initRTC	\ --
\ *G Initialise RTC - does not affect current state or contents.
\ ** Uses the 32kHz crystal.
  bit29 _SIM simSCGC6 + or!		\ power it up (may be the first time)
  _RTC
  bit8 over rtcCR + or!			\ OSCE - enable oscillator
  bit4 over rtcSR + or!			\ TCE - enable counter increment
  drop
;
' initRTC AtCold		\ performed at power up.

: xsecs>RTC	\ frac int --
\ *G Write the seconds and fraction of seconds values to the RTC.
  _RTC
  bit4 over rtcSR + bic!		\ disable TCE
  swap over rtcTSR + !			\ seconds
  swap #17 rshift over rtcTPR + !	\ fractional seconds
  bit4 over rtcSR + or!
  drop
;

: RTC>xsecs	\ -- frac int
\ *G Read the seconds and fractional seconds from the RTC.
  _RTC
  bit4 over rtcSR + bic!		\ disable TCE
  dup rtcTPR + @ #17 lshift swap	\ fraction
  dup rtcTSR + @ swap			\ seconds
  bit4 over rtcSR + or!			\ enable TCE
  drop
;

_RTC rtcTSR + constant lseconds	\ -- addr
\ *G The address of the seconds counter.

0 value dow	\ -- u
\ *G Return the day of week number where* 0 is Sunday, 1 is Monday
\ ** and so on. Set as a result of running *\fo{Time&Date} below.


\ ***********************
\ *S LSECONDS conversions
\ ***********************
\ *P When you have LSECONDS, you need conversion between Unix
\ ** seconds and the time/data functions supported by the RTC.
\ ** This code is taken from *\i{UnixTime.fth} and is compiled if
\ ** the equate *\fo{ConvertSecs?} is true.

#24 #60 * #60 * equ secs/day	\ -- 86400
\ *G Seconds per day.
#60 #60 * equ secs/hr		\ -- 3600
\ *G Seconds per hour
#60 equ secs/min		\ -- 60
\ *G Seconds per minute
#60 equ mins/hr			\ -- 60
\ *G Minutes per hour
#24 equ hrs/day			\ -- 24
\ *G Hours per day
#34698 equ uday0		\ -- 34698
\ *G The Julian day number for Unix day 0 at 1 Mar 1995.

compiler
: j>uday	\ jday -- uday
\ *G Convert a Julian day based at 0 for 1 March 1900 to a Unix
\ ** day number derived from *\fo{LSECONDS}.
  uday0 -
;

: u>jday	\ uday -- jday
\ *G Convert a Unix day number as used by *\fo{LSECONDS} to a Julian
\ ** day based at 0 for 1 March 1900.
  uday0 +
;
target

\ works with symmetric division
: >date		\ jday -- day month year
\ *G Convert a Julian day number to day month year format.
  1461 /mod				\ number of leap-year periods
   4 * >r				\ convert to number of years
   365 /mod				\ number of years and days
   dup 4 = if 2drop 365 3 then		\ correct if Feb 29th
   r> + >r				\ -- days | years
   123 + dup 5 153 */			\ compute month
   dup >r 153 5 */ - dup 0=		\ if 31 day month
   if  31 +  -1 else 0 then r> +	\ compute day
   1- dup 12 > if 12 - r> 1+ >r then	\ correct month and year
   r> 1900 +				\ -- day month year
;

: date>         \ day month year -- jday ; see month codes
\ *G Given day month and year generate the day number.
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if 12 + r> 1- >r then    	\ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

[undefined] /umod [if]
: /umod		\ u1 u2 -- urem uquot
\ *G Single precision division of unsigned U1 by U2 returning
\ ** both remainder and quotient.
  0 swap um/mod
;
[then]

: weekday       \ jday -- n ; 0=Sunday
\ *G Given a Julian day return the day of week number where
\ ** 0 is Sunday, 1 is Monday and so on.
  4 + 0 7 um/mod drop
;

: secs>td	\ lsecs -- sec min hr day mon yr
\ *G Convert a Unix seconds value to time and date.
  secs/min /umod
  mins/hr /umod
  hrs/day /umod				\ -- sec min hr day#
  u>jday >date				\ -- sec min hr day mon yr
;

: td>secs	\ sec min hr day mon yr -- secs
\ *G Use the given values for time and date, and return the Unix
\ ** *\fo{LSECONDS} value.
  date> j>uday secs/day *		\ -- sec min hr lsecs
  swap secs/hr * +
  swap secs/min * +
  +
;

: time&date	\ -- sec min hr day mon yr ; 10.6.2.2292
\ *G The ANS Forth word to return time and date. Sets *\fo{DOW}
\ ** as a side effect.
  lseconds @
  secs/min /umod
  mins/hr /umod
  hrs/day /umod				\ -- sec min hr day#
  u>jday dup weekday to dow
  >date					\ -- sec min hr day mon yr
;

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set *\fo{LSECONDS}
\ ** and any underlying RTC hardware.
  td>secs 0 swap xsecs>RTC
;


\ ***********************
\ *S Time and date output
\ ***********************

: .2d		\ u --
\ *G Display u as two digits with leading zero.
  0 <# # # #> type
;

: (.time)	\ secs mins hours --
\ *G Display given time.
  .2d  ." :" .2d  ." :" .2d
;

create months$	\ -- addr
\ *G String containing 3 character text for the months.
  ", JanFebMarAprMayJunJulAugSepOctNovDec"

: (.date)	\ day month year --
\ *G Display given date.
  rot .2d space				\ day
  swap 1- 3 * months$ 1+ + 3 type space	\ month
  4 .r					\ year
;

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow#		\ dow --
\ *G Display day of week by name.
  3 * days$ 1+ + 3 type  ;

: .time&date	\ --
\ *G Display the day of week, date and time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  dow .dow# space  (.date) space  (.time)
  r> base !
;

: .time		\ --
\ *G Read and display time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  drop drop drop  (.time)
  r> base !
;

: .date		\ --
\ *G Read and display date.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  (.date)  drop drop drop
  r> base !
;

: .zone		\ zone --
  if  3 spaces  else  ." GMT"  endif
;

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  dow .dow# ." , "  (.date) space  (.time)
  space .zone space
  r> base !
;


\ ======
\ *> ###
\ ======
