\ rtcSTM32fth - STM32F1 Real-Time Clock driver

((
Copyright (c) 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   htpp://www.mpeforth.com

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20120326 SFP003 STM32F1 version.
))

only forth definitions  decimal

\ ===============
\ *! rtcstm32
\ *T STM32 Real Time Clock
\ ===============

[undefined] lseconds? [if]
1 equ lseconds?	\ -- flag
\ *G Set this equate true to compile the Unix LSECONDS code.
\ ** This version is compiled if *\fo{lseconds?} has not
\ ** been defined yet.
[then]

[undefined] RTCint? [if]
1 equ RTCint?	\ -- flag
\ *G Set this equate true to compile the RTC one second interrupt
\ ** handler. This version is compiled if *\fo{RTCint?} has not
\ ** been defined yet.
[then]


\ ***********************
\ *S LSECONDS conversions
\ ***********************
\ *P If you have LSECONDS, you will need conversion between Unix
\ ** seconds and the time/data functions supported by the RTC.
\ ** This code is taken from *\i{UnixTime.fth}. This code is
\ ** always needed if you use PowerNet. It is also very useful
\ ** for time manipulations that must later be converted to
\ ** to time and day values.

lseconds? [if]

#24 #60 * #60 * equ secs/day	\ -- 86400
\ *G Seconds per day.
#60 #60 * equ secs/hr		\ -- 3600
\ *G Seconds per hour
#60 equ secs/min		\ -- 60
\ *G Seconds per minute
#60 equ mins/hr			\ -- 60
\ *G Minutes per hour
#24 equ hrs/day			\ -- 24
\ *G Hours per day
#34698 equ uday0		\ -- 34698
\ *G The Julian day number for Unix day 0 at 1 Mar 1995.

compiler
: j>uday	\ jday -- uday
\ *G Convert a Julian day based at 0 for 1 March 1900 to a Unix
\ ** day number derived from *\fo{LSECONDS}.
  uday0 -
;

: u>jday	\ jday -- uday
\ *G Convert a Unix day number as used by *\fo{LSECONDS} to a Julian
\ ** day based at 0 for 1 March 1900.
  uday0 +
;
target

\ works with symmetric division
: >date		\ jday -- day month year
\ *G Convert a Julian day number to day month year format.
  1461 /mod				\ number of leap-year periods
   4 * >r				\ convert to number of years
   365 /mod				\ number of years and days
   dup 4 = if 2drop 365 3 then		\ correct if Feb 29th
   r> + >r				\ -- days | years
   123 + dup 5 153 */			\ compute month
   dup >r 153 5 */ - dup 0=		\ if 31 day month
   if  31 +  -1 else 0 then r> +	\ compute day
   1- dup 12 > if 12 - r> 1+ >r then	\ correct month and year
   r> 1900 +				\ -- day month year
;

: date>         \ day month year -- jday ; see month codes
\ *G Given day month and year generate the day number.
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if 12 + r> 1- >r then    	\ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

: weekday       \ jday -- n ; 0=Sunday
\ *G Given a Julian day return the day of week number where
\ ** 0 is Sunday, 1 is Monday and so on.
  4 + 0 7 um/mod drop
;

[undefined] /umod [if]
: /umod		\ u1 u2 -- urem uquot
\ *G Single precision division of unsigned U1 by U2 returning
\ ** both remainder and quotient.
  0 swap um/mod
;
[then]

: secs>td	\ lsecs -- sec min hr day mon yr
\ *G Convert a Unix seconds value to time and date.
  secs/min /umod
  mins/hr /umod
  hrs/day /umod				\ -- sec min hr day#
  u>jday >date				\ -- sec min hr day mon yr
;

: td>secs	\ sec min hr day mon yr -- secs
\ *G Use the given values for time and date, and return the Unix
\ ** *\fo{LSECONDS} value.
  date> j>uday secs/day *		\ -- sec min hr lsecs
  swap secs/hr * +
  swap secs/min * +
  +
;

[undefined] lseconds [if]
variable lseconds	\ -- addr
\ *G If not already defined, create the variable holding the
\ ** Unix LSECONDS value.
[then]

[then]


\ *************
\ *S RTC access
\ *************

0 value dow	\ -- dow ; 0=Sunday
\ *G Returns the local day of the week, starting at 0=Sunday.
\ ** This value is updated when *\fo{TIME&DATE} below is called.

: unlockRTC	\ --
\ *G Enable access to the RTC.
  bit8 _PWR pwrCR + or!			\ disable protection
;

: lockRTC	\ --
\ *G Disable access to the RTC.
  bit8 _PWR pwrCR + bic!		\ enable protection
;

: waitRTOFF	\ --
\ *G Wait for the last RTC operation to complete.
  _RTC
  begin  dup rtcCRL + @ bit5 and  until
  drop
;

: +RTCconfig	\ --
\ *G Set the CNF bit in the CRL register.
  bit4 _RTC rtcCRL + or!
;

: -RTCconfig	\ --
\ *G Clear the CNF bit in the CRL register.
  bit4 _RTC rtcCRL + bic!
;

: setupRTC	\ --
\ *G Set up the RTC to generate a seconds count from the 32768 Hz
\ ** crystal.
  waitRTOFF  +RTCconfig
  _RTC
  0 over rtcPRLH + !			\ divide by 32768
  $7FFF swap rtcPRLL + !		\ divider = 32768-1
  -RTCconfig waitRTOFF
;

: init_RTC	\ --
\ *G Initialise RTC - does not affect current state or contents.
\ ** Uses the 32kHz crystal.
  bit28 bit27 or _RCC
    2dup rccAPB1rst + bic!		\ take RTC out of reset
    rccAPB1en + or!			\ and clock it
  unlockRTC
  _RCC rccBDCR + @ 0= if		\ RTC not initialised
    $8101 $8301 _RCC rccBDCR + setMask	\ RTCEN=1, RTCSEL=01, LSEON=1
    setupRTC
  endif
[ RTCint? ] [if]
  waitRTOFF
  bit0 _RTC rtcCRH + or!
  waitRTOFF
[then]
;
' init_RTC AtCold		\ performed at power up.

: (rtcSecs)	\ -- secs
  _RTC
  dup rtcCNTL + @ $FFFF and
  swap rtcCNTH + @ #16 lshift or
;

: setRTCsecs	\ secs --
\ *G Set the RTC counter value.
  waitRTOFF  +RTCconfig
  _RTC					\ -- secs *rtc
  over #16 rshift  over rtcCNTH + !
  swap $FFFF and  swap rtcCNTL + !
  -RTCconfig waitRTOFF
;

: getRTCsecs	\ -- secs
\ *G Return the RTC seconds counter.
  (rtcSecs)
  begin (rtcSecs) tuck =  until
;

: secs>dow	\ secs --
\ *G Set *\fo{DOW} from the seconds value.
  0 secs/day um/mod nip  u>jday weekday to dow
;

: time&date     \ -- seconds mins hours day month year
\ *G Return the local time and date, and* set *\fo{DOW} as a side
\ ** effect.
  getRTCsecs  dup secs>dow  secs>td
;

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set the RTC.
\ ** The current value of *\fo{DOW} is also used.
\ ** You can set the clock from AIDE's PowerTerm window by pressing
\ ** Control-D which will send the current day-of-week, date and
\ ** time, e.g.
\ *C   5 to dow  0 56 14 10 10 2008 SetTime&Date
  td>secs dup secs>dow  dup setRTCsecs  lseconds !
;


\ ***********************
\ *S Time and date output
\ ***********************

: .2d		\ u --
\ *G Display u as two digits with leading zero.
  0 <# # # #> type
;

: (.time)	\ secs mins hours --
\ *G Display given time.
  .2d  ." :" .2d  ." :" .2d
;

create months$	\ -- addr
\ *G String containing 3 character text for the months.
  ", JanFebMarAprMayJunJulAugSepOctNovDec"

: (.date)	\ day month year --
\ *G Display given date.
  rot .2d space				\ day
  swap 1- 3 * months$ 1+ + 3 type space	\ month
  4 .r					\ year
;

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow#		\ dow --
\ *G Display day of week by name.
  3 * days$ 1+ + 3 type  ;

: .time&date	\ --
\ *G Display the day of week, date and time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  dow .dow# space  (.date) space  (.time)
  r> base !
;

: .time		\ --
\ *G Read and display time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  drop drop drop  (.time)
  r> base !
;

: .date		\ --
\ *G Read and display date.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  (.date)  drop drop drop
  r> base !
;

: .zone		\ zone --
  if  3 spaces  else  ." GMT"  endif
;

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  dow .dow# ." , "  (.date) space  (.time)
  space .zone space
  r> base !
;

lseconds?  [if]
: rtc>lsecs	\ --
\ *G Executed at cold start to copy the RTC time into *\fo{LSECONDS}.
  time&date td>secs lseconds !
;
' rtc>lsecs AtCold
[then]


\ ************************
\ *S RTC interrupt handler
\ ************************

RTCint? [if]
\ *G The code here is compiled if *\fo{RTCint?} is non-zero.

defer RTCticker	\ --
\ *G By default executes *\fo{NOOP}, assign your own action as
\ ** required. Called in *\fo{RTC-isr} below.
  assign noop to-do RTCticker

: RTC-isr	\ --
\ *G The high level interrupt handler for the one second ticker.
  bit0 _RTC rtcCRL + bic!		\ clear SECF in CRL
[ lseconds? ] [if]
  1 lseconds +!				\ count seconds
[then]
  RTCticker				\ user action
;
' RTC-isr RTCvec# EXC: RTC-irq	\ -- addr
\ *G Creates the IRQ handler for the ticker.

: +RTCint	\ -- ; start RTC interrupt
\ *G Start the RTC 1 second interrupt.
  RTCvec# EnInt			 	\ set up and enable interrupt
;
' +RTCint AtCold

: -RTCint	\ -- ; stop clock interrupt
\ *G Stop the RTC periodic interrupt.
  RTCvec# DisInt
;

[then]


\ ************
\ Test code
\ ************

((
: dd   _RTC $40 pdump  ;
: tt   0 .ansidate  ;
))

\ ======
\ *> ###
\ ======
