\ rtcSTM32F4xx.fth - STM32F4xx Real-Time Clock driver

((
Copyright (c) 2005, 2010, 2011, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   htpp://www.mpeforth.com

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20121018 SFP004 Enables BB RAM.
20120214 SFP003 STM32F4 version.
20110728 MPE002 Converted from LPC23xx to LPC17xx.
20101119 SFP001 Made interrupt configurable.
		Refactored to ease use with UnixTime.fth.
))

only forth definitions  decimal

\ ===============
\ *! rtcstm32f4xx
\ *T STM32F4xx Real Time Clock
\ ===============

[undefined] lseconds? [if]
1 equ lseconds?	\ -- flag
\ *G Set this equate true to compile the Unix LSECONDS code.
\ ** This version is compiled if *\fo{lseconds?} has not
\ ** been defined yet.
[then]

[undefined] RTCint? [if]
1 equ RTCint?	\ -- flag
\ *G Set this equate true to compile the RTC one second interrupt
\ ** handler. This version is compiled if *\fo{RTCint?} has not
\ ** been defined yet.
[then]


\ ***********************
\ *S LSECONDS conversions
\ ***********************
\ *P If you have LSECONDS, you will need conversion between Unix
\ ** seconds and the time/data functions supported by the RTC.
\ ** This code is taken from *\i{UnixTime.fth}. This code is
\ ** always needed if you use PowerNet. It is also very useful
\ ** for time manipulations that must later be converted to
\ ** to time and day values.

lseconds? [if]

#24 #60 * #60 * equ secs/day	\ -- 86400
\ *G Seconds per day.
#60 #60 * equ secs/hr		\ -- 3600
\ *G Seconds per hour
#60 equ secs/min		\ -- 60
\ *G Seconds per minute
#60 equ mins/hr			\ -- 60
\ *G Minutes per hour
#24 equ hrs/day			\ -- 24
\ *G Hours per day
#34698 equ uday0		\ -- 34698
\ *G The Julian day number for Unix day 0 at 1 Mar 1995.

compiler
: j>uday	\ jday -- uday
\ *G Convert a Julian day based at 0 for 1 March 1900 to a Unix
\ ** day number derived from *\fo{LSECONDS}.
  uday0 -
;

: u>jday	\ jday -- uday
\ *G Convert a Unix day number as used by *\fo{LSECONDS} to a Julian
\ ** day based at 0 for 1 March 1900.
  uday0 +
;
target

\ works with symmetric division
: >date		\ jday -- day month year
\ *G Convert a Julian day number to day month year format.
  1461 /mod				\ number of leap-year periods
   4 * >r				\ convert to number of years
   365 /mod				\ number of years and days
   dup 4 = if 2drop 365 3 then		\ correct if Feb 29th
   r> + >r				\ -- days | years
   123 + dup 5 153 */			\ compute month
   dup >r 153 5 */ - dup 0=		\ if 31 day month
   if  31 +  -1 else 0 then r> +	\ compute day
   1- dup 12 > if 12 - r> 1+ >r then	\ correct month and year
   r> 1900 +				\ -- day month year
;

: date>         \ day month year -- jday ; see month codes
\ *G Given day month and year generate the day number.
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if 12 + r> 1- >r then    	\ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

[undefined] /umod [if]
: /umod		\ u1 u2 -- urem uquot
\ *G Single precision division of unsigned U1 by U2 returning
\ ** both remainder and quotient.
  0 swap um/mod
;
[then]

: secs>td	\ lsecs -- sec min hr day mon yr
\ *G Convert a Unix seconds value to time and date.
  secs/min /umod
  mins/hr /umod
  hrs/day /umod				\ -- sec min hr day#
  u>jday >date				\ -- sec min hr day mon yr
;

: td>secs	\ sec min hr day mon yr -- secs
\ *G Use the given values for time and date, and return the Unix
\ ** *\fo{LSECONDS} value.
  date> j>uday secs/day *		\ -- sec min hr lsecs
  swap secs/hr * +
  swap secs/min * +
  +
;

[undefined] lseconds [if]
variable lseconds	\ -- addr
\ *G If not already defined, create the variable holding the
\ ** Unix LSECONDS value.
[then]

[then]


\ *************
\ *S RTC access
\ *************

: unlockRTC	\ --
\ *G Enable access to the RTC.
  bit8 _PWR pwrCR + or!			\ disable protection
;

: lockRTC	\ --
\ *G Disable access to the RTC.
  bit8 _PWR pwrCR + bic!		\ enable protection
;

: +writeRTC	\ --
\ *G Enable access to protected RTC registers.
  _RTC  $CA over rtcWPR + !  $53 swap rtcWPR + !
;

: -writeRTC	\ --
\ *G Disable access to protected RTC registers.
  _RTC  $FF over rtcWPR + !  $00 swap rtcWPR + !
;

: waitRTCsync	\ --
\ *G Wait for RTC sync after a write to an RTC register.
  _RTC
  bit5 over rtcISR + bic!
  begin
    pause  dup rtcISR + @ bit5 and
  until
  drop
;

: init_RTC	\ --
\ *G Initialise RTC - does not affect current state or contents.
\ ** Uses the 32kHz crystal. Also enables battery backed RAM.
  unlockRTC
  $8101 $8301 _RCC rccBDCR + setMask	\ RTCEN=1, RTCSEL=01, LSEON=1
  bit18 _RCC rccAHB1ENR + or!		\ enable BB RAM
;
' init_RTC AtCold		\ performed at power up.

: getRTC	\ -- date time
\ *G Return the date and time registers.
  +writeRTC waitRTCsync
  _RTC
  dup rtcDR + @ $00FF:FF3F and
  swap rtcTR + @ $007F:7F7F and
  -writeRTC
;

0 value dow	\ -- dow ; 0=Sunday
\ *G Returns the local day of the week, starting at 0=Sunday.
\ ** This value is updated when *\fo{TIME&DATE} below is called.

: bcd2>bin	\ x -- u
\ *G Convert a two-digit BCD number into binary
  dup 4 rshift $0F and #10 *  swap $0F and +
;

: bin>2bcd	\ u -- x
\ *G Convert a binary value 0.99 into a two-digit BCD number.
  #10 /mod 4 lshift or
;

: RTC>TD	\ date time -- seconds mins hours day month year
\ *G Convert the consolidated date/time values to Forth standard form.
  dup $7F and bcd2>bin -rot		\ seconds
  dup 8 rshift $7F and bcd2>bin -rot	\ minutes
  #16 rshift $3F and bcd2>bin swap	\ hours ; -- secs mins hrs date
  dup $3F and bcd2>bin swap		\ day
  dup 8 rshift $1F and bcd2>bin swap	\ month
  dup #16 rshift bcd2>bin #2000 + swap	\ year
  #13 rshift $07 and 1- to dow		\ dow
;

: time&date     \ -- seconds mins hours day month year
\ *G Return the operating system local time and date, and
\ ** set *\fo{DOW} as a side effect.
  getRTC RTC>TD
;
((
  : time&date     \ -- seconds mins hours day month year
   \ *G Return the operating system local time and date, and
   \ ** set *\fo{DOW} as a side effect.
    getRTC
    begin					\ rollover check
  [ tasking? ] [if]  pause  [then]
      getRTC 2swap			\ -- dnew dold
      2over d=
    until					\ -- dnew
    RTC>TD
  ;
))

((
 To program the initial time and date calendar values, including
the time format and the prescaler configuration, the following
sequence is required:
1. Set INIT bit to 1 in the RTC_ISR register to enter
initialization mode. In this mode, the calendar counter is stopped
and its value can be updated.
2. Poll INITF bit of in the RTC_ISR register. The initialization
phase mode is entered when INITF is set to 1. It takes around 2
RTCCLK clock cycles (due to  clock synchronization).
3. To generate a 1 Hz clock for the calendar counter, program
first the synchronous prescaler factor in RTC_PRER register, and
then program the asynchronous prescaler factor. Even if only one
of the two fields needs to be changed, 2 separate write accesses
must be performed to the TC_PRER register.
4. Load the initial time and date values in the shadow registers
(RTC_TR and RTC_DR), and configure the time format (12 or 24
hours) through the FMT bit in the RTC_CR register.
5. Exit the initialization mode by clearing the INIT bit. The
actual calendar counter value is then automatically loaded and the
counting restarts after 4 RTCCLK clock cycles.
When the initialization sequence is complete, the calendar starts counting.
))

: +initRTC	\ --
\ *G Set up to initialise the RTC.
  _RTC
  bit6 over rtcISR + bic!
  bit7 over rtcISR + or!
  begin
    pause  dup rtcISR + @ bit6 and
  until
  drop
;

: -initRTC	\ --
\ *G Finish initialisation of the RTC.
  bit7 _RTC rtcISR + bic!
;

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set the RTC.
\ ** The current value of *\fo{DOW} is also used.
\ ** You can set the clock from AIDE's PowerTerm window by pressing
\ ** Control-D which will send the current day-of-week, date and
\ ** time, e.g.
\ *C   5 to dow  0 56 14 10 10 2008 SetTime&Date
  unlockRTC
  +writeRTC +initRTC
  #100 mod bin>2bcd #16 lshift		\ year
  swap bin>2bcd 8 lshift or		\ month
  swap bin>2bcd or			\ day
  dow 1+ #13 lshift or			\ dow
  $00FF:FF3F _RTC rtcDR + setMask	\ write date
  -initRTC -writeRTC
  waitRTCsync				\ after each reg write
  +writeRTC +initRTC
  bin>2bcd $3F and #16 lshift		\ hr, PM=0
  swap bin>2bcd 8 lshift or		\ min
  swap bin>2bcd or			\ sec
  $003F:7F7F _RTC rtcTR + setMask	\ write time
  -initRTC  -writeRTC
  waitRTCsync
;

_BKPSRAM constant _BKPSRAM	\ -- addr
\ *G Base address of battery backed RAM.


\ ***********************
\ *S Time and date output
\ ***********************

: .2d		\ u --
\ *G Display u as two digits with leading zero.
  0 <# # # #> type
;

: (.time)	\ secs mins hours --
\ *G Display given time.
  .2d  ." :" .2d  ." :" .2d
;

create months$	\ -- addr
\ *G String containing 3 character text for the months.
  ", JanFebMarAprMayJunJulAugSepOctNovDec"

: (.date)	\ day month year --
\ *G Display given date.
  rot .2d space				\ day
  swap 1- 3 * months$ 1+ + 3 type space	\ month
  4 .r					\ year
;

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow#		\ dow --
\ *G Display day of week by name.
  3 * days$ 1+ + 3 type  ;

: .time&date	\ --
\ *G Display the day of week, date and time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  dow .dow# space  (.date) space  (.time)
  r> base !
;

: .time		\ --
\ *G Read and display time.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  drop drop drop  (.time)
  r> base !
;

: .date		\ --
\ *G Read and display date.
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  (.date)  drop drop drop
  r> base !
;

: .zone		\ zone --
  if  3 spaces  else  ." GMT"  endif
;

: .AnsiDate	\ zone --
\ *G Display the day of week, date and time. If zone is 0
\ ** GMT is displayed. The format is:
\ *C   dow, hh:mm:ss dd Mmm yyyy [GMT]
  base @ >r  decimal
  time&date				\ -- seconds mins hours day month year
  dow .dow# ." , "  (.date) space  (.time)
  space .zone space
  r> base !
;

lseconds?  [if]
: rtc>lsecs	\ --
\ *G Executed at cold start to copy the RTC time into *\fo{LSECONDS}.
  time&date td>secs lseconds !
;
' rtc>lsecs AtCold
[then]


\ ************************
\ *S RTC interrupt handler
\ ************************

RTCint? [if]
\ *G The code here is compiled if *\fo{RTCint?} is non-zero.

defer RTCticker	\ --
\ *G By default executes *\fo{NOOP}, assign your own action as
\ ** required. Called in *\fo{RTC-isr} below.
  assign noop to-do RTCticker

: RTC-isr	\ --
\ *G The high level interrupt handler for the one second ticker.
  bit10 _RTC rtcISR + bic!		\ clear WUTF in ISR
  bit22 _EXTI extiPR + or!		\ clear EXTI PR reg bit
[ lseconds? ] [if]
  1 lseconds +!				\ count seconds
[then]
  RTCticker				\ user action
;
' RTC-isr RTCvec# EXC: RTC-irq	\ -- addr
\ *G Creates the IRQ handler for the ticker.

: tri  bit22  _EXTI extiSWIER + or!  ;
((
  PB14 dup isInput  dup notPulled
  $0100 _SYSCFG sysEXTICR4 + !		\ connect PB14 to EXTI
  bit14 _EXTI  2dup extiFTSR + !	\ happens on falling edge
  extiIMR + !				\ enable interrupt
))

: +RTCint	\ -- ; start RTC interrupt
\ *G Start the RTC 1 second interrupt.
  di
  +writeRTC
  _RTC
  bit10 over rtcCR + bic!		\ WUTE=0
  begin
    dup rtcISR + @ bit2 and		\ wait WUTWF=1
  until
  0 over rtcWUTR + !
  $0004 $0007 2 pick rtcCR + setMask	\ set WUCKSEL=100
  $4400 $4400 rot rtcCR + setMask	\ set WUTIE, WUTE
  -writeRTC
  RTCvec# EnInt			 	\ set up and enable interrupt
  bit22 _EXTI
    2dup extiRTSR + or!			\ trigger on rising edge
    2dup extiFTSR + bic!		\ not falling
    extiIMR + or!			\ enable EXT22=RTC wakeup
  bit10 _RTC rtcISR + bic!		\ clear WUTF to rise again
  ei
;
' +RTCint AtCold

: -RTCint	\ -- ; stop clock interrupt
\ *G Stop the RTC periodic interrupt.
  RTCvec# DisInt
;

[then]


\ ************
\ Test code
\ ************

((
: dd   _RTC $40 pdump  ;
: tt   0 .ansidate  ;
))

\ ======
\ *> ###
\ ======
