\ serK60p.fth - Kinetis polled serial driver

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20111111 MPE001 First release.
))

only forth definitions
decimal

\ ===========
\ *! serk60p
\ *T Kinetis K60 polled serial driver
\ ===========
\ *P The file *\i{Cortex\Drivers\serK60p.fth} provides polled
\ ** serial drivers for the on-chip UARTs.


\ *****************
\ *S Initialisation
\ *****************

\ *P The baud rate generator value is defined by:
\ *C   baudrate=moduleclock/(16*(SBR+BRFD))
\ *P where SBR is a 13 bit integer and BRFD is a 5 bit fraction.
\ ** To use this, we rarrange the equation to the form:
\ *C   (SBR+BRFD)=moduleclock/(16*baudrate)
\ *P In order to use integer operations, we scale this.
\ *C   brgval=2*moduleclock/baudrate
\ ** where SBR is bits 17:5 of brgval and BRFD is bits 4:0
\ ** of brgval.

interpreter
: genBRG	\ baud clock -- brgval
\ *G Generate the combined baud rate divisor
  2* swap /				\ 18 bit divisor
;

: brg>rlh	\ brgval -- brfa brl brh
\ *G Generate the three portions of the baud rate divisor.
  dup $01F and swap 5 rshift		\ -- brfa brx
  dup $FF and swap 8 rshift		\ -- brfa brl brh
  dup $1F > abort" UART baud rate overflow"
;
target

: genBRG	\ baud clock -- brgval
\ *G Generate the combined baud rate divisor
  2* swap /				\ 18 bit divisor
;

: brg>rlh	\ brgval -- brfa brl brh
\ *G Generate the three portions of the baud rate divisor.
  dup $01F and swap 5 rshift		\ -- brfa brx
  dup $FF and swap 8 rshift		\ -- brfa brl brh
  dup $1F > abort" UART baud rate overflow"
;

: (init-ser)	\ brgval ^uart --
\ *G Initialise the given UART to 8,N,1 format with no flow
\ ** control and with given baud rate divisor.
  >r brg>rlh r>				\ -- brfa brl brh ^uart
  0 over UnC2 + c!			\ disable UART and ints
  $88 over UnPFIFO + c!			\ enable FIFOs
  0 over UnC1 + c!			\ N81
  swap 0 or  over UnBDH + c!		\ high portion of baud rate integer
  swap  over UnBDL + c!			\ low portion of baud rate integer
  swap 0 or  over UnC4 + c!		\ fraction, normal operation
  $0C swap UnC2 + bor!			\ enable TX, Rx
;


\ ********************
\ *S Serial primitives
\ ********************

\ internal

: +FaultConsole	( -- )  ;
\ *G Because this is a polled driver, *\fo{+FaultConsole} for
\ ** fault handling is a *\fo{NOOP}. See *\i{Cortex\FaultCortex.fth}
\ ** for more details.

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup UnS1 + c@ bit7 and		\ check TDRE
  until
  UnD + c!
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)
;

: (serkey?)	\ base -- t/f
\ *G Return true if the given UART has a character avilable to read.
  UnS1 + c@ bit5 and 0<>		\ check RDRF
;

: (serkey)	\ base -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
[ tasking? ] [if]  pause  [then]
    dup UnS1 + c@ bit5 and
  until
  UnD + c@
;

external


\ ********
\ *S UART0
\ ********
\ *P Make sure that *\fo{init-ser0} initialises the pins to the
\ ** configuration that you are using.

useUART0? [if]
console0-speed system-speed GenBRG equ u0-brgval	\ -- n
\ *G UART0 baud rate divisor.

: init-ser0	\ --
\ *G Initialise serial port 0.
  bit10 _SIM simSCGC4 + or!		\ enable clock to UART
  $0300 _PORTD portPCR6 + !		\ Alt 3 on PTD6 for UART0 TxD
  $0300 _PORTD portPCR7 + !		\ Alt 3 on PTD7 for UART0 RxD
  u0-brgval _UART0 (init-ser)
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  _UART0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  _UART0 (serkey)  ;
create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ********
\ *S UART1
\ ********
\ *P Make sure that *\fo{init-ser1} initialises the pins to the
\ ** configuration that you are using.

useUART1? [if]

console1-speed bus-speed GenBRG equ u1-brgval	\ -- n
\ *G UART1 baud rate divisor.

: init-ser1	\ --
\ *G Initialise serial port 0
  bit11 _SIM simSCGC4 + or!		\ enable clock to UART
  $0300 _PORTC portPCR4 + !		\ Alt 3 on PTC4 for UART3 TxD
  $0300 _PORTC portPCR3 + !		\ Alt 3 on PTC3 for UART3 RxD
  u1-brgval _UART1 (init-ser)
;

: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 (seremit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 (sertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 (sercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  _UART1 (serkey?)  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  _UART1 (serkey)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ ********
\ *S UART2
\ ********
\ *P Make sure that *\fo{init-ser2} initialises the pins to the
\ ** configuration that you are using.

useUART2? [if]

console2-speed bus-speed GenBRG equ u2-brgval	\ -- n
\ *G UART2 baud rate divisor.

: init-ser2	\ --
\ *G Initialise serial port 2.
  bit12 _SIM simSCGC4 + or!		\ enable clock to UART
  $0300 _PORTD portPCR3 + !		\ Alt 3 on PTD3 for UART3 TxD
  $0300 _PORTD portPCR2 + !		\ Alt 3 on PTD2 for UART3 RxD
  u2-brgval _UART2 (init-ser)
;

: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 (seremit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 (sertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 (sercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  _UART2 (serkey?)  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  _UART2 (serkey)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ ********
\ *S UART3
\ ********
\ *P On the K60 Tower system, the console is implemented on UART3
\ ** connected through the DB9 connector on the TWR-SER board.

useUART3? [if]

console3-speed bus-speed GenBRG equ u3-brgval	\ -- n
\ *G UART3 baud rate divisor.

: init-ser3	\ --
\ *G Initialise serial port 3.
  bit13 _SIM simSCGC4 + or!		\ enable clock to UART
  $0300 _PORTC portPCR17 + !		\ Alt 3 on PTC17 for UART3 TxD
  $0300 _PORTC portPCR16 + !		\ Alt 3 on PTC16 for UART3 RxD
  u3-brgval _UART3 (init-ser)
;

: seremit3	\ char --
\ *G Transmit a character on UART3.
  _UART3 (seremit)  ;
: sertype3	\ c-addr len --
\ *G Transmit a string on UART3.
  _UART3 (sertype)  ;
: sercr3	\ --
\ *G Issue a CR/LF pair to UART3.
  _UART3 (sercr)  ;
: serkey?3	\ -- flag
\ *G Return true if UART3 has a character available.
  _UART3 (serkey?)  ;
: serkey3	\ -- char
\ *G Wait for a character on UART3.
  _UART3 (serkey)  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART3.
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line

console-port 3 = [if]
  console3 constant console
[then]

[then]


\ ********
\ *S UART4
\ ********
\ *P Make sure that *\fo{init-ser0} initialises the pins to the
\ ** configuration that you are using.

useUART4? [if]

console4-speed bus-speed GenBRG equ u4-brgval	\ -- n
\ *G UART4 baud rate divisor.

: init-ser4	\ --
\ *G Initialise serial port 4.
  bit10 _SIM simSCGC1 + or!		\ enable clock to UART
  $0300 _PORTE portPCR24 + !		\ Alt 3 on PTE24 for UART4 TxD
  $0300 _PORTE portPCR25 + !		\ Alt 3 on PTD7 for UART4 RxD
  u4-brgval _UART4 (init-ser)
;

: seremit4	\ char --
\ *G Transmit a character on UART4.
  _UART4 (seremit)  ;
: sertype4	\ c-addr len --
\ *G Transmit a string on UART4.
  _UART4 (sertype)  ;
: sercr4	\ --
\ *G Issue a CR/LF pair to UART4.
  _UART4 (sercr)  ;
: serkey?4	\ -- flag
\ *G Return true if UART4 has a character available.
  _UART4 (serkey?)  ;
: serkey4	\ -- char
\ *G Wait for a character on UART4.
  _UART4 (serkey)  ;

create Console4	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART4.
  ' serkey4 ,		\ -- char ; receive char
  ' serkey?4 ,		\ -- flag ; check receive char
  ' seremit4 ,		\ -- char ; display char
  ' sertype4 ,		\ caddr len -- ; display string
  ' sercr4 ,		\ -- ; display new line

console-port 4 = [if]
  console4 constant console
[then]

[then]


\ ********
\ *S UART5
\ ********
\ *P Make sure that *\fo{init-ser0} initialises the pins to the
\ ** configuration that you are using.

useUART5? [if]

console5-speed bus-speed GenBRG equ u5-brgval	\ -- n
\ *G UART5 baud rate divisor.

: init-ser5	\ --
\ *G Initialise serial port 5.
  bit11 _SIM simSCGC1 + or!		\ enable clock to UART
  $0300 _PORTE portPCR8 + !		\ Alt 3 on PTE8 for UART5 TxD
  $0300 _PORTE portPCR9 + !		\ Alt 3 on PTE9 for UART5 RxD
  u5-brgval _UART5 (init-ser)
;

: seremit5	\ char --
\ *G Transmit a character on UART5.
  _UART5 (seremit)  ;
: sertype5	\ c-addr len --
\ *G Transmit a string on UART5.
  _UART5 (sertype)  ;
: sercr5	\ --
\ *G Issue a CR/LF pair to UART5.
  _UART5 (sercr)  ;
: serkey?5	\ -- flag
\ *G Return true if UART5 has a character available.
  _UART5 (serkey?)  ;
: serkey5	\ -- char
\ *G Wait for a character on UART5.
  _UART5 (serkey)  ;

create Console5	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART5.
  ' serkey5 ,		\ -- char ; receive char
  ' serkey?5 ,		\ -- flag ; check receive char
  ' seremit5 ,		\ -- char ; display char
  ' sertype5 ,		\ caddr len -- ; display string
  ' sercr5 ,		\ -- ; display new line

console-port 5 = [if]
  console5 constant console
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
  [ useUART3? ] [if]  init-ser3  [then]
  [ useUART4? ] [if]  init-ser4  [then]
  [ useUART5? ] [if]  init-ser5  [then]
;


\ ======
\ *> ###
\ ======

