\ serK60qi.fth - Kinetis interrupt queued serial driver

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20111115 MPE001 First release.
))

only forth definitions
decimal

\ ===========
\ *! serk60qi
\ *T Kinetis K60 interrupt queued serial driver
\ ===========
\ *P The file *\i{Cortex\Drivers\serK60qi.fth} provides interrupt
\ ** and queued serial drivers for the on-chip UARTs.

\ *P Primary testing was performed using UART3 on a Tower system
\ ** with a TWR-K60N512 CPU module and a TWR-SER module.
\ ** Make sure that your *\fo{init-sern} words initialise the
\ ** pin registers to the configuration that you are using.


\ *****************
\ *S Initialisation
\ *****************
\ *P The baud rate generator value is defined by:
\ *C   baudrate=moduleclock/(16*(SBR+BRFD))
\ *P where SBR is a 13 bit integer and BRFD is a 5 bit fraction.
\ ** To use this, we rarrange the equation to the form:
\ *C   (SBR+BRFD)=moduleclock/(16*baudrate)
\ *P In order to use integer operations, we scale this.
\ *C   brgval=2*moduleclock/baudrate
\ ** where SBR is bits 17:5 of brgval and BRFD is bits 4:0
\ ** of brgval.

interpreter
: genBRG	\ baud clock -- brgval
\ *G Generate the combined baud rate divisor
  2* swap /				\ 18 bit divisor
;

: brg>rlh	\ brgval -- brfa brl brh
\ *G Generate the three portions of the baud rate divisor.
  dup $01F and swap 5 rshift		\ -- brfa brx
  dup $FF and swap 8 rshift		\ -- brfa brl brh
  dup $1F > abort" UART baud rate overflow"
;
target

: genBRG	\ baud clock -- brgval
\ *G Generate the combined baud rate divisor
  2* swap /				\ 18 bit divisor
;

: brg>rlh	\ brgval -- brfa brl brh
\ *G Generate the three portions of the baud rate divisor.
  dup $01F and swap 5 rshift		\ -- brfa brx
  dup $FF and swap 8 rshift		\ -- brfa brl brh
  dup $1F > abort" UART baud rate overflow"
;

: (init-ser)	\ brgval ^uart --
\ *G Initialise the given UART to 8,N,1 format with no flow
\ ** control and with given baud rate divisor.
  >r brg>rlh r>				\ -- brfa brl brh ^uart
  0 over UnC2 + c!			\ disable UART and ints
  $88 over UnPFIFO + c!			\ enable FIFOs
  0 over UnC1 + c!			\ N81
  swap 0 or  over UnBDH + c!		\ high portion of baud rate integer
  swap  over UnBDL + c!			\ low portion of baud rate integer
  swap 0 or  over UnC4 + c!		\ fraction, normal operation
  $0C over UnC2 + bor!			\ enable TX, RX
  bit5 swap UnC2 + bor!			\ enable receive interrupt
;


\ ********************
\ *S Serial primitives
\ ********************

\ internal

: >RxQ		\ char queue --
\ *G Put character from UART into queue.
\ ** INTERNAL.
  dup (cqFull?)				\ if queue full
  if  2drop  exit  endif		\  discard character
  (>cqueue)				\  queue character
;

: FIFO>RxQ      \ base queue --
\ *G Given a UART base address and a character queue,
\ ** empty the UART Rx FIFO into the queue.
\ ** INTERNAL.
  begin
    over UnS1 + c@ bit5 and
   while
    over UnD + c@  over >RxQ
  repeat
  2drop
;

: TxQ>FIFO	\ base queue --
\ *G Given a UART base address and a character queue,
\ ** put a transmit character into the UART Tx FIFO.
  dup cqEmpty? if
    drop
    bit7 swap UnC2 + bbic!		\ clear Transmit Interrupt Enable
    exit
  endif
  (cqueue>) swap UnD + c!		\ write character
;

0 value FaultConsole?	\ -- flag
\ *G When set non-zero, transmit is in polled mode. This is
\ ** required when fault exceptions provide debug information.

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
  -1 to FaultConsole?
  [ tasking? ] [if]  single  [then]
;

: (qseremit)	\ char base queue --
\ *G Send a character on the UART whose base address and queue are given.
  FaultConsole? 0= if
    begin  dup cqfull? while  pause  repeat
    [i
      swap >r
      (>cqueue)
      bit7 r> UnC2 + bor!
    i]
    exit
  endif
  drop					\ discard queue
  bit7 over UnC2 + bbic!		\ disable TX interrupt
  begin
    dup UnS1 + c@ bit7 and		\ check TDRE
  until
  UnD + c!
;

: (qsertype)	\ caddr len base queue --
\ *G Send a string on the UART whose base address and queue are given.
  2swap bounds				\ -- base queue
  ?do  2dup i c@ -rot (qseremit)  loop
  2drop
;

: (qsercr)	\ base queue --
\ *G Send a CR/LF on the UART whose base address and queue are given.
  2dup $0D -rot (qseremit)  $0A -rot (qseremit)
;

external


\ ********
\ *S UART0
\ ********
\ *P Make sure that *\fo{init-ser0} initialises the pins to the
\ ** configuration that you are using.

useUART0? [if]
console0-speed system-speed GenBRG equ u0-brgval	\ -- n
\ *G UART0 baud rate divisor.

#128 cqueue: SerInpQ0	\ -- addr
\ *G Receiver circular byte queue.
#64 cqueue: SerOutQ0	\ -- addr
\ *G Transmitter circular byte queue.

: ser0-isrh	\ --
\ *G UART0 high level ISR.
\ ** INTERNAL.
  _UART0
  dup UnS1 c@ bit5 and if		\ -- ^uart ; RX?
    dup SerInpQ0 FIFO>RxQ
  endif
  dup UnS1 c@ bit7 and if		\ -- ^uart ; TX?
    dup SerOutQ0 TxQ>FIFO
  endif
  drop
;
' ser0-isrh INT_UART0_RX_TX EXC: ser3-isr
\ *G UART0 interrupt service routine entry point.

: init-ser0	\ --
\ *G Initialise serial port 0.
  di
  SerInpQ0 init-cqueue			\ initialise character queues
  SerOutQ0 init-cqueue
  bit10 _SIM simSCGC4 + or!		\ enable clock to UART
  $0300 _PORTD portPCR6 + !		\ Alt 3 on PTD6 for UART0 TxD
  $0300 _PORTD portPCR7 + !		\ Alt 3 on PTD7 for UART0 RxD
  u0-brgval _UART0 (init-ser)
  $80 INT_UART0_RX_TX setPri		\ mid level priority
  INT_UART0_RX_TX EnInt			\ interrupt enable in NVIC
  ei
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 SerOutQ0 (qseremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 SerOutQ0 (qsertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 SerOutQ0 (qsercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  SerInpQ0 cqnotempty?  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  SerInpQ0 cqueue>  ;

create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line

console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ********
\ *S UART1
\ ********
\ *P Make sure that *\fo{init-ser1} initialises the pins to the
\ ** configuration that you are using.

useUART1? [if]

console1-speed GenBRG equ u1-brgval	\ -- n
\ *G UART1 baud rate divisor.


#128 cqueue: SerInpQ1	\ -- addr
\ *G Receiver circular byte queue.
#64 cqueue: SerOutQ1	\ -- addr
\ *G Transmitter circular byte queue.

: ser1-isrh	\ --
\ *G UART1 high level ISR.
\ ** INTERNAL.
  _UART1
  dup UnS1 c@ bit5 and if		\ -- ^uart ; RX?
    dup SerInpQ1 FIFO>RxQ
  endif
  dup UnS1 c@ bit7 and if		\ -- ^uart ; TX?
    dup SerOutQ1 TxQ>FIFO
  endif
  drop
;
' ser1-isrh INT_UART1_RX_TX EXC: ser1-isr
\ *G UART1 interrupt service routine entry point.

: init-ser1	\ --
\ *G Initialise serial port 1.
  di
  SerInpQ1 init-cqueue			\ initialise character queues
  SerOutQ1 init-cqueue
  bit11 _SIM simSCGC4 + or!		\ enable clock to UART
  $0300 _PORTC portPCR4 + !		\ Alt 3 on PTC4 for UART1 TxD
  $0300 _PORTC portPCR3 + !		\ Alt 3 on PTC3 for UART1 RxD
  u1-brgval _UART1 (init-ser)
  $80 INT_UART1_RX_TX setPri		\ mid level priority
  INT_UART1_RX_TX EnInt			\ interrupt enable in NVIC
  ei
;

: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 SerOutQ1 (qseremit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 SerOutQ1 (qsertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 SerOutQ1 (qsercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  SerInpQ1 cqnotempty?  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  SerInpQ1 cqueue>  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ ********
\ *S UART2
\ ********
\ *P Make sure that *\fo{init-ser2} initialises the pins to the
\ ** configuration that you are using.

useUART2? [if]

console2-speed bus-speed GenBRG equ u2-brgval	\ -- n
\ *G UART2 baud rate divisor.

#128 cqueue: SerInpQ2	\ -- addr
\ *G Receiver circular byte queue.
#64 cqueue: SerOutQ2	\ -- addr
\ *G Transmitter circular byte queue.

: ser2-isrh	\ --
\ *G UART2 high level ISR.
\ ** INTERNAL.
  _UART2
  dup UnS1 c@ bit5 and if		\ -- ^uart ; RX?
    dup SerInpQ2 FIFO>RxQ
  endif
  dup UnS1 c@ bit7 and if		\ -- ^uart ; TX?
    dup SerOutQ2 TxQ>FIFO
  endif
  drop
;
' ser2-isrh INT_UART2_RX_TX EXC: ser2-isr
\ *G UART2 interrupt service routine entry point.

: init-ser2	\ --
\ *G Initialise serial port 2.
  di
  SerInpQ2 init-cqueue			\ initialise character queues
  SerOutQ2 init-cqueue
  bit12 _SIM simSCGC4 + or!		\ enable clock to UART
  $0300 _PORTD portPCR3 + !		\ Alt 3 on PTD3 for UART2 TxD
  $0300 _PORTD portPCR2 + !		\ Alt 3 on PTD2 for UART2 RxD
  u2-brgval _UART2 (init-ser)
  $80 INT_UART2_RX_TX setPri		\ mid level priority
  INT_UART2_RX_TX EnInt			\ interrupt enable in NVIC
  ei
;

: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 SerOutQ2 (qseremit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 SerOutQ2 (qsertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 SerOutQ2 (qsercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  SerInpQ2 cqnotempty?  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  SerInpQ2 cqueue>  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ ********
\ *S UART3
\ ********
\ *P On the K60 Tower system, the console is implemented on UART3
\ ** connected through the DB9 connector on the TWR-SER board.

useUART3? [if]

console3-speed bus-speed GenBRG equ u3-brgval	\ -- n
\ *G UART3 baud rate divisor.

#128 cqueue: SerInpQ3	\ -- addr
\ *G Receiver circular byte queue.
#64 cqueue: SerOutQ3	\ -- addr
\ *G Transmitter circular byte queue.

: ser3-isrh	\ --
\ *G UART3 high level ISR.
\ ** INTERNAL.
  _UART3
  dup UnS1 c@ bit5 and if		\ -- ^uart ; RX?
    dup SerInpQ3 FIFO>RxQ
  endif
  dup UnS1 c@ bit7 and if		\ -- ^uart ; TX?
    dup SerOutQ3 TxQ>FIFO
  endif
  drop
;
' ser3-isrh INT_UART3_RX_TX EXC: ser3-isr
\ *G UART3 interrupt service routine entry point.

: init-ser3	\ --
\ *G Initialise serial port 3.
  di
  SerInpQ3 init-cqueue			\ initialise character queues
  SerOutQ3 init-cqueue
  bit13 _SIM simSCGC4 + or!		\ enable clock to UART
  $0300 _PORTC portPCR17 + !		\ Alt 3 on PTC17 for UART3 TxD
  $0300 _PORTC portPCR16 + !		\ Alt 3 on PTC16 for UART3 RxD
  u3-brgval _UART3 (init-ser)
  $80 INT_UART3_RX_TX setPri		\ mid level priority
  INT_UART3_RX_TX EnInt			\ interrupt enable in NVIC
  ei
;

: seremit3	\ char --
\ *G Transmit a character on UART3.
  _UART3 SerOutQ3 (qseremit)  ;
: sertype3	\ c-addr len --
\ *G Transmit a string on UART3.
  _UART3 SerOutQ3 (qsertype)  ;
: sercr3	\ --
\ *G Issue a CR/LF pair to UART3.
  _UART3 SerOutQ3 (qsercr)  ;
: serkey?3	\ -- flag
\ *G Return true if UART3 has a character available.
  SerInpQ3 cqnotempty?  ;
: serkey3	\ -- char
\ *G Wait for a character on UART3.
  SerInpQ3 cqueue>  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART3.
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line

console-port 3 = [if]
  console3 constant console
[then]

[then]


\ ********
\ *S UART4
\ ********
\ *P Make sure that *\fo{init-ser4} initialises the pins to the
\ ** configuration that you are using.

useUART4? [if]

console4-speed bus-speed GenBRG equ u4-brgval	\ -- n
\ *G UART4 baud rate divisor.

#128 cqueue: SerInpQ4	\ -- addr
\ *G Receiver circular byte queue.
#64 cqueue: SerOutQ4	\ -- addr
\ *G Transmitter circular byte queue.

: ser4-isrh	\ --
\ *G UART4 high level ISR.
\ ** INTERNAL.
  _UART4
  dup UnS1 c@ bit5 and if		\ -- ^uart ; RX?
    dup SerInpQ4 FIFO>RxQ
  endif
  dup UnS1 c@ bit7 and if		\ -- ^uart ; TX?
    dup SerOutQ4 TxQ>FIFO
  endif
  drop
;
' ser4-isrh INT_UART4_RX_TX EXC: ser4-isr
\ *G UART4 interrupt service routine entry point.

: init-ser4	\ --
\ *G Initialise serial port 4.
  di
  SerInpQ4 init-cqueue			\ initialise character queues
  SerOutQ4 init-cqueue
  bit10 _SIM simSCGC1 + or!		\ enable clock to UART
  $0300 _PORTE portPCR24 + !		\ Alt 3 on PTE24 for UART4 TxD
  $0300 _PORTE portPCR25 + !		\ Alt 3 on PTE25 for UART4 RxD
  u4-brgval _UART4 (init-ser)
  $80 INT_UART4_RX_TX setPri		\ mid level priority
  INT_UART4_RX_TX EnInt			\ interrupt enable in NVIC
  ei
;

: seremit4	\ char --
\ *G Transmit a character on UART4.
  _UART4 SerOutQ4 (qseremit)  ;
: sertype4	\ c-addr len --
\ *G Transmit a string on UART4.
  _UART4 SerOutQ4 (qsertype)  ;
: sercr4	\ --
\ *G Issue a CR/LF pair to UART4.
  _UART4 SerOutQ4 (qsercr)  ;
: serkey?4	\ -- flag
\ *G Return true if UART4 has a character available.
  SerInpQ4 cqnotempty?  ;
: serkey4	\ -- char
\ *G Wait for a character on UART4.
  SerInpQ4 cqueue>  ;

create Console4	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART4.
  ' serkey4 ,		\ -- char ; receive char
  ' serkey?4 ,		\ -- flag ; check receive char
  ' seremit4 ,		\ -- char ; display char
  ' sertype4 ,		\ caddr len -- ; display string
  ' sercr4 ,		\ -- ; display new line

console-port 4 = [if]
  console4 constant console
[then]

[then]


\ ********
\ *S UART5
\ ********
\ *P Make sure that *\fo{init-ser0} initialises the pins to the
\ ** configuration that you are using.

useUART5? [if]

console5-speed bus-speed GenBRG equ u5-brgval	\ -- n
\ *G UART5 baud rate divisor.

#128 cqueue: SerInpQ5	\ -- addr
\ *G Receiver circular byte queue.
#64 cqueue: SerOutQ5	\ -- addr
\ *G Transmitter circular byte queue.

: ser5-isrh	\ --
\ *G UART5 high level ISR.
\ ** INTERNAL.
  _UART5
  dup UnS1 c@ bit5 and if		\ -- ^uart ; RX?
    dup SerInpQ5 FIFO>RxQ
  endif
  dup UnS1 c@ bit7 and if		\ -- ^uart ; TX?
    dup SerOutQ5 TxQ>FIFO
  endif
  drop
;
' ser5-isrh INT_UART5_RX_TX EXC: ser5-isr
\ *G UART5 interrupt service routine entry point.

: init-ser5	\ --
\ *G Initialise serial port 5.
  di
  SerInpQ5 init-cqueue			\ initialise character queues
  SerOutQ5 init-cqueue
  bit11 _SIM simSCGC1 + or!		\ enable clock to UART
  $0300 _PORTE portPCR8 + !		\ Alt 3 on PTE8 for UART5 TxD
  $0300 _PORTE portPCR9 + !		\ Alt 3 on PTE9 for UART5 RxD
  u5-brgval _UART5 (init-ser)
  $80 INT_UART5_RX_TX setPri		\ mid level priority
  INT_UART5_RX_TX EnInt			\ interrupt enable in NVIC
  ei
;

: seremit5	\ char --
\ *G Transmit a character on UART5.
  _UART5 SerOutQ5 (qseremit)  ;
: sertype5	\ c-addr len --
\ *G Transmit a string on UART5.
  _UART5 SerOutQ5 (qsertype)  ;
: sercr5	\ --
\ *G Issue a CR/LF pair to UART5.
  _UART5 SerOutQ5 (qsercr)  ;
: serkey?5	\ -- flag
\ *G Return true if UART5 has a character available.
  SerInpQ5 cqnotempty?  ;
: serkey5	\ -- char
\ *G Wait for a character on UART5.
  SerInpQ5 cqueue>  ;

create Console5	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART5.
  ' serkey5 ,		\ -- char ; receive char
  ' serkey?5 ,		\ -- flag ; check receive char
  ' seremit5 ,		\ -- char ; display char
  ' sertype5 ,		\ caddr len -- ; display string
  ' sercr5 ,		\ -- ; display new line

console-port 5 = [if]
  console5 constant console
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
  [ useUART3? ] [if]  init-ser3  [then]
  [ useUART4? ] [if]  init-ser4  [then]
  [ useUART5? ] [if]  init-ser5  [then]
;


\ ======
\ *> ###
\ ======

