\ serLM3s9bxi.fth - LM3S9Bx interrupt RX, polled TX, serial communication drivers

((
Copyright (c) 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20100320 MPE001 First release.
))

only forth definitions


\ ==============
\ *! serlm3s9bxp
\ *T LM3S9Bx interrupt RX, polled TX serial driver
\ ==============
\ *P This driver provides polled serial comms for LM3S9Bx CPUs
\ ** using the internal UARTs.


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBaudRate} uses the system clock rate (in Hz), the
\ ** perpheral clock divider ratio, and the required serial speed
\ ** (in BPS) and returns the 16 bit divider value for the baud
\ ** rate generator
\ *E              system-speed
\ ** divisor = -------------------
\ **            clkdiv * baudrate
\ *P The calculation assumes that the fractional diver is used,
\ ** and that *\fo{clkdiv}=16, i.e. HSE=1, bit 5 set in UnCTL.

interpreter
: GenBRD22	\ bps-rate -- divisor22
\ *G Given a baud rate and the peripheral clock divider, calculate
\ ** the 16 bit baud rate integer divisor and the 6 bit fractional
\ ** divider. We assume HSE=1.
  system-speed 3 lshift swap /		\ bottom 6 bits = fraction
;
target


\ ********************
\ *S Serial primitives
\ ********************

useUART0? [if]
variable rxd0	\ -- addr
\ *G The first byte holds the character, the second byte is
\ ** non-zero if a character is available.
[then]

useUART1? [if]
variable rxd1	\ -- addr
\ *G The first byte holds the character, the second byte is
\ ** non-zero if a character is available.
[then]

useUART2? [if]
variable rxd2	\ -- addr
\ *G The first byte holds the character, the second byte is
\ ** non-zero if a character is available.
[then]

internal

: +FaultConsole	( -- )  ;
\ *G Because this is a polled driver, *\fo{+FaultConsole} for
\ ** fault handling is a *\fo{NOOP}. See *\i{Cortex/FaultCortex.fth}
\ ** for more details.

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup UnFR + @ bit5 and 0=		\ Tx FIFO full test
  until
  UnDR + !
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)  ;

: (serkey?)	\ addr -- t/f
\ *G Return true if the given UART has a character avilable to read.
  @ $FF00 and 0<>  ;

: (serkey)	\ addr -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
[ tasking? ] [if]  pause  [then]
    dup @ $FF00 and
  until
  dup @ $FF and  swap off
;

: initUART	\ divisor22 base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r				\ save divisor
  1 over UnCTL + bic!			\ UARTEN=0
  r@ 6 rshift $FFFF and over UnIBRD + !	\ integer baud rate
  r> $3F and over UnFBRD + !		\ fractional baud rate,
  $60 over UnLCRH + !			\ 8N1, non FIFOs
  $12 over UnIFLS + !			\ 1/2 FIFO level
  $50 over UnIM + !			\ RX timeout, RxFIFO
  0 over UnDMACTL + !			\ no DMA
  $0320 $CFFF 2 pick UnCTL + setMask	\ set RXE, TXE, HSE
  1 swap UnCTL + or!			\ UARTEN=1
;

external


\ ********
\ *S UART0
\ ********

useUART0? [if]
console0-speed GenBRD22 equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

: Rx0handler	\ --
\ *G Action of the UART receive interrupt handler.
  _UART0 UnDR + @ $FF and $FF00 or rxd0 !
;
' Rx0handler U0_vec# EXC: U0entry

: init-ser0	\ --
\ *G Initialise serial port 0, using U0Rx=PA0, U0Tx=PA1
  _SysCtl
  bit0 over sysRCGC1 + or!		\ enable clock and power to UART0
  bit0 swap sysRCGC2 + or!		\ enable clock and power to GPIOA
  noop noop noop noop			\ must wait a bit
  _GPIOA
  $03 over gpioAFSEL + or!		\ pins 0/1 are peripheral
  $03 over gpioDR2R + or!		\ pins 0/1 are 2mA drive
  $03 over gpioSLR + bic!		\ pins 0/1 have no slew control
  $03 over gpioDEN + or!		\ enable pins
  $03 over gpioODR + bic!		\ no open drain
  $03 over gpioPUR + bic!		\ no pull ups
  $03 over gpioPDR + bic!		\ no pull downs
  $11 $FF  rot gpioPCTL + setMask	\ peripheral 1 for PA0/1
  u0-divisor _UART0 initUART
  U0_vec# EnInt  ei			\ interrupt enable in NVIC
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  rxd0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  rxd0 (serkey)  ;

create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ********
\ *S UART1
\ ********

useUART1? [if]

console1-speed GenBRD22 equ u1-divisor	\ -- n
\ *G UART1 baud rate divisor.

: Rx1handler	\ --
\ *G Action of the UART receive interrupt handler.
  _UART1 UnDR + @ $FF and $FF00 or rxd1 !
;
' Rx0handler U1_vec# EXC: U1entry

: init-ser1	\ --
\ *G Initialise serial port 1, using U1Rx=PD2, U1Tx=PD3
  _SysCtl
  bit1 over sysRCGC1 + or!		\ enable clock and power to UART0
  bit3 swap sysRCGC2 + or!		\ enable clock and power to GPIOD
  noop noop noop noop			\ must wait a bit
  _GPIOD
  $0C over gpioAFSEL + or!		\ pins 2/3 are peripheral
  $0C over gpioDR2R + or!		\ pins 2/3 are 2mA drive
  $0C over gpioSLR + bic!		\ pins 2/3 have no slew control
  $1100 $FF00 rot gpioPCTL + setMask	\ peripheral 1 for PA2/3
  u1-divisor _UART1 initUART
  U1_vec# EnInt  ei			\ interrupt enable in NVIC
;

: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 (seremit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 (sertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 (sercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  rxd1 (serkey?)  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  rxd1 (serkey)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ ********
\ *S UART2
\ ********

useUART2? [if]

console2-speed GenBRD22 equ u2-divisor	\ -- n
\ *G UART2 baud rate divisor.

: Rx2handler	\ --
\ *G Action of the UART receive interrupt handler.
  _UART2 UnDR + @ $FF and $FF00 or rxd2 !
;
' Rx2handler U2_vec# EXC: U2entry

: init-ser2	\ --
\ *G Initialise serial port 2, using U2Rx=PG0, U2Tx=PG1
  _SysCtl
  bit2 over sysRCGC1 + or!		\ enable clock and power to UART2
  bit6 swap sysRCGC2 + or!		\ enable clock and power to GPIOG
  noop noop noop noop			\ must wait a bit
  _GPIOG
  $03 over gpioAFSEL + or!		\ pins 0/1 are peripheral
  $03 over gpioDR2R + or!		\ pins 0/1 are 2mA drive
  $03 over gpioSLR + bic!		\ pins 0/1 have no slew control
  $11 $FF rot gpioPCTL + setMask	\ peripheral 1 for PG0/1
  u2-divisor _UART2 initUART
  U2_vec# EnInt  ei			\ interrupt enable in NVIC
;

: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 (seremit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 (sertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 (sercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  _UART2 (serkey?)  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  _UART2 (serkey)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
;


\ ======
\ *> ###
\ ======

decimal
