\ SerLPC11xxqi.fth - LPC11xx queued interrupt RX and TX UART driver

((
Copyright (c) 2008, 2010, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20110628 MPE002 LPC11xx conversion.
20100125 MPE001 First release.
))

only forth definitions


\ =============
\ *! serlpc11xxqi
\ *T LPC11xx serial driver
\ =============
\ *P This driver provides queued interrupt-driven serial comms
\ ** for LPC11xx CPUs using the internal UART(s). Because there
\ ** are alternate pin assignments for some UARTs, you *\b{must}
\ ** check that the supplied code pin setup is correct for your
\ ** board.


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBaudRate} uses the system clock rate (in Hz), the
\ ** perpheral clock divider ratio, and the required serial speed
\ ** (in BPS) and returns the 16 bit divider value for the baud
\ ** rate generator
\ *E              system-speed
\ ** divisor = -------------------
\ **           16 * pcd * baudrate
\ *P The calculation assumes that the fractional diver is disabled.

interpreter
: GenBaudRate	\ bps-rate pcd -- divisor16
\ *G Given a baud rate and the peripheral clock divider
\ ** calculate the 16 bit baud rate divisor.
  system-speed swap / swap / #16 /mod
  swap 8 >=
  if  1+  endif
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal
: >RxQ		\ char queue --
\ *G Put character from UART into queue.
\ ** INTERNAL.
  dup (cqFull?)				\ if queue full
  if  2drop  exit  endif		\  discard character
  (>cqueue)				\  queue character
;

: FIFO>RxQ      \ base queue --
\ *G Given a UART base address and a character queue,
\ ** empty the UART Rx FIFO into the queue.
\ ** INTERNAL.
  begin
    over UnLSR + @ $01 and
   while
    over UnRBR + @ $FF and over >RxQ
  repeat
  2drop
;

: TxQ>FIFO	\ base queue
\ *G Given a UART base address and a character queue,
\ ** put transmit characters into the UART Tx FIFO.
  dup cqEmpty? if
    drop
    0 swap UnSCR + !			\ mark empty for EMIT routine
    exit
  endif
  (cqueue>) swap UnTHR + !		\ write character
;

0 value FaultConsole?	\ -- flag
\ *G When set non-zero, transmit is in polled mode. This is
\ ** required when fault exceptions provide debug information.

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
  -1 to FaultConsole?
  [ tasking? ] [if]  single  [then]
;

: (qseremit)	\ char base queue --
\ *G Send a character on the UART whose base address and queue are given.
  FaultConsole? 0= if
    begin				\ wait for non-full queue
      dup cqfull?
     while
      pause
    repeat
    [i
    over UnSCR + @ $FF and if		\ if TX in progress
      nip (>cqueue)  i]			\ just stash the character
      exit
    then
    drop				\ -- char base
    tuck UnTHR + !			\ write character
    $FF swap UnSCR + !			\ mark TX int started
    i]
    exit
  endif
  drop					\ discard queue
  $00 over UnIER + !			\ disable TX and RX interrupts
  0 over UnSCR + !			\ SCR used as TX int enabled flag
  begin					\ poll LSR
    dup UnLSR + c@ %00100000 and
  until
  UnTHR + c!				\ send
;

: (qsertype)	\ caddr len base queue --
\ *G Send a string on the UART whose base address and queue are given.
  2swap bounds				\ -- base queue
  ?do  2dup i c@ -rot (qseremit)  loop
  2drop
;

: (qsercr)	\ base queue --
\ *G Send a CR/LF on the UART whose base address and queue are given.
  2dup $0D -rot (qseremit)  $0A -rot (qseremit)
;

: initUART	\ divisor base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r
  0 over UnACR + !			\ turn off autobaud
  $10 over UnFDR + !			\ turn off fractional baud rate
  $81 over UnFCR + !			\ enable FIFOs - 8 char delay
  $83 over UnLCR + !			\ 8 data, 1 stop, no parity, enable DLAB
  r@ 8 rshift $FF and over UnDLM + !	\ set divisor
  r> $FF and over UnDLL + !
  $03 over UnLCR + !			\ clear DLAB
  $03 over UnIER + !			\ Enable TX and RX interrupts
  0 over UnSCR + !			\ SCR used as TX int enabled flag
  drop
;

external


\ ********
\ *S UART0
\ ********
\ *P This code only uses and configures RXD (PIO1.6) and
\ ** TXD (PIO1.7).

\ CTSi PIO0_7
\ RTSo PIO1_5
\ RXDi PIO1_6
\ TXDo PIO1_7
\ DTRo PIO2_0 PIO3_0
\ DSRi PIO2_1 PIO3_1
\ DCDi PIO2_2 PIO3_2
\ RIi  PIO2_3 PIO3_3

useUART0? [if]

1 equ pcdU0	\ -- n
\ *G Peripheral clock division ratio for UART0.
console0-speed pcdU0 GenBaudRate equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

#128 cqueue: SerInpQ0	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ0	\ -- addr
\ *G Transmitter circular byte queue.

: ser0-isrh	\ --
\ *G UART0 high level ISR.
\ ** INTERNAL.
  case  _UART0 dup UnIIR + @ 7 and
    0 of  UnMSR + @ drop  endof		\ No MSR for UART0
    2 of  SerOutQ0 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ0 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser0-isrh U0_vec# EXC: ser0-isr
\ *G UART0 interrupt service routine entry point.

: init-ser0	\ --
\ *G Initialise serial port 0
  di
  SerInpQ0 init-cqueue  SerOutQ0 init-cqueue	\ initialise character queues

  $1:0040 _SYSCON SYSAHBCLKCTRL + or!	\ Enable IOCON and GPIO clocks
  _IOCONFIG IOCON_PIO1_6 + dup @	\ RXD input, no pull up/down
    $03 invert and
    $01 or
  swap !
  _IOCONFIG IOCON_PIO1_7 + dup @	\ TXD output, no pull up/down
    $03 invert and
    $01 or
  swap !
  bit12 _SYSCON SYSAHBCLKCTRL + or!	\ Enable UART clock
  pcdU0 _SYSCON UARTCLKDIV + !		\ set clock divider

  u0-divisor _UART0 initUART
  U0_vec# EnInt				\ interrupt enable in NVIC
  ei
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 SerOutQ0 (qserEmit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 SerOutQ0 (qsertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 SerOutQ0 (qsercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  SerInpQ0 cqnotempty?  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  SerInpQ0 cqueue>  ;

create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line

console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]			\ use UART0?


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
[ useUART0? ] [if]  init-ser0  [then]
;


\ *********
\ Test code
\ *********

((
: txu1		\ --
  500 0 do  bl seremit1  loop  ;

: txu2		\ --
  500 0 do  bl seremit2  loop  ;

: txu3		\ --
  500 0 do  bl seremit3  loop  ;
))

\ *******
\ Perhaps
\ *******

((
: SetBaud	\ bps-rate uart --
 \ *G Takes the required serial comms rate (in BPS) and the UART
 \ ** base and sets the baud rate.
  swap system-speed pcdU0 / swap / #16 /mod
  swap 8 >=
  if  1+  endif
  dup 8 rshift $FF and  swap $FF and swap	\ -- uart divlo divhi
  [i
    rot						\ -- divlo divhi uart
    $83 over UnLCR + !				\ enable DLAB access
    tuck UnDLM + !				\ -- divlo uart ; high divisor ; MPE003
    tuck UnDLL + !				\ -- uart ; MPE003
    $03 swap UnLCR + !
  i]
;
))


\ ======
\ *> ###
\ ======

decimal
