\ serLPC17xxp.fth - LPC23xx polled RX, polled TX, serial communication drivers

((
Copyright (c) 2008, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20100122 MPE001 First release.
))

only forth definitions


\ ==============
\ *! serlpc17xxp
\ *T LPC17xx polled serial driver
\ ==============
\ *P This driver provides polled serial comms for LPC17xx CPUs
\ ** using the internal UARTs. Because there are alternate pin
\ ** assignments for some UARTs, you *\b{must} check that the
\ ** supplied code pin setup is correct for your board.


\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUART0? [if]
1 equ useUART0?	\ -- n
\ *G Compile UART0 code if non-zero.
[then]

[undefined] useUART1? [if]
0 equ useUART1?	\ -- n
\ *G Compile UART1 code if non-zero.
[then]

[undefined] useUART2? [if]
0 equ useUART2?	\ -- n
\ *G Compile UART2 code if non-zero.
[then]

[undefined] useUART3? [if]
0 equ useUART3?	\ -- n
\ *G Compile UART3 code if non-zero.
[then]


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBaudRate} uses the system clock rate (in Hz), the
\ ** perpheral clock divider ratio, and the required serial speed
\ ** (in BPS) and returns the 16 bit divider value for the baud
\ ** rate generator
\ *E              system-speed
\ ** divisor = -------------------
\ **           16 * pcd * baudrate
\ *P The calculation assumes that the fractional diver is disabled.

interpreter
: GenBaudRate	\ bps-rate pcd -- divisor16
\ *G Given a baud rate and the peripheral clock divider
\ ** (1, 2, 4, 8), calculate the 16 bit baud rate divisor.
  system-speed swap / swap / #16 /mod
  swap 8 >=
  if  1+  endif
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal

: +FaultConsole	( -- )  ;
\ *G Because this is a polled driver, *\fo{+FaultConsole} for
\ ** fault handling is a *\fo{NOOP}. See *\i{Cortex/FaultCortex.fth}
\ ** for more details.

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup UnLSR + c@ %00100000 and
  until
  UnTHR + c!
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)
;

: (serkey?)	\ base -- t/f
\ *G Return true if the given UART has a character avilable to read.
  UnLSR + c@ %00000001 and 0<>
;

: (serkey)	\ base -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
[ tasking? ] [if]  pause  [then]
    dup UnLSR + c@ %00000001 and
  until
  UnRBR + c@
;

: initUART	\ divisor base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r
  0 over UnACR + !			\ turn off autobaud
  $10 over UnFDR + !			\ turn off fractional baud rate
  1 over UnFCR + !			\ enable FIFOs
  $83 over UnLCR + !			\ 8 data, 1 stop, no parity, enable DLAB
  r@ 8 rshift $FF and over UnDLM + !	\ set divisor
  r> $FF and over UnDLL + !
  $03 over UnLCR + !			\ clear DLAB
  drop
;

external


\ ********
\ *S UART0
\ ********

useUART0? [if]
1 equ pcdU0	\ -- n
\ *G Peripheral clock division ratio for UART0.
pcdu0 PCLKdiv 6 lshift equ /pcdu0	\ -- mask
\ *G Bits for UART0 peripheral clock divider.
%11 6 lshift equ pcdU0mask		\ -- mask
\ *G Bit mask for clock bits PCLKSELx register.
console0-speed pcdU0 GenBaudRate equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

PCLKSELbug? [if]			\ set PCLKSELx now
  /pcdU0 pcdU0mask ^PCLKSEL0val setMask	\ set peripheral clock speed
[then]

: init-ser0	\ --
\ *G Initialise serial port 0
  ~pcUART0 _PCONP or!			\ enable clock and power
[ PCLKSELbug? 0= ] [if]			\ set PCLKSELx now
  /pcdU0 pcdU0mask _PCLKSEL0 setMask	\ set peripheral clock speed
[then]
  $0050 $00F0 _PINSEL0 setMask		\ PINSEL0 bits 7:4
  $00A0 $00F0 _PINMODE0 setMask		\ PINMODE0 bits 7:4
  u0-divisor _UART0 initUART
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  _UART0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  _UART0 (serkey)  ;
create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ********
\ *S UART1
\ ********

useUART1? [if]

1 equ pcdU1	\ -- n
\ *G Peripheral clock division ratio for UART.
pcdu1 PCLKdiv 8 lshift equ /pcdu1	\ -- mask
\ *G Bit mask for UART peripheral clock divider.
%11 8 lshift equ pcdU1mask		\ -- mask
\ *G Bit mask for clock bits PCLKSELx register.
console1-speed pcdU1 GenBaudRate equ u1-divisor	\ -- n
\ *G UART baud rate divisor.

PCLKSELbug? [if]			\ set PCLKSELx now
  /pcdU1 pcdU1mask ^PCLKSEL0val setMask	\ set peripheral clock speed
[then]

: init-ser1	\ --
\ *G Initialise serial port 1
  ~pcUART1 _PCONP or!				\ enable clock & pwr
[ PCLKSELbug? 0= ] [if]				\ set PCLKSELx now
  /pcdU1 pcdU1mask _PCLKSEL0 setMask		\ set peripheral clock speed
[then]
  $000A $000F _PINSEL4 setMask			\ TXD1, mode 2
  u1-divisor _UART1 initUART
;
: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 (seremit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 (sertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 (sercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  _UART1 (serkey?)  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  _UART1 (serkey)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ ********
\ *S UART2
\ ********

useUART2? [if]

1 equ pcdU2	\ -- n
\ *G Peripheral clock division ratio for UART.
pcdu2 PCLKdiv #16 lshift equ /pcdu2	\ -- mask
\ *G Bit mask for UART peripheral clock divider.
%11 #16 lshift equ pcdU2mask		\ -- mask
\ *G Bit mask for clock bits PCLKSELx register.
console2-speed pcdU2 GenBaudRate equ u2-divisor	\ -- n
\ *G UART baud rate divisor.

PCLKSELbug? [if]			\ set PCLKSELx now
  /pcdU2 pcdU2mask ^PCLKSEL1val setMask	\ set peripheral clock speed
[then]

: init-ser2	\ --
\ *G Initialise serial port 0
  ~pcUART2 _PCONP or!			\ enable clock & pwr
[ PCLKSELbug? 0= ] [if]			\ set PCLKSELx now
  /pcdU2 pcdU2mask _PCLKSEL1 setMask	\ set peripheral clock speed
[then]
  $0050:0000 $00F0:0000 _PINSEL0 setMask	\ PINSEL0 bits 23:20
  u2-divisor _UART2 initUART
;
: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 (seremit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 (sertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 (sercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  _UART2 (serkey?)  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  _UART2 (serkey)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ ********
\ *S UART3
\ ********

useUART3? [if]
1 equ pcdU3	\ -- n
\ *G Peripheral clock division ratio for UART.
pcdu3 PCLKdiv #18 lshift equ /pcdu3	\ -- mask
\ *G Bit mask for UART peripheral clock divider.
%11 #18 lshift equ pcdU3mask		\ -- mask
\ *G Bit mask for clock bits PCLKSELx register.
console3-speed pcdU3 GenBaudRate equ u3-divisor	\ -- n
\ *G UART baud rate divisor.

PCLKSELbug? [if]			\ set PCLKSELx now
  /pcdU3 pcdU3mask ^PCLKSEL1val setMask	\ set peripheral clock speed
[then]

: init-ser3	\ --
\ *G Initialise serial port 3.
  ~pcUART3 _PCONP or!				\ enable clock & pwr
[ PCLKSELbug? 0= ] [if]			\ set PCLKSELx now
  /pcdU3 pcdU3mask _PCLKSEL1 setMask	\ set peripheral clock speed
[then]
  $003C:0000 $003C:0000 _PINSEL1 setMask	\ PINSEL1 bits 21:18, mode 3
  u3-divisor _UART3 initUART
;
: seremit3	\ char --
\ *G Transmit a character on UART3.
  _UART3 (seremit)  ;
: sertype3	\ c-addr len --
\ *G Transmit a string on UART3.
  _UART3 (sertype)  ;
: sercr3	\ --
\ *G Issue a CR/LF pair to UART3.
  _UART3 (sercr)  ;
: serkey?3	\ -- flag
\ *G Return true if UART3 has a character available.
  _UART3 (serkey?)  ;
: serkey3	\ -- char
\ *G Wait for a character on UART3.
  _UART3 (serkey)  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART3.
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line

console-port 3 = [if]
  console3 constant console
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
  [ useUART3? ] [if]  init-ser3  [then]
;


\ ======
\ *> ###
\ ======

decimal
