\ serLPC40xxp.fth - LPC40xx polled RX, polled TX, serial communication drivers

((
Copyright (c) 2008, 2010, 2014, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20150624 SFP003 Updated for LPC40xx with enhanced pin selection.
20140107 BJC002 Converted IO configuration for LPC1788 family.
		LPC1788 does not have PCLKSELbug, so this fix is removed.
20100122 MPE001 First release.
))

only forth definitions


\ ==============
\ *! serlpc40xxp
\ *T LPC40xx polled serial driver
\ ==============
\ *P This driver provides polled serial comms for LPC40xx family CPUs
\ ** using the internal UARTs. Because there are alternate pin
\ ** assignments for some UARTs, you *\b{must} check that the
\ ** supplied code pin setup is correct for your board.

\ *P Code for UARTx (x=0..4) is compiled if the equate *\fo{useUARTx}
\ ** is non-zero. Each UART has a default pin configuration provided
\ ** by the word *\fo{SELIO-SERx}. To change the pin usage, provide
\ ** your own version of *\fo{SELIO-SERx} before the UART code is
\ ** compiled.


\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUART0? [if]
1 equ useUART0?	\ -- n
\ *G Compile UART0 code if non-zero.
[then]

[undefined] useUART1? [if]
0 equ useUART1?	\ -- n
\ *G Compile UART1 code if non-zero.
[then]

[undefined] useUART2? [if]
0 equ useUART2?	\ -- n
\ *G Compile UART2 code if non-zero.
[then]

[undefined] useUART3? [if]
0 equ useUART3?	\ -- n
\ *G Compile UART3 code if non-zero.
[then]

[undefined] useUART4? [if]
0 equ useUART4?	\ -- n
\ *G Compile UART4 code if non-zero.
[then]


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBaudRate} uses the peripheral clock rate (in Hz)
\ ** and the required serial speed (in BPS) to return the 16 bit
\ ** divider value for the baud rate generator
\ *E              periph-speed
\ ** divisor = -------------------
\ **           16 * baudrate
\ *P The calculation assumes that the fractional diver is disabled.

interpreter
: GenBaudRate	\ bps-rate -- divisor16
\ *G Given a baud rate, calculate the 16 bit baud rate divisor.
  periph-speed swap / #16 /mod
  swap 8 >=
  if  1+  endif
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal

: +FaultConsole	( -- )  ;
\ *G Because this is a polled driver, *\fo{+FaultConsole} for
\ ** fault handling is a *\fo{NOOP}. See *\i{Cortex/FaultCortex.fth}
\ ** for more details.

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup UnLSR + c@ %00100000 and
  until
  UnTHR + c!
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)
;

: (serkey?)	\ base -- t/f
\ *G Return true if the given UART has a character avilable to read.
  UnLSR + c@ %00000001 and 0<>
;

: (serkey)	\ base -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
[ tasking? ] [if]  pause  [then]
    dup UnLSR + c@ %00000001 and
  until
  UnRBR + c@
;

: initUART	\ divisor base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r
  0 over UnACR + !			\ turn off autobaud
  $10 over UnFDR + !			\ turn off fractional baud rate
  1 over UnFCR + !			\ enable FIFOs
  $83 over UnLCR + !			\ 8 data, 1 stop, no parity, enable DLAB
  r@ 8 rshift $FF and over UnDLM + !	\ set divisor
  r> $FF and over UnDLL + !
  $03 over UnLCR + !			\ clear DLAB
  drop

;

external


\ ********
\ *S UART0
\ ********

useUART0? [if]
console0-speed GenBaudRate equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

[undefined] selio-ser0 [if]
: selio-ser0	\ -- ; select I/O for UART
\ *G Initialise UART0 for operation on pins P0.2 (TX) and P0.3
\ ** (Rx).
\ ** All the words *\fo{SELIO-SERx} in this file are defaults used
\ ** if you have not provided your own.
  ~pcUART0 _SYSCON
    2dup PCONP + or!			\ enable clock and power
    RSTCON0 + bic!			\ clear reset
\ IO Pins for UART0 - modify to suit your hardware
  _IOCON				\ base
  AF1 over IOCON_P0_02 + !		\ P0.2 is U0_TxD (type D) ] compatible with boot loader
  AF1 over IOCON_P0_03 + !		\ P0.3 is U0_RxD (type D) ]
\ IO Pins for UART0 hardware handshake - optional, modify as required
\ -- if required --
  drop					\ .. base
;
[then]

\ : dl    10000000 0 do loop  ;

: init-ser0	\ --
\ *G Initialise serial port 0.
  selio-ser0  u0-divisor _UART0 initUART
\  begin  dl  [char] A seremit0  dl  [char] Z seremit0  again
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  _UART0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  _UART0 (serkey)  ;
create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ********
\ *S UART1
\ ********

useUART1? [if]

console1-speed GenBaudRate equ u1-divisor	\ -- n
\ *G UART baud rate divisor.

[undefined] selio-ser1 [if]
: selio-ser1	\ -- ; select I/O for UART
\ *G Initialise UART0 for operation on pins P3.16 (TX) and P3.17
\ ** (Rx).
  ~pcUART1 _PCONP or!				\ enable clock & pwr
\ IO Pins for UART1 - modify to suit your hardware
  _IOCON		\ base...
  IO_PERIPH_FUNC3 over IOCON_P3_16 + !		\ P3.16 is U1_TxD (type D)
  IO_PERIPH_FUNC3 over IOCON_P3_17 + !		\ P3.17 is U1_RxD (type D)
\  IO Pins for UART1 hardware handshake - optional, modify as required
  IO_PERIPH_FUNC3 over IOCON_P3_18 + !		\ P3.18 is U1_CTS (type D)
  IO_PERIPH_FUNC3 over IOCON_P3_20 + !		\ P3.20 is U1_DSR (type D)
  IO_PERIPH_FUNC3 over IOCON_P3_21 + !		\ P3.21 is U1_DTR (type D)
  drop			\ .. base
;
[then]

: init-ser1	\ --
\ *G Initialise serial port 1
  selio-ser1  u1-divisor _UART1 initUART
;

: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 (seremit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 (sertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 (sercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  _UART1 (serkey?)  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  _UART1 (serkey)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ ********
\ *S UART2
\ ********

useUART2? [if]

console2-speed GenBaudRate equ u2-divisor	\ -- n
\ *G UART baud rate divisor.

[undefined] selio-ser2 [if]
: selio-ser2	\ -- ; select I/O for UART
\ *G Initialise UART2 for operation on pins P0.10 (TX) and P0.11
\ ** (Rx).
\ ** All the words *\fo{SELIO-SERx} in this file are defaults used
\ ** if you have not provided your own.
  ~pcUART2 _PCONP or!			\ enable clock & pwr
\ IO Pins for UART2 - modify to suit your hardware
  _IOCON		\ base...
  IO_PERIPH_FUNC1 over IOCON_P0_10 + !		\ P0.10 is U2_TxD (type D)
  IO_PERIPH_FUNC1 over IOCON_P0_11 + !		\ P0.11 is U2_RxD (type D)
\ IO Pins for UART2 hardware handshake - optional, modify as required
\ -- if required --
  drop			\ .. base
;
[then]

: init-ser2	\ --
\ *G Initialise serial port 2
  selio-ser2  u2-divisor _UART2 initUART
;

: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 (seremit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 (sertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 (sercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  _UART2 (serkey?)  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  _UART2 (serkey)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ ********
\ *S UART3
\ ********

useUART3? [if]

console3-speed GenBaudRate equ u3-divisor	\ -- n
\ *G UART baud rate divisor.

[undefined] selio-ser3 [if]
: selio-ser3	\ -- ; select I/O for UART
\ *G Initialise UART3 for operation on pins P0.2 (TX) and P0.3
\ ** (Rx).
\ ** All the words *\fo{SELIO-SERx} in this file are defaults used
\ ** if you have not provided your own.
  ~pcUART3 _PCONP or!				\ enable clock & pwr
\ IO Pins for UART3 - modify to suit your hardware
  _IOCON			\ base ...
  IO_PERIPH_FUNC2 over IOCON_P0_2 + !		\ P0.2 is U3_TxD (type D)
  IO_PERIPH_FUNC2 over IOCON_P0_3 + !		\ P0.3 is U3_RxD (type D)
\ IO Pins for UART3 hardware handshake - optional, modify as required
\ -- if required --
  drop				\ ... base
;
[then]

: init-ser3	\ --
\ *G Initialise serial port 3.
  selio-ser3  u3-divisor _UART3 initUART
;

: seremit3	\ char --
\ *G Transmit a character on UART3.
  _UART3 (seremit)  ;
: sertype3	\ c-addr len --
\ *G Transmit a string on UART3.
  _UART3 (sertype)  ;
: sercr3	\ --
\ *G Issue a CR/LF pair to UART3.
  _UART3 (sercr)  ;
: serkey?3	\ -- flag
\ *G Return true if UART3 has a character available.
  _UART3 (serkey?)  ;
: serkey3	\ -- char
\ *G Wait for a character on UART3.
  _UART3 (serkey)  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART3.
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line

console-port 3 = [if]
  console3 constant console
[then]

[then]


\ ********
\ *S UART4
\ ********

useUART4? [if]

console4-speed GenBaudRate equ u4-divisor	\ -- n
\ *G UART baud rate divisor.

[undefined] selio-ser4 [if]
: selio-ser4	\ -- ; select I/O for UART
\ *G Initialise UART4 for operation on pins P5.4 (TX) and P5.3
\ ** (Rx).
\ ** All the words *\fo{SELIO-SERx} in this file are defaults
\ ** used if you have not provided your own.
  ~pcUART4 _PCONP or!			\ enable clock & pwr
\ IO Pins for UART4 - modify to suit your hardware
  _IOCON			\ base ...
  IO_PERIPH_FUNC4 over IOCON_P5_04 + !	\ P5.4 is U3_TxD (type D)
  IO_PERIPH_FUNC4 over IOCON_P5_03 + !	\ P5.3 is U3_RxD (type I)
\ IO Pins for UART4 hardware handshake - optional, modify as required
\ -- if required --
  drop				\ ... base
;
[then]

: init-ser4	\ --
\ *G Initialise serial port 4.
  selio-ser4  u4-divisor _UART4 initUART
;

: seremit4	\ char --
\ *G Transmit a character on UART4.
  _UART4 (seremit)  ;
: sertype4	\ c-addr len --
\ *G Transmit a string on UART4.
  _UART4 (sertype)  ;
: sercr4	\ --
\ *G Issue a CR/LF pair to UART4.
  _UART4 (sercr)  ;
: serkey?4	\ -- flag
\ *G Return true if UART4 has a character available.
  _UART4 (serkey?)  ;
: serkey4	\ -- char
\ *G Wait for a character on UART4.
  _UART4 (serkey)  ;

create Console4	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART4.
  ' serkey4 ,		\ -- char ; receive char
  ' serkey?4 ,		\ -- flag ; check receive char
  ' seremit4 ,		\ -- char ; display char
  ' sertype4 ,		\ caddr len -- ; display string
  ' sercr4 ,		\ -- ; display new line

console-port 4 = [if]
  console4 constant console
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
  [ useUART3? ] [if]  init-ser3  [then]
  [ useUART4? ] [if]  init-ser4  [then]
;


\ ======
\ *> ###
\ ======

decimal
