\ SerLPC40xxqi.fth - LPC40xx queued interrupt RX and TX UART driver

((
Copyright (c) 2008, 2010, 2014, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20150624 SFP003 Updated for LPC40xx with enhanced pin selection.
20140107 BJC002 Converted IO configuration for LPC1788 family.
		LPC1788 does not have PCLKSELbug, so this fix is removed.
20100122 MPE001 First release.
))

only forth definitions


\ =============
\ *! serlpc40xxqi
\ *T LPC40xx serial driver
\ =============
\ *P This driver provides queued interrupt-driven serial comms for
\ ** LPC40xx family CPUs using the internal UARTs. Because there are
\ ** alternate pin assignments for some UARTs, you *\b{must} check that
\ ** the supplied code pin setup is correct for your board.

\ *P Code for UARTx (x=0..4) is compiled if the equate *\fo{useUARTx}
\ ** is non-zero. Each UART has a default pin configuration provided
\ ** by the word *\fo{SELIO-SERx}. To change the pin usage, provide
\ ** your own version of *\fo{SELIO-SERx} before the UART code is
\ ** compiled.


\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUART0? [if]
1 equ useUART0?	\ -- n
\ *G Compile UART0 code if non-zero.
[then]

[undefined] useUART1? [if]
0 equ useUART1?	\ -- n
\ *G Compile UART1 code if non-zero.
[then]

[undefined] useUART2? [if]
0 equ useUART2?	\ -- n
\ *G Compile UART2 code if non-zero.
[then]

[undefined] useUART3? [if]
0 equ useUART3?	\ -- n
\ *G Compile UART3 code if non-zero.
[then]

[undefined] useUART4? [if]
0 equ useUART4?	\ -- n
\ *G Compile UART4 code if non-zero.
[then]


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBaudRate} uses the peripheral clock rate (in Hz)
\ ** and the required serial speed (in BPS) to return the 16 bit
\ ** divider value for the baud rate generator
\ *E              periph-speed
\ ** divisor = -------------------
\ **           16 * baudrate
\ *P The calculation assumes that the fractional diver is disabled.

interpreter
: GenBaudRate	\ bps-rate -- divisor16
\ *G Given a baud rate, calculate the 16 bit baud rate divisor.
  periph-speed swap / #16 /mod
  swap 8 >=
  if  1+  endif
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal
: >RxQ		\ char queue --
\ *G Put character from UART into queue.
\ ** INTERNAL.
  dup (cqFull?)				\ if queue full
  if  2drop  exit  endif		\  discard character
  (>cqueue)				\  queue character
;

: FIFO>RxQ      \ base queue --
\ *G Given a UART base address and a character queue,
\ ** empty the UART Rx FIFO into the queue.
\ ** INTERNAL.
  begin
    over UnLSR + @ $01 and
   while
    over UnRBR + @ $FF and over >RxQ
  repeat
  2drop
;

: TxQ>FIFO	\ base queue
\ *G Given a UART base address and a character queue,
\ ** put transmit characters into the UART Tx FIFO.
  dup cqEmpty? if
    drop
    0 swap UnSCR + !			\ mark empty for EMIT routine
    exit
  endif
  (cqueue>) swap UnTHR + !		\ write character
;

0 value FaultConsole?	\ -- flag
\ *G When set non-zero, transmit is in polled mode. This is
\ ** required when fault exceptions provide debug information.

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
  -1 to FaultConsole?
  [ tasking? ] [if]  single  [then]
;

: (qseremit)	\ char base queue --
\ *G Send a character on the UART whose base address and queue are given.
  FaultConsole? 0= if
    begin				\ wait for non-full queue
      dup cqfull?
     while
      pause
    repeat
    [i
    over UnSCR + @ $FF and if		\ if TX in progress
      nip (>cqueue)  i]			\ just stash the character
      exit
    then
    drop				\ -- char base
    tuck UnTHR + !			\ write character
    $FF swap UnSCR + !			\ mark TX int started
    i]
    exit
  endif
  drop					\ discard queue
  $00 over UnIER + !			\ disable TX and RX interrupts
  0 over UnSCR + !			\ SCR used as TX int enabled flag
  begin					\ poll LSR
    dup UnLSR + c@ %00100000 and
  until
  UnTHR + c!				\ send
;

: (qsertype)	\ caddr len base queue --
\ *G Send a string on the UART whose base address and queue are given.
  2swap bounds				\ -- base queue
  ?do  2dup i c@ -rot (qseremit)  loop
  2drop
;

: (qsercr)	\ base queue --
\ *G Send a CR/LF on the UART whose base address and queue are given.
  2dup $0D -rot (qseremit)  $0A -rot (qseremit)
;

: initUART	\ divisor base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r
  0 over UnACR + !			\ turn off autobaud
  $10 over UnFDR + !			\ turn off fractional baud rate
  $81 over UnFCR + !			\ enable FIFOs - 8 char delay
  $83 over UnLCR + !			\ 8 data, 1 stop, no parity, enable DLAB
  r@ 8 rshift $FF and over UnDLM + !	\ set divisor
  r> $FF and over UnDLL + !
  $03 over UnLCR + !			\ clear DLAB
  $03 over UnIER + !			\ Enable TX and RX interrupts
  0 over UnSCR + !			\ SCR used as TX int enabled flag
  drop
;

external


\ ********
\ *S UART0
\ ********

useUART0? [if]
console0-speed GenBaudRate equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

#128 cqueue: SerInpQ0	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ0	\ -- addr
\ *G Transmitter circular byte queue.

: ser0-isrh	\ --
\ *G UART0 high level ISR.
\ ** INTERNAL.
  case  _UART0 dup UnIIR + @ 7 and
\   0 of  UnMSR + @ drop  endof		\ No MSR for UART0
    2 of  SerOutQ0 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ0 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser0-isrh U0_vec# EXC: ser0-isr
\ *G UART0 interrupt service routine entry point.

[undefined] selio-ser0 [if]
: selio-ser0	\ -- ; select I/O for UART
\ *G Initialise UART0 for operation on pins P0.2 (TX) and P0.3
\ ** (Rx).
\ ** All the words *\fo{SELIO-SERx} in this file are defaults used
\ ** if you have not provided your own.
  ~pcUART0 _SYSCON
    2dup PCONP + or!			\ enable clock and power
    RSTCON0 + bic!			\ clear reset
\ IO Pins for UART0 - modify to suit your hardware
  _IOCON				\ base
  AF1 over IOCON_P0_02 + !		\ P0.2 is U0_TxD (type D) ] compatible with boot loader
  AF1 over IOCON_P0_03 + !		\ P0.3 is U0_RxD (type D) ]
\ IO Pins for UART0 hardware handshake - optional, modify as required
\ -- if required --
  drop					\ .. base
;
[then]

: init-ser0	\ --
\ *G Initialise serial port 0.
  di
  SerInpQ0 init-cqueue  SerOutQ0 init-cqueue	\ initialise character queues
  selio-ser0  u0-divisor _UART0 initUART
  U0_vec# EnInt				\ interrupt enable in NVIC
  ei
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 SerOutQ0 (qserEmit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 SerOutQ0 (qsertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 SerOutQ0 (qsercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  SerInpQ0 cqnotempty?  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  SerInpQ0 cqueue>  ;

create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line

console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ********
\ *S UART1
\ ********

useUART1? [if]

console1-speed GenBaudRate equ u1-divisor	\ -- n
\ *G UART baud rate divisor.

#128 cqueue: SerInpQ1	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ1	\ -- addr
\ *G Transmitter circular byte queue.

: ser1-isrh	\ --
\ *G UART1 high level ISR.
\ ** INTERNAL.
  case  _UART1 dup UnIIR + @ 7 and
    0 of  U1MSR + @ drop  endof		\ Modem Status interrupt
    2 of  SerOutQ1 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ1 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser1-isrh U1_vec# EXC: ser1-isr
\ *G UART1 interrupt service routine entry point.

[undefined] selio-ser1 [if]
: selio-ser1	\ -- ; select I/O for UART
\ *G Initialise UART0 for operation on pins P3.16 (TX) and P3.17
\ ** (Rx).
  ~pcUART1 _PCONP or!				\ enable clock & pwr
\ IO Pins for UART1 - modify to suit your hardware
  _IOCON		\ base...
  AF3 over IOCON_P3_16 + !		\ P3.16 is U1_TxD (type D)
  AF3 over IOCON_P3_17 + !		\ P3.17 is U1_RxD (type D)
\  IO Pins for UART1 hardware handshake - optional, modify as required
  AF3 over IOCON_P3_18 + !		\ P3.18 is U1_CTS (type D)
  AF3 over IOCON_P3_20 + !		\ P3.20 is U1_DSR (type D)
  AF3 over IOCON_P3_21 + !		\ P3.21 is U1_DTR (type D)
  drop			\ .. base
;
[then]

: init-ser1	\ --
\ *G Initialise serial port 1.
  di
  SerInpQ1 init-cqueue  SerOutQ1 init-cqueue	\ initialise character queues
  selio-ser1  u1-divisor _UART1 initUART
  $80 U1_vec# setPri				\ mid level priority
  U1_vec# EnInt					\ interrupt enable in NVIC
;

: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 SerOutQ1 (qserEmit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 SerOutQ1 (qsertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 SerOutQ1 (qsercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  SerInpQ1 cqnotempty?  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  SerInpQ1 cqueue>  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ ********
\ *S UART2
\ ********

useUART2? [if]

console2-speed GenBaudRate equ u2-divisor	\ -- n
\ *G UART baud rate divisor.

#128 cqueue: SerInpQ2	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ2	\ -- addr
\ *G Transmitter circular byte queue.

: ser2-isrh	\ --
\ *G UART2 high level ISR.
\ ** INTERNAL.
  case  _UART2 dup UnIIR + @ 7 and
\    0 of  U1MSR + @ drop  endof		\ Modem Status interrupt
    2 of  SerOutQ2 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ2 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser2-isrh U2_vec# EXC: ser2-isr
\ *G UART2 interrupt service routine entry point.

[undefined] selio-ser2 [if]
: selio-ser2	\ -- ; select I/O for UART
\ *G Initialise UART2 for operation on pins P0.10 (TX) and P0.11
\ ** (Rx).
\ ** All the words *\fo{SELIO-SERx} in this file are defaults used
\ ** if you have not provided your own.
  ~pcUART2 _PCONP or!			\ enable clock & pwr
\ IO Pins for UART2 - modify to suit your hardware
  _IOCON		\ base...
  AF1 over IOCON_P0_10 + !		\ P0.10 is U2_TxD (type D)
  AF1 over IOCON_P0_11 + !		\ P0.11 is U2_RxD (type D)
\ IO Pins for UART2 hardware handshake - optional, modify as required
\ -- if required --
  drop			\ .. base
;
[then]

: init-ser2	\ --
\ *G Initialise serial port 2.
  di
  SerInpQ2 init-cqueue  SerOutQ2 init-cqueue	\ initialise character queues
  selio-ser2  u2-divisor _UART2 initUART
  $80 U2_vec# setPri				\ mid level priority
  U2_vec# EnInt					\ interrupt enable in NVIC
  ei
;
: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 SerOutQ2 (qserEmit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 SerOutQ2 (qsertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 SerOutQ2 (qsercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  SerInpQ2 cqnotempty?  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  SerInpQ2 cqueue>  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]
[then]


\ ********
\ *S UART3
\ ********

useUART3? [if]

console3-speed GenBaudRate equ u3-divisor	\ -- n
\ *G UART baud rate divisor.

#128 cqueue: SerInpQ3	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ3	\ -- addr
\ *G Transmitter circular byte queue.
: ser3-isrh	\ --
\ *G UART3 high level ISR.
\ ** INTERNAL.
  case  _UART3 dup UnIIR + @ 7 and
\    0 of  U1MSR + @ drop  endof		\ Modem Status interrupt
    2 of  SerOutQ3 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ3 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser3-isrh U3_vec# EXC: ser3-isr
\ *G UART3 interrupt service routine entry point.

[undefined] selio-ser3 [if]
: selio-ser3	\ -- ; select I/O for UART
\ *G Initialise UART3 for operation on pins P0.2 (TX) and P0.3
\ ** (Rx).
\ ** All the words *\fo{SELIO-SERx} in this file are defaults used
\ ** if you have not provided your own.
  ~pcUART3 _PCONP or!				\ enable clock & pwr
\ IO Pins for UART3 - modify to suit your hardware
  _IOCON			\ base
  AF2 over IOCON_P0_02 + !		\ P0.2 is U3_TxD (type D)
  AF2 over IOCON_P0_03 + !		\ P0.3 is U3_RxD (type D)
\ IO Pins for UART3 hardware handshake - optional, modify as required
\ -- if required --
  drop			\ .. base
;
[then]

: init-ser3	\ --
\ *G Initialise serial port 3.
  di
  SerInpQ3 init-cqueue  SerOutQ3 init-cqueue	\ initialise character queues
  selio-ser3  u3-divisor _UART3 initUART
  $80 U3_vec# setPri				\ mid level priority
  U3_vec# EnInt					\ interrupt enable in NVIC
  ei
;
: seremit3	\ char --
\ *G Transmit a character on UART3.
  _UART3 SerOutQ3 (qserEmit)  ;
: sertype3	\ c-addr len --
\ *G Transmit a string on UART3.
  _UART3 SerOutQ3 (qsertype)  ;
: sercr3	\ --
\ *G Issue a CR/LF pair to UART3.
  _UART3 SerOutQ3 (qsercr)  ;
: serkey?3	\ -- flag
\ *G Return true if UART3 has a character available.
  SerInpQ3 cqnotempty?  ;
: serkey3	\ -- char
\ *G Wait for a character on UART3.
  SerInpQ3 cqueue>  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART3.
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line

console-port 3 = [if]
  console3 constant console
[then]

[then]

\ ********
\ *S UART4
\ ********

useUART4? [if]

console4-speed GenBaudRate equ u4-divisor	\ -- n
\ *G UART baud rate divisor.

#128 cqueue: SerInpQ4	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ4	\ -- addr
\ *G Transmitter circular byte queue.
: ser4-isrh	\ --
\ *G UART4 high level ISR.
\ ** INTERNAL.
  case  _UART4 dup UnIIR + @ 7 and
\    0 of  U1MSR + @ drop  endof		\ Modem Status interrupt
    2 of  SerOutQ4 TxQ>FIFO  endof	\ Tx interrupt
    4 of  SerInpQ4 FIFO>RxQ  endof	\ read characters (resets interrupt)
    6 of  UnLSR + @ drop  endof		\ Line Status interrupt
          nip
  endcase
;
' ser4-isrh U4_vec# EXC: ser4-isr
\ *G UART4 interrupt service routine entry point.

[undefined] selio-ser4 [if]
: selio-ser4	\ -- ; select I/O for UART
\ *G Initialise UART4 for operation on pins P5.4 (TX) and P5.3
\ ** (Rx).
\ ** All the words *\fo{SELIO-SERx} in this file are defaults
\ ** used if you have not provided your own.
  ~pcUART4 _PCONP or!			\ enable clock & pwr
\ IO Pins for UART4 - modify to suit your hardware
  _IOCON			\ base ...
  AF4 over IOCON_P5_04 + !		\ P5.4 is U3_TxD (type D)
  AF4 over IOCON_P5_03 + !		\ P5.3 is U3_RxD (type I)
\ IO Pins for UART4 hardware handshake - optional, modify as required
\ -- if required --
  drop				\ ... base
;
[then]

: init-ser4	\ --
\ *G Initialise serial port 4.
  di
  SerInpQ4 init-cqueue  SerOutQ4 init-cqueue	\ initialise character queues
  selio-ser4  u4-divisor _UART4 initUART
  $80 U4_vec# setPri				\ mid level priority
  U4_vec# EnInt					\ interrupt enable in NVIC
  ei
;

: seremit4	\ char --
\ *G Transmit a character on UART4.
  _UART4 SerOutQ4 (qserEmit)  ;
: sertype4	\ c-addr len --
\ *G Transmit a string on UART4.
  _UART3 SerOutQ4 (qsertype)  ;
: sercr4	\ --
\ *G Issue a CR/LF pair to UART4.
  _UART4 SerOutQ4 (qsercr)  ;
: serkey?4	\ -- flag
\ *G Return true if UART4 has a character available.
  SerInpQ4 cqnotempty?  ;
: serkey4	\ -- char
\ *G Wait for a character on UART4.
  SerInpQ4 cqueue>  ;

create Console4	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART4.
  ' serkey4 ,		\ -- char ; receive char
  ' serkey?4 ,		\ -- flag ; check receive char
  ' seremit4 ,		\ -- char ; display char
  ' sertype4 ,		\ caddr len -- ; display string
  ' sercr4 ,		\ -- ; display new line

console-port 4 = [if]
  console4 constant console
[then]

[then]

\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
  [ useUART3? ] [if]  init-ser3  [then]
  [ useUART4? ] [if]  init-ser4  [then]
;


\ *********
\ Test code
\ *********

((
: txu1		\ --
  500 0 do  bl seremit1  loop  ;

: txu2		\ --
  500 0 do  bl seremit2  loop  ;

: txu3		\ --
  500 0 do  bl seremit3  loop  ;
))

\ *******
\ Perhaps
\ *******

((
: SetBaud	\ bps-rate uart --
 \ *G Takes the required serial comms rate (in BPS) and the UART
 \ ** base and sets the baud rate.
  swap system-speed /vpb / swap / #16 /mod
  swap 8 >=
  if  1+  endif
  dup 8 rshift $FF and  swap $FF and swap	\ -- uart divlo divhi
  [i
    rot						\ -- divlo divhi uart
    $83 over UnLCR + !				\ enable DLAB access
    tuck UnDLM + !				\ -- divlo uart ; high divisor ; MPE003
    tuck UnDLL + !				\ -- uart ; MPE003
    $03 swap UnLCR + !
  i]
;
))


\ ======
\ *> ###
\ ======

decimal
