\ serNUC1xxp.fth - Nuvoton NUC1xx polled RX, polled TX, serial communication drivers

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20110705 MPE001 First release.
))

only forth definitions


\ ==============
\ *! sernuc1xxp
\ *T Nuvoton NUC1xx polled serial driver
\ ==============
\ *P This driver provides polled serial comms for NUC1xx CPUs
\ ** using the internal UARTs. Because there are alternate pin
\ ** assignments for some UARTs, you *\b{must} check that the
\ ** supplied code pin setup is correct for your board.


\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUART0? [if]
1 equ useUART0?	\ -- n
\ *G Compile UART0 code if non-zero.
[then]


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBaudRate} uses the internal high speed clock rate
\ ** (in Hz), the required serial speed (in BPS) and returns the
\ ** 16 bit divider value for the baud rate generator.
\ *E              ihso-speed
\ ** divisor = ---------------- - 2
\ **               baudrate
\ *P The calculation assumes Mode 2.

interpreter
: GenBR22Mclk	\ bps-rate -- divisor16
\ *G Given a baud rate, calculate the 16 bit baud rate divisor
\ ** assuming Mode 2 (DIV_X_EN=1, DIV_X_ONE=1)
  ihso-speed swap / 2-
  dup 3 $FFFF within? 0= if
    cr ." Can't set BRG" abort
  endif
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal

: +FaultConsole	( -- )  ;
\ *G Because this is a polled driver, *\fo{+FaultConsole} for
\ ** fault handling is a *\fo{NOOP}. See *\i{Cortex/FaultCortex.fth}
\ ** for more details.

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup Un_FSR + @ bit23 and 0=		\ TX FIFO not full
  until
  Un_THR + !
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)
;

: (serkey?)	\ base -- t/f
\ *G Return true if the given UART has a character avilable to read.
  Un_FSR + @ bit14 and 0=		\ RX FIFO not empty
;

: (serkey)	\ base -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
[ tasking? ] [if]  pause  [then]
    dup Un_FSR + @ bit14 and 0=		\ RX FIFO not empty
  until
  Un_RBR + @ $FF and
;

: initUART	\ divisor base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r
  0 over Un_IER + !			\ no interrupts
  0 over Un_FUN_SEL + !			\ UART mode
  $036 over Un_FCR + !			\ 14 byte FIFO, reset FIFOs
  $003 over UN_LCR + !			\ 8N1
\  0 over Un_MCR + !			\ no modem control lines used
  $28 over Un_TOR + !			\ $28=40=four character delay
  r> $3000:0000 or  over Un_BAUD + !	\ BRG Mode 2, baud = 22MHz/div+2
  drop
  10000 0 do loop			\ settling delay
;

: selU22Mclk	\ --
\ *G Select the internal 22MHz clock for the UARTs
  $0300:0000 _CLK CLKSEL1 + or!		\ select 22MHz osc for UARTs
;
external


\ ********
\ *S UART0
\ ********
\ *P This code only uses and configures RXD (PB.0) and TXD (PB.1).

useUART0? [if]

console0-speed GenBR22Mclk equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

: init-ser0	\ --
\ *G Initialise serial port 0. Clock source selection has already
\ ** been made, defaulting to the internal 22MHz oscillator.
  $03 _GCR GPB_MFP + or!		\ select PB0/1 to UART mode
  bit16 _CLK APBCLK + or!		\ enable UART0

  u0-divisor _UART0 initUART
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  _UART0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  _UART0 (serkey)  ;
create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  selU22Mclk				\ select 22MHz clock for UARTs
  [ useUART0? ] [if]  init-ser0  [then]
;


\ ======
\ *> ###
\ ======

decimal
