\ serSTM32F0xxp.fth - STM32F0xx polled serial driver

((
Copyright (c) 2009, 2011, 2014, 2017
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20170312 MAP004 Added OVRDIS selection to provent lockup on input
		overflow.
20140527 RDO003 RDoering - Add USART2 on PA2/3 for STM32F030 Nucleo
20120125 MPE002 Changed to use external pin selection words.
20111130 MPE001 Converted from STM32F103
))

only forth definitions
decimal

\ ===========
\ *! serstm32f0xxp
\ *T STM32F0xx polled serial driver
\ ===========
\ *P The serial driver *\i{Cortex/Drivers/serSTM32F0xxp.fth}
\ ** provides polled serial drivers for the on-chip USARTs.
\ ** The supplied driver implements USART1 on PA9/10 and
\ ** USART3 on PD8 (Tx) and PD9 (Rx).

\ *P When you copy this file for your own hardware do not forget
\ ** update the GPIO pin assignments and alternate function
\ ** selections.


\ ***********************
\ *S Baud rate generation
\ ***********************
\ *P The baud rate routines use 16 times sampling and the fractional
\ ** baud rate generator.

\ *E                pclkx
\ **  divisor = -------------
\ **            16 * baudrate
\ *P We then treat the divisor as an integer and four bit fraction,
\ ** effectively scaling it by 16 again.
\ *E             pclkx
\ **  BRRval = --------
\ **           baudrate

interpreter
: genBRRval	\ baud clock -- divisor
\ *G Generate the required BRR value.
  1 lshift				\ for rounding
  swap /				\ brr = clock / baud
  1+  1 rshift				\ rounding
;
target


\ **************
\ *S Shared code
\ **************

: (serKey?)	\ base -- flag
  usISR + @ bit5 and  ;

: (serkey)	\ base -- char
[ tasking? ] [if]
  begin
    dup usISR + @ bit5 and 0=
   while
    pause
  repeat
[else]
  begin
    dup usISR + @ bit5 and
  until
[then]
  usRDR + @ $FF and
;

: (seremit)	\ char base --
[ tasking? ] [if]
  begin
    dup usISR + @ bit7 and 0=
   while
    pause
  repeat
[else]
  begin
    dup usISR + @ bit7 and
  until
[then]
  usTDR + !
;

: (sertype)	\ caddr len base --
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
  #13 over (seremit)  #10 swap (seremit)  ;

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
[ tasking? ] [if]  single  [then]
;

: (init-ser)	\ brrval ^usart --
  $0000 over usCR1 + !			\ disable UART, 8 bits, no parity
  $0000 over usCR2 + !			\ 1 stop bit, UART mode
  $1000 over usCR3 + !			\ OVRDIS, no ints, no DMA, no LIN, no smartcard
  tuck usBRR + !			\ set baud rate
  $000C over usCR1 + !			\ 8 bit, no parity, TE, RE
  $000D over usCR1 + !			\ 8 bit, no parity, TE, RE, UE
  drop
;


\ *********
\ *S USART1
\ *********
\ *P USART1 is on the APB bus, so the baud rate is calculated
\ ** using that bus clock.

useUSART1? [if]

console1-speed APB-speed genBRRval equ /us1BRR	\ -- u
\ *G Baud rate divisor for USART1.

((
: selio-ser1	\ --
\ *G Example to perform clock and pin selection for USART1 on PA9/10.
\ ** Provide your own version, e.g. in the control file.
\ Clocks and reset
  _RCC
  $0000 $0003 2 pick rccCFGR3 + setMask	\ PCLK is UART clock
  bit17 over rccAHBen + or!		\ enable clock to GPIOA
  bit17 over rccAHBrst + bic!		\ take it out of reset
  bit14 over rccAPB2en + or!		\ enable clock to USART1
  bit14 over rccAPB2rst + bic!		\ take it out of reset
  drop
\ Default on PA9/10
  _GPIOA >r
  $0028:0000 $003C:0000 r@ gpioMODER + setMask	\ alternate function on 9/10
  $0000:0110 $0000:0FF0 r> gpioAFRH + setMask	\ AF1 for both
;
))

: init-ser1	\ --
\ *G Initialise USART1 for polled operation. Pin selection is
\ ** performed by *\fo{selio-ser1}.
  selio-ser1
  /us1BRR _USART1 (init-ser)
;

: serkey?1	\ -- flag
\ *G *\fo{KEY?} for USART1.
  _USART1 (serkey?)  ;

: serkey1	\ -- char
\ *G *\fo{KEY} for USART1.
  _USART1 (serkey)  ;

: seremit1	\ char --
\ *G *\fo{EMIT} for USART1.
  _USART1 (seremit)  ;

: serType1	\ c-addr len --
\ *G *\fo{TYPE} for USART1.
  _USART1 (sertype)  ;

: sercr1	\ --
\ *G *\fo{CR} for USART1.
  _USART1 (sercr)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Device structure for USART1
\ *[
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line
\ *]

console-port 1 = [if]
  console1 constant console
\ *G Device structure for console on USART1.
[then]

[then]			\ useUSART1?


\ *********
\ *N USART2
\ *********
\ *P USART2 is on the APB bus, so the baud rate is calculated
\ ** using that bus clock.

useUSART2? [if]

console2-speed APB-speed genBRRval equ /us2BRR	\ -- u
\ *G Baud rate divisor for USART2.

((
: selio-ser2	\ --
\ *G Example to perform clock and pin selection for USART2 on PA2/3.
\ ** You cannot use the easy GPIO code for this unless the GPIO ports
\ ** are clocked and taken out of reset in the start up code.
\ Clocks and reset
  _RCC
  $0000 $0003 2 pick rccCFGR3 + setMask	\ PCLK is UART clock
  bit17 over rccAHBen + or!		\ enable clock to GPIOA
  bit17 over rccAHBrst + bic!		\ take it out of reset
  bit17 over rccAPB1en + or!		\ enable clock to USART2
  bit17 over rccAPB1rst + bic!		\ take it out of reset
  drop
\ AF1 on PA2/3
  _GPIOA >r
  \ value    mask       _GPIO  MODER
  $0000:00A0 $0000:00F0 r@ gpioMODER + setMask      \ alternate function on 2/3
  $0000:1100 $0000:FF00 r> gpioAFRL + setMask   \ AF1 for 2/3

;
))

: init-ser2	\ --
\ *G Initialise USART2 for polled operation. Pin selection is
\ ** performed by *\fo{selio-ser2}.
  selio-ser2
  /us2BRR _USART2 (init-ser)
;

: serkey?2	\ -- flag
\ *G *\fo{KEY?} for USART2.
  _USART2 (serkey?)  ;

: serkey2	\ -- char
\ *G *\fo{KEY} for USART2.
  _USART2 (serkey)  ;

: seremit2	\ char --
\ *G *\fo{EMIT} for USART2.
  _USART2 (seremit)  ;

: serType2	\ c-addr len --
\ *G *\fo{TYPE} for USART2.
  _USART2 (sertype)  ;

: sercr2	\ --
\ *G *\fo{CR} for USART1.
  _USART2 (sercr)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Device structure for USART2
\ *[
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line
\ *]

console-port 2 = [if]
  console2 constant console
\ *G Device structure for console on USART1.
[then]

[then]			\ useUSART2?



\ *********
\ *N USART3
\ *********
\ *P USART3 is on the APB1 bus, so the baud rate is calculated
\ ** using that bus clock. This code is only compiled if the
\ ** equate *\fo{useUSART3} is non-zero}.

useUSART3? [if]

console3-speed APB1-speed genBRRval equ /us3BRR	\ -- u
\ *G Baud rate divisor for USART3.

((
: selio-ser3	\ --
\ *G Example to perform clock and pin selection for USART3
\ ** using GPIO pins PD8/9.
  _RCC
  bit3 over rccAHB1ENR + or!		\ enable clock to GPIOD
  bit3 over rccAHB1RSTR + bic!		\ take it out of reset
  bit18 over rccAPB1ENR + or!		\ enable clock to USART3
  bit18 over rccAPB1RSTR + bic!		\ take it out of reset
  drop
\ Default on PD8/9
  _GPIOD >r
  $000A:0000 $000F:0000 r@ gpioMODER + setMask	\ alternate function on 8/9
  $0000:0077 $0000:00FF r> gpioAFRH + setMask	\ AF7 for both
;
))

: init-ser3	\ --
\ *G Initialise USART3 for polled operation on GPIO pins PD8/9.
  selio-ser3
  /us3BRR _USART3 (init-ser)
;

: serkey?3	\ -- flag
\ *G *\fo{KEY?} for USART3.
  _USART3 (serkey?)  ;

: serkey3	\ -- char
\ *G *\fo{KEY} for USART3.
  _USART3 (serkey)  ;

: seremit3	\ char --
\ *G *\fo{EMIT} for USART3.
  _USART3 (seremit)  ;

: serType3	\ c-addr len --
\ *G *\fo{TYPE} for USART3.
  _USART3 (sertype)  ;

: sercr3	\ --
\ *G *\fo{CR} for USART3.
  _USART3 (sercr)  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Device structure for USART3.
\ *[
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line
\ *]

console-port 3 = [if]
  console3 constant console
\ *G Device structure if console is on USART3.
[then]

[then]			\ useUSART3?


\ *****************
\ *S Initialisation
\ *****************

: init-ser	\ --
\ *G Initialise primary serial ports.
[ useUSART1? ] [if] init-ser1 [then]
[ useUSART2? ] [if] init-ser2 [then]
[ useUSART3? ] [if] init-ser3 [then]
;


\ ======
\ *> ###
\ ======

decimal
