\ serSTM32F0xxqi.fth - STM32F0xx queued interrupt-driven serial driver

((
Copyright (c) 2009, 2011, 2014, 2017
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20171125 SFP005 Interrupt driven serial port
20170312 MAP004 Added OVRDIS selection to provent lockup on input
		overflow.
20140527 RDO003 RDoering - Add USART2 on PA2/3 for STM32F030 Nucleo
20120125 MPE002 Changed to use external pin selection words.
20111130 MPE001 Converted from STM32F103
))

only forth definitions
decimal

\ ===========
\ *! serstm32f0xxqi
\ *T STM32F0xx interrupt-driven serial driver
\ ===========
\ *P This driver provides queued interrupt-driven serial comms
\ ** for STM32F10x CPUs using the internal UARTs. Because there
\ ** are alternate pin assignments for some UARTs, you *\b{must}
\ ** check that the supplied code pin setup is correct for your
\ ** board.

\ *P When you copy this file for your own hardware do not forget
\ ** update the GPIO pin assignments and alternate function
\ ** selections.


\ ***********************
\ *S Baud rate generation
\ ***********************
\ *P The baud rate routines use 16 times sampling and the fractional
\ ** baud rate generator.

\ *E                pclkx
\ **  divisor = -------------
\ **            16 * baudrate
\ *P We then treat the divisor as an integer and four bit fraction,
\ ** effectively scaling it by 16 again.
\ *E             pclkx
\ **  BRRval = --------
\ **           baudrate

interpreter
: genBRRval	\ baud clock -- divisor
\ *G Generate the required BRR value.
  1 lshift				\ for rounding
  swap /				\ brr = clock / baud
  1+  1 rshift				\ rounding
;
target


\ **************
\ *S Shared code
\ **************

struct /UxData	\ -- len
\ *G The structure that is passed to UART handlers
  int ux.USART		\ base address of USART peripher
  int ux.RxQ		\ address of receive queue
  int ux.TxQ		\ address of transmit queue
  int ux.TxQne		\ address of transmit queue flag
end-struct

: UxHandler	\ struct --
  dup ux.USART @ usISR + @ bit5 and if	\ receive interrupt check
    dup ux.USART @ usRDR + @		\ get character
    over ux.RxQ @  dup (cqFull?)	\ -- struct char queue flag
    if  2drop  else  (>cqueue)  endif	\  discard or queue character
  endif
  dup ux.USART @ usCR1 + @ bit7 and if	\ transmit enabled
    dup ux.USART @ usISR + @ bit7 and if \   ready for more
      dup ux.TxQ @ cqchars if		\     and more available
        dup ux.TxQ @ (cqueue>)		\       get next character
          over ux.USART @ usTDR + !	\       send it
      else				\ no more characters
        bit7 over ux.USART @ usCR1 + bic! \ disable TX ints
      endif
    endif
  endif
  drop
;

0 value FaultConsole?	\ -- flag
\ *G When set non-zero, transmit is in polled mode. This is
\ ** required when fault exceptions provide debug information.

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
  -1 to FaultConsole?
  [ tasking? ] [if]  single  [then]
;

: (qseremit)	\ char struct --
\ *G Send a character on the UART whose structure is given.
  FaultConsole? 0= if
    begin				\ wait for non-full queue
      dup ux.TxQ @ cqfull?
     while
      pause
    repeat
    [i
    tuck ux.TxQ @ (>cqueue)		\ character onto queue
    bit7 over ux.USART @ usCR1 + or!	\ turn on transmitter interrupt
    i]
    drop  exit
  endif
  ux.USART @
  begin
    dup usISR + @ bit7 and
  until
  usTDR + !
;

: (qsertype)	\ caddr len struct --
\ *G Send a string on the UART whose structure is given.
  -rot bounds				\ -- base queue
  ?do  i c@ over (qseremit)  loop
  drop
;

: (qsercr)	\ base struct --
\ *G Send a CR/LF on the UART whose structure is given.
  $0D over (qseremit)  $0A swap (qseremit)
;

: (qserkey?)	\ struct -- flag
\ *G Check if a character is available on the UART whose structure is given.
  ux.RxQ @ cqnotempty?  ;

: (qserkey)	\ struct -- char
\ *G Read the next character from the UART whose structure is given.
  ux.RxQ @ cqueue>  ;

: (init-ser)	\ brrval ^usart --
  $0000 over usCR1 + !			\ disable UART, 8 bits, no parity
  $0000 over usCR2 + !			\ 1 stop bit, UART mode
  $1000 over usCR3 + !			\ OVRDIS, no ints, no DMA, no LIN, no smartcard
  tuck usBRR + !			\ set baud rate
  $002C over usCR1 + !			\ 8 bit, no parity, RXNEIE, TE, RE
  $002D over usCR1 + !			\ 8 bit, no parity, RXNEIE, TE, RE, UE
  drop
;


\ *********
\ *S USART1
\ *********
\ *P USART1 is on the APB bus, so the baud rate is calculated
\ ** using that bus clock.

useUSART1? [if]

console1-speed APB-speed genBRRval equ /us1BRR	\ -- u
\ *G Baud rate divisor for USART1.

#256 cqueue: SerInpQ1	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ1	\ -- addr
\ *G Transmitter circular byte queue.

variable U1TxQne	\ -- addr
\ *G Set true when the TxQ contains characters.

create U1struct	\ -- addr
  _USART1 ,  SerInpQ1 ,  SerOutQ1 ,  U1TxQne ,

: U1handler	\ --
\ *G Action of the USART1 interrupt handler.
  U1struct Uxhandler  ;
' U1handler USART1vec# EXC: U1entry


((
: selio-ser1	\ --
\ *G Example to perform clock and pin selection for USART1 on PA9/10.
\ ** Provide your own version, e.g. in the control file.
\ Clocks and reset
  _RCC
  $0000 $0003 2 pick rccCFGR3 + setMask	\ PCLK is UART clock
  bit17 over rccAHBen + or!		\ enable clock to GPIOA
  bit17 over rccAHBrst + bic!		\ take it out of reset
  bit14 over rccAPB2en + or!		\ enable clock to USART1
  bit14 over rccAPB2rst + bic!		\ take it out of reset
  drop
\ Default on PA9/10
  _GPIOA >r
  $0028:0000 $003C:0000 r@ gpioMODER + setMask	\ alternate function on 9/10
  $0000:0110 $0000:0FF0 r> gpioAFRH + setMask	\ AF1 for both
;
))

: init-ser1	\ --
\ *G Initialise USART1 for polled operation. Pin selection is
\ ** performed by *\fo{selio-ser1}.
  selio-ser1
  di
  SerInpQ1 init-cqueue
  SerOutQ1 init-cqueue
  U1TxQne off
  /us1BRR _USART1 (init-ser)
\  $80 USART1vec# SetPri			\ interrupt priority
  USART1vec# EnInt			\ interrupt enable in NVIC
  ei
;

: serkey?1	\ -- flag
\ *G *\fo{KEY?} for USART1.
  U1struct (qserkey?)  ;

: serkey1	\ -- char
\ *G *\fo{KEY} for USART1.
  U1struct (qserkey)  ;

: seremit1	\ char --
\ *G *\fo{EMIT} for USART1.
  U1struct (qseremit)  ;

: serType1	\ c-addr len --
\ *G *\fo{TYPE} for USART1.
  U1struct (qsertype)  ;

: sercr1	\ --
\ *G *\fo{CR} for USART1.
  U1struct (qsercr)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Device structure for USART1
\ *[
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line
\ *]

console-port 1 = [if]
  console1 constant console
\ *G Device structure for console on USART1.
[then]

[then]			\ useUSART1?


\ *********
\ *S USART2
\ *********
\ *P USART2 is on the APB bus, so the baud rate is calculated
\ ** using that bus clock.

useUSART2? [if]

console2-speed APB-speed genBRRval equ /us2BRR	\ -- u
\ *G Baud rate divisor for USART2.

#256 cqueue: SerInpQ2	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ2	\ -- addr
\ *G Transmitter circular byte queue.

variable U2TxQne	\ -- addr
\ *G Set true when the TxQ contains characters.

create U2struct	\ -- addr
  _USART2 ,  SerInpQ2 ,  SerOutQ2 ,  U2TxQne ,

: U2handler	\ --
\ *G Action of the USART2 interrupt handler.
  U2struct Uxhandler  ;
' U2handler USART2vec# EXC: U2entry
\ *G Entry point of the USART2 interrupt handler.

((
: selio-ser2	\ --
\ *G Example to perform clock and pin selection for USART2 on PA2/3.
\ ** You cannot use the easy GPIO code for this unless the GPIO ports
\ ** are clocked and taken out of reset in the start up code.
\ Clocks and reset
  _RCC
  $0000 $0003 2 pick rccCFGR3 + setMask	\ PCLK is UART clock
  bit17 over rccAHBen + or!		\ enable clock to GPIOA
  bit17 over rccAHBrst + bic!		\ take it out of reset
  bit17 over rccAPB1en + or!		\ enable clock to USART2
  bit17 over rccAPB1rst + bic!		\ take it out of reset
  drop
\ AF1 on PA2/3
  _GPIOA >r
  \ value    mask       _GPIO  MODER
  $0000:00A0 $0000:00F0 r@ gpioMODER + setMask      \ alternate function on 2/3
  $0000:1100 $0000:FF00 r> gpioAFRL + setMask   \ AF1 for 2/3

;
))

: init-ser2	\ --
\ *G Initialise USART2 for polled operation. Pin selection is
\ ** performed by *\fo{selio-ser2}.
  selio-ser2
  di
  SerInpQ2 init-cqueue
  SerOutQ2 init-cqueue
  U2TxQne off
  /us2BRR _USART2 (init-ser)
  USART2vec# EnInt			\ interrupt enable in NVIC
  ei
;

: serkey?2	\ -- flag
\ *G *\fo{KEY?} for USART2.
  U2struct (qserkey?)  ;

: serkey2	\ -- char
\ *G *\fo{KEY} for USART2.
  U2struct (qserkey)  ;

: seremit2	\ char --
\ *G *\fo{EMIT} for USART2.
  U2struct (qseremit)  ;

: serType2	\ c-addr len --
\ *G *\fo{TYPE} for USART2.
  U2struct (qsertype)  ;

: sercr2	\ --
\ *G *\fo{CR} for USART2.
  U2struct (qsercr)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Device structure for USART2
\ *[
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line
\ *]

console-port 2 = [if]
  console2 constant console
\ *G Device structure for console on USART1.
[then]

[then]			\ useUSART2?



\ *********
\ *N USART3
\ *********
\ *P USART3 is on the APB1 bus, so the baud rate is calculated
\ ** using that bus clock. This code is only compiled if the
\ ** equate *\fo{useUSART3} is non-zero}.

useUSART3? [if]

console3-speed APB1-speed genBRRval equ /us3BRR	\ -- u
\ *G Baud rate divisor for USART3.

#256 cqueue: SerInpQ3	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ3	\ -- addr
\ *G Transmitter circular byte queue.

variable U3TxQne	\ -- addr
\ *G Set true when the TxQ contains characters.

create U2struct	\ -- addr
  _USART3 ,  SerInpQ3 ,  SerOutQ3 ,  U3TxQne ,

: U3handler	\ --
\ *G Action of the USART3 interrupt handler.
  U3struct Uxhandler  ;
' U3handler USART3vec# EXC: U32entry
\ *G Entry point of the USART3 interrupt handler.

((
: selio-ser3	\ --
\ *G Example to perform clock and pin selection for USART3
\ ** using GPIO pins PD8/9.
  _RCC
  bit3 over rccAHB1ENR + or!		\ enable clock to GPIOD
  bit3 over rccAHB1RSTR + bic!		\ take it out of reset
  bit18 over rccAPB1ENR + or!		\ enable clock to USART3
  bit18 over rccAPB1RSTR + bic!		\ take it out of reset
  drop
\ Default on PD8/9
  _GPIOD >r
  $000A:0000 $000F:0000 r@ gpioMODER + setMask	\ alternate function on 8/9
  $0000:0077 $0000:00FF r> gpioAFRH + setMask	\ AF7 for both
;
))

: init-ser3	\ --
\ *G Initialise USART3 for polled operation. Pin selection is
\ ** performed by *\fo{selio-ser3}.
  selio-ser3
  di
  SerInpQ3 init-cqueue
  SerOutQ3 init-cqueue
  U3TxQne off
  /us3BRR _USART3 (init-ser)
  USART3vec# EnInt			\ interrupt enable in NVIC
  ei
;

: serkey?3	\ -- flag
\ *G *\fo{KEY?} for USART3.
  U3struct (qserkey?)  ;

: serkey3	\ -- char
\ *G *\fo{KEY} for USART3.
  U3struct (qserkey)  ;

: seremit3	\ char --
\ *G *\fo{EMIT} for USART3.
  U3struct (qseremit)  ;

: serType3	\ c-addr len --
\ *G *\fo{TYPE} for USART3.
  U3struct (qsertype)  ;

: sercr3	\ --
\ *G *\fo{CR} for USART3.
  U3struct (qsercr)  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Device structure for USART3.
\ *[
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line
\ *]

console-port 3 = [if]
  console3 constant console
\ *G Device structure if console is on USART3.
[then]

[then]			\ useUSART3?


\ *****************
\ *S Initialisation
\ *****************

: init-ser	\ --
\ *G Initialise primary serial ports.
[ useUSART1? ] [if] init-ser1 [then]
[ useUSART2? ] [if] init-ser2 [then]
[ useUSART3? ] [if] init-ser3 [then]
;


\ ======
\ *> ###
\ ======

decimal
