\ serSTM32F10xp.fth - STM32F10x polled serial driver

((
Copyright (c) 2009
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20120321 SFP001 Overhauled and refactored.
))

only forth definitions
decimal

\ ===========
\ *! serstm32f10xp
\ *T STM32F10x polled serial driver
\ ===========


\ ***********************
\ *S Baud rate generation
\ ***********************
\ *P *\fo{GenBRR} uses the system clock rate (in Hz), the
\ ** required serial speed (in BPS) and returns the 16 bit
\ ** divider value for the baud rate generator. The bottom
\ ** four bits of the divisor are used by the fractional part
\ ** of the baud rate generator.
\ *E              system-speed
\ ** divisor = -------------------
\ **                baudrate

interpreter
: genBRD	\ baud busclk -- divisor
\ *G Interpreter word to generate baud rate divisor. The clock must
\ ** be specified.
  1 lshift				\ for rounding
  swap /				\ brr = clock / baud
  1+  1 rshift				\ rounding
;

: genBRR	\ baud -- divisor
\ *G Interpreter word to generate baud rate divisor. The clock is
\ ** assumed to be the system clock.
  system-speed genBRD
;
target


\ **************
\ *S Shared code
\ **************

: (serKey?)	\ base -- flag
\ *G The equivalent of *\fo{KEY?}.
  usSR + @ bit5 and  ;

: (serkey)	\ base -- char
\ *G The equivalent of *\fo{KEY}.
[ tasking? ] [if]
  begin
    dup usSR + @ bit5 and 0=
   while
    pause
  repeat
[else]
  begin
    dup usSR + @ bit5 and
  until
[then]
  usRDR + @ $FF and
;

: (seremit)	\ char base --
\ *G The equivalent of *\fo{EMIT}.
[ tasking? ] [if]
  begin
    dup usSR + @ bit7 and 0=
   while
    pause
  repeat
[else]
  begin
    dup usSR + @ bit7 and
  until
[then]
  usTDR + !
;

: (sertype)	\ caddr len base --
\ *G The equivalent of *\fo{TYPE}.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G The equivalent of *\fo{CR}.
  #13 over (seremit)  #10 swap (seremit)
;

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
[ tasking? ] [if]  single  [then]
;

: (init-ser)	\ divisor base --
\ *G Given a baud rate divisor and a USART base address, set
\ ** the UART.
  $2000 over usCR1 + !			\ 8 bit, no parity, disable Tx/Rx
  $0000 over usCR2 + !			\ 1 stop bit, UART mode
  $0000 over usCR3 + !			\ no interrupts, no DMA, no LIN, no smartcard
  tuck usBRR + !			\ set baud rate
  $200C over usCR1 + !			\ 8 bit, no parity, TE, RE
  drop
;


\ *********
\ *S USART1
\ *********

useUSART1? [if]

console1-speed APB2-speed genBRD equ /us1BRD	\ -- u
\ *G Baud rate divisor for USART1.

[undefined] selio-ser1 [if]
bit14 bit2 or bit0 or equ u1_apb2enA	\ --
\ *G Initialisation value for USART1 in RCC_APB2ENR. Clocks for
\ ** USART1, GPIO port A, and Alternate function are enabled.
bit14 bit3 or bit0 or equ u1_apb2enB	\ -- ; remapped to GPIO port B
\ *G Initialisation value for USART1 in RCC_APB2ENR. Clocks for
\ ** USART1, GPIO port B, and Alternate function are enabled.

: selio-ser1	\ --
\ *G If this word has not already been defined, use this one that
\ ** selects USART1 I/O on pins PA9 and PA10.
  u1_apb2enA _RCC rccAPB2en + or!	\ set required bits
  bit2 _AFIO afMAPR + bic!		\ clear bit 2, no USART1 remap
  $04B0 $0FF0 _GPIOA gp.CRH + setmask	\ conf/mode RX=PA10, TX=PA9
((
  u1_apb2enB _RCC rccAPB2en + or!	\ set required bits
  bit2 _AFIO afMAPR + or!		\ set bit 2, USART1 remapped
  $4B00:0000 $FF00:0000 _GPIOB gp.CRL + setmask	\ conf/mode RX=PB6, TX=PB6
))
;
[then]

: init-ser1	\ --
\ *G Initialise USART1 for polled operation.
  selio-ser1				\ configure UART pins
  /us1BRD _USART1 (init-ser)		\ set up UART
;

: serkey?1	\ -- flag
\ *G The equivalent of *\fo{KEY?}.
  _USART1 (serkey?)  ;

: serkey1	\ -- char
\ *G The equivalent of *\fo{KEY}.
  _USART1 (serkey)  ;

: seremit1	\ char --
\ *G The equivalent of *\fo{EMIT}.
  _USART1 (seremit)  ;

: serType1	\ c-addr len --
\ *G The equivalent of *\fo{TYPE}.
  _USART1 (sertype)  ;

: sercr1	\ --
\ *G The equivalent of *\fo{CR}.
  _USART1 (sercr)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ *********
\ *S USART2
\ *********

useUSART2? [if]

console2-speed APB1-speed genBRD equ /us2BRD	\ -- u
\ *G Baud rate divisor for USART2.

[undefined] selio-ser2 [if]
: selio-ser2	\ --
\ *G If this word has not already been defined, use this one that
\ ** selects USART2 I/O on pins PD5 (Tx) and PD6 (Rx). On the
\ ** Olimex P107 board, these appear at RS232 level on the DB9
\ ** connector.
  bit17 _RCC
    2dup rccAPB1en + or!		\ enable clock to USART2
    rccAPB1rst + bic!			\ remove reset from USART2
  bit5 bit0 or _RCC
    2dup rccAPB2en + or!		\ enable clock to GPIOD and AFIO
    rccAPB2rst + bic!			\ remove reset from GPIOD and AFIO
  bit3 _AFIO afMAPR + or!		\ bit3=1, USART2 on PD5/6
  $04B0:0000 $0FF:0000 _GPIOD gp.CRL + setmask	\ conf/mode bits PD5 (Tx), PD6 (Rx)
;
[then]

((
: wastetime2	\ --
  4000000 0 do loop
;

: blinky2	\ --
  bit4 _RCC
    2dup rccAPB2en + or!		\ enable clock to GPIOC
    rccAPB2rst + bic!			\ remove reset from GPIOC
  $3300:0000 $FF00:0000 _GPIOC gp.CRL + setMask	\ 50MHz GP p-p o/p, PC6/7
  wastetime2
  begin
    $0040:0080 _GPIOC gp.BSRR + !
    [char] A _USART2 (seremit)
    wastetime2
    $0080:0040 _GPIOC gp.BSRR + !
    [char] Z _USART2 (seremit)
    wastetime2
  again
;
))

: init-ser2	\ --
\ *G Initialise USART for polled operation.
  selio-ser2				\ configure UART pins
  /us2BRD _USART2 (init-ser)		\ set up UART
\  blinky2
;

: serkey?2	\ -- flag
\ *G The equivalent of *\fo{KEY?}.
  _USART2 (serkey?)  ;

: serkey2	\ -- char
\ *G The equivalent of *\fo{KEY}.
  _USART2 (serkey)  ;

: seremit2	\ char --
\ *G The equivalent of *\fo{EMIT}.
  _USART2 (seremit)  ;

: serType2	\ c-addr len --
\ *G The equivalent of *\fo{TYPE}.
  _USART2 (sertype)  ;

: sercr2	\ --
\ *G The equivalent of *\fo{CR}.
  _USART2 (sercr)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ *********
\ *S USART3
\ *********

useUSART3? [if]

console3-speed APB1-speed genBRD equ /us3BRD	\ -- u
\ *G Baud rate divisor for USART3.

[undefined] selio-ser3 [if]
: selio-ser3	\ --
\ *G If this word has not already been defined, use this one that
\ ** selects USART3 I/O on pins PD8 (Tx) and PD9 (Rx). On the
\ ** Olimex P107 board, these appear on the UEXT connector..
  bit18 _RCC
    2dup rccAPB1en + or!		\ enable clock to USART3
    rccAPB1rst + bic!			\ remove reset from USART3
  bit5 bit0 or _RCC
    2dup rccAPB2en + or!		\ enable clock to GPIOD and AFIO
    rccAPB2rst + bic!			\ remove reset from GPIOD and AFIO
  $0030 _AFIO afMAPR + or!		\ bits 5:4 -> 11, full remap
  $004B $00FF _GPIOD gp.CRH + setmask	\ conf/mode bits PD8 (Tx), PD9 (Rx)
;
[then]

: init-ser3	\ --
\ *G Initialise USART for polled operation.
  selio-ser3				\ configure UART pins
  /us3BRD _USART3 (init-ser)		\ set up UART
;

: serkey?3	\ -- flag
\ *G The equivalent of *\fo{KEY?}.
  _USART3 (serkey?)  ;

: serkey3	\ -- char
\ *G The equivalent of *\fo{KEY}.
  _USART3 (serkey)  ;

: seremit3	\ char --
\ *G The equivalent of *\fo{EMIT}.
  _USART3 (seremit)  ;

: serType3	\ c-addr len --
\ *G The equivalent of *\fo{TYPE}.
  _USART3 (sertype)  ;

: sercr3	\ --
\ *G The equivalent of *\fo{CR}.
  _USART3 (sercr)  ;

create Console3	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device.
  ' serkey3 ,		\ -- char ; receive char
  ' serkey?3 ,		\ -- flag ; check receive char
  ' seremit3 ,		\ -- char ; display char
  ' sertype3 ,		\ caddr len -- ; display string
  ' sercr3 ,		\ -- ; display new line

console-port 3 = [if]
  console3 constant console
[then]

[then]


\ *****************
\ *S Initialisation
\ *****************

: init-ser	\ --
\ *G Initialise primary serial ports.
[ useUSART1? ] [if]  init-ser1  [then]
[ useUSART2? ] [if]  init-ser2  [then]
[ useUSART3? ] [if]  init-ser3  [then]
;


\ ======
\ *> ###
\ ======

decimal


