\ serSTM32F10xqi.fth - STM32F10x serial: queued interrupt RX and TX

((
Copyright (c) 2010, 2011, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20120326 SFP002 Overhauled and refactored.
20111129 MGS001 Removed interrupt priority setting
))

only forth definitions
decimal

\ ===========
\ *! serstm32f10xqi
\ *T STM32F10x serial driver
\ ===========
\ *P This driver provides queued interrupt-driven serial comms
\ ** for STM32F10x CPUs using the internal UARTs. Because there
\ ** are alternate pin assignments for some UARTs, you *\b{must}
\ ** check that the supplied code pin setup is correct for your
\ ** board.

\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUSART1? [if]
1 equ useUSART1?	\ -- n
\ *G Compile USART1 code if non-zero.
[then]

[undefined] useUSART2? [if]
0 equ useUSART2?	\ -- n
\ *G Compile USART2 code if non-zero.
[then]

[undefined] useUSART3? [if]
0 equ useUSART3?	\ -- n
\ *G Compile UART3 code if non-zero.
[then]


\ ***********************
\ *S Baud rate generation
\ ***********************
\ *P *\fo{GenBRR} uses the system clock rate (in Hz), the
\ ** required serial speed (in BPS) and returns the 16 bit
\ ** divider value for the baud rate generator. The bottom
\ ** four bits of the divisor are used by the fractional part
\ ** of the baud rate generator.
\ *E              system-speed
\ ** divisor = -------------------
\ **                baudrate

interpreter
: genBRD	\ baud busclk -- divisor
\ *G Interpreter word to generate baud rate divisor. The clock must
\ ** be specified.
  1 lshift				\ for rounding
  swap /				\ brr = clock / baud
  1+  1 rshift				\ rounding
;

: genBRR	\ baud -- divisor
\ *G Interpreter word to generate baud rate divisor. The clock is
\ ** assumed to be the system clock.
  system-speed genBRD
;
target


\ **************
\ *S Shared code
\ **************

struct /UxData	\ -- len
\ *G The structure that is passed to UART handlers
  int ux.USART		\ base address of USART peripher
  int ux.RxQ		\ address of receive queue
  int ux.TxQ		\ address of transmit queue
  int ux.TxQne		\ address of transmit queue flag
end-struct

: UxHandler	\ struct --
  dup ux.USART @ usSR + @ bit5 and if	\ receive interrupt check
    dup ux.USART @ usRDR + @		\ get character
    over ux.RxQ @  dup (cqFull?)	\ -- struct char queue flag
    if  2drop  else  (>cqueue)  endif	\  discard or queue character
  endif
  dup ux.USART @ usCR1 + @ bit7 and if	\ transmit enabled
    dup ux.USART @ usSR + @ bit7 and if \   ready for more
      dup ux.TxQ @ cqchars if		\     and more available
        dup ux.TxQ @ (cqueue>)		\       get next character
          over ux.USART @ usTDR + !	\       send it
      else				\ no more characters
        bit7 over ux.USART @ usCR1 + bic! \ disable TX ints
      endif
    endif
  endif
  drop
;

0 value FaultConsole?	\ -- flag
\ *G When set non-zero, transmit is in polled mode. This is
\ ** required when fault exceptions provide debug information.

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
  -1 to FaultConsole?
  [ tasking? ] [if]  single  [then]
;

: (qseremit)	\ char struct --
\ *G Send a character on the UART whose structure is given.
  FaultConsole? 0= if
    begin				\ wait for non-full queue
      dup ux.TxQ @ cqfull?
     while
      pause
    repeat
    [i
    tuck ux.TxQ @ (>cqueue)		\ character onto queue
    bit7 over ux.USART @ usCR1 + or!	\ turn on transmitter interrupt
    i]
    drop  exit
  endif
  ux.USART @
  begin
    dup usSR + @ bit7 and
  until
  usTDR + !
;

: (qsertype)	\ caddr len struct --
\ *G Send a string on the UART whose structure is given.
  -rot bounds				\ -- base queue
  ?do  i c@ over (qseremit)  loop
  drop
;

: (qsercr)	\ base struct --
\ *G Send a CR/LF on the UART whose structure is given.
  $0D over (qseremit)  $0A swap (qseremit)
;

: (qserkey?)	\ struct -- flag
\ *G Check if a character is available on the UART whose structure is given.
  ux.RxQ @ cqnotempty?  ;

: (qserkey)	\ struct -- char
\ *G Read the next character from the UART whose structure is given.
  ux.RxQ @ cqueue>  ;

\ *********
\ *S USART1
\ *********

useUSART1? [if]

console1-speed APB2-speed genBRD equ /us1BRD	\ -- u
\ *G Baud rate divisor for USART1.

#256 cqueue: SerInpQ1	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ1	\ -- addr
\ *G Transmitter circular byte queue.

variable U1TxQne	\ -- addr
\ *G Set true when the TxQ contains characters.

create U1struct	\ -- addr
  _USART1 ,  SerInpQ1 ,  SerOutQ1 ,  U1TxQne ,

: U1handler	\ --
\ *G Action of the USART1 interrupt handler.
  U1struct Uxhandler  ;
' U1handler USART1vec# EXC: U1entry

[undefined] selio-ser1 [if]
bit14 bit2 or bit0 or equ u1_apb2enA	\ --
\ *G Initialisation value for USART1 in RCC_APB2ENR. Clocks for
\ ** USART1, GPIO port A, and Alternate function are enabled.
bit14 bit3 or bit0 or equ u1_apb2enB	\ -- ; remapped to GPIO port B
\ *G Initialisation value for USART1 in RCC_APB2ENR. Clocks for
\ ** USART1, GPIO port B, and Alternate function are enabled.

: selio-ser1	\ --
\ *G If this word has not already been defined, use this one that
\ ** selects USART1 I/O on pins PA9 and PA10.
  u1_apb2enA _RCC rccAPB2en + or!	\ set required bits
  bit2 _AFIO afMAPR + bic!		\ clear bit 2, no USART1 remap
  $04B0 $0FF0 _GPIOA gp.CRH + setmask	\ conf/mode RX=PA10, TX=PA9
((
  u1_apb2enB _RCC rccAPB2en + or!	\ set required bits
  bit2 _AFIO afMAPR + or!		\ set bit 2, USART1 remapped
  $4B00:0000 $FF00:0000 _GPIOB gp.CRL + setmask	\ conf/mode RX=PB6, TX=PB6
))
;
[then]

: init-ser1	\ --
\ *G Initialise USART1 for operation on GPIO pins PA9/10.
\ ** Transmit is polled, receive is interrupt-driven.
  selio-ser1
  di
  SerInpQ1 init-cqueue
  SerOutQ1 init-cqueue
  U1TxQne off
  _USART1
  /us1BRD over usBRR + !		\ set baud rate
  $2000 over usCR1 + !			\ 8 bit, no parity
  $0000 over usCR2 + !			\ 1 stop bit, UART mode
  $0000 over usCR3 + !			\ no interrupts, no DMA, no LIN, no smartcard
  $202C over usCR1 + !			\ 8 bit, no parity, RXNEIE, TE, RE
  drop
\  $80 USART1vec# SetPri			\ interrupt priority
  USART1vec# EnInt			\ interrupt enable in NVIC
  ei
;

: serkey?1	\ -- flag
\ *G The equivalent of *\fo{KEY?}.
  U1struct (qserkey?)  ;

: serkey1	\ -- char
\ *G The equivalent of *\fo{KEY}.
  U1struct (qserkey)  ;

: seremit1	\ char --
\ *G The equivalent of *\fo{EMIT}.
  U1struct (qseremit)  ;

: serType1	\ c-addr len --
\ *G The equivalent of *\fo{TYPE}.
  U1struct (qsertype)  ;

: sercr1	\ --
\ *G The equivalent of *\fo{CR}.
  U1struct (qsercr)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ *********
\ *S USART2
\ *********

useUSART2? [if]

console2-speed APB1-speed genBRD equ /us2BRD	\ -- u
\ *G Baud rate divisor for USART2.

#256 cqueue: SerInpQ2	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ2	\ -- addr
\ *G Transmitter circular byte queue.

variable U2TxQne	\ -- addr
\ *G Set true when the TxQ contains characters.

create U2struct	\ -- addr
  _USART2 ,  SerInpQ2 ,  SerOutQ2 ,  U2TxQne ,

: U2handler	\ --
\ *G Action of the USART2 interrupt handler.
  U2struct Uxhandler  ;
' U2handler USART2vec# EXC: U2entry

[undefined] selio-ser2 [if]
: selio-ser2	\ --
\ *G If this word has not already been defined, use this one that
\ ** selects USART2 I/O on pins PD5 (Tx) and PD6 (Rx). On the
\ ** Olimex P107 board, these appear at RS232 level on the DB9
\ ** connector.
  bit17 _RCC
    2dup rccAPB1en + or!		\ enable clock to USART2
    rccAPB1rst + bic!			\ remove reset from USART2
  bit5 bit0 or _RCC
    2dup rccAPB2en + or!		\ enable clock to GPIOD and AFIO
    rccAPB2rst + bic!			\ remove reset from GPIOD and AFIO
  bit3 _AFIO afMAPR + or!		\ bit3=1, USART2 on PD5/6
  $04B0:0000 $0FF:0000 _GPIOD gp.CRL + setmask	\ conf/mode bits PD5 (Tx), PD6 (Rx)
;
[then]

: init-ser2	\ --
\ *G Initialise USART2 for interrupt-driven operation with queues.
  selio-ser2
  di
  SerInpQ2 init-cqueue
  SerOutQ2 init-cqueue
  U2TxQne off
  _USART2
  /us2BRD over usBRR + !		\ set baud rate
  $2000 over usCR1 + !			\ 8 bit, no parity
  $0000 over usCR2 + !			\ 1 stop bit, UART mode
  $0000 over usCR3 + !			\ no interrupts, no DMA, no LIN, no smartcard
  $202C over usCR1 + !			\ 8 bit, no parity, RXNEIE, TE, RE
  drop
  USART2vec# EnInt			\ interrupt enable in NVIC
  ei
;

: serkey?2	\ -- flag
\ *G The equivalent of *\fo{KEY?}.
  U2struct (qserkey?)  ;

: serkey2	\ -- char
\ *G The equivalent of *\fo{KEY}.
  U2struct (qserkey)  ;

: seremit2	\ char --
\ *G The equivalent of *\fo{EMIT}.
  U2struct (qseremit)  ;

: serType2	\ c-addr len --
\ *G The equivalent of *\fo{TYPE}.
  U2struct (qsertype)  ;

: sercr2	\ --
\ *G The equivalent of *\fo{CR}.
  U2struct (qsercr)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ *****************
\ *S Initialisation
\ *****************

: init-ser	\ --
\ *G Initialise primary serial ports.
  [ useUSART1? ] [if]  init-ser1  [then]
  [ useUSART2? ] [if]  init-ser2  [then]
  [ useUSART3? ] [if]  init-ser3  [then]
;


\ ======
\ *> ###
\ ======

decimal


