\ serTM4C12xxp.fth - TM4C12xx polled RX, polled TX, serial communication drivers

((
Copyright (c) 2010, 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20190214 MPE002 Converted to TM4C12xx.
20100317 MPE001 First release.
))

only forth definitions


\ ==============
\ *! sertm4c12xxp
\ *T TM4C12xx polled serial driver
\ ==============
\ *P This driver provides polled serial comms for TM4C12xx CPUs
\ ** using the internal UARTs. As of early 2019, it seems that
\ ** the UART baud rate unit requires a fairly low clock speed,
\ ** and is only reliable using the PIOSC as a source (16 MHz).


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBRD22} uses the PIOSC clock rate (in Hz), the
\ ** PIOSC clock speed, and the required serial speed
\ ** (in BPS) and returns the 16 bit divider value for the baud
\ ** rate generator plus a 6 bit fraction.
\ *E              PIOSC-speed
\ ** divisor = -------------------
\ **            clkdiv * baudrate
\ *P The calculation assumes that the fractional diver is used,
\ ** and that *\fo{clkdiv}=16, i.e. HSE=0, bit 5 clear in UnCTL.

interpreter
: GenBRD22	\ bps-rate -- divisor22
\ *G Given a baud rate, calculate the 16 bit baud rate integer
\ ** divisor and the 6 bit fractionaldivider. We assume HSE=0.
  #16000000  3 lshift  swap /		\ bottom 7 bits = fraction
  1+ 1 rshift				\ round fraction to 6 bits
\ cr ." Integer  " dup 6 rshift .
\ cr ." Fraction " dup $3F and .
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal

: +FaultConsole	( -- )  ;
\ *G Because this is a polled driver, *\fo{+FaultConsole} for
\ ** fault handling is a *\fo{NOOP}. See *\i{Cortex/FaultCortex.fth}
\ ** for more details.

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup Un.FR + @ bit5 and 0=		\ Tx FIFO full test
  until
  Un.DR + !
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)
;

: (serkey?)	\ base -- t/f
\ *G Return true if the given UART has a character avilable to read.
  Un.FR + @ bit4 and 0=			\ Rx FIFO empty test
;

: (serkey)	\ base -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
[ tasking? ] [if]  pause  [then]
    dup Un.FR + @ bit4 and 0=
  until
  Un.DR + @ $FF and
;

: initUART	\ divisor22 base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r				\ save divisor
  0 over Un.CTL + !			\ UARTEN=0
  5 over Un.CC + !			\ use AltClk=PIOSC for baud rate
  r@ 6 rshift $FFFF and over Un.IBRD + ! \ integer baud rate
  r> $3F and over Un.FBRD + !		\ fractional baud rate,
  $70 over Un.LCRH + !			\ 8N1, FIFOs
  $12 over Un.IFLS + !			\ 1/2 FIFO level
  0 over Un.IM + !			\ no interrupts
  0 over Un.DMACTL + !			\ no DMA
  $0300 $CFBF 2 pick Un.CTL + setMask	\ set RXE=1, TXE=1, HSE=0
  1 swap Un.CTL + or!			\ UARTEN=1
;

external


\ ********
\ *S UART0
\ ********

useUART0? [if]
console0-speed GenBRD22 equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.
[undefined] init-ser0 [if]
: init-ser0	\ --
\ *G Initialise serial port 0, using U0Rx=PA0, U0Tx=PA1
  _SysCtl
  bit0 over sc.RCGCUART + or!		\ enable clock and power to UART0
  bit0 swap sc.RCGCGPIO + or!		\ enable clock and power to GPIOA
  noop noop noop noop			\ must wait a bit
  _GPIOA
  $02 $03 2 pick gpio.DIR + setmask	\ bit 1 o/p, bit 0 i/p *
  $03 over gpio.AFSEL + or!		\ pins 0/1 are peripheral
  $03 over gpio.DR2R + or!		\ pins 0/1 are 2mA drive *
  $03 over gpio.SLR + bic!		\ pins 0/1 have no slew control *
  $03 over gpio.DEN + or!		\ enable pins
  $03 over gpio.ODR + bic!		\ no open drain *
  $03 over gpio.PUR + bic!		\ no pull ups *
  $03 over gpio.PDR + bic!		\ no pull downs *
  $11 $FF  rot gpio.PCTL + setMask	\ peripheral 1 for PA0/1
  u0-divisor _UART0 initUART
;
[then]

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  _UART0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  _UART0 (serkey)  ;
create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]	\ console port

[then]	\ useUART0?


\ ********
\ *S UART1
\ ********

useUART1? [if]

console1-speed GenBRD22 equ u1-divisor	\ -- n
\ *G UART1 baud rate divisor.

[undefined] init-ser1 [if]
: init-ser1	\ --
\ *G Initialise serial port 1, using U1Rx=PB0, U1Tx=PB1
  _SysCtl
  bit1 over sc.RCGCUART + or!		\ enable clock and power to UART1
  bit1 swap sc.RCGCGPIO + or!		\ enable clock and power to GPIOB
  noop noop noop noop			\ must wait a bit
  _GPIOB
  $03 over gpio.AFSEL + or!		\ pins 0/1 are peripheral
  $03 over gpio.DR2R + or!		\ pins 0/1 are 2mA drive
  $03 over gpio.SLR + bic!		\ pins 0/1 have no slew control
  $03 over gpio.DEN + or!		\ enable pins
  $03 over gpio.ODR + bic!		\ no open drain *
  $03 over gpio.PUR + bic!		\ no pull ups *
  $03 over gpio.PDR + bic!		\ no pull downs *
  $0011 $00FF rot gpioPCTL + setMask	\ peripheral 1 for PA0/1
  u1-divisor _UART1 initUART
;
[then]

: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 (seremit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 (sertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 (sercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  _UART1 (serkey?)  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  _UART1 (serkey)  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ ********
\ *S UART2
\ ********

useUART2? [if]

console2-speed GenBRD22 equ u2-divisor	\ -- n
\ *G UART2 baud rate divisor.

[undefined] init-ser2 [if]
: init-ser2	\ --
\ *G Initialise serial port 2, using U2Rx=PA6, U2Tx=PA7
  _SysCtl
  bit2 over sc.RCGCUART + or!		\ enable clock and power to UART2
  bit0 swap sc.RCGCGPIO + or!		\ enable clock and power to GPIOA
  noop noop noop noop			\ must wait a bit
  _GPIOA
  $C0 over gpio.AFSEL + or!		\ pins 6/7 are peripheral
  $C0 over gpio.DR2R + or!		\ pins 6/7 are 2mA drive
  $C0 over gpio.SLR + bic!		\ pins 6/7 have no slew control
  $C0 over gpio.DEN + or!		\ enable pins
  $C0 over gpio.ODR + bic!		\ no open drain *
  $C0 over gpio.PUR + bic!		\ no pull ups *
  $C0 over gpio.PDR + bic!		\ no pull downs *
  $1100:0000 $FF00:0000 rot gpioPCTL + setMask	\ peripheral 1 for PA6/7
  u2-divisor _UART2 initUART
;
[then]

: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 (seremit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 (sertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 (sercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  _UART2 (serkey?)  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  _UART2 (serkey)  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all used serial ports.
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
((
  begin
    PN0>1  [char] A seremit0
    1000000 0 do loop
    PN0>0  [char] Z seremit0
    1000000 0 do loop
  again
))
;


\ ======
\ *> ###
\ ======

decimal
