\ serTM4C12xxqi.fth - TM4C12xx queued interrupt RX/TX UART driver

((
Copyright (c) 2010, 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20190222 MPE002 Converted to TM4C12xx.
20100322 MPE001 First release.
))

only forth definitions


\ ==============
\ *! sertm4c12xxqi
\ *T TM4C12xx queued interrupt RX/TX UART driver
\ ==============
\ *P This driver provides serial comms for TM4C12xx CPUs
\ ** using the internal UARTs. As of early 2019, it seems that
\ ** the UART baud rate unit requires a fairly low clock speed,
\ ** and is only reliable using the PIOSC as a source (16 MHz).


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBRD22} uses the PIOSC clock rate (in Hz), the
\ ** PIOSC clock speed, and the required serial speed
\ ** (in BPS) and returns the 16 bit divider value for the baud
\ ** rate generator plus a 6 bit fraction.
\ *E              PIOSC-speed
\ ** divisor = -------------------
\ **            clkdiv * baudrate
\ *P The calculation assumes that the fractional diver is used,
\ ** and that *\fo{clkdiv}=16, i.e. HSE=0, bit 5 clear in UnCTL.

interpreter
: GenBRD22	\ bps-rate -- divisor22
\ *G Given a baud rate, calculate the 16 bit baud rate integer
\ ** divisor and the 6 bit fractionaldivider. We assume HSE=0.
  #16000000  3 lshift  swap /		\ bottom 7 bits = fraction
  1+ 1 rshift				\ round fraction to 6 bits
\ cr ." Integer  " dup 6 rshift .
\ cr ." Fraction " dup $3F and .
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal
: >RxQ		\ char queue --
\ *G Put character from UART into queue.
\ ** INTERNAL.
  dup (cqFull?)				\ if queue full
  if  2drop  exit  endif		\  discard character
  (>cqueue)				\  queue character
;

: FIFO>RxQ      \ base queue --
\ *G Given a UART base address and a character queue,
\ ** empty the UART Rx FIFO into the queue.
\ ** INTERNAL.
  begin
    over Un.FR + @ bit4 and 0=
   while
    over Un.DR + @ $FF and over >RxQ
  repeat
  2drop
;

: TxQ>FIFO	\ base queue --
\ *G Given a UART base address and a character queue,
\ ** put transmit characters into the UART Tx FIFO.
  begin
    dup cqNotEmpty?			\ more to send?
   while
    over Un.FR + @ bit5 and 0=		\ room?
   while
    2dup (cqueue>) swap Un.DR + !	\ write character
  repeat then
  2drop
;

0 value FaultConsole?	\ -- flag
\ *G When set non-zero, transmit is in polled mode. This is
\ ** required when fault exceptions provide debug information.

: +FaultConsole	\ --
\ *G For use in fault exception handlers, the multi-tasker must
\ ** be turned off and *\fo{EMIT) and friends must run in polled
\ ** mode.
  -1 to FaultConsole?
  [ tasking? ] [if]  single  [then]
;

: (qseremit)	\ char base queue --
\ *G Send a character on the UART whose base address and queue are given.
  FaultConsole? 0= if
    begin  dup cqfull? while  pause  repeat
    [i rot over (>cqueue) TxQ>FIFO i]	\ put char in buffer
    exit
  endif
  drop					\ discard queue
  $00 over Un.IM + !			\ disable TX and RX interrupts
  begin					\ poll LSR
    dup Un.FR + @ bit5 and 0=
  until
  Un.DR + !				\ send
;

: (qsertype)	\ caddr len base queue --
\ *G Send a string on the UART whose base address and queue are given.
  2swap bounds				\ -- base queue
  ?do  2dup i c@ -rot (qseremit)  loop
  2drop
;

: (qsercr)	\ base queue --
\ *G Send a CR/LF on the UART whose base address and queue are given.
  2dup $0D -rot (qseremit)  $0A -rot (qseremit)
;


: initUART	\ divisor22 base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r				\ save divisor
  0 over Un.CTL + !			\ disable everything
  5 over Un.CC + !			\ use AltClk=PIOSC for baud rate
  r@ 6 rshift $FFFF and over Un.IBRD + ! \ integer baud rate
  r> $3F and over Un.FBRD + !		\ fractional baud rate,
  $70 over Un.LCRH + !			\ 8N1, FIFOs
  $12 over Un.IFLS + !			\ 1/2 FIFO level
  $70 over Un.IM + !			\ RX timeout, RxFIFO, Tx
  0 over Un.DMACTL + !			\ no DMA
  $0300 $CFBF 2 pick Un.CTL + setMask	\ set RXE=1, TXE=1, HSE=0
  1 swap Un.CTL + or!			\ UARTEN=1
;

external


\ ********
\ *S UART0
\ ********

useUART0? [if]
console0-speed GenBRD22 equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

#128 cqueue: SerInpQ0	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ0	\ -- addr
\ *G Transmitter circular byte queue.

: ser0-isrh	\ --
\ *G UART0 high level ISR. INTERNAL.
  _UART0 dup Un.MIS + @			\ read status
  tuck swap Un.ICR + !			\ clear interrupts
  dup $050 and if
    _UART0 SerInpQ0 FIFO>RxQ
  endif
  $020 and if
    _UART0 SerOutQ0 TxQ>FIFO
  endif
;
' ser0-isrh UART0_vec# EXC: U0entry

[undefined] init-ser0 [if]
: init-ser0	\ --
\ *G Initialise serial port 0, using U0Rx=PA0, U0Tx=PA1
  di
  SerInpQ0 init-cqueue  SerOutQ0 init-cqueue	\ initialise character queues
  _SysCtl
  bit0 over sc.RCGCUART + or!		\ enable clock and power to UART0
  bit0 swap sc.RCGCGPIO + or!		\ enable clock and power to GPIOA
  noop noop noop noop			\ must wait a bit
  _GPIOA
  $02 $03 2 pick gpio.DIR + setmask	\ bit 1 o/p, bit 0 i/p *
  $03 over gpio.AFSEL + or!		\ pins 0/1 are peripheral
  $03 over gpio.DR2R + or!		\ pins 0/1 are 2mA drive *
  $03 over gpio.SLR + bic!		\ pins 0/1 have no slew control *
  $03 over gpio.DEN + or!		\ enable pins
  $03 over gpio.ODR + bic!		\ no open drain *
  $03 over gpio.PUR + bic!		\ no pull ups *
  $03 over gpio.PDR + bic!		\ no pull downs *
  $11 $FF  rot gpio.PCTL + setMask	\ peripheral 1 for PA0/1
  u0-divisor _UART0 initUART		\ set up UART
  UART0_vec# EnInt  ei			\ interrupt enable in NVIC
;
[then]

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 SerOutQ0 (qserEmit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 SerOutQ0 (qsertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 SerOutQ0 (qsercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  SerInpQ0 cqnotempty?  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  SerInpQ0 cqueue>  ;

create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line

console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]	\ console port

[then]	\ useUART0?


\ ********
\ *S UART1
\ ********

useUART1? [if]

console1-speed GenBRD22 equ u1-divisor	\ -- n
\ *G UART1 baud rate divisor.

#128 cqueue: SerInpQ1	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ1	\ -- addr
\ *G Transmitter circular byte queue.

: ser1-isrh	\ --
\ *G UART1 high level ISR. INTERNAL.
  _UART1 dup Un.MIS + @			\ read status
  tuck swap Un.ICR + !			\ clear interrupts
  dup $050 and if
    _UART1 SerInpQ1 FIFO>RxQ
  endif
  $020 and if
    _UART1 SerOutQ1 TxQ>FIFO
  endif
;
' ser1-isrh UART1_vec# EXC: U1entry

[undefined] init-ser1 [if]
: init-ser1	\ --
\ *G Initialise serial port 1, using U1Rx=PB0, U1Tx=PB1
  _SysCtl
  bit1 over sc.RCGCUART + or!		\ enable clock and power to UART1
  bit1 swap sc.RCGCGPIO + or!		\ enable clock and power to GPIOB
  noop noop noop noop			\ must wait a bit
  _GPIOB
  $03 over gpio.AFSEL + or!		\ pins 0/1 are peripheral
  $03 over gpio.DR2R + or!		\ pins 0/1 are 2mA drive
  $03 over gpio.SLR + bic!		\ pins 0/1 have no slew control
  $03 over gpio.DEN + or!		\ enable pins
  $03 over gpio.ODR + bic!		\ no open drain *
  $03 over gpio.PUR + bic!		\ no pull ups *
  $03 over gpio.PDR + bic!		\ no pull downs *
  $0011 $00FF rot gpioPCTL + setMask	\ peripheral 1 for PA0/1
  u1-divisor _UART1 initUART
  UART1_vec# EnInt  ei			\ interrupt enable in NVIC
;
[then]

: seremit1	\ char --
\ *G Transmit a character on UART1.
  _UART1 SerOutQ1 (qserEmit)  ;
: sertype1	\ c-addr len --
\ *G Transmit a string on UART1.
  _UART1 SerOutQ1 (qsertype)  ;
: sercr1	\ --
\ *G Issue a CR/LF pair to UART1.
  _UART1 SerOutQ1 (qsercr)  ;
: serkey?1	\ -- flag
\ *G Return true if UART1 has a character available.
  SerInpQ1 cqnotempty?  ;
: serkey1	\ -- char
\ *G Wait for a character on UART1.
  SerInpQ1 cqueue>  ;

create Console1	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART1.
  ' serkey1 ,		\ -- char ; receive char
  ' serkey?1 ,		\ -- flag ; check receive char
  ' seremit1 ,		\ -- char ; display char
  ' sertype1 ,		\ caddr len -- ; display string
  ' sercr1 ,		\ -- ; display new line

console-port 1 = [if]
  console1 constant console
[then]

[then]


\ ********
\ *S UART2
\ ********

useUART2? [if]

console2-speed GenBRD22 equ u2-divisor	\ -- n
\ *G UART2 baud rate divisor.

#128 cqueue: SerInpQ2	\ -- addr
\ *G Receiver circular byte queue.
#128 cqueue: SerOutQ2	\ -- addr
\ *G Transmitter circular byte queue.

: ser2-isrh	\ --
\ *G UART2 high level ISR. INTERNAL.
  _UART2 dup Un.MIS + @			\ read status
  tuck swap Un.ICR + !			\ clear interrupts
  dup $050 and if
    _UART2 SerInpQ2 FIFO>RxQ
  endif
  $020 and if
    _UART0 SerOutQ2 TxQ>FIFO
  endif
;
' ser2-isrh UART2_vec# EXC: U2entry

[undefined] init-ser2 [if]
: init-ser2	\ --
\ *G Initialise serial port 2, using U2Rx=PA6, U2Tx=PA7
  _SysCtl
  bit2 over sc.RCGCUART + or!		\ enable clock and power to UART2
  bit0 swap sc.RCGCGPIO + or!		\ enable clock and power to GPIOA
  noop noop noop noop			\ must wait a bit
  _GPIOA
  $C0 over gpio.AFSEL + or!		\ pins 6/7 are peripheral
  $C0 over gpio.DR2R + or!		\ pins 6/7 are 2mA drive
  $C0 over gpio.SLR + bic!		\ pins 6/7 have no slew control
  $C0 over gpio.DEN + or!		\ enable pins
  $C0 over gpio.ODR + bic!		\ no open drain *
  $C0 over gpio.PUR + bic!		\ no pull ups *
  $C0 over gpio.PDR + bic!		\ no pull downs *
  $1100:0000 $FF00:0000 rot gpioPCTL + setMask	\ peripheral 1 for PA6/7
  u2-divisor _UART2 initUART
  UART2_vec# EnInt  ei			\ interrupt enable in NVIC
;
[then]

: seremit2	\ char --
\ *G Transmit a character on UART2.
  _UART2 SerOutQ2 (qserEmit)  ;
: sertype2	\ c-addr len --
\ *G Transmit a string on UART2.
  _UART2 SerOutQ2 (qsertype)  ;
: sercr2	\ --
\ *G Issue a CR/LF pair to UART2.
  _UART2 SerOutQ2 (qsercr)  ;
: serkey?2	\ -- flag
\ *G Return true if UART2 has a character available.
  SerInpQ2 cqnotempty?  ;
: serkey2	\ -- char
\ *G Wait for a character on UART2.
  SerInpQ2 cqueue>  ;

create Console2	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART2.
  ' serkey2 ,		\ -- char ; receive char
  ' serkey?2 ,		\ -- flag ; check receive char
  ' seremit2 ,		\ -- char ; display char
  ' sertype2 ,		\ caddr len -- ; display string
  ' sercr2 ,		\ -- ; display new line

console-port 2 = [if]
  console2 constant console
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all used serial ports.
  [ useUART0? ] [if]  init-ser0  [then]
  [ useUART1? ] [if]  init-ser1  [then]
  [ useUART2? ] [if]  init-ser2  [then]
;


\ ======
\ *> ###
\ ======

decimal
