\ Virtual UART (version 2) for ARM/Cortex systems

((
Copyright (c) 2003, 2022
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

To do
=====

Change history
==============
Updated to 16 byte block.
))


\ ==============
\ *! serpsuart
\ *T Polled PSUART/VUART2 serial driver
\ ==============
\ *P This driver provides polled comms for systems using a Segger
\ ** J-Link, MSP-FET or other JTAG/SWD unit to implement the host side.


\ ****************
\ *S VUART introduction
\ ****************
\ *P The VUART is a virtual UART. It consists of a data structure accessed by
\ ** a JTAG debugger (Segger L-Link, or MSP-FET) on the up side, and by the target CPU
\ ** on the down side. The structure consists of two counted strings of
\ ** maximum data size seven bytes, for a total of 16 bytes.

struct /vuart2	\ -- len
\ *G Structure of virtual UART data block.
\ *[
  1 field v2.rfuup	\ RFU
  1 field v2.#up	\ num bytes to go up to host. Target write, host clear
  7 field v2.upbuff	\ data to go up to host
  1 field v2 rfudn	\ RFU
  1 field v2.#dn	\ num bytes down, available from host. Host write, target clear
  7 field v2.dnbuff	\ data down from host
end-struct
\ *]

/vuart2 buffer: VUART	\ -- addr
\ *G Virtual UART data block. 
	

\ *********
\ *S VUART code
\ *********

: initVuart	\ --
\ *G Initialise the VUART.
  VUART /vuart2 erase
;

: key?v		\ -- num ; 0..7
\ *G Return nz if the VUART has received one or more characters.
  VUART v2.#dn c@
;

: keyv		\ -- char
\ *G Wait for a character from the VUART.
  begin  key?v until
  VUART dup v2.dnbuff c@ >r		\ -- vuart ; R: -- char
  dup v2.#dn c@ 1 > if			\ -- vuart ; R: -- char
    dup v2.dnbuff 1+  over v2.dnbuff	\ copy data down
    2 pick v2.#dn c@ 1- cmove		\ -- vuart ; R: -- char
  then
  dup v2.#dn c@ 1- swap v2.#dn c!	\ -- ; decrement #chars
  r>					\ -- char	
;

: waitUPdone	\ vuart -- vuart 
\ *G Wait until all up characters have gone.
  begin	 dup v2.#up c@ 0=  until	\ wait for previous chars to go
;

: emitv		\ char --
\ *G Send a character to the VUART.
  VUART	waitUPdone			\ -- char VUART
  tuck v2.upbuff c!			\ put character in  
  1 swap v2.#up c!			\ count=1
;

: send7up	\ caddr len vuart -- caddr' len' vuart
\ *G Send seven characters to the VUART.
  waitUPdone				\ let previous block finish
  -rot  2dup 7 /string 2>r		\ -- vuart caddr len ; R: -- caddr' len'
  drop  over v2.upbuff  7 cmove		\ -- vuart ; 7 bytes to buffer
  7 over v2.#up c!			\ set buffer size
  2r> rot				\ -- caddr' len' vuart
;  

: sendNup	\ caddr len vuart -- 
  waitUPdone				\ let previous block finish
  2dup 2>r				\ -- caddr len vuart ; R: -- len vuart 
  v2.upbuff swap cmove			\ -- vuart ; len bytes to buffer
  2r> v2.#up c!				\ set buffer size
;  

: typev		\ c-addr len --
\ *G Type a string to the VUART
  VUART 				\ -- caddr len vuart 
  begin  
    over 7 >
   while				\ -- caddr' len' vuart
    send7up
  repeat
  over if  sendNup  else  3drop  then
;

: crv		\ --
\ *G Send CR/LF through the VUART.
  $0D emitv  $0A emitv
;


\ ***********************
\ *S Generic i/o assignments
\ ***********************

cdata
create ConsoleV	\ -- addr ; OUT managed by upper driver
\ *G Data structure holding function pointers.
  ' keyv ,		\ -- char ; receive char
  ' key?v ,		\ -- flag ; check receive char
  ' emitv ,		\ -- char ; display char
  ' typev ,		\ caddr len -- ; display string
  ' crv ,		\ -- ; display new line

console-port #15 = [if]
  ConsoleV constant Console
[then]

: init-ser	\ --
\ *G Initialise all serial ports
  initVuart
;


\ ======
\ *> ###
\ ======

decimal

decimal

