\ spiSTM32hw.fth - SPI for STM32F4

((
Copyright (c) 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20121016 MPE001 Initial release.
))

only forth definitions
decimal

\ =================
\ *! stm32hw
\ *T STM32 SPI driver
\ =================
\ *P STM32 devices and boards vary in their use of various pins
\ ** on the SD card, in particular the CS output and the CD (card
\ ** detect) and WP (write protect) inputs. Consequently this file
\ ** is not generic. The default configuration is for an
\ ** Olimex LPC1766-STK card.

\ *P The standard SD/MMC card pinout is as follows.
\ *E         SD mode                                 SPI mode
\ ** Pin     Name    Type    Description             Name    Type    Description
\ ** 1       CD/DAT3 I/O/PP  Card detect / data 3    CS      I       Chip select (low)
\ ** 2       CMD     PP      Command/Response line   DI      I       Data input
\ ** 3       Vss1    S       Supply voltage (gnd)    VSS     S       Supply voltage
\ ** 4       Vdd     S       Power supply            VDD     S       Power supply
\ ** 5       CLK     I       Clock                   SCLK    I       Clock
\ ** 6       Vss2    S       Supply voltage          VSS2    S       Supply voltage
\ ** 7       DAT0    I/O/PP  data line 0             O       O/PP    Data output
\ ** 8       DAT1    I/O/PP  data line 1             RSV
\ ** 9       DAT2    I/O/PP  data line 2             RSV
\ *P where the pins are numbered on the connector as below:
\ *E   9 1 2 3 4 5 6 7 8
\ **     * * * * * * * *
\ **   *


\ ***************************
\ *S Setup and initialisation
\ ***************************
\ *P The board uses I/O for SPI as follows.
\ *E   PI.0   AF5   SSEL1      pin 131   SPI2_NSS
\ **   PI.1   AF5   SCK1       pin 132   SPI2_SCK
\ **   PI.2   AF5   MISO1      pin 133   SPI2_MISO
\ **   PI.3   AF5   MOS11      pin 134   SPI2_MOSI

\ SPI1 - APB2
\ SPI2 - APB1
\ SPI3 - APB1

: genSPIbr	\ spi-speed bus-speed -- brx
  swap /  0 7 do			\ -- divisor
    dup 1 i lshift >
    if  drop i unloop exit  endif
  -1 +loop
  drop 0
;

interpreter
: genSPIbr	\ spi-speed bus-speed -- brx
  swap /  0 7 do			\ -- divisor
    dup 1 i lshift >
    if  drop i unloop exit  endif
  -1 +loop
  drop 0
;
target

: (initSPI)	\ freq busfreq *spi --
\ *G Set up the given SPI device, where *\i{freq} is the desired
\ ** SPI speed, *\i{busfreq} is the peripheral clock speed, and
\ ** *\i{*spi} is the SPI peripheral base address. The SPI bus is
\ ** set up in master mode, MSB first.
  >r
  genSPIbr 3 lshift $0004 +		\ initial value of spiCR1
  r@ spiCR1 + !				\ main mode control, CPOL=0, CPHA=0
  0 r@ spiCR2 + !
  bit6 r> spiCR1 + or!			\ set SPE bit
;


\ *******
\ *S SPI2
\ *******

[undefined] selio-spi2 [if]
8 0 pio: SPI2_NSS
8 1 pio: SPI2_SCK
8 2 pio: SPI2_MISO
8 3 pio: SPI2_MOSI

: selio-spi2	\ --
\ *G Example to perform clock and pin selection for SPI2 on PI0..3.
\ Clocks and reset
  _RCC
  bit8 over rccAHB1ENR + or!		\ enable clock to GPIOI
  bit8 over rccAHB1RSTR + bic!		\ take it out of reset
  bit14 over rccAPB1ENR + or!		\ enable clock to SPI2
  bit14 over rccAPB1RSTR + bic!		\ take it out of reset
  drop
\ Default on PI0,1,2,3
  5 SPI2_NSS  isFunction
  5 SPI2_SCK  isFunction
  5 SPI2_MISO isFunction
  5 SPI2_MOSI isFunction
;
[then]

: InitSPI2	\ --
\ *G Initialise SPI2.
  selio-SPI2
  1 MHz APB1-speed _SPI2 (initSPI)
;


\ ****************
\ *S SPI transfers
\ ****************
\ *P These routines have been optimised for speed over space
\ ** considerations. For high speed use with SD cards, you will
\ ** get best performance using interrupt driven or DMA driven
\ ** operation.

(( SPI status register
Bit	Symbol	Description
0	RXNE	Receive FIFO Not Empty. This bit is 0 if the Receive
		FIFO is empty, 1 if not.
1 	TXE	Transmit FIFO Empty. This bit is 1 is the Transmit FIFO is
		empty.
7	BSY	Busy. This bit is 0 if the SSP controller is idle,
		or 1 if it is currently sending/receiving a frame
		and/or the Tx FIFO is not empty.
))

compiler
: SPIXchg	\ bout *spi -- bin
\ *G Write data to SPIx and return the received byte.
  tuck spiDR + !		\ -- *spi ; Write to device
  begin				\ Wait for TXE=1, RXNE=1
    dup spiSR + @ 3 and 3 =
  until
  spiDR + @ $FF and		\ return received byte
;
target

IsLeaf : spiFlush	\ --
\ *G Flush the SPI2 input queue.
  _SPI2 begin
    dup spiSR + @ bit0 and
   while
    dup spiDR + @ drop
  repeat
  drop
;

IsLeaf : spi!	\ b --
\ *G Write data to SPI2 and discard the received byte.
  _SPI2 SPIXchg drop
;

IsLeaf : spi@	\ -- b
\ *G Write an idle byte to SPI2 and return the received byte.
  $FF _SPI2 SPIXchg
;

((
  IsLeaf : SPIidle	\ --
  \ *G Send an 0xFF byte to SPI2 and discard the received byte.
    $FF _SPI2 SPIXchg drop
  ;

  IsLeaf : spiwr	\ addr -- addr'
  \ *G Write byte at *\i{addr} to SPI2 and increment pointer.
    dup c@ _SPI2 SPIXchg drop  1+
  ;

  IsLeaf : spird	\ addr -- addr'
  \ *G Read byte from SPI2, put it at *\i{addr}, increment pointer.
    $FF _SPI2 SPIXchg over c!  1+
  ;

  : MMCon		\ --
  \ *G Assert CS.
    bit6 _FIO0 FIOCLR + !		\ Set pin low
  ;

  : MMCoff	\ --
  \ *G Negate CS.
    bit6 _FIO0 FIOSET + !		\ Set pin high
  ;

  : CDin?		\ -- flag
  \ *G Return non-zero if a card is present.
    -1
  ;

  : WPin?		\ -- flag
  \ *G Return non-zero if the write protect slider is set.
    0
  ;

  : initSDspi	\ --
  \ *G Initialise SPI2 for SD/MMC card use.
    #400000 initSPI2
  ;

  : spiSlow	\ --
  \ *G Set the SPI2 speed to the SD/MMC low speed of 400kHz.
    #400000 initSPI2
  ;

  : spiFast	\ --
  \ *G Set the SPI2 speed to the SD/MMC operating speed of 2 MHz.
  \ ** Note that the high speed is often limited by the quality
  \ ** of the hardware interface. At higher speeds termination
  \ ** resistor values and PCB tracking need attention.
  \ ** By default, the high speed is set to 2MHz, which works on
  \ ** most hardware. Tune this speed as required.
    #2000000 initSPI2
  ;
))


\ ======
\ *> ###
\ ======

decimal
