\ FlashProg.fth - Program LPC11xx Flash

((
Copyright (c) 2004, 2008, 2010, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
At end of programming, check for a partial block to program.

Change history
==============
20110628 MPE006 LPC11xx conversion
20100125 MPE005 Updated for LPC17xx
20081016 MPE004 Used intermediate buffer to avoid IAP problems
		with some devices.
20081008 MPE003 Converted for LPC23xx/24xx/28xx.
20050627 MPE002 Tables separated from FlashProg.fth
20040511 MPE001 First release
))

only forth definitions  decimal


\ ===========
\ *! flashlpc
\ *T LPC17xx Flash tools
\ ===========
\ *P These tools are provided for applications which reserve
\ ** part of the Flash for data. This code uses the IAP routines
\ ** in *\i{IAP17xx.fth}.

\ *P N.B. An erase causes the whole of the sector containing that
\ ** address to be erased.

\ *P N.B. All writes to Flash must be from RAM as the Flash
\ ** is unavailable at the time any part is being written.


\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the LPC1100 series, they are designed to be easily expanded
\ ** for future LPC11xx parts with as yet unknown sector sizes.

\ *P Sector tables contain the number of sectors and starting offset
\ ** of each sector, plus a dummy start address which enables
\ ** the size of the last sector to be calculated.
\ ** The boot block sector is *\b{not} included in the table.
\ ** The sector table is given by *\fo{SecTab} which is defined in
\ ** *\i{FlashTables.fth}.

: SectorN	\ n -- addr len
\ *G Convert sector number (zero based) to base address and length
  1+ cells sectab + 2@			\ -- end start
  tuck -
;

: FindSecN	\ addr -- n
\ *G Find the sector number containing address addr. If
\ ** addr is outside the internal Flash range, n is set to -1.
  sectab @ 0 ?do			\ -- addr
    dup i SectorN over + within if	\ test sector
      drop i  unloop exit
    endif
  loop
  drop -1
;

: .src/dest	\ src dest -- src dest
\ *G Display the addresses and contents of the source
\ ** and destination.
  ."  source: " over .lword  over c@ ." (" .byte ." )"
  ."  dest: " dup .lword  dup c@ ." (" .byte ." )"
;


\ ***************
\ *S Flash driver
\ ***************

cell buffer: FlDest	\ -- addr
\ *G Holds next destination address.

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count.

: (Prog512)	\ src dest --
\ *G Program 512 bytes at *\i{src} to *\i{dest}. The *\i{src}
\ ** address must be in RAM. On error, the variable *\fo{PrgErrs}
\ ** is incremented.
  #10 0 do
    dup FindSecN dup IAPprep 0= if	\ prep sector
      2dup #512 IAPcopy 0=		\ write data
      if  2drop  unloop exit  endif
    endif
  loop
  2drop  #PrgErrs incr
;

: Prog512	\ src dest --
\ *G Program 512 bytes at src to dest. Increment error
\ ** counter on error. *\fo{ISPram} is used as an intermediate
\ ** buffer to avoid avoid problems/bugs in some LPC devices.
\ cr ." Prog512 " over .dword ."  to " dup .dword
\  swap ISPram #512 cmove  ISPram swap	\ use temporary buffer
  (Prog512)
;

: (EraseFlash)	\ dest dlen --
\ *G Erase the flash for the given range. Note that complete
\ ** sectors in the range will be erased. If you want to write
\ ** partial sectors, check that they contain $FF in all bytes
\ ** before programming in order avoid having to erase them first.
  over + 1-				\ -- first last
  swap FindSecN swap FindSecN		\ -- nf nl
  #10 0 do
    2dup IAPprep 0= if
      2dup IAPerase 0=
      if  2drop  unloop exit  endif
    endif
  loop
  2drop
;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 512 byte boundary and
\ ** *\i{len} is rounded up to the next 512 byte unit. *\i{Src}
\ ** must be on a word boundary.
  rot -4 and -rot			\ force src
  swap #-512 and swap			\ force dest
  #511 + #-512 and			\ force len
  bounds ?do				\ -- src'
    dup i Prog512
    #512 +
  #512 +loop
  drop
;

: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over c@ over c@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    1 + swap  1 + swap
  1 +loop
  2drop
;

: ProgramFlash	\ src dest len --
\ *G Using the RAM memory buffer at *\i{src}, program the Flash
\ ** at *\i{dest} with *\i{len} bytes. The relevant Flash sectors
\ ** are erased, the Flash is programmed, and the result verified.
  cr ." Erasing " 2dup (EraseFlash)
  0 #PrgErrs !
  cr ." Programming "  3dup (ProgFlash)
  space #PrgErrs ? ." errors"
  cr ." Verifying "  (VerifyFlash)
;


\ ======
\ *> ###
\ ======

