\ ReprogApp.fth - reprogram LPC11xx Flash

((
Copyright (c) 2004, 2008, 2009, 2010, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
At end of download, check for a partial block to program.

Change max program size to depend on CPU.

Change history
==============
20110628 SFP006 LPC11xx conversion.
20100125 SFP005 LPC17xx conversion.
20090205 SFP004 LPC2388 does not erase last sector at $0007:D000
20081008 SFP003 Minor documentation changes.
20060701 SFP002 Added CopyFlash.
20050627 MPE002 Now uses sector tables.
20040212 MPE001 First release
))

only forth definitions  decimal


\ ****************
\ *S Flash handler
\ ****************

cell buffer: FlDest	\ -- addr
\ *G Holds next destination address

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count

0 equ Start#	\ -- n
\ *G First sector to erase/program

: End#		\ -- n
\ *G not LPC2388: Last (inclusive) sector to erase/program.
  sectab @ 1-  ;

: SectorRange	\ -- start end
\ *G The range of sectors to be handled.
  Start# End#
;

: SectorN	\ n -- addr len
\ *G Convert sector number (zero based) to base address and length
  1+ cells sectab + 2@			\ -- end start
  tuck -
;

: FindSecN	\ addr -- n
\ *G Find the sector number containing address addr. If
\ ** addr is outside the internal Flash range, n is set to -1.
  sectab @ 0 ?do			\ -- addr
    dup i SectorN over + within if	\ test sector
      drop i  unloop exit
    endif
  loop
  drop -1
;

: (EraseFlash)	\ nf nl --
\ *G Erase the Flash from first to last sector.
  #10 0 do			\ -- nf nl
    2dup IAPprep 0= if
      2dup IAPerase 0=
      if  2drop  unloop exit  endif
    endif
  loop
  2drop  #PrgErrs incr
;

: EraseFlash	\ --
\ *G Erase the Flash from start to end.
  SectorRange (EraseFlash)
;

: Prog512	\ src dest --
\ *G Program 512 bytes at src to dest. Increment error
\ ** counter on error.
  #10 0 do
    dup FindSecN dup IAPprep 0= if	\ prep sector
      2dup #512 IAPcopy 0=		\ write data
      if  2drop  unloop exit  endif
    endif
  loop
  2drop  #PrgErrs incr
;

: InitXBuff	\ --
\ *G Initialise the Flash buffer
  Xbase ^FlBuffer !
;

: toFlash	\ src dest --
\ *G Handle 128 bytes at src to go to dest in Flash.
  swap #128 +				\ update RAM pointer
  dup Xtop = if				\ got complete unit?
    2drop
    Xbase FlDest @ Prog512		\ program unit
    InitXbuff  #512 FlDest +!		\ update pointers
  else
    ^FlBuffer !  drop
  endif
;


\ *****************
\ *S Numeric output
\ *****************

: .nibble	\ n --
\ *G Convert a nibble to a hex ASCII digit and display it.
  $0F and  dup 9 >
  if  7 +  then
  $30 + emit
;

: .WORD		\ w --
\ *G Display w as a 4 digit unsigned hexadecimal number.
  dup #12 rshift .nibble
  dup 8 rshift .nibble
  dup 4 rshift .nibble
  .nibble
;

((
: .dword	\ x --
\ *G Display x as an 8 digit unsigned hexadecimal number.
  dup #16 rshift .word .word
;
))


\ ***************
\ *S Reprogrammer
\ ***************

l: m1
  ", Erasing "
l: m2
  ",  done "
l: m3
  ", Start Xmodem TX "
[defined] REBOOT [if]
l: m4
  ", Press R to reboot, other to reprogram "
[else]
l: m4
  ", Reset or press key to reprogram"
[then]
l: m5
  ", Errors"
l: m6
  ", No "

: $.	\ c-addr --
\ *G Display counted string.
  count type
;

: .PrgErrs
\ *G Display number of errors.
  cr  #PrgErrs @ .word  bl emit
  m5 $.
;

: initReprog	\ --
\ *G Get the setup data and calculate the requirements
  CLD_#FlSectors @ sectab !		\ number of sectors in the Flash
  initDelays				\ initialise the delays
  init-ser #50 ms
;

: ReprogFlash	\ --
\ *G Erase and program the Flash.
  initReprog
  begin
    InitXbuff  #PrgErrs off
    cr m1 $.  #20 ms
    EraseFlash
    m2 $.  .PrgErrs  #PrgErrs off
    cr m3 $.  $01A emit			\ send message and trigger
    0 dup FlDest !  $00400000 Bin-Up drop
    .PrgErrs
    #100 ms
    cr m4 $.
[defined] REBOOT [if]
    key $DF and dup emit  [char] R =
  until
  #100 ms
  reboot
[else]
    key drop
  again
[then]
;

((
\ SFP002...
l: msg7
  ", Copy "

: (CopyFlash)	\ src dest len --
\ *G Copy one Flash area to another. The parameters are as for *\fo{CMOVE}.
  initReprog
  cr m1 $.  #20 ms			\ "Erase" message
  #PrgErrs off
  2dup + 1- >r				\ -- src dest len ; R: -- last
  over FindSecN  r> FindsecN		\ -- src dest len nf nl
  (EraseFlash) .PrgErrs  #10 ms		\ -- src dest len
  #PrgErrs @
  if  3drop exit  endif
  cr msg7 $.				\ "Copy" message
  0 do					\ -- src dest
    over Xbase #512 cmove
    Xbase over Prog512
    #512 +  swap #512 + swap
  #512 +loop
  2drop
  .PrgErrs  #20 ms			\ error count
;

: CopyFlash	\ src dest len --
\ *G Copy the flash. The parameters are as for *\fo{CMOVE}.
\ ** The process is repeated until there are no errors, and
\ ** the system is then rebooted.
  begin
    3dup (CopyFlash)  #PrgErrs @ 0=
  until
  #100 ms  reboot
;
\ ...SFP002
))


\ ======
\ *> ###
\ ======

