\ Driver for on-board LCD Controller for the LPC1788 Family

((
	Barry Culver
	Culver Consultancy Ltd.
	46 Langley Crescent
	St. Albans
	Hertfordshire
	U.K.
	www.culverconsultancy.co.uk

	This driver is based on the C code supplied by Olimex for the LPC1888 demo
	board.

	It uses 8 bits per pixel mode with a LUT for 256 colours.  This is to reduce
	bus usage by LCD DMA.  Frame buffers are in external SDRAM using bank 0.

	The code assumes that the backlight controller is enabled by a low on port pin
	P0[10] and brightness controlled by a PWM output on P2[01], low duty cycle
	gives high brightness.

	The following equates, defined externally, are required.  This reduces the
	need to edit this file.
		glcd-clock-speed					(pixel clock speed in Hz)
		glcd-pixels-wide					(in pixels)
		glcd-pixels-high					(in pixels)
		(the following timings are in pixel clocks)
		glcd-h-pulse
		glcd-h-front-porch
		glcd-h-back-porch
		glcd-v-pulse
		glcd-v-front-porch
		glcd-v-back-porch


To do
=====

Change history
==============
20140117 BJC001	Created
))

only forth definitions
decimal


\ =======================================================================
\ PIN CONFIGURATION
\ =======================================================================

 0 10 pio: GLCD_BL_EN		\ high active backlight enable SEE ALSO GLCD-INIT-PINS
 IOCON_P0_10 equ GLCD_BL_EN_REG
 IOCON_P2_01 equ GLCD_BL_BRITE

\ =======================================================================
\ LCD DISPLAY CONFIGURATION
\ =======================================================================

\ Generate the LCD_CTLR register value - ASSUMES A TFT DISPLAY
  0 ~LCDCTRL_8BPP or			\ 8 bits per pixel
  ~LCDCTRL_TFT or			\ .. a TFT display
  ~LCDCTRL_RGB or			\ standard RGB colour order
  ~LCDCTRL_LE_BYTE or			\ little endian format frame store data
  ~LCDCTRL_LCD_OFF or			\ do not power up LCD display module
  equ LCDctrlRegVal

\ Calculate the LCD pixel clock prescaler value - we are using the internal clock
  system-speed glcd-clock-speed / 1-
  equ LCDcfgRegVal
  cr ." Actual GLCD clock speed = " system-speed LCDcfgRegVal 1+ / u.

\ Generate the LCD_POL register value - ASSUMES A TFT DISPLAY
  0 +LCD_POL_BCD			\ bypass the clock divider (for TFT)
  +LCD_POL_INV_VS			\ vertical sync is active low (not used?)
  +LCD_POL_INV_HS			\ horizontal sync is active low (not used?)
  glcd-pixels-wide 1- +LCD_POL_CPL	\ clocks per line
  +LCD_POL_INV_PCLK			\ output clock is low edge active
  equ LcdPolRegVal

\ Generate LCD_TIMH reg. horizontal timing  value
  0
  glcd-h-front-porch +LCD_TIMH_HF_PORCH
  glcd-h-back-porch +LCD_TIMH_HB_PORCH
  glcd-h-pulse +LCD_TIMH_HS_WIDTH
  glcd-pixels-wide +LCD_TIMH_PPL
  equ LcdTimhRegVal

\ Generate LCD_TIMV reg. vertical timing  value
  0
  glcd-v-front-porch +LCD_TIMV_VF_PORCH
  glcd-v-back-porch +LCD_TIMV_VB_PORCH
  glcd-v-pulse +LCD_TIMV_VS_WIDTH
  glcd-pixels-high +LCD_TIMV_LPP
  equ LcdTimvRegVal


\ =======================================================================
\ LCD FRAME BUFFERS & VARIABLES
\ =======================================================================

DMB0base equ GLCD_RAM_ADDR

glcd-pixels-wide glcd-pixels-high *	\ size of frame buffer in pixels
1 *					\ bytes per pixel; use 4 for true colour
  constant glcd-buffer-size

GLCD_RAM_ADDR daligned
dup constant GLCD_BUFF0
glcd-buffer-size + daligned
dup constant GLCD_BUFF1
glcd-buffer-size + daligned
dup constant GLCD_BUFF2
glcd-buffer-size + daligned
constant GLCD_BUFF3

4 constant GLCD_#PAGES

255 VALUE Ink
0 VALUE	Paper
VARIABLE ^DBuffer_L	\ points to buffer used for drawing


\ =======================================================================
\ INITIALISATION
\ =======================================================================
\ Initialise backlight using PWM brightness control
\ The backlight PWM runs at 1000Hz and the timer full scale is 1000 counts
\ so we need a 400kHz clock for the timer register.

#100 equ BrightPWMfs		\ using 100 gives us percent control value
periph-speed BrightPWMfs 1000 * / 1-
  equ BrightPWMprescale

: (BACKLIGHT-INIT)
  _IOCON			\ base
  IO_GPIO_FUNC over GLCD_BL_EN_REG + !	\ backlight enable driven as GPIO
  GLCD_BL_EN IsOutput
  GLCD_BL_EN clrPin			\ backlight is off
  IO_PERIPH_FUNC1 swap GLCD_BL_BRITE + ! \ brightness control pin PWM function
  ~pcPWM1 _PCONP or!			\ enable power to PWM block
  _PWM1					\ base
  bit1 over PWMnTCR + !			\ set disabled, reset - known state
  BrightPWMprescale over PWMnPR + !	\ set prescaler
  BrightPWMfs 1+ over PWMnMR0 + !	\ match register 0 sets full scale count
  #0 over PWMnMR2 + !			\ match register 2 controls the pin
  bit1 over PWMnMCR + !			\ reset timer on match with MR0
  bit10 over PWMnPCR + !		\ enable output 2
  bit0 bit3 or over PWMnTCR + !		\ start running
  bit0 bit2 or over PWMnLER + !		\ enable latch of match register values
  drop					\ base
;
' (backlight-init) AtCold

: SET-BACKLIGHT	\ bright | - ; set backlight brightness 0=off .. 100=max
  dup 0= if				\ .. turn it off
    drop GLCD_BL_EN clrPin		\ backlight is off
  else
    _PWM1				\ base
    swap 0 max 100 min			\ clamp to limits
    BrightPWMfs swap -			\ reverse due to hardware inversion
    over PWMnMR2 + !			\ set match register
    bit2 swap PWMnLER + !		\ update match reg at next match
    GLCD_BL_EN setPin			\ backlight is on
  endif
;

: GLCD-INIT-PINS	\ - | - ; initialise IO to drive LCD display
  _IOCON			\ base
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P0_04 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P0_05 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P0_06 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P0_07 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P0_08 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P0_09 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_20 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_21 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_22 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_23 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_24 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_25 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_26 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_27 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_28 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P1_29 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_00 + !		\ LCD power
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_01 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_02 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_03 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_04 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_05 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_06 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_07 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_08 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_09 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_12 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P2_13 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P4_28 + !
  IO_PERIPH_FUNC7 +HYSTERESIS over IOCON_P4_29 + !
\ set backlight, turn off
  drop			\ .. base
;
  ' glcd-init-pins atcold	\ note: we set up the pins only here
			\ only initialise LCD controller *after* EMC is running!
: PALETTE-VALUE		\ colour# | pvalue
  dup 6 rshift 1+ 5 * 2/ >r		\ intensity modifier now 2..10
  dup 3 and 				\ red - will occupy palette bits 0..4
  r@ *
  over $C and 2 rshift
  r@ *
  5 lshift or				\ green - will occupy palette bits 5..9
  swap $30 and 4 rshift
  r> *
  10 lshift or				\ blue  - will occupy palette bits 10..14
;

\ This word sets a basic palette which is suitable for simple graphics
\ that do not require many colours.
\ The 8bpp video ram value sets the colour as follows:
\ Bits 0..1 set the red intensity,
\ Bits 2..3 set the green intensity
\ Bits 4..5 set the blue intensity
\ Bits 6..7 overall intensity multiplier
\ A colour value of 0 is black, and 255 is white
: SET-DEF-PALETTE	\ - | - ; program controller with a basic palette
  _LCD					\ .. base
  LCD_PAL_BASE +			\ point to first palette reg.
  #255 0 do
  	i palette-value
  	i 1+ palette-value 16 lshift or	\ merge; 2 entries per reg.
  	over !
  	cell+				\ point to next reg
  2 +loop
  drop					\ .. base
;

\ COLOUR CONSTANTS
\ ~~~~~~~~~~~~~~~~
 0   constant #ColBLACK
 $C3 constant #ColRed
 $CC constant #ColGREEN
 $30 constant #ColDKBLUE
 $F0 constant #ColBLUE
 $7A constant #ColPaleBlue
 $4E constant #ColPaleGreen
 $7F constant #ColGREY
 $CF constant #ColYellow
 $FF constant #ColWHITE

: SET-INK		\ colour | - ;
  to Ink
;

: SET-PAPER		\ colour | - ;
  to Paper
;

: (GLCD-SET-VPAGE)	\ addr | - ; set viewed page address
  _LCD
  2dup LCD_UPBASE + !			\ set frame buffer address
  LCD_LPBASE + !			\ set frame buffer address
;

: GLCD-SET-VPAGE	\ page# -- ; set the viewed page
  glcd-page-addr
  (glcd-set-vpage)
;


\ High level controller initialisation
: GLCD-INIT-CONTROLLER	\ - | - ; start controller running ONLY after SDRAM available
  ~pcLCD _PCONP or!					\ enable power to LCD controller block
  _LCD											\ .. base
  0 over CRSR_CTRL + !			\ disable cursor
  LCDctrlRegVal over LCD_CTRL + !	\ set up control register, controller not enabled yet
  LCDcfgRegVal _LCD_CFG !					\ set pixel clock prescaler
  LcdPolRegVal over LCD_POL + !		\ set the signal polarities
  LcdTimhRegVal over LCD_TIMH + !	\ set horizontal timing
  LcdTimvRegVal over LCD_TIMV + !	\ set vertical timing
  #defaultpage glcd-set-vpage		\ set frame buffer addr
  GLCD_BUFF0 ^DBuffer_L !		\ set frame address used for drawing
  set-def-palette
  dup LCD_CTRL + @ $0003:FFFF and	\ mask unused bits
  ~LCDCTRL_EN or over LCD_CTRL + !	\ enable controller
  100 ms				\ allow start-up
  dup LCD_CTRL + @ $0003:FFFF and	\ mask unused bits
  ~LCDCTRL_LCD_ON or			\ enable LCD module power
  over LCD_CTRL + !
  150 ms				\ settling time
  drop					\ .. base
  #defaultpage glcd-set-page
  #defaultpage LastPage_l !		\ default previous page
;

\ ===================================================================
\ Display Page Selection
\ ~~~~~~~~~~~~~~~~~~~~~~
\ ===================================================================

\ Number of available pages is calculated above

variable LastPage_L

 0    constant   #DEFAULTPAGE
 1    constant   #ALTPAGE
-1    constant   #PREV-PAGE

: GLCD-PAGE?    \ - | p ; returns currently selected DRAWING page
  ^DBuffer_L @
  case
    GLCD_BUFF0 of 0 endof
    GLCD_BUFF1 of 1 endof
    GLCD_BUFF2 of 2 endof
    GLCD_BUFF3 of 3 endof
      0 swap				\ default to page 0
  endcase
;

: GLCD-PAGE-ADDR	\ page# -- addr
  case
    0 of GLCD_BUFF0 endof
    1 of GLCD_BUFF1 endof
    2 of GLCD_BUFF2 endof
      GLCD_BUFF3 swap			\ otherwise default to page 3
  endcase
;


: GLCD-SET-PAGE  \ p | -
\ select the memory page used for the display, -1 means previous page
  dup 0< if                 		\ selected the last set page
    drop
    LastPage_l @
  else
    glcd-page? LastPage_l !		\ remember the current page
  endif
  glcd-page-addr
  dup ^DBuffer_L !
  (glcd-set-vpage)
;

: GLCD-PAGE-COPY    \ source_p# dest_p# | - ; copy screen data between pages.
                    \ no change to the page number currently displayed
  glcd-page-addr			\ display RAM address of dest. page start
  swap glcd-page-addr			\ display RAM address of source page start
  swap
  glcd-buffer-size cmove		\ copy data
;

