\ TScreen_LPC1788.fth

((
LPC1788 4 Wire Resistive Touch Screen Driver
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

This code reads a touch screen connected to four GPIO pins, two of which are
configured as ADC inputs.  The touch screen electrodes are static when idle
to minimise RFI.  A state machine is used to calculate the X and Y position
of the touch and to debounce.  NOTE: Resistive touch screens cannot detect
multi-touches.

This code provides a simple keypad emulation, using rectangular screen regions
which return a character code when 'pressed'.  Auto-repeat is implemented.

Touch screens vary somewhat, and calibration functions are included to ensure
that the calculated touch position is correct.  It is necessary to calibrate
each screen and save the calibration factors in non-volatile memory.  The user
will need to provide a user interface for the calibration primitives.

When calibrated, the touch screen state machine calculates the centre position
of the touch in screen pixels (0, 0 is the top left).

The code below includes an interface to highlight a pressed region on the
screen.  The example word XORS the screen so highlighting the pressed
rectangular region while the press is maintained.  Implement the word

This code is loosely based on the Olimex demo C code supplied with the LPC1788
MOD-LCD 4.3 demo board and attributed to IAR.  The main change is to handle
the touch screen state machine on a timer tick, rather than by using multiple
(ADC, GPIO and Timer) interrupts!

EQUATES
~~~~~~~
#SHO-DIAGS - set non-zero, diagnostic text is output.  If not defined, this
file defines it false.

#PIXELSHIGH and #PIXELSWIDE must be defined for the screen resolution in pixels.

#MAX_TS_BUTTONS - specifies the maximum number of rectangular button regions
that can be specified at one time.  Code below sets a default value of 8 if
not previously defined.

BASIC INTERFACE
~~~~~~~~~~~~~~~
	word TS-ADD-BUTTON	- register a rectangular button region
	word TS-ERASE-BUTTONS - clear all button region defs.
	word TS-KEY - await touch, return press as character code
	word TS-KEY? - true if touch pending

---------------------------------------------------------------------------

Forth implementation by:
	Barry Culver,
	Culver Consultancy Ltd.
	www.culverconsultancy.co.uk
	Forth Implementation Copyright (c) 2014 Culver Consultancy Ltd.

Destributed for use with MPE Compiler Products by:
	MicroProcessor Engineering
	133 Hill Lane
	Southampton SO15 5AF
	England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)

))

DECIMAL			\ use '$' prefix for hex

[UNDEFINED]  #SHO-DIAGS [IF]
  0 equ #SHO-DIAGS
[ENDIF]

\ ADC used for touch screen
_ADC0 equ _TS_ADC

\ ADC Clock - maximum 12.4MHz, but lower is suggested if input impedance
\ is significant.
#500 KHz equ ADC_SPEED		\ 500kHz is as per Olimex/IAR code

periph-speed ADC_SPEED /mod
swap [if] 1+ [endif]			\ round up
1- 8 lshift 				\ adjust & shift for register merge
bit21 or				\ merge in the power-up bit
equ ADC_CLKDIV_VALUE

bit24 equ ADC_START			\ start a single conversion

\ GPIO Pins used for Touch Screen

\ IOCON registers
IOCON_P0_24 equ TS_Y1_IOCON
IOCON_P0_22 equ TS_Y2_IOCON
IOCON_P0_23 equ TS_X1_IOCON
IOCON_P0_21 equ TS_X2_IOCON

\ BIT Masks
bit24 equ TS_Y1_MASK
bit22 equ TS_Y2_MASK
bit23 equ TS_X1_MASK
bit21 equ TS_X2_MASK

\ FIO Registers
_FIO0 equ _TS_FIO

\ Bits
bit24 equ TS_Y1_BIT
bit22 equ TS_Y2_BIT
bit23 equ TS_X1_BIT
bit21 equ TS_X2_BIT

\ ADC Channel bits
bit1 equ ADC_CHL_Y1	\ Y1 is ADC 0, chl 1
bit0 equ ADC_CHL_X1	\ X1 is ADC 0, chl 0


\ Variables

0	VALUE	TSstate		\ holds state - see below
\ state codes:
0 equ TS_INTR_SETUP_DLY
1 equ TS_SETUP_DETECT
2 equ TS_WAIT_FOR_TOUCH
3 equ TS_SETUP_X_MEASURE
4 equ TS_X_MEASURE
7 equ TS_SETUP_Y_MEASURE
8 equ TS_Y_MEASURE
8 equ TS_SETUP_REL_DETECT
10 equ	TS_AWAIT_RELEASE

\ *** NOTE ***
\ The delays and time periods below are specified in system timer ticks.
\ The code was written assuming 1ms system timer period.  If the system
\ timer runs slower, adjust these times accordingly and expect a less
\ responsive screen.

\ Thresholds and delays
ADC_MAX_RDG 2 * 3 / equ	TS_TOUCH_THR	\ threshold to detect press
ADC_MAX_RDG 3 /	equ	TS_RELEASE_THR	\ threshold to detect press

\ delay times are in ticks
 2 equ #TS_SETUP_DLY
40 EQU #TS_SETTLE_DELAY
10 equ #TS_SAMPLES    	\ 10ms

\ These values are set in ticks.
 900 equ	#TS_1ST_REP_DELAY	\ delay until 1st auto-repeat
1200	equ	#TS_REP_DELAY		\ delay to subsequent repeats

variable  Touch			\ true if a screen press detected
variable  TSSamples
variable StateDelay
variable  Xaccum_l
variable  Yaccum_l
wvariable Xpos_w
wvariable Ypos_w
wvariable Xoffset_w
wvariable Yoffset_w
wvariable Xspan_w
wvariable Yspan_w

\ Values for Touch
1 equ #SHORT_TOUCH
2 equ#LONG_TOUCH

\ Trigger ADC Conversions - IO must be set up first
: (TS-TRIGGER-ADC)	\ _ADC chl_bits | - ; initiate a conversion for given channel
  ADC_CLKDIV_VALUE or  ADC_START or
  swap ADnCR + !
;

: TS-TRIGGER-X1		\ _ADC | -; initiate a conversion for X1
  ADC_CHL_X1 (ts-trigger-adc)
;

: TS-TRIGGER-Y1		\ _ADC | -; initiate a conversion for Y1
  ADC_CHL_Y1 (ts-trigger-adc)
;

: TS-READ-ADC		\ _ADC | rdg; get result of last conversion, right justified
  ADnGDR + @
  ADC_RES_MASK and
  ADC_RES_SHIFT rshift
;

: TS-ADC-HAS-RDG?	\ _ADC | flag; true is conversion result waiting
  ADnGDR + @
  ~ADCGDR_Done and 0<>
;

\ The four touch screen lines are set up for the given state
: (TS-SET-LINES)	\ state | -;
  case
    TS_SETUP_DETECT of
      _IOCON
      IO_PERIPH_FUNC1 -FILTER over TS_X1_IOCON + !
      IO_GPIO_FUNC +PULLDN +HYSTERESIS over TS_X2_IOCON + !
      IO_GPIO_FUNC +DIGITAL over TS_Y1_IOCON + !
      IO_GPIO_FUNC +HYSTERESIS swap TS_Y2_IOCON + !
      _TS_FIO
      TS_Y1_MASK TS_Y2_MASK or over FIODIR + or!	\ Y lines are outputs
      TS_Y1_MASK TS_Y2_MASK or over FIOSET + !		\ and set high
      dup FIODIR + @
      TS_X1_MASK TS_X2_MASK or not and swap FIODIR + !	\ X2 is input (pull down), X1 is ADC
    endof

    TS_SETUP_X_MEASURE of
      _IOCON
      IO_PERIPH_FUNC1 -FILTER over TS_X1_IOCON + !	\ ADC analogue
      IO_GPIO_FUNC +HYSTERESIS over TS_X2_IOCON + !
      IO_GPIO_FUNC +DIGITAL over TS_Y1_IOCON + !
      IO_GPIO_FUNC +HYSTERESIS swap TS_Y2_IOCON + !
      _TS_FIO
      TS_Y1_MASK TS_Y2_MASK or over FIODIR + or!	\ Y lines are outputs
      TS_Y1_MASK over FIOSET + !			\ Y1 set high
      TS_Y2_MASK over FIOCLR + !			\ Y2 set low
      dup FIODIR + @
      TS_X1_MASK TS_X2_MASK or not and swap FIODIR + !	\ X2 is input (HiZ), X1 is ADC
    endof

    TS_SETUP_Y_MEASURE of
      _IOCON
      IO_PERIPH_FUNC1 -FILTER over TS_Y1_IOCON + !	\ ADC analogue
      IO_GPIO_FUNC +HYSTERESIS over TS_X1_IOCON + !
      \ X lines are port outputs
      IO_GPIO_FUNC +HYSTERESIS +DIGITAL over TS_X1_IOCON + !
      IO_GPIO_FUNC +HYSTERESIS swap TS_X2_IOCON + !
      _TS_FIO
      TS_X1_MASK TS_X2_MASK or over FIODIR + or!	\ X lines are outputs
      TS_X1_MASK over FIOSET + !			\ and set X1 = high, X2 = low
      TS_X2_MASK over FIOCLR + !
      dup FIODIR + @
      TS_Y1_MASK TS_Y2_MASK or not and swap FIODIR + !	\ Y2 is input (HiZ), Y1 is ADC
    endof
  endcase
;

: (TS-SET-STATE)	\ state | - ; set new state, reset timer
  to TSstate
  StateDelay off
;

\ State machine to handle decoding touch screen presses.
: (TS-HANDLER)		\ --; called from the tick timer
  1 StateDelay +!			\ increment ms timer
  TSstate case
    TS_INTR_SETUP_DLY of
      Touch off
      Xpos_w woff
      Ypos_w woff
      StateDelay @ #TS_SETTLE_DELAY > if
        TS_SETUP_DETECT (ts-set-lines)
    	TS_SETUP_DETECT (ts-set-state)
      endif
    endof

    TS_SETUP_DETECT of
      _TS_ADC ts-trigger-x1		\ prepare first reading for next state
      TS_WAIT_FOR_TOUCH (ts-set-state)
    endof

    TS_WAIT_FOR_TOUCH of
      _TS_ADC ts-read-adc
      TS_TOUCH_THR > if			\ .. press detected
        StateDelay @ #TS_SETTLE_DELAY > if	\ .. press is maintained
          TS_SETUP_X_MEASURE (ts-set-lines)
          TS_SETUP_X_MEASURE (ts-set-state)
          Touch off			\ kill any pending press
          stat-led-on
        else
          _TS_ADC ts-trigger-x1		\ .. for next time
        endif
      else
        _TS_ADC ts-trigger-x1		\ .. for next time
        StateDelay off			\ reset the timer
      endif
    endof

    TS_SETUP_X_MEASURE of
      _TS_ADC ts-trigger-x1		\ prepare first reading for next state
      TS_X_MEASURE (ts-set-state)
      Xaccum_l off
      TSSamples off
    endof

    TS_X_MEASURE of
      1 TSSamples +!
      _TS_ADC ts-read-adc
      Xoffset_w w@ - 0 max		\ remove offset
      Xaccum_l +!
      TSSamples @ #TS_SAMPLES >= if
        Xaccum_l @
        #PIXELSWIDE *
        TSSamples @ /			\ take mean
        Xspan_w w@ /
        Xpos_w w!
        TS_SETUP_Y_MEASURE (ts-set-lines)
        TS_SETUP_Y_MEASURE (ts-set-state)
      else
        _TS_ADC ts-trigger-x1
      endif
    endof

    TS_SETUP_Y_MEASURE of
      _TS_ADC ts-trigger-y1		\ prepare first reading for next state
      TS_Y_MEASURE (ts-set-state)
      Yaccum_l off
      TSSamples off
    endof

    TS_Y_MEASURE of
      1 TSSamples +!
      _TS_ADC ts-read-adc
      Yoffset_w w@ - 0 max		\ remove offset
      Yaccum_l +!
      TSSamples @ #TS_SAMPLES >= if
        Yaccum_l @
        #PIXELSHIGH *
        TSSamples @ /			\ take mean
        Yspan_w w@ /
        Ypos_w w!
        TS_SETUP_DETECT (ts-set-lines)
        TS_AWAIT_RELEASE (ts-set-state)
        #SHORT_TOUCH Touch !										\ flag press
      else
        _TS_ADC ts-trigger-y1
      endif
    endof

    TS_SETUP_REL_DETECT of
      _TS_ADC ts-trigger-x1		\ prepare first reading for next state
      TS_AWAIT_RELEASE (ts-set-state)
      TSSamples off
    endof

    TS_AWAIT_RELEASE of
      _TS_ADC ts-read-adc
      TS_RELEASE_THR < if		\ .. release detected
        StateDelay @ #TS_SETTLE_DELAY > if	\ .. release is confirmed
          TS_WAIT_FOR_TOUCH (ts-set-state)
          stat-led-off
        else
          _TS_ADC ts-trigger-x1		\ .. for next time
        endif
      else
        StateDelay off	\ reset the timer
        _TS_ADC ts-trigger-x1		\ .. for next time
        1 TSSamples +!
        TSSamples @ #TS_1ST_REP_DELAY = if
          #LONG_TOUCH Touch !								\ flag a repeat
        endif
        TSSamples @ #TS_REP_DELAY = if
          #LONG_TOUCH Touch !
          #TS_1ST_REP_DELAY TSSamples !	\ reset timer for subsequent repeats
        endif
      endif
    endof
endcase
;

: TS-CLEAR-CAL	\ -- ; reset screen calibration factors - prior to new cal
  ADC_MAX_RDG Yspan_w w!
  Yoffset_w woff
  ADC_MAX_RDG Xspan_w w!
  Xoffset_w woff
;

: (TS-INIT)
  ts-clear-cal
  TS_INTR_SETUP_DLY (ts-set-state)
  \ init ADC - IO configuration is done by touchscreen-set-lines
  ~pcADC _PCONP or!			\ enable power
  ADC_CLKDIV_VALUE			\ set clock divider, power up
  _TS_ADC ADnCR + !
  ['] (ts-handler) 1 every drop		\ check touch screen every tick (ms)
;
' (ts-init) atcold



\ These two calibration functions should be called sequentially, and with
\   the user pressing a stylus in the exact corner of the screen.
\   First, TS-CAL-TOPLEFT to set the offset, then TS-CAL-BOTRIGHT to set span

\ Await stylus press at the top right of the screen.  Set cal factors.
: TS-CAL-TOPLEFT		\ - | flag; true if calibration was set OK
  Touch off
  ts-clear-cal
  begin  pause  Touch @ until		\ await press on top left
  Xaccum_l @ #TS_SAMPLES /
  Yaccum_l @ #TS_SAMPLES /
  2dup + ADC_MAX_RDG 3 / < if		\ .. possible
[ #SHO-DIAGS ] [if]
    >ser
    cr ." TS-CAL-TOPLEFT: OK  y=" dup . ."  x=" over .
    >glcd
[endif]
    Yoffset_w w!
    Xoffset_w w!
    true
  else
[ #SHO-DIAGS ] [if]
    >ser
    cr ." TS-CAL-TOPLEFT: Error  y=" . ."  x=" .
    >glcd
[else]
    2drop
[endif]
    false
  endif
;

\ Await stylus press for calibrating the bottom right screen region.
: TS-CAL-BOTRIGHT	\ - | flag; true if calibration was set OK
  Touch off
  begin  pause  Touch @ until		\ await press on bottom right
  Xaccum_l @ #TS_SAMPLES /
  Yaccum_l @ #TS_SAMPLES /
  2dup + ADC_MAX_RDG 2/ > if
[ #SHO-DIAGS ] [if]
    >ser
    cr ." TS-CAL-BOTRIGHT: OK  y=" dup . ."  x=" over .
    >glcd
[endif]
    Yspan_w w!
    Xspan_w w!
    true
  else
[ #SHO-DIAGS ] [if]
    >ser
    cr ." TS-CAL-BOTRIGHT: Error  y=" . ."  x=" .
    >glcd
[else]
    2drop
[endif]
    false
  endif
;


((
Example Touch Screen Calibration Interface
Re-code this to suit your own graphics system.
))

create TS_Text_
  ", Touch Screen Calibration"
create TS_Text2_
  ", Using a stylus, touch the indicated corner..."

: TS-DO-CALIBRATE		\ --
  >glcd
  begin														\ top left calibration
    #ColWhite set-ink
    glcd-cls
    #NarrowFont glcd-set-font		\ use narrower font to fit
    uline
    0 TS_Text_ glcd.c$			\ prompt touch
    n-uline
    2 TS_Text2_ glcd.c$
    #ColYellow set-ink
    0 0 4 4 solid-box			\ draw the target to touch with stylus
    ts-cal-topleft bip 			\ .. pressed somewhere close
  until
  begin														\ bottom right calibration
    #ColWhite set-ink
    glcd-cls
    uline
    0 TS_Text_ glcd.c$
    n-uline
    6 TS_Text2_ glcd.c$
    #ColYellow set-ink
    #PIXELSWIDE 5 -  #PIXELSHIGH 5 -
    4 4 solid-box			\ draw the target to touch with stylus
    ts-cal-botright bip			\ .. pressed somewhere close
  until
  save-ts-cal										\ .. to non-volatile memory
  #ColWhite set-ink
  glcd-cls
;


((
Screen Button Decoding
~~~~~~~~~~~~~~~~~~~~~~
A number of non-overlapping rectangular screen regions are defined and when
a touch is detected within one of these regions, an alphabetic character code
is returned.  The code inplements auto-repeat.  Repeated codes are upper case.

Use the word TS-ADD-BUTTON to add these rectangular regions as required, up to
#MAX_TS_BUTTONS.  Use the word TS-ERASE-BUTTONS to erase the region list,
to set up a new one.

This makes it easy to emulate a keypad using the touch screen.
))

[UNDEFINED] #MAX_TS_BUTTONS [IF]
8 equ #MAX_TS_BUTTONS		\ default the maximum number of on-screen button regions
[ENDIF]

STRUCT	ButtDef
  WORD: +TopY
  WORD: +LeftX
  WORD: +BotY
  WORD: +RightX
  BYTE: +Key
  3 FIELD +tsspare
END-STRUCT

#MAX_TS_BUTTONS ButtDef * equ Sizeof_TS_BUTTONS

ButtDef #MAX_TS_BUTTONS * Buffer: TS_Buttons_
variable NoOfTSbuttons_L

\ erase all button region definitions - call before defining a new set
: TS-ERASE-BUTTONS		\ --;
  TS_Buttons_ Sizeof_TS_BUTTONS erase
  NoOfTSbuttons_L off
;

\ Add a rectangular screen region to the list for keypad emulation.
\ If the keycode is a lower case character, auto-repeats due to sustained
\ press are returned in upper case.  The variable Touch holds the 'keypress'.
: TS-ADD-BUTTON		\ x y width height keycode | -; add a button region definition
  NoOfTSbuttons_L @ dup #MAX_TS_BUTTONS < if
    ButtDef * TS_Buttons_ +		\ point to free entry
    dup >r
    +Key c!
    2 pick +  r@ +BotY w!
    2 pick +  r@ +RightX w!
    r@ +TopY w!
    r> +LeftX w!
    1 NoOfTSbuttons_l +!
  else					\ .. table is full
    drop
  endif
;

((
Button press notification
~~~~~~~~~~~~~~~~~~~~~~~~~
Implement this to suit your own graphics driver, if user feedback is required.
))

: TS-HIGHLIGHT-BUTTON	\ butt_ | -; highlight the button on screen
  dup >r +LeftX w@ 1+			\ X
  r@ +TopY w@ 1+			\ Y
  r@ +RightX w@ 2 pick - 1-		\ width
  r> +BotY w@ 2 pick - 1-		\ height
  2over 2over inv-solid-box
  50 ms
  inv-solid-box
;


\ Call this word only if Xpos, Ypos and Press hold a valid touchscreen 'press'.
: TS-SCAN-BUTTONS		\ - | keycode/0; return keycode is press co-ords match a button
  0					\ assume no match
  TS_Buttons_
  NoOfTSbuttons_l @ 0 ?do		\ scan the list of screen regions
    Touch @ 0= ?leave			\ quit if no press
    dup +TopY w@ Ypos_w w@ <
    over +BotY w@ Ypos_w w@ > and	\ Y within bounds?
    over +LeftX w@ Xpos_w w@ < and
    over +RightX w@ Xpos_w w@ > and
    if					\ .. touch is on the button
      nip
      dup +Key c@
      Touch @ #LONG_TOUCH =
      if  upc  endif			\ long press is upper case key
      swap
      dup ts-highlight-button
      leave
    endif
    ButtDef +				\ index to next
  loop
  Touch off				\ clear flag
  drop					\ .. leaving code or 0
;

((
  The touch screen decoding is done under interrupt and there is a single
  key buffer. You could add a ring buffer to replace variable Touch
  if you need more.
))

: TS-CLR-KEY	\ -- ; discard any pending key
  Touch off
;

: TS-KEY?	\ - | flag ; true if touch press available
  Touch @ 0<>
;

: TS-KEY	\ - | keycode ; await, return ts press
  begin
    begin											\ await press
      ts-key? 0=
     while
      pause				\ NOTE: ASSUMES multi-tasker
    repeat
    ts-scan-buttons
  ?dup until
;

\ TEST CODE
\ ~~~~~~~~~

false [if]

: TS-SHOW	\ --
  cr ." X=" Xpos_w w@ .
  ."   Y=" Ypos_w w@ .
  Touch off
;

: TS-TEST	\ --
  begin
    Touch @ if
      Xpos_w w@ Ypos_w w@ pset
      ts-show
    endif
  key? until
  key drop
;

: TS-SHOW-BUTTONS	\ --;
  cr
  NoOfTSbuttons_l @ ?dup if
    ."  LeftX TopY RightX BotY  Code"
    cr
    TS_Buttons_
    swap 0 do
      dup +LeftX w@ 4 .r 3 spaces
      dup +TopY w@ 3 .r 2 spaces
      dup +RightX w@ 3 .r 4 spaces
      dup +BotY w@ 3 .r 4 spaces
      dup +Key c@ emit
      ButtDef +
      cr
    loop
    drop
  else
    ." no buttons!"
  endif
;

[endif]

