\ StartLPC40xx.fth - Cortex startup code for LPC40xx family

((
Copyright (c) 2009, 2010, 2014, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20150622 SFP003 LPC40xx updates.
20140402 BJC002 Modified for the minor hardware differences for the LPC1788
		plus some additions. This file uses an iterative scheme to
		set the PLL for best match to given SYSTEM-SPEED.  Also uses
		the equate USE-XTAL which if false sets the clock source to
		be the IRC oscillator.
20100112 MPE001 First release.
))

only forth definitions  decimal

\ ===========
\ *! startlpc40xx
\ *T Cortex start up for LPC40xx
\ ===========
\ *P The file *\i{Cortex/Hardware/LPC40xx/startLPC40xx.fth} contains
\ ** the start up code for LPC40xx devices. Start up code for other
\ ** Cortex-M3 implementations, e.g. STM32, can be found in parallel
\ ** directories. The absolute minimum code to start a Cortex-M3
\ ** is in *\i{Cortex/StartCortex.fth}.

[undefined] ExcVecs [if]
l: ExcVecs	\ -- addr ; start of vector table
\ *G The exception vector table is *\fo{/ExcVecs} bytes long. The
\ ** equate is defined in the control file. Note that this table
\ ** table must be correctly aligned as described in the Cortex-M3
\ ** Technical Reference Manual (ARM DDI0337, rev E page 8-21). If
\ ** placed at 0 or the start of Flash, you'll usually be fine.
  /ExcVecs allot&erase
[endif]

\ *G The equate USE-XTAL is set to zero to use the internal 12MHz RC
\ ** oscillator.  This is accurate to a couple of percent so fine for
\ ** general use - BUT NOT ACCURATE ENOUGH FOR USB!
[undefined] USE-XTAL [if]		\ .. not defined in the control file
  false equ USE-XTAL			\ default to compatible state
[endif]

interpreter
: SetExcVec	\ addr exc# --
\ *G Set the given exception vector number to the given address.
\ ** Note that for vectors other than 0, the Thumb bit is forced
\ ** to 1. *\b{Do not} use this to set the ISP checksum!
  dup if				\ If not the stack top
    swap 1 or swap			\   set the Thumb bit
  endif
  cells ExcVecs + !  ;
target

L: CLD1		\ holds xt of main word
  0 ,					\ fixed by MAKE-TURNKEY

proc DefExc2	b  $	end-code	\ NMI
proc DefExc3	b  $	end-code	\ HardFault
proc DefExc4	b  $	end-code	\ MemManage
proc DefExc5	b  $	end-code	\ BusFault
proc DefExc6	b  $	end-code	\ UsageFault
proc DefExc11	b  $	end-code	\ SVC
proc DefExc12	b  $	end-code	\ DbgMon
proc DefExc14	b  $	end-code	\ PendSV
proc DefExc15	b  $	end-code	\ SysTick


\ Calculate the initial value for the data stack pointer.
\ We allow for TOS being in a register and guard space.

[undefined] sp-guard [if]		\ if no guard value is set
0 equ sp-guard	\ -- +n
\ *G The number of cells reserved as guard space on the Forth
\ ** data stack. Usually set to two in the control file. This
\ ** value allows the data stack to underflow by the given amount
\ ** before other data is corrupted.
[then]

init-s0 tos-cached? sp-guard + cells - equ real-init-s0  \ -- addr
\ *G The data stack pointer value set at start up.

[undefined] LPCboost [if]
  system-speed 100 MHz >= 1 and
  equ LPCboost				\ default setting only - see control file
[endif]

\ *P We are assuming that the CPU clock runs at system clock rate
\ ** i.e. divide by 1
1 equ CCLKSELval	\ -- x
\ *G The CCLKCFG register initial value.

\ *P If the external xtal is NOT used, the clock source will be the int
\ ** RC oscillator.  We set xtal-speed to this frequency for PLL calcs use.
use-xtal 0= xtal-speed 0= or [if]	\ .. using int. RC osc.
  irc-speed equ xtal-speed
[endif]

\ *P We calculate the required division factor for peripheral clocks
system-speed periph-speed / 1 max equ PCLKSELval
\ *G Peripheral clock register initial value.

interpreter
[-short-branches
: gen-PLL0	\ -- Mval Pval
\ *G INTERPRETER: Generate the data for the PLL0CFG register.
\ ** The PLL input, Fin, must lie in the range 10 to 25 MHz.
\ ** The PLL osc. output, Fcco, must lie in the range 155 and 320 MHz.
\ ** Fcco = (Fin * M * 2 * P)
\ ** The final PLL output is (Fin * M)
\ ** Mval and Pval are the register values to use Mval = M-1
  { | P bestP bestM -- Mval Pval }
  system-speed xtal-speed u< if
    cr ." system speed less than xtal speed" abort
  endif
  xtal-speed #10 MHZ u< if
    cr ." Xtal frequency too low to use for PLL input!" abort
  endif
  1	to P				\ initial P
  -1 to bestP				\ best P is invalid
  -1 to bestM				\ best M invalid
  -1					\ maximum for best error difference
  4 0 do				\ scan possible Pvalues (1, 2, 4, 8)
    33 1 do				\ scan possible M values (1 .. 32)
      P 2* i * xtal-speed *		\ calc Fcco; 2 * P * M * Fin
      dup #156 MHz #320 MHz within? if	\ .. within PLL range
        P 2* /				\ calc PLL output frequency
        system-speed
      	2dup u<	if			\ absolute difference unsigned
          swap
      	endif
	-
      	dup 2 pick u< if		\ lower error than last guess
      	  nip				\ leave new error
      	  i to bestM
      	  P to bestP
      	else
          drop
      	endif
      else
      	drop
      endif
    loop
    dup 0= ?leave			\ stop if found a solution with low P value
    P 2* to P				\ .. next P
  loop
  bestM 0< if
    cr ." Acceptable Mval not found!"  abort
  endif
  cr ." Main PLL: best match found; P = " bestP 1 u.r
  ." , M = " bestM 1 u.r ."  (Fcco = " bestM xtal-speed * bestP 2* * 1 MHZ / 1 u.r
  ." MHz), error = " 1 u.r ." Hz"
  bestM 1-						\ Mval
  case bestP 						\ Pval
    1 of 0 endof
    2 of 1 endof
    4 of 2 endof
    8 of 3 endof
  endcase
;

: gen-FlTim	\ -- flcfg
\ *G Generate the Flash configuration register value.
  system-speed 1- #20 MHZ /
  LPCboost if				\ .. performance mode
    3 max
  endif
;
short-branches]
target

system-speed xtal-speed <> [if]
  gen-PLL0 equ Pval  equ Mval
  Pval #5 lshift  Mval or  equ PLL0CFGval
\ recalculate the system-speed
  xtal-speed 				\ PLL input clock
  Mval 1+  *				\ PLL output speed
    equ cco-real			\ CCO speed
  cco-real CCLKSELval /
  dup system-speed <> [if]
    equ system-speed			\ system speed this will give
    cr ." NOTE: Actual system speed is " system-speed 1 u.r ." Hz."
  [else]				\ requested system speed has been achieved exactly
    drop
  [endif]
  usb-speed [if]
    cco-real usb-speed /  equ USBCLKval		\ 0=off, 1=/1, 2=/2 etc.
    USBCLKval 0=  USBCLKval 3 >  or [if]
      cr ." Invalid USBCLKSET divider value!"
      abort
    [endif]
  [else]
    0 equ USBCLKval		\ clock is disabled
  [then]
[then]
  gen-FlTim #12 lshift equ FlCfgval

CCLKSELval system-speed * equ cco-speed	\ -- hz
\ *G The actual CCO frequency that will be generated by the
\ ** calculated configuration values.

internal
: DoPLL0Feed	\ _syscon | - ; send the feed sequence to pll0
  $AA over PLL0FEED + !			\ feed sequence to load PLL
  $55 swap PLL0FEED + !
;

: setClocks	\ --
\ *G Enable the clocks, set the bus dividers, and enable the PLLs
\ ** as required.
  _SYSCON				\ point at LPC System Control Regs
  1 over CCLKSEL + !			\ ensure not running from PLL, /1
  0 over CLKSRCSEL + !			\ ensure CPU+PLL running from IRC initially
  0 over USBCLKSEL + !			\ USB initially not clocked
  0 over SPIFICLKSEL + !		\ SPIFI not clocked
  0 over PLL0CON + !			\ PLL is initially disabled
  dup DoPLL0Feed			\ feed sequence for PLL disable
  [ use-xtal ] [if]
  \ Main oscillator set up using xtal oscillator
    [ xtal-speed #14 MHz > ] [if]	\ .. high speed xtal
      bit4 bit5 or over SCSreg + !	\ enable main oscillator high speed mode
    [else]
      bit5 over SCSreg + !		\ enable main oscillator std mode
    [endif]
    begin										  		\ wait for it to start
      dup SCSreg + @ bit6 and
    until
    bit0 over CLKSRCSEL + !		\ %01 for xtal as main oscillator
  [endif]

[ system-speed xtal-speed u> ] [if]	\ .. need to use PLL to multiply xtal frequency
  [ cr ." Configuring PLL0 with PLL0CFG val $" PLL0CFGval .hex ]
\	configure PLL0
\ (note: PLL1 for USB is set up by the USB code.)
  PLL0CFGval over PLL0CFG + !		\ set PLL divider
  dup DoPLL0Feed			\ feed sequence to load PLL
  bit0 over PLL0CON + !			\ enable PLL
  dup DoPLL0Feed			\ feed sequence to start PLL
  begin													\ wait for lock
    dup PLL0STAT + @ bit10 and		\ test PLOCK
  until
  CCLKSELval over CCLKSEL + !		\ set up core clock divider, still using SYSCLK source
  bit8 CCLKSELval or over CCLKSEL + !	\ switch core clock to use PLL0.
[else]
  [ cr ." NOTE: PLL *not* used as xtal-speed <= system-speed." ]
[endif]

  \ Set the Flash memory timing
  dup FlashCfg + @ $03FF and		\ read no-change bits
  FlCfgval or 				\ merge FlashTim bits
  over FlashCfg + !

  \ Set the peripheral clock division factor
  PCLKSELval over PCLKSEL + !

[ system-speed #100 MHZ >=  LPCboost and ] [if]	 \ enable performance mode if clk >= 100MHz
  3 over PBOOST + !		 \ full regulator voltage
  [ cr ." Boost mode is ENABLED" ]
[else]
  0 over PBOOST + !		 \ BOOST is off
[endif]
  drop
;

: setNVIC	\ --
\ *G Initialise the NVIC. By default it points into Flash.
  ExcVecs _SCS scsVTOR + !		\ Vector Table Offset Reg.
;


: StartCortex	\ -- ; never exits
\ *G Set up the Forth registers and start Forth. Other primary
\ ** hardware initialisation is also performed here.
  begin
    INT_STACK_TOP SP_main sys!		\ set SP_main for interrupts
    INIT-R0 SP_process sys!		\ set SP_process for tasks
    2 control sys!			\ switch to SP_process
    REAL-INIT-S0 set-sp			\ Allow for cached TOS and guard space
    INIT-U0 up!				\ USER area
\ here onwards we have a working forth system & can use stack and execute words
\    init-stat-led
\    stat-led-on
    setClocks
\    stat-led-off
    setNVIC
    CLD1 @ execute
  again
;
external

INT_STACK_TOP StackVec# SetExcVec	\ Define initial return stack
' StartCortex ResetVec# SetExcVec	\ Define startup word
DefExc2       2         SetExcVec	\ set default exception handlers
DefExc3       3         SetExcVec
DefExc4       4         SetExcVec
DefExc5       5         SetExcVec
DefExc11      11        SetExcVec
DefExc12      12        SetExcVec
DefExc14      14        SetExcVec
DefExc15      15        SetExcVec


\ ------------------------------------------
\ reset values for user and system variables
\ ------------------------------------------

[undefined] umbilical? [if]
\  true
\ [else]
\   umbilical? 0=
\ [then]
\ [if]
\ initial values of user variables
L: USER-RESET
  init-s0 tos-cached? sp-guard + cells - ,	\ s0
  init-r0 ,				\ r0
  0 ,  0 ,                              \ #tib, 'tib
  0 ,  0 ,                              \ >in, out
  $0A ,  0 ,                            \ base, hld

\ initial values of system variables
L: INIT-FENCE 0 ,                       \ fence
L: INIT-DP 0 ,                          \ dp
L: INIT-VOC-LINK 0 ,                    \ voc-link
[then]


\ ======
\ *> ###
\ ======

decimal

