\ RpiPico.ctl - Raspberry Pi Pico standalone Forth

((
Copyright (c) 2023
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20230622 MPE001 First release.
))


\ ================
\ *! rpipico
\ *T Control file for Raspberry Pi Pico board
\ ================
\ *P The control file *\i{Cortex/Hardware/RP2040/RpiPico.ctl}
\ ** produces a standalone Forth for the Raspberry Pi Pico board.
\ ** The serial port is connected to USART1 on pins PA9 (Tx) and
\ ** PA10 (Rx).
\ ** You cannot use the easy GPIO code for UART initialisation unless
\ ** the GPIO ports are clocked and taken out of reset in the start
\ ** up code.

\ *P When you modify this file for your own hardware do not forget
\ ** to update the GPIO pin assignments and alternate function
\ ** selections.

\ *P For debugging, You can easily use the low-cost Segger J-LINK
\ ** EDU unit.

only forth definitions decimal


\ **************************
\ *S Define directory macros
\ **************************
\ *P The MPE cross compilers contain a useful text macro system.
\ ** Macros allow easy porting of control files when projects are
\ ** moved from their default locations. Each directory macro
\ ** defines where a particular set of files are located.

1 constant MPE?
0 constant XYZ?

MPE? [if]	\ MPE macros
c" ."			setmacro AppDir		\ application files
c" ."			setmacro HwDir		\ Board hardware files
c" ../.."		setmacro CpuDir		\ CPU specific files
c" ../../../Common" 	setmacro CommonDir	\ where the common code lives
c" ../../../Examples"	setmacro ExampleDir	\ example files
[then]

XYZ? [if]
  ...
[then]


\ ***************************************
\ Load compiler extensions such as macros
\ ***************************************

include %CpuDir%/Macros		\ macros needed by the cross-compiler


\ *********************************************************
\ Turn on the cross compiler and define CPU and log options
\ *********************************************************

\ file: PROG.log                \ uncomment to send log to a file

CROSS-COMPILE			\ Turn host Forth into a cross-compiler

only forth definitions        	\ default search order

  no-log                       	\ uncomment to suppress output log
  rommed                    	\ split ROM/RAM target
  interactive                  	\ enter interactive mode at end
  +xrefs                        \ enable cross references
  align-long                    \ code is 32bit aligned
  Cortex-M0                     \ Thumb2 processor type and register usage
  +LongCalls                    \ permit standalone Forth to handle
                                \ calls outside 25 bit range.
  UF2-format                    \ also produce UF2 file
  $1000:0000 set-UF2dest        \ start of Pico Flash
  set-UF2-RP2040

  +SaveCdataOnly                \ no data area image files


0 equ false
-1 equ true


\ *******************
\ *S Configure target
\ *******************

\ ======================
\ *N RP2040 variant definitions
\ ======================

$2000:0000 equ FlashBase        \ -- addr
\ *G Start address of Flash. The bottom 4 kb (the vector area) is
\ ** reserved for bootloader(s).

#128 kb equ /Flash      \ -- len
\ *G Size of Flash.
4 kb equ /FlashPage     \ -- len
\ *G Size of a Flash Page.
0 equ KeepPages         \ -- u
\ *G Set this non-zero for the number of pages at the end of Flash
\ ** that are reserved for configuration data. Often set to 1 or 2
\ ** by systems that use PowerNet.

#64 cells equ /ExcVecs  \ -- len
\ *G Size of the exception/interrupt vector table. There are
\ ** 16 reserved by ARM.

12 MHz equ xtal-speed           \ -- hz
\ *G Master oscillator crystal clock rate in HZ. This is the
\ ** internal oscillator with an external crstal.
12 MHz equ system-speed         \ -- hz
\ *G Requested CPU clock speed in HZ. Note that you must
\ ** calculate the PLL values.

1 equ AHBdiv    \ -- u
\ *G Division ratio of the AHB clock from the system clock.
\ ** This may not be more than 48 MHz.
1 equ APBdiv    \ -- u
\ *G Division ratio of the APB clock from the system clock.
\ ** This may not be more than 48 MHz.

system-speed AHBdiv / equ AHB-speed     \ -- hz
\ *G AHB bus speed.
system-speed APBdiv / equ APB-speed	\ -- hz
\ *G APB bus speed.

\ =============
\ *N Memory map
\ =============
\ *P The Flash memory starts at $1000:0000. The bottom 4 kb is the
\ ** bootloader area.

  $2000:0000 4 kb - $2001:FFFF cdata section RP2040 \ code in Flash copied to RAM
  \ boot Flash and boot2 loader is at $1000:0000
  \ Forth code starts in Flash at     $1000:1000
  \ Forth code is copied to           $2000:0000
  $2002:0000 $2002:FFFF idata section PROGd     \ 64k IDATA
  $2003:0000 $2003:FFFF udata section PROGu     \ 64k UDATA

interpreter
: prog RP2040  ;                        \ synonym
target

PROG PROGd PROGu CDATA                  \ use Code for HERE , and so on

\ ============================
\ *N Stack and user area sizes
\ ============================
\ *P These equates calculate how much RAM is needed for the USER area,
\ ** return stack and data stack of each task, plus an area used
\ ** by Forth interrupts that share a common USER (and optionally)
\ ** common stacks. This can be done to reduce the overall amount
\ ** of RAM needed for interrupt stacks. Since interrupts can occur
\ ** in any task, switching stacks for high level interrupts will
\ ** reduce the total RAM requirements.

$0F0 equ UP-SIZE                \ size of each task's user area
$0F0 equ SP-SIZE                \ size of each task's data stack
$0100 equ RP-SIZE               \ size of each task's return stack
up-size rp-size + sp-size +
  equ task-size                 \ size of TASK data area
\ define the number of cells of guard space at the top of the data stack
#2 equ sp-guard                 \ can underflow the data stack by this amount

$0100 equ TIB-LEN               \ terminal i/p buffer length

\ define nesting levels for interrupts and SWIs.
1 equ #IRQs                     \ number of IRQ stacks,
                                \ shared by all IRQs (1 min)
0 equ #SVCs                             \ number of SVC nestings permitted
                                \ 0 is ok if SVCs are unused

\ ==========================
\ *N Serial and ticker rates
\ ==========================

1 equ useUART0? \ -- n
\ *G Set non-zero to compile code for UART0,
\ ** device *\fo{Console0}.
#115200 equ console0-speed
\ *G *\fo{Console0} speed in BPS.

0 equ useUART1? \ -- n
\ *G Set non-zero to compile code for UART1,
\ ** device *\fo{Console1}.
#115200 equ console1-speed
\ *G *\fo{Console1} speed in BPS.

0 equ console-port      \ -- n ; Designate serial port for terminal (0..n).
\ *G Ports 0,1 are the on-chip UARTs. The internal USB device
\ ** is port 10, and bit-banged ports are defined from 20 onwards.

#1 equ tick-ms          \ -- ms
\ *G Timebase tick in ms.

\ =====================
\ *N Software selection
\ =====================

\ Kernel components
0 equ BigKernel?	\ zero to reduce kernel size
1 equ ColdChain?        \ nz to use cold chain mechanism
0 equ tasking?          \ true if multitasker needed
6 cells equ tcb-size    \ for internal consistency check
0 equ event-handler?	\ true to include event handler
0 equ message-handler?	\ true to include message handler
1 equ semaphores?	\ true to include semaphores
0 equ timebase?		\ true for TIMEBASE code
0 equ softfp?		\ true for software floating point
0 equ FullCase?		\ true to include ?OF END-CASE NEXTCASE extensions
0 equ target-locals?	\ true if target local variable sources needed
0 equ romforth?		\ true for ROMForth handler
0 equ blocks?		\ true if BLOCK needed
0 equ sizeofheap	\ 0=no heap, nz=size of heap
1 equ heap-diags?       \ true to include diagnostic code
0 equ paged?            \ true if ROM or RAM is paged/banked
0 equ ENVIRONMENT?      \ true if ANS ENVIRONMENT system required
0 equ PowerNet?         \ -- flag
\ *G Set non-zero to compile the PowerNet TCP/IP stack.

\ *****************
\ default constants
\ *****************

\ cell equ cell                           \ size of a cell (32 bits)

\ ************
\ *S Boot block
\ ************

((
\ Next step: Rewrite boot block in Forth assembler.
cdata  chere $100 erase
hex
l: boot2	\ initialise 256 bytes
  4B11 W,  469D W,  4E11 W,  F000 W,  F816 W,  47B0 W,  4E10 W,  F000 W,
  F812 W,  47B0 W,  4E0F W,  F000 W,  F80E W,  47B0 W,  480E W,  490F W,
  4A0F W,  3A04 W,  5883 W,  508B W,  D1FB W,  4B0E W,  6019 W,  6803 W,
  469D W,  6843 W,  4718 W,  B500 W,  2200 W,  8A90 W,  0031 W,  8B12 W,
  4790 W,  0006 W,  BD00 W,  0000 W,  2000 W,  2004 W,  4649 W,  0000 W,
  4346 W,  0000 W,  5843 W,  0000 W,  1000 W,  1000 W,  0000 W,  2000 W,
  0000 W,  0002 W,  ED08 W,  E000 W,  0000 W,  0000 W,  0000 W,  0000 W,
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  0000 W,  
decimal
))

cdata  chere $100 erase
  include boot-simple/rp2040-boot2.fth
boot2 dup #251 +  boot2 #252 +  crc32RP2040  checksum	\ performed after FINIS

4 kb #256 - allot(t)  \ Align on next 4 kb page in flash memory


\ ************
\ *S Kernel files
\ ************

include %CpuDir%/CM0def                 \ Cortex generic equates and SFRs
include %CpuDir%/sfrRP2040              \ RP2040 special function registers
include %CpuDir%/StackDef               \ Reserve default task and stacks

PROGd  sec-top 1+ equ UNUSED-TOP  PROG  \ top of memory for UNUSED
include %HwDir%/startRpiPico            \ start up code
include %CpuDir%/CodeM0M1               \ low level kernel definitions
include %CommonDir%/kernel62            \ high level kernel definitions
include %CpuDir%/Drivers/rebootCM0      \ reboot using watchdog
include %CpuDir%/IntCortex              \ interrupt handlers for NVIC
\  include %CpuDir%/FaultCortex         \ fault exception handlers for NVIC
\  include %CommonDir%/CQueues          \ character queues
\  include %CpuDir%/Drivers/serRP2040p  \ polled serial driver
include %HwDir%/serRP2040p              \ polled serial driver
include %CommonDir%/Devtools            \ DUMP .S etc development tools
include %CommonDir%/Voctools            \ ORDER VOCS etc
include %CommonDir%/methods             \ target support for methods
include %CpuDir%/LocalCM0               \ local variables

tasking? [if]
  include %CpuDir%/MultiCM0		\ multi-tasker
[else]
  : pause  ;
[then]

include %CommonDir%/Delays
include %CpuDir%/Drivers/SysTickCM0
' start-clock AtCold

SIZEOFHEAP [if]
  include %CommonDir%/heap32		\ memory allocation set
[then]

environment? [if]
  include %CommonDir%/environ		\ ENVIRONMENT?
[then]

softfp? [if]
  include %CpuDir%/sfp32CM0		\ floating point primitives
  include %CommonDir%/sfp32com		\ common floating point code
[then]

\ *************
\ *S End of kernel
\ *************
\ *P After the main kernel has been built, some version data is
\ ** laid down for use by the sign-on code.

buildfile RP2040.no
\ l: version$
\    build$,
\ l: BuildDate$
\    DateTime$,

internal
: .banner       \ --
  cr ." ********************************"
;

: .CPU          \ -- ; display CPU type
  .banner
  cr ." MPE ROM PowerForth for RP2040"
 \  cr version$ $. space BuildDate$ $.
  .banner
;
external

: ANS-FORTH     \ -- ; marker
;


\ *******************
\ *S Application code
\ *******************

Synonym Ser-Key? key?
Synonym Ser-Key  key
Synonym Ser-Emit emit

\ ***************
\ *S Finishing up
\ ***************

libraries       \ to resolve common forward references
  include %CpuDir%/LibM0M1
  include %CommonDir%/LIBRARY
end-libs

decimal

update-build
FINIS


\ ======
\ *> ###
\ ======
