\ FlashProg.fth - Flash programming for STM32

((
Copyright (c) 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *! flashstm32
\ *T STM32 Flash programming
\ ===========

\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the STM32 series, they are designed to be easily expanded
\ ** for future STM32 parts with as yet unknown sector sizes.

((
FlashBase constant FlashBase
_FPEC constant _FPEC

: .FPEC		\ --
  _FPEC $24 pdump  ;
))

: ?Unlock	\ --
\ *G Unlock the Flash if protected.
  _FPEC flCR + @ bit7 and if
    _FPEC  flKEY1 over flKEYR + !	\ write keys
    flKEY2 over flKEYR + !
    bit7 swap flCR + bic!		\ clear LOCK bit
  endif
;

: waitFlNB	\ -- status
\ *G Wait until the Flash is not busy and return the status.
\ ** The top bit is set if the operation timed out.
  $0010:0000 _FPEC begin		\ -- cnt fpec
    swap 1- swap  over			\ check timeout counter
   while
    dup flSR + @ bit0 and 0=
  until then
  dup flSR + @				\ -- cnt fpec status ; get status
  dup $0034 and rot flSR + !		\ -- cnt status ; clear h/w status
  swap 0=
  if  $8000:0000 or  endif
;

: flw!		\ w addr -- status
\ *G Write a 16 bit word to flash. A status value of $20 indicates
\ ** success.
  ?Unlock
  bit0 _FPEC flCR + or!			\ set PG bit in control reg
  w!					\ write data
  waitFlNB				\ wait for completion/timeout
  bit0 _FPEC flCR + bic!		\ clear PG bit in control reg
;

: flPageErase	\ addr -- status
\ *G Erase the page containing addr. A status value of $20
\ ** indicates success.
  FlashBase -				\ relative to start of Flash
  ?Unlock  _FPEC
  bit1 over flCR + or!			\ set PER bit in control reg
  swap over flAR + !			\ set address
  bit6 swap flCR + or!			\ set STRT bit in control reg
  waitFlNB				\ wait for completion/timeout
  bit1 _FPEC flCR + bic!		\ clear PER bit in control reg
;

: .src/dest	\ src dest -- src dest
\ *G Display the addresses and contents of the source
\ ** and destination.
  ."  source: " over .lword  over c@ ." (" .byte ." )"
  ."  dest: " dup .lword  dup c@ ." (" .byte ." )"
;


\ ***************
\ *S Flash driver
\ ***************

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count.

: ?flError	\ status --
\ *G Increment error count if status is not $20.
  $20 <> if  #PrgErrs incr  endif  ;

: (EraseFlash)	\ dest dlen --
\ *G Erase the flash for the given range. Note that complete
\ ** sectors in the range will be erased. If you want to write
\ ** partial sectors, check that they contain $FF in all bytes
\ ** before programming in order avoid having to erase them first.
  bounds ?do
    i flPageErase ?flError
  /FlashPage +loop
;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 2 byte boundary and
\ ** *\i{len} is rounded up to the next 2 byte unit. *\i{Src}
\ ** must be on a 2 byte boundary.
  rot -2 and -rot			\ force src
  swap -2 and swap			\ force dest
  1+ -2 and				\ force len
  bounds ?do				\ -- src'
    dup w@ i flw! ?flError  2+
  2 +loop
  drop
;

: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over c@ over c@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    1 + swap  1 + swap
  1 +loop
  2drop
;

: ProgramFlash	\ src dest len --
\ *G Using the RAM memory buffer at *\i{src}, program the Flash
\ ** at *\i{dest} with *\i{len} bytes. The relevant Flash sectors
\ ** are erased, the Flash is programmed, and the result verified.
  cr ." Erasing " 2dup (EraseFlash)
  0 #PrgErrs !
  cr ." Programming "  3dup (ProgFlash)
  space #PrgErrs ? ." errors"
  cr ." Verifying "  (VerifyFlash)
;


\ ======
\ *> ###
\ ======

decimal
