\ FlashSTM32.fth - Program STM32 Flash in 16 bit units.

((
Copyright (c) 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

\ *************************
\ *! flashstm32
\ *T STM32 Flash operations
\ *************************
\ *P Although some of these routines are not the most efficient
\ ** for the STM32 series, they are designed to be easily expanded
\ ** for future STM32 parts with as yet unknown sector sizes. This
\ ** code is for Flash with a fixed sector/page size. The page size
\ ** is defined by *\fo{/FlashPage} before the file
\ ** *\i{FlashSTM32.fth} is compiled.

internal

((
FlashBase constant FlashBase
_FPEC constant _FPEC

: .FPEC		\ --
  _FPEC $24 pdump  ;
))

: ?Unlock	\ --
\ +G Unlock the Flash if protected.
  _FPEC flCR + @ bit7 and if
    _FPEC  flKEY1 over flKEYR + !	\ write keys
    flKEY2 over flKEYR + !
    bit7 swap flCR + bic!		\ clear LOCK bit
  endif
;

: waitFlNB	\ -- status
\ +G Wait until the Flash is not busy and return the status.
\ +* The top bit is set if the operation timed out.
  $0010:0000 _FPEC begin		\ -- cnt fpec
    swap 1- swap  over			\ check timeout counter
   while
    dup flSR + @ bit0 and 0=
  until then
  dup flSR + @				\ -- cnt fpec status ; get status
  dup $0034 and rot flSR + !		\ -- cnt status ; clear h/w status
  swap 0=
  if  $8000:0000 or  endif
;

: flw!		\ w addr -- status
\ +G Write a 16 bit word to flash. A status value of $20 indicates
\ +* success.
  ?Unlock
  bit0 _FPEC flCR + or!			\ set PG bit in control reg
  w!					\ write data
  waitFlNB				\ wait for completion/timeout
  bit0 _FPEC flCR + bic!		\ clear PG bit in control reg
;

: flPageErase	\ addr -- status
\ +G Erase the page containing addr. A status value of $20
\ +* indicates success.
  FlashBase -				\ relative to start of Flash
  ?Unlock  _FPEC
  bit1 over flCR + or!			\ set PER bit in control reg
  swap over flAR + !			\ set address
  bit6 swap flCR + or!			\ set STRT bit in control reg
  waitFlNB				\ wait for completion/timeout
  bit1 _FPEC flCR + bic!		\ clear PER bit in control reg
;

: (w!f)		\ w addr --
\ +G 16 bit Flash write primitive.
  flw! drop
;

: (c!f)		\ b addr --
\ +G 8 bit Flash write primitive.
  2dup c@ = if  2drop exit  endif
  dup -2 and  swap 1 and if		\ addr bit 0 = 1 (high byte
    swap 8 lshift  over c@ or		\ -- addr' b<<8+b0
  else
    swap  over 1+ c@ 8 lshift or	\ -- addr' b1+b
  then
  swap (w!f)
;

: (!f)		\ x addr --
  over $FFFF and over (w!f)
  swap 16 rshift swap 2+ (w!f)
;

: inFlash?	\ addr -- flag
\ +G Returns true if the address is in the Flash area.
  FLASHSTART FLASHEND within		\ Flash range
;

: c!f		\ b addr --
\ +G Program any address, including Flash, with an 8 bit value.
\ +* After programming one half of a 16 bit unit with anything
\ +* other than $FF, you cannot program the other half.
  dup InFlash?
  if  (c!f)  else c!  endif
;

: w!f		\ w addr --
\ +G Program any address, including Flash, with a 16 bit value.
  dup InFlash?
  if  (w!f)  else  w!  endif
;

: !f		\ x addr --
\ +G Program any address, including Flash, with a 32 bit value.
  dup InFlash?
  if  (!f)  else  !  endif
;

external

: FlErase	\ addr len --
\ *G Erase the Flash sectors in the given range.
  bounds ?do
    i inFlash? if
      i flPageErase drop  /FlashPage
    else
      -1 i !  4
    endif
  +loop
;

: FlProg	\ src dest len --
\ *G Program the Flash at *\i{dest} from memory at *\i{src}.
  bounds ?do
    dup w@ i (w!f)  2+
  2 +loop
  drop
;


\ ======
\ *> ###
\ ======

decimal


