\ Reflash.fth - handles reflashing the system

((
Copyright (c) 2004, 2006, 2008, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20100125 SFP002 Cortex conversion from LPC23xx code.
		Added FlashSize setting.
20060701 SFP001 Added CopyFlash.
))


\ ==========
\ *! reflashstm32
\ *T STM32 Reflashing
\ ==========
\ *S Introduction
\ *P If you destroy the application in the internal Flash, you
\ ** must use the ST ISP loader to reload an Intel Hex file.
\ ** A suitable baud rate is 38400 baud.

\ *P Once the Forth system is running again, you can use the word
\ ** *\fo{REFLASH ( -- )} to download a new binary image.
\ ** Note that *\fo{REFLASH} only handles binary memory image
\ ** files. These should transferred using the XModem 128
\ ** (checksum) protocol. If you are using AIDE with the file
\ ** server enabled, a file selection dialog will appear
\ ** automatically after you have executed *\fo{REFLASH}.

\ *P The STM32 can only be completely reflashed from an
\ ** application by a program running from RAM. A separate
\ ** application built by a control file *\i{Reprog/Reprog*.ctl}
\ ** is used to do this. A binary image, *\fo{REPROG*.IMG}, is
\ ** inserted into the application. When required, the code is
\ ** copied into the internal RAM and executed from RAM.

\ *P Return is by rebooting the system. The contents of the
\ ** internal RAM and Flash are destroyed, therefore any data
\ ** that must be preserved should be saved externally to the
\ ** chip. Alternatively, modify the reprogramming code to
\ ** use the last sector to hold data to be
\ ** preserved.


\ ***************************
\ *S Code in main application
\ ***************************
\ *P The layout of the RAM code is defined in the control file
\ ** for the RAM application. This defines the first cell as
\ ** the Forth word to execute.

create reprog.img	\ -- addr
\ *G The start address of the reprogramming code
  data-file %HwDir%/ReProg/REPROGSTM32F0.IMG equ /reprog	\ -- len
\ *G Generic STM32: The length of the reprogramming code loaded
\ ** into the Flash dictionary.

$2000:0000 equ reprogrun	\ -- addr
\ *G The run time address of the reprogramming code.
\ ** This *\b{must} match the start address of the *\fo{CDATA}
\ ** section defined in *\i{ReprogSTM32.ctl}, and that section
\ ** *\b{must not} overlap the run-time stacks and user area.

0 value ReflashDev	\ -- addr
\ *G Holds 0 or a UART base address. If set to 0, the reflash
\ ** code will use the default device for the XModem transfer.
\ ** Otherwise it will use the value here as the base address
\ ** of the UART to use. Since the same binary reprogramming
\ ** code is used for many systems, it is foolish to make any
\ ** assumptions about the default device.

useUSART1? [if]
console-port 1 = [if]
  _USART1 to ReflashDev
\ *G Set *\fo{ReflashDev} for USART1 if *\fo{console-port}=1.
[then]
[then]
useUSART2? [if]
console-port 2 = [if]
  _USART2 to ReflashDev
\ *G Set *\fo{ReflashDev} for USART2 if *\fo{console-port}=2.
[then]
[then]
useUSART3? [if]
console-port 3 = [if]
  _USART3 to ReflashDev
\ *G Set *\fo{ReflashDev} for USART3 if *\fo{console-port}=3.
[then]
[then]

\ *P The start of the RAM code block holds data about the
\ ** running Forth.
\ *(
\ *B Cell 0 - xt of *\fo{ReprogFlash} to run,
\ *B Cell 1 - xt of *\fo{CopyFlash} to run,
\ *B Cell 2 - 0 for default UART or base address of another UART,
\ *B Cell 3 - clock speed in Hz of the calling Forth,
\ *B Cell 4 - Flash size in bytes,
\ *B Cell 5 - size of each page in the Flash.
\ *)

[undefined] KeepPages [if]
0 equ KeepPages	\ -- u
\ *G Set this non-zero for the number of pages at the end of Flash
\ ** that are reserved for configuration data. Often set to 1 or 2
\ ** by systems that use PowerNet. If not already defined, we use
\ ** the value here.
[then]
/FlashPage KeepPages * equ /Keep	\ -- len
\ *G Number of bytes to keep at the end of Flash for configuration
\ ** data.

: callit	\ xt --
\ *G Load the reprogramming code and execute the xt.
  [defined] dfi [if] dfi [then]		\ disable FIQ
  di					\ disable IRQ [and FIQ]
  ReflashDev				\ get device selection BEFORE copy
  reprog.img reprogrun /reprog cmove	\ copy code to RAM
  reprogrun tuck 2 cells + !		\ set device base address
  system-speed over 3 cells + !		\ set device clock speed
  /Flash /Keep - over 4 cells + !	\ set Flash size
  /FlashPage swap 5 cells + !		\ set Flash page size
  execute
  [defined] reboot [if]  reboot  [else]  ExcVecs 4 + @ execute  [then]
;

: reflash	\ -- ; no exit
\ *G Copies the reprogramming code to the run-time address
\ ** and executes it.
  reprog.img @ callit
;

: CopyFlash	\ src dest len --
\ *G Copy the flash. The parameters are as for *\fo{CMOVE}.
\ ** The process is repeated until there are no errors, and
\ ** the system is then rebooted.
  reprog.img cell + @ callit
;


\ ======
\ *> ###
\ ======

