\ Mcp23S17.fth - read and write I/O pins.

((
Copyright (c) 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)

The minimum procedure to get it working as a simple output port is;
1. Reset
2. Send SPI bytes 0x40,0x0A,0xA0 (configures IOCON reg)
3. Send SPI bytes 0x40,0x10,0x00 (set PORT1 direction all outs)
4. Send SPI bytes 0x40,0x1A (sets PORT1 address ready to receive a byte)
5. (Now every SPI byte sent just writes direct to PORT1 outputs)


To do
=====

Change history
==============
20121218 SFP002 Used compatible SD driver.
20120508 SFP001 First cut.
))

only forth definitions
decimal

\ ===========
\ *! mcp23s17sd
\ *T MCP23S17 I/O Expander
\ ===========
\ *P The MCP23S17 I/O expander is an SPI device that provides two
\ ** 8-bit I/O ports. In this code, the extended addressing feature
\ ** is used so that up to eight devices can be used with the same chip
\ ** select line. The IOCON.BANK signal is left at the power up
\ ** value of 0.

\ *P The MCP23S17 errata sheet notes that if A2=1, the extended
\ ** addressing *\b{must} be used.

\ *P This code uses the low-level SPI driver in
\ ** *\i{Cortex/Hardware/STM32/spiSD-P107dma.fth}. On the Olimex
\ ** STM32-P107 board, SPI3 is used both for the SD card interface
\ ** and on the UEXT connector to which the MCP23S17 devices are
\ ** attached.


\ *********************
\ *S SPI access control
\ *********************

3 8 pio: SPI3_RST	\ PD8 GPIO output ; SFP002
1 15 pio: nCSUEXT	\ PB15 GPIO output 10MHz

: init-spi-uext	\ --
\ *G Set up to use the SPI line on the UEXT connector with SPI3.
\ ** The SPI3 initialisation is performed by the SD driver, which
\ ** uses a different chip select line.
  Output10MHz nCSUEXT isFunction  nCSUEXT setPin
  SPI3_RST isOutput  SPI3_RST setPin	\ Reset high
;
' init-spi-uext AtCold

: rst_low	\ --
\ *G Set the SPI reset line low.
  SPI3_RST clrPin
;

: rst_high	\ --
\ *G Set the SPI reset line high.
  SPI3_RST setPin
;

system-speed spi3-hispeed / 5 / 1 max equ /spi3CSdelay	\ -- u
\ *G Number of iterations for delays around CS operations.

: spi3delay	\ --
\ *G Because the CS line is controlled by software, software
\ ** delays are used around changing the CS line.
  /spi3CSdelay begin  1- dup 0=  until drop
;

: [spi3		\ --
\ *G Start a transaction by asserting CS.
  rst_high
  SecSem request			\ SD card interlock
  spi3delay  nCSUEXT clrPin  spi3delay
;

: spi3]		\ --
\ *G Finish a transaction by de-asserting CS.
  spi3delay  nCSUEXT setPin  spi3delay
  SecSem signal
;

\ *P The I/O
\ I/O addresses for BANK=0
$00 equ rIODIRA
$01 equ rIODIRB
$02 equ rIPOLA
$03 equ rIPOLB
$04 equ rGPINTENA
$05 equ rGPINTENB
$06 equ rDEFVALA
$07 equ rDEFVALB
$08 equ rINTCONA
$09 equ rINTCONB
$0A equ rIOCON
\ $0B equ rIOCON
$0C equ rGPPUA
$0D equ rGPPUB
$0E equ rINTFA
$0F equ rINTFB
$10 equ rINTCAPA
$11 equ rINTCAPB
$12 equ rGPIOA
$13 equ rGPIOB
$14 equ rOLATA
$15 equ rOLATB

0 value CurrS17	\ -- u ; 0..7
\ *G Holds the current device. Set by you as required.

: read8-S17	\ reg -- b
\ *G Read 8 bits from the register.
  [spi3
    $41 CurrS17 1 lshift or spi!	\ select read
    spi!				\ register
    spi@				\ data
  spi3]
;

: write8-S17	\ b reg --
\ *G Write 8 bits to the register.
  [spi3
     $40 CurrS17 1 lshift or spi!	\ select write
     spi!				\ register
     spi!				\ data
  spi3]
;

: resetS17	\ --
\ *G Apply a reset pulse to the UEXT connector and the MCP23S17s.
  rst_low  20 ms  rst_high 20 ms
;

: initS17	\ --
\ *G Initialise the current MCP23S17 device. Must be run after
\ ** the SPI driver has been initialised.
  $28 rIOCON write8-S17			\ SEQOP=1, HAEN=1
  $00 rIODIRA write8-S17		\ port A all outputs
  $FF rIODIRB write8-S17		\ port B all inputs
;

: initS17all	\ --
\ *G Initialise all MCP23S17s that may be on the bus.
  CurrS17  8 0
  do  i to CurrS17  initS17  loop
  to CurrS17
;


\ ********
\ *S Tests
\ ********

1 [if]

2 value msdelay

: tr		\ reg --
  $55 over write8-s17  read8-s17 .byte space
;

: trall		\ --
  cr  $16 0 do  i tr  loop  ;

: trax		\ --
  CurrS17  8 0 do
    cr ." Dev=" i .  i to CurrS17  trall
  loop
  to CurrS17
;

: ta		\ -- ; waggle port A
  0 to CurrS17
  begin
    $00 rGPIOA write8-S17  msdelay ms
    $FF rGPIOA write8-S17  msdelay ms
    key?
  until
;

: tb		\ -- ; waggle port A
  0 to CurrS17
  begin
    $55 rGPIOA write8-S17  200 ms
    $AA rGPIOA write8-S17  200 ms
    key?
  until
;

: tc		\ -- ; waggle port A
  begin
    256 0 do
      i rGPIOA write8-s17
      key? ?leave
      400 ms
    loop
    key?
  until
;


\ ======
\ *> ###
\ ======

decimal

