\ ReprogApp.fth - reprogram LPC2xxx Flash

((
Copyright (c) 2004, 2008, 2009, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
At end of download, check for a partial block to program.

Change max program size to depend on CPU.

Change history
==============
20100203 SFP006 STM32 conversion.
20100125 SFP005 LPC17xx conversion.
20090205 SFP004 LPC2388 does not erase last sector at $0007:D000
20081008 SFP003 Minor documentation changes.
20060701 SFP002 Added CopyFlash.
20050627 MPE002 Now uses sector tables.
20040212 MPE001 First release
))

only forth definitions  decimal


\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the STM32 series, they are designed to be easily expanded
\ ** for future STM32 parts with as yet unknown sector sizes.

((
FlashBase constant FlashBase
_FPEC constant _FPEC

: .FPEC		\ --
  _FPEC $24 pdump  ;
))

: ?Unlock	\ --
\ *G Unlock the Flash if protected.
  _FPEC flCR + @ bit7 and if
    _FPEC  flKEY1 over flKEYR + !	\ write keys
    flKEY2 over flKEYR + !
    bit7 swap flCR + bic!		\ clear LOCK bit
  endif
;

: waitFlNB	\ -- status
\ *G Wait until the Flash is not busy and return the status.
\ ** The top bit is set if the operation timed out.
  $0010:0000 _FPEC begin		\ -- cnt fpec
    swap 1- swap  over			\ check timeout counter
   while
    dup flSR + @ bit0 and 0=
  until then
  dup flSR + @				\ -- cnt fpec status ; get status
  dup $0034 and rot flSR + !		\ -- cnt status ; clear h/w status
  swap 0=
  if  $8000:0000 or  endif
;

: flw!		\ w addr -- status
\ *G Write a byte to flash. A status value of $20 indicates
\ ** success.
  ?Unlock
  bit0 _FPEC flCR + or!			\ set PG bit in control reg
  w!					\ write data
  waitFlNB				\ wait for completion/timeout
  bit0 _FPEC flCR + bic!		\ clear PG bit in control reg
;

: flPageErase	\ addr -- status
\ *G Erase the page containing addr. A status value of $20
\ ** indicates success.
  FlashBase -				\ relative to start of Flash
  ?Unlock  _FPEC
  bit1 over flCR + or!			\ set PER bit in control reg
  swap over flAR + !			\ set address
  bit6 swap flCR + or!			\ set STRT bit in control reg
  waitFlNB				\ wait for completion/timeout
  bit1 _FPEC flCR + bic!		\ clear PER bit in control reg
;


\ ****************
\ *S Flash handler
\ ****************

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count.

 : ?flError	\ status --
\ *G Increment error count if status is not $20.
  $20 <> if  #PrgErrs incr  endif  ;

: (EraseFlash)	\ dest len --
\ *G Erase the Flash range
  bounds ?do
    i flPageErase ?flError
  /FlashPage +loop
;

: EraseFlash	\ --
\ *G Erase the Flash from start to end.
  Flashbase /Flash (EraseFlash)  ;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 2 byte boundary and
\ ** *\i{len} is rounded up to the next 2 byte unit. *\i{Src}
\ ** must be on a 2 byte boundary.
  rot -2 and -rot			\ force src
  swap -2 and swap			\ force dest
  1+ -2 and				\ force len
  bounds ?do				\ -- src'
    dup w@ i flw! ?flError  2+
  2 +loop
  drop
;

((
: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over c@ over c@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    1 + swap  1 + swap
  1 +loop
  2drop
;
))

: toFlash	\ src dest --
\ *G Handle 128 bytes at src to go to dest in Flash.
  #128 (ProgFlash)
;


\ *****************
\ *S Numeric output
\ *****************

: .nibble	\ n --
\ *G Convert a nibble to a hex ASCII digit and display it.
  $0F and  dup 9 >
  if  7 +  then
  $30 + emit
;

: .WORD		\ w --
\ *G Display w as a 4 digit unsigned hexadecimal number.
  dup #12 rshift .nibble
  dup 8 rshift .nibble
  dup 4 rshift .nibble
  .nibble
;

((
: .dword	\ x --
\ *G Display x as an 8 digit unsigned hexadecimal number.
  dup #16 rshift .word .word
;
))


\ ***************
\ *S Reprogrammer
\ ***************

l: m1
  ", Erasing "
l: m2
  ",  done "
l: m3
  ", Start Xmodem TX "
[defined] REBOOT [if]
l: m4
  ", Press R to reboot, other to reprogram "
[else]
l: m4
  ", Reset or press key to reprogram"
[then]
l: m5
  ", Errors"
l: m6
  ", No "

: $.	\ c-addr --
\ *G Display counted string.
  count type
;

: .PrgErrs
\ *G Display number of errors.
  cr  #PrgErrs @ .word  bl emit
  m5 $.
;

: initReprog	\ --
\ *G Get the setup data and calculate the requirements
  init-ser #50 ms  ;

: ReprogFlash	\ --
\ *G Erase and program the Flash.
  initReprog
  begin
    #PrgErrs off
    cr m1 $.  #20 ms
    EraseFlash
    m2 $.  .PrgErrs  #PrgErrs off
    cr m3 $.  $01A emit			\ send message and trigger
    FlashBase /Flash Bin-Up drop
    .PrgErrs
    #100 ms
    cr m4 $.
[defined] REBOOT [if]
    key $DF and dup emit  [char] R =
  until
  #100 ms
  reboot
[else]
    key drop
  again
[then]
;

\ SFP002...
l: msg7
  ", Copy "

: (CopyFlash)	\ src dest len --
\ *G Copy one Flash area to another. The parameters are as for *\fo{CMOVE}.
  initReprog
  cr m1 $.  #20 ms			\ "Erase" message
  #PrgErrs off
  2dup (EraseFlash) .PrgErrs  #10 ms	\ -- src dest len
  #PrgErrs @
  if  3drop exit  endif
  cr msg7 $.				\ "Copy" message
  (ProgFlash)				\ --
  .PrgErrs  #20 ms			\ error count
;

: CopyFlash	\ src dest len --
\ *G Copy the flash. The parameters are as for *\fo{CMOVE}.
\ ** The process is repeated until there are no errors, and
\ ** the system is then rebooted.
  begin
    3dup (CopyFlash)  #PrgErrs @ 0=
  until
  #100 ms  reboot
;
\ ...SFP002


\ ======
\ *> ###
\ ======

