\ spiSD-P107.fth - 8 bit SPI driver for Olimex P107 SD card

((
Copyright (c) 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20121217 SFP003 Converted to SD card use.
20120508 SFP002 Added reset line.
20120501 SFP001 Started.
))

only forth definitions
decimal

\ ================
\ *! sppistm32p107
\ *T SPI Driver for Olimex STM32-P107 board
\ ================
\ *P This code was tested on an Olimex STM32-P107 board which uses
\ ** an STM32F107VCT6 CPU in LQFP 100 pin format.

\ *P The standard SD/MMC card pinout is as follows.
\ *E         SD mode                                 SPI mode
\ ** Pin     Name    Type    Description             Name    Type    Description
\ ** 1       CD/DAT3 I/O/PP  Card detect / data 3    CS      I       Chip select (low)
\ ** 2       CMD     PP      Command/Response line   DI      I       Data input
\ ** 3       Vss1    S       Supply voltage (gnd)    VSS     S       Supply voltage
\ ** 4       Vdd     S       Power supply            VDD     S       Power supply
\ ** 5       CLK     I       Clock                   SCLK    I       Clock
\ ** 6       Vss2    S       Supply voltage          VSS2    S       Supply voltage
\ ** 7       DAT0    I/O/PP  data line 0             O       O/PP    Data output
\ ** 8       DAT1    I/O/PP  data line 1             RSV
\ ** 9       DAT2    I/O/PP  data line 2             RSV
\ *P where the pins are numbered on the connector as below:
\ *E   9 1 2 3 4 5 6 7 8
\ **     * * * * * * * *
\ **   *

\ *P The micro SD card pinout is as follows.
\ *E         SD mode    SPI mode
\ ** Pin     Name       Name    Type    Description
\ ** 1       DAT2       nc
\ ** 2       DAT3       CS      I       Chip select (low)
\ ** 3       CMD        DI      I       Data input (MOSI)
\ ** 4       Vdd        VDD     S       Power supply (max 3v3)
\ ** 5       CLK        SCLK    I       Clock
\ ** 6       Vss        VSS     S       Supply voltage (gnd)
\ ** 7       DAT0       DO      O/PP    Data output (MISO)
\ ** 8       DAT1       nc

\ *E SD card connections
\ ** Pin     Function   CPU
\ ** 1       nc
\ ** 2       nCS        PA4
\ ** 3       MOSI       PC12
\ ** 4       Vdd
\ ** 5       SCLK       PC10
\ ** 6       Vss
\ ** 7       MISO       PC11
\ ** 8       nc


\ ****************
\ *S Configuration
\ ****************

interpreter
: genSPIBR	\ speed clock -- br
\ *G Generate an SPI baud rate value that is less than or equal
\ ** to the given speed.
  /  -1 swap  8 0 do			\ -- flag div
    dup 1 i 1+ lshift <=
    if  nip i swap  leave  endif
  loop
  drop  dup -1 = abort" Cannot generate SPI baud rate"
;
target


\ ****************
\ *S SPI interface
\ ****************
\ *P The port setup code requires *\i{Cortex/Drivers/gpioSTM32.fth}.
\ ** All GPIO ports are clocked and enabled at startup.

200 kHz equ spi3-lospeed	\ -- hz
\ *G SPI3 low speed at startup.
2 Mhz equ spi3-hispeed		\ -- hz
\ *G SPI3 high speed after set up.
apb1-speed spi3-lospeed genSPIBR 3 lshift equ spiBR3lo	\ -- x
\ *G Bit mask for SPI3 low bit rate.
apb1-speed spi3-hispeed genSPIBR 3 lshift equ spiBR3hi	\ -- x
\ *G Bit mask for SPI3 high bit rate.
system-speed spi3-hispeed / 5 / 1 max equ /spi3delay	\ -- u
\ *G Number of iterations for delays around CS operations.
$0304 spiBR3lo or equ /initSPI3_CR1	\ -- x
\ *G 2 lines unidirectional master, 8 bit MSB, no CRC, CPOL=0,
\ ** CPHA=0. N.B. SSM and SSI must be set in master mode.

[undefined] selio-spi3 [if]
\ 3 8 pio: SPI3_RST	\ PD8 GPIO output ; SFP002
2 11 pio: MISO3		\ PC11 AF input
2 12 pio: MOSI3		\ PC12 AF output 50MHz - may be >10MHz
2 10 pio: SCK3		\ PC10 AF output 50MHz - may be >10MHz
0  4 pio: nCSSD		\ PA4 GPIO output 10MHz

: selio-spi3	\ --
\ *G Example to perform clock and pin selection for SPI3.
\ ** This version is used if the word *\fo{selio-spi} has not
\ ** already been defined.
  bit28 _AFIO afMAPR + or!		\ SPI3_REMAP = 1
  MISO3 isInput
  AFOutput50MHz MOSI3 isFunction
  AFOutput50MHz SCK3 isFunction
  Output10MHz nCSSD isFunction
  nCSSD setPin			\ CS high
  \ SPI3 is on APB1
  _RCC
  bit15 over rccAPB1en + or!		\ enable SPI3 clock
  bit15 over rccAPB1rst + bic!		\ take it out of reset
  drop
;
[then]

((
: rst_low	\ --
\ *G Set the SPI reset line low.
  SPI3_RST clrPin
;

: rst_high	\ --
\ *G Set the SPI reset line high.
  SPI3_RST setPin
;
))

: spi3delay	\ --
\ *G Because the CS line is controlled by software, software
\ ** delays are used around changing the CS line.
  /spi3delay begin  1- dup 0=  until drop
;

: init-spi3	\ --
\ *G Initialise SPI3.
  selio-spi3
  _SPI3
  0 over spiCR1 + w!			\ disable SPI3
  /initSPI3_CR1 over spiCR1 + w!	\ 2 lines uni master, 8 bit MSB, no CRC
  0 over spiCR2 + w!			\ no interrupts
  bit6 /initSPI3_CR1 or over spiCR1 + w! \ SPE=1, 2 lines uni master, 8 bit MSB, no CRC
  drop
;
' init-spi3 AtCold

: spi3xchg	\ xout -- xin
\ *G Transmit one byte and read one byte.
  _SPI3
  tuck spiDR + w!			\ set data
  begin					\ wait for TXE=1 and RXNE=1
    dup spiSR + w@ $03 and $03 <>
   while
    pause
  repeat
  spiDR + w@				\ read data
;

: setSPI3br	\ 00xxx000 --
\ *G Write the three BR bits in SPI_CR1.
  _SPI3
  dup spiCR1 + w@
  $FFC7 and  or
  swap spiCR1 + w!
;


\ *************
\ *S SD/MMC API
\ *************
\ *P The file *\i{SDspi.fth} contains a generic driver for SD and
\ ** MMC cards in SPI mode. The low level SPI interface should
\ ** provide the following words:
\ *E MMCon     ( -- )       Drives CS low.
\ ** MMCoff    ( -- )       Drives CS high.
\ ** spiFlush  ( -- )       Flush SPI receive queue (if any).
\ ** spi!      ( c -- )     Write SPI, discard received byte.
\ ** spi@      ( -- c )     Write $FF byte, return received byte.
\ ** spiwr     ( a -- a' )  Reads next byte and writes it to SPI
\ ** spird     ( a -- a' )  Reads SPI and writes to next byte.
\ ** initSDspi ( -- )       Initialise SPI for SD/MMC use.
\ ** spiSlow   ( -- )       Set SPI clock to low speed
\ ** spiFast   ( -- )       Set SPI clock to high speed

: MMCon		\ --
\ *G Start an SD card transaction by asserting CS.
  spi3delay  nCSSD clrPin  spi3delay
;

: MMCoff	\ --
\ *G Finish a transaction by de-asserting CS.
  spi3delay  nCSSD setPin  spi3delay
;

: spiFlush	\ --
\ *G Flush the SPI receive queue
  _SPI3
  begin
    dup spiSR + w@ $01 and		\ RXNE set
   while
    dup spiDR + w@ drop  pause		\ read data
  repeat
  drop
;

: spi!		\ bout --
\ *G Write a byte to the SPI line.
  spi3xchg drop
;

: spi@		\ -- bin
\ *G Read a byte from the SPI line.
  $FF spi3xchg
;

: spiwr		\ caddr -- caddr'
\ *G Read next byte from buffer and write it to SPI, incrementing
\ ** the address.
  dup c@ spi3xchg drop  1+
;

: spird		\ caddr -- caddr'
\ *G Read one byte from the SPI and write it to the buffer, incrementing
\ ** the address.
  $FF spi3xchg over c!  1+
;

: initSDspi	\ --
\ *G Initialise SPI for SD/MMC use.
  init-spi3
;

: spiSlow	\ --
\ *G Set SPI clock to low speed.
  spiBR3lo setSPI3br
;

: spiFast	\ --
\ *G Set SPI clock to high speed.
  spiBR3hi setSPI3br
;


\ ************************
\ *S Card detect extension
\ ************************

: CDin?		\ -- flag
\ *G Return non-zero if a card is present. This is not a fast
\ ** operation as the SPI CS pin is switched to input mode, the
\ ** pin is read and then switched back to output mode. This word
\ ** should also not be used outside the task running the SD card.
  nCSSD isInput				\ floating input
  1000 0 do loop			\ wait a bit
  nCSSD getPin				\ read pin
  Output10MHz nCSSD isFunction		\ back to output
;

: WPin?		\ -- flag
\ *G Return non-zero if the write protect slider is set.
\ ** This always returns 0 as micro SD cards have no write
\ ** protect pin.
   0
;


\ *****
\ Tests
\ *****

0 [if]
_SPI3 constant _SPI3

: tcssd		\ -- ; waggle CS line
  begin
    MMCon  $AA spi3xchg drop  MMCoff  1 ms  key?
  until
;
[then]


\ ======
\ *> ###
\ ======

decimal
