\ FlashProg.fth - Flash programming for STM32F3

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)

Adapted to the STM32F30x by Kurt heinz
Synics AG
Querstrasse 8
CH-8105 Regensdorf

tel: +41 (0)44 840 40 60
net: info@synics.ch
web: www.synics.ch

To do
=====

Change history
==============
20171018 KH002  First fully usable and checked relase for the STM32F30x
20171009 KH001  First adaption trial to STM32F30x
20111205 SFP001 First release.
))

only forth definitions
decimal

\ ===========
\ *! flashstm32f3
\ *T STM32F3 Flash programming
\ ===========
\ *P When Flash is being written or erased, any read access will
\ ** stall the bus until the Flash operation has finished. Thus,
\ ** providing that you do not erase the code in *\i{FlashProg.fth}
\ ** itself, the Flash programming code can be run from Flash.
\ *P Programming of the option byte is not supported.
\ *P The flash is kept unlocked after an erase or a write operation

\ *E
\ **  Table 6. Flash module organization(1)
\ **
\ ** |--------------------|---------------------------|---------|---------------|
\ ** | Flash area         | Flash memory addresses    |  Size   | Name          |
\ ** |                    |                           | (bytes) |               |
\ ** |--------------------|---------------------------|---------|---------------|
\ ** |                    | 0x0800 0000 - 0x0800 07FF |  2 K    | Page 0        |
\ ** |                    | 0x0800 0800 - 0x0800 0FFF |  2 K    | Page 1        |
\ ** |                    | 0x0800 1000 - 0x0800 17FF |  2 K    | Page 2        |
\ ** |Main memory         | 0x0800 1800 - 0x0800 1FFF |  2 K    | Page 3        |
\ ** |                    |            -              |         |               |
\ ** |                    |            -              |         |               |
\ ** |                    |            -              |         |               |
\ ** |                    | 0x0803 F800 - 0x0807 FFFF |  2 K    | Page 255      |
\ ** |--------------------|---------------------------|-------------------------|
\ ** |Information block   | 0x1FFF D800 - 0x1FFF F7FF |  8 K    | System memory |
\ ** |                    | 0x1FFF F800 - 0x1FFF F80F |   16    | Option bytes  |
\ ** |--------------------|---------------------------|-------------------------|
\ ** |                    | 0x4002 2000 - 0x4002 2003 |    4    | FLASH_ACR     |
\ ** |Flash memory        | 0x4002 2004 - 0x4002 2007 |    4    | FLASH_KEYR    |
\ ** |interface registers | 0x4002 2008 - 0x4002 200B |    4    | FLASH_OPTKEYR |
\ ** |                    | 0x4002 200C - 0x4002 200F |    4    | FLASH_SR      |
\ ** |                    | 0x4002 2010 - 0x4002 2013 |    4    | FLASH_CR      |
\ ** |                    | 0x4002 2014 - 0x4002 2017 |    4    | FLASH_AR      |
\ ** |                    | 0x4002 2018 - 0x4002 201B |    4    |  Reserved     |
\ ** |                    | 0x4002 201C - 0x4002 201F |    4    | FLASH_OBR     |
\ ** |                    | 0x4002 2020 - 0x4002 2023 |    4    | FLASH_WRPR    |
\ ** |--------------------|---------------------------|-------------------------|

\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the STM32F3 series, they are designed to be easily expanded
\ ** for future STM32F3 parts with as yet unknown sector sizes.
\ ** Other than at STM32F4, the STM32f3xx only supports 32bit access to registers
\ ** and programming of 16bit values (half-word, no word or byte programming)

$800 equ MainFlashPageSize
\ *G Main flash page size

$045670123 equ fl.KEY1
\ *G STM32F30x specific flash protection key
$0CDEF89AB equ fl.KEY2
\ *G STM32F30x specific flash protection key

bit0  equ CR_PG        \ bit of CR register, Flash programming
bit1  equ CR_PER       \ bit of CR register, page erase
bit2  equ CR_MER       \ bit of CR register, mass earase
bit4  equ CR_OPTPG     \ bit of CR register, option byte programmimg
bit5  equ CR_OPTER     \ bit of CR register, option erase choosen
bit6  equ CR_START     \ bit of CR register, trigger an erase operation
bit7  equ CR_LOCK      \ bit of CR register
bit9  equ CR_OPTWRE    \ bit of CR register, option byte write enable
bit10 equ CR_ERRIE     \ bit of CR register, error interrupt enable
bit12 equ CR_EOPIE     \ bit of CR register, end of operation interrupt enable
bit13 equ CR_OBL_LAUCH \ bit of CR register, force oprtion byte loading

bit1  equ SR_BSY       \ bit of SR register
bit3  equ SR_PGERR     \ bit of SR register
bit4  equ SR_WRPRTERR  \ bit of SR register
bit5  equ SR_EOP       \ bit of SR register

bit31 equ FL_TIMEOUT   \ soft bit

-1 [if]
_Flash constant _Flash

: .FPEC		\ --
  _Flash $24 pdump  ;
[then]

: FlashUnlock	\ --
\ *G Unlock the Flash if it was locked
  _Flash fl.CR + @ CR_LOCK and if	\ locked?
    _Flash  fl.KEY1 over fl.KEYR + !	\ write keys
    fl.KEY2 swap fl.KEYR + !
  endif
;

: waitFlNB	\ -- status
\ *G Wait until the Flash is not busy and return the status.
\ ** The top bit is set if the operation timed out.
  $1000:0000                              \ use this for simple timeout
  _Flash begin		                  \ -- cnt fpec
     swap 1- swap over 0=        	  \ check timeout counter
     over fl.SR + @ SR_BSY and 0= or      \ -- cnt fpec
   until
   fl.SR + dup @                          \ -- cnt *sr status; get status
   dup SR_EOP = if drop 0 then            \ status set to 0 (no error) if=EndOfProgramm operation
   SR_EOP SR_WRPRTERR or SR_PGERR or rot ! \ -- cnt status ; clear h/w status
   swap 0=				  \ timeout check
   if FL_TIMEOUT or  endif                \ notify timeout
;

: wfl!		\ x addr -- status
\ *G Write a 16 bit word to Flash. A status value of zero indicates
\ ** success. 8 and 32 bit write are not supported on STM32F30x.
  FlashUnlock
  CR_PG _Flash fl.CR + !		\ set and PG bits in control reg
  w!					\ write data
  waitFlNB				\ wait for completion/timeout
  CR_PG _Flash fl.CR + bic!		\ clear PG bit in control reg
;

: flSecErase	\ u -- status
\ *G Erase the sector u. A status value of zero indicates success.
  FlashUnlock
  _Flash
  CR_PER over fl.CR + !			\ set page erase
  swap over fl.AR + !		        \ write address
  CR_PER CR_START or swap fl.CR + !	\ start page erase
  waitFlNB				\ wait for completion/timeout
  0 _Flash fl.CR + !		        \ clear CR reg
;

: .src/dest	\ src dest -- src dest
\ *G Display the addresses and contents of the source
\ ** and destination.
  ."  source: " over .lword  over c@ ." (" .byte ." )"
  ."  dest: " dup .lword  dup c@ ." (" .byte ." )"
;

\ ***************
\ *S Flash driver
\ ***************

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count.

: ?flError	\ status --
\ *G Increment error count if status is not zero.
  if  #PrgErrs incr  endif  ;


: (EraseFlash)	\ dest dlen --
\ *G Erase the Flash for the given range. Note that complete
\ ** sectors in the range will be erased.
  MainFlashPageSize + MainFlashPageSize / 0 ?do
    dup flSecErase ?flError
    MainFlashPageSize +
  loop
  drop
;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 2 byte boundary and
\ ** *\i{len} is rounded up to the next 2 byte unit. *\i{Src}
\ ** must be on a 2 byte boundary and must be in RAM.
  rot -2 and -rot			\ force src
  swap -2 and swap			\ force dest
  3 + -2 and				\ force len
  bounds ?do				\ -- src'
    dup @ i wfl! ?flError
    2+
  2 +loop
  drop
;

: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over c@ over c@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    1 + swap  1 + swap
  1 +loop
  2drop
;

: ProgramFlash	\ src dest len --
\ *G Using the RAM memory buffer at *\i{src}, program the Flash
\ ** at *\i{dest} with *\i{len} bytes. The relevant Flash sectors
\ ** are erased, the Flash is programmed, and the result verified.
  cr ." Erasing " 2dup (EraseFlash)
  0 #PrgErrs !
  cr ." Programming "  3dup (ProgFlash)
  space #PrgErrs ? ." errors"
  cr ." Verifying "  (VerifyFlash)
;


\ ======
\ *> ###
\ ======

decimal
