\ I2C bit-bang driver for STM32F30x board

\  Copyright (c) 2009-2016
\  Synics AG
\
\  tel: +41 (0)44 840 40 60
\  net: info@synics.ch
\  web: www.synics.ch
\

\ The  bit handling part is from  MPE i2cSTM32F4xxbb.fth
\ Copyright (c) 1999
\ MicroProcessor Engineering
\ 133 Hill Lane
\ Southampton SO15 5AF
\ England
\
\ tel: +44 (0)2380 631441
\ fax: +44 (0)2380 339691
\ net: mpe@mpeforth.com
\      tech-support@mpeforth.com
\ web: www.mpeforth.com

\ ---------------------------------------------------------

\ *! i2cSTM32F4xxbb
\ *T STM32F4xx software I2C driver
\ ===============
\ *S Introduction
\ ===============
\ *P The I2C interface is a software bit-banging system using
\ ** GPIO bits PB8 (SCL) and PB9 (SDA). The initialisation code
\ ** sets up these bits as open drain outputs.

\ *P Against the original MPE versionm, the timing is made by using TIMER@ (see direct_timer.fth)
\ ** instead of cycle counting mechanism. The proposed solution is for single master but with
\ ** a detection of too slow I2C device (low state stretching).
\ ** There is no handling of BUS BUSY as the CPU is always master.


decimal

\ ============================
\ *S I2C bit physical coupling
\ ============================
0 5 pio: SCLio	\ --
\ *G The I/O bit for SCL.
0 7 pio: SDAio	\ --
\ *G The I/O bit for SDA.

#400 kHz equ i2c-speed	\ -- hz
\ *G Define the I2C clock speed in Herz.

\ ====================
\ *S I2C bit functions
\ ====================

10usTxTICKS   equ I2C_BIT_TIME
\ *G Normal bit time for a 100kHz I2C

1usTxTICKS   equ I2C_SETLING_TIME
\ *G Setling time until a bit may react physically on positive edges

100usTxTICKS   equ I2C_MAX_BIT_TIME
\ *G Maximal allowed stretched time

: i2cdelay \ --
  I2C_BIT_TIME 2 / short_del
;

: i2csettling \ --
  I2C_SETLING_TIME short_del
;

: read_scl	\ -- bit
\ *G Read SCL bit and return in l.s. bit. No delays.
  SCLio getPin
;

: write_scl	\ b -- ; output clock bit
\ *G Write to SCL with no delays.
  SCLio swap
  if setPin else clrPin  endif
;

: SCL_low	\ --
\ *G Set SCL low. A quarter cycle delay is performed
\ ** before and after the transition.
  SCLio clrPin
;

: SCL_high	\ --
\ *G Set SCL high. A quarter cycle delay is performed
\ ** before and after the transition.
  SCLio setPin
;

: read_sda	\ -- bit
\ *G Read SCL bit and return in l.s. bit. No delays.
  SDAio getPin
;

: write_sda	\ b -- ; output to SDA
\ *G Write the SDA bit.
  SDAio swap
  if  setPin  else  clrPin  endif
;

: sda_low	\ --
\ *G Set SDA low.
  SDAio clrPin
;

: sda_high	\ --
\ *G Set SDA high.
  SDAio setPin
;

: initI2C	\ --
\ *G Initialise the I2C hardware. Performed after reset. This
\ ** code needs *\fo{Cortex/Drivers/gpioSTM32F30x.fth}.
  SCLio dup isOutput  1 over isPinOD  dup PulledUp setPin
  SDAio dup isOutput  1 over isPinOD  dup PulledUp setPin
;

' initI2C atCold

\ ------------------------ driver --------------------------

\ definitions

$301 equ I2C_STRETCHOVERFLOW
\ *G THROW code is case of time overflow while stretching
$302 equ I2C_ARBITRATION_LOST
\ *G THROW code is case of I2C arbitration lost.
\ ** This case is detected but not completely handled! And loss
\ ** of arbitration may lose the frame.
$303 equ I2C_BUS_BUSY_ERROR
\ *G THROW code is case of I2C arbitration loss.
\ ** This case is detected but not completely handled!
\ ** And loss of arbitration may lose the frame.
$306 equ I2C_DEVICE_NOTPRESENT
\ *G THROW code is case of a I2C device is not present.
   1 equ I2C_R_BIT

\ 2014-06-20 kh code rearranged. Comments added, devices example moved to i2c_devices.fth
\ 2012-06-07 kh code migrated on STM32 for a TL74 temperature sensor
\ 2011-10-27 kh first release written in C for a sauna sensor (Sensirion SHT21)

\ --------------------

: i2c_start \ --;
\ *G Generate a START or RESTART condition
  sda_high I2C_BIT_TIME 2 / short_del
  scl_high I2C_BIT_TIME 2 / short_del
  read_sda 0= I2C_BUS_BUSY_ERROR ?THROW \ data must have been high to produce a START CONDITION
  timer@
  begin
    I2C_SETLING_TIME short_del
    read_scl 0=      \ allow stretching time
  while
    timer@ over - I2C_MAX_BIT_TIME > I2C_STRETCHOVERFLOW ?throw
  repeat
  drop
  sda_low i2cdelay
  scl_low i2cdelay
;

: i2c_stop \ --;
\ *G Generate a STOP condition
  scl_low I2C_SETLING_TIME short_del
  sda_low i2cdelay
  scl_high
  timer@
  begin
    read_scl 0=      \ allow stretching time
    I2C_SETLING_TIME short_del
  while
    timer@ over - I2C_MAX_BIT_TIME > I2C_STRETCHOVERFLOW ?throw
  repeat
  drop
  i2cdelay
  sda_high
  timer@
  begin
    I2C_SETLING_TIME short_del
    read_sda 0=      \ allow stretching time
  while
    timer@ over - I2C_MAX_BIT_TIME > I2C_STRETCHOVERFLOW ?throw
  repeat
  drop
;

: i2c_master_transmit \ byte -- ackn;
\ *G Transmit one byte in MASTER TRANSMIT mode with ACKN as answer
  $80   \ use as sampler bit for a 8 bit transmission
  begin
    2dup and 0<> dup write_sda             \ convert to logical and write
    scl_high i2cdelay
    read_sda 0<> <> I2C_BUS_BUSY_ERROR ?throw  \ check for data on bus  (no collsion avoidance in this release, only detection)
    timer@
    begin
      read_scl 0=                           \ allow stretching time
    while
      timer@ over - I2C_MAX_BIT_TIME > I2C_STRETCHOVERFLOW ?throw
    repeat
    drop
    scl_low
    i2cdelay
    1 rshift
  dup 0= until
  2drop
  sda_high                                  \ quest for ACKN bit
  i2csettling
  scl_high
  i2cdelay
  timer@
  begin
    read_scl 0=                             \ allow stretching time
  while
    timer@ over - I2C_MAX_BIT_TIME > I2C_STRETCHOVERFLOW ?throw
  repeat
  drop
  read_sda                                  \  read the ACKN as 9th bit
  scl_low i2cdelay
;

: i2c_master_receive \ ackn -- byte;
\ *G Read one byte from I2c and drive the ACKN as 9th bit
  $80 0             \ sampler and result
  sda_high
  begin
    i2cdelay
    read_sda if
      over or          \ affect bit
    then
    scl_high i2cdelay
    timer@
    begin
      read_scl 0=                             \ allow stretching time
    while
      timer@ over - I2C_MAX_BIT_TIME > I2C_STRETCHOVERFLOW ?throw
    repeat
    drop
    scl_low
    swap 1 rshift swap
  over 0= until
  nip
  swap write_sda i2cdelay                     \ write ACKN
  scl_high i2cdelay
  timer@
  begin
    read_scl 0=                               \ allow stretching time
  while
    timer@ over - I2C_MAX_BIT_TIME > I2C_STRETCHOVERFLOW ?throw
  repeat
  drop
  scl_low
;

: i2c_c@ \ reg adr -- val;
  i2c_start
  dup  i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw
  swap i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw
  i2c_start              \ restart for reading
  I2C_R_BIT or i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw
  1 i2c_master_receive i2c_stop
;

: i2c_c! \ n reg adr -- ;
  i2c_start
  i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw
  i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw
  i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw
  i2c_stop
;

variable  ic2_field_p
$40 value ic2_field_size

: i2c_field@ \ field_adr size adrp adr -- val;
\ *P Read a data field of i2c data where:
\ ** adr       is the i2c device address (8bit, not 7bit rule)
\ ** adrp      is the data pointer of teh first byte to read
\ **           ATTENTION: usually the bit 7 of adrp should be set for autoincrement of the data pointer
\ ** size      teh number of read bytes
\ ** field_adr is the pointer of the result field
  i2c_start
  dup  i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw \ set device adr
  swap i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw \ set data pointer, ATTENTION: usually the bit 7 should be set for data pointer autoincrement
  i2c_start              \ restart for reading
  I2C_R_BIT or i2c_master_transmit I2C_DEVICE_NOTPRESENT ?throw
  bounds
  over 1- -rot           \ save last ptr
  ?do
    dup i =              \ 0 (=ACKN) for all but last (=NACKN)
    i2c_master_receive i c!
  loop
  drop
  i2c_stop
;


: dump_device \ start adr --;
  swap >r
  ic2_field_p dup @ free drop          \ free receiver area
  ic2_field_size allocate drop over !  \ define new one
  @ ic2_field_size rot r> swap i2c_field@
  ic2_field_p @ ic2_field_size dump
;

