\ methods. Ref: Terry Rayburn, euroForml 87
\ modified K. Heinz, Synics AG
\
\ Copyright (c) 2007-2010
\ Synics AG
\
\ tel: +41 (0)44 840 40 60
\ net: info@synics.ch
\ web: www.synics.ch

\ history:
\ 2012-11-12 kh IBUFFER corrected
\ 2010-11-11 kh IBUFFER: added
\ 2010-08-11 kh field size of safe variables corrected, alignment limitation solved
\ 2010-07-06 kh cell safe objects added, check error now uses THROW
\ 2005-08-14 kh final version of safety variables developped for the SSU project
\
\ concept: each object has a certain number of methods. These methods are standardized and coded
\ with a simple number. Advantage: simple (fast) runtime access, disadvantage: unable to recognize a wrong method
\ if no other mechanism is planed.

\ This version is based on late binding (method is searched at run time).

\ object layout:
\ struct
\    cell field object_adr
\    cell field object_method_table_adr
\ n cells field optional_parameter_field
\ end-struct

: method:
  create cell * ,
  does> @ over cell + @ + @ swap @ swap execute ;  \ I agree, it is obcure...

\ the method is compatible to following create does> word
\ : my_object
\ create xyz ,   \ the value xyz is the parameter on the stack for the method
\ method_list ,  \ the pointer method_list point on a cell structure with all methods
\ does> ;

decimal

0 method: get         \ n1 -- n2; n1: pointer to object structure, n2: result
1 method: put         \ n1 n2 -- ; n1: value, n2: object
2 method: get_adr     \ n1 -- n2 ; n1: object adr, n2: object data adr
3 method: object_init \ n1 -- ;    n1: object adr,
4 method: ?valid      \ n1 -- n2 ; n1: object adr, n2: flag with the validity of the object

\ ------------------------ data types: -----------------------------

: illegal_method drop ." unkown method for this object " ;

\ bus size RAM variable

align  here
  ' @ ,
  ' ! ,
  ' noop ,
  ' off  ,
  ' illegal_method ,
  equ variable_methods

interpreter

: obj_variable \ -- ;
  cdata create
  ihere(t) , cell iallot(t)                   \ variable pointer
  variable_methods ,
  does> ;                                     \ -- n1 , pointer to object structure

target

\ 8 bit RAM variable

: reset_byte \ n --
 0 c!
;

  here
  ' c@ ,
  ' c! ,
  ' noop ,
  ' reset_byte  ,
  ' illegal_method ,
  equ byte_variable_methods

interpreter

: obj_byte_variable \ -- ;
  cdata create
  ihere(t) , 1 iallot(t)                   \ variable pointer
  byte_variable_methods ,
  does> ;                                     \ -- n1 , pointer to object structure

target

obj_variable bidon              \ test for object variable

\ -------------- unaligned memory r/w ----------------------

2 constant wcell

\ ---- little endian ---

: le-l@ \ caddr -- n, like @ but without the alignment restriction
  dup c@
  swap 1+ swap over c@  8 lshift or
  swap 1+ swap over c@ 16 lshift or
  swap 1+           c@ 24 lshift or
;

: le-l! \ n caddr --
  2dup c!
  swap 8 rshift swap 1+ 2dup c!
  swap 8 rshift swap 1+ 2dup c!
  swap 8 rshift swap 1+ c!
;

: le-w@ \ caddr -- n, like @ but without the alignment restriction
  dup c@
  swap 1+ c@  8 lshift or
;

: le-w! \ n caddr --
  2dup c!
  swap 8 rshift swap 1+ c!
;

(( more elegant.... but 3 time slower and about the same code size in ARM7

: le-l@		\ caddr -- x
  dup le-w@  swap 2+ le-w@ #16 lshift or
;


: le-l!		\ w caddr --
  2dup le-w!  swap #16 rshift swap 2+ le-w!
;

))

: le-x! \ d adr --;
\ *G 64 bit operator, strictly little endian
  rot over le-l!
  cell+ le-l!
;

: le-x@ \ adr --;
\ *G 64 bit operator, strictly little endian
  dup le-l@ swap cell+ le-l@
;

\ ------ big endian -----

: be-w@ \ addr -- w ,
  count 8 lshift swap c@ or
;

: be-w! \ w addr --,
  over 8 rshift over c! 1+ c!
;

: be-l@ \ addr -- n ;
  dup be-w@ 16 lshift swap wcell + be-w@ or
;

: be-l! \ n addr -- ;
  over 16 rshift over be-w!
  wcell + be-w!
;

\ ------------------- safety relevant cells--------------------------

\ this array uses duplicated wcells (16bits)
\ each of it has its value complement
\ modified: 2009-11-28 kh

\ Usable prims in practice:
\   1   3   5   7  11  13  17  19  23  29  31  37  41  43  47  53  59  61  67  71
\  73  79  83  89  97 101 103 107 109 113 127 131 137 139 149 151 157 163 167 173
\ 179 181 191 193 197 199 211 223 227 229 233 239 241 251 257 263 269 271 277 281
\ 283 293 307 311 313 317 331 337 347 349 353 359 367 373 379 383 389 397 401 409
\ 419 421 431 433 439 443 449 457 461 463 467 479 487 491 499 503 509

#37 equ SAFETY_ZONE_SIZE    \ number should be prim !

101 equ SAFETY_CELL_ERROR   \ THROW code if safety cell failed

SAFETY_ZONE_SIZE 2* buffer: safety_ram
SAFETY_ZONE_SIZE equ SAFE_IMAGE_OFFSET

internal
variable safety_cell_pointer
external

safety_ram safety_cell_pointer !

: init_safe_area \ --
  safety_ram SAFETY_ZONE_SIZE 2dup 0 fill
  dup rot + swap $ff fill
;

' init_safe_area AtCold

\ each variable has a complemented copy in a offsetted array

((
: handle_error \ n --
  safety_ram - wcell / . ." cell integrity failure " cr ;

defer cell_error \ n --

assign handle_error to-do cell_error
))

\ ---------------- 16 bit safety cells -----------------------

$ffff equ W_MASK

: safe_w! \ n adr -- ,
  over invert \ build complement
  over SAFE_IMAGE_OFFSET + le-w! le-w!
;

: ?valid_w \ n -- t/f
  dup SAFE_IMAGE_OFFSET + le-w@
  W_MASK xor swap le-w@  =
;

: safe_w@ \ adr -- n ,
  dup >r SAFE_IMAGE_OFFSET + le-w@ W_MASK xor r@ le-w@ \ -- n n'
  over <> if
\     drop -1 r@ cell_error
     SAFETY_CELL_ERROR throw
  then
  r> drop
;

: safe_winit \ adr --,
  0 swap safe_w!
;


\ 16 bit safety variable

  align here
  ' safe_w@ ,
  ' safe_w! ,
  ' noop ,
  ' safe_winit  ,
  ' ?valid_w ,
  ' illegal_method ,
  equ safety_wvariable_methods

interpreter

: obj_safety_wvariable \ -- ;
  create
  wcell safety_cell_pointer dup @ , +!
  safety_wvariable_methods ,
  does> ;                                     \ -- n1 , pointer to object structure

target

\ ---------------- 32 bit safety cells -----------------------

: safe_! \ n adr -- ,
  over invert \ build complement
  over SAFE_IMAGE_OFFSET + le-l! le-l! ;

: ?valid_cell \ n -- t/f
  dup SAFE_IMAGE_OFFSET + le-l@ invert swap le-l@  = ;

: safe_@ \ adr -- n ,
  dup >r SAFE_IMAGE_OFFSET + le-l@ invert r@ le-l@ \ -- n n'
  over <> if
\     drop -1 r@ cell_error
     SAFETY_CELL_ERROR throw
  then
  r> drop ;

: safe_cellinit \ adr --,
  0 swap safe_! ;

\ 32 bit safety variable

  align here
  ' safe_@ ,
  ' safe_! ,
  ' noop ,
  ' safe_cellinit  ,
  ' ?valid_cell ,
  ' illegal_method ,
  equ safety_variable_methods

interpreter

: obj_safety_variable \ -- ;
  create
  cell safety_cell_pointer dup @ , +!
  safety_variable_methods ,
  does> ;                                     \ -- n1 , pointer to object structure

target

\ test cells

obj_safety_variable mycell1
obj_safety_variable mycell2

\ ----------------- ibuffer ------------------------

interpreter

: ibuffer: \ n --;
\ *G Create a buffer of size n bytes like BUFFER: but inside the initialised zone.
\ ** Advantage: no init
  variable
  cell - 0 max iallot
;

target

\ --- wcell values

: signed_w@ \ adr -- n
\ *G Read a signed w value pointed by
  w@
  dup 1 15 lshift and if \ transform to signed_w@
    -1 15 lshift or
  then
;

: w>l \ n1 -- n2
\ *G Convert a w value signed to a cell sized value
  dup 1 15 lshift and if \ transform to signed
    -1 15 lshift or
  then
;
