\ -----simple direct timers ----------------
\
\ Copyright (c) 2007-2012
\ Synics AG
\
\ tel: +41 (0)44 840 40 60
\ net: info@synics.ch
\ web: www.synics.ch
\

\ =================
\ *! Direct timers
\ =================
\ *T Direct timers - Direct_Timers.fth

\ *G The timing is based on a free running 32 bit up counter.
\ ** The time checking words must be able to deal with circular arithmetic
\ ** On CortexM4 STM32F407xx, TIM2 or TIM5 may be used for it

0
1  0 lshift or \ CEN
0  1 lshift or \ UDIS
0  2 lshift or \ URS
0  3 lshift or \ OPM
0  4 lshift or \ DIR
0  5 lshift or \ CMS
0  7 lshift or \ ARPE
0  8 lshift or \ CKD
equ TIMx_CR1_VALUE

\ -------- timer equates ---------
\ this part is CPU and implementation specific

_TIM2  equ _TIMERx
\ *G Define the physical timer

bit0   equ cken_bitx
APB1-speed APB1div 1 <> [if] 2 * [then] equ APBx-speed
\ *G Define the speed for the direct timer

2 equ TIMEx_PRESCALER
\ *G 1 is the best resolution of TIMEx_PRESCALER, but take care,
\ ** the resolution is a compromise against the precision due to
\ ** the integer approx

: start_freerun_timer \ --
\ *G Start the timer counter
  _RCC
  cken_bitx over rccAPB1ENR + or!       \ enable clock to TIM2
  drop
  _TIMERx
  TIMEx_PRESCALER  1- over timPSC + w!  \ set the prescaler
  TIMx_CR1_VALUE      over timCR1 + w!  \ enable the timer as up counter
  1000 0 do
    i drop                              \ consume a bit time
  loop
  1                   over timEGR + w!  \ create an event to load prescaler
  drop
;

' start_freerun_timer AtCold

1 [if]

: .prescs \ --; debugging
  _RCC rccCFGR +
  @ dup ." AHBdiv: " 4 rshift $0f and .
  dup ." APB1div: " 10 rshift 7 and .
  ." APB2div: " 13 rshift 7 and .
;

[then]

1000000000 equ DIRECT_TIMER_RESOLUTION   \ 1sec in ns
100 equ ANTI_OVERFLOW_FACT

DIRECT_TIMER_RESOLUTION ANTI_OVERFLOW_FACT /
TIMEx_PRESCALER
APBx-speed ANTI_OVERFLOW_FACT /
*/ equ TxTICKSIZE        \ tick size in ns

APBx-speed  50 1000000 TIMEx_PRESCALER * */ equ 50usTxTICKS  \ examples
APBx-speed  20 1000000 TIMEx_PRESCALER * */ equ 20usTxTICKS
APBx-speed  10 1000000 TIMEx_PRESCALER * */ equ 10usTxTICKS
APBx-speed   2 1000000 TIMEx_PRESCALER * */ equ 2usTxTICKS
APBx-speed 100 1000000 TIMEx_PRESCALER * */ equ 100usTxTICKS


: timer@ \ -- n
\ *G Get the current value of the free running 32 bit up counter.
\ ** On a LPC2xxx, the fastest direct timer tick is about 56ns @ 72 MHz,
\ ** on STM32F4xx (Cortex M4) it is 12ns @ 168 MHz (24ns in this example)
 _TIMERx timCNT + @
;

: short_del \ n --
\ *G Wait n cycles of the free running timer
  timer@ +
  begin
    timer@ over - 0>
  until drop
;

: do_delay \ n --
\ *G Wait n cycles of the free running timer including a *\fo{PAUSE} inside the loop
  timer@ +
  begin
    pause
    timer@ over - 0>
  until drop
 ;

: us>ticks \ n1 -- n2;
\ *G Convert us seconds to TxTICKS
  APBx-speed swap 1000000 TIMEx_PRESCALER * */
;

\ *G How to define a time in TxTICKs:
\ *C              .--- this value is in usec
\ *C              v
\ *C APBx-speed 100 1000000 TIMEx_PRESCALER * */ constant 100usTxTICK

interpreter

: ns>ticks \ n1 -- n2;
\ *G Convert ns seconds to TxTICKS
  APBx-speed ANTI_OVERFLOW_FACT / swap
  DIRECT_TIMER_RESOLUTION ANTI_OVERFLOW_FACT /
  TIMEx_PRESCALER * */
;

target

: ms_delay \ n -- ;
\ *G The delay consumed by this function is not very precise! -> not
\ ** suitable for watches. This word may be used for startup as
\ ** long as the interrupt used by *\fo{TICKS} is not initialized
\ ** and *\fo{MS} is not yet usable.
  1000 * us>ticks do_delay
;

: (exec_time) \ adr -- dt
\ *G Get the raw time of a word in TxTICKS
   timer@ >r
   execute
   timer@ r> -
 ;

: ticks>ns \ n1 -- n2
\ *G Convert a number of TxTICKS in ns
  DIRECT_TIMER_RESOLUTION ANTI_OVERFLOW_FACT /
  TIMEx_PRESCALER *
  system-speed ANTI_OVERFLOW_FACT /
  */
;

: .txticks \ n -- ;
\ *G Print a TxTICKS value in usecs
  DIRECT_TIMER_RESOLUTION ANTI_OVERFLOW_FACT /
  TIMEx_PRESCALER *
  APBx-speed ANTI_OVERFLOW_FACT /
  */
  100 / s>d <# # [char] , hold #s #> type ."  usec"
;

: .exec_time \ adr -- ; usage: ' <myfunction> exec_time
\ *G Print the time consumed by a function in usecs.
\ ** Usage example:
\ *C ' <myfunction> exec_time
  [i ['] noop (exec_time) i] >r
  (exec_time) r> -
  0 max
  cr
  .txticks
;

\ ----------------

: cyclic_event_sync? \ n adr -- t/f
\ *G Check if TIMERx value is bigger than the value in *\i{adr}
\ ** The volatile timer *\i{adr} is checked, if (actual timer - volatile value )> n,
\ ** then the flag is set. The volatile value is increased by *\i{n}.
  timer@ over @ - 2 pick >= if
    +! true                     \ the timer is increased by n since last update
  else
    2drop false
  then
;

: cyclic_event_async? \ n adr -- t/f, check if TIMERx value is bigger of the value in adr
\ *G The volatile timer adr is checked, if (actual timer - volatile value )> n,
\ ** then the flag is set. The actual timer is stored in the volatile value.
  timer@ over @ - 2 pick >= if
    nip timer@ swap ! true       \ the timer is refreshed from now
  else
    2drop false
  then
;
