\ FlashProg.fth - Flash programming for STM32F4

((
Copyright (c) 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20151015 SFP002 Added 128 byte routines and changed for later
                STM32F4 revisions.
20111205 SFP001 First release.
))

only forth definitions
decimal

\ ===========
\ *! flashstm32f4
\ *T STM32F4 Flash programming
\ ===========
\ *P When Flash is being written or erased, any read access will
\ ** stall the bus until the Flash operation has finished. Thus,
\ ** providing that you do not erase the code in *\i{FlashProg.fth}
\ ** itself, the Flash programming code can be run from Flash.

\ ****************
\ *S Sector tables
\ ****************
\ *P Sector tables contain the number of sectors and starting offset
\ ** of each sector, plus a dummy start address which enables
\ ** the size of the last sector to be calculated.

create sectab	\ -- addr
\ *G Generic STM32F4xx sector table for CPUs with up to 1 Mb of
\ ** Flash. Note that the addresses are offsets from the start of
\ ** the Flash.
  #12 ,
  $0000:0000 ,  $0000:4000 ,  $0000:8000 ,  $0000:C000 ,
  $0001:0000 ,  $0002:0000 ,  $0004:0000 ,  $0006:0000 ,
  $0008:0000 ,  $000A:0000 ,  $000C:0000 ,  $000E:0000 ,
  $0010:0000 ,

: SectorN	\ n -- offset len
\ *G Convert sector number (zero based) to sector offset and length
  1+ cells sectab + 2@			\ -- end start
  tuck -
;

: FindSecN	\ addr -- n
\ *G Find the sector number containing address addr. If
\ ** addr is outside the internal Flash range, n is set to -1.
\  FlashBase -				\ address to offset
  $0800:0000 -				\ address to offset
  sectab @ 0 ?do			\ -- addr
    dup i SectorN over + within if	\ test sector
      drop i  unloop exit
    endif
  loop
  drop -1
;


\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the STM32F4 series, they are designed to be easily expanded
\ ** for future STM32F4 parts with as yet unknown sector sizes.

((
FlashBase constant FlashBase
_FPEC constant _FPEC

: .FPEC		\ --
  _FPEC $24 pdump  ;
))

: ?Unlock	\ --
\ *G Unlock the Flash.
  _FPEC flCR + @ bit31 and if		\ locked?
    _FPEC  flKEY1 over flKEYR + !	\ write keys
    flKEY2 swap flKEYR + !
  endif
;

: waitFlNB	\ -- status
\ *G Wait until the Flash is not busy and return the status.
\ ** The top bit is set if the operation timed out.
  $0100:0000 _FPEC begin		\ -- cnt fpec
    swap 1- swap  over			\ check timeout counter
   while
    dup flSR + @ bit16 and 0=
  until then
  dup flSR + @				\ -- cnt fpec status ; get status
  dup $00F3 and rot flSR + !		\ -- cnt status ; clear h/w status
  swap 0=				\ timeout check
  if  bit31 or  endif
;

: fl!		\ x addr -- status
\ *G Write a 32 bit word to Flash. A status value of zero indicates
\ ** success.
  ?Unlock
  $0201 _FPEC flCR + !			\ set PSIZE and PG bits in control reg
  !					\ write data
  waitFlNB				\ wait for completion/timeout
  bit0 _FPEC flCR + bic!		\ clear PG bit in control reg
;

: flSecErase	\ u -- status
\ *G Erase the sector u. A status value of zero indicates success.
  ?Unlock
  $0F and 3 lshift $0202 or _FPEC	\ PSIZE=2 for 32 bit, sec#, SER
  2dup flCR + !				\ set required data
  swap bit16 or swap flCR + !		\ set STRT bit
  waitFlNB				\ wait for completion/timeout
  bit1 _FPEC flCR + bic!		\ clear SER bit in control reg
;

: .src/dest	\ src dest -- src dest
\ *G Display the addresses and contents of the source
\ ** and destination.
  ."  source: " over .lword  over c@ ." (" .byte ." )"
  ."  dest: " dup .lword  dup c@ ." (" .byte ." )"
;


\ ***************
\ *S Flash driver
\ ***************

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count.

: ?flError	\ status --
\ *G Increment error count if status is not zero.
  if  #PrgErrs incr  endif  ;

: (EraseFlash)	\ dest dlen --
\ *G Erase the Flash for the given range. Note that complete
\ ** sectors in the range will be erased.
  over + 1-				\ -- first last
  FindSecN swap FindSecN 		\ -- nl nf
  begin
    dup flSecErase ?flError
    2dup <>
   while
    1+
  repeat
  2drop
;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 4 byte boundary and
\ ** *\i{len} is rounded up to the next 4 byte unit. *\i{Src}
\ ** must be on a 4 byte boundary and must be in RAM.
  rot -4 and -rot			\ force src
  swap -4 and swap			\ force dest
  3 + -4 and				\ force len
  bounds ?do				\ -- src'
    dup @ i fl! ?flError  4+
  4 +loop
  drop
;

: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over c@ over c@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    1 + swap  1 + swap
  1 +loop
  2drop
;

: ProgramFlash	\ src dest len --
\ *G Using the RAM memory buffer at *\i{src}, program the Flash
\ ** at *\i{dest} with *\i{len} bytes. The relevant Flash sectors
\ ** are erased, the Flash is programmed, and the result verified.
  cr ." Erasing " 2dup (EraseFlash)
  0 #PrgErrs !
  cr ." Programming "  3dup (ProgFlash)
  space #PrgErrs ? ." errors"
  cr ." Verifying "  (VerifyFlash)
;


\ ********************************
\ *S Programming in 128 byte units
\ ********************************

: StartProg128	\ dest dlen -- flag ; 0=success
  $F3 _FPEC flSR + !			\ clear status reg by writing ones
  #PrgErrs off  (EraseFlash)  #PrgErrs @
;

: (Prog128)	\ src dest -- flag ; 0=success
\ *G Write 128 bytes from memory at *\i{src} to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 4 byte boundary. *\i{Src}
\ ** may be unaligned if the CPU supports it by default, and
\ ** must be in RAM.
  #PrgErrs @ >r				\ error counter
  -4 and				\ force dest to 4 byte boundary
  #128 bounds ?do			\ -- src'
    dup @ i fl! ?flError  4+
  4 +loop
  drop
  r> #PrgErrs @ <>			\ true if an error occurred
;

: (Verify128)	\ src dest -- flag ; 0=success
\ *G Verify 128 bytes from memory at *\i{src} to Flash at *\i{dest}.
  #128 s= 0=
;

: Prog128	\ src dest -- flag ; 0=success
\ *G Write 128 bytes from memory at *\i{src} to Flash at *\i{dest}.
\ ** A non-zero error code is returned if programming or verification
\ ** fails.
  2dup (Prog128)
  if  2drop -1  endif
  (Verify128)
;



\ ======
\ *> ###
\ ======

decimal
