\ FlashProg.fth - Flash programming for TM4C12xx

((
Copyright (c) 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20190224 SFP001 First release.
))

only forth definitions
decimal

\ ===========
\ *! flashtm4c12xx
\ *T TM4C12xx Flash programming
\ ===========
\ *P When Flash is being written or erased, any read access will
\ ** stall the bus until the Flash operation has finished. Thus,
\ ** providing that you do not erase the code in *\i{FlashProg.fth}
\ ** itself, the Flash programming code can be run from Flash.


\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the TM4C12xx series, they are designed to be easily expanded
\ ** for future TM4C12xx parts with as yet unknown sector sizes.

: fl!		\ x addr -- status ; 0=success
\ *G Write a 32 bit word to Flash. The address must be four byte aligned.
  _FLASH_CTRL				\ -- x addr base
  rot over flc.FMD + !			\ -- addr base ; write x
  tuck flc.FMA + !			\ -- base ; write addr
  $A442:0001 over flc.FMC + !		\ start write
  begin dup flc.FMC + @ 1 and 0= until
  flc.FCRIS + @ bit13 and 		\ check error bit
;

: flSecErase	\ addr -- status ; 0=success
\ *G Erase the sector at addr. The address is clipped to a 16 kb
\ ** unit.
  _FLASH_CTRL				\ -- addr base
  swap $FFFF:C000 and  over flc.FMA + !	\ -- base ; write addr
  $A442:0002 over flc.FMC + !		\ start erase
  begin dup flc.FMC + @ 2 and 0= until
  flc.FCRIS + @ bit11 and 		\ check error bit
;

: .src/dest	\ src dest -- src dest
\ *G Display the addresses and contents of the source
\ ** and destination.
  ."  source: " over .lword  over c@ ." (" .byte ." )"
  ."  dest: " dup .lword  dup c@ ." (" .byte ." )"
;


\ ***************
\ *S Flash driver
\ ***************

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count.

: ?flError	\ status --
\ *G Increment error count if status is not zero.
  if  #PrgErrs incr  endif  ;

: (EraseFlash)	\ dest dlen --
\ *G Erase the Flash for the given range. Note that complete
\ ** sectors in the range will be erased.
  swap $FFFF:C000 and swap		\ clip address to 16 kb unit
  $3FFF + $FFFF:C000 and		\ len -> 16 kb unit
  bounds ?do				\ over the range
    i flSecErase ?flError		\ erase sector
  $4000 +loop				\ step address by sector size
;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 4 byte boundary and
\ ** *\i{len} is rounded up to the next 4 byte unit. *\i{Src}
\ ** must be on a 4 byte boundary and must be in RAM.
  rot -4 and -rot			\ force src
  swap -4 and swap			\ force dest
  3 + -4 and				\ force len
  bounds ?do				\ -- src'
    dup @ i fl! ?flError  4+
  4 +loop
  drop
;

: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over c@ over c@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    1 + swap  1 + swap
  1 +loop
  2drop
;

: ProgramFlash	\ src dest len --
\ *G Using the RAM memory buffer at *\i{src}, program the Flash
\ ** at *\i{dest} with *\i{len} bytes. The relevant Flash sectors
\ ** are erased, the Flash is programmed, and the result verified.
  cr ." Erasing " 2dup (EraseFlash)
  0 #PrgErrs !
  cr ." Programming "  3dup (ProgFlash)
  space #PrgErrs ? ." errors"
  cr ." Verifying "  (VerifyFlash)
;


\ ********************************
\ *S Programming in 128 byte units
\ ********************************

: StartProg128	\ dest dlen -- flag ; 0=success
  $2E07 _FLASH_CTRL flc.FCMISC + !	\ clear status reg by writing ones
  #PrgErrs off  (EraseFlash)  #PrgErrs @
;

: (Prog128)	\ src dest -- flag ; 0=success
\ *G Write 128 bytes from memory at *\i{src} to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 4 byte boundary. *\i{Src}
\ ** may be unaligned if the CPU supports it by default, and
\ ** must be in RAM.
  #PrgErrs @ >r				\ error counter
  -4 and				\ force dest to 4 byte boundary
  #128 bounds ?do			\ -- src'
    dup @ i fl! ?flError  4+
  4 +loop
  drop
  r> #PrgErrs @ <>			\ true if an error occurred
;

: (Verify128)	\ src dest -- flag ; 0=success
\ *G Verify 128 bytes from memory at *\i{src} to Flash at *\i{dest}.
  #128 s= 0=
;

: Prog128	\ src dest -- flag ; 0=success
\ *G Write 128 bytes from memory at *\i{src} to Flash at *\i{dest}.
\ ** A non-zero error code is returned if programming or verification
\ ** fails.
  2dup (Prog128)
  if  2drop -1  exit  endif
  (Verify128)
;


\ ======
\ *> ###
\ ======

decimal
