\ MinXmodemRx.fth - XMODEM file receive

((
Copyright (c) 2001, 2002, 2004, 2005, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20100203 MPE005 Recoded for CPUs with slow DO.
		Reduced poll interval to 10us for CPUs without FIFOs.
20040209 SFP004 Modified for LPC2106 with 512 byte Flash writes,
		Receive only.
20030924 SFP003 Overhauled receive code to make	it minimal
		for single chip Flash programmers.
20021014 SFP002 Modified for COMMON directory.
20010717 EJB000 First release
))


\ ==================
\ *! minxmodemrx
\ *T XMODEM Receiver and Transmitter
\ ==================
\ *S Introduction
\ *P This file implements the XMODEM 128-byte checksum protocol
\ ** for receive only.

\ *P No test code is provided for this file as the system has
\ ** been tested by comparison of transferred binary files.

\ ***************************
\ *S Words in MinXmodemRx.fth
\ ***************************

decimal


\ ==========================
\ *N Constants and variables
\ ==========================

\ Characters
$001 equ START	$001 equ SOH
$004 equ EOT
$006 equ ACK
$015 equ NAK
$018 equ CAN

\ Error codes
$0101 equ blkerror			\ error codes
$0102 equ wrongblk
$0103 equ noreply
$0104 equ crcerror
$0105 equ overflow

$0100 equ maxerrs			\ max errors before we bomb

\ Variables
cell buffer: BLK#			\ Current Block No.
cell buffer: CUR-ADDRESS                \ Current host address
cell buffer: LAST-BLOCK                 \ last block we can receive
cell buffer: rxerrcnt			\ counter for rx errors
cell buffer: rxstatus			\ receiver status

compiler
: X-buffer	\ -- addr
  Xbase
;
target


\ ===================
\ *N XMODEM reception
\ ===================
\ *P This code is a much chopped-down version of that
\ ** provided in *\i{COMMON\XMODEMTXRX.FTH}. It is much smaller
\ ** and it is assumed that the word *\fo{TOFLASH ( src dest -- )}
\ ** is used to program each 128 byte block into Flash.

[undefined] ser-flush [if]
: ser-flush	\ -- ; flush the link input
\ *G Flush all input characters from the host/target link.
  begin
    ser-key?
  while
    ser-key drop
  repeat
;
[then]

: send-ack	\ -- ; send ACK
\ *G Transmit an ACK character.
  ACK ser-emit
;

: send-nak	\ -- ; Transmit a NAK character
\ *G Transmit a NAK character.
  NAK ser-emit
;

: send-can	\ -- ; send CAN character
\ *G Transmit a CAN character.
  CAN ser-emit  rxstatus on
;

(( \ original
: TimedKey	\ -- char|-1
  \ *G As *\fo{KEY} but returns -1 on timeout
  #800000 0 do				\ approx 100ms
    ser-key?
    if  ser-key  unloop exit  then
  loop
  -1
;
))
: TimedKey	\ -- char|-1
  \ *G As *\fo{KEY} but returns -1 on timeout.
  $400000 begin				\ approx 100ms
    ser-key?
    if  ser-key  nip exit  then
    1- dup 0=
  until
  drop -1
;

(( \ original
: Get-BlockData	\ -- cksum ; get block from host
 \ *G Receive a 128 byte XMODEM data block from the host.
  0
  X-buffer #128 bounds do		\ Get 128 byte Block
\    ser-key dup i c!			\ save byte to buffer
    TimedKey dup i c!			\ save byte to buffer
    + $0FF and                          \ Add to Check-Sum
  loop
;
))
: Get-BlockData	\ -- cksum ; get block from host
 \ *G Receive a 128 byte XMODEM data block from the host.
  0  X-buffer begin			\ -- crc caddr
    TimedKey 2dup swap c!		\ -- crc addr char ; save byte to buffer
    rot + $FF and swap 1+		\ -- crc' caddr'
    dup X-buffer #128 + u>=
  until
  drop
;


: CancelXfer	\ --
\ *G Cancel transfer by sending 2 CAN characters.
  send-can send-can			\ 2 CAN's aborts transfer
;

: ?toomanyerrs	\ --
\ *G Cancel reception if too many errors have occurred.
  ser-flush
  1 rxerrcnt +!
  rxerrcnt @ maxerrs > if
    CancelXfer
  else
    send-nak
  endif
;

: From-Buffer	\ -- ; copy 128 bytes from X-BUFFER
\ *G Move 128 bytes from *\fo{X-BUFFER} into the memory pointed to
\ ** by *\fo{CUR-ADDRESS}. The user must supply a routine *\fo{TOFLASH}
\ ** ( src dest -- ) which programs the data at src into Flash
\ ** at dest.
  X-buffer cur-address @ toFlash	\ program 128 bytes
  #128 cur-address +!
;

: Get-Block	\ --
\ *G Receive an XMODEM 128 byte data block from the host, processing
\ ** the header and checksum data.
  TimedKey TimedKey			\ get block and inverse
  over + $0FF <> if
    ?toomanyerrs  send-can
    drop  exit				\ lose incoming block number
  endif
  Blk# @ $00FF and <> if		\ correct block?
    ?toomanyerrs  exit
  endif
  Get-BlockData				\ get the data
  Timedkey <> if			\ get single cksum byte
    ?toomanyerrs  exit
  endif
  Blk# @ last-block @ > if
    CancelXfer  exit
  endif
  From-Buffer				\ copy data to memory
  send-ack				\ all OK send ACK
  1 Blk# +!				\ bump for next block
  rxerrcnt off				\ clear error count again
;

: key/1s	\ -- flag
\ *G Waits for one second or until a key is pressed,
\ ** returning true if a key was pressed.
  #100000 begin
    ser-key? 0= over and
   while
    delay10us  1-
  repeat
  drop
  ser-key?
;

: waitforresponse	\ -- ; wait for host to respond
\ *G Wait for the host to respond for up to 1 second. If the
\ ** host does not respond, a *\fo{NOREPLY} status is set in the
\ ** variable *\fo{RXSTATUS}.
  key/1s 0=
  if  rxstatus on  endif		\ flag no response
;

: Bin-Up	\ addr len -- status ; status 0 = GOOD
\ *G Upload (receive) a block of data of the given size into memory
\ ** using the XMODEM 128 byte block protocol. An error status is
\ ** returned, 0 indicating success.
\  cr ." Start Xmodem transmitter on host" cr
  rxstatus off				\ start with good status
  rxerrcnt off				\ and no errors
  ser-flush				\ flush input
  7 rshift last-block !			\ calculate last block number
  CUR-ADDRESS !				\ Start from addr
  1 BLK# !				\ first block number

  NAK ser-emit				\ NAK requests single byte checksum
  begin					\ wait for remote startup
    key/1s 0=				\ done if a reply within a second
   while
    NAK ser-emit			\ then send NAK if still no key
  repeat

  begin					\ outer loop handles blocks
    waitforresponse				\ check if we got a response
    rxstatus @ 0= if
      case  ser-key				\ get response
        EOT of  EOT rxstatus !  send-ack  endof	\ all finished
        CAN of  CAN rxstatus !  send-ack  endof	\ remote cancelled
        SOH of  Get-block  endof		\ start of data block
      endcase
    endif
    rxstatus @				\ loop until all done
  until

  rxstatus @ dup EOT = if
    drop 0				\ returns 0 for good transfer
  endif
;

((
: gogo	\ -- status ; for preliminary testing
  $01A emit  FlashBase /Flash Bin-up  ;
))


\ ======
\ *> ###
\ ======

