\ ReprogApp.fth - reprogram STM32F4 Flash

((
Copyright (c) 2004, 2008, 2009, 2010, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====
At end of download, check for a partial block to program.

Change max program size to depend on CPU.

Change history
==============
20111205 SFP006 STM32F4 conversion
20100203 SFP006 STM32 conversion
20100125 SFP005 LPC17xx conversion.
20090205 SFP004 LPC2388 does not erase last sector at $0007:D000
20081008 SFP003 Minor documentation changes.
20060701 SFP002 Added CopyFlash.
                Now uses sector tables.
20040212 MPE001 First release
))

only forth definitions  decimal


\ *****************
\ *S Numeric output
\ *****************

: .nibble	\ n --
\ *G Convert a nibble to a hex ASCII digit and display it.
  $0F and  dup 9 >
  if  7 +  then
  $30 + emit
;

: .WORD		\ w --
\ *G Display w as a 4 digit unsigned hexadecimal number.
  dup #12 rshift .nibble
  dup 8 rshift .nibble
  dup 4 rshift .nibble
  .nibble
;

: .dword	\ x --
\ *G Display x as an 8 digit unsigned hexadecimal number.
  dup #16 rshift .word .word
;



\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the TM4C series, they are designed to be easily expanded
\ ** for future TM4C parts with as yet unknown sector sizes.

: fl!		\ x addr -- status ; 0=success
\ *G Write a 32 bit word to Flash. The address must be four byte aligned.
  _FLASH_CTRL				\ -- x addr base
  rot over flc.FMD + !			\ -- addr base ; write x
  tuck flc.FMA + !			\ -- base ; write addr
  $A442:0001 over flc.FMC + !		\ start write
  begin dup flc.FMC + @ 1 and 0= until
  flc.FCRIS + @ bit13 and 		\ check error bit
;

: flSecErase	\ addr -- status ; 0=success
\ *G Erase the sector at addr. The address is clipped to a 16 kb
\ ** unit.
  _FLASH_CTRL				\ -- addr base
  swap $FFFF:C000 and  over flc.FMA + !	\ -- base ; write addr
  $A442:0002 over flc.FMC + !		\ start erase
  begin dup flc.FMC + @ 2 and 0= until
  flc.FCRIS + @ bit11 and 		\ check error bit
;


\ ****************
\ *S Flash handler
\ ****************

cell buffer: #PrgErrs	\ -- addr
\ *G Holds error count.

: ?flError	\ status --
\ *G Increment error count if status is not zero.
  if  #PrgErrs incr  endif  ;

: (EraseFlash)	\ dest dlen --
\ *G Erase the Flash for the given range. Note that complete
\ ** sectors in the range will be erased.
  swap $FFFF:C000 and swap		\ clip address to 16 kb unit
  $3FFF + $FFFF:C000 and		\ len -> 16 kb unit
  bounds ?do				\ over the range
    i flSecErase ?flError		\ erase sector
  $4000 +loop				\ step address by sector size
;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at *\i{dest}.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed. *\i{Dest} is forced to a 4 byte boundary and
\ ** *\i{len} is rounded up to the next 4 byte unit. *\i{Src}
\ ** must be on a 4 byte boundary and must be in RAM.
  rot -4 and -rot			\ force src
  swap -4 and swap			\ force dest
  3 + -4 and				\ force len
  bounds ?do				\ -- src'
    dup @ i fl! ?flError  4+
  4 +loop
  drop
;

: EraseFlash	\ --
\ *G Erase the Flash from start to end.
  Flashbase /Flash (EraseFlash)  ;

: toFlash	\ src dest --
\ *G Handle 128 bytes at src to go to dest in Flash.
  #128 (ProgFlash)
;


\ ***************
\ *S Reprogrammer
\ ***************

l: m1
  ", Erasing "
l: m2
  ",  done "
l: m3
  ", Start Xmodem TX "
[defined] REBOOT [if]
l: m4
  ", Press R to reboot, other to reprogram "
[else]
l: m4
  ", Reset or press key to reprogram"
[then]
l: m5
  ", Errors"
l: m6
  ", No "

: $.	\ c-addr --
\ *G Display counted string.
  count type
;

: .PrgErrs
\ *G Display number of errors.
  cr  #PrgErrs @ .word  bl emit
  m5 $.
;

: initReprog	\ --
\ *G Get the setup data and calculate the requirements
  init-ser #50 ms  ;

: ReprogFlash	\ --
\ *G Erase and program the Flash.
  initReprog
  begin
    #PrgErrs off
    cr m1 $.  #20 ms
    EraseFlash
    m2 $.  .PrgErrs  #PrgErrs off
    cr m3 $.  $01A emit			\ send message and trigger
    FlashBase /Flash Bin-Up drop
    .PrgErrs
    #100 ms
    cr m4 $.
[defined] REBOOT [if]
    key $DF and dup emit  [char] R =
  until
  #100 ms
  reboot
[else]
    key drop
  again
[then]
;

\ SFP002...
l: msg7
  ", Copy "

: (CopyFlash)	\ src dest len --
\ *G Copy one Flash area to another. The parameters are as for *\fo{CMOVE}.
  initReprog
  cr m1 $.  #20 ms			\ "Erase" message
  #PrgErrs off
  2dup (EraseFlash) .PrgErrs  #10 ms	\ -- src dest len
  #PrgErrs @
  if  3drop exit  endif
  cr msg7 $.				\ "Copy" message
  (ProgFlash)				\ --
  .PrgErrs  #20 ms			\ error count
;

: CopyFlash	\ src dest len --
\ *G Copy the flash. The parameters are as for *\fo{CMOVE}.
\ ** The process is repeated until there are no errors, and
\ ** the system is then rebooted.
  begin
    3dup (CopyFlash)  #PrgErrs @ 0=
  until
  #100 ms  reboot
;
\ ...SFP002


\ ======
\ *> ###
\ ======

