\ Cortex exception (interrupt) handlers

((
Copyright (c) 2010, 2011, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20160816 SFP003 Updates for separate FP stack.
20160619 SFP002 Updated for HFP.
20110627 SFP001 Cortex M0/M1 changes.
))

only forth definitions  decimal

\ ============
\ *! intcortex
\ *T Exception and Interrupt handlers
\ ============
\ *P The file *\i{Cortex\IntCortex.fth} contains generic interrupt
\ ** handlers for Cortex-M0/M1/M3/M4 processors.
\ ** *\i{IntCortex.fth} requires *\i{Cortex\CortexDef} to be compiled
\ ** before the *\i{SFRxxxx} file for your particular CPU. The file
\ ** *\i{Cortex/StackDef.fth} provides default main task and stack
\ ** layouts and should be compiled from the control file or copied
\ ** into your control file. Default stack initialisation code is
\ ** provided in *\i{Cortex/StackDef.fth}.

\ *P The high-level interrupt handlers all share a common "stack
\ ** of stacks". On entry to the interrupt handler, Forth system
\ ** registers are allocated. Your initialisation code *\b{must}
\ ** set up R13. This is normally provided by the MPE wrapper code
\ ** for each exception.

\ *P In order to support exception nesting the equate *\fo{#IRQs} in the
\ ** control file must be set to the maximum number of nestings
\ ** required.
\ ** The equate *\fo{#IRQs} is used to calculate the size  of the
\ ** required exception stack, together with the equate *\fo{#SVCs}.
\ ** in the control file.

\ *P Each interrupt has its own *\fo{USER} area. No user variables
\ ** are initialised except for *\fo{S0} and *\fo{R0} in SVC
\ ** handlers.

\ *P It is assumed that the banked stack pointers have
\ ** already been set up by the hardware initialisation code.

\ *P It is implicit throughout the code that the
\ ** system registers UP, PSP, FSP, RSP are always set so that:
\ *C   UP > PSP > FSP > RSP
\ *P Because of this layout, data stack underflows may corrupt
\ ** the first part of the *\fo{USER} area. You have been warned.
\ ** During testing, it may be as well to set the equate
\ ** *\fo{SP-GUARD} to 2 or 3 in your control file to leave a few
\ ** guard cells on the data stack.


\ ************************************
\ *S Cortex-M3 NVIC Exception handlers
\ ************************************
\ *P This section is not a treatise on the Nested Vectored
\ ** Interrupt Controller. These words provide a fairly basic
\ ** set of tools for using the NVIC, which is fully
\ ** documented in the Cortex-M3 Technical Reference Manual
\ ** (ARM DDI 0337)from *\i{www.arm.com}.

\ *P This code requires *\i{Kernel62.fth} and the equate
\ ** *\fo{COLDCHAIN?} must be set non-zero in the control file.
\ ** The NVIC address and register offsets are defined in the
\ ** file *\i{Cortex/CortexDef.fth}.

\ *P The MPE code works in terms of vector numbers, which are
\ ** 16 greater than the external interrupt numbers.

Not-M0/M1? [if]
PROC ExcEntry	\ --
\ *G This is the template code for M3+ vectored exception handlers.
\ --- CPU has saved ----
\   xPSR PC LR R12 R3 R2 R1 R0
\ --- save state ---
  push    { r4-r11, r14 }		\ registers not saved by core + link
  sub     rsp, rsp, # up-size sp-size + fp-size + \ make space for USER area, data stack and FP stack

\ --- set up Forth ---
  add   up, rsp, # sp-size fp-size +	\ USER area at top
[ FP-SIZE ] [if]			\ SFP002
  add   r8, rsp, # fp-size		\ new FP stack
[defined] FS0 [if]
  str   r8, [ up, # u# fs0 ]
[then]
[defined] FSP [if]
  str   r8, [ up, # u# fsp ]
[then]
[then]
  sub	psp, up, # sp-guard cells	\ generate new PSP
  sub	r4, psp, # tos-cached? cells	\ generate new S0
  str   r4, [ up, # s0-offset ]		\   that is compatible with SP!
  str   rsp, [ up, # r0-offset ]	\ set R0
\ --- call high level handler ---
l: ExcCall
  bl      ' noop			\ call the high level handler
\ --- restore state ---
  add     rsp, rsp, # up-size sp-size +	fp-size + \ remove space for USER area, data stack, FP stack
  pop     { r4-r11, pc }		\ restore and exit
end-code
chere ExcEntry - equ /ExcEntry		\ size of template
ExcCall ExcEntry - equ /ExcCall		\ offset to call instruction
[else]
sp-size 3 and [if]  cr ." bad SP-SIZE" abort  [then]
sp-size 2 rshift equ sp-cells
up-size 3 and [if]  cr ." bad UP-SIZE" abort  [then]
up-size 2 rshift equ up-cells

SP-SIZE #256 u< 0=  UP-SIZE #256 u< 0= or  equ LargeISR?  \ -- flag
\ *G The interrupt is often larger than 256 bytes, which requires
\ ** larger ISR entry code.

LargeISR? [if]
PROC ExcEntry	\ --
\ *G This is the template code for M0/M1 vectored exception handlers
\ ** with a large ISR frame.
\ --- CPU has saved ----
\   xPSR PC LR R12 R3 R2 R1 R0
\ --- save state ---
  push    { r4-r7, r14 }			\ registers not saved by core + link
  mov     r0, r8
  mov     r1, r9
  mov     r2, r10
  mov     r3, r11
  push    { r0-r3 }

\ --- set up Forth ---
\ calculate the offsets
  mov .s  r0, # up-cells		\ R0 = UP-SIZE
  lsl .s  r0, r0, # 2
  mov .s  r2, # sp-cells		\ R2 = SP-SIZE
  lsl .s  r2, r2, # 2
  add .s  r3, r0, r2			\ R3 = SP-SIZE + UP-SIZE
\ set new RSP
  mov     r1, rsp
  sub .s  r1, r3
  mov     rsp, r1			\ RSP = RSP - SP-SIZE - UP-SIZE
  push    { r3 }			\ how much we dropped the stack by
\ set new UP
  add .s  r1, r1, r2			\ UP = RSP + SP-SIZE
  mov     up, r1			\ will need R1 later
\ set new PSP
  mov     psp, r1			\ generate new PSP
  sub .s  psp, # sp-guard cells
\ set new S0 and R0 user vars
  sub .s  r4, psp, # tos-cached? cells	\ generate new S0
  str     r4, [ r1, # s0-offset ]	\   that is compatible with SP!
  mov     r0, rsp
  str     r0, [ r1, # r0-offset ]	\ set R0
\ --- call high level handler ---
l: ExcCall
  bl      ' noop			\ call the high level handler
\ --- restore state ---
\ restore RSP
  pop     { r3 }			\ amount by which stack was dropped
  mov     r1, rsp
  add .s  r1, r3
  mov     rsp, r1
\ restore regs
  pop     { r0-r3 }
  mov     r8, r0
  mov     r9, r1
  mov     r10, r2
  mov     r11, r3
  pop     { r4-r7, pc }			\ restore and exit
end-code
chere ExcEntry - equ /ExcEntry		\ size of template
ExcCall ExcEntry - equ /ExcCall		\ offset to call instruction
[else]
PROC ExcEntry	\ --
\ *G This is the template code for M0/M1 vectored exception handlers
\ ** with a small ISR frame. A separate floating point stack is not
\ ** currently supported in interrupts.
\ --- CPU has saved ----
\   xPSR PC LR R12 R3 R2 R1 R0
\ --- save state ---
  push    { r4-r7, r14 }		\ registers not saved by core + link
  mov     r0, r8
  mov     r1, r9
  mov     r2, r10
  mov     r3, r11
  push    { r0-r3 }

\ --- set up Forth ---
\ set new UP
  mov     r1, rsp
  sub .s  r1, # UP-SIZE
  mov     up, r1			\ will need R1 later
\ set new RSP
  mov .s  r2, r1
  sub .s  r2, # SP-SIZE
  mov     rsp, r2			\ RSP = RSP - SP-SIZE - UP-SIZE
\ set new PSP
  mov     psp, r1			\ generate new PSP = UP - abit
  sub .s  psp, # sp-guard cells
\ set new S0 and R0 user vars
  sub .s  r4, psp, # tos-cached? cells	\ generate new S0
  str     r4, [ r1, # s0-offset ]	\   that is compatible with SP!
  mov     r0, rsp
  str     r0, [ r1, # r0-offset ]	\ set R0
\ --- call high level handler ---
l: ExcCall
  bl      ' noop			\ call the high level handler
\ --- restore state ---
\ restore RSP
  mov     r1, rsp
  add .s  r1, # SP-SIZE
  add .s  r1, # UP-SIZE
  mov     rsp, r1
\ restore regs
  pop     { r0-r3 }
  mov     r8, r0
  mov     r9, r1
  mov     r10, r2
  mov     r11, r3
  pop     { r4-r7, pc }		\ restore and exit
end-code
chere ExcEntry - equ /ExcEntry		\ size of template
ExcCall ExcEntry - equ /ExcCall		\ offset to call instruction
[then]					\ LargeISR?

[then]					\ Not-M0/M1?

interpreter
: EXC:		\ xt vec# -- ; -- isr
\ *G Creates an exception handler that runs the given Forth word.
\ ** At run time the entry point of the ISR is returned. Use in
\ ** the form:
\ *C   ' <action> <vec#> EXC: <actionISR>
\ *P The example below will define a handler for the SysTick
\ ** interrupt/exception that runs the Forth word *\fo{SysTickHandler}.
\ *C   ' SysTickHandler #15 EXC: SysTickEntry
  swap align  chere /ExcEntry callot	\ -- vec# xt isr ; reserve space
  ExcEntry over /ExcEntry move		\ copy template
  tuck /ExcCall + !call			\ -- vec# isr ; form call instruction
  tuck swap setExcVec
  label					\ returns entry point
;
target

: int>bit	\ int# -- mask offset
\ Convert an interrupt number into its bit mask and ofset
\ into an array of 32 bit registers.
  dup $1F and 1 swap lshift
  swap 5 rshift 2 lshift
;

: EnInt		\ vec# --
\ *G Enable the requested NVIC interrupt source.
\ ** NVIC interrupts have vector numbers in the range 16..255.
  dup #15 u> if
    #16 - int>bit _SCS + nvSetEnR0 + !
    exit
  endif
  drop
;

: DisInt	\ vec# --
\ *G Disable the requested NVIC interrupt source.
\ ** NVIC interrupts have vector numbers in the range 16..255.
  dup #15 u> if
    #16 - int>bit _SCS + nvClrEnR0 + !
    exit
  endif
  drop
;

: SetPri	\ pri vec# --
\ *G Set the priority of the given NVIC interrupt source.
\ ** NVIC interrupts have vector numbers in the range 16..255.
\ ** The priority numbers are in the range 0..255, where 0 is
\ ** the highest priority. The number of bits actually used
\ ** is implementation dependent, but unused bits are always
\ ** the low-order bits. Hence, using $10, $20 and so on is
\ ** fairly portable.
  dup #15 u> if
    16 - _SCS + nvPR0 + c!  exit
  endif
  drop
;

: SWINT		\ vec# --
\ *G Generate interrupt from software.
  dup #15 u> if
    #16 - int>bit _SCS + nvSetPendR0 + !
    exit
  endif
  drop
;


\ ======
\ *> ###
\ ======

