\ MinCortex.fth - ARM Cortex-M3 Umbilical Forth minimal code definitions

((
Released for use with the MPE Forth Cross Compiler by:

Copyright (c) 2000, 2008, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


Change history
==============
20100204 SFP011 Cortex-M3 conversion.
20081128 SFP010 Added S=.
20000620 SFP009 Matched USER to updated compiler [build 561]
20000608 SFP008 Converted to v6.1 generic I/O model
))

only forth definitions  decimal


\ =========
\ *! mincortex
\ *T Minimal Umbilical code definitions
\ =========
\ *P The file *\i{Cortex/MinCortex.fth} contains the minimum code
\ ** definitions required to support Umbilical Forth. If additional
\ ** words are required, they may be copied to a new file from
\ ** *\i{Cortex/CodeCortex.fth} or *\i{Common/Kernel62.fth}.

\ *****************
\ *S Register usage
\ *****************
\ *P For Cortex-M3+ the following register usage is the default:
\ *E   r15         pc      program counter
\ **   r14         link    link register; bit0=1=Thumb, usually set
\ **   r13         rsp     return stack pointer
\ **   r12         psp     data stack pointer
\ **   r11         up      user area pointer
\ **   r10         --
\ **   r9          lp      locals pointer
\ **   r8          --
\ **   r7          tos     cached top of stack
\ **   r0-r6       scratch
\ *P The VFX optimiser reserves R0 and R1 for internal operations.
\ ** *\fo{CODE} definitions must use R10 as TOS with NOS pointed to by
\ ** R12 as a full descending stack in ARM terminology. R0..R8 are
\ ** free for use by *\fo{CODE} definitions and need not be preserved or
\ ** restored. You should assume that any register can be affected
\ ** by other words.

\ ****************
\ *S Configuration
\ ****************
\ *P These equates are set false (zero) if they have not already
\ ** been defined.

[undefined] DSQRT? [if]
false equ DSQRT?	\ -- flag
\ *G Set this non-zero to compile *\fo{DSQRT}.
[then]

[undefined] FastCmove? [if]
false equ FastCmove?	\ -- flag
\ *G Set this flag true to use a fast but vast (~1kb) version of
\ ** *\fo{CMOVE}. If your application uses either *\fo{CMOVE} or
\ ** *\fo{MOVE} in time-critical code, the fast version offers
\ ** an overall speed up of about four times. The code for the
\ ** fast but vast version was written by Rowley Associates,
\ ** whose permission to adapt and publish the code with the
\ ** MPE cross compiler is much appreciated.
[then]


\ ***************
\ *S Flow of control
\ ***************

CODE (DO)	\ limit index --
\ *G The run time action of *\fo{DO} compiled on the target.
\ ** INTERNAL.
  ldr     r1, [ psp ], # 4		\ get limit
L: PDO
  ldr     r2, [ link, # -1 ]		\ get LEAVE address
  orr .s  r2, r2, # 1			\ force Thumb bit
  add .s  link, link, # 4		\ update link
  add .s  r1, r1, # $80000000
  sub .s  r0, tos, r1

  stmfd   rsp ! { r0, r1, r2 }		\ push LEAVE then limit, then index on ret. stack
  ldr     tos, [ psp ], # 4		\ update tos
  next,
END-CODE

CODE (?DO) 	\ limit index --
\ *G The run time action of *\fo{?DO} compiled on the target.
\ ** INTERNAL.
  ldr     r1, [ psp ], # 4		\ get limit
  cmp     r1, tos			\ check not equal
  b .ne   pdo				\ take DO ?
  ldr     tos, [ psp ], # 4		\ update tos
  ldr     r0, [ link, # -1 ]		\ get leave destination
  orr .s  r0, r0, # 1			\ set Thumb bit
  bx      r0				\ and take leave branch
END-CODE

CODE EXECUTE 	\ xt --
\ *G Execute the code described by the XT. This is a Forth equivalent
\ ** to an assembler JSR/CALL instruction.
  orr .s  r0, tos, # 1		        \ move CFA, setting Thumb bit
  ldr     tos, [ psp ], # 4		\ get new TOS
  bx      r0				\ execute CFA - link contains ret addr of execute
END-CODE


\ **************************
\ *S Stack operations and maths
\ **************************

CODE NOOP	\ --
\ *G A NOOP, null instruction.                                          )
  next,
END-CODE

: DROP 		\ x --
\ *G Lose the top data stack item and promote NOS to TOS.
  drop  ;

CODE WITHIN? 	\ n1 n2 n3 -- flag
\ *G Return TRUE if N1 is within the range N2..N3.
\ ** This word uses signed arithmetic.
  ldmfd   psp ! { r0, r1 }
  mov .s  r2, # 0
  mov .s  r3, # 0
  cmp     r1, r0
  it .ge
    mov     r2, # 1
  cmp     r1, tos
  it .le
    mov     r3, # 1
  tst     r2, r3
  ite .ne
    mvn     tos, # 0
    mov     tos, # 0
  next,
END-CODE

CODE WITHIN 	\ n1|u1 n2|u2 n3|u3 -- flag
\ *G The ANS version of WITHIN?.
\ ** This word uses unsigned arithmetic, so that signed compares are
\ ** treated as existing on a number circle.
  ldmfd   psp ! { r0, r1 }              \ tos = n3, r0 = n2, r1 = n1
  sub .s  tos, tos, r0
  sub .s  r1, r1, r0
  cmp     r1, tos
  ite .lo
    mvn     tos, # 0
    mov     tos, # 0
  next,
END-CODE


\ *****************
\ *S Multiplication
\ *****************

: UM*		\ u1 u2 -- ud
\ *G Perform unsigned-multiply between two numbers and return double
\ ** result.
  um*  ;

: *		\ n1 n2 -- n3
\ *G Standard signed multiply. N3 = n1 * n2.
  *  ;

: m*		\ n1 n2 -- d
\ *G Signed multiply yielding double result.
  m*  ;


\ ***********
\ *S Division
\ ***********
\ *P ARM Cortex provides 32/32 division instructions, but no
\ ** 64/32 ones. Avoid 64/32 division routines if you can where
\ ** performance matters.

macro: udiv64_step	\ --
\ *G Cross compiler macro to perform one step of the unsigned
\ ** 64 bit by 32 bit division
    add .s  r1, r1, r1			\ 64 bit shift of dividend
    adc .s  r0, r0, r0
    adc .s  r3, r2, r2			\ preserve dividend bit 63, R2=0
    sub .s  r0, r0, tos			\ trial subtraction, carry set (no borrow) if ok
    adc .s  r3, r3, r3			\ success if bit 0 or 1 set
    ite .ne
      orr  r1, r1, # 1			\ succeeded, update quotient
      add  r0, r0, tos			\ failed, undo subtraction
;m

((
code um/mod	\ ud1 u2 -- urem uquot
\ *G Slow and short - Full 64 by 32 unsigned division subroutine.
\ ** This routine uses a loop for code size. This version is
\ ** commented out by default.
  ldr     r0, [ psp ], # 4		\ dividend high
  ldr     r1, [ psp ]			\ dividend low
  mov	  r4, # #32			\ loop counter
  mov     r2, # 0			\ always 0
  begin,
    udiv64_step
    sub .s  r4, r4, # 1
  eq, until,
  mov     tos, r1			\ move quotient
  str     r0, [ psp ]			\ move remainder
  next,
end-code
))

code um/mod	\ ud1 u2 -- urem uquot
\ *G Fast and big - Full 64 by 32 unsigned division subroutine.
\ ** Unrolled for speed. This routine uses 660 bytes of code space
\ ** using the Thumb-2 instruction set, whereas the ARM32 version
\ ** uses 920 bytes.
  ldr     r0, [ psp ], # 4		\ dividend high
  ldr     r1, [ psp ]			\ dividend low
  mov     r2, # 0			\ always 0

  udiv64_step	\ 1
  udiv64_step	\ 2
  udiv64_step	\ 3
  udiv64_step	\ 4
  udiv64_step	\ 5
  udiv64_step	\ 6
  udiv64_step	\ 7
  udiv64_step	\ 8
  udiv64_step	\ 9
  udiv64_step	\ 10
  udiv64_step	\ 11
  udiv64_step	\ 12
  udiv64_step	\ 13
  udiv64_step	\ 14
  udiv64_step	\ 15
  udiv64_step	\ 16
  udiv64_step	\ 17
  udiv64_step	\ 18
  udiv64_step	\ 19
  udiv64_step	\ 20
  udiv64_step	\ 21
  udiv64_step	\ 22
  udiv64_step	\ 23
  udiv64_step	\ 24
  udiv64_step	\ 25
  udiv64_step	\ 26
  udiv64_step	\ 27
  udiv64_step	\ 28
  udiv64_step	\ 29
  udiv64_step	\ 30
  udiv64_step	\ 31
  udiv64_step	\ 32

\ all done
  mov     tos, r1			\ move quotient
  str     r0, [ psp ]			\ move remainder
  next,
end-code

macro: udiv63_step	\ --
\ *G Cross compiler macro to perform one step of the unsigned
\ ** 63 bit by 31 bit division
  adc .s  r0, tos, r0 .lsl # 1		\ trial subtract from high (divisor negated)
  it .cc				\ restore
    sub     r0, r0, tos
  adc .s  r1, r1, r1			\ shift
;m

proc Udiv63/31	\ r0:r1/tos ; 63/31 unsigned divide -> tos=quot, r0=rem
\ *G Unsigned division primitive - unrolled for speed.
\ ** Note that this routine does not handle the top bit of the divisor
\ ** and dividend correctly. *\fo{Udiv63/31} is used for signed divide
\ ** operations for which the top bits are always zero.
  rsb .s  tos, tos, # 0          \ negate divisor
  add .s  r1, r1, r1
  udiv63_step			\ 1
  udiv63_step			\ 2
  udiv63_step			\ 3
  udiv63_step			\ 4
  udiv63_step			\ 5
  udiv63_step			\ 6
  udiv63_step			\ 7
  udiv63_step			\ 8
  udiv63_step			\ 9
  udiv63_step			\ 10
  udiv63_step			\ 11
  udiv63_step			\ 12
  udiv63_step			\ 13
  udiv63_step			\ 14
  udiv63_step			\ 15
  udiv63_step			\ 16
  udiv63_step			\ 17
  udiv63_step			\ 18
  udiv63_step			\ 19
  udiv63_step			\ 20
  udiv63_step			\ 21
  udiv63_step			\ 22
  udiv63_step			\ 23
  udiv63_step			\ 24
  udiv63_step			\ 25
  udiv63_step			\ 26
  udiv63_step			\ 27
  udiv63_step			\ 28
  udiv63_step			\ 29
  udiv63_step			\ 30
  udiv63_step			\ 31
  adc .s  r0, tos, r0 .lsl # 1	\ 32
  it .cc
    sub     r0, r0, tos
  adc     tos, r1, r1
  next,
end-code

CODE FM/MOD 	\  d1 n2 -- rem quot ; floored division
\ *G Perform a signed division of double number D1 by single number N2
\ ** and return remainder and quotient using floored division. See the
\ ** ANS Forth specification for more details of floored division.
  ldmfd psp ! { r0, r1 }                \ r0:r1=dividend
\ Make dividend and divisor absolute, saving signs
l: fm/mod1
  mov .s  r3, r0                        \ r3=sign of dividend
  b .pl   fm/mod2                       \ form |dividend|
  mov     r2, # 0			\ dnegate = subtract from zero
  sub .s  r1, r2, r1
  sbc .s  r0, r2, r0
l: fm/mod2
  mov .s r2, tos                        \ r2=sign of divisor
  it .mi
    rsb     tos, tos, # 0               \ form |divisor|

  mov     r6, link                      \ store link - r6 is free for use here
  bl Udiv63/31

  teq     r0, # 0                       \ remainder = 0?
  b .eq   fm/mod3                       \ NO
  cmp     r2, # 0			\   remainder takes sign of divisor
  it .mi
    rsb     r0, r0, # 0
  eor .s  r3, r3, r2                    \   r3=sign of quotient
  ittt .mi
    rsb     tos, tos, # 0               \   form quotient if neg.
    sub     r0, r2, r0                  \   form remainder if neg.
    sub     tos, tos, # 1               \   form quotient if neg.
  str     r0, [ psp, # -4 ] !		\   push remainder
  bx      r6                            \   exit
l: fm/mod3                              \ YES
  eor .s  r3, r3, r2                    \   r3=sign of quotient
  it .mi
    rsb     tos, tos, # 0               \   form quotient
  str     r0, [ psp, # -4 ] !		\   push remainder
  bx      r6				\   exit
END-CODE

CODE SM/REM 	\ d1 n2 -- rem quot ; symmetric division
\ *G Perform a signed division of double number D1 by single number N2
\ ** and return remainder and quotient using symmetric (normal) division.
  ldmfd psp ! { r0, r1 }		\ r0:r1 = d1
\ Make dividend and divisor absolute, saving signs
l: sm/rem1
  mov .s  r3, r0			\ sign of dividend
  b .pl   sm/rem2
  mov     r2, # 0			\ dnegate = subtract from zero
  sub .s  r1, r2, r1
  sbc .s  r0, r2, r0
l: sm/rem2
  mov .s  r2, tos			\ sign of divisor
  it .mi
    rsb     tos, tos, # 0

  mov     r6, link
  bl      Udiv63/31			\ tos=quot, r0=rem

  teq     r0, # 0			\ check remainder
  b .eq   sm/rem3			\ skip if rem=zero
  eor .s  r3, r3, r2			\ sign of result
  itt .mi
    rsb     tos, tos, # 0		\ negate quot and rem
    rsb     r0, r0, # 0
  cmp     r2, # 0			\ if divisor negative
  it .mi
    rsb     r0, r0, # 0			\ negate remainder
  str     r0, [ psp, # -4 ] !		\   push remainder
  bx      r6
l: sm/rem3
  eor .s  r3, r3, r2			\ sign of quotient
  it .mi
    rsb     tos, tos, # 0
  str     r0, [ psp, # -4 ] !
  bx      r6
END-CODE

CODE /MOD 	\ n1 n2 -- rem quot
\ *G Signed symmetric division of N1 by N2 single-precision
\ ** returning remainder and quotient.
  ldr     r0, [ psp ]			\ r0=dividend, tos=divisor
  sdiv    r2, r0, r1			\ r2=quotient
  mul     r3, tos, r2			\ r3=quotient*divisor
  sub .s  r3, r0, r3			\ r3=rem=dividend-(quotient*divisor)
  mov .s  tos, r2			\ quotient
  str     r3, [ psp ]			\ remainder
  next,
END-CODE

: / 		\ n1 n2  -- n3
\ *G Standard signed division operator. n3 = n1/n2.
  /  ;

: MOD 		\ n1 n2 -- n3
\ *G Return remainder of division of N1 by N2. n3 = n1 mod n2.
  /mod drop  ;

: */MOD 	\ n1 n2 n3 -- n4 n4
\ *G Multiply n1 by n2 to give a double precision result, and then
\ ** divide it by n3 returning the remainder and quotient. The point
\ ** of this operation is to avoid loss of precision.
  >r m* r> sm/rem
;

: */ 		\ n1 n2 n3 -- n4
\ *G Multiply n1 by n2 to give a double precision result, and then
\ ** divide it by n3 returning the quotient. The point
\ ** of this operation is to avoid loss of precision.
  */mod nip
;

: M/ 		\ d n1 -- n2
\ *G Signed divide of a double by a single integer.
  sm/rem nip
;


\ *********************
\ *S Miscellaneous math
\ *********************

CODE D+ 	\ d1 d2 -- d3
\ *G Add two double precision integers.
  ldmfd   psp ! { r0-r2 }		\ r0 = d2 low, r1 = d1 high, r2 = d1 low
  add .s  r0, r0, r2			\ d3l = d2l + d1l
  adc .s  tos, tos, r1			\ d3h = d2h + d1h
  str     r0, [ psp, # -4 ] !		\ push d3l
  next,
END-CODE

CODE D- 	\ d1 d2 -- d3
\ *G Subtract two double precision integers. D3=D1-D2.
  ldmfd   psp ! { r0-r2 }		\ r0 = d2 low, r1 = d1 high, r2 = d1 low
  sub .s  r0, r2, r0			\ d3l = d2l - d1l
  sbc .s  tos, r1, tos			\ d3h = d2h - d1h
  str     r0, [ psp, # -4 ] !		\ push d3l
  next,
END-CODE

CODE DNEGATE 	\ d1 -- -d1
\ *G Negate a double number.
L: DNEG1
  ldr     r0, [ psp ]			\ r0 = d1l, tos = d1h
  mov     r1, # 0
  rsb .s  r0, r0, # 0			\ negate low
  sbc     tos, r1, tos			\ negate high carrying through
  str     r0, [ psp ]
  next,
END-CODE

CODE ?NEGATE 	\ n1 flag -- n1|n2
\ *G If flag is negative, then negate n1.
  mov .s  tos, tos			\ set processor flags
  ldr     tos, [ psp ], # 4		\ discard flag
  it .mi
    rsb     tos, tos, # 0		\ negate n1 if flag=true
  next,
END-CODE

CODE ?DNEGATE 	\ d1 flag -- d1|d2
\ *G If flag is negative, then negate d1.
  mov .s  tos, tos			\ set processor flags
  ldr     tos, [ psp ], # 4		\ discard flag
  b .mi   DNEG1
  next,
END-CODE

CODE ABS	\ n -- u
\ *G If n is negative, return its positive equivalent (absolute value).
  cmp     tos, # 0
  it .mi
    rsb     tos, tos, # 0
  next,
END-CODE

CODE DABS 	\ d -- ud
\ *G If d is negative, return its positive equivalent (absolute value).
  mov .s  tos, tos			\ set processor flags
  b .mi   DNEG1				\ negate d if flag=true
  bx      r14				\ restore link and exit
END-CODE

CODE ROLL 	\ xu xu-1 .. x0 u -- xu-1 .. x0 xu
\ *G Rotate the order of the top N stack items by one place such that
\ ** the current top of stack becomes the second item and the Nth item
\ ** becomes TOS. See also *\fo{ROT}.
  mov     r0, tos, .lsl # 2		\ r0=position of xu in bytes
  cmp     r0, # 0			\ u is valid?
  itt .le
    ldr     tos, [ tos ], # 4		\   no - get new tos
    bx      link			\   no - exit

  ldr     tos, [ psp ++ r0 ]		\ put xu in tos
L: ROLL1
  sub .s  r1, r0, # 4
  ldr     r2, [ psp ++ r1 ]
  str     r2, [ psp ++ r0 ]
  mov     .s r0, r1
  b .ne   ROLL1
  add     psp, psp, # 4
  next,
END-CODE


\ **********
\ *S Strings
\ **********

FastCmove? [if]
  include %CpuDir%\fcmove.fth
[else]
code cmove	\ asrc adest len --
\ *G Copy *\i{len} bytes of memory forwards from *\i{asrc} to *\i{adest}.
\ ** If the performance of *\fo{CMOVE} is important in your
\ ** application, set the equate *\fo{FastCmove?} non-zero and
\ ** a much faster (four to five times) but much larger (~900 bytes)
\ ** version will be compiled.
\ ** See *\i{Cortex\fcmove.fth} for the details.
  mov .s  r3, tos                	\ r3=len
  ldmfd   psp ! { r0, r1, tos }		\ r0=addr2, r1=addr1, r3=len, tos restored
  it .eq
    bx      link			\ return if len=0, nothing to do
  orr     r2, r1, r0			\ check alignment
  orr     r2, r2, r3			\ R2 := adr1|addr2|len
  and .s  r4, r2, # 3			\ will be zero if aligned
  b .ne   cmovx3			\ if not aligned
L: CMOVX1
  ldr     r2, [ r1 ], # 4
  sub .s  r3, r3, # 4			\ hoisted for ARM9
  str     r2, [ r0 ], # 4
  b .ne   CMOVX1
  next,
L: CMOVX3
  ldrb    r2, [ r1 ], # 1
  sub .s  r3, r3, # 1			\ hoisted for ARM9
  strb    r2, [ r0 ], # 1
  b .ne   CMOVX3
  next,
END-CODE
[then]

CODE CMOVE>	\ c-addr1 c-addr2 u --
\ *G As *\fo{CMOVE} but working in the opposite direction,
\ ** copying the last character in the string first.
  mov .s  r3, tos                	\ r3=len
  ldmfd   psp ! { r0, r1, tos }		\ r0=addr2, r1=addr1, r3=len, tos restored
  it .eq
    bx      link			\ return if len=0, nothing to do
  add     r0, r0, r3			\ addr2+len
  add     r1, r1, r3			\ addr1+len
  orr     r2, r1, r0			\ check alignment
  orr     r2, r2, r3			\ R2 := adr1|addr2|len
  and .s  r4, r2, # 3			\ will be zero if aligned
  b .ne   CMV>3				\ byte by byte if not aligned
L: CMV>1	\ copy by 4 byte units
  ldr     r2, [ r1, # -4 ] !
  sub .s  r3, r3, # 4			\ hoisted for ARM9 and above
  str     r2, [ r0, # -4 ] !
  b .ne   CMV>1
  next,
L: CMV>3	\ copy byte by byte
  ldrb    r2, [ r1, # -1 ] !
  sub .s  r3, r3, # 1			\ hoisted for ARM9 and above
  strb    r2, [ r0, # -1 ] !
  b .ne   CMV>3
  next,
END-CODE

CODE FILL	\ c-addr u char --
\ *G Fill LEN bytes of memory starting at ADDR with the byte information
\ ** specified as CHAR.
  mov .s  r2, tos
  ldmfd   psp ! { r0, r1, tos }		\ r2=char, r0=count, r1=addr
  cbz     r0, FILX
L: FIL1
  sub .s  r0, r0, # 1
  strb    r2, [ r1 ], # 1
  b .ne FIL1
L: FILX
  next,
END-CODE

: erase		\ c-addr u -- ; wipe memory
\ *G Set U bytes of memory starting at C-ADDR with zeros.
  0 fill
;

CODE S=		\ c-addr1 c-addr2 u -- flag
\ *G Compare two same-length strings/memory blocks, returning TRUE if
\ ** they are identical.
  ldmfd   psp ! { r0, r1 }		\ r0=adr2, r1=adr1
  mov .s  tos, tos
  b .eq   S=X                           \ TRUE as len=0
L: S=1
  ldrb    r2, [ r0 ], # 1
  ldrb    r3, [ r1 ], # 1
  cmp     r2, r3
  b .ne   S=X                           \ FALSE as mismatched chars
  sub .s  tos, tos, # 1
  b .ne   S=1                           \ if count=0 then TRUE
L: S=X
  ite .eq
    mvn     tos, # 0			\ TRUE
    mov     tos, # 0			\ FALSE
  next,
END-CODE

CODE (") 	\ -- a-addr ; return address of string, skip over it
\ *G Return the address of a counted string that is inline after the
\ ** CALLING word, and adjust the CALLING word's return address to
\ ** step over the inline string. The adjusted return address will
\ ** be at a four byte boundary.
\ ** See the definition of *\fo{(.")} for an example.
  str     tos, [ psp, # -4 ] !		\ save TOS
  ldr     r0, [ rsp ]			\ get return address
  bic     tos, r0, # 1			\ remove Thumb bit
  ldrb    r1, [ tos ]			\ length byte
  add .s  r1, r1, tos			\ r1=address+length+1
  add .s  r1, # 4			\ +3 to align to four byte boundary
  bic .s  r1, r1, # 3			\ -4 and
  orr .s  r1, r1, # 1			\ restore Thumb bit
  str     r1, [ rsp ]			\ update return address
  next,
END-CODE

: (C")		\ -- c-addr
\ *G The run time action compiled by C".
  (")
;

: (S")		\ -- c-addr u
\ *G The run time action compiled by S".
  (") count
;


\ ***************
\ Runtime support
\ ***************

\ ************************************
\ *S Umbilical versions of defining words
\ ************************************

asmcode
here is-action-of constant
\ *G The runtime code for a CONSTANT.
  str     tos, [ psp, # -4 ] !		\ save TOS
  ldr     tos, [ link, # -1 ]		\ get const val/addr from afer CALL
  pop     { pc }
END-CODE

asmcode
here is-action-of variable
\ *G The runtime action for a VARIABLE.
  str     tos, [ psp, # -4 ] !		\ save TOS
  ldr     tos, [ link, # -1 ]		\ get const val/addr from afer CALL
  pop     { pc }
END-CODE

asmcode
here is-action-of value
\ *G The runtime action of a VALUE.
  str     tos, [ psp, # -4 ] !		\ save TOS
  ldr     tos, [ link, # -1 ]		\ get var addr from afer CALL
  ldr     tos, [ tos ]			\ get data
  pop     { pc }
END-CODE

asmcode
here is-action-of user
\ *G The runtime action of a USER variable.
  str     tos, [ psp, # -4 ] !		\ save TOS
  ldr     r0, [ link, # -1 ]		\ get ptr to user val from after CALL
  add .s  tos, up, r0			\ add user pointer to user offset
  pop     { pc }
end-code

interpreter
: u#		\ "<name>"-- u
\ *G An *\fo{INTERPRETER} word that returns the index of the
\ ** *\fo{USER} variable whose name follows, e.g.
\ *C   u# S0
  ' >body
;
target

: CRASH		\ --  ; used as action of DEFER
\ *G The default action of a DEFERed word. A NOOP.
;

asmcode
here is-action-of DEFER	\ Comp: "<spaces>name" -- ; Run: i*x -- j*x
\ *G The runtime action of a DEFERred word.
  ldr     r0, [ link, # -1 ]		\ get data address
  ldr     link, [ rsp ], # 4		\ allows return to 'top-level' from word
  ldr     r1, [ r0 ]			\ fetch xt
  orr     r1, r1, # 1			\ set T bit
  bx      r1
END-CODE


\ ****************
\ *S Display words
\ ****************

: SPACE         \ --
\ *G Output a blank space (ASCII 32) character.
  bl emit
;

: SPACES        \ n --
\ *G Output 'n' spaces, where 'n' > 0.
\ ** If 'n' < 0, no action is taken.
  0 max 0 ?do  space  loop
;

: .nibble	\ n --
\ *G Convert a nibble to a hex ASCII digit and display it.
  $0F and  dup 9 >
  if  7 +  then
  $30 + emit
;

: .BYTE         \ b --
\ *G Display the byte *\i{b} as a 2 digit hex number.
  dup 4 rshift .nibble  .nibble
;

: .WORD		\ w --
\ *G Display w as a 4 digit unsigned hexadecimal number.
  dup #12 rshift .nibble
  dup 8 rshift .nibble
  dup 4 rshift .nibble
  .nibble
;

: .dword	\ x --
\ *G Display x as an 8 digit unsigned hexadecimal number.
  dup #16 rshift .word .word
;

: .lword	\ x --
\ *G A synonym for *\fo{.DWORD} above.
  .dword
;


\ ======
\ *> ###
\ ======

