\ ARM Cortex v7 Multi-Tasker

((
Copyright (c) 2010, 2014, 2015, 2017, 2022
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20221031 RBS005 Corrected use of SUB RSP, Rn with updated assembler.
20170804 MAJ004 Corrected EVENT?
20151110 RBS003 Incorporated Robert Sexton's code for lazy
		register saves and use of WFI in PAUSE.
20150330 SFP002 Allowed for optional float stack.
20141121 SFP001 Reduced set of preserved registers.


********************
Resource definitions
********************

For Cortex-M3+ the following register usage is the default:
  r15         pc      program counter
  r14         link    link register; bit0=1=Thumb, usually set
  r13         rsp     return stack pointer
  r12         psp     data stack pointer
  r11         up      user area pointer
  r10         --
  r9          lp      locals pointer
  r8          --      will be float stack pointer in future
  r7          tos     cached top of stack
  r0-r6       scratch
))

\ ===========
\ *! multicortex
\ *T ARM Cortex multitasker
\ ===========
\ *P The ARM Cortex multitasker follows the model introduced
\ ** with the v6.1 compilers. A few extensions are also
\ ** provided.

\ *P The code in *\i{MultiCortex.fth} only works with the
\ ** Cortex-M3 register set. When using the a Cortex-M0 device
\ ** with its register allocation, use *\i{MultiCM0.fth}. If you
\ ** need the interworking (ARM32) register set, please
\ ** contact MPE or modify the code yourself.


\ ****************
\ *S Configuration
\ ****************
\ *P The configuration equates can be defined before this file
\ ** is compiled.

[undefined] lazy-save? [if]
1 equ lazy-save?	\ -- flag
\ *G If previously undefined, *\fo{LAZY-SAVE?} is set to one
\ ** and registers are only saved if they have to be.
[then]

[undefined] norun-sleep? [if]
0 equ norun-sleep?	\ -- flag
\ *G If previously undefined, *\fo{NORUN-SLEEP?} is set to zero.
\ ** If set to one, the processor is put to sleep using a WFI
\ ** instruction if there are no other tasks to run. This can
\ ** save a lot of power, but requires regular interrupts and
\ ** care that tasks are properly enabled. For this equate to
\ ** be used *\fo{LAZY-SAVE?} must be non-zero.
[then]

[undefined] test-multi? [if]
0 equ test-multi?	\ -- flag ; true to compile test code
\ *G If previously undefined, *\fo{TEST-MULTI?} is set to zero
\ ** and test code is not compiled.
[then]


\ ****************************
\ *S TCB data structure layout
\ ****************************
\ *E cell       LINK    link to next task
\ ** cell       SSP     Saved Stack Pointer
\ ** cell       STAT    Bit 0     1 = running, 0 = halted
\ **                    Bit 1     1 = message pending
\ **                    Bit 2     1 = event triggered
\ **                    Bit 3     1 = event handler run
\ **                    Bit 4..7  Reserved
\ ** 	        others  1 = set to run task, available to user
\ ** cell       TASK    Task that sent message here
\ ** cell       MESG    Message address
\ ** cell       EVNTw   CFA of word run as event handler
\ *P This structure is allocated at the start of the *\fo{USER}
\ ** area. Consequently the *\fo{TCB} of the current task is given
\ ** by *\fo{UP}.

struct /TCB	\ -- size
\ *G The structure used by the code that matches the
\ ** description above.
  ptr tcb.link		\ link to next task ; MUST BE FIRST
  ptr tcb.ssp	        \ Saved Stack Pointer
  int tcb.status	\ status word
  int tcb.msrc		\ message source
  int tcb.mesg		\ message
  ptr tcb.event		\ xt of word which is event handler
end-struct

\ bit masks for the status cell
$0001 equ run-mask  run-mask invert equ ~run-mask	\ running
  0 equ run-bit#
$0002 equ msg-mask  msg-mask invert equ ~msg-mask	\ message available
  1 equ msg-bit#
$0004 equ trg-mask  trg-mask invert equ ~trg-mask	\ event triggered
  2 equ trg-bit#
$0008 equ evt-mask  evt-mask invert equ ~evt-mask	\ event run
  3 equ evt-bit#


\ ******************
\ Consistency checks
\ ******************

/tcb tcb-size <> [if]
  cr ." Control file and tasker disagree as to TCB size"  abort
[endif]


\ ***************************
\ *S Task handling primitives
\ ***************************

cdata		\ so that comma is to code space

init-u0 constant main	\ -- addr ; tcb of main task
\ *G Returns the base address of the main task's *\fo{USER} area.
l: main-link
  0 ,					\ end of task chain

0 value multi?	\ -- flag
\ *G Returns true if the tasker is enabled.

: single	\ --
\ *G Disable scheduler.
  0 to multi?  ;

: multi		\ --
\ *G Enable scheduler.
  -1 to multi?  ;

lazy-save? [if]
CODE pause	\ -- ; the scheduler itself
\ *G The software scheduler itself - with lazy register save.
l: [schedule]
  mvl     r0, # ' multi? >body		\ inspect contents of MULTI?
  ldr     r1, [ r0 ]
  cmp     r1, # 0			\ single tasking if 0
  it .eq
    bx      lr				\ single tasking so exit
\
\ select next task to run
\
  mov     r2, up
l: [schedule]1
  ldr     r2, [ r2, # 0 tcb.link ]	\ get next task
  ldr     r0, [ r2, # 0 tcb.status ]	\ inspect status
  cmp     r0, # 0			\ 0 = not running
norun-sleep? [if]
  b .ne   [schedule]2
  cmp     r2, up			\ all way round the loop?
  it .eq
    wfi
  b [schedule]1
[else]
  b .eq   [schedule]1
[then]
l: [schedule]2

\ If we have just gone all the way round, there's no need
\ to save the registers as we are just going to reuse them.
  cmp     r2, up			\ in same task
  b .eq   [schedule]3
\
\ save previously running task state - TOS, LP, PSP, LINK
\
  push    { r7, r9, r12, link }		\ stack used registers ; SFP001
  str     rsp, [ up, # 0 tcb.ssp ]	\ save RSP in TCB
  mov     up, r2			\ load new UP
\
\ load selected task state - TOS, LP, PSP, LINK
\
  ldr     rsp, [ up, # 0 tcb.ssp ]	\ restore RSP
  pop     { r7, r9, r12, link }		\ restore registers
l: [schedule]3

event-handler? [if]	\ if user wants the event handler
\
\ event handler exit
\
  ldr     r0, [ up, # 0 tcb.status ]	\ look at status
  and .s  r1, r0, # trg-mask		\ inspect event trigger bit
  ne, if,
    bic     r0, r0, # trg-mask		\ clear trigger bit
    orr     r0, r0, # evt-mask		\ set event bit
    str     r0, [ up, # 0 tcb.status ]	\ restore status
    ldr     r1, [ up, # 0 tcb.event ]	\ run event handler
    orr     r1, r1, # 1			\ set Thumb bit
    bx      r1
  endif,
[then]
  next,
end-code
[else]
CODE pause	\ -- ; the scheduler itself
\ *G The software scheduler itself - without lazy register save.
l: [schedule]
  mvl     r0, # ' multi? >body		\ inspect contents of MULTI?
  ldr     r1, [ r0 ]
  cmp     r1, # 0			\ single tasking if 0
  it .eq
    bx      lr				\ single tasking so exit
\
\ save previously running task - TOS, LP, PSP, LINK
\
  push    { r7, r9, r12, link }		\ stack used registers ; SFP001
  str     rsp, [ up, # 0 tcb.ssp ]	\ save RSP in TCB
\
\ select next task to run
\
l: [schedule]1
  ldr     up, [ up, # 0 tcb.link ]	\ get next task
  ldr     r0, [ up, # 0 tcb.status ]	\ inspect status
  cmp     r0, # 0			\ 0 = not running
  b .eq   [schedule]1
\
\ run selected task - TOS, LP, PSP, LINK
\
  ldr     rsp, [ up, # 0 tcb.ssp ]	\ restore RSP
  pop     { r7, r9, r12, link }		\ restore registers

\
\ event handler
\
event-handler? [if]	\ if user wants the event handler
  ldr     r0, [ up, # 0 tcb.status ]	\ look at status
  and .s  r1, r0, # trg-mask		\ inspect event trigger bit
  ne, if,
    bic     r0, r0, # trg-mask		\ clear trigger bit
    orr     r0, r0, # evt-mask		\ set event bit
    str     r0, [ up, # 0 tcb.status ]	\ restore status
    ldr     r1, [ up, # 0 tcb.event ]	\ run event handler
    orr     r1, r1, # 1			\ set Thumb bit
    bx      r1
  endif,
[then]

\
\ exit
\
  next,
end-code
[then]
: status	\ -- task-status
\ *G Returns the current task's status cell, but with the run
\ ** bit masked out.
  up@ tcb.status @ run-mask invert and  ;

: restart	\ task -- ; mark task TCB as running
\ *G Sets the RUN bit in the task's status cell.
  run-mask swap tcb.status or!  ;

: halt		\ task -- ; reset running bit in TCB
\ *G Clears the RUN bit in the task's status cell.
  run-mask swap tcb.status bic!  ;

: stop          \ -- ; halt oneself
\ *G *\fo{HALT}s the current task, and executes *\fo{PAUSE}.
  self halt  pause
;


\ **************
\ *S Event handling
\ **************
\ *P Event handling is only compiled if the equate
\ ** *\fo{EVENT-HANDLER?} is set non-zero in the control
\ ** file.

event-handler? [if]

: set-event	\ task --
\ *G Set the event trigger in the TCB of *\i{task}.
  tcb.status dup @  trg-mask or  swap !
;

: event?	\ task -- flag
\ *G Returns true if true if *\i{task} has processed an event
\ ** trigger and the event-run flag has not been cleared yet.
\ ** Use *\fo{clr-event-run} below in the event handler to
\ ** clear the event-run flag.
  tcb.status @  evt-mask and 0<>	\ MAJ004
;

: clr-event-run	\ --
\ *G Reset the current task's *\fo{EVENT_RUN} flag.
  up@ tcb.status dup @  ~evt-mask and  swap !
;

: to-event      \ xt task -- ; define action of a task
\ *G Sets XT as the event handler for the task.
  tcb.event !				\ set into TCB of task
;
[then]


\ ****************
\ *S Message handling
\ ****************
\ *P Message handling is only compiled if the equate
\ ** *\fo{MESSAGE-HANDLER?} is set non-zero in the control
\ ** file.

message-handler? [if]
: msg?		\ task -- flag
\ *G Returns true if task has received a message.
  tcb.status @ msg-mask and 0<>
;

internal
: (send-msg)	\ mesg task -- ; store message in task
  self over tcb.msrc !			\ set source
  dup
  tcb.status dup @  msg-mask or  swap !	\ set message bit
  tcb.mesg !				\ set message
;

: (get-message)	\ -- mesg task
  up@ tcb.mesg @			\ message
  up@ tcb.msrc @
  up@ tcb.status
  dup @  ~msg-mask and  swap !		\ reset message bit
;
external

: send-message  \ addr task --
\ *G Send a message to a task.
  begin  dup msg?               	\ if task already has a message
    while  pause                	\ wait
  repeat                        	\ and try again
  (send-msg)				\ send the message
;

: get-message   \ -- addr task
\ *G Wait for any message and return the message and the task it
\ ** came from.
  begin  self msg? 0=           	\ check for a pending message
    while  pause                	\ wait for it
  repeat                        	\ try again
  (get-message)                 	\ read it
;
[then]

event-handler? message-handler? and [if]
: wait-event/msg        \ --
\ *G Wait for a message or an event trigger.
  begin
    self msg?  self event?  or 0=
   while
    pause
  repeat
;
[then]


\ *************************
\ *S Task structure management
\ *************************

internal

((
For Cortex-M3+ the following register usage is the default:
  r15         pc      program counter
  r14         link    link register; bit0=1=Thumb, usually set
  r13         rsp     return stack pointer
  r12         psp     data stack pointer
  r11         up      user area pointer
  r10         --
  r9          lp      locals pointer
  r8          --
  r7          tos     cached top of stack
  r0-r6       scratch
))
(( Extract from scheduler save
  push    { r7 r9 r12, link }		\ stack used registers
  str     rsp, [ up, # 0 tcb.ssp ]	\ save SP in TCB
))
code init-task	\ xt task -- ; Initialise a task stack
\ *G Initialise a task's stack before running it and
\ ** set it to execute the word whose XT is given.
  ldr     r1, [ psp ], # 4		\ get execution address
  orr     r1, r1, # 1			\ set Thumb bit
  mov     r2, rsp			\ save return stack pointer

\ Generate the RSP of the new task
\ the next line works because TASK-U0 is greater than TASK-R0
  sub     r3, tos, # task-u0 task-r0 -	\ generate new task RSP
  mov     rsp, r3

  mov     r0, # 0			\ will need this many times

  str     rsp, [ tos, # r0-offset ]	\ save new R0 (taskID=UP)
  str     r0, [ tos, # 0 tcb.status ]	\ clear new task status

  push    { r1 }			\ R14 LINK, push xt=link
\ the next line works because TASK-U0 is greater than TASK-S0
  sub     r1, tos, # task-u0 task-s0 -	\ calculate new PSP
  sub	  r1, r1, # sp-guard cells	\ generate new PSP ; SFP003
  push    { r1 }			\ R12, new PSP ; SFP003
  sub	  r1, r1, # tos-cached? cells	\ generate new S0 ; SFP003
  str     r1, [ tos, # s0-offset ]	\ that is compatible with SP!
\  push    { tos }			\ R11, new R11=UP
\  push    { r0 }			\ R10, unused
  push    { r0 }			\ R9,  new locals pointer
\  push    { r0 }			\ R8,  unused
  push    { r0 }			\ R7,  new TOS for Cortex
  str     rsp, [ tos, # 0 tcb.ssp ]	\ save RSP
[defined] FS0 [if]	\ separate float and data stacks
  sub     r1, tos, # TASK-U0 TASK-F0 -	\ offset from UP to FS
  str     r1, [ tos, # u# fs0 ]		\ FS0
  str     r1, [ tos, # u# fsp ]		\ FSP
[then]
  mov     rsp, r2			\ restore RSP
  ldr     tos, [ psp ], # 4		\ restore TOS
  next,
end-code

: add-task      \ task -- ; insert into list
\ *G Add the task to the list of tasks after the current task.
  self tcb.link @			\ save task currently pointed to
  over tcb.link !			\ and make new task point to it
  self tcb.link !			\ current task points to new task
;

: sub-task      \ task -- ; remove task from chain
\ *G Remove the task from the task list.
\ look for task which points to required task
  self					\ head of chain ; -- target current
  begin  2dup tcb.link @ <>		\ while curr. does not point to target
   while				\ step through chain
    tcb.link @  dup self =		\ but if task not in list, check for end
    if  2drop  exit  endif		\ and get out
  repeat                                \ -- target points-to-it
\ set task that points to target to point to task pointed to by the target
  swap tcb.link @			\ task pointed to by target
  swap tcb.link !			\ set into link field
;

external

: initiate	\ xt task -- ; start task from scratch
\ *G Start the given task executing the word whose XT is given, e.g.
\ *C   ['] <name> <task> INITIATE
  tuck init-task			\ reset TCB
  dup add-task                          \ add to chain
  run-mask swap tcb.status !		\ mark as running
  pause					\ let it run
;

: sleeper	\ xt task --
\ *G Start task from scratch, but leave it *\fo{HALT}ed. Use in
\ ** the form:
\ *C   ['] <action> <taskname> SLEEPER
\ *P to put a task on the active task list, but as if *\fo{HALT}ed.
\ ** *\fo{SLEEPER} allows you to make a task ready for waking up
\ ** later, perhaps by another task. This avoids having to put
\ ** *\fo{STOP} as the first word in a task. Note that *\fo{SLEEPER}
\ ** does not call *\fo{PAUSE}. See also *\fo{INITIATE}.
  tuck init-task			\ reset TCB
  dup add-task                          \ add to chain
  tcb.status off			\ mark as idle
;

: terminate	\ task --
\ *G Stop a task, and remove it from the list.
  sub-task pause			\ halt task, and remove from chain
;

: init-multi    \ -- ; initialisation with multi-tasking
\ *G Initialise the multitasker and start it. If tasking is
\ ** selected by setting the equate *\fo{TASKING?} in the control
\ ** file, *\i{KERNEL62.FTH} will automatically run this word.
\ ** Make sure that your initialisation code includes
\ ** *\fo{INIT-MULTI} or your code will crash.
  single				\ tasker disabled
  main /tcb erase			\ clean TCB
  main dup tcb.link !			\ initialise MAIN to point to itself
  run-mask main tcb.status !		\ and reset status field
  multi					\ mark tasker enabled
;

: his		\ task uservar -- addr
\ *G Given a task id and a *\fo{USER} variable, returns the
\ ** address of that variable in the given task. This word is
\ ** used to set up *\fo{USER} variables in other tasks.
  self - +
;


\ **********
\ *S Semaphores
\ **********
\ *P The semaphore code is only compiled if the equate
\ ** *\fo{SEMAPHORES?} is set non-zero in the control
\ ** file.

semaphores? [if]

\ *P A *\fo{SEMAPHORE} is an extended variable used for signalling
\ ** between tasks, and for resource allocation. It contains two
\ ** cells, a *\b{counter} and an *\b{arbiter}. The counter field
\ ** is used as a count of the number of times the resource may
\ ** be used, and the arbiter field contains the TCB of the task
\ ** that currently owns it. This field can be used for priority
\ ** arbitration and deadlock detection/arbitration. The count
\ ** field allows the semaphore to be used as a*\b{counted}
\ ** semaphore or as an exclusive access semaphore.

\ *P For example a character buffer may be used where the semaphore
\ ** counter contains the number of available characters.

\ *P An exclusive access semaphore is used to share resources.
\ ** The semaphore is initialised to one, usually by *\fo{SIGNAL}.
\ ** The first task to *\fo{REQUEST} it gains access, and all
\ ** other tasks must wait until the accessing task *\fo{SIGNAL}s
\ ** that it has finished with the resource.

interpreter
: semaphore     \ -- ; -- addr [child]
\ *G Creates a semaphore which returns its address at runtime.
\ ** The count field is initialised to zero for use as counted
\ ** semaphore. Use in the form:
\ *C Semaphore <name>
\ *P If you want this to be an exclusive access semaphore, follow
\ ** this with:
\ *C   1 <name> !
  idata create
    0 ,  0 ,				\ count and arbiter fields
  cdata
;
target

heads? [if]
: semaphore	\ -- ; -- addr [child]
  create
    0 ,  0 ,				\ count and arbiter fields
;
[then]

: signal        \ addr --
\ *G *\fo{SIGNAL} increments the counter field of a semaphore, indicating
\ ** either that another item has been allocated to the resource, or
\ ** that it is available for use again, 0 indicating in use by a task.
  [I					\ must be interrupt safe
    dup incr  cell+ off			\ inc. counter, release
  I]
;

: request       \ sem -- ; get access to resource, wait if count = 0
\ *G *\fo{REQUEST} waits until the counter field of a semaphore
\ ** is non-zero, and then decrements the counter field by one.
  begin
    [I  dup @ 0=			\ n.b. test and set
   while
    I]  pause				\ operations must be
  repeat                                \ non-interruptible and indivisible
  dup decr				\ got it, decrement counter
  self swap cell+ !			\ mark resource as mine
  I]					\ re-enable interrupts
;

[then]


\ ******************
\ *S TASK and START:
\ ******************
\ *P *\fo{TASK <name>} builds a named task user area.
\ ** The action of a task is assigned and the task started
\ ** by the word *\fo{INITIATE}
\ *C   ['] <action> <task> INITIATE

\ *P *\fo{START:} is used inside a colon definition. The code
\ ** before *\fo{START:} is the task's initialisation, performed
\ ** by the current task. The code after *\fo{START:} up to the
\ ** closing *\fo{;} is the action of the task. For example:

\ *E   TASK FOO
\ **   : RUN-FOO
\ **     ...
\ **     FOO START:
\ **     ...
\ **     begin ... pause again
\ **   ;

\ *P All tasks must run in an endless loop, except for initialisation
\ ** code.
\ ** When *\fo{RUN-FOO} is executed, the code after *\fo{START:}
\ ** is set up as the action of task *\fo{FOO} and started.
\ ** *\fo{RUN-FOO} then exits.

\ *P If you want to perform additional actions after starting the task, you
\ ** should use *\fo{INITIATE} to start the task.

variable task-chain	\ -- addr
\ *G Anchors list of all tasks created by *\fo{TASK} and friends.
  main-link task-chain !		\ MAIN already defined

interpreter	\ host version
: task		\ -- ; -- task ; TASK <name> builds a task
\ *G Note that the cross-interpreter's version of *\fo{TASK} has
\ ** been modified from v6.2 onwards to leave the current
\ ** section as *\fo{CDATA}.
  udata
  task-size reserve task-u0 + constant	\ build pointer ; SFP006
  cdata
  here  task-chain @ ,  task-chain !	\ link task
;
target

heads? [if]	\ target version
: task		\ -- ; -- task ; TASK <name> builds a task
\ *G Creates a new task and data area, returning the address
\ ** of the user area at run time. The task is also linked into
\ ** the task chain anchored by *\fo{TASK-CHAIN}.
  here task-u0 +			\ task's user area ; SFP006
  task-size allot
  constant				\ build pointer
  here  task-chain @ ,  task-chain !	\ link task
;
[then]

: start:	\ task -- ; exits from caller
\ *G Used inside a colon definition. The code following
\ ** *\fo{START:} up to the ending semi-colon forms the action
\ ** of the task. The word containing *\fo{START:} finishes at
\ ** *\fo{START:}.
  r> swap initiate
;


\ ******************
\ *S Debugging tools
\ ******************

: .task		\ task --
\ *G Display task's name if it has one, otherwise display its
\ ** address.
  task-chain
  begin					\ -- task link
    @ dup
   while				\ -- task link
    2dup cell - @ = if
      nip  #12 - >name .name space	\ ARM specific
      exit
    endif
  repeat
  drop .lword ."  is not in TASK chain"
;

: .tasks	\ -- ; display all task names
\ *G Display all the tasks anchored by *\fo{TASK-CHAIN}.
  task-chain
  begin
    @ dup
   while
    dup #12 - >name .name space		\ CMx specific
  repeat
  drop
;

: .running	\ --
\ *G Display all the running tasks.
  main
  begin
    cr dup .task
    tcb.link @
    dup main =
  until
  drop
;


\ ********
\ all done
\ ********

decimal


\ *********
\ Test code
\ *********

test-multi? [if]

task task1	\ -- addr ; build task area

500000 value rate1	\ -- n ; number of idle iterations

: action1	\ -- ; action of task1
  decimal  console opvec !
  begin
    [char] * emit
    rate1 0 do pause loop
  again
;

\ To start task use:
\   ' action1 task1 initiate

task task2	\ -- addr ; build task area

400000 value rate2	\ -- n ; number of idle iterations

: run-task2	\ -- ; example of using START:
  task2 start:
    console opvec !
    begin
      [char] / emit
      rate2 0 do pause loop
    again
;

\ To start task use:
\   run-task2

[then]


\ ======
\ *> ###
\ ======

