\ VFP64SX.fth - ARM/Cortex embedded hardware floating point package
\ double precision

((
Copyright (c) 2000-2009, 2010, 2011, 2015, 2016, 2019
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change History
==============
20190103 SFP029 Changed F.P constants to start "f%" to avoid
		conflicts with the binary prefix %.
20161228 SFP028 Embedded updates. FSP=R8.
20161115 SFP027 More changes for FSP=R8.
20160119 SFP026 Only set FPSYSTEM on load, not FPABI.
20160118 SFP025 Changed FSP from USER variable to R8.
20160117 SFP024 Corrected D>F and F>D for d=0.
20160114 SFP023 Changed FTOS to D8 because of the EABI.
20160108 SFP022 Set FPSYSTEM and FPABI on load.
20150529 SFP021 Corrected FSIGN.
20150506 SFP020 Big overhaul. Updates for ANS and Forth 2012.
20150428 SFP019 Converted to VFP double precision - VFP64S.fth.
20150423 SFP018 Converted to VFP single precision - VFP32S.fth.
20150330 SFP017 Overhauled to run on VFX Forth.
		Updated for systems whose shift instructions do not return
		0 for a shift count of 32 or more.
		Reordered for hosted systems.
		Converted to use separate float and data stacks.
20150122 BJC016 Previous change BJC014 did not work properly. Now using 5/4
                rounding for values printed in non exponent format using F.
                Added word NORM-FROUND which applies 5/4 rounding assuming that
                the value is printed 'normally'.
20111201 MPE015 Updated documentation.
20110302 BJC014 Corrected OP-PREPARE which assumes 8 sig figs regardless
                of the actual precision required.  It now uses PLACES to
                ensure that output from F. etc is now rounded to the places
                displayed.
20100630 MPE013 Improved compatibility with VFX Forth for Win/Lin/DOS.
                Minor refactoring.
20100628 BJC012 Changes to FNUMBER? allow use of HEX and other
                integers after REALS executed:
                  If BASE is not decimal FNUMBER? always does
                integer conversion.
                  FCHECK now recognises the integer prefixes
                ('#','$','0x' etc.) so these are converted as
                integers.
                The character used as a decimal separator is
                set in VALUE DPchar, defined here (if not previously).
                This allows use of ',' or indeed other characters.
                The US/UK convention '.' is the default.
20100526 BJC011 CDATA is selected prior to defining all FARRAYs
                for constant data (previously CDATA was assumed,
                and things broke if it wasn't).
20090729 SBD010 Fixed ROUND at limits.
		Improved accuracy in ?10PWR.
20090319 SFP009 Added INTERPRETER versions of F@ F! and F,
		Added FBUFF.
20090220 MPE008 Added Alberto Pasquale's IEEE conversion words.
20071105 SFP007 Made >FLOAT convert only in DECIMAL.
20071012 SFP006 Improved documentation.
20060328 SFP005 Reordered to remove forward references.
20050705 SFP004 Improved FVARIABLE for optimised systems.
20021112 SFP003 DOCGENned it.
20000713 SFP002 Renamed DINT to F>D for consistency. F>D is the ANS word.
		The original F>D was just a synonym. Similarly SINT was
		renamed to F>S
20000706 SFP001 Applied improvements supplied by Hiden Analytical.
		Added minor performance optimisations.
		Made F/ and DINT return clipped values on overflow.
19981029 MSD000 Brought the common stuff out of all the various versions of
         MSD000 SOFTFP.FTH into a single (that is, *this*) file.

Polynomials from
  Cody and Waite
    http://ebook30.com/science/mathematics/190013/william-j.-cody-software-manual-for-the-elementary-functions.html

  Abramowitz & Stegun
    http://www.math.ucla.edu/~cbm/aands/
    http://www.convertit.com/Go/Bioresearchonline/Reference/AMS55.ASP
))


\ ==========
\ *! vfp64sx
\ ==========
\ *T VFP Floating Point - double precision

\ ***************
\ *S Introduction
\ ***************
\ *P The Forth data stack and the floating point stack are separate.
\ ** As with the data and return stacks, the floating point stack
\ ** grows down.
\ ** The floating point data storage format is IEEE 64 bit (double
\ ** precision) format.The source code is in the file
\ ** *\i{Cortex/VFP64SX.fth}. As of January 2017, the  float stack
\ ** pointer is in R8 and the top of the float stack is in S8 for
\ ** performance. The file should also work with minimal
\ ** change on ARM32 CPUs but has not been tested for these.

\ *P A single precision version is provided in *\i{Cortex/VFP32SX.fth}.
\ ** Note that single-precision floating point only has 23 bits in
\ ** the mantissa and 8 bits in the exponent; and thus has a
\ ** severely restricted dynamic range.


\ *********************
\ *S Compiling the code
\ *********************
\ *P The source code for the ARM/Cortex version of the code is in
\ ** the file *\i{Cortex/VFP64SX.fth}. In order to provide full
\ ** initialisation in other files, the following equates must
\ ** be set in the control file before any code is compiled:
\ *E   0 equ softfp?     \ true for software floating point
\ **   1 equ hardfp?     \ true for hardware floating point
\ **   1 equ fpstack?    \ true for a separate floating point stack,
\ **                     \ in which case FP-SIZE must be non-zero
\ **   $0100 equ FP-SIZE \ size of FP stack in bytes. Must not be 0.
\ *P The separate floating point stack is used when *\fo{fpstack?}
\ ** is true and *\fo{FP-SIZE} is defined and is non-zero. Add
\ ** the line below to compile the code.
\ *E   include %CpuDir%/VFP64SX.fth

\ *P The code must be compiled with *\i{xArmCortexHfpDev}, which
\ ** has support for compiling IEEE F.P. numbers.


\ **********************************
\ *S Entering floating-point numbers
\ **********************************
\ *P Floating point number entry is enabled by *\fo{REALS} and
\ ** disabled by *\fo{INTEGERS}.

\ *P Floating-point numbers of the form 0.1234e1 are required
\ ** (see *\fo{FNUMBER?}) during interpretation and compilation
\ ** of source code. Floating-point numbers are compiled as
\ ** literal numbers when in a colon definition (compiling) and
\ ** placed on the stack when outside a definition (interpreting).
\ ** Inside a colon definition, a floating point literal number
\ ** must be preceded by *\fo{F#}.
\ *C : foo  ... f# 1.234e0 ...  ;

\ *P The more flexible word *\fo{>FLOAT} accepts numbers in two
\ ** forms, 1.234 and 0.1234e1. Both words are documented later
\ ** in this chapter. See also the section on *\i{Gotchas} later
\ ** in this chapter.


\ *******************************
\ *S Creating and using variables
\ *******************************
\ *P To create a variable, use *\fo{FVARIABLE}. *\fo{FVARIABLE} works in the
\ ** same way as *\fo{VARIABLE}. For example, to create a floating-point
\ ** variable called *\fo{VAR1} you code:
\ *C   FVARIABLE VAR1
\ *P When *\fo{VAR1} is used, it returns the address of the floating-point
\ ** number.

\ *P Two words are used to access floating-point variables,
\ ** *\fo{F@} and *\fo{F!}. These are analogous to *\fo{@} and
\ ** *\fo{!}.

\ *********************
\ *S Creating constants
\ *********************
\ *P To create a floating-point constant, use *\fo{FCONSTANT}, which
\ ** is analogous to *\fo{CONSTANT}. For example, to generate a
\ ** floating-point constant called *\fo{CON1} with a value of 1.234,
\ ** you enter:
\ *C   F# 1.234e0 FCONSTANT FCON1
\ *P When *\fo{FCON1} is executed, it returns 1.234 on the float
\ ** stack.

\ ***************************
\ *S Using the supplied words
\ ***************************
\ *P The supplied words split into several groups:
\ *(
\ *B sines, cosines and tangents
\ *B arc sines, cosines and tangents
\ *B arithmetic functions
\ *B logarithms
\ *B powers
\ *B displaying floating-point numbers
\ *B inputting floating-point numbers
\ *)

\ *P The following functions only exist as target words so you
\ ** cannot use them in calculations in your source code when
\ ** outside a colon definition.

\ *N Calculating sines, cosines and tangents
\ *P To calculate sine, cosine and tangent, use *\fo{FSIN}, *\fo{FCOS} and
\ ** *\fo{FTAN} respectively. Angles are expressed in radians.

\ *N Calculating arc sines, cosines and tangents
\ *P To calculate arc sine, cosine and tangent, use *\fo{FASIN}, *\fo{FACOS}
\ *P and *\fo{FATAN} respectively. They return an angle in radians.

\ *N Calculating logarithms
\ *P Two words are supplied to calculate logarithms, *\fo{FLOG} and *\fo{FLN}.
\ ** *\fo{FLOG} calculates a logarithm to base 10 (decimal).
\ ** *\fo{FLN} calculates a logarithm to base e. Both take a
\ ** floating-point number in the range from 0 to Einf.

\ *N Calculating powers
\ *P Three power functions are supplied:
\ *C   FEXP F10^X X^Y

\ *********************
\ *S Degrees or radians
\ *********************
\ *P The angular measurement used in the trigonometric functions
\ ** are in radians. To convert between degrees and radians use
\ ** *\fo{RAD>DEG} or *\fo{DEG>RAD}. *\fo{RAD>DEG} converts an angle from radians
\ ** to degrees. *\fo{DEG>RAD} converts an angle from degrees to radians.

\ ************************************
\ *S Displaying floating-point numbers
\ ************************************
\ *P Two words are available for displaying floating-point numbers,
\ ** *\fo{F.} and *\fo{E.}. The word *\fo{F.} takes a floating-point
\ ** number from the stack and displays it in the form xxxx.xxxxx
\ ** or x.xxxxxEyy depending on the size of the number. The word
\ ** *\fo{E.} displays the number in the latter form.

\ *************************************
\ *S Number formats, ANS and Forth-2012
\ *************************************
\ *P The ANS Forth standard specifies that floating point numbers
\ ** must  be entered in the form 1.234e5 and must contain a
\ ** point '.' and 'e' or 'E'. In order to distinguish between
\ ** double and floating point numbers, a floating point number
\ ** must contain 'e' and an exponent.

\ **********************
\ *S FP Stack primitives
\ **********************
\ *P The floating point stack pointer is register R8. The top float
\ ** is cached in register D8.

8 equ FPCELL	\ -- u
\ *G Size of a floating point number in bytes.

FPCELL constant FPCELL	\ -- u
\ *G Size of a floating point number in bytes.

FPCELL setFloatSize	\ --
\ *G Tell the cross compiler the size in memory of a floating
\ ** point number.
4 setFloatAlignment	\ must be aligned
\ *G Tell the cross compiler alignment in memory of a floating
\ ** point number.

finit			\ initialise host FP.
host previous
: (doFloatAlignment)	\ --
\ *G Host word that performs float alignment.
  ?SaveLink  chere(t) 2 and
  if  $BF00 w,c(t)  endif
;
' (doFloatAlignment) is doFloatAlignment
target

: finit		\ F: i*f -- ; resets FPU and FP stack
\ *G Reset the floating point stack.
\ ** Do not forget to use this in a task before using floating
\ ** point.
  $00F0:0000 $E000:ED88 or!		\ enable FPU, set bits 23:20 of CPACR
  dsb#0f isb#0f				\ flush
  fs0 @ fsp!
;
' finit AtCold

: fdepth        \ -- #f
\ *G Floating point equivalent of *\fo{DEPTH}. The result is returned
\ ** on the Forth data stack.
  FS0 @  fsp@ -  FPCELL /  ;		\ possible # of floats on FP stack

[undefined] clz [if]
code CLZ	\ x -- u
\ *G Return the number of leading zeros in x.
  clz		tos, tos
  next,
end-code
[then]

: DCLZ		\ dx -- u
\ *G Return the number of leading zeros in the double dx.
  clz dup #32 = if			\ -- sx cnt
    drop  clz #32 +
  else
    nip
  then
;

code >fs	\ f64 -- ; F: -- f64
\ *G Copy a float from the data stack to the floating point stack.
\ ** The top of the data stack contains bits 63:32.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ save current FTOS
  ldr		r3,  [ psp ], # 4	\ low portion of F64
  vmov		d8, r3, tos		\ copy TOS:R3 to float
  ldr		tos, [ psp ], # 4
  next,
end-code

host
2variable ftemp
: >fs  swap ftemp 2!  ftemp df@  ;
target
interpreter
: >fs  >fs  ;
target

code fs>	\ F: f64 -- ; -- f64
\ *G Copy a float from the float stack to the data stack.
\ ** The top of the data stack contains bits 63:32.
  str		tos, [ psp, # -4 ] !	\ save old TOS
  vmov		r3, tos, d8		\ TOS:R3 := FTOS
  str		r3, [ psp, # -4 ] !	\ save low part of FTOS
  vldr		d8, [ r8, # 0 ]		\ restore FTOS
  add .s	r8, r8, # FPCELL	\ update FSP
  next,
end-code

code fps@	\ -- fps
\ *G Read the floating point status/control register.
  str		tos, [ psp, # -4 ] !	\ save old TOS
  vmrs		tos, FPSCR
  next,
end-code

code fps!	\ fps --
\ *G Set the floating point status/control register.
  vmsr		FPSCR, tos
  ldr		tos, [ psp ], # 4
  next,
end-code

code exp@	\ F: f -- f ; -- exp(2)
\ *G Copy the exponent of the top float to the data stack.
\ ** The IEEE exponent offset is removed.
  str		tos, [ psp, # -4 ] !	\ preserve old TOS
  vmov		r3, tos, d8		\ TOS:R3 := FTOS
  mov		tos, tos lsr # #20	\ exponent is 11 bits at bit 52
  mov		r2,     # $07F0		\ mask = $07FF
  orr		r2, r2, # $000F
  and		tos, tos, r2		\ remove sign bit
  sub		tos, tos, # $03F0	\ remove offset 1022
  sub		tos, tos, # $000E
  mov		pc, lr
end-code

code exp!		\ exp(2) -- ; F: f -- f'
\ *G Change/Set the exponent of the top float.
\ ** The IEEE exponent offset is added.
  mov		r2,     # $07F0		\ exponent mask
  orr		r2, r2, # $000F
  mov		r3, r2 lsl # #20	\ 11 bits at bit 52
  vmov		r5, r4, d8		\ copy FTOS to R4:R5
  bic .s	r4, r4, r3		\ clear old exponent
  add .s	tos, tos, # $03F0	\ add exponent offset 1022
  add .s	tos, tos, # $000E
  mov .s	tos, tos lsl # #20	\ shift into place
  and .s	tos, tos, r3		\ clip
  orr .s	r4, r4, tos		\ merge in new exponent
  vmov		d8, r5, r4		\ put FTOS := R4:R5
  ldr		tos, [ psp ], # 4	\ restore TOS
  next,
end-code

code F!		\ F: r -- ; addr --
\ *G Stores *\i{r} at *\i{addr}.
  vstr		d8, [ tos, # 0 ]	\ store FTOS/S4
  vldr		d8, [ r8, # 0 ]		\ reload FTOS
  add .s	r8, r8, # FPCELL	\ update FSP
  ldr		tos, [ psp ], # 4	\ restore TOS
  next,
end-code

code F@		\ addr -- ; F: -- r
\ *G Fetches *\i{r} from *\i{addr}.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ save current FTOS
  vldr		d8, [ tos ]		\ fetch value
  ldr		tos, [ psp ], # 4	\ restore TOS
  next,
end-code

synonym DF! F!		\ F: r -- ; addr --
\ *G Stores *\i{r} at *\i{addr} in IEEE 64 bit format.

synonym DF@ F@		\ addr -- ; F: -- r
\ *G Fetches *\i{r} from *\i{addr}, which contains a float
\ ** in IEEE 64 bit format..

code SF!	\ F: r -- ; addr --
\ *G Stores *\i{r} at *\i{addr} in IEEE 32 bit format.
  vcvt .f32<f64	s1, d8			\ 32 bit version of FTOS
  vstr		s1, [ tos, # 0 ]	\ store FTOS/S1
  vldr		d8, [ r8, # 0 ]		\ reload FTOS
  add .s	r8, r8, # FPCELL	\ update FSP
  ldr		tos, [ psp ], # 4	\ restore TOS
  next,
end-code

code SF@	\ addr -- ; F: -- r
\ *G Fetches *\i{r} from *\i{addr} in IEEE 32 bit format.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ save current FTOS
  vldr		s1, [ tos ]		\ fetch value
  vcvt .f64<f32	d8, s1			\ convert 32 bit to 64 bit
  ldr		tos, [ psp ], # 4	\ restore TOS
  next,
end-code

: F,		\ F: r --
\ *G Lays a real number into the dictionary, reserving FPCELL bytes.
  here  FPCELL allot  f!  ;

synonym DF, F,
\ *G Lays a real number into the dictionary as an IEEE 64 bit number.

: SF,		\ F: r --
\ *G Lays a real number into the dictionary as an IEEE 32 bit number.
  here  4 allot  sf!  ;

code FDUP  	\ F: r -- r r
\ *G Floating point equivalent of *\fo{DUP}.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ save current FTOS
  next,
end-code

code FOVER	\ F: r1 r2 -- r1 r2 r1
\ *G Floating point equivalent of *\fo{OVER}.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ save r2=current FTOS
  vldr		d8, [ r8, # FPCELL ]	\ r1 := new FTOS
  next,
end-code

code FSWAP	\ F: r1 r2 -- r2 r1
\ *G Floating point equivalent of *\fo{SWAP}.
  vldr		d3, [ r8, # 0 ]		\ s3 := FNOS = r1
  vstr		d8, [ r8, # 0 ]		\ save old FTOS = r2
  vmov .f64	d8, d3			\ FTOS := r1
  next,
end-code

code FPICK	\ u -- ; F: fu..f0 -- fu..f0 fu
\ *G Floating point equivalent of *\fo{PICK}.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ save current FTOS
  mov .s	tos, tos .lsl # 3	\ 8 bytes per item
  add .s	r1, r8, tos		\ pointer to fu
  vldr		d8, [ r1, # 0 ]		\ new FTOS := fu
  ldr		tos, [ psp ], # 4	\ restore TOS
  next,
end-code

code FROT	\ F: r1 r2 r3 -- r2 r3 r1
\ *G Floating point equivalent of *\fo{ROT}.
  vmov .f64	d3, d8			\ s3 := r3
  vldr		d2, [ r8, # 0 ]		\ s2 := r2
  vldr		d8, [ r8, # FPCELL ]	\ s4/FTOS := r1
  vstr		d3, [ r8, # 0 ]		\ FNOS := r3
  vstr		d2, [ r8, # FPCELL ]	\ F3OS := r2
  next,
end-code

code F-ROT		\ F: r1 r2 r3 -- r3 r1 r2
\ *G Floating point equivalent of *\fo{-ROT}.
  vmov .f64	d3, d8			\ s3 := r3
  vldr		d2, [ r8, # 0 ]		\ s2 := r2
  vldr		d1, [ r8, # FPCELL ]	\ s1 := r1
  vmov .f64	d8, d2			\ FTOS := r2
  vstr		d1, [ r8, # 0 ]		\ FNOS := r1
  vstr		d3, [ r8, # FPCELL ]	\ F30S := r3
  next,
end-code

code FDROP	\ F: r --
\ *G Floating point equivalent of *\fo{DROP}.
  vldr		d8, [ r8, # 0 ]		\ restore FTOS
  add .s	r8, r8, # FPCELL	\ update FSP
  next,
end-code

code FNIP		\ F: r1 r2 -- r2
\ *G Floating point equivalent of *\fo{NIP}.
  add .s	r8, r8, # FPCELL	\ update FSP
  next,
end-code

code f>r	\ F: f -- ; R: -- f
\ *G Put float onto return stack.
  sub .s	rsp, rsp, # FPCELL
  vstr		d8, [ rsp, # 0 ]	\ FTOS -> return stack
  vldr		d8, [ r8, # 0 ]		\ restore FTOS
  add .s	r8, r8, # FPCELL	\ update FSP
  next,
end-code

code fr>	\ R: f -- ; F: -- f
\ *G Pull float from the return stack.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ FTOS -> float stack
  vldr		d8, [ rsp, # 0 ]	\ restore FTOS from r. stack
  add .s	rsp, rsp, # FPCELL
  next,
end-code

Thumb? [if]
code flit	\ F: -- f ; inline literal
\ *G Run-time routine for a floating point literal. Cortex version.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ save current FTOS
  bic		r1, lr, # 1		\ address of literal, ignore T bit
  vldr		d8, [ r1, # 0 ]		\ get literal
  add .s	lr, lr, # FPCELL	\ step over it
  next,
end-code
[else]
code flit	\ F: -- f ; inline literal
\ *G Run-time routine for a floating point literal. ARM32 version.
  sub .s	r8, r8, # FPCELL	\ update FSP
  vstr		d8, [ r8, # 0 ]		\ save current FTOS
  vldr		d8, [ lr, # 0 ]		\ get literal
  add .s	lr, lr, # FPCELL	\ step over it
  next,
end-code
[then]

[undefined] dlshift [if]
: DLSHIFT	\ d u -- d<<u
\ *G Double number left shift, with *\i{u} in the range 1..64. 
  dup #32 >=
  if -rot  drop 0 swap  rot #32 - then	\ -- 0 dl u-32
  >r swap dup r@ lshift  swap #32 r@ - rshift  rot r> lshift or
;

: DRSHIFT 	\ d u -- d>>u
  dup #32 >=
  if  -rot  nip 0  rot #32 -  then	\ -- dh 0 u-32
  >r dup r@ rshift  swap #32 r@ - lshift  rot r> rshift or swap
;
[then]


\ ********************************
\ *S Floating point defining words
\ ********************************

target-only
: FVARIABLE	\ "<spaces>name" -- ; Run: -- addr
\ *G Use in the form: *\fo{FVARIABLE <name>} to create a variable
\ ** that will hold a floating point number.
  create  FPCELL allot
;

: FCONSTANT	\ F: r -- ; "<spaces>name" -- ; Run: -- r
\ *G Use in the form: *\fo{<float> FCONSTANT <name>} to create a
\ ** constant that returns a floating point number.
  create  f,  does> f@  ;

: FARRAY	\ "<spaces>name" fn-1..f0 n -- ; Run: i -- ; F: -- ri
\ *G Create an initialised array of floating point numbers. Use
\ ** in the form:
\ *C   fn-1 .. f1 f0 n FARRAY <name>
\ *P to create an array of n floating point numbers. When the
\ ** array *\fo{name} is executed, the index i is used to return
\ ** the address of the i'th 0 zero-based element in the array.
\ ** For example:
\ *C   4e0 3e0 2e0 1e0 0e0 5 FARRAY TEST
\ *P will set up an array of five elements.
\ ** Note that the rightmost float (0e0) is element 0.
\ ** Then *\fo{i TEST} will return the *\{i}th element.
  create
    0 do  f,  loop
  does>
    swap FPCELL * + f@
;

: FBUFF		\ u "name" -- ; i -- addr
\ *G Creates a buffer for *\i{u} floats in the current memory
\ ** section. The child action is to return the address of the
\ ** *\i{i}th element (zero-based).
\ *C   10 fbuff foo
\ *P Creates an buffer for ten float elements.
\ *C   3 foo
\ *P Returns the address of element 3 in the buffer.
  create
    FPCELL * allot
  does>
    swap FPCELL * +
;
host&target

interpreter
: FVARIABLE	\ "<spaces>name" -- ; Run: -- addr
  ihere FPCELL iallot constant  ;

: FCONSTANT	\ F: r -- ; "<spaces>name" -- ; Run: -- r
  create  f,  does> f@  ;

: FARRAY	\ "<spaces>name" fn-1..f0 n -- ; Run: i -- ; F: -- ri
  create
    0 do  f,  loop
  does>
    swap FPCELL * + f@
;

: FBUFF		\ u "name" -- ; i -- addr
  create
    FPCELL * allot
  does>
    swap FPCELL * +
;

target


\ *******************
\ *S Type conversions
\ *******************

code FSIGN	\ F: fn -- |fn| ; -- flag ; true if negative
\ *G Return the absolute value of fn and a flag which is true
\ ** if fn is negative. F.P. stack operation.
  str		tos, [ psp, # -4 ] !
  vcmp		d8, # 0
  vmrs		APSR_nzcv FPSCR		\ copy flags
  mi, if,
    vneg .f64	d8, d8			\ FTOS := -FTOS
    mvn	.s	tos, # 0		\ tos := invert(0)
  else,
    mov	.s	tos, # 0
  then,
  next,
end-code

((
code *2^n	\ n -- ; F: f -- f*2^n
  vmov		r4, r0, d8		\ IEEE 64 bit float in r0:r4
  mov		r1, r0 .lsr # #20	\ r1 has exponent
  mov		r5,     # $7F0		\ 11 bit exponent mask
  orr		r5, r5, # $00F
  and		r1, r1, r5		\ mask out sign bit
  add		r1, r1, tos		\ add n to exponent
  and		r1, r1, r5		\ clip to 11 bits
  mov		r1, r1 .lsl # #20	\ put back in exponent position
  mov		r2,     # $7F00:0000	\ build mask
  orr		r2, r2, # $00F0:0000
  bic		r0, r0, r2		\ clear original exponent
  orr		r0, r0, r1		\ merge new exponent
  vmov		d8, r0			\ put back
  ldr		tos, [ psp ], # 4	\ restore TOS
  mov		pc, lr
end-code
))

: D>F		\ d -- ; F: -- fn
\ *G Converts a double integer to a float.
  2dup or if
    dup 0< >r  dabs			\ strip sign
    2dup dclz dup #63 swap - >r		\ form exponent ; -- d clz ; R: -- sign exp
    #11 - dup 0<			\ first '1' bit to bit 52 of upper word
    if  abs drshift  else  dlshift  then
    $000F:FFFF and 			\ clip
    r> #1023 + $7FF and #20 lshift  or	\ merge exponent
    r> $8000:0000 and or		\ bit 31 is sign bit
  then
  >fs
;

: S>F	\ n -- ; F: -- fn
\ *G Converts a single signed integer to a float.
  s>d d>f
;

: f>d		\ f -- ; -- dint(f)
\ *G Converts a float to a double integer.
\ ** Note that *\fo{F>D} truncates the number towards zero
\ ** according to the ANS specification.
  {: | exp temp[ FPCELL ] -- :}
  exp@ -> exp  temp[ f!
  temp[ 2@ swap				\ exponent in top half
  2dup or if				\ nz
    $000F:FFFF and			\ bits 51:0 are mantissa
    $0010:0000 or				\ bit 52 is hidden
    #53 exp -  dup 0<			\ find shift count
    if  dlshift  else  drshift  then	\ to integer
    temp[ cell + @ 0<
    if  dnegate  then
  then
;

: f>s		\ f -- ; -- dint(f)
\ *G Converts a float to a single integer.
\ ** Note that *\fo{F>S} truncates the number towards zero
\ ** according to the ANS specification.
  f>d drop
;

: FINT		\ F: f1 -- f2
\ *G Chop the number towards zero to produce a floating point
\ ** representation of an integer.
  f>d d>f
;


\ *************
\ *S Arithmetic
\ *************

code FNEGATE	\ F: r1 -- r2
\ *G Floating point negate.
  vneg .f64	d8, d8
  next,
end-code

: ?FNEGATE	\ n -- ; F: fn -- fn|-fn
\ *G If n is negative, negate fn.
  0< if  fnegate  then
;

code FABS	\ F: fn -- |fn|
\ *G Floating point absolute.
  vcmp		d8, # 0
  vmrs		APSR_nzcv FPSCR		\ copy flags
  mi, if,
    vneg .f64	d8, d8			\ FTOS := -FTOS
  then,
  next,
end-code

code F+		\ F: r1 r2 -- r3
\ *G Floating point addition.
  vldr		d3, [ r8, # 0 ]		\ r1
  add .s	r8, r8, # FPCELL
  vadd .f64	d8, d3, d8
  next,
end-code

code F-		\ F: r1 r2 -- r3
\ *G Floating point subtraction, r3=r1-r2
  vldr		d3, [ r8, # 0 ]		\ r1
  add .s	r8, r8, # FPCELL
  vsub .f64	d8, d3, d8
  next,
end-code

code F*		\ F: r1 r2 -- r3
\ *G Floating point multiply.
  vldr		d3, [ r8, # 0 ]		\ r1
  add .s	r8, r8, # FPCELL
  vmul .f64	d8, d3, d8
  next,
end-code

code F/		\ F: r1 r2 -- r3
\ *G Floating point divide.
  vldr		d3, [ r8, # 0 ]		\ r1
  add .s	r8, r8, # FPCELL
  vdiv .f64	d8, d3, d8
  next,
end-code

code fsqrt  	\ F: f1 -- f2
\ *G F2=sqrt(f1).
  vsqrt .f64	d8, d8
  next,
end-code

: FSEPARATE	\ F: f1 f2 -- f3 f4
\ *G Leave the signed integer quotient f4 and remainder f3 when
\ ** f1 is divided by f2. The remainder has the same sign as the
\ ** dividend.
  fover fover f/
  frot frot fabs
  fswap fsign >r fswap
  2 fpick fabs fint f* f-
  r>
  if  fnegate  then
  fswap
;

: FFRAC		\ F: f1 f2 -- f3
\ *G Leave the fractional remainder from the division f1/f2. The
\ ** remainder takes the sign of the dividend.
  fseparate fdrop
;


\ ***********************
\ *S Relational operators
\ ***********************

code F0<	\ F: f1 -- ; -- flag
\ *G Floating point *\fo{0<}.
  str		tos, [ psp, # -4 ] !	\ preserve old TOS
  vcmp		d8, # 0
  vmrs		APSR_nzcv FPSCR		\ copy flags
  mi, if,
    mvn		tos, # 0
  else,
    mov		tos, # 0
  then,
  vldr		d8, [ r8, # 0 ]		\ restore FTOS
  add .s	r8, r8, # FPCELL
  next,
end-code

code F0>	\ F: f1 -- ; -- flag
\ *G Floating point *\fo{0>}.
  str		tos, [ psp, # -4 ] !	\ preserve old TOS
  vcmp		d8, # 0
  vmrs		APSR_nzcv FPSCR		\ copy flags
  gt, if,
    mvn .s	tos, # 0
  else,
    mov	.s	tos, # 0
  then,
  vldr		d8, [ r8, # 0 ]		\ restore FTOS
  add .s	r8, r8, # FPCELL
  next,
end-code

code F0=  		\ F: f1 -- ; -- flag
\ *G Floating point *\fo{0=}.
  str		tos, [ psp, # -4 ] !	\ preserve old TOS
  vcmp		d8, # 0
  vmrs		APSR_nzcv FPSCR		\ copy flags
  eq, if,
    mvn	.s	tos, # 0
  else,
    mov	.s	tos, # 0
  then,
  vldr		d8, [ r8, # 0 ]		\ restore FTOS
  add .s	r8, r8, # FPCELL
  next,
end-code

code F0<>  	\ F: f1 -- ; -- flag
\ *G Floating point *\fo{0<>}.
  str		tos, [ psp, # -4 ] !	\ preserve old TOS
  vcmp		d8, # 0
  vmrs		APSR_nzcv FPSCR		\ copy flags
  ne, if,
    mvn .s	tos, # 0
  else,
    mov .s	tos, # 0
  then,
  vldr		d8, [ r8, # 0 ]		\ restore FTOS
  add .s	r8, r8, # FPCELL
  next,
end-code

: F=		\ F: f1 f2 -- ; -- flag
\ *G Floating point *\fo{=}.
  f- f0=
;

: F<  		\ F: r1 r2 -- ; -- flag
\ *G Floating point *\fo{<}.
  f- f0<
;

: F>  		\ F: f1 f2 -- ; -- flag
\ *G Floating point *\fo{>}.
  f- f0>
;

: FMAX  	\ F: r1 r2 -- r1|r2
\ *G Floating point *\fo{MAX}.
  fover fover f<
  if  fswap  then
  fdrop
;

: FMIN  	\ F: r1 r2 -- r1|r2
\ *G Floating point *\fo{MIN}.
  fover fover f>
  if  fswap  endif
  fdrop
;


\ ****************
\ *S Miscellaneous
\ ****************

: FALIGNED	\ addr -- f-addr
\ *G Aligns the address to accept an 8-byte float.
  FPCELL 1- +  FPCELL negate and  ;

: FALIGN  	\ --
\ *G Aligns the dictionary to accept an 8-byte float.
  here dup faligned swap - allot  ;

synonym DFALIGNED FALIGNED	\ addr -- f-addr
\ *G Aligns the address to accept an 8-byte float.

synonym DFALIGN FALIGN  	\ --
\ *G Aligns the dictionary to accept an 8-byte float.

Synonym SFALIGNED ALIGNED	\ addr -- f-addr
\ *G Aligns the address to accept a 4-byte float.

Synonym SFALIGN ALIGN	  	\ --
\ *G Aligns the dictionary to accept a 4-byte float.

: FLOAT+  	\ f-addr1 -- f-addr2
\ *G Increments addr by 8, the size of a float.
  FPCELL +  ;
: FLOATS  	\ n1 -- n2
\ *G Returns *\i{n2}, the size of *\i{n1} floats.
  FPCELL *  ;
Synonym DFLOAT+ FLOAT+ 	\ f-addr1 -- f-addr2
\ *G Increments *\i{addr} by 8, the size of a D-float.
Synonym DFLOATS FLOATS 	\ n1 -- n2
\ *G Returns *\i{n2}, the size of *\i{n1} D-floats.
Synonym SFLOAT+ 4+ 	\ f-addr1 -- f-addr2
\ *G Increments *\i{addr} by 4, the size of an S-float.
Synonym SFLOATS CELLS 	\ n1 -- n2
\ *G Returns *\i{n2}, the size of *\i{n1} S-floats.


\ ***************************
\ *S Powers of ten operations
\ ***************************
\ *P Floating point IEEE numbers have the following approximate
\ ** ranges:
\ *(
\ *B Single precision - 10^-44 to 10^+38
\ *B Double precision - 10^-323 to 10^308
\ *)
\ *P As a result, the input code is different for 32 bit and 64
\ ** bit floats.
f# 1.0e-256 fconstant f%10^-256
\ *G Floating point 1.0e-256.
f# 1.0e-32 fconstant f%10^-32
\ *G Floating point 1.0e-32.
f# 1.0e-16 fconstant f%10^-16
\ *G Floating point 1.0e-16.
f# 0.1e0 fconstant f%.1
\ *G Floating point 0.1.
f# 1.0e0 fconstant f%1
\ *G Floating point 1.0.
f# 10.0e0 fconstant f%10
\ *G Floating point 10.0.
f# 1.0e16 fconstant f%10^16
\ *G Floating point 1.0e16.
f# 1.0e32 fconstant f%10^32
\ *G Floating point 1.0e32.
f# 1.0e256 fconstant f%10^256
\ *G Floating point 1.0e256.

F# 1.0E15  F# 1.0E14  F# 1.0E13  F# 1.0E12
F# 1.0E11  F# 1.0E10  F# 1.0E09  F# 1.0E08
F# 1.0E07  F# 1.0E06  F# 1.0E05  F# 1.0E04
F# 1.0E03  F# 1.0E02  F# 1.0E01  F# 1.0e00
 16 FARRAY POWERS-OF-10E1
\ *G An array of 16 powers of ten starting at 10^0
\ ** in steps of 1.

F# 1.0E256  F# 1.0E240  F# 1.0E224  F# 1.0E208
F# 1.0E192  F# 1.0E176  F# 1.0E160  F# 1.0E144
F# 1.0E128  F# 1.0E112  F# 1.0E96   F# 1.0E80
F# 1.0E64   F# 1.0E48   F# 1.0E32   F# 1.0E16
F# 1.0E00
17 FARRAY POWERS-OF-10E16
\ *G An array of 17 powers of ten starting at 10^0
\ ** in steps of 16.

F# 1.0E-15 F# 1.0E-14 F# 1.0E-13 F# 1.0E-12 F# 1.0E-11
F# 1.0E-10 F# 1.0E-09 F# 1.0E-08 F# 1.0E-07 F# 1.0E-06
F# 1.0E-05 F# 1.0E-04 F# 1.0E-03 F# 1.0E-02 F# 1.0E-01
F# 1.0e-00
16 FARRAY POWERS-OF-10E-1
\ *G An array of 16 powers of ten starting at 10^0
\ ** in steps of -1.

F# 1.0E-256  F# 1.0E-240  F# 1.0E-224  F# 1.0E-208
F# 1.0E-192  F# 1.0E-176  F# 1.0E-160  F# 1.0E-144
F# 1.0E-128  F# 1.0E-112  F# 1.0E-96   F# 1.0E-80
F# 1.0E-64   F# 1.0E-48   F# 1.0E-32   F# 1.0E-16
F# 1.0E00
17 FARRAY POWERS-OF-10E-16
\ *G An array of 17 powers of ten starting at 10^0
\ ** in steps of -16.

: RAISE_POWER	\ exp(10) -- ; F: f -- f'
\ *G Raise the power in preparation for number formatting.
  DUP >R  #15 AND ?DUP
  IF  POWERS-OF-10E1 F*  THEN
  R@ 4 rshift  #15 AND ?DUP
  IF  #16 MIN POWERS-OF-10E16 F*  THEN
  R> 8 rshift  0
  ?DO  f%10^256 F*  LOOP
;

: SINK_FRACTION	\ exp(10) -- ; F: f -- f'
\ *G Reduce the power in preparation for number formatting.
  ABS DUP >R  #15 AND ?DUP
  IF  POWERS-OF-10E-1 F*  THEN
  R@ 4 rshift #15 AND ?DUP
  IF  POWERS-OF-10E-16 F*  THEN
  R> 8 rshift  0
  ?DO  f%10^-256 F*  LOOP
;

: *10^X		\  exp(10) -- ; F: f -- f'
\ *G Generate float' = float *10^dec_exp.
  DUP 0< IF
    SINK_FRACTION
  ELSE
    RAISE_POWER
  THEN
;


\ ***********************
\ *S Floating point input
\ ***********************
\ *P Note that number conversion takes place in *\fo{PAD}.

: CONVERT-EXP	\ c-addr --
\ *G If the character at c-addr is 'D' convert it to 'E'.
  count [char] D scan if
    [char] E swap c!
  else
    drop
  then
;

: CONVERT-FPCHAR	\ c-addr --
\ *G Convert the f.p. char '.' to the double char ',' for
\ ** conversion.
  count [char] . scan if
    [char] , swap c!
  else
    drop
  then
;

: ALL-BLANKS?	\ c-addr len -- flag
\ *G Return true if string is all blanks (spaces).
  true -rot
  bounds ?Do
    i c@ bl <>
    if  drop false leave  then
  Loop
;

[-short-branches
: FCHECK	\ -- am lm ae le e-flag .-flag
\ *G Check the input string at *\fo{PAD}, returning the separated
\ ** mantissa and exponent flags. The e-flag is returned true
\ ** if the string contained an exponent indicator 'E' and
\ ** the .-flag is returned true if a '.' was found.
\ returns NOT FP, if number has base prefix
\ e.g. '$','#','0x' etc.  (Base suffix NOT supported)
  base @ >r                             \ preserve number base
\  pad count 2dup OverrideBase d= if     \ not integer - NOTE: WORKS ONLY FOR PREFIXES
  pad count 2dup check-prefix d= if     \ not integer - NOTE: WORKS ONLY FOR PREFIXES
    pad count bounds ?do
      i c@ [char] e = if  [char] E i c!  then
    loop
    pad count 2dup [char] E scan dup >r	\ ap lp ae le --
    tuck >r >r - r> r>                  \ am lm ae le --
    dup if  1 /string  then             \ am lm ae' le' --
    pad count [char] . scan nip
    r> swap
  else                                  \ integer prefix detected
    0 0 0 0                             \ dummies
    false false                         \ do DP or exponent
  endif
  r> base !                             \ restore number base
;
short-branches]

: doMNUM	\ c-addr u -- d 2 | 0
\ *G Convert the mantissa string to a double number and 2. If
\ ** conversion fails, just return 0.
  (integer?) dup 1 =
  if  drop s>d 2  then
;

: doENUM	\ c-addr u -- n 1 | 0 ; str as above
\ *G Convert the exponent string to a single number and 1. If
\ ** conversion fails, just return 0.
  dup if
    (integer?)  dup 2 =
    if  2drop drop 0  then
  else
    2drop  0 1
  then
;

: FIXEXP     \ dmant exp(10) -- ; F: -- f
\ *G Convert a double integer mantissa and a single integer
\ ** exponent into a floating point number.
  >R d>f R> *10^X
;

[-short-branches
: FNUMBER?	\ addr -- 0 | n 1 | d 2 | -1 ; F: -- [f]
\ *G Behaves like the integer version of *\fo{NUMBER?} except that if
\ ** the number is in F.P. format and *\fo{BASE} is decimal, a floating
\ ** point conversion is attempted. If conversion is successful,
\ ** the floating point number is left on the float stack and
\ ** the result code is 2. This word only accepts text with an
\ ** 'E' as a floating point indicator, e.g, 1.2345e0.
\ ** If *\fo{BASE is not decimal all numbers are treated as integers.
\ ** The integer prefixes '#','$','0x' etc. are recognised and
\ ** cause integer conversion to be used.
  count pad place			\ copy string to pad
  pad count upper			\ force upper case
  fcheck drop                           \ valid f.p. number?
  BASE @ #10 = and if                   \ and base is DECIMAL
    pad convert-fpchar                  \   convert f.p. char ready for conversion to a double
    doEnum if                           \   convert exponent ok?
      -rot  over c@ [char] , <> ?undef  \     must be at least one digit for d.p.
      doMnum if                         \     convert mantissa ok?
        rot dpl @  dup 0>
	if  -  else  drop  then
        fixexp  -1
      else
        drop 0
      then
    else
      2drop 0
    then
    dup #l !  -1 dpl !
  else
    2drop 2drop  pad integer?           \ try to convert as integer
  then
;

: >FLOAT	\ c-addr u -- true|false ; F: -- [f]
\ *G Try to convert the string at c-addr/u to a floating point number.
\ ** If conversion is successful, flag is returned true, and a floating
\ ** number is returned on the float stack, otherwise just flag=0 is
\ ** returned. This word accepts several forms, e.g. 1.2345e0, 1.2345,
\ ** 12345 and converts them to a float. Note that double numbers
\ ** (containing a ',') cannot be converted.
\ ** Number conversion is decimal only, regardless of the current *\fo{BASE}.
  base @ >r decimal			\ SFP007
  pad place			        \ copy string to pad
  pad count upper			\ force upper case
  pad convert-exp                       \ make sure exponent char is an 'E'
  fcheck or if                          \ valid f.p. number?
    pad convert-fpchar                  \   convert f.p. char ready for conversion to a double
    doEnum if                           \   convert exponent ok?
      -rot doMnum if                    \     convert mantissa ok?
        rot dpl @  dup 0>
	if  -  else  drop  then
        fixexp  true
      else
        drop 0
      then
    else
      2drop 0
    then
    dup #l !  -1 dpl !
  else
    2drop				\ discard exponent string
    case  2dup 2>r (integer?)		\   try to convert integer to f.p.
      0 of  2r@ all-blanks? dup		\    all blanks is a special case returning f.p. 0
            if  0 >fs  then
	endof
      1 of  s>f true  endof		\    convert single integer to float
      2 of  d>f true  endof		\    convert double
    endcase
    2r> 2drop
  then
  r> base !				\ SFP007
;
short-branches]

: FLITERAL	\ Comp: F: r -- ; Run: F: -- r
\ *G Compiles a float as a literal into the current definition.
\ ** At execution time, a float is returned. For example,
\ ** *\fo{[ f%PI 2e0 F* ] FLITERAL} will compile 2PI as a floating point
\ ** literal. Note that *\fo{FLITERAL} is immediate.
  ['] flit compileAligned,  f,
; IMMEDIATE

: (F#)  	\ addr -- -1|0 ; F: -- [f]
\ *G The primitive for *\fo{F#} below.
  fnumber?
  case
    -1 of  -1  endof
     0 of  0      endof
     1 of  s>f -1  endof
     2 of  dpl @ -1 <>                  \ see if double
           if  d>f  then                \ convert to fp
           -1
       endof
  endcase
;

: F#  		\ F: -- [f] ; or compiles it (state smart)
\ *G If interpreting, takes text from the input stream and,
\ ** if possible converts it to a f.p. number on the stack.
\ ** Numbers in integer format will be converted to floating-point.
\ ** If compiling, the converted number is compiled.
  bl word (f#) ?undef  state @
  if  postpone fliteral  then
;
IMMEDIATE


\ ************************
\ *S Floating point output
\ ************************
\ *P Note that, because global conversion buffers are used, the
\ ** floating point output routines are not thread-safe.

8 value precision	\ -- u
\ *G Number of significant digits output.

: set-precision		\ u --
\ *G Set the number of significant digits used for output.
  1 max  #18 min  to precision  ;

: exp(10)  	\ F: f -- f ; -- exp[10]
\ *G Generate the power of ten corresponding to the float's power of two.
  fdup f0=
  if  precision exit  then
  exp@ dup 0< >r			\ F: -- f ; -- exp(2) ; R: -- sign
  abs 8651 28738 */mod nip		\ l10(2)= 0.30102999512840141972301482357854 ; SBD010
  r> if  negate 1-  then
;
((
: exp(10)	\ F: f -- f ; -- n ; return exponent base 10
  fdup f0= if
    precision				\ why?
  else
   fdup fabs flog floor f>s
  then
;
))

64 buffer: fopbuff	\ -- addr
\ *G Buffer in which output string is built.
32 buffer: frepbuff	\ -- addr
\ *G Buffer for use as the output of *\fo{REPRESENT}.

f# 0.5e0 fconstant f%ROUNDPOINT

: roundfp	\ F: +f -- +f'
\ *G Add 0.5e(exp-precision-1).
  exp(10) precision - 1-		\ will add 0.5e(exp-precision-1)
  f%roundpoint *10^x f+
;

: REPRESENT	\ F: r -- ; c-addr u -- n flag1 flag2
\ *G Assume that the floating number is of the form +/-0.xxxxEyy.
\ ** Place the significand xxxxx at c-addr with a maximum of u digits.
\ ** Return n the signed integer version of yy. Return flag1 true
\ ** if f is negative, and return flag2 true if the results are
\ ** valid. In this implementation all errors are handled by
\ ** exceptions, and so flag2 is always true.
  2dup [char] 0 fill			\ initialise buffer
  fdup f0=
  if  fdrop 2drop 1 0 -1 exit  then
  fsign >r				\ -- caddr u ; R: -- sgn ; F: -- |f|
  roundfp				\ round by adding 0.5e(exp-precision-1)
  dup exp(10) dup >r - 1-		\ -- caddr u u-exp-1 ; R: -- sgn exp ; F: -- f
  *10^x					\ -- caddr u ; R: -- sgn exp ; F: -- f' ; scale to integer
  f>d <# #S #>				\ -- caddr u caddr2 len2 ; R: -- sgn exp
  rot 2dup - >r				\ -- caddr caddr2 len2 u ; R: -- sgn exp len2-u
  min rot swap move			\ -- ; R: -- sgn exp len2-u
  r> r> + 1+  r>  -1
;

[undefined] append [if]
: append	\ c-addr u $dest --
\ *G Add the string described by C-ADDR U to the counted string at
\ ** $DEST. The strings must not overlap.
  >r
  tuck  r@ count +  swap cmove          \ add source to end
  r> c+!                                \ add length to count
;
[then]

: (.sign)	\ flag $out --
\ *G Add '-' or nothing to the output string.
  swap if  s" -" 2 pick append  then  drop
;

: (.mant)	\ binp $out n --
\ *G Add the mantissa string at *\i{binp), produced by *\fo{REPRESENT},
\ ** to a counted string at *\i{$out) with *\i{n} digits before the
\ ** decimal point.
  { binp $out n | u -- }
  precision n min to u
  binp u $out append  s" ." $out append	\ add chars before '.' and '.'
  binp u + precision u - $out append	\ add chars after '.'
;

: (.exp)	\ exp(10) $out --
\ *G Add the exponent to the output string.
  >r
  s" e" r@ append
  dup 0< if  s" -"  else  s" +"  then r@ append
  abs s>d <# # #s #> r> append		\ at least two digits
;

: (.initfop)	\ f -- ; -- exp(10)
\ *G initialise output conversion.
  0 fopbuff c!				\ zero string count
  frepbuff precision represent drop	\ -- n flag1 ; convert
  fopbuff (.sign)			\ -- n ; sign
;

: (fs.)		\ F: f -- ; -- caddr len
\ *G Produce a string containing the number in scientific notation.
  (.initfop)				\ initialise output conversion
  frepbuff fopbuff 1 (.mant)		\ convert mantissa
  1- fopbuff (.exp)			\ convert exponent
  fopbuff count
;

: (fe.)		\ F: f -- ; -- caddr len
\ *G Produce a string containing the number in engineering notation.
  (.initfop)				\ initialise output conversion
  1- s>d 3 fm/mod 3 * swap 1+ >r	\ -- exp' ; R: -- #leading
  frepbuff fopbuff r> (.mant)		\ convert mantissa
  fopbuff (.exp)
  fopbuff count
;

: ff?		\ f: f -- f ; -- flag
\ *G Return true if the number can be represented in free format.
  fdup f0=  exp(10) 0 precision within  or
;

: (ff.)		\ F: f -- ; -- caddr len
\ *G Produce a string containing the number in free notation.
\ ** If the number cannot be displayed in free notation,
\ ** scientific notation is uesed.
  ff? 0=
  if  (fs.) exit  then
  (.initfop)				\ initialise output conversion
  >r frepbuff fopbuff r> (.mant)	\ convert mantissa
  fopbuff count
;

: fs.		\ F: f --
\ *G Display *\i{f} in scientific notation:
\ *C   x.xxxxxE[-]yy
  (fs.) type space
;

: fe.		\ F: f --
\ *G Display *\i{f} in engineering notation:
\ *C   x.xxxxxE[-]yy
\ *P where the mantissa is 1 <= mantissa < 1000 and the exponent
\ ** is a multiple of three.
  (fe.) type space
;

: ff.		\ F: f --
\ *G Display *\i{f} in free notation:
\ *C   xxx.xxxxx
  (ff.) type space
;

: F.  		\ F: f --
\ *G Print the f.p. number in free format, xxxx.yyyy, if
\ ** possible. Otherwise display using the x.xxxxEyy format.
  ff?
  if  ff.  else  fs.  then
;


\ ***********
\ *S Rounding
\ ***********

f# 1.0e0 fconstant f%ONE
\ *G Floating point 1.0.
f# 2.0e0 fconstant f%two
\ *G Floating point 2.0.

: 1/f		\ F: f1 -- f2
\ *G Reciprocal; *\i{f2=1/f1}.
  f%one fswap f/  ;
: f2/		\ F: f1 -- f2
\ *G Divide by 2.0; *\i{f2=f1/2.0}.
  f%two f/  ;

[-short-branches
: FLOOR  	\ F: r1 -- r2
\ *G Floored round towards -infinity.
  fdup fint                             \ input, input rounded to zero
  fover f0> if				\ input was positive?
    fnip                                \   positive integral fine
  else
    fswap fover f- f0<>                 \   negative, but input not an integral?
    if  f%one f-  then                   \     round towards -ve infinity
  then
;

: FROUND  	\ F: r1 -- r2
\ *G Round the number to nearest or even.
  fdup fint                             \ input, input rounded to zero
  fover fover f- fabs f%roundpoint f< if	\ input fractional part closer to integral?
    fnip                                \   just return integral
  else
    fswap f0> if                        \   input positive?
      f%one                              \     round towards +ve infinity
    else
      f%one fnegate                      \     round towards -ve infinity
    then
    f+
  then
;
short-branches]


\ *************************
\ *S Trigonmetric functions
\ *************************
\ *P N.B. All angles are in radians.

((
\ XC mant/exp     IEEE64
(  1518500249 0 ) $6640:0000 $3FE6:A09E >fs fconstant f%SQR.5	\ 0.70710678
(  1686629716 2 ) $5500:0000 $4009:21FB >fs fconstant f%PI	\ 3.14159266
(  1686629715 1 ) $54C0:0000 $3FF9:21FB >fs fconstant f%PI/2	\ 1.57079633
(  1686629716 0 ) $5500:0000 $3FE9:21FB >fs fconstant f%PI/4	\ 0.78539816
( 1367130550 -1 ) $6D80:0000 $3FD4:5F30 >fs fconstant f%1/PI	\ 0.31830988
(  1367130550 0 ) $6D80:0000 $3FE4:5F30 >fs fconstant f%2/PI	\ 0.63661977
                                    180 s>f fconstant f%180
( 1686626304 18 ) $0000:0000 $4109:21F8 >fs fconstant f%SINMAX	\ 205887
(  1677721600 8 ) $0000:0000 $4069:0000 >fs fconstant f%ALNMAX	\ 200
(  1549082004 1 ) $6500:0000 $3FF7:1547 >fs fconstant f%1/LN2	\ 1.44269504
(  1488522236 0 ) $FF00:0000 $3FE6:2E42 >fs fconstant f%LN2	\ 0.6931472
(  1660944384 7 ) $0000:0000 $4058:C000 >fs fconstant f%ALGMAX	\ 99
(  1783446570 2 ) $0A80:0000 $400A:934F >fs fconstant f%1/LG2	\ 3.3219281
( 1292913983 -1 ) $4FC0:0000 $3FD3:4413 >fs fconstant f%LG2	\ 0.30102999
))

0.5e0 fsqrt     fconstant f%SQR.5
%pi             fconstant f%pi
%pi/2           fconstant f%pi/2
%pi/4           fconstant f%pi/4
1e0 %pi f/      fconstant f%1/pi
2e0 %pi f/      fconstant f%2/pi
180 s>f         fconstant f%180
205887 s>f      fconstant f%sinmax
200 s>f         fconstant f%alnmax
1e0 2e0 fln f/  fconstant f%1/ln2
2e0 fln         fconstant f%ln2
1e0 2e0 flog f/ fconstant f%1/lg2
2e0 flg         fconstant f%lg2

: DEG>RAD	\ F: n1 -- n2
\ *G Convert degrees to radians.
  f%180 f/ f%pi f*
;

: RAD>DEG	\ F: n1 -- n2
\ *G convert radians to degrees.
  f%180 f* f%pi f/
;

: ?TR-ERR	\ x --
  abort" Overflow in trig. function"
;

f# -1.6666666E-1
f# 8.3333315E-3
f# -1.98409E-4
f# 2.7526E-6
f# -2.39E-8
5 farray SINARRAY

: (SIN)  	\ F: f1 -- f2
  fdup fdup f* fdup 0 sinarray f*
  4 1
  do  i sinarray f+ fover f*  loop
  4 sinarray f+ f* f%1 f+ f*
;

: SINFN		\ F: f1 -- f2
  fdup f%1/pi f*				\ ang ang/pi --
  fint fdup f>s >r			\ ang int[ang/pi] --
  f%pi f* f-				\ ang-pi*int[ang/pi] --
  fdup f%pi/2 f>
  if  f%pi fswap f-  then
  (sin)
  r> 1 and
  if  fnegate  then
;

: (FSIN)	\ F: f1 -- f2
  fsign  >r
  fdup f%sinmax f> ?tr-err
  sinfn
  r> ?fnegate
;

: FSIN		\ F: f1 -- f2
\ *G f2=sin(f1).
  (FSIN)
;

: FCOS		\ F: f1 -- f2
\ *G f2=cos(f1).
  f%pi/2 f+  (fsin)
;

f# 3.3333140E-1
f# 1.3339240E-1
f# 5.3374060E-2
f# 2.4565089E-2
f# 2.9005250E-3
f# 9.5168091E-3
6 farray TANARRAY

: (TAN)		\ F: f1 -- f2
  fdup fdup f* fdup 0 tanarray f*
  5 1
  do  i tanarray f+ fover f*  loop
  5 tanarray f+ f* f%1 f+ f*
;

[-short-branches
: FTAN		\ F: f1 -- f2
\ *G f2=tan(f1).
  fsign >r
  fdup f%sinmax f> ?tr-err
  fdup f%2/pi f* fint fswap fover f%pi/2 f* f-
  fdup f%pi/4 f> if
    f%pi/2 f- (tan) fswap f%1 f+
  else
    (tan) fswap
  then
  f>s 1 and
  if  f%1 fswap f/ fnegate  then
  r>
  if  fnegate  then
;
short-branches]

f# 1.5707963E0
f# -2.1459880E-1
f# 8.8978987E-2
f# -5.0174305E-2
f# 3.0891881E-2
f# -1.7088126E-2
f# 6.6700901E-3
f# -1.2624911E-3
8 farray ASINARRAY

: ARCFN  	\ F: f1 -- f2
  fdup 0 asinarray f*
  7 1
  do  i asinarray f+ fover f*  loop
  7 asinarray f+
  fswap f%1 fswap f- fsqrt f*
;

f# -3.3333145E-1
f# 1.9993551E-1
f# -1.4208899E-1
f# 1.0656264E-1
f# -7.5289640E-2
f# 4.2909614E-2
f# -1.6165737E-2
f# 2.8662257E-3
8 farray ATANARRAY

: ATFN		\ F: f1 -- f2
  fdup fdup f* fdup 0 atanarray f*
  7 1
  do  i atanarray f+ fover f*  loop
  7 atanarray f+ f* f%1 f+ f*
;

: FASIN  	\ F: f1 -- f2
\ *G f2=arcsin(f1).
  fsign >r
  arcfn f%pi/2 fswap f-
  r> if  fnegate  then
;

: FACOS  	\ F: f1 -- f2
\ *G f2=arccos(f1).
  fsign >r
  arcfn
  r> if  f%pi fswap f-  then
;

[-short-branches
: FATAN  	\ F: f1 -- f2
\ *G f2=arctan(f1).
  fsign >r
  fdup f%1 f> if
    f%1 fswap f/  atfn f%pi/2 fswap f-
  else
    atfn
  then
  r> if  fnegate  then
;
short-branches]


\ ************************
\ *S Logarithms and Powers
\ ************************

f# 9.9999642E-1
f# -4.9987412E-1
f# 3.3179903E-1
f# -2.4073381E-1
f# 1.6765407E-1
f# -9.5329390E-2
f# 3.6088494E-2
f# -6.4535442E-3
8 farray LOGARRAY

(  1488522234 0 ) $FE80:0000 $3FE6:2E42 >fs fconstant f%D8		\ 0.6931472
( 1865280597 -1 ) $1540:0000 $3FDB:CB7B >fs fconstant f%D9		\ 0.43429448

: FLN  		\ F: f1 -- f2
\ *G Take the logarithm of f1 to base e and return the result.
  fdup f0> 0=
  abort" Invalid argument to FLN/FLOG"
  exp@ 1- >r  1 exp!
  f%1 f-
  fdup 0 logarray f*
  7 1 do
    i logarray f+ fover f*
  loop
  7 logarray f+ f*  r> s>f  f%d8 f* f+
;

: FLOG  	\ F: f1 -- f2
\ *G Take the logarithm of f1 to base 10 and return the result.
  fln f%d9 f*
;

f# 1.1512928E0
f# 6.6273088E-1
f# 2.5439357E-1
f# 7.2951737E-2
f# 1.7421120E-2
f# 2.5549180E-3
f# 9.3264267E-4
7 farray 10ARRAY

: (10^X)  	\ F: f1 -- f2
  fdup 0 10array f*
  6 1 do
    i 10array f+ fover f*
  loop
  6 10array f+ f*
  f%1 f+  fdup f*
;

f# -1.0000000E0
f# 4.9999992E-1
f# -1.6666530E-1
f# 4.1657348E-2
f# -8.3013598E-3
f# 1.3298820E-3
f# -1.4131610E-4
7 farray E-ARRAY

: (E^X)  	\ F: f1 -- f2
  fdup 0 e-array f*
  6 1 do
    i e-array f+ fover f*
  loop
  6 e-array f+ f*
  f%1 f+ f%1 fswap  f/
;

: FEXP  	\ F: f1 -- f2
\ *G f2=e^f1.
  fsign >r
  fdup f%alnmax f> abort" Overflow in FEXP"
  fdup  f%1/ln2 f*
  fdup f>s >r
  fint f%ln2 f* f-
  (e^x)
  r> exp@ + exp!			\ add to exponent
  r> if  f%1 fswap f/  then
;

Synonym FE^X FEXP  	\ F: f1 -- f2
\ *G Compatibility word.

f# 1.0e-5 fconstant f%minexpm1

: fexpm1	\ r1 -- r2
\ *G Raise e to the power *\i{r1} and subtract one, giving *\i{r2}.
  fdup fabs f%minexpm1 f<
  if  fdup fdup f* f%two f/ f+  else  fexp f%one f-  then
;

: F10^X  	\ F: f1 -- f2
\ *G f2=10^f1
  fsign >r
  fdup f%algmax f> abort" overflow in f10^x"
  fdup f%1/lg2 f*
  fdup f>s >r
  fint f%lg2 f* f-
  (10^x)
  r> exp@ + exp!			\ add to exponent
  r> if  f%1 fswap f/  then
;

[-short-branches
: FX^N  	\ n -- ; F: fx -- fx^n
\ *G fx^n=x^n where x is a float and n is an integer.
  dup 0= if				\ check for n=0, result=1.0
    drop fdrop  f%1  exit
  then
  dup >r abs				\ -- |n| ; R: -- n
  f%1 fswap
  begin					\ F: -- fx fconst ; -- |n| ; R: -- n
    dup 0>
  while
    dup 1 rshift dup 1 lshift rot -	\ F: -- fx fconst ; |n|/2 |n|'-|n|
    if
      fswap fover f* fswap
    then
    fdup f*
  repeat
  drop fdrop
  r> 0<
  if  f%1 fswap f/  then
;

: F**  	\ F: fx fy -- fx^fy
\ *G fn=X^Y where Y and Y are both floats.
  fdup f0= if
    fdrop fdrop  f%1  exit
  then
  fdup fdup fint f- f>r
  f>s >r fdup r>
  fx^n fswap
  fr> fdup f0<> if
    fover f0< abort" Result of F** is complex"
    fswap fdup f0< >r
    fabs flog f*  f10^x
    r> if  f%1 fswap f/  then
    f*
  else
    fdrop fdrop
  then
;
short-branches]

Synonym FX^Y F**	\ --
\ *G Compatibility word for old code.


\ **********************************
\ *S COSEC SEC COTAN and hyberbolics
\ **********************************

: fcosec        \ F: f -- cosec(f)
( *G Floating point cosecant. )
  fsin 1/f  ;

: fsec          \ F: f -- sec(f)
( *G Floating point secant. )
  fcos 1/f  ;

: fcotan        \ f: f -- cot(f)
( *G Floating point cotangent. )
  ftan 1/f  ;

: fsinh         \ F: f -- sinh(f) ; (e^x - 1/e^x)/2
( *G Floating point hyberbolic sine. )
  fexp  fdup 1/f f-  f2/  ;

: fcosh         \ F: f -- cosh(f) ; (e^x + 1/e^x)/2
( *G Floating point hyberbolic cosine. )
  fexp  fdup 1/f f+  f2/  ;

: ftanh         \ F: f -- tanh(f) ; (e^x - 1/e^x)/(e^x + 1/e^x)
( *G Floating point hyberbolic tangent. )
  fexp fdup 1/f  fover fover f-		\ GEK029
  frot frot f+  f/
;

: fasinh	\ F: f -- asinh(f) ; ln( f+sqrt(1+f*f) )
( *G Floating point hyberbolic arcsine. )
  fdup
  fdup f*  f%1 f+  fsqrt
  f+  fln  ;

: facosh	\ F: f -- acosh(f) ; ln( f+sqrt(f*f-1) )
( *G Floating point hyberbolic arccosine. )
  fdup
  fdup f*  f%1 f-  fsqrt
  f+  fln  ;

: fatanh	\ F: f -- atanh(f) ; ln( (1+f)/(1-f) )/2
( *G Floating point hyberbolic arctangent. )
  f%1 fover f+  f%1 frot f-  f/
  fln  f2/  ;


\ **********************************
\ *S Plugging floats into the system
\ **********************************

: reals		\ -- ; turn FP system on
\ *G Switch the system and *\fo{NUMBER?} to permit floating point
\ ** input using *\fo{FNUMBER?}. This action can be reversed by
\ ** *\fo{INTEGERS}.
  assign fnumber? to-do number?
;

: integers	\ -- ; turn FP system off
\ *G Switch the system and *\fo{NUMBER?} to restore integer-only
\ ** input.
  assign integer? to-do number?
;

DeferPrompt? [if]
: .fSysPrompt	\ --
  .sysprompt  fdepth ?dup
  if  ." +" .  then
; assign .fSysPrompt to-do .prompt
[then]


\ **********
\ *S Gotchas
\ **********

\ ===========================
\ *N Floating point constants
\ ===========================
\ *P As of January 2019, target floating point constants are defined as
\ ** *\fo{f%name} rather than the previous *\fo{%name}. The change
\ ** prevents name conflicts with binary integers that use a '%'
\ ** prefix.

\ =================
\ *N Number formats
\ =================
\ *P The ANS and Forth-2012 specifications define the format of
\ ** floating point numbers during text interpretation as:
\ *E Convertible string := <significand><exponent>
\ **
\ ** <significand> := [<sign>]<digits>[.<digits0>]
\ ** <exponent>    := E[<sign>]<digits0>
\ ** <sign>        := { + | - }
\ ** <digits>      := <digit><digits0>
\ ** <digits0>     := <digit>*
\ ** <digit>       := { 0 | 1 | 2 | 3 | 4 | 5 | 6 | 7 | 8 | 9 }

\ *P This format is handled by the word *\fo{FNUMBER?}. The word
\ ** *\fo{>FLOAT} accepts a more relaxed format.
\ *E Convertible string := <significand>[<exponent>]
\ **
\ ** <significand> := [<sign>]{<digits>[.<digits0>] | .<digits> }
\ ** <exponent>    := <marker><digits0>
\ ** <marker>      := {<e-form> | <sign-form>}
\ ** <e-form>      := <e-char>[<sign-form>]
\ ** <sign-form>   := { + | - }
\ ** <e-char>      := { D | d | E | e }

\ *P This restriction makes it difficult to use the text interpreter
\ ** during program execution as it requires floating point numbers
\ ** to contain 'D' or 'E' indicators, which is not profane practice.
\ ** A quick kluge to fix this is to change *\fo{FNUMBER?} as below.
\ *E Replace:
\ **   fcheck drop if                       \ valid f.p. number?
\ ** with:
\ **   fcheck or if                         \ valid f.p. number?
\ *P Note that this change can/will cause problems if number base
\ ** is not *\fo{DECIMAL}.

\ ==========================
\ *N Floating point literals
\ ==========================
\ *P Because we still have to support a variety of legacy floating
\ ** point packages, plus new ones for as yet undefined CPUs, the
\ ** handling of F.P. literals is far from perfect.

\ *P Our recommendation for use of floating point literals in both
\ ** cross-compiled and target code, is to use *\fo{F#} and/or
\ ** *\fo{FLITERAL}, e.g to compile 2.0e0, use one of the following
\ *E : foo  ... f# 2.0e0 ... ;
\ ** : foo  ... [ 2 s>f ] fliteral ... ;


\ *********
\ Test code
\ *********

0 [if]

: tf1   f# 123.0e0  ;
: tf2   [ 123 s>f ] fliteral ;

[then]


\ ======
\ *> ###
\ ======

