\ IAP11xx.fth - LPC11xx IAP routines

((
Copyright (c) 2004, 2010, 2011, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20151129 MPE004 Lite system conversion
20110617 MPE003 LPC11xx conversion.
20100125 MPE002 LPC17xx conversion.
20040809 MPE001 Released for use with LPC2xxx bootloaders
                v1.52/1.63.
))

decimal

\ ======
\ *! flashlpc11xx
\ *T NXP LPC11xx Flash routines
\ ======


\ *******************
\ *S Flash primitives
\ *******************
\ *P The IAP is accessed by calling a Thumb routine at the
\ ** IAPentry address with the address of a command block in R0
\ ** and the address of a status/result block (RAM) in R1.
\ ** All the IAPxxx words return a 0 result on success. Note
\ ** also that all interrupts are disabled for the duration
\ ** of an IAP call, and therefore that ticker interrupts will
\ ** not be serviced. In particular, the sector erase command
\ ** may take 400ms, and a write of a 512 byte line may take 1ms.

[undefined] FullIAP? [if]
0 equ FullIAP?		\ -- n
\ *G If this *\fo{EQU}ate is set non-zero, additional IAP routines
\ ** are compiled, e.g. to get the bootloader version number
\ ** and the device part number. The definition here is only
\ ** used if it has not been previously defined.
[then]

internal

5 cells buffer: IAPcmd	\ -- addr ; max 5 cells
\ *G Command input buffer for IAP routines.
4 cells buffer: IAPres	\ -- addr ; max 4 cells
\ *G Result output buffer from IAP routines.

code IAP	\ *cmd *res --
\ *G The primitive to call the IAP routines. Interrupts are
\ ** disabled for the duration of the IAP call.
\ save lr
  push    { lr }
\ Disable interrupts
  mrs     r2, PRIMASK			\ save and disable interrupts
  push    { r2 }
  cps     .id .i
\ save high regs
  push    { lr }
  mov     r0, r8
  mov     r1, r9
  mov     r2, r10
  mov     r3, r11
  mov     r4, r12
  push    { r0-r4 }			\ save state
\ Set up parameters and call the IAP Thumb routine
  ldmia   psp ! { r0 }			\ R0 = *cmd
  mov     r1, tos			\ R1 = *res
  push    { r6 }
  ldr     r3, ^IAP			\ get entry address
  blx     r3
  pop    { r6 }
  ldmia   psp ! { tos }			\ restore TOS
\ restore high regs
  pop     { r0-r4 }			\ restore state
  mov     r8, r0
  mov     r9, r1
  mov     r10, r2
  mov     r11, r3
  mov     r12, r4
  pop     { r0 }
  mov     lr, r0
\ restore interrupts
  pop     { r2 }
  msr     PRIMASK r2
\ restore lr/pc
  pop     { pc }
end-code

align l: ^IAP
  $1FFF:1FF1 ,				\ entry address, Thumb mode

prog-speed #1000 / equ cclk-khz		\ system speed in kHz

: IAPprep	\ start end -- res
\ *G Prepare sectors for erase/write.
  IAPcmd
  #50 over !				\ -- start end addr ; cmd#
  rot over cell + !			\ -- end addr ; start
  tuck  2 cells + !			\ -- addr ; end
  IAPres IAP  IAPres @
;

: IAPcopy	\ Rsrc Fdest len -- res
\ *G Copy/program len bytes from Rsrc in RAM to Fdest in Flash.
  IAPcmd
  #51 over !				\ cmd
  tuck  3 cells + !			\ len
  tuck  cell + !			\ Fdest
  tuck  2 cells + !			\ Rsrc
  cclk-khz over 4 cells + !		\ -- addr ; cclk/1000
  IAPres IAP  IAPres @
;

: IAPerase	\ start end -- res
\ *G Erase the inclusive range of sectors.
  IAPcmd
  #52 over !				\ cmd
  tuck  2 cells + !			\ end
  tuck  cell + !			\ start
  cclk-khz over 3 cells + !		\ cclk/1000
  IAPres IAP  IAPres @
;

FullIAP? [if]

: IAPcheck	\ start end -- res
\ *G Blank check the inclusive range of sectors.
  IAPcmd
  #53 over !				\ cmd
  tuck  2 cells + !			\ end
  tuck  cell + !			\ start
  IAPres IAP  IAPres @
;

: IAPcompare	\ Rsrc Fdest len -- res
\ *G Compare len bytes from Rsrc in RAM to Fdest in Flash.
  IAPcmd
  #56 over !				\ cmd
  tuck  3 cells + !			\ len
  tuck  cell + !			\ Fdest
  tuck  2 cells + !			\ Rsrc
  IAPres IAP  IAPres @
;

: IAPbootver	\ -- bootver
\ *G Return the 16 bit boot code version. The high byte is the
\ ** major version and the low byte is the minor version.
  IAPcmd
  #55 over !
  IAPres IAP  IAPres cell + w@
;

: IAPpartno	\ -- part#
\ *G Return the NXP part number.
  IAPcmd
  #54 over !
  IAPres IAP  IAPres cell + @
;

: IAPserialno	\ -- addr
\ *G Return a pointer to the 128 bit (4 cells) NXP LPC1xxx serial
\ ** number.
  IAPcmd
  #58 over !
  IAPres IAP  IAPres
;

[then]

external


\ **************
\ *S Flash tools
\ **************

((
cell buffer: #PrgErrs	\ --
  \ *G Holds number of programming errors.
))

internal

compiler
: FindSecN	\ addr -- sec#
\ *G Find the sector number containing address addr.
  #12 rshift				\ 4 kb sectors
;
((
: FindPageN	\ addr -- page#
  \ *G Find the sector number containing address addr.
  #6 rshift				\ 64 byte pages
;
))
target

: Prog256	\ src dest --
\ +G Program 256 bytes at *\i{src} to *\i{dest}. Increment error
\ +* counter on error. *\i{Dest} must be 256 byte aligned.
  #10 0 do
    dup FindSecN dup IAPprep 0= if	\ prep sector
      2dup #256 IAPcopy 0=		\ write data
      if  2drop  unloop exit  endif
    endif
  loop
  2drop  ( #PrgErrs incr )
;

external

: FlErase	\ addr len --
\ *G Erase the Flash sectors in the given range.
  over + 1-				\ -- startaddr endaddr
  swap FindSecN  swap FindSecN		\ -- secnf secnl
  10 0 do
    2dup IAPprep 0= if
      2dup IAPerase 0=
      if  2drop  unloop exit  endif
    endif
  loop
  2drop  ( #PrgErrs incr )
;

: FlProg	\ src dest len --
\ *G Program the Flash at *\i{dest} from RAM at *\i{src}.
\ ** *\i{Dest} must be 256 byte aligned.
  #255 + #-256 and  bounds ?do
    dup i prog256  256 +
  256 +loop
  drop
;


\ *********************
\ *S Code buffer in RAM
\ *********************
\ *P Many ARM chips cannot program their own Flash on a byte by
\ ** byte basis. Some chips will only program a single 32 bit
\ ** word set to $FFFF:FFFF. To avoid all the unpleasantness and
\ ** compromises that this enforces, code is compiled into RAM
\ ** and then copied to Flash. In most cases, this is all hidden
\ ** from the programmer, but in some cases you do need to be aware
\ ** of what is happening.

[undefined] /CodeRamBuff [if]
#256 equ /CodeRamBuff	\ -- len
\ *G Default size of the code buffer in RAM. This is also the
\ ** maximum size of a definition.
[then]
/CodeRamBuff buffer: CodeRamBuff	\ -- addr
\ *G Buffer used during compilation.

internal
0 value CodeOffset	\ -- n
\ +G During compilation to a RAM buffer, the offset between the
\ +* Flash and the buffer is stored here.

: CodeHere	\ -- addr
\ +G The value of *\fo{HERE} when RAM buffer compilation started.
  CodeRamBuff CodeOffset -
;

: CodeLen	\ -- len
\ +G The amount of code compiled since RAM buffer compilation started.
  here CodeHere -
;

((
: .crb		\ --
\ +G Display the code RAM buff data
  cr ." CRB = " CodeRamBuff .dword
  cr ." CO  = " CodeOffset .dword
  cr ." CH  = " CodeHere .dword
  cr ." CL  = " CodeLen .dword
;
))

: FlashAlign	\ --
\ +G Align the Flash as required for programming.
  dp @ #15 + #-16 and dp !
;

#256 buffer: FlashBuff	\ -- addr
\ +G Buffer used to program partial sectors.

: code>flash	\ --
\ +G Copy code to Flash. LPC11xx Flash must be written in blocks
\ +* of at least 256 bytes. Bytes of $FF cause no change to existing
\ +* Flash data. However, Flash error correction (ECC) require
\ +* changes to be applied in groups of 16 bytes. See the LPC111x
\ +* User Manual, section 26.3.7.
  CodeHere $FF and 0= if		\ on 256 byte boundary
    CodeRamBuff CodeHere CodeLen FlProg
    exit
  then
\ part 1
\ .crb
  FlashBuff #256 $FF fill		\ only non-FF bytes are programmed
  CodeRamBuff				\ source
  FlashBuff CodeHere $FF and +		\ dest - offset into buffer
  CodeLen #256 CodeHere $FF and - min	\ len is min of 256 and residue of 256 byte page
  dup >r cmove
  FlashBuff  CodeHere #-256 and  #256 FlProg
\ part 2
  CodeLen r@ - if			\ more to do
    FlashBuff #256 $FF fill		\ only non-FF bytes are programmed
    CodeRamBuff r@ +			\ start of next part of source
    Flashbuff				\ buffer is dest
    CodeLen r@ - cmove			\ copy
    FlashBuff				\ src  = buffer
    CodeHere #255 + #-256 and		\ dest = next page
    #256 FlProg
  then
  r> drop
;
external

: +BuffComp	\ --
\ *G Start compiling into the RAM buffer. *\fo{HERE} still points
\ ** into Flash. This word is already used inside words such as
\ ** *\fo{:} and *\fo{VARIABLE}.
  CodeOffset 0= if
    FlashAlign
    CodeRamBuff /CodeRambuff $FF fill
    CodeRamBuff here - to CodeOffset
  endif
;

: -BuffComp	\ --
\ *G Stop compiling into the RAM buffer and program the RAM buffer
\ ** to Flash. This word is already inside *\fo{;} and *\fo{VARIABLE}.
  CodeOffset if
    FlashAlign
    code>flash
    0 to CodeOffset
  endif
;


\ *********************************
\ *S Reading and writing code space
\ *********************************
\ *P When compiling into Flash from a code buffer, we use a simple
\ ** set of words to access the area being accessed.

internal
: RAM/Flash	\ addr -- addr'
\ +G If required adjust the address from Flash to the RAM buffer.
  CodeOffset if
    dup CodeHere dup /CodeRambuff + within
    if  CodeOffset +  endif
  endif
;

: ?NotFlash	\ addr -- addr
\ +G Error if the address is in the Flash area.
  dup FLASHSTART FLASHEND within	\ Flash range
  abort" 8/16/32 bit Flash write illegal"
;
external

: c@c		\ addr -- b
\ *G 8 bit fetch from Flash or the RAM buffer if active
  RAM/Flash c@
;

: w@c		\ addr -- w
\ *G 16 bit fetch from Flash or the RAM buffer if active
  RAM/Flash w@
;

: @c		\ addr -- x
\ *G 32 bit fetch from Flash or the RAM buffer if active
  RAM/Flash @
;

: c!c		\ b addr --
\ *G 8 bit store to Flash or the RAM buffer if active
  RAM/Flash ?NotFlash c!
;

: w!c		\ w addr --
\ *G 16 bit store to Flash or the RAM buffer if active
  RAM/Flash ?NotFlash w!
;

: !c		\ b addr --
\ *G 32 bit store to Flash or the RAM buffer if active
  RAM/Flash ?NotFlash !
;

internal
: >Info		\ src dest len --
\ +G Save an application or kernel memory information block to
\ +* Flash. The address *\i{src} must be aligned. The address
\ +* *\i{dest} must be the start address of empty (erased) Flash.
\ +* The kernel CRC is added at the end.
  3dup s= if				\ same as before?
    2dup + @ crcslot @ =		\ CRC matches?
    if  3drop exit  endif		\ nothing to do
  endif
  over @ $FFFF:FFFF <> 			\ if Flash not empty
  if  2dup FlErase  endif		\ erase it
  swap >r  dup >r			\ -- src len ; R: -- dest len
  CodeRamBuff swap cmove		\ -- ; copy to RAM buffer
  crcslot @ CodeRamBuff r@ + !		\ add crc to RAM buffer
  CodeRamBuff r> r> swap  cell +	\ -- src' dest len'
  FlProg
;
external


\ ======
\ *> ###
\ ======

decimal

