\ FlashSTM32F4.fth - Flash programming for STM32F4

((
Copyright (c) 2011, 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20140710 SFP002 update for RAM Buffer.
20111205 SFP001 First release.
))

only forth definitions
decimal

\ ===========
\ *! flashstm32f4
\ *T STM32F4 Flash programming
\ ===========
\ *P When Flash is being written or erased, any read access will
\ ** stall the bus until the Flash operation has finished. Thus,
\ ** providing that you do not erase the code in *\i{FlashProg.fth}
\ ** itself, the Flash programming code can be run from Flash.

\ ****************
\ *S Sector tables
\ ****************
\ *P Sector tables contain the number of sectors and starting offset
\ ** of each sector, plus a dummy start address which enables
\ ** the size of the last sector to be calculated.

create sectab	\ -- addr
\ *G Generic STM32F4xx sector table for CPUs with up to 1 Mb of
\ ** Flash. Note that the addresses are offsets from the start of
\ ** the Flash.
  #12 ,
  $0000:0000 ,  $0000:4000 ,  $0000:8000 ,  $0000:C000 ,
  $0001:0000 ,  $0002:0000 ,  $0004:0000 ,  $0006:0000 ,
  $0008:0000 ,  $000A:0000 ,  $000C:0000 ,  $000E:0000 ,
  $0010:0000 ,

: SectorN	\ n -- offset len
\ *G Convert sector number (zero based) to sector offset and length
  1+ cells sectab + 2@			\ -- end start
  tuck -
;

: FindSecN	\ addr -- n
\ *G Find the sector number containing address addr. If
\ ** addr is outside the internal Flash range, n is set to -1.
\  FlashBase -				\ address to offset
  $0800:0000 -				\ address to offset
  sectab @ 0 ?do			\ -- addr
    dup i SectorN over + within if	\ test sector
      drop i  unloop exit
    endif
  loop
  drop -1
;


\ *******************
\ *S Flash primitives
\ *******************
\ *P Although some of these routines are not the most efficient
\ ** for the STM32F4 series, they are designed to be easily expanded
\ ** for future STM32F4 parts with as yet unknown sector sizes.

((
FlashBase constant FlashBase
_FPEC constant _FPEC

: .FPEC		\ --
  _FPEC $24 pdump  ;
))

: ?Unlock	\ --
\ *G Unlock the Flash.
  _FPEC flCR + @ bit31 and if		\ locked?
    _FPEC  flKEY1 over flKEYR + !	\ write keys
    flKEY2 swap flKEYR + !
  endif
;

: waitFlNB	\ -- status
\ *G Wait until the Flash is not busy and return the status.
\ ** The top bit is set if the operation timed out.
  $0100:0000 _FPEC begin		\ -- cnt fpec
    swap 1- swap  over			\ check timeout counter
   while
    dup flSR + @ bit16 and 0=
  until then
  dup flSR + @				\ -- cnt fpec status ; get status
  dup $00F3 and rot flSR + !		\ -- cnt status ; clear h/w status
  swap 0=				\ timeout check
  if  bit31 or  endif
;

: (c!f)		\ w addr -- status
\ *G Write an 8-bit word to Flash. A status value of zero indicates
\ ** success.
  ?Unlock
  $0001 _FPEC flCR + !			\ set PSIZE and PG bits in control reg
  c!					\ write data
  waitFlNB				\ wait for completion/timeout
  bit0 _FPEC flCR + bic!		\ clear PG bit in control reg
;

: (w!f)		\ w addr -- status
\ *G Write a 16-bit word to Flash. A status value of zero indicates
\ ** success.
  ?Unlock
  $0101 _FPEC flCR + !			\ set PSIZE and PG bits in control reg
  w!					\ write data
  waitFlNB				\ wait for completion/timeout
  bit0 _FPEC flCR + bic!		\ clear PG bit in control reg
;

: (!f)		\ x addr -- status
\ *G Write a 32-bit word to Flash. A status value of zero indicates
\ ** success.
  ?Unlock
  $0201 _FPEC flCR + !			\ set PSIZE and PG bits in control reg
  !					\ write data
  waitFlNB				\ wait for completion/timeout
  bit0 _FPEC flCR + bic!		\ clear PG bit in control reg
;

: flSecErase	\ u -- status
\ *G Erase the sector *\i{u}. A status value of zero indicates success.
  ?Unlock
  $0F and 3 lshift $0202 or _FPEC	\ PSIZE=2 for 32 bit, sec#, SER
  2dup flCR + !				\ set required data
  swap bit16 or swap flCR + !		\ set STRT bit
  waitFlNB				\ wait for completion/timeout
  bit1 _FPEC flCR + bic!		\ clear SER bit in control reg
;

((
: .src/dest	\ src dest -- src dest
 \ *G Display the addresses and contents of the source
 \ ** and destination.
  ."  source: " over .lword  over c@ ." (" .byte ." )"
  ."  dest: " dup .lword  dup c@ ." (" .byte ." )"
;
))

: inFlash?	\ addr -- flag
\ +G Returns true if the address is in the Flash area.
  FLASHSTART FLASHEND within		\ Flash range
;


\ ***************
\ *S Flash driver
\ ***************

external

: c!f		\ b addr --
\ *G Program any address, including Flash, with an 8 bit value.
  dup InFlash?
  if  (c!f) drop  else c!  endif
;

: w!f		\ w addr --
\ *G Program any address, including Flash, with a 16 bit value.
  dup InFlash?
  if  (w!f) drop  else  w!  endif
;

: !f		\ x addr --
\ *G Program any address, including Flash, with a 32 bit value.
  dup InFlash?
  if  (!f) drop  else  !  endif
;

: FlErase	\ addr len --
\ *G Erase the Flash sectors in the given range.
\ ** Note that complete sectors in the range will be erased.
  over + 1-				\ -- first last
  FindSecN swap FindSecN 		\ -- nl nf
  begin
    dup flSecErase 0=
   while
    2dup <>
   while
    1+
  repeat then
  2drop
;

: FlProg	\ src dest len --
\ *G Program the Flash at *\i{dest} from memory at *\i{src}.
  bounds ?do
    dup @ i (!f) drop  cell +
  cell +loop
  drop
;

\ internal
: >Info		\ src dest len --
\ +G Save a memory block to Flash. The address *\i{src}must be
\ +* even. The address *\i{dest} must be the start address of
\ +* empty (erased) Flash. The kernel CRC is added at the end.
  3dup s= if
    2dup + @ crcslot @ = if
      3drop exit
    endif
  endif
  over @ $FFFF:FFFF <> if
    2dup FlErase
  endif
  2dup + crcslot @ swap !f		\ save kernel checksum
  bounds ?do
    dup @ i !f  cell +
  cell +loop
  drop
;
external


\ *********************
\ *S Code buffer in RAM
\ *********************
\ *P Many ARM chips cannot program their own Flash on a byte by
\ ** byte basis. Some chips will only program a single 32 bit
\ ** word set to $FFFF:FFFF. To avoid all the unpleasantness and
\ ** compromises that this enforces, code is compiled into RAM
\ ** and then copied to Flash. In most cases, this is all hidden
\ ** from the programmer, but in some cases you do need to be aware
\ ** of what is happening.

\ SFP007...
[undefined] /CodeRamBuff [if]
#512 equ /CodeRamBuff	\ -- len
\ *G Default size of the code buffer in RAM.
[then]
/CodeRamBuff buffer: CodeRamBuff	\ -- addr
\ *G Buffer used during compilation.

internal
0 value CodeOffset	\ -- n
\ +G During compilation to a RAM buffer, the offset between the
\ +* Flash and the buffer is stored here.

: CodeHere	\ -- addr
\ +G The value of *\fo{HERE} when RAM buffer compilation started.
  CodeRamBuff CodeOffset -
;

: CodeLen	\ -- len
\ +G The amount of code compiled since RAM buffer compilation started.
  here CodeHere -
;
external

: +BuffComp	\ --
\ *G Start compiling into the RAM buffer. *\fo{HERE} still points
\ ** into Flash. This word is already used inside words such as
\ ** *\fo{:} and *\fo{VARIABLE}.
  CodeOffset 0= if
    CodeRamBuff 512 $FF fill
    CodeRamBuff here - to CodeOffset
  endif
;

: -BuffComp	\ --
\ *G Stop compiling into the RAM buffer and program the RAM buffer
\ ** to Flash. This word is already inside *\fo{;} and *\fo{VARIABLE}.
  CodeOffset if
    align
    CodeRamBuff CodeHere CodeLen FlProg
    0 to CodeOffset
  endif
;
\ ...SFP007


\ *********************************
\ *S Reading and writing code space
\ *********************************
\ *P When compiling into Flash from a code buffer, we use a simple
\ ** set of words to access the area being accessed.

internal
: RAM/Flash	\ addr -- addr'
  CodeOffset if
    dup CodeHere dup /CodeRambuff + within
    if  CodeOffset +  endif
  endif
;
external

: c@c		\ addr -- b
\ *G 8 bit fetch from Flash or the RAM buffer if active
  RAM/Flash c@
;

: w@c		\ addr -- w
\ *G 16 bit fetch from Flash or the RAM buffer if active
  RAM/Flash w@
;

: @c		\ addr -- x
\ *G 32 bit fetch from Flash or the RAM buffer if active
  RAM/Flash @
;

: c!c		\ b addr --
\ *G 8 bit store to Flash or the RAM buffer if active
  RAM/Flash c!f
;

: w!c		\ w addr --
\ *G 16 bit store to Flash or the RAM buffer if active
  RAM/Flash w!f
;

: !c		\ b addr --
\ *G 32 bit store to Flash or the RAM buffer if active
  RAM/Flash !f
;


\ ======
\ *> ###
\ ======

decimal
