\ FlashXMC11xx.fth - Program XMC11xx Flash in 16 byte units.

((
Copyright (c) 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20151218 SFP001 First release.
))

\ ***************************
\ *! flashxmc11xx
\ *T XMC11xx Flash operations
\ ***************************
\ *P Although these words are not the most efficient for these
\ ** CPUs, they are designed to be easily expanded for future
\ ** parts with as yet unknown sector sizes. The code is for
\ ** Flash with fixed sector and page sizes. The page size
\ ** is defined by *\fo{/FlashPage}.

4 kb equ /FlashSector	\ -- u
\ *G Size of a Flash sector.
#256 equ /FlashPage	\ -- u
\ *G Size of a Flash page.
#16 equ /FlashBlock	\ -- u
\ *G Size of a Flash block.


\ *************
\ *S Primitives
\ *************

internal

: FlAction	\ action --
\ *G Start the given Flash action. INTERNAL.
  _NVM nvm.PROG + !
;

: FlFinish	\ --
\ *G Wait until the current action finishes and clear it. INTERNAL.
  _NVM begin 				\ -- _nvm
    dup nvm.STATUS + @ bit0 and 0=
  until
  0 swap nvm.PROG + !
;

: FlPageErase	\ addr --
\ *G Erase a 256 byte page. INTERNAL.
  $092 FlAction  0 swap !  FlFinish
;

: FlWriteBlock	\ src dest --
\ *G Write a 16 byte block to a 16 byte aligned address. INTERNAL.
  $091 FlAction
  /FlashBlock bounds do  		\ write 16 bytes
    dup @ i !  cell +
  cell +loop
  drop  FlFinish
;

16 buffer: BlockBuffer	\ -- addr
\ *G 16 byte buffer for partial block writes. INTERNAL.

: prepBlock	\ --
\ *G Fill the buffer with FF bytes. INTERNAL.
  BlockBuffer 16 $FF fill
;

: splitAddr	\ addr -- addr' addrb
\ *G Split address into 16 byte aligned address and an address in
\ ** *\fo{BlockBuffer}. INTERNAL.
  dup #-16 and  swap $0F and BlockBuffer +
;

: writeBlock	\ addr16 --
\ *G Write *\fo{BlockBuffer} to a 16 byte aligned address in Flash.
\ ** INTERNAL.
  BlockBuffer swap FlWriteBlock
;

: (c!f)		\ b addr --
\ *G 8 bit Flash write primitive. INTERNAL.
  prepBlock  splitaddr rot swap c!  writeBlock
;

: (w!f)		\ w addr --
\ *G 16 bit Flash write primitive. INTERNAL.
  prepBlock  splitaddr rot swap w!  writeBlock
;

: (!f)		\ x addr --
\ *G 32 bit Flash write primitive. INTERNAL.
  prepBlock  splitaddr rot swap !  writeBlock
;

: inFlash?	\ addr -- flag
\ *G Returns true if the address is in the Flash area. INTERNAL.
  FLASHSTART FLASHEND within		\ Flash range
;

: c!f		\ b addr --
\ *G Program any address, including Flash, with an 8 bit value.
\ ** INTERNAL.
  dup InFlash?
  if  (c!f)  else c!  endif
;

: w!f		\ w addr --
\ *G Program any address, including Flash, with a 16 bit value.
\ ** INTERNAL.
  dup InFlash?
  if  (w!f)  else  w!  endif
;

: !f		\ x addr --
\ *G Program any address, including Flash, with a 32 bit value.
\ ** INTERNAL.
  dup InFlash?
  if  (!f)  else  !  endif
;

external


\ ***********
\ *S User API
\ ***********

: FlErase	\ addr len --
\ *G Erase the Flash sectors in the given range.
  bounds ?do  i flPageErase  /FlashPage +loop
;

: FlProg	\ src dest len --
\ *G Program the Flash at *\i{dest} from memory at *\i{src}.
\ ** *\i{Src} must be 4-byte aligned.
\ ** *\i{Dest} is forced to a 16 byte boundary.
  swap #15 + #-16 and swap
  bounds ?do			\ -- src
    dup i FlWriteBlock  /FlashBlock +
  /FlashBlock +loop
  drop
;


\ *********************
\ *S Code buffer in RAM
\ *********************
\ *P Many ARM chips cannot program their own Flash on a byte by
\ ** byte basis. Some chips will only program a single 32 bit
\ ** word set to $FFFF:FFFF. To avoid all the unpleasantness and
\ ** compromises that this enforces, code is compiled into RAM
\ ** and then copied to Flash. In most cases, this is all hidden
\ ** from the programmer, but in some cases you do need to be aware
\ ** of what is happening.

[undefined] /CodeRamBuff [if]
#256 equ /CodeRamBuff	\ -- len
\ *G Default size of the code buffer in RAM. This is also the
\ ** maximum size of a definition.
[then]
/CodeRamBuff buffer: CodeRamBuff	\ -- addr
\ *G Buffer used during compilation.

internal
0 value CodeOffset	\ -- n
\ +G During compilation to a RAM buffer, the offset between the
\ +* Flash and the buffer is stored here.

: CodeHere	\ -- addr
\ +G The value of *\fo{HERE} when RAM buffer compilation started.
  CodeRamBuff CodeOffset -
;

: CodeLen	\ -- len
\ +G The amount of code compiled since RAM buffer compilation started.
  here CodeHere -
;

((
: .crb		\ --
\ +G Display the code RAM buff data
  cr ." CRB = " CodeRamBuff .dword
  cr ." CO  = " CodeOffset .dword
  cr ." CH  = " CodeHere .dword
  cr ." CL  = " CodeLen .dword
;
))

: FlashAlign	\ --
\ +G Align the Flash as required for programming.
  dp @ #15 + #-16 and dp !
;

: code>flash	\ --
\ +G Copy code to Flash. XMC11xx Flash must be written in blocks
\ +* of at least 16 bytes. Bytes of $FF cause no change to existing
\ +* Flash data. However, Flash error correction (ECC) require
\ +* changes to be applied in groups of 16 bytes. See the XMC1100
\ +* User Manual.
  CodeRamBuff CodeHere CodeLen FlProg
;
external

: +BuffComp	\ --
\ *G Start compiling into the RAM buffer. *\fo{HERE} still points
\ ** into Flash. This word is already used inside words such as
\ ** *\fo{:} and *\fo{VARIABLE}.
  CodeOffset 0= if
    FlashAlign
    CodeRamBuff /CodeRambuff $FF fill
    CodeRamBuff here - to CodeOffset
  endif
;

: -BuffComp	\ --
\ *G Stop compiling into the RAM buffer and program the RAM buffer
\ ** to Flash. This word is already inside *\fo{;} and *\fo{VARIABLE}.
  CodeOffset if
    FlashAlign
    code>flash
    0 to CodeOffset
  endif
;


\ *********************************
\ *S Reading and writing code space
\ *********************************
\ *P When compiling into Flash from a code buffer, we use a simple
\ ** set of words to access the area being accessed.

internal
: RAM/Flash	\ addr -- addr'
\ +G If required adjust the address from Flash to the RAM buffer.
  CodeOffset if
    dup CodeHere dup /CodeRambuff + within
    if  CodeOffset +  endif
  endif
;
external

: c@c		\ addr -- b
\ *G 8 bit fetch from Flash or the RAM buffer if active
  RAM/Flash c@
;

: w@c		\ addr -- w
\ *G 16 bit fetch from Flash or the RAM buffer if active
  RAM/Flash w@
;

: @c		\ addr -- x
\ *G 32 bit fetch from Flash or the RAM buffer if active
  RAM/Flash @
;

: c!c		\ b addr --
\ *G 8 bit store to Flash or the RAM buffer if active
  RAM/Flash c!f
;

: w!c		\ w addr --
\ *G 16 bit store to Flash or the RAM buffer if active
  RAM/Flash w!f
;

: !c		\ b addr --
\ *G 32 bit store to Flash or the RAM buffer if active
  RAM/Flash !f
;

internal
: >Info		\ src dest len --
\ +G Save an application or kernel memory information block to
\ +* Flash. The address *\i{src} must be aligned. The address
\ +* *\i{dest} must be the start address of empty (erased) Flash.
\ +* The kernel CRC is added at the end.
  3dup s= if				\ same as before?
    2dup + @ crcslot @ =		\ CRC matches?
    if  3drop exit  endif		\ nothing to do
  endif
  over @ $FFFF:FFFF <> 			\ if Flash not empty
  if  2dup FlErase  endif		\ erase it
  swap >r  dup >r			\ -- src len ; R: -- dest len
  CodeRamBuff swap cmove		\ -- ; copy to RAM buffer
  crcslot @ CodeRamBuff r@ + !		\ add crc to RAM buffer
  CodeRamBuff r> r> swap  cell +	\ -- src' dest len'
  FlProg
;
external


\ ======
\ *> ###
\ ======

decimal

