\ gpioSTM32F4xx.fth - easy access to STM32F4 GPIO pins.

((
Copyright (c) 2011, 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20120210 SFP003 Corrected ISFUNCTION.
20120130 SFP002 Refactored, added pin speed.
20111201 SFP001 Adapted from Kinetis version.
))

only forth definitions
decimal

\ ================
\ *! gpiostm32f4xx
\ *T STM32F4 GPIO utilities
\ ================
\ *P The code in *\i{Cortex\Drivers\gpioSTM32F4xx.fth} provides
\ ** utility words for accessing GPIO pins on a bit by bit basis
\ ** for the STM32F4 CPUs.
\ ** The code is written for ease of use rather than performance.


\ ********************
\ *S Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number. Bits 0..31 form
\ ** P0.0 to P0.31, bits 32..63 form P1.0 to P1.31 and so on.
\ ** Although ST name their ports from A, here we number
\ ** them from 0. Note also that ST usually only provide 16
\ ** I/O bits per port. Later we define constants for the port
\ ** numbers.

$0400 equ /Port		\ -- u
\ *G Separation in memory between GPIO port base addresses.
#10 equ PortShift	\ -- u
\ *\fo{/Port} as a shift.

target-only
: PIO:		\ port# bit# -- ; -- struct
\ *G Define a port I/O bit by name. For example
\ *C   3 12 pio: LED1   \ PD.12
\ *C   PD 12 pio: LED1  \ PD.12
\ *P defines an I/O bit on GPIOD, bit 12 as *\fo{LED1}. At run
\ ** time, the I/O bit structure is returned.
  create
    over PortShift lshift _GPIOA + ,	\ pins base
    1 over lshift ,			\ bit mask
    swap 5 lshift or ,			\ I/O bit number
;
host&target
interpreter
: PIO:		\ port# bit# -- ; -- struct
  create
    over PortShift lshift _GPIOA + ,	\ pins base
    1 over lshift ,			\ bit mask
    swap 5 lshift or ,			\ I/O bit number
;
target

\ *P Define the ports as constants.
\ *[
0 constant PA
1 constant PB
2 constant PC
3 constant PD
4 constant PE
5 constant PF
6 constant PG
7 constant PH
8 constant PI
9 constant PJ
\ *]


\ ******************
\ *S GPIO pin access
\ ******************

: setPin	\ struct --
\ *G Set the pin high.
  2@ gpioBSRRL + w!  ;

: clrPin	\ struct --
\ *G Set the pin low.
  2@ gpioBSRRH + w!  ;

: getPin	\ struct -- 0/1
\ *G Read the pin state.
  2@ gpioIDR + @ and 0<> 1 and
;


\ ***********************
\ *S IO pin configuration
\ ***********************

: mode>mask	\ mode -- modemask
\ Repeats the two bit mask 16 times.
  3 and
  dup 2 lshift or			\ 4 bits
  dup 4 lshift or			\ 8 bits
  dup 8 lshift or			\ 16 bits
  dup #16 lshift or			\ 32 bits
;

: >2bitmask	\ bit# -- bitmask
\ Converts a bit number 0..15 to a 2 bit mask for a 32 bit word,
  $0F and 2* $03 swap lshift  ;

: mode>valmask	\ mode bit# -- val bitmask
\ Convert a two bit mode and bit number (0..15) into a value and
\ bitmask to use with setMask.
  swap mode>mask 			\ -- bit# modemask
  swap >2bitmask			\ -- modemask bitmask
  tuck and swap 			\ -- val bitmask
;

: ms>vba	\ mode struct -- val bitmask *gpio
  tuck  2 cells + @ $0F and		\ -- struct mode bit#
  mode>valmask				\ -- struct val bitmask
  rot @					\ -- val bitmask *gpio
;

: isPinMode	\ mode struct --
\ *G Sets the GPIO mode for a pin. Use with the constants below.
  ms>vba gpioMODER + setMask		\ set pin's bits in MODER
;

: af#>mask	\ mode -- afbits
\ Repeats a four bit mask 8 times.
  $0F and
  dup 4 lshift or			\ 8 bits
  dup 8 lshift or			\ 16 bits
  dup #16 lshift or			\ 32 bits
;

: >4bitmask/off	\ bit# -- bitmask offset
\ Converts a bit number 0..15 to a 4 bit mask and offset for a
\ 32 bit word,
  dup 7 and 4* $0F swap lshift		\ -- bit# mask
  swap $08 and 1 rshift			\ -- mask offset
;

: af#>valmask	\ af# bit# -- val mask offset
\ Convert a 4 bit alternate function number and bit number
\ (0..15) into a value and bitmask to use with setMask.
  swap af#>mask 			\ -- bit# afmask
  swap >4bitmask/off 			\ -- afmask bitmask offset
  >r  tuck and swap  r>			\ -- val bitmask offset
;

\ *P Pins can be in one of four modes, 0..3.

0 constant InputMode	\ -- mode#
\ *G The pin is an input.
1 constant OutputMode	\ -- mode#
\ *G The pin is an output.
2 constant AFmode	\ -- mode#
\ *G The pin is for one of the alternate function modes.
3 constant AnalogMode	\ -- mode#
\ *G The pin is an analogue pin.

: isInput	\ struct --
\ *G Set the pin to GPIO mode and input.
  InputMode swap isPinMode
;

: isOutput	\ struct --
\ *G Set the pin as an output.
  OutputMode swap isPinMode
;

: isFunction	\ af# struct --
\ * Set the pin to be used with one of the alternate functions.
  >r
  AFmode r@ isPinMode			\ uses an alternate function.
  r@ 2 cells + @ $0F and af#>valmask	\ -- val mask offset ; -- R: struct
  r> @ gpioAFRL + + setMask		\ set bits in AFRL or AFRH
;

: isAnalog	\ struct --
\ *G Set the pin as an analog pin.
  AnalogMode swap isPinMode
;

: isPinOD	\ 0/1 struct --
\ *G Enable (nz) or disable (0) the open drain driver.
  swap if
    2@ gpioOTYPER + or!
  else
    2@ gpioOTYPER + bic!
  endif
;

: isPinPuPd	\ mode struct --
\ *G Sets the pull up/down for a pin.
  ms>vba gpioPUPDR + setMask		\ set pin's bits in PUPDR
;

: NotPulled	\ struct --
\ *G No pull up or pull down.
  0 swap isPinPuPd  ;

: PulledUp	\ struct --
\ *G Enable the pin pull up resistor.
  1 swap isPinPuPd  ;

: PulledDown	\ struct --
\ *G Enable the pin pull down resistor.
  2 swap isPinPuPd  ;

: isPinSpeed	\ mode struct --
\ *G Sets the output speed for a pin according to the constants
\ ** below.
  ms>vba gpioOSPEEDR + setMask		\ set pin's bits in PUPDR
;

0 constant 2MHZ_ls	\ -- mask
\ *G The default is 2MHz, low speed.
1 constant 25MHZ_ms	\ -- mask
\ *G 25 MHz, medium speed.
2 constant 50MHZ_fs	\ -- mask
\ *G 50 MHz, fast speed.
3 constant 100MHZ_hs	\ -- mask
\ *G 100 MHz, high speed.

: initGPIO	\ --
\ *G Enable clocks to all GPIO ports and take them out of reset.
\ ** performed in the cold chain.
  $1FF _RCC  2dup rccAHB1ENR + or!  rccAHB1RSTR + bic!
;
' initGPIO atCold


\ ========================================
\ *S Test code for STM32F4 Discovery board
\ ========================================
\ *P The user LEDs on the Discovery board are set as follows:
\ *E  LED1  green   PD12
\ **  LED2  orange  PD13
\ **  LED3  red     PD14
\ **  LED4  blue    PD15
\ *P Note that the LEDs are driven active high/on.
\ *[

0 [if]
decimal

3 12 pio: led1		\ PD.12
3 13 pio: led2		\ PD.13
3 14 pio: led3		\ PD.14
3 15 pio: led4		\ PD.15

: ledDelay	\ --
  200 ms
;

: tled1		\ --
  led1 isOutput
  begin
    led1 clrPin  ledDelay
    led1 setPin  ledDelay
  key? until
;

: tleds		\ --
  led1 isOutput  led2 isOutput  led3 isOutput  led4 isOutput
  begin
    led1 setPin  ledDelay  led1 clrPin
    led2 setPin  ledDelay  led2 clrPin
    led3 setPin  ledDelay  led3 clrPin
    led4 setPin  ledDelay  led4 clrPin
  key? until
;
[then]
\ *]


\ ======
\ *> ###
\ ======

decimal
