\ gpioXMC1xxx.fth - easy access to XMC1xxx GPIO pins.

((
Copyright (c) 2011, 2012, 2014, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20151221 SFP006 XMC1xxx conversion.
20151210 SFP005 Commented out INITGPIO which is done at power up.
20120326 SFP004 Conversion for STM32F0.
20120210 SFP003 Corrected ISFUNCTION.
20120130 SFP002 Refactored, added pin speed.
20111201 SFP001 Adapted from Kinetis version.
))

only forth definitions
decimal

\ ===========
\ *> drivers
\ *S XMC1xxx GPIO utilities
\ ===========
\ *P The code in *\i{CortexLite\Drivers\gpioXMC1xxx.fth} provides
\ ** utility words for accessing GPIO pins on a bit by bit basis
\ ** for Infineon XMC CPUs.
\ ** The code is written for ease of use rather than performance.


\ ********************
\ *N Defining I/O pins
\ ********************
\ *P FIO/GPIO acess is defined using a bit number. Bits 0..31 form
\ ** P0.0 to P0.31, bits 32..63 form P1.0 to P1.31 and so on.
\ ** Although ST name their ports from A, here we number
\ ** them from 0. Note also that ST usually only provide 16
\ ** I/O bits per port. Later we define constants for the port
\ ** numbers.

$0100 equ /Port		\ -- u
\ *G Separation in memory between GPIO port base addresses.
#8 equ PortShift	\ -- u
\ *\fo{/Port} as a shift.

target-only
: PIO:		\ port# bit# -- ; -- struct
\ *G Define a port I/O bit by name. For example
\ *C   1 12 pio: LED1   \ P1.12
\ *C   P1 12 pio: LED1  \ P1.12
\ *P defines an I/O bit on PORT1, bit 12 as *\fo{LED1}. At run
\ ** time, the I/O bit structure is returned.
  create				\ -- port# bit#
    over PortShift lshift _PORT0 + ,	\ pins base
    1 over lshift ,			\ bit mask
    swap 5 lshift or ,			\ I/O bit number
    -BuffComp
;
host&target
interpreter
: PIO:		\ port# bit# -- ; -- struct
  create
    over PortShift lshift _PORT0 + ,	\ pins base
    1 over lshift ,			\ bit mask
    swap 5 lshift or ,			\ I/O bit number
;
target

\ *P Define the ports as constants. So that we can use names
\ ** rather than numbers, e.g.
\ *C   PD 12 pio: LED1   \ PD.12
\ *[
0 constant P0
1 constant P1
2 constant P2
\ *]


\ ******************
\ *N GPIO pin access
\ ******************

: setPin	\ struct --
\ *G Set the pin high.
  2@ pn.OMR + w!  ;

: clrPin	\ struct --
\ *G Set the pin low.
  2@ pn.OMR + 2+ w!  ;

: getPin	\ struct -- 0/1
\ *G Read the pin state.
  2@ pn.IN + @ and 0<> 1 and
;


\ ***********************
\ *N IO pin configuration
\ ***********************

internal
: mode>val	\ mode -- modemask
\ Repeats the five bit mask four times.
  3 lshift				\ 5:3 occupies an 8 bit slot
  dup 8 lshift or
  dup 16 lshift or
;

: bm8&addr	\ bit# -- bitmask address
\ +G Converts a bit number 0..15 to a bitmask and the address
\ +* of the corresponding IOCRx register.
  dup $03 and 8 *  $FF swap lshift	\ -- bit# bitmask
  over $0F and -4 and  pn.IOCR0 +	\ -- bit# bitmask IOCRx
  rot 5 rshift Portshift lshift	+	\ -- bitmask offset
  _PORT0 +
;

: mode>vba	\ mode bit# -- val bitmask offset
\ +G Convert a five bit mode and bit number (0..15) into a value and
\ +* bitmask to use with the *\fo{pn.IOCRx} registers and *\fo{setMask}.
  >r  mode>val  r> bm8&addr
;
external

: isPinMode	\ mode struct --
\ *G Sets the GPIO mode for a pin. Use with the constants below.
  2 cells + @ mode>vba setMask		\ set pin's bits in IOCRx
;

\ *P Pins can be in one of eight modes or alternate functions.
\ ** Some modifiers can be applied.
\ *[
%00000 constant DirectIN	\ direct input
%00100 constant InvertIN	\ inverted input
: +PD ( x -- x' ) %00001 or  ;	\ with pull down (i/p)
: +PU ( x -- x' ) %00010 or  ;	\ with pull down (i/p)
: +PS ( x -- x' ) %00011 or  ;	\ with sample on OUT (i/p)

%10000 constant ModeOUT		\ output with direct input
%10001 constant AF1		\ alternate functions
%10010 constant AF2
%10011 constant AF3
%10100 constant AF4
%10101 constant AF5
%10110 constant AF6
%10111 constant AF7
: +PP ( x -- x' ) ;		\ select push pull
: +OD ( x -- x' ) %01000 or  ;	\ select open drain
\ *]

: isInput	\ struct --
\ *G Set the pin to be an input (*\fo{DirectIN}).
  DirectIN swap isPinMode
;

: isOutput	\ struct --
\ *G Set the pin as an output (*\fo{ModeOUT}).
  ModeOUT swap isPinMode
;


\ *****************************
\ *N Test code for XMC2Go board
\ *****************************

0 [if]

P1 0 pio: LED1	\ -- struct
\ *G Structure for pin 1.0 connected as LED1.
P1 1 pio: LED2	\ -- struct
\ *G Structure for pin 1.1 connected as LED2.

[then]


\ ======
\ *> ###
\ ======

decimal
