\ serLPC11xxp.fth - LPC23xx polled RX, polled TX, serial communication drivers

((
Copyright (c) 2008, 2010, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20110614 MPE001 First release.
))

only forth definitions


\ ==============
\ *! serlpc11xxp
\ *T LPC11xx polled serial driver
\ ==============
\ *P This driver provides polled serial comms for LPC11xx CPUs
\ ** using the internal UARTs. Because there are alternate pin
\ ** assignments for some UARTs, you *\b{must} check that the
\ ** supplied code pin setup is correct for your board.


\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUART0? [if]
1 equ useUART0?	\ -- n
\ *G Compile UART0 code if non-zero.
[then]


\ ********
\ *S Tools
\ ********

\ *P *\fo{GenBaudRate} uses the system clock rate (in Hz), the
\ ** perpheral clock divider ratio, and the required serial speed
\ ** (in BPS) and returns the 16 bit divider value for the baud
\ ** rate generator
\ *E              system-speed
\ ** divisor = -------------------
\ **           16 * pcd * baudrate
\ *P The calculation assumes that the fractional diver is disabled.

interpreter
: GenBaudRate	\ bps-rate pcd -- divisor16
\ *G Given a baud rate and the peripheral clock divider
\ ** (1..255), calculate the 16 bit baud rate divisor.
  system-speed swap / swap / #16 /mod
  swap 8 >=
  if  1+  endif
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal

: +FaultConsole	( -- )  ;
\ *G Because this is a polled driver, *\fo{+FaultConsole} for
\ ** fault handling is a *\fo{NOOP}. See *\i{Cortex/FaultCortex.fth}
\ ** for more details.

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup UnLSR + c@ %00100000 and
  until
  UnTHR + !
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)
;

: (serkey?)	\ base -- t/f
\ *G Return true if the given UART has a character avilable to read.
  UnLSR + c@ %00000001 and 0<>
;

: (serkey)	\ base -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
[ tasking? ] [if]  pause  [then]
    dup UnLSR + c@ %00000001 and
  until
  UnRBR + @ $FF and
;

: initUART	\ divisor base --
\ *G Initialise the UART with the given baud rate divisor.
  swap >r
  0 over UnACR + !			\ turn off autobaud
  $10 over UnFDR + !			\ turn off fractional baud rate
  1 over UnFCR + !			\ enable FIFOs
  $83 over UnLCR + !			\ 8 data, 1 stop, no parity, enable DLAB
  r@ 8 rshift $FF and over UnDLM + !	\ set divisor
  r> $FF and over UnDLL + !
  $03 over UnLCR + !			\ clear DLAB
  drop
;

external


\ ********
\ *S UART0
\ ********
\ *P This code only uses and configures RXD (PIO1.6) and
\ ** TXD (PIO1.7).

\ CTSi PIO0_7
\ RTSo PIO1_5
\ RXDi PIO1_6
\ TXDo PIO1_7
\ DTRo PIO2_0 PIO3_0
\ DSRi PIO2_1 PIO3_1
\ DCDi PIO2_2 PIO3_2
\ RIi  PIO2_3 PIO3_3


useUART0? [if]

1 equ pcdU0	\ -- n
\ *G Peripheral clock division ratio for UART0.
console0-speed pcdU0 GenBaudRate equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

: init-ser0	\ --
\ *G Initialise serial port 0
  $1:0040 _SYSCON SYSAHBCLKCTRL + or!	\ Enable IOCON and GPIO clocks
  _IOCONFIG IOCON_PIO1_6 + dup @	\ RXD input, no pull up/down
    $03 invert and
    $01 or
  swap !
  _IOCONFIG IOCON_PIO1_7 + dup @	\ TXD output, no pull up/down
    $03 invert and
    $01 or
  swap !

  bit12 _SYSCON SYSAHBCLKCTRL + or!	\ Enable UART clock
  pcdU0 _SYSCON UARTCLKDIV + !		\ set clock divider

  u0-divisor _UART0 initUART
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  _UART0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  _UART0 (serkey)  ;
create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
;


\ ======
\ *> ###
\ ======

decimal
