\ serLPC81xp.fth - LPC81x polled RX, polled TX, serial communication drivers

((
Copyright (c) 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20140709 MPE001 First release.
))

only forth definitions


\ ==============
\ *! serlpc81xp
\ *T LPC81x polled serial driver
\ ==============
\ *P This driver provides polled serial comms for LPC81x CPUs
\ ** using the internal UARTs. Because there are alternate pin
\ ** assignments for some UARTs, you *\b{must} check that the
\ ** supplied code pin setup is correct for your board.


\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUART0? [if]
1 equ useUART0?	\ -- n
\ *G Compile UART0 code if non-zero.
[then]


\ ********
\ *S Tools
\ ********
\ *P *\fo{GenBaudRate} uses the system clock rate (in Hz), the
\ ** perpheral clock divider ratio, and the required serial speed
\ ** (in BPS) and returns the 16 bit divider value for the baud
\ ** rate generator
\ *E              system-speed
\ ** divisor = -------------------
\ **           16 * pcd * baudrate
\ *P The calculation assumes that the fractional diver is disabled.

interpreter
: GenBaudRate	\ bps-rate pcd -- divisor16
\ *G Given a baud rate and the peripheral clock divider
\ ** (1..255), calculate the 16 bit baud rate divisor.
  system-speed 2* swap / swap / #16 /mod
  swap 8 >=
  if  1+  endif
;
target


\ ********************
\ *S Serial primitives
\ ********************

internal

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup UnSTAT + @ $0004 and
  until
  UnTXDAT + !
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)
;

: (serkey?)	\ base -- t/f
\ *G Return true if the given UART has a character avilable to read.
  UnSTAT + @ $0001 and 0<>
;

: (serkey)	\ base -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
\ [ tasking? ] [if]  pause  [then]
    dup UnSTAT + @ $0001 and
  until
  UnRXDAT + @ $FF and
;

: initUART	\ divisor base --
\ *G Initialise the UART with the given baud rate divisor.
  _SYSCON				\ U_PCLK = UARTCLKDIV/(1 + MULT/DIV)
  1 over UARTCLKDIV + !			\ divide by 1
  0 over UARTFRGMULT + !		\ 1 + mul/div = 1 + 0/256 = 1
  $FF over UARTFRGDIV + !		\ divisor = 256
  drop

  $0004 over UnCFG + !			\ 8N1
  0 over UnCTL + !			\ defaults
  $F920 over UnSTAT + !			\ clear error flags
  $F965 over UnINTENCLR + !		\ disable interrupts
  tuck UnBRG + !			\ BRG value
  $0005 over UNCFG + !			\ 8N1 + Enable
  drop
;

external


\ ********
\ *S UART0
\ ********
\ *P This code only uses and configures RXD and TXD.

useUART0? [if]

1 equ pcdU0	\ -- n
\ *G Peripheral clock division ratio for UART0.
console0-speed pcdU0 GenBaudRate equ u0-divisor	\ -- n
\ *G UART0 baud rate divisor.

internal

: init-ser0	\ --
\ *G Initialise serial port 0; TX on P0.4, RX on P0.0
  $FFFF:0004 _SM smPINASSIGN0 + !	\ CTS, RTS, RXD, TXD ; $FF=off, RTS=812o/p, CTS=

  $4:0040 _SYSCON SYSAHBCLKCTRL + or!	\ Enable IOCON and GPIO clocks
  pcdU0 _SYSCON UARTCLKDIV + !		\ set clock divider

  $0080 _IOCONFIG PIO0_4 + !		\ TXD output, no pull up/down
  $0080 _IOCONFIG PIO0_0 + !		\ RXD input, no pull up/down
  $0088 _IOCONFIG PIO0_9 + !		\ CTS output, pulled down

 \ set FTDI CTS line to 0 on P0-9
  $0000:0200 _GPIO0 gpDIR0 + or!	\ set bit 9 to output
  $0000:0200 _GPIO0 gpCLR0 + or!	\ set bit 9 to zero

  u0-divisor 1- _UART0 initUART
;

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _UART0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _UART0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _UART0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  _UART0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  _UART0 (serkey)  ;

external

create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line

console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
;


\ ======
\ *> ###
\ ======

decimal
