\ serXMC1xxxp.fth - XMC1xxx polled RX, polled TX, serial communication drivers

((
Copyright (c) 2008, 2010, 2011, 2015
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20151214 MPE002 XMC1xxx conversion.
20110614 MPE001 First release.
))

only forth definitions  decimal


\ ==============
\ *! serxmc1xxxp
\ *T XMC1xxx polled serial driver
\ ==============
\ *P This driver provides polled serial comms for XMC1xxx CPUs
\ ** using the internal UARTs. Because there are alternate pin
\ ** assignments for some UARTs, you *\b{must} check that the
\ ** supplied code pin setup is correct for your board.

\ *P The defaults in this file are for the Infineon XMC2Go board.


\ ****************
\ *S Configuration
\ ****************
\ *P The following equates control which UARTs are required
\ ** and initialised. If the equates are undefined, they are
\ ** generated and set.

[undefined] useUART0? [if]
1 equ useUART0?	\ -- n
\ *G Compile USIC0-CH0 code if non-zero.
\ ** Default P2.1 = TxD, P2.2 = RxD
[then]

[undefined] useUART1? [if]
0 equ useUART1?	\ -- n
\ *G Compile USIC0-CH1 code if non-zero.
[then]


\ ********
\ *S Tools
\ ********
\ *P Calculating BRG coefficients
\ *C fperiph -> fracdiv -> divmode -> quanta -> baud
\ *C fperiph / fracdiv / PDIV+1 / DCTQ+1

\ *P Quanta (below) is the number of samples per bit in the baud rate.
\ ** This is usually 16 in hardware UARTs, but here is programmable as
\ ** DCTQ+1.

\ *E normal divider (DM=01)
\ ** ==============
\ **                 fperiph
\ ** baud = ------------------------
\ **        (1024-step) * (pdiv+1) * quanta
\ **
\ **                  baud * (pdiv+1) * quanta
\ ** 1024 - step = -------------------------------
\ **                         fperiph
\ **
\ ** fractional divider (DM=10)
\ ** ==================
\ **                fperiph * step
\ ** baud = ----------------------------
\ **        1024 * (pdiv+1) * quanta * 2
\ **
\ **        baud * (pdiv+1) * 1024 * quanta * 2
\ ** step = -----------------------------------
\ **                    fperiph

\ *P Normally we use the fractional divider. If the STEP value is too
\ ** high, we introduce one or two stages of divide by two by
\ ** changing PDIV.


4 equ /PDIV	\ -- u
\ *G Initial value of PDIV divider ratio.

interpreter
: genStep	\ baud /pdiv -- step
\ *G Generate the STEP field in the USIC.FDR register using the
\ ** fractional divider. The quanta (samples/bit) are assumed to
\ ** be 16.
  *  #32768 periph-speed */
;

: genBRG	\ baud -- /pdiv /step
\ *G Given the required baud rate, generate the required PDIV and
\ ** STEP values for 16 samples/bit.
  4 begin
    2dup genStep dup #1022 >=
   while			\ -- baud /pdiv /step
    drop  2/ dup 0= if
      cr ." Cannot reduce PDIV enough" abort
    then
  repeat
  drop begin
    2dup genStep dup #256 u<
   while			\ -- baud /pdiv /step
    drop  2*
  repeat
  rot drop			\ discard baud
  over 1024 > if
    cr ." Cannot find baud rate values" abort
  then
;

: adjustStep	\ step -- step'
\ *G Over a number of boards, we have determined empirically that
\ ** the internal clock is not based exactly around 64 MHz. Use
\ ** this adjustment of the step value if an internal clock is
\ ** being used. Yes, it's a fudge factor!
  575 590 */
;
target


\ ********************
\ *S Serial primitives
\ ********************

\ internal

: +FaultConsole	( -- )  ;
\ *G Because this is a polled driver, *\fo{+FaultConsole} for
\ ** fault handling is a *\fo{NOOP}. See *\i{Cortex/FaultCortex.fth}
\ ** for more details.

: (seremit)	\ char base --
\ *G Transmit a character on the given UART.
  begin
    dup usic.PSR + @ bit9 and 0=
  until
  usic.IN0 + !
;

: (sertype)	\ caddr len base --
\ *G Transmit a string on the given UART.
  -rot bounds
  ?do  i c@ over (seremit)  loop
  drop
;

: (sercr)	\ base --
\ *G Transmit a CR/LF pair on the given UART.
  $0D over (seremit)  $0A swap (seremit)
;

: (serkey?)	\ base -- t/f
\ *G Return true if the given UART has a character avilable to read.
  usic.TRBSR + @ bit3 and 0=
;

: (serkey)	\ base -- char
\ *G Wait for a character to come available on the given UART and
\ ** return the character.
  begin
[ tasking? ] [if]  pause  [then]
    dup usic.TRBSR + @ bit3 and 0=
  until
  usic.OUTR + @ $FF and
;

external


\ ********
\ *S UART0
\ ********
\ *P by default, P2.1 = TxD, P2.2 = RxD.

useUART0? [if]

console0-speed genBRG adjustStep equ /Step equ /Pdiv
\ *G Calculate the required values of the STEP and PDIV divisors.

[undefined] init-ser0 [if]
/Step equ UART_FDR_STEP		\ Step value for the fractional divider
/Pdiv 1- equ UART_BRG_PDIV	\ divide by 4
15 equ UART_BRG_DCTQ		\ 16 ticks per UART bit time
0 equ UART_BRG_PCTQ		\ pre-divide by 1

$8000 UART_FDR_STEP or equ FDRval	\ -- x
\ *G Initial value of FDR register, set DM=10, STEP=575.

UART_BRG_PDIV 16 lshift
UART_BRG_DCTQ 10 lshift or
UART_BRG_PCTQ 8 lshift or
  equ BRGval	\ -- x
\ *G PDIV=3, DCTQ=9, PCTQ=0.
\ ** BAUDRATE = fCTQIN/(BRG.PCTQ x BRG.DCTQ).
\ ** CLKSEL = 0 (fPIN = fFD), CTQSEL = 00b (fCTQIN = fPDIV),
\ ** PPPEN = 0 (fPPP=fPIN).

$0707:0102 equ SCTRval	\ -- x
\ *G Configuration of USIC Shift Control.
\ ** SCTR.WLE = 8 (Word Length)
\ ** SCTR.FLE = 8 (Frame Length)
\ ** SCTR.TRM = 1 (Transmission Mode)
\ ** SCTR.PDL = 1 (This bit defines the output level at the shift
\ ** data output signal when no data is available for transmission).

$0000:0500 equ TCSRval	\ -- x
\ *G Configuration of USIC Transmit Control/Status Register.
\ ** TBUF.TDEN = 1 (TBUF Data Enable: A transmission of the data
\ ** word in TBUF can be started if TDV = 1.
\ ** TBUF.TDSSM = 1 (Data Single Shot Mode: allow word-by-word
\ ** data transmission which avoid sending the same data several
\ ** times.

$0002:0801 equ PCRval	\ -- x
\ *G Configuration of Protocol Control Register, page 345/347.
\ ** PCR.SMD = 1 (Sample Mode based on majority).
\ ** PCR.STPB = 0 (1x Stop bit).
\ ** PCR.SP = 8 (Sample Point) for 16 ticks/bit.
\ ** PCR.PL = 0 (Pulse Length is equal to the bit length).
\ ** PCR.TSTEN = 1 (PSR.9 is transmit complete)

: init-ser0	\ --
\ *G Initialise serial port 0. This version is used if *\fo{init-ser0}
\ ** has not already been defined.
  unlockBPS
  bit3 _CCU ccu.CGATCLR0 + !		\ enable (disable gating) USIC0
  lockBPS
  0 _PORT2 pn.PDISC + !			\ all P2 is digital
  _USIC0_CH0
    $0000 over usic.CCR + !		\ mode = 0 = disabled
    $0003 over usic.KSCFG + !		\ enable the module, BPMODEN/MODEN
    FDRval over usic.FDR + !		\ fractional divider
    BRGval over usic.BRG + !		\ baud rate
    SCTRval over usic.SCTR + !		\ shift for UART, 8 bits
    TCSRval over usic.TCSR + !		\ Transmit Control Status
    PCRval over usic.PCR + !		\ ASC Protocol register
    0 over usic.DX3CR + !		\ use DX0 for DX3, selects P2.2
    6 over usic.DX0CR + !		\ use DX0G for DX0.
    $0500:0020 over usic.RBCTR + !	\ 2^5 entries, pos=32 for Rx buffer
    $0500:0000 over usic.TBCTR + !	\ 2^5 entries, pos=0 for Tx buffer
    $0002 over usic.CCR + !		\ mode = 2 = ASC
  drop
  $0000:B000 _PORT2 pn.IOCR0 + !	\ P2.2=input, P2.1=o/p ALT6 = U0C0.DOUT0
;
[then]

: seremit0	\ char --
\ *G Transmit a character on UART0.
  _USIC0_CH0 (seremit)  ;
: sertype0	\ c-addr len --
\ *G Transmit a string on UART0.
  _USIC0_CH0 (sertype)  ;
: sercr0	\ --
\ *G Issue a CR/LF pair to UART0.
  _USIC0_CH0 (sercr)  ;
: serkey?0	\ -- flag
\ *G Return true if UART0 has a character available.
  _USIC0_CH0 (serkey?)  ;
: serkey0	\ -- char
\ *G Wait for a character on UART0.
  _USIC0_CH0 (serkey)  ;
create Console0	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for UART0.
  ' serkey0 ,		\ -- char ; receive char
  ' serkey?0 ,		\ -- flag ; check receive char
  ' seremit0 ,		\ -- char ; display char
  ' sertype0 ,		\ caddr len -- ; display string
  ' sercr0 ,		\ -- ; display new line
console-port 0 = [if]
  console0 constant console
\ *G *\fo{CONSOLE} is the device used by the Forth system for interaction.
\ ** It may be changed by one of the phrases of the form:
\ *C   <device>  dup opvec !  ipvec !
\ *C   <device>  SetConsole
[then]

[then]


\ ************************
\ *S System initialisation
\ ************************

: init-ser	\ --
\ *G Initialise all serial ports
  [ useUART0? ] [if]  init-ser0  [then]
;


\ ======
\ *> ###
\ ======

decimal
