\ MultiCM0lite.fth - ARM Cortex-M0 v7 Lite Multi-Tasker

((
Copyright (c) 2010, 2011, 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20141121 MPE003 Added LP to saved registers.
20140210 MPE002 Conversion for Lite kernel
20110627 MPE001 Cortex-M0/M1 conversion.


********************
Resource definitions
********************

  For Cortex-M0/M1 the following register usage is the default:
  r15         pc      program counter
  r14         link    link register; bit0=1=Thumb, usually set
  r13         rsp     return stack pointer
  r12         --
  r11         up      user area pointer
  r10         --
  r9          lp      locals pointer
  r8          --
  r7          tos     cached top of stack
  r6          psp     data stack pointer
  r0-r5       scratch
))

\ ===========
\ *! multicortex
\ *T ARM Cortex-M0/M1 multitasker
\ ===========
\ *P The ARM Cortex multitasker follows the model introduced
\ ** with the v6.1 compilers.


\ ****************************
\ *S TCB data structure layout
\ ****************************
\ *E cell       LINK    link to next task
\ ** cell       SSP     Saved Stack Pointer
\ ** cell       STAT    Bit 0     1 = running, 0 = halted
\ **                    Bit 1     1 = message pending
\ **                    Bit 2     1 = event triggered
\ **                    Bit 3     1 = event handler run
\ **                    Bit 4..7  Reserved
\ ** 	        others  1 = set to run task, available to user
\ ** cell       TASK    Task that sent message here
\ ** cell       MESG    Message address
\ ** cell       EVNTw   CFA of word run as event handler
\ *P This structure is allocated at the start of the *\fo{USER}
\ ** area. Consequently the *\fo{TCB} of the current task is given
\ ** by *\fo{UP}.

struct /TCB	\ -- size
\ *G The structure used by the code that matches the
\ ** description above.
  ptr tcb.link		\ link to next task ; MUST BE FIRST
  ptr tcb.ssp	        \ Saved Stack Pointer
  int tcb.status	\ status word
  int tcb.msrc		\ message source
  int tcb.mesg		\ message
  ptr tcb.event		\ xt of word which is event handler
end-struct

\ bit masks for the status cell
$0001 equ run-mask  run-mask invert equ ~run-mask	\ running
  0 equ run-bit#
$0002 equ msg-mask  msg-mask invert equ ~msg-mask	\ message available
  1 equ msg-bit#
$0004 equ trg-mask  trg-mask invert equ ~trg-mask	\ event triggered
  2 equ trg-bit#
$0008 equ evt-mask  evt-mask invert equ ~evt-mask	\ event run
  3 equ evt-bit#


\ ******************
\ Consistency checks
\ ******************

/tcb tcb-size <> [if]
  cr ." Control file and tasker disagree as to TCB size"  abort
[endif]


\ ***************************
\ *S Task handling primitives
\ ***************************

cdata		\ so that comma is to code space

init-u0 constant main	\ -- addr ; tcb of main task
\ *G Returns the base address of the main task's *\fo{USER} area.
l: main-link
  0 ,					\ end of task chain

CODE (pause)	\ -- ; the scheduler itself
\ *G The software scheduler itself.
l: [schedule]
\
\ save previously running task TOS, PSP, LP
\
  mov     r0, lp			\ MPE002
  push    { r0, r6, r7, link }		\ LP, PSP, TOS, LP ; MPE003
\  push    { r6, r7, link }		\ PSP, TOS
  mov     r0, rsp
  mov     r1, up
  str     r0, [ r1, # 0 tcb.ssp ]	\ save SP in TCB
\
\ select next task to run
\
l: [schedule]1
  ldr     r1, [ r1, # 0 tcb.link ]	\ get next task
  ldr     r0, [ r1, # 0 tcb.status ]	\ inspect status
  cmp     r0, # 0			\ 0 = not running
  b .eq   [schedule]1
\
\ run selected task - TOS, PSP, LP
\
  mov     up, r1
  ldr     r0, [ r1, # 0 tcb.ssp ]	\ restore SSP
  mov     rsp, r0
  pop     { r0, r6, r7 }		\ restore registers ; MPE003
  mov     lp, r0			\ MPE003
\  pop     { r6, r7 }			\ restore registers
  pop     { r0 }
  mov     link, r0

  next,
end-code

0 value multi?	\ -- flag
\ *G Returns true if the tasker is enabled.

: single	\ --
\ *G Disable scheduler.
  ['] noop to-do pause  0 to multi?  ;

: multi		\ --
\ *G Enable scheduler.
  ['] (pause) to-do pause  -1 to multi?  ;


: status	\ -- task-status
\ *G Returns the current task's status cell, but with the run
\ ** bit masked out.
  up@ tcb.status @ run-mask invert and  ;

: restart	\ task -- ; mark task TCB as running
\ *G Sets the RUN bit in the task's status cell.
  run-mask swap tcb.status or!  ;

: halt		\ task -- ; reset running bit in TCB
\ *G Clears the RUN bit in the task's status cell.
  run-mask swap tcb.status bic!  ;

: stop          \ -- ; halt oneself
\ *G *\fo{HALT}s the current task, and executes *\fo{PAUSE}.
  self halt  pause
;


\ *************************
\ *S Task structure management
\ *************************

internal

code init-task	\ xt task -- ; Initialise a task stack
\ *G Initialise a task's stack before running it and
\ ** set it to execute the word whose XT is given.
  ldmia   psp ! { r1 }			\ get execution address
  mov .s  r3, # 1
  orr .s  r1, r1, r3			\ set Thumb bit
  mov     r2, rsp			\ save return stack pointer

\ Generate the RSP of the new task
\ the middle line works because TASK-U0 is greater than TASK-R0
  mov     r0, tos
  sub .s  r0, r0, # task-u0 task-r0 -	\ generate new task RSP
  mov     rsp, r0

  str     r0, [ tos, # r0-offset ]	\ save new R0 (taskID=UP)
  mov .s  r0, # 0			\ will need this many times
  str     r0, [ tos, # 0 tcb.status ]	\ clear new task status

  push    { r1 }			\ R14 LINK, push xt=link
  push    { r0 }			\ R7, new TOS for ARM
\ the next but one line works because TASK-U0 is greater than TASK-S0
  mov     r1, tos
  sub .s  r1, r1, # task-u0 task-s0 -	\ calculate new PSP
  sub .s  r1, r1, # sp-guard cells	\ generate new PSP
  push    { r1 }			\ R6, new PSP
  push    { r0 }			\ LP - dummy for now ; MPE003
  sub .s  r1, r1, # tos-cached? cells	\ generate new S0
  str     r1, [ tos, # s0-offset ]	\ that is compatible with SP!
  mov     r0, rsp
  str     r0, [ tos, # 0 tcb.ssp ]	\ save RSP

  mov     rsp, r2			\ restore RSP
  ldmia   psp ! { tos }			\ restore TOS
  next,
end-code

: add-task      \ task -- ; insert into list
\ *G Add the task to the list of tasks after the current task.
  self tcb.link @			\ save task currently pointed to
  over tcb.link !			\ and make new task point to it
  self tcb.link !			\ current task points to new task
;

: sub-task      \ task -- ; remove task from chain
\ *G Remove the task from the task list.
\ look for task which points to required task
  self					\ head of chain ; -- target current
  begin  2dup tcb.link @ <>		\ while curr. does not point to target
   while				\ step through chain
    tcb.link @  dup self =		\ but if task not in list, check for end
    if  2drop  exit  endif		\ and get out
  repeat                                \ -- target points-to-it
\ set task that points to target to point to task pointed to by the target
  swap tcb.link @			\ task pointed to by target
  swap tcb.link !			\ set into link field
;

external

: initiate	\ xt task -- ; start task from scratch
\ *G Start the given task executing the word whose XT is given, e.g.
\ *C   ['] <name> <task> INITIATE
  tuck init-task			\ reset TCB
  dup add-task                          \ add to chain
  run-mask swap tcb.status !		\ mark as running
  pause					\ let it run
;

: terminate	\ task --
\ *G Stop a task, and remove it from the list.
  sub-task pause			\ halt task, and remove from chain
;

: init-multi    \ --
\ *G Initialise the multitasker and start it.
  single				\ tasker disabled
  main /tcb erase			\ clean TCB
  main dup tcb.link !			\ initialise MAIN to point to itself
  run-mask main tcb.status !		\ and reset status field
  multi					\ mark tasker enabled
;  ' init-multi AtCold

: his		\ task uservar -- addr
\ *G Given a task id and a *\fo{USER} variable, returns the
\ ** address of that variable in the given task. This word is
\ ** used to set up *\fo{USER} variables in other tasks.
  self - +
;


\ **********
\ *S Semaphores
\ **********
\ *P The semaphore code is only compiled if the equate
\ ** *\fo{SEMAPHORES?} is set non-zero in the control
\ ** file.

\ *P A *\fo{SEMAPHORE} is an extended variable used for signalling
\ ** between tasks, and for resource allocation. It contains two
\ ** cells, a *\b{counter} and an *\b{arbiter}. The counter field
\ ** is used as a count of the number of times the resource may
\ ** be used, and the arbiter field contains the TCB of the task
\ ** that currently owns it. This field can be used for priority
\ ** arbitration and deadlock detection/arbitration. The count
\ ** field allows the semaphore to be used as a*\b{counted}
\ ** semaphore or as an exclusive access semaphore.

\ *P For example a character buffer may be used where the semaphore
\ ** counter contains the number of available characters.

\ *P An exclusive access semaphore is used to share resources.
\ ** The semaphore is initialised to one, usually by *\fo{SIGNAL}.
\ ** The first task to *\fo{REQUEST} it gains access, and all
\ ** other tasks must wait until the accessing task *\fo{SIGNAL}s
\ ** that it has finished with the resource.

interpreter
: semaphore     \ -- ; -- addr [child]
\ *G Creates a semaphore which returns its address at runtime.
\ ** The count field is initialised to zero for use as counted
\ ** semaphore. Use in the form:
\ *C Semaphore <name>
\ *P If you want this to be an exclusive access semaphore, follow
\ ** this with:
\ *C   1 <name> !
  idata create
    0 ,  0 ,				\ count and arbiter fields
  cdata
;
target

: semaphore	\ -- ; -- addr [child]
  2 cells buffer:
  0 0 last @ name> execute 2!
;

: signal        \ addr --
\ *G *\fo{SIGNAL} increments the counter field of a semaphore, indicating
\ ** either that another item has been allocated to the resource, or
\ ** that it is available for use again, 0 indicating in use by a task.
  [I					\ must be interrupt safe
    dup incr  cell+ off			\ inc. counter, release
  I]
;

: request       \ sem -- ; get access to resource, wait if count = 0
\ *G *\fo{REQUEST} waits until the counter field of a semaphore
\ ** is non-zero, and then decrements the counter field by one.
  begin
    [I  dup @ 0=			\ n.b. test and set
   while
    I]  pause				\ operations must be
  repeat                                \ non-interruptible and indivisible
  dup decr				\ got it, decrement counter
  self swap cell+ !			\ mark resource as mine
  I]					\ re-enable interrupts
;


\ ******************
\ *S TASK and START:
\ ******************
\ *P *\fo{TASK <name>} builds a named task user area.
\ ** The action of a task is assigned and the task started
\ ** by the word *\fo{INITIATE}
\ *C   ['] <action> <task> INITIATE

\ *P *\fo{START:} is used inside a colon definition. The code
\ ** before *\fo{START:} is the task's initialisation, performed
\ ** by the current task. The code after *\fo{START:} up to the
\ ** closing *\fo{;} is the action of the task. For example:

\ *E   TASK FOO
\ **   : RUN-FOO
\ **     ...
\ **     FOO START:
\ **     ...
\ **     begin ... pause again
\ **   ;

\ *P All tasks must run in an endless loop, except for initialisation
\ ** code.
\ ** When *\fo{RUN-FOO} is executed, the code after *\fo{START:}
\ ** is set up as the action of task *\fo{FOO} and started.
\ ** *\fo{RUN-FOO} then exits.

\ *P If you want to perform additional actions after starting the task, you
\ ** should use *\fo{INITIATE} to start the task.

variable task-chain	\ -- addr
\ *G Anchors list of all tasks created by *\fo{TASK} and friends.
  main-link task-chain !		\ MAIN already defined

interpreter	\ host version
: task		\ -- ; -- task ; TASK <name> builds a task
\ *G Note that the cross-interpreter's version of *\fo{TASK} has
\ ** been modified from v6.2 onwards to leave the current
\ ** section as *\fo{CDATA}.
  udata
  task-size reserve task-u0 + constant	\ build pointer
  cdata
  here  task-chain @ ,  task-chain !	\ link task
;
target

heads? [if]	\ target version
: task		\ -- ; -- task ; TASK <name> builds a task
\ *G Creates a new task and data area, returning the address
\ ** of the user area at run time. The task is also linked into
\ ** the task chain anchored by *\fo{TASK-CHAIN}.
  ralign
  rp @ task-u0 +			\ task's user area
  task-size rallot
  constant				\ build pointer
  here  task-chain @ ,  task-chain !	\ link task
;
[then]

: start:	\ task -- ; exits from caller
\ *G Used inside a colon definition. The code following
\ ** *\fo{START:} up to the ending semi-colon forms the action
\ ** of the task. The word containing *\fo{START:} finishes at
\ ** *\fo{START:}.
  r> swap initiate
;


\ ******************
\ *S Debugging tools
\ ******************

: .task		\ task --
\ *G Display task's name if it has one, otherwise display its
\ ** address.
  task-chain
  begin					\ -- task link
    @ dup
   while				\ -- task link
    2dup cell - @ = if
      nip  #12 - >name .name space	\ ARM specific
      exit
    endif
  repeat
  drop .dword ."  is not in TASK chain"
;

: .tasks	\ task -- ; display all task names
\ *G Display all the tasks anchored by *\fo{TASK-CHAIN}.
  task-chain
  begin
    @ dup
   while
    dup #12 - >name .name space		\ ARM specific
  repeat
  drop
;

: .running	\ --
\ *G Display all the running tasks.
  main
  begin
    cr dup .task
    tcb.link @
    dup main =
  until
  drop
;


\ ********
\ all done
\ ********

decimal


\ *********
\ Test code
\ *********

0 [if]

task task1	\ -- addr ; build task area

50000 value rate1	\ -- n ; number of idle iterations

: action1	\ -- ; action of task1
  decimal  console opvec !
  begin
    [char] * emit
    rate1 0 do pause loop
  again
;

\ To start task use:
\   ' action1 task1 initiate

task task2	\ -- addr ; build task area

900000 value rate2	\ -- n ; number of idle iterations

: run-task2	\ -- ; example of using START:
  task2 start:
    console opvec !
    begin
      [char] / emit
      rate2 0 do pause loop
    again
;

\ To start task use:
\   run-task2

[then]


\ ======
\ *> ###
\ ======

