(( Time&Date1.fth                                          RV 10/07/2018

Version:      1.0

Copyright by: Reinhard Vogg
              Kiefernweg 8
              87448 Waltenhofen


\ ======================================================================
Description:

Simple time and date functions using the TICK counter.

On the stand-alone FORTH for the STM32F4 Discovery Board
TICKS counts from 0 at system start to $FFFFFFFF.
This is equivalent to a time duration of 49 days 17 hours and 1 minute.

TICKS is a value and can be set by the statement  nnnn TO TICKS.

The word DI (disable interrupt) stops the counting.
When entering EI (enable interrupt) TICKS resumes to count.

The calendar functions have been imported from CALENDAR.FTH written by
    Christophe Lavarenne
    16A, rue Charles Infroit
    92190 Meudon  FRANCE

NOTE:
The processor STM32F407VGT6 (Cortex M4) of the STM32F4 Discovery Board
has an internal HW RTC but the 32678 Hz crystal for the RTC (LSE clock)
isn't fitted.

The RTC of the processor STM32F407VGT6 (Cortex M4) allows to select
different clock sources for the RTC:
- LSE clock (32768 Hz using an external 32678 Hz crystal)
- HSE clock (high speed clock derived from the main crystal clock via prescaler)
- LSI clock (internal RC oscillator approx. 32kHz)
(refer to figure 21 and chapter 26.3.1 of the reference manual RM0090)

The initialisation and readout of the RTC seems to be very complicated.


\ ======================================================================
Changes/History:

10/01/18  Start of the development

10/05/18  Time functions implemented and tested with WinFORTH6.14

10/06/18  Date functions implemented and tested with WinFORTH6.14

10/07/18  WinFORTH6.14 adaptions removed and modifications for the
          stand-alone FORTH (MPE Cortex Target Lite) for the
          STM32F4 Discovery Board added and tested


\ ======================================================================
To Do:



))

DECIMAL
60 60 *        CONSTANT sec/h
sec/h 24 *     CONSTANT sec/day
sec/day 1000 * CONSTANT ms/day


\ USER variable BASE is internal on MPE Cortex Target Lite FORTH
\ temporary solution:
: BASE  OUT CELL+ ;


VARIABLE <lastTICKS>      \ holds the last read TICKS value
VARIABLE <time>           \ holds the time of the day in milliseconds
VARIABLE <date>           \ holds the date in days since 1.1.1980
VARIABLE <tdelta>         \ holds the read TICKS value when performing TStart
                          \ and the elapsed time after TStop

0 <date> !  0 <time> !    TICKS <lastTICKS> !   0 <tdelta> !


\ Outputs u as 2 digit decimal number, BASE keeps unchanged.
: .nn_d  ( u ---)  BASE @ >R  DECIMAL  0 <# # # #> TYPE  R> BASE ! ;

\ Outputs u as 3 digit decimal number, BASE keeps unchanged.
: .nnn_d  ( u ---)  BASE @ >R  DECIMAL  0 <# # # # #> TYPE  R> BASE ! ;

\ Outputs u as 4 digit decimal number, BASE keeps unchanged.
: .nnnn_d  ( u ---)  BASE @ >R  DECIMAL  0 <# # # # # #> TYPE  R> BASE ! ;


\ Converts the time in milliseconds since 00:00:00.000 to hour min sec
: Tms>HMS ( tms - ms sec min hour)
          S>D  1000 UM/MOD   \ rem(tms/1000)=ms  quot(tms/1000)
          DUP  sec/h  /MOD   \ ms  quot(tms/1000)  rem(tms/1000)/sec/h
                             \               hour=quot(tms/1000)/sec/h
          SWAP 60 /          \ ms  quot(tms/1000)  hour  min=rem(tms/1000)/sec/h / 60
          ROT                \ ms  hour  min  quot(tms/1000)
          60   MOD           \ ms  hour  min  sec=rem(quot(tms/1000)/60)
          SWAP               \ ms  hour  sec  min
          ROT   ;            \ ms  sec  min  hour


\ Reads TICKS and updates the variables.
: Update_T&D  ( ---)
              TICKS  DUP  <LastTICKS> @  -  ( ticks ticks-LastTICKS = delta)
              DUP
              0<  ABORT" Time & date error; TICKS run through $FFFFFFFF"
              ( ticks delta)  SWAP  <LastTICKS> !   ( delta)
              <time> +!                \ <time> = <time> + delta
              <time> @  ms/day  /MOD   \ quot = <time> / ms/day
                                       \  rem = remainder(<time> / ms/day)
              <date> +!                \ <date> = <date> + remainder(<time> / ms/day)
              <time> !                 \ <time> = <time> / ms/day
;


\ Displays the current time in the form hh:dd:ss.xxx (xxx = milliseconds)
: .T     ( ---)                                         ( not in ANS FORTH)
         Update_T&D
         <time> @
         Tms>HMS  .nn_d  $3A EMIT  .nn_d  $3A EMIT  .nn_d  $2E EMIT  .nnn_d ;

\ Displays the current time in the form hh:dd:ss         (not in ANS FORTH)
: .TIME  ( ---)
         Update_T&D
         <time> @
         Tms>HMS  .nn_d  $3A EMIT  .nn_d  $3A EMIT  .nn_d  DROP ;


\ Julian calendar using an unsigned 16 bit number to count the date in days
\ The time domain is from 3/1/1900 to 7/31/2079
\ Weekday codes: 0=Sun 1=Mon ...  6=Sat
\ Month   codes: 1=Jan 2=Feb ... 12=Dec
\ The year is always output complete (0000 ... 1990 ... 32767)
\ The year may be entered abreviated for XXth (0=1900 ... 99=1999)

\ Converts the given day month and year to a 16 bit number un in the
\ range 36159 to 65531 representing the date between 3/1/1900 and 7/31/2079
: date> ( day month year - un)          \ see month codes
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if 12 + r> 1- >r then    	\ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

\ Converts the unsigned number un to the weekday code ( see weekday codes)
: weekday ( un -- n)
  4 + 0 7 um/mod drop
;

\ Returns the weekday code ( see weekday codes)
: week-day  ( ---)  Update_T&D  <date> @  weekday ;


\ Converts the unsigned number un to day, month and year
\ (see description above)
: >date ( un - day month year)
  0 1461 um/mod                         \ no leap-year periods
  4 * >r                                \ convert to no years
  365 /mod                              \ no years and days
  dup 4 = if 2drop 365 3 then           \ correct if Febr.29th
  r> + >r                               \ -- days | years
  123 + dup 5 153 */                    \ compute month
  dup >r 153 5 */ - r>                  \ compute day
  1- dup 12 > if 12 - r> 1+ >r then     \ correct month and year
  r> 1900 +                        	\ -- day month year
;


\ Sets the current date.                                 (not in ANS FORTH)
: !TIME ( hh mm ss ---)      \ hh = hours; mm = minutes; ss = seconds
                             \ the milliseconds will be reset to 0
         SWAP  60 *  +
         SWAP  sec/h * +
         1000  *
         <time> !
         TICKS <lastTICKS> !
;

\ Sets the current date.                                 (not in ANS FORTH)
: !DATE ( dd mm yyyy ---)    \ dd = day; mm = month; yyyy = year
        date>  <date> !
\       TICKS <lastTICKS> !  \ TBC if needed

;

\ Fetches the time and date information.                 (ANS FORTH compatible>
: TIME&DATE ( --- n1 n2 n3 n4 n5 n6) \ n1 = seconds; n2 = minutes; n3 = hours;
            Update_T&D               \ n4 = day; n5 = month; n6 = year
            <time> @  Tms>HMS
            4 ROLL DROP              \ remove the milliseconds
            <date> @  >date
;


\ Displays the date in the form dd.mm.yyyy  or  mm/dd/yyyy
: (.date)  ( day month year ---)
           ROT  .nn_d  $2E EMIT  SWAP  .nn_d  $2E EMIT  .nnnn_d ; \ german, french, italian
\         SWAP  .nn_d  $2F EMIT  SWAP  .nn_d  $2F EMIT  .nnnn_d ; \ USA, UK, Canada


\ Displays the current date in the form dd/mm/yyyy       (not in ANS FORTH)
\ It uses the Julian Calender starting at 01.01.1980
: .DATE  ( ---)
        Update_T&D
        <date> @
        >date  (.date)
;

\ Returns the number of milliseconds since 00:00:00.000  (not in ANS FORTH)
: ms@      ( --- u)  Update_T&D  <time> @ ;


\ Starts the elapsed time counter.                       (not in ANS FORTH)
\ TStart saves the TICKS count in <tdelta>.
: TStart ( ---)   TICKS  <tdelta> !  ;

\ Stops the elapsed time counter.                        (not in ANS FORTH)
\ TStop calculates the time difference and saves the result in <tdelta>.
\ The time difference can be displayed by Elapsed
: TStop  ( ---)   TICKS  <tdelta> @  -  <tdelta> !  ;

\ Displays the elapsed time between TStart and TStop.    (not in ANS FORTH)
: Elapsed ( ---)
          <tdelta> @  ( tms = time difference in milliseconds)
          Tms>HMS
          ." Elapsed time: "
          .nn_d  $3A EMIT  .nn_d  $3A EMIT  .nn_d  $2E EMIT  .nnn_d
;


\ \s

\ testing

: .var  CR  ." <lastTICKS> "  <lastTICKS> @ .
        CR  ."      <time> "  <time> @ .
        CR  ."      <date> "  <date> @ .
        CR  ."    <tdelta> "  <tdelta> @ .
        CR  ;

\ hh mm ss !TIME   ( setup the current time)

: TEST1  CR .TIME  CR .T  CR ;

: TEST2  CR .T
         CR  TStart  1000 ms    TStop  Elapsed  CR .T  CR ;

\ dd mm yy !DATE   ( setup the current date)

: TEST3  CR  .DATE  CR ;

: TEST4  CR  .T   CR  .DATE ;

\ Simulation of a time duration of multiple days between display of the time.
\ Set the time and date before invoking this test !
\ This test should increment the date by 4 days
: TEST5  TEST4  86400000 4 * <lastTICKS> @ + TO TICKS  TEST4 ;



