\ Trig32.fth -- 32 bit integer trig functions

\ *************************************************
\ From Brad Eckert on comp.lang.forth - 16 Feb 2013
\ *************************************************

\ Cosine function in 32-bit ANS Forth

\ Angle input is 32-bit
\ Output range is 0 to +/-$7FFFFFFF
\ Accuracy is 28.7 bits, worst case error 10/2^32.
\ See Jack Ganssle's Guide to Approximations for FP C version.

HEX
  6487ED51 CONSTANT C1  7FFFFFFF CONSTANT C2
  80000001 CONSTANT C3  15555553 CONSTANT C4
  FE93E947 CONSTANT C5  000D00BD CONSTANT C6
  FFFFB61E CONSTANT C7  00000111 CONSTANT C8
VARIABLE X2

: ROUND  ( d -- n )     \ round to nearest
   80000000 0 D+  NIP
;
: TERM  ( y w -- y' )   \ y*X2/2 + w
   SWAP  X2 @ M* D2* D2* D2*  ROUND +
;
: COS1Q  ( angle -- n ) \ 1st quadrant
   C1 M* D2* D2* ROUND  \ scale to pi/2 = $40000000
   DUP M* NIP X2 !
   C8 C7 TERM  C6 TERM  C5 TERM
   C4 TERM  C3 TERM  C2 TERM
;
: COS  ( angle -- n )   \ angle is 32-bit circle
   DUP 0< IF INVERT THEN
   DUP 40000000 AND  IF  INVERT
       3FFFFFFF AND  COS1Q NEGATE
   ELSE COS1Q THEN
;
DECIMAL

1 [IF]
\ translate coefficients from
\ Jack Ganssle's Guide to Approximations.
\ also test for worst case error

REQUIRES FPMATH
VARIABLE TALLY
: FRAC ( r -- )     \ convert to 32-bit constants
   TALLY @ 1 AND 0= IF CR THEN
   $80000000 0 D>F F*  1 TALLY +!
   TALLY @ 2 > IF   \ they double as you go down
      TALLY @ 2 - 0 DO F2* LOOP
   THEN  ."   "
   F>D HEX <# # # # # # # # # #> TYPE
   ."  CONSTANT C"  DECIMAL TALLY @ .
;
: SHOW  ( -- ) CR   \ Output coefficients to screen
 1E0 FATAN FRAC
 0.99999999999925182E0 FRAC
-0.49999999997024012E0 FRAC
 0.041666666473384543E0 FRAC
-0.001388888418000423E0 FRAC
 0.000024801040648456E0 FRAC
-0.000000275246963843E0 FRAC
 0.000000001990785685E0 FRAC
;

\ Test sweeps through the circle to find the maximum error
\ On a typical PC, "20 WORST" tests a million points fast.

VARIABLE ERROR  VARIABLE SPAN
DFVARIABLE ANGLE
HEX
: TEST1  ( i -- n ) 20 SPAN @ - LSHIFT COS ;
: TEST2  ( i -- n ) S>F ANGLE DF@ F* FCOS 7FFFFFFF S>F F* F>S ;
DECIMAL
: SETANGLE ( span -- )
   1E0 FATAN 8E0 F* ( 2pi) 0 ?DO F2/ LOOP ANGLE DF! ;
: WORST  ( n -- u )  \ Brute force sweep for worst estimate
   DUP SPAN !  SETANGLE  0 ERROR !
   1 SPAN @ LSHIFT 0 DO
     I TEST1  I TEST2  - ABS
     ERROR @ MAX ERROR !
   LOOP
   ERROR @ ;
[THEN]

\ ***************************************************
\ From Gary Bergstom on comp.lang.forth - 16 Feb 2013
\ ***************************************************

\ I don't normally do 32bit so the below is not optimized.  The
\ polynomials should be minimax ones, like the originals from Ganssle.
\ But I didn't have the time - these are simple Taylor terms.  So I
\ included enough to make the error drop to a max of about 3 counts.
\ Chebyshev or minimax terms would probably be one term less with
\ similar accuracy.  The code should be rewritten with a call to the
\ polynomial evaluation mentioned above. This code runs somewhat
\ faster (~20%, timing the "worst" routine which means the sin/cos
\ routines run even faster) than the original.  Optimizing the
\ coefficients and implementing a polynomial evaluator should make
\ it go even faster.  It is larger but my work usually finds that
\ execution time is more valuable than code size.  I don't seem to
\ fill flash memory on my little embedded chips.

\ Jack Crenshaw's book should be on every embedded programmers shelf.
\ At least everyone who does more than adds and subtracts.

\ Just replace the similar code in Brad's post with that below. =20

\ Have fun.

\ Gary Bergstrom

\ Trig functions in 32-bit Forth
\ Angle input is 32-bit
\ Output range is 0 to +/-$7FFFFFFF, fractional 1.31 format
\ Accuracy is about 30-31 bits, worst case error is about 3 counts
\ See Jack W Crenshaw's book: Math Toolkit for Real-Time Programming
\ The polynomials below are simple Taylor terms (room for improvement)

HEX
  6487ED51 CONSTANT S1 295779CC NEGATE CONSTANT S3
  0519AF19 CONSTANT S5   4CB4B3 NEGATE CONSTANT S7
     2A0F0 CONSTANT S9      F18 NEGATE CONSTANT S11

  7FFFFFFF CONSTANT C0 4EF4F326 NEGATE CONSTANT C2
  103C1F08 CONSTANT C4  155D3C7 NEGATE CONSTANT C6
     F0FA8 CONSTANT C8     69B4 NEGATE CONSTANT C10

  40000000 Constant PI/2
  20000000 Constant PI/4
DECIMAL

: *. ( f1 f2 -- f=3Df1*f2 ) M* D2*  nip  ;
\ 80000000 0 D+  NIP ;  rounding doesn't seem to help

: <SIN> ( -PI/4<n<PI/4 -- sin<n> ) 2* DUP DUP *. >R
      S11 R@ *. S9 + R@ *. S7 + R@ *.  S5 + R@ *. S3 + R> *. S1 + *. 2* ;
: <COS> ( -PI/4<n<PI/4 -- cos<n> ) 2* DUP     *. >R
      C10 R@ *. C8 + R@ *. C6 + R@ *.  C4 + R@ *. C2 + R> *. 2*  C0 + ;

: SIN ( n -- sin<n> ) DUP PI/4 + DUP >R \ get segment, Jack's book
    $C0000000 AND -                     \ adjust value for segment
    R@ $40000000 AND                    \ the two bits, in number on R, select cos/sin and polarity
    IF  <COS> ELSE <SIN> THEN
    R> 0 < IF NEGATE THEN ;

: COS ( n -- COS<n> ) PI/2 + SIN ;

