\ AT49BV1614.FTH - Atmel AT49BV1614 Flash write utility

((
Copyright (c) 2002
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20021014 SFP001 First release.
))


\ ************
\ *! 29BV1614
\ *T AT49BV1614 Flash programming
\ ************

\ *P This code is used by the MPE boot loader to write to the Flash
\ ** for storing applications. This code is designed for use with
\ ** an Atmel 49BV1614 device on a 16 bit bus. The size of each
\ ** sector is given by the sector table.

\ *P This code is hardware and CPU dependent because of bus width
\ ** and cache considerations. This code was tested on an Atmel
\ ** EB55 board.

\ *P Care must be taken when modifying or using this code. The actual primitives
\ ** used to perform the write and erase operations are specifically written to
\ ** be executed from RAM or memory that is not in the Flash itself. The
\ ** Flash devices do not handle simultaneous read/write operation.
\ ** Because the flash may still present at 0000:0000, interrupts are
\ ** disabled during programming. The flash memory is copied into RAM
\ ** by the boot code and all code execution (except for interrupt
\ ** vector despatch) occurs from RAM.

\ *P The equate FLASHBASE must be set up before this file is
\ ** compiled. If it is not defined a base address of 0400:0000
\ ** will be used, ensuring that bit 26 of this address is set to
\ ** bypass the cache. Uncached addresses may be passed to the
\ ** programming routines as these force bit 26 of Flash addresses
\ ** to a '1' as required by the Samsung S3C4510.

\ *P Compiler and Interpreter macros must be set up before
\ ** this code is compiled to force an uncached address. Some
\ ** CPUs, e.g. MIPS and some ARMs use a particular address
\ ** range to determine whether the access should be cached
\ ** or uncached. During Flash programming, ALL accesses to
\ ** the flash should be uncached and write buffers should
\ ** be turned off.
\ ** The following is for for a Samsung S3C4510B CPU.
\ *E compiler
\ ** : ForceUncached		\ addr -- addr'
\ **   bit26 or  ;
\ ** target
\ ** interpreter
\ ** : ForceUncached		\ addr -- addr'
\ **   bit26 or  ;
\ ** target

\ *P If your CPU does not have a cache or if caching is
\ ** never enabled for Flash access, use the following code.
\ *E compiler
\ ** : ForceUncached	;	\ addr -- addr'
\ ** target
\ ** interpreter
\ ** : ForceUncached	;	\ addr -- addr'
\ ** target

\ *P Note that when programming these devices, programming part
\ ** of a sector (64kb) will leave the unprogrammed part
\ ** set to bytes of FF.

decimal


\ *************
\ *S Primitives
\ *************
\ *P By default, these words are compiled without heads to
\ ** reduce application size.

\ internal

create sectab	\ -- addr
\ *G Table containing the number of sectors and starting offset
\ ** of each sector, plus a dummy start address which enables
\ ** the size of the last sector to be calculated.
  #39 ,
  $00000000 ,  $00002000 ,  $00004000 ,  $00006000 ,
  $00008000 ,  $0000A000 ,  $0000C000 ,  $0000E000 ,
  $00010000 ,  $00020000 ,  $00030000 ,  $00040000 ,
  $00050000 ,  $00060000 ,  $00070000 ,  $00080000 ,
  $00090000 ,  $000A0000 ,  $000B0000 ,  $000C0000 ,
  $000D0000 ,  $000E0000 ,  $000F0000 ,  $00100000 ,
  $00110000 ,  $00120000 ,  $00130000 ,  $00140000 ,
  $00150000 ,  $00160000 ,  $00170000 ,  $00180000 ,
  $00190000 ,  $001A0000 ,  $001B0000 ,  $001C0000 ,
  $001D0000 ,  $001E0000 ,  $001F0000 ,  $00200000 ,

: SectorN	\ n -- addr len
\ *G Convert sector number to base address and length
  1+ cells sectab + 2@			\ -- end start
  tuck -
;

: FindSector	\ addr -- start len
\ *G Find the start and length of the sector containing address
\ ** addr. If len is zero, the sector was not found and start
\ ** is set to addr.
  flashbase - 0				\ convert to offset, assume failure
  sectab @ 0 ?do
    over i SectorN over + within if	\ test sector
      2drop i SectorN
      leave
    endif
  loop
  swap flashbase + swap			\ make offset absolute
;

: 100us		\ --
\ *G Wait 100 microseconds. A 50MHz cached CPU is assumed.
  #1600 0 do loop
;

: FlUnlock	\ --
\ *G Unlock the flash.
  $AAAA [ flashbase $5555 1 <<n + ] literal w!	\ unlock
  $5555 [ flashbase $2AAA 1 <<n + ] literal w!
;

: FlCmd		\ command --
\ *G Unlock the flash and write the command.
  $AA [ flashbase $5555 1 <<n + ] literal w!	\ unlock
  $55 [ flashbase $2AAA 1 <<n + ] literal w!
      [ flashbase $5555 1 <<n + ] literal w!	\ command
;

: Write16	\ w address -- flag ; 0 for success
\ *G Write a 16 bit word to the given absolute address.
  [I
  $A0A0 FlCmd				\ write command
  2dup w!				\ write data
  #10000 >r				\ counter
  begin
    2dup w@ <>  r@ 0<> and		\ while not same and not timed out
   while
    r> 1- >r				\ count down, go round again
  repeat
  2drop  r> 0=				\ success if counter non-zero
  I]
;

: FlReset	\ --
\ *G Reset device after AutoSelect command.
  $F0F0 [ flashbase $5555 1 <<n + ] literal w!	\ reset
;

: FlId		\ -- manid devid
\ *G Get manufacturer's ID, should be $20 $E2 for the ST M29F040B
\ ** device.
  [I
  $9090 FlCmd
  [ flashbase $000 + ] literal w@
  [ flashbase $002 + ] literal w@
  FlReset
  I]
;

: SecErase	\ secaddr -- ; erase sector
\ *G Erase the sector at the given address.
  [I
  $80 FlCmd  FlUnlock $30 over c!	\ Sector erase sequence
  dup >r  w@				\ first status read
  begin
    r@ w@ tuck xor $40 and		\ nz if DQ6 toggled
    0=
  until
  drop
  r> drop
  I]
;

[undefined] .lword [if]
: .lword	\ x --
\ *G Display x as an 8 digit hexadecimal number in the form
\ ** "abcd:efgh".
  base @ hex  swap
  0 <# # # # # ascii : hold # # # # #> type
  base !
;
[then]

[undefined] .word [if]
: .byte		\ w --
\ *G Display w as a 4 digit hexadecimal number in the form
\ ** "abcd".
  base @ hex  swap
  0 <# # # #> type
  base !
;
[then]

((
[undefined] .byte [if]
: .byte		\ x --
\ *G Display x as a 2 digit hexadecimal number in the form
\ ** "ab".
  base @ hex  swap
  0 <# # # #> type
  base !
;
[then]
))

: .src/dest	\ src dest -- src dest
\ *G Display the addresses and contents of the source
\ ** and destination.
  ."  source: " over .lword  over w@ ." (" .word ." )"
  ."  dest: " dup .lword  dup w@ ." (" .word ." )"
;


\ *************************
\ *S User/Application words
\ *************************
\ *P These words have heads by default and are in the target
\ ** dictionary.

\ external

: (EraseFlash)	\ dest dlen --
\ *G Erase the flash sectors covering dlen bytes at dest.
  ?dup if
    swap ForceUncached swap		\ *** force bypass of cache **
    over + 1- >r			\ -- dest ; R: -- end
    begin				\ -- addr ; R: -- end
      cr ."  Erasing sector at: " dup .lword
      dup SecErase			\ erase current sector
      FindSector +			\ step to start of next sector
      dup r@ u>				\ until out of range
    until
    drop  r> drop
  endif
;

: (ProgFlash)	\ src dest len --
\ *G Write len bytes from memory at src to Flash at dest.
\ ** Note that the Flash is assumed to be erased, and that
\ ** no verification is performed except when each byte is
\ ** programmed.
  swap ForceUncached swap		\ *** force bypass of cache **
  0 ?do
    over w@ over write16 if
      cr ."  Write failed:" .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    2 + swap  2 + swap
  2 +loop
  2drop
;

: (VerifyFlash)	\ src dest len --
\ *G Verify len bytes from memory at src to Flash at dest.
  swap ForceUncached swap		\ *** force bypass of cache **
  3dup s= if  3drop exit  endif
  cr ."  Verify failed"
  0 ?do
    over w@ over w@ <> if
      cr 2 spaces .src/dest
      key? ?leave
    then
    i $FFF and 0= if  ." ."  endif
    2 + swap  2 + swap
  2 +loop
  2drop
;

: ProgramFlash	\ src dest len --
\ *G Using the memory buffer at src, program the Flash at dest
\ ** with len bytes. The relevant Flash sectors are erased, the
\ ** Flash is programmed, and the result verified.
  cr ." Erasing " 2dup (EraseFlash)
  cr ." Programming "  3dup (ProgFlash)
  cr ." Verifying "  (VerifyFlash)
;


\ ******
\ *> ###
\ ******


