\ Morse code example for Olimex P2148 board

((
Copyright (c) 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ ===========
\ *! morsecode
\ *T Morse code example
\ ===========

\ =======
\ *S LEDs
\ =======
\ *P The LEDs are on P0.10 and P0.11.

0 10 pio: led1	\ -- struct
\ *G Defines LED1 as being attached to pin P0.10.
0 11 pio: led2	\ -- struct
\ *G Defines LED2 as being attached to pin P1.10.

: tled1		\ --
\ *G Test LED1.
  led1 isOutput
  begin
    led1 clrPin  200 ms
    led1 setPin  200 ms
  key? until
;

: tled2		\ --
\ *G Test LED2.
  led2 isOutput
  begin
    led2 clrPin  200 ms
    led2 setPin  200 ms
  key? until
;


\ *********
\ *S Buzzer
\ *********
\ *P The buzzer is driven by two GPIO lines to double the voltage
\ ** swing.

0 12 pio: buzz1	\ -- struct
\ *G Defines BUZZ1 as being attached to pin P0.12.
0 13 pio: buzz2	\ -- struct
\ *G Defines BUZZ2 as being attached to pin P0.13.

: initBuzz	\ --
\ *G Initialise buzzer.
  buzz1 isOutput  buzz2 isOutput
;

: termBuzz	\ --
\ *G Stop the buzzer.
  buzz1 isInput  buzz2 isInput
;

: b1hi		\ --
\ *G Set BUZZ1 high, BUZZ2 low.
  buzz1 setpin  buzz2 clrpin
;

: b1lo		\ --
\ *G Set BUZZ1 low, BUZZ2 high.
  buzz1 clrpin  buzz2 setpin
;

2 value BuzzDelay	\ -- ms
\ *G Number of milliseconds between switching the buzzer pins.

: buzz		\ ms --
\ *G Run the buzzer for the given number of milliseconds.
  initBuzz
  begin
    b1hi  BuzzDelay ms
    b1lo  BuzzDelay ms
    BuzzDelay 2* -  dup 0<
  until
  drop  termBuzz
;


\ *************
\ *S Morse code
\ *************

variable DotTime	\ -- addr
  100 DotTime !

: interpart	\ --
  DotTime @ ms
;

: dot		\ --
  DotTime @ buzz  interpart
;

: dash		\ --
  DotTime @ 3 * buzz  interpart
;

: mletter	\ caddr len --
  bounds ?do
    i c@ [char] . =
    if  dot  else  dash  then
  loop
  DotTime @ 2* ms			\ already one time in dot/dash
;


\ ******************************
\ *S Building a Morse code table - 1
\ ******************************
\ *P The Morse code entries are held as a list of ASCII characters
\ ** and Morse strings.
\ *E   char A c,  ", .-"
\ **   char B c,  ", -..."
\ **   ...
\ **   0 c, 0 c,

\ *P We use a bit of syntactic sugar to make this more readable:
\ *[
interpreter
: morse:	\ -- ; use as MORSE A .-
  char c,  bl word count $,
;
target

create MorseTable	\ -- addr
  morse: A   .-       morse: U   ..-
  morse: B   -...     morse: V   ...-
  morse: C   -.-.     morse: W   .--
  morse: D   -..      morse: X   -..-
  morse: E   .        morse: Y   -.--
                      morse: Z   --..
  morse: F   ..-.
  morse: G   --.      morse: 0   -----
  morse: H   ....     morse: 1   .----
  morse: I   ..       morse: 2   ..---
  morse: J   .---     morse: 3   ...--
                      morse: 4   ....-
  morse: K   -.-      morse: 5   .....
  morse: L   .-..     morse: 6   -....
  morse: M   --       morse: 7   --...
  morse: N   -.       morse: 8   ---..
  morse: O   ---      morse: 9   ----.

  morse: P   .--.     morse: .   .-.-.-
  morse: Q   --.-     morse: ,   --..--
  morse: R   .-.      morse: ?   ..--..
  morse: S   ...      morse: -   -....-
  morse: T   -        morse: '   .----.
                      morse: :   ---...
  morse: U   ..-      morse: "   .-..-.
  morse: V   ...-     morse: /   -..-.
  morse: W   .--      morse: @   .--.-.
  morse: X   -..-
  morse: Y   -.--
  morse: Z   --..
  0 c,  0 c,				\ end of table
\ *]

: FindMchar	\ char -- caddr len true | 0
\ Look up a character in the table. Lower case characters are
\ converted to upper case.
  upc
  MorseTable begin			\ -- char table
    dup c@				\ ends on a zero
   while
    2dup c@ = if			\ match
      nip  1+ count true exit
    endif
    1+ count +				\ step forward
  repeat
  2drop 0				\ failed
;

: emitMorse	\ char --
\ Transmit a Morse character. Invalid characters form an inter-word
\ gap. This allows us to put spaces in text.
  findMchar if
    mletter
  else
    DotTime @ 7 * ms
  endif
;

: sendMorse	\ caddr len --
\ Send a string as Morse code.
  bounds ?do  i c@ emitMorse  loop
;


\ ******************************
\ *S Building a Morse code table - 2
\ ******************************
\ *P We need to build a device that will take in ordinary
\ ** text and will output the corresponding Morse code.
\ ** We make a list of the defined Morse characters.
\ ** Each entry in the list consists of three bytes
\ *E Byte 0 - ASCII character
\ ** Byte 1 - count of dots/dashes
\ ** if byte 1 is non-zero
\ **   Byte 2=bitmask, 0=dot, 1=dash.
\ **   bits are processed most significant first.
\ ** if byte 1 is 0
\ **   byte 2 contains a delay in dot times.

\ *[

interpreter
: pattern	\ caddr len -- pattern
  $80 0 2swap				\ bit pattern
  bounds ?do				\ -- mask pattern
    i c@ [char] - =
    if  over or  then
    swap 1 rshift swap			\ next mask
  loop
  nip
;

: morse:	\ -- ; use as MORSE A .-
  char c,  bl word count dup c,  pattern c,
;
target

create MorseTable2	\ -- addr
  morse: A   .-       morse: U   ..-
  morse: B   -...     morse: V   ...-
  morse: C   -.-.     morse: W   .--
  morse: D   -..      morse: X   -..-
  morse: E   .        morse: Y   -.--
                      morse: Z   --..
  morse: F   ..-.
  morse: G   --.      morse: 0   -----
  morse: H   ....     morse: 1   .----
  morse: I   ..       morse: 2   ..---
  morse: J   .---     morse: 3   ...--
                      morse: 4   ....-
  morse: K   -.-      morse: 5   .....
  morse: L   .-..     morse: 6   -....
  morse: M   --       morse: 7   --...
  morse: N   -.       morse: 8   ---..
  morse: O   ---      morse: 9   ----.

  morse: P   .--.     morse: .   .-.-.-
  morse: Q   --.-     morse: ,   --..--
  morse: R   .-.      morse: ?   ..--..
  morse: S   ...      morse: -   -....-
  morse: T   -        morse: '   .----.
                      morse: :   ---...
  morse: U   ..-      morse: "   .-..-.
  morse: V   ...-     morse: /   -..-.
  morse: W   .--      morse: @   .--.-.
  morse: X   -..-
  morse: Y   -.--
  morse: Z   --..
  0 c,  0 c,				\ end of table
\ *]

: FindPchar	\ char -- pattern len true | 0
\ Look up a character in the table. Lower case characters are
\ converted to upper case.
  upc
  MorseTable2 begin			\ -- char table
    dup c@				\ ends on a zero
   while
    2dup c@ = if			\ match
      nip
      dup 2+ c@  swap 1+ c@ true
      exit
    endif
    3 +  				\ step forward
  repeat
  2drop 0				\ failed
;

: soundPattern	\ pattern len --
  $80 -rot
  0 ?do					\ -- mask pattern
    2dup and
    if  dash  else  dot then
    1 lshift
  loop
  2drop
;

: emitMorseP	\ char --
\ Transmit a Morse character. Invalid characters form an inter-word
\ gap. This allows us to put spaces in text.
  findPchar if
    soundPattern
  else
    DotTime @ 7 * ms
  endif
;

: sendMorseP	\ caddr len --
\ Send a string as Morse code.
  bounds ?do  i c@ emitMorseP  loop
;



\ ======
\ *> ###
\ ======

decimal
