\ NetCode.fth - Network operations

((
Copyright (c) 2002, 2003, 2007, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20100406 SFP004 Added Cortex-M3 specifics.
20070227 SFP003 Added pseudoheader checksum routine.
		Removed [CK and CK] from compilation.
20030621 SFP002 Recoded @(N) and !(N)
20030323 SFP001 Added interpreted W,(N) and ,(N)
))


\ **********
\ *! netcode
\ *T Network operations - CPU specific
\ **********
\ *P The file NETCODE.FTH contains code that will show
\ ** significant benefit from being coded, or is sensitive
\ ** to the byte order of the underlying CPU.

\ *P The code in this file is used by PowerNet v3 onwards,
\ ** and should only be compiled once if both PowerNet and
\ ** NETBOOT.FTH are both compiled.


\ ****************************************
\ *S Network order (big-endian) operations
\ ****************************************
\ *P Note that these functions have to be capable of fetching
\ ** 32 bit cells from 16 bit aligned addresses, not just from
\ ** 32 bit aligned addresses.

\ *P Note also that these routines assume a byte-addressed CPU.


[undefined] @(n) [if]	\ ** In case of coded definitions

little-endian? [if]
: w@(n)         \ addr -- u16
\ *G Network order 16 bit fetch.
  dup c@ 8 lshift swap 1+ c@ or
;

: w!(n)         \ u16 addr --
\ *G Network order 16 bit store.
\ ** Must write MSB first if writing to some hardware
  over 8 rshift  over c!	\ high byte
  1+ c!				\ low byte
\  over ( $FF and ) over 1+ c!
\  swap 8 rshift swap c!
;

interpreter
: w@(n)         \ addr -- u16
  dup c@ 8 lshift swap 1+ c@ or
;
: w!(n)         \ u16 addr --
  over $FF and over 1+ c!
  swap 8 rshift swap c!
;
target
[else]
compiler
: w@(n)		w@  ;
: w!(n)		w!  ;
target
interpreter
: w@(n)		w@  ;
: w!(n)		w!  ;
target
: w@(n)		w@  ;
: w!(n)		w!  ;
[then]

: @(n)          \ addr -- u32
\ *G Network order 32 bit fetch.
  dup c@ 8 lshift
  over 1 + c@ or 8 lshift
  over 2 + c@ or 8 lshift
  swap 3 + c@ or
;

: !(n)		\ u32 addr --
\ *G Network order 32 bit store.
  2dup 3 + c!  swap 8 rshift swap
  2dup 2 + c!  swap 8 rshift swap
  2dup 1 + c!  swap 8 rshift swap
  c!
;
interpreter
: @(n)          \ addr -- u32
  dup w@(n) #16 lshift swap 2 + w@(n) or
;

: !(n)		\ u32 addr --
  over $FFFF and over 2 + w!(n)
  swap #16 rshift swap w!(n)
;
target

[then]			\ ** In case of coded definitions

interpreter
: w,(n)		\ w --
\ *G Network order W,
  here w!(n)  2 allot
;

: ,(n)		\ x --
\ *G Network order version of , (comma).
  here !(n)  cell allot
;
target
XC? 0= [if]
: w,(n)		\ w --
\ *G Network order W,
  here w!(n)  2 allot
;

: ,(n)		\ x --
\ *G Network order version of , (comma).
  here !(n)  cell allot
;
[then]


\ ********************
\ *S Internet checksum
\ ********************
\ *P Because this code is used so much in the system, it is
\ ** worth coding it for maximum efficiency. One of the
\ ** following equates may be defined (any value) in the control
\ ** file to compile CPU specific code for the Internet
\ ** checksum:
\ *D CPU=Cortex    Cortex v7, e.g. M3, but not Cortex-M0.
\ *D CPU=ARM       All ARM variants using the 32 bit ISA
\ *D CPU=H8/300H   Hitachi H8/300H and H8S
\ *D CPU=68K       Motorola 68xxx CPUs
\ *D CPU=386       Intel 386 and above (i32)
\ *P If no CPU specific equate is defined, the high level code
\ ** will be used.

\ *P If the checksum words have been defined previously, this section
\ ** of code is ignored.

[undefined] (CKSUM) [if]	\ ** start of CPU specific code

[defined] CPU=Cortex [if]
\ *G If this equate is defined (any value will do), a Cortex
\ ** (not M0) machine code IP checksum routine will be used.
code (cksum)	\ cksum addr len -- cksum'
\ *G ARM IP Checksum routine.
\ ** This word is safe against alignment issues as the network
\ ** order word fetch is done as two byte fetches.
        ldr      r0, [ psp ], # 4	\ r0 = address
        add .s   r1, tos, r0		\ r1 = addr+len-1
        sub .s   r1, r1, # 1
        ldr      tos, [ psp ], # 4	\ tos = checksum-acc
        mov      r3, # $FFFF		\ r3 = $FFFF
l: ckl2
        cmp      r0, r1
        itt .le
          ldrb     r2, [ r0 ], # 1
          add      tos, tos, r2 .lsl # 8
        itt .lt
          ldrb     r2, [ r0 ], # 1
          add      tos, tos, r2
        itt .le
          add      tos, tos, tos .lsr # #16
          and      tos, tos, r3
        b .lt    ckl2
        next,
end-code
[then]

[defined] CPU=ARM [if]
\ *G If this equate is defined (any value will do), an ARM
\ ** machine code IP checksum routine will be used.
code (cksum)	\ cksum addr len -- cksum'
\ *G ARM IP Checksum routine.
\ ** This word is safe against alignment issues as the network
\ ** order word fetch is done as two byte fetches.
        ldmfd    r12 ! { r0 }           \ r0 = address
        add      r1, r10, r0            \ r1 = end-address-1
        sub      r1, r1, # 1
        ldmfd    r12 ! { r10 }		\ r10 = checksum-acc
        mvn      r3, # 0                \ r3 = $FFFF
        mov      r3, r3, lsr # #16
l: ckl2
        cmp      r0, r1
        ldrb .le r2, [ r0 ], # 1
        add  .le r10, r10, r2 lsl # 8
        ldrb .lt r2, [ r0 ], # 1
        add  .lt r10, r10, r2
        add  .le r10, r10, r10 lsr # #16
        and  .le r10, r10, r3
        b    .lt ckl2
        next,
end-code
[then]

[defined] CPU=H8/300H [if]
\ *G If this equate is defined (any value will do), an
\ ** H8/300H or H8S machine code IP checksum routine will be used.
code (cksum)	\ cksum addr len -- cksum'
\ Start by summing 16 bit words
        mov.l   er5     @-er7
        mov.l   @sp+    er0             \ ER0 = address pointer
        mov.l   @sp+    er2             \ er2 = checksum
        mov.l   tos     er1             \ er1 = byte length
        shlr.l  er1                     \ er1 = word length

        sub.l   er5     er5             \ clear curr word register
l: clp
        mov.w   @er0+   r5
        add.l   er5     er2
        dec.l   # 1     er1
        bne     clp

        mov.l   tos     er1             \ ER0 = addr, ER1 = len
        and.l   # 1     er1             \ check for odd byte reqd.
        beq     dv
        sub.l   er5     er5
        mov.b   @er0    r5h
        add.l   er5     er2
\ got 32bit summation of 16 bit words in ER2
\ now need to bring down to 16 bit
l: dv
        cmp.l   # #65535 er2
        ble     xdone
        mov.l   er2     tos
        mov.w   tos.wh  tos.wl
        and.l   # #65535 er2
        and.l   # #65535 tos
        add.l   tos     er2
        bra     dv

l: xdone                                \ now make 1's complement 16bit
        mov     er2     tos             \ return sum for now

        mov.l   @er7+   er5
        next,
end-code
[then]

[defined] CPU=68K [defined] CPU=Coldfire  or [if]
\ *G If this equate is defined (any value will do), a
\ ** 68xxx/Coldfire machine code IP checksum routine will be used.
code (cksum)	\ cksum addr len -- cksum'
        d7 d5 l. move,			\ D5 = len
        # 1 d5 l. andi,			\ odd/even indicator
        # #65534 d7 l. andi,		\ byte len clipped to words

        (a6)+ a2 movea,                 \ a2 = start add
        a2 a3 movea,                    \ a3 = end add
        d7 a3 l. adda,                  \ a3 = END ADD
        (a6)+ d7 move,                  \ d7 = cksum
        # 0 d6 moveq,
l: clp                                  \ sum 16 bit words into D7.l
        (a2)+ d6 w. move,
        d6 d7 l. add,
        a2 a3 cmpa,
        clp bne,

        # 1 d5 l. cmpi,                 \ no odd byte
        dv bne,
        # 0 d6 moveq,
        (a2)+ d6 b. move,
        # 8 d6 lsl,
        # 65280 d6 l. andi,
        d6 d7 l. add,

l: dv                                   \ turn 32bit into 16 bit
        # #65535 d7 l. cmpi,            \ using succesive shift and add
        xdone ble,
        d7 d6 l. move,
        # 8 d6 lsr,
        # 8 d6 lsr,
        # #65535 d6 l. andi,
        # #65535 d7 l. andi,
        d6 d7 l. add,
        dv bra,

l: xdone
        # 0 d0 moveq,			\ restore A3=0
        d0 a3 l. movea,
        next,
end-code
[then]

((
[defined] CPU=386 [if]
\ *G If this equate is defined (any value will do), a
\ ** i386+ machine code IP checksum routine will be used.
\ ** N.B. This code has not been tested!
code (cksum)	\ cksum addr len -- cksum'
        push    esi

        mov     esi, 0 [ebp]		\ ESI = addr
        mov     ecx, ebx		\ ECX = len
        mov     ebx, 4 [ebp]		\ EBX = checksum
        add     ebp, # 8		\ stack update
        ( esi = source, ecx = length, ebx = acc )

        mov     edx, ecx
        and     edx, # 1		\ EDX = odd/even indicator
        and     ecx, $FFFFFFFE		\ clip to words
        add     ecx, esi		\ end address
        ( esi = source, ecx = end, edx = odd?, ebx = acc )

        xor     eax, eax
l$1:
        mov     ax, 0 [esi]		\ little endian fetch
        add     esi, # 2
	xchg    al, ah			\ make big-endian
        shl     eax, # #16
        add     ebx, eax
        jnc     l$2
        add     ebx, # $10000
l$2:
        cmp     ecx, esi
        jne     l$1

        cmp     edx, # 0
        je      l$3

        mov     al, 0 [esi]
        shl     eax, # #24
        and     eax, # $FF000000
        add     ebx, eax
        jne     l$3
        add     ebx, # $10000
l$3:
        shr     ebx, # #16
\        xchg    bl, bh

        pop     esi
        next,
end-code
[then]
))
[then]			\ ** end of CPU specific code

[undefined] (CKSUM) [if]
: (cksum)	\ cksum addr len -- cksum'
\ *G Compute the partial internet checksum of a memory area.
\ ** This is a high level implementation that can be used by
\ ** any CPU.
\ ** The word (CKSUM) is useful when calculating header checksums, and
\ ** avoids the need to create the pseudoheaders described in
\ ** the TCP/IP literature.
  swap 2- swap				\ start address 2- so we can inc it more efficiently
  begin
    dup 1 >				\ check if more than 1 byte left
  while
    2- >r				\ dec & shift the length out of the way
    2+ tuck w@(n)			\ get 16 bits of data
    +					\ and add it into cksum
    dup $80000000 and if		\ check top bit
      dup $0FFFF and			\ get low 16 bits
      swap #16 rshift +			\ and add in the high 16 bits
    endif
    swap 				\ re-order values
    r>					\ retrieve length
  repeat
  if					\ odd byte left
    2+ c@ #8 lshift +
    dup $80000000 and if		\ check top bit
      dup $0FFFF and			\ get low 16 bits
      swap #16 rshift +			\ and add in the high 16 bits
    endif
  else					\ all done
    drop
  endif
  begin
    dup $FFFF0000 and
  while
    dup $0FFFF and			\ get low 16 bits
    swap #16 rshift +			\ and add in the high 16 bits
  repeat
;
[then]

[undefined] CKSUM [if]
: cksum		\ addr len -- cksum ; computed as 16 bit ints
\ *G Compute the internet checksum of a memory area. This a high level
\ ** implementation that can be used by any CPU.
  0 -rot				\ initial value of checksum
  (cksum)
  invert  $FFFF and			\ invert and clip to 16 bits
;
[then]


((
STRUCT /pseudohdr	\ -- size
\ *G The pseudoheader structure for checksum calculation. This
\ ** word is for illustration and is *\b{not} compiled.
  4 field ph_iploc			\ source IP address
  4 field ph_iprem			\ dest IP address
  2 field ph_prot			\ protocol
  2 field ph_pklen			\ packet length
END-STRUCT
))

: PHcksum	\ *hdr len ipsrc ipdest ipproto -- cksum
\ *G A generic TCP/UDP checksum function. The parameters supplied
\ ** are used to form a 'pseudoheader' from which the acual
\ ** checksum is generated or checked.
  { | ph[ #12 ] -- }
  ph[ 8 + w!(n)     			\ protocol -> pseudohdr
  ph[ 0 + !(n)	         	 	\ src ip -> pseudohdr
  ph[ 4 + !(n)		        	\ dest ip -> pseudohdr
  dup ph[ #10 + w!(n)			\ packet length -> pseudohdr
  0 ph[ #12 (cksum)			\ start cksum over pseudoheader
  -rot (cksum)				\ checksum over the header + data
  invert  $FFFF and			\ invert and clip to 16 bits
;


\ ======
\ *> ###
\ ======

decimal

