\ UnixTime.fth - Unix style time based on LSECONDS ticker.

((
Copyright (c) 2003, 2008
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20080510 MGS002 Corrected >DATE.
20080508 MPE001 Refactored in places for use with NTP seconds.
))


\ ===========
\ *! unixtime
\ *T Unix-style time handling
\ ===========
\ *P Unix time handling is based on a 32 bit seconds counter,
\ ** in this case held as a variable called *\fo{LSECONDS}. All
\ ** calculations derive from *\fo{LSECONDS}. Second 0 is midnight
\ ** at the start of 1 Mar 1995. This will roll over in 2131.
\ ** If longer timekeeping is required, extend the system to use
\ ** 64 bit *\fo{XSECONDS} where *\fo{LSECONDS} contains the low
\ ** 32 bits. It is the user's responsibility to update *\fo{LSECONDS}.

\ *P Since many RTCs are attached by I2C or similar serial
\ ** connections, we suggest that the hardware either use
\ ** a one second interrupt from the RTC after the initial read,
\ ** or periodically resynchronise the ticker interrupt and derive
\ ** *\fo{LSECONDS} from the ticker interrupt.

\ *P If you are using this code with NTP timestamps, these are
\ ** based from 1 Jan 1900, which has an *\fo{LSECONDS} value of
\ ** 1291876096+86400=1291962496, so add 1291962496 to the NTP
\ ** seconds value to produce a Unix time value.

\ *P If you are planning a product which needs to cope with dates
\ ** beyond 2035, note the following from RFC4330. The implication
\ ** is that for such products, you will have to modify the words
\ ** that convert between seconds and day numbers. However, the
\ ** values based at 1 Mar 1995 will not roll over until 2131.

\ *Q As the NTP timestamp format has been in use for over 20 years, it
\ ** is possible that it will be in use 32 years from now, when the
\ ** seconds field overflows.  As it is probably inappropriate to
\ ** archive NTP timestamps before bit 0 was set in 1968, a convenient
\ ** way to extend the useful life of NTP timestamps is the following
\ ** convention: If bit 0 is set, the UTC time is in the range 1968-
\ ** 2036, and UTC time is reckoned from 0h 0m 0s UTC on 1 January
\ ** 1900.  If bit 0 is not set, the time is in the range 2036-2104 and
\ ** UTC time is reckoned from 6h 28m 16s UTC on 7 February 2036.  Note
\ ** that when calculating the correspondence, 2000 is a leap year, and
\ ** leap seconds are not included in the reckoning.



\ ****************
\ *S Date handling
\ ****************
\ *P These routines are taken from the Julian calendar in
\ ** *\i{EXAMPLES\CALENDAR.FTH}. They handle dates by converting
\ ** them to a single precision Julian day number. Day 0 is
\ ** 1 March 1900.

: date>         \ day month year -- jday ; see month codes
\ Given day month and year generate the day number.
  dup 99 > if 1900 - then >r            \ year<100: XXth century
  dup 3 < if 12 + r> 1- >r then    	\ year origin -> March 1
  1+ 153 5 */ 123 - +                   \ compute m/d/y - 3/1/y
  r@ 365 um* drop + r> 4 / +            \ and add 3/1/y - 3/1/0
;

: weekday       \ jday -- n ; 0=Sunday
\ *G Given a Julian day return the day of week number where
\ ** 0 is Sunday, 1 is Monday and so on.
  4 + 0 7 um/mod drop
;

(( requires floored division
: >date         \ jday -- day month year
  \ *G Convert a Julian day number to day month year format.
  0 1461 um/mod                         \ number of leap-year periods
  4 * >r                                \ convert to number of years
  365 /mod                              \ number of years and days
  dup 4 = if 2drop 365 3 then           \ correct if Febr.29th
  r> + >r                               \ -- days | years
  123 + dup 5 153 */                    \ compute month
  dup >r 153 5 */ - r>                  \ compute day
  1- dup 12 > if 12 - r> 1+ >r then     \ correct month and year
  r> 1900 +                        	\ -- day month year
;
))

\ works with symmetric division
: >date		\ jday -- day month year
\ *G Convert a Julian day number to day month year format.
   1461 /mod				\ number of leap-year periods
   4 * >r				\ convert to number of years
   365 /mod				\ number of years and days
   dup 4 = if 2drop 365 3 then		\ correct if Feb 29th
   r> + >r				\ -- days | years
   123 + dup 5 153 */			\ compute month
   dup >r 153 5 */ - dup 0=		\ if 31 day month
   if  31 +  -1 else 0 then r> +	\ compute day
   1- dup 12 > if 12 - r> 1+ >r then	\ correct month and year
   r> 1900 +				\ -- day month year
;


\ ************************
\ *S Time/Date conversions
\ ************************

#24 #60 * #60 * equ secs/day	\ -- 86400
\ *G Seconds per day.
#60 #60 * equ secs/hr		\ -- 3600
\ *G Seconds per hour
#60 equ secs/min		\ -- 60
\ *G Seconds per minute
#60 equ mins/hr			\ -- 60
\ *G Minutes per hour
#24 equ hrs/day			\ -- 24
\ *G Hours per day
#34698 equ uday0		\ -- 34698
\ *G The Julian day number for Unix day 0 at 1 Mar 1995.

compiler
: j>uday	\ jday -- uday
\ *G Convert a Julian day based at 0 for 1 March 1900 to a Unix
\ ** day number derived from *\fo{LSECONDS}.
  uday0 -
;

: u>jday	\ jday -- uday
\ *G Convert a Unix day number as used by *\fo{LSECONDS} to a Julian
\ ** day based at 0 for 1 March 1900.
  uday0 +
;
target

[undefined] /umod [if]
: /umod		\ u1 u2 -- urem uquot
\ *G Single precision division of unsigned U1 by U2 returning
\ ** both remainder and quotient.
  0 swap um/mod
;
[then]

: secs>td	\ lsecs -- sec min hr day mon yr
\ *G Convert a Unix seconds value to time and date.
  secs/min /umod
  mins/hr /umod
  hrs/day /umod				\ -- sec min hr day#
  u>jday >date				\ -- sec min hr day mon yr
;

: td>secs	\ sec min hr day mon yr -- secs
\ *G Use the given values for time and date, and return the Unix
\ ** *\fo{LSECONDS} value.
  date> j>uday secs/day *		\ -- sec min hr lsecs
  swap secs/hr * +
  swap secs/min * +
  +
;

: time&date	\ -- sec min hr day mon yr ; 10.6.2.2292
\ *G The ANS Forth word to return time and date.
  lseconds @ secs>td
;

: dow		\ -- dow ; 0=Sunday
\ *G Return the day-of-week value last read from the RTC.
\ ** Sunday returns 0, Monday returns 1 and so on.
  lseconds @ 0 secs/day um/mod nip	\ -- uday
  u>jday weekday
;

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set *\fo{LSECONDS}
\ ** and any underlying RTC hardware.
  td>secs lseconds !
;


\ ************************
\ *S Time and date display
\ ************************
\ *P These functions rely on the ANS Forth word *\fo{TIME&DATE}
\ ** ( -- s m h dd mm yy ) and the non-standard *\fo{DOW}
\ ** ( -- dow, 0=Sun) to get the day of the week.

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow		\ --
\ *G Display day of week.
  dow 3 * days$ 1+ + 3 type
;

create months	\ -- addr
\ *G String containing 3 character text for the months.
  z", JanFebMarAprMayJunJulAugSepOctNovDec"

: .2r		\ n --
\ *G Display n as a two digit number with leading zeros.
  s>d <# # # #> type
;

: .4r		\ n --
\ *G Display n as a four digit number with leading zeros.
  s>d <# # # # # #> type
;

: .Time/Date	\ --
\ *G Display the system time.
\ ** The format is:
\ *C   hh:mm:ss dd Mmm yyyy
  time&date				\ -- s m h dd mm yy
  dup #100 <				\ correct two digit year clocks
  if  #2000 +  endif
  rot .2r space				\ day - generate date
  swap 1- 3 * months + 3 type space	\ month
  .4r space				\ year
  .2r ." :"			 	\ hours - generate time
  .2r ." :"				\ minutes
  .2r					\ seconds
;

: .zone		\ zone --
  if  3 spaces  else  ." GMT"  endif
;

: .AnsiDate	\ zone --
\ *G Display the day of week, time and date. If zone is 0
\ ** GMT is displayed. The format is Internet time:
\ *C   "dow, hh:mm:ss dd Mmm yyyy [GMT] "
  .dow ." , " .time/date space .zone space
;


\ ======
\ *> ###
\ ======

