\ DS1307drv.fth - DS1307 RTC driver

((
Copyright (c) 2002
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)2380 631441
fax: +44 (0)2380 339691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com
))

decimal

\ *! ds1307
\ *T Dallas DS1307 RTC driver
\ ===============
\ *S Introduction
\ ===============
\ *P The file DRIVERS\DS1307.FTH may be used with the I2C
\ ** drivers in the COMMON directory. Support is provided
\ ** for the ANS word TIME&DATE. The day of week format uses
\ ** 0=Sunday which is compatible with the file
\ ** EXAMPLES\CALENDAR.FTH for date handling.

\ *P The code for this RTC was tested on a Hitachi
\ ** evaluation board for the H8S/2148AF.


\ ====================
\ *S DS1307 primitives
\ ====================

$0D0 constant DS1307_id	\ -- $D0 ; 
\ *G The base device address (devaddr) of a DS1307. Only one
\ ** device may be connected.

: selwrite	\ reg --
\ *G Select a write to register reg.
  DS1307_id write_byte_i2c  write_byte_i2C
;

: selread	\ reg --
\ *G Select an RTC register for reading and start the transaction,
\ ** which must be terminated by LASTREAD.
  +ack_i2c  start_i2c selwrite stop_i2c	\ set reg
  100 0 do loop
  start_i2c  DS1307_id $01 or write_byte_i2c
;

: lastread	\ -- b
\ *G Perform the last read of the RTC and issue a STOP.
  -ack_i2c  read_byte_i2c  stop_i2c
;

: rtc!		\ b reg --
\ *G Write a byte to register reg.
  +ack_i2c start_i2c
  selwrite write_byte_i2c
  stop_i2c
;

: rtc@		\ reg -- b
\ *G Read a byte from register reg.
  selread lastread
;

: >rtc		\ addr reg len --
\ *G Write len bytes at addr to the RTC starting at register reg.
  +ack_i2c start_i2c
  swap selwrite				\ -- addr len; set register
  bounds				\ copy all bytes
  ?do  i c@ write_byte_i2c  loop
  stop_i2c
;

: rtc>		\ reg addr len --
\ *G Read len bytes from the RTC starting at register reg to addr.
  dup 0= if  drop 2drop  exit  endif	\ sanity check
  rot selread				\ -- addr len ; set reg
  1- 0 ?do				\ -- addr
    read_byte_i2c over c!  1+		\ -- addr+1
  loop
  lastread swap c!
;


\ ***************************
\ *S Time and date primitives
\ ***************************

: bcd>bin	\ bcd -- b
\ *G Convert a two digit BCD value to binary.
  dup $0F and				\ low nibble
  swap 4 rshift #10 *			\ high nibble
  +					\ merge
;

: bin>bcd	\ b -- bcd
\ *G convert a binary value to a two digit BCD value.
  0 #10 um/mod				\ -- rem quot
  4 lshift or				\ merge
;

8 buffer: RTCbuff	\ -- addr
\ *G Holds the data read from the RTC, and converted to
\ ** binary values.

: GetTD		\ --
\ *G Get time and date raw data from the RTC, and convert
\ ** it to binary values.
  0 RTCbuff 8 rtc>
  RTCbuff 7 bounds
  do  i c@ bcd>bin i c!  loop
;

: SetTD		\ --
\ *G Get time and date binary data from RTCbuff, convert
\ ** it to BCD values and write it to the RTC. The control
\ ** register is not written. A side effect is to start the
\ ** RTC if the seconds value is less than 80, otherwise it
\ ** will be halted.
  RTCbuff 7 bounds
  do  i c@ bin>bcd i c!  loop
  RTCbuff 0 7 >rtc
;


\ ************************
\ *G Application interface
\ ************************

: init_RTC	\ --
\ *G Clear the Clock Halt bit, enable the 1Hz square wave
\ ** output, and select 24 hour operation.
  0 rtc@  $7F and  0 rtc!		\ clear bit 7 of reg 0
  $90 7 rtc!				\ OUT=1, SQWE=1, RS1:0=0
  2 rtc@  bit6 invert and  2 rtc!	\ HRS.6=0
;

: time		\ -- sec min hr
\ *G Return the last read time. Sec is the second
\ ** {0...59}, min is the minute {0...59}, hr is the hour.
  RTCbuff
  dup c@  over 1+ c@  rot 2+ c@
;

: date		\ -- day mon yr
\ *G Return the last read date. Day is the day {1...31},
\ ** mon is the month {1...12}, and yr is the year (e.g., 1991).
  RTCbuff 4 +
  dup c@  over 1+ c@  rot 2+ c@ #2000 +
;

: time&date	\ -- sec min hr day mon yr ; 10.6.2.2292 
\ *G The ANS Forth word to return time and date.
  getTD time date
;

: settime&date	\ sec min hr day mon yr --
\ *G Use the given values for time and date, and set the RTC.
  #100 mod  RTCbuff tuck 6 + c!		\ year
  tuck 5 + c!				\ month
  tuck 4 + c!				\ day
  tuck 2 + c!				\ hour
  tuck 1 + c!				\ min
  c!					\ sec
  SetTD					\ write to chip
;

: dow@		\ -- dow ; 0=Sunday
\ *G Return the day-of-week value last read from the RTC.
\ ** Sunday returns 0, Monday returns 1 and so on.
  RTCbuff 3 + c@ 1-
;

: dow!		\ dow -- ; 0=Sunday
\ *G Set the day-of-week value in the RTC buffer. It will be
\ ** written to the RTC when SETTIME&DATE is next executed.
\ ** Set 0 for Sunday, 1 for Monday and so on.
  1+ RTCbuff 3 + c!
;

$3F equ lastRAM	\ -- rtcaddr

: ?init_RTC	\ --
\ *G Initialise the RTC if this has not already been done, as
\ ** indicated by the last location in RTC RAM holding the I character.
  lastRAM rtc@ [char] I <> if
    init_RTC
    [char] I lastRAM rtc!
  endif
;


\ ***********************
\ *S Time and date output
\ ***********************

: .2d		\ u --
\ *G Display u as two digits with leading zero.
  0 <# # # #> type
;

: .time		\ --
\ *G Display time from the RTC buffer.
  RTCbuff
  dup 2 + c@ .2d  ." :"
  dup 1 + c@ .2d  ." :"
  c@ .2d
;

create months$	\ -- addr
\ *G String containing 3 character text for the months.
  ", JanFebMarAprMayJunJulAugSepOctNovDec"

: .date		\ --
\ *G Display date from the RTC buffer.
  RTCbuff
  dup 4 + c@ 2 .r  space
  dup 5 + c@
  1- 3 * months$ 1+ + 3 type space
  6 + c@ #2000 + 4 .r
;

create days$	\ -- addr
\ *G String containing 3 character text for the days of the week.
  ", SunMonTueWedThuFriSat"

: .dow		\ --
\ *G Display day of week from the RTC buffer.
  dow@ 3 * days$ 1+ + 3 type
;

: .time&date	\ --
\ *G Display the day of week, date and time.
  .dow space .date space .time
;


\ ************
\ *S test code
\ ************

TestI2C? [if]

: tdump		\ addr len --
\ *G DUMP a region of RTC registers/memory.
  BASE @ -rot  HEX
  bounds ?do				\ get limits
    CR  i .WORD  i $10 bounds		\ print address
    do  space  i rtc@ .BYTE  loop	\ print byte
    2 SPACES  I $10 bounds		\ gap
    DO  i rtc@ .ASCII  LOOP		\ ascii o/p
    key? ?LEAVE				\ quit or halt
  $10 +LOOP				\ until done
  BASE !  CR				\ restore base
;

: .rtc		\ --
\ *G Display all the registers and RAM of the RTC.
  0 #64 tdump
;

: tfill  	\ c-addr u char -- 
\ *G FILL a region of RTC registers/memory.
  -rot  bounds
  ?do  dup i rtc!  loop
  drop
;

: tw		\ --
\ *G Write 0 to register 7 repeatedly until a key is pressed.
  start-timers  init_I2C
  begin
    #10 ms
    0 7 rtc!
  key? until ;

: tr
\ *G Read register 7 repeatedly until a key is pressed.
  start-timers  init_I2C
  begin
    #10 ms
    7 rtc@ drop
  key? until ;

[then]


\ ======
\ *> ###
\ ======

