\ wcmemb.fth

((  ==========================
    Tectime Data Systems, 2013
    ==========================
    File        :   wcm.fth
    Title       :
    Description :   Simple wildcard string pattern matching.
    Created     :
    Requirements:
    Author      :   G. C. Smith
))


\ ==========
\ *! WCMatch
\ *T Wild Card Match
\ ==========
\ *P This is a 'wildcard matching' routine, NOT a 'regex' routine.
\ ** Regex differs markedly from this very simple one. If you need
\ ** something more sophisticated, you are probably best off
\ ** interfacing to a regex library such as the one at
\ *C   www.pcre.org

\ **************
\ *S Explanation
\ **************

\ *P Take two strings, a 'source' string and a 'pattern'. The test
\ ** is to see if the source matches the pattern where the pattern
\ ** can contain the wildcard characters '?' and '*'. These two
\ ** characters can be 'escaped' using the character '\'.

\ *P The asterisk as a wildcard implies 'any of zero or more
\ ** characters match'. Thus '*' will match with each of 'a',
\ ** '12abxyz' and the zero length string ''. An asterisk then
\ ** matches anything. A pattern of "ab*12" will match any text
\ ** which starts with 'ab' and ends with '12'.

\ *P The question mark indicates any one character. Under
\ ** DOS/Windows the question mark can also match zero characters
\ ** but this behaviour seems inconsistant - see below for an
\ ** example. The code here insists that a question mark matches
\ ** exactly one of any one character. Thus '?' matches 'a', 'b'
\ ** and '%'. It does not match the zero-length string ''.

\ *E Source                  Pattern         Match
\ ** ---------------------   --------------  -----
\ ** "abc"                   "abc"           yes
\ ** "abcd"                  "abc"           no
\ ** "abc"                   "abc*"          yes
\ ** "abc"                   "abc?"          yes
\ ** "abc"                   "*abc"          yes
\ ** "abc"                   "?abc"          no
\ ** "ab"                    "a?b"           no
\ ** "a"                     "?"             yes
\ ** "123abc"                "*abc"          yes
\ ** "123abc"                "?abc"          no
\ ** "123abc"                "???abc"        yes
\ ** "123abc"                "1*c"           yes
\ ** ""                      ""              yes


\ =========
\ *S Method
\ =========

\ *P Compare source and pattern strings character by character
\ ** starting at the begining. If the two characters are equal
\ ** advance to the next characters to the end of either or both
\ ** strings.

\ *P The test for equality is complicated by the fact that the
\ ** pattern string may contain the 'wildcard' characters '?'
\ ** and '*' and the 'escape character' '\'.

\ *P The wildcard character '?' produces a match with any single
\ ** character in the source.

\ *P '*' produces a match with zero or more of any characers. So
\ ** we get a match between a source string of 'ab' and a pattern
\ ** of '*b'. Similarly a match occurs between a source string of
\ ** 'b' and pattern '*b'.

\ *P The third special pattern character is '\'. It is an 'escape'
\ ** character and  indicates that the next character in the
\ ** pattern string is to be treated as a non-special character.
\ ** Thus '\?' indicates a question mark which is not a wildcard
\ ** and '\*' indicates an asterisk which is not a wildcard. The two
\ ** characters '\\' indicates a single '\' character (which is not
\ ** an escape character). This code interprets any character
\ ** following '\' as that single character and so '\a' in a
\ ** pattern is equivalent to 'a' etc.

\ *P If any character comparison fails, the whole test fails, i.e.
\ ** the match returns success only if the end of the two strings
\ ** is reached together.

\ *P The primary word which performs the comparison is *\fo{wcMatch?} and
\ ** there are two versions of this: one using locals and one
\ ** without using locals. The one using locals is faster than
\ ** the other but each performs adequately.


\ ========
\ *S Words
\ ========

: *wcm  \ src slen ptn plen -- t/f
\ *G The action taken by *\fo{WCMatch?} when an asterisk character
\ ** is encountered during traversal through the pattern.
  1 /string             \ s sl p pl ; skip '*'
  \ since '*' matches anything, if there
  \ is no pattern left, it MUST be a match!
  dup 0= if  4drop -1 exit  then

  2>r					\ -- s sl ; R: -- p pl
  begin					\ s sl ; sl=0 is allowed here
    2dup 2r@ wcMatch?			\ -- s sl flag ; R: -- p pl
    if  2r> 4drop -1 exit  then
    dup
  while
    1 /string				\ -- s' sl' ; R: -- p pl
  repeat
  \ if we have reached here, there is no
  \ source string left, and no match occurred.
  2r> 4drop False
;

: wcMatch?  \ src slen ptn plen -- t/f
\ *G Wild Card Match. *\i{src} is the address of the start of a
\ ** souce string and *\i{slen} is its length. Similarly, *\i{ptn}
\ ** is the address of a pattern string and its length is *\i{plen}.
\ ** A value of TRUE is returned only if the source string matches
\ ** the pattern according to the rules described above.
					\ -- s sl p pl
  begin dup while			\ while more pattern
    case over c@
      '*' of *wcm exit      endof
      '?' of ( noop)           endof
    ( else)				\ -- s sl p pl ch ; the pattern char
      '\' = if				\ -- s sl p pl ; if it's backslash
        1 /string dup 0=		\ -- s sl p' pl' flag ; next one, exists?
        if  4drop 0 exit  then		\ -- ; if not, get out
      then				\ -- s sl p pl
      2 pick 0= if			\ -- s sl p pl  ; if no source
        4drop False exit		\ failed!
      then				\ -- s sl p pl
      3 pick c@ 2 pick c@ <> 		\ -- s sl p pl flag ; matched?
      if  4drop 0 exit  then		\ if not, we failed
    end-case				\ -- s sl p pl
    1 /string 2swap 1 /string 2swap	\ -- s' sl' p' pl'
  repeat
					\ -- s' sl' p' 0
  \ no pattern remains. A match only
  \ if there is also no source...
  2drop nip  0=				\ slen=0 ?
;

: $strmatch     \ src patt -- flag
\ *G Perform *\fo{wcMatch?} on two counted strings.
  swap count rot count wcMatch?
;

: zstrmatch     \ src patt -- flag
\ *G Perform *\fo{wcMatch?} on two zero-terminated strings.
  swap zcount rot zcount wcMatch?
;


\ ****************************
\ *S DOS/Windows inconsistency
\ ****************************
\ *P When using the wildcard character '?' in a file path/name
\ ** matching routine in DOS/Windows (such as the shell command
\ ** DIR for example), the question mark may match zero
\ ** characters. For instance, a pattern of 'ab?.*' matches the
\ ** file nme 'ab.txt'. However, placing the question mark in
\ ** another position causes the match to fail. For example,
\ ** the pattern '?ab.*' does not match 'ab.txt'.


\ ======
\ *> ###
\ ======

decimal


\ =====
\ tests
\ =====

0 [if]

256 buffer: src
256 buffer: ptn
0 value result


: -blanks   \ --
  bl parse-leading 			\ skip leading spaces
;

: tib?  \ -- flag ; any chars in text input buffer?
    #tib @ >in @ >
;

: quoted?  \ -- ch|0
    tib? if tib >in @ + c@ else 0 then
;

: getString \ "test" ++ ca u
  bl parse-leading 			\ skip leading spaces
  quoted? if 1 >in +! [char] " parse
          else bl parse then
;

: get-yes   \ "test" -- t/f
    bl word count 2dup upper \ ca u
    3 min s" YES" str=
;

: Test:
    getString src place
    getString ptn place
    get-yes to Result
    src $. space 10 >pos
    ptn $. space 20 >pos
    Src count ptn count WCMatch? \ flag
    dup if ." True  " else ." False " then
    Result = if ." OK" else ." *Wrong*" then
    0 parse-word 2drop
    cr
;
cr
test: "abc"                   "abc"           yes
test: "abcd"                  "abc"           no

test: "abc"                   "abc*"          yes
test: "abcxyz"                "abc*"          yes
test: "abc"                   "*abc"          yes
test: "123abc"                "*abc"          yes
test: "123abc"                "1*c"           yes

test: "abc"                   "abc?"          no
test: "abc"                   "?abc"          no
test: "ab"                    "a?b"           no
test: "a"                     "?"             yes
test: "axyz"                  "?xyz"          yes
test: "axyzb"                 "?xyz?"         yes
test: "123abc"                "?abc"          no
test: "123abc"                "???abc"        yes
test: ""                      ""              yes (i.e. zero-lengths match)

[then]


