\ FatStructs.fth - FAT format, mostly documentation

((
Copyright (c) 2007
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))

only forth definitions
decimal

\ =============
\ *! fatstructs
\ *T FAT file sytem structures
\ =============
\ *P This is not a definitive reference to FAT file systems. See
\ ** *\i{Docs\fatgen103.pdf} for the gory details.

\ *P For this implementation the host Forth has a cell size of
\ ** 32 bits, and the CPU is byte-addressed. The assumption
\ ** byte=char=au is made.

\ *P Note the following:
\ *(
\ *B All data on the disk is in little-endian format.
\ *B There are three variants of the FAT file system called FAT12,
\ ** FAT16, and FAT32.
\ *B Sectors on the disk are assumed to be 512 bytes. It doesn't
\ ** have to be 512 bytes, but almost always is. This implementation
\ ** assumes 512 bytes.
\ *B A cluster (also called extent) is a group of sectors
\ ** (1, 2, 4...). Cluster numbers 0 and 1 are reserved and have
\ ** special meanings. High cluster numbers also have special meanings.
\ *B The implementation described here does not support long file
\ ** names (yet).
\ *)

\ *P There are four contiguous regions of the disk.
\ *(
\ *B BPB and reserved area. The first sector contains the BIOS
\ ** Parameter Block (BPB) and for bootable disks, a primary boot
\ ** loader. Any additional sectors usually contain more of the
\ ** boot loader.
\ *B FAT (File Allocation Table). There are usually two copies of
\ ** the table. The second copy is usually updated every so often,
\ ** e.g. at shutdown or file close.
\ *B Root Directory (not necessary for FAT32).
\ *B Files and Directories.
\ *)


\ **************
\ *S Boot sector
\ **************
\ *P The layout of the boot sector varies between FAT versions.
\ ** The only way to distinguish them is the use the *\fo{BS.secs16}
\ ** and *\fo{BS.secs32} fields below.

\ *P Some non-bootable disks, e.g. SD/MMC cards, only contain a
\ ** partition table in the boot sector. The partition table is
\ ** at offset $01BE (decimal 446) and is 64 bytes long, containing
\ ** four 16 byte entries.

\ *P The bytes at offsets 510 and 511 must be $55 and $AA
\ ** respectively.

struct /BPBbase	\ -- len
\ *G Layout of the boot sector. Entries marked with '***' are misaligned
\ ** when read into an aligned buffer.
  3 field BS.jump	\   0: E9, off16, or EB, off8, 90
  8 field BS.name	\   3: ASCII identifier, not useful
  2 field BS.bps	\  11: bytes per sector, must be 512. ***
  1 field BS.spc	\  13: sectors per cluster; 1, 2, 4..128.
			\      Max bytes/cluster = 32K
  2 field BS.rsv	\  14: number of reserved sectors.
			\      1 for FAT12/16, 32 for FAT32
  1 field BS.fats	\  16: number of FAT tables. Always use 2.
  2 field BS.RootEnts	\  17: number of 32 byte root dir entries,
			\      0 for FAT32, 512 for FAT16. ***
  2 field BS.secs16	\  19: Total number of sectors on disk. ***
  1 field BS.media	\  21: F8=fixed, F0=removable.
  2 field BS.fat16	\  22: sectors per FAT table. 0 for FAT32.
  2 field BS.spt	\  24: sectors/track - mostly obsolete
  2 field BS.heads	\  26: number of heads - mostly obsolete
  4 field BS.hidden	\  28: sectors before this partition, or 0.
  4 field BS.secs32	\  32: new 32 bit sector count if secs16=0.
end-struct

struct /BPBfat16	\ -- len
\ *G Structure for FAT12 and FAT16 BPB.
  /BPBbase +
  1 field BS.Drv#	\  36: Int 0x13 drive#, 0=floppy, 0x80=hard
  1 +			\  37: NT specific, set to 0.
  1 field BS.sig16	\  38: 0x29
  4 field BS.VolID16	\  39: serial number, format time/date. ***
  11 field BS.VolLab16	\  43: ASCII volume label, blank filled.
  8 field BS.fstype16	\  54: file type info string, e.g. "FAT12"
end-struct

struct /BPBfat32	\ -- len
\ *G Structure for FAT32.
  /BPBbase +
  4 field BS.fat32	\  36: sectors per FAT table.
  2 field BS.ExtFlags	\  40: 0x008x=only use one FAT
  2 field BS.ver	\  42: FAT file version = 0
  4 field BS.RootClus	\  44: Cluster number of root directory (2).
  2 field BS.FSinfo	\  48: usually 1
  2 field BS.BkBoot	\  50: usually 6
  12 +			\  52: reserved for future use
  1 field BS.Drv32	\  64: as for FAT12/16 but in different place.
  1 +			\  65: reserved for future use
  1 field BS.sig32	\  66: 0x29, as for FAT12/16 but in different place.
  4 field BS.VolID32	\  67: serial number, format time/date. ***
  11 field BS.VolLab32	\  71: ASCII volume label, blank filled.
  8 field BS.fstype32	\  82: file type info string, e.g. "FAT12"
end-struct

struct /partition
\ *G Structure for a partition table entry. There may be four entries
\ ** starting at offset 446/$01BE.
  1 field pe.state	\ 00: State of Partition (00h=Inactive, 80h=Active)
  1 field pe.bHead	\ 01: Beginning of Partition - Head
  2 field pe.bCyl	\ 02: Beginning of Partition - Cylinder/Sector
  1 field pe.type 	\ 04: Type of Partition
  1 field pe.eHead	\ 05: End of Partition - Head
  2 field pe.eCyl	\ 06: End of Partition - Cylinder/Sector
  4 field pe.bOffset	\ 08: #Sectors Between MBR and First Sector in Partition
  4 field pe.#Sectors	\ 12: Number of Sectors in the Partition
end-struct


\ *************************
\ *S Directory entry layout
\ *************************

\ Directory entry attribute definitions
#define ATTR_READ_ONLY 0x01
#define ATTR_HIDDEN    0x02
#define ATTR_SYSTEM    0x04
#define ATTR_VOLUME_ID 0x08
#define ATTR_DIRECTORY 0x10
#define ATTR_ARCHIVE   0x20
\ top two bits always zero
#define ATTR_LONG_NAME 0x0F	\ continuation entry of long file name

struct /DirEntry	\ -- len
\ *G Structure of the 8.3 directory entries.
  11 field DE.name	\  0: 8.3 name
  1 field DE.Attr	\ 11: attribute byte
  1+			\ 12: reserved for Windows NT - set to 0
  1 field DE.centisecs	\ 13: centi-seconds of file creation, 0..199
  2 field DE.CrtTime	\ 14: File creation time
  2 field DE.CrtDate	\ 16: File creation date
  2 field DE.LastAcc	\ 18: last access date - also set to write date
  2 field DE.Clus0Hi	\ 20: First data cluster - high part; 0 for FAT12/16
  2 field DE.WrtTime	\ 22: Time of last write/creation
  2 field DE.WrtDate	\ 24: Date of last write/creation
  2 field DE.Clus0Lo	\ 26: First data cluster - low part
  4 field DE.FileSize	\ 28: File size in bytes
end-struct


\ ======
\ *> ###
\ ======

decimal

