(( ===================================================================

SDCardQSPI - primitive QSPI-based SD card I/O


  Copyright (c) 2010 OpenMRT
  Cape Town
  South Africa
  info@openmrt.co.za

This file contain the primitive drivers for the QSPI-based innterface
for all current flavours of the SD menory card. The interface is a
simple SPI-based connection to the QSPI connection of any Freescale
M52259 based circuit

 SDCard pin    CPU pin      Remarks
----------   -----------   ---------
9  DAT2          -         Add pull-up resistor 10K to 3V3
1  CD/DAT3     QSPI_CS0    Add pull-up resistor 10K to 3V3
2  CMD         QSPI_DOUT   Add pull-up resistor 10K to 3V3
3  VSS1        GND
4  VDD         3V3         add 10uH series inductor + 4u7 cap over VDD / VSS 
5  CLK         QSPI_SCK    Add pull-up resistor 10K to 3V3
6  VSS2        GND
7  DAT0        QSPI_DIN    Add pull-up resistor 10K to 3V3
8  DAT1          -         Add pull-up resistor 10K to 3V3

The code is polled-based and uses the full 16bit x 16 DIn and DOut
buffers of the QSPI device.

initSDCard will get the card into a ready - you might will need to run
it again if it fails.

SCGet1, HCGet1 will read any 512-byte block from the SD card into SDBuff
and SCPut1, HCPut1 will write a propperly set up block of data in SDBuff
to the SD card


To do
=====
  see *TODO:

Change history
==============
  201006230858 1.00      Basic I/O for ver 1.x, ver2.x SC and HC
  
===================================================================== ))

ONLY FORTH DEFINITIONS
HEX



(( --- General Tools ----------------------------------------------- ))

: /S  ( -- )  \ Reset the data stack 
  S0 @ SP!
;

: initMRAM ( -- )
\ *G initialize the MRAM module on the M52259EVB only 
  $FF _GPIO gpPTEPAR + C!     \ A7-A0
  $FF _GPIO gpPTFPAR + C!     \ A1-A8
  $FF _GPIO gpPTGPAR + C!     \ A19-A16
  $5555 _GPIO gpPTHPAR + W!   \ D7-D0
  $20 _GPIO gpPASPAR + C!     \ MB_ALE
  $00000D40 _SCM scmCSCR0 + !
  $80000000 _SCM scmCSAR0 + ! \ MRAM base address $8000:0000
  $000F0001 _SCM scmCSMR0 + !
;

: .HEX ( u1 u2 -- )   BASE @ >R  HEX  0 <#  SWAP 1-
  0 ?DO  #  LOOP  #S  #> TYPE  R> BASE ! ;


(( === QSPI Initilization ========================================== ))

VARIABLE SDOffset             \ Start Block Token offset
$230 BUFFER: SDBuff           \ SDCard R/W single block buffer, room for 32+512+16 bytes

\ use EQU's - much less hassle in the assembler sections
_QSPI QMR +   EQU xQMR    
_QSPI QDLYR + EQU xQDLYR  
_QSPI QWR +   EQU xQWR    
_QSPI QIR +   EQU xQIR    
_QSPI QAR +   EQU xQAR    
_QSPI QDR +   EQU xQDR    

\ Initialize the QSPI module
: initQSPI  ( -- )
  $1455 _GPIO gpPQSPAR + W!        \ select the primary QSPI functionality on port QS
  $83C8 xQMR W!                    \ Master, inactive 1, data 0/1, $C8=#200 divider (200kHz)
  $0404 xQDLYR W!                  \ disable, CS before 4, CS after 8+1 
  $D00F xQIR W!                    \ enable lockouts and clear flags
;

\ Enable the QSPI
CODE EnQSPI
  # 7 D0 MOVEQ,                  \ D0 to hold decimal 15
  D0 $L xQDLYR B. BSET,          \ set the SPE bit
NEXT, END-CODE

\ Halt the QSPI
CODE HaQSPI  ( -- )   
  # 7 D0 MOVEQ,
  D0 $L xQWR B. BSET,
NEXT, END-CODE


CODE WaitSPIF
  $L xQIR A0 LEA,                  \ A0 points to W. _QSPI:QIR 
  # 1 A0 ADDQ,                     \ A0 1+ to LOWER BYTE ( to test QIR[0] SPIF )
  BEGIN,
    # 0 (A0) B. BTST,
    NE, UNTIL,
  # 0 (A0) B. BSET,
NEXT, END-CODE

CODE WaitSPIFEn
  $L xQIR A0 LEA,                  \ A0 points to _QSPI:QIR 
  # 1 A0 ADDQ,                     \ A0 to LOWER BYTE ( to test QIR[0] SPIF )
  BEGIN,
    # 0 (A0) B. BTST,
    NE, UNTIL,
  # 0 (A0) B. BSET,
  # 7 D0 MOVEQ,                    \ D0 to hold decimal 15
  D0 $L xQDLYR B. BSET,            \ set the SPE bit
NEXT, END-CODE


(( --- Special Note ---------------------------------------------------
  The transmission is MS bit first. CRC7 calculation will be done 1 
  byte at a time, 'adding' the CRC and keeping the data byte intact
  CRC is the LEFTMOST 7 bits in the byte, bit0 does not count
  >> CRC is LEFT ALLIGNED and must be right alligned before sending

  CODE TEST ( -- x )  \ put something on the stack
    D7 -(A6) MOVE, # $AA55 D7 MOVE,    
  NEXT, END-CODE
  
  D7 <- TOS, (A6) <- NOS
-------------------------------------------------------------------- )) 

CODE CRC7  ( CRC7 b -- CRC7' b )
\ *G Calculate CRC7 from the byte-based data in b   
\ ** D7 is TOS, holds original data
\ ** remember : TOS = D7, NOS = 0 d(A6)
  D7 D1 MOVE,                      \ D1 is a working copy of the data
  # $18 D0 MOVEQ,  D0 D1 LSL,      \ D1 cell left justified (byte << 24)
  (A6) D2 MOVE,                    \ D2 copy the CRC7 into  
  # 7 D4 MOVEQ,                    \ D4 has loop counter
  BEGIN,                           \ start with the CRC7 calculation loop
    D1 D0 MOVE,                    \ D0 is a working copy of shifted data 
    # $80000000 D0 AND,            \ mask out the MSBit to XOR with existing CRC7
    D0 D2 EOR,
    # 1 D2 LSL,  CS,               \ D2 left shifted to inspect MSBit
      IF,  # $12000000 D0 MOVE,    \ D0 XOR modifier in TRUE
      ELSE,  D0 CLR, ENDIF,        \ no modifier
    D0 D2 EOR,                     \ D2 adjust the CRC7
    # 1 D1 ASL,                    \ D1 next bit to inspect
   # 1 D4 SUBQ,                    \ D4 loop counter decrement
    MI, UNTIL,                     \ 7 - 0 gives 8 times loop on minus check
  D2 (A6) MOVE,                    \ copy the CRC7' back onto the stack
NEXT, END-CODE                     \ D7 is TOS already there

CODE CRC16 ( addr -- CRC16 )
\ *G Calculate the 512-byte block CR 16 starting at address addr and 
\ ** leave the result on the stack
  D7 A0 MOVEA,  D2 CLR,            \ A0 to block start, D2 is CRC16 value, start with 0       
  # $FF D6 MOVE,  BEGIN,           \ 255 word block loop word at-a-time $200/2-1
    D1 CLR, (A0) D1 W. MOVE,       \ get the pointed-to data word
    # $10 D0 MOVE,  D0 D1 ASL,     \ left-align in 32-bits
    # 2 A0 ADDA,                   \ adjust pointer to next word
    \ 16-bit word loop
    # $F D5 MOVEQ,  BEGIN,          \ D5 is word loop counter, 16 bits in D1
      D1 D0 MOVE,                  \ working copy of data
      \ CRC16 calculation start here
      # $80000000 D0 AND,  D0 D2 EOR, \ mask data MS bit and EOR with existing CRC
      # 1 D2 LSL,  CS,             \ shift left and inspect ms+1 bit
        IF,  # $10210000 D0 MOVE,
        ELSE,  D0 CLR,  THEN,
      D0 D2 EOR,  # 1 D1 ASL,      \ adjust D2, the CRC16 value
      # 1 D5 SUBQ, MI, UNTIL,      \ loop the word bits
    # 1 D6 SUBQ, MI, UNTIL,        \ loop the data block
  # $10 D0 MOVE,  D0 D2 LSR,       \ Adjust 32-bit left-alligned CRC16 to 16-bit format
  D2 D7 MOVE,                      \ put the CRC16 calue on the stack
NEXT, END-CODE


(( --- Notes ----------------------------------------------------------
  A typical cycle:
    QAR <- $20, select first command RAM entry
    QDR <- load up to 16 commands, 16-bit max
    QAR <- $0, select first transmit RAM entry
    QDR <- up to 16 Tx data words
    QWR to adjust queue
    EnQSPI, enable the QSPI, will stop when finished
    Poll QIR[SPIF] to determine completion
    QAR <- $10, select first receive RAM entry
    QDR -> read the received data
-------------------------------------------------------------------- )) 
  

(( --- VFX compatability ------------------------------------------- ))

err_FATread     equ SD_READ_FAIL       \ SD card failure codes
err_FATread 1+  equ SD_CMD_FAIL    \ command hung
err_FATwrite    equ SD_WRITE_FAIL    \ write error
err_FATwrite 1+ equ SD_RESET_FAIL  \ failure to reset


0 value SDSC1?  \ -- flag
\ *G Returns true when the card is a Std Capacity ver1.x card.

0 value SDSC2?  \ -- flag
\ *G Returns true when the card is a Std Capacity ver1.x card.

0 value SDHC?
\ *G Returns true when the card is a High Capacity (SDHC) card.

: CMD0  (  -- R1 )
\ *G GO_IDLE_STATE (R1) Primitive
  0 xQAR W!  0  $40 CRC7 8 LSHIFT SWAP            \ TX RAM, new CRC7,
  0 CRC7 ROT XOR xQDR W!  0 CRC7 CRC7 xQDR W!     \ make CDMToken[2,1] and write
  0 CRC7 DROP  $18 RSHIFT $01 OR  xQDR W!         \ make CMDToken[0] and write
  $FFFF xQDR W!  ( 3+1 )                          \ reserve 16 clock cycles to get response
  $20 xQAR W!  4 0 DO $FE00 xQDR W! LOOP          \ Command RAM, and commands
  $1300 xQWR W!  EnQSPI  WaitSPIF                 \ set Wrap Register, go and wait for finish
  $13 xQAR W!  xQDR W@                            \ get CMD0 reply
;

: tryCMD0   ( -- ior)
\ *G Try 10 times at 100ms intervals to reset the card, returning
\ ** 0 for success.
  10 0 DO
    CMD0 $FF01 =  IF 0 UNLOOP EXIT  ENDIF
    100 MS
  LOOP
  SD_RESET_FAIL
;

: CMD8   ( -- b )
\ *G SEND_IF_COND (R7)
\ ** Send command 8 for range 2.7-3.6 volts, and return 0 for a good
\ ** response, a non-zero response byte for a illegal command, or -1 for
\ ** bad response data.
  0 xQAR W!  0  $48 CRC7 8 LSHIFT SWAP            \ TX RAM, new CRC7, CMD8
  $00 ( Arg[31:24]) CRC7 ROT XOR xQDR W!          \ CMD + ARG[31:24] and CMDToken2
  $00 ( Arg[23:16]) CRC7 $100 * SWAP  
  $01 ( Arg[15:8]) CRC7 ROT XOR xQDR W!           \ CMDToken1
  $AA ( Arg[7:0]) CRC7 $100 * SWAP                \ argument = $0000_01AA
  $18 RSHIFT $01 OR XOR  xQDR W!                  \ CMDToken0
  3 0 DO  $FFFF xQDR W!  LOOP  ( 3+3 )            \ response time
  $20 xQAR W!  6 0 DO $FE00 xQDR W! LOOP          \ Command RAM, and commands
  $1500 xQWR W!  EnQSPI  WaitSPIF                 \ set Wrap Register, go and wait for finish
  $13 xQAR W!  3 0 DO xQDR W@ LOOP                \ get reply
  ROT DUP $FF01 =
    IF  DROP $01AA <> IF DROP -1 THEN
    ELSE  >R 2DROP R> THEN
;

\ --- TESTING --------------------------------------------------------
\ SCSD ver1.x  : $01 | $FF05 $FFFF $FFFF  ( illegal command )
\ SCSD ver2.x  : $01 | $FF01 $0000 $01AA
\ SCHC ver2.x  : $01 | $FF01 $0000 $01AA
\ --------------------------------------------------------------------

: ACMD41
\ *G APP_CMD + SD_SEND_OP_CODE (R1)+(R1) (rev1.x)
\ ** wait for a ver 1.x card to become ready
  0 xQAR W!  
  0  $77 CRC7 8 LSHIFT SWAP                       \ TX RAM, new CRC7, CMD55
  $00 ( Arg[31:24]) CRC7 ROT XOR xQDR W!          \ CMD + ARG[31:24] and CMDToken2
  $00 ( Arg[23:16]) CRC7 $100 * SWAP  
  $00 ( Arg[15:8]) CRC7 ROT XOR xQDR W!           \ CMDToken1
  $00 ( Arg[7:0]) CRC7 $100 * SWAP                \ argument = $0000_0000
  $18 RSHIFT $01 OR XOR  xQDR W!  $FFFF xQDR W!   \ CMDToken0, ( 3+1 )
  0  $69 CRC7 8 LSHIFT SWAP                       \ TX RAM, new CRC7, CMD55
  $40 ( Arg[31:24]) CRC7 ROT XOR xQDR W!          \ CMD + ARG[31:24] and CMDToken2
  $00 ( Arg[23:16]) CRC7 $100 * SWAP  
  $00 ( Arg[15:8]) CRC7 ROT XOR xQDR W!           \ CMDToken1
  $00 ( Arg[7:0]) CRC7 $100 * SWAP                \ argument = $4000_0000
  $18 RSHIFT $01 OR XOR  xQDR W!   $FFFF xQDR W!  \ CMDToken0, ( 3+1 ) 
  $20 xQAR W!  $10 0 DO $FE00 xQDR W! LOOP        \ Command RAM, and commands
  $1700 xQWR W!  $200 0 DO                        \ set QSPI loop and wait $200 cycles
    EnQSPI  $17 xQAR W!  WaitSPIF  xQDR W@        \ the 8th word is the ACMD41 response
    $FF00 = IF  0  UNLOOP EXIT  ENDIF   LOOP
  sd_reset_fail
;

: CMD58   ( -- OCR )
\ *G READ_OCR (R3)
  0 xQAR W!  0  $7A CRC7 8 LSHIFT SWAP            \ TX RAM, new CRC7, CMD8
  $00 ( Arg[31:24]) CRC7 ROT XOR xQDR W!          \ CMD + ARG[31:24] and CMDToken2
  $00 ( Arg[23:16]) CRC7 $100 * SWAP  
  $00 ( Arg[15:8]) CRC7 ROT XOR xQDR W!           \ CMDToken1
  $00 ( Arg[7:0]) CRC7 $100 * SWAP                \ argument = $0000_0000
  $18 RSHIFT $01 OR XOR  xQDR W!                  \ CMDToken0
  3 0 DO  $FFFF xQDR W!  LOOP  ( 3+3 )            \ response time
  $20 xQAR W!  6 0 DO $FE00 xQDR W! LOOP          \ Command RAM, and commands
  $1500 xQWR W!  EnQSPI  WaitSPIF $14 xQAR W!     \ set Wrap Register, go and wait for finish
  xQDR W@ $10 LSHIFT  xQDR W@ XOR                 \ get reply, OCR is in words 4 and 5
;

\ Results:
\     HCSD : FF00 C0FF 8000 
\  SCSD(2) : FF00 80FF 8000
\  SCSD(1) : FF00 80FF 8000

: SD_reset  \ --
\ *G Reset the card, and perform the initialisation sequence.
\ ** On failure, an *\fo{sd_reset_fail throw} occurs.
  0 to SDSC1?  0 to SDSC2?  0 to SDHC?
  TryCMD0 throw                              \ force card into IDLE state
  CMD8 dup case
    $FF05 of  -1 to SDSC1? drop endof        \ check for rev 1.x card
    0 of  -1 to SDSC2? drop endof            \ check for rev 2.x card
    throw  endcase                           \ report the error code
  ACMD41 throw                               \ wait 512 cycles for the SD card to become ready                                     
  CMD58  DUP $C0000000 AND case              \ get the card OCR ( 32 bits )
    $C0000000 of                             \ this is a High Capacity card
      -1 to SDHC?  0 to SDSC2?  endof        \ set flags
    $80000000 of                             \ this is a Standard Capacity card
      SDSC1? IF 0 to SDSC2?
      ELSE -1 to SDSC2?  THEN  endof         \ set flags
    throw  endcase  DROP 
\  cr ." SD SC ver 1.x card? = " SDSC1? .
\  cr ." SD SC ver 2.x card? = " SDSC2? .
\  cr ." SD HC ver 2.x card? = " SDHC? .
  $8302 xQMR W!                              \ Master, inactive 1, data 0/1, (2x25ns=20MHz)  
;

: CMD13  (  -- R2 )
\ *G SEND_STATUS (R2) (1 6-bit )
  0 xQAR W!  0  $4D CRC7 8 LSHIFT SWAP            \ TX RAM, new CRC7,
  $00 CRC7 ROT XOR xQDR W!  $00 CRC7 CRC7 xQDR W! \ make CDMToken[2,1] and write
  0 CRC7 DROP  $18 RSHIFT $01 OR  xQDR W!         \ make CMDToken[0] and write
  $FFFF xQDR W!  $FFFF xQDR W!   ( 3+2 )          \ reserve 32 clock cycles to get response
  $20 xQAR W!  5 0 DO $FE00 xQDR W! LOOP          \ Command RAM, and commands
  $1400 xQWR W!  EnQSPI  WaitSPIF                 \ set Wrap Register, go and wait for finish
  $13 xQAR W!  xQDR W@  xQDR W@                   \ get CMD0 reply
  $FF00 AND SWAP $FF AND XOR                      \ leave 16-bit status on the stack
;

\ Primitive code to instigate a move of 16 data register words from QSPI data to addr
\ and seek the Start Block Token ($E0 ) after which Addr is incremented and TRUE returned
CODE 32?QSPI  ( addr -- addr' TRUE / FALSE )  
  D7 A0 MOVEA,  $L xQDR A1 LEA,                        \ A0->SDBuff(TOS), A1->QSPI Data Register 
  # $10 D0 MOVEQ,  D0 $L xQAR W. MOVE,                 \ reset Address Register to RxRAM offset 0
  (A1) D0 MOVE, D3 CLR, (A1) D3 W. MOVE, D0 D3 EOR,    \ read 2 words, make a cell
  (A1) D0 MOVE, D4 CLR, (A1) D4 W. MOVE, D0 D4 EOR, 
  (A1) D0 MOVE, D5 CLR, (A1) D5 W. MOVE, D0 D5 EOR, 
  (A1) D0 MOVE, D6 CLR, (A1) D6 W. MOVE, D0 D6 EOR, 
  { D3 D4 D5 D6 } (A0) MOVEM,                          \ dump 4 cells
  # $10 A0 ADDA,                                       \ move the SDBuff pointer 16 bytes
  (A1) D0 MOVE, D3 CLR, (A1) D3 W. MOVE, D0 D3 EOR,    \ read 2 words, make a cell
  (A1) D0 MOVE, D4 CLR, (A1) D4 W. MOVE, D0 D4 EOR, 
  (A1) D0 MOVE, D5 CLR, (A1) D5 W. MOVE, D0 D5 EOR, 
  (A1) D0 MOVE, D6 CLR, (A1) D6 W. MOVE, D0 D6 EOR, 
  { D3 D4 D5 D6 } (A0) MOVEM,                          \ dump 4 cells
  \ there is now 32 bytes of new data starting at TOS
  D7 A0 MOVEA,  # $20 D2 MOVE,  BEGIN,                 \ reset A0, D2 to check 32 bytes
    D1 CLR,  (A0)+ D1 B. MOVE,                         \ start to look at the just-read data
    # $FE D0 MOVE,  D0 D1 CMP, EQ,                     \ look for $FE
      IF, D7 A0 MOVEA,  $20 D(A0) A0 LEA,  A0 D7 MOVE, \ found it, adjust addr and Offset
        # $20 D0 MOVE,  D2 D0 SUB,
        D0 $L SDOffset B. MOVE,                        \ save the adjusted offset
        # -1 D0 MOVE,  D2 CLR,                         \ TRUE flag and kill the loop  
      ELSE, D0 CLR, THEN,
    # 1 D2 SUBQ,  MI, UNTIL,  
    D7 -(A6) MOVE,  D0 D7 MOVE,                        \ push flag onto the stack
NEXT, END-CODE

CODE 32>QSPI   ( addr -- addr' )
\ *G Move a block of 32 bytes to addr from the QSPI DOut RAM and increment the pointer address
\ ** with 32 bytes
  D7 A0 MOVEA,  $L xQDR A1 LEA,              \ A0 -> buffer, A1 -> QSPI Address Register 
  D0 CLR,  D0 $L xQAR W. MOVE,               \ set QAR for Data Out
  # $F D3 MOVE,  BEGIN,                      \ D3 is loop counter for readig SDBuff
    D1 CLR,  (A0)+ D1 B. MOVE,  # 8 D1 ASL,  \ get first byte and adjust
    D0 CLR,  (A0)+ D0 B. MOVE,  D1 D0 EOR,   \ get second byte and make word
    D0 (A1) W. MOVE,                         \ put the word into the QSPI
    # 1 D3 SUBQ,  MI, UNTIL,                 \ loop the data block
  A0 D7 MOVE,                                \ push the current buffer pointer 
NEXT, END-CODE

CODE 32*QSPI  ( -- T/F )
\ *G Wait for the QSPI to become not busy
\ ** This is normally executed directly after a WRITE_BLOCK has happened
  $L xQDR A0 LEA,                            \ A0 point to QSPI Buffer DI buffer 
  # $10 D0 MOVE,  $L D0 xQAR W. MOVE,        \ select Data In buffer
  # $20 D2 MOVE,  BEGIN,                     \ D2 is loop counter
    (A0) D0 W. MOVE,  D0 TST, NE,
      IF,  # -1 D0 MOVE,  D2 CLR,            \ found it, kill the loop
      ELSE,  D0 CLR,  THEN,
      # 1 D2 SUBQ,  MI, UNTIL,               \ loop the data block  
  D7 -(A6) MOVE,  D0 D7 MOVE,                \ push flag onto the stack
NEXT, END-CODE

\ Primitive code to instigate a move of 16 data register words from QSPI data buffer to addr
\ and increment the addr 16 bytes
CODE 32<QSPI  ( addr -- addr' )  
  D7 A0 MOVEA,  $L xQDR A1 LEA,                        \ A0->SDBuff(TOS), A1->QSPI Data Register 
  # $10 D0 MOVEQ,  D0 $L xQAR W. MOVE,                 \ reset Address Register to RxRAM offset 0
  (A1) D0 MOVE, D3 CLR, (A1) D3 W. MOVE, D0 D3 EOR,    \ read 2 words, make a cell
  (A1) D0 MOVE, D4 CLR, (A1) D4 W. MOVE, D0 D4 EOR, 
  (A1) D0 MOVE, D5 CLR, (A1) D5 W. MOVE, D0 D5 EOR, 
  (A1) D0 MOVE, D6 CLR, (A1) D6 W. MOVE, D0 D6 EOR, 
  { D3 D4 D5 D6 } (A0) MOVEM,                          \ dump 4 cells
  # $10 A0 ADDA,                                       \ move the SDBuff pointer 16 bytes
  (A1) D0 MOVE, D3 CLR, (A1) D3 W. MOVE, D0 D3 EOR,    \ read 2 words, make a cell
  (A1) D0 MOVE, D4 CLR, (A1) D4 W. MOVE, D0 D4 EOR, 
  (A1) D0 MOVE, D5 CLR, (A1) D5 W. MOVE, D0 D5 EOR, 
  (A1) D0 MOVE, D6 CLR, (A1) D6 W. MOVE, D0 D6 EOR, 
  { D3 D4 D5 D6 } (A0) MOVEM,                          \ dump 4 cells
  # $10 A0 ADDA,  A0 D7 MOVE,                          \ move SDBuff 10 bytes and return to stack
NEXT, END-CODE

: CMD17  ( n --  )     \ n holds 32-bit address/block number
\ *G READ_SINGLE_BLOCK (R1)
  \ construct the TX data
  $0 xQAR W!  0  $51 CRC7 8 LSHIFT SWAP           \ TX RAM, new CRC7, CMD55
  2 PICK $18 RSHIFT CRC7 ROT XOR xQDR W!          \ CMD + ARG[31:24] and CMDToken2
  OVER $10 RSHIFT  $FF AND  CRC7 8 LSHIFT SWAP    \ allign and CRC
  2 PICK 8 RSHIFT $FF AND CRC7 ROT XOR xQDR W!    \ CMDToken1
  SWAP $FF AND CRC7 8 LSHIFT SWAP                 \ ARG[23:8]
  $18 RSHIFT $01 OR XOR  xQDR W!                  \ ARG[7:0], 3
  $0D 0 DO  $FFFF xQDR W!  LOOP                   \ response time, 13, 3+13=16
  \ construct the CMD data
  $20 xQAR W!  $10 0 DO $FE00 xQDR W! LOOP        \ fill Command RAM, and commands
  $1300 xQWR W!  $13 xQAR W!  EnQSPI  WaitSPIF    \ set Wrap Register, go and wait for response
  xQDR W@   $FF00 <> ABORT" CMD17 Failure..."     \ get reply and test
  0 xQAR W! $FFFF xQDR 2DUP 2DUP W! W! W!         \ reset CMD17 command
  $1F00 xQWR W!  SDBuff  EnQSPI                   \ target addr on stack
  BEGIN WaitSPIFEn 32?QSPI UNTIL                  \ wait for the Start Block Token
  $10 0 DO WaitSPIFEn 32<QSPI LOOP                \ get 512 bytes
  WaitSPIF 32<QSPI DROP                           \ get the last 32 bytes
;

\ WRITE_BLOCK (R1) 
: CMD24  ( n -- )   \ n holds 32-bit address/block number
  \ construct the initial WRITE_BLOCK command
  $0 xQAR W!  0  $58 CRC7 8 LSHIFT SWAP           \ TX RAM, new CRC7, CMD24
  2 PICK $18 RSHIFT CRC7 ROT XOR xQDR W!          \ CMD + ARG[31:24] and CMDToken2
  OVER $10 RSHIFT  $FF AND  CRC7 8 LSHIFT SWAP    \ allign and CRC
  2 PICK 8 RSHIFT $FF AND CRC7 ROT XOR xQDR W!    \ CMDToken1
  SWAP $FF AND CRC7 8 LSHIFT SWAP                 \ ARG[23:8]
  $18 RSHIFT $01 OR XOR  xQDR W!                  \ ARG[7:0], 3
  $FFFF xQDR W!                                   \ response time, 3+1=4
  \ construct the COMMAND data
  $20 xQAR W!  $10 0 DO $FE00 xQDR W! LOOP        \ fill Command RAM with *QSPI_CS0
  $1300 xQWR W!  $13 xQAR W!  EnQSPI  WaitSPIF    \ set Wrap Register, go and wait for response
  xQDR W@   DUP $FF00 <> IF  CR ." Status : " U.  \ get reply and test
    -1 ABORT" CMD24 Command Failure..."
  ELSE DROP THEN 

  \ CMD24, Response, now construct the data block.
  \ The Start Block Token for the WRITE_BLOCK is $FE and if a read block is modified
  \ this will already be present in the SDBuff first 32 bytes. In total 1 + 512 + 2 bytes
  \ must be sent ($203) To wait for the card to finish, a contimeous stream of $FF must be
  \ sent. The Start Block Token is at offset SDOffset into SDBuff

  SDBuff SDOffset C@ +  DUP 1+ CRC16 OVER 
  $201 + W! $1F00 xQWR W!  $11 0 DO                \ start data block address, save CRC16
    32>QSPI EnQSPI WaitSPIF LOOP  DROP $11 xQAR W! \ put 544 bytes to the QSPI Dout register
  xQDR W@   DUP $FFE5 <> IF  CR ." Status : " U.   \ get reply and test
    -1 ABORT" CMD24 Data Response Failure..."
  ELSE DROP THEN 

  \ now wait for not busy ( busy is DIn low all the time = $00 )
  0 xQAR W! $FFFF xQDR 2DUP 2DUP W! W! W!          \ reset CMD24 command
  $1F00 xQWR W!  EnQSPI                            \ full 16 words command buffer and queue
  BEGIN WaitSPIFEn 32*QSPI UNTIL WaitSPIF          \ wait for QSPI Not Busy 
  $1300 xQWR W! EnQSPI WaitSPIF                    \ flush, 8 bytes $FF's
;

: SD_readblock   ( adest blockno -- )
  SDHC? 0= IF  $200 * THEN                         \ adjust for SDSC cards
  CMD17  SDBuff SDOffset C@ + 1+  
  SWAP $200 MOVE                                   \ Primitive, move data
;

: SD_writeblock   ( asrc blockno -- )
\ *G Write a 512-byte block from address *\i{asrc} to block *\i{blockno}.
  SDHC? 0= IF  $200 * THEN
  SDBuff $230 $FF FILL                             \ blank the block
  1 SDOffset C!  $FE SDBuff 1+ C!                  \ setup buffer to receive data
  SWAP SDBuff 2+ $200 MOVE  
  CMD24  
;


(( --- diagnostics only ----------------------------------------------

\ get 1 block of data off a Standard Capacity SD card
: SCGet1  ( n -- )   \ block #
  $200 * CMD17                \ READ_SINGLE_BLOCK (R1)  
  SDBuff SDOffset C@ + 1+ 
  $210 DUMP
;

\ get 1 block of data off a High Capacity SD card
: HCGet1  ( n -- )   \ block #
  CMD17                       \ READ_SINGLE_BLOCK (R1)  
  SDBuff SDOffset C@ + 1+ 
  $210 DUMP
;

\ put 1 block of data, contained in SDBuff with the Start Block Token offset
\ by SDOffset
: SCPut1 ( n -- )
  $200 * CMD24
;

\ put 1 block of data, contained in SDBuff with the Start Block Token offset
\ by SDOffset
: HCPut1 ( n -- )
  CMD24
;

-------------------------------------------------------------------- ))



(( === Second Phase Utilities ====================================== ))


: CMD9  ( --  )
\ *G SEND_CSD (R1)
\ ** Read the CSD into SDBuff + SDOffset
  0 xQAR W!  0  $49 CRC7 8 LSHIFT SWAP            \ DOut RAM, new CRC7,
  0 CRC7 ROT XOR xQDR W!  0 CRC7 CRC7 xQDR W!     \ make CDMToken[2,1] and write
  0 CRC7 DROP  $18 RSHIFT $01 OR  xQDR W!         \ make CMDToken[0] and write
  $0D 0 DO  $FFFF xQDR W!  LOOP                   \ response time, 13, 3+13=16
  \ construct the CMD data
  $20 xQAR W!  $10 0 DO $FE00 xQDR W! LOOP        \ fill Command RAM, and commands
  $1300 xQWR W!  $13 xQAR W!  EnQSPI  WaitSPIF    \ set Wrap Register, go and wait for response
  xQDR W@   $FF00 <> ABORT" CMD17 Failure..."     \ get reply and test
  0 xQAR W! $FFFF xQDR 2DUP 2DUP W! W! W!         \ reset CMD17 command
  $1F00 xQWR W! SDBuff  EnQSPI                    \ full 16 words command buffer
  BEGIN WaitSPIFEn 32?QSPI UNTIL                  \ wait for the Start Block Token
  WaitSPIF 32<QSPI DROP                           \ get the last 32 bytes
\  SDBuff SDOffset C@ 1+ + $20 DUMP                \ let's see it...
;

0 value #Sectors
\ *G Number of sectors on the card.

: GetBit  \ caddr u -- 0/1
\ *G Get bit number *\i{u} from a bit array. Bit 0 is the top bit
\ ** of the first byte. Bit 15 is the bottom bit of the second
\ ** byte.
  dup 7 and swap 3 rshift swap     \ -- caddr #bytes bit#
  1 7 rot - lshift -rot            \ -- mask caddr #bytes
  + c@ and 0<> 1 and
;  
  
: GetBits \ caddr start len -- x
\ *G Get *\i{len} bits starting at bit *\i{start} from the bitmap
\ ** starting at *\i{caddr}.
  rot 0 swap 2swap                 \ -- 0 caddr start len
  bounds ?do                       \ -- x caddr
    dup i GetBit                   \ -- x caddr 0/1
    rot 1 lshift or swap           \ -- x' caddr
  loop
  drop
;


: CheckCSD  ( -- )
\ *G Extract the CSD information, particularly the card capacity in
\ ** sectors.
  CMD9  SDBuff SDOffset C@ 1+ +
\ ** the CSD info is 128 buts long - mirror the start position so the reference
\ ** manual offsets ( SD Specification ver 2.0 p86 ) can be used
\ ** for example, SECTOR_SIZE is at 45:39 and is 7 bits wide
  dup #127 #45 - 7 getbits $7f <> abort" Sector size must be 512 bytes"
  dup 0 2 getBits if                    \ CSD 2.0
    #127 #69 - #22 getbits 1+           \ 22 bit C_SIZE
    #10 lshift -> #Sectors              \ fixed 2^10 multiplier
  else                                  \ CSD 1.0
    dup #127 #73 - #12 getBits 1+       \ 12 bit C_SIZE
    swap #127 #49 - 3 getBits           \ C_SIZE_MULT
    2+ lshift -> #Sectors               \ (C_SIZE+1)*2^(mult+2)*(2^9) ; SFP004
  endif
;


\ **********************
\ *S FAT file system API
\ **********************
\ *P The file system needs these words to access the mass storage:
\ *D SecRead  ( addr sector dev -- ) Reads a sector from the specified device.
\ **          *\fo{THROW}s on error.
\ *D SecWrite ( addr sector dev -- ) Writes a sector to the specified device.
\ **          *\fo{THROW}s on error.
\ *D MassInit ( -- ) Initializes mass storage access.
\ *D MassTerm ( -- ) Terminates mass storage access.

\ *P To simplify the code, the raw read/write interface treats all
\ ** read/write errors as fatal, and *\fo{THROW}s on error. Retries
\ ** should be accommodated within the sector read/write code.

: SecRead \ addr sector dev --
  drop          \ only support one card
  ['] SD_readblock catch err_FATread ?throw
;

: SecWrite  \ addr sector dev --
  drop          \ only support one card
  ['] SD_writeblock catch err_FATwrite ?throw
;

: MassInit  ( -- )
\ *G Initialize the mass storage manager. This is done when
\ ** a new card is detected.
  initMRAM          \ initialize the MRAM for M52259EVB
  initQSPI          \ initialize the QSPI device, slow clock
  ['] SD_reset catch abort" SD card couldn't be initialized"
  CheckCSD
;

: MassTerm  ( -- )
\ *G Shut down the mass storage manager. This is done when a
\ ** card is removed.
;


