\ SDspi.fth - generic SD/MMC card driver in SPI mode.

((
Copyright (c) 2007, 2008
Brad Eckert

and

Copyright (c) 2007, 2008
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20121218 SFP005 Added DMA options for block read/write.
20081120 SFP004 Corrected CSD check for CSD 1.0.
20081117 SFP003 Added semaphore option.
20080730 SFP002 Added speed setting routines.
20080721 SFP001 Updated with CSD reading and parsing for CSD 2.0.
		Added #SECTORS containing number of sectors on card.
))

only forth definitions
decimal

\ =========
\ *! sdspi
\ *T Generic SPI driver for SD cards
\ =========
\ *P The file *\i{SDspi.fth} contains a generic driver for SD and
\ ** MMC cards in SPI mode. The low level SPI interface should
\ ** provide the following words:
\ *E MMCon     ( -- )       Drives CS low.
\ ** MMCoff    ( -- )       Drives CS high.
\ ** spiFlush  ( -- )       Flush SPI receive queue (if any).
\ ** spi!      ( c -- )     Write SPI, discard received byte.
\ ** spi@      ( -- c )     Write $FF byte, return received byte.
\ ** spiwr     ( a -- a' )  Reads next byte and writes it to SPI
\ ** spird     ( a -- a' )  Reads SPI and writes to next byte.
\ ** initSDspi ( -- )       Initialise SPI for SD/MMC use.
\ ** spiSlow   ( -- )       Set SPI clock to low speed
\ ** spiFast   ( -- )       Set SPI clock to high speed

\ *P *\fo{SPIRD} and *\fo{SPIWR} are expected to be fast. Their
\ ** semantics were chosen to minimize threading overhead for
\ ** target Forths that use threaded code.

\ *P Please note the following restrictions.
\ *(
\ *B Sector size must be 512 bytes. This will not be changed by MPE,
\ ** so 2Gb standard capacity cards may not be supported, but SDHC cards
\ ** (usually 4Gb and up) are supported.
\ *B If the card contains a partition table, only the first
\ ** partition is accessible.
\ *)

\ *P For further details of SD and MMC cards see
\ *C   http://www.sdcard.org/about/memory_card/pls/
\ *C   http://wolverine.caltech.edu/referenc/SDMMCv52.pdf

\ *P The standard SD/MMC card pinout is as follows.
\ *E         SD mode                                 SPI mode
\ ** Pin     Name    Type    Description             Name    Type    Description
\ ** 1       CD/DAT3 I/O/PP  Card detect / data 3    CS      I       Chip select (low)
\ ** 2       CMD     PP      Command/Response line   DI      I       Data input
\ ** 3       Vss1    S       Supply voltage (gnd)    VSS     S       Supply voltage
\ ** 4       Vdd     S       Power supply            VDD     S       Power supply
\ ** 5       CLK     I       Clock                   SCLK    I       Clock
\ ** 6       Vss2    S       Supply voltage          VSS2    S       Supply voltage
\ ** 7       DAT0    I/O/PP  data line 0             O       O/PP    Data output
\ ** 8       DAT1    I/O/PP  data line 1             RSV
\ ** 9       DAT2    I/O/PP  data line 2             RSV
\ *P where the pins are numbered on the connector as below:
\ *E   9 1 2 3 4 5 6 7 8
\ **     * * * * * * * *
\ **   *


\ ****************
\ *S Configuration
\ ****************

decimal

[undefined] MciSem? [if]
1 equ MciSem?	\ -- x
\ *G Set this non-zero to use a an exclusive access semaphore in
\ ** the sector read/write routines *\fo{SecRead} and
\ ** *\fo{SecWrite}. *\fo{MciSem?} can be defined before this file
\ ** and will override this copy. This semaphore will be needed if
\ ** the disc can be accessed by other sources such as USB.
[then]


\ *******************
\ *S SD/MMC functions
\ *******************
\ *P The SD error/throw codes are defined in terms of the *\fo{err_FATread}
\ ** and *\fo{err_FATwrite} codes from *\i{XC6harn.fth}.
\ *[
err_FATread equ sd_read_fail		\ SD card failure codes
err_FATread 1+ equ sd_cmd_fail		\ command hung
err_FATwrite equ sd_write_fail		\ write error
err_FATwrite 1+ equ sd_reset_fail	\ failure to reset
\ *]

1000 equ sd_timeout	\ -- ms
\ *G Timeout in milliseconds. Slower targets may want a smaller
\ ** *\fo{SD_TIMEOUT} value, especially those whose SPI interface
\ ** takes more than a millisecond per transfer. This usually only
\ ** comes into play when using a bit-banged interface.

: SD_idle       \ cnt --
\ *G Send idle (0xFF) a specified number of times.
  0 ?do  $FF spi!  loop
;

: SD_sendArg	\ arg32 --
\ *G Send *\i{arg32} in big-endian format.
  word-split
  byte-split SPI! SPI!
  byte-split SPI! SPI!
;

: SD_prep	\ --
\ *G Prepare to send an SD command.
  spiFlush  MMCon  3 SD_idle
;

: SD_NB		\ -- res8|-1
\ *G Wait until the card is not busy or times out, returning -1
\ ** for no response or the 8-bit response. In SPI mode, all
\ ** commands generate at least one byte of response.
  sd_timeout later begin		\ -- ms ; wait for response
    dup timedout? 0=
   while
    SPI@ dup $80 and 0=			\ wait for non-busy
    if  nip exit  then
    drop  pause
  repeat
  drop -1
;

: SD_send       \ arg32 cmd6 -- result8 ; -1=timeout, other = result
\ *G Sends a command and tests for time-out, returning -1 for no
\ ** response or the 8-bit response. In SPI mode, all commands
\ ** generate at least one byte of response. The line is left with
\ ** CS active.
  SD_prep
  $40 or SPI!  SD_sendArg  $95 SPI!	\ 6 bit command, 32 bit argument, fake CRC
  SD_NB
;

: +spi@		\ x -- x<<8+b
\ *G Shift x 8 bits left, read the next SPI byte in a response and
\ ** merge it.
  8 lshift SPI@ or
;

: SD_ReadResp32	\ -- x32
\ *G Read four bytes in big-endian format.
  spi@ +spi@ +spi@ +spi@
;

: SD_sync	\ --
\ *G Get the card into a known state.
  MMCoff  60 SD_idle
  MMCon   60 SD_idle			\ get SPI interface into a known state
  MMCoff  60 SD_idle
;

0 value SD2card?	\ -- flag
\ *G Returns true when the card follows SD v2 specification.
0 value HCcard?
\ *G Returns true when the card is a High Capacity (SDHC) card.

: tryCMD0	\ -- ior
\ *G Try 10 times at 100 ms intervals to reset the card, returning
\ ** 0 for success.
  10 0 do
    SD_sync  0 0 SD_send $FE and 0=
    if  0 unloop exit  endif
    100 ms
  loop
  sd_reset_fail
;

: CMD8		\ -- b
\ *G Send command 8 for range 2.7-3.6 volts, and return 0 for a good
\ ** response, a non-zero response byte for a bad command, or -1 for
\ ** bad response data.
  SD_prep  $48 SPI!  $01AA SD_sendArg  $87 SPI!
  SD_NB dup $FE and 0= if
    drop  SD_ReadResp32 $01AA <>
  endif
;

: CheckInit	\ sd2? -- ior
\ *G Initialise the card. The parameter *\i{sd2} is -1 for an SD2
\ ** card or 0 for a SD1 card. *\i{Ior} is returned 0 for success.
\ ** *\b{N.B.} This operation may take several seconds.
  bit30 and  200 0 do			\ -- mask
    0 55 SD_send drop  dup 41 sd_send 0=
    if  drop 0  unloop exit  endif
    100 ms
  loop
  drop sd_reset_fail
;

: SD_reset	\ --
\ *G Reset the card, and perform the initialisation sequence.
\ ** On failure, an *\fo{sd_reset_fail throw} occurs.
  spiSlow 10 ms
  0 to SD2card?  0 to HCcard?
  TryCMD0 throw				\ reset card
  CMD8 bit2 and 0=  dup to SD2card?	\ Check for SD2 card
  dup 0= if  sd_sync 100 ms sd_sync  endif
  CheckInit throw			\ initialise card
  SD2card? if
    0 58 SD_send drop  SD_ReadResp32
    bit30 and 0<> to HCcard?
  endif
\ cr ." SD2card? = " SD2card? .
\ cr ." HCcard? = " HCcard? .
  spiFast  10 ms
;

: SD_cmd        \ arg32 cmd --
\ *G A more robust version of *\fo{SD_send}.
  2dup SD_send 0< if			\ failed
    \ card did not respond in a reasonable time - might be a PNY!
    \ hard reset the card and retry the command, but only once.
    250 ms  SD_reset			\ wait 250 ms to ensure buffers flushed
    SD_send 0< sd_cmd_fail and throw	\ retry once, must work this time
  else					\ valid response
    2drop
  then
;

: SD_readstatus \ -- status
\ *G Read the SD card status register.
  0 13 SD_send  SPI@ swap byte-join
;

: SD_wait       \ xt -- timeout?
\ *G Wait for condition specified by *\i{xt}, which has the stack
\ ** effect *\fo{( -- flag)} where flag is returned non-zero to
\ ** finish the operation before timeout.
  sd_timeout later begin		\ -- xt ms
    dup timedout? 0=
   while
    over execute
    if  2drop 0 exit  then
    1 ms
  repeat
  2drop -1
;

:noname ( -- f ) spi@ 255 =             \ bus<>FF when busy writing
; equ SD_writedone?	\ -- xt
\ *G Gives the xt of a *\fo{:NONAME} word used with *\fo{SD_wait}
\ ** while writing a data block.

:noname ( -- f ) spi@ 254 =             \ result = 0FE when ready to read
; equ SD_readready?	\ -- xt
\ *G Gives the xt of a *\fo{:NONAME} word used with *\fo{SD_wait}
\ ** before reading a data block.

: SD_waitread   \ --
\ *G Wait until ready to read.
  SD_readready? SD_wait sd_read_fail and throw
;

\ SD_READCSD may be used to extract information required to format the SD card.
\ You can assume the card supports 25 MHz SPI speed and a minimum block size
\ of 512 bytes. The size information is contained at the following byte offsets
\ on the buffer filled by SD_READCSD:

\ 05  ----rrrr  Minimum read block size = 2^r bytes
\ 06  ------ss
\ 07  ssssssss  s = 12-bit size
\ 08  ss------
\ 09  ------mm  Size multiplier = 2^(m+2)
\ 10  m-------

\ The number of 512-byte blocks in the SD card is s * 2^(m+2+r-9). Formatting
\ involves writing a new boot record and clearing the FAT table. Searching for
\ bad sectors is recommended but a brute force search will take a long time.
\ Other data in the CSD includes write-protect status and access times.

: SD_readcsd    \ addr --
\ *G Read the CSD - 16 bytes.
  0 9 SD_cmd  SD_waitread
  16 0 do spird loop drop
  spi@ spi@ 2drop
;


\ ************************
\ *S Sector read and write
\ ************************
\ *P The basic sector read and write routines use the simple API
\ ** above. If your SPI unit supports DMA operation, you can use
\ ** it for the 512 byte block transfers by providing the words
\ *C   sdWriteBlk  \ asrc --
\ *C   sdReadBlk   \ adest --
\ *P These routines write and read the 512 bytes of data to and
\ ** from the SPI bus.

: SD_addr	\ blockno -- sdaddr
\ *G Convert a sector number to an SD card address.
  HCcard? 0=
  if  9 lshift  endif
;

[defined] sdWriteBlk [if]	\ custom DMA version
: SD_writeblock	\ asrc blockno --
\ *G Write a 512-byte block from address *\i{asrc} to block *\i{blockno}.
  SD_addr 24 SD_cmd  $FF spi!		\ waits for NB, discard 2nd byte of R2
  $FE spi!				\ start write operation
  sdWriteBlk
  begin spi@ dup $FF = while drop repeat \ wait for status
  $1F and 5 <> sd_write_fail and throw	\ bogus status
  SD_writedone? SD_wait			\ wait for completion
  sd_write_fail and throw
;
[else]
: SD_writeblock	\ asrc blockno --
\ *G Write a 512-byte block from address *\i{asrc} to block *\i{blockno}.
  SD_addr 24 SD_cmd  $FF spi!		\ waits for NB, discard 2nd byte of R2
  $FE spi!				\ start write operation
  512 0 do
    spiwr spiwr spiwr spiwr spiwr spiwr spiwr spiwr
    spiwr spiwr spiwr spiwr spiwr spiwr spiwr spiwr
    spiwr spiwr spiwr spiwr spiwr spiwr spiwr spiwr
    spiwr spiwr spiwr spiwr spiwr spiwr spiwr spiwr
  32 +loop
  drop
  begin spi@ dup $FF = while drop repeat \ wait for status
  $1F and 5 <> sd_write_fail and throw	\ bogus status
  SD_writedone? SD_wait			\ wait for completion
  sd_write_fail and throw
;
[then]

[defined] sdReadBlk [if]	\ custom DMA version
: SD_readblock	\ adest blockno --
\ *G Read a 512-byte block to address *\i{adest} from block *\i{blockno}.
  SD_addr 17 SD_cmd  SD_waitread	\ start read, wait until ready
  sdReadBlk
  spi@ spi@ 2drop			\ discard CRC
;
[else]
: SD_readblock	\ adest blockno --
\ *G Read a 512-byte block to address *\i{adest} from block *\i{blockno}.
  SD_addr 17 SD_cmd  SD_waitread	\ start read, wait until ready
  512 0 do
    spird spird spird spird spird spird spird spird
    spird spird spird spird spird spird spird spird
    spird spird spird spird spird spird spird spird
    spird spird spird spird spird spird spird spird
  32 +loop
  drop  spi@ spi@ 2drop			\ discard CRC
;
[then]


\ ***************
\ *S CSD handling
\ ***************
\ *P The card CSD is read during initialisation into *\fo{CSDbuff}.
\ ** You can assume that the card supports a minimum block size
\ ** of 512 bytes. The size information is contained in the
\ ** following byte offsets in *\fo{CSDbuff}.
\ *E 05  ----rrrr  Minimum read block size = 2^r bytes
\ ** 06  ------ss
\ ** 07  ssssssss  s = 12-bit size
\ ** 08  ss------
\ ** 09  ------mm  Size multiplier = 2^(m+2)
\ ** 10  m-------

\ *P The number of 512-byte blocks in the SD card is s * 2^(m+2+r-9).
\ ** Formatting involves writing a new boot record and clearing
\ ** the FAT table. Searching for bad sectors is recommended but
\ ** a brute force search will take a long time. Other data in the
\ ** CSD includes write-protect status and access times.
\ ** Card formatting is not yet supported as it is assumed that
\ ** cards will be formatted on a PC.

: GetBit	\ caddr u -- 0/1
\ *G Get bit number *\i{u} from a bit array. Bit 0 is the top bit
\ ** of the first byte. Bit 15 is the bottom bit of the second
\ ** byte.
  dup 7 and swap 3 rshift swap		\ -- caddr #bytes bit#
  1 7 rot - lshift -rot			\ -- mask caddr #bytes
  + c@ and 0<> 1 and
;

: GetBits	\ caddr start len -- x
\ *G Get *\i{len} bits starting at bit *\i{start} from the bitmap
\ ** starting at *\i{caddr}.
  rot 0 swap 2swap			\ -- 0 caddr start len
  bounds ?do				\ -- x caddr
    dup i GetBit			\ -- x caddr 0/1
    rot 1 lshift or swap		\ -- x' caddr
  loop
  drop
;

(( faster but much bigger
: GetBitsFromByte	\ byte startbit numbits -- bits
 \ *G Return n bits starting at bit m from a given byte
  over +		\ byte startbit (startbit+numbits)
  8 swap -		\ byte startbit bitstoright
  over +		\ byte startbit bitstoshift
  rot rot lshift	\ bitstoright byte<<startbit
  $ff and		\ Only dealing with 8 bits
  swap rshift		\ bits
;

: GetBits		\ buffer startbit numbits -- value
 \ *G Get a number of bits from a given bit location in a buffer
  rot rot		\ numbits buffer startbit --
  dup 3 rshift		\ numbits buffer startbit index --
  swap over 3 lshift -	\ numbits buffer index ignorebits --
  rot rot		\ numbits ignorebits buffer index --
  +			\ numbits ignorebits startptr --
  swap rot over dup if	\ startptr ignorebits numbits ignorebits
    8 swap -		\ startptr ignorebits numbits bitsinfirst
    over min		\ startptr ignorebits numbits firstbitstoread
    swap over -		\ startptr ignorebits firstbitstoread numbits'
    rot rot 3 pick c@	\ startptr numbits' ignorebits firstbitstoread firstbyte
    rot rot GetBitsFromByte	\ startptr numbits' firstbits
    rot 1+ swap		\ numbits' nextptr firstbits
  else
    drop rot rot drop 0	\ numbits' nextptr 0
  then
  \ Get whole bytes in-between
  2 pick
  3 rshift 0 ?do
    swap dup c@		\ numbits' firstbits byteptr thisbyte --
    rot 8 lshift or	\ numbits' byteptr currentbits --
    swap 1+ swap	\ Point at next byte
  loop
  \ Get last partial byte
  rot 7 and		\ byteptr currentbits lastnumbits --
  dup if
    tuck lshift swap	\ byteptr currentbits' lastnumbits --
    rot c@ 0 rot	\ currentbits lastbyte 0 lastnumbits --
    GetBitsFromByte	\ currentbits lastbits --
    or			\ allbits --
  else
    rot 2drop		\ allbits --
  then
;
))

\ SFP001...
0 value #Sectors
\ *G Number of sectors on the card.

#16 buffer: CSDbuff	\ -- addr
\ *G Buffer to read CSD.

: CheckCSD	\ --
\ *G Extract the CSD information, particularly the card capacity in
\ ** sectors.
  CSDbuff SD_readcsd
  CSDbuff 44 4 getbits 9 <> abort" Sector size must be 512 bytes"
  CSDbuff dup 0 2 getBits if	\ CSD 2.0
    #58 #22 getbits 1+			\ 22 bit C_SIZE
    #10 lshift -> #Sectors		\ fixed 2^10 multiplier
  else				\ CSD 1.0
    dup #54 #12 getBits 1+		\ 12 bit C_SIZE
    swap 78 3 getBits			\ C_SIZE_MULT
    2+ lshift -> #Sectors		\ (C_SIZE+1)*2^(mult+2)*(2^9) ; SFP004
  endif
;
\ ...SFP001


\ **********************
\ *S FAT file system API
\ **********************
\ *P The file system needs these words to access the mass storage:
\ *D SecRead  ( addr sector dev -- ) Reads a sector from the specified device.
\ **          *\fo{THROW}s on error.
\ *D SecWrite ( addr sector dev -- ) Writes a sector to the specified device.
\ **          *\fo{THROW}s on error.
\ *D MassInit ( -- ) Initializes mass storage access.
\ *D MassTerm ( -- ) Terminates mass storage access.

\ *P To simplify the code, the raw read/write interface treats all
\ ** read/write errors as fatal, and *\fo{THROW}s on error. Retries
\ ** should be accommodated within the sector read/write code.

MciSem? [if]
Semaphore SecSem	\ -- addr
\ *G Exclusive access semaphore for sector read/write routines.
  1 SecSem !

: SecRead	\ addr sector dev --
\ *G Reads a sector from the specified device. *\fo{THROW}s on error.
  drop					\ only support one card
  SecSem request  ['] SD_readblock catch  SecSem signal
  err_FATread ?throw
;

: SecWrite	\ addr sector dev --
\ *G Writes a sector to the specified device. *\fo{THROW}s on error.
  drop					\ only support one card
  SecSem request  ['] SD_writeblock catch  SecSem signal
  err_FATwrite ?throw
;
[else]
: SecRead	\ addr sector dev --
  drop					\ only support one card
  ['] SD_readblock catch err_FATread ?throw
;

: SecWrite	\ addr sector dev --
  drop					\ only support one card
  ['] SD_writeblock catch err_FATwrite ?throw
;
[then]

: MassInit	\ --
\ *G Initialize the mass storage manager. This is done when
\ ** a new card is detected.
[ MciSem? ] [if]
  0 1 SecSem 2!
[then]
  initSDspi
  ['] SD_reset catch abort" SD/MMC card couldn't be initialized"
  CheckCSD
;

: MassTerm	\ --
\ *G Shut down the mass storage manager. This is done when a
\ ** card is removed.
;


\ ************
\ *S Test code
\ ************

0 [if]
\ *G To compile this test code, change the conditional compilation
\ ** in the file from *\fo{0 [if]} to *\fo{1 [if]}.

#512 buffer: SDSecBuff	\ -- addr
 \ *G Sector buffer for tests.

: .sector	\ u --
\ *G Read and dump a sector.
  SDSecBuff swap 0 SecRead
  SDSecBuff 512 dump
;
[then]

0 [if]
\ *G To compile this test code, change the conditional compilation
\ ** in the file from *\fo{0 [if]} to *\fo{1 [if]}.

: ParseCSD		\ *csd --
\ *G Parse the CSD structure and display the contents.
  cr
  ." CSD_STRUCTURE : " dup 0 2 getbits dup >r  . cr
  ." reserved (should be 0) : " dup 2 6 getbits . cr
  ." TAAC : " dup 8 8 getbits . cr
  ." NSAC : " dup 16 8 getbits . cr
  ." TRAN_SPEED : " dup 24 8 getbits . cr
  ." CCC : " dup 32 12 getbits . cr
  ." READ_BL_LEN : " dup 44 4 getbits . cr
  ." READ_BL_PARTIAL : " dup 48 1 getbits . cr
  ." WRITE_BLK_MISALIGN : " dup 49 1 getbits . cr
  ." READ_BLK_MISALIGN : " dup 50 1 getbits . cr
  ." DSR_IMP : " dup 51 1 getbits . cr
  r@ if		\ CSD 2.0 ; SFP001
    ." CSD 2.0" cr
    ." reserved (should be 0) : " dup 52 6 getbits . cr
    ." C_SIZE : " dup 58 22 getbits . cr
    ." reserved (should be 0) : " dup 80 1 getbits . cr
  else		\ CSD 1.0
    ." CSD 1.0" cr
    ." reserved (should be 0) : " dup 52 2 getbits . cr
    ." C_SIZE : " dup 54 12 getbits . cr
    ." VDD_R_CURR_MIN : " dup 66 3 getbits . cr
    ." VDD_R_CURR_MAX : " dup 69 3 getbits . cr
    ." VDD_W_CURR_MIN : " dup 72 3 getbits . cr
    ." VDD_W_CURR_MAX : " dup 75 3 getbits . cr
    ." C_SIZE_MULT : " dup 78 3 getbits . cr
  endif
  ." ERASE_BLK_EN : " dup 81 1 getbits . cr
  ." SECTOR_SIZE : " dup 82 7 getbits . cr
  ." WP_GRP_SIZE : " dup 89 7 getbits . cr
  ." WP_GRP_ENABLE : " dup 96 1 getbits . cr
  ." reserved (should be 0) : " dup 97 2 getbits . cr
  ." R2W_FACTOR : " dup 99 3 getbits . cr
  ." WRITE_BL_LEN : " dup 102 4 getbits . cr
  ." WRITE_BL_PARTIAL : " dup 106 1 getbits . cr
  ." reserved (should be 0) : " dup 107 5 getbits . cr
  ." FILE_FORMAT_GRP : " dup 112 1 getbits . cr
  ." COPY : " dup 113 1 getbits . cr
  ." PERM_WRITE_PROTECT : " dup 114 1 getbits . cr
  ." TMP_WRITE_PROTECT : " dup 115 1 getbits . cr
  ." FILE_FORMAT : " dup 116 2 getbits . cr
  ." reserved (should be 0) : " dup 118 2 getbits . cr
  ." CRC : " dup 120 7 getbits . cr
  ." unused (should be 1) : " dup 127 1 getbits . cr
  drop  r> drop
;

: .CSD		\ --
\ *G Read and display the CSD contents.
  CSDbuff SD_readcsd  CSDbuff ParseCSD
;

[then]


\ ======
\ *> ###
\ ======

decimal
