\ WinDriver.fth - Raw disk read/write for VFX Forth for Windows

((
Copyright (c) 2007
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
))


\ ============
\ *! windriver
\ *T Disk read/write for VFX Forth for Windows
\ ============
\ *P The file *\i{WinDriver.fth} is the VFX Forth for Windows driver
\ ** layer for testing the FAT file system with floppy drive A or a
\ ** Flash drive.

\ *P For demonstration on a host PC, this simple harness connects
\ ** to a floppy or USB drive on any Windows 2000/XP PC. You can set
\ ** the drive letter using *\fo{SetDrive} below.
\ ** Your C: drive won't work if it is formatted for NTFS. Also,
\ ** drives with multiple partitions are currently not supported.

\ *P The file system needs these words to access the mass storage:
\ *D SecRead  ( addr sector dev -- ) Reads a sector from the specified device.
\ **          *\fo{THROW}s on error.
\ *D SecWrite ( addr sector dev -- ) Writes a sector to the specified device.
\ **          *\fo{THROW}s on error.
\ *D MassInit ( -- ) Initializes mass storage access.
\ *D MassTerm ( -- ) Terminates mass storage access.

\ *P To simplify the code, the raw read/write interface treats all
\ ** read/write errors as fatal, and *\fo{THROW}s on error. Retries
\ ** should be accommodated within the sector read/write code.


\ ==================
\ *S Windows details
\ ==================
\ *P From Microsoft's documentation on how to read raw data from a disk:
\ ** To open a logical drive, direct access is of the form
\ *C   \\.\X:
\ *P where X: is a hard-drive partition letter, floppy disk drive, or CD-ROM drive.

\ *P You can open a physical or logical drive using the CreateFile() application
\ ** programming interface (API) with these device names provided that you have
\ ** the appropriate access rights to the drive (that is, you must be an
\ ** administrator). You must use both the CreateFile() FILE_SHARE_READ and
\ ** FILE_SHARE_WRITE flags to gain access to the drive.

\ *P Once the logical or physical drive has been opened, you can then perform
\ ** direct I/O to the data on the entire drive. When performing direct disk I/O,
\ ** you must seek, read, and write in multiples of sector sizes of the device
\ ** and on sector boundaries. Call DeviceIoControl() using
\ ** IOCTL_DISK_GET_DRIVE_GEOMETRY to get the bytes per sector, number of sectors,
\ ** sectors per track, and so forth, so that you can compute the size of the
\ ** buffer that you will need.

\ *P For this implementation, sectors are 512-byte, sector numbers
\ ** starting from 0. Addresses are RAM source/dest, and ior is
\ ** nonzero if an error occurs.


\ ******************************
\ *S Sector read/write interface
\ ******************************
\ *P This section depends on the target on which the FAT file
\ ** system is being hosted. Recode this for your hardware or host
\ ** operating system.

((
: CHS		\ LBA -- Sector Head Cylinder
\ *G Calculates the address on the drive of the given sector.
\ ** By default, this word is commented out as it is usually only
\ ** required by old IDE drives.
\ *E Sector   = (LBA mod SectorsPerTrack)+1
\ ** Cylinder = (LBA/SectorsPerTrack)/NumHeads
\ ** Head     = (LBA/SectorsPerTrack) mod NumHeads
  0 SectorsPerTrack @ um/mod N++ 0 NumHeads @ um/mod
;
))

1 constant devices
\ *G Number of physical mass storage devices supported.

create Drive$	\ -- addr
\ *G Holds a counted string holding the drive name string.
\ ** The drive letter is the fifth character.
  ", \\.\A:"

: SetDrive	\ char --
\ *G Select the required drive (0..n). Must be done before *\fo{OpenDrive}
\ ** below is run. The default is 'A'.
  Drive$ 5 + c!
;

-1 value hDrive	\ -- handle
\ *G Handle returned by Windows for the current raw drive.
\ ** Set to -1 when closed.

: openDrive	\ --
\ *G Open the raw drive.
  Drive$ count
  R/W FILE_SHARE_READ or FILE_SHARE_WRITE or open-file
  abort" Error opening drive"  to hDrive
;

: closeDrive	\ --
\ *G Close the raw drive.
  hDrive close-file drop  -1 to hDrive
;

: selDrive	\ dev --
\ *G Devices are numbered 0..n. In this case an error occurs
\ ** if the device is non-zero.
  0<> abort" Unsupported drive"
;

: brw		\ sector addr -- addr len handle
\ *G Given a sector number and a buffer address, return the address
\ ** sector length and the raw disc handle.
  swap secs>bytes 0 hDrive reposition-file throw	\ -- address
  bytes/sec hDrive			\ -- addr len handle
;

: SecRead	\ addr sector dev --
\ *G Read a sector from the specified device.
\ ** An *\fo{err_RawRead THROW} occurs on error.
  selDrive  swap brw read-file nip
  if  err_RawRead throw  then
;

: SecWrite 	\ addr sector dev --
\ *G Write a sector to the specified device.
\ ** A *\fo{err_RawWrite THROW} occurs on error.
  selDrive  swap brw write-file
  if  err_RawWrite throw  then
;

: MassInit	\ --
\ *G Initialize the mass storage manager. This is done at startup.
  -1 to hDrive  OpenDrive
;

: MassTerm	\ --
\ *G Shut down the mass storage manager. This is done at program exit.
  CloseDrive
;



\ ======
\ *> ###
\ ======

decimal

