\ mciLPC40xx.fth - SD card driver for LPC40xx CPUs.

((
Copyright (c) 2008, 2013, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20160208 SFP005 LPC40xx clock conversion.
20130316 SFP004 Added protection against sector read/write outside
		MCI DMA region.
20081120 SFP003 Corrected CSD check for CSD 1.0.
20081119 SFP002 Added busy poll before sector read and write.
                Improved error detection.
20081114 SFP001 Added semaphore option.
))

only forth definitions
decimal

\ =============
\ *! mcilpc40xx
\ *T SD card driver for LPC40xx
\ =============
\ *P The code in *\i{ARM\Drivers\mciLPC23xx.fth} provides an
\ ** SD card driver with DMA using the MCI unit in LPC23xx/24xx
\ ** and other devices. The MCI unit is an ARM Primecell PL180.
\ ** Full documentation is available from *\i{www.arm.com}.

\ *P Testing was performed on LPC2388 and LPC2468 boards.
\ ** At the time of writing (October 2008), the AHB bridge
\ ** configuration registers for the LPC2388 were not documented.
\ ** As a consequence, DMA cannot access the main CPU internal
\ ** RAM and the sector read/write buffers must be in USB RAM.
\ ** The USB RAM is split into two sections, *\fo{URAMP} (primary
\ ** for USB use) and *\fo{URAME} (extended, for sector buffers).
\ ** Unless additional code is added, this restriction applies to
\ ** the FAT file system as well.

\ *P Please note the following restrictions.
\ *(
\ *B Sector size must be 512 bytes. This will not be changed by MPE,
\ ** so 2Gb standard capacity cards may not be supported, but SDHC cards
\ ** (usually 4Gb and up) are supported.
\ *B If the card contains a partition table, only the first
\ ** partition is accessible.
\ *)

\ *P For further details of SD and MMC cards see
\ *C   http://www.sdcard.org/about/memory_card/pls/
\ *C   http://wolverine.caltech.edu/referenc/SDMMCv52.pdf

\ *P The standard SD/MMC card pinout is as follows.
\ *E         SD mode                                 SPI mode
\ ** Pin     Name    Type    Description             Name    Type    Description
\ ** 1       CD/DAT3 I/O/PP  Card detect / data 3    CS      I       Chip select (low)
\ ** 2       CMD     PP      Command/Response line   DI      I       Data input
\ ** 3       Vss1    S       Supply voltage (gnd)    VSS     S       Supply voltage
\ ** 4       Vdd     S       Power supply            VDD     S       Power supply
\ ** 5       CLK     I       Clock                   SCLK    I       Clock
\ ** 6       Vss2    S       Supply voltage (gnd)    VSS2    S       Supply voltage
\ ** 7       DAT0    I/O/PP  data line 0             O       O/PP    Data output
\ ** 8       DAT1    I/O/PP  data line 1             RSV
\ ** 9       DAT2    I/O/PP  data line 2             RSV
\ *P where the pins are numbered on the connector as below:
\ *E   9 1 2 3 4 5 6 7 8
\ **     * * * * * * * *
\ **   *



\ ****************
\ *S Configuration
\ ****************
\ *P The MCI unit's SD card pins can be assigned to more than one
\ ** set of pins. In addition, the Card Detect and Write Protect
\ ** pins are assigned to a wide range of GPIO lines on different
\ ** boards.

\ *P The *\i{hwMcixxx.fth} files contain the hardware
\ ** configurations for various boards. They are compiled before
\ ** the main MCI driver file *\i{mciLPC40xx.fth}. The examples
\ ** supplied can be compiled before the build file *\i{mci23xx.bld}.

decimal

[undefined] MciSem? [if]
tasking? equ MciSem?	\ -- x
\ *G Set this non-zero to use a an exclusive access semaphore in
\ ** the sector read/write routines *\fo{SecRead} and
\ ** *\fo{SecWrite}. *\fo{MciSem?} can be defined before this file
\ ** and will override this copy. This semaphore will be needed if
\ ** the disc can be accessed by other sources such as USB.
[then]

\ *P Because all SD cards must support 20 MHz operation, the SD
\ ** bus clock is set just below this.


\ ============================
\ *N Clock speeds and timeouts
\ ============================

interpreter
: Hz>DivMCI	\ hz -- divisor hz'
\ *G Convert a requested MCI clock speed to the nearest divider
\ ** value, which may produce a higher actual clock speed.
  periph-speed 2 /  swap / 1-
  dup $FFFF:FF00 and abort" MCI speed too low"
  periph-speed 2 /  over 1+ /
;
target

200 kHz Hz>DivMCI equ /mciSlowHz  equ /mciSlowDiv
\ *G Divider and actual frequency for the requested MCI bus slow
\ ** speed.
15 Mhz Hz>DivMCI  equ /mciFastHz  equ /mciFastDiv
\ *G Divider and actual frequency for the requested MCI bus high
\ ** speed. For an LPC23xx at 72 Mhz, the maximum appropriate
\ ** speed is 18 Mhz, which gives a value of 1 for *\fo{/mciFastDiv}.
interpreter
: >DataTimer	\ ms -- count
\ *G Given a period in milliseconds, generate the timeout count
\ ** for the *\fo{mciDatTim} register. This is calculated from
\ ** *\fo{/mciFastHz} as all sector read/writes are performed
\ ** after the switch to high speed.
  /mciFastHz #1000 /  *
;
target

100 >DataTimer equ ReadTOval	\ -- u
\ *G Time out value in the *\fo{mciDatTim} register for sector reads.
200 >DataTimer equ WriteTOval	\ -- u
\ *G Time out value in the *\fo{mciDatTim} register for sector writes.


\ ****************
\ *S Misc Control
\ ****************

: settle	\ --
\ *G A short wait used to allow MCI and DMA registers to settle.
  32 0 do loop
;

: PwrUpMCI	\ --
\ *G Take power to power up condition. Sets the Open Drain bit in
\ ** the *\fo{mciPwr} register.
  _MCI
  $42 over mciPwr + !
  begin  pause  dup mciPwr + @ bit1 and  until
  drop  2 ms
;

: PwrOnMCI	\ --
\ *G Take power to power on condition. Sets the Open Drain bit in
\ ** the *\fo{mciPwr} register.
  _MCI
  $43 over mciPwr + !			\ Open-drain CMD
  begin  pause  dup mciPwr + @ $03 and $03 =  until
  drop  50 ms
;

: PwrOffMCI	\ --
\ *G Take power to power off condition. Sets the Open Drain bit
\ ** in the*\fo{mciPwr} register.
  $40 _MCI mciPwr + !  settle		\ Open-drain CMD
;

: -OpenDrain
\ *G Clears the Open Drain bit in the*\fo{mciPwr} register.
\ ** You must remove Open Drain drive to operate at with a high
\ ** speed MCI clock, say greater than 2 MHz.
  bit6 _MCI mciPwr + bic!  settle
;

: +SDpower	\ --
\ *G Apply power to the SD card. Sets the Open Drain bit in the
\ ** *\fo{mciPwr} register.
  case  _MCI mciPwr + @ $03 and
    2 of  PwrOnMCI  endof
    3 of  endof
      PwrUpMCI PwrOnMCI
  endcase
;

: -SDpower	\ --
\ *G Remove power from the SD card, but leave the the PSU powered up.
  $42 _MCI mciPwr + !  settle
;

: SpeedMCI	\ div --
\ *G Set the MCI bus clock speed to the given divisor value.
  $FF and bit8 or $01FF _MCI mciClk + setMask
  settle				\ delay 3MCLK + 2PCLK before next write
;

: SDwidth?	\ -- n
\ *G Return 1 for 1 bit mode or 4 for 4 bit bus mode. Future
\ ** code may use 0 for SPI mode
  _MCI mciClk + @ bit11 and
  if  4  else 1  endif
;

: SDwidth	\ n --
\ *G Set the bus width. Use 1 for 1 bit bus mode or 4 for 4 bit
\ ** mode. Future code may use 0 for SPI mode.
  case
    1 of  bit11 _MCI mciClk + bic!  endof
    4 of  bit11 _MCI mciClk + or!  endof
  endcase
  settle				\ settling delay
;


\ ***********
\ *S Commands
\ ***********

$0000 equ NoResp
$0040 equ ShortResp
$00C0 equ LongResp

: waitInactive	\ --
\ *G Wait until the command channel is inactive.
  _MCI
  begin					\ wait for previous command to complete
    dup mciSt + @ bit11 and
   while
    0 over mciCmd + !			\ cancel command reg
    bit11 over mciClr + !		\ clear status
    pause
  repeat
  drop
;

: cmdStatus	\ status cmd -- status'
\ *G If the command *\i{cmd} was 1, 41, or 12, clear the CRC bit
\ ** of the status.
  dup 1 =
  if  drop  $FFFF:FFFE and  exit  endif
  dup 41 =
  if  drop  $FFFF:FFFE and  exit  endif
  12 =
  if  drop  $FFFF:FFFE and  endif
;

: SDIssCmd	\ arg32 cmd resp -- ior ; 0=success
\ *G Issue an SD command, which is defined by the 32 bit *\i{arg32},
\ ** the 6 bit command *\i{cmd} and the expected response type
\ ** *\i{resp}. The result *\fo{ior} is returned -1 on a timeout
\ ** or error, otherwise zero for success.
\ cr ." Cmd=" over .
  over >r  waitInactive
  swap $3F and or  $0400 or  swap	\ form command ; -- cmd' arg32
  _MCI
  $7FF over mciClr + !			\ clear all status bits
  20 0 do loop
  tuck mciArg + !  tuck mciCmd + !	\ stash arg before command
  begin					\ wait for status set
    dup mciSt + @ $00C5 and 0=
   while
    pause
  repeat
  dup mciSt + @ $00C5 and tuck		\ -- status mci status
  swap mciClr + !			\ clear status bits
  r> cmdStatus $0005 and 0<>		\ test error bits
;

: R1Resp	\ -- resp
\ *G Return the R1 response.
  _MCI mciResp0 + @
;

[defined] !(n) [if]
: cell>buff	\ addr x -- addr+4
\ *G Transfer a cell to memory such that the high byte is at
\ ** the low address.
  over !(n)  4 +
;
[else]
: cell>buff	\ addr x -- addr+4
  swap
  over #24 rshift  over c! 1+
  over #16 rshift  over c! 1+
  over 8 rshift  over c! 1+
  tuck c!  1+
;
[then]
: R2>buff	\ addr --
\ *G Transfer the R2 response into a 16 byte buffer. The address
\ ** must be cell aligned.
  _MCI mciResp0 + swap			\ -- src dest
  over @ cell>buff
  over 4 + @ cell>buff
  over 8 + @ cell>buff
  swap 12 + @ cell>buff drop
;


\ *******************
\ *S SD/MMC functions
\ *******************
\ *P The SD error/throw codes are defined in terms of the *\fo{err_FATread}
\ ** and *\fo{err_FATwrite} codes from *\i{XC6harn.fth}.
\ *[
err_FATread equ sd_read_fail		\ SD card failure codes
err_FATread 1+ equ sd_cmd_fail		\ command hung

err_FATwrite equ sd_write_fail		\ write error
err_FATwrite 1+ equ sd_reset_fail	\ failure to reset
\ *]

0 value SD2card?	\ -- flag
\ *G Returns true when the card follows SD v2 specification.
0 value HCcard?
\ *G Returns true when the card is a High Capacity (SDHC) card.
0 value cardRCA		\ -- rca
\ *G Holds the current Relative Card Address (RCA) in the upper
\ ** 16 bits. This is reset to zero at power up and by
\ ** *\fo{sd_reset} below. The RCA is required as part of the
\ ** 32 bit argument by the following commands:
\ *C  7 9 10 13 15 55

: tryCMD0	\ -- ior
\ *G Try 10 times at 100 ms intervals to reset the card, returning
\ ** 0 for success.
  10 0 do
    0 0 0 SDIssCmd 0=
    if  0 unloop exit  endif
    100 ms
  loop
  sd_reset_fail
;

: CmdR1		\ arg32 cmd -- r1|-1
\ *G Issue a command expecting an R1 (48/32 bit) response.
  ShortResp SDIssCmd
  if  -1 exit  endif
  R1Resp
;

: CmdR2		\ arg32 cmd -- ior
\ *G Issue a command expecting an R2 (136/128 bit) response.
\ ** The response data is left in the MCI registers.
  LongResp SDIssCmd
;

: CmdR2save	\ arg32 cmd buff16 -- ior
\ *G Issue a command expecting an R2 (136/128 bit) response.
\ ** On success the response is saved in the 16 byte buffer,
\ ** otherwise the buffer is cleared.
  >r  CmdR2 dup
  if  r> #16 erase  else  r> R2>buff  endif
;

: CMD8		\ -- b
\ *G Send command 8 for range 2.7-3.6 volts, and return 0 for a good
\ ** response, or -1 for an invalid command or bad response.
  $01AA 8 CmdR1 $01AA <>
;

: CMD55		\ -- r1|-1
\ *G Issue CMD55 (start of ACMD) and return the response.
  cardRCA 55 CmdR1
;

: ACMD41	\ arg32 -- r1|-1
\ *G Issue ACMD41 and return the response.
  CMD55 -1 =
  if  drop -1  exit endif
  41 CmdR1
;

: ACMD6		\ 1/4 -- r1|-1
\ *G Set the bus width to 1 or 4. The command can only be run
\ ** when the current card is in TRAN state, i.e. is selected
\ ** and ready to read/write data.
  CMD55 -1 =
  if  drop -1  exit endif
  dup 4 = if  $0002 else $0000 endif	\ arg32 generation
  6 CmdR1 dup >r -1 <>			\ -- 1/4 flag ; R: -- r1
  if  SDwidth  else  drop  endif	\ set if accepted
  r>
;


\ *****************
\ *S Initialisation
\ *****************

#16 buffer: CSDbuff	\ -- addr
\ *G Buffer to read CSD into.

: initMCI	\ -- ior ; 0 on success
\ *G Perform a full initialisation of the hardware, setting the
\ ** SD card clock speed to 400kHz or less.
  initDMA				\ enable DMA
  ~pcSDC _PCONP or!			\ power up
  initMCIconfig				\ initialise selected configuration
  _MCI
  $00FF over mciClk + !			\ start in 1 bit bus mode, clock disabled
  0 over mciPwr + !  10 ms		\ reset power
  0 over mciMask0 + !			\ no interrupts
  0 over mciMask1 + !
  0 over mciCmd + !			\ no command
  0 over mciDatCtrl + !			\ no data
  $07FF over mciClr + !			\ clear status bits
  drop
  PwrUpMCI				\ power up
  /mciSlowDiv speedMCI			\ slow bus speed
  0					\ success
;
:noname  initMCI drop ;  AtCold

: CheckInit	\ sd2? -- ior
\ *G Initialise the card. The parameter *\i{sd2} is -1 for an SD2
\ ** card or 0 for a SD1 card. *\i{Ior} is returned 0 for success.
\ ** *\b{N.B.} This operation may take several seconds.
  bit30 and  $00FF:8000 or  200 0 do	\ -- mask
    dup ACMD41 dup -1 <> if
      dup bit31 and if			\ bit 31 set if finished
        bit30 and SD2card? and 0<> to HCcard?	\ bit 30 set for HC card
        drop 0  unloop exit
      endif
      drop
    else
      drop
    endif
    100 ms
  loop
\ If we get here it might be an MMC card, but we don't handle these.
  drop sd_reset_fail
;

: CardIdent	\ -- ior
\ *G Issue CMD2 and CMD3 to address the card, CMD9 to read the
\ ** CSD and then CMD7 to set the card RCA and put the card into
\ ** transfer mode.
  0 2 CmdR2 -1 =			\ CMD2 = get CID register
  if  -1 exit  endif
  0 3 CmdR1 dup -1 =			\ CMD3 = get RCA
  if  drop -1 exit  endif
  $FFFF:0000 and			\ isolate RCA
  dup 9 CSDbuff CmdR2save drop		\ CMD9 = read CSD
  dup 7 CmdR1				\ CMD7 = select card
  -1 = dup 				\ -- rca flag flag
  if  nip  else  swap to cardRCA  endif
;

: SD_reset	\ --
\ *G Reset the card, and perform the initialisation sequence.
\ ** On failure, an *\fo{sd_reset_fail throw} occurs.
  PwrOffMCI  200 ms  +SDpower		\ cycle card power
  /mciSlowDiv SpeedMCI  1 SDwidth	\ bus speed and width
  10 ms
  0 to SD2card?  0 to HCcard?  0 to cardRCA
  CDin? 0= if
    cr ." SD Card not present"
    sd_reset_fail throw
  endif
  TryCMD0 sd_reset_fail ?throw		\ reset card
  CMD8 bit2 and 0=  dup to SD2card?	\ Check for SD2 card
  CheckInit sd_reset_fail ?throw	\ initialise card
\ Since we don't handle MMC cards, assume an SD card
  -OpenDrain				\ remove Open Drain drive - essential
  CardIdent sd_reset_fail ?throw	\ identify and select card
  5 0					\ put system in 4 bit mode
  do  4 ACMD6 -1 <> ?leave  10 ms  loop
  /mciFastDiv speedMCI  10 ms
((
cr ." Card reset good"
cr ." SD2card? = " SD2card? .
cr ." HCcard? = " HCcard? .
cr ." Card RCA = " cardRCA .dword
))
;

: reInitMCI	\ -- ior
\ *G Reinitialise the MCI unit after an error.
  ['] sd_reset catch
;


\ ************************
\ *S Sector read and write
\ ************************

: waitTFRready	\ -- ior ; 0=ready, 1=timeout, -1=fail
\ *G Wait until the SD card is ready to perform a transfer.
  #250 later begin			\ 250 ms timeout
    CardRCA 13 CmdR1 dup -1 =		\ failed?
    if  nip exit  endif
    bit8 and				\ ready for TX/RX?
    if  drop 0 exit  endif
    dup expired
  until
  drop 1
;

: initSDread	\ --
\ *G Initialise the SD data read system with DMA enabled.
  _MCI
  $00 over mciDatCtrl + !		\ disable data path
  $07FF over mciClr + !  settle		\ clear status
  ReadTOval over mciDatTim + !		\ read timeout
  #512 over mciDatLen + !		\ block length
  drop
;

: initSDwrite	\ --
\ *G Initialise the SD data write system with DMA enabled.
  _MCI
  $00 over mciDatCtrl + !		\ disable data path
  $07FF over mciClr + !  settle		\ clear status
  WriteTOval over mciDatTim + !		\ read timeout
  #512 over mciDatLen + !		\ block length
  drop
;

: SD_addr	\ blockno -- sdaddr
\ *G Convert a sector number to an SD card address.
  HCcard? 0=
  if  9 lshift  endif
;

[sections URAME cdata
UAlign #512 buffer: SDSecBuff	\ -- addr
 \ *G Sector buffer in the upper USB RAM section, *\fo{URAME}.
 \ ** This buffer is used when sector read/writes are to areas
 \ ** that cannot be accessed by DMA.
URAME
sec-base equ SdDmaStart			\ first address
sec-base sec-len + equ SdDmaEnd		\ last+1 address
cdata
sections]

: outSDdma?	\ addr -- flag
 \ *G Return true if the target address cannot use local DMA.
  SdDmaStart SdDmaEnd within 0=
;

: (SDreadsec)	\ adest blockno -- ior
\ *G Read a 512-byte block to address *\i{adest} from block
\ ** *\i{blockno}, returning 0 on success, -1 for command error,
\ ** -2 for not ready, or the MCI status code for other errors.
  waitTFRready				\ check for transfer
  if  2drop -2 exit  endif
  initSDread
  SD_addr 17 CmdR1 -1 = if		\ -- adest
    drop -1  exit
  endif
  dup >r  dup outSDdma? dup >r		\ adest outside DMA region
  if  drop SDSecBuff  endif		\ read to local buffer
  _MCI					\ -- adest mci
  dup mciFIFO + rot P2MvalMCI +p2m0	\ start DMA
\  #512 over mciDatLen + !		\ 512 bytes
  $9B over mciDatCtrl + !		\ enable, read, DMA, 512 bytes
  #500 later >r				\ SFP004
  begin					\ wait for MCI completion
    pause  dup mciSt + @ $072A and
    r@ timedout? or			\ SFP004
  until
  dup mciSt + @ $072A and		\ -- mci status
  r> timedout? if  bit29 or  endif	\ SFP004
  dma0Err? if  bit30 or  endif
  0 rot mciDatCtrl + !  -dma0		\ stop MCI data, stop DMA
  dup $0500 = if			\ check success ; SFP004
    drop 0  r> if			\ outside SD DMA region
      SDSecBuff r> #512 cmove		\ copy out
    else				\ local
      r> drop
    endif
  else
    r> drop  r> drop
  endif
;

: SDreadsec	\ adest blockno -- ior
\ *G Read a 512-byte block to address *\i{adest} from block
\ ** *\i{blockno}, returning 0 on success. Up to ten retries
\ ** are attempted.
  CDin? 0=				\ check for drive present
  if  2drop err_NoDrive  exit  endif
  10 0 do
    2dup (SDreadsec)
    case
       0 of  2drop 0 unloop  exit  endof
      -1 of  reInitMCI drop  endof
      -2 of  reInitMCI drop  endof
        initDMA
    endcase
    i 1+ 2* ms
  loop
  2drop sd_read_fail
;

: (SDwritesec)	\ asrc blockno -- ior
\ *G Write a 512-byte block from address *\i{asrc} to block
\ ** *\i{blockno}, returning 0 on success, -1 for command error,
\ ** -2 for not ready, or the MCI status code for other errors.
  over outSDdma? if			\ cannot use local DMA? ; SFP004
    swap SDSecBuff #512 cmove		\ copy data to local buffer
    SDSecBuff swap			\ write from local buffer
  endif
  waitTFRready				\ check for transfer
  if  2drop -2 exit  endif
  initSDwrite				\ prep SD DMA
  SD_addr 24 CmdR1 -1 =			\ issue CMD24
  if  drop -1 exit  endif
  _MCI swap				\ -- mci asrc
  over mciFIFO + M2PvalMCI +m2p0	\ -- mci ; start DMA
  $99 over mciDatCtrl + !		\ enable, write, DMA, 512 bytes
  begin					\ wait for MCI completion
    pause  dup mciSt + @ $071A and
  until
  dup mciSt + @ $071A and		\ -- mci status
  dma0Err? if  bit30 or  endif
  0 rot mciDatCtrl + !  -dma0		\ stop MCI data, stop DMA
  dup $0500 =				\ check success
  if  drop 0  endif
;

: SDwritesec	\ asrc blockno -- ior
\ *G Write a 512-byte block from address *\i{asrc} to block
\ ** *\i{blockno}, returning 0 on success. Up to ten retries
\ ** are attempted.
  CDin? 0=				\ check for drive present
  if  2drop err_NoDrive  exit  endif
  WPin?					\ check for write protect
  if  2drop err_r/o  exit endif
  10 0 do
    2dup (SDwritesec)
    case
       0 of  2drop 0 unloop exit  endof
      -1 of  reInitMCI drop  endof
      -2 of  reInitMCI drop  endof
        initDMA
    endcase
    i 1+ 2* ms
  loop
  2drop sd_write_fail
;


\ ***************
\ *S CSD handling
\ ***************
\ *P The card CSD is read during initialisation into *\fo{CSDbuff}.
\ ** You can assume that the card supports a minimum block size
\ ** of 512 bytes. The size information is contained in the
\ ** following byte offsets in *\fo{CSDbuff}.
\ *E 05  ----rrrr  Minimum read block size = 2^r bytes
\ ** 06  ------ss
\ ** 07  ssssssss  s = 12-bit size
\ ** 08  ss------
\ ** 09  ------mm  Size multiplier = 2^(m+2)
\ ** 10  m-------

\ *P The number of 512-byte blocks in the SD card is s * 2^(m+2+r-9).
\ ** Formatting involves writing a new boot record and clearing
\ ** the FAT table. Searching for bad sectors is recommended but
\ ** a brute force search will take a long time. Other data in the
\ ** CSD includes write-protect status and access times.
\ ** Card formatting is not yet supported as it is assumed that
\ ** cards will be formatted on a PC.

: GetBit	\ caddr u -- 0/1
\ *G Get bit number *\i{u} from a bit array. Bit 0 is the top bit
\ ** of the first byte. Bit 15 is the bottom bit of the second
\ ** byte.
  dup 7 and swap 3 rshift swap		\ -- caddr #bytes bit#
  1 7 rot - lshift -rot			\ -- mask caddr #bytes
  + c@ and 0<> 1 and
;

: GetBits	\ caddr start len -- x
\ *G Get *\i{len} bits starting at bit *\i{start} from the bitmap
\ ** starting at *\i{caddr}.
  rot 0 swap 2swap			\ -- 0 caddr start len
  bounds ?do				\ -- x caddr
    dup i GetBit			\ -- x caddr 0/1
    rot 1 lshift or swap		\ -- x' caddr
  loop
  drop
;

0 value #Sectors	\ -- u
\ *G Number of sectors on the card.

: ?GoodCSD	\ --
\ *G Check the CSD information in *\fo{CSDbuff}, particularly the
\ ** card capacity in sectors. Can cope with 2Gb CSD 1.0 cards.
  CSDbuff 44 4 getbits 9 <> abort" Sector size must be 512 bytes"
  CSDbuff dup 0 2 getBits if	\ CSD 2.0
    #58 #22 getbits 1+			\ 22 bit C_SIZE
    #10 lshift          		\ fixed 2^10 multiplier
  else				\ CSD 1.0
    dup #54 #12 getBits 1+		\ 12 bit C_SIZE
    swap 78 3 getBits			\ C_SIZE_MULT
    2 + lshift				\ (C_SIZE+1)*2^(mult+2)*(2^9) ; SFP003
    CSDbuff 44 4 getbits        	\ used for size calcs - size on stack is in terms of the block size
    dup 10 12 within?
    if  9 - lshift  else  drop  endif	\ scale for block size
  endif
  -> #Sectors
;



\ **********************
\ *S FAT file system API
\ **********************
\ *P The file system needs these words to access the mass storage:
\ *D SecRead  ( addr sector dev -- ) Reads a sector from the specified device.
\ **          *\fo{THROW}s on error.
\ *D SecWrite ( addr sector dev -- ) Writes a sector to the specified device.
\ **          *\fo{THROW}s on error.
\ *D MassInit ( -- ) Initializes mass storage access.
\ *D MassTerm ( -- ) Terminates mass storage access.

\ *P To simplify the code, the raw read/write interface treats all
\ ** read/write errors as fatal, and *\fo{THROW}s on error. Retries
\ ** should be accommodated within the sector read/write code.

MciSem? [if]
Semaphore SecSem	\ -- addr
\ *G Exclusive access semaphore for sector read/write routines.
  1 SecSem !

: SecRead	\ addr sector dev --
\ *G Reads a sector from the specified device. *\fo{THROW}s on error.
  drop  SecSem request  SDreadsec  SecSem signal  throw
;

: SecWrite	\ addr sector dev --
\ *G Writes a sector to the specified device. *\fo{THROW}s on error.
  drop  SecSem request  SDwritesec  SecSem signal  throw
;
[else]
: SecRead	\ addr sector dev --
  drop SDreadsec throw
;

: SecWrite	\ addr sector dev --
  drop SDwritesec throw
;
[then]

: MassInit	\ --
\ *G Initialize the mass storage manager. This is done when
\ ** a new card is detected.
[ MciSem? ] [if]
  0 1 SecSem 2!
[then]
  initMCI drop
  ['] SD_reset catch abort" bad SD card initialisation"
  ?GoodCSD
;

: MassTerm	\ --
\ *G Shut down the mass storage manager. This is done when a
\ ** card is removed.
;


\ ************
\ *S Test code
\ ************

0 [if]
\ *G To compile the main test code, change the conditional
\ ** compilation in the file from *\fo{0 [if]} to *\fo{1 [if]}.

_MCI constant _MCI	\ -- addr
\ *G Visible target constant for MCI unit base address.
_GPDMA constant _GPDMA	\ -- addr
\ *G Visible target constant for GPDMA unit base address.

[sections URAME cdata
#512 buffer: SecBuff	\ -- addr
\ *G Test sector buffer in the upper USB RAM section, *\fo{URAME}.
sections]

[then]

1 [if]
\ *G To compile the CSD parse code, change the conditional
\ ** compilation in the file from *\fo{0 [if]} to *\fo{1 [if]}.

: ParseCSD		\ *csd --
\ *G Parse the CSD structure and display the contents.
  cr
  ." CSD_STRUCTURE : " dup 0 2 getbits dup >r  . cr
  ." reserved (should be 0) : " dup 2 6 getbits . cr
  ." TAAC : " dup 8 8 getbits . cr
  ." NSAC : " dup 16 8 getbits . cr
  ." TRAN_SPEED : " dup 24 8 getbits . cr
  ." CCC : " dup 32 12 getbits . cr
  ." READ_BL_LEN : " dup 44 4 getbits . cr
  ." READ_BL_PARTIAL : " dup 48 1 getbits . cr
  ." WRITE_BLK_MISALIGN : " dup 49 1 getbits . cr
  ." READ_BLK_MISALIGN : " dup 50 1 getbits . cr
  ." DSR_IMP : " dup 51 1 getbits . cr
  r@ if		\ CSD 2.0 ; SFP001
  ." CSD 2.0" cr
  ." reserved (should be 0) : " dup 52 6 getbits . cr
  ." C_SIZE : " dup 58 22 getbits . cr
  ." reserved (should be 0) : " dup 80 1 getbits . cr
  else		\ CSD 1.0
  ." CSD 1.0" cr
  ." reserved (should be 0) : " dup 52 2 getbits . cr
  ." C_SIZE : " dup 54 12 getbits . cr
  ." VDD_R_CURR_MIN : " dup 66 3 getbits . cr
  ." VDD_R_CURR_MAX : " dup 69 3 getbits . cr
  ." VDD_W_CURR_MIN : " dup 72 3 getbits . cr
  ." VDD_W_CURR_MAX : " dup 75 3 getbits . cr
  ." C_SIZE_MULT : " dup 78 3 getbits . cr
  endif
  ." ERASE_BLK_EN : " dup 81 1 getbits . cr
  ." SECTOR_SIZE : " dup 82 7 getbits . cr
  ." WP_GRP_SIZE : " dup 89 7 getbits . cr
  ." WP_GRP_ENABLE : " dup 96 1 getbits . cr
  ." reserved (should be 0) : " dup 97 2 getbits . cr
  ." R2W_FACTOR : " dup 99 3 getbits . cr
  ." WRITE_BL_LEN : " dup 102 4 getbits . cr
  ." WRITE_BL_PARTIAL : " dup 106 1 getbits . cr
  ." reserved (should be 0) : " dup 107 5 getbits . cr
  ." FILE_FORMAT_GRP : " dup 112 1 getbits . cr
  ." COPY : " dup 113 1 getbits . cr
  ." PERM_WRITE_PROTECT : " dup 114 1 getbits . cr
  ." TMP_WRITE_PROTECT : " dup 115 1 getbits . cr
  ." FILE_FORMAT : " dup 116 2 getbits . cr
  ." reserved (should be 0) : " dup 118 2 getbits . cr
  ." CRC : " dup 120 7 getbits . cr
  ." unused (should be 1) : " dup 127 1 getbits . cr
  drop  r> drop
;

: .CSD		\ --
\ *G Display the CSD contents.
  CSDbuff ParseCSD
;

[then]


\ ======
\ *> ###
\ ======

decimal

((
[sections URAME cdata
uAlign #512 buffer: VerBuff	\ -- addr
 \ *G Verify sector buffer in the upper USB RAM section, *\fo{URAME}.
sections]

: SDverifysec	\ asrc blockno -- ior ; 0=ok
 \ *G Verify that the contents of the block match the given one.
 \ ** Return 0 if they match.
  VerBuff swap SDreadSec		\ -- asrc flag
  if  drop -1 exit  endif
  VerBuff #512 s= 0=
;
))

