\ sdioSTM32F4xx.fth - STM32F4xx SDIO driver, DMA, 4 or 1-bit I/O

((
Copyright (c) 2010, 2011, 2012, 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20130220 SFP004 Updated to select 4 or 1 bit bus width.
20120415 MRT003 Revision 1.02
20120327 MRT002 Revision 1.01
20100630 MRT001 Revision 1.00
))

only forth definitions
decimal

\ ===========
\ *! sdioSTM32F4xxd1
\ *T STM32F4xx SD card I/O with DMA
\ ===========
\ *P The SD card driver *\i{Cortex/Drivers/sdioSTM32F4xxd1.fth}
\ ** provides full speed 1-bit DMA-based I/O.

\ *P The 48 MHz clock for the SD card must be set up by the
\ ** system start up code.


\ ****************
\ *S Configuration
\ ****************

[undefined] SDwide? [if]
1 equ SDwide?	\ -- flag
\ *G If not previously defined, the SD card is run in wide
\ ** (4 bit) mode. In most STM32F4xx devices, wide mode conflicts
\ ** with USART3. If USART3 is being used, e.g. on the
\ ** STM3240G-EVAL board, the equate *\fo{SDwide?} should be set
\ ** to zero to avoid this conflict.
[then]

[undefined] MciSem? [if]
tasking? equ MciSem?	\ -- x
\ *G Set this non-zero to use a an exclusive access semaphore in
\ ** the sector read/write routines *\fo{SecRead} and
\ ** *\fo{SecWrite}. *\fo{MciSem?} can be defined before this file
\ ** and will override this copy. This semaphore will be needed if
\ ** the disc can be accessed by other sources such as USB.
[then]

[undefined] SDWhiteWash? [if]
0 equ SDWhiteWash?	\ -- flag
\ *G Set non-zero to compile code to wipe the whole card.
[then]

[undefined] SDtest? [if]
0 equ SDtest?	\ -- flag
\ *G Set non-zero to compile test code.
[then]


\ ****************
\ *S SDCard basics
\ ****************
\ *P On the STM3240G-EVAL board, the equate *\fo{SDwide?} should
\ ** be set to zero to avoid a conflict with USART3.

decimal

\ SDIO Commands

 #0 equ CMD0
 #8 equ CMD8
#55 equ CMD55
#41 equ ACMD41
 #2 equ CMD2
 #3 equ CMD3
 #9 equ CMD9
 #7 equ CMD7
 #6 equ ACMD6
#24 equ CMD24
#25 equ CMD25
#12 equ CMD12
#13 equ CMD13
#17 equ CMD17
#18 equ CMD18

 $00 equ NORESP
 $40 equ SHRESP
 $C0 equ LNRESP
$F40 equ R3RESP
\ Note:  this R3RESP is totaly out of standard. However, because
\ of the masking in SD_Command it will be processed as SHRESP as
\ R3 does not contain a valid CRC. Therefore, CCRCFAIL is set and
\ CMDREND is never set for R3. To properly process R3, exit the
\ loop CCRCFAIL condition and don't check CMDREND.

\ the response types

$01 equ R1
$02 equ R1b
$03 equ R2
$04 equ R3
$05 equ R6
$06 equ R7

\ space to save the extracted command responses

variable CMDRes
variable 1CSD
variable 2CSD
variable 3CSD
variable 4CSD

variable RCA
variable SDType
\ variable ReadAddr
\ variable response
variable SDIORxTx  0 SDIORxTx !         \ busy flag

\ *P I/O pin definitions
\ *[
PD 2  PIO: sdioPinCMD	\ -- struct ; CMD always needed
PC 12 PIO: sdioPinCK	\ -- struct ; CK always needed
PC 8  PIO: sdioPinD0	\ -- struct ; D0 always needed
SDwide? [if]		\ 4 bit mode
PC 9  PIO: sdioPinD1	\ -- struct
PC 10 PIO: sdioPinD2	\ -- struct ; also USART3 TX
PC 11 PIO: sdioPinD3	\ -- struct ; also USART3 RX
[then]
\ *]

: setSdioPin	\ struct --
\ *G Set the SDIO pin to AF12, 25MHz and push-pull o/p.
\ ** up/down.
  12 over isFunction  25MHZ_ms over isPinSpeed  0 swap isPinOD
;

: setSdioClk	\ struct --
\ *G Set the SDIO pin for CLK; AF12, 25MHz, push-pull o/p, no pull
\ ** up/down.
  dup setSdioPin  NotPulled
;

: setSdioData	\ struct --
\ *G Set the SDIO pin for data and CMD; AF12, 25MHz, push-pull o/p,
\ ** pulled up.
  dup setSdioPin  PulledUp
;

: SD_LInit	\ --
\ *G Low level initialize the SDIO device, SDIO uses PC12:8 and PD2.
  \ GPIOC and GPIOD peripheral clock enable (AHB1 bit2, bit3)
  $0040:000C _RCC rccAHB1ENR + or!	\ enable GPIOC, GPIOD and DMA2 clocks
  bit11 _RCC rccAPB2ENR + or!		\ enable SDIO clock
  [ SDwide? ] [if]
    sdioPinD3 setSdioData
    sdioPinD2 setSdioData
    sdioPinD1 setSdioData
  [then]
  sdioPinD0 setSdioData
  sdioPinCMD setSdioData
  sdioPinCK setSdioClk
  \ make it go ... starts in 1 bit mode at <= 400 kHz
  \ Assume that PLL48CK is set to 48 MHz - see PLLQ value
  $0000:0176 $0000:7FFF  _SDIO sdCLKCR + setMask  \ set SDIO clock divider and enable
  $0000:0003 $0000:0003  _SDIO sdPOWER + setMask  \ power up the SDIO
;

: SD_CMD	\ cmd resp arg -- status
\ *G Send a CMD compilation to the SD card.
  _SDIO >r
  $0000:00C5 $0000:00C5 r@ sdICR + setMask        \ clear the command flags
  ( arg TOS) $FFFF:FFFF r@ sdARG + setMask        \ --> argument
  ( resp TOS) dup $C0 and rot
  ( cmd TOS)  $3F and $0400 or or                 \ mask cmd, enable CPSM
  ( adj TOS) $0000:7FFF r@ sdCMD + setMask        \ --> command
  dup 0=  if                                      \ NoResponse, wait for CMDSENT
    begin                                         \ Bit2 CTIMEOUT, Bit7 CMDSENT
      r@ sdSTA + @  bit2 bit7 or and
    until
  else                                            \ AllElse, wait for CMDREND or CCRCFAIL
    begin                                         \ Bit0 CCRCFAIL, Bit2 CTIMEOUT, Bit6 CMDREND
      r@ sdSTA + @  bit0 bit2 or bit6 or and
    until
  endif
  \ Check to see if the response is valid
  \ We consider all R3 responses without a timeout as a valid response
  \ It seems CMDSENT and CMDREND are mutually exlusive. (though I am not sure. Check this later)
  r> sdSTA + @  dup                               \ fixing the return stack...
  bit2 and  abort" SD_CMD : Command Response Timeout"       \ Bit2 CTIMEOUT
  swap $F40 <>  if                                \ ignore resp = R3RERSP
    dup bit0 and  abort" SD_CMD: Command Response CRC Fail" \ Bit0 CCRCFAIL
  endif
  \ copy of sdSTA is on the stack
;

: SD_RESP	\ type -- value
\ *S Extract the response to the CMDResx variables.
  case
    R1  of          \ R1 responses
      _SDIO sdRESP1 + @  dup CMDRes !                      \ get it, save it and keep a copy
      dup $FDFF:E008 and  abort" SD_RESP : R1 failure"      \ All error bits must be zero
      $1F00 and 9 RSHIFT $0F and                    \ return card state
    endof
    R1b of
      _SDIO sdRESP1 + @  dup CMDRes !                      \ get it, save it and keep a copy
      dup $FDFF:E008 and  abort" SD_RESP : R1b failure"     \ All error bits must be zero
      $1F00 and  9 rshift $0F and                           \ return card state
    endof
    R2  of          \ CSD or CSI, 128 bits
      _SDIO >r
      r@ sdRESP1 + @ 1CSD !
      r@ sdRESP2 + @ 2CSD !
      r@ sdRESP3 + @ 3CSD !
      r> sdRESP4 + @ 4CSD !
      0
    endof
    R3  of          \ OCR
      _SDIO sdRESPCMD + @
      $3F <> abort" SDIO : R3 Unexpected command index"
      _SDIO sdRESP1 + @ CMDRes !
      0
    endof
    R6  of          \ RCA
      _SDIO sdRESPCMD + @
      $03 <> abort" SDIO : R6 Unexpected command index"
      _SDIO sdRESP1 + @ dup CMDRes !
      $FFFF:0000 and                                     \ RCA in RCA:0000 format!
    endof
    R7  of
      _SDIO sdRESPCMD + @
      $08 <> abort" SDIO : R7 Unexpected command index"
      _SDIO sdRESP1 + @  dup CMDRes !
      dup $FF and $AA <> abort" SDIO : R7 AA pattern mismatch"
      $FF00 and  8 rshift
    endof
      1 abort" Unknown Response Type"
  endcase
;

: SD_Init	\ -- TimeoutRemain
\ *G Initialise the card
  CMD0 NORESP 0  SD_CMD  drop                     \ go idle state, no response
  CMD8 SHRESP $0000:01AA  SD_CMD  drop            \ Non v2.0 compliant SDs will abort here due to the timeout
  R7 SD_RESP drop                                 \ ACMD41 loop
  $FF                                             \ timeout value for CMD41
  begin
    1 ms                                          \ inserted a little wait state
    CMD55 SHRESP 0  SD_CMD drop
    R1 SD_RESP  drop
    ACMD41 R3RESP $C0100000  SD_CMD drop
    R3 SD_RESP  drop
    1- dup 0< abort" SD_Init : ACMD41 Timeout"
    CMDRes @ dup bit31 and if			\ bit31 is busy bit, 0 = busy
      bit30 and if          			\ bit30 = 1, card is HC
        1 SDType !
\        cr ." SDCard is High Capacity (SDHC)" cr
      else
        0 SDType !
\        cr ." SDCard is Standard Capacity (SDSC)" cr
      endif
      -1                                           \ card is alive, get out
    else
      drop 0
    endif
  until drop
\ --- Now we are in the Ready State ---
  CMD2 LNRESP 0 SD_CMD drop             \ Ask for CID using CMD2
\ --- Now the card is in the identification mode
  CMD3 SHRESP 0 SD_CMD drop             \ Request the RCA number with CMD3
  R6 SD_RESP RCA !                      \ RCA is in RCA:0000 format!
\ --- Now the card is in stand-by mode. From this point on I can change frequency as I wish (max 24MHz)
  CMD9 LNRESP RCA @ SD_CMD drop         \ read CSD
  R2 SD_RESP drop
\ --- Put the Card in the tranfer mode
  CMD7 SHRESP RCA @  SD_CMD drop   R1 SD_RESP drop
[ SDwide? ] [if]			\ switch to 4 bit mode, high speed
\ --- change bus width mode
  CMD55 SHRESP RCA @  SD_CMD drop   R1 SD_RESP drop
  ACMD6 SHRESP $02  SD_CMD drop   R1 SD_RESP drop
\ --- configure SDIO wide-bus width and new 24MHz clock=48/(Div+2)=48/2=24
  $0000:0900 $0000:7FFF _SDIO sdCLKCR + setMask
[else]					\ stay in 1 bit mode, switch to high speed
  $0000:0100 $0000:7FFF _SDIO sdCLKCR + setMask
[then]
\ --- Now we can start issuing read/write commands
;

: SD_SIZE	\ -- ud
\ *G Return the size of the card (not the partition) in bytes.
  SDType @ 0= if
    2CSD @  $3FF and 2 lshift
    3CSD @  30 rshift or        \ this is C_SIZE
    3CSD @  15 rshift 7 and     \ this is C_SIZE_MULT
    2CSD @  16 rshift $F and    \ this is READ_BL_LEN
    1 swap lshift               \ this is BLOCK_LEN            c_size c_size_mult block_len --
    swap 2+ 1 swap lshift       \ this is MULT                        c_size block-len mult --
    rot 1+  *                   \ this is BLOCKNR                         block_len blocknr --
    m*  $400 0 d-               \ this is the MEMORY_CAPACITY
  else
    2CSD @ $3F and 16 lshift
    3CSD @ 16 rshift or
    1+ $80000 m*  $400 0 d-
  endif
;

: /SDIO		\ --
\ *G Initialise the SD driver.
  SD_LInit SD_Init
;

: *SDIO		\ --
\ *G Shut down the SD driver.
  CMD0 NORESP 0  SD_CMD  drop
  bit11 _RCC rccAPB2ENR + bic!                    \ disable SDIO clock
  $0000:0000 $0000:0003  _SDIO sdPOWER + setMask  \ power down the SDIO
  cr ." Ok to remove the SD card ..." cr
;


\ **********************
\ *S DMA Block Transfers
\ **********************
\ *P DMA for the SDIO is on DMA2, Channel 4, Stream 3.

  $00 equ P->M                          \ Transfer Direction
  $01 equ M->P
  $02 equ M->M

: SD_SBT  \ *buf cnt dir --
\ *S Start Block Transfer.

  \ cnt must be integer multiple of 512 - Enforce this inside this function
  \ Starts the actual data tranfer using the DMA.
  \ Prior to calling this command, the SDCard driver must have
  \ been initialised

  \ Make cnt an integer multiple of 512, then mask it with the maximum value allowed (2^24)
  swap 8 rshift 8 lshift $01FF:FFFF and           \  stack : ( *buf dir cnt -- )

  \ *** Part I - adjust DMA 2 Stream 3  ************************************************

  _DMA2 >r

\    1. If the stream is enabled, disable it by resetting the EN bit in the DMA_SxCR register,
\       then read this bit in order to confirm that there is no ongoing stream operation. Writing
\       this bit to 0 is not immediately effective since it is actually written to 0 once all the
\       current transfers have finished. When the EN bit is read as 0, this means that the
\       stream is ready to be configured. It is therefore necessary to wait for the EN bit to be
\       cleared before starting any stream configuration. All the stream dedicated bits set in the
\       status register (DMA_LISR and DMA_HISR) from the previous data block DMA transfer
\       should be cleared before the stream can be re-enabled.

  \ Reset the control register and make sure it is off
  bit0 r@ dmaS3CR + bic!
  begin  r@ dmaS3CR + @ bit0 and 0=  until

  \ Clear all the Stream 3 Interrupt Flag Registers
  $0F40:0000 $0F40:0000 r@ dmaLIFCR + setMask

\    2. Set the peripheral port register address in the DMA_SxPAR register. The data will be
\       moved from/ to this address to / from the peripheral port after the peripheral event.

  _SDIO sdFIFO +   r@ dmaS3PAR  +  !    \ point to the SDIO FIFO register

\    3. Set the memory address in the DMA_SxMA0R register (and in the DMA_SxMA1R
\       register in the case of a double buffer mode). The data will be written to or read from
\       this memory after the peripheral event.

  rot   r@ dmaS3M0AR +  !               \ point to buffer ( stack after rot : dir cnt *buf )

\    4. Configure the total number of data items to be transferred in the DMA_SxNDTR
\       register. After each peripheral event or each beat of the burst, this value is
\       decremented.

  0 r@ dmaS3NDTR +  !                   \ Peripheral controls, no need to indicate a size

\    5. Select the DMA channel (request) using CHSEL[2:0] in the DMA_SxCR register.
\    6. If the peripheral is intended to be the flow controller and if it supports this feature, set
\       the PFCTRL bit in the DMA_SxCR register.
\    7. Configure the stream priority using the PL[1:0] bits in the DMA_SxCR register.

\    8. Configure the FIFO usage (enable or disable, threshold in transmission and reception)

\    9. Configure the data transfer direction, peripheral and memory incremented/fixed mode,
\       single or burst transactions, peripheral and memory data widths, Circular mode, Double
\       buffer mode and interrupts after half and/or full transfer, and/or errors in the
\       DMA_SxCR register.
\   10. Activate the stream by setting the EN bit in the DMA_SxCR register.
\       As soon as the stream is enabled, it can serve any DMA request from the peripheral
\       connected to the stream.
\       Once half the data have been transferred on the AHB destination port, the half-transfer flag
\       (HTIF) is set and an interrupt is generated if the half-transfer interrupt enable bit (HTIE) is
\       set. At the end of the transfer, the transfer complete flag (TCIF) is set and an interrupt is
\       generated if the transfer complete interrupt enable bit (TCIE) is set.
\
\   *** Warning: To switch off a peripheral connected to a DMA stream
\       request, it is mandatory to, first, switch off the DMA stream to
\       which the peripheral is connected, then to wait for EN bit = 0.
\       Only then can the peripheral be safely disabled.

  0                 \ dmaS3CRVariable

  bit27          or \ Bits 27:25 CHSEL[2:0]: Channel selection 100 : channel 4 selected

  bit23          or \ Bits 24:23 MBURST: Memory burst transfer configuration - 01: INCR4 (incremental burst of 4 beats)
                    \   (FIFO size must be integer multiple of memory burst)
                    \   (FIFO is 4byte. Therefore we can only use 4 beat in this case)

  bit21          or \ Bits 22:21 PBURST[1:0]: Peripheral burst transfer configuration - 01: INCR4 (incremental burst of 4 beats)
                    \   ([Burst Size*Psize] must be equal to [FIFO size] to prevent FIFO underrun and overrun errors)
                    \   (burst also does not work in direct mode)

                    \ Bit 20 - Reserved

                    \ Bits 19 CT: Current target (only in double buffer mode)
                    \   0: The current target memory is Memory 0 (addressed by the DMA_SxM0AR pointer)
                    \   1: The current target memory is Memory 1 (addressed by the DMA_SxM1AR pointer)

                    \ Bit 18 DBM: Double buffer mode
                    \   0: No buffer switching at the end of transfer
                    \   1: Memory target switched at the end of the DMA transfer

  bit17             \ Bits 17:16 PL[1:0]:
  bit16 or       or \   Priority level, 11 very high

  bit15          or \ Bits 15 PINCOS: Peripheral increment offset size,
                    \   0: The offset size for the peripheral address calculation is linked to the PSIZE
                    \   1: The offset size for the peripheral address calculation is fixed to 4 (32-bit alignment).
                    \   This bit has no meaning if bit PINC = '0'.

  bit14          or \ Bits 14:13 MSIZE[1:0]: Memory data size
                    \   00: byte (8-bit)
                    \   01: half-word (16-bit)
                    \   10: word (32-bit)
                    \ In Direct mode, MSIZE is forced by hardware to the same value as PSIZE as soon as bit EN = '1'.

  bit12          or \ Bits 12:11 PSIZE[1:0]: Peripheral data size
                    \   00: Byte (8-bit)
                    \   01: Half-word (16-bit)
                    \   10: Word (32-bit)

  bit10          or \ Bits 10 MINC: Memory increment mode, 1: Memory address pointer is incremented after each data transfer
                    \   (increment is done according to MSIZE)

                    \ Bits 9 PINC: Peripheral increment mode, 1: Peripheral address pointer is incremented after each data transfer
                    \   (increment is done according to PSIZE)

\ Note: Ref manual p173 :
\   The par at the end of 8.3.11) says that burst mode is not allowed when Pinc=0.
\   However, it appears that this is not true at all.
\   In fact, when PBURST = $00, the SDIO's DMA control does not work at all.

                    \ Bits 8 CIRC: Circular mode
                    \   This bit is set and cleared by software and can be cleared by hardware.
                    \   0: Circular mode disabled
                    \   1: Circular mode enabled
                    \   When the peripheral is the flow controller (bit PFCTRL=1) and the stream is enabled
                    \   (EN=1), then this bit is automatically forced by hardware to 0.
                    \   It is automatically forced by hardware to 1 if the DBM bit is set, as soon as the stream is
                    \   enabled (EN = 1)

                    \ Bits 7:6 DIR[1:0]: Data transfer direction, will do this later
                    \   00: Peripheral-to-memory
                    \   01: Memory-to-peripheral
                    \   10: Memory-to-memory

  bit5           or \ Bits 5 PFCTRL: Peripheral flow controller
                    \   1: The peripheral is the flow controller
                    \ (The DMA tranfer ends when the data issues end of transfer signal regardless of NDTR value)

                    \ Bits 4:1 Interrup Bits - not now

                    \ Bits 0 EN: Stream enable / flag stream ready when read low
                    \   0: Stream disabled
                    \   1: Stream enabled

\ dmaS3CR is now made, TOS, except for the DIR bits and the EN bit ( dir cnt dmaS3CRVariable -- )

\ Set the Direction of transfer
  2over swap drop case                  \ Bits 7:6 DIR[1:0]: Data transfer direction
    P->M of  0     endof                \   00: Peripheral-to-memory
    M->P of  bit6  endof                \   01: Memory-to-peripheral
    endcase  or                         \ adjust dmaS3CR variable

  $0FFF:FFFF r@ dmaS3CR + setMask       \ write to the DMA2 S3CR register

\ Set the FIFO CR  (stack : dir cnt)
  $0000:0021        \ reset value
                    \ Bits 7 FEIE: FIFO error interrupt enable 0:disabled 1:enabled
                    \ Bit 6 RESERVED
                    \ Bits 5:3 FS[2:0]: FIFO status
                    \   000: 0 < fifo_level < 1/4
                    \   001: 1/4 = fifo_level < 1/2
                    \   010: 1/2 = fifo_level < 3/4
                    \   011: 3/4 = fifo_level < full
                    \   100: FIFO is empty
                    \   101: FIFO is full
                    \   others: no meaning
  bit2           or \ Bits 2 DMDIS: Direct mode disable, 0:enabled 1:disabled
  bit1           or \ Bits 1:0 FTH[1:0]: FIFO threshold selection
  bit0           or \   00: 1/4 full FIFO
                    \   01: 1/2 full FIFO
                    \   10: 3/4 full FIFO
                    \   11: full FIFO

  $0000:00FF r@ dmaS3FCR + setMask

\ Enable the DMA (When it is enabled, it starts to respond DMA requests)
  bit0 r> dmaS3CR + or!                 \ last r@ fix to r>

\ Bits 0 EN: Stream enable / flag stream ready when read low
\ This bit is set and cleared by software.
\   0: Stream disabled
\   1: Stream enabled
\ This bit may be cleared by hardware:
\    on a DMA end of transfer (stream ready to be configured)
\    if a transfer error occurs on the AHB master buses
\    when the FIFO threshold on memory AHB port is not compatible with the size of the burst
\ When this bit is read as 0, the software is allowed to program the Configuration and FIFO
\ bits registers. It is forbidden to write these registers when the EN bit is read as 1.

\ *** END of PART I  (stack : dir cnt )

\ *** PART II - Adjust and enable SDIO Peripheral ************************************************

  _SDIO >r

\ Clear the Data status flags in the Interrupt Register
  $0000:073A $0000:07FF r@ sdICR + setMask

\ First adjust the DTIMER ( $FFFF:FFFF for now )
  ( $FFFF:FFFF ) 48000000 r@ sdDTIMER + !	\ 2 second timeout

\ set the length of the transfer  (stack : dir cnt )
  $01FF:FFFF r@ sdDLEN + setMask

\ DPSM - Data Path State Machine (stack dir )

\ Now adjust DCTRL (and enable it at the same time) ( dir is TOS )
  $0000:0000                  \ make the sdDRCTLVariable

                              \ Bit 11 SDIOEN: SD I/O enable functions
                              \   If this bit is set, the DPSM performs an SD I/O-card-specific operation.

                              \ Bit 10 RWMOD: Read wait mode
                              \   0: Read Wait control stopping SDIO_D2
                              \   1: Read Wait control using SDIO_CK

                              \  Bit 9 RWSTOP: Read wait stop
                              \   0: Read wait in progress if RWSTART bit is set
                              \   1: Enable for read wait stop if RWSTART bit is set

                              \ Bit 8 RWSTART: Read wait start
                              \   If this bit is set, read wait operation starts.

  bit7                     or \ Bits 7:4 DBLOCKSIZE: Data block size
  bit4                     or \   1001 : (9 decimal) lock length = 29 = 512 bytes

  bit3                     or \ Bit 3 DMAEN: DMA enable bit
                              \   0: DMA disabled.
                              \   1: DMA enabled.

                              \ Bit 2 DTMODE: Data transfer mode selection 1: Stream or SDIO multibyte data transfer.
                              \   0: Block data transfer
                              \   1: Stream or SDIO multibyte data transfer

  swap case                   \ Bit 1 DTDIR - Data transfer direction selection  ( dir sdDRCTLVariable is TOS)
    M->P of  0    endof       \   0: from controller to card
    P->M of  bit1 endof       \   1: from card to controller
    endcase                or

  bit0                     or \ Bit 0 DTEN: Data transfer enabled bit
                              \ Data transfer starts if 1b is written to the DTEN bit. Depending on the direction bit, DTDIR,
                              \ the DPSM moves to the Wait_S, Wait_R state or Readwait if RW Start is set immediately at
                              \ the beginning of the transfer. It is not necessary to clear the enable bit after the end of a data
                              \ transfer but the SDIO_DCTRL must be updated to enable a new data transfer

  $0000:0FFF r> sdDCTRL + setMask       \ last r@ fix to r>

\ *** End of PART II  (

\ Warn everybody that there may be a transfer in progress
  1 SDIORxTx !
;

: SD_WTE	\ --
\ *G Checks if there is an ogoing tranmission and waits until it
\ ** ends. It then checks the data flags to see if there is an
\ ** error. In case of an error it aborts. Before the start of
\ ** data transmission the data flags are all cleared. Therefore,
\ ** calling this function after a real transmission works as
\ ** expected.

\ Check if there is an ongoing transmission and if the DMA is disabled (SDIO disables the DMA after it is done with it)

  begin
    _DMA2 dmaS3CR + @  $0000:0001 and  0=    \ mask out EN bit and wait till it goes 0
  until

  \ Wait for the DMA Interrupt flags if a SDIO transfer are in progress.
  SDIORxTx @ if
    _DMA2 dmaLISR + @ dup     \ get the current DMA low interrupt status register. For Stream3 :
    bit27                     \ TCIF3: S3 clear transfer complete interrupt flag
    bit25 or                  \ TEIF3: S3 clear transfer error interrupt
    bit24 or                  \ DMEIF3: S3 clear direct mode error interrupt flag
    bit22 or and              \ CFEIF3: S3 clear FIFO error interrupt flag
    if  bit27 ( TCIF3) and 0=  abort" SD_WTE : a DMA error has occured"  endif
  endif

  \ Wait till SDIO is not active
  begin
    _SDIO sdSTA + @ bit13 bit12 or and 0=        \ mask out RxACT and TxRCT and wait till both 0
  until

  \ Check if there is an ongoing transmission check if it has been completed without error
  SDIORxTx @ if
    \ wait here until for data response
    begin
      _SDIO sdSTA + @
      bit10                 \ Bit 10 DBCKEND: Data block sent/received (CRC check passed)
[ SDwide? ] [if]
      bit9 or               \ Bit 9  STBITERR: Start bit not detected on all data signals in wide bus mode
[then]
      bit3 or               \ Bit 3  DTIMEOUT: Data timeout
      bit1 or and           \ Bit 1  DCRCFAIL: Data block sent/received (CRC check failed)
    until
    _SDIO sdSTA + @ bit10 and 0= abort" SD_WTE : SDIO Data Transmission Error"
  endif
  0 SDIORxTx !                \ there are no more transmissions
;


\ ************************
\ *S Sector read and write
\ ************************

: SD_RSB	\ *buff blk --
\ *G Read Single Block, read 512 bytes from SD block *\i{blk} into
\ ** address *\i{*buff}.
  SDType @ 0=  if 512 *  endif                    \ blk now is ReadAddres  ( *buf ReadAddress is TOS )
  CMD17 SHRESP rot SD_CMD drop                    \ CMD17 - READ_SINGLE_BLOCK
  R1 SD_RESP drop
  ( *buf) 512 P->M SD_SBT                         \ Start Block Transfer, SD > memory
  SD_WTE                                          \ Wait Transmission End
  ;

: SD_WSB	\ *buff blk --
\ *G Write Single Block, write 512 bytes from address *\i{*buff}
\ ** to SD block *\i{blk}
  SDType @ 0=  if 512 *  endif                    \ blk now is WriteAddress  ( *buf WriteAddress is TOS )
  CMD24 SHRESP rot SD_CMD drop                    \ CMD24 - Write_SINGLE BLOCK
  R1 SD_RESP drop
  ( *buf) 512 M->P SD_SBT                         \ Start Block Transfer, memory > SD
  SD_WTE                                          \ Wait Transmission End
;



\ **********************
\ *S FAT file system API
\ **********************
\ *P The file system needs these words to access the mass storage:
\ *D SecRead  ( addr sector dev -- ) Reads a sector from the specified device.
\ **          *\fo{THROW}s on error.
\ *D SecWrite ( addr sector dev -- ) Writes a sector to the specified device.
\ **          *\fo{THROW}s on error.
\ *D MassInit ( -- ) Initializes mass storage access.
\ *D MassTerm ( -- ) Terminates mass storage access.

\ *P To simplify the code, the raw read/write interface treats all
\ ** read/write errors as fatal, and *\fo{THROW}s on error. Retries
\ ** should be accommodated within the sector read/write code.

0 value #Sectors
\ *G Number of sectors on the card.

: SD_readblock	\ adest blockno -
\ *G Read a 512-byte block from block *\i{blockno} to address
\ ** *\i{adest}.
  SD_RSB
;

: SD_writeblock	\ asrc blockno --
\ *G Write a 512-byte block from address *\i{asrc} to block
\ ** *\i{blockno}.
  SD_WSB
;

MciSem? [if]
Semaphore SecSem	\ -- addr
\ *G Exclusive access semaphore for sector read/write routines.
  1 SecSem !

: SecRead	\ addr sector dev --
\ *G Reads a sector from the specified device. *\fo{THROW}s on error.
  drop					\ only support one card
  SecSem request
  ['] SD_readblock catch
  SecSem signal
  err_FATread ?throw
;

: SecWrite	\ addr sector dev --
\ *G Writes a sector to the specified device. *\fo{THROW}s on error.
  drop					\ only support one card
  SecSem request
  ['] SD_writeblock catch
  SecSem signal
  err_FATwrite ?throw
;
[else]
: SecRead	\ addr sector dev --
  drop					\ only support one card
  ['] SD_readblock catch err_FATread ?throw
;

: SecWrite	\ addr sector dev --
  drop					\ only support one card
  ['] SD_writeblock catch err_FATwrite ?throw
;
[then]

: MassInit	\ --
\ *G Initialize the mass storage manager. This is done when
\ ** a new card is detected.
[ MciSem? ] [if]
  0 1 SecSem 2!
[then]
  /SDIO
  SD_Size #512 m/ -> #Sectors		\ # 512-byte blocks on card
;

: MassTerm	\ --
\ *G Shut down the mass storage manager. This is done when a
\ ** card is removed.
  *SDIO
;


\ *********************
\ *S Whitewash the card
\ *********************
\ *P Sometimes the card needs to be completely overwritten. This
\ ** is known as whitewashing.

SDWhiteWash? SDtest? or [if]
512  buffer: InBuf
512  buffer: OutBuf
1024 buffer: DatBuf
[then]

SDWhiteWash? [if]

: SD_SMBW	\ blk --
\ *G Start Multiple Block Write into block blk on the SD card )
  SDType @ 0=  if 512 *  endif                     \ blk now is WriteAddres
  CMD25 SHRESP rot SD_CMD drop                     \ CMD25 - WRITE_MULT_BLOCK
  R1 SD_RESP drop
  $00A0:07FF $00A0:07FF _SDIO sdICR + setMask      \ clear any flags
;

: SD_WD		\ *gBuf size --
\ *G Start copy multiple data, <size> bytes from address *gBuf into SD block blk
  0                                                \ this is the buffer contents count, bcnt
  DatBuf >r                                        \ working copy as *pBuf on rstack
  begin
    rot dup ( *gBuf TOS) @ r@ !                    \ get 32-bit data from *gBuf into *pBuf
    r> cell+ >r                                    \ *pBuf to next 32-bitword
    cell+                                          \ *gbuf to next 32-bit word
    rot cell-                                      \ dec data size
    rot cell+                                      \ inc buffer contents counter
    dup $200 ( 512) >=  if                         \ bufer 1/2 full?
      SD_WTE                                       \ wait for SD traffic to die down
      r@ $200 - 512 M->P SD_SBT                    \ write this buffer out
      r@ DatBuf $400 + >=
        if  r> drop DatBuf >r  endif               \ check for double buffer fill
      drop 0                                       \ reset bcnt
    endif
    over 0 <=                                       \ cnt terminate loop
  until
  r> 2drop 2drop
;

: SD_EMBW	\ --
\ *G End Multiple Block Write.
  SD_WTE                                           \ wait for the traffic to die down
  CMD12 SHRESP 0 SD_CMD drop                       \ STOP_TRANSMISSION
  R1b SD_RESP drop
  CMDRes @  $100 and 0= if                        \ check bit8
    begin
      CMD13 SHRESP RCA @ SD_CMD drop               \ SEND_STATUS
      R1b SD_RESP drop
      CMDRes @ $100 and                           \ bit8, READY_FOR_DATA
    until
  endif
;

: .ast		( -- )  [char] * emit  ;

: 25dots	( -- )  cr 25 0 do  .ast  loop cr cr ;

: SD_WW		\ --
\ *G In order to "whitewash" the disk a special version of SD_WD,
\ ** SD_WW is made which will write the contents of OutBuf to all
\ ** the blocks on the disk.
  SD_SIZE 512 m/                                  \ get the disks usable block count
  dup 100 /                                       \ get the 100th size
  0 begin                                         \ active block counter
    SD_WTE  OutBuf 512 M->P SD_SBT                \ wait & start the block transfer
    \ size 100 cnt
    1+ 2dup swap mod  0=                          \ adjust cnt and see if 100th count?
    if  .ast  endif                               \ write a * every % completed
    >r over r@ - r> swap 0 <=
  until
  drop drop drop
;

: WhWa  25dots 0 SD_SMBW SD_WW SD_EMBW ;

[then]


\ ************
\ *S Test Code
\ ************

SDtest? [if]

: /S  ( -- )  \ Reset the data stack
  S0 @ SP!
;

VARIABLE #PLACES              \ Counts number of digits

: ',' ( -- ) [CHAR] , HOLD ;  \ Inserts a comma

: (D.ENG) ( d -- c-addr n)    \ Formats the string
  SWAP OVER DABS              \ Set up stack
  0 #PLACES !  <#             \ Initialize place counter
    BEGIN                     \ Start the conversion
      # 1 #PLACES +!          \ Increment counter
      2DUP D0= 0=
    WHILE                     \ More significant digits?
      #PLACES @ 3 MOD 0=      \ Every 3 digits...
      if  ',' endif           \ ... insert a comma
      REPEAT
  ROT SIGN #> ;               \ Append sign, finish

: D.ENG ( d -- ) (D.ENG) TYPE SPACE ;

: .HEXX ( u1 n -- )
  BASE @ >R  HEX  0 <#  SWAP 1-
  0 ?DO  #  LOOP  #S  #> TYPE  R> BASE !
;

: ??   ( n -- n )   \ prettyprint the top of stack and retain
  DUP DUP $FFFF AND SWAP  #16 RSHIFT
  $24 EMIT 4 .HEXX  $3A EMIT 4 .HEXX  CR
;

: ?.   ( n -- )   \ prettyprint the top of stack and remove
  DUP $FFFF AND SWAP  #16 RSHIFT
  $24 EMIT 4 .HEXX  $3A EMIT 4 .HEXX  CR
;

: .. ( n -- )   \ prettyprint the top of stack and remove, no CR at the end
  DUP $FFFF AND SWAP  #16 RSHIFT
  $24 EMIT 4 .HEXX  $3A EMIT 4 .HEXX
;

: XDUMP		\ addr len --
\ *G Display (dump) len bytes of memory starting at addr.
[defined] FlushKeys [if] FlushKeys [then]
  bounds ?do				                    \ get limits
    cr  i 8 + i                         \ new line
    do  space i c@ .byte  loop          \ first eight bytes
    space  i $010 + i 8 +               \ gap
    do  space i c@ .byte  loop          \ second eight bytes
    2 spaces  i $010 + i                \ gap
    do  i c@ .ascii  loop               \ ascii o/p
    halt? ?leave			                  \ quit or halt
  $010 +loop                            \ until done
  cr
;

: ?SDSize SD_Size ." SD usable capacity : " d.eng ." bytes" cr ;

: FillInBuf  ( n -- ) $80 0 do dup InBuf I 4* + ! loop drop ;
: FillOutBuf  ( n -- ) $80 0 do dup OutBuf I 4* + ! loop drop ;

: ?IB  InBuf 512 dump ;
: ?OB  OutBuf 512 dump ;

: RSB ( n -- ) InBuf swap SD_RSB InBuf 512 xdump ;

: ?SDState	\ --
  CMD13 SHRESP RCA @  SD_CMD drop  R1 SD_RESP
  case  ."  [State = "
    0 of ." IDLE]" endof
    1 of ." READY]" endof
    2 of ." IDENT]" endof
    3 of ." STDBY]" endof
    4 of ." XFER]" endof
    5 of ." DATA]" endof
    6 of ." RCV]" endof
    7 of ." PRG]" endof
    8 of ." DIS]" endof
      ." *** UNKNOWN STATE ***"
  endcase
  ."  mode" CR
;

: ?SD		\ --
\ *G Display the SD peripheral registers.
  CR
  ."   SDIO_POWER : " _SDIO sdPOWER    + @  ?.
  ."   SDIO_CLKCR : " _SDIO sdCLKCR    + @  ?.
  ."     SDIO_ARG : " _SDIO sdARG      + @  ?.
  ."     SDIO_CMD : " _SDIO sdCMD      + @  ?.
  ." SDIO_RESPCMD : " _SDIO sdRESPCMD  + @  ?.
  ."   SDIO_RESP1 : " _SDIO sdRESP1    + @  ?.
  ."   SDIO_RESP2 : " _SDIO sdRESP2    + @  ?.
  ."   SDIO_RESP3 : " _SDIO sdRESP3    + @  ?.
  ."   SDIO_RESP4 : " _SDIO sdRESP4    + @  ?.
  ."  SDIO_DTIMER : " _SDIO sdDTIMER   + @  ?.
  ."    SDIO_DLEN : " _SDIO sdDLEN     + @  ?.
  ."   SDIO_DCTRL : " _SDIO sdDCTRL    + @  ?.
  ."  SDIO_DCOUNT : " _SDIO sdDCOUNT   + @  ?.
  ."     SDIO_STA : " _SDIO sdSTA      + @  ?.
  ."     SDIO_ICR : " _SDIO sdICR      + @  ?.
  ."    SDIO_MASK : " _SDIO sdMASK     + @  ?.
  ." SDIO_FIFOCNT : " _SDIO sdFIFOCNT  + @  ?.
  ."    SDIO_FIFO : " _SDIO sdFIFO     + @  ?.
  CR ;

: ?DMA2S3	\ --
\ *G Display the DMA2 S3 registers.
  CR
  ."    DMA2_LISR : " _DMA2 dmaLISR    + @  ?.
  ."    DMA2_S3CR : " _DMA2 dmaS3CR    + @  ?.
  ."     DMA2_FCR : " _DMA2 dmaS3FCR   + @  ?.
  CR ;

[then]
