\ spiLP17xxhard.fth - SPI for SD/MMC card on LPC17xx
\ Tested on Olimex LPC1766-STK board.

((
Copyright (c) 2007, 2008, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20110727 SFP003 Converted to LPC1766.
20081120 SFP002 Added card detect and write protect words.
20080731 SFP001 Correct initialisation of fast I/O.
))

only forth definitions
decimal

\ =================
\ *! spiLPC17xxhard
\ *T LPC17xx drivers for SD/MMC cards
\ =================
\ *P LPC17xx devices and boards vary in their use of various pins
\ ** on the SD card, in particular the CS output and the CD (card
\ ** detect) and WP (write protect) inputs. Consequently this file
\ ** is not totally generic. The default configuration is for an
\ ** Olimex LPC1766-STK card.

\ *P The standard SD/MMC card pinout is as follows.
\ *E         SD mode                                 SPI mode
\ ** Pin     Name    Type    Description             Name    Type    Description
\ ** 1       CD/DAT3 I/O/PP  Card detect / data 3    CS      I       Chip select (low)
\ ** 2       CMD     PP      Command/Response line   DI      I       Data input
\ ** 3       Vss1    S       Supply voltage (gnd)    VSS     S       Supply voltage
\ ** 4       Vdd     S       Power supply            VDD     S       Power supply
\ ** 5       CLK     I       Clock                   SCLK    I       Clock
\ ** 6       Vss2    S       Supply voltage          VSS2    S       Supply voltage
\ ** 7       DAT0    I/O/PP  data line 0             O       O/PP    Data output
\ ** 8       DAT1    I/O/PP  data line 1             RSV
\ ** 9       DAT2    I/O/PP  data line 2             RSV
\ *P where the pins are numbered on the connector as below:
\ *E   9 1 2 3 4 5 6 7 8
\ **     * * * * * * * *
\ **   *


\ ***************************
\ *S Setup and initialisation
\ ***************************
\ *P The Olimex board uses I/O for the SD card as follows.
\ *E   P0.21  MMC_PWR    pad 57
\ **   P0.6   SSEL1      pad 79    CD/DAT3/CS
\ **   P0.7   SCK1       pad 78    CLK/SCLK
\ **   P0.8   MISO1      pad 77    DAT0/DO
\ **   P0.9   MOS11      pad 76    CMD/DI
\ **   P1.28             pad 44    CP
\ **   P1.29             pad 45    WP

0 [ps		\ PINSEL1 values for SSP without SSEL
  6 +ps			\ CS
2 [psm]
  7 +ps			\ SCK
  8 +ps			\ MISO
  9 +ps			\ MOSI
ps] equ ssp_PS0val  equ ssp_PS0mask

: +sdPwr	\ --
\ *G Apply power to the card.
  bit21 _FIO0  2dup FIODIR + or!  FIOCLR + !
;

: -sdPwr	\ --
\ *G Remove power from the card.
  bit21 _FIO0  2dup FIODIR + bic!  FIOSET + !
;

: InitSSP1	\ freq --
\ *G Initialise SSP1 port at the given frequency.
  system-speed 4 /  swap /mod swap	\ clock divisor
  if  1+  then
  1- 8 lshift >r			\ bits 15:8 in sspCR0

  $0000:0000 $0000:0C00 _PINSEL1 setMask  \ P0.21 as GPIO
  $0000:0000 $0000:0C00 _PINMODE1 setMask  \ P0.21 with pull up
  bit21 _FIO0				\ P0.21=MMC_PWR
  2dup FIOSET + !			\ set output to 1 (high=off)
  FIODIR + or!				\ set as output
\  0 $0F00:0000 _PINSEL3 setMask		\ P1.28, P1.29 as GPIO
\  0 $0F00:0000 _PINMODE3 setMask	\ P1.28, P1.29 with pull ups
\  bit28 bit29 or _FIO1			\ P1.28=WP, P1.29=CD
\  FIODIR + bic!				\ force as inputs

  ~pcSSP1 _PCONP or!			\ power up SSP1
  ssp_PS0val ssp_PS0mask _PINSEL0 setMask	\ pin set up
  $0020:0000 $0030:0000 _PCLKSEL0 setMask	\ clock divide by two
  bit6 _FIO0  2dup FIOSET + !  FIODIR + or!	\ CS o/p high

  _SSP1					\ peripheral base address
  2 over sspCPSR + !			\ prescale by 2, system/4
  r> 7 or over sspCR0 + !		\ set divisor and 8 bit transfers
  2 over sspCR1 + !			\ enable SSP
  drop

  +sdPwr
;


\ ****************
\ *S SPI transfers
\ ****************
\ *P These routines have been optimised for speed over space
\ ** considerations. For high speed use with SD cards, you will
\ ** get best performance using interrupt driven or DMA driven
\ ** operation.

(( SSP status register
Bit	Symbol	Description			Reset value
0 	TFE	Transmit FIFO Empty. This bit is 1 is the Transmit FIFO is
		empty, 0 if not.		1
1 	TNF	Transmit FIFO Not Full. This bit is 0 if the Tx FIFO is
		full, 1 if not.			1
2	RNE	Receive FIFO Not Empty. This bit is 0 if the Receive
		FIFO is empty, 1 if not.	0
3	RFF	Receive FIFO Full. This bit is 1 if the Receive
		FIFO is full, 0 if not.		0
4	BSY	Busy. This bit is 0 if the SSP controller is idle,
		or 1 if it is currently sending/receiving a frame
		and/or the Tx FIFO is not empty. 0
))

compiler
: SPI1Xchg	\ bout -- bin
\ *G Write data to SPI1 and return the received byte.
  _SSP1				\ -- bout *ssp
  begin				\ Wait for FIFO to be empty
    dup SSPSR + @ bit0 and
  until
  tuck SSPDR + !		\ -- *ssp ; Write to device
  begin				\ Wait for bus activity to finish
    dup SSPSR + @ bit4 bit2 or and bit2 =
  until
  SSPDR + @ $FF and		\ return received byte
;
target

IsLeaf : spiFlush	\ --
\ *G Flush the input queue.
  _SSP1
  begin					\ Wait for not busy
    dup SSPSR + @ bit4 and 0=
  until
  begin					\ Flush RXq
    dup SSPSR + @ bit2 and
   while
    dup SSPDR + @ drop
  repeat
  drop
;

IsLeaf : spi!	\ b --
\ *G Write data to SPI1 and discard the received byte.
  SPI1Xchg drop
;

IsLeaf : spi@	\ -- b
\ *G Write an idle byte to the SPI and return the received byte.
  $FF SPI1Xchg
;

IsLeaf : SPIidle	\ --
\ *G Send an 0xFF byte to the SSP and discard the received byte.
  $FF SPI1Xchg drop
;

IsLeaf : spiwr	\ addr -- addr'
\ *G Write byte at *\i{addr} to SPI and increment pointer.
  dup c@ SPI1Xchg drop  1+
;

IsLeaf : spird	\ addr -- addr'
\ *G Read byte from SPI, put it at *\i{addr}, increment pointer.
  $FF SPI1Xchg over c! 1+
;

: MMCon		\ --
\ *G Assert CS.
  bit6 _FIO0 FIOCLR + !		\ Set pin low
;

: MMCoff	\ --
\ *G Negate CS.
  bit6 _FIO0 FIOSET + !		\ Set pin high
;

\ SFP002...
: CDin?		\ -- flag
\ *G Return non-zero if a card is present.
  -1
\  _FIO1 FIOPIN + @ bit28 and 0=	\ P1.28
;

: WPin?		\ -- flag
\ *G Return non-zero if the write protect slider is set.
  0
\  _FIO1 FIOPIN + @ bit29 and 0<>	\ P1.29
;
\ ...SFP002

: initSDspi	\ --
\ *G Initialise SPI1 for SD/MMC card use.
  #400000 initSSP1
;

: spiSlow	\ --
\ *G Set the SPI speed to the SD/MMC low speed of 400kHz.
  #400000 initSSP1
;

: spiFast	\ --
\ *G Set the SPI speed to the SD/MMC operating speed of 2 MHz.
\ ** Note that the high speed is often limited by the quality
\ ** of the hardware interface. At higher speeds termination
\ ** resistor values and PCB tracking need attention.
\ ** By default, the high speed is set to 2MHz, which works on
\ ** most hardware. Tune this speed as required.
  #2000000 initSSP1
;


\ ======
\ *> ###
\ ======

decimal
