\ spiLPC214hard.fth - SPI for SD/MMC card on LPC2148

((
Copyright (c) 2007, 2008, 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20130206 SFP003 Corrected SPI1dev table PINSEL entry.
20081120 SFP002 Added card detect and write protect words.
20080731 SFP001 Correct initialisation of fast I/O.
))

only forth definitions
decimal

\ =================
\ *! spiLPC2148hard
\ *T LPC2148 SPI drivers for SD/MMC cards
\ =================
\ *P LPC2xxx devices and boards vary in their use of various pins
\ ** on the SD card, in particular the CS output and the CD (card
\ ** detect) and WP (write protect) inputs. Consequently this file
\ ** is not totally generic. The default configuration is for an
\ ** Olimex P2148 card.

\ ***************************
\ *S Setup and initialisation
\ ***************************
\ *P PINSELx register calculation is eased by the following three
\ ** *\fo{INTERPRETER} words. Pin numbers are clipped to 0..15
\ ** so that the resulting masks can be used with any of the
\ ** PINSELx registers.

interpreter
: [ps		\ mode -- mode mask orval
\ *G Start a pin select calculation by specifiying the mode (0..3)
\ ** for a group of pins.
  3 and
  dup 2 lshift or			\ 4 bits
  dup 4 lshift or			\ 8 bits
  dup 8 lshift or			\ 16 bits
  dup #16 lshift or			\ 32 bits
  0 0
;

: +ps		\ mode mask orval pin# -- mode mask' orval'
\ *G Given a *\i{pin#} in the range 0..31, update the *\i{mask}
\ ** and *\i{orval} values.
  $0F and				\ clip to 0..15
  1 swap 2* lshift dup 1 lshift or	\ 2 bit mask
  >r					\ -- mode mask orval ; R: -- pinmask
  swap r@ or swap			\ merge in mask
  2 pick r> and or			\ merge mode
;

: ps]		\ mode mask orval -- mask orval
  rot drop
;
target

\ *P Use the mask value to clear the PINSELx bits and then use
\ ** the value to set the required bits, e.g.
\ *E   _PINSELx dup @          \ read PINSELx
\ **   <mask> invert and       \ clear bits to 0 (GPIO mode)
\ **   <orval> or              \ set selected mode bits
\ **   swap !                  \ write PINSELx

1 [ps		\ PINSEL0 values for SPI0 without SSEL
  4 +ps			\ SCK
  5 +ps			\ MISO
  6 +PS			\ MOSI
ps] equ spi0_PS0val  equ spi0_PS0mask

1 [ps		\ PINSEL0 values for SPI0 with SSEL
  4 +ps			\ SCK
  5 +ps			\ MISO
  6 +ps			\ MOSI
  7 +ps			\ SSEL
ps] equ spi0ssel_PS0val  equ spi0ssel_PS0mask

2 [ps		\ PINSEL1 values for SSP without SSEL
  17 +ps		\ SCK
  18 +ps		\ MISO
  19 +ps		\ MOSI
ps] equ spi1_PS1val  equ spi1_PS1mask

2 [ps		\ PINSEL1 values for SSP with SSEL
  17 +ps		\ SCK
  18 +ps		\ MISO
  19 +ps		\ MOSI
  20 +ps		\ SSEL
ps] equ spi1ssel_PS1val  equ spi1ssel_PS1mask

struct /spiDev	\ -- len
\ *S SPI device structure.
\ *[
  1 field spi.flags	\ flags field: for SD/MMC = 00000110
			\ bit 7 = 1 = SSP, 0 = SPI
			\ bit 2 = 1 = set pins
			\ bit 1 = 1 = msb first
			\ bit 0 = 1 = use SSEL
  aligned		\ skip padding
  int spi.Base		\ base address of SPI port I/O
  int spi.PCONP		\ PCONP bit mask
  int spi.psel		\ PINSELx register to use
  int spi.psmask	\ PINSELx mask value
  int spi.psset		\ value to or in
end-struct
\ *]

struct /sspDev	\ -- len
\ *S SSP device structure.
\ *[
  1 field ssp.flags	\ flags field: for SD/MMC = 10000110
			\ bit 7 = 1 = SSP, 0 = SPI
			\ bit 2 = 1 = set pins
			\ bit 1 = 1 = msb first
			\ bit 0 = 1 = use SSEL
  aligned
  int ssp.Base		\ base address of SPI port I/O
  int ssp.PCONP		\ PCONP bit mask
  int ssp.psel		\ PINSELx register to use
  int ssp.psmask	\ PINSELx mask value
  int ssp.psset		\ value to or in
  int ssp.CR0		\ SSPCR0
  int ssp.CR1		\ SSPCR1
end-struct
\ *]


create SPI0dev	\ -- addr
\ *G Device table for SPI0 device - must match *\fo{/spiDev} above.
  $06 c,		\ flags
  0 c,  0 c,  0 c,	\ padding
  _SPI0 ,		\ base address
  bit8 ,		\ PCONP enable bit
  _PINSEL0 ,		\ PINSEL register
  spi0_PS0mask ,	\ PINSEL mask
  spi0_PS0val ,		\ PINSEL value to or in

create SPI1dev	\ -- addr
\ *G Device table for SPI1 device - must match *\fo{/sspDev} above.
  $86 c,		\ flags
  0 c,  0 c,  0 c,	\ padding
  _SSP ,		\ base address
  bit10 ,		\ PCONP enable bit
  _PINSEL1 ,		\ PINSEL register
  spi1_PS1mask ,	\ PINSEL mask
  spi1_PS1val ,		\ PINSEL value to or in
  #7 ,			\ Default SSPCR0
  #2 ,			\ Default SSPCR1

[undefined] setMask [if]	\ now in CodeARM.fth
: setMask	\ value mask addr --
\ *G Clear the *\i{mask} bits at *\i{addr} and set (or) the
\ ** bits defined by *\i{value}.
  tuck @				\ -- value addr mask x
  swap invert and			\ -- value addr x'
  rot or				\ -- addr x''
  swap !
;
[then]

: initSPI	\ freq devtab -- freq'|0
\ *G Initialise an LPC21xx SPI port.
  dup spi.Base @ >r			\ base address of peripheral
  dup spi.PCONP @ _PCONP or!		\ power up (enable) peripheral
  dup spi.psset @ over spi.psmask @	\ freq devtab psset psmask
  2 pick spi.psel @ rot rot 2 pick @	\ freq devtab _PINSELn psval psmask PINSELn
  swap invert and or			\ freq devtab _PINSELn PINSELn'
  swap !				\ write PINSELx
  system-speed /vpb / rot /		\ calculate divider
  8 max  #254 min  -2 and		\ even value in range 8..254
  dup r@ SPCCR + !			\ set SPI clock divider
  system-speed /vpb / swap / swap	\ -- freq' devtab ; real frequency
  bit5  over spi.flags c@ bit0 and 0=	\ master
  if  bit6 or  endif			\ flag bit0=0 for LSB first
		\ *** Need to check state of control reg here!
  r@ SPCR + or!				\ write bits into control reg
  r@ SPDR + @ drop			\ read to clear
  r@ SPSR + @ drop			\ read to clear
  bit0 r> SPINT + !			\ clear SPIE
  drop					\ discard devtab
;

: InitSSP	\ freq devtab -- freq'|0
\ *G Initialise an LPC21xx SSP port. Note that *\i{freq}
\ ** must be less than PCLK/2.
  bit0 bit1 or _SYSCON SCS + or!	\ enable fast I/O for GPIO0 and GPIO1 ; SFP002
  bit20 _FIO0				\ P0.20=CS
  2dup FIOSET + !			\ set output to 1 (high=off)
  FIODIR + or!				\ set as output ; SFP001
  bit24 bit25 or _FIO1			\ P1.24=WP, P1.25=CD
  FIODIR + bic!				\ force as inputs

  dup ssp.Base @ >r			\ base address of peripheral
  dup ssp.PCONP @ _PCONP or!		\ power up (enable) peripheral
  dup ssp.psset @ over ssp.psmask @	\ -- freq devtab psset psmask
  2 pick ssp.psel @ setMask		\ -- freq devtab
  system-speed /vpb / rot /mod		\ -- devtab div rem ; calculate divider
  swap if  1+  endif
  tuck 1 rshift 1-			\ -- div devtab div' ; SCR for CPSDVSR=2
  8 lshift				\ Bits 15:8 hold SCR
  over ssp.CR0 @ or r@ SSPCR0 + !	\ -- div devtab ; SSPCR0
  2 r@ SSPCPSR + !			\ Set CPSDVSR to 2
  ssp.CR1 @ r> SSPCR1 + !		\ -- div ; SSPCR1

  system-speed /vpb / swap /		\ -- freq'
;

: SPIdevInit	\ freq devtab -- freq'|0
\ *G Initialise the SPI port specified by the *\i{devtab} table
\ ** to the given frequency *\i{freq}. On success, the actual
\ ** frequency used is returned. On failure, zero is returned.
  dup spi.flags c@ bit7 and if
    initSSP
  else
    initSPI
  endif
;


\ ****************
\ *S SPI transfers
\ ****************
\ *P These routines have been optimised for speed over space
\ ** considerations. For high speed use with SD cards, you will
\ ** get best performance using interrupt driven or DMA driven
\ ** operation.

(( SSP status register
Bit	Symbol	Description			Reset value
0 	TFE	Transmit FIFO Empty. This bit is 1 is the Transmit FIFO is
		empty, 0 if not.		1
1 	TNF	Transmit FIFO Not Full. This bit is 0 if the Tx FIFO is
		full, 1 if not.			1
2	RNE	Receive FIFO Not Empty. This bit is 0 if the Receive
		FIFO is empty, 1 if not.	0
3	RFF	Receive FIFO Full. This bit is 1 if the Receive
		FIFO is full, 0 if not.		0
4	BSY	Busy. This bit is 0 if the SSP controller is idle,
		or 1 if it is currently sending/receiving a frame
		and/or the Tx FIFO is not empty. 0
))

compiler
: SPI1Xchg	\ bout -- bin
\ *G Write data to SPI1 and return the received byte.
  _SSP				\ -- bout *ssp
  begin				\ Wait for FIFO to be empty
    dup SSPSR + @ bit0 and
  until
  tuck SSPDR + !		\ -- *ssp ; Write to device
  begin				\ Wait for bus activity to finish
    dup SSPSR + @ bit4 bit2 or and bit2 =
  until
  SSPDR + @ $FF and		\ return received byte
;
target

IsLeaf : spiFlush	\ --
\ *G Flush the input queue.
  _SSP
  begin					\ Wait for not busy
    dup SSPSR + @ bit4 and 0=
  until
  begin					\ Flush RXq
    dup SSPSR + @ bit2 and
   while
    dup SSPDR + @ drop
  repeat
  drop
;

IsLeaf : spi!	\ b --
\ *G Write data to SPI1 and discard the received byte.
  SPI1Xchg drop
;

IsLeaf : spi@	\ -- b
\ *G Write an idle byte to the SPI and return the received byte.
  $FF SPI1Xchg
;

IsLeaf : SPIidle	\ --
\ *G Send an 0xFF byte to the SSP and discard the received byte.
  $FF SPI1Xchg drop
;

IsLeaf : spiwr	\ addr -- addr'
\ *G Write byte at *\i{addr} to SPI and increment pointer.
  dup c@ SPI1Xchg drop  1+
;

IsLeaf : spird	\ addr -- addr'
\ *G Read byte from SPI, put it at *\i{addr}, increment pointer.
  $FF SPI1Xchg over c! 1+
;

system-speed #1000000 / equ clks/us	\ -- u
\ *G Number of clock cycles in a microsecond.
5 clks/us * 2 / 1- equ loops/2.5us	\ -- u
\ *G Number of loops required for a 2.5us delay.

((
IsLeaf : us2.5	\ --
 \ *G A delay of 2.5 microseconds. This is used around *\fo{MMCon}
 \ ** and *\fo{MMCoff} below when the CS pin is a GPIO pin. This
 \ ** delay is chosen because it corresponds to one bit time at the
 \ ** SD low speed of 400kHz.
 \ ** By default this word is commented out. It is here for testing
 \ ** *\fo{MMCon} and *\fo{MMCoff} below.
  loops/2.5us
  begin  1- dup 0= until
  drop
;
))
: MMCon		\ --
\ *G Assert CS.
\  us2.5
  bit20 _FIO0 FIOCLR + !		\ Set pin low
\  us2.5
;

: MMCoff	\ --
\ *G Negate CS.
\  us2.5
  bit20 _FIO0 FIOSET + !		\ Set pin high
\  us2.5
;

\ SFP002...
: CDin?		\ -- flag
\ *G Return non-zero if a card is present.
  _FIO1 FIOPIN + @ bit25 and 0=
;

: WPin?		\ -- flag
\ *G Return non-zero if the write protect slider is set.
  _FIO1 FIOPIN + @ bit24 and 0<>
;
\ ...SFP002

: initSDspi	\ --
\ *G Initialise SPI1 for SD/MMC card use.
  #400000 Spi1dev SPIdevInit 0= abort" Can't initialise SPI1"
;

: spiSlow	\ --
\ *G Set the SPI speed to the SD/MMC low speed of 400kHz.
  #400000 Spi1dev SPIdevInit drop
;

: spiFast	\ --
\ *G Set the SPI speed to the SD/MMC operating speed of 2 MHz.
\ ** Note that the high speed is often limited by the quality
\ ** of the hardware interface. At higher speeds termination
\ ** resistor values and PCB tracking need attention.
\ ** By default, the high speed is set to 2MHz, which works on
\ ** most hardware. Tune this speed as required.
  #2000000 Spi1dev SPIdevInit drop
;


\ ======
\ *> ###
\ ======

decimal
