\ FatCore.fth - FAT file system core code

((
Copyright (c) 2006, 2007 Brad Eckert

Copyright (c) 2007, 2008, 2009, 2010, 2011, 2012, 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


Test platforms:
  VFX Forth for Windows
  Olimex LPC-P2148 board
  STM3240G-EVAL board

To do
=====
Some nice features to have:
  Multiple file paths

Change history
==============
20150222 BJC026 Fixed _RESIZE.
20130314 MPE025 Moved CWD and PWD here from PlainDir.fth.
		Changed TEST-FILE to use attibutes $18 so as not
		to find directories and volume IDs.
20130313 SFP024 DELETE-FILE will accept names with paths.
20130301 SFP023	Added RENAME-FILE.
		_TEST-FILE saves directory entry and sector
		Allowed ".." in GOODNAME.
		Rewrote NAME.EXT for pretties.
		Removed CD.. as it is unused.
20130227 SFP022 Renamed $CD to $CWD to reduce parsing errors.
20121214 SFP021 Corrected CD+.
20110816 SFP020 Corrected writing of attribute on disk on create.
20110318 BJC019 Fixed DIR].
		Enhanced robustness on error by preventing locks.
		Removed NODUBL.
20101125 SFP018 More configuration equates.
20101124 SFP017 Reset current directory at initialisation.
20101109 SFP016	Forced use of entry 0 when extending directory.
20101104 SFP015 Fixed directory bug in getdir.
20090216 SFP014 Added more configuration options.
20090105 SFP013 Renamed INITCACHE to INITSECCACHE - too many caches!
20081212 SFP012 Changed to make the system task safe.
20081211 MPE011 Renamed CD to $CD to avoid usage problems.
20081208 SFP010 Fixed bug in initialisation of CD.
		Fixed bug in failed OPEN-FILE.
		Added MAX_PATH.
20081024 SFP009 Added FILEEXISTS? and the ANS FILE-STATUS.
		Fixed TESTLINE to avoid AIDE triggers.
20070514 SFP008 Fixed REPO-FILE bug.
		WIPECLUS no longer returns an ior as SECWRITE throws.
		Changed some < or > to U< or U>.
20070513 SFP007 Changed sector primitives to SECREAD and SECWRITE.
                Added LRU cache system using transaction numbers.
		Removed restriction for binary numbers of buffers.
20070510 MPE006 Moved directory listing code to PlainDir.fth.
		Improved portability with uaLExx@/! words.
20070505 BRE005 Merged latest changes with improved error reporting.
20070501 MPE004 Made file name search case-insensitive.
		Made file name creation force upper case.
		DIR resets to start of directory.
20070419 MPE003 Hardware layer isolated and DocGenned.
20060802 BRE002 Received update from author.
		fixed DIRUP bug (added drop at end)
		made NBUF 13 bytes, not 12.
		added recursive file properties
		fixed dir++ and dir-- subdirectory traversal bug
20060726 BRE001 Received from author.

\ 004:  renumbered IORs:
\       1/2/3 = error reading device (hardware specific from rawread)
\       4/5/6 = error writing device (hardware specific from rawwrite)
\       7 = not enough free file handles
\       8 = no room left in the root directory
\       9 = error in FAT parameters
\       10 = file size longer than 32-bit
\       11 = file was not found
\       12 = not a valid directory
\       13 = directory was not found
\       14 = attempt to read W/O file
\       15 = attempt to write R/O file
\       16 = invalid created pathname
\       17 = invalid created filename
\       18 = full directory
\       19 = invalid directory name
))

only forth definitions
decimal

\ ==========
\ *! fatintro
\ *T FAT file system introduction
\ ==========
\ *P The file *\i{FatCore.fth} contains the core FAT filing system.
\ ** before you compile this file you must provide some configuration
\ ** information, a prortability harness for your Forth kernel,
\ ** and a sector read/write interface for your disk system.
\ ** Some example interfaces are provided in the issue code.

\ *P To build the code, there are example build files, including
\ ** amongst others:
\ *D FatVfxW.bld  Builds a demonstration system on VFX Forth for
\ **              Windows. This can be used with any drive formatted
\ **              with a FAT file system.
\ *D SSP2148.bld  Builds code for an Olimex LPC-P2148 board using the
\ **              SSP (SPI1) driver from the ARM cross compiler.
\ *D mci23xx.bld  Builds code for an NXP device using the MCI
\ **              controller (4 bit SD card) on LPC23xx/24xx and
\ **              some LPC17xx devices.
\ *D qspi52259.bld Builds code for the QSPI device on Coldfire
\ **              MCF522xx devices.
\ *D STM32F4eval.bld Builds code for an STM32F4xx on the ST
\ **              STM3240G-EVAL board.
\ *D STM32F4disco.bld Builds code for an STM32F4xx on the STM32F4
\ **              Discovery board.

\ *P The *\i{Docs} folder includes useful documentation about
\ ** FAT file systems, including *\i{fatgen103.pdf} from Microsoft.

\ *P Some useful links:
\ *C http://www.seas.ucla.edu/classes/mkampe/cs111.sq05/docs/dos.html

\ ====================
\ *S Design objectives
\ ====================
\ *P This is a filing system for embedded systems, in particular for
\ ** use with SD/SDHC/MMC and other Flash cards. You must provide the
\ ** sector read and write routines.

\ *P Because the filing system is designed for use with removable
\ ** media, it is assumed that they will be formatted on a PC.

\ *P RAM usage is more important than performance, although the
\ ** use of the LRU sector cache considerably improves performance
\ ** with both hard and floppy drives, as well as with slow
\ ** Flash drives.

\ ===============
\ *S Capabilities
\ ===============
\ *(
\ *B FAT112, FAT16 and FAT32 are all supported.
\ *B Directories and subdirectories are supported.
\ *B The ANS Forth File wordset is supported.
\ *B Can be used with any media type having a sector
\ ** read and write interface.
\ *)

\ ===============
\ *S Restrictions
\ ===============
\ *(
\ *B A 32 bit host Forth is assumed.
\ *B Sectors must be 512 bytes in length. As a result, 2Gb standard
\ ** capacity SD cards may not work, but SDHC cards will work.
\ *B Sector numbers are 32 bit unsigned numbers. This restricts
\ ** drive size to 2048Gb (2Tb) before considering restrictions
\ ** in the FAT32 data.
\ *B Only one drive is supported, although future expansion is
\ ** provided for.
\ *B File size is internally held as a 32 bit number, and is thus
\ ** restricted to 4Gb per file.
\ *B The file system is not thread safe. Enable the interlock (e.g. a
\ ** semaphore) to the ANS file layer if you need thread safe
\ ** operation.
\ *B Directory deletion must be performed file by file, unless
\ ** you use the extension in *\i{DelDirs.fth}.
\ *B File names must be in the old DOS 8.3 format. Long file
\ ** names are not supported, but drives with long file names
\ ** can be handled using the short alias.
\ *B Drive formatting is not supported.
\ *B The file system must be explicitly initialised at each power
\ ** up and insertion of a card/drive.
\ *)


\ **************
\ *S Build files
\ **************
\ *P Each supported configuration has a build file with a *\i{.bld}
\ ** extension. The build file contains equates describing the
\ ** maximum number of open files, number of sector buffers and
\ ** some CPU characteristics. The build file than loads a the
\ ** required files, e.g. for an STM32F4 board:
\ *E include %FFDir%\XC6harn               \ harness
\ ** include %FFDir%\Drivers\sdioSTM32F4xx \ STM32F4 SDIO (1-bit DMA)
\ ** include %FFDir%\FatCore               \ core file system code
\ ** include %FFDir%\PlainDir              \ simple directory listing
\ ** include %FFDir%\DelDirs               \ directory and disk deletion
\ *P Note the use of text macros to define the directory containing
\ ** the file system code.

\ *P The file *\i{sdioSTM32F4xx.fth} is the CPU specific driver.
\ ** Where there are hardware differences, e.g. for the card
\ ** detect pin, write protectpin, power control or pin assignments,
\ ** there may also be a file describing the hardware. This is
\ ** often called *\i{hwxxxxxx.fth) and may be compiled before
\ ** the build file.

\ ************************
\ *S Configuration options
\ ************************
\ *P The following equates control compilation of some features.
\ ** The defaults here are used if the equates have not previously
\ ** been defined, e.g.in your control file.

[undefined] FATopen? [if]
1 equ FATopen?	\ -- flag
\ *G Set non-zero to compile the FAT code with all heads present.
[then]

[undefined] FATtext? [if]
1 equ FATtext?	\ -- flag
\ *G Set non-zero to compile the FAT code for *\fo{READ-LINE}
\ ** and *\fo{WRITE-LINE}.
[then]

[undefined] FATio? [if]
1 equ FATio?	\ -- flag
\ *G Set non-zero to compile the FAT code with a file generic I/O
\ ** device.
[then]

[undefined] FATspace? [if]
1 equ FATspace?	\ -- flag
\ *G Set non-zero to compile the FAT code for used space.
[then]

[undefined] FATtest? [if]
0 equ FATtest?	\ -- flag
\ *G Set non-zero to compile the FAT code with test code.
[then]


\ ==========
\ *! fatcore
\ *T FAT file system core
\ ==========
\ *P The file *\i{FatCore.fth} contains the bulk of the file
\ ** system code.

decimal

FATopen? 0= [if]  internal  [then]


\ *****************************
\ *S Buffers, globals and tools
\ *****************************
\ *P File reads and writes use sector cache buffers for file I/O.
\ ** One buffer gives the lowest performance, since accessing
\ ** more than one sector causes disk thrashing. *\fo{WRITE-FILE}
\ ** accesses the FAT table and the data sectors of the disk so
\ ** there should be at least two sector buffers for each file
\ ** being actively written. *\fo{READ-LINE} tends to like more
\ ** than two buffers - six are recommended, especially with
\ ** slow drives.

#64 equ MAX_PATH	\ -- len
\ *G The maximum size of a pathname including the file name.
\ ** This equate is not used here, but is simply a suggested
\ ** value used elsewhere.

#buffers bytes/sec * buffer: bufs	\ -- addr
\ *G Sector cache.
#buffers cells buffer: csects	\ -- addr
\ *G Holds sector numbers of the currently buffered sectors.
#buffers cells buffer: ctr#s	\ -- addr ; SFP007
\ *G Holds transaction numbers of the currently buffered sectors.
\ ** The lowest number identifies the oldest sector.
#buffers buffer: wrpendings	\ -- addr
\ *G Holds the write-pending flags for the currently buffered sectors.
#buffers buffer: pdrives	\ -- addr
\ *G Holds the physical devices for the currently buffered sectors.
variable buff#	\ -- addr
\ *G Holds the next available sector buffer number (0, 1, ..).
variable tr#	\ -- addr ; SFP007
\ *G Holds the current transaction number.
cvariable cdrive	\ -- addr
\ *G Holds the current drive number, which must be 0 for the
\ ** moment.
  0 cdrive c!


\ *P The following variables hold the disk characteristics.
\ ** These will be changed at a later date to support multiple
\ ** drives by providing a structure per drive.

variable SectorsPerTrack	\ -- addr
\ *G Holds the number of sectors per track for the raw disk.
variable NumHeads	\ -- addr
\ *G Holds the number of heads for the raw disk.
variable rootentries	\ -- addr
\ *G Number of 32-byte directory entries in the root directory.
variable rootdsecs	\ -- addr
\ *G Number of sectors in the root directory.
variable rootsector	\ -- addr
\ *G Sector of root directory.
variable sec/cluster	\ -- addr
\ *G Sectors per cluster.
variable clusshift	\ -- addr
\ *G Equivalent shift count for sectors per cluster.
variable reservedsec	\ -- addr
\ *G Reserved sectors: 1 for FAT12/16, usually 32 for FAT32.
variable totalsec	\ -- addr
\ *G Total number of sectors.
variable FATsize	\ -- addr
\ *G FAT size in sectors.
variable basesector     \ -- addr
\ *G Start sector of the current partition, currently 0
  0 basesector !

variable dcurs	\ -- addr
\ *G Holds the display cursor position. This is an aid for later
\ ** directory listing tools.
2variable savedDirEnt	\ -- addr
\ *G Directory sectory and entry saved when a file/directory is found.

\ *P The following words manipulate the data above.

: initSecCache	\ --
\ *G Clear the sector cache buffers.
  0 buff# !
  csects #buffers cells -1 fill
  ctr#s #buffers cells erase  0 tr# !	\ SFP007
  wrpendings #buffers erase
  pdrives #buffers erase
;

: buf		\ -- addr
\ *G return the address of the current/next sector cache buffer.
  buff# @ secs>bytes bufs +
;

: csect		\ -- addr
\ *G Return the address holding the current sector number.
  buff# @ cells csects +
;

: ctr#		\ -- addr
\ *G Return the address holding the current transaction number.
  buff# @ cells ctr#s +
;

: writePend	\ -- addr
\ *G Return the address of the current sector write-pending flags (bytes).
  buff# @ wrpendings +
;

: pdrive	\ -- addr
\ *G Return the address of the device number (byte) in the cache line.
  buff# @ pdrives +
;

: bumpBuff#	\ --
\ *G Step the current sector to the next sector cache buffer.
\  buff# @ 1+  #buffers 1- and  buff# !
  buff# @ 1+ dup #buffers < and buff# !
;

: wmark		\ --
\ *G Mark the current sector cache buffer as modified, so that
\ ** it will be written back.
  1 writepend c!
;

: RootDirSectors	\ -- n
\ *G Number of sectors for the root directory.
  rootentries @ $0F + 4 rshift
;

: StartSector	\ -- n
\ *G Return the sector number of the first data sector.
  RootDsecs @  FATsize @ 2* +  reservedsec @ +  basesector @ +
;

FastShift? 0= [if]
: c>s		\ nclus -- nsec
\ *G Convert a number of clusters to the corresponding number of sectors.
\ ** The *\fo{EQU FastShift?} determines whether this word is implemented
\ ** as a multiply by *\fo{sec/cluster} or a shift by *\fo{clusshift}.
  sec/cluster @ *
;

: s>c		\ nsec -- nclus
\ *G Convert a number of sectors to the corresponding number of clusters.
\ ** The *\fo{EQU FastShift?} determines whether this word is implemented
\ ** as a divide by *\fo{sec/cluster} or a shift by *\fo{clusshift}.
  sec/cluster @ /
;
[else]
: c>s		\ nclus -- nsec
  clusshift @ lshift
;

: s>c		\ nsec -- nclus
  clusshift @ rshift
;

: s>cn		\ nsec -- nclus noff
  clusshift @ 2dup rshift -rot  1- and
;
[then]

: cluster>sec	\ cluster -- sector
\ *G Converts a cluster number to a physical sector.
  2 - c>s StartSector +
;

: sec>cluster	\ sector -- cluster
\ *G Converts a physical sector to a cluster number.
  StartSector - s>c 2 +
;

: b>c		\ bytes -- clusters
\ *G Returns the number of clusters needed to hold a number of bytes.
  secsizemask + bytes>secs  s>c
;

: b>cn		\ bytes -- nclus noff
\ *G Convert a number of sectors to a number of complete clusters and the
\ ** sector offset (0..noff) within the following partially filled cluster.
  bytes>secs dup s>c swap 		\ -- nclus nsecs
  sec/cluster @ 1- and			\ -- nclus noff
;

: c>b		\ clusters -- bytes
\ *G Converts clusters to a byte count.
  c>s secs>bytes
;

: CountofClusters	\ -- n
\ *G Number of clusters available for data. According to Microsoft,
\ ** the FAT type (12/16/32) depends solely on this number of
\ ** clusters, and not on the ASCII string in sector 0.
  totalsec @  reservedsec @ FATsize @ 2* +
  RootDsecs @ +  -
  s>c
;

: nodir		\ --
\ *G Reset the directory display cursor.
  0 dcurs !
;


\ ********************
\ *S Sector operations
\ ********************

: wflush	\ --
\ *G If the current sector cache buffer is marked as modified,
\ ** write it back.
  writepend c@ if
    buf csect @ cdrive c@ SecWrite
    0 writepend c!
  then
;

: flushall	\ --
\ *G Write all marked sector cache buffers back.
  #buffers FOR bumpbuff# wflush NEXT
;

: seeksector	\ sector -- found? ; look for a current sector
\ *G Return true if the given sector is already cached.
  #buffers FOR
    bumpbuff#  dup csect @ =
    if  drop r> drop  true  exit then
  NEXT
  drop 0  bumpbuff#
;

: setCtr#	\ --
\ *G Update the global transaction number and set the current
\ ** sector's transaction number from it.
  tr# @ 1+ dup tr# !  ctr# !
;

: oldest	\ --
\ *G Set *\fo{buff#} to the oldest cached sector.
  buff# @ ctr# @  #buffers FOR		\ -- buff# trdiff
    bumpbuff#  ctr# @ 2dup - 0< if	\ -- buff# trdiff trdiff'
      drop
    else
      nip nip  buff# @ swap		\ -- buff#' trdiff'
    then
  NEXT
  drop  buff# !
;

: readsector	\ sector --
\ *G Ensure sector is in the cache buffers.
  dup csect @ =
  if  drop setCtr# exit  then		\ no need to change sectors
  dup seeksector
  if  drop setCtr# exit  then		\ desired sector is in cache
  oldest wflush setCtr#			\ -- sector
  buf  swap dup csect !			\ -- addr sector
  cdrive c@ dup pdrive c!  SecRead
;


\ *****************************
\ *S FAT type specific routines
\ *****************************
\ *P Several routines are vectored according to the FAT implementation
\ ** on the drive. These routines use a table of xts to hold the action
\ ** required for FAT12, FAT16 and FAT32. The variable *\fo{FATtype}
\ ** holds the offset in bytes used to index the table.

\ *P Note that this code assumes a 32 bit Forth implementation.

variable FATtype	\ -- addr
\ *G Holds the offset (0/4/8) in the FAT function tables.

: FATexec	\ ?? table -- ??
\ *G From the function table, executes the function indexed
\ ** from *\fo{FATtype}.
  FATtype @ + @C execute
;

: *3/2		( n -- n' )	dup 2/ +  ;
\ *G Signed multiply by 3/2. For FAT12 operations.
: *2/3		( n -- n' )	$AAAAAAAA um* nip ;
\ *G Unsigned multiply by 2/3. For FAT12 operations.

[undefined] 4/ [if]
: 4/		( n -- n' )  2/ 2/  ;
[then]

\ FAT@ and FAT! read/write the FAT table. FAT12, FAT16 and FAT32 are supported.
\ |CL marks an end cluster.

create /fats	\ -- addr
\ *G Vector table for *\fo{/FAT} below.
  ' *2/3 ,  ' 2/ ,  ' 4/ ,

: /fat		\ #bytes -- #entries
\ *G Given a memory size in bytes, returns the number
\ ** of complete FAT entries in it.
  /fats FATexec
;

create FAToffsets	\ -- addr
\ *G Vector table for *\fo{FAToffset} below.
  ' *3/2 ,  ' 2* ,  ' cells ,

: FAToffset	\ cluster -- offset sector
\ *G Get sector number and offset for the given FAT entry.
  FAToffsets FATexec bytes>offSecs reservedsec @ +
  basesector @ +
\ cr ." Offset = " over .dword  ."  sector = " dup .dword
;

: FAT@32	\ cluster -- n
\ *G Return the 28 bit FAT32 entry for the given cluster.
  FAToffset readsector buf + uaLE@ $0FFFFFFF and
;

: FAT@16	\ cluster -- n
\ *G Return the 16 bit FAT16 entry for the given cluster.
  FAToffset readsector buf + uaLEw@
;

: fat@12bndry	\ offset sector -- w
\ *G Read a 16 bit FAT12 entry that crosses a sector boundary.
\ ** *\i{Offset} is *\fo{Bytes/Sector-1} and *\i{sector} is the
\ ** sector number of the first sector.
  tuck readsector  buf + c@		\ -- sector lo
  swap 1+ readsector buf c@		\ -- lo hi
  8 lshift or				\ -- w
;

: FAT@12	\ cluster -- n
\ *G Return the 12 bit FAT12 entry for the given cluster.
  dup  FAToffset over secsizemask = if	\ -- cluster offset sector
    tuck readsector  buf + c@		\ -- cluster sector lo
    swap 1+ readsector buf c@		\ -- cluster lo hi
    8 lshift or				\ -- cluster w
  else
    readsector  buf + uaLEw@		\ -- cluster w
  endif
  swap 1 and
  if  4 rshift  else  $0FFF and  then
;

create FAT@s	\ -- addr
\ *G Vector table for *\fo{FAT@}.
  ' FAT@12 ,  ' FAT@16 ,  ' FAT@32 ,

: FAT@		\ cluster -- n
\ *G Return the FAT entry for the given cluster.
  FAT@s FATexec
;

: FAT!32	\ n cluster --
\ *G Set the FAT32 entry for the given cluster.
  FAToffset readsector buf + uaLE!  wmark
;

: FAT!16	\ n cluster --
\ *G Set the FAT16 entry for the given cluster.
  FAToffset readsector buf + uaLEw!  wmark
;

: fat!12bndry	\ w offset sector --
\ *G Write a 16 bit FAT12 entry that crosses a sector boundary.
\ ** *\i{Offset} is *\fo{Bytes/Sector-1} and *\i{sector} is the
\ ** sector number of the first sector.
  rot >r  				\ -- offset sector ; R: -- w
  tuck readsector			\ -- sector offset ; R: -- w ; first sector
  r@ swap buf + c!  wmark		\ -- sector ; R: -- w ; write low byte
  1+ readsector				\ -- ; R: -- w
  r> 8 rshift buf c!  wmark
;

: wmerge12	\ n cluster nr -- nw
\ *G Merge two FAT12 entries before writing. *\i{N} is the new
\ ** entry, *\i{nr} is the entry from the disc, and *\i{cluster}
\ ** is the entry number to which the value *\i{nw} will be written.
  swap 1 and if				\ -- n nr
    $0F and swap 4 lshift +		\ nnnnnnnnnnnnxxxx
  else
    $0F000 and +			\ xxxxnnnnnnnnnnnn
  then
;

: FAT!12	\ n cluster --
\ *G Set the FAT12 entry for the given cluster.
  dup  FAToffset over secsizemask = if	\ -- n cluster offset sector
    2dup 2>r fat@12bndry		\ -- n cluster nr ; R: -- offset sector
    wmerge12 2r> fat!12bndry
  else
    readsector  buf + dup >r uaLEw@	\ -- n cluster nr ; R: -- addr
    wmerge12 r> uaLEw!  wmark
  endif
;

create FAT!s	\ -- addr
\ *G Vector table for *\fo{FAT@}.
  ' FAT!12 ,  ' FAT!16 ,  ' FAT!32 ,

: FAT!		\ n cluster --
\ *G Set the FAT entry for the given cluster.
  FAT!s FATexec
;

create |cls	\ -- addr
\ *G Table holding last-cluster markers for each FAT type.
  $0FF8 ,  $0FFF8 ,  $0FFFFFF8 ,

: |cl		\ -- n
\ *G Last-cluster marker.
  |cls FATtype @ + @C
;


\ *********************************
\ *S FAT File system initialisation
\ *********************************

variable bptr	\ -- addr
\ *G Stream pointer into the current sector cache buffer.
variable bperror	\ -- addr
\ *G If set non-zero, an error exists in the BPB.

: newb		\ --
\ *G Reset the input stream.
  buf bptr !
;

: skipb		\ --
\ *G Step to the next stream byte.
  1 bptr +!
;

: getb		\ -- byte
\ *G Get stream byte and step to next.
  bptr @ c@ skipb
;

: getw		\ -- word
\ *G Get 16 bit stream item and step past it. The item is read in
\ ** little-endian format.
  getb getb byte-join
;

: getl		\ -- long
\ *G Get 32 bit stream item and step past it. The item is read in
\ ** little-endian format.
  getw getw word-join
;

: wantb		\ byte --
\ *G Read the next stream byte and if it is not the given byte,
\ ** set *\fo{bperror} to 1.
  getb <>
  if  1 bperror !  then
;

: wantw		\ word --
\ *G Read the next stream word and if it is not the given word,
\ ** set *\fo{bperror} to 1.
  byte-split swap wantb wantb
;

: lengthen	\ w -- u
\ *G Read the next 32 bit item. If w is non-zero, discard the new
\ ** item and return w, otherwise return the new item.
  getl  over
  if  drop  else  nip  then
;

: bp_init	\ --
\ *G Initialise the system, read the drive characteristics and
\ ** set up for file operations.
   initSecCache
   basesector @ readsector  0 bperror !	\ read BPB sector
   newb $0B bptr +!			\ skip to bytes/sector
( 11) bytes/sec wantw			\ Expect 512 bytes/sector.
( 13) getb 1 max dup sec/cluster !	\ sectors per cluster
      qlog2 clusshift !			\ equivalent shift
( 14) getw reservedsec !
( 16) 2 wantb				\ expect 2 FAT tables, even with flash drives
( 17) getw dup rootentries !		\ number of entries in root directory - 32 bytes/entry
      $000F + 4 rshift RootDsecs !	\ number of sectors in root directory - 16 entries/sector
( 19) getw skipb			\ total sectors
( 22) getw >r				\ 16-bit FAT size
( 24) getw SectorsPerTrack !		\ physical disk attributes
( 26) getw NumHeads !
      getl drop
( 32) lengthen totalsec !
\ The rest of the BPB depends on the FAT type
( 36) r> lengthen dup FATsize !		\ 32-bit FAT size
   2* reservedsec @ + basesector @ + rootsector !
   CountofClusters
   dup $0FF5 < if			\ FAT12 test
     drop  0				\ FAT12 offset = 0
   else
     $0FFF5 <				\ FAT16/32 test
     if  1 cells  else  2 cells  then 	\ FAT16=4, FAT32=8
   then
   FATtype !
;


\ **************
\ *S Directories
\ **************
\ *P Each sector of a directory contains 16 entries indexed by
\ ** *\fo{DIRENTRY}. The sector number *\fo{DIRSECTOR} is a 32-bit
\ ** value addressing up to 2Tbytes. The root directory is a
\ ** contiguous run of *\fo{ROOTDSECS} sectors starting at the sector
\ ** in *\fo{ROOTSECTOR}. Subdirectories use a chain in the FAT table
\ ** just like files. So, the root directory has a fixed number of
\ ** possible directory entries. *\fo{DIRCLUSTER} is 0 when the
\ ** root directory is selected.

\ *P For most drives, cluster 0 and cluster 1 are occupied by
\ ** two copies of the FAT. Cluster 2 is *\b{usually} the start
\ ** of the root directory. For FAT12 and FAT16, the root
\ ** directory is of fixed size.

decimal

: dt		\ x acc n bits -- acc' x
\ *G Shift accumulator left by *\i{bits}, add in *\i{n}, and
\ ** bring the next item *\i{x} to the top. Ugh!
  >r swap r> lshift + swap
;

: datetime	\ -- u
\ ** Convert the current date and time into the packed 32 bit
\ ** format for directory entries:
\ *C   yyyyyyymmmmddddd:hhhhhmmmmmmsssss
  0 time&date #1980 - swap 4 dt 5 dt 5 dt 6 dt 2/ 5 dt drop
;

hex

variable dircluster	\ -- addr
\ *G Holds first cluster of the current directory, or 0 for the
\ ** root directory.
variable dirsector	\ -- addr
\ *G Holds current sector number of directory.
variable direntry	\ -- addr
\ *G Holds current entry number.

: dir[]		\ offset -- addr
\ *G Index into the current directory entry.
\ ** Ensures that the entry is in the sector cache.
  direntry @ 5 lshift +  dirsector @ readsector  buf +
;

: secmask	\ -- sector mask
\ *G Return the sector and the mask for testing the current
\ ** directory sector.
  dirsector @ 0 cluster>sec - sec/cluster @ 1-
;

: firstsec?	\ -- flag
\ *G Return true if we are at the first sector of a cluster.
  secmask and 0=
;

: lastsec?	\ -- flag
\ *G Return true if we are at the last sector of a cluster.
  secmask tuck and =
;

: attrib?	\ mask -- flag
\ *G Test file attribute bits, returning non-zero if any bits in
\ ** the mask are set. The attributes are a bit mask
\ *(
\ *B bit 0, $01 - read only
\ *B bit 1, $02 - hidden
\ *B bit 2, $04 - system
\ *B bit 3, $08 - volume ID
\ *B bit 4, $10 - directory
\ *B bit 5, $20 - archive
\ *B $0F - long file name
\ *)
  $0B dir[] c@ and
;

: dir?     	\ -- b
\ *G Return the first byte of the directory entry.
  0 dir[] c@
;

: dirsmudged?	\ -- flag
\ *G Return true if the current directory entry is "smudged".
  0 dir[] c@  dup $0E5 = swap $005 = or
;
\ According to Fatgen103.pdf, 5 is a valid Kanji character.
\ I don't believe we need the second test.

: dirsmudge	\ --
\ *G Mark the current directory entry as "smudged".
  $0E5 0 dir[] c!  wmark
;

: get-es	\ -- entry sector
\ *G Return the current directory entry and sector.
  direntry @ dirsector @
;

: set-es	\ entry sector --
\ *G Set the current directory entry and sector.
  dirsector !  direntry !
;

: dircluster@	\ -- clus
\ *G Return the current directory cluster.
  dirsector @ sec>cluster
;

: dircluster!	\ clus --
\ *G Sect the current directory cluster, and hence start sector.
  cluster>sec dirsector !
;

: DirValid?	\ -- flag
\ *G Are we at a valid directory entry for display?
  dirsmudged?  $0A attrib? or		\ smudged or hidden or a volume name
  0 dir[] uaLEw@ $0202E = or  		\ or '.' folder name
  0=					\ none of the above
;

: sbackward	\ --
\ *G Go to the previous directory cluster, use 1- for root.
  dircluster@  dircluster @ if
    >r  dircluster @			\ -- start ; R: -- end
    begin				\ find the previous cluster
      dup FAT@ r@ over <>
     while
      nip
    repeat
    r> 2drop
    cluster>sec sec/cluster @ 1- + dirsector !    \ last sector in this cluster
  else
    1- dircluster!
  then
;

: sforward	\ --
\ *G Go to the next directory cluster, use FAT for subfolders,
\ ** 1+ for root
  dircluster@  dircluster @
  if  FAT@  else  1+  then
  dircluster!
;

: sector0?	\ -- flag
\ *G Are we at the first sector of a directory?
  dircluster @ dircluster@ =  firstsec? and	\ beginning of first subfolder
  dirsector @ rootsector @ = or		\ or beginning of root directory
;

: dsecup	\ --
\ *G Step to the next directory sector.
  lastsec? if
    sforward  0 direntry !		\ SFP016
  else
    1 dirsector +!
  then
;

: dsecdn	\ --
\ *G Step to the previous directory sector.
  firstsec?				\ first sector in cluster?
  if  sbackward  else  -1 dirsector +!  then
;

: entbump	\ offset -- n
\ *G Go to next/prev entry, where offset = +/-1
  direntry @ + $0F and dup direntry !
;

: dir++		\ --
\ *G Go to the next directory sector/entry.
  1 entbump 0= if  dsecup  then
;

: dir--		\ --
\ *G Go to the previous directory sector/entry.
  -1 entbump $0F = if  dsecdn  then
;

: noend?	\ clus -- notlast?
\ *G Return true if the cluster number is not an end-chain-marker.
  |cl tuck and  xor			\ XOR used as <>
\ *** Since there can be other uses for high sector numbers, should
\ *** we use  "|cl u<"  instead?
;

: OKprev	\ -- flag
\ *G True if the directory pointer is not at the very beginning
\ ** of the directory.
  direntry c@ 0= if
    dircluster @ dircluster@ =  firstsec? and	\ beginning of first subfolder
    dirsector @ rootsector @ = or  0=		\ or beginning of root directory
  else  1  then
;

: OKnext	\ -- flag ; okay to go forward?
\ *G True if the directory pointer is not at the very end of
\ ** the directory.
  direntry c@ $0F = if
    dircluster @ if
      lastsec?
      if  dircluster@ FAT@ noend?  else  1  then
    else
      rootsector @ rootdsecs @ + 1-
      dirsector @ <>			\ last root sector
    then
  else 1 then
;

: dirnext	\ -- flag
\ *G Jump to the next valid directory entry. Return non-zero
\ ** if the next entry is valid.
  begin
    OKnext  dir? 0<> and			\ next entry exists?
   while
    dir++  DirValid?				\ viewable entry: okay
    if  1 exit  then
  repeat
  0
;

: todir		\ --
\ *G Skip to first valid entry.
  DirValid? 0= if  dirnext drop  then
;

: dirprev	\ -- flag
\ *G Jump to previous valid directory entry.
  begin
    OKprev
   while				\ previous entry exists?
    dir-- DirValid?			\ viewable entry: okay
    if  1 exit  then
  repeat
  0 todir
;

decimal

#13 buffer: npad	\ -- addr
\ *G Temporary buffer for building 8.3 filenames.

: padc		\ char --
\ *G Add one character to *\fo{npad}.
  npad dup >r count + c!  r@ c@ 1+ r> c!
;

: pad$		\ caddr len --
\ *G Add string to *\fo{npad}.
  FOR count padc NEXT drop
;

: dfname	\ -- caddr len
\ *G Return the main part of the file name with trailing spaces
\ ** removed.
  0 dir[] 8 -trailing
;

: dfext		\ -- caddr len
\ *G Return the extension part of the file name with trailing spaces
\ ** removed.
  8 dir[] 3 -trailing
;

: dfullname	\ -- caddr len
\ *G Return filename with extension.
  0 npad c!  dfname pad$  dfext dup
  if  [char] . padc  pad$  else  2drop  then
  npad count
;

: parent?	\ -- flag
\ *G Return true if the current directory entry is a parent
\ ** directory, i.e. "..".
  dfname 2 = swap uaLEw@ #11822 = and	\ '..' = $2E2E = #11822
;

\ Offset 20 = high, 26 = low.
: cluster0	\ -- n
\ *G Get the first cluster number of the data for this entry.
  #26 dir[] uaLEw@  #20 dir[] uaLEw@ word-join
;

: setClus0	\ u dirent --
\ *G Set the cluster number into the given directory entry.
  >r  word-split r@ #20 + uaLEw!  r> #26 + uaLEw!
;

: cluster0!	\ n --
\ *G Set the first cluster number of the data for this entry.
  0 dir[] setClus0  wmark
;

: folder?	\ -- flag
\ *G Is the current entry a directory/folder?
  $10 attrib?
;

: dsize		\ -- n
\ *G The file size in bytes of the file for the current
\ ** directory entry.
  #28 dir[] uaLE@
;

: dirsect0	\ -- sect
\ *G Return the starting sector of the current directory.
  dircluster @ dup
  if  cluster>sec  else  drop  rootsector @  then
;

: dirhome	\ --
\ *G Select the first entry in the current directory.
  0 dirsect0 set-es
;


\ *************
\ *S FAT chains
\ *************

: freechain	\ clus0 --
\ *G Writes zeros to a chain in the FAT table.
  begin  dup noend? while  dup FAT@ 0 rot FAT!  repeat
  drop
;

: _freesecs	\ --
\ *G Frees the chain of the current file.
  cluster0 ?dup if  freechain  then
;


: _delete	\ --
\ *G Deletes the current file.
  dirsmudge _freesecs
;

: getcluster	\ clus -- clus'
\ *G Gets the next free cluster number after this one and marks
\ ** it as the last.
  begin  dup FAT@ while  1+  repeat
\ Should we check the incremented cluster# against |CL
\ and start at the beginning of the FAT table? THROW
\ if disk is full?
  |cl over FAT!
;

: chain+	\ cl -- cl'
\ *G Appends a cluster to the end of the chain.
  dup getcluster 2dup =
  if  nip  else  tuck swap FAT!  then
;

: shrink	\ clus0 --
\ *G Shrinks a FAT chain and marks the last cluster.
  dup >r freechain |cl r> FAT!
;

: expand	\ clus0 n --
\ *G Expands a FAT chain and marks the last cluster.
  FOR chain+ NEXT drop
;


\ **************************
\ *S File/Path name handling
\ **************************
\ *P File names are checked against allowable characters. Dot (.)
\ ** is allowable once to separate filename and ext fields. The
\ ** widths of these fields may not exceed 8 and 3 characters
\ ** respectively.

hex
create goodchars	\ -- addr
\ *G An array of packed bit flags: 1=allowed: 0..7, 8..9, ...127
   000 c, 000 c, 000 c, 000 c, 05F c, 0C4 c, 0FF c, 0C0 c,
   0FF c, 0FF c, 0FF c, 0E3 c, 0FF c, 0FF c, 0FF c, 0F6 c,
decimal

: goodch	\ char -- flag
\ *G Check if the character is allowable in a file/path name.
  dup 128 and
  if exit then                                             \ 80..FF
  dup 7 and >r
  3 rshift goodchars + c@c
  r> lshift 128 and
;

: goodstr	\ caddr len limit -- flag
\ *G Check that string has only allowable chars and is not too long.
  over < 0=  swap
  FOR  over c@ goodch and  N++  NEXT
  nip
;

: name.ext	\ a u -- a1 u1 a2 u2
\ *G Separate filename string into name and extension strings.
  2dup [char] . scan  dup
  if  rot over - -rot  1 /string  then	\ SFP023
;
\ a alen b blen -- a alen-blen b blen
\   rot over - -rot

: goodname	\ caddr len -- flag
\ *G Are filename and extension (could be in a path) valid?
  2dup s" .." compare 0=		\ SFP023
  if  2drop true exit  endif
  name.ext  3 goodstr >r  8 goodstr r>  and
\  FATOK c@ 0<> and			\ say "no" if filesystem isn't operational
;

\ *P A filename that contains one or more slashes (either '/' or' \')
\ ** is assumed to be a file path ending in a filename. When one
\ ** of these is encountered, the path string is stripped off and
\ ** used it to set the file path. The current directory is not
\ ** changed. If all of your files are in the same directory, you
\ ** can use *\fo{CWD} and friends to set the current directory.
\ ** Begin with a slash to start at the root directory.

: /scan		\ a u -- a' u'
\ A version of *\fo{SCAN} that allows either '\' or '/' in paths.
  2dup  [char] / scan
  2swap [char] \ scan
  2 pick over <
  if  2swap  then
  2drop
;

: /parse	\ addr len -- addr' len' a1 u1
\ *G Pick the file name off left end of path. The string *\i{a1/u1}
\ ** contains the next directory element
  2dup /scan  dup >r dup
  if  1 /string  then
  2swap r> -
;


\ ****************************
\ *S Current Working Directory
\ ****************************
\ *P The current file path is tracked by a path string. The
\ ** file tree is navigated by selecting the cluster for the
\ ** subdirectory and adding its name to the path. A ".." folder
\ ** name trims the path name to go back a level. This path string
\ ** is for display use.

\ *P Because this is not a fully thread-safe file system and only
\ ** has a single "working directory" for all tasks/threads, you
\ ** will need to be careful with external access systems such as
\ ** HTTP, FTP and USB.

pathmax buffer: pathbuf		\ -- addr
\ *G Path name buffer.

: nopath	\ --
\ *G Empty the path name buffer.
  0 pathbuf !
;

: path		\ -- caddr len
\ *G Return the current path string.
  pathbuf count
;

: bpath		\ b --
\ *G Bump the pathname string length.
  pathbuf c@ + pathmax min pathbuf c!
;

: pc+		\ char --
\ *G Add *\i{char} to the path string.
  path + c!  1 bpath
;

: /?		\ -- flag
\ *G Return true if the last+1 character is '/'.
  path tuck + c@ [char] / = and
;

: trimp		\ --
\ *G Trim path to just before the last '/' character.
  begin
    path nip
   while
    -1 bpath  /?
    if  exit  then
  repeat
;

: addp		\ caddr len --
\ *G Add string to path, separated by a '/' character.
  path + 1- c@ [char] / <>		\ even ok if len=0, 0<>'/'
  if  [char] / pc+  endif
  FOR count pc+ NEXT drop
;


\ ***********************
\ *S Directory management
\ ***********************

: dirstart	\ cluster --
\ *G Starts referencing a new directory at the given cluster.
  dup dircluster !  0=
  if  nopath  then
  0 dirsect0 set-es  0 dcurs !
  todir
;

: rootdir	\ --
\ *G Start at the root directory.
  nodir  0 dirstart  s" /" pathbuf place
;

: dirnest	\ --
\ *G Start at the folder pointed to by the current directory entry.
  cluster0 dirstart
;

: sel-folder	\ --
\ *G Select the current directory entry as a folder.
  parent?
  if  trimp  else  dfname addp  then
  nodir dirnest
;

(( \ del SFP023
: cd..		\ --
  \ *G Step to the parent folder, or the root directory if no
  \ ** parent exists.
  dirhome				\ go back to parent folder
  begin					\ loop to find it
    parent?
    if  sel-folder exit  then
    dirnext 0=
  until
  rootdir				\ not found, go back to root
;
))

((
: sel-file	\ --
  cr ." File " d-name ."  selected."
;

: dirsel	\ --
 \ *G Opens a directory entry.
  zcurs DirValid? if
    folder?
    if  sel-folder  else  sel-file  then
  then
;
))

: _test-file	\ caddr len attr -- found?
\ *G Search in the current directory for an existing filename
\ ** matching none of the given* attribute bits. Return non-zero
\ ** if found. If found, the directory sector and entry are saved
\ ** in *\fo{savedDirEnt} and the directory pointers address the
\ ** found filename entry. The attribute bits settings are:
\ *(
\ *B bit 0, $01 - read only
\ *B bit 1, $02 - hidden
\ *B bit 2, $04 - system
\ *B bit 3, $08 - volume ID
\ *B bit 4, $10 - directory
\ *B bit 5, $20 - archive
\ *B $0F - long file name
\ *)
  >r  2dup goodname if
    dirhome  begin
      r@ attrib? 0= if			\ ignore hidden/vol/folder...
        2dup dfullname ucompare 0= if
          get-es savedDirEnt 2!
          r> 2drop  exit		\ found  *** assume addr non-zero.
        then
      then
      dirnext 0=
    until
  then
  r> drop  2drop 0
;

: test-file	\ caddr len -- found?
\ *G Search for a file that is neither a directory nor a volume ID.
\ ** Upon exit when found, the directory pointers address the
\ ** found file.
  $18 _test-file			\ not folder or volume
;

: cd+		\ caddr len -- ior
\ *G Open new directory. *\fo{CD+} changes the directory up or down
\ ** a level.
  $08 _test-file if			\ not a volume ID, but hidden is ok.
    folder? if
      sel-folder 0			\ SFP021
    else
      err_BadDir			\ 1 = not a directory
    then
  else
    err_DirNotFound			\ 2 = directory not found
  then
;

\ renamed MPE011 SFP022
: $cwd		\ addr len -- ior
\ *G Open a subdirectory given a file path. *\i{Ior}=0 if okay.
  /parse dup				\ begin directory parse
  if  cd+  else  2drop 0  rootdir  then	\ SFP010
  begin
\ pathbuf 16 dump
    over
   while
    >r /parse dup
    if  cd+  else  2drop -1  then
    r> or
  repeat
\ pathbuf 16 dump
  nip nip
;

: setdir	\ addr len -- addr' len' ior
\ *G Select a subdirectory given a file path, stripping off the filename.
  2dup
  begin  /parse 2 pick while  2drop  repeat
  2>r 2drop r@ -  dup
  if  $cwd  else  2drop 0  then
  2r> rot
;

: wipeclus	\ clus --
\ *G Fill a cluster with zeros, *\fo{THROW} on error.
  wflush buf bytes/sec erase -1 csect !	\ save current buffer and invalidate it
  cluster>sec  sec/cluster @ FOR	\ -- sec ; R: -- x ; start at the first sector
    buf over cdrive c@ SecWrite
    1+
  NEXT
  drop
;

: getdir	\ -- ior
\ *G Finds the first directory entry that may be overwritten.
  dirhome  begin
    dir?
   while
    dirsmudged? 0=
   while
    OKnext
   while
    dir++
  repeat then then
  dirsmudged? dir? 0= or
  if  0 exit  then			\ normal termination
  dircluster @ if
    dircluster@ chain+ dup dircluster!	\ next cluster ; SFP015
    0 direntry !			\ start at first entry ; SFP016
    wipeclus 0				\ erase the new directory cluster
  else
    err_RootFull			\ can't expand the root directory ; SFP015
  then
;


\ ***************************
\ *S Handles and File support
\ ***************************
\ *P File support is ANS standard, after the stream model used
\ ** by mainstream computing. File status is held in array of
\ ** records:
\ *E 8-bit filemode           3=R/W, 1=R/O, 2=W/O
\ **                          bit 2 = 1 = updated
\ **                          bit 1 = 1 = writable
\ **                          bit 0 = 1 = readable
\ ** 8-bit device             reserved for file device/drive
\ ** 8-bit directory entry    directory index (0..15) for this file
\ ** 8-bit                    reserved
\ ** 32-bit directory sector  directory index for this file
\ ** 32-bit position          file position for up to 4GB files
\ ** 32-bit start cluster     FAT location
\ ** 32-bit current sector    active sector
\ ** 32-bit file size         length in bytes

\ *P The following data types are used:
\ *D fam        "File Access Method", describes read/write permission
\ **            etc.
\ *D ior        "IO Result", A return result from most IO calls, this
\ **            value is 0 for success or non-zero as an error-code.
\ *D fileid     "File Identifier", a handle for a file.

FATopen? 0= [if]  external  [then]
1 constant R/O	\ -- fam
\ *G Get ReadOnly fam. This *\b{must} be bit 0 to match DOS.
2 constant W/O	\ -- fam
\ *G Get WriteOnly fam.
3 constant R/W	\ -- fam
\ *G Get ReadWrite fam.
: bin		\ fam -- fam'
\ *G Modify a file-access method to include BINARY.
; immediate
FATopen? 0= [if]  internal  [then]


\ *P The corresponding FAT attribute bits at byte 11 in each directory
\ ** entry are:
\ *E #define ATTR_READ_ONLY 0x01
\ ** #define ATTR_HIDDEN    0x02
\ ** #define ATTR_SYSTEM    0x04
\ ** #define ATTR_VOLUME_ID 0x08
\ ** #define ATTR_DIRECTORY 0x10
\ ** #define ATTR_ARCHIVE   0x20
\ **   top two bits are always zero
\ ** #define ATTR_LONG_NAME 0x0F   \ continuation entry of long file name

6 cells constant bytes/handle	\ -- len
\ *G Number of bytes needed for each handle. Must be an integer
\ ** number of cells for portability.

#files bytes/handle * buffer: filehandles	\ -- addr
\ *G Array of file handle records.

cell +user handle	\ -- addr
\ *G Holds file handle during some operations.

((
: h>rec		\ n -- addr
 \ *G Convert a handle to a record address.
  bytes/handle * filehandles +
;
))

: +handle[]	\ n -- addr
\ *G Given a handle, return the address in the handle record array.
\ ** Why not just use the address as the handle?
  handle @ bytes/handle * + filehandles +
;

\ *[
\ H.FAM must be first.
: h.fam	  ( -- addr )  0 +handle[]  ;			\ r/o, w/o, r/w, etc.
: h.dev   ( -- addr )  1 +handle[] ;			\ device of this file, 0..15
: h.dire  ( -- addr )  2 +handle[]  ;			\ directory entry 0..15
: h.dirs  ( -- addr )  cell +handle[]  ;		\ directory sector
: h.pos   ( -- addr )  [ 2 cells ] literal +handle[] ;	\ file position
: h.clus0 ( -- addr )  [ 3 cells ] literal +handle[] ;	\ starting cluster
: h.sect  ( -- addr )  [ 4 cells ] literal +handle[] ;	\ current sector
: h.len   ( -- addr )  [ 5 cells ] literal +handle[] ;	\ file length
\ *]

: closeall	\ --
\ *G Wipes out the file structure to ensure files are closed.
  filehandles #files bytes/handle * erase
;

: h.clus!	\ cluster --
\ *G Set the current file sector.
  cluster>sec h.sect !
;

: h:dir		\ --
\ *G Points the directory pointers at the file addressed by
\ ** current handle.
  h.dire c@ h.dirs @ set-es ;

: after/before	\ -- after before
\ *G Get the number of bytes before and after the current sector
\ ** position.
  h.pos @ secsizemask and		\ bytes in sector before posn
  bytes/sec over -			\ bytes after posn
  swap
;

: _resize	\ len --
\ *G Given a file size in bytes, allocate clusters in FAT for the
\ ** file.
  dup b>c  h.len @ b>c -  swap h.len !	\ clusters to allocate ; BJC026
  dup 0=
  if  drop exit  then			\ no change in chain length
  h.clus0 @ swap dup 0<
  if  drop shrink  else  expand  then
;

: nexthan	\ -- fileid ior
\ *G Find available file handle, and make it the current handle.
  #files FOR
    r@ handle !  h.fam c@ 0=
    if  r> 0 exit  then
  NEXT
  -1 err_NoHandles
;

: (repo-file)	\ u -- ior
\ *G Repositions the current file. Must not be past the end.
\ ** Unprotected.
  h.len @ over u<			\ can't be past the end. *** U<
  if  drop err_BadPos exit  then	\ *** change to return fixed ior
  dup h.pos !  b>cn >r			\ #clus -- ; R: -- noff
  h.clus0 @ swap			\ step n clusters
  FOR  FAT@  NEXT
  cluster>sec r> + h.sect !		\ reset current sector
  0
;

: repo-file	\ u -- ior
\ *G Repositions the current file. Must not be past the end.
\ ** protected.
  ['] (repo-file) catch ?dup
  if  nip  endif
;

: [dir		\ -- ; R: -- i*x
\ *G Saves current directory information on the return stack
\ ** for later restoration by *\fo{DIR]}.
\ ** This word depends on the return address being a single cell
\ ** on the Forth return stack.
  r>
  get-es 2>r  dircluster @ >r  dcurs @ >r
  >r
;

: dir]		\ -- ; R: i*x --
\ *G Restores directory information previously saved on the
\ ** return stack by *\fo{[DIR}.
\ ** This word depends on the return address being a single cell
\ ** on the Forth return stack.
  r>
  r> dcurs !  r> dircluster !  2r> set-es
  >r
  begin
    OKprev DirValid? 0= and
   while
    dirprev drop			\ BJC019
  repeat
;


\ ************
\ *S File Read
\ ************

: samesect?	\ -- flag
\ *G Checks if the next read or write is in a new sector.
  after/before nip
;

: sameclus?	\ -- cluster same?
\ *G Checks if the next sector is in the same cluster.
  h.sect @ dup sec>cluster
  swap 1+ sec>cluster over =
;

: bumppos	\ n -- overflow
\ *G Bump file position, check for wrap to next sector.
   h.pos @ dup >r + dup h.pos ! r> xor bytes/sec and
;

: bumpsec	\ len --
\ *G Bump file position and maybe sector/cluster.
  bumppos if
    sameclus?
    if  drop  1 h.sect +!  else  FAT@ h.clus!  then
  then
;

: read-1sec	\ addr len -- addr' len' n
\ *G Read from current sector.
  2>r  after/before			\ -- after before ; R: -- addr len
  over r@ min bytes/sec = if		\ read a whole sector
    drop
    2r@ drop h.sect @ h.dev c@ SecRead
  else
    h.sect @ readsector			\ reading less than a whole sector
    buf + swap 2r@ rot min dup >r cmove r>
  then					\ -- len ; R: -- DA DL
  2r> rot dup >r /string  r> dup bumpsec
;

: _read-file	\ baddr blen -- len ior
\ *G File read primitive.
  h.fam c@ 1 and if			\ -- baddr blen ; read permission?
    [dir
    h.len @ h.pos @ - min 0		\ -- baddr len 0 ; clip to file size
    begin
      over
     while
      >r read-1sec r> +
    repeat
    nip nip 0
    dir]
  else
    2drop  0 err_r/o			\ reading W/O file
  then
;


\ *************
\ *S File write
\ *************

: timestamp	\ --
\ *G Adds the time and date to the current directory entry.
  datetime dup #22 dir[] uaLE!
  #16 rshift  #18 dir[] uaLEw!
  wmark
;

: copytime	\ --
\ *G Copy the creation date.
  #22 dir[] #14 dir[] 1 cells move  wmark
;

: archive+	\ --
\ *G Sets the current directory entry's archive bit.
  11 dir[] dup c@ $20 or swap c!  wmark
;

: filestamp	\ --
\ *G Stamps the directory entry addressed by the current handle.
  h:dir timestamp archive+
;

: stamp		\ --
\ *G Mark the file as stamp-pending.
  h.fam dup c@ 4 or swap c!
;

: newchain	\ --
\ *G Begin a new chain.
  2 getcluster dup h.clus0 ! h.clus!  stamp
;

: nextwsec	\ --
\ *G Find next writable sector in file.
  sameclus? if
    drop  1 h.sect +!			\ same cluster, just bump sector
  else
    dup FAT@ noend? 0=			\ if at last sector, allocate a new one
    if  dup 1 expand  then
    FAT@ h.clus!			\ get next writable cluster in FAT
  then
;

: write-1sec	\ a u -- a' u'
\ *G Write 512 bytes or less to file.
  h.clus0 @ 0=
  if  newchain  then			\ empty: assign starting cluster
  over >r after/before			\ -- a u after before ; R: -- src
  h.sect @ readsector			\ preread is almost always needed
  buf + >r over min 2r> 2 pick move	\ -- a u len'
  wmark  dup >r /string r> bumppos
  if  nextwsec  then
  h.pos @ h.len @ max h.len !		\ stretch filesize if needed
;

FATtext? [if]
create crlf	\ -- addr
\ *G Holds a CR/LF pair for *\fo{WRITE-LINE} below.
  #13 c,  #10 c,
[then]

((
: newdir     ( Aname Nname Aext Next clus0 attributes -- )  \ make new dir entry
   2>r 0 dir[] 11 32 fill                               \ blank filename and ext
   8 dir[] swap umove  0 dir[] swap umove              \ lay in upcased filename
   2r> 11 dir[] tuck c!  1+ 20 erase                             \ and attribute
   word-split 20 dir[] sw! 26 dir[] sw!               \ and 32-bit start cluster
   timestamp copytime ;
))

: name>dir	\ caddr len dirent --
\ *G Copy the given file name into the given directory entry.
  dup #11 bl fill			\ wipe name
  >r  name.ext r@ 8 + swap 3 min ucmove	\ extension
  r> swap 8 min ucmove			\ name
;

: prepdir	\ caddr len clus0 attribs --
\ *G Set up a new directory entry using the given name, cluster
\ ** and attributes. Note that *\i{attribs} is in DOS form, not
\ ** in the form used by *\i{h.fam}.
  0 dir[] dup #12 + #20 erase		\ -- caddr len clus0 attribs dirent
  tuck #11 + c!  tuck setClus0  name>dir
  timestamp copytime			\ includes WMARK
;

: _newfold	\ addr len cluster --
\ *G add a new folder entry to the current folder, and step to the
\ ** next directory entry.
  s"  " rot $10 prepdir  0 dir[] swap cmove
  dir++
;

: es^		\ --
\ *G Set the current handle's entry and sector fields.
  get-es h.dirs ! h.dire c!
;

: ?open_err	\ a b c flag err# -- a b c | -- -1 err# [exits caller]
\ *G If flag is false/0, just the first three parameters are
\ ** returned. If flag is true, the first three parameters
\ ** are discarded and -1 and the error number are returned
\ ** and exit is from the caller. Provided to reduce code space
\ ** and dependent on the return address being on the Forth
\ ** return address. The caller *\b{must not} use local
\ ** variables.
  swap if
    nip nip nip  -1 swap  r> drop
  else
    drop
  then
;

: _open-file	\ caddr len fam -- fileid ior
\ *G File open primitive. Performs no directory restoration.
\ ** This word will *\b{not} open a directory.
  -rot setdir err_BadPathName ?open_err	\ path check
  test-file 0= if			\ -- fam ; existence check ; SFP010 MPE025
    drop  -1 err_FileNotFound exit	\ SFP010
  endif					\ SFP010
  nexthan dup				\ -- fam fileid ior ior
  if  rot drop  exit then
  rot h.fam c!  dsize h.len !
  cluster0 h.clus0 !
  es^  0 repo-file if			\ close file on position error
    2drop  -1 err_BadPos  0 h.fam c!
  then
;

: _create-file	\ caddr len fam -- fileid ior
\ *G Create a file on disk, returning a 0 ior for success and
\ ** a file id. If the file already exists, it is truncated to
\ ** zero length. Performs no directory restoration.
  -rot setdir err_BadPathName ?open_err	\ -- fam caddr' len' ior
  2dup goodname 0= err_BadFileName ?open_err	\ -- fam caddr' len' ior
  2dup test-file if			\ -- fam caddr' len' ; MPE025
    _freesecs				\ erase if existing
  else
    getdir err_DirFull ?open_err	\ get a new directory entry
  then
  2 pick r/o = 1 and  0 swap prepdir	\ -- fam ; set up directory entry ; SFP020
\  2 pick r/o and  0 swap prepdir	\ -- fam ; set up directory entry
  nexthan dup				\ -- fam fileid ior ior
  if  rot drop  exit  then
  rot h.fam dup bytes/handle erase c!	\ wipe handle, set fam
  es^  flushall
;

: _mkdir      \ c-addr u -- ior
\ *G *\i{C-addr/u} is an 8.3 directory name with no directory
\ ** separators. *\fo{MKDIR} creates a new directory of that name
\ ** in the current directory. Directory removal is not properly
\ ** supported without the code in *\i{DelDirs}.fth.
\ ** You can use *\fo{DELETE-FILE} to remove an empty directory;
\ ** but if the directory is not emptied the disk will be
\ ** corrupted.
\ ** Performs no directory restoration.
\  setdir				\ walk pathname
\  if  2drop err_BadPathName exit  then
  2dup goodname 0=			\ invalid directory name?
  if  2drop err_NotDir exit  then
  getdir				\ full directory?
  if  2drop err_DirFull exit  then
  nexthan nip				\ handle not available?
  if  2drop  err_NoHandles exit then
  newchain				\ make new chain
  h.clus0 @ dup >r $10 prepdir		\ initialise directory entry
  r> dup wipeclus 			\ -- clus ; make empty dir cluster
  cluster>sec 0 swap set-es		\ make new dir current
  s" ."  h.clus0 @ _newfold
  s" .." dircluster @ _newfold
  flushall  0 h.fam c!			\ flush disk, close dir
  0
;

: _mkDirEx	\ c-addr u -- ior
\ *G As *\fo{-mkdir}, but supports full pathnames.
  setdir
  if  2drop err_BadPathName  else  _mkdir  then
;

: (read-file)	\ caddr len fileid -- #read ior
\ *G Read data from a file. The number of characters actually
\ ** read is returned as *\i{#read}, and *\i{ior} is returned 0
\ ** for a successful read. Unprotected.
  handle ! _read-file
;

: (write-file)	\ caddr len fileid -- ior
\ *G Write a block of memory to a file. Unprotected.
  handle !  h.fam c@ 2 and if		\ writable?
    stamp
    begin				\ -- caddr len
      dup
     while
      write-1sec
    repeat
    nip
  else
    2drop err_r/o			\ writing R/O file
  then
;

: (close-file)	\ fileid -- ior
\ *G Close an open file. Unprotected.
  handle !  h:dir  h.fam c@ 4 and if	\ need to time/date file cos written
    filestamp  h.clus0 @ cluster0!
    h.len @ #28 dir[] uaLE!
  then
  wmark flushall  0 dup h.fam c!	\ flush buffers and free this file handle
;

FATtext? [if]
: (read-line)	\ caddr u1 fileid -- u2 flag ior
\ *G Read an line of text from a file into a buffer, without EOL.
\ ** Unprotected.
  handle !  h.pos @ >r  dup >r		\ -- caddr len ; R: -- posn len
  over swap 2+ _read-file		\ -- caddr #read ior ; R: -- posn len
  over 0= over or if			\ if nothing read, or if error
    nip nip  0 0 rot			\ -- 0 0 ior
    r> drop  r> drop  exit		\ get out now
  endif
  drop					\ -- caddr #read ; R: -- posn len
  over swap #10 scan 0<> 1 and		\ -- caddr caddr' LF? ; R: -- posn len
  swap over + swap			\ -- caddr caddr'' LF? ; R: -- posn len
  2dup - 1- c@ #13 = 1 and +		\ -- caddr caddr'' #terms ; R: -- posn len
  -rot  swap -				\ -- #terms #used ; R: -- posn len
  tuck  swap - r> min  swap		\ -- u2 #used ; R: -- posn
  -1 swap r> + repo-file		\ -- u2 true ior
;
[then]

: (delete-file)	\ caddr len -- ior
\ *G Delete a named file from disk, and return ior=0 on success.
\ ** Unprotected.
  setdir				\ walk pathname, if any
  if  2drop err_BadPathName exit  then
  test-file				\ not folder or volume
  if  _delete 0  else  err_FileNotFound  then
  flushall
;

: (rename-file)	\ caddr1 len1 caddr2 len2 -- ior
\ *G Rename a named file on the disk, and return ior=0 on success.
\ ** Unprotected.
  2swap test-file if			\ -- caddr2 len2 ; MPE025
    savedDirEnt 2@ set-es
    0 dir[] name>dir
    wmark 0
  else
    err_FileNotFound
  then
  flushall
;

: (initFATfs)	\ -- ior
\ *G Initialize the FAT file system. Unprotected.
  MassInit				\ open mass storage
  closeall				\ clear file handles
  0 basesector !			\ assume no partitions
  bp_init bperror @ dup if		\ read FAT parameters, not a BPB
    drop  newb  454 bptr +!		\ point to first sector of first partition
    getl basesector !			\ read and set
    bp_init  bperror @
  endif
  dup 0=
  if  rootdir  endif
;

: (termFATfs)	\ --
\ *G Shut down the FAT file system. Unprotected.
  flushall closeall MassTerm
;

((
: sameFile?	\ h1 h2 -- flag
 \ *G Return true if the two handles point to the same file.
  swap h>rec swap h>rec			\ -- addr1 addr2
  over c@ 0=  over c@ 0= or		\ fam=0=closed
  if  2drop 0 exit  endif
  over 1+ c@ over 1+ c@ <>		\ check drives
  if  2drop 0 exit  endif
  3 cells + @ swap 3 cells + @ =	\ check start cluster
;

: #FileOpen	\ fileid -- u
 \ *G Return the number of times the file is open.
  0 swap  #files 0 do			\ -- u fileid
    dup i sameFile?
    if  swap 1+ swap  endif
  loop
  drop
;
))


\ *******************
\ *S Protection tools
\ *******************

tasking? [if]
Semaphore FileSem	\ --
\ *G Interlocks the file system for multitasking.
  1 FileSem !

: +FileLock	\ --
\ *G Wait until the file system is available and lock access
  FileSem request
;

: -FileLock	\ --
\ *G Unlock access to the file system.
  FileSem signal
;
[else]
compiler
: +FileLock  ;
: -FileLock  ;
target
[then]



\ *****************************
\ *S User API, mostly ANS Forth
\ *****************************
\ *P The API layer is made safe for multitasking systems by a
\ ** semaphore. Because the underlying hardware driver may
\ ** *\fo{THROW}, e.g. for a fatal read/write error, these words
\ ** include a *\fo{CATCH}.

FATopen? 0= [if]  external  [then]

: open-file	\ caddr len fam -- fileid ior
\ *G Open an existing file on disk.
\ ** This word will *\b{not} open a directory.
  +FileLock
\ [con ." Oi"  con]
  [dir
  ['] _open-file catch ?dup
  if  nip nip nip  -1 swap  endif
  dir]
\ [con ." oO"  con]
  -FileLock
;

: create-file	\ caddr len fam -- fileid ior
\ *G Create a file on disk, returning a 0 ior for success and
\ ** a file id. If the file already exists, it is truncated to
\ ** zero length.
  +FileLock [dir
  ['] _create-file catch ?dup
  if  nip nip nip  -1 swap  endif
  dir] -FileLock
;

: read-file	\ caddr len fileid -- #read ior
\ *G Read data from a file. The number of characters actually
\ ** read is returned as *\i{#read}, and *\i{ior} is returned 0
\ ** for a successful read.
  +FileLock
\ [con ." Ri"  con]
  ['] (read-file) catch ?dup
  if  nip nip nip  0 swap  endif
\ [con ." oR"  con]
  -FileLock
;

: write-file	\ caddr len fileid -- ior
\ *G Write a block of memory to a file.
  +FileLock [dir
  ['] (write-file) catch ?dup
  if  nip nip nip  endif
  dir] -FileLock
;

: close-file	\ fileid -- ior
\ *G Close an open file.
  +FileLock
\ [con ." Ci"  con]
  [dir  ['] (close-file) catch if nip endif  dir]
\ [con ." oC"  con] 1 ms
  -FileLock
;

FATtext? [if]
: write-line	\ caddr len fileid -- ior
\ *G Write data followed by CR/LF pair. IOR=0 for success.
  dup >r write-file  crlf 2 r> write-file or
;

: read-line     \ caddr u1 fileid -- u2 flag ior
\ *G Read an line of text from a file into a buffer, without EOL.
\ ** The EOL marker may be either CR/LF (DOS) or LF (Unix).
\ ** Read the next line from the file specified by fileid into memory
\ ** at the address *\i{caddr}. At most *\i{u1} characters are read.
\ ** Up to two line-terminating characters may be read into memory
\ ** at the end of the line, but are not included in the count *\i{u2}.
\ ** The line buffer provided by *\i{caddr} should be at least *\i{u1+2}
\ ** characters long.
\ *P If the operation succeeds, *\i{flag} is true and *\i{ior} is zero. If a
\ ** line terminator was received before *\i{u1} characters were read,
\ ** then *\i{u2} is the number of characters, not including the line
\ ** terminator, actually read (0 <= u2 <= u1). When u1 = u2, the
\ ** line terminator has yet to be reached.
\ *P If the operation is initiated when the value returned by
\ ** *\fo{FILE-POSITION} is equal to the value returned by *\fo{FILE-SIZE}
\ ** for the file identified by *\i{fileid}, *\i{flag} is false, *\i{ior} is zero,
\ ** and *\i{u2} is zero. If *\i{ior} is non-zero, an exception occurred
\ ** during the operation and *\i{ior} is the I/O result code.
\ *P An ambiguous condition exists if the operation is initiated
\ ** when the value returned by *\fo{FILE-POSITION} is greater than the
\ ** value returned by *\fo{FILE-SIZE} for the file identified by fileid,
\ ** or if the requested operation attempts to read portions of the
\ ** file not written.
\ *P At the conclusion of the operation, *\fo{FILE-POSITION} returns the
\ ** next file position after the last character read.
  +FileLock
\ [con ." Li"  con]
  ['] (read-line) catch ?dup
  if  nip nip nip 0 0 rot  endif
\ [con ." oL"  con]
  -FileLock
;
[then]

\ BJC019...
: resize-file	\ len-ud fileid -- ior
\ *G Set the size of the file to *\i{ud}, an unsigned double number.
\ ** After using *\fo{RESIZE-FILE}, the result returned by
\ ** *\fo{FILE-POSITION} may be invalid.
  +FileLock  handle !			\ -- len-ud
  if					\ file too big?
    drop err_FileSize
  else
    ['] _resize catch
  endif
  -FileLock  ;

: reposition-file	\ len-ud fileid -- ior
\ *G Set file position, and return ior=0 on success.
\ ** See the ANS Forth document.
  +FileLock  handle !			\ -- len-ud
  if					\ file too big?
    drop err_FileSize
  else
    repo-file				\ repo-file is now protected
  endif
  -FileLock
;
\ ...BJC019

: file-position	\ fileid -- len-ud ior
\ *G Return file position, and return ior=0 on success.
  +FileLock  handle !  h.pos @ 0 dup  -FileLock
;

: file-size	\ fileid -- len-ud ior
\ *G Get size in bytes of an open file as a double number,
\ ** and return ior=0 on success.
  +FileLock  handle !  h.len @ 0 dup  -FileLock
;

: delete-file	\ caddr len -- ior
\ *G Delete a named file from disk, and return ior=0 on success.
  +FileLock [dir
  ['] (delete-file) catch ?dup
  if  nip nip  endif
  dir] -FileLock
;

: FileExist?    \ c-addr u -- flag
\ *G Look to see if a specified file exists, returning TRUE
\ ** if the file exists.
  r/o Open-File if
    drop FALSE
  else
    close-file drop  TRUE
  then
;

: file-status	\ caddr len -- x ior
\ *G Return the status of the file identified by the character
\ ** string *\i{c-addr/len}. If the file exists, *\i{ior} is zero;
\ ** otherwise *\i{ior} is the implementation-defined I/O result
\ ** code. *\i{X} contains implementation-defined information about
\ ** the file (always zero for FATfiler).
   FileExist? 0= 0 swap
;

: rename-file	\ caddr1 len1 caddr2 len2 -- ior
\ *G Rename the file named by the character string *\i{c1addr/len1}
\ ** to the name in the character string *\i{caddr2/len2}. *\i{Ior}
\ ** is zero on success.
  +FileLock [dir
  ['] (rename-file) catch ?dup
  if  nip nip nip nip  endif
  dir] -FileLock
;

: mkdir		\ c-addr u -- ior
\ *G *\i{C-addr/u} is an 8.3 directory name with no directory
\ ** separators. *\fo{MKDIR} creates a new directory of that name
\ ** in the current directory. Directory removal is not supported
\ ** unless the code in *\i{DelDirs.fth} is compiled.
\ ** You can use *\fo{DELETE-FILE} to remove a directory but this
\ ** is not recommended unless you know that the directory is
\ ** empty. Otherwise, you may end up with lost chains.
  +FileLock [dir
  ['] _mkdir catch ?dup
  if  nip nip  endif
  dir] -FileLock
;

: mkDirEx	\ c-addr u -- ior
\ *G As *\fo{mkdir} but accepts directory separators. Directories
\ ** can be created outside the current directory. Directory removal
\ ** is not supported unless the code in *\i{DelDirs.fth} is compiled.
\ ** You can use *\fo{DELETE-FILE} to remove a directory but this
\ ** is not recommended unless you know that the directory is
\ ** empty. Otherwise, you may end up with lost chains.
  +FileLock [dir
  ['] _mkDirEx catch ?dup
  if  nip nip  endif
  dir] -FileLock
;

\ MPE025...
: cwd		\ "<dir>" -- ; CWD <dirname>
\ *G An equivalent to the DOS CD or Unix cwd command.
  bl word count $cwd
  if  ."  Can't do this"  endif
;

: pwd		\ -- ; PWD
\ *G Displays the current working directory name.
\ ** Because this is not a fully thread-safe file system and only
\ ** has a single "working directory" for all tasks/threads, you
\ ** will need to be careful with external access systems such as
\ ** HTTP, FTP and USB.
  path type
;
\ ...MPE025


tasking? [if]
: SyncDrive	\ --
\ *G Force all updated sectors to the disk and empty the sector
\ ** cache. This operation is necessary when directories are
\ ** modified by a separate task, e.g. when USB makes direct sector
\ ** read/write accesses.
  ['] flushall catch drop  initSecCache
;
[then]

: initFATfs	\ -- ior
\ *G Initialize the FAT file system.
  ['] (initFATfs) catch
  dup 0= if  drop  endif
[ tasking? ] [if]
  1 FileSem !				\ enable access
[then]
;

: termFATfs	\ --
\ *G Shut down the FAT file system.
  ['] (termFATfs) catch drop
;


\ **************************
\ *S Generic I/O file device
\ **************************
\ *P This device allows files to be used with *\fo{KEY}, *\fo{EMIT}
\ ** and friends. To avoid having to create a device for each
\ ** file, one device is built that takes the file id/handle from
\ ** the *\fo{USER} variable *\fo{MyFID}.

\ *P The application programmer is responsible for opening the
\ ** file, setting the user variable and closing the file.

FATio? [if]

cell +user MyFID	\ -- addr
\ *G User variable holding the file handle (0..n-1) in the low
\ ** 16 bits. The four bytes are used as follows:
\ *D 0/1	file handle
\ *D 2		character buffer for file read/write
\ *D 3		flags
\ *D 3.0	reserved
\ *D 3.1	set if EOF
\ *D 3.2	set after file error

\ *P *\fo{KEY?} always returns true. After an EOF or file error,
\ ** *\fo{KEY} always returns an LF character. This prevents
\ ** words such as *\fo{ACCEPT} blocking on error.


: FileKey?	\ -- flag
\ *G Returns true. Does not call *\fo{PAUSE}.
  true
;

: FileKey	\ -- char
\ *G The file version of *\fo{KEY}. Executes *\fo{PAUSE} if
\ ** LF is returned.
  MyFID 3 + c@ $06 and
  if  pause $0A exit  endif
  MyFID 2+ 1 MyFID w@ read-file if	\ read one character, error?
    drop  bit2 MyFID 3 + c!  pause $0A exit
  endif
  1 <> if				\ EOF check
    bit1 MyFID 3 + c!  pause $0A exit
  endif
  MyFID 2+ c@ dup $0A =
  if  pause  endif
;

: FileType	\ caddr len --
\ *G The file equivalent of *\fo{TYPE}. After any file error,
\ ** the string is discarded.
  MyFID 3 + c@ bit2 and			\ error check
  if  2drop exit  endif
  MyFID w@ write-file			\ write string
  if  bit2 MyFID 3 + c!  endif		\ error check
;

: FileEmit	\ char --
\ *G The file equivalent of *\fo{EMIT}. Uses *\fo{FileType}.
  MyFID 2 + tuck c!  1 FileType
;

: FileCr	\ char --
\ *G The file equivalent of *\fo{CR}. Uses *\fo{FileType} and
\ ** calls *\fo{PAUSE}.
  crlf 2 FileType  pause
;

create FileCon	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device for files.
  ' FileKey ,		\ -- char ; receive char
  ' FileKey? ,		\ -- flag ; check receive char
  ' FileEmit ,		\ -- char ; display char
  ' Filetype ,		\ caddr len -- ; display string
  ' FileCr ,		\ -- ; display new line

: FileAccept	\ c-addr +n1 -- +n2 ; read up to LEN chars into ADDR
\ *G As *\fo{ACCEPT} with no echoing or flow control.
  MyFID 3 + c@ $06 and			\ Error/EOF check
  if  2drop 0 exit  endif
  MyFID w@ read-line			\ read line from file, error?
  if  2drop 0  bit2 MyFID 3 + c!  exit  endif
  0=
  if  bit1 MyFID 3 + c!  endif
;

: FileEOF?	\ -- flag
\ *G Return true if the file is at EOF.
  MyFID 3 + c@ bit1 and
;

: FileErr?	\ -- flag
\ *G Return true if the file has returned an error.
  MyFID 3 + c@ bit2 and
;

: ?SetFID	\ fileid ior -- ior
\ *G Set *\fo{MyFID} according to the results of a file open
\ ** or create.
  tuck
  if  drop -1  endif
  MyFID w!  0 MyFID 2+ w!
;

: OpenFileCon	\ caddr len -- ior
\ *G Open the given file path in read-only mode as the current
\ ** task's file input device.
\ ** The path is given by *\i{caddr/len} and the mode by *\i{fam}.
\ ** The I/O pointers *\fo{IPVEC} and *\fo{OPVEC} are not set.
  r/o open-file ?SetFID
;

: CreateFileCon	\ caddr len -- ior
\ *G Create the given file path in read/write mode as the current
\ ** task's file output device.
\ ** The path is given by *\i{caddr/len} and the mode by *\i{fam}.
\ ** The I/O pointers *\fo{IPVEC} and *\fo{OPVEC} are not set.
  r/w create-file ?SetFID
;

: CloseFileCon	\ --
\ *G Close the file handle in *\fo{MyFID}.
  MyFID w@ close-file drop  -1 MyFID !
;

[then]


\ ****************
\ *S Drive details
\ ****************

FATspace? [if]
\ *G The code below will be compiled if the equate is non-zero.

: .DriveSize	\ --
   totalsec @ bytes/sec #1000000 */	\ drive size in Mb
   . ." Mb"
;

: .FATtype	\ --
\ *G Display FAT type.
  case FATtype @
    0 of  ." FAT12"  endof
    4 of  ." FAT16"  endof
    8 of  ." FAT32"  endof
      ." ???"
  endcase
;

: .Drive	\ --
\ *G Display drive details. *\fo{InitFATfs} must have been
\ ** successfully run.
  cr ." Drive Size       = " .DriveSize
  cr ." FAT type         = " .FATtype
  cr ." Start sector     = " basesector ?
  cr ." Reserved sectors = " reservedsec ?
  cr ." Sectors/Cluster  = " sec/cluster ?
  cr ." Root Dir sector  = " rootsector ?
  cr ." Root Dir entries = " rootentries ?
  cr ." Sectors/Track    = " SectorsPerTrack ?
  cr ." Number of heads  = " NumHeads ?
  cr ." Total sectors    = " totalsec ?
  cr ." FAT size         = " FATsize ?
  cr ." Cluster count    = " CountOfClusters dup . ." (0x" .dword ." )"
;

: TotalClusters	\ -- n
\ *G Returns the
  FATsize @ secs>bytes /fat
;

: UsedClusters	\ -- n
  0 TotalClusters
  FOR  r@ FAT@ if 1+ then  NEXT
;

[then]


\ ************
\ *S Test code
\ ************

FATtest? [if]
\ *G The code below will be compiled if the equate is non-zero.

#512 buffer: SBscr	\ -- addr
\ *G Scratch sector buffer.

-1 value hFile	\ --
\ *G File handle for tests.

: go		\ --
\ *G Start up the file system.
  initFATfs abort" Couldn't start FAT file system"
  rootdir
;

: testline	\ n -- caddr len
\ *G A line of text to output.
  55 mod 55 0				\ SFP009
  do  dup i + bl + pad i + c!  loop
  drop  pad 55
;

: testw		\ #lines "<filename>" --
\ *G Test file writing, use in the form:
\ *C   #lines testw file.ext
  bl word count r/w create-file abort" Can't create file" to hFile
  1 max 1000 min  0
  do  i testline hFile write-line abort" Write error"  loop
  hFile close-file abort" Error closing file"
;

variable lineno

[-short-branches

: testr		\ "<filename>" --
\ *G Test file reading with *\fo{read-line}. Use in the form:
\ *C   testr file.ext
   1 lineno !
   bl word count r/w open-file abort" Can't open file" to hFile
   begin
     SBscr 128 hFile read-line abort" Read error"
     lineno @ -1 <> and
    while
     SBscr swap  cr lineno @ 4 .r space  type
     1 lineno +!
     pause  key? if
       key drop
       cr ." ESC to exit, other to continue" key $1B =
       if  -1 lineno !  endif
     endif
  repeat
  drop
  hFile close-file abort" Error closing file"
;

: ?eol		\ caddr len -- caddr' len'
\ *G Strip the lst character from the line if it is CR or LF.
  2dup + 1- c@				\ test last char
  dup $0D = swap $0A = or		\ CR/LF test
  if  1-  endif
;

: rdLnCc	\ caddr len -- len' ; 0=eof
\ *G Read a line from the file on a character by character basis.
\ ** The line terminator is a line feed character.
  { | tmp[ cell ] -- }
  over 0  2swap bounds ?do		\ -- caddr len'
     tmp[ 1 hFile read-file  abort" Read error"
     1 <> ?leave
     1 +  tmp[ c@ dup i c!  $0A = ?leave
  loop
  ?eol ?eol nip
;

: testcr	\ "<filename>" --
\ *G Test file reading character by character. Use in the form:
\ *C   testcr file.ext
   1 lineno !
   bl word count r/w open-file abort" Can't open file" to hFile
   begin
     SBscr 128 rdLnCc
     dup lineno @ -1 <> and
    while				\ -- len
     SBscr swap  cr lineno @ 4 .r space  type
     1 lineno +!
     pause  key? if
       key drop
       cr ." ESC to exit, other to continue" key $1B =
       if  -1 lineno !  endif
     endif
  repeat
  drop
  hFile close-file abort" Error closing file"
;
short-branches]

0 value <s>	\ -- u
\ *G Last sector dumped.

: secdump	\ u --
\ *G Display the contents of the given sector.
  cr ." Sector " dup .
  dup to <s>
  SBscr swap cdrive c@ SecRead
  SBscr bytes/sec dump
;
: ns		\ --
\ *G Dump next sector.
  <s> 1+ secdump  ;
: ps		\ --
\ *G Dump previous sector.
  <s> 1- secdump  ;
: ss		\ --
\ *G Dump same sector again.
  <s> secdump  ;
[then]


\ ======
\ *> ###
\ ======

decimal

