\ DHCP.fth - Contains code for BOOTP/DHCP packet handling

((
Copyright (c) 2007, 2008,2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)

Substantial portions of this code have been derived from code
provided by:
Copyright (c) 2002, 2003, 2004
Hiden Analytical Ltd.
Tel:	01925 445225
Fax:	01925 416518

To do
=====

Change history
==============
20101122 SFP007 DNS changes
20081124 SFP006 Changed for SNTP.
20080305 SFP005 Fixed network order problem.
20080225 SFP004 Forced timeout on DHCP packet reception.
20070828 VRW003 Now uses UDP service chain.
		Moved SendDatagram to UDP.fth.
20070214 MPE002 Adapted from Hiden code.
20031022 SBD001 AddDHCPOptions, ParseDHCPOptions.
		Fetch BOOTPMAGIC in network order.
))

only forth definitions
decimal


\ =======
\ *! dhcp
\ *T DHCP and BOOTP
\ =======
\ *P The code in *\i{<PNet>\DHCP.fth} implements a DHCP *\b{client}
\ ** according to RFC2131. See also RFC2132 and RFC1542. Do not attempt
\ ** to modify this code until you have absorbed these RFCs.

\ *P If you do not have a DHCP server on you network, there are many
\ ** available. For Windows PCs, several users have recommended
\ ** the *\b{haneWin DHCP Server} from *\i{www.hanewin.net}.

0 equ DHCPdebug?	\ -- x
\ *G Set this non-zero to display DHCP debug information during
\ ** execution.


\ ***************************
\ *S DHCP/BOOTP state machine
\ ***************************
\ *P The DHCP system is controlled by a state machine, which is
\ ** run whenever the timer *\fo{DHCPtimer) times out. The action
\ ** performed is responsible setting the next state and updating
\ ** the timer. The state machine actions are performed by
\ ** *\fo{DHCPidle ( -- )} which is called in the service task.

\ *P States of the DHCP state machine. See RFC2131 Figure 5 below
\ ** for the transition diagram.

\ *E  --------                               -------
\ ** |        | +-------------------------->|       |<-------------------+
\ ** | INIT-  | |     +-------------------->| INIT  |                    |
\ ** | REBOOT |DHCPNAK/         +---------->|       |<---+               |
\ ** |        |Restart|         |            -------     |               |
\ **  --------  |  DHCPNAK/     |               |                        |
\ **     |      Discard offer   |      -/Send DHCPDISCOVER               |
\ ** -/Send DHCPREQUEST         |               |                        |
\ **     |      |     |      DHCPACK            v        |               |
\ **  -----------     |   (not accept.)/   -----------   |               |
\ ** |           |    |  Send DHCPDECLINE |           |                  |
\ ** | REBOOTING |    |         |         | SELECTING |<----+            |
\ ** |           |    |        /          |           |     |DHCPOFFER/  |
\ **  -----------     |       /            -----------   |  |Collect     |
\ **     |            |      /                  |   |       |  replies   |
\ ** DHCPACK/         |     /  +----------------+   +-------+            |
\ ** Record lease, set|    |   v   Select offer/                         |
\ ** timers T1, T2   ------------  send DHCPREQUEST      |               |
\ **     |   +----->|            |             DHCPNAK, Lease expired/   |
\ **     |   |      | REQUESTING |                  Halt network         |
\ **     DHCPOFFER/ |            |                       |               |
\ **     Discard     ------------                        |               |
\ **     |   |        |        |                   -----------           |
\ **     |   +--------+     DHCPACK/              |           |          |
\ **     |              Record lease, set    -----| REBINDING |          |
\ **     |                timers T1, T2     /     |           |          |
\ **     |                     |        DHCPACK/   -----------           |
\ **     |                     v     Record lease, set   ^               |
\ **     +----------------> -------      /timers T1,T2   |               |
\ **                +----->|       |<---+                |               |
\ **                |      | BOUND |<---+                |               |
\ **   DHCPOFFER, DHCPACK, |       |    |            T2 expires/   DHCPNAK/
\ **    DHCPNAK/Discard     -------     |             Broadcast  Halt network
\ **                |       | |         |            DHCPREQUEST         |
\ **                +-------+ |        DHCPACK/          |               |
\ **                     T1 expires/   Record lease, set |               |
\ **                  Send DHCPREQUEST timers T1, T2     |               |
\ **                  to leasing server |                |               |
\ **                          |   ----------             |               |
\ **                          |  |          |------------+               |
\ **                          +->| RENEWING |                            |
\ **                             |          |----------------------------+
\ **                              ----------
\ **           Figure 5: State-transition diagram for DHCP clients

\ *[
[sm
  smState smDHidle	\ idle, do nothing (pseudostate)
  smState smDHinit	\ initialisation, o/p DHCPDISCOVER
  smState smDHselect	\ process DHCPOFFER(s)
  smState smDHrequest	\ process ACK, discard offers
			\ ACK/decline -> o/p DHCPDECLINE, smDHinit
			\ ACK/accept -> record lease, smDHbound
  smState smDHarpcheck	\ go ACK, check -> DHCPdecline/init or bound
  smState smDHbound	\ discard messages, time out ->
  smState smDHrenew	\ wait ACK, T2 expiry
  smState smDHrebind	\ wait ACK/NAK
  smState smDHrebooting	\ warm restart, DHCPREQUEST, wait ACK/NAK
sm] #smDHactions	\ number of states in this machine
\ *]

variable DHCPTimer	\ -- addr
\ *G Holds the *\fo{TICKS} time at which the next DHCP action will
\ ** be performed, 0 for no action.
  0 DHCPtimer !

variable smDHcount	\ -- addr
\ *G Down counter used to provide repetitions and timeouts.

variable smDHstate	\ -- n
\ *G Holds current DHCP state.

create smDHactions	\ -- addr
\ *G Execution table containing the xts of the action words.
  #smDHactions cells allot&erase

XC? [if]
interpreter
: setDHstate	\ xt state --
\ *G *\fo{INTERPRETER} word that sets the action of the given state.
  cells smDHactions + !
;
target
[else]
: setDHstate	\ xt state --
  cells smDHactions + !
;
[then]

: DHCPidle	\ --
\ *G Performed periodically to do DHCP actions and lease checks.
  DHCPtimer @ dup if			\ Perform action?
    dup timedout? if
      smDHstate @ cells			\ table index
      smDHactions + @ execute		\ execute state action
      smDHcount decr			\ repetition count down
    endif
  endif
  drop
;


\ ***************************
\ *S State machine utilitiles
\ ***************************

: SetDHCPtimer	\ ms --
\ *G Set the DHCP timer to time out *\i{ms} later.
  later 1 or DHCPtimer !
;

DHCPdebug? [if]
: .DHstate	\ --
\ *G Debug tool to display the current state of the DHCP state machine.
  cr ." State: "
  case  smDHstate @
    smDHidle of  ." idle "  endof
    smDHinit of  ." init "  endof
    smDHselect of  ." select "  endof
    smDHrequest of  ." request "  endof
    smDHarpcheck of  ." ARPcheck"  endof
    smDHbound of  ." bound "  endof
    smDHrenew of  ." renew "  endof
    smDHrebind of  ." rebind "  endof
    smDHrebooting of  ." rebooting "  endof
      ." unknown " dup .
  endcase
;
[then]

: toDHCPstate	\ ms n state --
\ *G The given *\i{state} will be executed after the given
\ ** period in *\i{ms} and *\i{n} times if sensitive to the number of
\ ** times. Note that this specifies the interval until the
\ ** state is executed, not the repetition rate of the state.
  smDHstate !  smDHcount !  setDHCPtimer
[ DHCPdebug? ] [if]  .DHstate  [then]
;


\ ************************
\ *S DHCP Data definitions
\ ************************

\ ====================
\ *N DHCP control data
\ ====================
\ *P These variables contain the working data used to
\ ** access the DHCP server.

variable BOOTP_XID	\ -- addr
\ *G Transaction ID number.
variable DHCPLease	\ -- addr
\ *G Length of DHCP lease in ms.
variable DHCPLeaseTimer	\ -- addr
\ *G *\fo{TICKS} at which lease times out.
variable DHCP_T1	\ -- addr
\ *G Time to RENEW in ms.
variable DHCP_T2	\ -- addr
\ *G time to REBIND in ms.
variable DHCPServer	\ -- addr
\ *G DHCP server IP address.
wvariable DHCPflags	\ -- addr
\ *G Holds the DHCP flags, set to $8000 to force broadcast
\ ** responses, or 0 when we can accept unicast responses.
  $8000 DHCPflags w!


\ ======================
\ *N DHCP transient data
\ ======================
\ *P These variables contain data generated by the DHCP
\ ** transactions. It is later manipulated to set system
\ ** variables and the DHCP control data after validation.

variable GivenIP	\ --
\ *G IP address given by DHCP or BOOTP.
variable GivenMask	\ --
\ *G IP mask given by DHCP or BOOTP.
variable GivenGateway	\ --
\ *G IP gateway given by DHCP or BOOTP.
variable GivenServer	\ --
\ *G IP address of DHCP server given by DHCP or BOOTP.
variable GivenLease	\ -- addr
\ *G Length of DHCP lease in seconds given by DHCP or BOOTP.
variable GivenT1	\ -- addr
\ *G Time to RENEW in ms given by DHCP or BOOTP.
variable GivenT2	\ -- addr
\ *G time to REBIND in ms given by DHCP or BOOTP.
variable GivenReply?	\ -- addr
\ *G Set non-zero when a DHCP reply has been received, but the
\ ** data has not yet been been processed. During this time any
\ ** following DHCP packets will be discarded.
SNTPauto? [if]
variable GivenSNTP	\ -- addr
\ *G SNTP server returned by DHCP.
[then]
DNSauto? [if]
variable GivenDNS	\ -- addr
\ *G DNS server returned by DHCP.
[then]

\ reserved for future use
\ variable DomainNameServer
\ variable BroadcastAddr
\ variable NBNSCounter
\ variable NBNSLeaseTimer
\ Wvariable APIPAsubnetaddr	\ APIPA next ip addr for ARP
\ Wvariable IP_MTU_VAR
\ Cvariable IP_TTL_VAR
\ Cvariable TCP_TTL_VAR


\ =====================
\ *N DHCP packet layout
\ =====================

STRUCT /dhcp	\ -- len
\ *G BOOTP and DHCP payload structure after the IP and UDP headers.
  1 FIELD dh.op		\ op code
  1 FIELD dh.htype	\ ARP hardware address type
  1 FIELD dh.hlen	\ hardware address length
  1 FIELD dh.hops	\ 0, for use by gateways
  4 FIELD dh.xid	\ transaction ID
  2 FIELD dh.secs	\ seconds since start of boot
  2 FIELD dh.flags	\ See RFC2131, top bit set for broadcasts
  4 FIELD dh.ciaddr	\ client IP address - client->server
  4 FIELD dh.yiaddr	\ your IP address   - server->client
  4 FIELD dh.siaddr	\ server's ip address
  4 FIELD dh.giaddr	\ gateway ip address
  16 FIELD dh.chaddr	\ client hardware address
  64 FIELD dh.sname	\ server host name
  128 FIELD dh.file	\ path of boot file
  312 FIELD dh.options	\ BOOTP: vendor info, 64; DHCP options, 312 min
END-STRUCT

CREATE BOOTPMAGIC	\ -- addr
\ *G Magic number used to identify BOOTP and DHCP packets.
   99 C,  130 C,  83 C,  99 C,


\ *************
\ *S DHCP tools
\ *************

[undefined] umin [if]
: umin		\ u1 u2 -- u1|u2
\ *G Minimum of two unsigned values.
  2dup u>
  if  swap  then
  drop
;
[then]

: TxMs		\ secs -- ms
\ *G Return the number of seconds required for expiry of the
\ ** T1/T2 timer.
  2/  DHCPlease @ umin   #1000 * $7FFFFFFF and
;


: NoDHCPserver	\ --
\ *G Set up for no known DHCP server.
  INADDR_BROADCAST  dup GivenServer !(n)  DHCPServer !(n)
;

: AcceptDHCPin	\ --
\ *G Accept incoming DHCP packets by clearing *\fo{GivenReply?}.
  0 GivenReply? !
;

: RejectDHCPin	\ --
\ *G Reject incoming DHCP packets by setting *\fo{GivenReply?}
\ ** to an non-valid value.
  -1 GivenReply? !
;

: AcceptGiven	\ --
\ *G Transfer the data given by DHCP transactions to the working
\ ** data.
  GivenIp @(n) IPaddress !(n)
  GivenMask @(n) EnetIPmask !(n)
  GivenGateway @(n) IPgateway !(n)
  GivenServer @(n) DHCPserver !(n)
  GivenLease @ DHCPLease !
  GivenT1 @ DHCP_T1 !
  GivenT2 @ DHCP_T2 !
[ SNTPauto? ] [if]
  GivenSNTP @(n) SNTPserver !(n)
[then]
[ DNSauto? ] [if]
  GivenDNS @(n) DNSserver !(n)
[then]
;


\ =======================
\ *S DHCP state selection
\ =======================
\ *P Each state is controlled by three values defined here.
\ *(
\ *B Entry delay - delay before state is executed for the first time.
\ *B Period - period between repetitions of the same state.
\ *B Count - number of repetitions before *\fo{smDHcount} reaches zero.
\ *)

\ smDHidle

#1000 equ idlePreMs
#1000 equ idleMs
1 equ #idle

: goIdle	\ --
\ *G Go to the *\fo{smDHidle} state.
  idlePreMs  #idle smDHidle toDHCPstate
;


\ smDHinit
#10000 equ initPreMs
#4000 equ initMs
4 equ #init

: goInit	\ --
\ *G Go to the *\fo{smDHnit} state, so starting the DHCP process.
  NoDHCPserver  $8000 DHCPflags w!	\ no server, need broadcast reply
  INADDR_BROADCAST
  dup IPgateway !(n)  GivenGateway !(n)
  IN_CLASSC_NET
  dup EnetIPMask !(n)  GivenMask !(n)
  DHCP_INFINITE
  dup DHCPLease !  dup DHCP_T1 !  dup DHCP_T2 !
  dup GivenLease !  dup GivenT1 !  GivenT2 !
  0 dup IpAddress !(n)  GivenIP !
  ticks BOOTP_XID !
  AcceptDHCPin
  initPreMs  #init smDHinit toDHCPstate
;


\ smDHselect

#100 equ selectPreMs
#100 equ selectMs
2 equ #select

: goSelect	\ --
\ *G Go to the *\fo{smDHselect} state.
  selectPreMs  #select smDHselect toDHCPstate
;


\ smDHrequest

#100 equ requestPreMs
#4000 equ requestMs
4 equ #request

: goRequest	\ --
\ *G Go to the *\fo{smDHrequest} state.
  AcceptGiven				\ accept acquired parameters, why?
  requestPreMs  #request smDHrequest toDHCPstate
;

\ smDHarpcheck
\ ARP is strictly local - not routed - so 200ms should do.

#200 equ arpcheckPreMs 	\ -- ms
#200 equ arpcheckMs 	\ -- ms
5 equ #arpcheck

: goARPcheck	\ --
\ *G Go to the *\fo{smDHselect} state.
  RejectDHCPin				\ ignore all incoming DHCP packets
  arpcheckPreMs  #arpcheck smDHarpcheck toDHCPstate
;

\ smDHbound

#1000 equ boundPreMs
#1000 equ boundMs
1 equ #bound

: goBound	\ --
\ *G Go to the *\fo{smDHbound} state.
  AcceptGiven				\ accept acquired parameters
  RejectDHCPin				\ ignore all incoming DHCP packets
  DHCPlease @ DHCP_INFINITE = if
    boundPreMs
  else
    DHCP_T1 @ TxMs
  endif
  #bound smDHbound toDHCPstate
;

\ smDHrenew - uses values from smdHrequest

: goRenew	\ --
\ *G Go to the *\fo{smDHrenew} state.
  $0000 DHCPflags w!  AcceptDHCPin	\ accept incoming packets.
  requestPreMs  #request smDHrenew toDHCPstate
;

\ smDHrebind - uses values from smdHbound

: goRebind	\ --
\ *G Go to the *\fo{smDHrebind} state.
  DHCPlease @ DHCP_INFINITE = if
    boundPreMs
  else
    DHCP_T2 @ DHCP_T1 @ - TxMs
  endif
  #request smDHrebind toDHCPstate
;


\ smDHrebooting - uses values from smdHrequest

: goRebooting	\ --
\ *G Go to the *\fo{smDHreboot} state.
  AcceptDHCPin				\ accept incoming packets.
  requestPreMs  #request smDHrebooting toDHCPstate
;


\ *******************
\ *S UDP transmission
\ *******************

: (BOOTPSend)	{ *pb ipaddr  -- }
\ *G Send BOOTP data as a UDP packet. The parameters are *\i{*pb},
\ ** a pointer to a PBuff containing the UDP data, and *\i{ipaddr},
\ ** the IP address to send to.
  udp_hdr *pb unbumpPBuffData		\ point data to udp header
  IPPORT_BOOTPCLIENT			\ get local port number
  *pb PBDataStart uh_sport w!(n)	\ into udp header
  IPPORT_BOOTPSERVER			\ and remote port number
  *pb PBDataStart uh_dport w!(n)	\ into udp header
  *pb ipaddr SendDatagram
;


\ *************************
\ *S Outgoing message tools
\ *************************
\ *P Some data is carried as DHCP options in the DHCP packet.
\ ** Options that may contain an odd number of bytes are forced
\ ** to a 16 bit boundary.

XC? [if]
compiler
: c!++		\ addr b -- addr+1
\ *G *\fo{COMPILER} word, saves *\i{b} at *\i{addr} and
\ ** increments *\i{addr}.
  over c!  1+
;

: !(n)++	\ addr x -- addr+4
\ *G *\fo{COMPILER} word, saves *\i{x} at *\i{addr} and
\ ** increments *\i{addr}.
  over !(n)  4+
;
target
[else]
: c!++		\ addr b -- addr+1
  over c!  1+
;

: !(n)++	\ addr x -- addr+4
  over !(n)  4+
;
[then]

: haligned	\ *opt -- *opt'
\ *G Align option - assumes that options start 16 bit aligned.
\ ** If necessary a zero (pad) byte is added.
  dup 1 and
  if  0 c!++  endif			\ align
;

: INCLUDE_SERVERID_OPT	\ *opt -- *opt'
\ *G Add served IP address to the options list.
  DHCP_OPT_SERVERID c!++  4 c!++
  GivenServer @(n) !(n)++
;

: INCLUDE_REQ_INF_LEASE_OPT	\ *opt -- *opt'
\ *G Add request for an ininite lease to the options list.
  DHCP_OPT_LEASE c!++  4 c!++
  DHCP_INFINITE !(n)++
;

: INCLUDE_OFFERED_LEASE_OPT	\ *opt -- *opt'
\ *G Add the offered lease to the options list.
  DHCP_OPT_LEASE c!++  4 c!++
  GivenLease @ !(n)++
;

: INCLUDE_REQUESTADDR_OPT	\ *opt -- *opt'
\ *G Add the offered IP address to the options list.
  DHCP_OPT_REQUESTADDR c!++  4 c!++
  GivenIP @(n) !(n)++
;

CREATE ParamRequestList
\ *G Parameter Request List that indicates which responses we want.
  5					\ base number of options
SNTPauto? [if] 1+ [then]
DNSauto? [if] 1+ [then]
  c,					\ total number of options that follow
(  1 ) DHCP_OPT_SUBNETMASK c,
(  3 ) DHCP_OPT_ROUTER c,
( 54 ) DHCP_OPT_SERVERID c,
( 58 ) DHCP_OPT_T1 c,
( 59 ) DHCP_OPT_T2 c,
SNTPauto? [if]
( 42 ) DHCP_OPT_SNTP c,
[then]
DNSauto? [if]
(  6 ) DHCP_OPT_DOMAINSERVERS C,	\ server
[then]

\ (  6 ) DHCP_OPT_DOMAINSERVERS C,	\ server
\ ( 15 ) DHCP_OPT_DOMAINNAME C,		\ my domain name

: INCLUDE_PARAMREQ_OPT	\ *opt -- *opt'
\ *G Add parameter request list to the options list.
  DHCP_OPT_PARAMREQ c!++
  ParamRequestList dup c@ 1+ >R
  over R@ cmove  R> +
  haligned
;

-short-branches
\ was AddDHCPOptions
: addDHCPopts	\ msgtype *bootp --
\ *G Add the relevant options for the message type to the options
\ ** section of the DHCP message.
  dh.options				\ -- msgtype *opt
  BOOTPMAGIC @(n) !(n)++		\ BOOTP/DHCP magic number
  DHCP_OPT_MSGTYPE c!++  1 c!++		\ message type option
  over c!++
  DHCP_OPT_CLIENTID c!++		\ Client ID = Ethernet MAC
  ETHER_ADDRESS_LEN 1+ c!++
  ARPHARD_ETHER c!++			\ Hardware type - Not v. explicit
  EtherAddress over CopyEthAdd		\ Ethernet address (6 bytes)
  ETHER_ADDRESS_LEN +

  CASE  swap				\ -- *opt msgtype
\ Client -> server messages
    DHCPDISCOVER OF
      INCLUDE_REQ_INF_LEASE_OPT
    ENDOF
    DHCPREQUEST OF
      smDHstate @ dup smDHrenew <>	\ not when renewing
      swap smDHrebind <> and  if	\ not when rebinding
        INCLUDE_REQUESTADDR_OPT
        INCLUDE_SERVERID_OPT
      endif
      INCLUDE_OFFERED_LEASE_OPT
      INCLUDE_PARAMREQ_OPT
    ENDOF
    DHCPDECLINE OF
      INCLUDE_REQUESTADDR_OPT
      INCLUDE_SERVERID_OPT
    ENDOF
\ we do not do DHCPINFORM or DHCPRELEASE yet
  ENDCASE
  255 swap c!				  \ end of options
;
+short-branches

((
Option                     DHCPDISCOVER  DHCPREQUEST      DHCPDECLINE,
                           DHCPINFORM                     DHCPRELEASE
------                     ------------  -----------      -----------
Requested IP address       MAY           MUST (in         MUST
                           (DISCOVER)    SELECTING or     (DHCPDECLINE),
                           MUST NOT      INIT-REBOOT)     MUST NOT
                           (INFORM)      MUST NOT (in     (DHCPRELEASE)
                                         BOUND or
                                         RENEWING)
IP address lease time      MAY           MAY              MUST NOT
                           (DISCOVER)
                           MUST NOT
                           (INFORM)
Use 'file'/'sname' fields  MAY           MAY              MAY
DHCP message type          DHCPDISCOVER/ DHCPREQUEST      DHCPDECLINE/
                           DHCPINFORM                     DHCPRELEASE
Client identifier          MAY           MAY              MAY
Vendor class identifier    MAY           MAY              MUST NOT
Server identifier          MUST NOT      MUST (after      MUST
                                         SELECTING)
                                         MUST NOT (after
                                         INIT-REBOOT,
                                         BOUND, RENEWING
                                         or REBINDING)
Parameter request list     MAY           MAY              MUST NOT
Maximum message size       MAY           MAY              MUST NOT
Message                    SHOULD NOT    SHOULD NOT       SHOULD
Site-specific              MAY           MAY              MUST NOT
All others                 MAY           MAY              MUST NOT
))


\ It was found that the Allied Telesyn Router DHCP server
\ always uses the IP address of its Ethernet 0 (ETH0) port
\ as its Server ID and does not fill in its siaddr field
\ We have decided to play safe and if there is any doubt to
\ broadcast the reply.

: UseBroadcast?	\ ServerAddr IPAddr -- ServerAddr
\ *G If ServerAddr and IPAddr are not the same and
\ ** IPAddr is not INADDR_BROADCAST (the initialised value)
\ ** then substitute INADDR_BROADCAST for ServerAddr.
\ ** So if siaddr is left at 0 in the DHCP offer
\ ** the Request will be broadcast.
  2dup <>  swap INADDR_BROADCAST <>  and
  if  drop INADDR_BROADCAST  then
;

: GetBOOTPServerAddr	\ -- ipaddr
\ *G If there is any doubt about the server IP address then
\ ** broadcast the reply. At this point the routing table
\ ** only contains entries for the received BOOTP reply and
\ ** the broadcast address so there is no point in replying to
\ ** the siaddr or the address in DHCPServerID if it differs from
\ ** BOOTPServer.
  GivenServer @(n) DHCPServer @(n) UseBroadcast?
;

DHCPdebug? [if]
: .DHCPtype	\ msgtype --
\ *G Display the message type as text.
   ." DHCP"
   case
     DHCPDISCOVER of  ." discover "  endof
     DHCPOFFER    of  ." offer "  endof
     DHCPREQUEST  of  ." request "  endof
     DHCPDECLINE  of  ." decline "  endof
     DHCPACK      of  ." ack "  endof
     DHCPNAK      of  ." nak "  endof
     DHCPRELEASE  of  ." release "  endof
     DHCPINFORM   of  ." inform "  endof
       ." unknown " dup .
  endcase
;
[then]

: SendDHCP	{ msgtype | *pb *bootp -- }
\ *G Send a DHCP message. The fields and options are filled in as
\ ** appropriate for the message type.
[ DHCPdebug? ] [if]
cr ." OUT: " message-type .DHCPtype
[then]
  AcceptDHCPin				\ mark no response yet
  GetPbuf ?dup 0=			\ must grab a buffer
  if  exit  endif
  -> *pb				\ save pbuf address for use

  0 p_dat *pb p_data w!			\ initial packet data pointer
  ether_hdr ip_hdr + udp_hdr +          \ calculate header size
  dup *pb p_data w+!  *pb p_len w!	\ move dp by header size

  *pb >UdpDataPos + -> *bootp
  *bootp /dhcp erase			\ set to zeros

  BOOTPREQUEST *bootp dh.op c!		\ op code = request
  ARPHARD_ETHER *bootp dh.htype c!	\ ARP hardware address type
  ETHER_ADDRESS_LEN *bootp dh.hlen c!	\ Ethernet MAC address length
  EtherAddress *bootp dh.chaddr CopyEthAdd

  BOOTP_XID @ *bootp dh.xid !(n)	\ transaction ID
  TICKS 1000 / *bootp dh.secs w!(n)	\ seconds since start of boot
  DHCPflags w@ *bootp dh.flags w!(n)	\ broadcast flag
  GivenIP @(n) *bootp dh.ciaddr !(n)	\ client IP address

  msgtype *bootp addDHCPopts		\ add DHCP options

[ DHCPdebug? ] [if]
  cr ." Header"  *bootp #48 dump
  cr ." Options"  *bootp dh.options #64 dump
[then]
												\ to length
  /dhcp  dup *pb p_datalen w! 		\ set data length
  *pb p_len w+!  		  	\ modify packet length

  *pb GetBOOTPServerAddr (BOOTPSend)
;

((
Field      DHCPDISCOVER          DHCPREQUEST           DHCPDECLINE,
           DHCPINFORM                                  DHCPRELEASE
-----      ------------          -----------           -----------
'op'       BOOTREQUEST           BOOTREQUEST           BOOTREQUEST
'htype'    (From "Assigned Numbers" RFC)
'hlen'     (Hardware address length in octets)
'hops'     0                     0                     0
'xid'      selected by client    'xid' from server     selected by
                                 DHCPOFFER message     client
'secs'     0 or seconds since    0 or seconds since    0
           DHCP process started  DHCP process started
'flags'    Set 'BROADCAST'       Set 'BROADCAST'       0
           flag if client        flag if client
           requires broadcast    requires broadcast
           reply                 reply
'ciaddr'   0 (DHCPDISCOVER)      0 or client's         0 (DHCPDECLINE)
           client's              network address       client's network
           network address       (BOUND/RENEW/REBIND)  address
           (DHCPINFORM)                                (DHCPRELEASE)
'yiaddr'   0                     0                     0
'siaddr'   0                     0                     0
'giaddr'   0                     0                     0
'chaddr'   client's hardware     client's hardware     client's hardware
           address               address               address
'sname'    options, if           options, if           (unused)
           indicated in          indicated in
           'sname/file'          'sname/file'
           option; otherwise     option; otherwise
           unused                unused
'file'     options, if           options, if           (unused)
           indicated in          indicated in
           'sname/file'          'sname/file'
           option; otherwise     option; otherwise
           unused                unused
'options'  options               options               (unused)
))


: SendDiscover	\ --
\ *G Send a DHCPDISCOVER message.
  NoDHCPserver  DHCPDISCOVER SendDHCP
;

: SendRequest	\ --
\ *G Send a DHCPREQUEST message.
  DHCPREQUEST SendDHCP
;

: SendDecline	\ --
\ *G Send a DHCPDECLINE message.
  NoDHCPserver  DHCPDECLINE SendDHCP
;

: doDHidle	\ --
\ *G The idle action is just to restart the timer.
  IdleMs SetDHCPtimer			\ restart timer
;

: doDHinit	\ --
\ *G Send DHCPDISCOVER periodically until we get a response
  InitMs SetDHCPtimer			\ restart timer
  GivenReply? @ DHCPOFFER = if
    GivenServer @(n) INADDR_BROADCAST <> if
      GivenServer @(n) DHCPserver !(n)
      goSelect  exit
    endif
  endif
  smDHcount @ 0=			\ unless used all tries
  if  goInit  else  SendDiscover  endif
;

: doDHselect	\ --
\ *G Select is currently a no-op as we take the first offer we
\ ** receive.
  goRequest
;

: doDHrequest	\ --
\ *G Send DHCPREQUEST periodically until we get a response
  RequestMs SetDHCPtimer		\ restart timer
  GivenReply? @ DHCPACK = if
    goARPcheck				\ we have a good response
  else
    smDHcount @ 0=			\ unless used all tries
    if goInit else SendRequest endif
  endif
;

: doDHarpcheck	\ --
\ *G The DHCP specification says that clients *\b{SHOULD} check
\ ** that the IP address given to them is not used elsewhere. If
\ ** a route is found, the process is restarted.
  ARPcheckMs setDHCPtimer
  GivenIP @(n) (find_route) if		\ DHCP failed if route found
    SendDecline  goInit
  else
    smDHcount @ if
      GivenIP @(n) SendARPrequest
    else
      .PowerNet goBound			\ display state and accept
    endif
  endif
;

: doDHbound	\ --
\ *G Wait until the T1 timer expires, and go to the RENEWING state.
  DHCPlease @ DHCP_INFINITE <> if	\ if we have a finite lease
    goRenew
  else
    boundMs setDHCPtimer		\ just maintain state
  endif
;

: doDHrenew	\ --
\ *G Send DHCPREQUEST periodically until we get a response.
  RequestMs SetDHCPtimer		\ restart timer
  DHCPlease @ DHCP_INFINITE <> if
    GivenReply? @ DHCPACK = if
      goBound				\ we have a good response
    else
      smDHcount @ 0=			\ unless used all tries
      if  goRebind  else  SendRequest  endif
    endif
  endif
;

: doDHrebind	\ --
\ *G Send DHCPREQUEST periodically until we get a response.
  RequestMs SetDHCPtimer		\ restart timer
  DHCPlease @ DHCP_INFINITE <> if
    case  GivenReply? @
      DHCPACK of  goBound  endof	\ good response
      DHCPNAK of  goInit  endof		\ bad response
        smDHcount @ 0=			\ unless used all tries
        if  goInit  else  SendRequest  endif
    endcase
  endif
;

: doDHrebooting	\ --
\ *G Send DHCPREQUEST periodically until we get a response.
  doDHrebind
;


\ ****************************
\ *S Receive BOOTP/DHCP packet
\ ****************************
\ *P DHCP packet reception is treated as a special case because
\ ** DHCP uses two ports and must operate before the unit's IP
\ ** address has been set.

: doDHCPopt	\ *option -- *option'
\ *G Process the given option if we know about it and step over it.
\ ** Unknown options are ignored. Values for known options are saved
\ ** in the *\fo{Givenxxx} variables.
  dup 2+  swap c@			\ pick up and step over length
  case
    0 of  1- exit  endof		\ padding, has no length
    DHCP_OPT_SUBNETMASK of  dup @(n) GivenMask !(n)  endof
    DHCP_OPT_ROUTER of  dup @(n) GivenGateway !(n)  endof
\    DHCP_OPT_BROADCASTADDR of  dup SetMyBroadcastAddr  endof
    DHCP_OPT_MSGTYPE of  dup c@ GivenReply? !  endof
    DHCP_OPT_SERVERID of  dup @(n) GivenServer !(n)  endof
    DHCP_OPT_LEASE of  dup  @(n) GivenLease !  endof
    DHCP_OPT_T1	of  dup @(n) GivenT1 !  endof
    DHCP_OPT_T2	of  dup @(n) GivenT2 !  endof
[ SNTPauto? ] [if]
    DHCP_OPT_SNTP of  dup @(n) GivenSNTP !(n)  endof	\ SFP005
[then]
[ DNSauto? ] [if]
    DHCP_OPT_DOMAINSERVERS of  dup @(n) GivenDNS !(n)  endof	\ SFP007
[then]
  endcase
  dup 1- C@ +				\ get length & skip
;

: scanDHCPopts	\ *bootp len --
\ *G Given the DHCP data area, parse the DHCP options.
  over + >r  dh.options			\ -- *opt ; R: -- end
  dup @(n) BOOTPMAGIC @(n) = if
    4+					\ step over magic number
    begin
      dup c@ 255 <>			\ stop at END option
     while
      dup r@ u<				\ stop at end of packet
     while
      doDHCPopt				\ process option
    repeat then
  then
  drop  r> drop				\ -- type
;

-short-branches
: RxDHCP	{ *pb | *udp *bootp len -- }
\ *G Process received DHCP and BOOTP packets.
\ ** When we get here we know this is from port 67 to port 68
\ ** and that the Ethernet packet was either addressed to us or
\ ** was broadcast.

\  cr ." In RxDHCP"
\  *pb PBHdrStart .ippkt

  GivenReply? @ 0= if
\ The packet will be discarded if previous one has not been processed.
\ This is necessary in networks with multiple DHCP servers.
    *pb GetUDPAddress -> *udp
    *udp >UDPdata -> *bootp

    \ Check this is a BOOTP or DHCP packet
    *bootp dh.op c@ BOOTPREPLY =		\ op code = reply
    *bootp dh.htype c@ ARPHARD_ETHER = AND	\ ARP hardware address type
    *bootp dh.hlen c@ ETHER_ADDRESS_LEN = AND	\ hardware address length
    if
      *bootp dh.chaddr				\ check this is for our Ethernet
      EtherAddress ETHER_ADDRESS_LEN s= if
        *bootp dh.yiaddr @(n) GivenIP !(n)
        *bootp *udp GetUDPlen scanDHCPopts	\ scan options
[ DHCPdebug? ] [if]
cr ." IN: " GivenReply? @ .DHCPtype
cr ." Header"  *bootp #48 dump
cr ." Options"  *bootp dh.options #64 dump
[then]
        DHCPtimer @				\ if timeout set ; SFP004
        if  ticks 1- 1 or DHCPtimer !  then	\ make timeout now ; SFP004
      then
    then
[ DHCPdebug? ] [if]
else
cr ." IN: dumped"
[then]
  then
  *pb FreeQB				\ give it back
;
+short-branches


: DHCPpacket?	\ *udp -- flag
\ *G Given a UDP header, return non-zero if the packet is a
\ ** DHCP packet. Used in the UDP layer to route DHCP packets.
  dup uh_sport w@(n) IPPORT_BOOTPSERVER =	\ check source port number
  swap uh_dport w@(n) IPPORT_BOOTPCLIENT =	\ check destination port number
  and
;


\ ***********************
\ *S State machine initialisation and startup
\ ***********************
\ *[
  ' doDHidle       smDHidle      setDHstate
  ' doDHinit       smDHinit      setDHstate
  ' doDHselect     smDHselect    setDHstate
  ' doDHrequest    smDHrequest   setDHstate
  ' doDHarpcheck   smDHarpcheck  setDHstate
  ' doDHbound      smDHbound     setDHstate
  ' doDHrenew      smDHrenew     setDHstate
  ' doDHrebind     smDHrebind    setDHstate
  ' doDHrebooting  smDHrebooting setDHstate
\ *]

#68 equ DHCPport	\ -- port
\ *G DHCP standard port.

create DHCPServiceStruct	\ -- addr
\ *G Holds the UDP Service info to run the DHCP code on receipt
\ ** of DHCP packets, or idle on timeout. Note that this *\b{must}
\ ** match the *\fo{/UDPservice} structure in *\i{UDP.fth}.
  UDPServiceChain link,  DHCPport ,  ' RxDHCP ,  ' DHCPidle ,

: runDHCP	\ --
\ *G Start the DHCP process and wait until a valid IP address has
\ ** been set.
[ diags? ] [if]  cr ." DHCP started"  [then]
  goInit
\  WaitForIPaddress			\
\ [ diags? ] [if]  ." done."  [then]
;


\ ======
\ *> ###
\ ======

decimal
