\ DebugTools.fth - debug tools for the PowerNet stack

((
Copyright (c) 1996-2003, 2007, 2009, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999
and:
  CEM Systems Ltd, Belfast, 2000-2002
  Hiden Analytical, 2002

To do
=====

Change history
==============
20100512 MPE003 Added DICTTOP
20090519 MPE002 Added [CON .CONLINE RDEPTH and .RS
20071024 MPE001 Replaced ASCII with [CHAR] and CHAR.
))


\ =============
\ *! debugtools
\ *T Debugging tools
\ =============

decimal


\ ****************
\ *S Miscellaneous
\ ****************

: ?dump		\ addr len --
\ *G Now redundant, but does nothing for "addr 0 DUMP".
  1 max dump
;

[undefined] .decimal [if]
: .decimal      \ n --
\ *G display a value in decimal.
  base @ >r  decimal .  r> base !
;
[then]

: .hex          \ n --
\ *G display a value in hexadecimal.
  base @ >r  hex u.  r> base !
;


\ ***************************
\ *S Using the system console
\ ***************************

[undefined] consoleIO [if]
: consoleIO	\ --
\ *G Select debug console for output. By default this
\ ** is the CONSOLE device.
  console opvec !  console ipvec !
;
[then]

: [con		\ -- ; R: -- ipvec opvec
\ *G Save the current I/O devices on the return stack,
\ ** and set I/O to the console. Restore the previous
\ ** state with *\fo{io]}.
\ *C   [con ... io]
  r>  ipvec @ >r opvec @ >r  >r
  console dup opvec ! ipvec !
;

: con]		\ -- ; R: ipvec opvec --
\ *G Restore the I/O devices from the return stack.
  r>  r> opvec ! r> ipvec !  >r  ;

: .ConLine	\ caddr len --
\ *G Display text with leading CR on Forth console.
  [con
    cr ." Socket " my_hsocket @ . ." :"  type  ." :"
  io]
;

variable debug_out

: debug_emit    \ char --
\ *G As EMIT but always uses the CONSOLE device.
  opvec @ >r  ipvec @ >r  ConsoleIO
  emit  1 debug_out +!
  r> ipvec !  r> opvec !
;

: debug_cr      \ --
\ *G As CR but always uses the CONSOLE device.
  opvec @ >r  ipvec @ >r  ConsoleIO
  cr  debug_out off
  r> ipvec !  r> opvec !
;

: debug_space	\ --
\ *G As SPACE but always uses the CONSOLE device.
  #32 debug_emit
;

: debug_spaces  \ n --
\ *G As SPACES but always uses the CONSOLE device.
  0 ?do  debug_space  loop
;

: debug_>pos    \ n --
\ *G Position the console output to column n.
  out @ - 1 max debug_spaces
;

: debug_type    \ addr len --
\ *G As TYPE but always uses the CONSOLE device.
  bounds ?do i c@ debug_emit loop
;

: debug_count&type      \ c-addr --
\ *G As COUNT TYPE but always uses the CONSOLE device.
  count debug_type
;

: debug_.       \ n --
\ *G As . but always uses the CONSOLE device.
  s>d <# #S #> debug_type
  #32 debug_emit
;

: debug_.hex    \ n --
\ *G As .HEX but always uses the CONSOLE device.
  base @ >r
  hex [char] $ debug_emit debug_.
  r> base !
;

: debug_.decimal        \ n --
\ *G As .DECIMAL but always uses the CONSOLE device.
  base @ >r
  decimal [char] # debug_emit debug_.
  r> base !
;

: debug_.lword  \ n --
\ *G As .LWORD but always uses the CONSOLE device.
  base @ >r
  hex s>d <# # # # # [char] : hold # # # # #> debug_type
  r> base !
;

: debug_.byte   \ n --
\ *G As .BYTE but always uses the CONSOLE device.
  base @ >r
  hex s>d <# # # #> debug_type
  r> base !
;

: debug_.ASCII  \ char --
\ *G As .ASCII but always uses the CONSOLE device.
  $7F and  dup $20 $7E within? 0=
  if  drop $2E  then
  debug_emit
;

: debug_.s      \ --
\ *G As .S but always uses the CONSOLE device.
  debug_cr  depth ?dup if
    S"      top" debug_type
    dup 0< 1 ?throw  0 do
        debug_cr
        i pick
        dup debug_.decimal #16 debug_>pos debug_.hex
    loop
  else
    S" empty stack" debug_type
  then
  debug_cr
;

: debug_dump	\ addr len --
\ *G As DUMP but always uses the CONSOLE device.
  bounds ?do						\ get limits
    debug_cr i debug_.lword  debug_space   i 8 + i      \ print address
    do  debug_space i c@ debug_.byte  loop              \ first eight bytes
    debug_space  i $010 + i 8 +                         \ gap
    do  debug_space i c@ debug_.byte  loop              \ second eight bytes
    2 debug_spaces  i $010 + i                          \ gap
    do  i c@ debug_.ascii  loop                         \ ascii o/p
  $010 +loop                                            \ until done
  debug_cr
;


\ *****************
\ *S Stack checking
\ *****************

decimal

: ?StackEmpty	\ --
\ *G If the stack depth is non-zero, issue a console
\ ** warning message and clear the stack. This is often
\ ** used at the end of a task loop.
  depth 0= if  exit  endif
  opvec @ >r  ipvec @ >r  ConsoleIO
  cr ." *** Stack fault: depth = " depth .decimal ." (d) "
  ." in task " self .task
  s0 @ sp!
  cr ."   Stack cleared."
  r> ipvec !  r> opvec !
;

[undefined] DictTop [if]
compiler
: DictTop  here  ;
\ *G Used by some debug tools as the highest permitted address in
\ ** the dictionary. If not previously defined, *\fo{HERE} is used.
target
[then]

[undefined] ip>nfa [if]
: name?         \ addr -- flag                                   MPE.0000
\ *G Check to see if the supplied address is a valid NFA,
\ ** returning true if the address appears to be a valid NFA.
\ ** This word is implementation dependent. For MPE cross compilers,
\ ** a valid NFA for MPE embedded systems satisfies the following:
\ *(
\ *B All characters within string are printable ASCII within range 33..126
\ *B String Length is non-zero in range 1..31 and bit 7 is set, ignore bits 6, 5
\ *)
  count					\ c-addr u --
  dup  $BF and  $81 $9F within? 0=	\ NFA first byte = 1SIxxxxx, count = xxxxx
  					\ mask           = 10111111
  if  2drop  0  exit  then
  $01F and bounds ?do
    i c@ #33 #126 within? 0=		\ check all ascii chars
    if  unloop  FALSE exit  then
  loop
  TRUE
;

: ip>nfa	\ addr -- nfa
\ *G Attempt to move backwards from an address within a definition
\ ** to the relevant NFA.
  2-				\ NFA must be at least 'n' bytes backwards
  begin
    dup name? 0=
   while
    1-
  repeat
;
[then]

: SF{		\ n -- ; R: -- depth
\ *G You can check for stack faults with:
\ *C  n SF{ .... }SF
\ *P where *\i{n} describes the stack change between SF{ and }SF.
\ ** If the stack change is different, an error message is generated.
\ ** This word will work on most systems in which the return address
\ ** is held on the return stack.
  r>  swap depth 2- + >r  >r
;

: }SF		\ -- ; R: depth -- ; perform stack check
\ *G The end of an SF{ ... }SF structure. This word is
\ ** not strictly portable as it assumes that the Forth
\ ** return stack holds a valid return address. In the
\ ** vast majority of cases the assumption is true, but
\ ** beware of some 8051 implementations.
\ ** See SF{
  r>
  r> depth 2- <> if
    cr ." Stack assertion failed at " dup .dword
    ."  in " dup ip>nfa .name
  endif
  >r
;

: rdepth	\ -- n
\ *G Return the number of items on the return stack.
  r0 @ rp@ - 4- 4/
;

[-short-branches
: .rs           \ --
\ *G Display to the console the current contents of the return
\ ** stack. Where possible a word name is also displayed with
\ ** the data value.
  cr  rdepth dup 0= if			\ get depth of return stack
    drop ." Return stack empty" exit
  endif
  dup 0< if
    drop ." return stack underflow" exit
  endif
  ." *** top: " dup . ." items"
  0 do
    i 3 + rpick				\ DO LOOP uses 3 items
    cr ."  #" dup 8 u.r			\ display stack item as number
    dup $0400 u>			\ if not a small number
    over DictTop u< and if		\ and not beyond the dictionary
      2 spaces ip>nfa .name		\ show word name
    else
      drop
    endif
    halt? ?leave
  loop
;
short-branches]


\ *************
\ *S Cold Chain
\ *************

: showExecChain	\ item --
\ *G Show the contents of a chain, given an item in the chain.
  cr ." xt        probably"
  begin
    dup
   while
    dup cell + @
    cr dup .dword space ip>nfa .name	\ display xt and name
    @					\ get next entry
  repeat
  drop
;

: showColdChain	\ --
\ *G Display the cold chain.
  ColdChainFirst showExecChain  ;


\ ======
\ *> ###
\ ======

