\ ETHTASK.FTH - definitions for a Forth Ethernet Interface

((
Copyright (c) 1997-2005, 2008
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999
and:
  CEM Systems Ltd, Belfast, 2000-2002

To do
=====

Change history
==============
20081124 SFP005 Added Ethernet link check.
		Added Ethernet link up/down chains.
20080318 SFP004 Added packet discard option to simulate lost packets
		on transmission.
20080307 SFP003 Changed FLUSH to FLUSHDEBUG to avoid name conflict.
20050216 SFP002 Run INITETHER in RUNETHERTASK
20030312 SFP001 Incorporated CEM changes.
19990705 SJC000 Added call to DISCARDRX when there is no heap space to store
                incoming ether packet. The driver needs to be called to
                release the frame from driver/chip memory.
))


\ ==========
\ *! ethtask
\ *T Ethernet processing task
\ ==========
\ *P The Ethernet ports are handled by a task which despatches
\ ** *\fo{PBUF}s to and from other layers of the stack.

only forth definitions

0 equ LoseTX?	\ -- u
\ *G When this *\fo{EQU} is set non-zero, every *\i{u}th transmit
\ ** packet will be discarded. This equate is used when testing
\ ** the TCP retry mechanism and for simulating lossy links.
0 equ LoseRX?	\ -- u
\ *G When this *\fo{EQU} is set non-zero, every *\i{u}th receive
\ ** packet will be discarded. This equate is used when testing
\ ** the TCP retry mechanism and for simulating lossy links.

\ #29 equ LoseTX?	\ -- u
\ #31 equ LoseRX?	\ -- u


\ ***************************
\ *S Ethernet packet handlers
\ ***************************

(( \ debug code for testing
\ the following hold a pair of ethernet addresses
\ as a temporary measure to test out the interface.

variable etherTo    2 allot-ram
  etherTo 6 erase
variable etherFrom  2 allot-ram
  etherFrom 6 erase

#1550 Buffer: Ether_RxBuff
#1550 Buffer: Ether_TxBuff
))

XC? [if]
: Send>Ether	\ *pb *rt -- len
\ *G Send the data in the supplied PBuf (chain).
\ *D params	*pb is a pbuf set to the IP layer
\ **		*rt points to the required routing table entry
\ *D returns 	#bytes sent if OK else SOCKET_ERROR
[else]
: rtSend>Ether	\ *pb *rt -- len
[then]
  iproute_ethadd 			\ this will form the destination
  over p_dat ether_dest copyEthAdd	\ move it in
  EtherAddress 				\ and my address as the source
  over p_dat ether_src copyEthAdd	\ move it in
  ETHERTYPE_IP
  over p_dat ether_type w!(n)		\ set ether type
  ether_hdr over unbumpPBuffData	\ add ethernet header length
  dup PBDataLen over p_len w!
  dup TxEnetQ enqueue			\ *pbuf's are queued
  p_len w@				\ return length
;
XC? 0= [if]  assign rtSend>Ether to-do Send>Ether  [then]

: EtherIpPacket	\ *pb --
\ *G Process a PBUF which holds a received Ethernet IP packet.
\ cr ." In EtherIpPacket"
\ dup p_dat  over p_len w@ #64 min dump
  dup p_len w@ ether_hdr -		\ get length - ehdr
  over p_datalen w!			\ and set data len
  pbuf_hdr ether_hdr +			\ point past ethernet header
  over p_pkthdr w!			\ and set as the data pointer
  dup PBHdrStart			\ -- *pb *ip ; start of ip packet
  dup ip_dest @(n) IPAddress @(n) = if	\ directly addressed to me? ; SFP001
    over p_dat ether_src		\ pointer to Ethernet address in PBUF
    swap ip_src @(n) AddEtherRoute	\ add IP source address to route tables
  else
    drop				\ discard *IP
  endif
  ether_hdr over p_data w+!		\ bump start of data
  RxIpPacket				\ go handle the ip packet
;

: RxEtherPacket		\ *pbuf --
\ *G The PBUF holds a received Ethernet packet which is
\ ** processed.
  case  dup p_dat ether_type w@(n)	\ get packet type
    ETHERTYPE_IP of  EtherIpPacket  endof
    ETHERTYPE_ADDRESS_RESOLUTION of  EtherArpPacket  endof
      swap FreeQB			\ unknown, discard it
  endcase
;

: RxEtherPkt	\ -- ; SFP001
\ *G Get waiting packet from the Ethernet driver layer.
  GetPbuf ?dup if			\ grab a buffer
    dup p_dat PSIZE get_ether_pkt	\ go read the packet, 0=bad or none
    dup #60 < if			\ check for runt packets, etherhdr+data+FCS<64
      drop FreeQB
    else
      over p_len w!			\ set buffer length
      RxEtherPacket			\ go handle the incoming packet
    endif
  else					\ memory allocate failed
[ diags? ] [if]  cr ." No Ether Rx Buffer space!" [then]
    DiscardRx				\ SJC000
  endif
;


\ *************************
\ *S Link failure detection
\ *************************

0 value Linked?	\ --
\ *G Returns true if PowerNet is linked to the world by an active
\ ** connection.

create LinkUpChain	\ -- addr
\ *G Anchors the chain of words executed when the link is
\ ** established.
\ ** This chain must be defined at compile time.
  0 ,
create LinkDnChain	\ -- addr
\ *G Anchors the chain of words executed when the link fails.
\ ** This chain must be defined at compile time.
  0 ,

interpreter
: AtLinkUp	\ xt --
\ *G Add the word whose *\i{xt} is given to link up chain.
\ ** *\fo{INTERPRETER} word.
  align  LinkUpChain link, ,  ;
: AtLinkDn	\ xt --
\ *G Add the word whose *\i{xt} is given to link up chain.
\ ** *\fo{INTERPRETER} word.
  align  LinkDnChain link, ,  ;
target

useDHCP? [if]  ' runDHCP AtLinkUp  [then]
\ *G Start DHCP when the link comes up.
useSNTP? [if]  ' runSNTP AtLinkUp  [then]
\ *G Start SNTP when the link comes up.


: checkLink	\ -- flag
\ *G Check PowerNet's connection and set *\fo{Linked?}. Returns
\ ** true if PowerNet is linked to the world by an active
\ ** connection.
  EtherLink? dup -> Linked?
;

variable CheckTime	\ -- addr
\ *G Holds the time at which the Ethernet link should next be checked.

: nextCheck	\ --
\ *G Set the next link check time. This is normally every 200 ms.
  #200 later 1 or CheckTime !
;

: CheckEther	\ --
\ *G The Ethernet link is checked every so often for an
\ ** established link. If the link fails, the Ethernet
\ ** task halts until the link is re-established. When it
\ ** is re-established a user-extensible chain of actions
\ ** takes place. This permits the system to restart actions
\ ** such as DHCP and SNTP.
  CheckTime @ timedout? if
    checkLink 0= if
[ diags? ] [if] cr ." Link down" [then]
      LinkDnChain execChain
      begin
        5000 ms initEther  5000 ms checkLink
      until
[ diags? ] [if] cr ." Link up" [then]
      LinkUpChain execChain
    endif
    nextCheck
  endif
;


\ ****************
\ *S Ethernet task
\ ****************

LoseTX? [if]
variable #TXcount
[then]
LoseRX? [if]
variable #RXcount
[then]

: DoRunEther	\ --
\ *G The task action that handles all general background Ethernet
\ ** processing.
  ConsoleIO decimal			\ default debug I/O device
  cr ." Ethernet Task running"  100 ms
\ SFP005 ...
  InitEther  #5000 ms  checkLink	\ start Ethernet
  if  LinkUpChain execChain  endif
  nextCheck
\ ...SFP005

[ slip? ] [if]  ETHER_PORT portid !  [then]
  begin
[ diags? ] [if]  ?StackEmpty  [then]	\ check stack depth
    isRx? if				\ Ethernet packet received?
[ LoseRX? ] [if]
      #RXcount incr  #RXcount @ LoseRX? >=
      if  #RXcount off  DiscardRX  else  RxEtherPkt  endif
[else]
      RxEtherPkt			\ go read the packet in
[then]
    endif
    TxEnetQ dequeue ?dup if		\ *pbuf's are queued
[ LoseTX? ] [if]
      #TXcount incr  #TXcount @ LoseTX? >=
      if  #TXcount off  else
[then]
      dup p_dat				\ packet start addr
      over p_len w@			\ packet length
\ #10 ms cr ." EtherOut# " dup . over #64 dump
      send_ether_pkt 	 		\ and send it
\ cr ." Ether sent"
[ LoseTX? ] [if]
      endif
[then]
      FreeQB
    endif
    pause  flushDebug  checkEther	\ SFP005
  again
;

task EtherTask	\ -- addr
\ *G The Ethernet handler task.
cdata

: RunEtherTask	\ --
\ *G Launch the Ethernet support task.
\  InitEther				\ SFP002 ; SFP005
  ['] DoRunEther EtherTask initiate	\ assign action and start task
;


\ **********
\ *S Routing
\ **********

((
#2000	equ ROUTE_SEARCH_TIME	\ max time allowed for route searches
3	equ ARP_REQ_TRIES
))

: (AskForEtherRoute)	\ ip -- addr|0
\ *G See if we can get a route using ARP.
  ROUTE_SEARCH_TIME later >r		\ start a timer so we don't stay here too long
  0 					\ initial route
  over SendArpRequest
  begin					\ -- *ip rt
    pause				\ allow other tasks to still run
    drop dup (find_route) dup 0=	\ -- *ip rt' ; still none?
   while
    r@ timedout?
  until then
  r> drop				\ discard timer target
  nip					\ finally return the route if we got one
;

XC? [if]
: AskForEtherRoute	\ ip -- addr|0
\ *G See if we can get a route using ARP, but retrying several
\ ** times.
[else]
: rtAskForEtherRoute	\ ip -- addr|0
[then]
  0
  ARP_REQ_TRIES 0 ?do			\ loop for retries
    drop
    dup (AskForEtherRoute) dup ?leave	\ go do ARP, leave if route good
  loop
  nip					\ return the result addr|0
;
XC? 0= [if] assign rtAskForEtherRoute to-do AskForEtherRoute  [then]


\ ======
\ *> ###
\ ======

decimal


