\ ClientTools.fth - Tools for writing client code

((
Copyright (c) 2014, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

To do
=====

Change history
==============
20160104 SFP002 Converted to a toolbox.
20140131 SFP001 First cut.
))

decimal

\ ==============
\ *! clienttools
\ ==============
\ *P The code in *\i{<pnet>\Examples\ClientTools.fth} implements
\ ** code to access a server, to manage the connection and to
\ ** read/write. Most of this code assumes that a a valid socket
\ ** handle is in the *\fo{USER} variable *\fo{my_hsocket}.


\ **************
\ *S Connections
\ **************

: initConn	\ ipaddr port# -- socket|0
\ *G Open a connection to the poll server, returning the socket
\ ** number or zero on error. The return value is also placed in
\ ** *\fo{MY_HSOCKET}. A maximum of five attempts are made to
\ ** connect.
  5 0 do				\ -- ipaddr port#
    2dup 0 swap TCPconnect 0= if	\ -- ipaddr port# hsock
      nip nip				\ -- hsock
      dup my_hsocket !  unloop exit
    then
    drop  #2000 ms
  loop
  2drop 0
;

: termConn	\ --
\ *G Close the poll connection. The socket handle is in *\fo{MY_HSOCKET}.
  my_Hsocket @ disconnect drop
;


\ **************
\ *S Generic I/O
\ **************
\ *P We define a generic I/O device to handle most client data
\ ** transfers.

: skKey?	\ -- n
\ *G Return non-zero if the socket has data waiting to be read.
  my_HSocket @ pollSocket 0 max
;

: skKey		\ -- byte
\ *G Read a single character from the socket. -1 is returned on
\ ** error.
  {: | buff[ cell ] -- :}
  begin  skKey? 0= while  pause repeat
  my_hsocket @ buff[ 1 0 recv 1 =
  if  buff[ c@  else  -1  then
;

: skType	\ caddr len --
\ *G Send a block through the socket.
  my_hsocket @ -rot 0 send drop
;

: skEmit	\ char --
\ *G Send a single character through the socket.
  {: | buff[ cell ] -- :}
  buff[ c!  buff[ 1 skType
;

l: skCRchars	\ -- addr
  #13 c,  #10 c,

: skCr		\ --
\ *G Send a CR/LF pair through the socket.
  skCRchars 2 skType
;

create SocketConn	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device structure for sockets. The socket handle
\ ** is in the *\fo{USER} variable *\fo{my_hsocket}. As a result
\ ** we only need one instance of this device.
  ' skKey ,		\ -- char ; receive char
  ' skKey? ,		\ -- flag ; check receive char
  ' skEmit ,		\ -- char ; display char
  ' skType ,		\ caddr len -- ; display string
  ' skCr ,		\ -- ; display new line
  
\ *P You make this the current actions of *\fo{KEY} and *\fo{EMIT} 
\ ** and friends by setting *\fo{IPVEC} and *\fo{OPVEC} as required,
\ ** e.g. with
\ *C   SocketConn setConsole


\ *************
\ *S More tools
\ *************

: checkSock	\ hs -- ior
\ *G The *\i{ior} is returned non-zero if the socket is invalid
\ ** or not in TCPS_ESTABLISHED state.
  get_socket_addr dup 0=		\ socket is valid and has struct or not
  if  drop -3 exit  endif		\ error if 0
  so_*cb @ dup 0=			\ check for a valid control block
  if  drop -2 exit  endif		\ error if 0
  t_state w@ TCPS_ESTABLISHED <>	\ get tcp STATE
;

: checkConn	\ -- ior
\ *G The *\i{ior} is returned non-zero if the socket is invalid
\ ** or not in TCPS_ESTABLISHED state.
  my_hsocket @ checkSock
;

: waitData	\ ms -- flag ; 0=success
\ *G Wait for a response (socket has input data) for the given
\ ** period. Return true if the socket times out.
  later begin
    skKey? 0=
   while
    dup expired
  until then
  timedout?
;

: ReadData	\ caddr len -- ior
\ *G Read *\i{len} bytes from *\fo{my_hsocket} into the buffer
\ ** *\i{caddr}. The result *\ior} is 0 on success.
  begin
    dup
   while
    2dup
    my_hsocket @ -rot #1024 min 0 recv	\ -- caddr len len'
    dup SOCKET_ERROR =
    if  drop 2drop  -1 exit  then
    /string
  repeat
  nip
;

: WriteData	\ caddr len -- ior
\ *G Write *\i{len} bytes to *\fo{my_hsocket} into the buffer
\ ** *\i{caddr}. The result *\ior} is 0 on success.
  my_hsocket @ -rot 0 send SOCKET_ERROR =
;


\ ************
\ *S Test code
\ ************

0 [if]

variable escs
2 value ESC_LIMIT

: tterm \ --;
\ *G Simple terminal to an open TCP channel, ESC_LIMIT times ESC to leave.
\ ** ESC_LIMIT >1 allow vt-100 escape sequences to pass through (cursor, etc.) without leaving.
  begin
    my_hsocket @ 0= if
      ." no valid socket" exit
    then
    key? if
      key dup #27 = if
        1 escs +!
        escs @ ESC_LIMIT > if
          drop exit
        then
      else
        escs off
      then
      skemit
    then
    pause
    skkey? if
      skkey emit
    then
    pause
  again
;

[then]


\ Example how to start a connection on a Telnet or an other simple TCP connection
\
\ 192   8 lshift \ MOXA box setup, internal Synics
\ 168 + 8 lshift
\ 0   + 8 lshift
\ 253 +          \ init for 192.168.0.53, port 23
\ 23  initconn .
\
\ tterm          \ the terminal is now started, 2 times ESC to leave it
\
\ 192   8 lshift \ MOXA box terminal, internal Synics
\ 168 + 8 lshift
\ 0   + 8 lshift
\ 253 +
\ 4001  initconn
\
\ tterm          \ the same as above for serial link over MOXA box

\ ******
\ *> ###
\ ******

