\ ClientTools.fth - Tools for writing client code

((
Copyright (c) 2014, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

tterm and telnet negociation state machine:

Copyright (c) 2016
Synics AG, CH-8105 Regensdorf

tel: +41 (0)44 840 40 60
net: info@synics.ch
web: www.synics.ch


To do
=====

Change history
==============

20160111 kh002  Minimal Telnet negociation added to connect to a Debian Wheezy telnet server
20160106 kh001  tterm as dump TCP terminal added
20160104 SFP002 Converted to a toolbox.
20140131 SFP001 First cut.
))

decimal

\ ==============
\ *! clienttools
\ ==============
\ *P The code in *\i{<pnet>\Examples\ClientTools.fth} implements
\ ** code to access a server, to manage the connection and to
\ ** read/write. Most of this code assumes that a a valid socket
\ ** handle is in the *\fo{USER} variable *\fo{my_hsocket}.


\ **************
\ *S Connections
\ **************

#2000 value INITCONN_DELAY
\ *G Time delay before retrying

: initConn	\ ipaddr port# -- socket|0
\ *G Open a connection to the poll server, returning the socket
\ ** number or zero on error. The return value is also placed in
\ ** *\fo{MY_HSOCKET}. A maximum of five attempts are made to
\ ** connect.
  5 0 do				\ -- ipaddr port#
    2dup 0 swap TCPconnect 0= if	\ -- ipaddr port# hsock
      nip nip				\ -- hsock
      dup isMySocket unloop exit
    then
    drop  INITCONN_DELAY ms
  loop
  2drop 0
;

: termConn	\ --
\ *G Close the poll connection. The socket handle is in *\fo{MY_HSOCKET}.
  my_Hsocket @ disconnect drop
;


\ **************
\ *S Generic I/O
\ **************
\ *P We define a generic I/O device to handle most client data
\ ** transfers.

: skKey?	\ -- n
\ *G Return non-zero if the socket has data waiting to be read.
  my_HSocket @ pollSocket 0 max
;

: skKey		\ -- byte
\ *G Read a single character from the socket. -1 is returned on
\ ** error.
  { | buff[ cell ] -- }
  begin  skKey? 0= while  pause repeat
  my_hsocket @ buff[ 1 0 recv 1 =
  if  buff[ c@  else  -1  then
;

: skType	\ caddr len --
\ *G Send a block through the socket.
  my_hsocket @ -rot 0 send drop
;

: skEmit	\ char --
\ *G Send a single character through the socket.
  { | buff[ cell ] -- }
  buff[ c!  buff[ 1 skType
;

l: skCRchars	\ -- addr
  #13 c,  #10 c,

: skCr		\ --
\ *G Send a CR/LF pair through the socket.
  skCRchars 2 skType
;

create SocketConn	\ -- addr ; OUT managed by upper driver
\ *G Generic I/O device structure for sockets. The socket handle
\ ** is in the *\fo{USER} variable *\fo{my_hsocket}. As a result
\ ** we only need one instance of this device.
  ' skKey ,		\ -- char ; receive char
  ' skKey? ,		\ -- flag ; check receive char
  ' skEmit ,		\ -- char ; display char
  ' skType ,		\ caddr len -- ; display string
  ' skCr ,		\ -- ; display new line


\ *************
\ *S More tools
\ *************

: checkSock	\ hs -- ior
\ *G The *\i{ior} is returned non-zero if the socket is invalid
\ ** or not in TCPS_ESTABLISHED state.
  get_socket_addr dup 0=		\ socket is valid and has struct or not
  if  drop -3 exit  endif		\ error if 0
  so_*cb @ dup 0=			\ check for a valid control block
  if  drop -2 exit  endif		\ error if 0
  t_state w@ TCPS_ESTABLISHED <>	\ get tcp STATE
;

: checkConn	\ -- ior
\ *G The *\i{ior} is returned non-zero if the socket is invalid
\ ** or not in TCPS_ESTABLISHED state.
  my_hsocket @ checkSock
;

: waitData	\ ms -- flag ; 0=success
\ *G Wait for a response (socket has input data) for the given
\ ** period. Return true if the socket times out.
  later begin
    skKey? 0=
   while
    dup expired
  until then
  timedout?
;

: ReadData	\ caddr len -- ior
\ *G Read *\i{len} bytes from *\fo{my_hsocket} into the buffer
\ ** *\i{caddr}. The result *\ior} is 0 on success.
  begin
    dup
   while
    2dup
    my_hsocket @ -rot #1024 min 0 recv	\ -- caddr len len'
    dup SOCKET_ERROR =
    if  drop 2drop  -1 exit  then
    /string
  repeat
  nip
;

: WriteData	\ caddr len -- ior
\ *G Write *\i{len} bytes to *\fo{my_hsocket} into the buffer
\ ** *\i{caddr}. The result *\ior} is 0 on success.
  my_hsocket @ -rot 0 send SOCKET_ERROR =
;

\ ------ testing words ------

decimal

\ Client negociation state machine
\ --------------------------------
\ The basic Telnet negociation is specified by RFC-854, a more elaborated implementation help may be found in RFC-1143
\ This very basic Telnet Client implementation simply refuses all WILL xxx tentatives by answering DON'T xxx
\ This behavior is foreseen as the very minimal requirement
\ A minimal basic negaciation is requested for a Telnet Client for example to a Raspberry Pi (Raspian), as it otherwise
\ don't start at all (no "login:" proposed)

variable tclient_option_val

240 constant TELNET_SE        \ Sub negociation field End
250 constant TELNET_SB        \ Sub negociation field Begin
251 constant TELNET_WILL
252 constant TELNET_WONT
253 constant TELNET_DO
254 constant TELNET_DONT
255 constant TELNET_IAC

variable escs                 \ ESC counter to exit TTERM
2 value ESC_LIMIT

[undefined] append_char [if]

: append_char \ adr n char -- adr n+1;
  over 3 pick + c! 1+
;
[then]

[undefined] dump_raw [if]

: dump_raw \ addr n --;
\ *G Hex dump small fields without any formatting
  base @ >r hex
  bounds ?do
    i c@ 0 <# # # bl hold #> type
  loop
  r> base ! ;

[then}

[undefined] _ESC [if] 27 equ _ESC [then]

: do_tclient_option \ option --;
\ *G Client side option handling
  dup cr ." WILL option: " .
  dup case
    1 of                            \ accept echo=1 proposal
      pad 0
      TELNET_IAC  append_char
      TELNET_WILL append_char
      rot append_char               \ accept proposed option
    endof
\    31 of                          \ accept windows size proposal
\      pad 0
\      TELNET_IAC  append_char
\      TELNET_WILL append_char
\      rot append_char              \ accept proposed option
\    endof
    pad 0                           \ refuses all other options
    TELNET_IAC  append_char
    TELNET_WONT append_char
    rot append_char
    rot
  endcase
\  2dup cr dump_raw                 \ debug
  skType
;

: will_tclient_option \ optiom --;
\ *G Client side option handling
  dup cr ." DO option: " .
  pad 0
  TELNET_IAC  append_char
  TELNET_DONT append_char
  rot append_char          \ refuses proposed option
\  2dup cr dump_raw
  skType
;

: tclient_unknown \ byte --;
  cr ." Telnet client command is not implemented: " .
;

variable sub_fsize         \ negociation data size
\ *G Dynamic size of a IAC SB ........... IAC SE sub-negociation byte field

10 constant MAX_SUB_SIZE
MAX_SUB_SIZE buffer: sub_field
\ *G Sub-negociation data field

: init_tsub \ n --;
  drop
  0 sub_fsize !
;

: append_subfield \ n --;
\ *G Append sub negociation byte to SUB_FIELD
  sub_field sub_fsize @ MAX_SUB_SIZE 1- min + c!  \ avoid overflow
  1 sub_fsize +!           \ keep the real size, also if overflowed
;

: close_subfield  \ n --;
  drop
  cr ." Sub nogaciat. : "
  sub_field sub_fsize @
  dump_raw                 \ display the content of the proposed subnegociation
                           \ the sub-negociation effects should be place here
;

\ ---- state machine declarations ----------

1  dup equ tclient_data     \ default data transmission state
1+ dup equ tclient_iac
1+ dup equ tclient_swill    \ WILL from server
1+ dup equ tclient_sdo      \ DO from server
1+ dup equ tclient_ssub     \ sub negociation field
1+ dup equ tclient_iac_ssub \ sub negociation field

fsm: tclient_handler

  tclient_data state:
    TELNET_IAC ' = tclient_iac    action: drop
    tclient_data                 default: emit

  tclient_iac state:
    TELNET_DO ' =  tclient_swill  action: drop
    TELNET_WILL ' = tclient_sdo   action: drop
    TELNET_SB ' =  tclient_ssub   action: init_tsub
                                                              \ to be fully complete, a DONT and a WONT action state should be added here
    tclient_data                 default: tclient_unknown     \ unkown or not implemented function

  tclient_swill state:
    tclient_data default: do_tclient_option

  tclient_sdo state:
    tclient_data default: will_tclient_option

  tclient_ssub state:
    TELNET_IAC ' = tclient_iac_ssub action: drop
    tclient_ssub                   default: append_subfield   \

  tclient_iac_ssub state:
    TELNET_SE ' = tclient_data      action: close_subfield
    TELNET_IAC ' = tclient_iac_ssub action: append_subfield   \ IAC as data
    tclient_data                   default: tclient_unknown   \ unkown function

: tterm \ --;
\ *G Simple terminal to an open TCP channel, ESC_LIMIT times ESC to leave.
\ ** ESC_LIMIT >1 allow vt-100 escape sequences to pass through (cursor, etc.) without leaving.
  tclient_data tclient_handler set-state                       \ initiate state-machine
  begin
    my_hsocket @ 0= if
      ." no valid socket" exit
    then
    key? if
      key dup _ESC = if               \ the ESC sequense is not solved within the state machine
        1 escs +!
        escs @ ESC_LIMIT > if         \ check for escaping
          drop exit
        then
      else
        escs off
      then
      skemit
    then
    pause
    skkey? if
      skkey
\      emit
      tclient_handler cycle
    then
    pause
  again
;

\ Example how to start a connection on a Telnet or an other simple TCP connection
\ -------------------------------------------------------------------------------
\
\ \ The server device is a MOXA box making a terminal over IP, its service port is 23, the data port is 4001
\ \ The MOXA box seems no needing any negociation
\
\ 192   8 lshift \ MOXA box setup, internal Synics
\ 168 + 8 lshift
\ 0   + 8 lshift
\ 253 +          \ init for 192.168.0.253, port 23
\ 23  initconn . \ open the connection
\ tterm          \ the terminal is now started, 3 times ESC to leave it
\
\
\ 192   8 lshift \ MOXA box terminal, internal Synics
\ 168 + 8 lshift
\ 0   + 8 lshift
\ 253 +
\ 4001  initconn \ init for 192.168.0.253, port 4001
\ tterm          \ the same as above for serial link over MOXA box

\ ******
\ *> ###
\ ******

