\ TargetClient.fth - client code for the target server

((
Copyright (c) 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

To do
=====

Change history
==============
20140131 SFP001 First cut.
))

decimal

\ ===============
\ *! targetclient
\ ===============
\ *P The code in *\i{<pnet>\Examples\TargetServer\TargetServer.fth}
\ ** implements a server used to manage embedded systems. The primary
\ ** purpose is to allow embedded targets to obtain software updates.
\ ** The command sequence is text-based and a single socket is used for
\ ** commands, data and status responses. The code is written so that
\ ** it can be used as a model by code in other languages. It requires
\ ** little more than a Winsock/BSD socket API.

\ *P The code in *\i{<pnet>\Examples\TargetServer\TargetClient.fth}
\ ** implements code to access the server, to check for a new binary
\ ** image, to download a new image and to copy this from RAM to
\ ** Flash. The client code is compiled with PowerNet for an
\ ** embedded system.

\ *E Command sequence - bad command
\ **     HOST                 TARGET
\ **               <-         <cmd+params>
\ **     nack      ->
\ **               <-         <disconnect>

\ *E Command sequence - data to target
\ **     HOST                 TARGET
\ **               <-         <cmd+params>
\ **     ack len   ->
\ **               <-           ready
\ **     len data  ->
\ **               <-         <disconnect>

\ *E Command sequence - data to host
\ **     HOST                 TARGET
\ **               <-         <cmd+params+len>
\ **     ready     ->
\ **               <-           len data
\ **     status    ->
\ **               <-         <disconnect>


1 value TCdiags?	\ -- n
\ *G Set this non-zero to get diagnostic information about the
\ ** server.

compiler
: [tc		\ --
\ *G A compiler macro used to surround diagnostic code, and terminated
\ ** by *\fo{TC]}.
\ *E   [TC  ." debug message"  TC]
  postpone TCdiags?  postpone if
;
: tc]
\ *G Terminates a *\fo{[TC ... TC]} structure.
  postpone endif
;
target


\ ***************
\ *S Target tools
\ ***************

variable HostIPaddr	\ -- address
\ *G Holds the host IPaddress in network order
  HostIPaddr
  192 over 0 + c!
  168 over 1 + c!
  0   over 2 + c!
  200 over 3 + c!
  drop

2033 constant TargetPort#	\ --
\ G Port number used by the target server.

[undefined] crlf [if]
create crlf	\ -- addr
\ *G Holds a CR/LF pair.
  #13 c,  #10 c,
[then]

64 buffer: Cmd$	\ -- addr
\ *G Buffer in which to build commands. Counted string
128 buffer: Response$
\ *G Buffer for responses. Counted string.

: OpenTargetConn	\ -- socket|0
\ *G Open a connection to the poll server, returning the socket
\ ** number or zero on error. The return value is also placed in
\ ** *\fo{MY_HSOCKET}. A maximum of five attempts are made to
\ ** connect.
  5 0 do
    HostIPaddr @(n) 0 TargetPort# TCPconnect 0=
    if  dup my_hsocket !  unloop exit  then
    drop  #2000 ms
  loop
  0
;

: CloseConn	\ --
\ *G Close the poll connection. The socket handle is in *\fo{MY_HSOCKET}.
  my_Hsocket @ disconnect drop
;

: GotResponse?	\ -- n
\ *G Return non-zero if the socket has data waiting to be read.
  my_HSocket @ pollSocket 0 max
;

: waitResponse	\ -- flag ; 0=success
\ *G Wait for a response (socket has input data) for 2000 ms.
\ ** Return true if the socket times out.
  2000 later begin
    GotResponse? 0=
   while
    dup expired
  until then
  timedout?
;

: getResponse	\ --
\ *G Read the response.
Response$ off
  my_hsocket @ pollsocket
  0 max 127 min Response$ c!
  my_hsocket @ Response$ count 0 recv drop
cr ." Response=" Response$ dup c@ 1+ dump
;

: extractDec	\ caddr len -- caddr' len' u
\ *G Extract a number in decimal from the start of the
\ ** string, returning the remaining string starting at the first
\ ** non-numeric character and the converted number.
  base @ >r  decimal
  0 0 2swap >number 2swap drop
  r> base !
;

: -leading	\ caddr len -- caddr' len'
\ *G Remove leading blanks from the string.
  begin
    dup
   while
    over c@ bl =
   while
    1 /string
  repeat then
;

: ack+len?	\ -- len 0 | ior
\ *G Parse the response. It is either "ACK len" or NAK". For "ACK",
\ ** return len and ior=0, otherwise just return ior=true.
  response$ count dup 3 < if		\ -- caddr len
    2drop true exit
  then
  over s" ACK" is= if			\ process ACK
    3 /string -leading extractDec	\ get number
    nip nip 0  exit			\ discard string residue, ior=0
  then
  2drop true
;

: sendCmd	\ caddr len --
\ *G Send the text plus IP address to the server.
  Cmd$ place				\ command text
  s"  " Cmd$ append			\ add 1 space
  IPaddress @(n) >inet_ntoa Cmd$ append	\ IP address
  crlf 2 Cmd$ append			\ CR/LF
  my_hsocket @ Cmd$ count 0 send drop	\ send
\ cr ." Cmd sent:" Cmd$ count type
;

: sendReady	\ --
\ *G Send "READY\m" to the server.
  my_hsocket @ s\" READY\m" 0 send drop
;

: CmdResp	\ caddr len -- flag
\ *G Send the text plus IP address to the server. Wait for a
\ ** response. If there is a response within 2 seconds, return
\ ** *\i{flag=true}, otherwise return *\i{flag=false}.
  sendCmd waitResponse 0=
;

: ReadyResp	\ -- flag
\ *G Send "READY\m" to the server. Wait for a
\ ** response. If there is a response within 2 seconds, return
\ ** *\i{flag=true}, otherwise return *\i{flag=false}.
  sendReady waitResponse 0=
;

: ReadData	\ caddr len -- ior
\ *G Read *\i{len} bytes from *\fo{my_hsocket} into the buffer
\ ** *\i{caddr}. The result *\ior} is 0 on success.
  begin
    dup
   while
    2dup
    my_hsocket @ -rot #1024 min 0 recv	\ -- caddr len len'
    dup SOCKET_ERROR =
    if  drop 2drop  -1 exit  then
    /string
  repeat
  nip
;

: ReadProgram	\ len -- ior
\ *G Read a program image of length *\i{len} into the buffer
\ ** at *\fo{ImageBuffer}. The result *\ior} is 0 on success.
\ ** The first four cells of *\fo{ImageBuffer} are reserved.
\ ** The first cell contains the image length, the next three
\ ** are reserved for future use.
  ImageBuffer 4 cells erase
  dup ImageBuffer !
  ImageBuffer 4 cells + swap ReadData
;


\ ***********
\ *S User API
\ ***********

: NewProgFile?	\ -- flag
\ *G Return true if the server has a new program image file
\ ** for us.
  0					\ assume no file
  OpenTargetConn if	\ 0 on error
    s" QUERYNEWPROG" CmdResp if
      getResponse ack+len? 0= if		\ ACK + len
        1 = if
          ReadyResp if
            getResponse
            drop  Response$ 1 + c@	\ 0/1
          then
        then
      then
    then
    CloseConn
  then
;

: getNewProg	\ -- flag
\ *G Read a new program into the external RAM area. Return
\ ** true if the program is successfully loaded.
  0					\ -- flag ; assume no file
  OpenTargetConn if			\ 0 on error
    s" NEWPROG" CmdResp if
      getResponse ack+len? 0= if		\ ACK + len
        >r ReadyResp if			\ -- flag ; R: -- len
          drop r@ ReadProgram 0=		\ -- flag' ; R: -- len
        then
        r> drop
      then
    then
    CloseConn
  then
;

\ *P After loading the new program, it can be validated, for
\ ** example by checking a CRC at some address. The cross compiler
\ ** can generate a CRC over a range and store it at a target
\ ** address. Note that the client sends the client's IP address
\ ** with the command so that the server can identify the client.
\ ** You could also use MAC addresses.

\ *P After validation has succeeded, you can copy the RAM
\ ** image into the Flash using
\ *C  : CopyFlash	\ src dest len --
\ *P This word ends by rebooting the system.


\ ******
\ *> ###
\ ******

