\ TargetServer.fth - simple single-threaded server

((
Copyright (c) 2003, 2005, 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

To do
=====

Change history
==============
20140107 SFP001 Derived from PowerNet server.
))

decimal

include %vfxpath%\Lib\Win32\Genio\SocketIo	\ socket interface


\ ===============
\ *! targetserver
\ ===============
\ *P The code in *\i{<pnet>\Examples\TargetServer\TargetServer.fth}
\ ** implements a server used to manage embedded systems. The primary
\ ** purpose is to allow embedded targets to obtain software updates.
\ ** The command sequence is text-based and a single socket is used for
\ ** commands, data and status responses. The code is written so that
\ ** it can be used as a model by code in other languages. It requires
\ ** little more than a Winsock/BSD socket API.

\ *P The code in *\i{<pnet>\Examples\TargetServer\TargetClient.fth}
\ ** implements code to access the server, to check for a new binary
\ ** image, to download a new image and to copy this from RAM to
\ ** Flash.

\ *P The server is built by compiling *\i{TargetServer.fth} on
\ ** VFX Forth for Windows.

\ *P The server is single-threaded, dealing with one connection at
\ ** a time. Commands are sent from the target to the host.

\ *E Command sequence - bad command
\ **     HOST                 TARGET
\ **               <-         <cmd+params>
\ **     nack      ->
\ **               <-         <disconnect>

\ *E Command sequence - data to target
\ **     HOST                 TARGET
\ **               <-         <cmd+params>
\ **     ack len   ->
\ **               <-           ready
\ **     len data  ->
\ **               <-         <disconnect>

\ *E Command sequence - data to host
\ **     HOST                 TARGET
\ **               <-         <cmd+params+len>
\ **     ready     ->
\ **               <-           len data
\ **     status    ->
\ **               <-         <disconnect>


1 value TSdiags?	\ -- n
\ *G Set this non-zero to get diagnostic information about the
\ ** server.

: [ts		\ --
\ *G A macro used to surround diagnostic code, and terminated
\ ** by *\fo{TS]}.
\ *E   [TS  ." debug message"  TS]
  postpone TSdiags?  postpone if
; immediate

: ts]
\ *G Terminates a *\fo{[TS ... TS]} structure.
  postpone endif
; immediate

cell +user my_hSocket	\ -- addr
\ *G A user variable holding the current socket handle.
\ ** Equivalent to Thread-Local Storage (TLS).
cell +user my_hListen	\ -- addr
\ *G A user variable holding the current listening socket handle.
cell +user my_hConn	\ -- addr
\ *G A user variable holding the current connection socket handle.
cell +user my_close	\ -- addr
\ *G Set true when a task or socket should be closed.


\ ********************************
\ *S Service creation and deletion
\ ********************************

[-sin

: .SRVRmessage	\ caddr len --
\ *G Display a message to the current output (usually the console)
\ ** in the form:
\ *C   server <given text> on socket <n>
\ *P The *\fo{USER} variable *\fo{MY_HSOCKET} must contain valid data.
  base @ >r  decimal
  cr ." TS " type ."  on socket "  my_hsocket @ .
  r> base !
  #100 ms
;

: CloseServerSocket	\ --
\ *G Close the server's socket.
  my_hsocket @ closesocket drop
  0 my_hsocket !
;

: (CreateServerSocket)	{ #port -- res }
\ *G Create a listening TCP socket on the given port.
\ ** On success, the *\fo{USER} variable *\fo{MY_HSOCKET} contains
\ ** the socket number.
\ *D #port     port number to listen on.
\ *D res       socket number (1..n) or SOCKET_ERROR.
  my_hSocket off  my_close off		\ assume the worst !!

  AF_INET SOCK_STREAM IPPROTO_TCP socket \ create TCP socket
  dup INVALID_SOCKET =
  if  drop SOCKET_ERROR  exit  endif
  dup my_hsocket !

  AF_INET #port IPaddress @(n) bindTo	\ bind to our IP address
  if  CloseServerSocket SOCKET_ERROR  exit  endif
  my_hsocket @ SOMAXCONN Listen		\ go into Listen state
  if  CloseServerSocket SOCKET_ERROR  exit  endif

  my_hsocket @
;

: CreateServerSocket	\ port# --
\ *G Create a server socket on the given port.
  begin
    dup (CreateServerSocket) SOCKET_ERROR =
   while
    [ts  cr ." Can't create server socket on port " dup .  ts]
    #1000 Sleep  pause
  repeat
  [ts  s" listening" .SRVRmessage  ." port " dup . ts]
  drop
;

: SafeSAccept	\ -- hsock|0|-1
\ *G Waits until a connection is made to the server, returning
\ ** the socket handle or an error code.
  { | addr[ #64 ] addrlen[ cell ] -- hsock }
[defined] Target_386_Windows [if]
  0 FD_CLOSE FD_ACCEPT or		\ block, close or accept
  my_hsocket @ SetNotify >r		\ start notification
  0 begin				\ -- n ; last notify value
    drop  5 Sleep  pause
    my_hsocket @ GetNotify		\ check for close/accept
    my_close @ 				\ check for server closing
    if  FD_CLOSE or  endif
    dup
  until
  FD_CLOSE and if			\ get out on close
    0
  else
    #64 addrlen[ !
    my_hsocket @ addr[ addrlen[ Saccept
  endif
  r> CloseNotify			\ MUST be after accept()
[else]
  my_hsocket @ addr[ addrlen[ Saccept
[then]
;

: WaitAccept	\ -- socket|0
\ *G Wait in the server until a connection is established,
\ ** return the accepted socket.
  begin					\ wait for valid connection
    SafeSAccept dup 0=			\ will return 0 on listen close
    if  exit  endif
    dup INVALID_SOCKET =
   while
    [ts  s" bad connection" .SRVRmessage  ts]
    drop pause
  repeat
  [ts  s" got connection" .SRVRmessage  ts]
;

sin]


\ *******************************
\ *S Socket read/write primitives
\ *******************************

ErrDef err_BadSocket	"SOCKET_ERROR returned"

: sock-key?	\ -- flag
\ *G Return true if the socket defined by *\fo{MY_HSOCKET} has
\ ** data ready to read. A *\fo{THROW) occurs for a failed socket.
  my_hsocket @ pollSock
  dup SOCKET_ERROR = err_BadSocket ?THROW
;

: sock-key	\ -- char
\ *G Return the next character from the socket defined by *\fo{MY_HSOCKET}.
  {: | temp[ cell ] -- :}
  begin  sock-key? 0= while  1 ms  repeat
  $0A temp[ c!				\ returns LF on error
  my_hsocket @ temp[ 1 0 recv drop
  temp[ c@
;

: sock-readline	\ c-addr len -- len'
\ *G Read a line of text from the socket defined by *\fo{MY_HSOCKET}
\ ** into a buffer, without the EOL, which *\b{must} be CR/LF.
\ ** The length read is returned.
  0 tuck ?do				\ -- buff #read
    case  sock-key
      $0D of  0  endof
      $0A of  nip unloop exit  endof	\ -- #read
        2 pick c!  			\ save character
        swap 1+ swap 1+			\ step buffer and #read
        1
    end-case
  +loop
  nip
;

: sock-type	\ caddr len --
\ *G Write a block of data to the socket.
  begin
    dup
   while
    2dup my_hsocket @ writesock throw	\ -- caddr len #written
    /string				\ -- caddr' len'
  repeat
  2drop
;

: sock-emit	\ char --
  {: | tmp[ cell ] -- :}
  tmp[ c!  tmp[ 1 sock-type
;

: sendLine	\ caddr len --
\ *G Send text *\i{caddr/len} followed by CR/LF. The data is
\ ** buffered to prevent fragmentation.
  {: | temp[ 512 ] -- :}
  temp[ place  crlf$ count temp[ append
  temp[ count sock-type
;

: WaitData	\ ms -- ior
\ *G Wait for data for the given time in milliseconds
\ ** and return 0 if data is ready.
  later begin
    dup timedout? 0=
   while
    my_hsocket @ pollsock 0=
   while
    10 ms
  repeat then
  timedout?
  if  -1 exit  endif
  my_hsocket @ pollsock SOCKET_ERROR =
  if  -2 exit  endif
  0
;

256 constant /ReadyBuff		\ -- len
\ *G Size of the Buffer for the READY response.
/ReadyBuff buffer: ReadyBuff	\ -- addr
\ *G Buffer for the READY response.

: WaitReady	\ -- ior
\ *G Waits up to 2000 ms for the target to send "READY". Returns
\ ** true for errors and 0 on success.
  2000 WaitData dup if  exit  endif
  drop  my_hsocket @ ReadyBuff /ReadyBuff 0 recv
  5 < if  -3 exit  endif	\ must be at least 5 characters for READY
  ReadyBuff s" ready" is= 0=	\ case insensitive check of start of buffer
;


\ ********************
\ *S Server operations
\ ********************

2033 constant TargetPort#	\ --
\ G Port number used by the target server.

256 buffer: CmdBuffer	\ -- addr
\ *G Input buffer for command input.

vocabulary Cmds.Voc	\ --
\ *G Vocabulary for target commands.

[undefined] ConsoleIO [if]
: ConsoleIO	\ --
\ *G Define the default debug device for PowerNet tasks.
  ip-default @ ip-handle !
  op-default @ op-handle !
;
[then]

: CmdInterpret	\ --
  begin
    >in @ saved>in !
    ?stack  bl word dup c@
  while					\ -- caddr
    dup >r count ['] Cmds.Voc voc>wid
    search-wordlist ?dup if		\ R: -- *name ; xt +/-1
      r> drop
      0< state @ and
      if  compile,  else  Cexecute  then
    else				\ R: -- *name
      r@ Donumber? ?dup if		\ ?? ; R: caddr --
        r> drop  state @ if
          dup 0<
          if  C#Fliteral  else  C#literal  then
        else
          drop
        then
      else
        r> count CUndefined
      then
    then
  repeat
  drop
;
+short-branches

: TargetEval	\ i*x c-addr u -- j*x
\ *G Process the supplied string as though it had been entered via the
\ ** interpreter.
  save-input n>r
  to-source  -1 to source-id  >in off
  'SourceFile off
  ['] CmdInterpret catch
  nr> restore-input drop
  throw
;

: TargetConnection	\ --
\ *G Read and execute an incoming command
  my_hConn @ my_hSocket !
cr ." my_hListen = " my_hListen @ .
cr ." my_hConn   = " my_hConn @ .
cr ." my_hSocket = " my_hSocket @ .
  CmdBuffer 256 sock-readline
  CmdBuffer swap			\ -- caddr len
cr ." Command: " 2dup type
  TargetEval
;

[-short-branches
: TargetServer	\ 0 -- 0
  consoleIO decimal
." TargetServer started" cr
  begin
    TargetPort# createServerSocket
." Server socket created" cr
    my_hsocket @ my_hListen !
    begin
      WaitAccept dup my_hConn !
." SACCEPT returns " dup . cr
     while
      ['] TargetConnection catch ?dup if \ error
." Command error" cr
        s" NAK" sendLine		\ NAK response
        .throw
      endif
      100 ms				\ allow final XMIT to target
      my_hConn @ closesocket drop	\ we're done now
." Command completed" cr
      my_hListen @ my_hSocket !
    repeat
." SACCEPT failed" cr
    my_hListen @ closesocket drop
  again
;
short-branches]

Task TStask	\ -- addr
\ *G Task that is the Target Server.

: +TargetServer	\ --
\ *G Start the server if it is not running.
  TStask taskState 0= if
    ['] TargetServer TStask initiate
  endif
;

: -TargetServer	\ --
\ *G Stop the server if it is running.
  TStask taskState if
    TStask terminate
  endif
;


\ ******************
\ *S Target commands
\ ******************

1 value NewProg?

((
: NewProg?	\ -- flag
\ *G Return true when a new file is available.
  ...
;
))

: NewFile$	\ -- caddr len
\ *G Returns the name of the new file.
\  s" NewFile.img"
  s" TargetServer.fth"
;

: sendNAK	\ --
\ *G Send a NAK rsponse.
  s" NAK" sendLine
;

: sendACK+u	\ u --
\ *G Send an ACK followed by a decimal text number and a CR/LY.
  {: | temp[ 64 ] -- :}
  base @ >r  decimal
  s" ACK " temp[ place
  (u.) temp[ append
  r> base !
  temp[ count sendLine
;

: sendImage	{: hFile fsize buff -- :}
  buff fsize hFile read-file 0= if
    fsize = if
      buff fsize sock-type
    then
  then
;

: >EOL		\ --
\ *G Discard the remainder of the input line.
  0 word drop
;

also Cmds.Voc definitions

: QueryNewProg	\ -- ; QueryNewProg <ipaddress>
\ *G The client sends the query string. The parameters are used
\ ** to select whether a new file is available. The server returns
\ ** "ACK 1", waits for the READY response and then sends a single
\ ** byte, 0 for no file and non-zero if a file is available.
cr ." QueryNewProg "
  >eol					\ discard IP address
  S" ACK 1" sendLine			\ 1 character of data
  WaitReady 0= if			\ 0=success
    NewProg? 0<> 1 and sock-emit	\ send 0/1
  then
;

: NewProg	\ -- ; NewProg <ipaddress>
cr ." NewProg "
  >eol					\ discard IP address
  NewProg? 0=
  if  sendNAK exit  then
  NewFile$ r/o open-file
  if  drop  sendNAK exit  then		\ can't open file
  dup file-size if
    2drop  close-file drop		\ can't get file size
    sendNAK exit
  then
  drop					\ -- handle fsize
  dup sendACK+u				\ send ACK and file size
  WaitReady 0= if			\ 0=success
    dup allocate 0= if
      >r
      2dup r@ ['] sendImage catch drop
      r> free drop
    then
  then
  drop  close-file drop
;

previous definitions

+targetserver


\ ******
\ *> ###
\ ******

