\ WebPost.fth - POST handlers for PowerNet v4 onwards

((
Copyright (c) 2008
Guralp Systems
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20090109 SFP002 Added file storage example.
20081217 MPE001 Moved tools to PNmisc.fth.
		Moved header scanning to HTTP.FTH.
		Added NOT-- to avoid alignment issues.
))

only forth definitions
decimal

\ ==========
\ *! webpost
\ *T POST handlers and HTTP updates
\ ==========
\ *P The file *\i{<Pnet>/Examples/WebPost.fth} provides example
\ ** handlers for HTTP POST requests. According to the RFCs, GET
\ ** requests are made when the server is left unchanged by the
\ ** request. POST requests are made when the state of the server
\ ** is (or may be) changed by the request.

\ *P The example here is of uploading a new application binary
\ ** image to the server. *\fo{DEFER}red words are used to handle
\ ** the target specific actions. This example can be used as the
\ ** basis of other POST handlers, especially those using multipart
\ ** forms. For details of form encoding, see RFCs 2045 and 2046.
\ ** These can be obtained from *\f{http://www.rfc-editor.org}.

\ *P To use the system, you need PowerNet version 4.62 or later.
\ ** Compile the file *\i{<Pnet>/Examples/WebPost.fth} after
\ ** the PowerNet build file. When the system is running, point
\ ** your browser at the page *\i{Reflash.htm}. When you
\ ** have selected the required file, press the Submit button.
\ ** This returns the form results and new binary to a page called
\ ** *\i{NewApp.asp}, which is actually a Forth word that performs
\ ** the process. When it is complete, the page *\i{/naresp.asp}
\ ** is served to display the results to the user.

\ *P To provide flexibility in how the new binary is handled, a
\ ** simple interface is provided that can be used to save the
\ ** new binary code to Flash or a file system, e.g. on an SD
\ ** card.


\ *************
\ *S Discussion
\ *************
\ *P The requirement is to be able to point any browser at the
\ ** PowerNet web server, and to be able to upload a new binary
\ ** image to it using a form for data entry. What is done with
\ ** this file is application dependent, but it may replace the
\ ** existing application, or it may be saved to an SD card.

\ *******
\ *N Form
\ *******
\ *P In order to send binary data to a server (file upload), a
\ ** POST request must be made by the form. Here is the HTML for
\ ** an example form. Following sections indicate the response,
\ ** in our case using Firefox.
\ *E <html>
\ ** <head><title>Select new application</title></head>
\ ** <body>
\ ** <h4>Select application</h4>
\ ** <p>Select the application using the Browse button,<br>
\ ** then press the "Submit" button to send it to me.<br>
\ ** Once you have pressed "Submit" you are committed.<br>
\ ** To avoid sending a file, return to the previous page.
\ ** </p>
\ ** <form enctype="multipart/form-data" method=POST action="newapp.asp">
\ **   <table>
\ **   <tr>
\ **     <td>New application file</td>
\ **     <td colspan=4><input type=file name=appfile></td>
\ **   </tr>
\ **   <tr>
\ **     <td>Upload to remote:</td>
\ **     <td colspan=2><input type=submit name=send value=Submit></td>
\ **   </tr>
\ **   </table>
\ ** </form>
\ ** </body>
\ ** </html>

\ **********
\ *N Headers
\ **********
\ *P The following is a response by Firefox to the form above.
\ ** Note that whenever you press a button of the submit type,
\ ** you get the complete response, including any selected file.

\ *E POST /newapp.asp HTTP/1.1
\ ** Host: 192.168.0.227
\ ** User-Agent: Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.14) Gecko/20080404 Firefox/2.0.0.14
\ ** Accept: text/xml,application/xml,application/xhtml+xml,text/html;q=0.9,text/plain;q=0.8,image/png,*/*;q=0.5
\ ** Accept-Language: en-us,en;q=0.5
\ ** Accept-Encoding: gzip,deflate
\ ** Accept-Charset: ISO-8859-1,utf-8;q=0.7,*;q=0.7
\ ** Keep-Alive: 300
\ ** Connection: keep-alive
\ ** Referer: http://192.168.0.227/Reflash.htm
\ ** Content-Type: multipart/form-data; boundary=---------------------------805274455224
\ ** Content-Length: 9386
\ **
\ ** -----------------------------805274455224

\ *P The *\fo{Content-Type} and *\fo{Content-Length} headers are
\ ** essential. The first tells us that this is a form response
\ ** in the style we need and the second tells us the overal size
\ ** of the response. If you need security, you can use the
\ ** *\fo{Referer} field to indentify which host and file contained
\ ** the form.

\ *P In particular the *\fo{Content-Type} field contains a
\ ** *\fo{boundary} string. This is the only data that identifies
\ ** where one part of the form starts and ends.

\ ******************
\ *S Form boundaries
\ ******************
\ *P The boundary string is defined by the browser. The PowerNet
\ ** server has *\b{no} control over it. However, it is mandatory
\ ** that the string is *\b{not} contained in the data.

\ *P The example below shows three versions of the boundary string.
\ *E ---------------------------805274455224
\ ** -----------------------------805274455224
\ ** -----------------------------805274455224--

\ *P The first is the separator defined by the *\fo{boundary}
\ ** portion of the *\fo{Content-Type} header.
\ ** Under some circumstances, this may be delimited by '"'
\ ** characters which are not part of the boundary string.

\ *P The second is applied to all part separators except the last one.
\ ** The second form is the same as the first but with two leading '-'
\ ** characters. It is always preceded by a CR/LF pair and terminated
\ ** by a CR/LF pair.

\ *P The third marks the end of the form. It is identical to the
\ ** second type but has two additional '-' characters at the end
\ ** of the line.

\ ************
\ *N Form data
\ ************
\ *P Using the form above, we receive two blocks of form data
\ ** between boundary markers.

\ *E Content-Disposition: form-data; name="appfile"; filename="memcopy.s"
\ ** Content-Type: application/octet-stream
\ **
\ ** /******************** (C) COPYRIGHT
\ ** ...

\ *P Note that in our form, the submit button is after the file
\ ** browser, so the file is sent first! The order in which the
\ ** elements of the form are sent is the order in which they appear
\ ** in the HTML of the form.

\ *P The *\fo{Content-Disposition} header tells us which element
\ ** of the form is being returned, and the *\fo{Content-Type}
\ ** header tells us the file is being transferred as 8 bit
\ ** binary. There is blank line (CR/LF pair) between the last
\ ** header and the start of the binary data.

\ *P Note that there is nothing to tell us the size of the data!
\ ** We *\fo{must} rely on detecting the boundary separator unless
\ ** there is some magic data at the start of the binary data
\ ** section.

\ *E Content-Disposition: form-data; name="send"
\ **
\ ** Submit

\ *P This represents the Submit button. You can have several
\ ** buttons of type *\fo{submit}. They will each cause the form
\ ** response. You can use them to decide where to place the data,
\ ** e.g. primary or secondary application.

\ *P If you want a "Cancel" button, your page will need a second
\ ** form pointing to a different page.

\ *****************
\ *N After the form
\ *****************
\ *P After the last item of form data will come the residue, which is
\ ** usually null. Note that more than one form may be contained in
\ ** a web page.

\ ===============
\ *N Restrictions
\ ===============
\ *P The following assumptions and restrictions should be noted.
\ *(
\ *B Only the last boundary string defined in the message headers
\ ** is used. Thus nested boundary strings are not supported.
\ *B It is assumed that a boundary string is followed by CR/LF
\ ** or -/-/CR/LF.
\ *B Very little checking of the message is performed.
\ *B For simplicity, compatibility with previous versions of
\ ** PowerNet and to save RAM, global variables are used.
\ *)


\ *******************************
\ *S Parsing multipart boundaries
\ *******************************
\ *P There are two conditions we have to deal with, parsing text
\ ** lines and detecting the marker at the end of a binary file.
\ ** The second situation is coded within the binary file handler.

: Not--		\ caddr -- flag ; true for not '--'
\ *G Return true if the two characters at *\i{caddr} are not '--'.
  dup c@ 8 lshift swap 1+ c@ or $2D2D <>
;

: Bdry$?	\ caddr len -- 0|-1
\ *G Returns true if the string matches the boundary string plus
\ ** two leading dashes.
  over Not--			\ first two chars <> '--' ?
  if  2drop 0  exit  endif
  drop 2+ Boundary$ count s=
;

: Boundary?	\ caddr len -- -1|0|1
\ *G Returns -1 if the string is a normal boundary marker, 0 if it
\ ** is not a boundary marker, and 1 if it is the last boundary
\ ** marker.
  dup Boundary$ c@ 2+ =	if		\ check length - normal
    Bdry$? exit
  endif
  dup Boundary$ c@ 4 + = if		\ check length - last
    2dup + 2- Not--			\ last two chars <> '--' ?
    if  2drop 0  exit  endif
    Bdry$? 1 and exit
  endif
  2drop 0
;

: NextBoundary	\ -- flag
\ *G Read text up to the next boundary marker. *\i{Flag} is
\ ** returned true if the boundary was the last one.
  0 begin
    drop  SVquery
    source Boundary? dup
  until
  1 =
;


\ ***************************
\ *S Flash update application
\ ***************************

\ ===================
\ *N System interface
\ ===================
\ *P The flash interface with the underlying system is handled
\ ** by three *\fo{DEFER}red words.

#1024 equ /FlashBlock	\ -- len
\ *G Unit size passed to the application.

Defer InitUpd	\ caddr len --
\ *G Initialise the app's binary update system. The input string
\ ** is the file name to be received. This is passed for validation
\ ** purposes and in case the file is saved in a local file system.
  assign 2drop to-do InitUpd
Defer AddUpd	\ caddr len --
\ *G Add the given memory block to the update. All blocks except
\ ** the last contain *\fo{/FlashBlock} bytes of data.
    assign 2drop to-do AddUpd
Defer TermUpd	\ --
\ *G Terminate the app's binary update system.
  assign noop to-do TermUpd


\ ===================
\ *N Receiving a file
\ ===================

#256 buffer: Filename$	\ -- addr
\ *G Buffer holding file name as a counted string.

#80 equ /BndryBlock	\ -- len
\ *G Size of additional buffer for boundary detection. The maximum
\ ** size of a boundary string is 70 characters (RFC2046). These
\ ** are prefixed by CR/LF/-/- at the start and will be followed
\ ** by -/- in the last boundary line.
/FlashBlock buffer: BinBuff	\ -- addr
\ *G Buffer for storing binary data.
/BndryBlock buffer: BndryBuff$	\ -- addr
\ *G Buffer for storing potential boundary string data.
\ ** Holds a counted string starting with a CR.

variable NextByte	\ -- addr
\ *G Holds the address (in *\fo{BinBuff}) of the next character
\ ** to be received.
variable #FileSize	\ -- addr
\ *G Holds the number of bytes stored as binary data.
variable #AppSize	\ -- addr
\ *G Holds the number of bytes transferred to the application.
variable BinDone	\ -- addr
\ *G Set if the file has been transferred.
variable BinStatus	\ -- addr
\ *G Holds the file receive status (0=good).

: ResetBlock	\ --
\ *G Reset the next byte pointer.
  BinBuff NextByte !
;

: StartRxApp	\ --
\ *G Initialise variables and buffers for file reception.
  #AppSize off  #FileSize off  0 BndryBuff$ c!
  ResetBlock  Filename$ count InitUpd
\ [io consoleio  cr ." StartRxApp, file=" Filename$ count type  io]
;

: Block>App	\ --
\ *G Send the data block to the application. All blocks except
\ ** the last contain *\fo{/FlashBlock} bytes of data.
  BinBuff NextByte @ over -  dup #AppSize +!  dup
  if  AddUpd  else  2drop  endif
  ResetBlock
;

: StopRxApp	\ --
\ *G Clean up after receiving a file.
  TermUpd  ResetBlock
\ [io consoleio  cr ." StopRxApp"  io]
;

: addData	\ byte --
\ *G Add the byte to the binary buffer.
  1 #FileSize +!			\ update count
  NextByte @ tuck c!  1+ dup NextByte !	\ update storage
  BinBuff /FlashBlock + u< not		\ check for pass to Flash
  if  Block>App ( SendRxBlock )  endif
;

: NotBoundary	\ --
\ *G Add the partial boundary string to the data and clear the
\ ** buffer.
  BndryBuff$ count bounds
  ?do  i c@ addData  loop
  0 BndryBuff$ c!
;


create CrLf--	\ -- caddr
\ *G Holds the four characters CR/LF/-/-.
  #13 c,  #10 c,  char - c,  char - c,

: FullBndry?	\ -- flag
\ *G Check the contents of *\fo{BndryBuff$} against the start of
\ ** *\fo{Boundary$} which holds the received boundary separator.
\ ** If the data does not match copy the data to the binary buffer.
\ ** If the data matches the complete boundary string, return
\ ** true, otherwise return false.
\ *P By definition the marker string does not appear in the
\ ** the binary. The code reads the binary into a buffer that is
\ ** 80 characters longer than the required Flash block size to
\ ** accommodate the boundary string (70 chars maximum). We check
\ ** for the sequence CR/LF/-/-, and if found check for the
\ ** following boundary string.
  BndryBuff$ c@ 4 <= if			\ first part
    CrLf-- BndryBuff$ count s= 0=
    if  NotBoundary  endif
    0
  else
    BndryBuff$ count 4 /string		\ region to compare
    Boundary$ 1+ swap s= if		\ matches
      BndryBuff$ c@ 4- Boundary$ c@ =	\ returns true for full match
    else
      NotBoundary 0
    endif
  endif
;

: ReceiveBinary	\ --
\ *G The separator line between the part headers and the binary
\ ** file has been read. Read the binary up to the concluding
\ ** part boundary. Later code will check if this is the final
\ ** boundary.
\ [hd  cr ." Starting binary receive"  hd]
  begin
    SVkey  BndryBuff$ c@ if
      BndryBuff$ addChar  FullBndry?
      if  BinDone on  exit  endif
    else
      dup #13 = if			\ start checking at CRs
        BndryBuff$ addChar
      else
        addData
      endif
    endif
  again
;

: BinTail	\ -- flag
\ ** If the boundary tail matches the last boundary condition,
\ ** return true, otherwise return false.
\ [hd  cr ." Checking tail"  hd]
   SVkey SVkey [char] - dup d= dup	\ last if -/-/CR/LF
   if  SVkey drop  SVkey drop  endif
;

: ReadBinFile	\ -- flag
\ *G The separator line between the part headers and the binary
\ ** file has been read. Read the binary up to the concluding
\ ** part boundary. If the boundary string is the last one,
\ ** return true, otherwise return false.
  StartRxApp  ReceiveBinary Block>App  BinTail  StopRxApp
;


\ ****************
\ *N Part scanning
\ ****************
\ *P In this example, the only form part we need to process
\ ** is the one containing a binary file. We detect this by
\ ** parsing the headers for the pair:
\ *E Content-Disposition: form-data; name="appfile"; filename="foo"
\ ** Content-Type: application/octet-stream
\ *P We assume that a valid file is present when
\ *(
\ *B the "form-data" marker is present,
\ *B the filename is non-null
\ *B and the "application/octet-stream" marker is present.
\ *)

variable OctetStream	\ -- addr
\ *G Set true when *\f{Content-Type: application/octet-stream}
\ ** has been received.
variable FormData	\ -- addr
\ *G Set true when the "form-data" marker has been received.

: ?formData	\ caddr len --
\ *G Check for the "form-data" marker.
  s" form-data" isearch
  if  FormData on  endif
  2drop
;

: ?appfile	\ caddr len --
\ *G Check for 'name="appfile"'.
  s" filename" dup >r isearch if	\ look for filename string
    r> /string -leading -trailing
    Filename$ ExtractValue
  else
    2drop  r> drop
  endif
;

: doPdisp	\ caddr len --
\ *G Check the given content disposition line.
  2dup ?formData ?appfile
;

: doPtype	\ caddr len --
\ *G Check the given content type line.
  s" application/octet-stream" isearch
  if  OctetStream on  endif
  2drop
;

create PartHdrs	\ -- addr
\ *G Table holding the headers to be processed by part header
\ ** handlers.
  ' doPdisp ,	", Content-Disposition" align
  ' doPtype ,	", Content-Type" align
  0 , 0 ,

: CheckPartHdrs	\ -- x
\ *G Check the data received in the part headers, and return
\ ** *\i{x}, indicating what to do with the part. If *\i{x}
\ ** is zero, just skip the part.
  FormData @  OctetStream @ and  FileName$ c@ 0<> and
\ [io consoleio  cr ." Check=" dup .  io]
;

: ProcessPart	\ -- flag
\ *G After receiving a boundary, process a part up to and including
\ ** the next boundary, returning true if it is the last one.
  FormData off  OctetStream off
  PartHdrs ProcessHeaders  CheckPartHdrs
  if  ReadBinFile  else  NextBoundary  endif
\ [io consoleio  cr ." ProcessPart=" dup .  io]
;

: ProcessParts	\ --
\ *G Read the parts up to the last boundary
    0 FileName$ c!  NextBoundary 0= if
    begin  ProcessPart  until
  endif
;


\ **********************
\ *S File update handler
\ **********************

: doNewApp	\ caddr len --
\ *G This is the handler for the update request.
  BinDone off  BinStatus off		\ SFP002
  ReadHeaders  ProcessParts
  2drop  s" /nareply.asp" ServePage
;
' doNewApp xtPage: /NewApp.asp
\ *G Defines the page which performs the upload.

: .UploadDone	\ --
\ *G Used in the body of an ASP page to indicate the response
\ ** to the upload.
  cr ." File upload "  BinDone @ if
    BinStatus @ if			\ SFP002
      ." failed with status=" BinStatus @ .
    else
      ." successful,"
    endif
    ." <br>received " #FileSize @ .decimal
    ." bytes from file " FileName$ count type
    ." ,<br>copied " #AppSize @ .decimal ." bytes to the application."
  else
    ." failed."
  endif
  cr
;

MemPages? [if]
MemPage: /Reflash.htm %IPSTACK%\TestPages\Reflash.htm
\ *G A form used to request a new application.
MemPage: /nareply.asp %IPSTACK%\TestPages\nareply.asp
\ *G The response delivered after an upload.
align
[then]


\ ====================================
\ *N Example for pages stored in files
\ ====================================
\ *P This example can be used to upload binary files, e.g. a new
\ ** application image, to the file system. In order to compile this
\ ** code the FAT file system must be present, and the equate
\ ** *\fo{FilePages?} (used in *\i{Pages.fth}} must be set non-zero.

((
Defer InitUpd	\ caddr len --
\ *G Initialise the app's binary update system. The input string
\ ** is the file name to be received. This is passed for validation
\ ** purposes and in case the file is saved in a local file system.
  assign 2drop to-do InitUpd
Defer AddUpd	\ caddr len --
\ *G Add the given memory block to the update. All blocks except
\ ** the last contain *\fo{/FlashBlock} bytes of data.
    assign 2drop to-do AddUpd
Defer TermUpd	\ --
\ *G Terminate the app's binary update system.
  assign noop to-do TermUpd
))


FilePages?  [defined] create-file and [if]

[undefined] FileDir$ [if]
create FileDir$	\ -- addr
\ *G The base directory for files. This will be prepended to
\ ** the page name given to *\fo{InitUpd}. The string is held
\ ** as a counted string in the buffer. Make sure that your
\ ** application sets *\fo{FileDir$} to point to its own pages.
\ ** The directory name must not end in a separator. The directory
\ ** must exist before use.
  ", \Uploads"
[then]

-1 value hPostFile	\ -- handle
\ *G Handle of the file used to save the download data.

: InitFileUpd	\ caddr len --
\ *G Initialise the file reception.
  -1 to hPostFile			\ assume it'll end badly
  MAX_PATH allocate			\ buffer space for file name
  if  1 BinStatus !  drop exit  endif
  >r
  filedir$ count r@ place		\ base directory
  over c@ dirSep? 0=			\ check file name
  if  [char] \ r@ addchar  endif	\ add separator
  r@ append				\ add file name
  r@ count 2dup PrepFilename		\ apply OS requirements
  2dup delete-file drop			\ delete existing
  r/w create-file  r> free drop		\ create file
  if  2 BinStatus !  drop exit  endif
  to hPostFile
;

: AddFileUpd	\ caddr len --
\ *G Add the given memory block to the update. All blocks except
\ ** the last contain *\fo{/FlashBlock} bytes of data.
  hPostFile -1 <> if
    hPostFile write-file if
      hPostFile close-file drop
      3 BinStatus !  -1 to hPostFile
    endif
  else
    2drop
  endif
;

: TermFileUpd	\ --
\ *G Terminate the app's binary file update system.
  hPostFile -1 <> if
    hPostFile close-file
    if  4 BinStatus !  endif
  endif
;

assign InitFileUpd to-do InitUpd
assign AddFileUpd  to-do AddUpd
assign TermFileUpd to-do TermUpd

[then]


\ ======
\ *> ###
\ ======

decimal
