\ PBUFFERS.FTH - definitions for queued buffer handling

((
Copyright (c) 1997-2005
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999, 2005
and:
  CEM Systems Ltd, Belfast, 2000-2002

To do
=====

Change history
==============
20050301 EGS003 Converted to pure PBuffs, removed iobuffer definitions etc
))


\ ===========
\ *! pbuffers
\ ===========
\ *T PBUF buffers


\ ***************
\ *S Introduction
\ ***************
\ *P PBUFs are used to hold data
\ ** which is received or transmitted as IP packets. Rather than
\ ** copy data repeatedly, the system passes pointers to these
\ ** structures. In order to avoid heavy heap use, these buffers
\ ** are preallocated.

\ *P Collections of buffers are handled by the QUEUE data structures.

\ *P The function of PBUFs is to hold data that will be received or
\ ** transmitted to a device. Headers are added to application data.
\ ** The worst case occurs for Ethernet systems:
\ *E Ethernet header
\ ** IP header
\ ** TCP, UDP or other protocol header(s)
\ ** Application data

\ *P Traditionally, TCP/IP stacks based on the BSD implementations
\ ** use separate buffers for each data area. In order to avoid
\ ** additional heap allocations and data copying, PowerNet uses
\ ** a single buffer and moves pointers as the PBUF is moved up
\ ** and down the TCP/IP stack.


\ ******************
\ *S Data structures
\ ******************

\ the PBuf structure
\ multiple entries are defined dependent on the pbuf useage
\ N.B. The first 10 bytes are common with a simple data buffer
\ this is to allow management by common code
struct pbuf_hdr			\ -- size ; size of pbuf_hdr structure
\ *G PBUF buffer header definition. The first four fields
\ ** (10 bytes) are common with the obsolete BF_HDR definition.
  cell field *p_next	\ points to the next PBUF in a PBUF chain
  2 field p_type	\ the type & size of the pbuf
			\ bits 13-0  = size (up to 16383/$3FFF)
			\ bits 15-14 = type
\ -- the next two fields can be changed by the xxBump words so
\ -- that software can deal with a "current" UDP or TCP data area.
  2 field p_datalen	\ length of current data, e.g. TCP or UDP data
  2 field p_data	\ offset to current data from start of pbuf
\ -- the next two fields are used to define a current header,
\ -- IP, UDP, or TCP header
  2 field p_pkthdr	\ offset to packet data header, e.g IP, UDP or TCP header
  2 field p_pkthdrlen	\ length of packet header (never read!)
\ -- the next two fields describe the total payload
  2 field p_len		\ length of payload (but not in device layer)
  0 field p_dat		\ start of payload is here
end-struct

(( old style names - left in for converting old code.
struct /buffer_header	\ -- size
\ Tx/Rx I/O buffer header definition. The first four fields
\ (10 bytes) are common to the PBUF definition below.
  cell field >bf_*next		\ pointer to next buffer
  2    field >bf_type		\ type & size of allocated buffer
				\ bits 11-0  = size (up to 4095/FFF)
				\ bits 15-12 = type
  2    field >bf_datalen	\ number of chars in buffer
  2    field >bf_data		\ offset to start of the data
  0    field >bf_dat		\ the start of the data
end-struct

struct /pbuf_hdr	\ -- size ; size of pbuf_hdr structure
\ PBUF buffer header definition. The first four fields
\ (10 bytes) are common to the *\fo{BF_HDR} definition above.
  /buffer_header FIELD >p_header>
\  cell field *p_next		\ points to the next pbuf in a pbuf chain
\  2    field p_type		\ the type & size of the pbuf
\				\ bits 11-0  = size (up to 4095/FFF)
\				\ bits 15-12 = type
\  2    field p_datalen		\ length of current data, e.g. TCP or UDP data
\  2    field p_data		\ offset to current data from start of pbuf
\				\ end of common data
  2    field p_pkthdr		\ offset to packet data header, e.g IP or TCP header
  2    field p_pkthdrlen	\ length of packet header (unused)
  2    field p_len		\ length of payload (but not in device layer)
  0    field p_dat		\ start of payload is here
end-struct
))


PSIZE pbuf_hdr + equ pbuf	\ -- n
\ *G Size of complete pbuf structure.

pbuf CELL 1- and [if]
  cr ." PBUF must be cell aligned. " abort
[then]

$3FFF equ BUFFSIZEMASK	\ -- mask
\ *G The buffer size is held in the low 14 bits of the 16 bit
\ ** type field. This permits a maximum data size of *\fo{BUFFSIZEMASK}
\ ** bytes.
$C000 equ BUFFTYPEMASK	\ -- mask
\ *G The buffer type is held in the top 2 bits of the 16 bit
\ ** type field.

PSIZE BUFFSIZEMASK u> [if]
  cr ." Urk! PSIZE too big for PBUF" abort
[then]

\ Defined buffer types

$0000 equ DATABUFFER	\ -- n
\ *G Set into the type bits for I/O buffers.
$4000 equ PBUFFER	\ -- n
\ *G Set into the type bits for PBUFs.


\ ****************
\ *S PBUF handling
\ ****************

decimal

: !pb_size		\ n *pb -- ; set buffer size
\ *G Set the size of a buffer and mark it as a PBUF.
  swap PBUFFER or  swap p_type w!
;
compiler
: !pb_size		\ n *pb -- ; set buffer size
  swap PBUFFER or  swap p_type w!
;
target

: PBHdrStart	\ *pbuf -- addr ; of packet header
\ *G Return the address of the current header.
  dup p_pkthdr w@ +
;

: PBHdrLen	\ *pbuf -- len ; of packet header
\ *G Return the length of the current header.
  p_len w@
;

: PBDataStart	\ *pbuf -- addr ; of data
\ *G Return the address of the current data.
  dup p_data w@ +
;

: PBDataLen	\ *pbuf -- len ; of data
\ *G Return the length of the current data.
  p_datalen w@
;

: !PBDataLen	\ len *pbuf -- ; of data
\ *G Set the length of the current data.
  p_datalen w!
;

: !*PBData	\ offset addr --
\ *G Set offset to buffer data.
  p_data w!
;

compiler
: PBHdrStart	\ *pbuf -- addr ; of packet header
  dup p_pkthdr w@ +
;

: PBHdrLen	\ *pbuf -- len ; of packet header+data
  p_len w@
;

: PBDataStart	\ *pbuf -- addr ; of data
  dup p_data w@ +
;

: PBDataLen	\ *pbuf -- len ; of data
  p_datalen w@
;

: !PBDataLen	\ len *pbuf -- ; of data
  p_datalen w!
;

: !*PBData	\ offset addr --
  p_data w!
;
target

: bumpPBuffData		\ n *pbuf --
\ *G Moves the data pointer/count along the buffer,
\ ** incrementing the address, decrementing the count.
  2dup p_data w+!
  swap negate swap p_datalen w+!
;

: unbumpPBuffData	\ n *pbuf --
\ *G Moves the data pointer/count back up the buffer,
\ ** decrementing the address, incrementing the count.
  swap negate swap bumpPBuffData
;


\ **************************************
\ *S Queue buffer allocation and release
\ **************************************

: PBinit	\ *pb --
\ *G Initialise a PBUF.
  PSIZE over !pb_size			\ set PBUFFER type and size
  0 p_dat over p_data w!		\ p_data holds offset to data
  0 over p_datalen w!			\ set no data in PB
  0 p_dat over p_pkthdr w!		\ set start of packet
  drop
;

: AllocPbuf	\ -- *pb|0
\ *G *\fo{ALLOCATE} and initialise a PBUF from the heap and return
\ ** the PBUF address. If the buffer cannot be allocated, 0 is
\ ** returned without waiting.
  pbuf allocate if			\ grab a buffer and flag the type
    drop 0  exit			\ exit if failed
  endif
  dup pbuf erase			\ wipe all; why not just the header?
  dup PBinit				\ initialise it
;

#TXpbufs [if]
0 value TxPbuf		\ -- addr
\ *G Points to the dedicated transmit PBUF.
0 value NextTxPbuf	\ -- addr
\ *G Holds *\fo{TxPbuf} when the transmit PBUF is available, or
\ ** zero when not available.
[then]

: InitIOQueues	\ --
\ *G Initialise the predefined queues and their I/O buffers
\ ** and PBUFs.
  FirstQ LastQ over - erase		\ wipe queues
  NUMFREEPBUFS 0 do
    AllocPbuf ?dup 0= if		\ grab a PBUF from the heap with initialisation
      cr ." Failed to allocate requested PBUFs" leave
    else
      FreePBufQ Enqueue			\ add this to the free PBuf queue
    endif
  loop
[ #TXpbufs ] [if]
  AllocPbuf  dup -> TxPbuf  dup -> NextTxPbuf  0= if
    cr ." Failed to allocate TX PBUF"
  endif
[then]
;

: GetPbuf	\ -- *pb|0
\ *G Get a PBUF from the free PBUF queue and return
\ ** the PBUF address. If there are no free buffers, 0 is
\ ** returned without waiting.
  FreePBufQ dequeue dup if
    dup PBinit
  endif
;

#TXpbufs [if]
: GetTxPbuf	\ -- *pb|0
\ *G Get a PBUF from the free PBUF queue and return the PBUF address.
\ ** If there are no free buffers, try the dedicated transmit PBUF.
  GetPbuf dup 0= if
    drop NextTxPbuf dup if
      0 -> NextTxPbuf
    endif
  endif
;
[else]
XC? [if] compiler [then]
: GetTxPbuf	GetPbuf  ;
XC? [if] target [then]
[then]

0 equ QBdiags?	\ -- n
\ *G When this equate is set non-zero, additional test code is
\ ** compiled and included in *\fo{FREEQB} below.

QBdiags? [if]

variable qbug	\ -- addr
\ *G When set to non-zero, *\fo{FREEQB} will display diagnostics if
\ ** the buffer is bad.

PBUFFER PSIZE or equ PBcheck	\ -- x
\ *G Check value for a PBUF.

: checkQB	\ *qb -- *qb
  qbug @ if
    dup p_type w@
    dup PBcheck =  swap IOcheck = and
    if  exit  endif
    opvec @ >r  ipvec @ >r  ConsoleIO
    cr ." Bad buffer in task: " self .task
    dup $80 dump
    r> ipvec !  r> opvec !
  endif
;

[then]

#TXpbufs [if]
: FreeQB	\ *pb/bf --
\ *G Place a PBUF or I/O buffer on the relevant free queue.
\ ** N.B. *\fo{FreeQB} cleans up the buffers before requeueing
\ ** them.
[ QBdiags? ] [if] checkQB [then]	\ for testing only
  dup PBinit  dup TxPbuf = if
    -> NextTxPbuf
  else
    FreePBufQ Enqueue
  endif
;
[else]
: FreeQB	\ *pb/bf --
\ *G Place a PBUF or I/O buffer on the relevant free queue.
\ ** N.B. *\fo{FreeQB} cleans up the buffers before requeueing
\ ** them.
[ QBdiags? ] [if] checkQB [then]	\ for testing only
  dup PBinit FreePBufQ Enqueue
;
[then]

((
: GetPbufWait	\ n -- *pb|0
 \ *G Get a PBUF from the free PBUF queue and return
 \ ** the PBUF address. If there are no free buffers, 0 is
 \ ** returned after n milliseconds.
  later
  begin				\ -- time
    GetPbuf ?dup
    if  nip  exit  endif
    dup expired
  until
  drop 0
;
))


\ ======
\ *> ###
\ ======

decimal

