\ Servers.fth - Support for multi-threaded servers.

((
Copyright (c) 2003, 2005, 2007, 2008, 2009, 2011, 2012, 2013
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999


To do
=====
Add a STOPSERVER word. This must close the listening socket at
least before killing the task. It may be desirable to force a
reset on all active connections. It's not as easy as it looks!

Change history
==============
20130321 SFP019 Added application/service specific shutdown
		action for FTP.
20130306 SFP018 Added WAITSOCKETSENT
20130118 SFP017 Used TCPSOCKET from BSD.FTH.
20120215 SFP016 Restored the kill chain.
20110729 SFP015 More initialisation.
20110717 SFP014 Added more PAUSEs on error.
		Removed kill chain.
20090528 SFP013 Removed service creation for listening sockets.
		Overhauled for new LISTEN and SACCEPT.
20090505 SFP012 Add service length to socket structure.
20081201 SFP011 Overhauled service data structure defaults.
		Renamed SENDSVEMITBUFF to SVFLUSHOP.
		Added low RAM option.
20080603 SFP010 Moved Telnet TNACCEPT and TNQUERY here and renamed
		them SVACCEPT and SVQUERY.
		Changed the error return character from SVKEY to
		line feed to match the above.
20080303 SFP009 Refactored to restore SVCHAIN and friends to this file.
20080228 SFP008	Updated shutdown system to match changes in TCP.FTH.
20080225 SFP007 Refactored INITSERVERSOCKET.
20080222 SFP006 Fixed problem with socket closure during listen.
20071113 SFP005 Closed socket on failure to launch service.
20070828 VRW004 Moved chain mechanism to Primitiv.fth.
20050302 SFP003 Made COPYUSER a macro.
		Made SERVICEIO a synonym.
20030708 SFP002 Corrected race problem in ?SERVICECLOSE.
		Changed EMIT structure to avoid using queues,
		which cause clean-up problems and additional
		task switches. This also improved output
		performance by 20%.
20030416 SFP001 Derived from Telnet server.
))

decimal


\ ==========
\ *! servers
\ *T Support for TCP services
\ ==========
\ *P PowerNet supports multithreaded TCP servers that can accept
\ ** multiple connections. The model used is that one task listens
\ ** on a port, establishes a connection, passes the socket to a
\ ** service task to run the service, and then creates a new
\ ** listening socket. When a connection to a service task is no
\ ** longer established, a support task cleans up the service
\ ** specific data, closes the service socket, and terminates the
\ ** service task.

\ *P The focus of the tools is for servers using text, e.g. HTTP
\ ** and Telnet. To enable use of standard Forth words and parsing
\ ** tools, a Generic I/O device and a *\fo{TIB} are established
\ ** for each connection .

\ *P From PowerNet v4.6 onwards and then again in v4.8, much
\ ** attention to reducing RAM usage has taken place. This comes
\ ** at the expense of performance. Use of the the low RAM
\ ** configuration may be necessary for single-chip applications.
\ ** From v4.8 onwards, you can use the equates *\fo{SVlowRAM?}
\ ** and*\fo{SVsingle?} in the PowerNet configuration file to
\ ** reduce RAM consumption in listen state by up to 2kb per
\ ** listening socket above the reductions already achieved in
\ ** the TCP layer.


\ ******************
\ *S Service numbers
\ ******************
\ *P These are MPE defined for use in the socket structure.

0 equ service_none	\ -- n
\ *G Defines a null service.
1 equ service_Telnet	\ -- n
\ *G Defines a Telnet service.
2 equ service_HTTP	\ -- n
\ *G Defines an HTTP service.
3 equ service_FTP	\ -- n
\ *G Defines an FTP service.

8 equ service_Echo       \ -- n
\ *G Defines an Echo service.
9 equ service_MultiChat	\ -- n
\ *G Defines a MultiChat service. Windows version only.
10 equ service_ModBus
\ *G Defines a ModBus TCP service.

20 equ service_App
\ *G Service number for an application-specific service.

: .service	\ n --
\ *G Display the service type corresponding to n.
  case
    service_none of  ." No "  endof
    service_telnet of  ." Telnet "  endof
    service_HTTP of  ." HTTP "  endof
    service_FTP of  ." FTP "  endof
    service_Echo of  ." Echo "  endof
    service_Modbus of ." ModBus "  endof
    service_App of ." Application "  endof
      ." Unknown "
  endcase
;


\ ************************
\ *S Service specific data
\ ************************
\ *P These data definitions are required by each server task. The
\ ** data is allocated at the start of the task and released when
\ ** the task is *\fo{TERMINATE}d. The chain *\fo{SVchain} links
\ ** all the service tasks. Each task has *\fo{USER} variables
\ ** *\fo{MY_SOCKET} and *\fo{MY_HSOCKET} which hold the socket
\ ** address and socket number. From these, the service specific
\ ** data can be found.

\ *P The first part of a service data area is common to all
\ ** services, and additional fields can be added as required.
\ ** See *\I{TELNET.FTH} and *\i{HTTP.FTH} for examples. The
\ ** common data area includes service managment data and facilities
\ ** for routing input and output through the normal Forth *\fo{KEY}
\ ** and *\fo{EMIT} wordset.

variable SVchain	\ -- addr
\ *G Holds the tail of the service chain.

#64 equ /SVOB		\ -- n
\ *G Size of a service output buffer. A value of 64 or 128 bytes
\ ** is sufficient to provide satisfactory output for Telnet.

#64 equ /SVIB		\ -- n
\ *G Size of a service input buffer.

#256 equ /SVtib	\ -- n
\ *G Size of the service's TIB area.

struct /SVdata	\ -- len
\ *G Defines the common data in a service specific data area.
\ *[
  int SVlink			\ link to previous service
  				\ N.B. MUST BE FIRST
  int SVtask			\ task that runs this service
  int SVsk#			\ socket# used by this service
  int SVsendflags		\ TCP override flags for lower layers
  int SVdone			\ set nz to close service
  int SVappClean		\ *sv -- *sv ; application specific clean up xt
dup equ /CSVdata		\ size of core service data
\ end of core structure
  /SVtib field SVtibBuff	\ service TIB, e.g. for Telnet
dup equ /MSVdata		\ size of core+TIB service data
\ end of core plus input buffer
SVlowRAM? 0= [if]
  int #SVOB			\ number of characters in SVOB ; SFP002
  /SVOB field SVopBuff		\ service output buffer ; SFP002
  int #SVIB			\ number of characters in SVIB
  int ^SVIB			\ offset of next character in SVIB
  /SVIB field SVipBuff		\ service raw input buffer
[then]
end-struct
\ *]

\ *P The field *\fo{SVappClean} contains the xt of a word
\ *C   cleanFTP ( *sv -- *sv )
\ *P that frees any aqdditional resources that are not released
\ ** by free the service data area. For an example, see *\fo{cleanFTP}
\ ** in SERVICES/FTP.fth.

XC? [if] compiler [then]
: MySVD		\ -- addr
\ *G Returns the address of the task's service data.
  my_socket @ so_*service @
;
XC? [if] target [then]

: SVTib		\ -- tib
\ *G Returns the address of the TIB buffer for this task.
  MySVD SVtibBuff
;

: SVdone?	\ -- flag
\ *G Return true if the socket can be closed.
  MySVD SVdone @
;

: SVbye		\ --
\ *G Set the *\fo{SVdone} exit flag.
  MySVD SVdone on
;


\ ********************
\ *S Server assistance
\ ********************

: CheckSV	\ -- flag
\ *G Run by the service tasks to check whether the service task should
\ ** be closed. Flag is returned true if the service task should be closed.
  my_hsocket @ get_socket_addr if	\ validity check
    my_socket @ if			\ safety check
      my_socket @ so_*cb @ if		\ has control block?
        my_socket @ so_*service @ if	\ has service block
          my_socket @ so_*service @ SVdone @ 0= if
            my_socket @ TcpState@ TCPS_ESTABLISHED = if  \ ESTABLISHED?
              false exit
            endif
          endif
\          SVbye
        endif
      endif
    endif
  endif
  true
;


\ ********************************
\ *S Service KEY, EMIT and friends
\ ********************************

$0A equ AcceptChar	\ -- char
\ *G The character returned by*\fo{SVkey} below when an error
\ ** has occurred. This should be the character that
\ ** *\fo{SVaccept} uses as a line terminator, usually LF.


\ ==================
\ *N Low RAM version
\ ==================
\ *P In this version all buffering is performed in the socket
\ ** layer. This saves RAM at the cost of performance. On a local
\ ** area network the penalty may be significant - it certainly
\ ** is for Telnet.

SVlowRAM? [if]

: SVkey?	\ -- flag
\ *G Return true if a character is avilable from the service's
\ ** client or an error has occurred.
  CheckSV if  -1 exit  endif
  my_hsocket @ pollSocket		\ no PAUSE in poll
  dup SOCKET_ERROR =			\ so do not need CHECKSV again
  if  SVbye  endif
;

: SVkey 	\ -- char ; receive char
\ *G Return a character from the service's client.
\ ** If the service's *\fo{SVdone} flag is set, an LF is
\ ** returned.
  { | char[ cell ] -- ch }
  begin  SVkey? 0= while  pause  repeat
  CheckSV				\ SFP005
  if  pause  AcceptChar  exit  endif	\ SFP014
  my_hsocket @ char[ 1 0 recv SOCKET_ERROR =
  if  pause  AcceptChar  exit  endif	\ SFP014
  char[ c@
;

: SVtype	\ caddr len --
\ *G Send a string/buffer to the service's client.
  CheckSV if  2drop exit  endif
  my_hsocket @ -rot 0 send
  SOCKET_ERROR = if			\ may PAUSE ; RMS002 SFP005
    CheckSV 0=
    if  SVbye  endif
  endif
;

: SVemit	\ char --
\ *G Send a character to the service's client.
  { | char[ cell ] -- }
  char[ c!  char[ 1 SVtype		\ SFP005
;

l: crlf		\ -- caddr
  #13 c,  #10 c,

: SVcr		\ --
\ *G Send a CR/LF pair to the service's client.
  crlf 2 SVtype
;

: SVflushOP	\ --
\ *G A compatibility *\fo{NOOP} in the low RAM version.
;

[then]


\ ===========================
\ *N High performance version
\ ===========================

SVlowRAM? 0= [if]

: SVIBuffer	\ -- addr
\ *G Returns the address of the input buffer for this task.
  MySVD SVipBuff
;

: SVOBuffer	\ -- addr
\ *G Returns the address of the output buffer for this task.
  MySVD SVipBuff
;

: isSVInput	\ --
\ *G Read any available characters from the incoming TCP stream
\ ** into the service input buffer.
  my_hsocket @ pollsocket 0 > if
    my_hsocket @			\ socket
    SVIBuffer /SVIB 0	 		\ *buffer len flags
    recv	 			\ go try to receive some data
    dup SOCKET_ERROR =			\ invalid?
    if  drop 0  endif			\  treat as 0
    my_socket @ so_*service @		\ -- n *sd
    0 over ^SVIB !			\ set offset to zero
    #SVIB !				\ set number of characters
  endif
;

: (SVkey?)	\ -- flag ; check receive char
\ *G Return true if a character has been received by the server
\ ** or the service must be closed.
  CheckSV				\ check for closing
  if  true  exit  endif
  my_socket @ so_*service @		\ addr
  dup #SVIB @ over ^SVIB @ =		\ empty or all read?
  if  isSVinput  endif			\ so try for more
  dup #SVIB @ swap ^SVIB @ -		\ no of chars to be read
;

: SVflushOP	\ -- ; SFP002
\ *G Send current buffer if not empty by passing it to the
\ ** socket. *\fo{SVflushOP} is used by *\fo{SVkey?} and
\ ** *\fo{SVkey} so that pending output is transmitted. If
\ ** your code does not call either of these, e.g. through the
\ ** Generic I/O, you should add *\fo{SVflushOP} to your code
\ ** where appropriate.
  CheckSV if  exit  endif
  MySVD dup #SVOB @ if
    >r
    my_hsocket @			\ socket to send to
    r@ SVopBuff  r@ #SVOB @		\ buffer and length
    0 send drop				\ send with no flags
    0 r> #SVOB !			\ reset buffer count
    pause
  else
    drop
  endif
;

: (SVemit)	\ char -- ; SFP002
\ *G Send a character to a client of this service.
  MySVD tuck				\ -- *sv char *sv
  dup #SVOB @ swap SVopBuff + c!	\ stash character
  dup #SVOB @ 1+ tuck			\ -- #chars *sv #chars
  swap #SVOB !				\ stash updated count
  /SVOB >=				\ if full send buffer
  if  SVflushOP  endif
;


\ =======================
\ *H Service vectored I/O
\ =======================

: SVkey?	\ -- flag ; check receive char
\ *G Return true if a character has been received by the server.
\ ** Any pending output characters are sent.
  SVflushOP  (SVkey?)
;

: SVkey 	\ -- char ; receive char
\ *G Return a character from the service's client. Any pending
\ ** output characters are sent first. If the service's *\fo{SVdone}
\ ** flag is set or a socket error has occurred, an LF character
\ ** is returned.
  SVflushOP
  begin  (SVkey?) 0= while  pause  repeat
  SVdone?
  if  pause  AcceptChar  exit  endif	\ SFP014
  my_socket @ so_*service @
  dup SVipBuff over ^SVIB @ + c@	\ -- *sv char
  swap ^SVIB incr
;

: SVemit	\ char -- ; emit char
\ *G Send a character to a service's client.
\  SVdone?				\ check for closing
  CheckSV
  if  pause  drop  exit  endif		\ SF014
  (SVemit)
;

: SVtype	\ caddr len -- ; display string
\ *G Send a string to a service's client
  bounds ?do  i c@ SVemit  loop
;

: SVcr		\ -- ; display new line
\ *G Send a new line sequence to a service's client
  $0D SVemit  $0A SVemit
;

[then]


\ =====================
\ *N Generic I/O device
\ =====================

create ConsoleSV	\ -- addr ; OUT managed by upper driver
\ *G Function despatch table for service I/O.
\ ** *\fo{OUT} is managed by the upper level driver.
  ' SVkey ,		\ -- char ; receive char
  ' SVkey? ,		\ -- flag ; check receive char
  ' SVemit ,		\ -- char ; display char
  ' SVtype ,		\ caddr len -- ; display string
  ' SVcr ,		\ -- ; display new line

: Console=SV	\ --
\ *G Select the service I/O as the console.
  ConsoleSV dup opvec ! ipvec !
;

: Init-ConsoleSV	\ --
\ *G Initialise for console I/O by the service. Note that the
\ ** service's socket must have been set up and and the private
\ ** service area initialised.
  MySVD
[ SVlowRAM? 0= ] [if]
  0 over #SVOB !			\ initialise o/p buffer
[then]
  SVtibBuff 'tib ! 			\ initialise TIB
  #tib off
  echoing off  xon/xoff off		\ SFP015
;


\ ==========================
\ *N Service console support
\ ==========================

: +SV_responsive	\ --
\ *G Set the socket to use the TCP_PSH flag when sending
\ ** to improve interactivity.
  CheckSV 0= if
    my_socket @ so_*cb @ t_override
    dup w@  TCP_PSH or  swap w!
  endif
;

: -SV_responsive	\ --
\ *G Set the socket not to use the TCP_PSH flag when sending.
  CheckSV 0= if
    my_socket @ so_*cb @ t_override
    dup w@  TCP_PSH invert and  swap w!
  endif
;

: SV_responsive?	\ -- res
\ *G Return non-zero if the socket's TCP_PSH flag override is set.
  CheckSV if  0 exit  endif
  my_socket @ so_*cb @ t_override
  w@ TCP_PSH and
;

\ SFP010...
: SVaccept	\ c-addr +n1 -- +n2 ; read up to LEN chars into ADDR
\ *G Read a string of maximum size n1 characters to the buffer at
\ ** c-addr, returning n2 the number of characters actually read.
\ ** Input is terminated by LF, and CR is ignored. This satisfies
\ ** the requiresments of DOS, Windows, Unices and the TCP/IP NVT
\ ** (Network Virtual Terminal).
\ ** If *\fo{ECHOING} is non-zero, characters are echoed.
\ ** If *\fo{XON/XOFF} is non-zero, an XON character is sent at the
\ ** start and an XOFF character is sent at the the end.
  dup 0=				\ check for pathological case
  if  nip exit  then
  xon/xoff @ if  XON emit  then         \ enable server
  0 -rot bounds                         \ form address limits
  ?do
    case SVkey				\ get character
      8 of ?bs               endof	\ BS
      9 of bl i save-ch 1+ 1 endof	\ TAB, convert to space
      $0D of  CheckSV ?leave  0  endof	\ CR, ignore
      $0A of bl i save-ch  leave  endof	\ LF, terminator
      $0C of bl i save-ch  leave  endof	\ FF, treat like CR
      $7F of ?bs               endof	\ DEL
             i save-ch  1+ 1 0		\ stash char, echo it?
    endcase
  +loop                                 \ round again
  xon/xoff @ if  XOFF emit  then        \ disable server
;

: SVquery	\ -- ; fetch line into TIB
\ *G Reset the input source specification to the console and
\ ** accept a line of text into the input buffer.
\  0 to source-id                        \ mark source as keyboard
  'tib @ /SVtib 2- SVaccept #tib !	\ "tib-len 1-" was #80
  >in off
;
\ ...SFP010


\ ********************************
\ *S Service creation and deletion
\ ********************************

: isMySocket	\ hs --
\ *G Set up *\fo{USER} variables for this socket.
  dup my_hsocket !  get_socket_addr my_socket !
;

\ SFP007...
: NoSocket	\ --
\ *G Zero the *\fo{USER} variables *\fo{MY_HSOCKET} and *\fo{MY_SOCKET}.
  0 my_hsocket !  0 my_socket !
;

: ShutErrSocket	\ -- socket_error
\ *G Close the current server socket and return *\fo{SOCKET_ERROR}.
\ ** All allocated socket and service memory is released by the
\ ** close.
  my_hsocket @ CloseSocket drop
  NoSocket  SOCKET_ERROR
;

: InitServerSocket	{ #service /data #port -- res }
\ *G Create a listening socket and the private data for
\ ** a TCP service conversation. On success, the *\fo{USER}
\ ** variables *\fo{MY_HSOCKET} and *\fo{MY_SOCKET} contain the
\ ** socket number and socket address. On failure, these variables
\ ** contain zero.
\ *D #service  service type number.
\ *D /data     size of the service private data area, at least /SVDATA.
\ *D #port     port number to listen on.
\ *D res       socket number (1..n) or SOCKET_ERROR.

\ Create a socket
  NoSocket
  TCPsocket ?dup 0=			\ SFP017
\  AF_INET SOCK_STREAM IPPROTO_TCP socket ?dup 0=
  if  SOCKET_ERROR exit  endif
  isMySocket				\ SFP013

\ Initialise service specific data.
  #service my_socket @ so_service !	\ set service type number
  /data my_socket @ so_/service !	\ set service data length

\ initialise and bind the socket
  my_hsocket @ AF_INET #port INADDR_ANY bindTo SOCKET_ERROR =
  if  ShutErrSocket exit  endif
  my_hsocket @ Listen
  if  ShutErrSocket exit  endif
  my_hsocket @
;
\ ...SFP007

: SVinitiate	\ xt -- task|0
\ *G Allocates memory for a server task and *\fo{INITIATE}s it with the
\ ** given xt, returning the task's TCB address. If memory
\ ** cannot be allocated, zero is returned.
  task-size allocate if
    cr ." Cannot allocate task memory"
    2drop 0
  else					\ -- xt task
    dup task-size erase			\ wipe data area
    task-u0 +				\ user area=TCB=task id
    tuck initiate			\ start task
  endif
;

: (SVterminate)	\ task --
\ *G Terminates a server or service task and frees off the task
\ ** memory. Does not *\fo{PAUSE}
  dup sub-task				\ remove task from task chain
  task-u0 - free drop			\ free task data
;

: SVterminate	\ task --
\ *G Terminates a server or service task and frees off the task
\ ** memory.
  (SVterminate) pause
;


\ *************************
\ *S Service listening task
\ *************************

: .SVmessage	\ caddr len --
\ *G Display a message to the current output (usually the console)
\ ** in the form:
\ *C   <service> <given text> on socket <n>
\ *P The *\fo{USER} variables *\fo{MY_SOCKET} and *\fo{MY_HSOCKET}
\ ** must contain valid data.
  base @ >r  decimal
  cr  my_socket @ so_service @ .service
  type ."  on socket "  my_hsocket @ .
  r> base !
;

: ServiceCreate	\ #service /data port# --
\ *G Create a new service listening socket.
  begin
    3dup InitServerSocket SOCKET_ERROR =
   while
    cr  2 pick .service ." initialisation failed "
    #5000 ms
  repeat
  2drop drop
  s" listening" .SVmessage
;

: StartService	\ hs xt --
\ *G Initialise and launch a new service task on socket *\i{hs}.
\ ** The action of the new task is given by *\i{xt}
\ ** given by xt.
\ [con  cr over . ." Starting service"  con] 2 ms
  swap  task-size allocate if		\ -- xt hs addr
    2drop  closeSocket drop		\ --
\    [con  cr ." Cannot allocate service task, connection dropped"  con] 10 ms
    exit
  endif					\ -- xt hs addr
  dup task-size erase			\ wipe data area
[defined] SetupTCB [if]			\ needed by PFDv3 to use common code
  dup SetupTCB				\ use addr as TCB and set pointers in TCB
[then]
  task-u0 +				\ -- xt hs task ; user area=TCB=task id
  2dup my_hsocket his !			\ set new task's socket#
  swap get_socket_addr			\ -- xt task *sk
  2dup swap my_socket his !		\ set new task's socket address
  so_*service @ 2dup SVtask !		\ -- xt task SVdata ; identify task
  SVchain AddLink			\ link SVdata into service chain
  initiate				\ assign action and start task
\ [con  cr ." service started"  con] 2 ms
;


\ ************************
\ *S Service support tools
\ ************************

: waitSocketSent	\ hs --
\ *G Wait until all transmit data for a socket has been sent and
\ ** acked.
  get_socket_addr dup if		\ good socket number
    so_*cb @ dup if
      begin				\ -- *cb
        dup #unsent  over #unacked or
        over t_state w@ TCPS_ESTABLISHED = and
       while
        pause				\ needed to flush data
      repeat
    endif
  endif
  drop
;

: wait-socket-empty	\ --
\ *G Wait until all output has been sent from the socket.
  my_hsocket @ waitSocketSent
;

: SVdisconnect	\ --
\ *G *\fo{Disconnect} the current service socket and run *\fo{SVbye}.
  wait-socket-empty
  my_hsocket @ disconnect drop
  SVbye
;

: SVstartup	\ --
\ *G Performs the default actions when a service starts.
\ ** I/O is set to the console, *\fo{BASE} to decimal, and
\ ** a console message is issued.
  ConsoleIO decimal
  s" started" .SVmessage
;

: SVshutdown	\ --
\ *G Performs actions required when a service finishes.
  ConsoleIO decimal
  s" halting" .SVmessage
  50 ms
[ SVsingle? ] [if]
  SVdisconnect
[else]
  SVdone? 0= if
    SVdisconnect
  endif
  begin  pause  again
[then]
;

: SRVRstartup	\ caddr len --
\ *G The first action a server task should perform. The string
\ ** is displayed on the console with a startup message.
  ConsoleIO decimal			\ default debug I/O device
  WaitForIPaddress			\ SFP007
  cr type ."  server started"
;


\ *****************
\ *S Service output
\ *****************
\ *P In order to avoid race conditions, a separate task handles
\ ** testing whether a service should be closed.

variable SVkillChain	\ -- addr
\ *G Holds the tail of service tasks to be destroyed.

XC? [if]
: ?SVkill	\ *SVdata --
\ *G If the service has a task, i.e. is a service rather than a
\ ** server, move it to the kill chain, otherwise just free the
\ ** memory.
[else]
: rt?SVkill	\ *SVdata --
[then]
  dup SVtask @ if			\ if task is active, i.e. service is running
    dup SVchain DelLink			\ remove from service chain
    SVkillChain AddLink			\ add to kill chain
  else
    free drop				\ no task, it's a server, so just free it.
  endif
;
XC? 0= [if]  assign rt?SVkill to-do ?SVkill [then]

: ?ServiceClose	\ --
\ *G Close any service tasks with a non-zero *\fo{SVDONE} field
\ ** in the private service area.
  SVchain @
  begin					\ -- curr
    dup
   while
    dup SVdone @ if
\ cr ." ?SVC on socket " dup SVsk# @ . cr
      dup @ swap			\ -- next cur
      dup SVchain DelLink		\ remove from service chain
      SVsk# @ CloseSocket drop		\ clean up socket, will put on kill chain ; SFP008
    else
      @
    endif				\ -- next
  repeat
  drop
;

\ SFP008...
: ?ServiceKill	\ --
\ *G Kill any service tasks that are on the kill chain and free
\ ** the task and service memory. The kill chain is extended by
\ ** *\fo{CloseSocket}. The service task clean up action (the xt
\ ** in the *\fo{SVappClean} field) is run before memory is freed.
  SVkillChain @
  begin
    dup
   while
\ cr ." ?SVK on service " dup .dword cr
    dup @ swap				\ -- next curr
    dup SVkillChain delLink		\ remove from chain
    dup SVappClean @ if			\ if there's a clean up action
      dup SVappClean @ execute		\ run it
    endif
    dup SVtask @ ?dup if		\ if task is defined
\ ."  task " dup .dword cr
      SVterminate			\ kill it and free it
    endif
    free drop				\ free service memory
  repeat
  drop
;

\ *P The field *\fo{SVappClean} contains the xt of a word
\ *C   cleanFTP ( *sv -- *sv )
\ *P that frees any aqdditional resources that are not released
\ ** by free the service data area. For an example, see *\fo{cleanFTP}
\ ** in SERVICES/FTP.fth.

: ServiceIO
\ *G Handle any queued output - called in the service task.
  ?ServiceClose ?ServiceKill
;
\ ...SFP008


\ **************
\ *S Diagnostics
\ **************
\ *P Diagnostic code is only compiled if the *\fo{EQU}ate *\fo{DIAGS?}
\ ** is set non-zero.

diags? [if]

[then]


\ ******
\ *> ###
\ ******

