\ Telnet.fth - Telnet Server

((
Copyright (c) 1997-2003, 2006, 2008
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999
and
  CEM Systems Ltd, Belfast, 2000-2002
  Hiden Analytical, 2002

To do
=====

Change history
==============
20081124 SFP007	Added wait for IP address.
20080603 SFP006 Changed to use SVACCEPT and SVQUERY.
		Added more error checking.
20080331 SFP005 Enhanced service shut down checking.
20080222 SFP004 Added check for listen failure.
20061113 MGS003 Fixed missing @ in .TELNETCHAIN.
20060706 RMS002 Fail socket if transmission fails
20020416 SFP001 Generic server code moved to SERVERS.FTH and
		refactored.
))

decimal


\ =========
\ *! telnet
\ =========
\ ****************
\ *T Telnet Server
\ ****************
\ *P The Powernet Telnet server is a multithreaded server
\ ** that can accept multiple Telnet connections. Local
\ ** echo is not required on the client.
\ ** The model used is described in *\i{SERVERS.FTH}.

\ *P For testing, please be aware that the standard Windows
\ ** Telnet client is very slow. A nuch faster alternative is
\ ** HyperTerminal Personal Edition from:
\ *C   http://www.hilgraeve.com


\ ***********************
\ *S Telnet specific data
\ ***********************
\ *P These data definitions are required by each Telnet server task. The data
\ ** is allocated at the start of the task and released when the task is
\ ** *\fo{TERMINATE}d. The chain *\fo{SVCHAIN} links all the service tasks.

((
#5023 equ TelnetPort#	\ -- n ; standard is 23
\ *G Define the port used for the Telnet server. The standard
\ ** port is 23, but 5023 is the default set for PowerNet as
\ ** most application Telnet servers are private.
\ ** Moved to *\i{PNconfig.fth}.
))


\ ***************
\ *S IAC handling
\ ***************
\ *P Not yet implemented.

((
\ define some Telnet char codes
#240 equ TNET_SE	\ Subnegotiation end
#241 equ TNET_NOP	\ No-op
#242 equ TNET_DM	\ Data Mark
#243 equ TNET_BRK	\ Break
#244 equ TNET_IP	\ Interrupt Process
#245 equ TNET_AO	\ Abort Output
#246 equ TNET_AYT	\ Are you there ??
#247 equ TNET_EC	\ Erase char
#248 equ TNET_EL	\ Erase Line
#249 equ TNET_GA	\ Go ahead
#250 equ TNET_SB	\ following is sub-negotiation
#251 equ TNET_WILL	\ option command accept
#252 equ TNET_WONT	\ option command refuse
#253 equ TNET_DO	\ Request to DO an option
#254 equ TNET_DONT	\ Request to NOT DO an option
#255 equ TNET_IAC	\ Interrupt escape char

: DoEscapeSequence	\ addr -- n
\ *G Process a character sequence with the TNET_IAC escape
\ ** character at addr, returning n = decoded length.
\ ** At present these sequences are ignored.
  drop 1
;

: ParseTelnetBuffer	\ numchars --
\ *G Process incoming Telnet data.
  SVipBuff swap bounds ?do		\ MUST handle these char-by-char
    i c@ TNET_IAC = if
      i DoEscapeSequence		\ skips n chars, returning n
    else
      i c@ *KeyQueue @ (>cqueue)	\ go put char into buffer
      1
    endif
  +loop
;
))


\ **********************
\ *S Telnet vectored I/O
\ **********************
\ *P The Telnet server establishes its own generic I/O
\ ** based on that in *\i{SERVERS.FTH} in order to handle IAC
\ ** processing in the future.

create ConsoleTN	\ -- addr ; OUT managed by upper driver
\ *G Function despatch table for Telnet I/O.
\ ** *\fo{OUT} is managed by the upper level driver.
  ' SVkey ,		\ -- char ; receive char
  ' SVkey? ,		\ -- flag ; check receive char
  ' SVemit ,		\ -- char ; display char
  ' SVtype ,		\ caddr len -- ; display string
  ' SVcr ,		\ -- ; display new line

: Init-ConsoleTN	\ --
\ *G Initialise for console I/O by Telnet. Note that the
\ ** Telnet socket must have been set up and and the private
\ ** service area initialised.
  Init-ConsoleSV
;

/SVdata equ /TNdata	\ -- len
\ *G Required size of console service area.

: Console=Telnet	\ --
\ *G Select Telnet as the console.
  ConsoleTn dup opvec ! ipvec !
;


\ ***********************
\ *S Telnet service tasks
\ ***********************

: tn_announce	\ --
\ *G Issues Telnet signon message.
  cr cr .PNET ."  Telnet server"
  cr ." Use TNBYE to disconnect"
;

: (tn_login)	\ -- ior
\ *G Perform the default login procedure and return non-zero
\ ** if successful so that the return code can be used as a
\ ** user identifier by higher level code if required
  cr cr ." Add secure login to TN_LOGIN"
  #100 ms
  -1
;

defer tn_login	\ -- ior
\ *G Perform the assigned login procedure and return non-zero
\ ** if successful so that the return code can be used as a
\ ** user identifier by higher level code if required. When
\ ** zero is returned for unsuccessful login, the Telnet
\ ** session is terminated. See *\fo{TelnetService}.
  assign (tn_login) to-do tn_login

: tn_quit	\ --
\ *G Empty the return stack, store 0 in *\fo{SOURCE-ID}, and
\ ** enter interpretation state. *\fo{TN_QUIT} repeatedly
\ ** *\fo{SVaccept}s a line of input and *\fo{INTERPRET}s it, with
\ ** a prompt if interpreting and *\fo{ECHOING} is on. Note that
\ ** any task that uses *\fo{TN_QUIT} must initialise *\fo{'TIB},
\ ** *\fo{BASE}, *\fo{IPVEC}, and *\fo{OPVEC}. Note that
\ ** *\fo{TN_QUIT} clears the stack on exit.
[ blocks? ] [if]
  blk off
[then]
  xon/xoff off  echoing on              \ No Xon/Xoff, do Echo
  0 to source-id  [compile] [		\ set up
  begin
    +sv_responsive			\   immediate output
    echoing @ if  cr  then              \   if echoing enabled issue new line
    SVquery				\   get user input
    -sv_responsive			\   delayed output
    CheckSV 0=				\ in case closed ; SFP006
   while				\ SFP006
    ['] interpret catch ?dup 0= if	\   interpret line
      state @ 0=  echoing @ 0<>  and if	\   if interpreting & echoing
        ."  ok"  depth ?dup		\     prompt user
        if  ." -" .  then
      then
    else
      .throw  s0 @ sp!  [compile] [	\    display error, clean up
      cr source type			\    display input line
      cr >in @ 1- spaces ." ^"		\    display pointer to error
    then
    ( SVdone? ) CheckSV		        \ Until can be closed, SFP005
  until then				\ do next line ; SFP006
  s0 @ sp!				\ clear stack
;

: TNbye		\ --
\ *G Use this to close the connection in Telnet.
  my_hsocket @ disconnect drop
;

: TelnetService	\ --
\ *G The Telnet service task launched for each established
\ ** Telnet connection.
  SVstartup
  Init-ConsoleTN Console=Telnet
  #100 ms
  TN_announce				\ Powernet sign on
  TN_login if				\ log in, nz for success
    >in off  tn_quit			\ go run the Forth loop
    s0 @ sp!				\ clean up the data stack
  endif
  SVshutdown
;


\ ************************
\ *S Telnet listening task
\ ************************

: TNserverPass	\ --
\ *G One iteration through the Telnet server.
  my_hsocket @ >r			\ preserve listening socket#
  begin
    r@ isMySocket
    my_hsocket @ 0 0 Saccept		\ wait for connection or failure
    dup SOCKET_ERROR <>
   while
[ SVsingle? ] [if]
    isMySocket  TelnetService
[else]
    ['] TelnetService StartService	\ launch another service
[then]
  repeat
  drop  r> closeSocket drop		\ socket error
;

TelnetPort# [if]	\ fixed port number

: TelnetServer	\ -- ; stay here forever
\ *G The Telnet listening task.
  s" Telnet" SRVRstartup
  begin
[ diags? ] [if]  ?StackEmpty  [then]	\ check stack depth
    service_telnet /TNdata TelnetPort# ServiceCreate
    TNserverPass  pause
  again
;

[else]

0 value DynTnPort#

: TelnetServer	\ -- ; stay here forever
\ *G The Telnet listening task.
  s" Telnet" SRVRstartup
  begin
[ diags? ] [if]  ?StackEmpty  [then]	\ check stack depth
    begin  DynTnPort# 0= while  #50 ms  repeat
    service_telnet /TNdata DynTnPort# ServiceCreate
    TNserverPass  pause
  again
;

[then]

0 value TelnetTask	\ -- 0|task
\ *G Returns 0 or the Telnet server task if running.

: RunTelnetSockTask	\ --
\ *G Start the Telnet server task.
  TelnetTask 0= if
    ['] TelnetServer SVinitiate to TelnetTask
  endif
;

: StopTelnet	\ --
\ *G Stop the Telnet server.
  TelnetTask if
    TelnetTask SVterminate  0 -> TelnetTask
    cr ." Telnet server stopped"
  endif
;


\ **************
\ *S Diagnostics
\ **************
\ *P This code is only compiled if the *\fo{EQU}ate *\fo{DIAGS?}
\ ** is set non-zero.

diags? [if]

: .TelnetChain	\ --
\ *G Display data about the Telnet sockets.
   TelnetTask 0= if
     cr ." Telnet not running"  exit
   endif
   cr ." Listening on socket " TelnetTask my_hsocket his @ .
   cr ." Serving on sockets "
   SVchain
   begin
     @ dup
    while
     dup SVsk# @ get_socket_addr	\ @ missing ; MGS003
     so_service @ service_telnet =
     if  dup SVsk# @ .  endif
   repeat
   drop
;

: .tn		\ -- ; synonym
\ *G Display data about the Telnet sockets.
  .TelnetChain
;

[then]


\ ******
\ *> ###
\ ******

