\ SLIPCOM.FTH - definitions for a Forth Comms port SLIP Interface

((
Copyright (c) 1997-2003, 2008
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999
and:
  CEM Systems Ltd, Belfast, 2000-2001

To do
=====

Change history
==============
20080307 SFP003 Changed FLUSH to FLUSHDEBUG to avoid name conflict.
20030313 SFP002 Moved SlipPortTask from POWERNET.BLD
20010503 SFP001 Timeouts handled by v6.1 timebase mechanism
))


\ ==========
\ *! slipcom
\ *T SLIP interface
\ ==========
\ *P The SLIP interface is handled by a task using a
\ ** state machine. See SLIP\SLIPCOM.FTH.


\ ***************
\ *S SLIP equates
\ ***************

$C0 equ FRAME_END		\ -- char
\ *G Frame end character.
$DB equ FRAME_ESCAPE		\ -- char
\ *G Frame escape character.
$DC equ TRANS_FRAME_END		\ -- char
\ *G Trans Frame end character.
$DD equ TRANS_FRAME_ESCAPE	\ -- char
\ *G Trans Frame escape character.

#1000 equ SLIP_RX_TIMEOUT	\ -- ms
\ *G Receive timout in milliseconds.


\ ***********************
\ *S Slip input functions
\ ***********************

: (gotSlipChar)	\ char --
\ *G Char is saved in the input buffer.
  rxspace @ if
    *rxnow @ c!				\ so add to buffer
    1 *rxnow +!				\ bump pointer for next
    1 rxcount +!			\ also bump bytes rx'd
    -1 rxspace +!			\ and dec space left
  else
\    cr ." Slip Rx buffer full!"
    drop
  endif
;

: gotTransFrameEsc	\ --
\ *G Received a TRANS_FRAME_ESCAPE.
  rxstate @ 2 = if
    FRAME_ESCAPE 			\ save char as frame_esc
    1 rxstate !
  else
    TRANS_FRAME_ESCAPE 			\ save char as trans_frame_esc
  endif
  (gotSlipChar)
;

: gotTransFrameEnd	\ --
\ *G Received a TRANS_FRAME_END.
  rxstate @ 2 = if
    FRAME_END 				\ save char as frame_end
    1 rxstate !
  else
    TRANS_FRAME_END 			\ save char as trans frame_end
  endif
  (gotSlipChar)
;

: gotFrameEsc	\ --
\ *G Received a FRAME_ESCAPE.
  rxstate @ $FF <> if
    2 rxstate !				\ just set the escape flag
  endif
;

: StartOfPacket	\ --
\ *G Set up for a new packet.
  GetPbuf ?dup if			\ grab a buffer
    >r
    r@ *rxpbuf !			\ save this buffer
    r@ p_dat ether_hdr + *rxnow !	\ and set pointer to start of data
    0 p_dat ether_hdr +
    dup r@ p_data w!			\ start of data
    r@ p_pkthdr w!			\ start of packet
    rxcount off				\ start with no chars
    PSIZE rxspace !			\ & set max space available
    1 rxstate !				\ so set receiving state
    r> drop
  else					\ memory allocate failed
    *rxpbuf off
    rxcount off				\ start with no chars
    rxspace off				\ & set max space available
    $FF rxstate !			\ so set receiving state
    ." No Slip Rx Buffer space!!"
  endif
;

: EndOfPacket	\ --
\ *G We got the end of a packet.
  *rxpbuf @ if
    rxcount @ *rxpbuf @ p_len w!	\ set buffer length
    rxcount @ *rxpbuf @ p_datalen w!	\ and data length
    *rxpbuf @ PBHdrStart		\ start of ip packet
    ip_src @(n) add_route ?dup if	\ add ip address to routing table
      SLIP_MASK 1 or over iproute_port w! \ set port identifier
\      ROUTE_LIFE swap iproute_life w!	\ restart lifetime
      ROUTE_LIFE swap SetRouteLife	\ restart lifetime
    endif
\    ." Slip in"
    *rxpbuf @ RxIpPacket		\ now go handle the incoming packet
  endif
  rxstate off				\ and clean up current state
;

: gotFrameEnd	\ --
\ *G We received a FRAME_END character.
  rxstate @ if				\ this should be the end of a packet
\    0 portid @ set-port-timer		\ stop pkt timeout
    RxExpSlip off			\ stop receive timer
    EndOfPacket
  else					\ this must be the start of a packet
\    #10 portid @ set-port-timer	\ start pkt timeout (#10 * 100ms = 1 sec)
    SLIP_RX_TIMEOUT later 1 or RXExpSlip !	\ start receive packet timeout
    StartOfPacket
  endif
;

: GotSlipChar	\ char --
\ *G Handle an incoming SLIP char
\  dup .hex
  case
    TRANS_FRAME_ESCAPE of  gotTransFrameEsc  endof
    TRANS_FRAME_END    of  gotTransFrameEnd  endof
    FRAME_ESCAPE       of  gotFrameEsc       endof
    FRAME_END          of  gotFrameEnd       endof
			   dup (gotSlipChar)
  endcase
;

: GetSlipChars	\ --
\ *G Receive characters from a SLIP port until all are in.
  begin
    key?
  while
    key GotSlipChar
  repeat
;


\ ************************
\ *S Slip output functions
\ ************************

: Send>Slip	{ *pb *rt | ln -- int }
\ *G Send the data in the supplied PBuf (chain).
\ *D *PB the first pbuf in the chain (only 1 allowed)
\ *D *RT points to the required routing table entry
\ *D int number of bytes sent if OK else SOCKET_ERROR.
  *pb p_len w@ -> ln			\ set length as default
  *pb
  *rt iproute_port w@ $000F and		\ the destination serial port
  case
    1 of  TxSlipQ enqueue  endof
          swap FreeQB
          0 -> ln			\ set length as 0 (error)
  endcase
  ln
;

: (SendSlipChar)	\ char --
\ *G Sends the char to the SLIP port.
  emit
;

: send_frame_esc_char	\ --
\ *G Sends the escape sequence to the SLIP port.
  FRAME_ESCAPE (SendSlipChar)
  TRANS_FRAME_ESCAPE (SendSlipChar)
;

: send_frame_end_char	\ --
\ *G Sends the escape sequence to the SLIP port.
  FRAME_ESCAPE (SendSlipChar)
  TRANS_FRAME_END (SendSlipChar)
;

: SendSlipChar	\ char --
\ *G Sends the character to the SLIP port, processing
\ ** escape characters.
\  1 txcount +!				\ bump the counter
  case
    FRAME_END		of send_frame_end_char	endof
    FRAME_ESCAPE	of send_frame_esc_char	endof
			(SendSlipChar) 0
  endcase
;

: SlipTx	\ *pbuf --
\ *G Send the pbuf chain out of the slip port.
  FRAME_END (SendSlipChar)			\ send start of frame
  dup PBHdrStart  over PBHdrLen
  bounds ?do
    i c@ SendSlipChar				\ send char by char
  loop
  FRAME_END (SendSlipChar)			\ send end of frame
  flushDebug pause				\ finally queue the buffers ; SFP003
  FreeQB					\ and now we can free up the buffer chain
;


\ ********************
\ *S SLIP support task
\ ********************

 0 value SlipDevice	\ -- addr
\ *G You MUST define a REAL serial port device for SLIP.
\ This will be used to set *\fo{OPVEC} and *\fo{IPVEC}.

: DoSlipPort	\ -- ; task action
\ *G The action of the slip tasks.
  ConsoleIO decimal			\ default debug I/O device
  cr ." SlipPortTask running"
  TXSlipQ initQ				\ initialise SLIP transmit queue
\  *emitNext off				\ clear pointer
\  *emitBuff off				\ clear pointer
\  *keyNext off				\ clear pointer
\  *keyBuff off				\ clear pointer
\  emitmode off				\ non-block mode
  rxstate off				\ start waiting for incoming frame
  rxspace off				\ and start with a clean slate
  *rxnow off
  *rxpbuf off
  rxcount off
  RXExpSlip off				\ clear receive timer
  SlipDevice opvec !			\ set the serial port
  SlipDevice ipvec !

  begin
[ diags? ] [if] ?StackEmpty [then]	\ check stack depth
    RXExpSlip @ if
      RXExpSlip @ timedout? if		\ yes
        *rxpbuf @ ?dup if
          FreeQB			\ free up any buffer
        endif
        RXExpSlip off			\ stop timer
        rxstate off
      endif
    endif

    GetSlipChars			\ receive handler

    TxSlipQ dequeue ?dup if
      SlipTx				\ go format and send it
    endif

    pause
  again
;

task SlipPortTask	\ -- addr
\ *G Slip port task.
cdata

: RunSlipTask	\ --
\ *G Run the SLIP port task. Changes will be necessary
\ ** to support multiple SLIP devices. The technique used
\ ** by the TELNET launcher is appropriate.
\ ** See SERVICES\TELNET.FTH.
  ['] DoSlipPort SlipPortTask initiate	\ assign action
;


\ ======
\ *> ###
\ ======

decimal


