\ SNTP.fth - Contains code for SNTP packet handling

((
Copyright (c) 2007, 2008
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)

Substantial portions of this code have been derived from code
provided by:
Copyright (c) 2002, 2003, 2004
Hiden Analytical Ltd.
Tel:	01925 445225
Fax:	01925 416518

To do
=====
2011mar10 - mgs - from RFC1361
 The following table is a summary of the SNTP client operations.
   There are three recommended error checks shown in the table. In all NTP
   versions, if the Leap Indicator field is 3 or the Transmit Timestamp
   is zero (unsynchronized), the server has never synchronized or not
   synchronized to a valid timing source within the last 24 hours. If
   the Stratum field is 0 (unspecified or unavailable), the server has
   never synchronized, has lost reachability with all timing sources or
   is synchronized by some protocol other than NTP. Whether to believe
   the transmit timestamp or not in this case is at the discretion of
   the client implementation.

      Field Name              Request        Reply
      -------------------------------------------------------------
      Leap Indicator (LI)     0              if 3 (unsynchronized),
                                             disregard
      Version Number (VN)     (see text)     ignore
      Mode                    3 (client)     ignore
      Stratum                 0              if 0 (unspecified),
                                             disregard
      Poll                    0              ignore
      Precision               0              ignore
      Root Delay              0              ignore
      Root Dispersion         0              ignore
      Reference Identifier    0              ignore
      Reference Timestamp     0              ignore
      Originate Timestamp     0              ignore
      Receive Timestamp       0              ignore
      Transmit Timestamp      0              time of day (seconds only);
                                             if 0 (unsynchronized),
                                             disregard
      Authenticator           (not used)     ignore

5. SNTP Server Operations
   The timestamp fields are set as follows. The Reference Timestamp,
   Receive Timestamp and Transmit Timestamp fields are set to the time
   of day at the server. The Originate Timestamp field is copied
   unchanged from the request. The following table summarizes these
   actions.

      Field Name              Request        Reply
      ----------------------------------------------------------
      Leap Indicator (LI)     ignore         0 (normal), 3
                                             (unsynchronized)
      Version Number (VN)     ignore         copied from request
      Mode                    (see text)     (see text)
      Stratum                 ignore         server stratum (1)
      Poll                    ignore         copied from request
      Precision               ignore         server precision
      Root Delay              ignore         0
      Root Dispersion         ignore         0 (see text)
      Reference Identifier    ignore         source identifier or 0
      Reference Timestamp     ignore         time of day or 0
      Originate Timestamp     ignore         copied from request
      Receive Timestamp       ignore         time of day or 0
      Transmit Timestamp      ignore         time of day or 0
      Authenticator           ignore         (not used)


Change history
==============
20110729 SFP008 Default action is to set clock.
20110310 MGS007 Client sends its time in sn.TXTimestamp - returned by server
                as 'Originate'. We have made this correction too often!
20101213 HAR007 previous by CDS.
20101220 SFP006 Changed initialisation in DOSNINIT just to wait
		for manual or DHCP configuration.
20080508 SFP005	Miscellaneous unmarked code tidies.
		Documentation update for seconds wrap in 2036.
20080507 MGS004 Changed default SNTP configuration.
		Removed authenticator option.
20080215 MGS003 Corrected timestamp order and location.
20070829 VRW002 Second release, enhanced chains.
20070821 MPE001 First Issue, based on DHCP.FTH
))

only forth definitions
decimal

\ =======
\ *! sntp
\ *T SNTP client
\ =======
\ *P As of PowerNet v4.30 (8 May 2008) the user interface to the
\ ** SNTP data has changed and is not compatible with the previous
\ ** one.

\ *P The code in *\i{<PNet>\SNTP} implements most of an SNTP *\b{client}
\ ** according to RFC1361 and RFC4330. See also RFC1305. Do not attempt
\ ** to modify this code until you have absorbed these RFCs.

\ *P If you do not have an NTP server on your network, there are
\ ** "public" ones available. See *\i{www.ntp.org} for details.
\ ** Note when testing that advertised public servers are not
\ ** always available. Your ISP probably maintains a functioning
\ ** NTP server.

[undefined] SNTPdebug? [if]
0 equ SNTPdebug?	\ -- x
\ *G Set this non-zero to display SNTP debug information during
\ ** execution.
[then]

[undefined] SNTPactions? [if]
1 equ SNTPactions? 	\ -- x
\ *G Defined the actions performed by default for timestamps.
\ *E  0     Dummy action
\ **  1     Set clock
\ **  ...   user defined action
[then]


\ *****************************
\ *S SNTP equates and structure
\ *****************************

3 equ SNTP_VN		\ Claim version 3 - not quite maximum compatability ; MGS004
0 equ SNTP_LI		\ We're a client - we don't know the LI
3 equ SNTP_CLIENT	\ Mode setting for Client v3, use 0 for VN=1
0 equ SNTP_STRATUM	\ 0 for a client
0 equ SNTP_PRECISION	\ Not used for SNTP, only for NTP

SNTP_LI 6 lshift  SNTP_VN 3 lshift or  SNTP_CLIENT or
  equ SNTP_FLAGS	\ precalculate the sn.flags field below

#123 equ SNTPport#	\ -- port#
\ *G The standard SNTP port number. The client (us) and the
\ ** server use the same port number.

STRUCT /sntp	\ -- len
\ *G SNTP payload structure after the IP and UDP headers.
\ ** The *\i{Timestamp} fields contain a 32.32-bit fractional
\ ** time in network (big-endian) format. The first cell contains
\ ** seconds, measured since 1st Jan 1900, as per RFC1361. The
\ ** second cell contains a fractional second.
  1 FIELD sn.flags	\ Leap Indicator, Version Number, Mode
  1 FIELD sn.stratum	\ Server NTP Stratum
  1 FIELD sn.poll	\ Max time between SNTP Polls
  1 FIELD sn.precision	\ Local clock precison (power of two, signed)
  4 FIELD sn.RootDelay
  4 FIELD sn.RootDispersion
  4 FIELD sn.RefIdentifier
  8 FIELD sn.RefTimestamp
  8 FIELD sn.OrigTimestamp
  8 FIELD sn.RXTimestamp
  8 FIELD sn.TXTimestamp
\ 12 FIELD sn.Authenticator	\ optional in SNTPv3 ; MGS004
END-STRUCT

\ *P If you are using the SNTP code with the Unix calendar code
\ ** in *\i{Examples/UnixTime.fth\}, NTP timestamps are
\ ** based from 1 Jan 1900, which has an *\fo{LSECONDS} value of
\ ** 1291876096+86400=1291962496, so add 1291962496 to the NTP
\ ** seconds value to produce a Unix time value.

\ *P If you are using this code with NTP timestamps, these are
\ ** based from 1 Jan 1900, which has an *\fo{LSECONDS} value of
\ ** 1291876096+86400=1291962496, so add 1291962496 to the NTP
\ ** seconds value to produce a Unix time value. Note that the
\ ** Unix *\fo{LSECONDS} counter rolls over in 2036. Suitable
\ ** code to convert *\fo{LSECONDS} to time and date can be found
\ ** in your cross compiler *\i{Examples/UnixTime.fth}.

\ *P If you are planning a product which needs to cope with dates
\ ** beyond 2035, note the following from RFC4330.

\ *Q As the NTP timestamp format has been in use for over 20 years, it
\ ** is possible that it will be in use 32 years from now, when the
\ ** seconds field overflows.  As it is probably inappropriate to
\ ** archive NTP timestamps before bit 0 was set in 1968, a convenient
\ ** way to extend the useful life of NTP timestamps is the following
\ ** convention: If bit 0 is set, the UTC time is in the range 1968-
\ ** 2036, and UTC time is reckoned from 0h 0m 0s UTC on 1 January
\ ** 1900.  If bit 0 is not set, the time is in the range 2036-2104 and
\ ** UTC time is reckoned from 6h 28m 16s UTC on 7 February 2036.  Note
\ ** that when calculating the correspondence, 2000 is a leap year, and
\ ** leap seconds are not included in the reckoning.



\ *********************
\ *S SNTP Configuration
\ *********************

struct /SNTPparams		\ -- len
\ *G A set of parameters to configure the SNTP service
\ ** This structure is the simplest way to interact with the SNTP
\ ** module. The elements have the following meaings :-
\ *D *\fo{sp.SNTPserver} - the IPv4 address of the SNTP server
\ **                       to use.
\ *D *\fo{sp.PollInterval} - log2 of the interval between polls
\ **                         in seconds, e.g. 5 for a 32 second
\ **                         interval. This value is set to 5 by
\ **                         default, and you should probably
\ **                         increase it when you are satisfied
\ **                         with your SNTP handling.
\ *D *\fo{sp.xtSetTime - the xt of a callback to your
\ ** application to process the received SNTP packet. This is
\ ** called whenever an SNTP packet is received.
\ ** The word receives the address and length of the UDP data,
\ ** i.e. a *\fo{/SNTP} structure described below, and should
\ ** return nothing *\fo{( caddr len -- )}.
\ ** Note that the data is in a *\fo{pbuff} that is owned by
\ ** the SNTP engine and will be discarded. Copy required data
\ ** out of the *\fo{/SNTP} structure.
\ *D *\fo{sp.xtGetTime} - an optional configuration, and is
\ ** not required for pure SNTP. If set, it must put the current
\ ** timestamp on the stack as a 32.32-bit fractional RFC1361 time.
\ ** The stack effect is *\fo{( -- TimeLo TimeHi )}.
\ ** The default action is a word that returns a zero timestamp,
\ ** which is SNTP-compliant behaviour. This configuration option
\ ** is supplied to give the system better resolution if the clock
\ ** controller can generate an NTP-compliant timestamp.
\ *[
  int sp.SNTPserver	\ IP address of the server to use. Must be first.
  int sp.PollInterval	\ Time between SNTP polls, in seconds, as a power of two
  int sp.xtSetTime	\ Callback to alert caller to a new NTP packet
  int sp.xtGetTime	\ xt of word to provide current timestamp, or zero if n/a
end-struct
\ *]

(( before v4.30
: DropSNTPPacket	\ ReceiveTimeHi ReceiveTimeLo *udp --
 \ *G A default action to drop the timestamp and packet info
  drop 2drop
;

: GenZeroZero	 \ -- ReceiveTimeHi ReceiveTimeLo
 \ *G A default action to return a non-existent timestamp
  0 0
;
))
(( before 4.92
: DiscardSNTP	\ caddr len --
 \ *G The default action of the *\fo{sp.xtSetTime} field.
 \ ** Discard the packet.
\ cr ." NTP seconds = " over sn.TXTimestamp @(n)
\ dup u. 1291962496 + secs>td . . . . . .	\ Requires UnixTime.fth
  2drop
;

: Gen00		 \ -- TimeLo TimeHi
 \ *G The default action of the *\fo{sp.xtGetTime} field.
 \ ** Returns a non-existent timestamp.
\ [con  cr ." SNTP GEN00"  con]
  0 0
;
))
: useRxSNTP	\ caddr len --
 \ *G The default action of the *\fo{sp.xtSetTime} field
 \ ** is to set the clock.
  drop sn.TXTimestamp @(n)		\ first part of the timestamp is seconds
  1291962496 + dup lseconds !		\ see Examples/UnixTime.fth
  secs>td settime&date
;

: getTxSNTP	 \ -- TimeLo TimeHi
 \ *G The default action of the *\fo{sp.xtGetTime} field.
 \ ** Returns a timestamp from the clock.
  0 Time&Date td>secs 1291962496 -
;

[undefined] SNTPparams [if]
idata create SNTPparams	\ -- addr
\ *G The *\fo{/SNTPparams} parameter block in IDATA space (RAM).
\ ** Note: This layout *\b{must} match the layout of the
\ ** *\fo{/SNTPparams} structure above. Patch this table with
\ ** the actions you supply.
  4 cells allot
cdata
SNTPparams equ SNTPserver
\ *G Address holding SNTP server IP address.
[then]

SNTPparams
  0 over sp.SNTPserver !(n)		\ IP address of the server to use
  8 over sp.PollInterval !		\ 2^n seconds between SNTP polls
SNTPactions? 0= [if]
  ' DiscardSNTP over sp.xtSetTime !	\ Callback for a new NTP packet
  ' Gen00 over sp.xtGetTime !		\ xt to provide current timestamp, or zero if n/a
[then]
SNTPactions? 1 = [if]
  ' useRxSNTP over sp.xtSetTime !	\ Callback for a new NTP packet
  ' getTxSNTP over sp.xtGetTime !	\ xt to provide current timestamp, or zero if n/a
[then]
drop


\ *********************
\ *S SNTP state machine
\ *********************
\ *P The SNTP system is controlled by a state machine, which is
\ ** run whenever the timer *\fo{SNTPtimer) times out. The action
\ ** performed is responsible for setting the next state and updating
\ ** the timer. The state machine actions are performed by
\ ** *\fo{SNTPidle ( -- )} which is called in the service task.

\ *P States of the SNTP state machine. See RFC1361

\ *E  -------                                   ------------
\ ** |       |                                 |            |
\ ** | INIT- |----Parameter set from caller--->| CONFIGURED |<-------+
\ ** | REBOOT|                             +---|            |        |
\ ** |       |                             |    ------------         |
\ **  -------                              |         |               |
\ **                                       |   NTPREQUEST Sent       |
\ **                          NTP Packet Received    |               |
\ **                                       |         +---------------+
\ **                                   --------                      ^
\ **                                  |        |                     |
\ **                                  | ACTIVE |--- Timeout----------+
\ **                                  |        |
\ **                                   --------


\ *[
[sm
  smState smSNinit		\ initialisation, wait for params from caller
  smState smSNconfigured	\ Send SNTP Request, set up timeout
  smState smSNactive		\ SNTP answer received within timeout period
sm] #smSNactions		\ number of states in this machine
\ *]

variable SNTPTimer	\ -- addr
\ *G Holds the *\fo{TICKS} time at which the next SNTP action will
\ ** be performed, 0 for no action.
  0 SNTPtimer !

variable smSNcount	\ -- addr
\ *G Down counter used to provide repetitions and timeouts.

variable smSNstate	\ -- n
\ *G Holds current SNTP state.

create smSNactions	\ -- addr
\ *G Execution table containing the xts of the action words.
\ ** This in *\fo{CDATA} space and is filled in later.
  #smSNactions cells allot&erase

interpreter
: setSNstate	\ xt state --
\ *G *\fo{INTERPRETER} word that sets the action of the given state.
  cells smSNactions + !
;
target

: SNTPidle	\ --
\ *G Performed periodically to do SNTP actions.
  SNTPtimer @ dup if			\ Perform action?
    dup timedout? if
      smSNstate @ cells			\ table index
      smSNactions + @ execute		\ execute state action
      smSNcount decr			\ repetition count down
    endif
  endif
  drop
;


\ **************************
\ *S State machine utilities
\ **************************

: SetSNTPtimer	\ ms --
\ *G Set the SNTP timer to time out *\i{ms} later.
  later 1 or SNTPtimer !
;

SNTPdebug? [if]
: .SNstate	\ --
\ *G Debug tool to display the current state of the SNTP state machine.
  cr ." State: "
  case  smSNstate @
    smSNinit of  ." init "  endof
    smSNconfigured of  ." configured "  endof
    smSNactive of  ." active "  endof
      ." unknown " dup .
  endcase
;
[then]

: toSNTPstate	\ ms n state --
\ *G The given *\i{state} will be executed after the given
\ ** period in *\i{ms} and *\i{n} times if sensitive to the number of
\ ** times. Note that this specifies the interval until the
\ ** state is executed, not the repetition rate of the state.
  smSNstate !  smSNcount !  setSNTPtimer
[ SNTPdebug? ] [if]  .SNstate  [then]
;


\ *************************
\ *S Outgoing message tools
\ *************************

variable SNTPreply?	\ -- addr
\ *G Set non-zero when an SNTP reply has been received, but the
\ ** data has not yet been been processed. During this time any
\ ** following SNTP packets will be discarded.

: (SNTPPSend)	{ *pb ipaddr -- }
\ *G Send SNTP data as a UDP packet. The parameters are *\i{*pb},
\ ** a pointer to a PBuff containing the UDP data, and *\i{ipaddr},
\ ** the IP address to send to.
  udp_hdr *pb unbumpPBuffData		\ point data to udp header
  SNTPport#				\ get local port number
  *pb PBDataStart uh_sport w!(n)	\ into udp header
  SNTPport#				\ and remote port number
  *pb PBDataStart uh_dport w!(n)	\ into udp header
  *pb ipaddr SendDatagram
;

: SendSNTP	{ | *pb *sntp -- }
\ *G Send an SNTP message.
[ SNTPdebug? ] [if]  cr ." Sending SNTP request "  [then]
  0 SNTPreply? !			\ mark no response yet
  GetPbuf ?dup 0=			\ must grab a buffer
  if  exit  endif
  -> *pb				\ save pbuf address for use

  0 p_dat *pb p_data w!			\ initial packet data pointer
  ether_hdr ip_hdr + udp_hdr +          \ calculate header size
  dup *pb p_data w+!  *pb p_len w!	\ move dp by header size

  *pb >UdpDataPos + -> *sntp
  *sntp /sntp erase			\ set to zeros

  SNTP_FLAGS *sntp sn.flags c!		\ Setup Leap Indicator, Version Number and Mode
  SNTP_STRATUM *sntp sn.stratum c!	\ Stratum
  SNTPParams sp.PollInterval @ 		\ Poll Interval as configured by application
  *sntp sn.poll c!
  SNTP_PRECISION *sntp sn.precision c!	\ SNTP Precision

  SNTPparams sp.xtGetTime @ execute 	\ Get timestamp function - even if it's our dummy one
  *sntp sn.TXTimestamp tuck !(n) 4 + !(n)	\ Put timestamp in outgoing packet ; MGS007

[ SNTPdebug? ] [if]  cr ." Header"  *sntp #48 dump  [then]

  /sntp  dup *pb p_datalen w! 		\ set data length
  *pb p_len w+!  		  	\ modify packet length

  *pb SNTPparams sp.SNTPServer @(n) (SNTPPSend)
;


\ =======================
\ *S SNTP state selection
\ =======================
\ *P Each state is controlled by three values defined here.
\ *(
\ *B Entry delay - delay before state is executed for the first time.
\ *B Period - period between repetitions of the same state.
\ *B Count - number of repetitions before *\fo{smSNcount} reaches zero.
\ *)

\ smSNinit
#10000 equ SNinitPreMs
#4000 equ SNinitMs
4 equ #SNinit

\ smSNconfigured
#10000 equ SNconfiguredPreMs
#4000 equ SNconfiguredMs
4 equ #SNconfigured

\ smSNactive
#10000 equ SNactivePreMs
#4000 equ SNactiveMs
4 equ #SNactive

\ Request timeout - revisit this when actual requirement is clearer...
#20000 equ SNrequestMs

: goSNTPInit	\ --
\ *G Go to the *\fo{smSNnit} state - stop SNTP activity and wait
\ ** for parameters from the calling app.
  -1 SNTPreply? !			\ reject SNTP input
  SNinitPreMs  #SNinit smSNinit toSNTPstate
;

: goSNTPConfigured	\ --
\ *G Go to the *\fo{smSNconfigured} state; start the SNTP process.
  0 SNTPreply? !			\ accept SNTP input
  SNconfiguredPreMs  #SNconfigured smSNconfigured toSNTPstate
  SendSNTP
;

: goSNTPActive	\ --
\ *G Go to the *\fo{smSNactive} state - SNTP data has been received
  1 SNTPparams sp.PollInterval @ lshift	\ Convert poll interval to seconds
  1000 *				\ Now ms
  #SNactive smSNactive toSNTPstate
;

: doSNinit	\ --
\ *G *\fo{SNTPauto?}=nz: Do nothing until SNTP server has been
\ ** set, either manually or by DHCP, and then start SNTP.
  SNTPparams sp.SNTPserver @(n) if
    goSNTPconfigured			\ start SNTP actions
  else
    SNInitMs SetSNTPtimer		\ restart timer
  endif
;
((
SNTPauto? [if]
 : doSNinit	\ --
 \ *G *\fo{SNTPauto?}=nz: Do nothing until SNTP is configured by
 \ ** DHCP, and then start SNTP.
   SNTPparams sp.SNTPserver @(n) if
     goSNTPconfigured			\ start SNTP actions
   else
     SNInitMs SetSNTPtimer		\ restart timer
   endif
 ;
[else]
 : doSNinit	\ --
 \ *G *\fo{SNTPauto?}=zero: Do nothing. SNTP must be started by
 \ ** the application using *\fo{goSNTPConfigured} above.
   SNInitMs SetSNTPtimer			\ restart timer
 ;
[then]
))

: doSNconfigured	\ --
\ *G "Configured" state has timed out - send SNTP request and
\ ** wait for response or timeout
  SNRequestMs SetSNTPtimer		\ restart timer
  SendSNTP				\ Send SNTP request
;

: doSNactive	\ --
\ *G "active" state has timed out; revert to "configured"
  goSNTPconfigured
;

: runSNTP	\ --
\ *G Start the SNTP process and wait until a valid time has been
\ ** returned.
[ diags? ] [if]  cr ." SNTP started"  [then]
  goSNTPInit
\ [ diags? ] [if]  ." done."  [then]
;


\ **********************
\ *S Receive SNTP packet
\ **********************

: RxSNTP	\ *pb --
\ *G Process received SNTP packets.
\ ** When we get here we know this is from port 123
\ ** and that the Ethernet packet was either addressed to us or
\ ** was broadcast.
  SNTPreply? @ 0= if			\ accept if no previous response
    dup PBHdrStart			\ returns IP header
    ip_dest @(n) IpAddress @(n) = if	\ check this is for our IP address
      dup GetUDPAddress			\ returns UDP header
      dup udp_hdr + swap		\ -- *pb *sntp *udp
      uh_ulen w@(n)			\ -- *pb caddr len
      SNTPparams sp.xtSetTime @ execute
      goSNTPactive			\ Switch to "active" mode
    then
  then
  FreeQB				\ give it back
;

create SNTPServiceStruct	\ -- addr
\ *G Holds the UDP Service info to run the SNTP code on receipt
\ ** of SNTP packets, or idle on timeout. Not that this *\b{must}
\ ** match the *\fo{/UDPService} structure in *\i{UDP.fth}.
  UDPServiceChain link,  SNTPport# ,  ' RxSNTP ,  ' SNTPidle ,


\ ***********************
\ *S Set up state machine
\ ***********************
\ *[
  ' doSNinit		smSNinit	setSNstate
  ' doSNconfigured	smSNconfigured	setSNstate
  ' doSNactive		smSNactive	setSNstate
\ *]


\ ======
\ *> ###
\ ======

decimal
