\ BSD.FTH - BSD sockets layer

((
Copyright (c) 1997-2003, 2006, 2007, 2008, 2009, 2013, 2014
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999
and:
  CEM Systems Ltd, Belfast, 2000-2002
  Hiden Analytical, Warrington UK, 2002

To do
=====
Ensure that disconnect sends all pending output.

Make BINDTO the primitive for BIND for efficiency.

De-local and Forth the code.


Change history
==============
20140204 SFP013 Corrected CONNECTTO error handling.
20130118 SFP012 Added server extensions.
20090605 SFP011 Improved robustness in WAITCONNQ.
20080327 SFP010 Refactored (IoctlRead) for added security checks
		and UDP handling.
20080325 SFP009 Updated DISCONNECT to handle UDP sockets.
20080319 SFP008 Updated (PollSocket) for UDP sockets.
20080317 SFP007 Migrated changes from ADD_HSOCKET.
20080228 SFP006 Changes implemented to support new shutdown system.
		Moved FreeupSocket to TCP.FTH.
20070228 SFP005 Replaced SOCKADDR with SOCKADDR_IN.
20060703 MPE004 Overhauled RECV and RECVFROM to preserve unread data.
                Added POLLSOCKET ( hs -- #bytes|-1 ).
20030621 SFP003 Fetches from IPADDRESS are network order.
20030324 SFP002 Refactored and overhauled.
20030313 SFP001 Added BINDTO.
20000221 SJC000 Change for consistency. Use (n) definitions for fetch
                and store when we mean network order. Even just for
                a data move.
))


\ ======
\ *! bsd
\ *T BSD API layer
\ ======
\ *P Despite its complexity, this layer is unashamedly provided
\ ** to ease porting applications whose description in the
\ ** TCP/IP literature is in terms of the well known BSD API.
\ ** The Windows Winsock API is similar to the BSD API. Note that the
\ ** PowerNet version does not provide all the facilities of the
\ ** full BSD API. See *\i{SOCKETS\BSD.FTH}.


\ ***********************
\ *S SOCKET_ERROR returns
\ ***********************
\ *P These words are provided to factor out error handling
\ ** in the BSD layer. By default they are implemented as COMPILER
\ ** macros. If implemented as discrete words, remove the
\ ** comments around the "R> DROP" phrases.

XC? [if]
compiler
: ?se0X		\ flag -- ; flag -- SOCKET_ERROR ; exits caller
\ *G If flag is true,
\ ** returns SOCKET_ERROR and exits the CALLING WORD; otherwise
\ ** does nothing.
\ ** This is provided as a factor for parameter testing.
  if  socket_error  ( r> drop ) exit  endif
;

: ?se1X		\ n flag -- n | SOCKET_ERROR ; exits caller
\ *G If flag is true, drops 1 item,
\ ** returns SOCKET_ERROR and exits the CALLING WORD; otherwise
\ ** does nothing.
\ ** This is provided as a factor for parameter testing.
  if  drop  socket_error  ( r> drop )  exit  endif
;

: ?se2X		\ n1 n2 flag -- n1 n2 | SOCKET_ERROR ; exits caller
\ *G If flag is true, drops 2 items,
\ ** returns SOCKET_ERROR and exits the CALLING WORD; otherwise
\ ** does nothing.
\ ** This is provided as a factor for parameter testing.
  if  2drop  socket_error  ( r> drop )  exit  endif
;

: ?se3X		\ n1 n2 n3 flag -- n1 n2 n3 | SOCKET_ERROR ; exits caller
\ *G If flag is true, drops 3 items,
\ ** returns SOCKET_ERROR and exits the CALLING WORD; otherwise
\ ** does nothing.
\ ** This is provided as a factor for parameter testing.
  if  3drop  socket_error  ( r> drop )  exit  endif
;
target
[else]
: ?se0X		\ flag -- ; flag -- SOCKET_ERROR ; exits caller
  if  socket_error  r> drop  exit  endif
;

: ?se1X		\ n flag -- n | SOCKET_ERROR ; exits caller
  if  drop  socket_error  r> drop  exit  endif
;

: ?se2X		\ n1 n2 flag -- n1 n2 | SOCKET_ERROR ; exits caller
  if  2drop  socket_error  r> drop  exit  endif
;

: ?se3X		\ n1 n2 n3 flag -- n1 n2 n3 | SOCKET_ERROR ; exits caller
  if  3drop  socket_error  r> drop  exit  endif
;
[then]


\ **************
\ *S BSD factors
\ **************
\ *P These words are not part of the official BSD interface. They
\ ** are either factors or useful in low-level code.

\ MPE004...
: (pollSocket)	\ *sk -- *pb|0|SOCKET_ERROR
\ *G Non-BSD function to poll a socket and return the next input
\ ** packet buffer if available.
  dup so_*inq @ dup 0= ?se2x		\ -- *sk *q
  PeekQ dup if				\ -- sk *pb|0
    nip  exit				\ has packet, use it
  endif
  swap dup skTCP? 0=			\ -- 0 *sk flag
  if  drop  exit  endif
  so_*cb @ dup 0= ?se2x			\ -- 0 *cb
  t_state w@ TCPS_ESTABLISHED <>	\ it is not established.
  if  drop SOCKET_ERROR  endif
;
\ MPE004...

\ SFP010...
: (ioctlRead)	\ *arg *sk -- 0|SOCKET_ERROR
\ *G Return the amount of data that can be read from a socket,
\ ** storing it at address *\i{*arg}.
  (pollSocket)				\ -- *arg *pb|0|SOCKET_ERROR
  dup SOCKET_ERROR =			\ error check
  if  nip exit  endif			\ just return SOCKET_ERROR
  dup					\ check *pb or 0
  if  p_datalen w@  endif		\ -- *arg len
  swap !  0
;
\ ...SFP010

: (ioctlState)	\ *arg *sk -- res|SOCKET_ERROR
\ *G Return the current socket state to the address *arg.
  so_*cb @ t_state w@  swap !		\ set tcp STATE
  0
;

\ Move to UDP.FTH?
: sendtoUDP	{ hsocket *buffer bufflen flags | *pb res -- res }
\ *G A factor used to send the *\i{*buffer/bufflen} block of
\ ** memory by UDP.
  \ Attempt to get pbuf
  GetTxPbuf dup -> *pb 0=
  if  SOCKET_ERROR exit  then		\ failed to get pbuf, bail

  \ Copy data into pbuf (yeah, I know, a COPY)
  0 p_dat *pb p_data w!			\ initial packet data pointer
  ether_hdr ip_hdr + udp_hdr +		\ calculate header size
  dup *pb p_data w+! *pb p_len w!	\ move dp by header size
  *buffer *pb dup p_data w@ + bufflen move	\ stash data from source buff
  bufflen dup *pb p_datalen w! *pb p_len w+!	\ set datalen, modify packetlen

  hsocket *pb UDPSend			\ send pbuf
  dup 0< not				\ < 0 = error !!!
  if
    ether_hdr - ip_hdr - udp_hdr -      \ modify returned size for res
  then
;

: SaveSktRem	\ *sk -- remip remport
\ *G A factor to return the socket's current remote IP address and port.
  dup so_iprem @  swap so_portrem w@
;

: RestoreSktRem	\ remip remport *sk --
\ *G A factor to restore the socket's current remote IP address and port.
  tuck  so_portrem w!  so_iprem !
;

\ MPE004...
: recvInfo	\ *sk *name --
\ *G A factor to extract the remote IP and port details from a
\ ** received TCP or UDP packet.
\ ** N.B. Replaces the now obsolete *\fo{RecvUDPinfo} and
\ ** *\fo{RecvTCPinfo}. *\fo{RecvInfo} uses a socket structure as
\ ** input.
\ This code will become more intelligent when we support more than
\ UDP and TCP.
  over so_portrem w@ over sin_port w!(n)  \ source port
  swap so_iprem @ over sin_addr !(n)   	\ INET Addr
  AF_INET swap sin_family w!(n)		\ a-family
;
\ ...MPE004

: getSocketInfo	\ hs *name *len --
\ *G If *\i{*len} contains at least *\fo{SOCKADDR_IN} and
\ ** *\i{*name} is non-zero, *\i{*name} is the address of a
\ ** *\fo{SOCKADDR_IN} structure. Copy the remote IP address and
\ ** port details to *\fo{*name}.
  dup 0=
  if  3drop exit endif
  @ SOCKADDR_IN u< not if		\ -- hs *name
    dup if
      swap get_socket_addr dup if	\ -- *name *sk
        swap recvInfo exit
      endif
    endif
  endif
  2drop
;


\ ****************
\ *S BSD Style API
\ ****************
\ *P This is a loose implementation of the BSD sockets interface.
\ ** It is not exactly compliant but will run most BSD style
\ ** code without too many nasty surprises.

$04004667F constant FIONREAD		\ -- x
\ *G IOCTLSOCKET command code to obtain the number of bytes
\ ** available.
$080000001 constant GET_TCPSTATE	\ -- x
\ *G IOCTLSOCKET command code to get the TCP state..

: ioctlsocket	\ hs command *arg -- res|SOCKET_ERROR
\ *G Perform socket operations selected by the *\i{command}
\ ** parameter. Data is returned at *\i{*arg}. The supported
\ ** operations are FIONREAD and GET_TCPSTATE. FIONREAD works
\ ** with both UDP and TCP sockets.
  rot get_socket_addr dup 0= ?se3X	\ get out if bad hs
  \ -- command *arg *sk
[ tcp? ] [if]
  dup so_*cb @ 0= ?se3X			\ get out if no control block
[then]
  case rot				\ select on command
    FIONREAD     of  (ioctlRead)  endof
[ tcp? ] [if]
    GET_TCPSTATE of  (ioctlState)  endof
[then]
      3drop SOCKET_ERROR
  end-case
;

\ ...MPE004
: pollSocket	\ hsock -- #bytes|SOCKET_ERROR
\ *G Non-BSD function to poll a socket and return the number of
\ ** bytes available to be read. Can be used with both UDP and TCP
\ ** sockets.
  get_socket_addr dup 0= ?se1X		\ get out if bad hs
  (pollSocket)				\ -- *pb|0|SOCKETERROR
  dup SOCKET_ERROR = ?se1x		\ check error
  dup
  if  p_datalen w@  endif
;
\ ...MPE004

: bind		\ hs *name namelen -- res
\ *G Associate a socket with a family, protocol and port. The
\ ** parameters *\i{*name} and *\i{namelen} (addr/len) describe a
\ ** *\fo{SOCKADDR_IN} structure. The family must be AF_INET. The
\ ** port is the port number that will be listened to. The IP
\ ** address is usually 0, in which case the system's IP address
\ ** will be used. This allows use with systems which obtain an
\ ** IP address dynamically, e.g. through DHCP.
\ Parameter checking
  SOCKADDR_IN <> ?se2X			\ -- hs *name ; exit if bad
  dup sin_family w@(n) AF_INET <> ?se2X	\ -- hs *name ; exit if not AF_INET
  swap get_socket_addr dup 0= ?se2X	\ -- *name *sk ; get out if bad socket

  \ Check requested ip/port is unique - probably not BSD
\  ip pcol port
  swap					\ -- *sk *name
  dup sin_addr @(n)			\ -- *sk *name ipaddr
  over dup so_protocol w@		\ -- *sk *name ipaddr *name pcol
  swap sin_port w@(n)			\ -- *sk *name ipaddr pcol port
  find_hsocket_port ?se2X		\ -- *sk *name ; exit if not unique

  \ The bind seems to be a reasonable request so override socket's
  \ local port and ip
  2dup sin_port w@(n) swap so_portloc w!
  sin_addr @(n) dup INADDR_ANY = if	\ -- *sk ipaddr
    drop  IPAddress @(n)		\ SFP003
  then
  swap so_iploc !

  0
;

XC? [if]
: bindTo	\ hs af port ipaddr -- res
\ *G A non-BSD function that binds a socket to the given set of
\ ** address family (af), port (port) and IP address (ipaddr).
\ ** The returned result (res) is 0 for success, otherwise
\ ** SOCKET_ERROR. N.B. subject to change. See *\fo{BIND}.
[else]
: rtbindTo	\ hs af port ipaddr -- res
[then]
  { | sai[ SOCKADDR_IN ] -- }
  sai[ SOCKADDR_IN erase		\ wipe buffer
  sai[ sin_addr !(n)			\ set ip address
  sai[ sin_port w!(n)			\ set port
  sai[ sin_family w!(n)			\ set af
  sai[ SOCKADDR_IN bind			\ bind it
;
XC? 0= [if]  assign rtbindTo to-do bindTo  [then]

XC? [if]
: closesocket	\ hs -- 0|SOCKET_ERROR
\ *G Close the socket. This is the close of last resort
\ ** as it simply reclaims socket memory without performing
\ ** any notification to the other end. Use *\fo{DISCONNECT} in
\ ** preference.
[else]
: rtclosesocket	\ hs -- 0|SOCKET_ERROR
[then]
  get_socket_addr dup 0= ?se1X
  (CloseSocket) 0
;
XC? 0= [if]  assign rtclosesocket to-do closesocket  [then]

((
#0 equ	TCPS_CLOSED		\ closed
#1 equ	TCPS_LISTEN		\ listening for connection (passive open)
#2 equ	TCPS_SYN_SENT		\ have sent syn (active open)
#3 equ	TCPS_SYN_RECEIVED	\ have sent & received syn : awaiting ACK
#4 equ	TCPS_ESTABLISHED	\ established (data transfer)
#5 equ	TCPS_CLOSE_WAIT		\ received FIN, waiting for application close
#6 equ	TCPS_FIN_WAIT_1		\ have closed, sent FIN : awaiting ACK & FIN
#7 equ	TCPS_CLOSING		\ simultaneous close : awaiting ACK
#8 equ	TCPS_LAST_ACK		\ received FIN, have closed : awaiting ACK
#9 equ	TCPS_FIN_WAIT_2		\ have close : awaiting FIN
#10 equ	TCPS_TIME_WAIT		\ 2MSL wait state after active close

#2 equ	TCPS_SYN_SENT		\ have sent syn (active open)
#3 equ	TCPS_SYN_RECEIVED	\ have sent & received syn : awaiting ACK
#4 equ	TCPS_ESTABLISHED	\ established (data transfer)

If closesocket is applied to a socket in one of the states above,
we should really do
  *sk TCP_RST SendTcpFlags
))

((
create resetTab	\ -- addr
\ Holds a list of states in which a reset is required.
  0 c,  0 c,  1 c,  1 c,
  1 c,  1 c,  1 c,  1 c,
  1 c,  1 c,  0 c,  0 c,

\ Firefox doesn't like this! or was it the state machine?
: resetSocket	\ hs --
  \ *G Not BSD: Close the socket, sending an RST if this is a TCP
  \ ** socket and the state is appropriate. Finally close the
  \ ** socket. Used when we need to recover RAM quickly.
  get_socket_addr ?dup 0=
  if  exit  endif
  dup skValid? if			\ valid TCP socket
    dup so_*cb @ t_state w@		\ get state
    #11 min resetTab + c@ if		\ should send RST?
      dup TCP_RST SendTcpFlags  #50 ms
    endif
  endif
  (CloseSocket)
;
))

: connect	\ hs *name namelen -- res
\ *G Open a connection to a destination defined by the
\ ** SOCKADDR structure described by *\i{*name/namelen}.
\ ** The adress family in the structure must be AF_INET.
\ ** On success (res=0), the socket is ready to send and
\ ** receive data.
  SOCKADDR_IN <> ?se2X			\ exit if bad namelen
  dup sin_family w@(n) AF_INET <> ?se2X	\ exit if bad AF
  swap get_socket_addr dup 0= ?se2x	\ -- *name *sk ; exit if bad socket
  \ The connect seems to be a reasonable request so override socket's
  \ remote port and ip
  over sin_port w@(n) over so_portrem w!
  swap sin_addr @(n) over so_iprem !	\ -- *sk
  dup so_protocol w@ IPPROTO_TCP =	\ if a TCP socket
  if  Tcp_Open  else  drop 0  endif	\ Go open the Tcp socket
;

: disconnect	\ hs -- res
\ *G Disconnect the socket. Returns 0 on success.
  get_socket_addr dup 0= ?se1X		\ -- *sk ; exit if bad socket
[ tcp? ] [if]
  dup so_protocol w@ IPPROTO_TCP =
  if  Tcp_Close  else  (CloseSocket) 0  endif	\ go close it then ; SFP009
[else]
  (CloseSocket) 0
[then]
;

: DiscAllSockets	\ --
\ *G Disconnect *\b{all} socket connections.
  MAXSOCKETS 1+ 1
  do  i disconnect drop  loop		\ yes, disconnect it
;

(( for DocGen
: socket	\ address-family socket-type proto-group -- hs|0
\ *G Create a new socket of the given characteristics, returning
\ ** the new socket number (1..n) on success, or zero for failure.
\ ** *\b{DO NOT CHANGE} the return of *\fo{INVALID_SOCKET}=zero
\ ** for failure!
))

: sendto	\ hsocket *buffer bufflen flags *name namelen -- len/err
\ *G A factor used to send the *\i{*buffer/bufflenlen} block of
\ ** memory by TCP. If *\i{*name} is a valid *\fo{SOCKADDR_IN}
\ ** structure then save the current remote address for the socket
\ ** and override with supplied settings using *\fo{CONNECT}.
  { hsocket *buffer bufflen flags *name namelen | *sk res -- len/err }
  hsocket get_socket_addr dup -> *sk	\ first locate the socket structure
  0= if  SOCKET_ERROR exit  endif

  *name if
    *sk SaveSktRem			\ save remote ipaddr/port
    hSocket *name namelen connect	\ attempt connect
    SOCKET_ERROR = if
      *sk RestoreSktRem			\ restore and barf
      SOCKET_ERROR exit			\ connect failed.
    then
    2>r					\ preserve remote ipaddr/port
  then

  case  *sk so_protocol w@
    IPPROTO_UDP of			\ clip bufflen to maximum data size ; SFP009
      hsocket *buffer bufflen PDATA_MAX min flags sendtoUDP
    endof
[ tcp? ] [if]
    IPPROTO_TCP of
      hsocket *buffer bufflen flags sendtoTCP
    endof
[then]
      drop  SOCKET_ERROR
  end-case

  \ If *name was valid then we should have a saved remote address on
  \ stack so this will pop it back in place
  *name
  if  2r> *sk RestoreSktRem  then
;

: send		\ hsocket *buffer bufflen flags -- #sent|socket_error
\ *G The more common form of *\fo{sendto} when no address override
\ ** is required.
  0 0 SendTo
;

\ MPE004...
: recvfrom	\ hsocket *buffer bufflen flags *name *namelen -- res
\ *G Receive data from a socket. The parameters are as for
\ ** *\fo{SENDTO}, *\b{except} that *\i{*namelen} is a pointer to the
\ ** length. If *\i{*name/*namelen} is a valid *\fo{SOCKADDR_IN}
\ ** structure, on return it will contain the sender's address and port.
\ ** Note that only one packet is read, and if you do not read
\ ** all of it, the remaining data is *\b{not} discarded.
  { hsocket *buffer bufflen flags *name *namelen | amount *pb *sk -- res }

  hsocket get_socket_addr dup -> *sk	\ first locate the socket structure
  0= if  SOCKET_ERROR exit  endif

  begin					\ spin until there is some data
    *sk (pollSocket) dup -> *pb 0=	\ check for new packet
   while
    pause
  repeat
  *pb SOCKET_ERROR =			\ check poll result
  if  SOCKET_ERROR exit  endif
  *pb p_datalen w@ -> amount		\ get #data bytes in packet

  *name if				\ Request for sender's address?
    *namelen @ SOCKADDR_IN <> if	\ a "Matteo" error has occured
      *sk so_*inq @ dequeue FreeQB	\ dequeue and free packet ; SFP006
      SOCKET_ERROR exit			\ get out
    then
    *sk *name recvInfo			\ return sender info
  then

  bufflen amount >= if	\ buffer same/longer than data
    *pb dup p_data w@ +			\ start of data
    *buffer amount move			\ copy
    *sk so_*inq @ dequeue FreeQB	\ dequeue and free packet
    amount
  else			\ buffer shorter than data
    *pb dup p_data w@ +			\ start of data
    *buffer bufflen move		\ copy
    bufflen *pb BumpPBuffData		\ step over data read
    bufflen
  endif
;
\ ...MPE004

: recv		\ hsocket *buffer bufflen flags -- len|err
\ *G Receive up to bufflen bytes of memory from the current packet
\ ** returning the length read.
\ ** Note that only one packet is read, and if you do not read
\ ** all of it, the remaining data is *\b{not} discarded.
  0 0 recvfrom
;

tcp? [if]

: Listen	\ hs -- 0|SOCKET_ERROR
\ *G Starts a bound socket listening on the port specified
\ ** by the previous *\fo{BIND} or *\fo{BINDTO} operation.
\ ** *\fo{LISTEN} just changes the mode of the socket; it does
\ ** not wait for anything to happen.
\ ** The socket must be a TCP socket.
  get_socket_addr dup 0= ?se1X
  dup so_protocol w@ IPPROTO_TCP <> ?se1X
  TCPS_LISTEN swap so_*cb @ t_state w!	\ just set state to 'listen'
  0					\ return a GOOD response
;

: -Listen?	\ hs -- state true | 0
\ *G Check the socket state, and return the state and true if it
\ ** not listening. Otherwise, just return false.
  get_socket_addr dup 0=		\ check for closed socket ; SFP008
  if  drop TCPS_CLOSED true exit  endif
  TcpState@
  dup TCPS_LISTEN =			\ treat listen states as equivalent
  if  drop 0  exit  endif
  dup TCPS_SYN_SENT =
  if  drop 0  exit  endif
  dup TCPS_SYN_RECEIVED =
  if  drop 0  exit  endif
  -1
;

: waitListener	\ hs -- flag
\ *G Wait until a connection is established or closed/closing.
\ ** Return *\i{flag=false} if established. If the socket is not
\ ** established, there has been an error in the listening socket.
  begin  pause dup -Listen?  until
  nip  TCPS_ESTABLISHED <>
;

: waitConnQ	\ hs -- ior
\ *G Wait until the listening socket has a connection pending.
\ ** Return 0 on success, or non-zero if the listening socket
\ ** has failed.
  begin					\ -- hs
    dup get_socket_addr dup 0=
    if  2drop -1 exit  endif		\ -- hs *sk
    dup skValid? 0=
    if  2drop -2 exit  endif
    TCPConnQ? 0=
   while				\ -- hs
    TcpIdleMs ms
  repeat
  drop 0
;

: repSocket	\ hs1 -- hs2|0
\ *G Given a socket, make another of the same type using *\i{hs1}
\ ** as a template.
  get_socket_addr dup 0=
  if  drop 0 exit  endif
  AF_INET  swap dup so_type w@
  swap so_protocol w@  socket
;

: repService	\ *sk1 *sk2 -- ior
\ *G create a new service data structure for *\i{sk2} using
\ ** the details in *\i{sk1} as a template.
  over so_/service @ 0=			\ 0 if no service structure needed
  if  2drop 0 exit  endif
  over so_/service @ allocate		\ -- *sk1 *sk2 caddr flag
  if  3drop -1 exit  endif		\ bad allocate
  >r
  over so_service @ over so_service !	\ copy service type
  swap so_/service @			\ -- *sk2 /srvc ; R: -- caddr
  r@ over erase				\ wipe service structure
  over so_/service !			\ -- *sk2 ; R: -- caddr ; set service size
  dup so_handle w@ r@ SVsk# !		\ set socket#
  r> swap so_*service !			\ set service data address
  0
;

: WaitNextConn	\ hsl -- hsa 0 | reason -1
\ *G Wait for the next connection on listening socket *\i{hsl},
\ ** create a socket and service for the connection, and return
\ ** the new socket *\i{hsa} and 0. If a failure occurs, return
\ ** *\i{reason} and -1. If the listening socket failed *\i{reason}
\ ** is non-zero.
  dup waitConnQ if			\ wait for a connection
    drop -1 -1 exit			\ listening socket failed
  endif
  dup repSocket dup 0= if		\ get a new socket
    2drop 0 -1 exit			\ probably no RAM, temp fault
  endif					\ -- hsold hsnew
  over get_socket_addr
  over get_socket_addr
  2dup 2>r TCPConnQ>			\ get connection from listener
  dup waitListener if			\ wait for established or bad
    closesocket drop
    2r> 2drop drop  0 -1 exit		\ new socket fault
  endif
  2r> repService if			\ replicate service structure
    closesocket drop  drop
    0 -1 exit
  endif
  nip 0					\ -- hsa 0
;

: Saccept	\ hs *skaddr *sklen -- hsnew|SOCKET_ERROR
\ *G The PowerNet version of the BSD *\b{accept()} call is named
\ ** *\fo{Saccept} to prevent a name conflict with the input
\ ** routine *\fo{accept}. Wait until a connection has been made
\ ** to the listening TCP socket *\i{hs} or there is an error in the
\ ** socket. If a connection has been established, transfer it to
\ ** a new socket *\i{hsnew}. If *\i{*skaddr} is non-zero the
\ ** contents of *\i{*sklen} are at least SOCKADDR_IN, *\i{*skaddr}
\ ** is filled in with the IP address and port of the remote end.
\ *P After a successful operation, the new socket is used for the
\ ** connection just made, and the old socket remains listening.
\ *P If SOCKET_ERROR is returned, no new socket is available and
\ ** previous socket *\i{hs} should be checked to see if it should
\ ** be closed and remade.
  2>r
  begin
    dup WaitNextConn
   while				\ -- hs hsnew | -- hs reason
    if  2r> 2drop drop SOCKET_ERROR exit  endif	\ listening socket failed
    pause
  repeat
  nip  dup 2r> getSocketInfo		\ return requested information
;

((
: Open		\ hs -- res|SOCKET_ERROR
  \ *G Actively open the socket.
  \ ** UNUSED and commented out.
  get_socket_addr dup 0= ?se1x		\ ONLY if we have a valid handle
  dup so_protocol w@ IPPROTO_TCP =
  if  Tcp_Open  else  drop 0  endif 	\ so go open it
;
))
[then]


\ *************
\ *S Extensions
\ *************
\ *P These extensions make life a bit simpler when connecting
\ ** to servers.


: TCPsocket	\ -- hs|0
\ *G Create a TCP socket. Return the socket handle on success
\ ** or *\fo{INVALID_SOCKET} (0) on error.
  AF_INET SOCK_STREAM IPPROTO_TCP socket
;

: connectTo	\ caddr u port# socket -- socket ior
\ *G Attempt to connect to a server. The socket has already been
\ ** created in the appropriate mode. The value of *\i{socket} must
\ ** be a positive non-zero number.
\ ** *\i{Caddr/u} describes the server address either as a name
\ ** or an IPaddress string and *\i{port#} is the requested port.
\ ** If *\i{u} is zero, *\i{caddr} is treated as a 32 bit number
\ ** representing an IPv4 address.
\ ** In this implementation, *\i{u} must be zero.
\ ** On success, the socket and zero are returned, otherwise
\ ** SOCKET_ERROR and an error code are returned.
  { c-addr len p# hsock | sin[ SOCKADDR_IN ] -- }

  hsock 0 <=				\ check socket positive
  if  hsock -2 exit  then
  len if				\ must be zero
    hsock closesocket drop		\ SFPxxx
    hsock -3 exit
  then

  sin[ SOCKADDR_IN erase                \ fill in address type
  AF_INET sin[ sin_family w!(n)		\ and port number for
  p# sin[ sin_port w!(n)		\ server address struct
  c-addr sin[ sin_addr !(n)		\ store it

  hsock sin[ SOCKADDR_IN connect if	\ Attempt to connect socket using server
    hsock closesocket drop
    hsock -4 exit
  then

  hsock 0
;

: TCPConnect	\ c-addr u port# -- socket ior
\ *G Attempt to create a TCP socket and connect to a server.
\ ** *\i{Caddr/u} describes the server address either as a name
\ ** or an IPaddress string and *\i{port#} is the requested port.
\ ** If *\i{u} is zero, *\i{caddr} is treated as a 32 bit number
\ ** representing an IPv4 address.
\ ** In this implementation, *\i{u} must be zero.
\ ** On success, the socket and zero are returned, otherwise
\ ** SOCKET_ERROR and an error code are returned.
  TCPsocket connectTo
;

: UDPConnect	\ c-addr u port# -- socket ior
\ *G Attempt to create a UDP socket and connect to a server.
\ ** *\i{Caddr/u} describes the server address either as a name
\ ** or an IPaddress string and *\i{port#} is the requested port.
\ ** If *\i{u} is zero, *\i{caddr} is treated as a 32 bit number
\ ** representing an IPv4 address.
\ ** In this implementation, *\i{u} must be zero.
\ ** On success, the socket and zero are returned, otherwise
\ ** SOCKET_ERROR and an error code are returned.
  AF_INET SOCK_DGRAM IPPROTO_UDP socket	\ Create a socket
  connectTo
;


\ ======
\ *> ###
\ ======

decimal

