\ SOCKPRIM.FTH - Socket handling primitives for the Forth Sockets Interface

((
Copyright (c) 1997-2003, 2008, 2009, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999
and
  CEM Systems Ltd, Belfast, 2000-2002
  Hiden Analytical, Warrington UK, 2002

To do
=====

Change history
==============
20120215 SFP010 Undid SFP009 in restoring the kill chain.
20110715 SFP009 Reduced FREEUPSOCKET to (FREEUPSOCKET).
20110531 SFP008 Removed references to unused RCV_*RQ and SND_*RQ.
		Removed CBALLOC.
20091103 SFP007 Set default MSS.
20090524 SFP006 Added listen queue handlers.
20080317 SFP005 Set/Clear socket handle in ADD_HSOCKET and
		SUB_HSOCKET.
20080228 SFP004 Socket numbers are now allocated by next fit
		rather than first fit to reduce problems with
		close detection.
20080227 SFP003 Added SUB_HSOCKET.
20030621 SFP002 Fetches from IPADDRESS are network order.
20030312 GS_001	Fixed FIND_HSOCKET_TCPPORT.
))

\ ===========
\ *! sockprim
\ ===========
\ ********************
\ *T Socket Primitives
\ ********************
\ *P The file SOCKETS\SOCKPRIM.FTH contains primitives for
\ ** the sockets layers.

decimal

: GetSocketError	\ -- ErrorCode
\ *G Return the last socket error code.
  SocketErrorCode @
;

: SetSocketError	\ ErrorCode --
\ *G Set the last socket error code.
\  ." Socket Error!" dup .decimal cr
  SocketErrorCode !
;

: ClrSocketError	\ --
\ *G Clear the last socket error code.
  SocketErrorCode off
;

\ create an array of socket pointers (handle is index)
MAXSOCKETS cells equ SOCKLIST_LEN	\ -- size
\ *G The size of the sockets array.

SOCKLIST_LEN buffer: socket_list	\ -- addr
\ *G The static sockets array.
socket_list SOCKLIST_LEN + cell -
  equ socket_list_last	\ -- addr
\ *G The address of the last entry in the socket list.

: initsockets	\ --
\ *G Initialise the socket list.
  socket_list MAXSOCKETS cells erase
;

\ SFP004...
0 value NextSock#	\ -- socknum
\ *G Returns the next socket number to be used. This value is
\ ** is ranged 0..n-1, not 1..n.

: +Sock#	\ --
\ *G Step to the next socket number, wrapping as required.
  NextSock# 1+				\ next one
  dup MAXSOCKETS 1- u>			\ range check
  if  drop 0  endif			\ wrap to start
  -> NextSock#				\ set
;

: NextSockEntry	\ -- addr
\ *G Return the socket list entry address for *\fo{NextSock#}.
  NextSock# cells socket_list +
;

: add_hsocket	\ *socket -- HSOCKET|0
\ *G See if we can get a socket handle.
  0  MAXSOCKETS 0 do			\ -- *sk 0
    NextSockEntry @ 0=
    if  drop NextSock# 1+ leave  endif
    +Sock#				\ bump socket number
  loop					\ -- *sk hs|0
  dup if				\ if we got a handle
    swap  2dup so_handle w!		\ save handle in socket structure
    NextSockEntry !  +Sock#		\ save *sk in entry & return handle
  else
    nip					\ return no handle
  endif
;
\ ...SFP004

\ SFP003...
: sub_hsocket	\ *socket --
\ *G If the given socket address is in the table, mark its socket
\ ** as unused. If the socket has already been removed, no action
\ ** is taken.
  socket_list SOCKLIST_LEN bounds do	\ -- *sk
    dup i @ = if
      0 over so_handle w!		\ mark socket structure
      0 i !				\ mark socket table
      leave
    endif
  cell +loop
  drop
;
\ ...SFP003

: get_socket_addr	\ HSOCKET -- addr|0
\ *G Given a socket handle, return the socket address, or
\ ** 0 if the socket is not in use.
\  dup 0> if
\    1- cells socket_list + @		\ get address
  1- dup MAXSOCKETS u< if		\ rebase to 0, test ; SFP004
    cells socket_list + @		\ get address
  else
    drop 0				\ return 0
  endif
;

: get_socket_inq	\ HSOCKET -- *q|0 ;
\ *G Return the socket's input queue.
  get_socket_addr dup if
    so_*inq @				\ convert to *q
  endif
;

: get_socket_outq	\ HSOCKET -- *q|0 ;
\ *G Return the socket's output queue.
  get_socket_addr dup if
    so_*outq @				\ convert to *q
  endif
;

: find_hsocket_port	{ ip pcol port | hs *sk -- hsocket|0 }
\ *G Find a socket handle with matching IP, protocol and port.
  0 -> hs				\ default fail code
  MAXSOCKETS 1+ 1 do			\ SFP004
    i get_socket_addr ?dup if		\ socket valid?
      -> *sk
      *sk so_iploc @ ip = if		\ get socket protocol
        *sk so_protocol w@ pcol = if	\ OK so protocol matches
          *sk so_portloc w@ port = if	\ OK so port matches as well
            i -> hs leave		\ return handle
          endif
        endif
      endif
    endif
  loop
  hs					\ & return handle
;

: (find_hsocket_tcpport)	{ ips ps ipd pd | hs *sk -- hsocket|0 }
\ *G Find a socket handle with matching IP and protocol and port
  0 -> hs				\ default fail code
  MAXSOCKETS 1+ 1 do
    i get_socket_addr ?dup if		\ is valid socket ??
      -> *sk
      *sk so_protocol w@ IPPROTO_TCP = if \ only if a TCP socket
        *sk so_iploc @ ipd = if		\ same local ipadd
          *sk so_portloc w@ pd = if	\ same local port
            *sk so_iprem @ ips = if	\ same remote ipadd
              *sk so_portrem w@ ps = if	\ same remote port
                i -> hs  leave		\ return handle
              endif
            endif
          endif
        endif
      endif
    endif
  loop
  hs					\ & return handle
;

: find_hsocket_tcpport	{ ips ps ipd pd | hs *sk -- hsocket|0 }
\ *G Find a socket handle with matching IP and protocol and port.
  0 -> hs				\ default fail code
  ips ps ipd pd (find_hsocket_tcpport) 	\ find full match
  -> hs  hs 0= if			\ no match
    0 0 ipd pd (find_hsocket_tcpport) 	\ find local unmapped match
    -> hs
  endif
  hs					\ & return handle
;

: WaitForIpAddress	\ --
\ *G Waits until we have a valid IP Address, i.e. one that
\ ** is not 0.0.0.0.
  begin
    pause  IPAddress @(n)		\ SFP002
  until
;

: numsockets	{ | res -- numsockets }
\ *G Count the number of socket handles in use.
  0 -> res
  socket_list				\ starting address
  MAXSOCKETS 0 do
    dup @ if
      res 1+ -> res			\ bump count
    endif
    cell+				\ bump address
  loop
  drop					\ clean up address
  res					\ return result
;

: RebindAllSockets	{ ip -- }
\ *G Rebind sockets to new local IP address.
  socket_list				\ starting address
  MAXSOCKETS 0 do
    dup @ ?dup if
      ip swap so_iploc !		\ set local socket address
    endif
    cell+				\ bump address
  loop
  drop					\ clean up address
;


\ ******************
\ *S Ephemeral ports
\ ******************
\ *P Ephemeral ports are port numbers allocated by TCP/IP
\ ** when you make a connection. These are in a range which
\ ** does not conflict with the "well known" ports. Each
\ ** socket connection uses a different port number which
\ ** allows TCP to identify multiple connections from
\ ** one machine to the same port on another machine.

((
#6000 equ FirstPortNumber	\ -- n
\ *G First ephemeral port number.
\ ** Moved to *\i{PNconfig.fth}.
#9000 equ LastPortNumber	\ -- n
\ *G Last+1 ephemeral port number.
\ ** Moved to *\i{PNconfig.fth}.
))

variable LastPortUsed	\ -- addr
\ *G Holds the last ephemeral port number used.
  FirstPortNumber LastPortUsed !

: (GetFreePort)	\ -- port
\ *G Get a free port number; the free port numbers
\ ** count up.
  LastPortUsed @ ?dup if
    1+ dup LastPortNumber >= if
      drop FirstPortNumber
    endif
  else
    FirstPortNumber
  endif
  dup LastPortUsed !
;

: GetFreePort	\ *socket -- port#
\ *G Get an ephemeral port for a socket.
  begin
    dup so_iploc @
    over so_protocol w@
    (GetFreePort) find_hsocket_port 0=
  until
  drop  LastPortUsed @
;


\ ******************************************
\ *S TCP control block creation and deletion
\ ******************************************

: FlushQ	\ queue --
\ *G Flush the contents of a queue, return them to the free
\ ** queue, and *\fo{FREE} the queue itself. If queue=0, no
\ ** action is taken.
  ?dup 0= if  exit  endif
  begin
    dup Dequeue dup
   while				\ -- queue buff|0
    FreeQB
  repeat
  drop
  free drop				\ release queue
;

: freeConnQ	\ *cb --
\ *G Free up any outstanding items on the connection queue.
  dup if
    dup listen_q @  swap listen_q off
    begin				\ -- item
      dup
     while				\ --  item
      dup @  swap free drop		\ must get next before FREE
    repeat
  endif
  drop
;

\ SFP008...
: DiscardTcpTcb	\ *cb --
\ *G Free-up all resources in the TCP control block, and then
\ ** free the control block itself.
  ?dup 0= if  exit  endif
  dup *tx_buffer @ free drop		\ got a data buffer?
  dup freeConnQ				\ got a connection queue? ; SFP006
  free drop				\ free up the control block
;

: MakeTcpCb	\ -- addr|0
\ *G Create a new TCP control block from the heap. On error,
\ ** zero is returned.
  tcpcb allocate			\ get the control block
  if  drop 0  exit  endif		\ failed
  dup tcpcb erase			\ simply wipe it for now
  ticks #250 * over			\ -- *cb iss *cb
  2dup snd_iss !  snd_nxt !		\ -- *cb ; set initial sequence numbers
  $FFFF over snd_win !			\ initialise send window for first Tx packet ; EGS028
  defMSS over t_maxseg !		\ set default MSS ; SFP007
;
\ ...SFP008

XC? [if] compiler [then]
: skTCP?	\ *sk -- flag
\ *G Returns true if the socket protocol is TCP.
  so_protocol w@ IPPROTO_TCP =		\ is this a TCP socket?
;
XC? [if] target [then]

: FreeupSocket	\ *sk --
\ *G Frees up all allocated memory for a socket. *\i{*sk} is assumed
\ ** to be valid.
  dup sub_hsocket			\ remove from socket table
  dup so_*inq @ FlushQ			\ flush inQ
  dup so_*outq @ FlushQ			\ flush outQ
  dup skTCP? if				\ is this a TCP socket ??
    dup so_*cb @ DiscardTcpTcb		\ so also free control block
  endif
[ Servers? ] [if]
  dup so_*service @ ?dup		\ if there is an allocated service packet
  if  ?SVkill  endif			\ in Servers.fth
[then]
  free drop				\ finally free up the socket structure
;

: ?BadSocket	\ *sk x flag -- *sk x | -- 0 and exit caller
\ *G A factor used by *\fo{SOCKET} for error recovery. If *\i{flag}
\ ** is non-zero, *\i{x} is discarded, the socket is freed,
\ ** and exit is from the calling word.
  if					\ -- *sk x
    drop FreeupSocket 			\ free up socket struct ; SFP009 SFP010
    INVALID_SOCKET			\ return Failed
    r> drop				\ exit the CALLER
  endif
;

XC? [if]
: socket	\ address-family socket-type proto-group -- hs|0
\ *G Create a new socket of the given characteristics, returning
\ ** the new socket number (1..n) on success, or zero for failure.
\ ** *\b{DO NOT CHANGE} the return of *\fo{INVALID_SOCKET}=zero
\ ** for failure!
[else]
: rtsocket	\ address-family socket-type proto-group -- hs|0
[then]
  rot AF_INET <> 			\ -- st pg ; check for AF_INET
  if  2drop INVALID_SOCKET  exit  endif

  (socket) allocate 			\ Allocate new socket struct
  if  3drop INVALID_SOCKET  exit  endif
  dup (socket) erase			\ -- st pg *sk ; erase struct
  tuck so_protocol w!			\ -- st *sk ; store protocol group
  tuck so_type w!			\ -- *sk ; store socket type
  queue allocate ?BadSocket		\ -- *sk *q ; allocate In Queue
  dup InitQ  over so_*inq !		\ -- *sk ; clean and set
  queue allocate ?BadSocket		\ -- *sk *q ; allocate Out Queue
  dup InitQ  over so_*outq !		\ -- *sk ; clean and set
  IPAddress @(n) over so_iploc !	\ Initial Local Address (me!)
  dup GetFreePort dup 0= ?BadSocket 	\ -- *sk port ; get ephemeral port
  over so_portloc w!			\ set local port
  dup add_hsocket			\ -- *sk hs
  dup INVALID_SOCKET = ?BadSocket	\ if bad socket
  drop					\ -- *sk
[ tcp? ] [if]
  dup so_protocol w@ IPPROTO_TCP = if	\ -- *sk ; TCP socket?
    MakeTcpcb dup 0= ?BadSocket		\ -- *sk *cb ; build new control block
    over so_*cb !
  endif
[then]
  so_handle w@				\ return hs
;
XC? 0= [if]  assign rtsocket to-do socket  [then]


\ ======
\ *> ###
\ ======

decimal
