\ primitiv.fth - Socket handling primitives for the Forth Sockets Interface

((
Copyright (c) 1997-2001, 2002, 2007, 2018
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691

Portions of this code have been licensed from:
  DelaComms Ltd. and Oxford Network Solutions 1996-1999
and:
  CEM Systems Ltd, Belfast, 2000-2002
  Hiden Analytical, 2002

To do
=====

Change history
==============
20180517 GAA004	Fixed typo in >INET_ATON.
20070828 VRW003 Moved state machine management words here
20070822 VRW002 Moved chain management words here
20070228 SFP001 Moved FILL_IPHDR and IPID words to IP.FTH.
))


\ ===========
\ *! primitiv
\ *T Helpers and primitives
\ ===========

[undefined] AddLink [if]
: AddLink	\ item anchor --
\ *G Used when a new item in the chain already exists, e.g. it
\ ** has been *\fo{ALLOCATE}d. The item is added to the chain. Note
\ ** that this word requires the link to be at offset 0 in the
\ ** item being added.
  2dup @ swap !				\ place old anchor at item+0
  !					\ item becomes new anchor
;
interpreter
: AddLink	\ item anchor --
  2dup @ swap !				\ place old anchor at item+0
  !					\ item becomes new anchor
;

: link,		\ var-addr -- ; lay a link in a chain whose head is at var-addr
\ *G Add a link to a chain anchored at address var-addr. The old contents of
\ ** var-addr are added to the dictionary as the new link, and the address
\ ** of the new link is placed at var-addr.
\ ** INTERPRETER only.
  here  over @ ,  swap !
;
target
[then]

: AddEndLink	\ item anchor --
\ *G Add an *\i{item} (a structure) to the end of of the chain
\ ** anchored at *\i{anchor}. The link field must be at offset 0
\ ** in *\i{item}.
  begin
    dup @
  while					\ -- item prev
    @
  repeat
  over swap !  off			\ add at end of chain, zero item
;

[undefined] DelLink [if]
: DelLink	\ item anchor -- ; remove item from chain
\ *G Delete/Remove an item from a chain achored at address anchor.
\ ** Note that this word requires the link to be at offset 0
\ ** in the item being removed.
  begin					\ -- item curr
    dup @ 0=				\ get out if at end of chain
    if  2drop  exit  endif
    2dup @ <>				\ while curr. does not point to target
   while				\ step through chain
    @ 					\ get next item
  repeat                                \ -- item link
\ set item that points to target to point to item pointed to by the target
  swap @  swap !			\ item's link into link
;
[then]

[undefined] ExecChain [if]
: ExecChain	\ anchor --
\ *G Execute the contents of a chain with the following structure:
\ *C   link | xt | ...
\ *P Each word that is run has the stack effect
\ *C   link -- link
\ *P Where link is the address of the link field in the structure.
\ ** Thus, data that follows the xt can easily be accessed.
  @
  begin
    dup
  while
    dup cell + @ execute
    @
  repeat
  drop
;
[then]

: ShowPacket	\ *pbuf -- ; display the packet
\ *G A diagnostic.
  $80 dump
;

: >inet_aton	\ caddr len -- ipaddr
\ *G Convert a dotted Internet address string, e.g. 192.168.0.1,
\ ** into an IPv4 address. If the string cannot be
\ ** converted, *\fo{INADDR_NONE} (-1) is returned.
  { | #dots bad? -- }
  0 -> #dots  0 -> bad?
  base @ >r decimal
  0 0					\ start off with total = 0 and current value = 0
  2swap bounds ?do			\ -- total curr
    i c@ dup [char] . = if		\ see if we are a dot
      drop 				\ OK so lose the dot
      dup #255 u>			\ check current
      if  -1 -> bad?  endif
      swap $100 * + 0			\ add current into total & start new current
      addr #dots incr			\ count number of dots
    else				\ NO so are we a digit ??
      dup [char] 0 [char] 9 within? if
	[char] 0 -			\ yes so add digit into current value
	swap #10 * +			\ shift up current value & add new digit
      else
        drop  -1 -> bad?		\ illegal char so mark it ; GAA004
      endif
    endif
  loop
  swap $100 * + 			\ add latest current into total
  r> base !
  bad? #dots 3 <> or
  if  drop INADDR_NONE  endif
;

(( removed - unused
: >inet_addr	\ c-addr -- ipaddr
 \ *G Convert dotted Internet address string into an IN_ADDR.
 \ ** *\i{C-addr} is a Forth counted string. See *\fo{>inet_aton}
  count >inet_aton
;
))

: (>inet_digit)	\ n -- n>>8
\ *G Perform number conversion of the low byte, and shift *\i{n}
\ ** right by 8 bits.
  0 $100 um/mod swap s>d #s 2drop [char] . hold
;

: >inet_ntoa	\ ipaddr -- c-addr len
\ *G Convert an IP address into a text string c-addr/len
\ ** in dotted quad notation aaa.bbb.ccc.ddd (the standard
\ ** text form).
  base @ >r decimal
  <#
  (>inet_digit)
  (>inet_digit)
  (>inet_digit)
  s>d #s #>
  r> base !
;

((
: .IPaddress	\ ipaddr --
\ *G Display a v4 IP address in dotted quad notation. Now obsolete.
  >inet_ntoa type
;
))
: .IPv4		\ ipaddr --
\ *G Display a v4 IP address *\i{ipaddr} in dotted quad notation.
  >inet_ntoa type
;

: .IPnet	\ addr --
\ *G Display the v4 IP address stored in network order at *\i{addr}.
  @(n) .IPv4
;

: .IPloc	\ addr --
\ *G Display the v4 IP address stored in native/local order at *\i{addr}.
  @ .IPv4
;

(( unused
: GetChainLength	\ pbuf -- int
\ \ *G Return the total length of data in the chain.
  dup p_len @ swap			\ start with length of this pbuf
  begin
    dup *p_next @ ?dup			\ another pbuf in chain?
  while
    nip					\ lose previous pbuf
    dup p_len @ rot + swap		\ add this length to the total
  repeat
  drop
;

: ?free		\ addr|0 --
\ *G Perform *\fo{FREE DROP} if the input is non-zero.
  dup if  free  then  drop
;
))

: CopyEthAdd	\ src dest -- ; copy ethernet address
\ *G Copy an Ethernet address. This is an ugly bit of code
\ ** which is faster than using *\fo{ETHER_ADDRESS_LEN CMOVE}
\ ** with most VFX compilers. If your CPU does not support
\ ** *\fo{W@} and *\fo{W!} as machine instructions, a simple
\ ** *\fo{CMOVE} version may be faster.
  over w@ over w!
  swap 2+ swap 2+
  over w@ over w!
  swap 2+ swap 2+
  over w@ over w!
  2drop
;

\ State Machine management

XC? [if]
interpreter
: [sm		\ -- 0
\ *G Starts the definition of a state machine's states.
  0
;

: smState	\ n -- n+1
\ *G Defines the next state as an *\fo{EQU} and increments the state number.
  dup equ  1+
;

: sm]		\ n --
\ *G Finishes the state machine and defines an equate of the number of states.
  equ
;
target
[else]
: [sm		\ -- 0
  0
;
: smState	\ n -- n+1
  dup equ  1+
;
: sm]		\ n --
  equ
;
[then]

ethernet? [if]
: xx-		\ caddr len -- caddr' len' digit|-1
  dup 0=
  if  -1 exit  endif
  base @ >r hex
  0 0 2swap >number dup if		\ -- ud caddr' len'
    over c@ [char] - <>
    if  2swap 2drop -1 -1 2swap  endif
    1 /string
  endif
  2swap drop  r> base !
;

: $>maca	\ caddr len dest -- flag ; 0=good
\ *G Convert an Ethernet MAC address string into six bytes at
\ ** *\i{dest}. If the string cannot be converted, *\i{flag} is
\ ** returned non-zero. The string *\i{caddr/len} is in the form:
\ *C  aa-bb-cc-dd-ee-ff
\ *P where the number base is hexadecimal.
  { | dest conv[ 6 ] -- }
  -> dest  6 0 do
    xx- dup -1 =
    if  drop 2drop -1 unloop exit  endif
    conv[ i + c!
  loop
  conv[ dest 6 cmove  2drop 0
;

[then]


\ ======
\ *> ###
\ ======

decimal

