\ spiSTM32-P107-9.fth - 9 bit SPI driver for STM32-P107 UEXT

((
Copyright (c) 2012
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20120501 SFP001 Started.
))

only forth definitions
decimal

\ ================
\ *! spistm32p107bb9
\ *T SPI Driver for Olimex STM32-P107 board
\ ================
\ *P This code was tested on an Olimex STM32-P107 board which uses
\ ** an STM32F107VCT6 CPU in LQFP 100 pin format. The display is
\ ** a MOD-NOKIA6610 from Olimex and uses an Epson S1D15G00
\ ** controller. Because this requires 9 bit SPI transfers, the
\ ** SPI port is bit-banged through the UEXT connector.

\ *P Note that there are two versions of the Nokia display in
\ ** circulation. The ones marked GE-8 on the sticker (that you
\ ** have already discarded) use the Epson controller, and the
\ ** ones marked GE-12 use an incompatible Philips controller.

\ *P Good introductions to programming these displays may be found
\ ** in *\i{Nokia 6100 LCD Display Driver.pdf} by Jim Lynch and in
\ ** the Seiko Epson S1D15G00 Series data sheet.

\ *E STM32-P107 UEXT Connector
\ ** Pin  Port  Function     Pin  Port  Function
\ ** 1          3v3          2          0v
\ ** 3    PD8   USART3 Tx    4    PD9   USART3 Rx
\ ** 5    PB8   I2C1 SCL     6    PB9   I2C1 SDA
\ ** 7    PC11  SPI3 MISO    8    PC12  SPI3 MOSI
\ ** 9    PC10  SPI3 SCK     10   PB15  CS_UEXT

\ *E MOD-NOKIA6610 UEXT Connector
\ ** Pin        Function     Pin        Function
\ ** 1          3v3          2          0v
\ ** 3          LCD_RST      4          LCD_BL
\ ** 5          --           6          --
\ ** 7          MISO         8          MOSI
\ ** 9          SCK          10         CS


\ ****************
\ *S SPI interface
\ ****************
\ *P This code requires *\i{Cortex/Drivers/gpioSTM32.fth}. All
\ ** GPIO ports are clocked and enabled at startup.

\ *P To drive the Nokia 6610 display using an Epson S1D5G00
\ ** controller the SPI is set to 9 bits, MSB first, CPOL=1
\ ** (CLK inactive high), CPHA=1 (capture on second edge).
\ ** Because the STM32 hardware cannot handle 9 bit SPI, the
\ ** interface is bit-banged in software.

3 8 pio: LCD_RST	\ PD8 GPIO output
3 9 pio: LCD_BL		\ PD9 GPIO output
2 11 pio: MISO3		\ PC11 AF input
2 12 pio: MOSI3		\ PC12 AF output 50MHz - may be >10MHz
2 10 pio: SCK3		\ PC10 AF output 50MHz - may be >10MHz
1 15 pio: nCSUEXT	\ PB15 GPIO output 10MHz

2 value /spiDelay	\ -- x
\ *G Value used for SPI half clock delay. The loop takes about
\ ** 5 clocks per iteration on a Cortex-M3 and 6 on an ARM.

: spiDelay	\ --
\ *G Wait half an SPI clock.
  /spiDelay begin 1- dup 0=  until  drop
;

: sck_low	\ --
\ *G Set SCK pin low.
  SCK3 clrPin
;

: sck_high      \ --
\ Set SCK pin high.
  SCK3 setPin
;

: mosi_low	\ --
\ *G Set MOSI pin low.
  MOSI3 clrPin
;

: mosi_high	\ --
\ *G Set MOSI pin high.
  MOSI3 setPin
;

: ss_low	\ --
\ *G Set SPI CS pin low.
  nCSUEXT clrPin
;

: ss_high	\ --
\ *G Set SPI CS pin high.
  nCSUEXT setPin
;

: rst_low	\ --
\ *G Set LCD RST pin low.
  LCD_RST clrPin
;

: rst_high	\ --
\ *G Set LCD RST pin high.
  LCD_RST setPin
;

: miso@		\ -- 0/1
\ *G Read MISO pin state.
  MISO3 getPin
;

: read_spi9	\ -- 9b ; msb first
\ *G Read 9 bits from the SPI bus.
  0  9 0 do
    1 lshift
    sck_low spiDelay  miso@ or
    sck_high spiDelay
  loop
;

: write_spi9	\ 9b -- ; msb first
\ *G Write 9 bits to the SPI bus.
  9 0 do
    sck_low
    dup $100 and
    if  mosi_high  else  mosi_low  endif
    spiDelay  sck_high spiDelay
    1 lshift
  loop
  drop
;

: BL_high	\ --
\ *G Enable backlight.
  LCD_BL setPin
;

: BL_low	\ --
\ *G Disble backlight.
  LCD_BL clrPin
;

: initUEXTpins	\ --
\ *G Initialise the pin selections and configurations.
\  bit28 _AFIO afMAPR + or!		\ SPI3_REMAP = 1
  LCD_RST isOutput
  LCD_BL isOutput
  MISO3 isInput
  Output50MHz MOSI3 isFunction
  Output50MHz SCK3 isFunction
  Output10MHz nCSUEXT isFunction
  BL_low
  rst_high  ss_high  sck_high  mosi_high
;
' initUEXTpins AtCold


\ *********************
\ *S Display primitives
\ *********************

: +dispCS	\ --
\ *G Set CS low and wait a bit.
  spiDelay  ss_low  spiDelay
;

: -dispCS	\ --
\ *G Wait a bit and set CS high
  spiDelay  ss_high  spiDelay
;

: writeDispCmd	\ b --
\ *G Clear bit 8 and write 9 bits to the SPI.
  $FF and  write_spi9
;

: writeDispData	\ b --
\ *G Set bit 8 and write 9 bits to the SPI.
  $FF and  $0100 or  write_spi9
;

  #define DISON     0xAF      // Display on
  #define DISOFF    0xAE      // Display off
  #define DISNOR    0xA6      // Normal display
  #define DISINV    0xA7      // Inverse display
  #define COMSCN    0xBB      // Common scan direction
  #define DISCTL    0xCA      // Display control
  #define SLPIN     0x95      // Sleep in
  #define SLPOUT    0x94      // Sleep out
  #define PASET     0x75      // Page address set
  #define CASET     0x15      // Column address set
  #define DATCTL    0xBC      // Data scan direction, etc.
  #define RGBSET8   0xCE      // 256-color position set
  #define RAMWR     0x5C      // Writing to memory
  #define RAMRD     0x5D      // Reading from memory
  #define PTLIN     0xA8      // Partial display in
  #define PTLOUT    0xA9      // Partial display out
  #define RMWIN     0xE0      // Read and modify write
  #define RMWOUT    0xEE      // End
  #define ASCSET    0xAA      // Area scroll set
  #define SCSTART   0xAB      // Scroll start set
  #define OSCON     0xD1      // Internal oscillation on
  #define OSCOFF    0xD2      // Internal oscillation off
  #define PWRCTR    0x20      // Power control
  #define VOLCTR    0x81      // Electronic volume control
  #define VOLUP     0xD6      // Increment electronic control by 1
  #define VOLDOWN   0xD7      // Decrement electronic control by 1
  #define TMPGRD    0x82      // Temperature gradient set
  #define EPCTIN    0xCD      // Control EEPROM
  #define EPCOUT    0xCC      // Cancel EEPROM control
  #define EPMWR     0xFC      // Write into EEPROM
  #define EPMRD     0xFD      // Read from EEPROM
  #define EPSRRD1   0x7C      // Read register 1
  #define EPSRRD2   0x7D      // Read register 2
  #define DISNOP    0x25      // NOP instruction

: dispNOP	\ --
\ *G Issue a NOOP command.
  +dispCS  DISNOP writeDispCmd  -dispCS
;

: +Display	\ --
\ *G Turn the display on.
  +dispCS  DISON writeDispCmd  -dispCS
;

: -Display	\ --
\ *G Turn the display off.
  +dispCS  DISOFF writeDispCmd  -dispCS
;

: setXrange	\ x1 x2 --
\ *G The X range is from column x1 to x2.
\ ** Note that the range is inclusive, so that if x1=x2
\ ** the range is one pixel.
  PASET writeDispCmd			\ Address set
  swap writeDispData			\ x1
  writeDispData				\ x2
;

: setYrange	\ y1 y2 --
\ *G The Y range is from row y1 to y2.
\ ** Note that the range is inclusive, so that if y1=y2
\ ** the range is one pixel.
  CASET writeDispCmd			\ Page Column set
  swap writeDispData			\ y1
  writeDispData				\ y2
;

: wrPixPair	\ color --
\ *G Write a pair of pixels in the given color.
  dup 4 rshift $FF and writeDispData	\ top 8 bits
  dup $0F and 4 lshift
  over 8 rshift $0F and or writeDispData  \ bottom 4 top 4
  $FF and writeDispData			\ bottom 8
;

: setPixel	\ x y color --
\ *G Plot a pixel at *\i{x,y} in the given *\i{color}.
  +dispCS
  rot dup setXrange
  swap dup setYrange
  RAMWR writeDispCmd			\ RAM write
  wrPixPair
  -dispCS
;

: initLCD	\ --
\ *G Initialise the LCD controller. If you change this word, we
\ ** recommend that you power cycle the system.
  rst_low  50 ms  rst_high  50 ms

  +dispCS

  DISCTL writeDispCmd		\ display control command
  $00 writeDispData		\ default, 2 divs, switch period=8
  $20 writeDispData		\ (32+1)*4 = 132 lines, 130 visible
  $10 writeDispData		\ no inversely highlighted lines

  COMSCN writeDispCmd		\ COM scan
  $00 writeDispData		\ Scan 1-80, 81-160

  OSCON writeDispCmd		\ Internal oscilator ON
  100 ms			\ wait aproximetly 100ms

  SLPOUT writeDispCmd		\ Sleep off

  VOLCTR writeDispCmd		\ Voltage control
  $1F writeDispData		\ middle value of V1
  $03 writeDispData		\ middle value of resistance value
  100 ms			\ settling time

  TMPGRD writeDispCmd		\ Temperature gradient
  $00 writeDispData		\ default

  PWRCTR writeDispCmd		\ Power control
  $0F writeDispData		\ reference voltage regulator on, circuit voltage follower on, BOOST ON

  DISNOR writeDispCmd		\ Normal display
  DISINV writeDispCmd		\ Inverse display

  PTLOUT writeDispCmd		\ Partial area off

  DATCTL writeDispCmd		\ Data control
  0x00 writeDispData		\ all inversions off, column direction
  0x00 writeDispData		\ RGB sequence
\  0x03 writeDispData		\ BGR sequence
  0x02 writeDispData		\ Grayscale -> 16

  PASET writeDispCmd		\ Address set
  0 writeDispData
  131 writeDispData

  CASET writeDispCmd		\ Page Column set
  0 writeDispData
  131 writeDispData

  -dispCS

  +Display BL_high		\ enable display and backlight
;


\ ****************
\ *S Display tools
\ ****************
\ *P The driver code here was derived from three sources:
\ *(
\ *B Nokia 6100 LCD Display driver, by James P. Lynch;
\ *B Bressenham's line algorithm, start at Wikipedia;
\ *P Midpoint circle algorithm, start at Wikipedia.
\ *)

variable X	\ -- addr
\ *G Holds the current X position
variable Y	\ -- addr
\ *G Holds the current Y position
variable dX	\ -- addr
\ *G Holds the current X increment, usually -1, 0, or 1.
variable dY	\ -- addr
\ *G Holds the current Y increment, usually -1, 0, or 1.
variable FColor	\ -- addr
\ *G Holds the current foreground colour. Colours are 12 bit
\ ** RGB values of the form *\i{rrrrggggbbbb}.
  $FFF Fcolor !
variable BColor	\ -- addr
\ *G Holds the current background colour. Colours are 12 bit
\ ** RGB values of the form *\i{rrrrggggbbbb}.
  $000 BColor !

: XY>		\ -- x y
\ *G Return the current X and Y coordinates.
  X @ Y @
;

: >XY		\ x y --
\ *G Set the current X and Y coordinates
  Y !  X !
;

: fillRect		\ w h --
\ *G Fill the area of given width and height at the present X,Y
\ ** position using the current foreground colour.
  { w h -- }
  w 0= h 0= or				\ sanity check
  if  exit  endif
  +dispCS
  X @ dup w + 1- setXrange
  Y @ dup h + 1- setYrange
  RAMWR writeDispCmd			\ RAM write
  w h * 1+ -2 and  0			\ force even number of pixels
  ?do  FColor @ wrPixPair  2 +loop	\ write two pixels
  -dispCS
;

: cls		\ --
\ *G Clears the screen to the current background colour.
\ ** The X-Y position is reset to 0,0.
  FColor @  BColor @ FColor !
  0 X !  0 Y !  132 132 fillRect
  Fcolor !
;

: gline		\ xd yd --
\ *G Draw a line from the current position to *\i{Xd,Yd}.
\ ** This is the general purpose line drawing routine
\ ** with no optimisations for specific orientations.
  { | x0 y0 x1 y1 stepx stepy -- }
  to y1  to x1
  X @ to x0  Y @ to Y0
  y1 y0 - dY !  x1 x0 - dX !
  dX @ 0<
  if  dX @ negate dX !  -1 to stepX  else  1 to stepX  endif
  dY @ 0<
  if  dY @ negate dY !  -1 to stepY  else  1 to stepY  endif
  dX @ 1 lshift dX !  dY @ 1 lshift dY !	\ scale
  dX @ dY @ > if
    dY @ dX @ 1 rshift -		\ initial error
    begin
      x0 y0 FColor @ setPixel
      x0 x1 <>
     while
      dup 0 >= if
        y0 stepY + to y0  dX @ -
      endif
      x0 stepX + to x0  dy @ +
    repeat
    drop
  else
    dX @ dY @ 1 rshift -
    begin
      x0 y0 FColor @ setPixel
      y0 y1 <>
     while
      dup 0 >= if
        x0 stepX + to x0  dY @ -
      endif
      y0 stepY + to y0  dX @ +
    repeat
    drop
  endif
  x0 X !  y0 Y !
;

: hline		\ x --
\ *G Draw a horizontal line until the destination X is reached.
\ ** This word is much faster than *\i{gline} above.
  dup X @ > if
    dup X @ -				\ -- x w
  else
    dup  X @ over -  swap X !		\ -- x w
  endif
  1 fillRect  X !
;

: vline		\ y --
\ *G Draw a vertical line until the destination Y is reached.
\ ** This word is much faster than *\i{gline} above.
  dup Y @ > if
    dup Y @ -				\ -- y h
  else
    dup  Y @ over -  swap Y !		\ -- y h
  endif
  1 swap fillRect  Y !
;

: line		\ xd yd --
\ *G Draw a line from the current position to *\i{Xd,Yd}.
\ ** This word uses whichever of the previous line words
\ ** gives the best results.
  over X @ =				\ vertical line
  if  nip vline exit  endif
  dup Y @ = 				\ horizontal line
  if  drop hline exit  endif
  gline
;

: rect		\ w h --
\ *G Draw a rectangle of given width and height at the current
\ ** position.
  { | x0 y0 x1 y1 -- }
  X @ dup to x0  rot + to x1
  Y @ dup to y0  + to y1
  x1 hline  y1 vline  x0 hline  y0 vline
;

: quadrant	{ cx cy rad fclr -- }
\ Draw 4 points for the circle centred at *\i{cx,cy} with
\ offset *\i{rad} in colour *\i{fclr}.
  cx cy rad + fclr setPixel
  cx cy rad - fclr setPixel
  cx rad + cy fclr setPixel
  cx rad - cy fclr setPixel
;

: octant	{ cx cy dx dy fclr -- }
\ Draw 8 points for the circle centred at *\i{cx,cy} with
\ offsets *\i{dx,dy} in colour *\i{fclr}.
  cx dx +  cy dy +  fclr setPixel
  cx dx -  cy dy +  fclr setPixel
  cx dx +  cy dy -  fclr setPixel
  cx dx -  cy dy -  fclr setPixel
  cx dy +  cy dx +  fclr setPixel
  cx dy -  cy dx +  fclr setPixel
  cx dy +  cy dx -  fclr setPixel
  cx dy -  cy dx -  fclr setPixel
;

: circle	\ radius --
\ *G Draw a circle of the given radius, using the current
\ ** position as the centre. The current position is not
\ ** changed.
  { rad | ddfX ddfY currX currY -- }

  0 -> ddfX  rad 2* negate -> ddfY
  0 -> currX  rad -> currY

  X @ Y @ rad FColor @ quadrant
  1 rad -  begin			\ -- f ; fractional error
    currX currY <
   while
    dup 0 >= if
      currY 1- -> currY
      ddfY 2+ dup -> ddfY  +
    endif
    currX 1+ -> currX
    ddfX 2+ dup -> ddfX  + 1+
    X @ Y @ currX currY Fcolor @ octant
  repeat
  drop
;


\ *********
\ Test Code
\ *********

0 [if]

: tcs		\ --
  begin +dispcs -dispcs key? until
;

: tnop		\ --
  begin dispNOP  key? until
;

: vlines	\ len --
  0 ?do
    X @     Y @  FColor @ setPixel
    X @ 1 + Y @  FColor @ setPixel
    X @ 2 + Y @  FColor @ setPixel
    1 Y +!
  loop
;

: hlines	\ len --
  0 ?do
    X @ Y @     FColor @ setPixel
    X @ Y @ 1+  FColor @ setPixel
    X @ Y @ 2+  FColor @ setPixel
    1 X +!
  loop
;

[then]


\ ======
\ *> ###
\ ======

decimal
