\ GUI_SP429I.fth - GUI code for embedded systems

((
Copyright (c) 1999, 2000, 2005, 2012, 2013, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20160223 HPO004 Converted to STM32F429 Discovery Sockpuppet
		interface.
20130514 MPE003 DocGen overhaul.
20120815 MPE002 Overhauled for PowerView v2.
20050303 MPE001 Overhauled.
))

decimal

\ ======
\ *! gui
\ *T PowerView Embedded GUI system
\ ======
\ *P The PowerView embedded GUI system in the *\i{Examples\PowerView}
\ ** directory is derived from several years work with industrial displays.
\ ** The code is designed to be a reasonable compromise between
\ ** code space, RAM space and facilities. It is not intended as
\ ** a replacement for a desktop windowing system.

\ *P Drivers are provided for several CPU and LCD panel
\ ** combinations:
\ *(
\ *B STM32F429I Discovery board using the C Sockpuppet interface:
\ ** see *\i{STM32F429disco.bld}
\ *B STM3240G-EVAL board with 16 bit colour panel:
\ ** see *\i{MB785_STM32F4.bld}.
\ *B Sharp LH77790B ARM and QVGA mono panel:
\ ** see *\i{LH77790_x2.bld}.
\ *B Nokia 6610 with Epson controller: limited support for
\ ** Olimex LCD module on UEXT connector.
\ *)

\ *P The supplied example code in this manual is for the MPE ARM
\ ** Development Kit, which uses a Sharp LH77790B ARM with an
\ ** on-chip LCD controller. To
\ ** compile the code, set up the text macro *\i{GuiDir} to point
\ ** to the PowerView folder, and compile the following files:
\ *E include %GuiDir%\FontDefs             \ font structs and data
\ ** include %GuiDir%\Font\Font12x8x2      \ 12x8 font
\ ** include %GuiDir%\Font\Font16x12x2     \ 16x12 font
\ ** include %GuiDir%\Font\Font16x16x2     \ 16x16 font
\ ** include %GuiDir%\Drivers\LcdQvgaMono5 \ hardware specific driver
\ ** include %GuiDir%\GUI                  \ GUI kernel
\ ** include %GuiDir%\DemoMono2            \ GUI demo


decimal


\ ****************
\ *S Configuration
\ ****************
\ *P The configuration equates below are only used if they have
\ ** not already been defined, e.g. in your control file.

[undefined] GUIdiags? [if]
1 equ GUIdiags?	\ -- n
\ *G Set this non-zero to compile some diagnostics.
[then]

[undefined] ShadedButtons? [if]
0 equ ShadedButtons?	\ -- n
\ *G Set this non-zero to emphasise the 3D effect of buttons defined
[then]


\ ********
\ *S Tools
\ ********

[undefined] c+! [if]
: c+!		\ n c-addr --
\ *G Add n to the byte at c-addr.
  tuck c@ + swap c!
;
[then]

[undefined] addchar [if]
: addchar       \ char string --
\ *G Add the character to the end of the counted string.
  tuck count + c!
  1 swap c+!
;
[then]

[undefined] append [if]
: append	\ c-addr u $dest --
\ *G Add the string described by C-ADDR U to the counted string at
\ ** $DEST. The strings must not overlap.
  >r
  tuck  r@ count +  swap cmove          \ add source to end
  r> c+!                                \ add length to count
;
[then]

[undefined] -leading [if]
: -leading	\ caddr len -- caddr' len'
\ *G Ignore leading spaces.
  bl skip
;
[then]

[undefined] split [if]
: split		\ addr len char -- laddr llen raddr rlen
\ *G Extract a substring at the start of addr/len, returning
\ ** the string remaining after char and the substring addr/sslen
\ ** which does not include char. If the string does not contain the
\ ** character, raddr is addr+len and rlen=0.
  rot dup 2swap scan			\ -- addr addr' len'
  dup 0= if				\ can't split
    -rot tuck over -			\ -- 0 addr' addr len
    2swap swap				\ -- addr len addr' 0
  else
    over swap				\ -- addr addr' addr' len'
    1 /string 2swap			\ -- addr'' len'' addr addr'
    over -				\ -- addr'' len'' addr sslen
    2swap				\ -- addr sslen addr'' len''
  endif
;
[then]

compiler
: 4dup		2over 2over  ;
target


\ ***********************
\ *S Unpacking rectangles
\ ***********************
\ *P In the GUI system, much use is made of rectangles stored
\ ** in *\fo{/Rect} structures.

struct /Rect	\ -- size
\ *G Rectangle structure.
\ *[
  int r.x		\ x,y position
  int r.y
  int r.w		\ width
  int r.h		\ height
end-struct
\ *]

: r>coords	\ rect -- x y w h
\ *G Return the coordinates of the given rectangle.
  dup r.x @ swap  dup r.y @ swap
  dup r.w @ swap  r.h @
;


\ **********
\ *S Buttons
\ **********

: centred	\ rect len -- xleft ytop
\ *G Given the length of a string and a rectangle in which to
\ ** display it, form the x/y start point at which the string
\ ** should be displayed.
  /fwidth * >r				\ -- rect ; R: -- pw
  dup r.x @  over r.w @ 2/ +		\ -- rect xcentre ; R: -- pw
  r> 2/ - swap				\ -- xleft rect
  dup r.y @  swap r.h @ 2/ +		\ -- xleft ycentre
  /fheight 2/ - 			\ -- xleft ytop
;

: button        \ caddr len rect --
\ *G Display a button - a rectangle with centred text.
  over /fwidth *  over r.w @ >		\ check width
  abort" String too long for button"
  dup r.h @ /fheight <			\ check height
  abort" Font too tall for button"

  LCD_COLOR_WHITE setFColor
  dup r>coords absRect			\ draw box
  LCD_COLOR_BLACK setTColor
  over centred 2swap sout		\ draw string
;

compiler
: x+i		\ rect i -- x rect i
  over r.x @  over + -rot
;

: y+i		\ rect i -- x rect i
  over r.y @  over + -rot
;

: h-i-1		\ rect i -- len
  swap r.h @ swap - 1-
;

: w-i-1		\ rect i -- len
  swap r.w @ swap - 1-
;
target

: 3D-Button	\ caddr len rect --
\ *G Display a 3D effect button. This is achieved by varying the
\ ** colour of the borders, and so is specific to the display
\ ** hardware.
  LCD_COLOR_LTGRAY setFColor
  dup r>coords absFilledRect		\ background
  LCD_COLOR_GREEN setFColor
  dup r>coords absRect			\ frame
[ ShadedButtons? ] [if]
  5 1 do				\ -- caddr len rect
    LCD_COLOR_WHITE setFColor
    dup i  x+i  over r.y @ 1+ -rot  h-i-1  v-line
    dup i  over r.x @ 1+ -rot  y+i  w-i-1  h-line
    LCD_COLOR_GRAY setFColor
    dup i  over dup r.x @ swap r.w @ + i - 1- -rot  y+i  h-i-1  v-line
    dup i  x+i  over dup r.y @ swap r.h @ + i - 1- -rot  w-i-1  h-line
  loop
[then]
  LCD_COLOR_BLACK setTColor
  r>coords 2drop 10 +  swap 16 + swap  2swap sout	\ draw string
;


\ ******************
\ *S Text formatting
\ ******************

: nInter        \ c-addr u -- n
\ *G Gets number of interword spaces in a string by counting the
\ ** number of spaces in the string.
  0 -rot  bounds ?do
    i c@ bl =
    if  1+  endif
  loop
;

#40 buffer: jText	\ -- addr
\ *G Temporary buffer that holds justified text.

: AddNext	\ caddr len -- caddr' len'
\ *G Extract the leftmost string (delimited by a space), and
\ ** add it to the justified text buffer, returning the string
\ ** to the right of the first space.
  bl split  2swap jText append
;

: AddSpaces	\ n --
\ *G Add n spaces to the justified text buffer.
  0
  ?do  bl jText addchar  loop
;

: FindLine	\ caddr1 len1 n -- caddr2 len2
\ *G Return a line of maximum length n containing complete words
\ ** of text.
  2dup > if
    dup >r
    min over swap  0 -rot  0 ?do	\ -- caddr n caddr'
      dup c@ bl =
      if  nip i swap  endif
      1+
    loop
    over 0= if				\ if word is too long
      nip  r> swap			\ replace 0 by n to split word
    else
      r> drop
    endif
  endif
  drop
;

: BuildJustified	\ caddr len w -- caddr' w
\ *G Build a justified line of text of size w in a buffer, and
\ ** return the buffer address and length.
  0 jText c!				\ empty string
  >r					\ -- caddr len ; R: -- w
  -leading -trailing 2dup nInter >r	\ -- caddr len ; R: -- w n
  dup >r				\ -- caddr' len' ; R: -- w n len
  AddNext
\ number of spaces is w-(len-nInter), gaps=nInter
  r> r>	tuck -  r> swap -  swap		\ -- caddr len #sp n
\ calculate integer and fractional number of spaces per gap
  dup >r /mod				\ -- caddr len r q ; R: -- n
\ initial error = fraction/rem
  over swap		        	\ -- caddr len r e q ; R: -- n
\ for each gap, issue q or q+1 spaces and update error
  r> tuck 0 ?do				\ -- caddr len r e n q
    >r					\ -- caddr len r e n ; R: -- q
    over 2* over >= if
      r@ 1+ AddSpaces			\ -- caddr len r e' n ; R: -- q
      tuck - swap			\ e':=e-n
    else
      r@ AddSpaces  			\ -- caddr len r e' n ; R: -- q
    endif
    -rot  over +  rot			\ e':=e+r
    >r >r >r  AddNext  r> r> r>
    r>					\ -- caddr' len' r e n q
  loop
  2drop 2drop 2drop
;

: ShowJLines	\ x y caddr len w --
\ *G Display a block of justified text starting at pixel position
\ ** x,y in a width of w pixels. The text must contain
\ ** single spaces as word separators. N.B. No word may be longer
\ ** than *\i{w/fontwidth-1} characters, otherwise the display may
\ ** hang.
  /fwidth / >r
  begin					\ -- x y caddr len ; R: -- #chars
    -leading dup
   while
    2dup r@ FindLine			\ find next line
    r@  over >r BuildJustified r>	\ -- x y caddr len len' ; -- #chars
    /string				\ step over string
    2over jText count sout		\ output string to display
    rot /fheight + -rot
  repeat
  2drop 2drop  r> drop
;


\ ******************
\ *S Drawing the GUI
\ ******************
\ *P The drawing routines for the GUI are all given the item's
\ ** data structure.

struct /GUIdef	\ -- len
\ *G All elements of the GUI are controlled by this data structure
\ ** which is usually built at compile time.
\ *[
\ -- Chains
  int pNext		\ points to next child
  int pParent		\ points to parent
  int pChild		\ points to first child
\ -- Processing
  int idSel		\ user specified identifier or flags
  			\ top bit set = exit menu/dialog
  int xtDraw		\ holds xt of routine that draws this
  int xtAction		\ holds xt of associated action, e.g. when pressed
\ -- Rectangle, text and font
  /Rect field Grect	\ holds X Y W H
  int ppFont		\ points to a pointer to the required font
  int pText		\ Points to counted string text for buttons etc
\ -- Type specific data
  0 field Gprivate	\ Item specific data follows.
\ MENU only
  int ppMenuFont	\ points to pointer to the font for the MENU
  4 -
end-struct
\ *]

\ *P The following two words are compiled if the equate
\ ** *\fo{GUIdiags?} is non-zero.

GUIdiags? [if]

: .GuiItem	\ item --
\ *G Display the data structure for a GUI element.
  cr dup .dword
     ."  N:" dup pNext @ .dword space
     ."  P:" dup pParent @ .dword space
     ."  C:" dup pChild @ .dword space
  cr ."  I:" dup idSel @ .dword space
     ."  D:" dup xtDraw @ >name .name
     ."  A:" dup xtAction @ >name .name space
     dup Grect dup r.x @ .  dup r.y @ .  dup r.w @ .  r.h @ .
  dup ppFont @ @  ."  F: " .dword space
  dup pText @
  if  cr space dup pText @ count type  endif
  drop
;

: .GuiChain	\ item --
\ *G Display the chain of data structures for a compund GUI
\ ** structure such as a menu.
  dup .GuiItem  pChild
  begin
    @ dup
   while
    dup .GuiItem  pNext
  repeat
  drop
;

[then]

: SetItemFont	\ item -- font
\ *G Return the current font and set the font for this item.
  CurrFont @  swap ppFont @ @ CurrFont !
;

: DrawButton	\ item --
\ *G Draw a 3D button with centred text.
  dup SetItemFont >r
  dup pText @ count  rot Grect
  3D-button
  r> CurrFont !
;

variable CurrentTextMode

: DrawTextMode	\ item mode --
\ *G Draw text in the rectangle using the given mode
\ ToDo: fix centering of text in C system
  CurrentTextMode !
  Font24 BSP_LCD_SetFont
  dup SetItemFont >r    \ save the current font onto the return stack
  >r  \ save the item object
  LCD_COLOR_WHITE setTColor

  r@ Grect r>coords 2drop  r@ pText @ count CurrentTextMode @  LCD_DisplayStringAtMode    \ draw string
  r> drop
  r> CurrFont !
;

: DrawCtext	\ item --
\ *G Draw text centred in the rectangle.
   CENTER_MODE DrawTextMode
;

: DrawLtext	\ item --
\ *G Draw text left justified in the rectangle.
   LEFT_MODE DrawTextMode
;

: DrawRtext	\ item --
\ *G Draw text right justified in the rectangle.
   RIGHT_MODE DrawTextMode
;

: DrawJText	\ item --
\ *G Draw justified text in the rectangular area.
   CENTER_MODE DrawTextMode
;

: DrawBox	\ item --
\ *G Draw a box (rectangle).
  LCD_COLOR_BLACK setFColor
  Grect r>coords absRect
;

: DrawTextBox	\ item --
\ *G Draw a box filled with justified text.
  dup drawBox  drawJtext
;

: DrawImage	\ item --
\ *G Display an image. The image data follows the item.
  dup Grect r.x @ swap
  dup Grect r.y @ swap
  Gprivate load-image
;

: DrawPicture	\ item --
\ *G Display an image. The address of the image data follows
\ ** the item.
  dup Grect r.x @ swap
  dup Grect r.y @ swap
  Gprivate @ load-image
;

: DrawMenuFrame	\ item --
\ *G Draw the menu frame and caption.
  dup DrawBox				\ frame
  dup pText @ c@ if
    dup SetItemFont >r
    dup >r
    r@ Grect r.x @  r@ Grect r.y @
    r@ Grect r.w @  /fheight
    4dup LCD_COLOR_LTGRAY setFColor absFilledRect
    LCD_COLOR_BLACK setFColor  absRect
    LCD_COLOR_LTGRAY setBColor
    LCD_COLOR_YELLOW setTColor
    r@ Grect r.x @ 30 +  r@ Grect r.y @ 2 +
    r> pText @ count sout
    r> CurrFont !
  endif
  drop
;

: DrawMenu	\ item --
\ *G Draw a menu and its first level children.
  CurrFont @ >r  dup ppMenuFont @ @ CurrFont !
  Font12 BSP_LCD_SetFont
  dup DrawMenuFrame
  pChild
  begin
    @ dup
   while
    dup dup xtDraw @ execute
    pNext
  repeat
  drop
  r> CurrFont !
;

: areaTouched?	\ rect -- flag
\ *G Return true if the last screen touch was in the given
\ ** rectangle.
  >r
  tsLastPos 2@				\ -- x y ; R: rect
  swap r@ r.x @ dup r@ r.w @ + within	\ -- y flag ; R: rect ; x range
  swap r@ r.y @ dup r> r.h @ + within	\ -- flag flag ; y range
  and
;

0 value LastItemTouched	\ -- struct
\ *G Holds the GUI structure corresponding to the last item touched.

: ExitTouched?	\ -- flag
\ *G Returns true if the last item touched causes an exit from
\ ** the menu/dialog handler.
  LastItemTouched idSel @ 0<
;

: ScanMenuActions    \ menu --
\ *G Scan the chain of data structures of a compound GUI
\ ** structure such as a menu, checking the area of each element against
\ ** the coordinate of the current press. If the area matches the current
\ ** screen touch, the menu action is executed.
  pChild
  begin
    @ dup
   while
    dup Grect areaTouched? if
      dup to LastItemTouched
      dup xtAction @ ?dup if  execute  then
    then
    pNext
  repeat
  drop
;

: TouchScreenAction  \ menu --
\ *G Wait for the touch screen to be pressed, and perform the
\ ** touch screen action scan on the given menu item. This action is
\ ** repeated until an item with the exit bit (top bit) set in
\ ** the *\fo{idSel} field.
  begin
    pause
    tsWaitPress  dup ScanMenuActions  tsWaitUp
    ExitTouched?
  until
  drop
;

Defer ActOnMenu	\ id item --
\ *G Perform the action specified by ID (usually a GUI item
\ ** structure or an x,y pair from a touch screen) for the
\ ** given menu. The default action is *\fo{2DROP}.
\ ** Define your own action for this word. It should walk the
\ ** menu chain, select which item if any that ID refers to,
\ ** and execute the selected action word in the item data
\ ** structure.
  assign 2drop to-do ActOnMenu


\ *****************
\ *S Defining a GUI
\ *****************

variable LastItem	\ -- addr
\ *G Holds the address of the last item defined
variable CurrItem	\ -- addr
\ *G Holds the address of the item being defined
variable ParentItem	\ -- addr
\ *G Holds the address of the item's parent
[undefined] Null$ [if]
create Null$  0 ,	\ -- addr
\ *G A null string which may be used as a counted
\ ** string or a zero-terminated string.
[then]

\ *P Please note that the following words are only available
\ ** during cross-compilation. If you need them on the target,
\ ** just make copies outside the *\fo{INTERPRETER ... TARGET}
\ ** structure.

interpreter

: pNext,	\ -- ; lay next pointer
\ *G INTERPRETER: lay the pointer to the next item. The list is
\ ** anchored at the parent's *\fo{pChild} field.
  here  0 ,
  LastItem @ if
    dup LastItem @ pNext !
  else
    ParentItem @ if
      dup ParentItem @ pChild !
    endif
  endif
  dup CurrItem !  LastItem !
;

: parent,	\ -- ; lay pointer to parent
\ *G INTERPRETER: Lay the pointer to the parent item.
  ParentItem @ ,
;

: child,	\ -- ; lay pointer to children
\ *G INTERPRETER: Lay a pointer to any children.
  0 ,
;

: +ParX		\ x -- x'
\ *G INTERPRETER: Add the parent's x offset.
  ParentItem @
  if  ParentItem @ Grect r.x @ +  endif
;

: +ParY		\ y -- y'
\ *G INTERPRETER: Add the parent's Y offset.
  ParentItem @
  if  ParentItem @ Grect r.y @ +  endif
;

: ident,	\ --
\ *G Lay default user identifier.
  0 ,
;

: Rect,		\ x y w h -- ; lay rectangle
\ *G INTERPRETER: Add the parent's x offset.
  2swap  swap +ParX , +ParY ,  swap , ,
;

: GuiDef,	\ x y w h xt --
\ *G INTERPRETER: Lay down the basic *\fo{/GuiDef} structure
\ ** using the given rectangle data and the xt of the word that
\ ** draws the item.
  align  pNext, parent, child,		\ lay chains
  ident,  ,  ['] drop ,			\ lay processing
  Rect,  CurrFont ,  Null$ ,		\ lay rect, font, text pointer
;

: text",	\ -- ; followed by delimited text
\ *G INTERPRETER: Lay down the following text string and set
\ ** the *\fo{pText} field in the current item being defined.
  align  here CurrItem @ pText !
  ",  align
;

target


\ ************************
\ *S GUI application words
\ ************************
\ *P These words are used to generate menus which can contain
\ ** buttons, text, boxes and graphics. The menu is displayed
\ ** by the word *\fo{DrawMenu ( menu -- )} and actions can be
\ ** performed by the user-defined action of
\ ** *\fo{ActOnMenu ( x menu -- )}.

\ *P With the exception of *\fo{MENU} the application words
\ ** construct data structures without names. If you want to
\ ** refer to a specific item, use
\ *C create <name>
\ *P before the item. The data structures are held in linked
\ ** lists, so having additional code and data between items is
\ ** permitted.

interpreter

: MENU		\ x y w h "<name>" -- par last ; -- addr
\ *G INTERPRETER: Use in the form:
\ *C  x y w h MENU <name>
\ *P to start a *\fo{MENU ... END-MENU} definition. When
\ *\fo{<name>} executes, it returns the address of the GUI item
\ ** data structure.
\ ** The positions of the internal elements are relative to
\ ** the top left hand corner of the menu. A title bar is
\ ** generated only if a caption is defined (see below).
  create
    LastItem @ >r  ParentItem @ >r
    ['] DrawMenu GuiDef,  CurrFont ,
    LastItem off
    CurrItem @ ParentItem !
    r> r>				\ -- par last
;

: CAPTION"	\ -- ; followed by delimited text
\ *G INTERPRETER: Use in the form:
\ *C  CAPTION" <text>"
\ *P to add a caption and title bar to a menu.
  text",
;

: MENUFONT	\ fonttable --
\ *G INTERPRETER: Use in the form:
\ *C  <fonttable> MENUFONT
\ *P to define the default font used by all items within a menu.
\ ** If no font is set, the current font will be used.
  align  here ParentItem @ ppMenuFont !
  ,
;

: FONT		\ fonttable --
\ *G INTERPRETER: Use in the form:
\ *C  <fonttable> FONT
\ *P to define the font used by an item. The defined font
\ ** will only apply to that item. If no font is set, the
\ ** current font will be used.
  align  here CurrItem @ ppFont !
  ,
;

: END-MENU      \ par last --
\ *P INTERPRETER: Ends a *\fo{MENU ... END-MENU} definition.
  LastItem !  ParentItem !
;

: SetID		\ x --
\ *G Give the current item an identifier that can be used to select
\ ** the item. This may be a character, a set of flags, or a unique
\ ** identifier for the current item. If the top bit is set, the
\ ** handler *\fo{TouchScreenAction} above exits after performing
\ ** the required action.
  CurrItem @ IdSel !
;

: BUTTON"	\ x y w h <"text"> --
\ *G INTERPRETER: Use in the form below to create a button:
\ *C  x y w h BUTTON" <text>"
   ['] DrawButton GuiDef,  text",
;

: RUNS          \  -- ; name follows inline
\ *G INTERPRETER: Use in the form:
\ *C  RUNS <name>
\ *P to set the *\fo{xtAction} field of the item being defined.
  LastItem @ 0= abort" No GUI item defined"
  ' LastItem @ xtAction !
;

: CTEXT"	\ x y w h <"text"> --
\ *G INTERPRETER: Use in the form:
\ *C  x y w h CTEXT" <text>"
\ *P to produce text centred in the given rectangle.
   ['] DrawCtext GuiDef,  text",
;

: LTEXT"	\ x y w h <"text"> --
\ *G INTERPRETER: Use in the form:
\ *C  x y w h LTEXT" <text>"
\ *P to produce text left justified in the given rectangle.
   ['] DrawLText GuiDef,  text",
;

: RTEXT"	\ x y w h <"text"> --
\ *G INTERPRETER: Use in the form:
\ *C  x y w h RTEXT" <text>"
\ *P to produce text right justified in the given rectangle.
   ['] DrawRtext GuiDef,  text",
;

: BOX		\ x y w h --
\ *G INTERPRETER: Use in the form:
\ *C  x y w h BOX
\ *P to produce a rectangular outline.
   ['] DrawBox GuiDef,
;

: JTEXT"	\ x y w h <"text"> --
\ *G INTERPRETER: Use in the form:
\ *C  x y w h JTEXT" <lots of text on one source line>"
\ *P to produce fully justfied text that may spread over several
\ ** lines. Note that there should be only one space between each
\ ** word for best results.
   ['] DrawJText GuiDef,  text",
;

: JTEXTBOX"	\ x y w h <"text"> --
\ *G INTERPRETER: Use in the form:
\ *C  x y w h JTEXTBOX" <lots of text on one source line>"
\ *P to produce a rectangular outline filled with justfied text.
\ ** Like *\fo{JTEXT"} but with a box around it.
   ['] DrawTextBox GuiDef,  text",
;

: GRAPHIC	\ x y w h --
\ *G INTERPRETER: Use in the form:
\ *C  x y w h GRAPHIC <filename>
\ *P to produce an image loaded with the top left hand corner
\ ** at x,y. The width and height are at present unused.
\ ** The image file is loaded into the dictionary at compile
\ ** time.
  ['] DrawImage GuiDef,
  data-file drop  align
;

: PICTURE	\ x y w h --
\ *G INTERPRETER: Use in the form:
\ *C  x y w h addr PICTURE
\ *P to produce an image loaded with the top left hand corner
\ ** at x,y. The width and height are at present unused.
\ ** The image data is at addr.
  >r
  ['] DrawPicture GuiDef,
  r> ,
;

: Image:	\ -- ; -- addr
\ *G INTERPRETER: Create a dictionary entry and load a graphics
\ ** file to the current section, e.g.
\ *C  Image: demo1 %GuiDir%\Pics\demo1.bin
\ *P The image can be displayed using code of the for:
\ *C  x y demo1 load-image
  create  data-file drop  align
;

target


\ *******************
\ *S Image Conversion
\ *******************
\ *P Before use, images created on a PC must be converted to
\ ** the format required by the particular LCD panel. The
\ ** The conversion tools are now part of the AIDE compiler
\ ** front-end.

\ *P Although it is possible to convert standard graphics formats
\ ** "on the fly" for display, it usually requires considerably
\ ** more image storage space, is usually slower, and requires a
\ ** more complicated (larger and slower) image display routine.
\ ** If your display system is running at 8 or 16 bits per pixel,
\ ** and/or images are frequently changed,
\ ** it may well be worthwhile to consider this approach.

\ *P Note that if you are interfacing to a C library through the
\ ** Sockpuppet interface, the library will determine what
\ ** conversion facilities are provided and/or required.



\ =======
\ *N AIDE
\ =======
\ *P The tool in the AIDE front end is the current version.
\ ** Source code for AIDE is available on request. In AIDE,
\ ** select *\i{Utilities -> BMP to LCD converter}. There
\ ** are options to flip the image vertically, generate little
\ ** or big-endian data, and to add the header needed by
\ ** PowerView to display the image.

\ ==============
\ *N Legacy code
\ ==============
\ *P The older tool for the QVGA 2-bit mono is in the
\ ** folder *\i{Examples\PowerView\BMPcnv} folder,
\ ** together with full source code to compile on VFX Forth
\ ** for Windows.

\ *P When generating images for use with a 2 bit LCD panel, ensure
\ ** that the image width is a multiple of four bits. Save the
\ ** image as 4 bits per pixel grayscale if this is available, or
\ ** as a 16-colour image, e.g. in MS Paint (file *\i{mspaint.exe}}
\ ** the selection of output file type is "16 Colour Bitmap".

\ *P *\i{BMPcnv.exe} is a Windows-32 tool for converting 16 colour
\ ** (4-bit) images to the four-level (2-bit) grayscale format
\ ** required by the GiantPlus GPG3224TWE3 and equivalent panels
\ ** for which we have provided a driver. It is assumed that the
\ ** bitmap being converted is a .BMP file with a 4 bit Windows
\ ** standard pallette (preferably grayscale) and has been flipped
\ ** vertically prior to conversion. To convert a file:
\ *(
\ *B Run the application *\i{bmpcnv.exe}.
\ *B Either enter the file name in the "16bit Grey BMP File" box,
\ ** or select the file using the "Browse" button.
\ *B To preview the image use the "Preview" button
\ *B Either enter the output file name in the "Output Binary" box,
\ ** or select the file using the "Browse" button.
\ *B Click the "Ok" Button
\ *)


\ ***************************
\ *S Using different hardware
\ ***************************
\ *P If you are not using a panel with the same RAM layout as one of
\ ** the supplied example drivers, or you are not using one of the
\ ** example LCD controllers, you will have to modify the files:
\ *D Fontxxxx.fth  contains the font used for text display.
\ *D LCDxxxx.fth   contains the LCD driver and graphics primitives.
\ *D GUI.fth       contains the high level code.

\ *P If your hardware requires an LCD layout that is not currently
\ ** supported, you will need to add a new output format to AIDE's
\ ** *\i{BmpCnv} utility. Source code for AIDE is available on request.
\ ** If the panel
\ ** operates at 8 or 16 bits per pixel, you may find it more
\ ** convenient to modify *\fo{LOAD-IMAGE} in the video driver
\ ** to operate with BMP file images.


\ ======
\ *> ###
\ ======

decimal

