((
NT35510.fth - driver for NT35510 LCD attached to DISC1 board

Code by Keith Maries - Ultra Sound Advice - October 2021

This code is for a 800x480 pixel LCD (Novatek code MRB3973), LANDSCAPE only,
with the 16-bit RGB565 protocol.
To change write direction and screen orientation, see Register command $3600
and p.193 of Novatek NT35510_Preliminary_V0.80 document.

There are several similar-looking LCDs which use different drivers (eg, OTM8009A).
This code is for Novatek NT35510 Driver only. Check driver code and
connections (before initialisation) by inputting
$0400 lcdReg@ :  this should output #85 = $55.

For pin connections, see "3.97inch 16BIT Module MRB3973 User Manual".
For 16-bit operation, check that resistors R3&R44 are shorted, and R2&R6 are open,
on the back of the LCD.

This code uses the STM32F4xx FSMC (Flexible Static Memory Controller) Bank 3,
which starts at $6800:0000 = _ELCD.
FSMC addresses external LCD memory in 16-bit or 2-byte words, so with
A0 connected to the RegisterSelect (RS) pin on the LCD, a command
using _ELCD triggers FSMC_NE3 with address A0 low (DATA).
A command using _ELCD +2 triggers FSMC_NE3 with address A0 high (REGISTER).


LCD Back-light
==============
The BL-CTRL pin on this LCD requires a PWM signal from the microcontroller.
To prevent full-on backlight at system switch-on, connect a 3k3 resistor to ground.

))

variable X	\ -- addr; current X position
variable Y	\ -- addr; current Y position

\ Colour values related to gamma correction tables below

$0000 constant BLACK
$07E0 constant GREEN
$7FFF constant CYAN
$052F constant BLUE2
$011F constant BLUE
$F81F constant MAGENTA
$F800 constant RED
$FD20 constant ORANGE
$FA00 constant TANGERINE
$FB00 constant HONEY
$FC00 constant PEACH
$FFE0 constant YELLOW
$FEE0 constant MIMOSA
$FFFF constant WHITE
$528A constant MIDGREY

$FC18 constant LIGHTPINK
$AFE5 constant LIGHTGREEN
$A987 constant DARKPINK
$39E7 constant DARKGREY
$DEF8 constant GRAY
$F6DE constant LIGHTGRAY
$000F constant NAVY
$03EF constant DARKCYAN
$780F constant PURPLE
$03E0 constant DARKGREEN
$7BE0 constant OLIVE

\ Basic words

: XY>		\ -- x y ; return current X and Y coordinates.
  X @ Y @
;

: >XY		\ x y -- ; set current X and Y coordinates
  Y !  X !
;

compiler
: XY>		X @  Y @  ;
: >XY		Y !  X !  ;
target


\ Writing & reading to the LCD Graphics RAM
\ =========================================

$6800:0000 constant _ELCD

: lcdReg@		\ reg -- w; Read a 16 bit value from an LCD register.
   _ELCD tuck w! 2+ w@
;

\ Note : At switch-on (before initialisation)  $0400 lcdReg@ produces $55,
\ a useful check that the LCD is working.

: lcdReg!		\ w reg --; Write a 16 bit value to an LCD register.
   _ELCD tuck w! 2+ w!
;

compiler
: lcdRAM! ( w -- )  _ELCD 2+  w! ;	\ w -- ; Write w to GRAM
: lcdRAM@ ( -- w )  _ELCD 2+  w@ ;	\ -- w ; read w from GRAM
: lcdRAMprep      $2C00 _ELCD w! ;	\ --   ; prepare GRAM for R/W
target

: setStartCol  ( x --)	\ set start column
   dup $00FF and  $2A01 lcdReg!
        8 rshift  $2A00 lcdReg!
;

: setStartRow  ( y --)	\ set start row
   dup $00FF and  $2B01 lcdReg!
        8 rshift  $2B00 lcdReg!
;

: setCursor  ( x y --) 	\ set position of next pixel
   setStartRow
   setStartCol
;

: LCDclear ( color y1 y2 --) \ fill screen from row y1 to y2 [max=480]
   /SCRN			\ enable writing to whole screen
   over   setStartRow
   lcdRAMprep		\ ( color y2 y1)
   swap - 800 * 0 do	\ write upwards
   dup lcdRAM!  loop
   drop
;

: /lcd	\ clear screen to black
   BLACK 0 480 LCDclear
;

: setPixel	\ color x y -- ; Plot a pixel at x,y in the given color.
   setCursor  lcdRAMprep  lcdRAM!
;

: pel ( color --)
   xy> setPixel
;

: fillX		\ u -- ; Write u pixels in the current foreground colour.
   lcdRAMprep
   0 ?do  FColor @ lcdRAM!  loop
;


\ Initialisation
\ ==============
\ GAMMA CORRECTION TABLES

hex

D100 value GMRCTR1	\ Gamma correction for RED   (+ve)
D200 value GMGCTR1	\ Gamma correction for GREEN (+ve)
D300 value GMBCTR1	\ Gamma correction for BLUE  (+ve)
D400 value GMRCTR2	\ Gamma correction for RED   (-ve)
D500 value GMGCTR2	\ Gamma correction for GREEN (-ve)
D600 value GMBCTR2	\ Gamma correction for BLUE  (-ve)

create gammatable	\ Gamma correction table - same for all the above
  00 w, 37 w, 00 w, 52 w, 00 w, 7B w, 00 w, 99 w, 00 w, B1 w, 00 w, D2 w,
  00 w, F6 w, 01 w, 27 w, 01 w, 4E w, 01 w, 8C w, 01 w, BE w, 02 w, 0B w,
  02 w, 48 w, 02 w, 4A w, 02 w, 7E w, 02 w, BC w, 02 w, E1 w, 03 w, 10 w,
  03 w, 31 w, 03 w, 5A w, 03 w, 73 w, 03 w, 94 w, 03 w, 9F w, 03 w, B3 w,
  03 w, B9 w, 03 w, C1 w,


: Wgamma  ( Regbase  --) \ write gamma table, 52 2-byte values
\ Regbase is the base address of the set of 52 values
   #52 0 do 		\ write 52 values
    dup i +  		\ increment Reg number
    gammatable i 2* + w@	\ fetch gamma value
    swap lcdReg!
   loop drop
;

: RegW ( nn .. n0 n Regbase --)	\ write n consecutive registers
   swap 0 do 		\ nn ... n0 Reg
    dup i +
    2 roll
    swap lcdReg!
   loop drop
;


: initNT35510	\ NT35510 Driver Initialisation

\ ***** SETTINGS FOR PAGE1 *****
 01 08 52 AA 55 	5 F000 RegW	\ MAUCCTR - Enable Page1
 34 34 34       	3 B600 RegW	\ BT1CTR - BT1 power control for AVDD
 0D 0D 0D	3 B000 RegW	\ SETAVDD - set AVDD voltage to 5.2V
 24 24 24	3 B700 RegW	\ BT2CTR - BT2 power control for AVEE, -6V
 0D 0D 0D	3 B100 RegW	\ SETAVEE - set AVDD voltage to 5.2V
 24 24 24       	3 B800 RegW	\ BT3CTR - BT3 power control for VCL
 00 00 00       	3 B200 RegW	\ SETVCL - set VCL voltage
 24 24 24       	3 B900 RegW	\ BT4CTR - BT4 power cntrl for VGH (clamp on)
 05 05 05      	3 B300 RegW	\ SETVGH - set VGH voltage
 34 34 34	3 BA00 RegW	\ BT5CTR - BT5 power control for VGLX (LVGL)
 0B 0B 0B	3 B500 RegW	\ SETVGL_REG - set VGL_REG (VGLO)
 00 A3 00	3 BC00 RegW	\ SETVGP - set VGMP & VGSP voltages
 00 A3 00	3 BD00 RegW	\ SETVGN - set VGMN & VGSN voltages
 A3 00	        	2 BE00 RegW	\ SETVCMOFF - set VCOM offset voltage (-0.1)
\ Gamma correction tables
  GMRCTR1 Wgamma		\ red   +ve
  GMGCTR1 Wgamma		\ green +ve
  GMBCTR1 Wgamma		\ blue  +ve
  GMRCTR2 Wgamma		\ red   -ve
  GMGCTR2 Wgamma		\ green -ve
  GMBCTR2 Wgamma		\ blue  -ve

\ ***** SETTINGS FOR PAGE0 ****
\  55  F000  	lcdReg!	\ MAUCCTR - enable Page0
\  AA  F001  	lcdReg!	\ these first 34 commands are
\  52  F002  	lcdReg!	\ identical to those for Page1
\  08  F003  	lcdReg!	\ so no need to repeat
  00  F004  lcdReg!	\ this is different

 02 05 02 05 08 5 B000 RegW	\ RGBCTR - RGB Interface Signals Control
 08  B600          lcdReg!	\ SDHDTCTR - Source Output Data Hold Time Cntrl
 50  B500          lcdReg!	\ DPRSLCTR - Display Resolution Control
 00 00        	2 B700 RegW	\ GSEQCTR - EQ Control Function for Gate Signals
 05 05 05 01  	4 B800 RegW	\ SDEQCTR - EQ Control Function for Source Driver
 00 00 00      	3 BC00 RegW	\ INVCTR - Inversion Driving Control (Columns)
 00 00 03	3 CC00 RegW	\ DPTMCTR12 - Display Timing Control
 00 31 07 84 01 	5 BD00 RegW	\ RGBCTR - RGB Interface Signals Control
 01  BA00  	  lcdReg!	\ SDVPCTR - Source Cntrl in Vert. Porch Time
 01 25 55 AA 	4 FF00 RegW	\ MAUCCTR - Command Set Disable
 00  3500  	  lcdReg!	\ TEON - Tearing Effect Line ON
 20  3600  	  lcdReg!	\ MADCTL - Memory Data Access Control
\ bit5 = $20 = exchange col/row - see p.193 - for LANDSCAPE

 55    3A00  	  lcdReg!	\ COLMOD - Interface Pixel Format (16-bit/pixel)
 #100  1100  	  lcdReg!	\ SLPOUT - Sleep Out
 #120 ms 			\ delay
 0     2900  	  lcdReg!	\ DISPON - Display On
;

decimal

: /SCRN	\ set 800x480 screen
   $0 $2A00 lcdReg!  $00 $2A01 lcdReg!	\ col start
   $3 $2A02 lcdReg!  $20 $2A03 lcdReg!	\ col end [$320 = #800]
   $0 $2B00 lcdReg!  $00 $2B01 lcdReg!	\ row start
   $1 $2B02 lcdReg!  $E0 $2B03 lcdReg!	\ row end [$1E0 = #480]
   $0 $2C00 lcdReg!    			\ ready for data
;

