\ DyLink.fth - Shared data in a C library accessed from Forth.

((
Copyright(C) 2011, 2012, 2013, 2014
Robert Sexton

Copyright (c) 2015, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)

To do
=====

Change history
==============
20160416 SFP003 Refactored.
20160229 SFP002 DocGen overhaul.
		Structure overhaul.
		Files refactored.
20151231 HPO001 Initial creation from RBS000.
))

decimal

\ =============
\ *> sockpuppet
\ =============

\ ***********************************
\ *S Sharing data between Forth and C
\ ***********************************
\ *P In the MPE environment, *\fo{VALUE}s work very well for this.
\ ** *\fo{VALUE}s are defined at compile time and can be updated
\ ** at runtime. So the basic logic is - walk the dynamic list,
\ ** and if you find something that is already defined, assume
\ ** it's a *\fo{VALUE} otherwise, generate a *\fo{CONSTANT}.

\ *P The code is based around 32 byte records. They are accessed
\ ** from both Forth and C.

\ ======================
\ *N C Linkage structure
\ ======================
\ *E #define DYNLINKNAMEMLEN 22
\ **
\ ** typedef struct {
\ **   // This union is a bit crazy, but it's the simplest way of
\ **   // getting the compiler to shut up.
\ **   union {
\ **      void (*fp) (void);
\ **      int*  ip;
\ **      unsigned int    ui;
\ **      unsigned int*  uip;
\ **      unsigned long* ulp;
\ **   } p;            ///< Pointer to the object of interest (4)
\ **   int16_t size;   ///< Size in bytes (6)
\ **   int16_t count;  ///< How many (8)
\ **   int8_t kind;    ///< Is this a variable or a constant? (9)
\ **   uint8_t strlen; ///< Length of the string (10)
\ **   const char name[DYNLINKNAMEMLEN]; ///< Null-Terminated C string.
\ ** } runtimelink_t;

\ *P When the Forth system powers up it runs the Forth word
\ ** *\fo{dy-populate} which uses SVC call 01 to get the address
\ ** of the *\f{dynamiclinks[]} table, and walks through the table
\ ** creating Forth named variables whose addresses match those in
\ **  the C system.

\ *P A Forth word *\fo{dy-show} is provided to list the entries
\ ** in the table.


\ ==========================
\ *N Forth Linkage structure
\ ==========================

\ *[
interpreter
: hword  2 field  ;
: byte   1 field  ;
target

struct /runtimelink	\ -- len
\ Forth equivalent of the C structure above.
  int   fdy.val		\ usually a pointer     0, 4
  hword fdy.size	\ size in bytes	        4, 2
  hword fdy.count	\ how many              6, 2
  byte  fdy.type	\ variable or constant  8, 1
  byte  fdy.nlen	\ name length           9, 1
  22 field fdy.zname	\ zero terminated name  10, 22
end-struct
\ *]

\ *P The accessor words just read the fields defined above. They
\ ** are defined as compiler macros. For interaction on the target,
\ ** use the field names above.
\ *[
compiler
: dy.val	fdy.val @ ;	\ addr -- n
: dy.size       fdy.size w@ ;	\ addr -- w
: dy.count      fdy.count w@ ;	\ addr -- w
: dy.type       fdy.type c@ ;	\ addr -- c
: dy.name       fdy.nlen  ;	\ addr -- addr'
target
\ *]


\ ===================================
\ *N Accessing shared data from Forth
\ ===================================

: dy-recordlen  \ -- n
\ *G Return the recordlength. Its the first thing.
   getsharedvars @ ;

: dy-first  \ -- addr
\ *G return the first entry in the dynamic variable list
   getsharedvars dy-recordlen +
;

: dy-next  \ addr -- addr
\ *G return the next entry in the dynamic variable list
   dy-recordlen +
;

: dy-string  \ addr -- caddr len
  dy.name count
;

: dy-stuff  \ n xt --
\ *G Store n in the VALUE defined by xt
   $8 + @ ! ;

: make-const  \ n addr len --
\ *G Create a *\fo{CONSTANT} of name *\i{addr/len} and value
\ ** *\i{n} by laying down a header and some machine code.
\ ** A key trick is that we have to lay down a pointer to
\ ** the first character of the definition after we finish.
  here cell + >r	\ This should point to the newly-created header.
  makeheader
\ ** This is what a constant looks like after emerging from the compiler.
\ *E ( 0002:0270 4CF8047D Lx.} )  str r7, [ r12, # $-04 ]!
\ ** ( 0002:0274 004F .O )        ldr r7, [ PC, # $00 ] ( @$20278=$7D04F84C )
\ ** ( 0002:0276 7047 pG )        bx LR
  $7D04F84C ,		\ str r7, [ r12, # $-04 ]!
  $4f00 w,		\ ldr r7, [ PC, # $00 ]
  $4770 w,		\ bx LR
  ,			\ Lay down the value n
  r> ,			\ lay down the link field for the next word - required
;

: dy-create  \ addr --
\ *G Look for an entry and if its a value, update it.
  dup dy.name count pad place		\ Make a counted string out of it.
  pad find 0= if			\ If failure, cleanup and make-const
    2DROP				\ Drop the length and pointer.
    dup dy.val				\ addr n
    over dy.name count			\ n addr b
    make-const
  else           			\ If ok, get value and stuff it in.
    SWAP dy.val				\ Fetch the value
    SWAP dy-stuff
  then
;

: dy-populate  \ --
\ *G Walk the table and make the constants.
  dy-first
  begin
    dup dy.val 0<>
  while
    dup dy-create
    dy-next
  repeat
  drop
;
' dy-populate AtCold

: dy-print  \ addr --
\ *G Dump out an entry as a set of constants
   cr
   dup dy.type [CHAR] V =   \ check the type
   if
     ." Value    "
     dup @ dy.val  ." $" .DWORD  space  \ display the value
   else
     ." Constant "
      dup dy.val  ." $" .DWORD  space  \ display the address
   then
   dup dy.size . ." x " dup dy.count .
   dy-string type
;

: dy-show  \ --
\ *G Walk the table and show the entries
  dy-first
  begin
     dup dy.val 0<>
  while
     \ dup cr dy-string type
     dup dy-print
     dy-next
  repeat
  drop
  cr
;

: dy-compare  \ addr n addr -- n
\ *G Given a record address and a string, figure out if it is
\ ** the one we're looking for.
  dy-string compare
;

: dy-find  \ addr n -- addr|0
\ *G Walk the dynamic variable list and find a string, return its
\ ** address if found, else 0.
  dy-first >R				\ Put it onto the return stack
  BEGIN
   R@ dy.val 0<>
  WHILE
   2DUP R@ dy-compare
   \ If we find it, clean the stack and leave the address on R
   0= IF 2DROP R> EXIT THEN
   R> dy-recordlen + >R
  REPEAT
  \ If we fall out, there will be a string addr/n on the stack
  2DROP R> DROP
  0 					\ Leave zero to indicate failure
;


\ ======
\ *> ###
\ ======
