// SockPuppetInterface.c
// Based on the sockpuppet interface by Robert Sexton

/// \file SockPuppetInterface.c
/// \brief Sockpuppet API System Call functions
/// \defgroup SAPI Sockpuppet
//! \addtogroup SAPI
//! @{
///
// Copyright(C) 2011-2013, 2016 Robert Sexton
//
// Copyright(C) 2015, 2016 MicroProcessor Engineering
//
// Change History.
// ==============
// 20160503 SFP004 Updated to v4 with AAPCS SVC call.
// 20160418 SFP003 Kluged in DocGen documentation
// 20160411 SFP002 Changed to MPE calling convention for SVC calls.
// 20160123 HPO001 HPO release

/*
\ =============
\ *> sockpuppet
\ *S The Sockpuppet C code
\ =============
\ *P The demonstration system runs on an STM32F429I Discovery board
\ ** with a QVGA colour display. The Forth demonstration code uses a
\ ** mixture of SVC calls and direct calls into the C jump table.
\ ** This section documents the C code. The examples are taken from
\ ** the file *\i{SockPuppetInterface.c}.


\ ==============
\ *N SVC handler
\ ==============
\ *P The SVC handler runs a short piece of assembler that permits
\ ** the use of both the main and process stacks. This version of
\ ** the code is AAPCS compliant for up to four arguments and a
\ ** 32-bit return value. The structure of the code is imposed by
\ ** the requirements of the Cortex-M interrupt/exception handler
\ ** for tail-chaining and late arrivals.

\ *E void SVC_Handler(void)
\ ** {
\ ** // on entry, R0..R3, R12, LR, PC, PSR have been saved
\ **   __asm( "tst    lr, #0x4" );         // Figure out which stack
\ **   __asm( "ite    eq" );
\ **   __asm( "mrseq  r0,msp" );           // Main stack
\ **   __asm( "mrsne  r0,psp" );           // Process/Thread Stack
\ **
\ **   __asm( "push   { r4, lr }" );       // now we can preserve more
\ **   __asm( "mov    r4, r0" );           // copy SP; R4 is preserved
\ **   __asm( "ldr    r1,[r0,#24]" );      // get the stacked PC
\ **   __asm( "ldrb   r1,[r1,#-2]" );      // extract the svc call number
\ **                                       // range checking goes here
\ **   __asm( "ldr    r2 ,=SVC_syscall_table" );
\ **   __asm( "ldr    r12, [r2, r1, LSL #2]" );  // R12 = function address
\ **   __asm( "ldm    r4, { r0-r3 }" );    // restore AAPCS args from R0..R3
\ **   __asm( "blx    r12" );              // call function
\ **   __asm( "str    r0, [ r4, #0 ]" );   // save return value
\ **   __asm( "pop    { r4, pc }" );       // restore
\ ** }


\ ======
\ *> ###
\ ======
*/

#include <stdint.h>
#include "SockPuppetInterface.h"
#include "stm32f429i_discovery.h"
#include "stm32f429i_discovery_ts.h"
#include "stm32f429i_discovery_lcd.h"
#include "stlogo.h"

extern void WaitForPressedState(uint8_t Pressed);
extern void Touchscreen_Calibration (void);
extern void Touchscreen_demo (void);
extern void EnableAllClocks( void );
extern void SystemClock_Config(void);

void __SAPI_31_SetLedsHW( uint32_t ledmask );

// in SockPuppetInterface.h
// __attribute__((section(".text")))
// const void*  SVC_syscall_table[];

void Touchscreen_CalibrationPlus( void )
{
//    HAL_Init();
//    EnableAllClocks();
//    SystemClock_Config();

    BSP_LCD_Init();
    BSP_LCD_LayerDefaultInit(0, (uint32_t) LCD_FRAME_BUFFER);
//    BSP_LCD_SetLayerWindow(0, 0, 0, 240, 160);
    BSP_LCD_SetLayerVisible(0, ENABLE);

    BSP_LCD_SelectLayer(0);
    BSP_LCD_SetBackColor(LCD_COLOR_BLACK);
    BSP_LCD_SetTextColor(LCD_COLOR_GREEN);
    BSP_LCD_DisplayOn();

    BSP_LCD_Clear(LCD_COLOR_BLACK);

    __SAPI_31_SetLedsHW( 0 );

    Touchscreen_Calibration();

    BSP_LCD_Init();
    BSP_LCD_LayerDefaultInit(0, (uint32_t) LCD_FRAME_BUFFER);
//    BSP_LCD_SetLayerWindow(0, 0, 0, 240, 160);
    BSP_LCD_SetLayerVisible(0, ENABLE);

    BSP_LCD_SelectLayer(0);
    BSP_LCD_SetBackColor(LCD_COLOR_BLACK);
    BSP_LCD_SetTextColor(LCD_COLOR_GREEN);
    BSP_LCD_DisplayOn();

    BSP_LCD_Clear(LCD_COLOR_BLACK);
}


int pendsv_uart = 0;  //!< Indicates the cause of PendSV is the UART.
//!< Store the address of the UART in question here.

/// Test variables
unsigned long g_ulSystemTimeMS = 12345678;
unsigned long MyVar1 = 1;
unsigned long MyVar2 = 2;
unsigned long MyVar3 = 3;

unsigned long getMSFunction( void );

unsigned long getMSFunction( void )
{
    return( HAL_GetTick() );
}

/// Dynamic Linking table.
/// This lets Forth add key variables into the dictionary.
/// Pointer, size (bytes), count, type: Variable, Constant, and a pointer to a string
/// Note that there may be funny stuff with regards to case and Forth,
/// so the names are upper-cased.

const runtimelink_t dynamiclinks[] __attribute__((aligned( sizeof(runtimelink_t) ))) =
 {
 { { .ui  = sizeof(runtimelink_t)  },                0, 0, 'C', FORTHNAME("RECORDLEN") },
 { { .ulp = &g_ulSystemTimeMS      }, sizeof(uint32_t), 1, 'V', FORTHNAME("SYSTIMEMS") },
 { { .fp = (void (*) (void)) &getMSFunction}, sizeof(uint32_t), 1, 'C', FORTHNAME("TEST-FN") },
 { { .ulp = &MyVar1                }, sizeof(uint32_t), 1, 'V', FORTHNAME("MYVAR1") },
 { { .ulp = &MyVar2                }, sizeof(uint32_t), 1, 'V', FORTHNAME("MYVAR2") },
 { { .ulp = &MyVar3                }, sizeof(uint32_t), 1, 'V', FORTHNAME("MYVAR3") },
 { { .ui = 0 } ,0,0,0,FORTHNAME("") }
 };


// *******************************************************************
// MANDATORY CALLS
// *******************************************************************

/// @brief SVC 0: Return the API Version
uint32_t __SAPI_00_ABIVersion( void )
{
    return 0x0400;
}

/// @brief SVC 1: Return the address of the runtime link list
uint32_t __SAPI_01_GetLinkList( void )
{
    return (uint32_t) dynamiclinks;
}

/// @brief SVC 15: Return the address of the syscall table
uint32_t __SAPI_15_GetFnTable( void )
{
    return (uint32_t)SVC_syscall_table;
}

/// @brief SVC 29: Get the Key Button value
uint32_t __SAPI_29_GetKeyButton( void )
{
    return BSP_PB_GetState( BUTTON_MODE_GPIO );
}

/// @brief SVC 30: Set LED3 ( bit0 ) and LED4 ( bit1 ) via BSP
void __SAPI_30_SetLeds( uint32_t ledmask )
{
    BSP_LED_Off( LED3 );
    BSP_LED_On( LED4 );
    if( ledmask & 0x01 )
    {
        BSP_LED_On( LED3 );
    }
    else
    {
        BSP_LED_Off( LED3 );
    }

    if( ledmask & 0x02 )
    {
        BSP_LED_On( LED4 );
    }
    else
    {
        BSP_LED_Off( LED4 );
    }
}

/// @brief SVC 31: Set LED3 ( bit0 ) and LED4 ( bit1 ) via direct hardware access
void __SAPI_31_SetLedsHW( uint32_t ledmask )
{
    uint16_t* MyGPIOG_BSRR; // pointer to LEDs GPIO port G

    if( ledmask & 0x01 )
    {
        MyGPIOG_BSRR = (uint16_t*)0x40021818;
        *MyGPIOG_BSRR = 0x2000;    // turn on LED3, PortG.13
    }
    else
    {
        MyGPIOG_BSRR = (uint16_t*)0x4002181A;
        *MyGPIOG_BSRR = 0x2000;    // turn off LED3, PortG.13
    }

    if( ledmask & 0x02 )
    {
        MyGPIOG_BSRR = (uint16_t*)0x40021818;
        *MyGPIOG_BSRR = 0x4000;    // turn on LED4, PortG.14
    }
    else
    {
        MyGPIOG_BSRR = (uint16_t*)0x4002181A;
        *MyGPIOG_BSRR = 0x4000;    // turn off LED4, PortG.14
    }
}

#if ( SAPI_COMMS )
// N.B. This section has not been converted to the AAPCS
// compliant call system.

/// SVC 2: StreamPutChar
/// There are two situations:
/// The FIFO is full.  Thats the bad scene.  Cut off the supply of CPU Cycles, possibly WDT Timeout.
/// Situation Two - We have pushed the FIFO into Highwater.  Tell them.
///
/// @param frame[0]/R0 Port number/Descriptor
/// @param frame[1]/R1 Character to output
/// @returns Success:  0
/// @returns Quench:  -1

unsigned int count_sapi_putchar_highwater=0;
unsigned int count_sapi_putchar_pendsv=0;
unsigned int count_sapi_putchar_success=0;

void __SAPI_StreamPutChar( int32_t *frame)
{
    // Physical UARTS are 0-9

#if (PLATFORM_LM3S)
    // Physical UARTS are 0-9
    if (frame[0] < 10 )
    {
        uint32_t tx_level;
        uint32_t rx_level;

        // There are four cases:
        // 3: Room before(1) -> Room After(1)
        // 2: Room before(1) -> High After(0)
        // 1: High before(0) -> Room After(1) !!!! Unpossible
        // 0: High before(0) -> High After(0)

        // This will block if the FIFO is full.
        UARTCharPut(uart_ports[frame[0]],frame[1]);

        // The RIS bit gets set when you have room for more.
        // If the TX bit was set before, there was room
        // If its still set now, there is still room.
        if ( (HWREG(uart_ports[frame[0]] + UART_O_RIS) & UART_RIS_TXRIS) != 0 )
        {
            count_sapi_putchar_success++;
            frame[0] = 0;
            return;
        }

        // If the FIFO is full, or we have triggered highwater, we will adjust the
        // FIFO Trigger level and enable the IRQ
        UARTFIFOLevelGet(uart_ports[frame[0]],&tx_level,&rx_level);

        // If we were below before, and we are above now, its highwater.
        if ( (HWREG(uart_ports[frame[0]] + UART_O_RIS) & UART_RIS_TXRIS) == 0 )
        {
            count_sapi_putchar_highwater++;
            UARTFIFOLevelSet(uart_ports[frame[0]],UART_FIFO_TX2_8,rx_level);
        }

        // There should be a check here for a full FIFO.  If its full whack PendSV
        // so that we WFI until there is room for the next character.
        // We'll return a pause to the app, but set the IRQ level high so that
        // we don't have to drain too much before letting the App have cycles.
        // Do this check second so that two calls to LevelSet don't conflict.
        if (HWREG(uart_ports[frame[0]] + UART_O_FR) & UART_FR_TXFF)
        {
            count_sapi_putchar_pendsv++;
            UARTFIFOLevelSet(uart_ports[frame[0]],UART_FIFO_TX7_8,rx_level);

            HWREG(NVIC_INT_CTRL) = NVIC_INT_CTRL_PEND_SV;
            pendsv_uart = uart_ports[frame[0]];
        }

        // Enable the IRQ so that the handler can wake us.
        HWREG(uart_ports[frame[0]] + UART_O_IM) |= UART_IM_TXIM;

        // In either case, we tell the App.
        frame[0] = -1;

        return;
    }
#endif

    // For now, assume that > 10 means networking

    return;
}

/// SVC 3: StreamGetChar
///
/// Pull a character out of the stream of choice
/// This gets called when CharsAvail says there is work to do.
unsigned int count_sapi_streamgetchar=0;
unsigned int count_sapi_streamgetchar_xon=0;

void __SAPI_StreamGetChar( int32_t *frame)
{
    count_sapi_streamgetchar++;
    uint32_t the_char;

#if (PLATFORM_LM3S)
    if (frame[0] < 10 )
    {
        uint32_t tx_level;
        uint32_t rx_level;

        the_char = UARTCharGet(uart_ports[frame[0]]);

        // If the UART isn't empty, we're done
        if ( (HWREG(uart_ports[frame[0]] + UART_O_FR) & UART_FR_RXFE) == 0 ) {
            frame[0] = the_char;
            return;
        }

        // If we're here, the FIFO is empty now.

        // Figure out if we need to issue the XON.
        UARTFIFOLevelGet(uart_ports[frame[0]],&tx_level,&rx_level);
        if ( rx_level != UART_FIFO_RX1_8 ) { // Send the XON and lower the level
            UARTFIFOLevelSet(uart_ports[frame[0]],tx_level,UART_FIFO_RX1_8);
            UARTCharPut(uart_ports[frame[0]],XON);
            count_sapi_streamgetchar_xon++;
        }

        frame[0] = the_char;
        return;
    }
#endif
}

/// SVC 4: StreamCharsAvail
///
/// Check to see if there is a char available
void __SAPI_StreamCharsAvail( uint32_t *frame )
{
    if (frame[0] < 10 )
    {
        frame[0] = 0;
#if (PLATFORM_LM3S)
        frame[0] = UARTCharsAvail(uart_ports[frame[0]]);
#endif
        return;
    }
}

#endif // SAPI_COMMS

// ------- SVC 5: Reserved -------
// ------- SVC 6: Reserved -------
// ------- SVC 7: Reserved -------

#if (SAPI_WATCHDOG)
// N.B. This section has not been converted to the AAPCS
// compliant call system.

/// @brief SVC 8: Reset the watchdog timer
/// @param Number of clock ticks until next reset
///
/// This one will probably need a little cleanup to be
/// cross-platform.
void __SAPI_WatchdogKick( uint32_t *frame )
{
    WatchdogKick(frame[0]);
}

#endif // SAPI_WATCHDOG

/// @brief SVC 9: GetTimeMS - Return the current value of the ms counter.
///
/// This could be replaced with a direct read, but this
/// allows us to enforce monotonic behavour for updates
// void __SAPI_GetTimeMS( uint32_t *frame )
// {
//     frame[0] = HAL_GetTick();
// }

// *******************************************************************
// Optional calls
// *******************************************************************

#if (SAPI_USAGE)
// N.B. This section has not been converted to the AAPCS
// compliant call system.

extern unsigned int usage;
/// @brief SVC 10: Return the number of CPU cycles in the last second.
void __SAPI_GetUsage( uint32_t *frame )
{
    frame[0] = usage;
}

#endif

// ------- SVC 11: DNSLookup - Defined in sapi-lwip-dns.c -------

// ------- SVC 12: pbuf_get - Defined in sapi_pbuf.c -------
// ------- SVC 13: pbuf_release - Defined in sapi_pbuf.c -------

// ------- SVC 14: udp_new - Defined in sapi-lwip-udp.c -------
// ------- SVC 15: udp_ - Defined in sapi-lwip-udp.c -------
// ------- SVC 16: udp_bind - Defined in sapi-lwip-udp.c -------
// ------- SVC 17: udp_connect - Defined in sapi-lwip-udp.c -------
// ------- SVC 18: udp_disconnect - Defined in sapi-lwip-udp.c -------
// ------- SVC 19: udp_send - Defined in sapi-lwip-udp.c -------
// ------- SVC 20: udp_sendto - Defined in sapi-lwip-udp.c -------

// *******************************************************************
// Helper Functions
// *******************************************************************

#if (0)
/// Depending on how conditional compilation is set, unused/disabled
/// SVC calls are set either to 0 (causes a crash?), or to
/// __SAPI_Disabled which just returns 0xdead and 0xbeef in
/// CPU registers R0 and R1.
/// Catch Disabled API Calls, and return a nonsense value
/// This one was a hang loop, but its better to do something
/// harmless
/// @brief Return a bogus value for disabled calls.
uint64_t __SAPI_Disabled( void )
{
//    uint16_t* MyGPIOG_BSRR; // test using the LED
//    MyGPIOG_BSRR = (uint16_t*)0x40021818;
//    *MyGPIOG_BSRR = 0x2000;    // turn on LED3, PortG.13
    return (0xbeef << 32) | 0xdead;
}
#else
#define __SAPI_Disabled (0)
#endif

#if (SAPI_COMMS)
// N.B. This section has not been converted to the AAPCS
// compliant call system.

// *******************************************************************
// PendSV - Target specific.
// *******************************************************************
static unsigned sapi_count_pendsv = 0; ///< Performance/Validation Counter

/// Low-Priority Handler to spin-wait until buffers drain.
///
/// Hang out here and WFI until both ringbuffers drain to an acceptable level.
/// Don't return control to thread-mode code.
///
/// NOTE NOTE NOTE:  If the NVIC doesn't use the correct priority none
/// of this will work, and it may in fact fail spectacularly
void PendSVHandler ()
{

    if ( pendsv_uart != 0 )
    {
        // Spin and WFI until its not full

#if (PLATFORM_LM3S)
        while(HWREG(pendsv_uart + UART_O_FR) & UART_FR_TXFF)
        {
            sapi_count_pendsv++;
            __asm("wfi");
        }
        pendsv_uart = 0;
#endif // PLATFORM_LM3S
    }
}

#endif // SAPI_COMMS

// ****************************************************************************
// Graphics and TouchScreen interface
// ****************************************************************************

/// @brief SVC 32: wait for the touchscreen to be pressed ( 1 ) or released ( 0 )
// void __SAPI_32_WaitForPressedState( uint32_t *frame )   // SVC 32
// {
//     WaitForPressedState( frame[0] );
// }

void __SAPI_tsWaitUp( void )     // SVC 81
{
    WaitForPressedState( 0 );
}

void __SAPI_tsWaitDown( void )   // SVC 82
{
    WaitForPressedState( 1 );
}

/// @brief SVC 46:
void __SAPI_BSP_LCD_SetLayerWindow( uint16_t Xpos, uint16_t Ypos, uint16_t Width, uint16_t Height )
{
    // ToDo: add index parameter ( fixed to 1 )
    BSP_LCD_SetLayerWindow( 1, Xpos, Ypos, Width, Height );    // SVC 46 ; SFP002
}


void __SAPI_BSP_LCD_FillTriangle( pPoint P1 , pPoint P2 , pPoint P3 )   // SVC 73
{
    BSP_LCD_FillTriangle(P1->X, P2->X, P3->X, P1->Y, P2->Y, P3->Y);
}

extern sFONT Font24;
extern sFONT Font20;
extern sFONT Font16;
extern sFONT Font12;
extern sFONT Font8;

sFONT* FontTable[ 0x08 ] =
{
    &Font24,  // LCD_DEFAULT_FONT
    &Font20,
    &Font16,
    &Font12,
    &Font8,
    0,
    0,
    0
};

/// @brief SVC 78:
uint32_t __SAPI_BSP_LCD_GetFontTable()
{
    return (uint32_t)&FontTable;     // SVC 78
}

/*
uint8_t  BSP_LCD_Init(void);
uint32_t BSP_LCD_GetXSize(void);
uint32_t BSP_LCD_GetYSize(void);
void     BSP_LCD_LayerDefaultInit(uint16_t LayerIndex, uint32_t FrameBuffer);
void	 BSP_LCD_LayerDefaultInitPixelFormat(uint16_t LayerIndex, uint32_t FB_Address, uint32_t PixelFormat);
void     BSP_LCD_SetTransparency(uint32_t LayerIndex, uint8_t Transparency);
void     BSP_LCD_SetLayerAddress(uint32_t LayerIndex, uint32_t Address);
void     BSP_LCD_SetColorKeying(uint32_t LayerIndex, uint32_t RGBValue);
void     BSP_LCD_ResetColorKeying(uint32_t LayerIndex);
void     BSP_LCD_SetLayerWindow(uint16_t LayerIndex, uint16_t Xpos, uint16_t Ypos, uint16_t Width, uint16_t Height);
void     BSP_LCD_SelectLayer(uint32_t LayerIndex);
void     BSP_LCD_SetLayerVisible(uint32_t LayerIndex, FunctionalState state);
void     BSP_LCD_SetTextColor(uint32_t Color);
void     BSP_LCD_SetBackColor(uint32_t Color);
uint32_t BSP_LCD_GetTextColor(void);
uint32_t BSP_LCD_GetBackColor(void);
void     BSP_LCD_SetFont(sFONT *pFonts);
sFONT*   BSP_LCD_GetFont(void);
uint32_t BSP_LCD_ReadPixel(uint16_t Xpos, uint16_t Ypos);
void     BSP_LCD_DrawPixel(uint16_t Xpos, uint16_t Ypos, uint32_t pixel);
void     BSP_LCD_Clear(uint32_t Color);
void     BSP_LCD_ClearStringLine(uint32_t Line);
void     BSP_LCD_DisplayStringAtLine(uint16_t Line, uint8_t *ptr);
void 	 BSP_LCD_DisplayStringAtLineMode(uint16_t Line, uint8_t *ptr, Text_AlignModeTypdef mode);
void     BSP_LCD_DisplayStringAt(uint16_t X, uint16_t Y, uint8_t *pText, Text_AlignModeTypdef mode);
void     BSP_LCD_DisplayChar(uint16_t Xpos, uint16_t Ypos, uint8_t Ascii);
void     BSP_LCD_DrawHLine(uint16_t Xpos, uint16_t Ypos, uint16_t Length);
void     BSP_LCD_DrawVLine(uint16_t Xpos, uint16_t Ypos, uint16_t Length);
void     BSP_LCD_DrawLine(uint16_t X1, uint16_t Y1, uint16_t X2, uint16_t Y2);
void     BSP_LCD_DrawRect(uint16_t Xpos, uint16_t Ypos, uint16_t Width, uint16_t Height);
void     BSP_LCD_DrawCircle(uint16_t Xpos, uint16_t Ypos, uint16_t Radius);
void     BSP_LCD_DrawPolygon(pPoint Points, uint16_t PointCount);
void     BSP_LCD_DrawEllipse(int Xpos, int Ypos, int XRadius, int YRadius);
void     BSP_LCD_DrawBitmap(uint32_t X, uint32_t Y, uint8_t *pBmp);
void     BSP_LCD_FillRect(uint16_t Xpos, uint16_t Ypos, uint16_t Width, uint16_t Height);
void     BSP_LCD_FillCircle(uint16_t Xpos, uint16_t Ypos, uint16_t Radius);
void     BSP_LCD_FillTriangle(uint16_t X1, uint16_t X2, uint16_t X3, uint16_t Y1, uint16_t Y2, uint16_t Y3);
void     BSP_LCD_FillPolygon(pPoint Points, uint16_t PointCount);
void     BSP_LCD_FillEllipse(int Xpos, int Ypos, int XRadius, int YRadius);
void     BSP_LCD_DisplayOff(void);
void     BSP_LCD_DisplayOn(void);
*/


// ****************************************************************************
// ****************************************************************************

#define SIZE_OF_SVC_SYSCALL_TABLE (128)

// This list must be the same as the equivalents in the Forth in SockPuppetInterface.fth
__attribute__((section(".text")))
const void* SVC_syscall_table[ SIZE_OF_SVC_SYSCALL_TABLE ] =
{
    __SAPI_00_ABIVersion,       // SVC 00  - Required
    __SAPI_01_GetLinkList,      // SVC 01  - Required
#if (SAPI_COMMS)
    __SAPI_02_PutChar,          // SVC 02
    __SAPI_03_GetChar,          // SVC 03
    __SAPI_04_GetCharAvail,     // SVC 04
    __SAPI_05_PutCharHasRoom,   // SVC 05
    __SAPI_06_SetIOCallback,    // SVC 06
    __SAPI_07_GetTimeMS,        // SVC 07
    __SAPI_08_NVICReset,        // SVC 08
    __SAPI_Disabled,            // SVC 09
    __SAPI_10_LauchUserApp,     // SVC 10
    __SAPI_11_MPULoad,          // SVC 11
    __SAPI_12_GetPrivs,         // SVC 12
    __SAPI_13_GetUsageCPU,      // SVC 13
    __SAPI_14_PetWatchdog,      // SVC 14
#else
    __SAPI_Disabled,            // SVC 02
    __SAPI_Disabled,            // SVC 03
    __SAPI_Disabled,            // SVC 04
    __SAPI_Disabled,            // SVC 05
    __SAPI_Disabled,            // SVC 06
    __SAPI_Disabled,            // SVC 07
    __SAPI_Disabled,            // SVC 08
    __SAPI_Disabled,            // SVC 09
    __SAPI_Disabled,            // SVC 10
    __SAPI_Disabled,            // SVC 11
    __SAPI_Disabled,            // SVC 12
    __SAPI_Disabled,            // SVC 13
    __SAPI_Disabled,            // SVC 14
#endif
    __SAPI_15_GetFnTable,       // SVC 15
    __SAPI_Disabled,            // SVC 16
    __SAPI_Disabled,            // SVC 17
    __SAPI_Disabled,            // SVC 18
    __SAPI_Disabled,            // SVC 19
#if (SAPI_LWIP)
    __SAPI_20_GetIP,            // SVC 20
    __SAPI_21_PBuf_Ram_Alloc,   // SVC 21
    __SAPI_22_PBuf_Free,        // SVC 22
    __SAPI_23_DNS_GetHostByName,// SVC 23
    __SAPI_24_UDP_GetPCB,       // SVC 24
    __SAPI_25_UDP_Connect,      // SVC 25
    __SAPI_26_UDP_Recv,         // SVC 26
    __SAPI_27_UDP_Send,         // SVC 27
#else
    __SAPI_Disabled,            // SVC 20
    __SAPI_Disabled,            // SVC 21
    __SAPI_Disabled,            // SVC 22
    __SAPI_Disabled,            // SVC 23
    __SAPI_Disabled,            // SVC 24
    __SAPI_Disabled,            // SVC 25
    __SAPI_Disabled,            // SVC 26
    __SAPI_Disabled,            // SVC 27
#endif
    HAL_GetTick,                // SVC 28
//    __SAPI_28_GetTimeMS,        // SVC 28
    __SAPI_29_GetKeyButton,     // SVC 29
    __SAPI_30_SetLeds,          // SVC 30
    __SAPI_31_SetLedsHW,        // SVC 31

// Graphics and Touchscreen
    WaitForPressedState,        // SVC 32
    BSP_TS_GetState,            // SVC 33
    Touchscreen_CalibrationPlus, // SVC 34
    Touchscreen_demo,           // SVC 35
    __SAPI_Disabled,            // SVC 36
    BSP_LCD_Init,                              // SVC 37
    BSP_LCD_GetXSize,                          // SVC 38
    BSP_LCD_GetYSize,                          // SVC 39
    BSP_LCD_LayerDefaultInit,                  // SVC 40
    BSP_LCD_LayerDefaultInitPixelFormat,       // SVC 41
    BSP_LCD_SetTransparency,                   // SVC 42
    BSP_LCD_SetLayerAddress,                   // SVC 43
    BSP_LCD_SetColorKeying,                    // SVC 44
    BSP_LCD_ResetColorKeying,                  // SVC 45
    __SAPI_BSP_LCD_SetLayerWindow,             // SVC 46
    BSP_LCD_SelectLayer,                       // SVC 47
    BSP_LCD_SetLayerVisible,                   // SVC 48
    BSP_LCD_SetTextColor,                      // SVC 49
    BSP_LCD_SetBackColor,                      // SVC 50
    BSP_LCD_GetTextColor,                      // SVC 51
    BSP_LCD_GetTextColor,                      // SVC 52
    BSP_LCD_SetFont,                           // SVC 53
    BSP_LCD_GetFont,                           // SVC 54
    BSP_LCD_ReadPixel,                         // SVC 55
    BSP_LCD_DrawPixel,                         // SVC 56
    BSP_LCD_Clear,                             // SVC 57
    BSP_LCD_ClearStringLine,                   // SVC 58
    BSP_LCD_DisplayStringAtLine,               // SVC 59
    BSP_LCD_DisplayStringAtLineMode,           // SVC 60
    BSP_LCD_DisplayStringAt,                   // SVC 61
    BSP_LCD_DisplayChar,                       // SVC 62
    BSP_LCD_DrawHLine,                         // SVC 63
    BSP_LCD_DrawVLine,                         // SVC 64
    BSP_LCD_DrawLine,                          // SVC 65
    BSP_LCD_DrawRect,                          // SVC 66
    BSP_LCD_DrawCircle,                        // SVC 67
    BSP_LCD_DrawPolygon,                       // SVC 68
    BSP_LCD_DrawEllipse,                       // SVC 69
    BSP_LCD_DrawBitmap,                        // SVC 70
    BSP_LCD_FillRect,                          // SVC 71
    BSP_LCD_FillCircle,                        // SVC 72
    __SAPI_BSP_LCD_FillTriangle,               // SVC 73
    BSP_LCD_FillPolygon,                       // SVC 74
    BSP_LCD_FillEllipse,                       // SVC 75
    BSP_LCD_DisplayOff,                        // SVC 76
    BSP_LCD_DisplayOn,                         // SVC 77
    __SAPI_BSP_LCD_GetFontTable,               // SVC 78

    __SAPI_Disabled,            // SVC 79
    __SAPI_Disabled,            // SVC 80

    __SAPI_tsWaitUp,            // SVC 81
    __SAPI_tsWaitDown,          // SVC 82

    __SAPI_Disabled,            // SVC 83
    __SAPI_Disabled,            // SVC 84
    __SAPI_Disabled,            // SVC 85
    __SAPI_Disabled,            // SVC 86
    __SAPI_Disabled,            // SVC 87
    __SAPI_Disabled,            // SVC 88
    __SAPI_Disabled,            // SVC 89
    __SAPI_Disabled,            // SVC 90
    __SAPI_Disabled,            // SVC 91
    __SAPI_Disabled,            // SVC 92
    __SAPI_Disabled,            // SVC 93
    __SAPI_Disabled,            // SVC 94
    __SAPI_Disabled,            // SVC 95
    __SAPI_Disabled,            // SVC 96
    __SAPI_Disabled,            // SVC 97
    __SAPI_Disabled,            // SVC 98
    __SAPI_Disabled,            // SVC 99
    __SAPI_Disabled,            // SVC 100
    __SAPI_Disabled,            // SVC 101
    __SAPI_Disabled,            // SVC 102
    __SAPI_Disabled,            // SVC 103
    __SAPI_Disabled,            // SVC 104
    __SAPI_Disabled,            // SVC 105
    __SAPI_Disabled,            // SVC 106
    __SAPI_Disabled,            // SVC 107
    __SAPI_Disabled,            // SVC 108
    __SAPI_Disabled,            // SVC 109
    __SAPI_Disabled,            // SVC 110
    __SAPI_Disabled,            // SVC 111
    __SAPI_Disabled,            // SVC 112
    __SAPI_Disabled,            // SVC 113
    __SAPI_Disabled,            // SVC 114
    __SAPI_Disabled,            // SVC 115
    __SAPI_Disabled,            // SVC 116
    __SAPI_Disabled,            // SVC 117
    __SAPI_Disabled,            // SVC 118
    __SAPI_Disabled,            // SVC 119
    __SAPI_Disabled,            // SVC 120
    __SAPI_Disabled,            // SVC 121
    __SAPI_Disabled,            // SVC 122
    __SAPI_Disabled,            // SVC 123
    __SAPI_Disabled,            // SVC 124
    __SAPI_Disabled,            // SVC 125
    __SAPI_Disabled,            // SVC 126
    __SAPI_Disabled             // SVC 127

}; // SVC_syscall_table end

/**
  * @brief  This function handles the SVCall exception.
  * @param  None
  * @retval None
  */
void SVC_Handler(void)
{
// on entry, R0..R3, R12, LR, PC, PSR have been saved
    __asm( "tst	   lr, #0x4" );         // Figure out which stack
    __asm( "ite	   eq" );
    __asm( "mrseq  r0,msp" );           // Main stack
    __asm( "mrsne  r0,psp" );           // Process/Thread Stack

    __asm( "push   { r4, lr }" );	// now we can preserve more
    __asm( "mov    r4, r0" );		// copy SP; R4 is preserved by AAPCS
    __asm( "ldr    r1,[r0,#24]" );      // get the stacked PC
    __asm( "ldrb   r1,[r1,#-2]" );      // extract the svc call number
                                        // range checking goes here
    __asm( "ldr    r2 ,=SVC_syscall_table" );
    __asm( "ldr    r12, [r2, r1, LSL #2]" );  // R12 = function address
    __asm( "ldm    r4, { r0-r3 }" );    // restore AAPCS arguments from R0..R3
    __asm( "blx    r12" );              // call function
    __asm( "str    r0, [ r4, #0 ]" );   // save return value
    __asm( "str    r1, [ r4, #4 ]" );   // save return value
    __asm( "pop    { r4, pc }" );       // restore
}

//! // Close the Doxygen group.
//! @}

