\ SockPuppetInterface.fth - access a C library from Forth.

((
Copyright(C) 2011, 2012, 2013, 2014
Robert Sexton

Copyright (c) 2015, 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)

To do
=====

Change history
==============
20160411 SFP003 Used AAPCS compiler for SVC calls.
20160229 SFP002 DocGen overhaul.
		Structure overhaul.
		Files refactored.
20151231 HPO001 Initial creation from RBS000.
))

decimal

\ =============
\ *! sockpuppet
\ *T Mixed Language Programming using SockPuppet
\ =============

\ ***************
\ *S Introduction
\ ***************
\ *P With the ever increasing complexity of microcomputers such
\ ** as the Cortex cores and systems, manufacturers are providing
\ ** development hardware and software systems based on C libraries
\ ** to make using their chips easier. Such libraries reduce the
\ ** requirement for chip documentation.
\ ** The conventional approach to providing support for development
\ ** boards in Forth has been to manually port the C library sources
\ ** to Forth. The *\b{SockPuppet} system takes a different approach
\ ** by providing an interface solution between Forth and C; the
\ ** Forth system calls the underlying C libraries. In turn, this
\ ** allows the details of the hardware to be abstracted away by
\ **  the C libraries, whilst allowing the Forth system to provide
\ ** a powerful, uniform and interactive user interface.

\ *P The MPE ARM/Cortex Forth cross-compiler supports calling
\ ** functions in C or any language that can provide functions
\ ** that use the AAPCS calling convention. This is an ARM
\ ** convention documented in *\i{IHI0042F_aapcs.pdf}.
\ ** Calls with a variable number of parameters (*\b{varargs}) are
\ ** not supported.

\ *P The example code in both Forth and C is available for the
\ ** Professional versions of the ARM Cortex cross compiler.
\ ** The example code provides a simple GUI for an STM32F429I
\ ** Discovery board using sample C code provided by ST and others.
\ ** The interface is defined for Cortex-M CPUs only. If you need
\ ** an ARM32 interface, contact MPE.

\ *P This work is directly inspired by Robert Sexton's Sockpuppet
\ ** interface:
\ *C   https://github.com/rbsexton/sockpuppet
\ *P His contribution and permission are gratefully acknowledged.


\ *************************************
\ *S How the Forth to C interface works
\ *************************************
\ *P Each function that is exported from the C world to the
\ ** Forth world appears as one of a number of types of call.
\ ** These words are called *\fo{externs}. You can handcraft
\ ** these words in assembler, but the compiler includes code
\ ** generators for several techniques. The call format and return
\ ** values match the AAPCS standard used by ARM C compilers.

\ *P There are several ways to call *\fo{externs}. They have their
\ ** pros and cons and are discussed in following sections.
\ *(
\ *B *\b{SVC calls}. You just need to know the SVC numbers. SVC calls
\ ** provide the greatest isolation between sections of code
\ ** written in other languages. The functions foreign to
\ ** Forth are accessed by SVC calls and/or jump tables. The
\ ** example solution uses SVC calls for most foreign functions.
\ ** Regardless of the primary call technique used, all techniques
\ ** rely on a small number of SVC calls.
\ *B *\b{Jump table}. The base address of the table can be set
\ ** at run time, e.g. by making a specific SVC call. The calling
\ ** words fetch the run-time address from the table, given an
\ ** index.
\ *B *\b{Double indirect call}. A primary jump table is at a fixed
\ ** address and contains the addresses of secondary tables,
\ ** which hold the actual routine addresses. The fixed address
\ ** and both indices must be known at compile time. This
\ ** technique is used by TI's Stellaris parts and some NXP parts
\ ** to accesss driver code in ROM.
\ *B *\b{Direct calls} to the address of the routine. You need to
\ ** know the address at compile time.
\ *)

\ *P There is a practical limit of four arguments if you use SVC
\ ** calls for the insulation between Forth and C. If you want
\ ** this limit changed, call MPE! The other interface methods
\ ** do not suffer from this limit.

\ *P It is a matter of convention between the Forth and C code
\ ** as to parameter passing order. It can be changed by either
\ ** side. MPE convention is for the left-most Forth parameter
\ ** to be passed in R0. This matches the AAPCS code used by the
\ ** hosted Forth compilers such as VFX Forth for ARM Linux.
\ ** We strongly suggest that you use the MPE convention in
\ ** order to take advantage of future Sockpuppet developments.


\ ============
\ *N SVC calls
\ ============
\ *P The examples use the MPE calling convention and are
\ ** illustrated in assembler as well as by using the code
\ ** generator.

\ *E SVC( 67 ) void BSP_LCD_DrawCircle( int x, int y, int r );
\ ** \ SVC 67 draw a circle of radius r at position (x,y).

\ *E CODE BSP_LCD_DrawCircle \ x y r --
\ ** \ SVC 67 draw a circle of radius r at position (x,y).
\ **   mov r2, tos                     \ r
\ **   ldr r1, [ psp ], # 4            \ y
\ **   ldr r0, [ psp ], # 4            \ x
\ **   svc # __SAPI_BSP_LCD_DrawCircle
\ **   ldr     tos, [ psp ], # 4       \ restore TOS
\ **   next,
\ ** END-CODE

\ *P When the SVC call occurs, the Cortex CPU stacks registers
\ ** R0-R3, R12, LR, PC, xPSR on the calling R13 stack with R0 at
\ ** the lowest address. The SVC handler places the address of
\ ** this frame in R0/R4, extracts the SVC call number, reloads
\ ** the AAPCS parameters from the frame and jumps to
\ ** the appropriate C function. In this case
\ *E void BSP_LCD_DrawCircle(
\ **   uint16_t Xpos, uint16_t Ypos, uint16_t Radius
\ ** );

\ *P SVC calls provide the highest insulation between Forth and
\ ** C, but suffer from several issues.
\ *(
\ *B The SVC call mechanism is part of the Cortex interrupt and
\ ** exception system. The assembler and/or C side of this uses
\ ** code written in assembler to allow the C routines called
\ ** from a jump table to to be AAPCS compliant.
\ *B The SVC mechanism is inefficient compared to a direct
\ ** AAPCS handler.
\ *B Because SVC calls are part of the CPU interrupt mechanism,
\ ** you have to care how long a call takes. Playing games with
\ ** Cortex interrupt mechanism can fix this, but is complex.
\ *)


\ =============
\ *N Jump table
\ =============
\ *P In order to avoid the penalties of the SVC call mechnism,
\ ** you can make an array of function pointers in C or assembler
\ ** and call functions using an index into the table.
\ *E jumptable:
\ **   dd func0         ; address of function 0
\ **   dd func1         ; address of function 1
\ **   ..

\ *P We still need to know the address of the jump table. This
\ ** is found using an SVC call (15) and stored in a variable.
\ ** The jump table address could be hard-coded, but given the
\ ** horrors of perverting link map files and the like, the
\ ** overhead of a single SVC call is preferable.

\ *E SVC( 15 ) void * GetDirFnTable( void );
\ ** \ Define SVC call that returns the address of the
\ ** \ jump table.
\ **
\ ** variable JT     \ -- addr
\ ** \ Holds the address of the jump table.
\ ** JT holdsJumpTable
\ ** \ Tell the cross compiler where the jump table address
\ ** \ is held.
\ **
\ ** : initJTI       \ -- ; initialise jump table calls
\ **   GetDirFnTable JT !  ;
\ **
\ ** JTI( n ) int open(
\ **   const char * pathname, int flags, mode_t mode
\ ** );

\ *P If constructed in assembler, the SVC despatch table and
\ ** the main jump table can be the same table; it's just a
\ ** question of what you put in the table.


\ ==============================
\ *N Double indirect call tables
\ ==============================
\ ** This form is used when calling ROM routines in devices
\ ** from TI, NXP and others. In these the primary table entries
\ ** hold the address of other tables. The secondary index
\ ** is then used to fetch the address of the actual routine
\ ** to be called.

\ *P Before use, you have to declare the base address of the
\ ** primary ROM table used for calling ROM functions. For
\ ** Luminary/TI CPUs, this will probably be:
\ *C  $0100:0010 setPriTable

\ *P Now you can define a set of ROM calls, for example,
\ ** again for a TI CPU.
\ *E DIC( 4, 0 ) void ROM_GPIOPinWrite(
\ **    uint32 ui32Port, uint8 ui8Pins, uint8 ui8Val
\ ** );
\ *P where:
\ *(
\ *B *\fo{ROM_APITABLE} is an array of pointers located at
\ ** *\fo{0x0100.0010}.
\ *B *\fo{ROM_GPIOTABLE} is an array of pointers located at
\ ** *\fo{ROM_APITABLE[4]}.
\ *B *\fo{ROM_GPIOPinWrite} is a function pointer located at
\ ** *\fo{ROM_GPIOTABLE[0]}.
\ *)
\ *P Parameters:
\ *(
\ *B *\fo{ui32Port} is the base address of the GPIO port.
\ *B *\fo{ui8Pins} is the bit-packed representation of the pin(s).
\ *B *\fo{ui8Val} is the value to write to the pin(s).
\ *)
\ *P Description:*\br{}
\ ** Writes the corresponding bit values to the output pin(s)
\ ** specified by *\fo{ui8Pins}. Writing to a pin configured as an
\ ** input pin has no effect. The pin(s) are specified using a
\ ** bit-packed byte, where each bit that is set identifies the
\ ** pin to be accessed, and where bit 0 of the byte represents
\ ** GPIO port pin 0, bit 1 represents GPIO port pin 1, and so
\ ** on.
\ *P Returns:*\br{}
\ ** None.
\ *P To call this function, use the Forth form:
\ *C   port pins val ROM_GPIOPinWrite


\ ===============
\ *N Direct calls
\ ===============
\ *P Where the address of the routine is known at the Forth
\ ** compile time, you can use a direct call.
\ *C   DIR( addr ) int foo( int a, char *b, char c );
\ *P The Forth word marshalls the parameters and calls the
\ ** subroutine at target address *\i{addr}.


\ ***********************
\ *S SVC call number list
\ ***********************
\ *P The demonstration code provided here mostly uses SVC calls.
\ ** To call the SVC, a small Forth word puts the arguments
\ ** into registers, calls the SVC, and then returns a value if
\ ** required. SVCs are identified by a number, which corresponds
\ ** here to an index into a a table.

\ *P The list must match the equivalents in the C code
\ ** *\f{SVC_syscall_table[]}. To ease sharing of data between
\ ** the Forth and C systems, you can use the *\fo{enum} parser
\ ** described in the "Interpreter directives" chapter of the
\ ** generic cross compiler manual.

\ *P The first 16 entries (0..15) are defined by the Sockpuppet
\ ** system. Each application is free to use entries 16 onwards
\ ** as suits the application.

\ *P This list will be different for every application.

\ *[
const=equ	\ constants invisible on target
\ const=constant  \ constants visible on target
#00 const __SAPI_00_ABIVersion          \ SVC 00  - Required
#01 const __SAPI_01_GetLinkList         \ SVC 01  - Required
#02 const __SAPI_02_PutChar             \ SVC 02
#03 const __SAPI_03_GetChar             \ SVC 03
#04 const __SAPI_04_GetCharAvail        \ SVC 04
#05 const __SAPI_05_PutCharHasRoom      \ SVC 05
#06 const __SAPI_06_SetIOCallback       \ SVC 06
#07 const __SAPI_07_GetTimeMS           \ SVC 07 - Required
#08 const __SAPI_08_NVICReset           \ SVC 08

#10 const __SAPI_10_LauchUserApp        \ SVC 10 - Reserved
#11 const __SAPI_11_MPULoad             \ SVC 11 - Reserved
#12 const __SAPI_12_GetPrivs            \ SVC 12 - Reserved
#13 const __SAPI_13_GetUsageCPU         \ SVC 13 - Reserved
#14 const __SAPI_14_PetWatchdog         \ SVC 14 - Reserved
#15 const __SAPI_15_GetFnTable          \ SVC 15 - Required

#20 const __SAPI_20_GetIP               \ SVC 20
#21 const __SAPI_21_PBuf_Ram_Alloc      \ SVC 21
#22 const __SAPI_22_PBuf_Free           \ SVC 22
#23 const __SAPI_23_DNS_GetHostByName   \ SVC 23
#24 const __SAPI_24_UDP_GetPCB          \ SVC 24
#25 const __SAPI_25_UDP_Connect         \ SVC 25
#26 const __SAPI_26_UDP_Recv            \ SVC 26
#27 const __SAPI_27_UDP_Send            \ SVC 27
#28 const __SAPI_28_GetTimeMS           \ SVC 28
#29 const __SAPI_29_GetKeyButton        \ SVC 29
#30 const __SAPI_30_SetLeds             \ SVC 30
#31 const __SAPI_31_SetLedsHW           \ SVC 31
\ Graphics and Touchscreen
#32 const __SAPI_32_WaitForPressedState \ SVC 32
#33 const __SAPI_33_BSP_TS_GetState     \ SVC 33
#34 const __SAPI_34_Touchscreen_Cal     \ SVC 34
#35 const __SAPI_35_Touchscreen_demo    \ SVC 35

#37 const __SAPI_BSP_LCD_Init           \ SVC 37
#38 const __SAPI_BSP_LCD_GetXSize       \ SVC 38
#39 const __SAPI_BSP_LCD_GetYSize       \ SVC 39
#40 const __SAPI_BSP_LCD_LayerDefaultInit           \ SVC 40
#41 const __SAPI_BSP_LCD_LayerDefaultInitPixelForm  \ SVC 41
#42 const __SAPI_BSP_LCD_SetTransparency            \ SVC 42
#43 const __SAPI_BSP_LCD_SetLayerAddress            \ SVC 43
#44 const __SAPI_BSP_LCD_SetColorKeying             \ SVC 44
#45 const __SAPI_BSP_LCD_ResetColorKeying           \ SVC 45
#46 const __SAPI_BSP_LCD_SetLayerWindow             \ SVC 46
#47 const __SAPI_BSP_LCD_SelectLayer                \ SVC 47
#48 const __SAPI_BSP_LCD_SetLayerVisible            \ SVC 48
#49 const __SAPI_BSP_LCD_SetTextColor   \ SVC 49
#50 const __SAPI_BSP_LCD_SetBackColor   \ SVC 50
#51 const __SAPI_BSP_LCD_GetTextColor   \ SVC 51
#52 const __SAPI_BSP_LCD_GetBackColor   \ SVC 52
#53 const __SAPI_BSP_LCD_SetFont        \ SVC 53
#54 const __SAPI_BSP_LCD_GetFont        \ SVC 54
#55 const __SAPI_BSP_LCD_ReadPixel      \ SVC 55
#56 const __SAPI_BSP_LCD_DrawPixel      \ SVC 56
#57 const __SAPI_BSP_LCD_Clear          \ SVC 57
#58 const __SAPI_BSP_LCD_ClearStringLine      \ SVC 58
#59 const __SAPI_BSP_LCD_DisplayStringAtLine  \ SVC 59
#60 const __SAPI_BSP_LCD_StrAtLineMode        \ SVC 60
#61 const __SAPI_BSP_LCD_DisplayStringAt      \ SVC 61
#62 const __SAPI_BSP_LCD_DisplayChar    \ SVC 62
#63 const __SAPI_BSP_LCD_DrawHLine      \ SVC 63
#64 const __SAPI_BSP_LCD_DrawVLine      \ SVC 64
#65 const __SAPI_BSP_LCD_DrawLine       \ SVC 65
#66 const __SAPI_BSP_LCD_DrawRect       \ SVC 66
#67 const __SAPI_BSP_LCD_DrawCircle     \ SVC 67
#68 const __SAPI_BSP_LCD_DrawPolygon    \ SVC 68
#69 const __SAPI_BSP_LCD_DrawEllipse    \ SVC 69
#70 const __SAPI_BSP_LCD_DrawBitmap     \ SVC 70
#71 const __SAPI_BSP_LCD_FillRect       \ SVC 71
#72 const __SAPI_BSP_LCD_FillCircle     \ SVC 72
#73 const __SAPI_BSP_LCD_FillTriangle   \ SVC 73
#74 const __SAPI_BSP_LCD_FillPolygon    \ SVC 74
#75 const __SAPI_BSP_LCD_FillEllipse    \ SVC 75
#76 const __SAPI_BSP_LCD_DisplayOff     \ SVC 76
#77 const __SAPI_BSP_LCD_DisplayOn      \ SVC 77
#78 const __SAPI_BSP_LCD_GetFontTable   \ SVC 78

#81 const __SAPI_tsWaitUp               \ SVC 81
#82 const __SAPI_tsWaitDown             \ SVC 82
\ *]


\ *******************************
\ *S Words containing an SVC call
\ *******************************

-saveR9		\ R9 does not need to be saved

svc( 0 ) int SAPI-Version( void );
\ *G SVC 00: Return the version of the API in use.
svc( 1 ) int GetSharedVars( void );
\ *G SVC 01: Get the address of the shared variable list.
svc( 15 ) int GetSvcFnTable( void );
\ *G SVC 15: Get the address of the SVC function table.

: FN  \ n --
\ *G Call a function in the SVC function table directly.
\ ** SVC calls that take a long time may/will block interrupts
\ ** such as the system ticker, and thus will fail. To avoid
\ ** this, such calls should be called directly so that they
\ ** do not affect the interrupt priority level. The functions
\ ** *\b{must} be declared as
\ *C   void function( void );
\ *P as the normal SVC parameter passing mechanism is completely
\ ** bypassed. Note also that these functions inhibit the
\ ** Forth multitasker for the duration of the call. It is
\ ** usually better to refactor the C library if you wish to
\ ** use the Forth multitasker.
   cells GetSvcFnTable + @ execute
;

svc( 28 ) int ticks( void );
\ *G SVC 28: Get the ms counter. Returns a value in milliseconds
\ ** that eventually wraps around.
svc( 29 ) int SAPI_GetKeyButton( void );
\ *G SVC 29: Get the Key Button value
svc( 30 ) void SAPI_SetLeds( int mask );
\ *G SVC 30: Set LED3 ( bit0 ) and LED4 ( bit1 ) via BSP.
svc( 31 ) void SAPI_SetLedsHW( int mask );
\ *G SVC 31: Set LED3 ( bit0 ) and LED4 ( bit1 ) via direct hardware access
svc( 33 ) void SAPI_33_BSP_TS_GetState( void * TSstruct );
\ *G SVC 33: Get the touchscreen coordinates into a structure.


\ *****************
\ *S Reserved calls
\ *****************
\ *P These calls are commented out by default and are not present
\ ** in all systems. However, the SVC numbers are reserved. Note
\ ** that some stack manipulation is required in some words - they
\ ** may not be direct translations.

0 [if]

CODE SetIOCallback  \ addr iovec read/write -- n
\ *G Set the IO Callback address for a stream.
\ ** *\i{Iovec}, read/write (r=1), address to set as zero.
\ ** Note the minor parameter swizzle here to keep the old value on TOS.
    mov r1, tos
    ldr r0, [ psp ], # 4 !
    ldr r2, [ psp ], # 4 !
    svc # __SAPI_06_SetIOCallback
    mov tos, r0
    next,
END-CODE

CODE RestartForthApp ( addr ) \ --
\ *G Do a stack switch and startup the user App.  Its a one-way
\ ** trip, so don't worry about stack cleanup.
    mov r0, tos
    svc # __SAPI_10_LauchUserApp
    next,
END-CODE

CODE MPULoad  \ --
\ *G Ask for MPU entry updates.
    mov r0, tos
    svc # __SAPI_11_MPULoad
    ldr tos, [ psp ], # $04
    next,
END-CODE

CODE privmode  \ --
\ *G Request Privileged Mode. In some systems, this is a huge
\ ** security hole.
    svc # __SAPI_12_GetPrivs
    next,
END-CODE

CODE PetWatchDog  \ --
\ *G Refresh the watchdog
    svc # __SAPI_14_PetWatchdog
    next,
END-CODE

CODE GetUsage  \ -- n
\ *G The number of CPU cycles consumed in the last second.
    svc # __SAPI_13_GetUsageCPU
    str tos, [ psp, # -4 ] !
    mov tos, r0
    next,
END-CODE

[then]


\ ****************
\ *S GUI SVC calls
\ ****************
\ *P These calls are defined for the demonstration code. Your
\ ** application code may reuse the numbers for other functions.

SVC( 49 ) void BSP_LCD_SetTextColor( int color );
\ *G SVC 49: Set text and line drawing colour.
SVC( 50 ) void BSP_LCD_SetBackColor( int color );
\ *G SVC 50: Set background colour for text
SVC( 53 ) void BSP_LCD_SetFont( int font );
svc( 56 ) void BSP_LCD_DrawPixel( int x, int y, int colour );
\ *G SVC 56: Draw a pixel at a at screen position (x,y)
svc( 57 ) void SAPI_BSP_LCD_Clear( int colour );
\ *G SVC 57: Set the LCD to the given colour.
svc( 60 ) void BSP_LCD_DisplayStringAtLineMode( int line, char * string, int mode);
\ *G SVC 60 Display the text at a on line a using the given mode
svc( 61 ) void BSP_LCD_DisplayStringAt( int x, int y, char * string, int mode );
\ *G Draw a string at position (x,y) in the given alignment mode.
\ *[
$01 constant CENTER_MODE   \ center mode
$02 constant RIGHT_MODE    \ right mode
$03 constant LEFT_MODE     \ left mode
\ *]
svc( 63 ) void BSP_LCD_DrawHLine( int x, int y, int length );
\ *G SVC 63: Draw a horizontal line at position (x,y) of length.
svc( 64 ) void BSP_LCD_DrawVLine( int x, int y, int length );
\ *G SVC 64: Draw a vertical line at position (x,y) of length.
svc( 65 ) void BSP_LCD_DrawLine( int x1, int y1, int x2, int y2 );
\ *G SVC 65: Draw a line between two points.
svc( 66 ) void BSP_LCD_DrawRect( int x, int y, int w, int h );
\ *G SVC 66: Draw a rectangle.
svc( 67 ) void BSP_LCD_DrawCircle( int x, int y, int radius );
\ *G SVC 67: Draw a circle of the given radius at screen position (x,y)
SVC( 70 ) void BSP_LCD_DrawBitmap( int x, int y, void * image );
\ *G SVC 70: draw the bitmap image at addr at the screen position (x,y).
svc( 71 ) void BSP_LCD_FillRect( int x, int y, int w, int h );
\ *G SVC 71: Draw a filled rectangle at (x,y) of size (w,h).
svc( 72 ) void BSP_LCD_FillCircle( int x, int y, int radius );
\ *G SVC 72 draw a filled circle of the given radius at screen position (x,y)
SVC( 78 ) void * BSP_LCD_GetFontTable( void );
: Font24 ( -- font )   BSP_LCD_GetFontTable 0 cells + @ ;
: Font20 ( -- font )   BSP_LCD_GetFontTable 1 cells + @ ;
: Font16 ( -- font )   BSP_LCD_GetFontTable 2 cells + @ ;
: Font12 ( -- font )   BSP_LCD_GetFontTable 3 cells + @ ;
: Font8  ( -- font )   BSP_LCD_GetFontTable 4 cells + @ ;


\ **************
\ *S Debug tools
\ **************

: mainSP@	\ -- addr
\ *G Return the main stack pointer.
  SP_main sys@  ;
: procSP@	\ -- addr
\ *G Return the process stack pointer.
  SP_process sys@  ;


\ **********
\ Test code
\ **********

1 [if]

: ttfr  123 10 10 10 10  BSP_LCD_FillRect  . ;

: ttrect    \ --
   LCD_COLOR_DARKBLUE lcdClear
   LCD_COLOR_GREEN BSP_LCD_SetTextColor
   10 10 60 60 BSP_LCD_DrawRect
   10 10 70 70 BSP_LCD_DrawLine

   LCD_COLOR_YELLOW BSP_LCD_SetTextColor
   40 110 60 60 BSP_LCD_FillRect

   LCD_COLOR_LTGRAY setFColor
   70 180 60 60 absFilledRect
   LCD_COLOR_LTGRAY setFColor
   70 180 90 BSP_LCD_DrawVLine
   70 180 90 BSP_LCD_DrawHLine

   LCD_COLOR_GREEN BSP_LCD_SetTextColor
   200 220 SetCursor
   30 hline
   -30 hline
   30 vline
   -30 vline

   LCD_COLOR_RED BSP_LCD_SetTextColor
   180 120 SetCursor
   xy> 40 + swap  0 + swap  gline
   xy> 40 + swap 40 + swap  gline
   xy>  0 + swap 40 + swap  gline
\   xy> -40 + swap -40 + swap  gline

   LCD_COLOR_YELLOW BSP_LCD_SetTextColor
   180 240  30 h-line
   180 240  -30 h-line
   180 240  30 v-line
   180 240  -30 v-line

   LCD_COLOR_WHITE BSP_LCD_SetTextColor
   180 180 SetCursor
   40  0 line
   40 40 line
    0 40 line
\   -40 -40 line

   LCD_COLOR_RED setFColor
   170 30 60 60 absFilledRect
   LCD_COLOR_YELLOW setFColor
   180 40 40 40 absRect
;

((
: ttstk  \ -- a
   $0 $0 $0800aa88 BSP_LCD_DrawBitmap
   \ Note the number is the address of the stlogo bitmap array in the C section - YMMV
;

: ttstk2  \ -- a
   #20 $0 image2 BSP_LCD_DrawBitmap
;
))

[then]


\ ======
\ *> ###
\ ======
