\ CdcDrv.fth - USB Communications Device Class operations

((
Copyright (c) 2008, 2009, 2010
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
fax:   +44 (0)23 8033 9691
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====
Correct CDC DMA OUT so that packets can be stored until there is
room in the serial queue.

Perform CDC compliance tests.

Change history
==============
20100906 SFP003 Added support for USBmin?.
20090723 SFP002 Isolated CDC UART from this code.
		See VCOMdev.fth.
20081030 SFP001 Initial release.
))

only forth definitions
decimal

\ =========
\ *! cdcdrv
\ *T Communications Device Class (CDC) driver
\ =========
\ *P The Communications Device Class (CDC) driver uses the ACM
\ ** subclass. Only a limited set of control functions are
\ ** implemented beyond providing a stub. If you need them,
\ ** expand them to suit your requirements.

\ *P There are two USB interfaces, a control interface with an
\ ** interrupt in endpoint, and a data interface with a bulk in
\ ** and and bulk out endpoint per COM channel.

\ *P The COM channel(s) are implemented as the ends of *\fo{CQUEUE}
\ ** structures. These queues should be large enough to hold the
\ ** number of bytes accumulated during a USB poll interval. The
\ ** poll interval is defined in the endpoint descriptor for the
\ ** control interface. By default, this is set to 1, indicating
\ ** a poll interval of 1 ms.


\ *******************
\ *S Data and buffers
\ *******************

-1 value RunCDC?	\ -- x
\ *G Set non-zero when CDC operation is enabled.

#10 aligned buffer: NotifyBuf	\ -- addr
\ *G Holds CDC notification data.
7 aligned buffer: LineCodingBuf	\ -- addr
\ *G Holds current line coding information.
2 buffer: SerialState		\ -- addr
\ *G Holds serial state.
2 buffer: LineControlState	\ -- addr
\ *G Holds Line Control State info.

create InitNotifyBuf	\ -- addr
\ *G Initialisation data for the notify buffer
  0xA1 c,		\ 0: bmRequestType
  0x20 c,		\ 1: bNotification (SERIAL_STATE)
  0x00 lew, 		\ 2: wValue
  0x00 lew,		\ 4: wIndex (Interface #, LSB first)
  0x02 lew,		\ 6: wLength (Data length = 2 bytes, LSB first)
  0x00 lew,		\ 8: UART State Bitmap (16bits, LSB first)
create InitLineCoding
\ *G Initialisation data for the line coding,
\ ** usbcdc11.pdf 6.2.13 p57.
  #115200 lel,		\ 0: baud rate
  0 c,			\ 4: 1 stop bit
  0 c,			\ 5: no parity
  8 c,			\ 6: 8 data bits


\ *************************
\ *S CDC Class RTI requests
\ *************************

: CDCdoEP0ClassRTI	\ --
\ *G Handle the CDC class REQUEST_TO_INTERFACE operations.
  case  SetupPacket usp.bRequest c@
    SEND_ENCAPSULATED_COMMAND of	\ 0x00
      EP0Buf EP0Data ep.*Data !
    endof
    GET_ENCAPSULATED_RESPONSE of	\ 0x01
      USBStatusInStage			\ just return a zero length packet
    endof
    SET_COMM_FEATURE of			\ 0x02
    endof
    SET_LINE_CODING of			\ 0x20
      LineCodingBuf EP0Data ep.*Data !	\ for data out stage
    endof
    GET_LINE_CODING of			\ 0x21
      $80 LineCodingBuf 7 USBWriteEP +EP0Data
    endof
    SET_CONTROL_LINE_STATE of		\ 0x22
      USBStatusInStage			\ just return a zero length packet
    endof
    SEND_BREAK of
    endof
      stall_i				\ unknown/unhandled
  endcase
;


\ *****************************
\ *S CDC Bulk endpoint handling
\ *****************************

variable #LastIn0	\ -- addr
\ *G Holds the number of bytes in the last packet sent to the
\ ** host. This is used to prevent a Windows problem that occurs
\ ** when the last OUT packet contains *\fo{/MaxPacket} bytes.
\ ** In this case a following zero-length packet is required.


\ ====================
\ *N Non DMA operation
\ ====================
\ *P Operation without DMA drivers is portable across a wide range
\ ** of CPUs and USB engines.

usbDMA? 0= [if]

/MaxPacket buffer: CdcBulkBuf	\ -- addr
\ *G Packet buffer for CDC bulk transfers.

: CdcBulkOut	\ --
\ *G Handle bulk OUT actions to the device. Note that this
\ ** includes NAK interrupts.
  UsbEPhalt CdcBoutEPMask and 0= if	\ if the endpoint is not stalled
    /MaxPacket cdcRoomIn?0 if		\ queue full check ; SFP002
      CDCBoutEP CdcBulkBuf USBReadEP
      CdcBulkBuff swap cdcRcv0		\ transfer to UART ; SFP002
    else
      -BulkNAKout
    endif
  endif
;

: CdcBulkIn	\ --
\ *G Handle bulk IN actions to the host. Note that this includes
\ ** NAK interrupts.
  UsbEPhalt CdcBinEPMask and 0= if	\ if the endpoint is not stalled
    #cdcTxOut0 /MaxPacket Min dup if	\ anything to send ; SFP002
      >r
      CdcBulkBuf r@ cdcTxmit0		\ send to host ; SFP002
      CDCBinEP CdcBulkBuf r@ USBWriteEP drop
      r> #LastIn0 !
    else
      #LastIn0 @ /MaxPacket =		\ avoid Windows problem
      if  CDCBinEP sendZLPin  endif
      #LastIn0 !  -BulkNAKin		\ no more to send
    endif
  endif
;

: doEP5		\ event --
\ *G The action of endpoint 5, which is for CDC bulk IN and OUT.
  case
    USB_EVT_OUT of  CdcBulkOut  endof
    USB_EVT_IN  of  CdcBulkIn  endof
  endcase
;
' doEP5 5 SetEPhandler

((
: CdcFrameCheck	\ --
\ *G Executed in the frame interrupt. If there is data to be sent
\ ** to the host, Bulk NAK IN interrupts are enabled.
\ ** No longer needed as Bulk NAK IN interrupts are always enabled
\ ** in the frame interrupt.
  #cdcTxOut0 if
    +BulkNAKin
  endif
;
' CdcFrameCheck AtUSBframe
))

[then]


\ ================
\ *N DMA operation
\ ================
\ *P Because of the variety of DMA engines used for DMA, isolation
\ ** of a hardware dependent DMA layer is more complex. As yet, we
\ ** do not guarantee that the hardware dependent DMA layer is
\ ** portable to all CPUs and USB engines.

usbDMA? [if]

[sections URAMP  uAlign
CDCBoutEP BulkDD:
\ *G Bulk OUT DMA descriptor.
CDCBinEP BulkDD:
\ *G Bulk IN DMA descriptor.
/MaxPacket buffer: CdcDmaOutBuff	\ -- addr
\ *G DMA buffer for data OUT from host.
/MaxPacket buffer: CdcDmaInBuff	\ -- addr
\ *G DMA buffer for data IN to host.
sections]

CDCBoutEP EPaddr equ CDCBoutPhys	\ -- u
\ *G Physical address of the CDC Bulk OUT channel.
CDCBinEP EPaddr equ CDCBinPhys	\ -- u
\ *G Physical address of the CDC Bulk IN channel.

: initCDCbo	\ --
\ *G Start the next CDC Bulk out DMA transfer.
  CdcDmaOutBuff /MaxPacket CDCBoutEP setEPdma
;

: initCDCbi	\ --
\ *G Initialise the CDC Bulk IN DMA transfer.
  CdcDmaInBuff 0 CDCBinEP setEPdma
;

: CdcBulkDmaOut	\ --
\ *G Handle CDC bulk DMA OUT actions to the device.
  UsbEPhalt CdcBoutEPMask and 0= if	\ if the endpoint is not stalled
    UDCA CDCBoutPhys cells + @		\ -- dd
    3 cells + @ #16 rshift		\ -- lenr
    dup cdcRoomIn?0 if			\ -- lenr ; queue full check ; SFP002
      CdcDmaOutBuff swap cdcRcv0	\ send to UART ; SFP002
    else
      drop
    endif
    initCDCbo				\ reinitialise
  endif
;

: CdcBulkDmaIn	\ --
\ *G Handle CDC bulk IN actions to the host.
  UsbEPhalt CdcBinEPMask and 0= if	\ if the endpoint is not stalled
    #cdcTxOut0 /MaxPacket Min dup if
      >r
      CdcDmaInBuff r@ cdcTxmit0		\ take characters from UART ; SFP002
      CdcDmaInBuff r@ CDCBinEP setEPdma
      r> #LastIn0 !
    else
      #LastIn0 @ /MaxPacket = if	\ avoid Windows problem
        CdcDmaInBuff 0 CDCBinEP setEPdma \ send zero length packet
      else
        CDCBinEP disEPdma		\ nothing to send, disable DMA
      endif
      #LastIn0 !			\ no more to send
    endif
  endif
;

: doEP5		\ event --
\ *G The action of endpoint 5, which is for CDC bulk IN and OUT.
  case
    USB_EVT_OUT_DMA_EOT of  CdcBulkDmaOut  endof
    USB_EVT_IN_DMA_EOT of  CdcBulkDmaIn  endof
    USB_EVT_IN_DMA_NDR of  CdcBulkDmaIn  endof
  endcase
;
' doEP5 5 SetEPhandler

: CdcFrameCheck	\ --
\ *G Executed in the frame interrupt. If there is data to be sent
\ ** to the host, Bulk NAK IN interrupts and channel DMA are
\ ** enabled.
  #cdcTxOut0 if				\ SFP002
    +BulkNAKin  CDCBinEP enEPdma
  endif
;
' CdcFrameCheck AtUSBframe

[then]


\ *************************
\ *S CDC Interrupt endpoint
\ *************************

: doEP1		\ event --
\ *G The action of endpoint 1, which is for CDC interrupt IN.
  case
    USB_EVT_IN  of
      SerialState w@ NotifyBuf 8 + w!	\ set up notify buffer
      CDCBinEP NotifyBuf USBWriteEP drop \ write it out
    endof
  endcase
;
' doEP1 1 SetEPhandler


\ ***************************
\ *S Reset and initialisation
\ ***************************

: resetCDCdata	\ --
\ *G Reset the CDC driver data, initialising CDC specific data to
\ ** start/restart operation. Performed at start up and as an EP0
\ ** action.
  InitNotifyBuf NotifyBuf 10 cmove
  InitLineCoding LineCodingBuf 7 cmove
  0 SerialState w!
  0 LineControlState w!
  #LastIn0 off
  initVCOM0				\ initialise serial device
;

: DevCDCreset	\ --
\ *G Reset the CDC driver, initialising the CDC specific data to
\ ** start/restart operation. Performed as an EP0 action.
  resetCDCdata
[ usbDMA? ] [if]
  initCDCbo initCDCbi
[then]
;
' DevCDCreset AtUSBreset

: CDCreset	\ -- flag ; 0=failed
\ *G Reset the CDC driver. EP0 action.
  DevCDCreset 1
;


\ ****************************
\ *S Diagnostics and Test code
\ ****************************

USBmin? [if]  external  [then]

: UsbCon	\ --
\ *G Switch the Forth console to the USB channel.
  usbCr0  UsbConsole0 dup ipvec ! opvec !
;

: UsbDisCon	\ --
\ *G Switch to the default console (usually a serial device)
\ ** and disconnect USB.
  Console dup ipvec ! opvec !  UsbDisconnect
;

(( moved to UsbEP0.fth
MSC? 0= [if]
: StartCDC	\ --
\ *G Start the USB and CDC systems. Used for CDC only devices.
  resetCDCdata startUSB  ;		\ initialise CDC, start USB
[else]
: StartCompo	\ --
\ *G Start the USB CDC and MSC systems. Used for composite devices.
  resetCDCdata initMSCdata startUSB  ;
[then]
))


\ *********************************
\ *S Testing with operating systems
\ *********************************

\ ==========
\ *N Windows
\ ==========
\ *P USB serial devices can be determined using: *\br{}
\ ** Control Panel -> System *\br{}
\ ** Hardware -> Device Manager -> Ports

\ *P Once a USB serial device has been found, it will usually
\ ** use the same port number when reconnected. To use two or
\ ** more of the same device, they must have different serial
\ ** numbers.

\ *P For testing we use HyperTerm and AIDE's PowerTerm. HyperTerm
\ ** is a very old program written before USB serial ports were
\ ** available. It does not have sufficient error recovery for
\ ** USB serial ports, and so is not suitable for production use.
\ ** Despite this, it is available on all Windows PCs and your
\ ** clients will use it.

\ *P Under Windows, a .INF file is required for any USB device
\ ** that includes a CDC class driver. The PID and VID (see
\ ** below) must match those in your USB device descriptors, as
\ ** must the manufacturer string. Sample .INF files are provided,
\ ** which you can use as the basis of your own.

\ *P The Windows USB subsytem is a little fragile with CDC class
\ ** operations. Note that HyperTerminal was written before USB
\ ** serial ports were available and its error handling is
\ ** inadequate for serious use with USB serial ports. TeraTerm
\ ** is widely recommended by USB developers.

\ *P We are told that the version of *\i{usbser.sys} issued with
\ ** Windows XP SP3 and Windows Vista is more robust
\ ** with composite devices.
\ ** Vista is fine, but XP SP2 is not. For XP, either upgrade to
\ ** SP3 or install hotfixes. It seems you need:
\ *(
\ *B Hotfix KB935892 (which brings usbccgp.sys 5.1.2600.3116)
\ *B Hotfix KB918365 (which brings usbser.sys 5.1.2600.2930)
\ *)

\ *P We installed SP3 for testing and had to copy the new version
\ ** of *\i{usbser.sys} manually from *\i{SP3.cab}. Use Windows
\ ** search and copy the new version to replace the one in
\ ** *\i{Windows\System32\Drivers}.


\ ========
\ *N Linux
\ ========
\ =====================
\ *H USB serial devices
\ =====================
\ *P When using USB serial devices, the name used varies according
\ ** to your distribution. The most common names appear to be:
\ *C   /dev/ttyUSBx
\ *C   /dev/ttyACMx

\ *P There are several methods of finding USB serial
\ ** ports. The simplest seems to be to unplug the device, then
\ ** reconnect it, then type the following incantation:
\ *C   dmesg | grep tty
\ *P where you must have root access. On many systems, e.g. Ubuntu
\ *C   sudo dmesg | grep tty
\ *P is required. The last few lines should then tell you which
\ ** USB serial port, e.g. *\f{/dev/ttyUSB0} was selected for
\ ** your device. If the last tells you that the device is now
\ ** disconnected, it is probably because of the "brltty bug".
\ ** Unless you need the Braille TTY access, remove the package
\ ** *\i{brltty}. Repeat:
\ *C   sudo dmesg | grep tty
\ *P to check that device remains connected. Some forums suggest
\ ** that you may also need to create the *\i{/dev/ttyUSBx} entries.
\ ** Do this with:
\ *C   sudo mknod /dev/ttyUSB0 c 188 0
\ *C   sudo mknod /dev/ttyUSB1 c 188 1
\ *C   sudo mknod /dev/ttyUSB2 c 188 2

\ ==================================
\ *H Linux serial terminal emulators
\ ==================================
\ *P The most widely used Linux equivalent to Windows' HyperTerm
\ ** appears to be *\i{minicom}. It isn't pretty, but it works
\ ** and is easy to use. There are plenty of others, including
\ ** GUI ones, but *\i{minicom} is the one we come back to as it
\ ** is available for nearly all distributions.


\ ===========
\ *N Mac OS/X
\ ===========
\ *P For automatic recognition by Mac OS X, the bit 0 of the
\ ** protocol byte in the relevant descriptors *\b{must} be set.
\ ** After enumeration, your USB serial sevices will appear in
\ ** the forms:
\ *C   /dev/tty.usbmodem000031FD1
\ *C   /dev/cu.usbmodem000031FD1

\ *P Tty is for "incoming calls" and cu for "outgoing calls".
\ ** It has to do with behaviour when opening the port, though
\ ** you can achieve any behaviour you want on either device by
\ ** opening it non-blocking and reconfiguring the tty settings.

\ *P To test a serial device, run a terminal
\ *C   Applications -> Utilities -> Terminal
\ *P Within the terminal run the *\b{screen} program.
\ *C   screen /dev/tty.usbmodem000031FD1
\ *P From a terminal you can get the documentation:
\ *C   man screen

\ ======
\ *> ###
\ ======

decimal
