\ UsbBase.fth - USB base operations, hardware independent

((
Copyright (c) 2008, 2017, 2018
Synics AG
and
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20180620 SFP003 Clipped responses to size in setup packet.
20171201 SFP002 Updated for STM32F0.
20081121 MVD002 Corrected RTEst.
20081031 SFP001 Updated for unknown descriptors.
))

only forth definitions
decimal

\ ========
\ *! usbbase
\ *T USB base layer
\ ========
\ *P The file *\i{UsbBase.fth} contains an (almost) hardware
\ ** independent USB code base. The hardware layer files,
\ ** *\i{UsbHwxxxx.fth} provide a standard interface that isolates
\ ** hardware from the base layer. Class drivers such as *\i{MscDrv.fth}
\ ** should be compiled after the core layer and can reference the
\ ** hardware layer. After the class files have been compiled,
\ ** compile *\fo{UsbEP0.fth}, which contain the USB control and
\ ** initialisation.

((
\ *P The five *\fo{VALUE}s below are used to provide status information
\ ** for the core layer, but are actually contained in *\i{UsbDefs.fth}
\ ** to avoid forward references and to permit easy reference by the
\ ** class drivers. Note that the endpoint status values are bit masks
\ ** in a hardware independent form. See *\fo{EP>mask} below.

0 value UsbDevAddr	\ -- u
\ *G USB device address
0 value UsbConfig	\ -- x
\ *G USB Configuration set. Non-zero when the USB device has been
\ ** configured.
0 value UsbEPmask	\ -- x
\ *G Endpoint mask.
0 value UsbEPhalt	\ -- x
\ *G Endpoint halt status.
0 value Usb#IfSet	\ -- n
\ *G Number of interfaces set.

: EP>mask	\ ep# -- mask
\ *G Decode an endpoint number to a bit mask such that if bit 7
\ ** is clear (OUT endpoints), the bit is in the low 16 bits, and
\ ** if set (IN endpoints) in the upper 16 bits.
  dup $0F and  swap $80 and		\ if bit 7 set (IN endpoint)
  if  #16 +  endif			\ shift by 16 more bits
  1 swap lshift				\ -- mask
;
))

#USBifs buffer: USBAltSets	\ -- addr
\ *G Alternate settings array

struct /UsbEPdata	\ -- len
\ *G USB Endpoint transfer data structure.
  int ep.Count			\ MUST be first to use 2@ and 2!
  int ep.*Data
end-struct

UAlign  /UsbEPdata buffer: EP0Data	\ -- addr
\ *G Endpoint 0 transfer control structure.
UAlign  /USBSetupPacket buffer: SetupPacket	\ -- addr
\ *G Current setup packet.
UAlign  /MaxPacket0 buffer: EP0Buf	\ -- addr
\ *G Packet buffer for endpoint 0 data.

compiler
: ReqType	\ -- b
\ *G Return the first byte in the global *\fo{SetupPacket}.
  SetupPacket usp.bmRequestType c@
;
: ReqDir	\ -- $80|$00
\ *G Return the direction bit of the *\fo{usp.bmRequestType}
\ ** field in the global *\fo{SetupPacket}.
  ReqType $80 and
;
: ReqTypeBits	\ -- %0xx00000
\ *G Return the type bits of the *\fo{usp.bmRequestType}
\ ** field in the global *\fo{SetupPacket}.
  ReqType $60 and
;
: ReqRecipient	\ -- %000xxxxx
\ *G Return the recipient bits of the *\fo{usp.bmRequestType}
\ ** field in the global *\fo{SetupPacket}.
  ReqType $1F and
;

: EP0Addr	\ -- addr|0
\ *G The address contained in *\fo{EP0Data ep.*Data} field.
  EP0Data ep.*Data @
;
: /EP0Data	\ -- len
\ *G The size of the data in *\fo{EP0Data} as given by the
\ ** *\fo{ep.Count} field.
  EP0Data ep.Count @
;
: NoEP0Data	\ --
\ *G Reset *\fo{EP0data}.
  EP0Buf 0 EP0Data 2!
;
target

: DevResetCore	\ --
\ *G Reset the core software configuration. This word is part of
\ ** USB device reset chain.
  UsbPowered? -> UsbDevSt
  0 -> UsbDevAddr  0 -> UsbConfig
  $0001:0001 -> UsbEPmask  0 -> UsbEPhalt
  NoEP0Data
[defined] DevDescReqLen [if] 0 -> DevDescReqLen [then]

;
' DevResetCore AtUSBreset

: UsbSetupStage	\ --
\ *G USB Request - Setup Stage
  $00 SetupPacket UsbReadEP drop	\ Read endpoint 0
;

: +EP0Data	\ len --
\ *G Update the *\fo{EP0Data} structure by *\i{len} bytes.
   EP0Data tuck 2@  rot /string  rot 2!
;

: UsbDataInStage	\ --
\ *G USB Request - Data In Stage.
\ ** Data is taken from the *\fo{EP0Data} global structure.
  $80 EP0Data 2@ /MaxPacket0 min USBWriteEP +EP0Data
;

2variable MoreIn	\ -- addr

: doDataIn	\ caddr len --
\ *G Set up the data and do the DataIn stage.
  $80 -rot  2dup EP0Data 2!  /MaxPacket0 min USBWriteEP +EP0Data
  EP0Data 2@ Morein 2!
;

[undefined] EP0inFrameCheck? [if]
0 equ EP0inFrameCheck?
[then]

EP0inFrameCheck? [if]
: ?MoreDataIn	\ --
\ *G Performed in the USB frame check and sends more data if required.
  MoreIn @ if
    MoreIn 2@ doDataIn
  then
;
' ?MoreDataIn AtUSBframe
[then]

: USBDataOutStage	\ --
\ *G USB Request - Data Out Stage.
\ ** Data is taken from the *\fo{EP0Data} global structure.
  $00 EP0Data ep.*Data @ USBReadEP +EP0Data
;

: sendZLPin	\ ep# --
\ *G Send a zero length packet from the given endpoint.
  0 0 USBWriteEP drop
;

: USBStatusInStage	\ --
\ *G USB Request - Status In Stage.
  $80 sendZLPin
;

: USBStatusOutStage	\ --
\ *G USB Request - Status Out Stage.
  $00 EP0Buf USBReadEP drop
;

: stall_i	\ --
\ *G Stall EP0 input.
  $80 USBSetStallEP  EP0Data ep.Count off
;

: stall_o	\ --
\ *G Stall EP0 output.
  $00 USBSetStallEP  EP0Data ep.Count off
;

: RTDst		\ -- flag ; true=success
\ *G Get REQUEST_TO_DEVICE status.
  addr UsbDevSt EP0Data ep.*Data !
  USBDataInStage 1
;

: RTIst		\ -- flag ; true=success
\ *G Get REQUEST_TO_INTERFACE status.
  UsbConfig if
    SetupPacket usp.wIndex c@ Usb#IfSet < if
      0 EP0Buf !  EP0Buf EP0Data ep.*Data !
      USBDataInStage 1  exit
    endif
  endif
  0
;

: RTEst		\ -- flag ; true=success
\ *G Get REQUEST_TO_ENDPOINT status.
  SetupPacket usp.wIndex c@ dup EP>mask	\ -- ep# mask
  dup UsbEPmask and if			\ if EP mask enabled
    over 0= UsbConfig or if		\ if EP0 or config set
      nip				\ -- mask
      UsbEPhalt and 0<> 1 and EP0Buf lew!
      EP0Buf EP0Data ep.*Data !
      USBDataInStage 1  exit		\ MVD002
    endif
  endif
  2drop 0
;

: USBGetStatus	\ -- flag ; true=success
\ *G Get Status USB Request. The data is in the global
\ ** *\fo{SetupPacket}.
  case  ReqRecipient			\ recipient field
    REQUEST_TO_DEVICE of  RTDst  endof
    REQUEST_TO_INTERFACE of  RTIst  endof
    REQUEST_TO_ENDPOINT of  RTEst  endof
       0 swap				\ return false
  endcase
;

: RTDsetf	\ sc -- flag ; sc, true=set; flag, true=success
\ *G Set/Clear Device feature.
  SetupPacket usp.wValue lew@ USB_FEATURE_REMOTE_WAKEUP = if
    if
      USB_GETSTATUS_REMOTE_WAKEUP addr UsbDevSt or!
    else
      USB_GETSTATUS_REMOTE_WAKEUP addr UsbDevSt bic!
    endif
    1
  else
    drop 0
  endif
;

: RTEsetf	\ sc -- flag ; sc, true=set; flag, true=success
\ *G Set/Clear Endpoint feature.
  SetupPacket usp.wIndex c@		\ get endpoint number
  dup EP>mask UsbEPmask and if		\ -- sc ep# ; if EP mask enabled
    dup $0F and 0= UsbConfig or if	\ if EP0 or config set
      SetupPacket usp.wValue lew@ USB_FEATURE_ENDPOINT_STALL = if
        swap if				\ -- ep#
          USBSetStallEP			\ Sets bit in UsbEpHalt
        else
          USBClrStallEP			\ Clears bit in UsbEpHalt
        endif
        NoEP0Data  1 exit
      endif
    endif
  endif
  2drop 0
;

: USBSetClrFeature	\ sc -- flag ; sc, true=set; flag, true=success
\ *G Set/Clear Feature USB Request. The data is in the global
\ ** *\fo{SetupPacket}.
  case  ReqRecipient			\ recipient field
    REQUEST_TO_DEVICE of  RTDsetf  endof
    REQUEST_TO_INTERFACE of  drop 0  endof
    REQUEST_TO_ENDPOINT of  RTEsetf  endof
      nip 0 swap
  endcase
;

: maxEP0resp	\ len -- len' ; SFP003
\ *G Clip the given length to the wLength value in the setup packet.
  SetupPacket usp.wLength lew@ min
;

: GetCfgDesc	\ -- flag ; true=success
\ *G Get Configuration descriptor. The data is in the global
\ ** *\fo{SetupPacket}.
  CfgDesc  SetUpPacket usp.wValue c@
  1- 0 max 0 ?do			\ -- desc ; step to required configuration
    dup c@ 0= ?leave  dup 2+ lew@ +
  loop
  dup c@ if
   dup 2+ lew@  ( SetupPacket usp.wLength lew@ min ) maxEP0resp doDataIn 1
  else
    drop 0
  endif
;

[defined] USBStr@ [if]
: GetStrDesc	 \ -- flag
\ *G Get String descriptor. The data is in the global *\fo{SetupPacket}.
\ ** Used when *\fo{USBStr@} has been defined and the string index
\ ** is a small integer. *\fo{USBStr@ ( index -- caddr len )} is usually
\ ** defined in the configuration file. See *\i{CdcMscConfig.fth} for an
\ ** example.
   SetupPacket usp.wValue c@ USBStr@
   maxEP0resp ( SetupPacket usp.wLength lew@ min ) doDataIn 1
;
[else]
: GetStrDesc	 \ -- flag
\ *G Get String descriptor. The data is in the global *\fo{SetupPacket}.
\ ** Used when *\fo{USBStr@} has not been defined and the string index
\ ** is the byte offset into the string table.
   StrDesc SetupPacket usp.wValue c@ + dup c@
   maxEP0resp ( SetupPacket usp.wLength lew@ min ) doDataIn 1
;
[then]

: GetRTDDesc	\ -- flag ; true=success
\ *G Get Device Descriptor USB Request. The data is in the global
\ ** *\fo{SetupPacket}.
  case  SetupPacket usp.wValue 1+ c@
    USBDevDesc# of
      DevDesc dup c@ maxEP0resp doDataIn 1	\ SFP003
    endof
    USBCfgDesc# of  GetCfgDesc  endof
    USBStrDesc# of  GetStrDesc  endof
      0 swap
  endcase
;

: USBGetDescriptor	\ -- flag ; true=success
\ *G Get Descriptor USB Request. The data is in the global
\ ** *\fo{SetupPacket}.
[ dbgCore? ] [if] ." GetDesc " [then]
  case  ReqRecipient			\ recipient field
    REQUEST_TO_DEVICE of  getRTDdesc  endof
[ HID? ] [if]
    REQUEST_TO_INTERFACE of  getRTIdesc  endof
[then]
       0 swap				\ return false
  endcase
;

: DisableAllEPs	\ --
\ *G Disable all endpoints except 0.
  $0002:0000 $0000:0002  #16 1 do	\ -- desc maska maskb
    dup UsbEPmask and
    if  i USBDisableEP  endif
    over UsbEPmask and
    if  i $80 or USBDisableEP  endif
    1 lshift swap 1 lshift swap		\ update masks
  loop
  2drop
  $0001:0001 -> UsbEPmask  0 -> UsbEPhalt	\ EP0 is always in use
;

: SetCfgConfig	\ desc -- desc'
\ *G Process the base configuration descriptor, returning the
\ ** next descriptor.
  dup 5 + c@ SetupPacket usp.wValue c@ = if	\ right config?
    SetupPacket usp.wValue c@ -> UsbConfig	\ set config
    dup 4 + c@ -> Usb#IfSet			\ Config number in use
    USBAltSets #UsbIfs erase			\ wipe alternates
    DisableAllEPs				\ except EP0
    1 USBConfigure
    dup 7 + c@ 0x40 and if
      USB_GETSTATUS_SELF_POWERED addr UsbDevSt or!
    else
      USB_GETSTATUS_SELF_POWERED addr UsbDevSt bic!
    endif
    dup c@ +				\ step over descriptor
  else
    dup 2+ lew@ +			\ skip complete configuration
  endif
;

: initEP	\ desc ep# mask -- desc
\ *G Initialise an endpoint for use.
  dup addr UsbEPmask or!
  addr UsbEPhalt bic!			\ -- desc ep#
  USBConfig2EP
\  over USBConfigEP
\  dup USBEnableEP  USBResetEP
;

((
: termEP	\ ep# mask --
\ *G Terminate using an endpoint.
\ ** Only needed if *\fo{USBSetIface} is provided.
   dup addr UsbEPmask bic!
   addr UsbEPhalt bic!
   USBDisableEP
;
))

: SetEPConfig	\ desc -- desc
\ *G Process an endpoint descriptor, returning the
\ ** same descriptor.
  dup 2+ c@ dup EP>mask initEP
;

: SetConfig	\ --
\ *G Set the configuration from the descriptor. The *\fo{CfgDesc}
\ ** Configuration descriptor (see *\i{Descriptors.fth} is walked
\ ** and processed.
  0 >r					\ assume no alternate settings
  CfgDesc
  begin
    dup c@				\ look at length
   while
    case  dup 1+ c@			\ look at type, each type handler steps
      USBCfgDesc# of  SetCfgConfig  endof
      USBIfaceDesc# of
        r> drop  dup 3 + c@ >r		\ change alternate settings
        dup c@ +			\ step over descriptor
      endof
      USBEPdesc# of
        r@ 0= if  SetEPConfig  endif
        dup c@ +			\ step over descriptor
      endof
        swap dup c@ + swap		\ step over descriptor
    endcase
  repeat
  drop  r> drop
;

: ClrConfig	\ --
\ *G Clear the existing configuration.
  0 -> USBConfig
  DisableAllEPs				\ except EP0
  0 USBConfigure
;

: USBSetConfig	\ -- flag ; true=success
\ *G Set Configuration USB Request.
  SetupPacket usp.wValue c@
  if  SetConfig  else  ClrConfig  endif
  SetupPacket usp.wValue c@ UsbConfig = 1 and
;

((
[-short-branches
: USBSetIface	\ -- flag ; true=success
\ *G Set Interface USB Request. The data is in the global
\ ** *\fo{SetupPacket}.
\ ** This word is now unused and is commented out.
  { | ifn altn old msk -- flag }
  UsbConfig 0= if  0 exit  endif
  0 dup -> altn  dup  -> msk  -> old
  0 CfgDesc begin			\ -- set desc
    dup c@				\ end of descriptor(s)
   while
    case  dup 1+ c@			\ inspect type
      USBCfgDesc# of	\ Configuration descriptor
        dup 5 + c@ UsbConfig = if	\ this is the one for us?
          dup c@ +			\ so step over base header
        else
          dup 2+ lew@ +			\ no, skip whole configuration
        endif
      endof		\ Interface descriptor
      USBIfaceDesc# of
        dup 3 + c@ -> altn  0 -> msk	\ alternate and mask
        dup 2 + c@ dup -> ifn		\ interface number
        SetupPacket usp.wIndex c@ = if	\ is ours?
          SetupPacket usp.wValue c@ altn = if
             nip 1 swap			\ -- set desc
             USBAltSets ifn +		\ -- set desc arr
             dup c@ -> old  altn swap c!	\ preserve and set
          endif
        endif
        dup c@ +			\ to next descriptor
      endof
      USBEPDesc# of	\ Endpoint descriptor
        ifn SetupPacket usp.wIndex c@ = if
          dup 2+ c@ dup EP>mask		\ -- set desc ep# mask
          altn SetupPacket usp.wValue c@ = if
            initEP
          else
            altn old = if
              dup msk and 0= if		\ -- set desc ep# mask
                2dup termEP
              endif
            endif
            2drop
          endif
        endif
        dup c@ +			\ to next descriptor
      endof
        swap  dup c@ +  swap		\ unknown descriptor
    endcase
  repeat				\ -- set desc
  drop
;
short-branches]
))


\ ======
\ *> ###
\ ======

decimal
