\ UsbHwSTM32F0x2.fth - USB driver for STM32F0x2 parts.

((
Copyright (c) 2017, 2018
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone numbers are:
       011 44 23 8063 1441
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20180991 MJB004 Added txstat@ and USBWrBusy? 
20180302 SFP003 CDC working cleanly.
20180223 SFP002 CDC working.
20171130 SFP001 Started
))

only forth definitions
decimal

\ ===========
\ *! usbhwstm32f0x2
\ *T USB driver for STM32F0x2 parts
\ ===========
\ *P The file *\i{UsbHwSTM32F0x2.fth} provides a USB driver
\ ** for STM32F0x2 devices. Some primitive code was taken
\ ** from a public-domain source, and updated for correctness
\ ** and interrupt operation.
\ *C https://github.com/jeelabs/embello/tree/master/explore/1608-forth/suf

/MaxPacket 64 <> [if]
  cr ." /MaxPacket must be 64" abort
[then]


\ *************************************
\ *S Endpoint packet memory and control
\ *************************************

: even		\ addr -- addr'
\ *G Force input address to be even.
  1+ -2 and
;

: usb-pma	( pos -- addr ) _USBSRAM + ;
\ *G Convert an offset/position in the packet RAM to an absolute
\ ** address.
compiler
: usb-pma	( pos -- addr ) _USBSRAM + ;
\ *G Convert an offset/position in the packet RAM to an absolute
\ ** address. Compiler macro.
target

: ep-reg	\ ep n -- addr
\ *G Return the address in packet RAM of reg n (0..3) of ep (0..7)
  2* swap $0F and 8 * + usb-pma ;

\ *P Each endpoint has an 8 byte control area of four buffers for
\ ** transmit and receive control.
\ *[
0 equ EPR_ADDR_TX
1 equ EPR_COUNT_TX
2 equ EPR_ADDR_RX
3 equ EPR_COUNT_RX
\ *]

: ep-addr	( ep# -- addr )  $0F and  cells _USBdev usbEP0R + + ;
\ *G Return the address in the peripheral block of the peripheral
\ ** register for endpoint ep.
compiler
: ep-addr	( ep# -- addr )  $0F and  cells _USBdev usbEP0R + + ;
target

\ *[
\ endpoint status values, both TX and RX
0 equ epDISABLED
1 equ epSTALL
2 equ epNAK
3 equ epVALID

\ endpoint type, both TX and RX
0 equ epBULK
1 equ epCONTROL
2 equ epISO
3 equ epINTERRUPT
\ *]

: rxstat!	\ ep u --
\ *G Set stat_rx without toggling/setting any other fields.
  swap ep-addr >r
  3 and 12 lshift  r@ w@ tuck  xor	\ -- rval rval^mask ; R: -- *ep
\  R^rrseekT^ttnnnn
\  5432109876543210
  %0011:0000:0000:0000 and swap		\ -- rval^mask' rval ; R: -- *ep
  %0000:1111:0000:1111 and		\
  %1000:0000:1000:0000 or
  or r> w!
;

: txstat!	\ ep u --
\ *G Set stat_tx without toggling/setting any other fields.
  swap ep-addr >r
  3 and 4 lshift  r@ w@ tuck  xor	\ -- orig mask ; R: -- addr
\  R^rrseekT^ttnnnn
\  5432109876543210
  %0000:0000:0011:0000 and swap		\ -- mask' orig
  %0000:1111:0000:1111 and		\ zero bits that toggle on 1
  %1000:0000:1000:0000 or		\ set bits that clear on zero
  or r> w!
;

: txstat@   \ ep -- n ;
\ *G return the transmit status in two lower bits
  ep-addr w@  4 rshift 3 and 
;

: USBWrBusy?    \ ep# -- f ;
\ *G returns true if the endpoint is in the process 
\ ** of transmitting 
  txstat@ 3 =
;

: rxclear	\ ep --
\ *G Clear the endpoint receive CTR_RX bit.
  ep-addr dup w@  $0F0F and  $0080 or  swap w!  ;

: txclear	\ ep --
\ *G Clear the endpoint transmit CTR_TX bit.
  ep-addr dup w@  $0F0F and  $8000 or  swap w!  ;

: ep-reset-rx#	\ ep --
\ *G Clear the RX count register.
  $8400 over EPR_COUNT_RX ep-reg w!
  epVALID rxstat!
;

: ep-reset-tx#	\ ep --
\ *G Clear the TX count register.
  $0000 over EPR_COUNT_TX ep-reg w!
  epNAK txstat!
;

\ *P Packet memory is used as shown below. Memory use is
\ ** inefficient as two 64 byte buffers are reserved for each
\ ** endpoint, whether needed or not.
\ *E 000..03F   Endpoint Table
\ ** 040..0BF   EP0 buffers, offset $00 for RX, $40 for TX
\ ** 0C0..13F   EP1 ...
\ ** 140..1BF   EP2 ...
\ ** 1C0..23F   EP3 ...
\ ** 240..2BF   EP4 ...
\ ** 2C0..33F   EP5 ...
\ ** 340..3BF   EP6 ...

: EPioBufRX	\ ep# -- pma
\ *G Return offset address of the receive packet memory buffer.
  $0F and $80 * $40 +
;

: EPioBufTX	\ ep# -- pma
\ *G Return offset address of the transmit packet memory buffer
  EPioBufRX $40 +
;

: set-EPtab-n	\ ep# --
\ *G Set up EP buffer area in the packet memory.
  dup EPioBufTX  over EPR_ADDR_TX ep-reg w!
  0 over EPR_COUNT_TX ep-reg w!
  dup EPioBufRX  over EPR_ADDR_RX ep-reg w!
  dup $0F and 0= if			\ EP0 control endpoint
    $8400 swap EPR_COUNT_RX ep-reg w!
  else
    $8400 swap EPR_COUNT_RX ep-reg w!
  then
;


\ *********************
\ *S Handling endpoints
\ *********************
\ *P This section contains an array of endpoint handlers and default
\ ** actions for other operations which are hardware dependent.

compiler
: USBDirCtrlEP	\ dir -- ; 0=out
\ *G This word is a dummy for this hardware.
  drop
;
: USBResetEP	\ dir -- ; 0=out
\ *G This word is a dummy for this hardware.
  drop
;
: USBEnableEP	\ dir -- ; 0=out
\ *G This word is a dummy for this hardware.
  drop
;
target

create EPxts	\ -- addr
\ *G Holds the xts of the 16 endpoint handler words. The action
\ ** of each handler must have the stack effect below:
\ *C   event --
  ' drop ,  ' drop ,  ' drop ,  ' drop ,
  ' drop ,  ' drop ,  ' drop ,  ' drop ,
  ' drop ,  ' drop ,  ' drop ,  ' drop ,
  ' drop ,  ' drop ,  ' drop ,  ' drop ,

: EPhandler	\ event u --
\ *G Run the endpoint handler associated with endpoint *\i{u},
\ ** where endpoints are numbered 0..15. Each endpoint handler
\ ** is passed an event type which allows it to determine what
\ ** it has to do. The action of each handler must have the stack
\ ** effect below.
\ *C   event --
  $0F and cells EPxts + @ execute
;

interpreter
: SetEPhandler	\ xt ep# --
\ *G Install the endpoint handler for endpoint *\i{ep#}.
\ ** Used during interpretation.
  $0F and cells EPxts + !
;
target

0 value UsbDevSt	\ -- x
\ *G USB device Status.

: USBsuspend	\ --
\ *G Suspend operations are performed by the hardware. This
\ ** is a hook for installing actions such as toggling an LED.
;

: USBresume	\ --
\ *G Resume operations are performed by the hardware. This
\ ** is a hook for installing actions such as toggling an LED.
;

: usbHWconnect	\ --
\ *G Connect pull up on USB DP line.
  bit15 _USBdev usbBCDR + or!
;

: usbHWdisconnect	\ --
\ *G Disconnect pull up on USB DP line.
  bit15 _USBdev usbBCDR + bic!
;

: USBconnect	\ --
\ *G Connect to the USB bus.
  USBhwconnect
;

: USBdisconnect	\ --
\ *G Disconnect from the USB bus.
  USBhwdisconnect  0 -> UsbConfig
;

: USBWakeUp	\ --
\ *G Called automatically on USB Remote Wakeup.
  %1000 _USBdev usbCNTR + wbic!
;

: USBSetAddress	\ addr --
\ *G Set the USB assigned address and enable the device. This
\ ** should be done during a SOF (frame) interrupt. Note that
\ ** *\fo{USBSetAddress} may not actually set the device address,
\ ** but may just signal that an action must be taken, e.g. in
\ ** the next frame interrupt
  $7F and  $80 or  -> DevAddr
;

: USBConfigure	\ cfg -- ; 0=unconfigure
\ *G Set the SIE configuration state.
  drop
\  0<> -> UsbConfig
;

\ USB Endpoint Callback Events
#define USB_EVT_SETUP       1   /* Setup Packet */
#define USB_EVT_OUT         2   /* OUT Packet */
#define USB_EVT_IN          3   /*  IN Packet */
#define USB_EVT_OUT_NAK     4   /* OUT Packet - Not Acknowledged */
#define USB_EVT_IN_NAK      5   /*  IN Packet - Not Acknowledged */
#define USB_EVT_OUT_STALL   6   /* OUT Packet - Stalled */
#define USB_EVT_IN_STALL    7   /*  IN Packet - Stalled */
#define USB_EVT_OUT_DMA_EOT 8   /* DMA OUT EP - End of Transfer */
#define USB_EVT_IN_DMA_EOT  9   /* DMA  IN EP - End of Transfer */
#define USB_EVT_OUT_DMA_NDR 10  /* DMA OUT EP - New Descriptor Request */
#define USB_EVT_IN_DMA_NDR  11  /* DMA  IN EP - New Descriptor Request */
#define USB_EVT_OUT_DMA_ERR 12  /* DMA OUT EP - Error */
#define USB_EVT_IN_DMA_ERR  13  /* DMA  IN EP - Error */

((
  $3210 0 set-EPnR		\ RX valid, CONTROL, TX stall, EP0
  $0021 1 set-EPnR		\ RX dis, BULK, TX NAK, EP1
  $0622 2 set-EPnR		\ RX dis, INTERRUPT, TX NAK, EP2
  $3003 3 set-EPnR		\ RX valid, BULK, TX dis, EP3
))

: cfg-EPnR	\ ep# x --
\ *G Configure the USB endpoint register,
  over $0F and or  swap set-EPnR  ;

: cfgCtrl	\ desc ep# -- desc
\ Configure a control endpoint.
  $3210 cfg-EPnR		 	\ -- desc ; RX valid, CONTROL, TX stall, EP0
;

: cfgIso	\ desc ep# -- desc
\ *G Configure an isochronous endpoint. UNTESTED.
  dup $80 and if			\ Iso IN
    $0420 cfg-EPnR			\ RX dis, ISO, TX NAK, IN
  else
    $3400 cfg-EPnR			\ RX valid, ISO, TX dis, OUT
  then
;

: cfgBulk	\ desc ep# -- desc
\ *G Configure a bulk endpoint.
  dup $80 and if			\ Bulk IN
    $0020 cfg-EPnR			\ RX dis, BULK, TX NAK, IN
  else
    $3000 cfg-EPnR			\ RX valid, BULK, TX dis, OUT
  then
;

: cfgInt	\ desc ep# -- desc
\ *G Configure an interrupt endpoint.
  $0620 cfg-EPnR			\ RX dis, INTERRUPT, TX NAK,
;

: USBConfigEP	\ desc --
\ *G Configure USB Endpoint register according to Descriptor.
  dup epd.Addr c@ $8F and		\ -- desc EP#
  over epd.Attribs c@ $03 and		\ -- desc EP# type
  case
    0 of  cfgCtrl  endof
    1 of  cfgIso  endof
    2 of  cfgBulk  endof
    3 of  cfgInt  endof
  endcase
  drop
;

: USBConfig2EP	\ desc ep# -- desc
\ *G Configure USB Endpoint according to Descriptor and EndPoint.
  dup set-EPtab-n
  over USBConfigEP
  dup USBEnableEP  USBResetEP
;

: USBDisableEP	\ ep# -- ; bit7=dir, bits3:0=num
\ *G Disable the endpoint.
  epDISABLED 2dup rxstat! txstat!
;

: USBSetStallEP	\ ep# -- ; bit7=dir, bits3:0=num
\ *G Stall the endpoint. This word *\b{must} clear the relevant
\ ** bit in *\fo{UsbEpHalt}.
  dup epSTALL over $080 and 		\ IN endpoint use TX, OUT use RX
  if  txstat!  else  rxstat!  then
  EP>mask addr UsbEpHalt or!
;

: epBulk?	\ ep# -- flag
\ *G Return true if the endpoint is BULK.
  ep-addr @ $0600 and  epBULK 9 lshift =
;

: USBClrStallEP	\ ep# -- ; bit7=dir, bits3:0=num
\ *G Clear the endpoint stall. This word *\b{must} clear the relevant
\ ** bit in *\fo{UsbEpHalt}, and should ensure that bulk IN/OUT NAKs cause
\ ** interrupts.
  dup EP>mask addr UsbEpHalt bic!	\ clear halt status bit
  dup $080 and if			\ IN endpoint?
    epNAK txstat!
  else
    epVALID rxstat!
  endif
;


\ =========
\ Main code
\ =========

: USBReadEP	\ ep# addr -- #read ; bit7=dir, bits3:0=num
 \ *G Read endpoint data from endpoint *\i{ep#} to *\i{addr}.
 \ ** Return the number of bytes read.
 \ ** Note that data is transferred in units of two bytes.
  swap $0F and dup >r			\ -- addr ep# ; R: -- ep#
  dup EPR_ADDR_RX ep-reg w@		\ -- addr ep# rxpma ; R: -- ep#
  usb-pma -rot				\ -- src addr ep# ; R: -- ep#
  EPR_COUNT_RX ep-reg w@ $03FF and 	\ -- src addr len ; R: -- ep#
  dup >r even  bounds ?do		\ -- src ; R: -- ep# len
    dup w@ i w!  2+
  2 +loop
  drop  r> r>				\ -- len ep#
  dup ep-reset-rx# rxclear		\ -- len
;

: USBWriteEP	\ ep# addr len -- cnt
 \ *G Write data *\i{addr/len} to host from endpoint *\i{ep#}.
 \ ** Return the number of bytes written.
  dup >r				\ ep# addr len ; R: -- len
  rot $0F and dup >r			\ addr len ep# ; R: -- len ep#
  epr_addr_tx ep-reg w@ usb-pma -rot	\ txpma addr len ; R: -- len ep#
  even bounds ?do			\ -- txpma
    i w@  over w! 2+
  2 +loop				\ -- txpma'
  drop					\ -- ; R: -- len ep#
  r> r> swap				\ -- len ep#
  2dup EPR_COUNT_TX ep-reg w!		\ set TX length in EPn data
  epVALID txstat!			\ -- len ; start write
;

: .ep0/istr	\ --
\ *G Display the EP0 endpoint register and the ISTR
  ." EP0R " 0 ep-addr @ .word ."  ISTR " _usbdev usbISTR + w@ .word space
\  ." BCNTR " _usbdev usbCNTR + w@ .word space
;

: .ep0-pma	\ --
\ *G Display the four EP0 packet memory registers
  cr ." EP0-PMA "
  _usbDev usbBTABLE + w@ usb-pma	\ table address
  8 bounds
  do  i w@ .word space  2 +loop
;

: set-EPnR	\ x ep --
\ *G Set the USB endpoint register
  over $070F and  over ep-addr w!	\ set bits that are directly written
  over #12 rshift over swap rxstat!	\ set RX status bits
  swap #4 rshift txstat!		\ set TX status bits
;

create usb:init
  $0080 w,  \ ADDR0_TX   control - rx: 64b @ +040, tx: 64b @ +080
  $0000 w,  \ COUNT0_TX
  $0040 w,  \ ADDR0_RX
  $8400 w,  \ COUNT0_RX

: USBswReset	\ --
\ *G The action performed for a software reset of the USB, either
\ ** at reset or in response to a USB reset.
  512 0 do  0 i usb-pma w!  2 +loop	\ wipe region
  0 _USBdev usbBTABLE + w!		\ EP table starts at packet buffer 0
  usb:init  8 0 do			\ initialise EP0 buffers
    dup w@  i _USBSRAM + w!  2+
  2 +loop  drop
  $3210 0 set-EPnR		\ RX valid, CONTROL, TX stall, EP0
\  $0021 1 set-EPnR		\ RX dis, BULK, TX NAK, EP1
\  $0622 2 set-EPnR		\ RX dis, INTERRUPT, TX NAK, EP2
\  $3003 3 set-EPnR		\ RX valid, BULK, TX dis, EP3
  $80 _USBdev usbDADDR + w! 	\ bit 7 = USB enable, address=0
;

: ep-setup	\ ep# --
\ *G Trigger a SETUP packet handler.
  USB_EVT_SETUP swap EPhandler
;

: ep-out	\ ep# --
\ *G Trigger an OUT packet handler.
  USB_EVT_OUT swap EPhandler
;

: ep-in		\ ep# --
\ *G Trigger an IN packet handler.
  $80 or  dup $7F and 0= if			\ EP0
    $80 or  dup EPR_COUNT_TX ep-reg w@
    $03FF and if				\ last packet len<>0?
      USB_EVT_IN swap EPhandler
    else
      dup txclear  ep-reset-tx#
    then
  else
    USB_EVT_IN swap EPhandler
  then
;

: usb-ctr	\ istr --
\ *G Given an interrupt status reading, process the packet.
  dup $0F and  swap $10 and if		\ -- ep# ; OUT or SETUP packet
    dup ep-addr w@ $800 and if
      ep-setup
    else
      ep-out
    then
  else
    ep-in
  then
\ cr ." F " .ep0/istr
\ cr \ .ep0-pma
;

: usb-poll	\ --
\ *G Main USB driver polling routine. Factored this way to allow
\ ** *\fo{EXIT}s from the branches.
  _USBdev usbISTR + w@
  $9E1F and
((
dup $8000 and
if  cr dup .word  then
))
  dup $8000 and if			\ good endpoint transfer
    usb-ctr  exit
  then
  dup $1000 and if			\ bit 12; USB wake up
    %1000 _USBdev usbCNTR + wbic!
    USBwakeup USBresume
    $EFFF _USBdev usbISTR + w!		\ clear interrupt
    drop  exit
  then
  dup $0800 and if			\ bit 11; USB suspend
    %1000 _USBdev usbCNTR + wor!	\ suspend and low power modes
    USBsuspend
    $F7FF _USBdev usbISTR + w!		\ clear interrupt
    drop  exit
  then
  dup $0400 and if			\ bit 10; USB reset
    USBswReset  UsbResetChain ExecChain	\ reset core layer and class drivers
    $FBFF _USBdev usbISTR + w!		\ clear interrupt
    drop exit
  then
  dup $0200 and if			\ bit 9; SOF
    DevAddr if
      DevAddr $80 or _USBdev usbDADDR + w! \ really set device address
      DevAddr $7F and -> UsbDevAddr
      0 -> DevAddr
    then
    UsbFrameChain ExecChain
    $FDFF _USBdev usbISTR + w!		\ clear interrupt
    drop exit
  then
  drop
;

: USBinterrupt	\ --
\ *G The USB interrupt handler. It calls *\fo{usb-poll} above to
\ ** permit better factoring. LED1 is enabled around the handler
\ ** to give the user or oscilloscope an idea of overheads.
  console1 dup ipvec ! opvec !  decimal
  usbLED1on
  usb-poll
  usbLED1off
;
' USBinterrupt USB_EXTI_18vec# EXC: USB_ISR	\ -- addr
\ *G The entry point for the USB interrupt.

: usbIntInit	\ --
\ *G Initialise USB hardware interrupt handler.
  LedActive off				\ no flashing LEDs
  $9E00 _USBdev usbCNTR + w! 		\ enable USB ints
  USB_EXTI_18vec# EnInt			\ enable USB handler
;

: InitUSB	\ --
\ *G Initialise the USB hardware.
  USBhwInit USBswReset USBintInit	\ all primary initialisation
  UsbResetChain ExecChain		\ reset core layer and class drivers
;

: startUSB	\ --
\ *G Start the USB system. This word gives more control to devices
\ ** that have VBUS control. If such devices are rebooted without
\ ** power down, they may/will need to perform a *\fo{USBdisconnect}
\ ** operation as part of the start up sequence. A disconnect,
\ ** delay, connect sequence is part of *\fo{startUSB}.
  initUSB  USBdisconnect  #1000 ms  USBconnect  ;


\ ======
\ *> ###
\ ======

decimal

