\ otgISP1301.fth - USB OTG handling for LPC23xx devices

((
Copyright (c) 2008, 2010, 2011
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel:   +44 (0)23 8063 1441
email: mpe@mpeforth.com
       tech-support@mpeforth.com
web:   http://www.mpeforth.com
Skype: mpe_sfp

From North America, our telephone and fax numbers are:
       011 44 23 8063 1441
       011 44 23 8033 9691
       901 313 4312 (North American access number to UK office)


To do
=====

Change history
==============
20110307 SFP002 Removed PAUSE from OTGI2CXCHG to fix FVD005 in
                USBCONNECT.
20100903 FVD001 Allow for state of ADR/PSW pin.
))

only forth definitions
decimal

\ =============
\ *! otgisp1301
\ *T USB OTG with LPC23xx and ISP1301
\ =============
\ *P The file *\i{USBotgLPC23xx} provides USB OTG facilities for
\ ** LPC23xx CPUs that are connected to USB through an ISP1301
\ ** device. The default configuration is for use with a custom
\ ** LPC2388 board, so modify this file for your hardware.
\ ** It will require minor changes for other LPC devices and
\ ** boards, in particular in the USB Connect function and LEDs.


\ **********************
\ *S USB interrupt level
\ **********************

4 equ USBpri#	\ -- u
\ *G Priority (0..15, 0=high) used for the USB interrupt handler.


\ *******
\ *S LEDs
\ *******
\ *P The USB software uses two LEDs to indicate USB activity
\ ** (LED1) and disk activity (LED2).

bit18 equ LED1mask			\ P1.18, PINSEL3
0 equ LED2mask				\ not present
LED1mask equ LEDiomask			\ P1.18 only
0 [ps  18 +ps  ps]
  equ LEDpinval  equ LEDpinmask		\ bits in PINSEL3.

: initLEDs	\ --
\ *G Initialise the two LEDs to off, output=high.
  bit0 _SYSCON SCS + or!		\ enable fast I/O for GPIO0
  LEDpinval LEDpinmask _PINSEL3 setMask	\ configure I/O
  LEDIOmask _FIO1
  2dup FIODIR + or!			\ set DIR bits
  FIOSET + !				\ set high
;
' initLEDs AtCold

: usbLED1on	\ --
\ *G Turn LED1 on.
  LED1mask _FIO1 FIOCLR + !  ;
: usbLED1off	\ --
\ *G Turn LED1 off.
  LED1mask _FIO1 FIOSET + !  ;
compiler
: usbLED1on	\ --
  LED1mask _FIO1 FIOCLR + !  ;
: usbLED1off	\ --
  LED1mask _FIO1 FIOSET + !  ;
: usbLED2on	\ --
  ;
: usbLED2off	\ --
  ;
target


\ **********
\ *S ISP1301
\ **********

1 [ps
                    ( 18 +ps ) 19 +ps	\ 18 for auto LED drive
    20 +ps   21 +ps   22 +ps   23 +ps
    24 +ps                   ( 27 +ps )	\ 27/OINT polled - ISP1301 bug
    28 +ps   29 +ps
\                      2 [psm] 19 +ps	\ PPWR
ps] equ otgPS3val  equ otgPS3mask
\ *G Value and mask to set *\fo{PINSEL3} for USB operation with the
\ ** ISP1301.
\ \	$05415550 $f03c000f $e002c00c andor!	\ configure PINSEL3
\ 	$05415540 $f03c003f $e002c00c andor!	\ configure PINSEL3 (no led)

1 equ otgADR/PSW			\ reset state of ADR/PSW pin
$58 otgADR/PSW 2* + equ otgI2caddr	\ -- devaddr
\ *G I2C device address (including R/W bit) for the ISP1301.
\ ** Depends on level of ADR/PSW pin at reset.

100 kHz equ otgI2c-speed	\ -- hz
\ *G OTG I2C speed in Hz.
usb-speed otgI2c-speed / 2/ equ #otgI2cClks	\ -- u
\ *G Number of USB clocks in an I2C half-clock period.

otgI2caddr $100 or equ otgDevSelW	\ -- devmask
\ *G The device address with START=1 and R/W=0 for a write.
otgI2caddr $101 or equ otgDevSelR	\ -- devmask
\ *G The device address with START=1 and R/W=1 for a read.

: otgI2cSt@	\ -- res
\ *G Read the OTG status register into the top 16 bits of the
\ ** returned result, and clear the bottom three bits, TDI, AFI
\ ** or NAI.
  _USB usbI2cSt +  dup @ tuck		\ -- st addr st
  $07 and swap !
  dup $2 and abort" OTG I2C AFI"
  dup $4 and abort" OTG I2C NAK"
  16 lshift
;

: otgI2cXchg	\ input -- output
\ *G The OTG I2C channel works on exchanges. The input value is
\ ** the byte data with STOP and START triggers in bits 9 and 8.
\ ** The output is the status in the top 16 bits and the received
\ ** data in the low 8 bits. You must set the START and STOP
\ ** condition requests correctly
  _USB usbI2cTx + !  begin
    ( pause )  otgI2cSt@		\ SFP002
    $0028:0000 and  $0020:0000 <>
  until
  otgI2cSt@ dup $0020:0000 and 0= if
    _USB usbI2cRx + @ $0FF and or
  endif
;

: otgI2c!	\ b reg --
\ *G Write a byte to an ISP1301 register.
  otgDevSelW otgI2cXchg drop		\ send device addr + START + WRITE
  $0FF and otgI2cXchg drop		\ send register number
  $0FF and $200 or otgI2cXchg drop	\ send byte and STOP
;

: otgI2c@	\ reg -- b
  otgDevSelW otgI2cXchg drop		\ send device addr + START + WRITE
  $0FF and $200 or otgI2cXchg drop	\ send register and STOP
  otgDevSelR otgI2cXchg drop		\ send device addr + START + READ
  $2FF otgI2cXchg			\ read byte and STOP
  dup $200:0000 and abort" OTG I2C RFE"	\ check RFE bit
  $0FF and				\ -- b
;


\ *P ISP 1301 registers
\ *[
$00 equ ispVIDlo
$01 equ ispVIDhi
$02 equ ispPIDlo
$03 equ ispPIDhi
$04 equ ispMC1set
$05 equ ispMC1clr
$06 equ ispOTGset
$07 equ ispOTGclr
$08 equ ispINTsrc
$0A equ ispILset
$0B equ ispILclr
$0C equ ispIELOset
$0D equ ispIELOclr
$0E equ ispIEHIset
$0F equ ispIEHIclr
$10 equ ispOTGst
$12 equ ispMC2set
$13 equ ispMC2clr
\ *]

: USBhwInit	\ --
\ *G Enable access to the USB peripheral. Initialise the OTG
\ ** system. The USB clock PLL is set in the start up code. The
\ ** ISP1301 is set to VP_VM unidirectional mode.
  ~pcUSB _PCONP or!			\ power up
  _USB
  $01F over USBClkCtrl + !		\ enable all USB and OTG clocks, inc. host
  begin					\ wait for all to settle
    pause  dup USBClkSt + @
    $01F and $01F =
  until
  otgPS3val otgPS3mask _PINSEL3 setMask	\ configure pins
  0 over otgStCtrl + !			\ configure as DEVICE, U1, in base state
  0 over otgIntEn + !			\ no OTG interrupts
  $0F over otgIntClr + !		\ clear any pending interrupts
  #otgI2cClks over			\ set up for 50:50 I2C clock
  2dup usbI2cClkHi + !  usbI2cClkLo + !
  drop
  $01 ispMC1set otgI2c!			\ full-speed mode
  $02 ispMC2set otgI2c!			\ unidirectional, software speed/susp
\  $b3 ispOTGclr otgI2c!  $4c ispOTGset otgI2c!	\ pull down D+, D-, Vbus
  $B6 ispOTGclr otgI2c!			\ pull up D-, pull down D+, Vbus
  $49 ispOTGset otgI2c!
  $7F ispIELOset otgI2c!		\ enable interrupts
  $7F ispIEHIset otgI2c!
;
\  $b6 ispOTGclr otgI2c!  $49 ispOTGset otgI2c!	\ pull up D+, pull down D-, Vbus

: otgI2cClrInt	\ mask --
\ *G Clear OTG I2C interrupts according to the mask.
  ispILclr otgI2c! ;

: otgI2cWaitInt	\ --
\ *G Wait for an OTG I2C interrupt. Currently polled on P1.27.
  begin
    pause  _FIO1 FIOPIN + @ bit27 and 0=
  until
;

0 [if]

: testOtg	\ --
\ *G OTG test code that is conditionally compiled.
  cr
  begin
    $FF otgI2cClrInt
    ispINTsrc otgI2c@ 1 and 0=			\ -- vbus?
   while
    otgI2cWaitInt
  repeat
  ." connect" cr
  $b6 ispOTGclr otgI2c!  $49 ispOTGset otgI2c!	\ pull up D+, pull down D-, Vbus
  begin
    $FF otgI2cClrInt
    ispINTsrc otgI2c@ 1 and			\ -- vbus?
   while
    ispINTsrc otgI2c@ $14 and 0=
    if ." reset" cr then
    otgI2cWaitInt
  repeat
  ." disconnect" cr
  $b3 ispOTGclr otgI2c!  $4c ispOTGset otgI2c!	\ pull down D+, D-, Vbus
;

[then]


\ ======
\ *> ###
\ ======

decimal

