\ RTC1_ms_timer.fth - NRF51xxx 1ms timer using RTC1

((
Derived from RTC1_ms_timer.c
COPYRIGHT FLIPTRONICS (C)   2013

Copyright (c) 2016
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20161025 MPE001 First release.
))

\ ==============
\ *! rtc1mstimer
\ *T NRF51xxx 1ms timer using RTC1
\ ==============
\ *P This library module sets up RTC1 to interrupt every millisecond
\ ** (actually .0009765625 seconds). It maintains a global
\ ** variable, *\fo{ticks}, which is a 32 bit unsigned value
\ ** that overflows every 49 days.

\ *P Since the RTC1 is an RC ocillator, it is not very accurate.
\ ** The RTC interrupt occurs in slightly under 1 ms, even if the
\ ** oscillator is perfect. The code tries to compensate for the
\ ** deficiencies and this can lead to occasions when *\fo{RTC1ms}
\ ** is not incremented in the interrupt routine. This occurs 24
\ ** times per second, at every 42nd interrupt. This gets us
\ ** pretty close. Overall the goal is to have *\fo{RTC1ms}
\ ** be incremented 1000 times per second.

[undefined] ticks [if]
0 value ticks		\ -- ms
\ *G Return current clock value in milliseconds. This value
\ ** can treated as a 32 bit unsigned value that will wrap
\ ** when it overflows. Compiled if not already defined.
[then]
0 value RTC1skip	\ -- u

: 1msTicker	\ --
\ *G High level version of ticker ISR.
\  tick-ms addr ticks +!			\ update ticker
  _RTC1 dup rtc.EVENTS_TICK + @ 1 and if	\ if ticker event
    0 over rtc.EVENTS_TICK + !			\ clear event
    1 addr RTC1skip +!  RTC1skip #42 < if
      1 addr ticks +!
    else
      0 to RTC1skip
    endif
  endif
  drop
[ timebase? ] [if] do-timers [then]
;
' 1msTicker RTC1_IRQn EXC: 1msTickISR	\ -- addr
\ *G Setting the high level ISR.

: start-clock	\ --
\ *G Start the the RTC1 timer to produce a 1 ms ticker.
\ Set up clock source for RTC (both RTC0 and RTC1
  _CLOCK
  0 over clk.LFCLKSRC + !		\ select RC oscillator
  0 over clk.EVENTS_LFCLKSTARTED + !	\ clear start event
  1 over clk.TASKS_LFCLKSTART + !	\ trigger start
  begin					\ wait for start
    dup clk.EVENTS_LFCLKSTARTED + @ 1 and
  until
  0 over clk.EVENTS_LFCLKSTARTED + !	\ clear start event
  drop
\ Configure RTC1 to interrupt every .0009765625 seconds
  0 to ticks
  _RTC1
  1 over rtc.TASKS_STOP + !		\ stop
  #31 over rtc.PRESCALER + !		\ 32768hz / 32 = 1024 Hz
  1 over rtc.EVTENSET + !		\ TICK event enable
  1 over rtc.INTENSET + !		\ TICK interrupt enable
  1 over rtc.TASKS_START + !		\ start
  drop
  RTC1_IRQn EnInt			\ enable interrupt
;

: stop-clock	\ --
\ *G Stop the RTC1 1ms interrupt.
  1 _RTC1 rtc.TASKS_STOP + !		\ stop
  RTC1_IRQn DisInt			\ enable interrupt
;

[undefined] ms [if]
: ms            \ n --
\ *G Waits for n milliseconds.
  ticks +
  begin
[ tasking? ] [if]  pause  [then]
    dup ticks - 0<
  until
  drop
;
[then]


\ ********************
\ Time base extensions
\ ********************

timebase? [if]
: start-timers  \ -- ; Start internal time clock
\ *G Initialise the TIMEBASE system.
  stop-clock init-timers start-clock
;

: stop-Timers   \ -- ; disable timer system
\ *G Stop the TIMEBASE ticker.
  stop-clock
;
[then]
