\ ARM Cortex Forth cross assembler

((
Copyright (c) 2008, 2009, 2010, 2011, 2012, 2013, 2014, 2015, 2017, 2022, 2023
MicroProcessor Engineering
133 Hill Lane
Southampton SO15 5AF
England

tel: +44 (0)23 8063 1441
fax: +44 (0)23 8033 9691
net: mpe@mpeforth.com
     tech-support@mpeforth.com
web: www.mpeforth.com

From North America, our telephone and fax numbers are:
  011 44 23 8063 1441
  011 44 23 8033 9691


To do
=====

Change history
==============
20230712 SFP015 Added LDR @= <imm32> which loads from the
                literal pool
20221031 SFP014 Enhanced detection of d=R13/15.
20170419 SFP013 Enhanced Thumb short move generation.
20150617 SFP012 Added DMB, DSB, ISB to ARM.
20150612 SFP011 Updated documentation.
20141118 SFP010 Added some VFP instructions.
20130902 SFP009 Corrected some three register operations.
20130301 SFP008	Added some DSP formats.
20120209 SFP007 Base corrections.
20110513 SFP006 Cortex M0x instruction selection and checks.
		Fixed SEV_T.
20100506 SFP005 Added ARMv5 switches.
20100430 SFP004 Corrected STRH_T.
20100427 SFP003 Corrected EOR_T.
20100207 SFP002 Corrected FIX_ADR.
20100205 SFP001 DocGen update for first release.
))


\ ============
\ *! xasmcortex
\ *T ARM Cortex Cross assembler
\ ============
\ *P The MPE cross compiler has a built-in cross-assembler.  This gives
\ ** you the ability to define new Forth words in assembler as well as
\ ** in Forth.  You can also assemble code to anywhere in memory.
\ ** This section is *\b{not} a treatise on ARM Cortex assembly language
\ ** programming. The essential document for Cortex-M is the ARMv7-M
\ ** Architecture Reference Manual, which is available from *\f{www.arm.com}
\ ** as a PDF file document reference *\i{ARM DDI 0403}. You may have to register
\ ** to download it. The ARM 32 bit instruction set is documented in
\ ** *\i{ARM DDI 0100}. A copy is provided in the *\i{Docs\ARM} folder
\ ** in PDF format. The ARMv6 instruction set used in the Raspberry Pi
\ ** is documented as an appendix to the ARMv7-A Architecture
\ ** Reference Manual.

\ *P The full Thumb-2 instruction set for Cortex-M3/M4 is supported.
\ ** The instruction notation is very similar to that provided
\ ** with the MPE ARM compiler. The Cortex assembler notation is
\ ** very close to that referred to by ARM as UAL (Unified
\ ** Assembly Language), which was designed to improve portability
\ ** between ARM and Cortex at the assembler source code level.

\ *P This assembler can be switched between Cortex and ARM
\ ** instruction sets in order to support the Cortex-A profile.
\ ** Note that the Thumb-2 instruction set can be regarded as an
\ ** encoding of the ARM instruction set with better code density
\ ** and features to support system programming. That ARM have
\ ** achieved this objective is indicated by the fact that with
\ ** only minor extensions to the compiler (see "Intrinsics" in
\ ** the code generator chapter), there is no assembly code in
\ ** the Forth start-up files.

\ *P By default, the assembler and VFX code generator are set to
\ ** use the legacy ARM instruction set with the TOS register set
\ ** to R10. This is the configuration used by VFX Forth for ARM
\ ** Linux.


\ **************************
\ *S Why write in assembler?
\ **************************
\ *P Forth is compact and quick, so why write in assembler? An
\ ** assembler definition is normally faster than a group of
\ ** corresponding Forth words. For Cortex CPUs, hand coding can
\ ** improve performance by keeping more data in registers in
\ ** loops and by taking more advantage of conditional execution.

\ *P That having been said, MPE does not write Cortex or ARM
\ ** code (even interrupt handlers) in assembler except in very rare cases.


\ ************************************
\ *S Creating Forth words in assembler
\ ************************************
\ *P Forth words can easily be defined in assembler. They increase the
\ ** execution speed of your code and can sometimes make your code
\ ** smaller.

\ ===========================
\ *N Defining assembler words
\ ===========================
\ *P Forth words written in assembler follow a similar form to a
\ ** word written in Forth. Instead of a colon you have *\fo{CODE}.
\ ** Instead of semi-colon you have *\fo{END-CODE}. For example:
\ *E CODE <name>
\ **   ...
\ **   ...
\ **   NEXT,
\ ** END-CODE
\ *P creates a word called *\fo{<name>}. Any assembler code
\ ** between the *\fo{CODE} and *\fo{END-CODE} will be assembled
\ ** into the word. When executed, the macro *\fo{NEXT,} will
\ ** stop the execution of the assembler and return to the calling
\ ** word.

\ ==========================
\ *N Writing assembler words
\ ==========================
\ *P The syntax used for the opcodes has been kept similar to the
\ ** standard ARM syntax.  See the list at the end of the chapter for a
\ ** comparison of ARM versus Forth syntax.

\ *P As a company, ARM changes or extends the instruction
\ ** sets in use at any time as new cores are designed. By normal
\ ** Forth standards, this assembler is a *\b{huge} piece of code.
\ ** Forgive us for deviations from the expected instruction
\ ** syntax, in places our assembler notation reflects ease of
\ ** implementation.

\ =================
\ *N Register names
\ =================
\ *P Registers are defined as follows.
\ *D Rn      The familiar integer registers R0..R15
\ *D CRn     Coprocessor registers CR0..CR15
\ *D xPSR    Processor status registers, e.g. CPSR and SPSR
\ *D Sn      Single-precision (32 bit) floating point or vector
\ **         register S0..S31
\ *D Dn      Double-precision (64 bit) floating point or vector
\ **         register D0..D15/31
\ *D Qn      Quadruple (128 bit) vector register, Q0..Q15

\ *P Note that the floating point and vector registers overlap.
\ ** The mapping between the registers is as follows:
\ *(
\ *B S<2n> maps to the least significant half of D<n>
\ *B S<2n+1> maps to the most significant half of D<n>
\ *B D<2n> maps to the least significant half of Q<n>
\ *B D<2n+1> maps to the most significant half of Q<n>
\ *)

\ *P Where the instruction cannot distinguish between floating
\ ** point and integer operation, use the


\ =================================
\ *N Preserving the Forth registers
\ =================================
\ *P The Forth interpreter and compiler use some of the target
\ ** processor's registers. These must be preserved if they are
\ ** used in the assembler. They can be saved on the stack, in
\ ** memory or in other registers and restored at the end of
\ ** the word. The ARM registers that are used are shown in the
\ ** table below.

\ ------------------------
\ *H Cortex register usage
\ ------------------------
\ *P *\table{ "Cortex Register usage" 20 20 60 }
\ ** *\hrow{} *\col{} Cortex register *\col{} Forth register *\col{} Notes
\ ** *\row{} *\col{} R15 or PC *\col{} IP
\ ** *\col{} The program counter. Altering this register will
\ ** cause the processor to jump to a new address.
\ ** *\row{} *\col{} R14 or LINK *\col{} _
\ ** *\col{} The link register. When a subroutine is entered
\ ** via the BL instruction, the return address is cached in R14.
\ ** If a further BL is to be executed within the subroutine,
\ ** remember to save the contents of R14, usually on the return
\ ** stack. Note that in Thumb mode, bit 0 will be set to 1 to
\ ** indicate that the CPU is in Thumb mode.
\ ** *\row{} *\col{} R13 or SP *\col{} RSP
\ ** *\col{} Forth return stack pointer, do not change this
\ ** without good reason. This stack holds return addresses. Note
\ ** that when entering a subroutine or word via the BL instruction
\ ** the return address is cached in R14, the link register. R13
\ ** is used in many ARM systems as a stack pointer.
\ ** *\row{} *\col{} R12 *\col{} PSP (1)
\ ** *\col{} Forth data stack pointer, do not change this without
\ ** good reason. Use it for passing parameters between words.
\ ** When writing assembler code, use *\fo{PSP} rather than
\ ** *\fo{R12}. Future versions of the compiler may use a
\ ** different register for the data stack pointer.
\ ** *\row{} *\col{} R11 *\col{} UP
\ ** *\col{} Pointer to the base of the current User Area.
\ ** *\row{} *\col{} R10 *\col{} TOS (1)
\ ** *\col{} Currently the default register for TOS for ARM32, but this
\ ** may change when interworking ARM and Cortex code.
\ ** *\row{} *\col{} R9 *\col{} LP *\col{} Local variable frame pointer.
\ ** *\row{} *\col{} R8 *\col{} --
\ ** *\col{} Currently unused, but we have plans for it.
\ ** *\row{} *\col{} R7 *\col{} TOS (2)
\ ** *\col{} Instead of holding the top item of the Forth data
\ ** stack in main memory, it is held in a register. This allows
\ ** many simple operations to execute faster, and it also reduces
\ ** the amount of memory traffic. For hosted systems such as VFX
\ ** Forth for Linux, TOS will be in R10 by default, but for code
\ ** density in the Thumb-2 instruction set, R7 is a better
\ ** choice.
\ ** *\row{} *\col{} R6 *\col{} PSP (2)
\ ** *\col{} For best code density with the Thumb-2 instruction
\ ** set, R6 is a better choice than R10 as the PSP.
\ ** *\row{} *\col{} R0..R6 *\col{} _ *\col{} Scratch
\ ** *\endtable{}

\ ---------------------
\ *H ARM register usage
\ ---------------------
\ *P *\table{ "ARM Register usage" 20 20 60 }
\ ** *\hrow{} *\col{} ARM register *\col{} Forth register *\col{} Notes
\ ** *\row{} *\col{} R15 or PC *\col{} IP
\ ** *\col{} The ARM program counter. Altering this register will
\ ** cause the processor to jump to a new address.
\ ** *\row{} *\col{} R14 or LINK *\col{} _
\ ** *\col{} The ARM link register. When a subroutine is entered
\ ** via the BL instruction, the return address is cached in R14.
\ ** If a further BL is to be executed within the subroutine,
\ ** remember to save the contents of R14, usually on the return
\ ** stack.
\ ** *\row{} *\col{} R13 or SP *\col{} RSP
\ ** *\col{} Forth return stack pointer, do not change this
\ ** without good reason. This stack holds return addresses. Note
\ ** that when entering a subroutine or word via the BL instruction
\ ** the return address is cached in R14, the link register. R13
\ ** is used in many ARM systems as a stack pointer.
\ ** *\row{} *\col{} R12 *\col{} PSP
\ ** *\col{} Forth data stack pointer, do not change this without
\ ** good reason. Use it for passing parameters between words.
\ ** *\row{} *\col{} R11 *\col{} UP
\ ** *\col{} Pointer to the base of the current User Area.
\ ** *\row{} *\col{} R10 *\col{} TOS
\ ** *\col{} Instead of holding the top item of the Forth data
\ ** stack in main memory, it is held in a register. This allows
\ ** many simple operations to execute faster, and it also reduces
\ ** the amount of memory traffic.
\ ** *\row{} *\col{} R9 *\col{} LP *\col{} Local variable frame pointer.
\ ** *\row{} *\col{} R0..R8 *\col{} _ *\col{} Scratch
\ ** *\endtable{}

\ ==============================
\ *N Executing an assembler word
\ ==============================
\ *P A Forth word written in assembler is executed in the same way as a
\ ** word written in Forth.  It is executed in the same way as a normal
\ ** word, by stating its name.

\ =========================
\ *S Assembling into memory
\ =========================
\ *P Assembler code can be assembled into memory and not in a Forth
\ ** word.  To do this you need to:
\ *(
\ *B turn on the assembler
\ *B write your assembler code
\ *B turn off the assembler
\ *)

\ *P To turn on the assembler, use the word *\fo{AsmCode}. To
\ ** switch back to Forth use the word *\fo{End-Code}. Between
\ ** these two words, any assembler will be assembled. The
\ ** assembled code will be placed in the dictionary without a
\ ** header. The code can be executed by the use of labels. This
\ ** is often used to define low-level interrupts. See the
\ ** chapter on Interrupts for more details on writing low-level
\ ** interrupts.

\ ***************************************
\ *S Creating defining words in assembler
\ ***************************************
\ *P The cross compiler allows you to define the run-time (*\fo{DOES>})
\ ** part of a defining word in assembler. To do this use
\ ** *\fo{;CODE} in the form:
\ *E : <name>
\ **   CREATE
\ **     ...
\ **   ;CODE
\ **     ...
\ ** END-CODE
\ *P An example is shown below:
\ *E : VARIABLE    \ <spaces>name -- ; -- addr
\ **   CREATE                     \ Create header
\ **     0 ,                      \ Initial value
\ **   ;CODE                      \ Run-time action
\ ** \ Cortex version
\ **     str tos, [ psp, # -4 ] ! \ save TOS
\ **     ldr tos, [ link, # -1 ]  \ get const val/addr from afer BL
\ **     pop     { pc }
\ ** END-CODE
\ ** \ ARM version
\ **     stmfd psp ! { tos }      \ Save TOS
\ **     ldr tos, [ link ], # 4   \ get pointer to data
\ **     ldr tos, [ tos ]         \ get variable address
\ **     NEXT,
\ ** END-CODE

\ *************************
\ *S Structured programming
\ *************************
\ *P Three facilities are available to give you the advantages of
\ ** structured programming, in assembler:
\ *(
\ *B control structures
\ *B labels
\ *B local labels
\ *)

\ =====================
\ *N Control structures
\ =====================
\ *P There are assembler equivalents to the Forth control structures.
\ ** The available structures are:
\ *E AHEAD, ... THEN, or ENDIF,
\ ** cc IF, ... THEN, or ENDIF,
\ ** cc IF, ... ELSE, ... THEN, or ENDIF,
\ ** BEGIN, ... cc UNTIL,
\ ** BEGIN, ... cc WHILE, ... REPEAT
\ ** BEGIN, ... AGAIN,
\ *P where cc is one of the condition codes in the table below.
\ ** *\table{"Condition code mnemonics" 15 15 20 15 15 20}
\ ** *\hrow{}
\ ** *\col{} ARM *\col{} Forth *\col{} Condition
\ ** *\col{} ARM *\col{} Forth *\col{} Condition
\ ** *\row{}
\ ** *\col{} .CS *\col{} CS, *\col{} carry set
\ ** *\col{} .NE *\col{} NE, *\col{} not equal or non-zero
\ ** *\row{}
\ ** *\col{} .CC *\col{} CC, *\col{} carry clear
\ ** *\col{} .GE *\col{} GE, *\col{} greater than or equal
\ ** *\row{}
\ ** *\col{} .PL *\col{} PL, *\col{} plus - positive or zero
\ ** *\col{} .LT *\col{} LT, *\col{} less than
\ ** *\row{}
\ ** *\col{} .MI *\col{} MI, *\col{} minus - negative
\ ** *\col{} .GT *\col{} GT, *\col{} greater than
\ ** *\row{}
\ ** *\col{} .VS *\col{} VS, *\col{} overflow set
\ ** *\col{} .LS *\col{} LS, *\col{} unsigned less than or equal (same)
\ ** *\row{}
\ ** *\col{} .VC *\col{} VC, *\col{} overflow clear
\ ** *\col{} .HS *\col{} HS, *\col{} unsigned greater than or equal
\ ** (same). Same as CS
\ ** *\row{}
\ ** *\col{} .LE *\col{} LE, *\col{} less than or equal
\ ** *\col{} .LO *\col{} LO, *\col{} unsigned less than. Same as CC
\ ** *\row{}
\ ** *\col{} .EQ *\col{} EQ, *\col{} equal or zero
\ ** *\col{} .HI *\col{} HI, *\col{} unsigned greater than
\ ** *\row{}
\ ** *\col{} .AL *\col{}* Always (default)
\ ** *\endtable{}

\ ------------------------
\ *H Cortex IT instruction
\ ------------------------
\ *P The Thumb-2 instruction set does not support the conditional
\ ** execution facilities of the ARM instruction set. Instead,
\ ** it provides the *\fo{IT} instruction.

\ *P In order to avoid performance penalties caused by taken
\ ** branches and associated cache flushes, the Thumb-2
\ ** instruction set provides the *\fo{ITxxx <cond>} instruction.
\ ** This permits up to four instructions to be executed
\ ** depending on the condition flags at the start. The first
\ ** instruction is executed if <cond> is true. The next three
\ ** instructions are executed if <cond> is true and x=T or
\ ** if <cond> is false and x=E. The following example illustrates
\ ** the use of the IT instruction.
\ *E CODE WITHIN?     \ n1 n2 n3 -- flag
\ ** \ Return TRUE if N1 is within the range N2..N3.
\ ** \ This word uses signed arithmetic.
\ **   ldmfd   psp ! { r0, r1 }
\ **   mov .s  r2, # 0
\ **   mov .s  r3, # 0
\ **   cmp     r1, r0
\ **   it .ge  \ next instruction if condition met
\ **     mov     r2, # 1
\ **   cmp     r1, tos
\ **   it .le  \ next instruction if condition met
\ **     mov     r3, # 1
\ **   tst     r2, r3
\ **   ite .ne
\ **     mvn     tos, # 0  \ if condition met
\ **     mov     tos, # 0  \ if condition not met
\ **   next,
\ ** END-CODE

\ -------------------------------
\ *H ARM conditional instructions
\ -------------------------------
\ *P The ARM is different from many processors in that many instructions
\ ** can be executed conditionally depending on the processor status
\ ** flags, by appending one of the mnemonics in the table above to the
\ ** instruction. An instruction without a condition suffix is assumed
\ ** to use *\fo{.AL}. Note that most instructions (except the test and
\ ** compare instructions) do not set the status flags by default. This
\ ** has to be done with the *\fo{.S} suffix:
\ *E ADD .S R0, R1, R2          \ Add, set condition codes
\ ** ADD .NE .S R0, R1, R2      \ if NE and set condition codes
\ **
\ ** CS, IF,                    \ do between IF, and ENDIF, if CS set
\ **   ...
\ ** ENDIF,

\ *P It is often quicker to avoid short jumps in code such
\ ** as those typically generated by *\fo{IF,} statements, by the use of
\ ** conditionally executed instructions. Skipping several instructions
\ ** is generally faster than using a branch instruction as this
\ ** involves flushing the processor pipeline. See the file *\i{CODEARM.FTH}
\ ** for examples of conditional execution.

\ =========
\ *N Labels
\ =========
\ *P Labels can be used to mark a place in assembler code.  That place
\ ** can then be referenced in other areas of code.

\ -------------------
\ *H Creating a label
\ -------------------
\ *P Labels can be defined by using the command *\fo{L: <name>}.
\ ** It is used in the form:
\ *C l: <name>
\ *P where *\fo{<name>} is what you want to call the label.

\ ----------------------
\ *H Referencing a label
\ ----------------------
\ *P A label is referenced by stating its name. For example,
\ *C   B .EQ <name>

\ ===============
\ *N Local labels
\ ===============
\ *P If you need to use labels within a code definition, you may use
\ ** the local labels provided.  These are used just as normal labels
\ ** in the assembler, but some restrictions apply:
\ *(
\ *B there is a maximum of ten labels
\ *B the names are in the form L$n where n is in the range 1 to 10
\ *B a reference is valid until the next occurrence of *\fo{CODE}
\ ** or *\fo{;CODE}.
\ *)

\ -------------------------
\ *H Creating a local label
\ -------------------------
\ ** To define a local label use *\fo{L$n:}, where *\fo{n} is a
\ ** number from one to ten.  For example:
\ *C L$1:

\ ----------------------------
\ *H Referencing a local label
\ ----------------------------
\ *P To reference a local label, type its name. For example,
\ *C B L$1
\ *P assembles code for a branch to L$1:.

\ ******************
\ *S Creating macros
\ ******************
\ *P A macro is a word that lays down code 'in-line' within an
\ ** assembler definition. Macros are used when there is a
\ ** repetitive use of a series of opcodes.

\ ===================
\ *N Defining a macro
\ ===================
\ *P The easiest way to create a macro is by using *\fo{MACRO:}.
\ ** The macro below can be used as a divide step operation.
\ *E macro: Udiv63/31_step   \ --
\ **   adc .s  r1, r1, r1
\ **   adc .s  r0, tos, r0, lsl # 1
\ **   sub .cc r0, r0, tos
\ ** ;m
\ *P The assembler's prefix notation leads to some peculiarities
\ ** when writing macros.
\ *( 1
\ *B Opcode words, e.g. *\fo{ADC} above, assemble the code for
\ ** the *\b{previous} instruction, so you may not know the stack
\ ** conditions. Put literal data on the return stack in the macro,
\ ** and retrieve it as needed.
\ *B The assembler wordlist is first in the search order, so be
\ ** careful that a normal Forth word name is not a name in the
\ ** assembler.
\ *B Register names are executable words. To pass registers by
\ ** number use *\fo{R# ( n -- )} for general purpose registers,
\ ** *\fo{CR# ( n -- )} for coprocessor registers, and *\fo{SR# ( n -- )}
\ ** for status registers (0=CPSR, 1=SPSR).
\ *)

\ *P A macro can also be defined using colon and semi-colon before
\ ** *\fo{CROSS-COMPILE} is executed. It must be defined in the
\ ** cross compiler's *\fo{ASM-ACCESS} vocabulary.
\ ** The place to create a macro is in the control file and it must
\ ** be defined before *\fo{CROSS-COMPILE}. As an example, the
\ ** macro *\fo{NEXT,} is shown below. *\fo{NEXT,} is defined as
\ ** a macro, so each time it is used, its code is laid down.
\ ** This makes it quicker than calling a subroutine.
\ *E \ switch to cross compilers assembler vocab
\ ** FORTH ALSO C-C ALSO ASSEMBLER
\ ** ALSO ASM-ACCESS DEFINITIONS
\ ** \ define NEXT
\ ** : NEXT,   \ -- ; lay in-line next code
\ **   bx  lr
\ ** ;
\ ** \ switch back to normal forth vocabulary
\ ** ONLY FORTH DEFINITIONS

\ ================
\ *N Using a macro
\ ================
\ *P A macro is used by stating its name. For example, in a *\fo{CODE}
\ ** definition, *\fo{NEXT,} is a macro.


\ ************
\ *S Debugging
\ ************
\ *P It is possible to disassemble compiled words using:
\ *C   XDASM <name>
\ *C   DIS <name>
\ *P This can be done during compilation by including an *\fo{XDASM}
\ ** statement in the control file, or interactively after compilation
\ ** by including the word *\fo{INTERACTIVE} before *\fo{FINIS}.


\ ****************
\ *S CPU selection
\ ****************
\ *P The instruction set of the processor is extended on various
\ ** processor cores. The selection available in this assembler
\ ** is for ARM32 with/without Thumb-1 and the Cortex-M0, M1, M3
\ ** and M4.

\ ***************
\ *S Number bases
\ ***************
\ *P The number base in the Forth assembler can be indicated by
\ ** *\fo{BINARY}, *\fo{DECIMAL}, and *\fo{HEX}. In addition,
\ ** numbers prefixed by the '$' '#' and '%' characters are
\ ** treated as special cases.
\ ** These characters affect the number base for that number only.
\ ** Note that the characters '$' and '%' follow Motorola usage.
\ ** Note also that the '#' symbol attached to a number is not the same
\ ** as the word *\fo{#} word that indicates immediate addressing.

\ ** *\table{"Number bases" 33 33 34 }
\ ** *\hrow{} *\col{} Symbol *\col{} Base *\col{} Example
\ ** *\row{} *\col{} $ *\col{} hex *\col{} $55AA
\ ** *\row{} *\col{} # *\col{} decimal *\col{} #1234
\ ** *\row{} *\col{} % *\col{} binary *\col{} %1011001
\ ** *\endtable{}


\ **************************
\ *S Thumb-2 instruction set
\ **************************
\ *P Thumb-2 literals are encoded in a number of ways. The term
\ ** *\i{imm} refers to an immediate value that is directly
\ ** encoded across a number of bit fields. The term *\i{const}
\ ** refers to an immediate value that comes from a restricted
\ ** but wide range. See *\i{ARM DDI 0403} for the gory details.
\ ** Items in square brackets are optional, e.g. *\fo{[.s]} indicates
\ ** that *\fo{.s} to set the conditions flags is optional.

\ *P When one of the *\fo{ITxxx} instruction applies, *\b{do not} apply
\ ** *\fo{.s}. The assembler will do what it has to.

\ *P The term <shift> means one of
\ *E .LSL # n    logical left shift by n
\ ** .LSR # n    logical right shift by n
\ ** .ASR # n    arithmetic right shift by n
\ ** .ROR # n    rotate right by n
\ ** .RRX        1 bit rotate right, carry in to new bit 31,
\ **             old bit 0 to carry out.
\ *P You can leave out <shift> and it will be encoded as
\ ** LSL # 0. Note that when .S applies, the rules for the
\ ** final value of the carry flag are a bit arcane.

\ *P You can force an instruction to use the 16 bit form by
\ ** using the *\fo{.n} indicator. You can force the 32 bit form
\ ** using the *\fo{.w} indicator. Without either of these, the
\ ** assembler will choose the shortest form. Using R0..R7 (the
\ ** low registers) and .S usually generates the shortest code.
\ ** Except in a few cases, using R8..R15 (the high registers)
\ ** will generate a 32 bit instruction.

\ *P Where a register appears twice in an instruction (usually Rd),
\ ** that particular encoding (usually a 16 bit form) is only
\ ** generated when the same register appears twice.

\ *P Use of the PC (R15) or SP (R13) registers may cause assembler
\ ** errors as these register fields are frequently used to handle
\ ** special instructions. Consult *\i{ARM DDI 0403}.

\ =======================
\ *N Base instruction set
\ =======================
\ *P This is the Cortex-M3 instruction set. Note that the
\ ** Cortex-M0/M1 instruction set is a subset of this. In ARM
\ ** terminology the Cortex-M0/M1 is defined in ARMv6-M.

\ *P In some cases using *\fo{.s} leads to shoter code because a
\ ** 16 bit Thumb-1 encoding is available. When coding for Cortex-M0
\ ** remember that the *\fo{.s} is often required.

\ *E ADC [.s]    Rd, Rn, # <const>
\ ** ADC [.s]    Rd, Rn, Rm
\ ** ADC [.s]    Rd, Rn, Rm <shift>
\ **
\ ** ADD .s      Rd, Rn, # <imm3>
\ ** ADD .s      Rd, Rd, # <imm8>
\ ** ADD [.s]    Rd, Rn, # <const>
\ ** ADD [.s]    Rd, Rn, # <imm12>
\ ** ADD .s      Rd, Rn, Rm
\ ** ADD         Rd, Rd, Rm
\ ** ADD [.s]    Rd, Rn, Rm <shift>
\ ** ADD         Rd, SP, # <imm8>
\ ** ADD         SP, SP, # <imm7>
\ ** ADD [.s]    Rd, SP, # <const>
\ ** ADD [.s]    Rd, SP, # <imm12>
\ ** ADD         Rd, SP, Rd
\ ** ADD         SP, SP, Rm
\ ** ADD         Rd, SP, Rm <shift>
\ **
\ ** ADR         Rd, <label>
\ **
\ ** AND         Rd, Rn, # <const>
\ ** AND         Rd, Rd, Rm
\ ** AND [.s]    Rd, Rn, Rm <shift>
\ **
\ ** ASR .s      Rd, Rn, # <imm5>
\ ** ASR [.s]    Rd, Rn, # <imm5>
\ ** ASR .s      Rd, Rd, Rm
\ ** ASR [.s]    Rd, Rn, Rm

\ *E B <cond>    <label>
\ ** B           <label>
\ **
\ ** BFC         Rd, # <lsb> # <width>
\ ** BFI         Rd, Rn, # <lsb> # <width>
\ **
\ ** BIC [.s]    Rd, Rn, # <const>
\ ** BIC .s      Rd, Rd, Rm
\ ** BIC [.s]    Rd, Rn, Rm <shift>
\ **
\ ** BKPT        # <imm8>
\ **
\ ** BL          <label>
\ **
\ ** BLX         Rm
\ **
\ ** BX          Rm

\ *E CBNZ        Rn, <label>
\ ** CBZ         Rn, <label>
\ **
\ ** CDP         <copro> <opc1> CRd, CRn, CRm, <opc2>
\ ** CDP2        <copro> <opc1> CRd, CRn, CRm, <opc2>
\ **
\ ** CLREX
\ **
\ ** CLZ         Rd, Rm
\ **
\ ** CMN         Rn, # <const>
\ ** CMN         Rn, Rm
\ ** CMN         Rn, Rm <shift>
\ **
\ ** CMP         Rn, # <imm8>
\ ** CMP         Rn, # <const>
\ ** CMP         Rn, Rm
\ ** CMP         Rn, Rm <shift>
\ **
\ ** CPS .ie     [.i] [.f]
\ ** CPS .id     [.i] [.f]

\ *E DBG         # <opt4>
\ ** DMB         # <opt4>
\ ** DSB         # <opt4>
\ **
\ ** EOR [.s]    Rd, Rn, # <const>
\ ** EOR .s      Rd, Rd, Rm
\ ** EOR [.s]    Rd, Rn, Rm <shift>
\ **
\ ** ISB         # <opt4>
\ **
\ ** IT <cond>
\ ** IT <cond>
\ ** ITT <cond>
\ ** ITE <cond>
\ ** ITTT <cond>
\ ** ITET <cond>
\ ** ITTE <cond>
\ ** ITEE <cond>
\ ** ITTTT <cond>
\ ** ITETT <cond>
\ ** ITTET <cond>
\ ** ITEET <cond>
\ ** ITTTE <cond>
\ ** ITETE <cond>
\ ** ITTEE <cond>
\ ** ITEEE <cond>

\ *E LDC         <copro> CRd, [ Rn, # +/-<imm8> ]
\ ** LDC         <copro> CRd, [ Rn, # +/-<imm8> ] !
\ ** LDC         <copro> CRd, [ Rn ], # +/-<imm8>
\ ** LDC         <copro> CRd, [ Rn ], {} <option>
\ ** LDC         <copro> CRd, <label>
\ ** LDCL        <copro> CRd, [ Rn, # +/-<imm8> ]
\ ** LDCL        <copro> CRd, [ Rn, # +/-<imm8> ] !
\ ** LDCL        <copro> CRd, [ Rn ], # +/-<imm8>
\ ** LDCL        <copro> CRd, [ Rn ], {} <option>
\ ** LDCL        <copro> CRd, <label>
\ ** LDC2        <copro> CRd, [ Rn, # +/-<imm8> ]
\ ** LDC2        <copro> CRd, [ Rn, # +/-<imm8> ] !
\ ** LDC2        <copro> CRd, [ Rn ], # +/-<imm8>
\ ** LDC2        <copro> CRd, [ Rn ], {} <option>
\ ** LDC2        <copro> CRd, <label>
\ ** LDC2L       <copro> CRd, [ Rn, # +/-<imm8> ]
\ ** LDC2L       <copro> CRd, [ Rn, # +/-<imm8> ] !
\ ** LDC2L       <copro> CRd, [ Rn ], # +/-<imm8>
\ ** LDC2L       <copro> CRd, [ Rn ], {} <option>
\ ** LDC2L       <copro> CRd, <label>
\ **
\ ** LDM         Rn, [!] { ra, rb ... rn }
\ ** LDMIA       Rn, [!] { ra, rb ... rn }
\ ** LDMFD       Rn, [!] { ra, rb ... rn }
\ ** LDMDB       Rn, [!] { ra, rb ... rn }
\ ** LDMEA       Rn, [!] { ra, rb ... rn }

\ *E LDR         Rt, [ Rn, # <imm5*4> ]
\ ** LDR         Rt, [ SP, # <imm8*4> ]
\ ** LDR         Rt, [ Rn, # <imm12> ]
\ ** LDR         Rt, [ Rn, # -<imm8> ]
\ ** LDR         Rt, [ Rn ], # +/-<imm8>
\ ** LDR         Rt, [ Rn, # +/-<imm8> ] !
\ ** LDR         Rt, <label>
\ ** LDR         Rt, [ Rn ++ Rm ]
\ ** LDR         Rt, [ Rn ++ Rm, LSL # <imm2> ]
\ ** LDR         Rt, @= <imm32>    \ loads from literal pool
\ **                               \ Use FLUSHLITPOOL to lay pool.
\ **
\ ** LDRB        Rt, [ Rn, # <imm5> ]
\ ** LDRB        Rt, [ Rn, # <imm12> ]
\ ** LDRB        Rt, [ Rn, # -<imm8> ]
\ ** LDRB        Rt, [ Rn ], # +/-<imm8>
\ ** LDRB        Rt, [ Rn, # +/-<imm8> ] !
\ ** LDRB        Rt, <label>
\ ** LDRB        Rt, [ Rn ++ Rm ]
\ ** LDRB        Rt, [ Rn ++ Rm, LSL # <imm2> ]
\ **
\ ** LDRBT       Rt, [ Rn, # +<imm8> ]
\ **
\ ** LDRD        Rt, Rt2, [ Rn, # -<imm8> ]
\ ** LDRD        Rt, Rt2, [ Rn ], # +/-<imm8>
\ ** LDRD        Rt, Rt2, [ Rn, # +/-<imm8> ] !
\ ** LDRD        Rt, Rt2, <label>
\ **
\ ** LDREX       Rt, [ Rn, # +<imm8> ]
\ ** LDREXB      Rt, [ Rn ]
\ ** LDREXH      Rt, [ Rn ]
\ **
\ ** LDRH        Rt, [ Rn, # <imm5*2> ]
\ ** LDRH        Rt, [ Rn, # <imm12> ]
\ ** LDRH        Rt, [ Rn, # -<imm8> ]
\ ** LDRH        Rt, [ Rn ], # +/-<imm8>
\ ** LDRH        Rt, [ Rn, # +/-<imm8> ] !
\ ** LDRH        Rt, <label>
\ ** LDRH        Rt, [ Rn ++ Rm ]
\ ** LDRH        Rt, [ Rn ++ Rm, LSL # <imm2> ]
\ **
\ ** LDRHT       Rt, [ Rn, # +<imm8> ]

\ *E LDRSB       Rt, [ Rn, # <imm12> ]
\ ** LDRSB       Rt, [ Rn, # -<imm8> ]
\ ** LDRSB       Rt, [ Rn ], # +/-<imm8>
\ ** LDRSB       Rt, [ Rn, # +/-<imm8> ] !
\ ** LDRSB       Rt, <label>
\ ** LDRSB       Rt, [ Rn ++ Rm ]
\ ** LDRSB       Rt, [ Rn ++ Rm, LSL # <imm2> ]
\ **
\ ** LDRSBT      Rt, [ Rn, # +<imm8> ]
\ **
\ ** LDRSH       Rt, [ Rn, # <imm12> ]
\ ** LDRSH       Rt, [ Rn, # -<imm8> ]
\ ** LDRSH       Rt, [ Rn ], # +/-<imm8>
\ ** LDRSH       Rt, [ Rn, # +/-<imm8> ] !
\ ** LDRSH       Rt, <label>
\ ** LDRSH       Rt, [ Rn ++ Rm ]
\ ** LDRSH       Rt, [ Rn ++ Rm, LSL # <imm2> ]
\ **
\ ** LDRSHT      Rt, [ Rn, # +<imm8> ]
\ **
\ ** LDRT        Rt, [ Rn, # +<imm8> ]
\ **
\ ** LSL         Rd, Rm, # <imm5>
\ ** LSL [.s]    Rd, Rm, # <imm5>
\ ** LSL .s      Rd, Rd, Rm
\ ** LSL [.s]    Rd, Rn, Rm
\ **
\ ** LSR         Rd, Rm, # <imm5>
\ ** LSR [.s]    Rd, Rm, # <imm5>
\ ** LSR .s      Rd, Rd, Rm
\ ** LSR [.s]    Rd, Rn, Rm

\ *E MCR         <copro> <opc1> Rt, CRn, CRm, <opc2>
\ ** MCR2        <copro> <opc1> Rt, CRn, CRm, <opc2>
\ ** MCRR        <copro> <opc1> Rt, Rt2, CRm, <opc2>
\ ** MCRR2       <copro> <opc1> Rt, Rt2, CRm, <opc2>
\ **
\ ** MLA         Rd, Rn, Rm, Ra
\ ** MLS         Rd, Rn, Rm, Ra
\ **
\ ** MOV .s      Rd, # <imm8>
\ ** MOV [.s]    Rd, # <const>
\ ** MOV         Rd, # <imm16>
\ ** MOV         Rd, Rm
\ ** MOV [.s]    Rd, Rm
\ ** MOV [.s]    Rd, Rm <shiftop> # n
\ ** MOV [.s]    Rd, Rm <shiftop> Rs
\ ** MOV [.s]    Rd, Rm .RRX
\ ** MOVT        Rd, # <imm16>
\ **
\ ** MRC         <copro> <opc1> Rt, CRn, CRm, <opc2>
\ ** MRC2        <copro> <opc1> Rt, CRn, CRm, <opc2>
\ ** MRRC        <copro> <opc1> Rt, Rt2, CRm, <opc2>
\ ** MRRC2       <copro> <opc1> Rt, Rt2, CRm, <opc2>
\ **
\ ** MRS         Rd, <SYSm8>
\ ** MSR         <SYSm8> Rn
\ ** <SYSm8> is a special register number in the range 0..255
\ **
\ ** MUL .s      Rd, Rn, Rd
\ ** MUL [.s]    Rd, Rn, Rm
\ **
\ ** MVN [.s]    Rd, # <const>
\ ** MVN .s      Rd, Rm
\ ** MVN [.s]    Rd, Rm <shift>

\ *E NEG .s      Rd, Rm
\ **
\ ** NOP [.n/.w]
\ **
\ ** ORN [.s]    Rd, Rm, # <const>
\ ** ORN [.s]    Rd, Rn, Rm <shift>
\ **
\ ** ORR [.s]    Rd, Rm, # <const>
\ ** ORR .s      Rd, Rd, Rm
\ ** ORR [.s]    Rd, Rn, Rm <shift>
\ **
\ ** PLD         [ Rn, # +<imm12> ]
\ ** PLD         [ Rn, # -<imm8> ]
\ ** PLD         <label>
\ ** PLD         [ Rn ++ Rm, LSL # <imm2> ]
\ ** PLDW        [ Rn, # +<imm12> ]
\ ** PLDW        [ Rn, # -<imm8> ]
\ ** PLI         [ Rn, # +<imm12> ]
\ ** PLI         [ Rn, # -<imm8> ]
\ ** PLI         <label>
\ ** PLI         [ Rn ++ Rm, LSL # <imm2> ]
\ **
\ ** POP         { ra, rb ... rn }
\ ** PUSH        { ra, rb ... rn }
\ **
\ ** RBIT        Rd, Rm
\ ** REV         Rd, Rm
\ ** REV16       Rd, Rm
\ ** REVSH       Rd, Rm
\ **
\ ** ROR [.s]    Rd, Rm, # <imm5>
\ ** ROR .s      Rd, Rd, Rm
\ ** ROR [.s]    Rd, Rn, Rm
\ **
\ ** RRX         Rd, Rm

\ *E RSB .s      Rd, Rn, # 0
\ ** RSB [.s]    Rd, Rn, # <const>
\ ** RSB [.s]    Rd, Rn, Rm <shift>
\ **
\ ** SBC [.s]    Rd, Rn, # <const>
\ ** SBC .s      Rd, Rd, Rm
\ ** SBC [.s]    Rd, Rn, Rm <shift>
\ **
\ ** SBFX        Rd, Rn, # <lsb> # <width>
\ **
\ ** SDIV        Rd, Rn, Rm
\ **
\ ** SEV
\ **
\ ** SMLAL       Rdlo, Rdhi, Rn, Rm
\ ** SMULL       Rdlo, Rdhi, Rn, Rm
\ **
\ ** SSAT        Rd, # <imm5> Rn <shift>
\ **
\ ** STC         <copro> CRd, [ Rn, # +/-<imm8> ]
\ ** STC         <copro> CRd, [ Rn, # +/-<imm8> ] !
\ ** STC         <copro> CRd, [ Rn ], # +/-<imm8>
\ ** STC         <copro> CRd, [ Rn ], {} <option>
\ ** STCL        <copro> CRd, [ Rn, # +/-<imm8> ]
\ ** STCL        <copro> CRd, [ Rn, # +/-<imm8> ] !
\ ** STCL        <copro> CRd, [ Rn ], # +/-<imm8>
\ ** STCL        <copro> CRd, [ Rn ], {} <option>
\ ** STC2        <copro> CRd, [ Rn, # +/-<imm8> ]
\ ** STC2        <copro> CRd, [ Rn, # +/-<imm8> ] !
\ ** STC2        <copro> CRd, [ Rn ], # +/-<imm8>
\ ** STC2        <copro> CRd, [ Rn ], {} <option>
\ ** STC2L       <copro> CRd, [ Rn, # +/-<imm8> ]
\ ** STC2L       <copro> CRd, [ Rn, # +/-<imm8> ] !
\ ** STC2L       <copro> CRd, [ Rn ], # +/-<imm8>
\ ** STC2L       <copro> CRd, [ Rn ], {} <option>
\ **
\ ** STM         Rn, [!] { ra, rb ... rn }
\ ** STMIA       Rn, [!] { ra, rb ... rn }
\ ** STMEA       Rn, [!] { ra, rb ... rn }
\ ** STMDB       Rn, [!] { ra, rb ... rn }
\ ** STMFD       Rn, [!] { ra, rb ... rn }

\ *E STR         Rt, [ Rn, # <imm5*4> ]
\ ** STR         Rt, [ SP, # <imm8*4> ]
\ ** STR         Rt, [ Rn, # <imm12> ]
\ ** STR         Rt, [ Rn, # -<imm8> ]
\ ** STR         Rt, [ Rn ], # +/-<imm8>
\ ** STR         Rt, [ Rn, # +/-<imm8> ] !
\ ** STR         Rt, [ Rn ++ Rm ]
\ ** STR         Rt, [ Rn ++ Rm, LSL # <imm2> ]
\ **
\ ** STRB        Rt, [ Rn, # <imm5> ]
\ ** STRB        Rt, [ Rn, # <imm12> ]
\ ** STRB        Rt, [ Rn, # -<imm8> ]
\ ** STRB        Rt, [ Rn ], # +/-<imm8>
\ ** STRB        Rt, [ Rn, # +/-<imm8> ] !
\ ** STRB        Rt, [ Rn ++ Rm ]
\ ** STRB        Rt, [ Rn ++ Rm, LSL # <imm2> ]
\ **
\ ** STRBT       Rt, [ Rn, # <imm8> ]
\ **
\ ** STRD        Rt, Rt2, [ Rn, # -<imm8> ]
\ ** STRD        Rt, Rt2, [ Rn ], # +/-<imm8>
\ ** STRD        Rt, Rt2, [ Rn, # +/-<imm8> ] !
\ **
\ ** STREX       Rd, Rt, [ Rn, # +<imm8> ]
\ ** STREXB      Rd, Rt, [ Rn ]
\ ** STREXH      Rd, Rt, [ Rn ]
\ **
\ ** STRH        Rt, [ Rn, # <imm5*2> ]
\ ** STRH        Rt, [ Rn, # <imm12> ]
\ ** STRH        Rt, [ Rn, # -<imm8> ]
\ ** STRH        Rt, [ Rn ], # +/-<imm8>
\ ** STRH        Rt, [ Rn, # +/-<imm8> ] !
\ ** STRH        Rt, [ Rn ++ Rm ]
\ ** STRH        Rt, [ Rn ++ Rm, LSL # <imm2> ]
\ **
\ ** STRHT       Rt, [ Rn, # +<imm8> ]
\ **
\ ** STRT        Rt, [ Rn, # <imm8> ]

\ *E SUB .s      Rd, Rn, # <imm3>
\ ** SUB .s      Rd, Rd, # <imm8>
\ ** SUB [.s]    Rd, Rn, # <const>
\ ** SUB [.s]    Rd, Rn, # <imm12>
\ ** SUB .s      Rd, Rn, Rm
\ ** SUB [.s]    Rd, Rn, Rm <shift>
\ ** SUB         SP, SP, # <imm7*4>
\ ** SUB [.s]    Rd, SP, # <const>
\ ** SUB [.s]    Rd, SP, # <imm12>
\ ** SUB         Rd, SP, Rm <shift>
\ **
\ ** SVC         # <imm8>
\ **
\ ** SXTB        Rd, Rm
\ ** SXTB        Rd, Rm, <rotation>  ( 0/8/16/24 bits )
\ ** SXTH        Rd, Rm
\ ** SXTH        Rd, Rm, <rotation>  ( 0/8/16/24 bits )
\ **
\ ** TBB         [ Rn, Rm ]
\ ** TBH         [ Rn, Rm ]

\ *E TEQ         Rn, # <const>
\ ** TEQ         Rn, Rm <shift>
\ **
\ ** TST         Rn, # <const>
\ ** TST         Rn, Rm
\ ** TST         Rn, Rm <shift>
\ **
\ ** UBFX        Rd, Rn, # <lsb> # <width>
\ **
\ ** UDIV        Rd, Rn, Rm
\ **
\ ** UMLAL       Rdlo, Rdhi, Rn, Rm
\ ** UMULL       Rdlo, Rdhi, Rn, Rm
\ **
\ ** USAT        Rd, # <imm5> Rn <shift>
\ **
\ ** UXTB        Rd, Rm
\ ** UXTB        Rd, Rm, <rotation>  ( 0/8/16/24 bits )
\ ** UXTH        Rd, Rm
\ ** UXTH        Rd, Rm, <rotation>  ( 0/8/16/24 bits )
\ **
\ ** WFE
\ ** WFI
\ ** YIELD


\ ==============
\ *N Integer DSP
\ ==============
\ *P These instructions are integer DSP instructions added to
\ ** Cortex-M4.

\ *E PKHBT       Rd, Rn, Rm
\ ** PKHBT       Rd, Rn, Rm .lsl # <imm5>
\ ** PKHTB       Rd, Rn, Rm
\ ** PKHTB       Rd, Rn, Rm .asr # <imm5>

\ *E QADD        Rd, Rn, Rm
\ ** QADD16      Rd, Rn, Rm
\ ** QADD8       Rd, Rn, Rm
\ ** QASX        Rd, Rn, Rm
\ ** QDADD       Rd, Rn, Rm
\ ** QDSUB       Rd, Rn, Rm
\ ** QSAX        Rd, Rn, Rm
\ ** QSUB        Rd, Rn, Rm
\ ** QSUB16      Rd, Rn, Rm
\ ** QSUB8       Rd, Rn, Rm

\ *E SADD16      Rd, Rn, Rm
\ ** SADD8       Rd, Rn, Rm
\ ** SASX        Rd, Rn, Rm
\ ** SEL         Rd, Rn, Rm
\ ** SHADD16     Rd, Rn, Rm
\ ** SHADD8      Rd, Rn, Rm
\ ** SHASX       Rd, Rn, Rm
\ ** SHSAX       Rd, Rn, Rm
\ ** SHSUB16     Rd, Rn, Rm
\ ** SHSUB8      Rd, Rn, Rm

\ *E SMLABB      Rd, Rn, Rm, Ra
\ ** SMLABT      Rd, Rn, Rm, Ra
\ ** SMLATB      Rd, Rn, Rm, Ra
\ ** SMLATT      Rd, Rn, Rm, Ra
\ ** SMLAD       Rd, Rn, Rm, Ra
\ ** SMLADX      Rd, Rn, Rm, Ra
\ ** SMLALBB     Rd, Rn, Rm, Ra
\ ** SMLALBT     Rd, Rn, Rm, Ra
\ ** SMLALTB     Rd, Rn, Rm, Ra
\ ** SMLALTT     Rd, Rn, Rm, Ra
\ ** SMLALD      RdLo, RdHi, Rn, Rm
\ ** SMLALDX     RdLo, RdHi, Rn, Rm
\ ** SMLAWB      Rd, Rn, Rm, Ra
\ ** SMLAWT      Rd, Rn, Rm, Ra
\ ** SMLSD       Rd, Rn, Rm, Ra
\ ** SMLSDX      Rd, Rn, Rm, Ra
\ ** SMLSLD      RdLo, RdHi, Rn, Rm
\ ** SMLSLDX     RdLo, RdHi, Rn, Rm
\ *E SMMLA       Rd, Rn, Rm, Ra
\ ** SMMLAR      Rd, Rn, Rm, Ra
\ ** SMMLS       Rd, Rn, Rm, Ra
\ ** SMMLSR      Rd, Rn, Rm, Ra
\ ** SMMUL       Rd, Rn, Rm
\ ** SMMULR      Rd, Rn, Rm
\ ** SMUAD       Rd, Rn, Rm
\ ** SMUADX      Rd, Rn, Rm
\ ** SMULBB      Rd, Rn, Rm
\ ** SMULBT      Rd, Rn, Rm
\ ** SMULTB      Rd, Rn, Rm
\ ** SMULTT      Rd, Rn, Rm
\ ** SMULWB      Rd, Rn, Rm
\ ** SMULWT      Rd, Rn, Rm
\ ** SMUSD       Rd, Rn, Rm
\ ** SMUSDX      Rd, Rn, Rm
\ *E SSAT16      Rd, # <imm4> Rn
\ ** SSAX        Rd, Rn, Rm
\ ** SSUB16      Rd, Rn, Rm
\ ** SSUB8       Rd, Rn, Rm
\ ** SXTAB       Rd, Rn, Rm
\ ** SXTAB       Rd, Rn, Rm, <rotation>  ( 0/8/16/24 bits )
\ ** SXTAB16     Rd, Rn, Rm
\ ** SXTAB16     Rd, Rn, Rm, <rotation>  ( 0/8/16/24 bits )
\ ** SXTAH       Rd, Rn, Rm
\ ** SXTAH       Rd, Rn, Rm, <rotation>  ( 0/8/16/24 bits )
\ ** SXTB16      Rd, Rm
\ ** SXTB16      Rd, Rm, <rotation>  ( 0/8/16/24 bits )

\ *E UADD16      Rd, Rn, Rm
\ ** UADD8       Rd, Rn, Rm
\ ** UASX        Rd, Rn, Rm
\ ** UHADD16     Rd, Rn, Rm
\ ** UHADD8      Rd, Rn, Rm
\ ** UHASX       Rd, Rn, Rm
\ ** UHSAX       Rd, Rn, Rm
\ ** UHSUB16     Rd, Rn, Rm
\ ** UHSUB8      Rd, Rn, Rm
\ ** UMAAL       RdLo, RdHi, Rn, Rm
\ *E UQADD16     Rd, Rn, Rm
\ ** UQADD8      Rd, Rn, Rm
\ ** UQASX       Rd, Rn, Rm
\ ** UQSAX       Rd, Rn, Rm
\ ** UQSUB16     Rd, Rn, Rm
\ ** UQSUB8      Rd, Rn, Rm
\ *E USAD8       Rd, Rn, Rm
\ ** USADA8      Rd, Rn, Rm, Ra
\ ** USAT16      Rd, # <imm4> Rn
\ ** USAX        Rd, Rn, Rm
\ ** USUB16      Rd, Rn, Rm
\ ** USUB8       Rd, Rn, Rm
\ *E UXTAB       Rd, Rn, Rm
\ ** UXTAB       Rd, Rn, Rm, <rotation>  ( 0/8/16/24 bits )
\ ** UXTAB16     Rd, Rn, Rm
\ ** UXTAB16     Rd, Rn, Rm, <rotation>  ( 0/8/16/24 bits )
\ ** UXTAH       Rd, Rn, Rm
\ ** UXTAH       Rd, Rn, Rm, <rotation>  ( 0/8/16/24 bits )
\ ** UXTB16      Rd, Rm
\ ** UXTB16      Rd, Rm, <rotation>  ( 0/8/16/24 bits )


\ ==============
\ *N Floating point for Cortex-Mx and ARM32
\ ==============
\ *P These are single-precision floating point instructions added
\ ** for the Cortex-M4F instruction set upwards and for the ARM32
\ ** instruction set. Double precision instructions are available
\ ** for systems that support them.

\ *E VABS .f32   Sd, Sm
\ ** VABS .f64   Dd, Dm
\ ** VADD .f32   Sd, Sn, Sm
\ ** VADD .f64   Dd, Dn, Dm
\ ** VCMP .f32   Sd, Sm
\ ** VCMP .f32   Sd, # 0
\ ** VCMP .f64   Dd, Dm
\ ** VCMP .f64   Dd, # 0
\ ** VCMPE .f32  Sd, Sm
\ ** VCMPE .f32  Sd, # 0
\ ** VCMPE .f64  Dd, Dm
\ ** VCMPE .f64  Dd, # 0

\ *E VCVT .si32<f32     Sd, Sm
\ ** VCVT .ui32<f32     Sd, Sm
\ ** VCVT .f32<si32     Sd, Sm
\ ** VCVT .f32<ui32     Sd, Sm
\ ** VCVT .si32<f64     Sd, Dm
\ ** VCVT .ui32<f64     Sd, Dm
\ ** VCVT .f64<si32     Dd, Sm
\ ** VCVT .f64<ui32     Dd, Sm
\ ** VCVT .xs32<f32     Sd, Sd, # fbits
\ ** VCVT .xu32<f32     Sd, Sd, # fbits
\ ** VCVT .xs16<f32     Sd, Sd, # fbits
\ ** VCVT .xu16<f32     Sd, Sd, # fbits
\ ** VCVT .xs32<f64     Dd, Dd, # fbits
\ ** VCVT .xu32<f64     Dd, Dd, # fbits
\ ** VCVT .xs16<f64     Dd, Dd, # fbits
\ ** VCVT .xu16<f64     Dd, Dd, # fbits
\ ** VCVT .f32<xs32     Sd, Sd, # fbits
\ ** VCVT .f32<xu32     Sd, Sd, # fbits
\ ** VCVT .f32<xs16     Sd, Sd, # fbits
\ ** VCVT .f32<xu16     Sd, Sd, # fbits
\ ** VCVT .f64<xs32     Dd, Dd, # fbits
\ ** VCVT .f64<xu32     Dd, Dd, # fbits
\ ** VCVT .f64<xs16     Dd, Dd, # fbits
\ ** VCVT .f64<xu16     Dd, Dd, # fbits
\ ** VCVT .f32<f16      Qd, Dm             \ ASIMD
\ ** VCVT .f16<f32      Dd, Qm             \ ASIMD
\ ** VCVT .f64<f32      Dd, Sm
\ ** VCVT .f32<f64      Sd, Dm

\ *E VCVTR .si32<f32    Sd, Sm
\ ** VCVTR .ui32<f32    Sd, Sm
\ ** VCVTR .si32<f64    Sd, Dm
\ ** VCVTR .ui32<f64    Sd, Dm
\ ** VCVTB .f32<f16     Sd, Sm
\ ** VCVTB .f16<f32     Sd, Sm
\ ** VCVTT .f32<f16     Sd, Sm
\ ** VCVTT .f16<f32     Sd, Sm

\ *E VDIV .f64          Dd, Dn, Dm
\ ** VDIV .f32          Sd, Sn, Sm
\ ** VFMA .f64          Dd, Dn, Dm
\ ** VFMA .f32          Sd, Sn, Sm
\ ** VFMS .f64          Dd, Dn, Dm
\ ** VFMS .f32          Sd, Sn, Sm
\ ** VFNMA .f64         Dd, Dn, Dm
\ ** VFNMA .f32         Sd, Sn, Sm
\ ** VFNMS .f64         Dd, Dn, Dm
\ ** VFNMS .f32         Sd, Sn, Sm
\ *E VLDMIA             Rn, { Dx-Dy }
\ ** VLDMIA             Rn ! { Dx-Dy }
\ ** VLDMDB             Rn ! { Sx-Sy }
\ ** VLDR               Dd, [ Rn, # imm ]
\ ** VLDR               Dd, label
\ ** VLDR               Sd, [ Rn, # imm ]
\ ** VLDR               Sd, label
\ *E VMLA .f64          Dd, Dn, Dm
\ ** VMLA .f32          Sd, Sn, Sm
\ ** VMLS .f64          Dd, Dn, Dm
\ ** VMLS .f32          Sd, Sn, Sm
\ ** VMOV .f64          Dd, # imm
\ ** VMOV .f32          Sd, # imm
\ ** VMOV .f64          Dd, Dm
\ ** VMOV .f32          Sd, Sm
\ ** VMOV .32           Dd [0/1] Rt
\ ** VMOV .32           Rt, Dn [0/1]
\ ** VMOV               Sn, Rt
\ ** VMOV               Rt, Sn
\ ** VMOV               Sm, Sm1, Rt, Rt2       \ Sm1=Sm+1
\ ** VMOV               Rt, Rt2, Sm, Sm1       \ Sm1=Sm+1
\ ** VMOV               Dm, Rt, Rt2
\ ** VMOV               Rt, Rt2, Dm
\ ** VMRS               Rt, FPSCR
\ ** VMSR               Rt, FPSCR
\ ** VMUL .f64          Dd, Dn, Dm
\ ** VMUL .f32          Sd, Sn, Sm
\ *E VNEG .f64          Dd, Dm
\ ** VNEG .f32          Sd, Sm
\ ** VNMLA .f64         Dd, Dn, Dm
\ ** VNMLA .f32         Sd, Sn, Sm
\ ** VNMLS .f64         Dd, Dn, Dm
\ ** VNMLS .f32         Sd, Sn, Sm
\ ** VNMUL .f64         Dd, Dn, Dm
\ ** VNMUL .f32         Sd, Sn, Sm
\ *E VPOP               { Dx-Dy }
\ ** VPOP               { Sx-Sy }
\ ** VPUSH              { Dx-Dy }
\ ** VPUSH              { Sx-Sy }
\ *E VSQRT .f64         Dd, Dm
\ ** VSQRT .f32         Sd, Sm
\ ** VSTMIA             Rn, { Dx-Dy }
\ ** VSTMIA             Rn ! { Dx-Dy }
\ ** VSTMDB             Rn ! { Sx-Sy }
\ ** VSTMIA             Rn, { Dx-Dy }
\ ** VSTR               Dd, [ Rn, # imm ]
\ ** VSTR               Dd, label
\ ** VSTR               Sd, [ Rn, # imm ]
\ ** VSTR               Sd, label
\ ** VSUB .f64          Dd, Dn, Dm
\ ** VSUB .f32          Sd, Sn, Sm

\ **********************
\ *S ARM instruction set
\ **********************
\ *P The ARM instruction set is mostly highly orthogonal. All data processing
\ ** instructions work on the contents of registers and immediate
\ ** constants only. Any data held in memory has to be loaded into a
\ ** register, manipulated, then saved back to memory using one of the
\ ** memory transfer instructions. This may appear to be restrictive,
\ ** but due to the large number of general-purpose registers available
\ ** for scratch storage, memory read/writes can be kept to a
\ ** minimum. The assembler is of the prefix variety, with the
\ ** instruction mnemonic preceding its parameters. Valid instructions
\ ** are:
\ *E B | BL <<cond>> expression
\ **
\ ** BLX expression
\ ** BLX Rm
\ **
\ ** MOV | MVN <<cond>> <<S>> Rd op2
\ ** CMN | CMP | TEQ | TST <<cond>> <<P>> Rn op2
\ ** ADC | ADD | AND | BIC | EOR | ORR | RSB | RSC | SBC | SUB <<cond>>
\ ** <<S>> Rd Rn op2
\ **
\ ** MRS <<cond>> Rd psr
\ ** MSR <<cond>> psr Rm
\ ** MSR <<cond>> psrf Rm
\ ** MSR <<cond>> psrf #expression
\ **
\ ** MUL <<cond>> <<S>> Rd Rm Rn
\ ** MLA <<cond>> <<S>> Rd Rm Rs Rn
\ **
\ ** UMULL | SMULL | UMLAL | SMLAL <<cond>> <<S>> RdLo RdHi Rm Rs
\ **
\ ** LDR | LDRB | LDRH | STR | STRB | STRH <<cond>> Rd address <<!>>
\ **
\ ** LDMFD | LDMED | LDMFA | LDMEA | LDMIA | LDMIB | LDMDA | LDMDB |
\ ** STMFD | STMED | STMFA | STMEA | STMIA | STMIB | STMDA | STMDB
\ ** <<cond>> Rn <<!>> Rlist <<^>>
\ **
\ ** SWP | SWPB <<cond>> Rd Rm [ Rn ]
\ **
\ ** SWI <<cond>> expression
\ **
\ ** CDP <<cond>> CP# operation CRd CRn CRm info
\ **
\ ** LDC | LDCL | STC | STCL <<cond>> CP# CRd address
\ **
\ ** MCR | MRC <<cond>> CP# operation Rd CRn CRm info

\ *P Two pseudo instructions *\fo{MVL} and *\fo{ADR} are also
\ ** available. *\fo{NOP} is supported as a synonym for:
\ *C   MOV  R0, R0
\ *P No switches are accepted for *\fo{NOP}.

\ ** *\table{"ARM instruction notation" 20 80 }
\ ** *\hrow{} *\col{} Parameter *\col{} Explanation
\ ** *\row{}  *\col{} <<cond>> *\col{}
\ ** Optional conditional execution code, i.e. .NE. See Control
\ ** Structures.
\ ** *\row{} *\col{} <<S>> *\col{}
\ ** Optional suffix .S to set the processor status flags.
\ ** *\row{} *\col{} <<P>> *\col{}
\ ** Optional suffix .P to  modify the PSR in 26-bit modes.
\ ** *\row{} *\col{} <<!>> *\col{}
\ ** Optional *\fo{!} enables write-back of the base register in loads and
\ ** stores.
\ ** *\row{} *\col{} <<^>> *\col{}
\ ** Optional *\fo{^} sets the status flags when loading the PC from memory
\ ** with the *\fo{LDMxx} instructions. Can also be used to force loading and
\ ** storing of user bank registers in non-user modes.
\ ** *\row{} *\col{} Rd, RdLo, RdHi, Rm, Rn, Rs *\col{}
\ ** Equates to a valid register number. See Register naming. Some
\ ** instructions, notably the multiplication set, place restrictions
\ ** on the combinations of registers allowed.
\ ** *\row{} *\col{} op2 *\col{}
\ ** Is either one of the operands produced by the ARM's barrel shifter
\ ** or an immediate constant. See Shift operations and Immediate
\ ** constants.
\ ** *\row{} *\col{} expression *\col{}
\ ** For the *\fo{B} and *\fo{BL} instructions this is a label name or an
\ ** expression evaluating to a branch address. The address is
\ ** converted to make it pc relative, allowing for the effects of
\ ** pipelining on the program counter. For the *\fo{SWI} instruction it is
\ ** the number of the SWI to be called.
\ ** *\row{} *\col{} #expression *\col{}
\ ** Evaluates to a 32-bit value. This has to be a valid Immediate
\ ** constant.
\ ** *\row{} *\col{} address *\col{}
\ ** A valid address specification. See Addressing modes.
\ ** *\row{} *\col{} Rlist *\col{}
\ ** A list of registers enclosed by braces. See Register lists.
\ ** *\row{} *\col{} psr *\col{}
\ ** Is the CPSR or SPSR register names.
\ ** *\row{} *\col{} psrf *\col{}
\ ** Is the CPSR_flag or SPSR_flag register names. Only the N, Z, C and
\ ** V flags are written into the status register. Use the forms _C _X
\ ** _S _F to indicate which sections are to be restored.
\ ** *\row{} *\col{} CP# *\col{}
\ ** The unique number of the required coprocessor.
\ ** *\row{} *\col{} CRd, CRn, CRm *\col{}
\ ** Equates to a valid coprocessor register number. See "Register
\ ** Naming" below.
\ ** *\row{} *\col{} operation *\col{}
\ ** Is evaluated to a constant.
\ ** *\row{} *\col{} info *\col{}
\ ** Is evaluated to a constant.
\ ** *\endtable{}

\ ******************
\ *S Register naming
\ ******************
\ *P There are fifteen general-purpose registers available in user
\ ** mode, plus the program counter. These are named *\fo{R0} through
\ ** *\fo{R15}. *\fo{R15} is the program counter. Coprocessor
\ ** registers are named
\ ** *\fo{CR0} through *\fo{CR15}. The Current Program Status
\ ** Register and Saved Program Status Register are named *\fo{CPSR}
\ ** and *\fo{SPSR} respectively. If transferring just the status
\ ** flags then *\fo{CPSR_flg} and *\fo{SPSR_flg} can be used.
\ ** As of v6.2 the notation
\ *C   CPSR_flag
\ *P is superceded by
\ *C CPSR _c _x _s _f
\ *P where the valid field definers are:
\ *C   _C _X _S _F _CXSF _FSXC _ALL

\ *P Standard ARM names are also available. *\fo{SP} refers to *\fo{R13}
\ ** (commonly used as a stack pointer), *\fo{LINK} refers to *\fo{R14}
\ ** (the link register), and *\fo{PC} refers to *\fo{R15} (the
\ ** program counter).

\ *P Forth register names can be used in place of the standard register
\ ** names. These are *\fo{TOS}, *\fo{LP}, *\fo{UP}, *\fo{RSP}
\ ** and *\fo{PSP}. As mentioned earlier these can be assigned to
\ ** different ARM registers.

\ *P All register names can be used with or without a trailing comma.
\ ** This makes for code that is more readable to the seasoned ARM
\ ** programmer. Character case is not important.


\ **********************
\ *S Immediate constants
\ **********************
\ *P Rather then specify the name of a register whose contents are to
\ ** be used in an operation, it is possible with many instructions to
\ ** specify a numeric value which is encoded with the opcode mnemonic
\ ** at assembly time.

\ *P When the *\fo{#} is encountered, the assembler recognises that the
\ ** following input is to be interpreted as a numeric value. The value
\ ** itself can be prefixed with the usual number base selectors such
\ ** as *\fo{#} for decimal, *\fo{$} for hexadecimal, *\fo{%} for
\ ** binary and *\fo{@} for octal:
\ *E ADD R2, R3, # $32   \ Add $32 to contents of R3
\ **                     \ and place result in R2

\ *P Note that in the UK, there may be confusion with some printers
\ ** between the hash symbol *\fo{#} and the pound symbol.

\ *P There are restrictions regarding the range of immediate constants
\ ** that can be used. As mentioned before each instruction and its
\ ** operands are encoded as a single 32-bit value on the ARM.
\ ** Obviously some of the 32 bits are going to be given over to the
\ ** instruction type, suffixes, and destination register etc. leaving
\ ** only 12-bits to represent the constant. 12 bits does not allow
\ ** many immediate constants to be used, so this is split into two
\ ** fields. One, 8-bits wide, specifies the constant while the other
\ ** 4-bit wide field specifies a value to shift the constant by (this
\ ** is actually a rotate right by the shift value times two places).
\ ** This widens the range of immediate constants that can be used, but
\ ** has the restriction that not every number in the full 32-bit range
\ ** can be used. Note that the range of negative immediate constants
\ ** that can be represented is very limited as these appear to the ARM
\ ** to be very large numbers i.e. -1 = $FFFFFFFF, and the larger a
\ ** number is the harder it is to represent using the method described
\ ** above. Judicial use of such instructions as *\fo{CMN} (compare
\ ** negative), *\fo{MVN} (move inverted data - not negated!) and *\fo{RSB}
\ ** (reverse subtract) can get around this problem.

\ *P Cortex only: If the literal pool is enabled you can load a 
\ ** register with a 32-bit immediate value using the form:
\ *E   LDR   Rx, @= imm32
\ *P e.g.
\ *E   LDR   R4, @= $12345678
\ *P You must flush the literal pool yourself using *\fo{FLUSHLITPOOL ( -- )
\ ** after *\fo{END-CODE} or *\fo{ENDPROC}.


\ *******************
\ *S Shift operations
\ *******************
\ *P Most data processing instructions allow operand two (the second
\ ** source operand) to be specified as a shifted register. Here the
\ ** contents of the register can be shifted at run-time by either a
\ ** fixed amount or by the contents of another register. This can be
\ ** done with one of the ARM's shift instructions, e.g.
\ *E ADD R0, R2, R7 LSL # 4   \ R7 logically shifted left by
\ **                          \ 4 places
\ ** BIC R2, R4, R7 ASR R6    \ R7 arithmetically shifted
\ **                          \ right by the contents of R6
\ *P Note that the contents of the register being shifted are not
\ ** changed by the shift. The shifted value is only used during the
\ ** instruction to calculate the new value to be stored in the
\ ** destination register.

\ *P Note also that a shift by zero bits causes no change in the
\ ** carry flag!

\ *P Shift operations supported by the ARM are:
\ ** *\table{"ARM shift instructions" 30 70 }
\ ** *\hrow{} *\col{} Instruction *\col{} Purpose
\ ** *\row{} *\col{} LSL # n or LSL Rn *\col{} Logical shift left
\ ** *\row{} *\col{} ASL # n or ASL Rn *\col{} Arithmetic shift
\ ** left (identical to LSL)
\ ** *\row{} *\col{} LSR # n or LSR Rn *\col{} Logical shift right
\ ** *\row{} *\col{} ASR # n or ASR Rn *\col{} Arithmetic shift right
\ ** *\row{} *\col{} ROR # n or ROR Rn *\col{} Rotate right
\ ** *\row{} *\col{} RRX *\col{} Rotate right with extend - (no
\ ** shift value or register is needed as the shift is by one place only)
\ ** *\endtable{}

\ *P Note that as with immediate constants, if the shift is by a fixed
\ ** amount it should be preceded by the *\fo{#} symbol to inform the
\ ** assembler that it is not dealing with a register.


\ *******************
\ *S Addressing modes
\ *******************
\ *P The ARM data processing instructions all work on the contents of
\ ** registers and immediate operands. To transfer data to and from
\ ** single registers and memory either the *\fo{LDR}, *\fo{STR},
\ ** *\fo{LDC} or *\fo{STC} instructions and their variants have
\ ** to be used. Addresses can be specified in three ways.

\ =========================
\ *N Pre-indexed addressing
\ =========================
\ *P Pre-indexed addressing allows an offset to be added to (or
\ ** subtracted from) an address held in a base register to form the
\ ** address from which data is to be transferred. The address has the
\ ** following format:
\ *C [ Rn <<offset>> ]
\ *P Where *\fo{Rn} is the base register name and the optional offset is
\ ** either:
\ *(
\ *B A simple register
\ *B An immediate constant
\ *B A shifted register
\ *)

\ *P The address expression must be terminated by a *\fo{]}. The
\ ** initial *\fo{[} is not strictly necessary but leads to code
\ ** that is more readable for experienced ARM programmers.

\ *P A simple or shifted register offset needs to be prefixed with
\ ** *\fo{++} or *\fo{--} indicating whether the contents of the
\ ** register should be added to or subtracted from the base
\ ** register. Immediate constants do not use the 8/4-bit field
\ ** format but rather range from -4095 to 4095. Shifted registers
\ ** can only be shifted by a constant preceded by the *\fo{#} symbol
\ ** and not by the contents of another register.

\ *P The address calculated by combining the base and offset registers
\ ** is often useful in subsequent loads and stores, especially when a
\ ** sequence of memory locations are to be accessed. Use the *\fo{!}
\ ** operator after the closing *\fo{]} to enable the write back feature of
\ ** the ARM. This will write the calculated address back into the base
\ ** register for subsequent instructions to use.

\ ** *\table{"Pre-indexed addressing" 40 60 }
\ ** *\hrow{} *\col{} Instruction *\col{} Address
\ ** *\row{} *\col{} LDR Rd, [ Rn ] *\col{}
\ ** Load from Rn. Treated as LDR Rd, [ Rn, # 0 ]
\ ** *\row{} *\col{} LDR Rd, [ Rn, ++ Rm ] *\col{}
\ ** Load from Rn plus Rm
\ ** *\row{} *\col{} LDR Rd, [ Rn, -- Rm ] ! *\col{}
\ ** Load from Rn minus Rm with write back
\ ** *\row{} *\col{} LDR Rd, [ Rn, ++ Rm LSL # 5 ] ! *\col{}
\ ** Load from Rn plus Rm shifted logically left five places with write
\ ** back
\ ** *\row{} *\col{} LDR Rd, [ Rn, # 20 ] *\col{}
\ ** Load from Rn plus twenty
\ ** *\row{} *\col{} LDR Rd, [ Rn, # -40 ] ! *\col{}
\ ** Load from Rn minus forty with write back
\ ** *\endtable{}

\ ==========================
\ *N Post-indexed addressing
\ ==========================
\ *P Post-indexed addresses have the following form:
\ *C   [ Rn ], <<offset>>

\ *P Post-indexed addressing adds the offset to the base register *\fo{Rn}
\ ** after the data has been transferred from the address held in the
\ ** base register. This implies that write back always occurs so it is
\ ** not necessary to specify it. It can be used however to force
\ ** non-privileged mode for the transfer cycle (same as the T suffix
\ ** on some ARM assemblers).

\ *P The offset is specified in exactly the same way as for pre-indexed
\ ** addressing. Examples of post-indexed addressing are:
\ ** *\table{"Post-indexed addressing" 40 60 }
\ ** *\hrow{} *\col{} Instruction *\col{} Address
\ ** *\row{} *\col{} LDR Rd, [ Rn ], ++ Rm *\col{}
\ ** Load from Rn then add Rm to Rn
\ ** *\row{} *\col{} LDR Rd, [ Rn ], -- Rm *\col{}
\ ** Load from Rn then subtract Rm from Rn
\ ** *\row{} *\col{} LDR Rd, [ Rn ], ++ Rm LSL # 5 *\col{}
\ ** Load from Rn then add Rm, shifted logically left five places, to
\ ** Rn
\ ** *\row{} *\col{} LDR Rd, [ Rn ], -- Rm LSL # 5 *\col{}
\ ** Load from Rn then subtract Rm, shifted logically left five places,
\ ** from Rn
\ ** *\row{} *\col{} LDR Rd, [ Rn ], # 20 *\col{}
\ ** Load from Rn then add 20 to Rn
\ ** *\row{} *\col{} LDR Rd, [ Rn ], # -40 *\col{}
\ ** Load from Rn then subtract 40 from Rn
\ ** *\endtable{}

\ =========================
\ *N PC relative addressing
\ =========================
\ *P The assembler also recognises addresses specified as either an
\ ** absolute number or an assembler label, e.g.
\ *C LDR R2, # $600		\ Load from memory location $600
\ *C LDR R2, label		\ Load from the address marked by label

\ *P Addresses specified using PC relative addressing are actually
\ ** converted into pre-indexed addresses that load from the program
\ ** counter (R15) plus or minus an immediate constant. This means that
\ ** the address of the desired memory location has to lie within +/-4096
\ ** bytes of the address of the instruction referencing it. The
\ ** assembler will take into account the effects of pipelining on the
\ ** program counter when calculating the value of the offset.

\ ================================
\ *N Byte and half word addressing
\ ================================
\ *P The instructions *\fo{LDRB} and *\fo{STRB} plus *\fo{LDRH} and
\ ** *\fo{STRH} can be used to transfer bytes or half words between
\ ** memory and registers. Byte loads and stores only utilise the
\ ** bottom 8-bits of the destination register and half words only the
\ ** bottom 16-bits. The contents of the rest of the register are
\ ** ignored on a store, and zeroed on a load from memory. Unlike word
\ ** memory transfers, byte loads and stores do not have to be aligned,
\ ** but half word transfers should be aligned to a two-byte boundary.

\ *****************
\ *S Register lists
\ *****************
\ *P Multiple registers can be loaded from and stored to memory using
\ ** the *\fo{LDM} and *\fo{STM} instructions. The format is:
\ *C LDMxx Rd, <<!>> { Ra, Rb, Rx-Ry, ... } <<^>>
\ *C STMxx Rd, <<!>> { Ra, PC, LINK, Re-Rf } <<^>>
\ ** *\fo{Rd} contains the base address where registers are
\ ** stored to or loaded from, followed by an optional *\fo{!}
\ ** to indicate that write back is required. A register list
\ ** enclosed by { and } follows. The order of the registers to
\ ** be stored is of no significance (they are always stored so
\ ** that the lowest register is at the lowermost address and
\ ** the highest register at the uppermost address) and up
\ ** to sixteen can be specified. Any register name, with
\ ** or without a trailing comma, can be used. Ranges can be specified
\ ** with a dash -. The following are all legal ways of specifying the
\ ** same list of registers:
\ *E { R0 R1 R2 R6 R12 }
\ ** { R0-R2, R6, R12 }
\ ** { R6, R12, R0-R2 }
\ *P Each register can only be specified once.
\ *P The optional final *\fo{^} sets the status flags when
\ ** loading the PC from memory with the *\fo{LDMxx} instruction.
\ ** It can also be used to force loading and storing of user bank
\ ** registers in non-user modes.

\ **************
\ ** MVL and ADR
\ **************
\ *P As indicated earlier, a common source of problems when programming
\ ** with ARM assembler is the restriction placed on the range of
\ ** immediate constants that can be used with the data processing
\ ** instructions. To get around this the pseudo instruction *\fo{MVL} can be
\ ** used to move any signed/unsigned 32-bit number into a register.
\ *C   MVL R2, # 127653
\ *P The *\fo{MVL} pseudo instruction will attempt to use a single
\ ** *\fo{MOV} or *\fo{MVN} instruction if possible, but may
\ ** generate up to four ARM instructions or two Cortex instructions
\ ** to get the value into the register.

\ *P For Cortex, you can also use
\ *C   MVL32 R2, # $12345678
\ *P to load a 32 bit value into a register. This is useful when
\ ** you are referencing the data address of a *\fo{VALUE} or
\ ** a Forth word, e.g.
\ *E $12345678 value foo
\ ** Proc MyISR
\ **   ...
\ **   mvl32  r0, # ' foo >body  \ load data address

\ *P For Cortex, branch destination addresses that are loaded
\ ** into the PC must have bit0 set to 1. To this, use:
\ *C   MVL32+1 R2, # <value>
\ *P The value can be forward referenced.

\ *P *\fo{ADR} is a pseudo instruction (macro) for ARM, but is a
\ ** real instruction for Thumb-2. The ARM *\fo{ADR} pseudo
\ ** instruction performs is used to move a 32-bit address into
\ ** a register.
\ *C   ADR label
\ *P Due to the possibility that a label might be forward referenced
\ ** and need 'fixing up' later on in the compilation, the *\fo{ADR}
\ ** pseudo instruction will always generate a *\fo{MOV} and three
\ ** *\fo{ORR} instructions.

\ ****************************************
\ *S Switching between Forth and Assembler
\ ****************************************
\ *P The compiler allows you to add in-line assembler inside colon
\ ** definitions, and to add high level phrases inside code
\ ** definitions.

\ *P Inline assembler code can be compiled inside a colon definition
\ ** using *\fo{[ASM} and *\fo{ASM]}. Use these in the form:
\ *E   : <name>
\ **     ... [ASM <assembler-code> ASM] ...
\ **   ;
\ *P High level code can be compiled inside a *\fo{CODE} definition using
\ ** *\fo{[FORTH} and *\fo{FORTH]}. Use these in the form:
\ *E   CODE <name>
\ **     ... [FORTH <high-level> FORTH] ...
\ **   END-CODE
\ *P Note that the optimiser is not flushed by the switches into
\ ** assembler. This can (and should) be achieved by placing
\ ** *\fo{[O/F]} before *\fo{[ASM} and *\fo{FORTH]}.

\ ***********
\ *S Glossary
\ ***********
\ *P This glossary details the lwords provided within the
\ ** cross-assembler to control the use of the assembler.

cr cr  ." Loading ARM Cortex assembler ..."  cr cr

defer c_asm-litpool
defer asm-flushlitpool


: 0>=   0< not  ;

only forth  also c-c  also assembler definitions


\ *************
\ Error numbers
\ *************

also c-c definitions
ErrDef err_immval               "immediate integer won't fit format"
ErrDef err_FPimmval             "immediate float won't fit 8-bit format"
ErrDef err_unaligned32          "branch or load/store to unaligned address"
ErrDef err_i12                  "12 bit offset out of range"
ErrDef err_nocc                 "condition code not set"
ErrDef err_short                "not in range -128..+127"
ErrDef err_gpr                  "not a general purpose register"
ErrDef err_cpr                  "not a coprocessor register"
ErrDef err_psr                  "processor status register"
ErrDef err_i44                  "8 bit offset out of range"
ErrDef err_off8                 "8 bit offset out of range"
ErrDef err_regno                "invalid register"
ErrDef err_cputype              "invalid for this CPU variant"
ErrDef err_ise                  "internal software error - report to MPE"
ErrDef err_uncons               "unconsumed reference"
ErrDef err_depth                "stack depth changed"
ErrDef err_medium               "error if not in range -32768..32767"
ErrDef err_fwdreftype           "invalid forward reference type - report to MPE"
ErrDef err_notenough            "not enough or too many registers defined"
ErrDef err_rgt                  "register not of right type"
ErrDef ERR_IMMSHIFT             "immediate shift count out of range"
ErrDef err_shift                "shift must be immediate"
ErrDef err_addrmode             "invalid addressing mode"
ErrDef err_notreg               "registers needed here"
ErrDef err_regorder             "invalid register range"
ErrDef err_reglist              "bad addressing mode before register list"
ErrDef err_reglistNull		"empty register list"
ErrDef err_badRegList		"invalid register list"
ErrDef err_r15                  "R15 not permitted here"
ErrDef err_26bit		"26 bit branch range exceeded"

ErrDef err_hreg			"Invalid use of high register"
ErrDef err_T2Const12		"Cannot encode constant for Thumb instruction"
ErrDef err_SP/PC		"Register must be SP/R13 or PC/R15"
ErrDef err_UNPREDICTABLE	"Result is UNPREDICTABLE according to ARM ARM"
ErrDef err_badTlit		"Invalid literal value"
ErrDef err_SPreg		"Register must be SP"
ErrDef err_badTargs		"Invalid set of operands or values"
ErrDef err_unaligned16		"Misaligned Thumb branch"
ErrDef err_TB6			"6 bit Thumb branch out of range"
ErrDef err_TB8			"8 bit Thumb branch out of range"
ErrDef err_TB11			"11 bit Thumb branch out of range"
ErrDef err_TB20			"20 bit Thumb branch out of range"
ErrDef err_TB22			"22 bit Thumb branch out of range"
ErrDef err_TB24			"24 bit Thumb branch out of range"
ErrDef err_writeback		"Misuse of writeback !"
ErrDef err_memoff		"Misaligned or out of range offset"
ErrDef err_PreIndex		"PreIndexed addressing required [ reg, # imm ]"
ErrDef err_notThumb		"Not a Thumb-1 or Thumb-2 instruction"
ErrDef err_notThumb2		"Not a Thumb-2 instruction"
ErrDef err_bit#			"Bit number must be 0..31"
ErrDef err_badReg		"Register MUST NOT be R13/SP or R15/PC"
ErrDef err_shiftop		"Invalid shift operation"
ErrDef err_shiftcnt3		"Invalid shift count, must be 0..3"
ErrDef err_ror#0		"Invalid shift count, ROR Rd, Rm, #0 = RRX"
ErrDef err_specReg		"Special register out of range 0..255"
ErrDef err_PLDoffset		"+12 or -8 bit offset out of range"
ErrDef err_satbit		"Bit number must be in range 1..32"
ErrDef err_bitwidth		"Width must be in range 1..32 bits"
ErrDef err_mov32s		"MOV/MOVT pair must not have .S"
ErrDef err_SP7			"R13 update out of range 0..508"
ErrDef err_SP12			"R13 update out of range 0..4095"
ErrDef err_BadBranchOpcode	"Unrecognised branch opcode"
ErrDef err_ldmMask1		"At least 1 register required in list"
ErrDef err_ldmMask2		"At least 2 registers required in list"
ErrDef err_NotThisCPU		"Invalid instruction or addressing mode for this CPU"
ErrDef err_ImmNot0		"Immediate value must be zero"
ErrDef err_ImmNeeded		"# and Immediate value needed"
ErrDef asm_invLabelRef		"ASM: Label reference number out of range"
ErrDef asm_invLabelDef		"ASM: Label definition number out of range"

ErrDef err_badVFPoption		"Invalid conversion option"
ErrDef err_SdSmNotSame		"Sd and Sm registers must be the same"
ErrDef err_NeedFPreg		"Floating point register Sx/Dx needed"
ErrDef err_NeedASIMDreg		"ASIMD register Dx/Qx needed"
ErrDef err_FPonly		"Instruction only has floating point forms"
ErrDef err_QregInvalid		"Qx register invalid here"
ErrDef err_VFP/SIMDsize		"invalid data size for VFP or SIMD register"
ErrDef err_VFP/SIMDindex	"invalid index for VFP or SIMD register"
ErrDef err_VFP_Sxpair		"VFP regs must be Sx and Sx+1"
previous definitions

only forth  also c-c  also assembler definitions

\ short? ?short                                    sfp 18/01/90

hex

: short?
  $-080 $07F within? ;

: ?short        \ n -- ; error if not in range -128..127
  short? 0= err_short ?a-error  ;

: medium?
  $-08000 $07FFF within? ;

: ?medium       \ n -- ; error if not in range -32768..32767
  medium? 0= err_medium ?a-error  ;

: 26bit?	\ n -- flag ; true if in range +/- 32Mb ; SFP018
  $-0200:0000 $01FF:FFFF within?  ;

: ?26bit        \ n -- ; error if not in range +/-32Mb ; SFP018
  26bit? 0= err_26bit ?a-error  ;

: alignPC+4	\ addr -- addr'
\ Add four to the address and align down to a four byte address.
  4 + -4 and  ;

: 25bitT2?	\ n -- flag ; true if in range +/- 16Mb
  $-0100:0000 $00FF:FFFF within?  ;



\ ******************************
\ ARM forward reference handling
\ ******************************

also c-c definitions
0 constant fwd_cfa                      \ cfa type
1 constant fwd_abs                      \ absolute 32 bit
2 constant fwd_r24                      \ relative 24 bit branch offset
3 constant fwd_p12                      \ PC relative 12 bit signed offset
4 constant fwd_p44                      \ PC relative 8 bit in 4+4
5 constant fwd_p08                      \ PC relative 8 bit bits 7..0
6 constant fwd_adr                      \ ADR instruction sequence
7 constant fwd_Trel8			\ Thumb1 8 bit branch
8 constant fwd_Trel11			\ Thumb1 11 bit branch
9 constant fwd_Trel22			\ Thumb1 22 bit branch, high part
#10 constant fwd_r24x			\ BLX relative 24+1 bit branch offset
#11 constant fwd_Trel24			\ 24 bit Thumb2 B/BL with J1, J2
#12 constant fwd_Trel20			\ 20 bit Thumb2 B .cc with J1, J2
#13 constant fwd_Trel1+5		\ Thumb2: 6 bit unsigned forward
#14 constant fwd_Trel8PCl		\ Thumb2: 8 bit PC offset to cell
#15 constant fwd_Trel+8PCn		\ Thumb1: 8 bit +ve PC offset to cell
\ #16 constant fwd_Trel12PCl		\ Thumb2: 12 bit PC offset to cell
#16 constant fwd_Trel12PCb		\ Thumb2: 12 bit PC offset to cell
#17 constant fwd_mov32			\ Thumb2: MOV/MOVT pair
#18 constant fwd_mov32+1		\ Thumb2: MOV/MOVT pair with bit 0 set
#19 constant fwd_absT2+1		\ Thumb2: absolute value with bit 0 set
#20 constant fwd_T2adr			\ Thumb2: 32 bit ADR, ADD/SUB rd, PC, # x
previous definitions

: forward-set   \ addr mode -- addr
  >r  cHERE(T) OVER cell+ !(s)          \ put address in packet ; SFP017
  r> OVER 2 cells+ !(s)  ;              \ put mode in packet

: ?Aligned32b	\ addr --
\ Error if addr is not a 32 bit aligned address.
  3 and err_unaligned32 ?a-error	\ check alignment
;

: ?Aligned16b	\ addr --
\ Error if addr is not a 16 bit aligned address.
  1 and err_unaligned16 ?a-error	\ check alignment
;

: ?unpred	\ flag -- ; true=error
  err_UNPREDICTABLE ?a-error  ;

: NotFwdRef?	\ imm/label -- flag
\ Return true if the reference is not a forward reference.
  dup if
    dup ref1 @(m) =
    if  drop false exit  endif
    dup ref2 @(m) =
    if  drop false exit  endif
  endif
  drop true
;

: FwdRef?	\ imm/label -- flag
\ Return true if the reference is a forward reference.
  NotFwdRef? 0=  ;


\ ========================
\ 32 bit absolute, e.g. xt
\ ========================

: ABSOLUTE      \ addr/n -- ; 32 bit absolute data
  DUP IF
    DUP REF1 @(m) = IF
      REF1 off(m)  1 forward-set	\ mark reference type
      drop  $FFFFFFFF l,c(t)  exit	\ $FF is empty Flash
    ENDIF
    DUP REF2 @(m) = IF
      REF2 off(m)  1 forward-set
      drop  $FFFFFFFF l,c(t)  exit
    ENDIF
  ENDIF
  L,C(T)  ;

: fix_abs	\ value/dest location/orig --
\ Fix up 32 bit absolute address forward reference.
\  ."  Fixing absolute address "
  dup ?Aligned32b l!c(t)		\ check alignment
;


\ ===============================
\ 24 bit PC relative for B and BL
\ ===============================

: +REL24,       \ opcode addr/n --
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m) fwd_r24 forward-set
      drop  $00FFFFFF or l,c(t)  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m) fwd_r24 forward-set
      drop  $00FFFFFF or l,c(t)  exit
    ENDIF
  ENDIF                                 \ -- opcode addr/n
  chere(t) 8 + -                        \ branch to PC+8+offset
  dup ?aligned32b			\ check alignment
  dup ?26bit				\ check range ; SFP018
  2 rshift  $00FFFFFF and               \ long word align, trim to 24 bits
  or l,c(t)  ;

: fix_r24       \ value/dest location/orig -- ; fix up forward reference
  tuck                                  \ save location
  8 + -                                 \ form offset
  dup ?aligned32b			\ check alignment
  2 rshift  $00FFFFFF and               \ long word align, trim to 24 bits
  over l@c(t) $FF000000 and  or         \ merge with opcode
  swap l!c(t)  ;                        \ put back


\ ============================
\ 24+1 bit PC relative for BLX in ARM mode
\ ============================

: CalcRel24x	\ dest orig -- field
  8 + -	dup				\ branch to PC+8+offset
  dup 1 and err_unaligned16 ?a-error    \ check alignment
  dup ?26bit				\ check range ; SFP018
  2 rshift  $00FFFFFF and               \ long word align, trim to 24 bits
  swap 2 and #23 lshift or		\ calculate H bit 24 from offset bit 1
;

: +REL24x,	\ opcode addr/n --
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m) fwd_r24x forward-set
      drop  $01FFFFFF or l,c(t)  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m) fwd_r24x forward-set
      drop  $01FFFFFF or l,c(t)  exit
    ENDIF
  ENDIF                                 \ -- opcode dest
  chere(t) CalcRel24x or l,c(t)
;

: fix_r24x	\ dest orig -- ; fix up forward reference
  tuck CalcRel24x			\ -- orig field
  over l@c(t) $FE000000 and  or         \ merge with opcode
  swap l!c(t)				\ put back
;


\ ==================
\ 12 bit PC relative
\ ==================

: +REL12        \ opcode addr/n -- opcode'
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_p12 forward-set
      drop  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_p12 forward-set
      drop  exit
    ENDIF
  ENDIF
  chere(t) 8 + -                        \ offset is PC+8+offset
  dup >r                                \ save sign
  abs dup $0FFF u> err_i12 ?a-error     \ error if out of range
  or                                    \ merge offset with opcode
  r> 0>= $00800000 and or               \ merge U bit 23 if positive
;

: fix_p12       \ value location -- ; fix up forward reference
  tuck                                  \ save location
  8 + -                                 \ form offset
  dup >r                                \ save sign
  abs dup $0FFF u> err_i12 ?a-error     \ error if out of range
  r> 0>= $00800000 and or               \ merge in U bit 23 if positive
  over l@c(t)  or  swap l!c(t)  ;	\ merge with opcode


\ ================
\ 4+4 bit relative
\ ================

: set_4/4       \ n opcode -- opcode'
  $FFFF:F0F0 and			\ strip previous offset
  over abs $0FF u> err_i44 ?a-error     \ error if n out of range
  over 0>= $00800000 and or             \ set U bit 23 according to sign
  swap abs                              \ make offset positive
  dup 4 lshift  $0F00 and  swap $0F and \ split into high and low nibbles
  or or  ;                              \ merge nibbles into opcode

: +REL4/4       \ opcode addr/n -- opcode'
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_p44 forward-set
      drop  exit
    ENDIF
    DUP REF2 @(m) =                   \ if forward ref #2
      IF  REF2 off(m)  fwd_p44 forward-set
          drop  exit
      ENDIF
  ENDIF
  chere(t) 8 + -                        \ offset is PC+8+offset
  swap set_4/4
;

: fix_4/4       \ value location -- ; fix up forward reference
  tuck                                  \ save location
  8 + -                                 \ form offset
  over l@c(t) set_4/4 swap l!c(t)  ;    \ merge with opcode


\ ==============
\ 8 bit relative
\ ==============

: set_8         \ n opcode -- opcode'
  $FFFF:FF00 and			\ strip previous offset
  over ?aligned32b			\ check alignment
  over abs $03FF u> err_off8 ?a-error   \ error if n out of range
  over 0>= $00800000 and or             \ set U bit 23 according to sign
  swap abs                              \ make offset positive
  2 rshift $0FF and or  ;               \ mask and merge with opcode

: +REL8         \ opcode addr/n -- opcode'
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_p08 forward-set
      drop  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_p08 forward-set
      drop  exit
    ENDIF
  ENDIF
  chere(t) 8 + -                        \ offset is PC+8+offset
  swap set_8                            \ merge offset with opcode
;

: fix_8         \ value location -- ; fix up forward reference
  tuck                                  \ save location
  8 + -                                 \ form offset
  over l@c(t) set_8 swap l!c(t)  ;      \ merge with opcode


\ ===============
\ ADR xt literals
\ ===============

: check-adr	\ label -- label'
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_adr forward-set  drop $FFFF:FFFF
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref #2
      REF2 off(m)  fwd_adr forward-set  drop $FFFF:FFFF
    ENDIF
  ENDIF
;

: set7:0c(t)	\ value location --
\ ORs the bottom 7 bits to a target addr, clearing the previous
\ bottom 8 bits.
  tuck l@c(t) $FFFF:FF00 and		\ -- addr(t) val opc'
  swap $FF and or			\ -- addr(t) opc''
  swap l!c(t)
;

: fix_adr       \ value location --
\ Fix up forward reference from ADR pseudo instruction (ARM32 mode).
  2dup set7:0c(t)			\ bits 7..0
  over #8 rshift over $4 + set7:0c(t)	\ bits 15..8
  over #16 rshift over $8 + set7:0c(t)	\ bits 23..16
  swap #24 rshift swap $C + set7:0c(t)	\ bits 31..24
;


\ ************************
\ Thumb forward references
\ ************************

: T2x32,	\ x --
\ +G Lay a 32 bit Thumb2 opcode.
  dup #16 rshift w,c(t)			\ first opcode word
  $FFFF and w,c(t)			\ second opcode word
;

: T2x32@(t)	\ addr(t) -- x32
\ +G Fetch a 32 bit opcode at a half-word address.
  dup w@c(t) #16 lshift  swap 2+ w@c(t) or
;

: T2x32!(t)	\ x32 addr(t) --
\ +G Store a 32 bit opcode at a half-word address.
  over #16 rshift over w!c(t)
  2+ w!c(t)
;

: CheckTaddr16	\ dest -- dest ; error if unaligned
  dup 1 and err_unaligned16 ?a-error	\ check alignment
;


\ ===================
\ 8 bit branch offset
\ ===================

: CheckRange8	\ offset -- offset ; error if not +/-8bit
  dup #-256 #254 within? 0= err_TB8 ?a-error
;

: CalcTrel8	\ dest orig -- field
\ calculate field for 8 bit Thumb branch
  4 + -  CheckTaddr16 CheckRange8	\ branch to PC+4+offset, check branch
  1 rshift  $00FF and			\ word align, trim to 8 bits
;

: +Trel8,	\ opcode addr/n --
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m) fwd_Trel8 forward-set
      drop  $00FF or w,c(t)  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m) fwd_Trel8 forward-set
      drop  $00FF or w,c(t)  exit
    ENDIF
  ENDIF                                 \ -- opcode dest
  chere(t) CalcTrel8  or w,c(t)
;

: fix_Trel8	\ dest orig -- ; fix up forward reference
  tuck CalcTrel8			\ save branch location, generate field
  over w@c(t) $FF00 and or swap w!c(t)	\ merge with opcode
;


\ ====================
\ 11 bit branch offset
\ ====================

: CheckRange11	\ offset -- offset ; error if not +/-11bit
  dup #-2048 #2046 within? 0= err_TB11 ?a-error
;

: CalcTrel11	\ dest orig -- field
\ calculate field for 8 bit Thumb branch
  4 + -  CheckTaddr16 CheckRange11	\ branch to PC+4+offset, check branch
  1 rshift  $07FF and			\ word align, trim to 11 bits
;

: +Trel11,	\ opcode addr/n --
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m) fwd_Trel11 forward-set
      drop  $07FF or w,c(t)  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m) fwd_Trel11 forward-set
      drop  $07FF or w,c(t)  exit
    ENDIF
  ENDIF                                 \ -- opcode dest
  chere(t) CalcTrel11  or w,c(t)
;

: update11	\ addr field --
\ Use field as the bottom 11 bits of the 16 bit opcode at addr.
  $07FF and				\ clip field to 11 bits
  over w@c(t) $F800 and or swap w!c(t)	\ merge with opcode
;

: fix_Trel11	\ dest orig -- ; fix up forward reference
  tuck CalcTrel11			\ save branch location, generate field
  update11				\ merge with opcode
;


\ ============================
\ Thumb-1 22 bit branch offset
\ ============================

variable Rel22fwd	\ -- addr ; set when forward reffed
variable Rel22off	\ -- addr ; holds calculated 22 bit field

: CheckRange22	\ offset -- offset ; error if not +/-22bit
  dup #-4194304 #4194302 within? 0= err_TB22 ?a-error
;

: CalcTrel22	\ dest orig -- field22
\ calculate field for 22 bit Thumb branch
  4 + -  CheckTaddr16 CheckRange22	\ branch to PC+4+offset, check branch
  1 rshift  $003FFFFF and		\ half-word align, trim to 22 bits
;

: +Trel22h,	\ opcode addr/n --
  Rel22fwd off				\ assume not forward referenced
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_Trel22 forward-set  Rel22fwd on
      drop  $07FF or w,c(t)  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_Trel22 forward-set  Rel22fwd on
      drop  $07FF or w,c(t)  exit
    ENDIF
  ENDIF                                 \ -- opcode dest
  chere(t) CalcTrel22 dup Rel22off !	\ calculate and save offset
  #11 rshift or w,c(t)			\ high part of field
;

: +Trel22l,	\ opcode addr/n --
  drop  Rel22fwd @ if			\ offset in Rel22off
    Rel22fwd off  $7FF
  else
    Rel22off @ $7FF and
  endif
  or w,c(t)  Rel22off off
;

: fix_Trel22	\ dest orig -- ; fix up forward reference
  tuck CalcTrel22			\ -- orig field ; save branch location, generate field
  2dup #11 rshift update11		\ high part opcode1
  swap 2+ swap				\ step to next opcode
  over w@c(t) $1800 and $0800 =		\ BLX: H=01?
  if  $FFFFFFFE and  endif		\ force bit0=0, i.e. dest bit1=0
  update11				\ low part opcode 2
;

\ ============================
\ Thumb-2 20 bit branch offset
\ ============================

: CheckRange20	\ offset -- offset ; error if not +/-20bit
  dup #-1048576 #1048574 within? 0= err_TB20 ?a-error
;

: CalcTrel20	\ dest orig -- field20
\ calculate field for 20 bit Thumb branch
  4 + -  CheckTaddr16 CheckRange20	\ branch to PC+4+offset, check branch
  1 arshift				\ half word align
  dup $7FF and 				\ imm11 ; bits 10..0   -> 10..0
  over #11 rshift $3F and #16 lshift or	\ imm6  ;      16..11  -> 21:16
  over $0002:0000 and 0<> $0000:2000 and or \ J1 ;     17      -> 13
  over $0004:0000 and 0<> $0000:0800 and or \ J2 ;     18      -> 11
  swap 0< $0400:0000 and or		\ merge S to bit26
;

$043F:2FFF constant Trel20mask		\ mask for Trel20 bits

: +Trel20,	\ opcode32 addr/n --
  DUP IF				\ check for forward references
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_Trel20 forward-set
      drop  Trel20mask or T2x32,  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_Trel20 forward-set
      drop  Trel20mask or T2x32,  exit
    ENDIF
  ENDIF                                 \ -- opcode dest
  chere(t) CalcTrel20 or T2x32,		\ form and lay opcode
;

: fix_Trel20	\ dest orig -- ; fix up forward reference
  tuck CalcTrel20			\ -- orig field ; generate field
  over T2x32@(t)
  Trel20mask invert and  or		\ merge field
  swap T2x32!(t)
;

\ ============================
\ Thumb-2 24 bit branch offset
\ ============================

cellbits #32 = [if]
: CheckRange24	\ offset -- offset ; error if not +/-22bit
  dup $FF00:0000 $00FF:FFFE within? 0= err_TB24 ?a-error
;
[then]
cellbits #64 = [if]
: CheckRange24	\ offset -- offset ; error if not +/-22bit
  dup $-100:0000 $00FF:FFFE within? 0= err_TB24 ?a-error
;
[then]

: GenJ		\ offset mask -- Jx ; J := (notI) xor S
  over and 0=  swap 0< xor
;

: CalcTrel24	\ dest orig -- field24
\ calculate field for 24 bit Thumb branch
  4 + -  CheckTaddr16 CheckRange24	\ branch to PC+4+offset, check branch
  1 arshift				\ half word align
  dup $7FF and 				\ offset imm11 ;    bits 10..0  -> 10..0
  over #11 rshift $3FF and #16 lshift or \ merge imm10 ;     bits 20..11 -> 25..16
  over $0020:0000 GenJ $0000:2000 and or \ merge J1/bit13 ; I1=bit21    -> 13
  over $0040:0000 GenJ $0000:0800 and or \ merge J2/bit11 ; I2=bit22    -> 11
  swap 0< $0400:0000 and or		\ merge S to bit26
;

: offT2rel25	\ opcode -- offset ; disassembler support
\ Form a 25 bit offset from the current instruction.
\ Page A6-40 encoding T4, branches with a 25 bit offset.
\ Page A6-49 encoding T1, BL
\ The last bit of 25 is an implicit 0.
\   Offset=S:I1:I2:imm10:imm11:'0'.
\   I1=not( J1 eor S )
\   I2=not( J2 eor S )
  dup #26 rshift 1 and >r			    \ save sign bit
  dup $07FF and 1 lshift			    \ bits 10:0  -> 11:1   imm11
  over #16 rshift $03FF and #12 lshift or	    \ bits 25:16 -> 21:12  imm10
  over #11 rshift 1 and r@ xor 1 xor #22 lshift or  \ bit  11    -> 22     I2 -> J2
  swap #13 rshift 1 and r@ xor 1 xor #23 lshift or  \ bit  13    -> 23     I1 -> J1
  r> #24 lshift negate or			    \ bit  26    -> 24     S extended
;

$07FF:2FFF constant Trel24mask		\ mask for Trel24 bits

: +Trel24,	\ opcode32 addr/n --
  DUP IF				\ check for forward references
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_Trel24 forward-set
      drop  Trel24mask or T2x32,  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_Trel24 forward-set
      drop  Trel24mask or T2x32,  exit
    ENDIF
  ENDIF                                 \ -- opcode dest
  chere(t) CalcTrel24 or T2x32,		\ form and lay opcode
;

: fix_Trel24	\ dest orig -- ; fix up forward reference
  tuck CalcTrel24			\ -- orig field ; generate field
  over T2x32@(t)
  Trel24mask invert and  or		\ merge field
  swap T2x32!(t)
;


\ ===================================
\ Thumb2 6 bit forward - CBNZ and CBZ
\ ===================================

: CheckRange1+5	\ offset -- offset
  dup $7E u> err_TB6 ?a-error
;

: CalcTrel1+5	\ dest orig -- field1+5
\ calculate field for 6 bit CBxx Thumb2 branch
  4 + -  CheckTaddr16 CheckRange1+5	\ branch to PC+4+offset, check branch
  1 rshift				\ half word align
  dup $1F and 3 lshift			\ 5 bit field
  swap $20 and 4 lshift or		\ 1 bit field
;

$02F8 constant Trel1+5mask		\ mask for Trel1+5 bits

: +Trel1+5,	\ opcode16 addr/n --
  DUP IF				\ check for forward references
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_Trel1+5 forward-set
      drop  Trel1+5mask or w,c(t)  exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_Trel1+5 forward-set
      drop  Trel1+5mask or w,c(t)  exit
    ENDIF
  ENDIF                                 \ -- opcode dest
  chere(t) CalcTrel1+5 or w,c(t)	\ form and lay opcode
;

: fix_Trel1+5	\ dest orig -- ; fix up forward reference
  tuck CalcTrel1+5			\ -- orig field ; generate field
  over w@c(t)
  Trel1+5mask invert and  or		\ merge field
  swap w!c(t)
;


\ ==============================================
\ Thumb2 8 bit PC offset to cell - 32 bit opcode
\ ==============================================

: +Trel8PCl	\ opcode addr/n -- opcode'
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_Trel8PCl forward-set
      drop  $0FF or  exit		\ Flash needs 1 bits
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_Trel8PCl forward-set
      drop  $0FF or  exit		\ Flash needs 1 bits
    ENDIF
  ENDIF
  chere(t) alignPC+4 -			\ offset is PC+4+offset
  swap set_8				\ merge offset with opcode
;

: fix_Trel8PCl         \ value location -- ; fix up forward reference
  tuck  alignPC+4 -			\ save location and form offset
  over T2x32@(t) set_8 swap T2x32!(t)	\ merge with opcode
;


\ ================================================
\ Thumb2 12 bit PC offsets to cell - 32 bit opcode
\ ================================================

: set_12b	\ offset opcode -- opcode'
  $FFFF:F000 and			\ strip previous offset
  over abs $0FFF u> err_i12 ?a-error	\ error if n out of range
  over 0>= $0080:0000 and or		\ set U bit 23 according to sign
  swap abs                              \ make offset positive
  $0FFF and or  ;			\ mask and merge with opcode

: +Trel12PCb	\ opcode addr/n -- opcode'
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_Trel12PCb forward-set
      drop  $0FFF or  exit		\ Flash needs 1 bits
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_Trel12PCb forward-set
      drop  $0FFF or  exit		\ Flash needs 1 bits
    ENDIF
  ENDIF
  chere(t) alignPC+4 -			\ dest is PC+4+offset
  swap set_12b				\ merge offset with opcode
;

: fix_Trel12PCb         \ value location -- ; fix up forward reference
  tuck  alignPC+4 -			\ save location and form offset
  over T2x32@(t) set_12b swap T2x32!(t)	\ merge with opcode
;


\ ===========================
\ Thumb2 ADR label references - 32 bit opcodes, offset in Lit138 format
\ ===========================

$0400:70FF constant Lit138mask
Lit138mask invert constant ~Lit138mask

: enc12>138	\ enc -- mask
\ +G For a 32 bit T2 instruction, generate the 1:3:8 32 bit
\ +* mask from a T2 12 bit immediate encoding.
  dup $0FF and				\ low 8 bits: 7..0      -> 7:0
  over $700 and 4 lshift or		\ middle 3 bits, 10..8  -> 14:12
  swap $800 and #15 lshift or		\ top bit, 11           -> 26
;

: set_T2adr	\ offset opcode -- opcode'
  ~Lit138mask and swap			\ -- opc off ; strip previous offset
  dup abs $0FFF u> err_i12 ?a-error	\ error if n out of range
  over 0< $00A0:0000 and or		\ set SUB opcode if negative
  abs enc12>138 or			\ make offset positive and merge
;

: +T2adr	\ opcode addr/n -- opcode'
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_T2adr forward-set
      drop  $0FFF or  exit		\ Flash needs 1 bits
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_T2adr forward-set
      drop  $0FFF or  exit		\ Flash needs 1 bits
    ENDIF
  ENDIF
  chere(t) alignPC+4 -			\ dest is PC+4+offset
  swap set_T2adr			\ merge offset with opcode
;

: fix_T2adr	\ value location -- ; fix up forward reference
  tuck  alignPC+4 -			\ save location and form offset
  over T2x32@(t) set_T2adr swap T2x32!(t)	\ merge with opcode
;


\ ==================================================
\ Thumb1 8 bit +ve PC offset to cell - 16 bit opcode
\ ==================================================

: set_+8n	\ n opcode -- opcode'
  $FF00 and				\ strip previous offset
  over ?aligned32b			\ check alignment
  over $03FF u> err_off8 ?a-error	\ error if n out of range
  swap 2 rshift $0FF and or  ;		\ mask and merge with opcode

: +Trel+8PCn	\ opcode addr/n -- opcode'
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_Trel+8PCn forward-set
      drop  $0FF or exit		\ Flash needs 1 bits
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref # 2
      REF2 off(m)  fwd_Trel+8PCn forward-set
      drop  $0FF or  exit		\ Flash needs 1 bits
    ENDIF
  ENDIF
  chere(t) alignPC+4 -			\ offset is PC+4+offset
  swap set_+8n 				\ merge offset with opcode
;

: fix_Trel+8PCn	\ value location -- ; fix up forward reference
  tuck alignPC+4 -			\ save location and form offset
  over w@c(t) set_+8n swap w!c(t)	\ merge with opcode
;


\ ====================
\ Thumb2 MOV/MOVT pair
\ ====================
\ MVL/ADR 32 bit literal as
\   MOV Rd, # immlo
\   MOVT Rd, # immhi

$040F:70FF constant Imm16mask	\ -- x
\ Bit mask for 16 bit immediates.
Imm16mask invert constant ~Imm16mask
\ Inverted bit mask for 16 bit immediates.

: encImm16	\ n -- xxxx
\ +G Encode n as a 16 bit literal to be ORred into a 32 bit opcode.
  dup $FF and				\ -- n x'  ;     7:0 -> 7:0
  over $0700 and 4 lshift or		\ -- n x'' ;    10:8 -> 14:12
  over $0800 and #15 lshift or		\ -- n x''' ;     11 -> 26
  swap $F000 and 4 lshift or		\ -- n x'''' ; 15:12 -> 19:16
;

: decodeImm16	\ opcode -- Imm16
  dup $FF and				\ 7:0	-> 7:0
  over $7000 and 4 rshift or		\ 14:12	-> 10:8
  over $0400:0000 and #15 rshift or	\ 26	-> 11
  swap $000F:0000 and 4 rshift or	\ 19:16	-> 15:12
;

: check-mov32	\ label -- label' flag
\ Check if label is forward referenced, return flag=true
\ for a forward reference.
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_mov32 forward-set
      drop $FFFF:FFFF  -1 exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref #2
      REF2 off(m)  fwd_mov32 forward-set
      drop $FFFF:FFFF  -1 exit
    ENDIF
  ENDIF
  0
;

: fix_mov16	\ value location --
  dup T2x32@(t) ~Imm16mask and
  rot encImm16 Imm16mask and or
  swap T2x32!(t)
;

: fix_mov32	\ value location -- ; fix up forward reference
  over $FFFF and over fix_mov16			\ low part in MOV
  swap #16 rshift $FFFF and  swap 4+ fix_mov16	\ high part in MOVT
;

: check-mov32+1	\ label -- label' flag
\ Check if label is forward referenced, return flag=true
\ for a forward reference.
  DUP IF
    DUP REF1 @(m) = IF			\ if forward ref #1
      REF1 off(m)  fwd_mov32+1 forward-set
      drop $FFFF:FFFF  -1 exit
    ENDIF
    DUP REF2 @(m) = IF			\ if forward ref #2
      REF2 off(m)  fwd_mov32+1 forward-set
      drop $FFFF:FFFF  -1 exit
    ENDIF
  ENDIF
  $0001 or  0
;

: fix_mov32+1	\ value location -- ; fix up forward reference
  swap $0001 or swap fix_mov32
;

: absT2+1	\ addr/n -- ; 32 bit absolute data
\ Call destinations need bit 0 set to stay in Thumb mode.
  DUP IF
    DUP REF1 @(m) = IF
      REF1 off(m)  fwd_absT2+1 forward-set	\ mark reference type
      drop  $FFFFFFFF l,c(t)  exit	\ $FF is empty Flash
    ENDIF
    DUP REF2 @(m) = IF
      REF2 off(m)  fwd_absT2+1 forward-set
      drop  $FFFFFFFF l,c(t)  exit
    ENDIF
  ENDIF
  $0001 or L,C(T)  ;

: fix_absT2+1	\ value/dest location/orig --
\ Fix up 32 bit absolute address forward reference.
\  ."  Fixing absolute address "
  swap $0001 or swap			\ force bit 0
  dup ?Aligned32b l!c(t)		\ check alignment
;


\ ==================
\ Forward resolution
\ ==================

((
: OFFSET        \ x "<spaces>name" -- ; Exec: n -- n+x*4
-1 offset status+

See REF! in Xdtc41-3.fth
Forward reference packet:
  cell  link to next fwd ref to same symbol
  cell  address of forward ref, CHERE(T) by default
  cell  forward reference mode, 0=CFA, 1=LABEL, 2+ = CPU dependent
))
((
: (resolve)	\ n field symbol -- ; resolve field in symbol with n
  status+ mark-resolved			\ -- n field ; mark as resolved
  2dup !(s)  cell+                      \ -- n fwd-link ; set n as value of field, step to forward link
  DUP @(s)				\ -- n fwd-link fwd-pkt
  swap off(s)				\ -- n fwd-pkt ; zero link chain pointer
  SWAP >R				\ -- fwd-pkt ; R: -- n
  BEGIN
    ?DUP
   WHILE				\ -- fwd-pkt ; R: -- n
    DUP @(s)				\ -- fwd-pkt next-pkt
    SWAP cell+				\ -- next-pkt fwd-pkt'
    DUP @(s)				\ -- next-pkt fwd-pkt' fwd-loc
    R@					\ -- next-pkt fwd-pkt' fwd-loc n
    SWAP				\ -- next-pkt fwd-pkt' n fwd-loc
    ROT cell+ @(s)			\ -- next-pkt n fwd-loc mode
    CASE	\ -- label-addr reference-location mode
      fwd_cfa OF  fix_r24  ENDOF	\ CFA/xt
      fwd_abs OF  L!C(T)  ENDOF         \ absolute
\ ARM
      fwd_r24 OF  fix_r24  ENDOF	\ BL label
      fwd_p12 OF  fix_p12  ENDOF	\ +/- 12 bit offset
      fwd_p44 OF  fix_4/4  ENDOF	\ +/- 8 bit offset
      fwd_p08 OF  fix_8  ENDOF		\ +/- 8 bit offset
      fwd_adr OF  fix_adr  ENDOF	\ ADR ; MRP002
      fwd_r24x OF  fix_r24x  ENDOF	\ BLX relative 24+1 bit branch
\ Thumb 1 and 2
      fwd_Trel8 OF  fix_Trel8  ENDOF	\ Thumb1 8 bit branch
      fwd_Trel11 OF  fix_Trel11  ENDOF	\ Thumb1 11 bit branch
      fwd_Trel22 OF  fix_Trel22  ENDOF	\ Thumb1 22 bit branch
      fwd_Trel24 of  fix_Trel24  endof	\ 24 bit Thumb2 B/BL
      fwd_Trel20 of  fix_Trel20  endof  \ 20 bit Thumb2 B .cc
      fwd_Trel1+5 of fix_Trel1+5 endof	\ 6 bit Thumb2 - CBxx
      fwd_Trel8PCl OF  fix_Trel8PCl  ENDOF	\ Thumb2 8 bit offset from PC to cell
      fwd_Trel+8PCn OF  fix_Trel+8PCn  ENDOF	\ Thumb1 8 bit offset from PC to cell
\      fwd_Trel12PCl OF  fix_Trel12PCl  ENDOF	\ Thumb2 12 bit offset from PC to cell
      fwd_Trel12PCb OF  fix_Trel12PCb  ENDOF	\ Thumb2 12 bit offset from PC
      fwd_mov32 OF  fix_mov32  ENDOF		\ Thumb2 MOV/MOVT pair
      fwd_mov32+1 OF  fix_mov32+1  ENDOF	\ Thumb2 MOV/MOVT pair, bit 0 set
      fwd_absT2+1 OF  fix_absT2+1  ENDOF	\ Thumb2 absolute, bit 0 set
      fwd_T2adr OF  fix_T2adr  ENDOF		\ Thumb2 ADR instruction
        err_fwdreftype throw
    ENDCASE
  REPEAT
  R> DROP  ;
))
: (resolve)
  status+ mark-resolved
  2dup !(s)  cell+                      \ resolve & mark res.
  DUP @(s)  swap off(s)  SWAP >R
  BEGIN
    ?DUP
   WHILE
    DUP @(s)  SWAP cell+  DUP @(s) R@ SWAP ROT cell+ @(s)
    CASE   ( label-addr reference-location mode )
      fwd_cfa OF  fix_r24  ENDOF	\ CFA/xt
      fwd_abs OF  L!C(T)  ENDOF         \ absolute
\ ARM
      fwd_r24 OF  fix_r24  ENDOF	\ BL label
      fwd_p12 OF  fix_p12  ENDOF	\ +/- 12 bit offset
      fwd_p44 OF  fix_4/4  ENDOF	\ +/- 8 bit offset
      fwd_p08 OF  fix_8  ENDOF		\ +/- 8 bit offset
      fwd_adr OF  fix_adr  ENDOF	\ ADR ; MRP002
      fwd_r24x OF  fix_r24x  ENDOF	\ BLX relative 24+1 bit branch
\ Thumb 1 and 2
      fwd_Trel8 OF  fix_Trel8  ENDOF	\ Thumb1 8 bit branch
      fwd_Trel11 OF  fix_Trel11  ENDOF	\ Thumb1 11 bit branch
      fwd_Trel22 OF  fix_Trel22  ENDOF	\ Thumb1 22 bit branch
      fwd_Trel24 of  fix_Trel24  endof	\ 24 bit Thumb2 B/BL
      fwd_Trel20 of  fix_Trel20  endof  \ 20 bit Thumb2 B .cc
      fwd_Trel1+5 of fix_Trel1+5 endof	\ 6 bit Thumb2 - CBxx
      fwd_Trel8PCl OF  fix_Trel8PCl  ENDOF	\ Thumb2 8 bit offset from PC to cell
      fwd_Trel+8PCn OF  fix_Trel+8PCn  ENDOF	\ Thumb1 8 bit offset from PC to cell
\      fwd_Trel12PCl OF  fix_Trel12PCl  ENDOF	\ Thumb2 12 bit offset from PC to cell
      fwd_Trel12PCb OF  fix_Trel12PCb  ENDOF	\ Thumb2 12 bit offset from PC
      fwd_mov32 OF  fix_mov32  ENDOF		\ Thumb2 MOV/MOVT pair
      fwd_mov32+1 OF  fix_mov32+1  ENDOF	\ Thumb2 MOV/MOVT pair, bit 0 set
      fwd_absT2+1 OF  fix_absT2+1  ENDOF	\ Thumb2 absolute, bit 0 set
      fwd_T2adr OF  fix_T2adr  ENDOF		\ Thumb2 ADR instruction
        err_fwdreftype throw
    ENDCASE
  REPEAT
  R> DROP  ;


\ ******************************
\ start of assembler - variables
\ ******************************

variable LDM-first		\ first register in ( xn-xm } ; SFP010
variable LDM-len		\ number of registers in ( xn-xm } ; SFP010
VARIABLE LDM-MASK               \ bit masks for LDM/STM
variable ldm-count		\ number of set bits in LDM-MASK
variable msr_mask		\ field mask for MSR instruction ; SFP016
variable cc-set                 \ true if condition code set
variable cc-code                \ condition code
variable imm-flag               \ set by #
variable @=-flag		\ set by @=
variable shiftop                \ set by shift e.g. LSL, bit 12 non-zero
variable indexing               \ set if indexed by [
variable <itcount>		\ set by IT instruction to #ins+1, counted by A_RESET
				\ # of instructions in complete IT block
variable <itcond>		\ condition used by last IT instruction
variable <s>                    \ set by .S
variable <psr>                  \ set by .P                                             \ MRP007
variable <w>                    \ set by ! or ], (Thumb-2), normally set if <p>=0
variable <b>                    \ set for BYTE
variable <u>                    \ set for up/add
variable <p>                    \ set by ] for PRE or cleared by ], for POST
				\ N.B. Post-Indexed without writeback
				\ indicates T version of LDRx/STRx.
variable reg2                   \ true for Rx ++ Ry or Rx -- Ry
variable <s22>                  \ true for status/User in STM/LDM
variable <option?>		\ set by {} for T2 LDCx MRCx
variable <wide>			\ 16/32 instruction select, 0=default, -1=32 bit, 1=16 bit
  0 <wide> !
variable <cpsMask>		\ used by CPS instruction
  0 <cpsmask> !
variable <VFPoption>		\ VFP option, e.g. VCVT .FP>SI, .SI>FP
variable <VFPdatatype>		\ 0 = integer, nz = float
variable <VFPunsigned>		\ 0 = signed, 1 = unsigned
variable <VFPdatasize>		\ 0 = 8 bit, 1 = 16 bit, 2 = 32 bit
variable <VFPindex>		\ data 32 - 0..7, 16 - 0..3, 32 - 0..1
management-var PRO-TYPE         \ processor type
  0 pro-type !(m)

#32 constant #regs
create reg-array                \ registers held in an array
  #regs cells allot&erase
variable reg-count              \ counts number of registers used in this instruction


( reset assembler variables )
HEX

( -- )
: (A-RESET)	\ -- ; resets assembler variables
  LDM-MASK off  ldm-count off
  LDM-first off  LDM-len off		\ SFP010
  msr_mask off				\ SFP016
  cc-set off  $0E0000000 cc-code !      \ default condition code
  reg-array #regs cells erase           \ registers
  reg-count off  <wide> off
  <cpsMask> off
  <VFPoption> off			\ SFP010
  <VFPdatatype> off			\ 0 = integer, nz = float ; SFP010
  <VFPunsigned>	off			\ 0 = signed, 1 = unsigned ; SFP010
  2 <VFPdatasize> !			\ 0 = 8 bit, 1 = 16 bit, 2 = 32 bit, 3 = 64 bit ; SFP010
  <VFPindex> off			\ index = 0
  <s> off  <w> off  <b> off  <u> on     \ flags
  <psr> off  <option?> off
  <p> off  imm-flag off  shiftop off
  @=-flag off
  indexing off  reg2 off  <s22> off
  <itcount> @				\ if IT count in progress
  if  <itcount> decr  endif

  showing @ out @ 0<> and               \ prettier when debugging
  if  cr  endif
  save-pos                              \ for error checking
;
assign (a-reset) to-do a-reset

also c-c definitions
: RESET         A-RESET  ;

previous definitions

also asm-access definitions
: RESET         A-RESET  ;
previous definitions

: asm-init      \ -- ; needed by optimiser ;    MRP011
  a-reset  initfwd
;

\ *******************************************
\ Dealing with CPU instruction set variations
\ *******************************************

hex

also c-c definitions

0 constant ArmISA	\ -- 0
\ +G Use ARM 32 bit instructions.
1 constant Thumb1ISA	\ -- 1
\ +G Use Thumb1 instructions.
2 constant Thumb2ISA	\ -- 2
\ +G Use Thumb2 instructions, e.g. for Cortex.
3 constant JazelleISA	\ -- 3
\ +G Use Jazelle instructions.

( Cross compiler uses Thumb2ISA, -1 )
( ARM Linux uses ArmISA, $003F )
Thumb2ISA value TargetISA	\ -- n
\ +G Returns the current target instruction set used by the
\ +* disassembler and assembler.
-1 value ISAflags	\ -- x
\ +G Bit mask of ISA capabilities. Provided for mapping ISA
\ +* changes, e.g. Cortex A/R/M, M0, M1, M3, M4 ...
\ +* For Cortex, the bits are used as follows:
\ +(
\ +B Bit0 - M0
\ +B Bit1 - M1
\ +B Bit2 - M2
\ +B Bit3 - M3
\ +B Bit4 - M4
\ +B Bit5 - M7
\ +B ...
\ +B Bit16 - has Jazelle
\ +B ...
\ +B Bit27 - SIMD present
\ +B Bit28 - 64 bit VFP present
\ +B Bit29 - 32 bit VFP present
\ +B Bit30 - Integer DSP present
\ +B Bit31 - always 1
\ +)

\ +P For ARM32, the bits are used as follows:
\ +(
\ +B Bit0 - always 1
\ +B Bit1 - always 1
\ +B Bit2 - always 1
\ +B Bit3 - always 1
\ +B Bit4 - 1 = full ARM7TDMI
\ +B Bit5 - 1 = ArmArch5
\ +B Bit6 - 1 = ArmArch5E with DSP
\ +B ...
\ +B Bit16 - has Jazelle
\ +B ...
\ +B Bit27 - SIMD present
\ +B Bit28 - 64 bit VFP present
\ +B Bit29 - 32 bit VFP present
\ +B Bit30 - Integer DSP present
\ *B Bit31 always 0
\ +)

$001F constant ARM32flags
$003F constant ArmArch5flags
$007F constant ArmArch5Eflags

: setTargetISA	\ n mask --
\ +G Sets the current target instruction set used by the
\ +* disassembler and assembler.
  -> ISAflags -> TargetISA  ;

: ARM32		\ --
\ *G Select ARM7 mode for assembler
  ArmISA ARM32flags setTargetISA  ;

: ArmArch5	\ --
\ *G Select ARMv5 mode for assembler
  ArmISA ArmArch5flags setTargetISA  ;

: Thumb-1	\ --
\ *G Select full Thumb-1 mode for assembler and code generator.
  Thumb1ISA -1 setTargetISA  ;

: Thumb-2	\ --
\ *G Select full Thumb-2 mode for assembler and code generator.
  Thumb2ISA -1 setTargetISA  ;

: Cortex-M0	\ --
\ *G Select Cortex-M0 for assembler and code generator.
  Thumb2ISA $8000:0001 setTargetISA  ;
: Cortex-M1	\ --
\ *G Select Cortex-M1 for assembler and code generator.
  Thumb2ISA $8000:0002 setTargetISA  ;
: Cortex-M3	\ --
\ *G Select Cortex-M3 for assembler and code generator.
  Thumb2ISA $8000:0008 setTargetISA  ;
: Cortex-M4	\ --
\ *G Select Cortex-M4 (includes integer DSP) for the assembler
\ ** and code generator.
  Thumb2ISA $C000:0010 setTargetISA  ;
: Cortex-M4F	\ --
\ *G Select Cortex-M4F (includes integer DSP and single-precision
\ ** VFP) for the assembler and code generator.
  Thumb2ISA $E000:0010 setTargetISA  ;
: Cortex-M7	\ --
\ *G Select Cortex-M7 (includes integer DSP, single and double
\ ** precision VFP) for the assembler and code generator.
  Thumb2ISA $F800:0020 setTargetISA  ;

: ARM32?	\ -- flag
\ *G Return true if in 32 bit ARM mode.
  TargetISA ArmISA =  ;

: ArmArch5?	\ -- flag
\ *G Return true if in 32 bit ARMv5 mode.
  TargetISA ArmISA =			\ ARM ISA?
  ISAflags $8000:0020 and $0020 =	\ ARMv5?
  and
;

: Thumb1?	\ -- flag
\ *G Return true if in Thumb-1 mode.
  TargetISA Thumb1ISA =  ;

: Thumb2?	\ -- flag
\ *G Return true if in Thumb-2 mode.
  TargetISA Thumb2ISA =  ;

: Thumb?	\ -- flag
\ *G Return true if in either Thumb mode.
  Thumb1? Thumb2? or  ;

: Cortex-M0?	\ -- flag
\ *G Return true if the Cortex-M0 instruction set has been selected.
  $8000:0001 ISAflags =  ;
: Cortex-M1?	\ -- flag
\ *G Return true if the Cortex-M1 instruction set has been selected.
  $8000:0002 ISAflags =  ;
: Cortex-M0/M1?	\ -- flag
\ *G Return true if the Cortex M0 or M1 instruction sets have been
\ ** selected.
  ISAflags dup $8000:0001 =  swap $8000:0002 = or  ;
: Cortex-M3?	\ -- flag
\ *G Return true if the Cortex-M3 instruction set has been selected.
  $8000:0008 ISAflags =  ;
: Cortex-M4?	\ -- flag
\ *G Return true if the Cortex-M4 instruction set has been selected.
  ISAflags $C000:0010 and  $C000:0010 =  ;
: Cortex-M4F?	\ -- flag
\ *G Return true if the Cortex-M4 instruction set has been selected.
  ISAflags $E000:0010 and  $E000:0010 =  ;
: Cortex-M7?	\ -- flag
\ *G Return true if the Cortex-M7 instruction set has been selected.
  ISAflags $F800:0020 and  $F800:0020 =  ;
: IDSP?		\ -- flag
\ *G Return true if the Cortex integer DSP instructions are present.
  ISAflags $4000:0000 and 0<>  ;
: Not-M0/M1?	\ -- flag
\ *G Return true if the Cortex M0 or M1 instruction sets have not
\ ** been selected.
  ISAflags dup $8000:0001 <>  swap $8000:0002 <> and  ;
: M0/M1?	\ -- flag
\ *G Return true if Thumb-2 and Cortex M0 or M1 has been selected.
  Thumb2? Cortex-M0/M1? and  ;
: Not-M0/M1	\ --
\ +G Error if M0 and M1 selected.
  Cortex-M0/M1? err_NotThisCPU ?a-error  ;
: IDSPonly	\ --
\ +G Error if integer DSP not present.
  ISAflags $C000:0000 and $C000:0000 <> err_NotThisCPU ?a-error  ;

previous definitions

: NotThumb	\ --
\ +G Error if not in ARM mode.
  TargetISA err_notThumb ?a-error  ;

: NotARM	\ --
\ +G Error if not in Thumb mode.
  TargetISA 0= err_notThumb ?a-error  ;

: Thumb2only	\ --
\ +G Error if not Thumb-2 instruction set.
  TargetISA Thumb2ISA <> err_notThumb2 ?a-error  ;
\ ...SFP022


\ ********************
\ Processor variations
\ ********************

\ MRP005...
: set-protype   \ n -- ; -- [child]
\ set up processor type
  create  ,  does> @ pro-type !(m)  ;

: protype?      \ n -- ; -- t/f [child]
\ test if processor has ALL or more features of defined processor
  create  ,  does> @ dup pro-type @(m) and = ;

: ?protype      \ n -- ; -- [child]
\ error if defined processor does NOT have all features of processor
  create  ,  does> @ dup pro-type @(m) and <> err_cputype ?throw  ;

: must-be-cpu   \ n --
\ error if not this cpu or one with exactly the same properties
  pro-type @(m) <> err_cputype ?throw  ;

\ ***************
\ Condition codes
\ ***************

: i-code        \ code -- ; -- [child] ; process condition codes in instructions
  create
    #28 lshift ,
  does>
    @ cc-code !  cc-set on
;

: c-code        \ code -- ; -- [child] ; process condition codes in conditionals
  create
    #28 lshift ,
  does> ?prefix
    @ cc-code !  cc-set on
;

: +cond         \ opcode -- opcode' ; add condition code into the opcode
  cc-code @off or  cc-set off  ;

: ?A+cond	\ opcode -- opcode' ; add condition code if in ARM32 mode
  Thumb? 0=
  if  +cond  then
;

: +sbit20       \ opcode -- opcode' ; add S into bit 20
  <s> @ $0010:0000 and or  ;

: +psrflags     \ opcode -- opcode' ; add P into bits 12:15
  <psr> @ $0000:F000 and or  ;

also asm-access definitions
hex
\ instructions          conditionals
0 i-code .EQ            0 c-code EQ,
  0 i-code .Z            0 c-code Z,
1 i-code .NE            1 c-code NE,
  1 i-code .NZ            1 c-code NZ,
2 i-code .CS            2 c-code CS,
  2 i-code .HS            2 c-code HS,
3 i-code .CC            3 c-code CC,
  3 i-code .LO            3 c-code LO,
4 i-code .MI            4 c-code MI,
5 i-code .PL            5 c-code PL,
6 i-code .VS            6 c-code VS,
7 i-code .VC            7 c-code VC,
8 i-code .HI            8 c-code HI,
9 i-code .LS            9 c-code LS,
0A i-code .GE           0A c-code GE,
0B i-code .LT           0B c-code LT,
0C i-code .GT           0C c-code GT,
0D i-code .LE           0D c-code LE,
0E i-code .AL

: .stack        \ -- ; permit stack to be displayed during debugging
  .s  ;

: .stk		\ --
  null-op .stack  ;

: .S            \ -- ; demand condition codes be set
  <s> on  ;
: .SX		\ -- ; demand condition codes be set ; avoid name clash
  <s> on  ;

: .P            \ -- ; demand processor status flags be set (TEQ, TST, CMN, CMP only)   \ MRP007
  <psr> on ;

previous definitions

: s/it?		\ -- flag
\ +G Return true if the S bit is set or an IT instruction is active.
  <s> @ 0<>  <itcount> @ 0<> or
;


\ *********************
\ Register declarations - low byte = reg #, next = reg type
\ *********************

: r[]		\ u -- addr
\ Address of uth register in the array
  cells reg-array +
;

: reg#@		\ index -- reg#
  r[] @ $FF and  ;

: 1stReg	\ -- reg#
  0 reg#@  ;

: 2ndReg	\ -- reg#
  1 reg#@  ;

: 3rdReg	\ -- reg#
  2 reg#@  ;

: 4thReg	\ -- reg#
  3 reg#@  ;

: loReg?	\ u -- flag ; true if low register
  8 u<  ;
: hiReg?	\ u -- flag ; true if high register
  7 u>  ;

: 1loReg?	\ -- flag
\ +G Return true if the first register is a low register.
  1stReg loReg?  ;

: 2loRegs?	\ -- flag
\ +G Return true if the first two registers are low regs.
  1stReg loReg?  2ndReg loReg? and  ;

: 3loRegs?	\ -- flag
\ +G Return true if the first three registers are low regs.
  1stReg loReg?  2ndReg loReg? and  3rdReg loReg? and  ;

: (r)           \ n type -- ; set a register
  swap $0FF and or                      \ merge type and reg#
  reg-count @ r[] !  reg-count incr	\ store into array
;

\ register types
$0100 constant gpr-type		\ R0..R15
$0200 constant cpr-type		\ CR0..CR15, Coprocessor registers
$0300 constant psr-type		\ CPSR, SPSR ...
$0400 constant sfp-type		\ S0..S31
$0500 constant dfp-type		\ D0..D31
$0600 constant qfp-type		\ Q0..Q15 single word vector

: r[]type?	\ reg# type -- flag
  swap r[] @ $FF00 and =
;

: Rx?		\ reg# -- flag
  r[] @ $FF00 and gpr-type =  ;
: Sx?		\ reg# -- flag
  r[] @ $FF00 and sfp-type =  ;
: Dx?		\ reg# -- flag
  r[] @ $FF00 and dfp-type =  ;
: Qx?		\ reg# -- flag
  r[] @ $FF00 and qfp-type =  ;

\ set register types
: r             \ n -- ; declare a general purpose register
  gpr-type (r)  ;                          \ set type

: cpr           \ n -- ; declare a coprocessor register
  cpr-type (r)  ;                          \ set type

: psr           \ n -- ; declare a processor status register
  psr-type (r)   ;                         \ set type

: sfpr		\ n -- ; declare a 32 bit FP register
  sfp-type (r)  ;                          \ set type

: dfpr		\ n -- ; declare a 64 bit FP/vector register
  dfp-type (r)  ;                          \ set type

: qfpr		\ n -- ; declare a 128 bit vector register
  qfp-type (r)  ;                           \ set type


\ create named registers
: reg           \ n -- ; -- [child] ; named general register
  create  ,  does> @ gpr-type (r)  ;

: cpreg         \ n -- ; -- [child] ; named coprocessor register
  create  ,  does> @ cpr-type (r)  ;

: psreg         \ n -- ; -- [child] ; named processor status register
  create  ,  does> @ psr-type (r)  ;

: sfpreg	\ n -- ; -- [child] ; named 32 bit FP register
  create  ,  does> @ sfp-type (r)  ;

: dfpreg	\ n -- ; -- [child] ; named 64 bit FP/vector register
  create  ,  does> @ dfp-type (r)  ;

: qfpreg	\ n -- ; -- [child] ; named 128 bit FP/vector register
  create  2* ,  does> @ qfp-type (r)  ;


\ Accessing nth regs
\ rewrite ...

: 1reg-mustbe	\ -- ; error if reg count <> 1
  reg-count @ 1 <> err_notenough ?a-error
;

: 2regs-mustbe	\ -- ; error if reg count <> 2
  reg-count @ 2 <> err_notenough ?a-error
;

: 3regs-mustbe	\ -- ; error if reg count <> 3
  reg-count @ 3 <> err_notenough ?a-error
;

: 4regs-mustbe	\ -- ; error if reg count <> 4
  reg-count @ 4 <> err_notenough ?a-error
;

: (last-r)      \ type -- n ; return last register used, checking type
  reg-count decr                        \ get offset to last defined
  reg-count @ 0< err_notenough ?a-error \ check
  reg-array reg-count @ cells + @       \ get register
  dup $FF00 and rot <> err_rgt ?a-error \ must be of right type
  $0FF and  ;                           \ strip out type info

: peek-regData	\ -- x ; return last register data, no step ; SFP010
  reg-count @ 1- r[] @
;

: last-regData	\ -- x ; return last register data, step back ; SFP010
  reg-count @ 1- dup reg-count !  r[] @
;

: last-reg      \ -- n ; return last register used, and step back
  gpr-type (last-r)  ;

: peek-reg	\ -- n ; return last register used
  last-reg  reg-count incr  ;

: peek-prevReg	\ -- n ; return previous register used
  reg-count decr  last-reg  2 reg-count +!  ;

: add-reg	\ n -- ; add register to end of list
  gpr-type (r)  ;

: drop-reg	\ -- ; drop last register defined
  last-reg drop  ;

: last-cpr      \ -- n ; return last copro register used, and step back
  cpr-type (last-r)  ;

: last-psr      \ -- n ; return last register used, and step back
  psr-type (last-r)  ;

: last-sfp	\ -- n ; return last 32 bit FP used, and step back
  sfp-type (last-r)  ;

: last-dfp	\ -- n ; return last 64 bit FP used, and step back
  dfp-type (last-r)  ;
\ ... rewrite

also asm-access definitions
: r#             \ n -- ; declare a general purpose register
  gpr-type (r)  ;                          \ set type

: cr#           \ n -- ; declare a coprocessor register
  cpr-type (r)  ;                          \ set type

: sr#           \ n -- ; declare a processor status register
  psr-type (r)   ;                         \ set type

: sfp#		\ n -- ; declare a 32 bit FP register
  sfp-type (r)   ;                         \ set type

: dfp#		\ n -- ; declare a 64 bit FP/vector register
  dfp-type (r)   ;                         \ set type

: qfp#		\ n -- ; declare a 128 bit FP/vector register
  qfp-type (r)   ;                         \ set type

0 reg R0        1 reg R1        2 reg R2        3 reg R3
4 reg R4        5 reg R5        6 reg R6        7 reg R7
8 reg R8        9 reg R9        0A reg R10      0B reg R11
0C reg R12      0D reg R13      0E reg R14      0F reg R15
0D reg SP       0E reg LINK     0F reg PC
                0E reg LR

0 reg R0,       1 reg R1,       2 reg R2,       3 reg R3,
4 reg R4,       5 reg R5,       6 reg R6,       7 reg R7,
8 reg R8,       9 reg R9,       0A reg R10,     0B reg R11,
0C reg R12,     0D reg R13,     0E reg R14,     0F reg R15,
0D reg SP,      0E reg LINK,    0F reg PC,
                0E reg LR,	$0F reg APSR_nzcv

0 cpreg CR0     1 cpreg CR1     2 cpreg CR2     3 cpreg CR3
4 cpreg CR4     5 cpreg CR5     6 cpreg CR6     7 cpreg CR7
8 cpreg CR8     9 cpreg CR9     0A cpreg CR10   0B cpreg CR11
0C cpreg CR12   0D cpreg CR13   0E cpreg CR14   0F cpreg CR15

0 cpreg CR0,    1 cpreg CR1,    2 cpreg CR2,    3 cpreg CR3,
4 cpreg CR4,    5 cpreg CR5,    6 cpreg CR6,    7 cpreg CR7,
8 cpreg CR8,    9 cpreg CR9,    0A cpreg CR10,  0B cpreg CR11,
0C cpreg CR12,  0D cpreg CR13,  0E cpreg CR14,  0F cpreg CR15,

00 sfpreg S0	01 sfpreg S1	02 sfpreg S2	03 sfpreg S3
04 sfpreg S4	05 sfpreg S5	06 sfpreg S6	07 sfpreg S7
08 sfpreg S8	09 sfpreg S9	0A sfpreg S10	0B sfpreg S11
0C sfpreg S12	0D sfpreg S13	0E sfpreg S14	0F sfpreg S15
10 sfpreg S16	11 sfpreg S17	12 sfpreg S18	13 sfpreg S19
14 sfpreg S20	15 sfpreg S21	16 sfpreg S22	17 sfpreg S23
18 sfpreg S24	19 sfpreg S25	1A sfpreg S26	1B sfpreg S27
1C sfpreg S28	1D sfpreg S29	1E sfpreg S30	1F sfpreg S31

00 sfpreg S0,	01 sfpreg S1,	02 sfpreg S2,	03 sfpreg S3,
04 sfpreg S4,	05 sfpreg S5,	06 sfpreg S6,	07 sfpreg S7,
08 sfpreg S8,	09 sfpreg S9,	0A sfpreg S10,	0B sfpreg S11,
0C sfpreg S12,	0D sfpreg S13,	0E sfpreg S14,	0F sfpreg S15,
10 sfpreg S16,	11 sfpreg S17,	12 sfpreg S18,	13 sfpreg S19,
14 sfpreg S20,	15 sfpreg S21,	16 sfpreg S22,	17 sfpreg S23,
18 sfpreg S24,	19 sfpreg S25,	1A sfpreg S26,	1B sfpreg S27,
1C sfpreg S28,	1D sfpreg S29,	1E sfpreg S30,	1F sfpreg S31,

00 dfpreg D0	01 dfpreg D1	02 dfpreg D2	03 dfpreg D3
04 dfpreg D4	05 dfpreg D5	06 dfpreg D6	07 dfpreg D7
08 dfpreg D8	09 dfpreg D9	0A dfpreg D10	0B dfpreg D11
0C dfpreg D12	0D dfpreg D13	0E dfpreg D14	0F dfpreg D15
10 dfpreg D16	11 dfpreg D17	12 dfpreg D18	13 dfpreg D19
14 dfpreg D20	15 dfpreg D21	16 dfpreg D22	17 dfpreg D23
18 dfpreg D24	19 dfpreg D25	1A dfpreg D26	1B dfpreg D27
1C dfpreg D28	1D dfpreg D29	1E dfpreg D30	1F dfpreg D31

00 dfpreg D0,	01 dfpreg D1,	02 dfpreg D2,	03 dfpreg D3,
04 dfpreg D4,	05 dfpreg D5,	06 dfpreg D6,	07 dfpreg D7,
08 dfpreg D8,	09 dfpreg D9,	0A dfpreg D10,	0B dfpreg D11,
0C dfpreg D12,	0D dfpreg D13,	0E dfpreg D14,	0F dfpreg D15,
10 dfpreg D16,	11 dfpreg D17,	12 dfpreg D18,	13 dfpreg D19,
14 dfpreg D20,	15 dfpreg D21,	16 dfpreg D22,	17 dfpreg D23,
18 dfpreg D24,	19 dfpreg D25,	1A dfpreg D26,	1B dfpreg D27,
1C dfpreg D28,	1D dfpreg D29,	1E dfpreg D30,	1F dfpreg D31,

00 qfpreg Q0	01 qfpreg Q1	02 qfpreg Q2	03 qfpreg Q3
04 qfpreg Q4	05 qfpreg Q5	06 qfpreg Q6	07 qfpreg Q7
08 qfpreg Q8	09 qfpreg Q9	0A qfpreg Q10	0B qfpreg Q11
0C qfpreg Q12	0D qfpreg Q13	0E qfpreg Q14	0F qfpreg Q15

00 qfpreg Q0,	01 qfpreg Q1,	02 qfpreg Q2,	03 qfpreg Q3,
04 qfpreg Q4,	05 qfpreg Q5,	06 qfpreg Q6,	07 qfpreg Q7,
08 qfpreg Q8,	09 qfpreg Q9,	0A qfpreg Q10,	0B qfpreg Q11,
0C qfpreg Q12,	0D qfpreg Q13,	0E qfpreg Q14,	0F qfpreg Q15,

0 psreg CPSR    1 psreg SPSR    \ 2 psreg CPSR_flg  3 psreg SPSR_flg	\ SFP016

0 psreg CPSR,   1 psreg SPSR,   \ 2 psreg CPSR_flg, 3 psreg SPSR_flg,	\ SFP016

: FPSCR  ;	: FPSCR,  ;	\ syntactic sugar

\ Forth model registers
((
\ Previous ARM model
09 reg LP	09 reg LP,
0A reg TOS      0A reg TOS,
0B reg UP       0B reg UP,
0C reg PSP      0C reg PSP,
0D reg RSP      0D reg RSP,
))

\ Current Cortex/ARM model
$07 reg TOS     $07 reg TOS,
$09 reg LP	$09 reg LP,
$0B reg UP      $0B reg UP,
$0C reg PSP     $0C reg PSP,	$0C reg xPSP
$0D reg RSP     $0D reg RSP,	$0D reg xRSP

previous definitions


\ ******************
\ Thumb instructions
\ ******************
\ This section supports Thumb-1 and Cortex-M instructions.

\ =================
\ Instruction width
\ =================

also asm-access definitions
: .w		\ --
\ Use 32 bit (wide) instruction formats.
  -1 <wide> !  ;
: .n		\ --
\ Use 16 bit (narrow) instruction formats.
  1 <wide> !  ;
previous definitions

: WideOnly?	\ -- flag
\ Return true if instruction must use wide form.
  <wide> @ -1 =  ;
: NarrowOnly?	\ -- flag
\ Return true if instruction must use narrow form.
  <wide> @ 1 =  ;

: WideOK?	\ -- flag
\ Return true if the wide form is permitted.
  <wide> @ 1 <>  Thumb1? 0= and  ;
: NarrowOK?	\ -- flag
\ Return true if the narrow form is permitted.
  <wide> @ -1 <>  ;


\ ====================
\ Const/Imm12 handling
\ ====================
\ In line with ARMv7 Reference Manual, we denote encoded 12 bit
\ values as constants (const) and unencoded values as literals
\ (lit). Immediate refers to the addressing mode, not to the
\ encoding of the value.

cellbits #32 = [if]
: lrol	rol  ;
: lror  ror  ;
[then]
cellbits #64 = [if]
$1:0000:0000 constant bit32
$0:8000:0000 constant bit31
: lrol		\ x u -- x' ; 32 bit rotate left
  0 ?do
    1 lshift dup bit32 and
    if  1 or  then			\ bit 31 -> bit 0
  loop
  $FFFF:FFFF and
;

: lror		\ x u -- x' ; 32 bit rotate right
  0 ?do
    dup 1 and  swap 1 rshift swap	\ -- x' bit0
    if  bit31 or  then			\ bit 0 -> bit 31
  loop
  $FFFF:FFFF and
;
[then]

: T2s8imm?	\ imm -- enc true | 0
  8 lrol  #32 8 do
    dup $FFFFFF80 and $00000080 = if
      $7F and  i 7 lshift or
      true  unloop exit
    endif
    1 lrol
  loop
  drop 0
;

: 00xy00xy?	\ imm -- flag
  dup $FFFF and over #16 rshift $FFFF and =
  swap $FF00 and 0= and
;

: xy00xy00?	\ imm -- flag
  dup $FFFF and over #16 rshift $FFFF and =
  swap $00FF and 0= and
;

: xyxyxyxy?	\ imm -- flag
  dup $FF and swap			\ -- lsb value
  2dup 8 rshift $FF and <>
  if  2drop 0 exit  endif
  2dup #16 rshift $FF and <>
  if  2drop 0 exit  endif
  2dup #24 rshift $FF and <>
  if  2drop 0 exit  endif
  2drop -1
;

: T2Const12	\ imm -- enc
\ +G Convert a 32 bit literal into its Thumb2 12 bit encoding.
  dup 0 #255 within?
  if  exit  endif
  dup T2s8imm?
  if  nip  exit  endif
  dup 00xy00xy?
  if  $00FF and  $0100 or  exit  endif
  dup xy00xy00?
  if  8 rshift $00FF and  $0200 or  exit  endif
  dup xyxyxyxy?
  if  $00FF and  $0300 or  exit  endif
  err_T2Const12 ?a-error
;

: T2Const12?	\ imm -- flag
\ +G Return true if the value can be encoded in T2 const12 format.
  dup 0 #255 within?
  if  drop true exit  endif
  dup T2s8imm?
  if  2drop true exit  endif
  dup 00xy00xy?
  if  drop true exit  endif
  dup xy00xy00?
  if  drop true exit  endif
  xyxyxyxy?
;

: T2Lit3?	\ n -- flag
\ +G Return true if n is in the range 0..7.
  8 u<  ;

: T2Lit12?	\ n -- flag
\ +G Return true if n is in the range $0..FFF.
  $1000 u<  ;

: T2memOff?	\ n -- flag
\ +G Return true if n is an index/offset in the range
\ +* -255..4095.
  #-255 #4096 within  ;

: decodeLit138	\ opcode -- lit12
\ Decode the 1:3:8 12 bit immediate/constant.
\ For disassembler and code generator support.
  dup #26 rshift 1 and #11 lshift	\ bit 26     -> 11
  over #12 rshift $07 and 8 lshift or	\ bits 14:12 -> 10:8
  swap $FF and or			\ bits 7:0   -> 7:0
;

: decodeConst138	\ opcode -- const12
\ Display 12 bit constant. See page A5-16.
\ For disassembler and code generator support.
  decodeLit138 dup $0C00 and if		\ inspect bits 11:10
    dup $007F and $080 or		\ unrotated value in bits 6:0
    swap $0F80 and 7 rshift		\ shift count in bits 11:7
    lror
  else
    case dup $0300 and			\ inspect bits 9:8
      $0000 of  $0FF and  endof
      $0100 of  $0FF and dup #16 lshift or  endof
      $0200 of  $0FF and dup #24 lshift swap #8 lshift or  endof
      $0300 of  $0FF and dup 8 lshift or  dup #16 lshift or  endof
    endcase
  endif
;


\ =====================
\ Other immediate forms
\ =====================

: Imm7*4?	\ n -- flag
\ +G Return true if the value fits into Imm7*4 form.
  dup 3 and 0=  swap $200 u< and
;

: Imm8*4?	\ n -- flag
\ +G Return true if n is positive and fits into Imm8*4 form.
  dup 3 and 0=  swap $400 u< and
;

: Imm16?	\ n -- flag
\ +G Return true if the constant fits IMM16, e.g. MOV Rd, # imm16
  $0001:0000 u<
;


\ ==============
\ Shift handling
\ ==============

: ?ShiftRange	\ n a b -- n'
\ Check that n is in the range a<=n<=b and clip n to 5 bits.
  2 pick >r  within? 0= err_immshift ?a-error  r> $01F and
; doNotSin

: CheckLSL#	\ n -- n
\ Error if not 0..31
  0 #31 ?ShiftRange
;

: CheckLSR#	\ n --
\ Error if not 1..32
  1 #32 ?ShiftRange
;

: CheckASR#	\ n --
\ Error if not 1..32
  1 #32 ?ShiftRange
;

: CheckROR#	\ n --
\ Error if not 0..31
  0 #31 ?ShiftRange
;

: checkTshift#  \ [n] -- n'
  case  shiftop @
        0 of  $1000 shiftop !  0  endof	\ convert <none> to .LSL # 0
    $1000 of  CheckLSL#  endof
    $1020 of  CheckLSR#  endof
    $1040 of  CheckASR#  endof
    $1060 of  CheckROR#  endof
    $1160 of  $1060 shiftop !  0  endof	\ convert .RRX to .ROR # 0
       err_immshift a-error
  endcase	\ -- n flag
;

: SetShiftOp	\ n -- ; --
  create
    5 lshift  $1000 or ,
  does>
    @ shiftop !
;

: shift         \ n -- ; sets shift indicator
  create
    5 lshift  $1000 or  ,
  does>
    @ shiftop !
;

: T2ShiftOp	\ n -- ; --
  create
    5 lshift  $1000 or ,
  does>
    @ shiftop ! ( Thumb2Only )
;

0 SetShiftOp SetLSL
1 SetShiftOp SetLSR
2 SetShiftOp SetASR
3 SetShiftOp SetROR


\ =================
\ Opcode generation
\ =================

: Topc16,	\ opcode --
\ +G Lay a 16 bit Thumb opcode. Includes A-RESET.
  w,c(t) a-reset
;

: T2opc32,	\ opcode --
\ +G Lay a 32 bit Thumb2 opcode. Includes A-RESET.
  T2x32, a-reset
;

: Aopc32,	\ opcode --
\ +G Lay a 32 bit ARM opcode. Includes A-RESET.
  l,c(t) a-reset
;

: Last-Treg	\ -- n
\ +G Return last register and error if not 0..7
  last-reg dup 7 u> err_hreg ?a-error
;

: splitReg	\ reg -- regl regh
\ +G Split register into low/hi, low 0..7, high 0/1.
  dup 7 and swap 8 and 3 rshift
;

: ?SP/PC	\ reg --
\ +G Error if *\i{reg} is not SP=13 or PC=15.
  dup #13 <> swap #15 <> and err_SP/PC ?a-error
;

: ?BadReg	\ reg -- reg
\ +G Error if *\i{reg} is SP=13 or PC=15.
  dup dup #13 = swap #15 = or err_badReg ?a-error
;

: ?BadRdSP/PC	\ --
\ +G Error?BadRd if Rd is SP or PC.
  1stReg ?BadReg drop
;

: ?BadRdSP/PC-M4M7	\ --
  Cortex-M4? Cortex-M7? or
  if  ?BadRdSP/PC  then
;

: ?SP		\ reg --
\ +G Error if *\i{reg} is not SP=13.
  #13 <> err_SPreg ?a-error
;

: ?SP/PClit	\ n --
\ +G Error if n is not word aligned or outside 0..1020.
  dup $3 and  swap 0 $3FF within? 0= or err_badTLit ?a-error
;

: ?SPlit	\ n --
\ +G Error if n is not word aligned or outside 0..508
  dup $3 and 0<>
  swap 0 $1FF within? 0=
  or err_badTLit ?a-error
;


\ ====================================
\ Data Processing Instruction handling
\ ====================================

: T1dpiF1	\ opcode --
\ +G Thumb 1 data processing instruction: op Rd, Rn, Rm
  last-Treg 6 lshift or  last-Treg 3 lshift or  last-Treg or Topc16,
;

: T1dpiF2	\ imm3 opcode --
\ +G Thumb 1 data processing instruction: op Rd, Rn, # imm3
  last-Treg 3 lshift or  last-Treg or
  swap 7 and 6 lshift or  Topc16,
;

: T1dpiF3	\ imm8 opcode --
\ +G Thumb 1 data processing instruction: op Rd, # imm8
  last-Treg 8 lshift or  swap $FF and or  Topc16,
;

: T1dpiF4	\ opcode --
\ +G Thumb 1 data processing instruction: op Rd, Rn, # shift
  last-Treg 3 lshift or  last-Treg or
  swap $1F and 6 lshift or  Topc16,
;

: T1dpiF5	\ opcode --
\ +G Thumb 1 data processing instruction: op Rd, Rm
  last-Treg 3 lshift or  last-Treg or  Topc16,
;

: T1dpiF6	\ imm8*4 opcode --
\ +G Thumb 1 data processing instruction: op Rd, SP|PC # imm8*4
  last-reg dup ?SP/PC #13 = $0800 and or	\ bit11=1 for SP
  last-Treg 8 lshift or				\ Rd
  swap dup ?SP/PClit  $3FC and 2 rshift or	\ imm8*4
  Topc16,
;

: T1dpiF7	\ imm7*4 opcode --
\ +G Thumb 1 data processing instruction: op SP, SP # imm7*4
  last-reg ?SP  last-reg ?SP
  swap dup ?SPlit  $1FC and 2 rshift or		\ imm7*4
  Topc16,
;

: T1dpiF8	\ opcode --
\ +G Thumb 1 data processing instruction: op rn, rm - Rn must be high reg.
\ +*       2                                          Rn can be any
  \ checked elsewhere
  last-reg 3 lshift or			\ H2:Rm
  last-reg splitreg 7 lshift or or	\ H1:Rn
  Topc16,
;

: T2dpiF8	\ opcode --
\ +G Thumb 2 data processing instruction: op reg, reg - includes high regs
  last-reg 3 lshift or				\ H2:Rm
  last-reg splitreg 7 lshift or or		\ H1:Rn
  Topc16,
;

: T1dpiF8u	\ opcode --
\ +G Thumb 1 data processing instruction: op reg, reg.
\ +* Always generates the high register form without checking the
\ +* register numbers. Although formally UNPREDICTABLE for two low
\ +* registers, this is the only way to copy a register without
\ +* setting the flags and is legal for Thumb-2. Used by the CPY_T
\ +* pseudo instruction.
  last-reg 3 lshift or				\ H2:Rm
  last-reg splitreg 7 lshift or or		\ H1:Rn
  Topc16,
;

: @=?		\ -- flag
\ +G True if litpool needed - @= used with LDR.
  @=-flag @ 0<>
;

: imm?		\ -- flag
\ +G True if an immeduate (#) has been set.
  imm-flag @ 0<>
;

: shift?	\ -- flag
\ +G True if a shift, e.g. .lsl, has been set.
  shiftop @ 0<>
;

: just1Treg?	\ -- flag
\ +G Return true if the instruction uses one reg with no
\ +* immediate or shift components.
  reg-count @ 1 =  shiftop @ 0= and  imm-flag @ 0= and
;

: just2Tregs?	\ -- flag
\ +G Return true if the instruction uses two regs with no
\ +* immediate or shift components.
  reg-count @ 2 =  shiftop @ 0= and  imm-flag @ 0= and
;

: just3Tregs?	\ -- flag
\ +G Return true if the instruction uses three regs with no
\ +* immediate or shift components.
  reg-count @ 3 =  shiftop @ 0= and  imm-flag @ 0= and
;

: just4Tregs?	\ -- flag
\ +G Return true if the instruction uses four regs with no
\ +* immediate or shift components.
  reg-count @ 4 =  shiftop @ 0= and  imm-flag @ 0= and
;

: 1Treg+imm?	\ -- flag
\ +G Return true if the instruction uses one reg with
\ +* an immediate value.
  reg-count @ 1 =  imm-flag @ 0<> and
;

: 2Tregs+imm?	\ -- flag
\ +G Return true if the instruction uses two regs with
\ +* an immediate value.
  reg-count @ 2 =  imm-flag @ 0<> and
;

: 3Tregs?	\ -- flag
\ +G Return true if the instruction uses three regs.
  reg-count @ 3 =
;

: 2Tregs?	\ -- flag
\ +G Return true if the instruction uses two regs.
  reg-count @ 2 =
;

: 1Treg?	\ -- flag
\ +G Return true if the instruction uses one reg.
  reg-count @ 1 =
;

((
: Rd=Rn?	\ -- flag ; SFP013
\ +G Return true if Rd=Rn.
  0 r[] @ 1 r[] @ =
;
))

: 1>2Regs	\ --
\ +G Convert Rd,... to Rd,Rn,... where Rn=Rd.
  0 r[] @ 1 r[] !			\ copy Rd
  reg-count incr
;

: 2>1Regs?	\ -- flag
\ +G Convert Rd,Rn,... to Rd,... if Rd=Rn
  0 r[] @ 1 r[] @ = dup
  if  reg-count decr  endif
;

: 2>3regs	\ --
\ +G Convert Rd,Rm to Rd,Rn,Rm where Rn=Rd
  1 r[] @ 2 r[] !			\ move Rm
  0 r[] @ 1 r[] !			\ copy Rd
  reg-count incr
;

: 3>2regsCom?	\ -- flag ; true=done
\ +G Convert a 3 register operation to a 2 register operation
\ +* if possible. This is for commutative operations, and changes
\ +* form if Rd=Rn or Rd=Rm.
  0 r[] @ 1 r[] @ = if			\ Rd=Rn
    2 r[] @ 1 r[] !  reg-count decr
    -1 exit
  endif
  0 r[] @ 2 r[] @ = if			\ Rd=Rm
    reg-count decr
    -1 exit
  endif
  0
;

: 3>2regsNonCom?	\ -- flag ; true=done
\ +G Convert a 3 register operation to a 2 register operation
\ +* if possible. This is for non-commutative operations, and
\ +* changes form only if Rd=Rn.
  0 r[] @ 1 r[] @ = if			\ Rd=Rn
    2 r[] @ 1 r[] !  reg-count decr
    -1 exit
  endif
  0
;

: +2regsNB	\ opcode -- opcode'
\ +G Merge 2 registers into a 32 bit Thumb2 instruction.
\ +* Rd in bits 11:8, Rm in bits 3:0.
  last-reg or				\ Rm
  last-reg #8 lshift or			\ Rd
;

: +3regs	\ opcode -- opcode'
\ +G Merge 3 registers into a 32 bit Thumb2 instruction.
  last-reg or				\ Rm
  last-reg #16 lshift or		\ Rn
  last-reg #8 lshift or			\ Rd
;

: +3regsNB	\ opcode -- opcode'
\ +G Merge 3 registers into a 32 bit Thumb2 instruction. Check
\ +* for bad (R13/15) registers
  last-reg ?BadReg or			\ Rm
  last-reg ?BadReg #16 lshift or	\ Rn
  last-reg ?BadReg #8 lshift or		\ Rd
;

: +4regsNB	\ opcode -- opcode'
\ +G Merge 4 registers into a 32 bit Thumb2 instruction. Check
\ +* for bad (R13/15) registers
  last-reg ?BadReg or			\ Rm
  last-reg ?BadReg #16 lshift or	\ Rn
  last-reg ?BadReg #8 lshift or		\ RdHi
  last-reg ?BadReg #12 lshift or	\ RdLo
;

: T2SRdRn#Lit138	\ imm opcode --
\ +G Lay a 32 bit Thumb 2 instruction: op [.s] Rd, Rn, # lit12
  swap enc12>138 or			\ immediate data
  last-reg #16 lshift or		\ Rn
  last-reg #8 lshift or			\ Rd
  <s> @ $00100000 and or		\ merge S bit
  T2opc32,				\ lay 32 bit opcode
;

: T2SRdRn#Const138	\ imm opcode --
\ +G Lay a 32 bit Thumb 2 instruction: op [.s] Rd, Rn, # const12
  swap T2Const12 swap T2SRdRn#Lit138	\ convert const, then lay
;

: T2Rn#Const138		\ imm opcode -- ; CMN CMP
\ +G Lay a 32 bit Thumb 2 instruction: op Rn, # const12
  swap T2Const12 enc12>138 or		\ -- opc'
  last-reg #16 lshift or		\ Rn
  T2opc32,				\ lay 32 bit opcode
;

: T2SRd#Const138	\ imm opcode -- ; MVN ...
  swap T2Const12 enc12>138 or    	\ -- opc' ; opc + const12
  +sbit20				\ S bit
  1stReg ?BadReg 8 lshift or		\ Rd
  T2opc32,
;

: (T2Rd#Imm16)	\ imm opcode -- ; MOV, MOVT
  ~Imm16mask and
  swap encImm16 Imm16mask and or	\ opc + imm16
  1stReg ?BadReg 8 lshift or  T2x32,
;

: T2Rd#Imm16	\ imm opcode -- ; MOV, MOVT
  (T2Rd#Imm16) a-reset  ;

: S/IT3>2loRegs?	\ -- flag
\ +G If possible, convert a three reqister operation op Rd, Rn, Rm
\ +* to a two register form. For this to happen Rd=Rn and all
\ +* registers must be low ones, the S flag must be set or an IT
\ +* block is active. If possible, do it and return true.
  just3Tregs? if
    3loRegs? if
      1stReg 2ndReg = if
        2 r[] @ 1 r[] !  reg-count decr  -1
      endif
    endif
  endif
  0
;

: T2SRdRnRm	\ opcode --
\ +G Lay a 32 bit Thumb three register instruction: top [.s] Rd, Rn, Rm
  Thumb2only  +sbit20 +3regs T2opc32,
;

: u>i3:i2	\ u -- mask
\ +G Convert a literal shift to its encoded form: imm3:imm2
  dup $1C and #10 lshift		\ bits 4:2 -> 14:12
  swap $03 and 6 lshift or		\ bits 1:0 -> 7:6
;

: shift>i3:i2:t	\ n -- mask
\ +G Convert a literal shift to its encoded form: imm3:imm2:type
  u>i3:i2
  shiftop @ $60 and 1 rshift or		\ type 6:5 -> 5:4
;

: +T2ShImm5	\ [imm] opcode -- opcode'
\ +G Merge a shift type and immediate shift value with the opcode.
  shiftop @ if
    >r  checkTshift# shift>i3:i2:t  r> or	\ merge shift
  endif
;

: T2SRdRnRmSh	\ [imm] opcode --
\ +G Lay a 32 bit Thumb 2 instruction: op [.s] Rd, Rn, Rm [ sh # n ]
  +T2ShImm5 +sbit20 +3regs T2opc32,
;

: +RnRm		\ opcode -- opcode' ; for instructions with no Rd
  last-reg or				\ Rm
  last-reg #16 lshift or		\ Rn
;

: T2RnRmSh	\ [imm] opcode -- ; CMN CMP
\ +G Lay a 32 bit Thumb 2 instruction: op [.s] Rn, Rm [ sh # n ]
  +T2ShImm5 +RnRm T2opc32,
;

: T2SRdRm	\ opcode --
\ +G Lay a 32 bit Thumb2 opcode, e.g. MOV  Rd, Rm
  last-reg or				\ Rm
  last-reg #8 lshift or			\ Rd
  <s> @ $0010:0000 and or		\ merge S bit
  T2opc32,
;

: T2SRdRmSh	\ imm opcode --
\ +G Lay a 32 bit Thumb 2 shift instruction: op [.s] Rd, Rm sh # n
  +t2ShImm5 T2SRdRm  ;

: +T2SshRdRmImm5,	\ imm opcode --
  +sbit20  swap u>i3:i2 or
  1stReg 8 lshift or  2ndReg or
  T2opc32,
;

: badTins	\ --
\ +G Error for a badly formed instruction.
  err_badTargs a-error
;

: T2mlx,	\ opc --
\ Used for MLA and MLS.
  4regs-mustbe
  1stReg 8 lshift or			\ Rd
  2ndReg #16 lshift or			\ Rn
  3rdReg or				\ Rm
  4thReg #12 lshift or			\ Ra
  T2opc32,
;


\ ===========================
\ Branch instruction handling
\ ===========================

: +TBcond	\ opcode -- opcode'
\ +G Add branch condition code into the opcode in bits 11..8
  cc-code @off #20 rshift or  cc-set off
;

: +TBcond25..22	\ opcode -- opcode'
\ +G Add branch condition code into the opcode in bits 25..22
  cc-code @off 6 rshift or  cc-set off
;

: (blx1)	\ dest opcode --
  2dup $E7FF and $1000 or swap +Trel22h,	\ high part, H=%10
  $F800 and swap $7FF and +Trel22l,		\ low part, H=%11 for BL, %01 for BLX
  a-reset
;


\ =========================================
\ LDM/STM and POP/PUSH instruction handling
\ =========================================

: ?NoWriteBack	\ --
\ Error if Writeback set.
  <w> @ err_writeback ?a-error
;

: ?WriteBack	\ --
\ Error if Writeback not set.
  <w> @ 0= err_writeback ?a-error
;

: ?BadList	\ flag --
\ Bad register list if flag=nz.
  err_regorder ?a-error
;

: T1ldm/stm	\ opcode --
  ?WriteBack
  ldm-mask @ $FF00 and ?BadList		\ check reg list
  ldm-count @ 1 < err_LdmMask1 ?a-error
  ldm-mask @ $00FF and or		\ merge reg list
  last-Treg 8 lshift or			\ merge base reg
  Topc16,
;

: T1pop/push	\ opcode -- ; can PUSH R14, can POP PC
  dup $0800 and				\ set for POP/$BC00 clear for PUSH/$B400
  if  $7F00  else  $BF00  endif
  ldm-mask @ and ?BadList		\ check reg list
  ldm-count @ 1 < err_LdmMask1 ?a-error
  ldm-mask @ $00FF and or		\ merge R0..R7 reg list
  ldm-mask @ $C000 and 0<> $0100 and or	\ merge R14 or R15
  Topc16,
;

: T2ldm/stm32	\ opcode --
  dup $0010:0000 and			\ set for LDM, clear for STM
  if  $2000 else $A000 endif
  ldm-mask @ and ?BadList		\ check reg list
  ldm-mask @ $C000 and $C000 = ?Unpred	\ can't have PC and LR
  ldm-count @ 2 < err_LdmMask2 ?a-error
  ldm-mask @ $FFFF and or		\ register list
  last-reg #16 lshift or		\ Rn
  <w> @ $0020:0000 and or		\ W
  T2opc32,
;

: T2pop/push	\ opcode -- ; can PUSH R14, can POP PC
  dup $0010:0000 and if			\ set for POP/$E8BD:x clear for PUSH/$E92D:x
    ldm-mask @ $C000 and $C000 = ?BadList
    ldm-mask @ $2000 and ?BadList
  else
    ldm-mask @ $A000 and ?BadList
  endif
  ldm-count @ 2 < err_LdmMask2 ?a-error
  ldm-mask @ $FFFF and or T2opc32,
;


\ =====================================
\ LDRx/STRx memory instruction handling
\ =====================================

: ?PreIndex	\ --
\ Error if not preindexed addressing: [ reg, # imm ]
  indexing @ 0=  <p> @ 0= or  err_PreIndex ?a-error
;

: Imm5?		\ imm scale -- flag
\ +G Return true if imm can be a 5 bit immediate value. Scale=4/2/1.
  case
    4 of  dup 3 and 0=  swap 2 arshift  endof
    2 of  dup 1 and 0=  swap 1 arshift  endof
    1 of  true swap  endof
      err_ise a-error
  endcase
  #32 u< and
;

: scaleImm5	\ imm5 scale -- 5b
\ scale=4/2/1
  case
    4 of  dup 3 and err_memoff ?a-error  2 rshift  endof
    2 of  dup 1 and err_memoff ?a-error  1 rshift  endof
    1 of  endof
      err_ise a-error
  endcase
  dup 0 #31 within? 0= err_memoff ?a-error
;

: CheckImm8*4	\ offset -- field
\ Check and scale immediate imm8*4
  dup 3 and err_memoff ?a-error
  2 rshift dup $FF u> err_memoff ?a-error
;

: T1memF1	\ imm5 scale opcode -- ; LDR Rd, [Rn, #imm5*scale]
  ?PreIndex ?NoWriteBack
  -rot scaleImm5 6 lshift or		\ imm5*scale
  last-Treg 3 lshift or			\ Rn
  last-Treg or  Topc16,			\ Rd
;

: T1memF2	\ opcode -- ; LDR Rd, [Rm, Rn]
  ?PreIndex ?NoWriteBack
  last-Treg 6 lshift or			\ Rm
  last-Treg 3 lshift or			\ Rn
  last-Treg or  Topc16,			\ Rd
;

: T1memF3	\ imm8*4 opcode -- ; LDR Rd, [PC, #imm8]
\ Note that PC address bits 1:0 are '00' for this calculation, and
\ the offset is always positive. Thus literal pools must be after
\ the reference.
  ?PreIndex ?NoWriteBack
  swap CheckImm8*4 or			\ imm8*4
  last-reg #15 <> err_SP/PC ?a-error	\ PC
  last-Treg 8 lshift or  Topc16,	\ Rd
;

: T1memF4	\ imm8*4 opcode -- ; LDR Rd, [SP, #imm8]
  ?PreIndex ?NoWriteBack
  swap CheckImm8*4 or			\ imm8*4
  last-reg #13 <> err_SP/PC ?a-error	\ SP
  last-Treg 8 lshift or  Topc16,	\ Rd
;

: ?Bit#		\ n -- ; error if not 0..31
  0 #31 within? 0= err_bit# ?a-error
;

: >MsbLsb	\ #lsb #width -- #msb #lsb
  over + 1- swap  over ?Bit#  dup ?Bit#
;

: mergeBFxImms	\ #lsb #width opcode -- opcode'
  -rot >MsbLsb				\ -- opcode #msb #lsb
  dup 3 and 6 lshift			\ convert #lsb to imm3:imm2
  swap $01C and #10 lshift or		\ -- opcode #msb imm
  or or					\ -- opcode'
;

: T2BfRdLsbWid		\ #lsb #width opcode -- ; BFC
  mergeBFxImms
  last-reg ?BadReg 8 lshift or		\ Rd
  T2opc32,
;

: T2BfRdRnLsbWid	\ #lsb #width opcode -- ; BFI
  mergeBFxImms
  last-reg ?BadReg 8 lshift or		\ Rd
  last-reg ?BadReg #16 lshift or	\ Rn
  T2opc32,
;

: T2xbfxRdRnLsbWid	\ #lsb #width opcode --
  rot u>i3:i2 or			\ merge #lsb
  swap dup 1 32 within? 0= err_bitwidth ?a-error
  1- or					\ merge width-1
  last-reg ?BadReg #16 lshift or	\ Rn
  last-reg ?BadReg 8 lshift or		\ Rd
  T2opc32,
;

: T2BfxRdRnLsbWid	\ #lsb #width opcode --
  mergeBFxImms
  last-reg ?BadReg #16 lshift or	\ Rn
  last-reg 8 lshift or			\ Rd
  T2opc32,
;

: cdp_tIns	\ opcode -- ; cp# opc1 opc2 -- ; CDP, CDP2
  create
    ,
  does> ?prefix Not-M0/M1
    @ 					\ cp# opc1 opc2 opcode --
    swap $07 and 5 lshift or		\ opc2
    swap $0F and #20 lshift or		\ opc1
    swap $0F and 8 lshift or		\ cp#
    last-cpr or                         \ CRm
    last-cpr #16 lshift or		\ CRn
    last-cpr #12 lshift or		\ CRd
    T2opc32,
;

: 0op32_T2Ins	\ -- ; no operand instructions
  create  ,  does> ?prefix Not-M0/M1 @ T2opc32,
;

: hint32_ins	\ opcode -- ; imm --
  create ,  does> ?prefix Thumb2only
  @  swap $0F and or  T2opc32,
;

: hint32_insA	\ opcode -- ; imm --
  create ,  does> ?prefix
  @  swap $0F and or  Aopc32,
;

\ ================
\ Label references
\ ================

: FwdLab32-8?	\ label -- flag
\ Return true if label is aligned and within range.
  dup 3 and
  if  drop false exit  endif
  chere(t) alignPC+4 - $0400 u<		\ offset < $0400
;

: T1RdLab32-8,	\ label opcode --
\ ADR or LDR to a cell aligned label
  1stReg 8 lshift or			\ -- label opc ; opc+Rt/Rd
  swap +Trel+8PCn Topc16,
;

: T2RdLab138	\ label opcode --
\ ADR to a non-aligned address
  1stReg 8 lshift or			\ -- label opc ; opc+Rt/Rd
  swap +T2adr T2opc32,  ;


\ =================
\ IT block handling
\ =================

: NotInIT	\ --
\ +G Error if in IT block.
  <itcount> @ ?unpred  ;

: LastInIT	\ --
\ +G Error if in IT block and not the last instruction.
  <itcount> @ if
    <itcount> @ 1 <> ?unpred
  endif
;

: ccBits	\ -- abcd
  cc-code @ #28 rshift $0F and
;

: ccbit0mask	\ count+1 -- 0/F
  ccBits 1 and 0<> $0F and		\ count+1  0/-1
  swap 1-				\ 0/-1 count
  case
    1 of  %0000  endof
    2 of  %1000  endof
    3 of  %1100  endof
    4 of  %1110  endof
        err_ise ?a-error
  endcase				\ -- count invbits
  and					\ -- invbits'
;

: it_Tins	\ mask invmask n --
  create  rot , swap , 1+ ,  does> ?prefix Not-M0/M1	\ SFP006
  Thumb2only  NotInIT			\ -- struct
  dup 2 cells + @ dup <itcount> !	\ +1 above allows for IT instruction itself
  ccbit0mask over cell + @ xor		\ -- struct invmask'
  ccBits $0F = ?unpred
  ccBits $0E = if			\ error check for .AL
    dup ?unpred
  endif
  swap @ or				\ mask in the guaranteed '1' bit
  ccBits 4 lshift or			\ firstcond
  $BF00 or Topc16,			\ opcode
;


\ ==========================
\ Memory instruction parsing
\ ==========================
\ <p> set for pre-index
\ <u> set for up/add
\ <b> set for byte
\ <w> set by ! for pre-indexed with write-back
\ <d> implementation defined, we always use 0.

: MustbeIndexed	\ --
\ +G Error if not using [ Rn ]
  indexing @ 0= err_addrmode ?a-error
;

: 0IndexOnly	\ [0] --
\ Intruction must be indexed and the offset must be 0.
  MustBeIndexed
  imm? if
    0<> err_addrmode ?a-error
  endif
;

: >off8		\ offset -- offset'
\ Convert an offset to 8 bit form for memory addressing.
\ cr ." off = " dup .xword
  dup 0<			\ direction, 0 -> +0
  if  abs  <u> off  else  <u> on  endif
  dup #255 u> err_off8 ?a-error
;

: scaleOffS8	\ imm8 scale -- off8
\ Scale an offset. The final offset is returned. Scale=4/2/1.
\ The <U> flag is set if the offset is positive/zero, cleared
\ if negative.
  case
    4 of  dup 3 and err_memoff ?a-error  2 arshift >off8  endof
    2 of  dup 1 and err_memoff ?a-error  1 arshift >off8  endof
    1 of  >off8  endof
      err_ise a-error
  endcase
;

: +offS8gen	\ [imm] opc scale -- opc'
\ Scale an offset and merge it with the opcode.
  imm? if
    swap >r  scaleOffS8  r> or
  else
    drop  <u> on
  endif
;

: +offS8L	\ [imm] opc -- opc'
\ Merge in an optional offset to a long-word. The offset is
\ placed in the bottom 8 bits of the opcode.
  4 +offS8gen
;

: +offS8H	\ [imm] opc -- opc'
\ Merge in an optional offset to a half-word. The offset is
\ placed in the bottom 8 bits of the opcode.
  2 +offS8gen
;

: +offS8B	\ [imm] opc -- opc'
\ Merge in an optional offset to a byte. The offset is
\ placed in the bottom 8 bits of the opcode.
  1 +offS8gen
;

: >off12	\ offset -- offset'
\ Convert an offset to 12 bit form for memory addressing.
  dup 0<			\ direction, 0 -> +0
  if  negate  <u> off  else  <u> on  endif
  dup #4095 u> err_i12 ?a-error
;

: +PUWlow	\ --
\ Merge in the P,U,W flags in bits 10,9,8 of the opcode.
  <p> @ $0000:0400 and  or		\ Preindexed
  <u> @ $0000:0200 and  or		\ merge Up/Add
  <w> @ $0000:0100 and  or		\ Write back
;

: +T2RdRn	\ --
\ +G Merge in the first two regs used, Rd/Rt12 and Rn16
  1stReg #12 lshift or			\ merge Rd
  2ndReg #16 lshift or			\ merge Rn
;

: +T2memInd8,	\ [imm] opc --
\ Uses an unscaled 8 bit signed offset.
  imm? if				\ -- imm opc
    swap >off8 or			\ -- opc'
  else					\ -- opc
    <u> on				\ -- opc
  endif
  +PUWlow +T2RdRn  T2opc32,
;

: +T2memIndP8S,	\ [imm] opc scale -- ; Rt, [ Rn, # +imm8 ]
\ For an 8 bit scaled positive offset
  +offS8gen
  <p> @ 0<>  <u> @ 0<> and  <w> @ 0= and
  0= err_addrmode ?a-error
  +T2RdRn  T2opc32,
;

: +T2memIndP8T,	\ [imm] opc --
\ For an 8 bit unscaled positive offset.
  1 +T2memIndP8S,
;

: +T2memIndP12,	\ +imm12 opc --
  over #4095 u> err_i12 ?a-error
  or +T2RdRn  T2opc32,
;

: +T2memIndPC12, \ +/-imm12 opc
  1stReg #12 lshift or			\ Rt
  swap +Trel12PCb T2opc32,		\ imm12
;

: +ShiftInd	\ [imm] opc -- opc'
\ Merge in the optional 2 bit scale factor, e.g. LDRH Rt, [ Rn, Rm .LSL # 2 ]
  >r
  shiftop @ $FFF and err_shiftop ?a-error \ must be LSL
  imm? if
    dup 3 u> err_shiftcnt3 ?a-error	\ must be 0..3
  else
    0
  endif
  4 lshift r> or			\ opc + shift
;


\ ========================
\ Coprocessor instructions
\ ========================

: ldc_Tins	\ opcode -- ; cp# [imm] --
  create
    ,
  does> ?prefix @ Thumb2only Not-M0/M1	\ SFP006
  <option?> @ if			\ -- cp# option opc
    $0080:0000 or			\ <p> off  <u> on  <w> off
    swap $0FF and or			\ merge option
    last-reg #16 lshift or		\ merge Rn
  else
    indexing @ if			\ -- cp# [imm] opc ; Reg+offset
      <p> @ 0=				\ if post-indexed
      if  <w> on  endif			\ force W bit
      +offS8L				\ Add 8 bit offset to long word
      <p> @ $0100:0000 and  or		\ Preindexed
      <u> @ $0080:0000 and  or		\ merge Up/Add
      <w> @ $0020:0000 and  or		\ Write back
      last-reg #16 lshift or		\ merge Rn
    else				\ -- cp# label opc ; abs addr
      $010F:0000 or			\ treat as pre-indexed PC relative
      swap +Trel8PCl			\ -- cp# opc'
    endif
  endif
  last-cpr #12 lshift or		\ merge CRd
  swap $0F and 8 lshift or		\ merge cp#
  T2opc32,
;

: T2mcr6	\ -- cp# opc1 opc2 ins
\ Used for MCR(2) cp# opc1 Rt Crn Crm opc2
  swap $07 and 5 lshift or		\ opc2
  swap $07 and #21 lshift or		\ opc1
  swap $0F and 8 lshift or		\ cp#
  last-cpr or				\ CRm
  last-cpr #16 lshift or		\ Crn
  last-reg #12 lshift or		\ Rt
  T2opc32,
;

: T2mcrr5	\ -- cp# opc1 in
\ Used for MCRR(2) <coproc>,<opc1>,<Rt>,<Rt2>,<CRm>
  swap $0F and 4 lshift or		\ opc1
  swap $0F and 8 lshift or		\ cp#
  last-cpr or				\ CRm
  last-reg #16 lshift or		\ Rt2
  last-reg #12 lshift or		\ Rt
  T2opc32,
;


\ ==================
\ Shift instructions
\ ==================

: (lsl_t)	\ [imm] --
  2Tregs+imm? if
    s/it? 2loRegs? and if
      $0000 T1dpiF4  exit
    endif
    Not-M0/M1				\ SFP006
    ?BadRdSP/PC-M4M7 			\ SFP014
    SetLSL $EA4F:0000 T2SRdRmSH  exit	\ errors if not T2
  endif
  just3Tregs? if
    3loRegs? if
      3>2regsNonCom? drop
    endif
  endif
  just2Tregs? if
    s/it? 2loregs? and if
      $4080 T1dpiF5 exit
    endif
    2>3Regs
  endif
  just3Tregs? if
    Not-M0/M1				\ SFP006
    ?BadRdSP/PC-M4M7 			\ SFP014
    $FA00:F000 T2SRdRnRm exit
  endif
  BadTins
;

: (lsr_t)	\ [imm] --
  2Tregs+imm? if
    s/it? 2loRegs? and if
      $0800 T1dpiF4  exit
    endif
    Not-M0/M1				\ SFP006
    ?BadRdSP/PC-M4M7 			\ SFP014
    SetLSR $EA4F:0000 T2SRdRmSH  exit	\ errors if not T2
  endif
  just3Tregs? if
    3loRegs? if
      3>2regsNonCom? drop
    endif
  endif
  just2Tregs? if
    s/it? 2loregs? and if
      $40C0 T1dpiF5 exit
    endif
    2>3regs
  endif
  just3Tregs? if
    Not-M0/M1				\ SFP006
    ?BadRdSP/PC-M4M7 			\ SFP014
    $FA20:F000 T2SRdRnRm exit
  endif
  badTins
;

: (asr_t)	\ [imm] --
  2Tregs+imm? if
    s/it? 2loRegs? and if
      $1000 T1dpiF4  exit
    endif
    Not-M0/M1				\ SFP006
    ?BadRdSP/PC-M4M7 			\ SFP014
    SetASR $EA4F:0000 T2SRdRmSH  exit	\ errors if not T2
  endif
  just3Tregs? if
    3>2regsNonCom? drop
  endif
  just2Tregs? if
    s/it? 2loRegs? and if
      $4100 T1dpiF5 exit
    endif
    2>3Regs
  endif
  just3Tregs? if
    Not-M0/M1? WideOk? and if		\ SFP006
      ?BadRdSP/PC-M4M7 			\ SFP014
      $FA40:F000 T2SRdRnRm exit
    endif
  endif
  badTins
;

: (ror_t)	\ [imm] --
  2Tregs+imm? if
    dup 0= err_ror#0 ?a-error		\ ROR Rd, Rm, #0 = RRX
    Not-M0/M1				\ SFP006
    ?BadRdSP/PC-M4M7 			\ SFP014
    SetROR $EA4F:0000 T2SRdRmSH  exit	\ errors if not T2
  endif
  just3Tregs? if
    3loRegs? if
      3>2regsNonCom? drop
    endif
  endif
  just2Tregs? if
    s/it? 2loRegs? and if
     $41C0 T1dpiF5 exit
    endif
    2>3regs
  endif
  just3Tregs? if
    Not-M0/M1				\ SFP006
    ?BadRdSP/PC-M4M7 			\ SFP014
    $FA60:F000 T2SRdRnRm exit
  endif
  badTins
;

: (rrx_t)
  Not-M0/M1				\ SFP006
  ?BadRdSP/PC-M4M7 			\ SFP014
  $EA4F:0030 T2SRdRm
;

: selShiftIns	\ ??? --
\ Select shift operation based on shiftop.
  case  shiftop @off
    $1000 of  (lsl_t)  endof
    $1020 of  (lsr_t)  endof
    $1040 of  (asr_t)  endof
    $1060 of  (ror_t)  endof
    $1160 of  (rrx_t)  endof
      err_shiftop a-error
  endcase
;


\ =========================
\ Miscellaneous and special
\ =========================

: ?SpecReg	\ u -- u
\ Error if out of range.
  dup #255 u> err_specReg ?a-error
;

: T2RnImm12	\ imm opc --
  swap abs $FFF and or			\ -- opc'
  1stReg #16 lshift or
  T2opc32,
;

: T2RnImm8	\ imm opc --
  swap abs $FF and or
  1stReg #16 lshift or
  T2opc32,
;

: T2RnRmsh#	\ imm opc --
  +ShiftInd				\ -- opc'
  2ndReg or				\ Rm
  1stReg #16 lshift or			\ Rn
  T2opc32,
;

: PLDins	\ opci12 opci8 opcrr -- ; [imm] --
  create rot , swap , ,  does> ?prefix Thumb2only Not-M0/M1	\ SFP006
  indexing @ if
    1Treg+Imm? if
      over 0 4095 within? if		\ imm addr --
        @ T2RnImm12  exit
      endif
      over -255 -1 within? if
        cell + @ $0080 or T2RnImm8  exit
      endif
      err_PLDoffset a-error
    endif
    dup 2 cells + @ 0<> 2Tregs? and if	\ -- [imm] addr ; check register form
      2 cells + @ T2RnRmsh#  exit
    endif
  endif
  reg-count @ 0= if			\ -- label struct ; PLD <label>
    dup 2 cells + @ if			\ PLD/PLI check
      2 cells + @ $000F:0000 or
      swap +Trel12PCb T2opc32, exit
    endif
  endif
  BadTins
;

: LMUL4ins	\ opcode -- ; -- ; SFP006
  create  ,  does> ?prefix Thumb2only Not-M0/M1
  @ +4regsNB T2opc32,
;

: LMUL4insAll	\ opcode -- ; -- ; SFP006
  create  ,  does> ?prefix Thumb2only
  @ +4regsNB T2opc32,
;

: xsat_Tins	\ opcode -- ; imm5 [shift5] --
  create  ,  does> @ ?prefix Thumb2only Not-M0/M1 >r	\ SFP006
  case  shiftop @			\ check shift type and counts
        0 of  SetLSL  0  endof
    $1000 of  CheckLSL#  endof
    $1040 of  CheckASR#  endof
       err_shiftop a-error
  endcase
  u>i3:i2 r> or				\ opc + shift count
  shiftop @ $1040 =			\ opc + shift type
  if  $0020:0000 or  endif
  swap dup 1 32 within? 0= err_satbit ?a-error	\ check imm5
  1- $01F and or			\ opc + shift + imm5
  last-reg ?BadReg #16 lshift or	\ merge Rn
  last-Reg ?BadReg 8 lshift or		\ merge Rd
  T2opc32,
;

: +rotn3	\ rotn opc -- opc'
\ Given a rotation 0..3 or 0,8,16,24, merge the rotation into
\ bits 5:4 of the opcode.
  case  swap
    0 of  endof
    1 of  $010 or  endof
    2 of  $020 or  endof
    3 of  $030 or  endof
    8 of  $010 or  endof
    #16 of  $020 or  endof
    #24 of  $030 or  endof
      err_shiftcnt3 a-error
  endcase
;

: xxtx_Tins	\ opc16 opc32 -- ; [rotn] --
  create  swap , ,  does> ?prefix Thumb2only
  just2Tregs? 2loRegs? and if		\ -- addr
    @ last-Treg 3 lshift or  last-Treg or
    Topc16, exit
  endif
  just2Tregs? if			\ inspecified, rotation=0
    imm-flag on  0 swap
  endif
  Not-M0/M1				\ SFP006
  2Tregs+Imm? if			\ -- rotn addr
    cell + @ last-reg ?BadReg or	\ opc + Rm
    last-reg ?BadReg 8 lshift or	\ opc + Rm + Rd
    +rotn3				\ -- opc' ; add in rotation
    T2opc32, exit
  endif
  BadTins
;

: tbx_Tins	\ opc -- ; --
  create  ,  does> @ ?prefix Thumb2only Not-M0/M1	\ SFP006
  2Tregs? if
    last-reg ?BadReg or  last-reg #16 lshift or
    T2opc32, exit
  endif
  BadTins
;

: hint16/32	\ opc16 opc32 -- ; --
  create  swap , ,  does> ?prefix Thumb2Only
  WideOnly? if
    Not-M0/M1
    cell + @ T2opc32,
  else
    @ Topc16,
  endif
;

: T2revxx	\ opcode --
  last-reg ?BadReg
  dup #16 lshift  swap or  or		\ Rm Rm
  last-reg ?BadReg 8 lshift or
  T2opc32,
;

: T2RdRmRm_Tins	\ opcode ; -- ; RBIT
  create  ,  does> @ ?prefix Thumb2Only	Not-M0/M1	\ -- opcode ; SFP006
  2Tregs? if
    ?BadRdSP/PC-M4M7 			\ SFP014
    T2revxx exit
  endif
  BadTins
;

: T2revxx_Tins	\ opc16 opc32 ; -- ; REV and friends
  create  swap , ,  does> ?prefix Thumb2Only	\ -- opcode
  2Tregs? if
    2loRegs? if
      @  last-reg 3 lshift or  last-reg or Topc16,
    else
      Not-M0/M1				\ SFP006
      ?BadRdSP/PC-M4M7 			\ SFP014
      cell + @ T2revxx
    endif
    exit
  endif
  BadTins
;

: T2movRdImm32	\ imm32 -- ; lay MOV/MOVT pair
  <s> @ err_mov32s ?a-error		\ bad if S bit set
  dup $FFFF and
  $F240:0000 (T2Rd#Imm16)		\ MOV Rd, # imm16 ; low part
  #16 rshift $FFFF and
  $F2C0:0000 (T2Rd#Imm16)		\ MOVT Rd, #imm16 ; high part
  a-reset
;

: T1>UAL3	\ --
\ Convert Thumb1 notation to 3 instruction UAL format
  1Treg+imm? if				\ sub rd, # imm
    1>2regs				\ ->  sub rd, rd, # imm
  endif
  just2Tregs? if			\ sub rd, rn
    2>3regs ( last-reg  peek-reg add-reg  add-reg )	\ ->  sub rd, rd, rn
  endif
;

: SP,SP,?	\ -- flag ; true if both regs SP
  peek-reg #13 =  peek-prevReg #13 = and  ;
: Rlo,SP/PC,?	\ -- flag ; true if lo reg and SP or PC
  1stReg loReg?  2ndReg dup #13 = swap #15 = or  and ;

: ?0offset	\ --
\ Convert a memory reference [ reg ] to [ reg, # 0 ]
\ This leaves  LDR reg, label alone
  just2Tregs? if
    indexing @
    if  imm-flag on  0  endif
  endif
;


\ =================
\ IDSP instructions
\ =================

: T2RdRnRmW_ins	\ opc -- ; SFP008
\ Generate T2 instruction with three 4-bit registers:
\ Rd at bit8, Rn at bit 16, Rm at bit 0.
  create  ,  does> @ ?prefix 		\ -- opcode
  Thumb2Only Not-M0/M1
  just3Tregs? 0= err_addrmode ?throw
  +3regsNB T2opc32,
;

: T2RdRnRmW_IDSPins	\ opc -- ; SFP008
\ Generate T2 DSP instruction with three 4-bit registers:
\ Rd at bit8, Rn at bit 16, Rm at bit 0.
  create  ,  does> @ ?prefix 		\ -- opcode
  IDSPonly
  just3Tregs? 0= err_addrmode ?throw
  ?BadRdSP/PC-M4M7 			\ SFP014
  +3regsNB T2opc32,
;

: +4regsNB-2	\ opcode -- opcode'
\ +G Merge 4 registers into a 32 bit Thumb2 instruction. Check
\ +* for bad (R13/15) registers. Used by SMLAxy and friends.
  last-reg ?BadReg #12 lshift or	\ Ra
  last-reg ?BadReg or			\ Rm
  last-reg ?BadReg #16 lshift or	\ Rn
  last-reg ?BadReg #8 lshift or		\ Rd
;

: HMUL4_IDSPins	\ opcode -- ; -- ; SFP006
  create  ,  does> ?prefix  IDSPonly
  @ +4regsNB-2 T2opc32,
;

: pkhxx_IDSPins	\ opcode -- ; [imm] --
  create  ,  does> @ ?prefix 		\ -- opcode
  IDSPonly
  3Tregs? 0= err_addrmode ?throw
  <s> @ err_addrmode ?throw		\ no .S
  shiftop @ if
    dup $020 and if			\ PKHTB
      shiftop @ $060 and $040 <>	\ .ASR
      err_addrmode ?throw
    else				\ PKHBT
      shiftop @ $060 and 0<>		\ .LSL
      err_addrmode ?throw
    endif
    +T2ShImm5
  endif
  +3regsNB T2opc32,
;

: MUL4lohi_IDSPins	\ opcode -- ; --
  create  ,  does> ?prefix IDSPonly
  @ +4regsNB T2opc32,
;

: xsat16_IDSPins	\ opcode -- ; imm5 --
  create  ,  does> @ ?prefix IDSPonly
  swap dup 1 16 within? 0= err_satbit ?a-error	\ check imm4
  1- $0F and or				\ opc + imm5
  last-reg ?BadReg #16 lshift or	\ merge Rn
  last-Reg ?BadReg 8 lshift or		\ merge Rd
  T2opc32,
;

: xxtax_IDSPins	\ opc32 -- ; [rotn] --
  create  ,  does> @ ?prefix IDSPonly
  3Tregs? 0= err_addrmode ?throw
  just3Tregs? if			\ unspecified, rotation=0
    imm-flag on  0 swap
  endif					\ -- rotn opc
  +3regsNB +rotn3			\ -- opc ; merge in regs and rotation
  T2opc32,
;

: xxtax2_IDSPins	\ opc32 -- ; [rotn] --
  create  ,  does> @ ?prefix IDSPonly
  2Tregs? 0= err_addrmode ?throw
  just2Tregs? if			\ unspecified, rotation=0
    imm-flag on  0 swap
  endif					\ -- rotn opc
  +2regsNB +rotn3			\ -- opc ; merge in regs and rotation
  T2opc32,
;


\ ===========================
\ Floating point instructions (T2 and ARM32), ASIMD and VFP
\ ===========================

((
variable <VFPdatatype>		\ 0 = integer, nz = float
variable <VFPunsigned>		\ 0 = signed, 1 = unsigned
variable <VFPdatasize>		\ 0 = 8 bit, 1 = 16 bit, 2 = 32 bit, 3 = 64 bit
variable <VFPindex>		\ data 32 - 0..7, 16 - 0..3, 32 - 0..1
))
also asm-access definitions
: .f32		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is 32 bit.
  <VFPdatatype> on  <VFPunsigned> off  2 <VFPdatasize> !
;
: .f64		\ --
\ *G Indicates that the data in the Dn/Qm registers is 64 bit.
  <VFPdatatype> on  <VFPunsigned> off  3 <VFPdatasize> !
;
: .s8		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is signed 8 bit.
  <VFPdatatype> off  <VFPunsigned> off  0 <VFPdatasize> !
;
: .s16		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is signed 16 bit.
  <VFPdatatype> off  <VFPunsigned> off  1 <VFPdatasize> !
;
: .s32		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is signed 32 bit.
  <VFPdatatype> off  <VFPunsigned> off  2 <VFPdatasize> !
;
: .s64		\ --
\ *G Indicates that the data in the Dn/Qm registers is signed 64 bit.
  <VFPdatatype> off  <VFPunsigned> off  3 <VFPdatasize> !
;
: .u8		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is unsigned 8 bit.
  <VFPdatatype> off  <VFPunsigned> on  0 <VFPdatasize> !
;
: .u16		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is unsigned 16 bit.
  <VFPdatatype> off  <VFPunsigned> on  1 <VFPdatasize> !
;
: .u32		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is unsigned 32 bit.
  <VFPdatatype> off  <VFPunsigned> on  2 <VFPdatasize> !
;
: .u64		\ --
\ *G Indicates that the data in the Dn/Qm registers is unsigned 64 bit.
  <VFPdatatype> off  <VFPunsigned> on  3 <VFPdatasize> !
;
: .i8		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is signed 8 bit.
  <VFPdatatype> off  <VFPunsigned> off  0 <VFPdatasize> !
;
: .i16		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is signed 16 bit.
  <VFPdatatype> off  <VFPunsigned> off  1 <VFPdatasize> !
;
: .i32		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is signed 32 bit.
  <VFPdatatype> off  <VFPunsigned> off  2 <VFPdatasize> !
;
: .i64		\ --
\ *G Indicates that the data in the Dn/Qm registers is signed 64 bit.
  <VFPdatatype> off  <VFPunsigned> off  3 <VFPdatasize> !
;
: .8		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is 8 bit.
  0 <VFPdatasize> !
;
: .16		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is 16 bit.
  1 <VFPdatasize> !
;
: .32		\ --
\ *G Indicates that the data in the Sn/Dn/Qm registers is 32 bit.
  2 <VFPdatasize> !
;
: .64		\ --
\ *G Indicates that the data in the Dn/Qm registers is 64 bit.
  3 <VFPdatasize> !
;
: [n]		( u -- )  <VFPindex> !  ;
\ *G Set an index for VMOV.
: [0]		( -- )  0 [n]  ;
\ *G Set index 0 for VMOV.
: [1]		( -- )  1 [n]  ;
\ *G Set index 1 for VMOV.
: [2]		( -- )  2 [n]  ;
\ *G Set index 2 for VMOV.
: [3]		( -- )  3 [n]  ;
\ *G Set index 3 for VMOV.
: [4]		( -- )  4 [n]  ;
\ *G Set index 4 for VMOV.
: [5]		( -- )  5 [n]  ;
\ *G Set index 5 for VMOV.
: [6]		( -- )  6 [n]  ;
\ *G Set index 6 for VMOV.
: [7]		( -- )  7 [n]  ;
\ *G Set index 7 for VMOV.
previous definitions
: .f32?		\ -- flag
  <VFPdatatype> @   <VFPunsigned> @ 0= and  <VFPdatasize> @ 2 = and  ;
: .f64?		\ -- flag
  <VFPdatatype> @   <VFPunsigned> @ 0= and  <VFPdatasize> @ 3 = and  ;
: .s8?		\ -- flag
  <VFPdatatype> @ 0=  <VFPunsigned> @ 0= and  <VFPdatasize> @ 0= and  ;
: .s16?		\ -- flag
  <VFPdatatype> @ 0=  <VFPunsigned> @ 0= and  <VFPdatasize> @ 1 = and  ;
: .s32?		\ -- flag
  <VFPdatatype> @ 0=  <VFPunsigned> @ 0= and  <VFPdatasize> @ 2 = and  ;
: .s64?		\ -- flag
  <VFPdatatype> @ 0=  <VFPunsigned> @ 0= and  <VFPdatasize> @ 3 = and  ;
: .u8?		\ -- flag
  <VFPdatatype> @ 0=  <VFPunsigned> @ and  <VFPdatasize> @ 0= and  ;
: .u16?		\ -- flag
  <VFPdatatype> @ 0=  <VFPunsigned> @ and  <VFPdatasize> @ 1 = and  ;
: .u32?		\ -- flag
  <VFPdatatype> @ 0=  <VFPunsigned> @ and  <VFPdatasize> @ 2 = and  ;
: .u64?		\ -- flag
  <VFPdatatype> @ 0=  <VFPunsigned> @ and  <VFPdatasize> @ 3 = and  ;
: .i8?		\ -- flag
  <VFPdatatype> @ 0=  <VFPdatasize> @ 0= and  ;
: .i16?		\ -- flag
  <VFPdatatype> @ 0=  <VFPdatasize> @ 1 = and  ;
: .i32?		\ -- flag
  <VFPdatatype> @ 0=  <VFPdatasize> @ 2 = and  ;
: .i64?		\ -- flag
  <VFPdatatype> @ 0=  <VFPdatasize> @ 3 = and  ;
: .8?		\ -- flag
  <VFPdatasize> @ 0=  ;
: .16?		\ -- flag
  <VFPdatasize> @ 1 =  ;
: .32?		\ -- flag
  <VFPdatasize> @ 2 =  ;
: .64?		\ -- flag
  <VFPdatasize> @ 3 =  ;

: last-FPreg	\ -- u flag ; u=reg#, flag=true=Du, flag=false=Su
  last-regData dup $FF00 and
  case
    sfp-type of  $FF and 0  endof
    dfp-type of  $FF and -1  endof
      err_NeedFPreg ?a-error
  endcase
;

: last-VFPreg	\ -- u ; ignores type
  last-regData $00FF and
;

: last-SIMDreg	\ -- u flag ; u=reg#, flag=true=Du, flag=false=Su
  last-regData dup $FF00 and
  case
    dfp-type of  $FF and 0  endof
    qfp-type of  $FF and -1  endof
      err_NeedASIMDreg ?a-error
  endcase
;

: last-ASIMDreg	\ -- u ; ignores type
  last-SIMDreg drop
;

\ Lay Sx register format - VFP only

: +sfr-d	\ opc u -- opc' ; add Sd to the opcode as Vd:D (bits 15:12,22)
  dup 1 rshift $0F and #12 lshift	\ bits 15:12
  swap 1 and #22 lshift or		\ bit 22
  or
;

: +sfr-n	\ opc u -- opc' ; add Sn to the opcode as Vn:N (bits 19:16,7)
  dup 1 rshift $0F and #16 lshift	\ bits 19:16
  swap 1 and 7 lshift or		\ bit 7
  or
;

: +sfr-m	\ opc u -- opc' ; add Sm to the opcode as Vm:M (bits 3:0,5)
  dup 1 rshift $0F and			\ bits 3:0
  swap 1 and 5 lshift or		\ bit 5
  or
;

((
: +SdSm,	\ opc -- ; add Sd, Sm regs to opcode and lay it.
  last-sfp +sfr-m  last-sfp +sfr-d  T2opc32,
;
))

: +Sd		\ opc -- opc' ; add Sd reg to opcode.
  last-sfp +sfr-d
;

: +Sm		\ opc -- opc'
  last-sfp +sfr-m
;

: +Sn		\ opc -- opc'
  last-sfp +sfr-n
;


\ Lay Dx/Qx register format without setting a size bit - ASIMD only

: (+dfr-d)	\ opc u -- opc' ; add Sd to the opcode as D:Vd (bits 22,15:12)
  dup $10 and #18 lshift		\ bit 4 -> bit 22
  swap $0F and #12 lshift or		\ bits 3:0 -> 15:12
  or
;

: (+dfr-n)	\ opc u -- opc' ; add Sn to the opcode as N:Vn (bits 7,19:16)
  dup $10 and 3 lshift			\ bit 4 -> bit 7
  swap $0F and #16 lshift or		\ bits 3:0 -> 19:16
  or
;

: (+dfr-m)	\ opc u -- opc' ; add Sm to the opcode as M:Vm (bits 5,3:0)
  dup $10 and 1 lshift			\ bit 4 -> bit 5
  swap $0F and  or			\ bits 3:0 -> 3:0
  or
;

: +Dd		\ opc -- opc' ; add Dd to opcode
  last-VFPreg (+dfr-d)
;

: +Dm		\ opc -- opc' ; add Dm to opcode
  last-VFPreg (+dfr-m)
;

: +Ad		\ opc -- opc' ; add ASIMDd reg to opcode.
  last-ASIMDreg (+dfr-d)
;

: +Am		\ opc -- opc' ; add ASIMDd reg to opcode.
  last-ASIMDreg (+dfr-m)
;

: +An		\ opc -- opc' ; add ASIMDd reg to opcode.
  last-ASIMDreg (+dfr-n)
;

: +AdAm		\ opc -- opc' ; add ASIMDd, ASIMDm regs to opcode.
  last-ASIMDreg (+dfr-m)  last-ASIMDreg (+dfr-d)
;

: +AdAnAm	\ opc -- opc' ; add ASIMDd, ASIMDn, ASIMDm regs to opcode.
  last-ASIMDreg (+dfr-m)  last-ASIMDreg (+dfr-n)  last-ASIMDreg (+dfr-d)
;

: +AdAmAm	\ opc -- opc' ; add ASIMDd, ASIMDn, ASIMDm regs to opcode.
  last-ASIMDreg dup (+dfr-m)  (+dfr-n)  last-ASIMDreg (+dfr-d)
;


\ Lay Dx register format, setting bit 8 (sz) to indicate S/D format - VFP only

: +dfr-d	\ opc u -- opc' ; add Sd to the opcode as D:Vd (bits 22,15:12) ; sz=bit8
  (+dfr-d)  $0100 or			\ sz bit
;

: +dfr-n	\ opc u -- opc' ; add Sn to the opcode as N:Vn (bits 7,19:16) ; sz=bit8
  (+dfr-n)  $0100 or			\ sz bit
;

: +dfr-m	\ opc u -- opc' ; add Sm to the opcode as M:Vm (bits 5,3:0) ; sz=bit8
  (+dfr-m)  $0100 or			\ sz bit
;

: +last-fr-d	\ opc -- opc' ; sdd last Su/Du to opc in the D slot ; sz=bit8
  last-FPreg if  +dfr-d  else  +sfr-d  endif
;

: +last-fr-n	\ opc -- opc' ; sdd last Su/Du to opc in the N slot ; sz=bit8
  last-FPreg if  +dfr-n  else  +sfr-n  endif
;

: +last-fr-m	\ opc -- opc' ; sdd last Su/Du to opc in the M slot ; sz=bit8
  last-FPreg if  +dfr-m  else  +sfr-m  endif
;

: +Vd		\ opc -- opc' ; add Vd reg to opcode.
  +last-fr-d
;

: +VdVm		\ opc -- opc' ; add Vd, Vm regs to opcode.
  +last-fr-m  +last-fr-d
;

: +VdVnVm	\ opc -- opc' ; add Vd, Vn, Vm regs to opcode.
  +last-fr-m  +last-fr-n  +last-fr-d
;

\ -------------------------------------

: imm-mustbe	\ -- ; error if immediate flag not set
  imm-flag @ 0= err_ImmNeeded ?a-error
;

: +Imm1:4	\ imm opc -- opc'
  imm-mustbe  swap  dup $0F and swap $10 and 1 lshift or  or
;


((
create fbitsTab	\ -- addr
  32 ,  32 ,  16 ,  16 ,		\ 5..8
  32 ,  32 ,  16 ,  16 ,		\ 9..12
))
: fbits>imm5	\ imm opc -- imm' opc ; encoded vaue is size-fbits
  <VFPoption> @ 5 - dup 7 u> err_badVFPoption
  cells + @ rot - swap			\ size - fbits
;

: +VdVdImm5	\ imm opc --
  0 r[] @  1 r[] @ <> err_SdSmNotSame ?a-error
  fbits>imm5 +Imm1:4  +VdVm
;
((
: +SdSdImm5,	\ imm opc --
  0 r[] @  1 r[] @ <> err_SdSmNotSame ?a-error
  fbits>imm5 +Imm1:4  +SdSm,
;
))
((
: +SdSnSm,	\ opc -- ; add Sd, Sn, Sm regs to opcode and lay it.
  last-sfp +sfr-m  last-sfp +sfr-n  last-sfp +sfr-d  T2opc32,
;
))
((
: T2FP_SdSm_ins	\ opc --
  create  ,  does> ?prefix @  2regs-mustbe  +SdSm,
;
))
((
: T2FP_SdSnSm_ins	\ opc --
  create  ,  does> ?prefix @  3regs-mustbe  +SdSnSm,
;
))
((
: T2FP_SdSm/Imm_ins	\ opcreg opcimm -- ; [ 0 ] --
  create  swap , ,  does> ?prefix 	\ -- addr
  1Treg+imm? if
    cell + @  1reg-mustbe		\ [ 0 ] opcSdImm
    last-sfp +sfr-d  T2opc32,
  else
    @  2regs-mustbe  +SdSm,		\ opcSdDm
  endif
;
))

: LastReg-S?	\ -- flag ; true if last register Sx
  peek-regData $FF00 and sfp-type =
;

: LastReg-D?	\ -- flag ; true if last register Dx
  peek-regData $FF00 and dfp-type =
;

: LastReg-Q?	\ -- flag ; true if last register Qx
  peek-regData $FF00 and qfp-type =
;

: +Q6		\ opc -- opc' ; set bit 6 if Qx regs - ASIMD only
  LastReg-Q? if  $0000:0040 or  endif
;

: ?NotQ		\ -- ; error if Q register used
  LastReg-Q? err_QregInvalid ?a-error
;

: +size2@18	\ opc -- opc' ; add data size in bits 19:18
  <VFPdatasize> @ #18 lshift or
;

: +FPsz1@8	\ opc -- opc' ; add FP size in bit 8
  <VFPdatasize> @ 3 =		\ Double precision?
  if  $0000:0100 or  endif
;

: +Fbit1@10	\ opc -- opc'
  .f32? if  $0000:0400 or  then
;

: T2_S/FP_VdVm_ins	\ opcSimd opcVFP --
  create  swap , ,   does> ?prefix  2regs-mustbe
  .f64?	 .f32? LastReg-S? and  or
   if				\ FP?
    cell+ @			\ -- opcVFP
    +FPsz1@8 +VdVm T2opc32,
  else				\ SIMD
    @ +size2@18 +Fbit1@10 +Q6 +AdAm T2opc32,
  endif
;

: ARM32_S/FP_VdVm_ins	\ opcSimd opcVFP --
  create  swap , ,   does> ?prefix  2regs-mustbe
  .f64?	 .f32? LastReg-S? and  or
   if				\ FP?
    cell+ @ +cond		\ -- opcVFP
    +FPsz1@8 +VdVm Aopc32,
  else				\ SIMD
    @ +size2@18 +Fbit1@10 +Q6 +AdAm Aopc32,
  endif
;

: FPonly	\ -- ; error if not FP instruction.
  <VFPdatatype> @ 0= err_FPonly ?a-error
;

: T2_FP_VdVm_ins	\ opcVFP --
  create  ,   does> ?prefix  2regs-mustbe FPonly
  @ +FPsz1@8 +VdVm T2opc32,
;

: ARM32_FP_VdVm_ins	\ opcVFP --
  create  ,   does> ?prefix  2regs-mustbe FPonly
  @ +cond +FPsz1@8 +VdVm Aopc32,
;

: +size2@20	\ opc -- opc' ; add data size in bits 21:20
  <VFPdatasize> @ 3 and #20 lshift or
;

: LastReg-S?	\ -- flag ; true if last register Sx
  peek-regData $FF00 and sfp-type =
;

: T2_S/FP_VdVnVm_ins	\ opcSimd opcF32 opcVFP --
  create  rot , swap , ,   does> ?prefix  3regs-mustbe
  <VFPdatatype> @ if		\ FP?
    <VFPdatasize> @ 2 =		\ F32
    LastReg-S? 0= and if	\ and not Sx
      cell+ @ +Q6 +AdAnAm T2opc32,
    else
      2 cells + @  ?NotQ	\ -- opcVFP
      +FPsz1@8 +VdVnVm T2opc32,
    endif
  else				\ integer SIMD
    @ +size2@20 +Q6 +AdAnAm T2opc32,
  endif
;

: ARM32_S/FP_VdVnVm_ins	\ opcSimd opcF32 opcVFP --
  create  rot , swap , ,  does> ?prefix  3regs-mustbe
  <VFPdatatype> @ if		\ FP?
    <VFPdatasize> @ 2 =		\ F32
    LastReg-S? 0= and if	\ and not Sx
      cell+ @ +Q6 +AdAnAm Aopc32,
    else
      2 cells + @  +cond  ?NotQ	\ -- opcVFP
      +FPsz1@8 +VdVnVm Aopc32,
    endif
  else				\ integer SIMD
    @ +size2@20 +Q6 +AdAnAm Aopc32,
  endif
;

: T2_VCMPx_ins	\ opc -- ; set bit 16 for # 0.0 version
  create  ,  does> @ ?prefix		\ opc --
  1Treg+imm? if				\ -- imm opc
    nip
    $0001:0000 or
    +FPsz1@8  +Vd  T2opc32,
  else
    2regs-mustbe  +VdVm  T2opc32,	\ opcSdDm
  endif
;

: ARM32_VCMPx_ins	\ [imm] opc -- ; set bit 16 for # 0.0 version
  create  ,  does> @ ?prefix		\ opc --
  +cond  1Treg+imm? if			\ -- imm opc
    nip
    $0001:0000 or
    +FPsz1@8  +Vd  Aopc32,
  else					\ -- opc
    2regs-mustbe  +VdVm  Aopc32,	\ opcSdDm
  endif
;

: checkRegList	\ u -- len ; error if LDM-LEN out of range of u
  ldm-len @ 1 rot within? 0= err_badRegList ?a-error
  ldm-len @ ldm-count @ <> err_badRegList ?a-error
  ldm-len @
;

: +VFPreglist	\ opc -- opc' ; add the VFP reg list to the opcode
  ldm-first @ $FF00 and sfp-type = if	\ -- opc
    ldm-first @ $FF and +sfr-d
    32 checkReglist $FF and or		\ -- opc'
    exit
  endif
  ldm-first @ $FF00 and dfp-type = if
    $0000:0100 or			\ mark as FP64 reg
    ldm-first @ $FF and +dfr-d
    16 checkReglist 2* $FF and or	\ -- opc'
    exit
  endif
  badTins
;

: T2_Vldm/stm_ins	\ opc -- ; --
  create  ,  does> ?prefix @		\ -- opc
  Thumb2only
  <w> @ $0020:0000 and or		\ write back option
  1stReg #16 lshift or			\ Rn
  +VFPreglist  T2opc32,
;

: ARM32_Vldm/stm_ins	\ opc -- ; --
  create  ,  does> ?prefix @		\ -- opc
  +cond
  <w> @ $0020:0000 and or		\ write back option
  1stReg #16 lshift or		\ Rn
  +VFPreglist  Aopc32,
;

: T2_Vpop/push_ins	\ opc -- ; --
  create  ,  does> ?prefix @		\ -- opc
  Thumb2only
  +VFPreglist  T2opc32,
;

: ARM32_Vpop/push_ins	\ opc -- ; --
  create  ,  does> ?prefix @		\ -- opc
  +cond  +VFPreglist  Aopc32,
;
((
: VFPreglist,	\ opc -- ; add the VFP reg list to the opcode and lay it
  ldm-first @ $FF00 and sfp-type = if	\ -- opc
    ldm-first @ $FF and +sfr-d
    32 checkReglist $FF and or		\ -- opc'
    T2opc32, exit
  endif
  ldm-first @ $FF00 and dfp-type = if
    $0000:0100 or			\ mark as FP64 reg
    ldm-first @ $FF and +dfr-d
    16 checkReglist 2* $FF and or	\ -- opc'
    T2opc32, exit
    exit
  endif
  badTins
;

: T2Vldm/stm_ins	\ opc -- ; --
  create  ,  does> ?prefix @		\ -- opc
  Thumb2only  1reg-mustbe
  <w> @ $0020:0000 and or		\ write back option
  last-reg #16 lshift or		\ Rn
  VFPreglist,
;

: T2Vpop/push_ins	\ opc -- ; --
  create  ,  does> ?prefix @		\ -- opc
  Thumb2only
  VFPreglist,
;
))

: (Vldr/str_ins)	\ [imm/dest] opc -- opc'
  just2Tregs? if			\ if offset not defined, force 0
    imm-flag on  0 swap
  endif
  2Tregs+Imm? if			\ -- imm opc
    +offS8L				\ Add 8 bit offset to long word
    <u> @ $0080:0000 and  or		\ merge Up/Add
    last-reg #16 lshift or		\ Rn
    +last-fr-d 				\ Sd/Dd
    exit
  endif
  1Treg? if				\ -- label opc
    Thumb? if  swap +Trel8PCl  else  +rel8  endif	\ process label
    +last-fr-d				\ add Sd/Dd
    $000F:0000 or			\ merge Rn=PC
    exit
  endif
  BadTins
;

: T2_Vldr/str_ins	\ opc -- ; [imm/dest] --
  create  ,  does> ?prefix @		\ -- opc
  Thumb2only Not-M0/M1
  (Vldr/str_ins) T2opc32,
;

: ARM32_Vldr/str_ins	\ opc -- ; [imm/dest] --
  create  ,  does> ?prefix @		\ -- opc
  +cond (Vldr/str_ins) Aopc32,
;
((
: T2Vldr/str_ins	\ opc -- ; [imm/dest] --
  create  ,  does> ?prefix @		\ -- opc
  Thumb2only Not-M0/M1
  just2Tregs? if			\ if offset not defined, force 0
    imm-flag on  0 swap
  endif
  2Tregs+Imm? if			\ -- imm opc
    +offS8L				\ Add 8 bit offset to long word
    <u> @ $0080:0000 and  or		\ merge Up/Add
    last-reg #16 lshift or		\ Rn
    +last-fr-d 				\ Sd/Dd
    T2opc32, exit
  endif
  1Treg? if				\ -- label opc
    swap +Trel8PCl			\ process label
    +last-fr-d				\ add Sd/Dd
    $000F:0000 or			\ merge Rn=PC
    T2opc32, exit
  endif
  BadTins
;
))

: T2_FP_VdVnVm_ins	\ opc --
  create  ,  does> ?prefix @  3regs-mustbe
  +FPsz1@8  +VdVnVm  T2opc32,
;

: ARM32_FP_VdVnVm_ins	\ opc --
  create  ,  does> ?prefix @  3regs-mustbe
  +cond  +FPsz1@8  +VdVnVm  Aopc32,
;

: +imm8ASMI	\ opc imm8 -- opc'
  dup >r  $0F and or			\ bits 3:0
  r@ 4 rshift $07 and #16 lshift or	\ bits 6:4
  r> 7 rshift 1 and dup #24 lshift	\ bit  7 for ARM
  swap #28 lshift or  or		\ bit  7 for T2
;

: +imm8VFP	\ opc imm8 -- opc'
  $FF and dup >r  $0F and or		\ bits 3:0 -> bits 3:0
  r> 4 rshift $0F and #16 lshift or	\ bits 7:4 -> bits 19:16
;

: ?Imm8		\ imm -- imm ; error if not 8 bit
  dup $FFFF:FF00 and err_immval ?a-error \ check 8 bit immediate
;

also forth definitions
defer fdropa	\ F: f -- ; -- ; used when FP pack installed
defer fdupa	\ F: f -- f f ; -- ; used when FP pack installed
defer f0=a	\ F: f -- ; -- flag ; used when FP pack installed
defer sf!a	\ F: f -- ; -- ; used when FP pack installed
previous definitions

: ?FImm8	\ F: f -- f ; error if f will not fit 8 bit format
\ S1:E8:M23
  { | ftmp[ cell ] -- }
  fdupa ftmp[ sf!a			\ force 32 bit float
  ftmp[ @ $0007:FFFF and err_FPimmval ?a-error  \ check low 19 bits=0
  ftmp[ @ dup $3E00:0000 and
  swap $4000:0000 and if		\ B=1
    0<> err_FPimmval ?a-error		\ 00bb_bbb0 = 0000_0000
  else					\ B=0
    $3E00:0000 <> err_FPimmval ?a-error	\ 00bb_bbb0 = 0011_1110 = $3E
  endif
;

: f>imm8	\ F: f -- ; -- imm8
  { | ftmp[ cell ] -- }
  ftmp[ sf!a				\ force 32 bit float
  ftmp[ @ $8000:0000 and  #24 rshift	\ a = sign bit is bit 7
  ftmp[ @ $4000:0000 and  $4000:0000 xor  #24 rshift or \ b = not B is bit 6
  ftmp[ @ $01F8:0000 and  #19 rshift or	\ low six bits
;
((
: +fimmASMI	\ opc -- opc' ; F: f --
  { | ftmp[ cell ] -- }
  ftmp[ sf!a				\ force 32 bit float
  ftmp[ @ $8000:0000 and  #24 rshift or	\ a = sign bit is bit 7
  ftmp[ @ $4000:0000 and  $4000:0000 xor  #24 rshift or \ b = not B is bit 6
  ftmp[ @ $01F8:0000 and  #19 rshift or	\ low six bits
;
))

: rep4*8?	\ aabbccdd -- flag ; true if aa=bb=cc=dd
  dup $FF and				\ -- x dd
  over 8 rshift $FF and over <>		\ cc <> dd
  if  2drop 0 exit  endif
  over #16 rshift $FF and over <>	\ bb <> dd
  if  2drop 0 exit  endif
  swap #24 rshift =			\ aa <> dd
;

: imm>ASMI	\ [imm] opc -- opc' ; R: [imm] --
  .f32? if
    fdupa f0=a if
      fdropa 				\ cmode=0000, op=0
    else
      ?Fimm8 f>imm8 +imm8ASMI  $0F00 or	\ cmode=1111, op=0
    endif
    exit
  endif
  swap  dup 0=
  if  drop  exit  endif			\ cmode=0000, op=0
  .i64? if
    ?imm8  +imm8ASMI  $0E20 or  exit	\ cmode=1110, op=1
  endif
  .i8? if
    ?Imm8  +imm8ASMI  $0E00 or  exit	\ cmode=1110, op=0
  endif
  .i16? if
    dup $FFFF:0000 and err_immval ?a-error \ check 16 bit immediate
    dup $FF00 and 0= if
      +imm8ASMI  $0800 or  exit		\ cmode=1000, op=0
    endif
    dup $00FF and 0= if
      8 rshift +imm8ASMI  $0A00 or  exit	\ cmode=1010, op=0
    endif
    err_immval a-error
  endif
\ the rest is I32
  dup $FFFF:FF00 and 0= if
    +imm8ASMI exit  			\ cmode=0000, op=0
  endif
  dup $FFFF:00FF and 0= if
    8 rshift +imm8ASMI  $0200 or exit	\ cmode=0010, op=0
  endif
  dup $FF00:FFFF and 0= if
    #16 rshift +imm8ASMI  $0400 or exit	\ cmode=0100, op=0
  endif
  dup $00FF:FFFF and 0= if
    #24 rshift +imm8ASMI  $0600 or exit	\ cmode=0110, op=0
  endif
  dup $FFFF:00FF and $0000:00FF = if
    8 rshift +imm8ASMI  $0C00 or exit	\ cmode=1100, op=0
  endif
  dup $FF00:FFFF and $0000:FFFF = if
    #16 rshift +imm8ASMI  $0D00 or exit	\ cmode=1101, op=0
  endif
  dup rep4*8? if
    $FF and +imm8ASMI  $0E00 or exit	\ cmode=1110, op=0
  endif
  ." Bad SIMD immediate = " .dword
  err_immval a-error		\ can't convert
;

: opc32,	\ opc --
  Thumb? if  T2opc32,  else  Aopc32,  endif
;

: +Rt		\ opc -- opc'
  last-reg ?BadReg #12 lshift or
;

: +Rt2		\ opc -- opc'
  last-reg ?BadReg #16 lshift or
;

: doVMOVI	\ [imm] -- ; F: [fimm] --
  .f64?  0 Dx? and
  .f32?  0 Sx? and
  or 0= if				\ SIMD immediate
    Thumb? if  $EF80:0010  else  $F280:0010  endif
    +Q6  +Ad +imm8ASMI opc32,
  else					\ VFP immediate
    Thumb? if  $EEB0:0A00  else  $0EB0:0A00 +cond  endif
    +FPsz1@8 +Vd ?FImm8 f>imm8 +imm8VFP opc32,
  endif
;

: doVMOV2r-sd-VFP	\ -- ; move between two VFP registers
  Thumb? if  $EEB0:0A40  else  $0EB0:0A40 +cond  endif
  +FPsz1@8 +VdVm opc32,
;

: doVMOV2r-dq-SIMD	\ -- ; move between two SIMD registers
  Thumb? if  $EF20:0110  else  $F220:0110  endif
  +Q6 +AdAmAm opc32,
;

: +index8	\ opx -- opx'
  <VFPindex> @ dup 7 u> err_VFP/SIMDindex ?a-error
  $07 and or
;

: +index16	\ opx -- opx'
  <VFPindex> @ dup 3 u> err_VFP/SIMDindex ?a-error
  $03 and 1 lshift or
;

: +index32	\ opx -- opx'
  <VFPindex> @ dup 1 u> err_VFP/SIMDindex ?a-error
  $01 and 2 lshift or
;

: +u4		\ opx -- opx' ; merge unsigned bit into U:opc1:opc2
  <VFPunsigned> @ if  %10000 or  endif
;

: +Uopc1opc2	\ opc opx -- opc' ; merge 5 bit value into opcode
  $001F and
  dup 2 rshift #21 lshift		\ U:opc1
  swap 3 and 5 lshift or		\ opc2
  or
;

: +U:opc1:opc2	\ opc -- opc'
  case  <VFPdatasize> @
    0 of  %01000 +u4 +index8 +Uopc1opc2  endof
    1 of  %00001 +u4 +index16 +Uopc1opc2  endof
    2 of  %00000 +index32 +Uopc1opc2  endof
      err_VFP/SIMDsize a-error
  endcase
;

: doVMOV2r-dir-SIMD	\ -- ; move from Rx to Dx
  Thumb? if  $EE00:0B10  else  $0E00:0B10 +cond  endif
  +U:opc1:opc2 +Rt +An opc32,
;

: doVMOV2r-rdi-SIMD	\ -- ; move from Dx to Rx
  Thumb? if  $EE10:0B10  else  $0E10:0B10 +cond  endif
  +U:opc1:opc2 +An +Rt opc32,
;

: doVMOV2r-sr-VFP
  Thumb? if  $EE00:0A10  else  $0E00:0A10 +cond  endif
  +Rt +Sn opc32,
;

: doVMOV2r-rs-VFP	\ -- ; VMOV Rx Sx
  Thumb? if  $EE10:0A10  else  $0E10:0A10 +cond  endif
  +Sn +Rt opc32,
;

: doVMOV4r-ssrr-VFP	\ -- ; VMOV Sm Sm1 Rt Rt2
  Thumb? if  $EC40:0A10  else  $0C40:0A10 +cond  endif
  1 reg#@ 0 reg#@ - 1 <> err_VFP_Sxpair ?a-error
  +Rt2  +Rt  last-VFPreg drop  +Sm  opc32,
;

: doVMOV4r-rrss-VFP	\ -- ; VMOV Rt Rt2 Sm Sm1
  Thumb? if  $EC50:0A10  else  $0C50:0A10 +cond  endif
  3 reg#@ 2 reg#@ - 1 <> err_VFP_Sxpair ?a-error
  last-VFPreg drop  +Sm  +Rt2  +Rt  opc32,
;

: doVMOV3r-drr	\ -- ; VMOV Dm Rt Rt2
  Thumb? if  $EC40:0B10  else  $0C40:0B10 +cond  endif
  +Rt2 +Rt +Dm opc32,
;

: doVMOV3r-rrd	\ -- ; VMOV Rt Rt2 Dm
  Thumb? if  $EC50:0B10  else  $0C50:0B10 +cond  endif
  +Dm +Rt2 +Rt opc32,
;

: doVMOV	\ [imm] --
  1Treg+imm?
  if  doVMOVI exit  endif
  just2Tregs? if
    .f64?  0 Dx? and  1 Dx? and
    .f32?  0 Sx? and  1 Sx? and  or
    if  doVMOV2r-sd-VFP exit  endif	\ 2 VFP regs operation
    0 Dx?  1 Dx? and
    0 Qx?  1 Qx? and  or
    if  doVMOV2r-dq-SIMD exit  endif	\ 2 SIMD registers
    0 Dx?  1 Rx? and
    if  doVMOV2r-dir-SIMD exit  endif	\ Dx [n] Rt
    0 Rx?  1 Dx? and
    if  doVMOV2r-rdi-SIMD exit  endif	\ Rt Dx [n]
    0 Sx?  1 Rx? and
    if  doVMOV2r-sr-VFP exit  endif
    0 Rx?  1 Sx? and
    if  doVMOV2r-rs-VFP exit  endif
  endif
  just3Tregs? if
    0 Dx?  1 Rx? and  2 Rx? and
    if  doVMOV3r-drr exit  endif
    0 Rx?  1 Rx? and  2 Dx? and
    if  doVMOV3r-rrd exit  endif
  endif
  just4Tregs? if
    0 Sx?  1 Sx? and  2 Rx? and  3 Rx? and
    if  doVMOV4r-ssrr-VFP exit  endif
    0 Rx?  1 Rx? and  2 Sx? and  3 Sx? and
    if  doVMOV4r-rrss-VFP exit  endif
  endif
  BadTins
;

: vmrs_ins	\ opcT opcA -- ; -- ; VMRS Rt, FPSCR ; VMSR FPSCR Rt
  create swap , , does> ?prefix
  Thumb?
  if  @  else  cell+ @ +cond  endif
  last-reg #12 lshift or		\ PC is legal as APSR_nzcv
  opc32,
;


\ ==============
\ VCTx run times
\ ==============

also asm-access definitions
: .SI32<F32	1 <VFPoption> !  <VFPunsigned> off  2 <VFPdatasize> !  ;
\ 32 bit FP to 32 bit signed int
: .UI32<F32	2 <VFPoption> !  <VFPunsigned> on  2 <VFPdatasize> !  ;
\ 32 bit FP to 32 bit unsigned int
: .F32<SI32	3 <VFPoption> !  <VFPunsigned> off  2 <VFPdatasize> !  ;
\ 32 bit signed int to 32 bit FP
: .F32<UI32	4 <VFPoption> !  <VFPunsigned> on  2 <VFPdatasize> !  ;
\ 32 bit unsigned int to 32 bit FP
: .SI32<F64	1 <VFPoption> !  <VFPunsigned> off  3 <VFPdatasize> !  ;
\ 64 bit FP to 32 bit signed int
: .UI32<F64	2 <VFPoption> !  <VFPunsigned> on  3 <VFPdatasize> !  ;
\ 64 bit FP to 32 bit unsigned int
: .F64<SI32	3 <VFPoption> !  <VFPunsigned> off  3 <VFPdatasize> !  ;
\ 32 bit signed int to 64 bit FP
: .F64<UI32	4 <VFPoption> !  <VFPunsigned> on  3 <VFPdatasize> !  ;
\ 32 bit unsigned int to 64 bit FP

: .XS32<F32	5 <VFPoption> !  <VFPunsigned> off  2 <VFPdatasize> !  ;
\ 32 bit FP to 32 bit signed fixed, op=1, U=0, sx=1
: .XU32<F32	6 <VFPoption> !  <VFPunsigned> on  2 <VFPdatasize> !  ;
\ 32 bit FP to 32 bit unsigned fixed, op=1, U=1, sx=1
: .XS16<F32	7 <VFPoption> !  <VFPunsigned> off  2 <VFPdatasize> !  ;
\ 32 bit FP to 16 bit signed fixed, op=1, U=0, sx=0
: .XU16<F32	8 <VFPoption> !  <VFPunsigned> on 2 <VFPdatasize> !  ;
\ 32 bit FP to 16 bit unsigned fixed, op=1, U=1, sx=0
: .XS32<F64	5 <VFPoption> !  <VFPunsigned> off  3 <VFPdatasize> !  ;
\ 64 bit FP to 32 bit signed fixed, op=1, U=0, sx=1
: .XU32<F64	6 <VFPoption> !  <VFPunsigned> on  3 <VFPdatasize> !  ;
\ 64 bit FP to 32 bit unsigned fixed, op=1, U=1, sx=1
: .XS16<F64	7 <VFPoption> !  <VFPunsigned> off  3 <VFPdatasize> !  ;
\ 64 bit FP to 16 bit signed fixed, op=1, U=0, sx=0
: .XU16<F64	8 <VFPoption> !  <VFPunsigned> on  3 <VFPdatasize> !  ;
\ 64 bit FP to 16 bit unsigned fixed, op=1, U=1, sx=0

: .F32<XS32	9 <VFPoption> !  <VFPunsigned> off  2 <VFPdatasize> !  ;
\ 32 bit signed fixed to 32 bit FP, op=0, U=0, sx=1
: .F32<XU32	#10 <VFPoption> !  <VFPunsigned> on  2 <VFPdatasize> !  ;
\ 32 bit unsigned fixed to 32 bit FP, op=0, U=1, sx=1
: .F32<XS16	#11 <VFPoption> !  <VFPunsigned> off  2 <VFPdatasize> !  ;
\ 16 bit signed fixed to 32 bit FP, op=0, U=0, sx=0
: .F32<XU16	#12 <VFPoption> !  <VFPunsigned> on  2 <VFPdatasize> !  ;
\ 16 bit unsigned fixed to 32 bit FP, op=0, U=1, sx=0
: .F64<XS32	9 <VFPoption> !  <VFPunsigned> off  3 <VFPdatasize> !  ;
\ 32 bit signed fixed to 64 bit FP, op=0, U=0, sx=1
: .F64<XU32	#10 <VFPoption> !  <VFPunsigned> on  3 <VFPdatasize> !  ;
\ 32 bit unsigned fixed to 64 bit FP, op=0, U=1, sx=1
: .F64<XS16	#11 <VFPoption> !  <VFPunsigned> off  3 <VFPdatasize> !  ;
\ 16 bit signed fixed to 64 bit FP, op=0, U=0, sx=0
: .F64<XU16	#12 <VFPoption> !  <VFPunsigned> on  3 <VFPdatasize> !  ;
\ 16 bit unsigned fixed to 64 bit FP, op=0, U=1, sx=0

: .F32<F16	#13 <VFPoption> !  ;
\ 13 = 16 bit FP to 32 bit FP
: .F16<F32	#14 <VFPoption> !  ;
\ 14 = 32 bit FP to 16 bit FP

: .F64<F32	#15 <VFPoption> !  ;
\ 15 = 32 bit FP to 64 bit FP
: .F32<F64	#16 <VFPoption> !  ;
\ 16 = 64 bit FP to 32 bit FP
previous definitions

: +opc3@16	\ opc lit -- opc' ; merge 3 bits at bit 16
  7 and  #16 lshift or  ;

: +sgn1@7	\ opc lit -- opc' ; merge 1 bit at bit 7
  1 and  7 lshift or  ;

: +sx1@7	\ opc lit -- opc' ; merge size of fixed point unit 1 bit at bit 7
  1 and  7 lshift or  ;

: +Ux1@16	\ opc 0/1 -- opc' ; merge U flag, 1 bit at bit 16
  1 and #16 lshift or  ;

: +op1@18	\ opc 0/1 -- opc'
  1 and #18 lshift or  ;

create VCVTtab_t	\ -- addr ; Thumb-2 opcode table
  $EEB8:0A40 ,		\ integer to float
  $EEBA:0A40 ,		\ fixed point to float
  $FFB6:0600 ,		\ half to single FP
  $EEB7:0AC0 ,		\ single to double FP

create VCVTtab_a	\ -- addr ; ARM32 opcode table
  $0EB8:0A40 ,		\ integer to float
  $0EBA:0A40 ,		\ fixed point to float
  $F3B6:0600 ,		\ half to single FP
  $0EB7:0AC0 ,		\ single to double FP

: 1stopc	( addr -- opc )  @  ;
: 2ndopc	( addr -- opc )  cell + @  ;
: 3rdopc	( addr -- opc )  2 cells + @  ;
: 4thopc	( addr -- opc )  3 cells + @  ;

: doVCVT	\ addr -- opc
  case  <VFPoption> @
    \ integer to float conversions, VFP operation
    1 of	\ .SI32<F32/64, op=1=round to zero
      1stopc ?A+cond  %101 +opc3@16  +FPsz1@8  1 7 lshift or
      .64? if  +Dm  else  +Sm  then  +Sd
    endof
    2 of	\ .UI32<F32/64, op=1=round to zero
      1stopc ?A+cond  %100 +opc3@16  +FPsz1@8  1 7 lshift or
      .64? if  +Dm  else  +Sm  then  +Sd
    endof
    3 of	\ .F32/64<SI32
      1stopc ?A+cond  %000 +opc3@16  +FPsz1@8  1 7 lshift or
      +Sm  .64? if  +Dd  else  +Sd  then
    endof
    4 of	\ .F32/64<UI32
      1stopc ?A+cond  %000 +opc3@16  +FPsz1@8  0 7 lshift or
      +Sm  .64? if  +Dd  else  +Sd  then
    endof
    \ fixed point to float conversions, VFP operation
    5 of	\ .XS32<F32/64, 32 bit FP to 32 bit signed fixed, op=1, U=0, sx=1
      2ndopc  1 +op1@18  0 +Ux1@16  1 +sx1@7  +VdVdImm5
    endof
    6 of	\ .XU32<F32/64, 32 bit FP to 32 bit unsigned fixed, op=1, U=1, sx=1
      2ndopc ?A+cond  1 +op1@18  1 +Ux1@16  1 +sx1@7  +VdVdImm5
    endof
    7 of	\ .XS16<F32/64, 32 bit FP to 16 bit signed fixed, op=1, U=0, sx=0
      2ndopc ?A+cond  1 +op1@18  0 +Ux1@16  0 +sx1@7  +VdVdImm5
    endof
    8 of	\ .XU16<F32/64, 32 bit FP to 16 bit unsigned fixed, op=1, U=1, sx=0
      2ndopc ?A+cond  1 +op1@18  1 +Ux1@16  0 +sx1@7  +VdVdImm5
    endof
    9 of	\ .F32/64, <XS32, 32 bit signed fixed to 32 bit FP, op=0, U=0, sx=1
      2ndopc ?A+cond  0 +op1@18  0 +Ux1@16  1 +sx1@7  +VdVdImm5
    endof
    #10 of	\ .F32/64<XU32, 32 bit unsigned fixed to 32 bit FP, op=0, U=1, sx=1
      2ndopc ?A+cond  0 +op1@18  1 +Ux1@16  1 +sx1@7  +VdVdImm5
    endof
    #11 of	\ .F32/64<XS16, 16 bit signed fixed to 32 bit FP, op=0, U=0, sx=0
      2ndopc ?A+cond  0 +op1@18  0 +Ux1@16  0 +sx1@7  +VdVdImm5
    endof
    #12 of	\ .F32/64<XU16, 16 bit unsigned fixed to 32 bit FP, op=0, U=1, sx=0
      2ndopc ?A+cond  0 +op1@18  1 +Ux1@16  0 +sx1@7  +VdVdImm5
    endof
    \ half to single FP conversions, ASIMD operation
    #13 of	\ .F32<F16, 16 bit FP to 32 bit FP
      3rdopc  $0000:0100 or  +AdAm
    endof
    #14 of	\ .F16<F32, 32 bit FP to 16 bit FP
      3rdopc  $0000:0000 or  +AdAm
    endof
    \ single to double FP conversions, VFP operation
    #15 of	\ 32 bit FP to 64 bit FP
      4thopc ?A+cond  $0000:0000 or  +Sm +Dd
    endof
    #16 of	\ 64 bit FP to 32 bit FP
      4thopc ?A+cond  $0000:0100 or  +Dm +Sd
    endof
      badTins
  endcase
;

: VCVT_t	\ --
  0 ?prefix drop  2regs-mustbe
  VCVTtab_t doVCVT T2opc32,
;

: VCVT_a	\ --
  0 ?prefix drop  2regs-mustbe
  VCVTtab_a doVCVT Aopc32,
;

: doVCVTR	\ --
  case  <VFPoption> @
    1 of	\ S32<F32/64, op=0=round to FPSCR
      1stopc  %101 +opc3@16  +FPsz1@8  0 7 lshift or
      .64? if  +Dm  else  +Sm  then  +Sd
    endof
    2 of	\ U32<F32/64, op=0=round to FPSCR
      1stopc  %100 +opc3@16  +FPsz1@8  0 7 lshift or
      .64? if  +Dm  else  +Sm  then  +Sd
    endof
      badTins
  endcase
;

: VCVTR_t	\ --
  0 ?prefix drop  2regs-mustbe
  VCVTtab_t doVCVTR T2opc32,
;

: VCVTR_a	\ --
  0 ?prefix drop  2regs-mustbe
  VCVTtab_a doVCVTR Aopc32,
;

: doVCVTB	\ opc -- opc'
  case  <VFPoption> @
    #13 of	\ 13 = 16 bit FP to 32 bit FP
      0 #16 lshift or  0 7 lshift or  +Sm +Sd
    endof
    #14 of	\ 14 = 32 bit FP to 16 bit FP
      1 #16 lshift or  0 7 lshift or  +Sm +Sd
    endof
      badTins
  endcase
;

: VCVTB_t	\ --
  0 ?prefix drop  2regs-mustbe
  $EEB2:0A40 doVCVTB T2opc32,
;

: VCVTB_a	\ --
  0 ?prefix drop  2regs-mustbe
  $0EB2:0A40 +cond doVCVTB Aopc32,
;

: doVCVTT	\ --
  case  <VFPoption> @
    #13 of	\ 13 = 16 bit FP to 32 bit FP
      0 #16 lshift or  1 7 lshift or  +Sm +Sd
    endof
    #14 of	\ 14 = 32 bit FP to 16 bit FP
      1 #16 lshift or  1 7 lshift or  +Sm +Sd
    endof
      badTins
  endcase
;

: VCVTT_t	\ --
  0 ?prefix drop  2regs-mustbe
  $EEB2:0A40 doVCVTT T2opc32,
;

: VCVTT_a	\ --
  0 ?prefix drop  2regs-mustbe
  $0EB2:0A40 +cond doVCVTT Aopc32,
;

\ ==================
\ Thumb instructions
\ ==================

: adc_t		\ [imm] --
  0 ?prefix drop
  just2Tregs? if
    NarrowOk? if			\ SFP006
      2LoRegs? if
        s/it? if
          $4140 T1dpiF5  exit
        endif
      endif
    endif
    2>3regs				\ convert to 3reg form
  endif
  2Tregs+imm? if
    WideOk? if
      Not-M0/M1? if			\ SFP006
        ?BadRdSP/PC-M4M7		\ SFP014
        $F140:0000 T2SRdRn#Const138  exit
      endif
    endif
  endif
  3Tregs? if
    just3Tregs? if			\ can we convert 3 regs to two regs?
      3LoRegs? if
        s/it? if
          3>2regsCom? if
            $4140 T1dpiF5  exit
          endif
        endif
      endif
    endif
    Not-M0/M1? if			\ SFP006
      ?BadRdSP/PC-M4M7			\ SFP014
      $EB400000 T2SRdRnRmsh  exit
    endif
  endif
  badTins
;

: add_t		\ [imm] --
  0 ?prefix drop  T1>UAL3
\ Check narrow forms
  NarrowOk? if
    2Tregs+imm? if
      <s> @ 0= if
        dup Imm7*4?  SP,SP,? and
        if  $B000 T1dpiF7 exit  endif	\ A6-26 T2
        dup Imm8*4?  Rlo,SP/PC,? and
        if  $A000 T1dpiF6 exit  endif	\ A6-26 T1
      endif
      s/it? if
        dup T2lit3?  2loRegs? and
        if  $1C00 T1dpiF2 exit  endif	\ A6-22 T1
        dup 0 #255 within?  1stReg loReg? and if
          1stReg 2ndReg =
          if  $3000 T1dpiF3 exit  endif	\ A6-22 T2
        endif
      endif
    endif
    just3Tregs? if
      s/it? if
        3loRegs? if
          $1800 T1dpiF1  exit		\ A6-24 T1
        endif
      endif
      <s> @ 0= if
        1stReg 2ndReg = if
          1stReg hiReg? 3rdReg hiReg? or  Thumb2? or if
            last-reg  drop-reg  add-reg	\ discard 2nd register
            $4400 T1dpiF8 exit		\ A6-24 T2
          endif
        endif
      endif
    endif
  endif
  Cortex-M0/M1? Thumb1? or		\ SFP006
  NarrowOnly? or
  if  BadTins  endif
  2Tregs+imm? if
    <s> @ 0= if
      dup T2lit12? if			\ Must be checked before Const
        $F200:0000 T2SRdRn#Lit138 exit
      endif
    endif
    dup T2Const12? if
      $F100:0000 T2SRdRn#Const138 exit
    endif
  endif
  3Tregs? if
    ?badRdSP/PC-M4M7 $EB00:0000 T2SRdRnRmsh  exit
  endif
  badTins
;

: adr_t		\ label -- ; ADR  Rd, label
  0 ?prefix drop
  1Treg? 0=  <s> @ or			\ must be ADR Rx, <label>
  if  BadTins  exit  endif
  NarrowOk? if
    1stReg loReg? if
      dup FwdRef? if
        $A000 T1RdLab32-8,  exit	\ -- label opc ; opc+Rt
      endif
      dup FwdLab32-8? if
        $A000 T1RdLab32-8,  exit	\ -- label opc ; opc+Rt
      endif
    endif
  endif
  WideOk? if
    Not-M0/M1? if			\ SFP006
      ?BadRdSP/PC-M4M7 			\ SFP014
      $F20F:0000 T2RdLab138 exit	\ ADD/SUB Rd, PC, # imm12
    endif
  endif
  badTins
;

: and_t		\ [imm] --
  0 ?prefix drop
  just3Tregs? if
    3>2RegsCom? drop
  endif
  just2Tregs? if
    NarrowOk? if			\ SFP006
      s/it? if
        $4000 T1dpiF5 exit
      endif
    endif
    2>3regs
  endif
  Not-M0/M1? WideOk? and if		\ SFP006
    ?BadRdSP/PC-M4M7 			\ SFP014
    2Tregs+imm? if
      $F000:0000 T2SRdRn#Const138 exit
    endif
    3Tregs? if
      $EA00:0000 T2SRdRnRmSh  exit
    endif
  endif
  badTins
;

: asr_t		\ [imm] --
  0 ?prefix drop  (asr_t)
;

: b_t		\ dest --
  0 ?prefix drop
  cc-code @ $F000:0000 =		\ never is ILLEGAL
  if  badTins  endif
  cc-code @ $E000:0000 = if		\ unconditional branch
    LastInIT
    NarrowOk? if
      dup chere(t) 4 + - abs #2048 <  over FwdRef? or if
        $E000 swap +Trel11, a-reset exit
      endif
    endif
    Not-M0/M1? WideOk? and if		\ if in Wide mode ; SFP006
      $F000:9000 swap +Trel24, a-reset exit
    endif
  else					\ conditional branch
    NotInIT
    NarrowOk? if
      dup chere(t) 4 + - abs #256 <  over FwdRef? or if
        $D000 +TBcond swap +Trel8, a-reset  exit
      endif
    endif
    Not-M0/M1? WideOk? and  if		\ if in Wide mode ; SFP006
      $F000:8000 +TBcond25..22 swap +Trel20, a-reset exit
    endif
  endif
  BadTins
;

: bfc_t		\ #lsb #width --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  $F36F:0000 T2BfRdLsbWid
;

: bfi_t		\ #lsb #width --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  $F360:0000 T2BfRdRnLsbWid
;

: bic_t		\ --
  0 ?prefix drop
  just2Tregs? if
    2LoRegs? if
      s/it? if
        $4380 T1dpiF5 exit
      endif
    endif
  endif
  2Tregs+imm? if
    Not-M0/M1? if			\ SFP006
      ?BadRdSP/PC-M4M7 			\ SFP014
      $F020:0000 T2SRdRn#Const138  exit
    endif
  endif
  3Tregs? if
    just3Tregs? if			\ can we convert 3 regs to two regs?
      3LoRegs? if
        s/it? if
          3>2regsCom? if
            $4380 T1dpiF5  exit
          endif
        endif
      endif
    endif
    Not-M0/M1? if			\ SFP006
      ?BadRdSP/PC-M4M7 			\ SFP014
      $EA20:0000 T2SRdRnRmsh  exit
    endif
  endif
  badTins
;

: bkpt_t	\ imm --
  0 ?prefix drop
  $00FF and  $BE00 or  Topc16,
;

: bl_t		\ dest --
  0 ?prefix drop
  Thumb1? if
    $F800 (blx1)  exit
  endif
  $F000:D000 swap +Trel24, a-reset
;

: blx_t		\ [dest] --
  0 ?prefix drop
  reg-count @ 0= if
    Not-M0/M1? if			\ SFP006
      $E800 (blx1) exit
    endif
  endif
  reg-count @ 1 = if
    $4780 last-reg 3 lshift or Topc16,  exit
  endif
  badTins
;

: bx_t		\ --
  0 ?prefix drop
  $4700 last-reg 3 lshift or  Topc16,
;

: cbnz_t	\ dest --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  $B900 last-Treg or  swap +Trel1+5, a-reset
;

: cbz_t		\ dest --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  $B100 last-Treg or  swap +Trel1+5, a-reset
;

$EE00:0000 cdp_tIns cdp_t
$FE00:0000 cdp_tIns cdp2_t

$F3BF:8F2F 0op32_T2Ins clrex_t

: clz_t		\ --
  0 ?prefix drop  Thumb2only Not-M0/M1
  ?BadRdSP/PC-M4M7 			\ SFP014
  $FAB0:F080
  last-reg ?BadReg			\ Rm
  dup #16 lshift or or			\ Rm2
  last-reg ?BadReg 8 lshift or		\ Rd
  T2opc32,
;

: cmn_t		\ --
  0 ?prefix drop
  Thumb1? if
    just2Tregs? if
      $42C0 T1dpiF5 exit
    endif
  endif
  Thumb2? if
    just2Tregs? if
      2LoRegs? <wide> @ 0= and if
        $42C0 T1dpiF5 exit
      endif
      Not-M0/M1? if			\ SFP006
        $EB10:0F00 T2RnRmSh exit
      endif
    endif
    Not-M0/M1? if			\ SFP006
      shiftop @ if
        $EB10:0F00 T2RnRmSh exit
      endif
      1Treg+imm? if
        $F110:0F00 T2Rn#Const138 exit
      endif
    endif
  endif
  badTins
;

: cmp_t		\ [imm] --
  0 ?prefix drop
  1Treg+imm? if				\ cmp rn, # imm8
    dup #256 u<
    if  $2800 T1dpiF3 exit  endif
    Thumb2? Not-M0/M1? and if		\ SFP006
      $F1B0:0F00 T2Rn#Const138 exit
    endif
  endif
  just2Tregs? if			\ cmp rn, rm
    2loRegs? 0=				\ check for high regs
    if  $4500 T1dpiF8 exit  endif
    $4280 T1dpiF5 exit
  endif
  Thumb2? Not-M0/M1? and if		\ cmp rn, rm sh # n
    2Tregs? if
      shiftop @ if
        $EBB0:0F00 T2RnRmSh exit
      endif
    endif
  endif
  badTins
;

: cps_t		\ --
  0 ?prefix drop  Thumb2only NotInIT
  Cortex-M0? if
   <cpsMask> @ 1 and err_cputype ?a-error	\ F invalid for M0/M1
  endif
  $B660 <cpsmask> @ or Topc16,
;

: cpy_t		\ --
  0 ?prefix drop
  Just2TRegs?
  if  $4600 T1dpiF8u exit  endif
  badTins
;

$F3AF:80F0 hint32_ins DBG_T
$F3BF:8F50 hint32_ins DMB_T
$F3BF:8F40 hint32_ins DSB_T

: eor_t		\ --
  0 ?prefix drop
  Thumb1? if
    just2Tregs? if
      $4040 T1dpiF5 exit
    endif
    badTins
  endif
  just2Tregs? if
    2>3regs
  endif
  3Tregs? if
    just3Tregs? if
      3loregs? if
        s/it? if
          3>2RegsCom? if
            $4040 T1dpiF5 exit
          endif
\          2>3regs
        endif
      endif
    endif
    Not-M0/M1? if			\ SFP006
      ?BadRdSP/PC-M4M7 			\ SFP014
      $EA80:0000 T2SRdRnRmSh  exit
    endif
  endif
  Not-M0/M1? if				\ SFP006
    2Tregs+imm? if
      ?BadRdSP/PC-M4M7 			\ SFP014
      $F080:0000 T2SRdRn#Const138 exit
    endif
  endif
  badTins
;

$F3BF:8F60 hint32_ins isb_t

$08 $00 1 it_Tins IT_t
$04 $00 2 it_Tins ITT_t
$04 $08 2 it_Tins ITE_t
$02 $00 3 it_Tins ITTT_t
$02 $08 3 it_Tins ITET_t
$02 $04 3 it_Tins ITTE_t
$02 $0C 3 it_Tins ITEE_t
$01 $00 4 it_Tins ITTTT_t
$01 $08 4 it_Tins ITETT_t
$01 $04 4 it_Tins ITTET_t
$01 $0C 4 it_Tins ITEET_t
$01 $02 4 it_Tins ITTTE_t
$01 $0A 4 it_Tins ITETE_t
$01 $06 4 it_Tins ITTEE_t
$01 $0E 4 it_Tins ITEEE_t

$EC10:0000 ldc_Tins LDC_t
$EC50:0000 ldc_Tins LDCL_t
$FC10:0000 ldc_Tins LDC2_t
$FC50:0000 ldc_Tins LDC2L_t

: ldm_t		\ --
  0 ?prefix drop
  Thumb1? if
    $C800 T1ldm/stm exit
  endif
  ldm-mask @ $FF00 and 0= if
    <w> @ 1stReg loReg? and if
      $C800 T1ldm/stm exit
    endif
  endif
  Not-M0/M1				\ SFP006
  $E890:0000 T2ldm/stm32
;
: ldmia_t  ldm_t  ;
: ldmfd_t  ldm_t  ;

: ldmdb_t	\ --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  $E910:0000 T2ldm/stm32
;
: ldmea_t  ldmdb_t  ;

\ rewrite this
: ldr_t		\ [label/imm] --
  0 ?prefix drop
  ?0offset
  @=? if				\ load from literal pool
    1stReg  c_asm-litpool  
    a-reset exit
  then
  NarrowOk? if
    3Tregs? shiftop @ 0= and  imm? 0= and if
      3LoRegs? if
        $5800 T1memF2  exit		\ ldr rd, [ rn ++ rm ]
      endif
    endif
    2Tregs+imm? if			\ imm --
      <p> @ <w> @ 0= and if		\ must be preindexed with no writeback
        1stReg loReg? if			\ Rd must be lo
          dup Imm8*4? if
            peek-reg #15 =		\ ldr rd, [ PC, # imm8 ]
            if  $4800 T1memF3 exit  endif
            peek-reg #13 =		\ ldr rd, [ SP, # imm8 ]
            if  $9800 T1memF4 exit  endif
          endif
          2ndReg loReg? if
            dup 4 Imm5? if
              4 $6800 T1memF1 exit	\ ldr rd, [ rn, # imm5 ]
            endif
          endif
        endif
      endif
    endif
    1Treg? if				\ Must be a label reference ; LDR Rt, <label>
      1stReg loReg? if
        dup FwdRef? if
          $4800 T1RdLab32-8,  exit	\ -- label opc ; opc+Rt
        endif
        dup FwdLab32-8? if
          $4800 T1RdLab32-8,  exit	\ -- label opc ; opc+Rt
        endif
      endif
    endif
  endif
  WideOK? Not-M0/M1? and if		\ SFP006
    3Tregs? if
      2ndReg #15 <> if
        3rdReg ?BadReg drop		\ Rm may not be R13/15
        $F850:0000 +ShiftInd		\ opc + shift
        +T2RdRn 3rdReg or		\ Rd, Rn, Rm
        T2opc32, exit
      endif
    endif
    2Tregs+imm? <p> @ and <w> @ 0= and if \ imm -- ; preindexed only
      dup #4096 u< if
        $F8D0:0000 +T2memIndP12, exit
      endif
      dup abs #4096 u< 2ndReg #15 = and if \ imm -- ; LDR Rt, [ PC # imm ]
        $F85F:0000 +T2memIndPC12, exit
      endif
    endif
    2Tregs? indexing @ and if		\ -- [imm] ; Reg+offset
      $F850:0800 +T2memInd8, exit
    endif
    1Treg? if				\ Must be a label reference ; LDR Rt, <label>
      $F85F:0000 +T2memIndPC12,  exit
    endif
  endif
  BadTins
;

: ldrb_t		\ [imm] --
  0 ?prefix drop
  ?0offset
  NarrowOk? if
    just3Tregs? 3loRegs? and		\ SFP009
    if  $5C00 T1memF2 exit  endif
    2Tregs+imm? 2loRegs? and if
      dup 1 imm5? if
        <p> @ if
          1 $7800 T1memF1 exit
        endif
      endif
    endif
  endif
  WideOk? Not-M0/M1? and if		\ SFP006
    3Tregs? if
      $F810:0000 +ShiftInd		\ opc + shift
      +T2RdRn 3rdReg or			\ Rd, Rn, Rm
      T2opc32, exit
    endif
    2Tregs+imm? <p> @ and <w> @ 0= and if	\ imm -- ; preindexed only
      dup #4096 u< if
        $F890:0000 +T2memIndP12, exit
      endif
      dup abs #4096 u< 2ndReg #15 = and if \ imm -- ; LDR Rt, [ PC # imm ]
        $F81F:0000 +T2memIndPC12, exit
      endif
    endif
    2Tregs? indexing @ and if		\ -- [imm] ; Reg+offset
\ <p> @ <u> @ and <w> @ 0= and #-1000 ?throw
      $F810:0800 +T2memInd8, exit
    endif
    1Treg? if				\ Must be a label reference ; LDR Rt, <label>
      $F81F:0000 +T2memIndPC12,  exit
    endif
  endif
  BadTins
;

: ldrbt_t	\ [imm] --
  0 ?prefix drop  Not-M0/M1
  2Tregs? if
    $F810:0E00 1 +T2memIndP8S, exit
  endif
  BadTins
;

: ldrd_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  3Tregs? if
    MustBeIndexed
   $E850:0000 +offS8L			\ Add 8 bit offset to long-word
    <p> @ $0100:0000 and  or		\ Preindexed
    <u> @ $0080:0000 and  or		\ merge Up/Add
    <w> @ $0020:0000 and  or		\ Write back
    1stReg #12 lshift or		\ Rt
    2ndReg #8 lshift or			\ Rt2
    3rdReg #16 lshift or		\ Rn
    T2opc32, exit
  endif
  2Tregs? if				\ -- label
    $E95F:0000				\ treat as pre-indexed PC relative
    1stReg #12 lshift or		\ Rt
    2ndReg #8 lshift or			\ Rt2
    swap +Trel8PCl T2opc32, exit
  endif
  BadTins
;

: ldrex_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1 MustBeIndexed	\ SFP006
  2Tregs? if				\ -- [imm]
    $E850:0F00 4 +T2memIndP8S, exit
  endif
  BadTins
;

: ldrexb_t	\ [0] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  2Tregs? if
    0IndexOnly
    $E8D0:0F4F +T2RdRn			\ opc, Rt, Rn
    T2opc32, exit
  endif
  BadTins
;

: ldrexh_t	\ [0] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  2Tregs? if
    0IndexOnly
    $E8D0:0F5F +T2RdRn			\ opc, Rt, Rn
    T2opc32, exit
  endif
  BadTins
;

: ldrh_t		\ [imm] --
  0 ?prefix drop
  ?0offset
  NarrowOk? if
    just3Tregs? 3loRegs? and		\ SFP009
    if  $5A00 T1memF2 exit  endif
    2Tregs+imm? 2loRegs? and <w> @ 0= and if
      dup 2 imm5? if
        <p> @ if
          2 $8800 T1memF1 exit
        endif
      endif
    endif
  endif
  WideOk? Not-M0/M1? and if		\ SFP006
    indexing @ if			\ not a label
      2Tregs? if			\ two register form
        dup #4096 u< if			\ use 12 bit form
          <p> @ <w> @ 0= and if
            $F8B0:0000 +T2memIndP12, exit
          endif
        endif
        $F830:0800 +T2memInd8,  exit	\ A6-110 T3
      endif
      3Tregs? if			\ three register form
        $F830:0000 +ShiftInd		\ opc + shift
        +T2RdRn 3rdReg or		\ Rd, Rn, Rm
        T2opc32, exit
      endif
    endif
    1Treg? if				\ Must be a label reference ; LDRH Rt, <label>
      $F83F:0000 +T2memIndPC12,  exit
    endif
  endif
  BadTins
;

: ldrht_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  2Tregs? if
    $F830:0E00 +T2memIndP8T, exit
  endif
  BadTins
;

: ldrsb_t		\ [imm] --
  0 ?prefix drop
  ?0offset
  just3Tregs? 3loRegs? and  shiftop @ 0= and	\ SFP009
  if  $5600 T1memF2 exit  endif
  Thumb1? if  BadTins  endif
  Cortex-M0/M1? if  BadTins  endif	\ SFP006
  indexing @ if				\ not a label
    2Tregs? if				\ two register form
      dup #4096 u< if			\ use 12 bit form
        <p> @ <w> @ 0= and if
          $F990:0000 +T2memIndP12, exit
        endif
      endif
      $F910:0800 +T2memInd8,  exit	\ 8 bit signed unscaled offset PUWlow
    endif
    3Tregs? if				\ three register form
      $F910:0000 +ShiftInd		\ opc + shift
      +T2RdRn 3rdReg or			\ Rd, Rn, Rm
      T2opc32, exit
    endif
  endif
  1Treg? if				\ Must be a label reference ; LDRSB Rt, <label>
    $F91F:0000 +T2memIndPC12,  exit
  endif
  BadTins
;

: ldrsbt_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  2Tregs? if
    $F910:0E00 +T2memIndP8T, exit
  endif
  BadTins
;

: ldrsh_t		\ [imm] --
  0 ?prefix drop
  just3Tregs? 3loRegs? and		\ SFP009
  if  $5E00 T1memF2 exit  endif
  Thumb1? if  BadTins  endif
  Cortex-M0/M1? if  BadTins  endif	\ SFP006
  indexing @ if				\ not a label
    2Tregs? if				\ two register form
      imm? if
        dup #4096 u< if			\ use 12 bit form
          <p> @ <w> @ 0= and if
            $F9B0:0000 +T2memIndP12, exit
          endif
        endif
      endif
      $F930:0800 +offS8B +PUWlow
      +T2RdRn T2opc32, exit
    endif
    3Tregs? if				\ three register form
      $F930:0000 +ShiftInd		\ opc + shift
      +T2RdRn 3rdReg or			\ Rd, Rn, Rm
      T2opc32, exit
    endif
  endif
  1Treg? if				\ Must be a label reference ; LDRSH Rt, <label>
    $F93F:0000 +T2memIndPC12,  exit
  endif
  BadTins
;

: ldrsht_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1
  2Tregs? if
    $F930:0E00 +T2memIndP8T, exit
  endif
  BadTins
;

: ldrt_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1
  2Tregs? if
    $F850:0E00 +T2memIndP8T, exit
  endif
  BadTins
;

: lsl_t		\ [imm] --
  0 ?prefix drop (lsl_t)
;

: lsr_t		\ [imm] --
  0 ?prefix drop (lsr_t)
;

: mcr_t		\ cp# opc1 opc2 --
  0 ?prefix drop  Thumb2only Not-M0/M1  $EE00:0010 T2mcr6
;

: mcr2_t	\ cp# opc1 opc2 --
  0 ?prefix drop  Thumb2only Not-M0/M1  $FE00:0010 T2mcr6
;

: mcrr_t	\ cp# opc1 opc2 --
  0 ?prefix drop  Thumb2only Not-M0/M1 $EC40:0000 T2mcrr5
;

: mcrr2_t	\ cp# opc1 opc2 --
  0 ?prefix drop  Thumb2only Not-M0/M1 $FC40:0000 T2mcrr5
;

: mla_t		\ --
  0 ?prefix drop  Thumb2only Not-M0/M1
  ?BadRdSP/PC-M4M7 			\ SFP014
  $FB00:0000 T2mlx,
;

: mls_t		\ --
  0 ?prefix drop  Thumb2only Not-M0/M1
  ?BadRdSP/PC-M4M7 			\ SFP014
  $FB00:0010 T2mlx,
;

: mov_t		\ [imm] --
  0 ?prefix drop
\ cr ." MOV_T " reg-count ?  imm? .  s/it? . 1stReg .  2ndReg . 3rdReg . <itcount> ?
  NarrowOk? if
    1Treg+imm? 1loReg? and s/it? and if
      dup #256 u< if
        $2000 T1dpiF3 exit
      endif
    endif
(( \ three operand MOV instructions only occur for shifts
    2Tregs+imm? 1loReg? and s/it? and if  \ if Rd=Rn=lo can use short form
      1stReg 2ndReg =  shiftop @ 0= and if
        dup #256 u< if
.source-line
          $2000 T1dpiF3 exit
        endif
      endif
    endif
))
    just2Tregs? if
      <s> @ 0= if
        peek-reg 8 and  peek-prevReg 8 and or	\ check for high regs
        if  $4600 T1dpiF8 exit  endif
      endif
      2loRegs? <s> @ 0<> and if
        $1C00 T1dpiF5 exit		\ ADD .S Rd, Rn, # 0
      endif
    endif
  endif
  Thumb1? if  BadTins  endif
  1Treg+imm? if				\ -- imm
    Not-M0/M1				\ SFP006
    dup Imm16? if
      <s> @ 0= if
        $F240:0000 T2Rd#Imm16 exit	\ opc + Rd + imm16
      endif
    endif
    $F04F:0000 T2SRd#Const138 exit	\ -- imm opc ; MOV Rd, # const12
  endif
  just2Tregs? if
    <s> @ 0= WideOnly? 0= and if
      $4600 T1dpiF8 exit
    endif
    Not-M0/M1				\ SFP006
    $EA4F:0000 T2SRdRm  exit
  endif
  shiftop @ if
    2Tregs+Imm?
    if  selShiftIns exit  endif		\ MOV Rd, Rm .LSL # x
    2Tregs?
    if  selShiftIns exit  endif		\ MOV Rd, Rm .RRX
    3Tregs?
    if  selShiftIns exit  endif		\ MOV Rd, Rm .LSL Rs
  endif
  BadTins
;

: movt_t	\ imm --
  0 ?prefix drop Thumb2only Not-M0/M1	\ SFP006
  ?BadRdSP/PC-M4M7 			\ SFP014
  $F2C0:0000 T2Rd#Imm16
;

: mrc_t		\ cp# opc1 opc2 --
  0 ?prefix drop Thumb2only Not-M0/M1	\ SFP006
  $EE10:0010 T2mcr6
;

: mrc2_t	\ cp# opc1 opc2 --
  0 ?prefix drop Thumb2only Not-M0/M1	\ SFP006
  $FE10:0010 T2mcr6
;

: mrrc_t	\ cp# opc1  --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  $EC50:0000 T2mcrr5
;

: mrrc2_t	\ cp# opc1 --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  $FC50:0000 T2mcrr5
;

: mrs_t		\ sr# --
  0 ?prefix drop Thumb2only
  $F3EF:8000
  last-reg 8 lshift or
  swap ?SpecReg or T2opc32,
;

: msr_t		\ sr# --
  0 ?prefix drop Thumb2only
  $F380:8800
  last-reg #16 lshift or
  swap ?SpecReg or T2opc32,
;

: mul_t		\ --
  0 ?prefix drop
  just2Tregs? if
    2loregs? s/it? and if
      $4340 T1dpiF5 exit
    endif
    2>3regs
  endif
  just3Tregs? if
    s/it? if
      3loregs? if
        3>2regsCom? if
          $4340 T1dpiF5 exit
        endif
      endif
    endif
    <s> @ 0= if
      Thumb2Only Not-M0/M1		\ SFP006
      ?BadRdSP/PC-M4M7 			\ SFP014
      $FB00:F000
      1stReg ?BadReg 8 lshift or
      2ndReg ?BadReg #16 lshift or
      3rdReg ?BadReg or
      T2opc32, exit
    endif
  endif
  badTins
;

: mvl_t		\ imm -- ; MVL [ .s ] reg, # xxx
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  1Treg+imm? if
    dup FwdRef? if
      check-mov32 drop T2movRdImm32	\ lay MOV/MOVT pair
      exit
    endif
    1loReg? s/it? and if
      dup #256 u< if
        $2000 T1dpiF3 exit		\ MOV .s Rlo, # imm8
      endif
    endif
    dup Imm16? <s> @ 0= and if
      $F240:0000 T2Rd#Imm16 exit	\ MOV Rd, # imm16
    endif
    dup T2Const12? if
      $F04F:0000 T2SRd#Const138 exit	\ -- imm opc ; MOV [ .s ] Rd, # const12
    endif
    dup invert T2Const12? if
      invert
      $F06F:0000 T2SRd#Const138 exit	\ MVN [ .s ]
    endif
    T2movRdImm32 exit			\ lay MOV/MOVT pair
  endif
  badTins
;

: mvl32_t		\ imm -- ; MVL reg, # xxx
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  check-mov32 drop T2movRdImm32
;

: mvl32+1_t		\ imm -- ; MVL reg, # xxx
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  check-mov32+1 drop T2movRdImm32
;

: mvn_t		\ --
  0 ?prefix drop
  just2Tregs? if
    2loRegs? s/it? and if
      $43C0 T1dpiF5 exit
    endif
  endif
  Not-M0/M1				\ SFP006
  2Tregs? if
    $EA6F:0000 T2SRdRmSH  exit		\ errors if not T2
  endif
  1Treg+Imm? if
    $F06F:0000 T2SRd#Const138 exit
  endif
  badTins
;

: neg_t		\ --
  0 ?prefix drop
  just1Treg?
  if  1>2regs  endif
  just2Tregs? 2loRegs? and s/it? and
  if  $4240 T1dpiF5 exit  endif
  badTins
;

: nop_t		\ --
  0 ?prefix drop
  Thumb1? if
    $46C0 Topc16,  exit			\ mov R8, R8
  endif
  WideOnly? if
    $F3AF:8000 T2opc32, exit
  endif
  $BF00 Topc16,
;

: orn_t		\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  2Tregs+imm? if
    ?BadRdSP/PC-M4M7 			\ SFP014
    $F060:0000 T2SRdRn#Const138  exit
  endif
  2Tregs?
  if  2>3regs  endif
  3Tregs? if
    ?BadRdSP/PC-M4M7 			\ SFP014
    $EA60:0000 T2SRdRnRmsh  exit
  endif
  BadTins
;

: orr_t		\ --
  0 ?prefix drop
  just3Tregs? if
    3>2regsNonCom? drop
  endif
  just2Tregs? if
    2loRegs? s/it? and if
      $4300 T1dpiF5 exit
    endif
    2>3regs
  endif
  Not-M0/M1				\ SFP006
  3Tregs? if
    ?BadRdSP/PC-M4M7 			\ SFP014
    $EA40:0000 T2SRdRnRmsh  exit
  endif
  2Tregs+imm? if
    ?BadRdSP/PC-M4M7 			\ SFP014
    $F040:0000 T2SRdRn#Const138  exit
  endif
  badTins
;

$EAC0:0000 pkhxx_IDSPins PKHBT_t
$EAC0:0020 pkhxx_IDSPins PKHTB_t

$F890:F000 $F810:FC00 $F810:F000 PLDins PLD_t
$F8B0:F000 $F830:FC00 0          PLDins PLDW_t
$F990:F000 $F910:FC00 $F910:F000 PLDins PLI_t

: pop_t		\ --
  0 ?prefix drop
  Thumb1? NarrowOk? or if
    ldm-mask @ $7F00 and 0= if		\ only high register permitted is 15
      $BC00 T1pop/push exit
    endif
  endif
  Not-M0/M1				\ SFP006
  WideOk? if
    $E8BD:0000 T2pop/push exit
  endif
  BadTins
;

: push_t	\ --
  0 ?prefix drop
  Thumb1? NarrowOk? or if
    ldm-mask @ $BF00 and 0= if		\ only high register permitted is 14
      $B400 T1pop/push exit
    endif
  endif
  Not-M0/M1				\ SFP006
  WideOk? if
    $E92D:0000 T2pop/push exit
  endif
  BadTins
;


$FA80:F080 T2RdRnRmW_IDSPins qadd_t
$FA90:F010 T2RdRnRmW_IDSPins qadd16_t
$FA80:F010 T2RdRnRmW_IDSPins qadd8_t
$FAA0:F010 T2RdRnRmW_IDSPins qasx_t
$FA80:F090 T2RdRnRmW_IDSPins qdadd_t
$FA80:F0B0 T2RdRnRmW_IDSPins qdsub_t
$FAE0:F010 T2RdRnRmW_IDSPins qsax_t
$FA80:F0A0 T2RdRnRmW_IDSPins qsub_t
$FAD0:F010 T2RdRnRmW_IDSPins qsub16_t
$FAC0:F010 T2RdRnRmW_IDSPins qsub8_t

$FA90:F0A0 T2RdRmRm_Tins rbit_t
$BA00 $FA90:F080 T2revxx_Tins rev_t
$BA40 $FA90:F090 T2revxx_Tins rev16_t
$BAC0 $FA90:F0B0 T2revxx_Tins revsh_t

: ror_t		\ --
  0 ?prefix drop (ror_t)
;

: rrx_t		\ --
  0 ?prefix drop (rrx_t)
;

: rsb_t		\ [imm] --
  0 ?prefix drop  Thumb2only
  2Tregs+Imm? if
    2loRegs? s/it? and if
      dup 0= if
        drop  $4240 T1dpiF5 exit	\ NEG = RSB Rd, Rn # 0
      endif
    endif
    Not-M0/M1				\ SFP006
    $F1C0:0000 T2SRdRn#Const138 exit
  endif
  3Tregs? if
    Not-M0/M1				\ SFP006
    $EBC0:0000 T2SRdRnRmSh exit
  endif
  BadTins
;

$FA90:F000 T2RdRnRmW_IDSPins sadd16_t
$FA80:F000 T2RdRnRmW_IDSPins sadd8_t
$FAA0:F000 T2RdRnRmW_IDSPins sasx_t

: sbc_t		\ [imm] --
  0 ?prefix drop
  NarrowOk? Thumb1? or if
    just3Tregs? if
      3>2regsNonCom? drop
    endif
    just2Tregs? if
      2loRegs? s/it? and if
        $4180 T1dpiF5 exit
      endif
      2>3regs
    endif
  endif
  WideOk? Not-M0/M1? and if		\ SFP006
    2Tregs+Imm? if
      Thumb2only  ?BadRdSP/PC-M4M7  $F160:0000 T2SRdRn#Const138 exit
    endif
    3Tregs? if
      ?BadRdSP/PC-M4M7  $EB60:0000 T2SRdRnRmSh exit
    endif
  endif
  badTins
;

: sbfx_t	\ #lsb #width --
  0 ?prefix drop Thumb2only Not-M0/M1	\ SFP006
  $F340:0000 T2xbfxRdRnLsbWid
;

: sdiv_t	\ --
  0 ?prefix drop Thumb2only Not-M0/M1	\ SFP006
  $FB90:F0F0 +3regsNB T2opc32,
;

$FAA0:F080 T2RdRnRmW_IDSPins sel_t

: sev_t		\ -- ; SFP006
  0 ?prefix drop Thumb2only
  NarrowOk? Cortex-M0/M1? or if
    $BF40 Topc16, exit
  endif
  WideOk? if
    $F3AF:8004 T2opc32, exit
  endif
  BadTins
;

$FA90:F020 T2RdRnRmW_IDSPins shadd16_t
$FA80:F020 T2RdRnRmW_IDSPins shadd8_t
$FAA0:F020 T2RdRnRmW_IDSPins shasx_t
$FAE0:F020 T2RdRnRmW_IDSPins shsax_t
$FAD0:F020 T2RdRnRmW_IDSPins shsub16_t
$FAC0:F020 T2RdRnRmW_IDSPins shsub8_t

$FB10:0000 HMUL4_IDSPins smlabb_t
$FB10:0010 HMUL4_IDSPins smlabt_t
$FB10:0020 HMUL4_IDSPins smlatb_t
$FB10:0030 HMUL4_IDSPins smlatt_t

$FB20:0000 HMUL4_IDSPins smlad_t
$FB20:0000 HMUL4_IDSPins smladx_t

$FBC0:0000 LMUL4ins smlal_t

$FBC0:0080 MUL4lohi_IDSPins smlalbb_t
$FBC0:0090 MUL4lohi_IDSPins smlalbt_t
$FBC0:00A0 MUL4lohi_IDSPins smlaltb_t
$FBC0:00B0 MUL4lohi_IDSPins smlaltt_t

$FBC0:00C0 MUL4lohi_IDSPins smlald_t
$FBC0:00D0 MUL4lohi_IDSPins smlaldx_t

$FB30:0000 HMUL4_IDSPins smlawb_t
$FB30:0010 HMUL4_IDSPins smlawt_t
$FB40:0000 HMUL4_IDSPins smlsd_t
$FB40:0010 HMUL4_IDSPins smlsdx_t
$FBD0:00C0 MUL4lohi_IDSPins smlsld_t
$FBD0:00D0 MUL4lohi_IDSPins smlsldx_t
$FB50:0000 HMUL4_IDSPins smmla_t
$FB50:0010 HMUL4_IDSPins smmlar_t
$FB60:0000 HMUL4_IDSPins smmls_t
$FB60:0010 HMUL4_IDSPins smmlsr_t

$FB50:F000 T2RdRnRmW_IDSPins smmul_t
$FB50:F010 T2RdRnRmW_IDSPins smmulr_t
$FB20:F000 T2RdRnRmW_IDSPins smuad_t
$FB20:F010 T2RdRnRmW_IDSPins smuadx_t
$FB10:F000 T2RdRnRmW_IDSPins smulbb_t
$FB10:F010 T2RdRnRmW_IDSPins smulbt_t
$FB10:F020 T2RdRnRmW_IDSPins smultb_t
$FB10:F030 T2RdRnRmW_IDSPins smultt_t

$FB80:0000 LMUL4ins smull_t

$FB30:F000 T2RdRnRmW_IDSPins smulwb_t
$FB30:F010 T2RdRnRmW_IDSPins smulwt_t
$FB40:F000 T2RdRnRmW_IDSPins smusd_t
$FB40:F010 T2RdRnRmW_IDSPins smusdx_t

$F300:0000 xsat_Tins ssat_t	\ imm5 [shift] --

$F320:0000 xsat16_IDSPins ssat16_t
$FAE0:F000 T2RdRnRmW_IDSPins ssax_t
$FAD0:F000 T2RdRnRmW_IDSPins ssub16_t
$FAC0:F000 T2RdRnRmW_IDSPins ssub8_t

$EC00:0000 ldc_Tins STC_t
$EC40:0000 ldc_Tins STCL_t
$FC00:0000 ldc_Tins STC2_t
$FC40:0000 ldc_Tins STC2L_t

: stm_t		\ --
  0 ?prefix drop
  Thumb1? if
    $C000 T1ldm/stm exit
  endif
  ldm-mask @ $FF00 and 0= if
    <w> @ 1stReg loReg? and if
      $C000 T1ldm/stm exit
    endif
  endif
  Not-M0/M1				\ SFP006
  $E880:0000 T2ldm/stm32
;
: stmia_t  stm_t  ;
: stmea_t  stm_t  ;

: stmdb_t	\ --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFPxx
  $E900:0000 T2ldm/stm32
;
: stmfd_t  stmdb_t  ;

: str_t		\ [imm] --
  0 ?prefix drop
  just2Tregs? if
    imm-flag on  0
  endif
  NarrowOk? if
    just3Tregs? if			\ SFP009
      3LoRegs? if
        $5000 T1memF2 exit		\ str rd, [ rn ++ rm ]
      endif
    endif
    2Tregs+imm? if
      <p> @ if				\ must be preindex.
        1stReg loReg? if		\ Rd must be lo
          dup Imm8*4? if
            peek-reg #13 =		\ str rd, [ SP, # imm8 ]
            if  $9000 T1memF4 exit  endif
          endif
          2ndReg loReg? if
            dup 4 Imm5? if
              4 $6000 T1memF1 exit	\ str rd, [ rn, # imm5 ]
            endif
          endif
        endif
      endif
    endif
  endif
  Not-M0/M1				\ SFP006
  WideOk? if
    3Tregs? if
      $F840:0000 +ShiftInd		\ opc + shift
      +T2RdRn 3rdReg or			\ Rd, Rn, Rm
      T2opc32, exit
    endif
    2Tregs+imm? <p> @ and <w> @ 0= and if	\ imm -- ; preindexed only
      dup #4096 u< if
        $F8C0:0000 +T2memIndP12, exit
      endif
    endif
    2Tregs? indexing @ and if		\ -- [imm] ; Reg+offset
      $F840:0800 +T2memInd8, exit
    endif
  endif
  BadTins
;

: strb_t	\ [imm] --
  0 ?prefix drop
  just2Tregs? if
    imm-flag on  0
  endif
  NarrowOk? if
    <p> @ if
      just3Tregs? 3loRegs? and		\ SFP009
      if  $5400 T1memF2 exit  endif	\ strb rd, [ rn ++ rm ]
      2Tregs+imm? 2loRegs? and if
        dup $20 u< if
          1 $7000 T1memF1 exit		\ strb rd, [ rn, # imm5 ]
        endif
      endif
    endif
  endif
  Not-M0/M1				\ SFP006
  WideOk? if
    3Tregs? if
      $F800:0000 +ShiftInd		\ opc + shift
      +T2RdRn 3rdReg or			\ Rd, Rn, Rm
      T2opc32, exit
    endif
    2Tregs+imm? <p> @ and <w> @ 0= and if	\ imm -- ; preindexed only
      dup #4096 u< if
        $F880:0000 +T2memIndP12, exit
      endif
    endif
    2Tregs? indexing @ and if		\ -- [imm] ; Reg+offset
      $F800:0800 +T2memInd8, exit
    endif
  endif
  BadTins
;

: strbt_t	\ [imm] --
  0 ?prefix drop  Not-M0/M1		\ SFP006
  2Tregs? if
    $F800:0E00 +T2memIndP8T, exit
  endif
  BadTins
;

: strd_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  3Tregs? if
    MustBeIndexed
   $E840:0000 +offS8L			\ Add 8 bit offset to long-word
    <p> @ $0100:0000 and  or		\ Preindexed
    <u> @ $0080:0000 and  or		\ merge Up/Add
    <w> @ $0020:0000 and  or		\ Write back
    1stReg #12 lshift or		\ Rt
    2ndReg #8 lshift or			\ Rt2
    3rdReg #16 lshift or		\ Rn
    T2opc32, exit
  endif
  2Tregs? if				\ -- label
    $E840:0000 $010F:0000 or		\ treat as pre-indexed PC relative
    swap +Trel8PCl T2opc32, exit
  endif
  BadTins
;

: strex_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1 MustBeIndexed	\ SFP006
  3Tregs? if				\ -- [imm]
    $E840:0000
    last-reg #16 lshift or		\ Rn
    last-reg ?BadReg #12 lshift or	\ Rt
    last-Reg ?BadReg 8 lshift or	\ Rd
    +offS8L T2opc32, exit
  endif
  BadTins
;

: strexb_t	\ [0] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  3Tregs? if
    0IndexOnly  $E8C0:0F40
    last-reg #16 lshift or		\ Rn
    last-reg ?BadReg #12 lshift or	\ Rt
    last-Reg ?BadReg or			\ Rd
    T2opc32, exit
  endif
  BadTins
;

: strexh_t	\ [0] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  3Tregs? if
    0IndexOnly  $E8C0:0F50
    last-reg #16 lshift or		\ Rn
    last-reg ?BadReg #12 lshift or	\ Rt
    last-Reg ?BadReg or			\ Rd
    T2opc32, exit
  endif
  BadTins
;

: strh_t	\ [imm] --
  0 ?prefix drop  ?0offset		\ SFP006
  NarrowOk? if
    <p> @ if				\ must be preindexed
      just3Tregs? 3LoRegs? and if	\ SFP009
        $5200 T1memF2 exit		\ strb rd, [ rn ++ rm ]
      endif
      2Tregs+imm? 2loregs? and if
        dup 2 Imm5? if
          2 $8000 T1memF1 exit		\ strb rd, [ rn, # imm5 ]
        endif
      endif
    endif
  endif
  Not-M0/M1				\ SFP006
  WideOk? if
    indexing @ if			\ not a label
      2Tregs+imm? if			\ two register form ; SFP006
        dup #4096 u< if			\ use 12 bit form
          <p> @ <w> @ 0= and if
            $F8A0:0000 +T2memIndP12, exit
          endif
        endif
        $F820:0800 +T2memInd8, exit
      endif
      3Tregs? if			\ three register form
        $F820:0000 +ShiftInd		\ opc + shift
        +T2RdRn 3rdReg or		\ Rd, Rn, Rm
        T2opc32, exit
      endif
    endif
  endif
  BadTins
;

: strht_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  2Tregs? if
    $F820:0E00 +T2memIndP8T, exit
  endif
  BadTins
;

: strt_t	\ [imm] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  2Tregs? if
    $F840:0E00 +T2memIndP8T, exit
  endif
  BadTins
;

: sub_t		\ [imm] --
  0 ?prefix drop  T1>UAL3
\ cr ." SUB_T " reg-count ?  imm? .  s/it? . 1stReg .  2ndReg . 3rdReg .
\ Check narrow forms
  NarrowOK? if
    2Tregs+imm? if
      <s> @ 0= if
        dup Imm7*4?  SP,SP,? and if
          $B080 T1dpiF7 exit		\ sub SP, SP, # imm7*4 ; A6-248 T1
        endif
      endif
      2loregs? s/it? and if
        dup 0 7 within? if
          $1E00 T1dpiF2 exit		\ sub Rd, Rn, # imm3 ; A6-244 T1
        endif
        peek-reg peek-prevReg = if
          dup 0 #255 within? if
            drop-reg
            $3800 T1dpiF3 exit		\ sub rd, rd, # imm8 ; A6-244 T2
          endif
        endif
      endif
    endif
    just3Tregs? 3loRegs? and if		\ SFP009
      $1A00 T1dpiF1  exit		\ sub Rd, Rn, Rm ; A6-246 T1
    endif
  endif
  Thumb1? NarrowOnly? or Cortex-M0/M1? or
  if  BadTins  endif
  2Tregs+Imm? if
    2ndReg #13 = if			\ can use R13 as dest
      <s> @ 0= if
        dup T2lit12? if			\ Enc T3: Must be checked before Const
          $F2A0:0000 T2SRdRn#Lit138 exit  \ sub Rd, Rn, # lit12 ; A6-244 T4
        endif
      endif
      dup T2Const12? if			\ Enc T2:
        $F1A0:0000 T2SRdRn#Const138  exit \ sub Rd, Rn, # const12 ; A6-244 T3
      endif
    else				\ cannot use R13 as dest
      <s> @ 0= if
        dup T2lit12? if			\ Enc T3: Must be checked before Const
          ?BadRdSP/PC-M4M7
          $F2A0:0000 T2SRdRn#Lit138 exit	\ sub Rd, Rn, # lit12 ; A6-244 T4
        endif
      endif
      dup T2Const12? if			\ Enc
        ?BadRdSP/PC-M4M7
        $F1A0:0000 T2SRdRn#Const138  exit	\ sub Rd, Rn, # const12 ; A6-244 T3
      endif
    then
  endif
  3Tregs? if
    $EBA0:0000 T2SRdRnRmsh  exit
  endif
  badTins
;

: swi_t		\ imm --
  0 ?prefix drop
  $FF and  $DF00 or  Topc16,
;
: svc_t		\ imm --
  swi_t  ;

$FA40:F080 xxtax_IDSPins sxtab_t
$FA20:F080 xxtax_IDSPins sxtab16_t
$FA00:F080 xxtax_IDSPins sxtah_t

$B240 $FA4F:F080 xxtx_Tins sxtb_t	\ [rotn] --

$FA2F:F080 xxtax2_IDSPins sxtb16_t

$B200 $FA0F:F080 xxtx_Tins sxth_t	\ [rotn] --

$E8D0:F000 tbx_Tins tbb_t
$E8D0:F010 tbx_Tins tbh_t

: teq_t		\ [shift] --
  0 ?prefix drop  Thumb2only Not-M0/M1	\ SFP006
  1Treg+imm? if
    $F090:0F00 T2Rn#Const138 exit
  endif
  2Tregs? if
    shiftop @ 0=
    if  setLSL 0  endif
    $EA90:0F00 T2RnRmSh exit
  endif
  BadTins
;

: tst_t		\ --
  0 ?prefix drop
  just2Tregs? 2loRegs? and shiftop @ 0= and if
    $4200 T1dpiF5 exit
  endif
  Thumb1? NarrowOnly? or Cortex-M0/M1? or	\ SFP006
  if  BadTins  endif
  1Treg+imm? if
    $F010:0F00 T2Rn#Const138 exit
  endif
  2Tregs? if
    shiftop @ 0=
    if  setLSL 0  endif
    $EA10:0F00 T2RnRmSh exit
  endif
  BadTins
;

$FA90:F040 T2RdRnRmW_IDSPins uadd16_t
$FA80:F040 T2RdRnRmW_IDSPins uadd8_t
$FAA0:F040 T2RdRnRmW_IDSPins uasx_t

: ubfx_t	\ #lsb #width --
  0 ?prefix drop Thumb2only Not-M0/M1	\ SFP006
  $F3C0:0000 T2xbfxRdRnLsbWid
;

: udiv_t	\ --
  0 ?prefix drop Thumb2only Not-M0/M1	\ SFP006
  $FBB0:F0F0 +3regsNB T2opc32,
;

$FA90:F060 T2RdRnRmW_IDSPins uhadd16_t
$FA80:F060 T2RdRnRmW_IDSPins uahdd8_t
$FAA0:F060 T2RdRnRmW_IDSPins uhasx_t
$FAE0:F060 T2RdRnRmW_IDSPins uhsax_t
$FAD0:F060 T2RdRnRmW_IDSPins uhsub16_t
$FAC0:F060 T2RdRnRmW_IDSPins uhsub8_t
$FBE0:0060 MUL4lohi_IDSPins umaal_t

$FBE0:0000 LMUL4ins umlal_t
$FBA0:0000 LMUL4ins umull_t

$FA90:F050 T2RdRnRmW_IDSPins uqadd16_t
$FA80:F050 T2RdRnRmW_IDSPins uqadd8_t
$FAA0:F050 T2RdRnRmW_IDSPins uqasx_t
$FAE0:F050 T2RdRnRmW_IDSPins uqsax_t
$FAD0:F050 T2RdRnRmW_IDSPins uqsub16_t
$FAC0:F050 T2RdRnRmW_IDSPins uqsub8_t
$FB70:F000 T2RdRnRmW_IDSPins usad8_t
$FB70:0000 HMUL4_IDSPins usada8_t

$F380:0000 xsat_Tins usat_t
$F3A0:0000 xsat16_IDSPins usat16_t
$FAE0:F040 T2RdRnRmW_IDSPins usax_t
$FAD0:F040 T2RdRnRmW_IDSPins usub16_t
$FAC0:F040 T2RdRnRmW_IDSPins usub8_t
$FA50:F080 xxtax_IDSPins uxtab_t
$FA30:F080 xxtax_IDSPins uxtab16_t
$FA10:F080 xxtax_IDSPins uxtah_t
$FA3F:F080 xxtax2_IDSPins uxtb16_t

$B2C0 $FA5F:F080 xxtx_Tins uxtb_t	\ [rotn] --
$B280 $FA1F:F080 xxtx_Tins uxth_t	\ [rotn] --

$FFB1:0300 $EEB0:0AC0 T2_S/FP_VdVm_ins VABS_t	\ M4 version - FP only
$EF00:0800 $EF00:0D00 $EE30:0A00 T2_S/FP_VdVnVm_ins VADD_t

$EEB4:0A40 T2_VCMPx_ins VCMP_t
$EEB4:0AC0 T2_VCMPx_ins VCMPE_t

\ VCVTx elsewhere

$EE80:0A00 T2_FP_VdVnVm_ins VDIV_t

$EEA0:0A00 T2_FP_VdVnVm_ins VFMA_t
$EEA0:0A40 T2_FP_VdVnVm_ins VFMS_t
$EE90:0A00 T2_FP_VdVnVm_ins VFNMA_t
$EE90:0A40 T2_FP_VdVnVm_ins VFNMS_t
$EC90:0A00 T2_Vldm/stm_ins VLDMIA_t	\ PU=01=IA
$ED10:0A00 T2_Vldm/stm_ins VLDMDB_t	\ PU=10=DB
$ED10:0A00 T2_Vldr/str_ins VLDR_t
$EE00:0A00 T2_FP_VdVnVm_ins VMLA_t
$EE00:0A40 T2_FP_VdVnVm_ins VMLS_t
\ VMOV elsewhere
\ VMRS elswhere
\ VMSR elsewhere
$EF00:0910 $EF00:0D10 $EE20:0A00 T2_S/FP_VdVnVm_ins VMUL_t

$FFB1:0380 $EEB1:0A40 T2_S/FP_VdVm_ins VNEG_t
$EE10:0A40 T2_FP_VdVnVm_ins VNMLA_t
$EE10:0A00 T2_FP_VdVnVm_ins VNMLS_t
$EE20:0A40 T2_FP_VdVnVm_ins VNMUL_t
$ECBD:0A00 T2_Vpop/push_ins VPOP_t
$ED2D:0A00 T2_Vpop/push_ins VPUSH_t
$EEB1:0AC0 T2_FP_VdVm_ins VSQRT_t
$EC80:0B00 T2_Vldm/stm_ins VSTMIA_t	\ PU=01
$ED00:0B00 T2_Vldm/stm_ins VSTMDB_t	\ PU=10
$ED00:0A00 T2_Vldr/str_ins VSTR_t
$FF00:0800 $EF20:0D00 $EE30:0A40 T2_S/FP_VdVnVm_ins VSUB_t
\ $EE30:0A40 T2_FP_VdVnVm_ins VSUB_t

$BF20 $F3AF:8002 hint16/32 wfe_t
$BF30 $F3AF:8003 hint16/32 wfi_t
$BF10 $F3AF:8001 hint16/32 yield_t


\ ****************
\ ARM instructions
\ ****************

\ ****************************
\ Data processing instructions
\ ****************************
\ e.g. ADC r0, r1, r2
\      ADC r0, r1, # 5
\      ADC r0, r1, r2 lsl r3
\      ADC r0, r1, r2 lsl # 5

: findshift     \ imm -- imm' n
  #16 0 do
    dup $FFFFFF00 and 0=                \ value fits in low 8 bits?
    if  #16 i - $0F and                 \ shift left = 16 - r-shift#
        unloop  exit                    \ then get out
    endif
    2 lror				\ circular shift right
  loop
  err_immval a-error
;

: imm12?        \ imm -- flag ; true if immediate value fits in 12 bits
  #16 0 do
    dup $FFFFFF00 and 0=                \ value fits in low 8 bits?
    if  drop true unloop exit  endif    \ then get out
    2 lror				\ circular shift right
  loop
  drop false
;

: +immop        \ n opcode -- opcode'
  $02000000 or                          \ set I bit
  swap findshift  8 lshift or  or  ;	\ merge in rrx/imm

: check-shift#  \ n -- n'
  dup 0 #32 within? 0= err_immshift ?a-error
  $01F and  ;

: +shiftop      \ [n] opcode -- opcode' ; deal with shifts
  shiftop @ $FFF and or                 \ merge shift type
  imm-flag @                            \ true if immediate shift
  if
    swap check-shift# 7 lshift or       \ merge immediate shift value
  else
    last-reg 8 lshift or                \ merge shift register
    $010 or                             \ merge register shift bit
  endif
  last-reg or                           \ merge Rm
;

: +operand2     \ [n] opcode -- opcode'
  shiftop @                           \ shift required?
  if
    +shiftop
  else
    imm-flag @                          \ immediate?
    if                                  \ y:
      +immop
    else                                \ n:
      last-reg or                       \ merge Rm
    endif
  endif
;

: dpins         \ minorop -- ; [n] -- ; data processing instructions
  create
    #21 lshift ,
  does> ?prefix
    @ +cond +sbit20
    +operand2                           \ operand 2
    last-reg #16 lshift or              \ Rn
    last-reg #12 lshift or              \ Rd
    l,c(t)  a-reset
;

: cmpins        \ minorop -- ; [n] -- ; compare instructions
  create
    #21 lshift ,
  does> ?prefix
    @ +cond  $00100000 or               \ S bit always set
    +psrflags                           \ .P suffix                             \ MRP007
    +operand2                           \ operand 2
    last-reg #16 lshift or              \ Rn
    l,c(t)  a-reset
;

: movins        \ minorop -- ; [n] -- ; move instructions
  create
    #21 lshift ,
  does> ?prefix
    @ +cond +sbit20
    +operand2                           \ operand 2
    last-reg #12 lshift or              \ Rd
    l,c(t)  a-reset
;

: mrsins        \ op -- ; -- ; MRS instruction
  create
    ,
  does> ?prefix
    @ +cond
    last-psr 1 and                      \ ignore flag bit indicator
    0<> $00400000 and or                \ merge into opcode
    last-reg #12 lshift or              \ Rd
    l,c(t)  a-reset
;

: msrins        \ op -- ; -- ; MSR instruction ; SFP016
  create
    ,
  does> ?prefix
    @ +cond  imm-flag @                 \ immediate or register?
    if  +immop  else  last-reg or  endif
    last-psr 0<> $00400000 and or	\ merge CPSR/SPSR into opcode
    msr_mask @ dup 0=			\ merge in new style field mask ; SFP016
    if  drop  $00080000  endif		\ replace "no fields" with _F
    or
    l,c(t)  a-reset
;

: msr_flag	\ -- ; -- ; define field names for MSR instruction
  create
    ,
  does>
    @ msr_mask @ or msr_mask !
;

also asm-access definitions
$00010000 msr_flag _c
$00020000 msr_flag _x
$00040000 msr_flag _s
$00080000 msr_flag _f
$000F0000 msr_flag _cxsf
$000F0000 msr_flag _fsxc
$000F0000 msr_flag _all
previous definitions


\ ***********************
\ Load/Store instructions
\ ***********************

: +immoff12    \ n opcode -- opcode'
\  $02000000 or                         \ set I bit 25
  swap dup abs $FFF u> err_i12 ?a-error \ error if out of range
  dup 0<                                \ check for neg/pos offset
  if
    abs $FFF and  or                    \ make positive, merge
  else
    $FFF and  or                        \ merge
    $00800000 or                        \ set U bit 23
  endif  ;

: +shiftoff     \ [n] opcode -- opcode' ; deal with shifts
  shiftop @ $FFF and or                 \ merge shift type
  imm-flag @ 0= err_shift ?a-error      \ must be immediate shift
  swap check-shift# 7 lshift or         \ merge immediate shift value
  last-reg or                           \ merge Rm
  <u> @ $00800000 and  or               \ set U bit
  $02000000 or                          \ set I bit 25
;

: +index        \ [n] opcode -- opcode'
  indexing @
  if
    <w> @ $00200000 and  or             \ Write back
    <p> @ $01000000 and  or             \ Preindexed
    imm-flag @ 0<>  shiftop @ 0<>  and  \ [Rn, ++ Rm LSL # n] [Rn], -- Rm LSL # n
    if
      +shiftoff
    else
      imm-flag @ 0<>  shiftop @ 0=  and \ [Rn, # n]           [Rn], # n
      if
        +immoff12
      else
        imm-flag @ 0=  shiftop @ 0= and \ [Rn] or [Rn, ++ Rm]
        if
          reg2 @                        \ [Rn ++ Rm]          [Rn], Rm
          if                            \ treat as Rm LSL # 0
            last-reg or                 \ merge Rm
            <u> @ $00800000 and  or     \ set U bit 23
            $02000000 or                \ set I bit 25
          else                          \ [Rn], treat as [Rn, # -0]
            $00800000 or                \ offset already 0 and set U bit 23
          endif
        else
          err_addrmode a-error          \ invalid addressing mode
        endif
      endif
    endif
    last-reg #16 lshift or              \ Rn
  else
    $010F0000 or                        \ force P,R15 = [PC,#(offset-8)]
    swap +rel12                         \ allow forward referencing
  endif
;

: ldrins        \ opcode -- ; [n] -- ; load/store
  create
    ,
  does> ?prefix
    @ +cond
    +index                              \ address
    last-reg #12 lshift or                 \ Rd
    l,c(t)  a-reset
;


\ ****************************
\ LD/ST SB and SH instructions
\ ****************************

: +immoff4/4   \ n opcode -- opcode'
  $00400000 or  set_4/4  ;              \ set I bit 22 , merge offset

: +imm/reg     \ [n] opcode -- opcode '
  indexing @
  if
    <w> @ $00200000 and  or             \ Write back
    <p> @ $01000000 and  or             \ Preindexed
    reg2 @                              \ [Rn, ++ Rm]   [Rn], -- Rm
    if                                  \ [Rn, ++ Rm]
      last-reg or                       \ merge Rm
      <u> @ $00800000 and  or           \ set U bit 23
    else
      imm-flag @                        \ [Rn,#imm]     [Rn],#imm
      if
        +immoff4/4
      else                              \ [Rn]
        $00C00000 or                    \ set U bit 23, and I bit 22 for [Rn,#0]
      endif
    endif
    last-reg #16 lshift or              \ Rn
  else
    $014F0000 or                        \ force I,P,R15 = [PC,#(offset-8)]
    swap +rel4/4                        \ allow forward referencing
  endif
;

: ldrhins       \ opcode -- ; [n] -- ; load/store SBH
  create
    ,
  does>
    ?prefix
    @ +cond
    +imm/reg                            \ address
    last-reg #12 lshift or              \ Rd
    l,c(t)  a-reset
;

: swpins        \ opcode -- ; -- ; SWP and SWPB
  create
    ,
  does> ?prefix
    @ +cond
    last-reg #16 lshift or              \ Rn
    last-reg or                         \ Rm
    last-reg #12 lshift or              \ Rd
    l,c(t)  a-reset
 ;


\ **********************
\ LDM/STMxx instructions
\ **********************

: $exec-reg     \ string -- regdata ; process as assembler register
  dup count upper                       \ force to upper case
  ['] asm-access v-find 0= #-13 ?throw	\ error if not found
  reg-count @ swap                      \ save for checking
  execute
  1+ reg-count @ <> err_notreg ?throw   \ check its a register of some type
  last-regData				\ return register value ; SFP009
;

: compound-reg  \ string -- ; process as R0-R10
  pocket $move                          \ copy away from HERE
  'tib @ >r  #tib @ >r                  \ save i/p stream
  >in @ >r  blk @ >r
  >in off  blk off                      \ set new i/p stream
  pocket count #tib ! 'tib !
  [char] - word $exec-reg               \ first register
  dup ldm-first !  $FF and		\ -- first
  [char] - word $exec-reg $FF and	\ -- first last
  2dup u> err_regorder ?throw           \ check order
  1+  swap                              \ last+1 first
  2dup - ldm-len !  do			\ length of group
    i ldm-first @ $0000:FF00 and (r)	\ add the registers
  loop
  r> blk !  r> >in !                    \ restore i/p stream
  r> #tib !  r> 'tib !
  pocket here $move                     \ copy text back to here
;

also asm-access definitions
: }             \ -- ; exits from the caller
  r> drop  ;
previous definitions

: get-regs      \ -- ; process register list until } found
  begin
    get-token ['] asm-access v-find     \ look for word in voc
    if
      execute                           \ set register
    else
      compound-reg                      \ try as R0-R10, S1-5, D7-20
    endif
  again
;

: set-mask      \ n -- ; set bits in data masks for LDM/STM
  1 swap lshift ldm-mask or!  ;

: set-list	\ rcprev --
\ Generate LDM-MASK and LDM-COUNT. The input is the register count
\ before the list started.
  reg-count @ swap -			\ number of registers defined
  dup 0= err_reglistNull ?a-error	\ 0 is invalid
  dup ldm-count !  0
  do  last-regData $FF and set-mask  loop	\ set bit number of register in mask
\  do  last-reg set-mask  loop		\ set bit number of register in mask
;

: get-list      \ -- ; process register list { rx, ry, ra-rb, rc-rd }
  reg-count @ >r			\ number of registers before list
  get-regs
  r> set-list
;

: store-mask    \ n -- ; set LDM-MASK - needed for macro handling
  ldm-mask !  ;

: +writeback    \ opcode -- opcode'
  <w> @ $00200000 and or  ;

: +ubit23       \ opcode -- opcode'
  <s22> @ $00400000 and or  ;

: ldmins        \ opcode -- ; -- ; LDMxx and STMxx
  create
    ,
  does> ?prefix
  @ +cond +writeback +ubit23
  ldm-mask @ $0FFFF and or            \ mask in register list
  last-reg #16 lshift or              \ RN
  l,c(t)  a-reset
;

\ opcode -- opcode' ; used to build STM/LDM instructions
: +p    $01000000 or  ;
: +u    $00800000 or  ;
: +l    $00100000 or  ;


\ ****************************
\ Coprocessor loads and stores
\ ****************************

: +imm8         \ [n] opcode -- opcode '
  indexing @
  if
    <w> @ $00200000 and  or             \ Write back
    <p> @ $01000000 and  or             \ Preindexed
    imm-flag @                          \ [Rn,#imm]     [Rn],#imm
    if  set_8  endif
    last-reg #16 lshift or		\ Rn
  else
    $010F0000 or                        \ force P,R15 = [PC,#(offset-8)]
    swap +rel8                          \ allow forward referencing
  endif
;

: ldcins        \ opcode -- ; [n] -- ; LDC/STC
  create
    ,
  does> ?prefix
    @ +cond
    +imm8                               \ address
    last-cpr #12 lshift or              \ CRd
    swap $0F and 8 lshift or            \ CP#
    l,c(t)  a-reset
;


\ ************
\ Opcode types
\ ************

: bxins         \ opcode -- ; -- ; for BX instruction
  create
    ,
  does> ?prefix
    @                                   \ opcode
    +cond                               \ add condition code
    last-reg or                         \ add reg #
    l,c(t)
    a-reset
;

: blins         \ opcode -- ; label -- ; for B and BL
  create  ,  does> ?prefix @
  +cond  swap +rel24,  a-reset  ;

: ?r15          \ n -- n ; error if n=15
  dup #15 = err_r15 ?a-error  ;

: blxins	\ opl opr -- ; [label] --
  create  , ,  does> ?prefix
  reg-count @ 0= if
    cell+ @  swap +rel24x,
  else
    @ +cond  last-reg ?r15 or l,c(t)
  endif
  a-reset
;

: mulins        \ opcode -- ; label -- ; for MUL
  create  ,  does> ?prefix
  @ +cond  +sbit20
  last-reg ?r15 8 lshift or             \ Rs
  last-reg ?r15 or                      \ Rm
  last-reg ?r15 #16 lshift or           \ Rd
  l,c(t)  a-reset  ;

: mlains        \ opcode -- ; label -- ; for MLA
  create  ,  does> ?prefix
  @ +cond  +sbit20
  last-reg ?r15 #12 lshift or           \ Rn
  last-reg ?r15 8 lshift or             \ Rs
  last-reg ?r15 or                      \ Rm
  last-reg ?r15 #16 lshift or           \ Rd
  l,c(t) a-reset  ;

: mullins       \ opcode -- ; label -- ; for MULL types
  create  ,  does>
  ?prefix
  @ +cond  +sbit20
  last-reg ?r15 8 lshift or             \ Rs
  last-reg ?r15 or                      \ Rm
  last-reg ?r15 #16 lshift or           \ Rdhi
  last-reg ?r15 #12 lshift or           \ Rdlo
  l,c(t) a-reset  ;

: swiins        \ opcode -- ; swi# -- ; SWI
  create  ,  does> ?prefix
  @ +cond
  swap $00FFFFFF and or
  l,c(t) a-reset
;

: cdpins        \ opcode -- ; cp# cpopc cp -- ; CDP
  create
    ,
  does> ?prefix
    @ +cond
    swap $07 and 5 lshift or            \ CP/exp2
    last-cpr or                         \ CRm
    last-cpr #16 lshift or              \ CRn
    last-cpr #12 lshift or              \ CRd
    swap $0F and #20 lshift or          \ CPOpC
    swap $0F and 8 lshift or            \ cp#
    l,c(t)  a-reset
;

: mrcins        \ opcode -- ; cp# cpopc cp -- ; MRC MCR
  create
    ,
  does> ?prefix
    @ +cond
    swap $07 and 5 lshift or            \ CP/exp2
    last-cpr or				\ CRm
    last-cpr #16 lshift or              \ CRn
    last-reg #12 lshift or              \ Rd
    swap $07 and #21 lshift or          \ CPOpC
    swap $0F and 8 lshift or            \ cp#
    l,c(t)  a-reset
;


\ **************************
\ Architecture v5 extensions
\ **************************

\ SFP018...
: bkptins	\ opcode -- ; #imm -- ; BKPT
  create
    ,
  does> ?prefix
    @ swap				\ -- opcode imm
    dup $0F and				\ bottom four bits
    swap $0FFF0 and 4 lshift		\ next 12 bits
    or or l,c(t)  a-reset
;
\ ...SFP018

: clzIns	\ opcode -- ; -- ; SFP019
  create
    ,
  does> ?prefix				\ CLZ .cc Rd, Rm
    @ +cond				\ .cc
    last-reg or				\ Rm
    last-reg #12 lshift or		\ Rd
    l,c(t)  a-reset
;

: 0opIns	\ opcode -- ; -- ; SFP020
  create
    ,
  does> ?prefix
    @ l,c(t)  a-reset
;


\ ******************
\ Macro instructions
\ ******************

\ MRP002...
: +rdrn         \ opcode rn -- opcode'
\ masks in rn as the rd in the correct place for the
\ orr instructions that make up the ADR instruction
  tuck
  #16 lshift or
  swap #12 lshift or
;

variable adrtotal
  0 adrtotal !

: adrins        \ -- ; label -- ; for ADR instruction - builds 4 instructions ( mov + 3 orrs) for full 32bit offsets
  create
  does> ?prefix
    drop
    last-reg >r                         \ get destination reg.
    0 +cond +sbit20			\ get condition code and set status flags if necessary.
    swap check-adr			\ set up forward ref if needed
    over $03A00000 or r@ #12 lshift or over $FF and or l,c(t)
    over $03800C00 or r@ +rdrn over #8 rshift  $FF and or l,c(t)
    over $03800800 or r@ +rdrn over #16 rshift $FF and or l,c(t)
    swap $03800400 or r> +rdrn swap #24 rshift $FF and or l,c(t)
adrtotal incr
    a-reset
;
\ ...MRP002

variable mvlcount	\ number of bytes used by multiple ins. MVLs
  0 mvlcount !
variable mvltotal	\ number of multiple ins. MVLs
  0 mvltotal !
variable mvl1total	\ number of single ins. MVLs
  0 mvl1total !
variable mvltemp	\ scratch

: (mvl32)	\ opbase val reg --
chere(t) mvltemp !
  >r
  \ -- opbase val
					\ lay MOV reg, lit7:0
  over $03A00000 or  r@ #12 lshift or  over $FF and or l,c(t)
  dup $0FF00 and if			\ check need ORR reg, reg, # lit15:8
    over $03800C00 or  r@ +rdrn  over #8 rshift $FF and or l,c(t)
  endif
  dup $0FF0000 and if			\ check need ORR reg, reg, # lit23:16
    over $03800800 or  r@ +rdrn  over #16 rshift $FF and or l,c(t)
  endif
  dup $FF000000 and if			\ check need ORR reg, reg, # lit31:24
    over $03800400 or  r@ +rdrn over #24 rshift $FF and or l,c(t)
  endif
  2drop  r> drop
chere(t) mvltemp @ - mvlcount +!
mvltotal incr
;

\ MRP004...
: mvlins        \ -- ; val -- ; for mpe mvl instruction
\ builds upto 4 instructions ( mov + 3 orrs) for full 32bit moves
  create
  does> ?prefix
    drop
    last-reg >r                         \ get destination reg.
    0 +cond +sbit20 swap                \ -- op val
    dup imm12? if                          \ fits in 12 bits?
      swap $03A00000 or  r@ #12 lshift or  \ -- val op
      swap findshift 8 lshift or or l,c(t)
mvl1total incr
    else
      dup 0< over 1+ abs imm12? and if  \ if -ve and val+1 fits in 12bits, use single MVN       \ MRP012
        1+ abs                                                                                  \ MRP012
        swap $03E00000 or  r@ #12 lshift or \ -- val op                                              \ MRP012
        swap findshift 8 lshift or  or  l,c(t)                                                       \ MRP012
mvl1total incr
      else
        r@ (mvl32)			\ four instruction MVL
      endif
    endif
    r> drop  a-reset
;
\ ...MRP004
$5 dpins ADC_a
$4 dpins ADD_A
adrins ADR_A
$012FFF10 bxins BX_A
$0A000000 blins B_a
$0B000000 blins BL_a
$FA000000 $012FFF30 blxins blx_a
$0E000000 cdpins CDP_a
$0 dpins AND_a
$1 dpins EOR_a
$E1200070 bkptins BKPT_A		\ SFP018
$A cmpins CMP_A
$B cmpins CMN_A
$F57F:F050 hint32_insA DMB_a
$F57F:F040 hint32_insA DSB_a
$F57F:F060 hint32_insA ISB_a
$C dpins ORR_A
$D movins MOV_A
$E dpins BIC_A
$0C100000 ldcins LDC_a
$0C500000 ldcins LDCL_a
$08000000
dup +l +p +u    dup ldmins LDMED_a      ldmins LDMIB_a
dup +l    +u    dup ldmins LDMFD_a	ldmins LDMIA_A
dup +l +p       dup ldmins LDMEA_a      ldmins LDMDB_a
dup +l          dup ldmins LDMFA_a      ldmins LDMDA_a
dup    +p +u    dup ldmins STMFA_a      ldmins STMIB_a
dup       +u    dup ldmins STMEA_a      ldmins STMIA_A
dup    +p       dup ldmins STMFD_a	ldmins STMDB_a
dup             dup ldmins STMED_a      ldmins STMDA_a
drop

$00200090 mlains MLA_a
$0E000010 mrcins MCR_a
$0E100010 mrcins MRC_a
$010F0000 mrsins MRS_a
$0120F000 msrins MSR_a
$F movins MVN_A
$016F0F10 clzIns CLZ_A			\ SFP019
$E1A00000 0opIns NOP_A			\ SFP020
$00000090 mulins MUL_A
mvlins MVL_a
: mvl32_a	\ imm32 --
  $E000:0000 ?prefix swap last-reg (mvl32) a-reset  ;
$04100000 ldrins LDR_A
$04000000 ldrins STR_A
$04500000 ldrins LDRB_A
$3 dpins RSB_a
$7 dpins RSC_a
$04400000 ldrins STRB_A
$000000B0 ldrhins STRH_A
$001000B0 ldrhins LDRH_A
$001000D0 ldrhins LDRSB_A
$001000F0 ldrhins LDRSH_A
$00800090 mullins UMULL_a
$00A00090 mullins UMLAL_a
$00C00090 mullins SMULL_a
$00E00090 mullins SMLAL_a
$0C000000 ldcins STC_a
$0C400000 ldcins STCL_a
$6 dpins SBC_A
$2 dpins SUB_A
$01000090 swpins SWP_a
$01400090 swpins SWPB_a
$9 cmpins TEQ_a
$8 cmpins TST_A
$0F000000 swiins SWI_A

0 shift LSL_A
1 shift LSR_A
2 shift ASR_A
3 shift ROR_A

$F3B1:0300 $0EB0:0AC0 ARM32_S/FP_VdVm_ins VABS_a
$F200:0800 $F200:0D00 $0E30:0A00 ARM32_S/FP_VdVnVm_ins VADD_a

$0EB4:0A40 ARM32_VCMPx_ins VCMP_a
$0EB4:0AC0 ARM32_VCMPx_ins VCMPE_a
\ VCVTx elsewhere

$0E80:0A00 ARM32_FP_VdVnVm_ins VDIV_a

$0EA0:0A00 ARM32_FP_VdVnVm_ins VFMA_a
$0EA0:0A40 ARM32_FP_VdVnVm_ins VFMS_a
$0E90:0A00 ARM32_FP_VdVnVm_ins VFNMA_a
$0E90:0A40 ARM32_FP_VdVnVm_ins VFNMS_a

$0C90:0A00 ARM32_Vldm/stm_ins VLDMIA_a	\ PU=01=IA
$0D10:0A00 ARM32_Vldm/stm_ins VLDMDB_a	\ PU=10=DB
$0D10:0A00 ARM32_Vldr/str_ins VLDR_a

$0E00:0A00 ARM32_FP_VdVnVm_ins VMLA_a
$0E00:0A40 ARM32_FP_VdVnVm_ins VMLS_a
$F200:0910 $F300:0D10 $0E20:0A00 ARM32_S/FP_VdVnVm_ins VMUL_a

$F3B1:0380 $0EB1:0A40 ARM32_S/FP_VdVm_ins VNEG_a
$0E10:0A40 ARM32_FP_VdVnVm_ins VNMLA_a
$0E10:0A00 ARM32_FP_VdVnVm_ins VNMLS_a
$0E20:0A40 ARM32_FP_VdVnVm_ins VNMUL_a

$0CBD:0A00 ARM32_Vpop/push_ins VPOP_a
$0D2D:0A00 ARM32_Vpop/push_ins VPUSH_a

$0EB1:0AC0 ARM32_FP_VdVm_ins VSQRT_a
$0C80:0A00 ARM32_Vldm/stm_ins VSTMIA_a	\ PU=01=IA
$0D00:0A00 ARM32_Vldm/stm_ins VSTMDB_a	\ PU=10=DB
$0D00:0A00 ARM32_Vldr/str_ins VSTR_a
$F300:0800 $F220:0D00 $0E30:0A40 ARM32_S/FP_VdVnVm_ins VSUB_a

also asm-access definitions  hex

: #             \ -- ; immediate operand marker
  imm-flag on  ;
: x#		\ -- ; immediate operand marker
  imm-flag on  ;
: @=		\ -- ; literal pool load marker
  @=-flag on  ;

: adc	0 if adc_t else adc_a endif  ;
: add	Thumb? if add_t else add_a endif  ;
: addx	Thumb? if add_t else add_a endif  ;
: adr	Thumb? if adr_t else adr_a endif  ;
: and	Thumb? if and_t else and_a endif  ;
: andx	Thumb? if and_t else and_a endif  ;	\ avoid name clash
: asl	Thumb? if lsl_t else lsl_a endif  ;
: asr	Thumb? if asr_t else asr_a endif  ;
: b	Thumb? if b_t else b_a endif  ;
: bfc	NotARM bfc_t  ;
: bfi   NotARM bfi_t  ;
: bic   Thumb? if bic_t else bic_a endif  ;
: bkpt  Thumb? if bkpt_t else bkpt_a endif  ;
: bl	Thumb? if bl_t else bl_a endif  ;
: blxx	Thumb? if bl_t else bl_a endif  ;	\ avoid name clash with BL = SPACE
: blx	Thumb? if blx_t else blx_a endif  ;
: bx	Thumb? if bx_t else bx_a endif  ;
: cbnz	NotARM cbnz_t  ;
: cbz   NotARM cbz_t  ;
: cdp	Thumb? if cdp_t else cdp_a endif  ;
: cdp2	NotARM cdp2_t  ;
: clrex	NotARM clrex_t  ;
: clz	thumb? if clz_t else clz_a endif  ;
: cmn	Thumb? if cmn_t else cmn_a endif  ;
: cmp	Thumb? if cmp_t else cmp_a endif  ;
: cps   NotARM cps_t  ;
: cpy   NotARM cpy_t  ;
: dbg   NotARM  DBG_T  ;
: dmb   Thumb? if DMB_T else DMB_A endif  ;
: dsb   Thumb? if DSB_T else DSB_A endif  ;
: eor	Thumb? if eor_t else eor_a endif  ;
: isb   Thumb? if ISB_T else ISB_A endif  ;
: IT    NotARM IT_t  ;
: ITT   NotARM ITT_t  ;
: ITE   NotARM ITE_t  ;
: ITTT  NotARM ITTT_t  ;
: ITET  NotARM ITET_t  ;
: ITTE  NotARM ITTE_t  ;
: ITEE  NotARM ITEE_t  ;
: ITTTT NotARM ITTTT_t  ;
: ITETT NotARM ITETT_t  ;
: ITTET NotARM ITTET_t  ;
: ITEET NotARM ITEET_t  ;
: ITTTE NotARM ITTTE_t  ;
: ITETE NotARM ITETE_t  ;
: ITTEE NotARM ITTEE_t  ;
: ITEEE NotARM ITEEE_t  ;
: ldc   Thumb? if ldc_t else ldc_a endif  ;
: ldc2  NotARM ldc2_t  ;
: ldcl  Thumb? if ldcl_t else ldcl_a endif  ;
: ldc2l NotARM ldc2l_t  ;

: ldm	NotARM ldm_t  ;
: ldmda NotThumb LDMDA_a  ;
: ldmdb Thumb? if ldmdb_t else LDMDB_a endif  ;
: ldmea	Thumb? if ldmea_t else LDMEA_a endif  ;
: ldmed	NotThumb LDMED_a  ;
: ldmfa NotThumb LDMFA_a  ;
: ldmfd Thumb? if ldmfd_t else ldmfd_a endif  ;
: ldmia Thumb? if ldmia_t else ldmia_a endif  ;
: ldmib	NotThumb LDMIB_a  ;

: ldr	Thumb? if ldr_t else ldr_a endif  ;
: ldrb	Thumb? if ldrb_t else ldrb_a endif  ;
: ldrbt NotARM ldrbt_t  ;
: ldrd	NotARM ldrd_t  ;
: ldrex NotARM ldrex_t  ;
: ldrexb NotARM ldrexb_t  ;
: ldrexh NotARM ldrexh_t  ;
: ldrh	Thumb? if ldrh_t else ldrh_a endif  ;
: ldrht NotARM ldrht_t  ;
: ldrsb	Thumb? if ldrsb_t else ldrsb_a endif  ;
: ldrsbt NotARM ldrsbt_t  ;
: ldrsh	Thumb? if ldrsh_t else ldrsh_a endif  ;
: ldrsht NotARM ldrsht_t  ;
: ldrt  NotARM ldrt_t  ;
: lsl	Thumb? if lsl_t else lsl_a endif  ;
: lsr	Thumb? if lsr_t else lsr_a endif  ;

: mcr   Thumb? if mcr_t else mcr_a endif  ;
: mcr2  NotARM mcr2_t  ;
: mcrr  NotARM mcrr_t  ;
: mcrr2 NotARM mcrr2_t  ;
: mla	Thumb? if mla_t else mla_a endif  ;
: mls   NotARM mls_t  ;
: mov	Thumb? if mov_t else mov_a endif  ;
: movt	NotARM movt_t  ;
: mrc   Thumb? if mrc_t else mrc_a endif  ;
: mrc2  NotARM mrc2_t  ;
: mrrc  NotARM mrrc_t  ;
: mrrc2 NotARM mrrc2_t  ;
: mrs   Thumb? if mrs_t else mrs_a endif  ;
: msr   Thumb? if msr_t else msr_a endif  ;
: mul	Thumb? if mul_t else mul_a endif  ;
: mvl	Thumb? if mvl_t else mvl_a endif  ;
: mvl32 Thumb? if mvl32_t else mvl32_a endif ;
: mvl32+1 NotARM mvl32+1_t ;
: mvn	Thumb? if mvn_t else mvn_a endif  ;

: neg	NotARM neg_t  ;
: nop	Thumb? if nop_t else nop_a endif  ;

: orn	NotARM orn_t  ;
: orr	Thumb? if orr_t else orr_a endif  ;

: pkhbt	NotARM pkhbt_t  ;
: pkhtb	NotARM pkhtb_t  ;

: pld	NotARM pld_t  ;
: pldw	NotARM pldw_t  ;
: pli	NotARM pli_t  ;
: pop	NotARM pop_t  ;
: push	NotARM push_t  ;

: qadd		NotARM qadd_t  ;
: qadd16	NotARM qadd16_t  ;
: qadd8		NotARM qadd8_t  ;
: qasx		NotARM qasx_t  ;
: qdadd		NotARM qdadd_t  ;
: qdsub		NotARM qdsub_t	;
: qsax		NotARM qsax_t  ;
: qsub		NotARM qsub_t  ;
: qsub16	NotARM qsub16_t  ;
: qsub8		NotARM qsub8_t  ;

: rbit	NotARM rbit_t  ;
: rev	NotARM rev_t  ;
: rev16	NotARM rev16_t  ;
: revsh	NotARM revsh_t  ;
: ror	Thumb? if ror_t else ror_a endif  ;
: rorx	Thumb? if ror_t else ror_a endif  ;	\ avoid name clash
: RRX	Thumb? if  rrx_t exit  endif  ROR_a # 0  ;
: rsb	Thumb? if rsb_t else rsb_a endif  ;
: rsc   NotThumb rsc_a  ;

: sadd16	NotARM sadd16_t  ;
: sadd8		NotARM sadd8_t  ;
: sasx		NotARM sasx_t  ;

: sbc	Thumb? if sbc_t else sbc_a endif  ;
: sbfx	NotARM sbfx_t  ;
: sdiv	NotARM sdiv_t  ;
: sel	NotARM sel_t  ;
: sev	NotARM sev_t  ;

: shadd16	NotARM shadd16_t  ;
: shadd8	NotARM shadd8_t  ;
: shasx		NotARM shasx_t	;
: shsax		NotARM shsax_t  ;
: shsub16	NotARM shsub16_t  ;
: shsub8	NotARM shsub8_t  ;

: smlal	Thumb? if smlal_t else smlal_a endif  ;

: smlabb	NotARM smlabb_t  ;
: smlabt	NotARM smlabt_t  ;
: smlatb	NotARM smlatb_t  ;
: smlatt	NotARM smlatt_t  ;

: smlald	NotARM smlald_t  ;
: smlaldx	NotARM smlaldx_t  ;

: smlawb	NotARM smlawb_t  ;
: smlawt	NotARM smlawt_t  ;
: smlsd		NotARM smlsd_t  ;
: smlsdx	NotARM smlsdx_t  ;
: smlsld	NotARM smlsld_t  ;
: smlsldx	NotARM smlsldx_t  ;
: smmla		NotARM smmla_t  ;
: smmlar	NotARM smmlar_t  ;
: smmls		NotARM smmls_t  ;
: smmlsr	NotARM smmlsr_t  ;

: smmul		NotARM smmul_t  ;
: smmulr	NotARM smmulr_t  ;
: smuad		NotARM smuad_t  ;
: smuadx	NotARM smuadx_t  ;
: smulbb	NotARM smulbb_t  ;
: smulbt	NotARM smulbt_t  ;
: smultb	NotARM smultb_t  ;
: smultt	NotARM smultt_t  ;

: smull	Thumb? if smull_t else smull_a endif  ;

: smulwb	NotARM smulwb_t  ;
: smulwt	NotARM smulwt_t  ;
: smusd		NotARM smusd_t  ;
: smusdx	NotARM smusdx_t  ;

: ssat		NotARM ssat_t  ;

: ssat16	NotARM ssat16_t  ;
: ssax		NotARM ssax_t  ;
: ssub16	NotARM ssub16_t  ;
: ssub8		NotARM ssub8_t  ;

: stc	Thumb? if stc_t else stc_a endif  ;
: stc2  NotARM stc2_t  ;
: stc2l NotARM stc2l_t  ;
: stcl  NotARM stcl_t  ;

: stm	NotARM stm_t  ;
: stmda NotThumb STMDA_a  ;
: stmdb Thumb? if stmdb_t else STMDB_a endif  ;
: stmea	Thumb? if stmea_t else STMEA_a endif  ;
: stmed	NotThumb STMED_a  ;
: stmfa NotThumb STMFA_a  ;
: stmfd Thumb? if stmfd_t else stmfd_a endif  ;
: stmia Thumb? if stmia_t else stmia_a endif  ;
: stmib	NotThumb STMIB_a  ;

: str	Thumb? if str_t else str_a endif  ;
: strb	Thumb? if strb_t else strb_a endif  ;
: strbt	NotARM strbt_t  ;		\ ARM too!
: strd	NotARM strd_t  ;
: strex	NotARM strex_t  ;
: strexb  NotARM strexb_t  ;
: strexh  NotARM strexh_t  ;
: strh	Thumb? if strh_t else strh_a endif  ;
: strht	NotARM strht_t  ;
: strt	NotARM strt_t  ;		\ ARM too!
: sub	Thumb? if sub_t else sub_a endif  ;
: subx	Thumb? if sub_t else sub_a endif  ;
: svc	NotARM svc_t  ;
: swi	Thumb? if swi_t else swi_a endif  ;
: swp	NotThumb SWP_a  ;
: swpb	NotThumb SWPB_a  ;

: sxtab		NotARM sxtab_t  ;
: sxtab16	NotARM sxtab16_t  ;
: sxtah		NotARM sxtah_t  ;

: sxtb	NotARM sxtb_t  ;
: sxtb16	NotARM sxtb16_t  ;
: sxth	NotARM sxth_t  ;

: tbb	NotARM tbb_t  ;
: tbh	NotARM tbh_t  ;
: teq	Thumb? if teq_t else teq_a endif  ;
: tst	Thumb? if tst_t else tst_a endif  ;

: uadd16	NotARM uadd16_t  ;
: uadd8		NotARM uadd8_t  ;
: uasx		NotARM uasx_t  ;

: ubfx	NotARM ubfx_t  ;
: udiv	NotARM udiv_t  ;

: uhadd16	NotARM uhadd16_t  ;
: uhadd8	NotARM uahdd8_t  ;
: uhasx		NotARM uhasx_t  ;
: uhsax		NotARM uhsax_t  ;
: uhsub16	NotARM uhsub16_t  ;
: uhsub8	NotARM uhsub8_t  ;
: umaal		NotARM umaal_t  ;

: umlal	Thumb? if umlal_t else umlal_a endif  ;
: umull	Thumb? if umull_t else umull_a endif  ;

: uqadd16	NotARM uqadd16_t  ;
: uqadd8	NotARM uqadd8_t  ;
: uqasx		NotARM uqasx_t  ;
: uqsax		NotARM uqsax_t  ;
: uqsub16	NotARM uqsub16_t  ;
: uqsub8	NotARM uqsub8_t  ;
: usad8		NotARM usad8_t  ;
: usada8	NotARM usada8_t  ;

: usat		NotARM usat_t  ;

: usat16	NotARM usat16_t  ;
: usax		NotARM usax_t  ;
: usub16	NotARM usub16_t  ;
: usub8		NotARM usub8_t  ;
: uxtab		NotARM uxtab_t  ;
: uxtab16	NotARM uxtab16_t  ;
: uxtah		NotARM uxtah_t  ;
: uxtb16	NotARM uxtb16_t  ;

: uxtb		NotARM uxtb_t  ;
: uxth		NotARM uxth_t  ;

: vabs		Thumb? if vabs_t else vabs_a endif  ;
: vadd		Thumb? if vadd_t else vadd_a endif  ;
: vcmp		Thumb? if vcmp_t else vcmp_a endif  ;
: vcmpe		Thumb? if vcmpe_t else vcmpe_a endif  ;
: vcvt		Thumb? if VCVT_t else VCVT_a endif  ;
: vcvtr		Thumb? if VCVTR_t else VCVTR_a endif  ;
: vcvtb		Thumb? if VCVTB_t else VCVTB_a endif  ;
: vcvtt		Thumb? if VCVTT_t else VCVTT_a endif  ;

: vdiv		Thumb? if vdiv_t else vdiv_a endif  ;

: vfma		Thumb? if vfma_t else vfma_a endif  ;
: vfms		Thumb? if vfms_t else vfms_a endif  ;
: vfnma		Thumb? if vfnma_t else vfnma_a endif  ;
: vfnms		Thumb? if vfnms_t else vfnms_a endif  ;

: vldmia	Thumb? if vldmia_t else vldmia_a endif  ;
: vldmdb	Thumb? if vldmdb_t else vldmdb_a endif  ;
: vldr		Thumb? if vldr_t else vldr_a endif  ;

: vmla		Thumb? if vmla_t else vmla_a endif  ;
: vmls		Thumb? if vmls_t else vmls_a endif  ;
: vmov		0 ?prefix drop  doVMOV  ;
$EEF1:0A10 $0EF1:0A10 vmrs_ins vmrs
$EEE1:0A10 $0EE1:0A10 vmrs_ins vmsr
: vmul		Thumb? if vmul_t else vmul_a endif  ;

: vneg		Thumb? if vneg_t else vneg_a endif  ;
: vnmla		Thumb? if vnmla_t else vnmla_a endif  ;
: vnmls		Thumb? if vnmls_t else vnmls_a endif  ;
: vnmul		Thumb? if vnmul_t else vnmul_a endif  ;

: vpop		Thumb? if vpop_t else vpop_a endif  ;
: vpush		Thumb? if vpush_t else vpush_a endif  ;

: vsqrt		Thumb? if vsqrt_t else vsqrt_a endif  ;
: vstmia	Thumb? if vstmia_t else vstmia_a endif  ;
: vstmdb	Thumb? if vstmdb_t else vstmdb_a endif  ;
: vstr		Thumb? if vstr_t else vstr_a endif  ;
: vsub		Thumb? if vsub_t else vsub_a endif  ;

: wfe	NotARM wfe_t  ;
: wfi	NotARM wfi_t  ;
: yield	NotARM yield_t  ;


\ **************
\ Register lists
\ **************

: {             \ -- ; register list ; N.B. overrides { for locals
  state @ if				\ we're inside a macro
    reg-count @  get-list  reg-count !
    ldm-mask @off postpone literal	\ compile mask as literal
    postpone store-mask
\    compile lit  ldm-mask @off ,        \ literal value of mask
\    compile store-mask                  \ set at macro run-time
  else
    get-list
  endif
; immediate

: {{		\ -- #regs ; alternate register list notation
  reg-count @  ldm-mask off  ;

: }}		\ #regs -- ; alternate register list notation
  set-list  ;

: ^             \ -- ; status/User flag for LDM/STM
  <s22> on  ;

\ ******
\ Shifts
\ ******

0 T2ShiftOp .LSL
1 T2ShiftOp .LSR
2 T2ShiftOp .ASR
3 T2ShiftOp .ROR
#11 T2ShiftOp .RRX

: .ie		\ -- ; for CPS
  $0010 <cpsMask> bic!  ;
: .id		\ -- ; for CPS
  $0010 <cpsMask> or!  ;
: .i		\ -- ; for CPS
  $0002 <cpsMask> or!  ;
: .f		\ -- ; for CPS
  Cortex-M0/M1? err_cputype ?a-error
  $0001 <cpsMask> or!  ;

\ *************
\ Miscellaneous
\ *************

: ++            \ -- ; marks second register coming, sets +/- flag
  reg2 on  <u> on  ;

: --            \ -- ; marks second register coming, clears +/- flag
  reg2 on  <u> off  ;

: [             \ -- ; start indexed operation - really just syntactic sugar
  indexing on  <p> on  ;
: ]             \ -- ; pre indexed operation
  indexing on  <p> on  ;
: ],            \ -- ; post indexed operation
  indexing on  <p> off
  Thumb2? if  <w> on  endif
;

: x[             \ -- ; start indexed operation - really just syntactic sugar
  indexing on  <p> on  ;
: x]             \ -- ; pre indexed operation
  indexing on  <p> on  ;
: x],            \ -- ; post indexed operation
  indexing on  <p> off
  Thumb2? if  <w> on  endif
;

: {}		\ -- ; option mode for T2 LDCx MRRCx
  <option?> on  ;

: dxb            \ b -- ; lay byte
\ *G Lay an 8-bit byte into the instruction stream.
\ ** No alignment is performed. Use in the form:
\ *C   dxb $55
  0 ?prefix drop  c,c(t) reset  ;
: dxw            \ w -- ; lay 16 bits
\ *G Lay a 16-bit word into the instruction stream.
\ ** No alignment is performed. Use in the form:
\ *C   dxw $55AA
  0 ?prefix drop  w,c(t) reset  ;
: dxl            \ l -- ; lay 32 bit long
\ *G Lay a 32-bit dword into the instruction stream.
\ ** No alignment is performed. Use in the form:
\ *C   dxl $11223344
  0 ?prefix drop  absolute  reset  ;

: db            \ b -- ; lay byte
\ *G Lay a single byte inline.
\ ** *\b{Obsolete}, will be removed in a future release,
\ ** use *\fo{DXB} instead.
  0 ?prefix drop  c,c(t) a-reset  ;
: dw            \ b -- ; lay 16 bits
\ *G Lay a 16 bit item inline. No alignment is performed.
\ ** *\b{Obsolete}, will be removed in a future release,
\ ** use *\fo{DXW} instead.
  0 ?prefix drop  w,c(t) a-reset  ;
: dd            \ l -- ; lay 32 bit long
\ *G Lay a 32 bit item inline. No alignment is performed.
\ ** *\b{Obsolete}, will be removed in a future release,
\ ** use *\fo{DXL} instead.
  0 ?prefix drop  l,c(t) a-reset  ;
: dl            \ l -- ; lay 32 bit long
\ *G Lay a 32 bit item inline. No alignment is performed.
\ ** *\b{Obsolete}, will be removed in a future release,
\ ** use *\fo{DXL} instead.
  0 ?prefix drop  l,c(t) a-reset  ;

: align4	\ --
\ *G Forces the PC to a four byte boundary.
  0 ?prefix drop  calign(t)  ;
(( breaks, why? because it finds the assembler AND.
  Thumb? if
    chere(t) 2 and			\ <<<< wrong AND
    if  $46C0 Topc16,  endif
  else
    calign(t)
  endif
;
))

(( for DocGen - defined in PreAsm.fth
: $             \ -- chere
\ *G Returns the current value of the PC.
  chere(t)  ;
))

: !             \ -- ; set <w> flag for writeback
  <w> on  ;

: xx!           \ -- ; set <w> flag for writeback, used in optimiser
  <w> on  ;

: .regs         \ -- ; display register array
  cr reg-count ?  reg-array $20 dump  ;

((
: r"            \ -- ; lays down a string of registers and evaluate at run time
  [interp] c"
  compile count
  compile also
  compile asm-access
  compile evaluate
  compile previous
; immediate
))

previous definitions

also c-c definitions

: fix_T2branch32	\ dest ^ins --
\ Fix up a T2 wide branch. The opcode is at ^ins and the target
\ is dest.
  case  dup T2x32@(t) $F800:D000 and
    $F000:8000 of  fix_Trel20  endof	\ A6-40 T3, B .cc
    $F000:9000 of  fix_Trel24  endof	\ A6-40 T4, B
    $F000:D000 of  fix_Trel24  endof	\ A6-49 T1, BL
      dup cr .dword space  T2x32@(t) .dword
      err_BadBranchOpcode throw
  endcase
;

: fix_T2branch	\ ^ins dest --
\ Fix up a T2 branch. The opcode is at ^ins and the target
\ is dest.
  swap					\ -- dest orig
  case  dup w@c(t) $F800 and
    $D000 of  fix_Trel8  endof		\ A6-40 T1
    $D800 of  fix_Trel8  endof		\ A6-40 T1
    $E000 of  fix_Trel11  endof		\ A6-40 T2
    $F000 of  fix_T2branch32  endof	\ A6-40 T3/4
      dup cr .dword space  w@c(t) .word
      err_BadBranchOpcode throw
  endcase
;

: fix_ARMbranch	\ ^ins dest --
\ Fix up an ARM branch. The opcode is at ^ins and the target
\ is dest.
  swap fix_r24  ;

: fix_ATbranch	\ sbrid --
\ Fix up an ARM or Thumb branch. The opcode is at ^ins and the
\ target is dest.
  Thumb? if
    chere(t) fix_T2branch
  else
    chere(t) fix_ARMbranch
  endif
;

previous definitions

: con-bra       \ -- ; compile conditional branch
  cc-set @ 0= err_nocc ?throw		\ check condition code was set
  Thumb? if
    $D0FF +TBcond $0100 xor w,c(t)	\ b .n .cc  condition reversed
  else
    $0A00:0000  +cond $1000:0000 xor	\ B opcode + cond, invert condition
    l,c(t)                              \ lay opcode
  endif
;

: uncon-bra     \ -- ; compile unconditional branch
  Thumb? if
    $E7FF w,c(t)			\ b .n
  else
    $EA000000 l,c(t)
  endif
;

: fix-bra	\ dest ^ins --
  Thumb? if
    swap fix_T2branch
  else
    fix_r24
  endif
;


\ assembler structures

also asm-access definitions

\ SFP021...  ANS version
: ahere         \ -- addr                                                       \ MSD006
  null-op chere(t) ;                                                            \ MSD006

: ahead,	\ -- orig
  null-op  chere(t) uncon-bra  ;
: if,           \ -- orig
  null-op  chere(t) con-bra  ;
: endif,        \ orig --
  null-op  chere(t) swap fix-bra  ;
: then,         \ orig --
  null-op  chere(t) swap fix-bra  ;
: else,         \ orig1 -- orig2
  null-op  chere(t) uncon-bra  chere(t) rot fix-bra  ;

: begin,        \ -- dest
  null-op  chere(t)  ;
: until,        \ dest --
  null-op  chere(t) con-bra fix-bra  ;
: again,        \ dest --
  null-op  chere(t) uncon-bra fix-bra  ;

: while,        \ dest -- orig dest
  null-op  chere(t) swap con-bra  ;
: repeat,       \ orig dest --
  null-op
  chere(t) uncon-bra  fix-bra	\ -- orig
  chere(t) swap fix-bra  ;
\ ...SFP021

previous definitions


\ =============================
\ Tool and disassembler support
\ =============================

only forth also c-c definitions  also assembler

: printmvl	\ -- ; diagnostic
  base @  decimal
  cr
  cr ." in ARM32 code:"
  cr ." MVL generates one instruction " mvl1total @ .  ." times."
  cr ." MVL generates more than one instruction " mvltotal @ .  ." times,"
  cr ." generating " mvlcount @ 4/ . ." instructions, "
  ." averaging " mvlcount @ 4/	#10 * mvltotal @ /
  0 <# # [char] . hold # #> type ."  instructions per MVL,"
  cr ." not counting single instruction MVLs."
  cr
  cr ." The ADR instruction was used " adrtotal @ .  ." times."
  cr
  base !
;

: >callDestARM	\ ^ins -- dest
\ Given the address of a B/BL opcode, return the branch target address.
  dup l@c(t)
  $00FF:FFFF and  2 lshift		\ Extract offset
  + 8 +					\ Convert to absolute address
;

: >callDestT2	\ ^ins -- dest
\ Given the address of a B/BL opcode, return the branch target address.
  dup T2x32@(t) offT2rel25 + 4 +  ;

: >Lit138T2	\ ^ins -- lit12
\ Given the address of an instruction, extract its 12 bit literal.
  T2x32@(t) decodeLit138  ;

: >Const138T2	\ ^ins -- lit12
\ Given the address of an instruction, extract its 12 bit constant.
  T2x32@(t) decodeConst138  ;

: >Lit16	\ ^ins -- imm16
\ Given the address of a MOV or MOVT instruction, extract its 16 bit literal.
  T2x32@(t) decodeImm16  ;

: >Lit32	\ ^ins -- imm32
\ Given the address of a MOV/MOVT instruction pair, extract its 32 bit literal.
  dup >lit16  swap 4 + >lit16 #16 lshift or  ;

: >Dest32	\ ^ins -- imm32
\ Given the address of a MOV/MOVT instruction pair, extract its
\ 32 bit literal as a branch target, clearing bit 0.
  >Lit32 -2 and  ;

: patchT1Imm7*4	\ lit7*4 ^ins --
  over Imm7*4? 0= err_SP7 ?throw
  dup w@c(t)				\ -- lit7*4 addr opc
  $FF80 and  rot 2 rshift $7F and or	\ -- addr opc'
  swap w!c(t)
;

: patchT2Imm138	\ lit12 ^ins --
\ Patch the literal field of the instruction.
  dup T2x32@(t) ~Lit138mask and		\ -- lit12 addr opc
  rot enc12>138 Lit138mask and or	\ -- addr opc'
  swap T2x32!(t)
;

: patchT2Const138	\ lit12 ^ins --
\ Patch the literal field of the instruction.
  swap T2Const12 swap patchT2Imm138  ;

(( Register models
\ Previous ARM model
0A reg TOS      0A reg TOS,
09 reg LP	09 reg LP,
0B reg UP       0B reg UP,
0C reg PSP      0C reg PSP,
0D reg RSP      0D reg RSP,

\ Current Cortex/ARM model
$07 reg TOS     $07 reg TOS,
$09 reg LP	$09 reg LP,
$0B reg UP      $0B reg UP,
$0C reg PSP     $0C reg PSP,
$0D reg RSP     $0D reg RSP,

Registers are defined using CREATE , DOES> and the code below
assumes that
  ' <foo> >BODY !
will patch the register number/mode for register <foo>.
))

: markReg	\ x -- x'
\ Convert a register number to the form used by the assembler.
  $FF and gpr-type or  ;

: ['A]		\ "name" --
\ perform ['] on a word in the ASM-ACCESS vocabulary.
  also asm-access  postpone [']  previous  ; immediate

: set-LPreg	\ n -- ; select LP register
  markReg dup ['A] LP >body !  ['A] LP, >body !  ;
: set-TOSreg	\ n -- ; select TOS register
  markReg dup ['A] TOS >body !  ['A] TOS, >body !  ;
: set-UPreg	\ n -- ; select UP register
  markReg dup ['A] UP >body !  ['A] UP, >body !  ;
: set-PSPreg	\ n -- ; select PSP register
  markReg dup ['A] PSP >body !  ['A] PSP, >body !  ;
: set-RSPreg	\ n -- ; select RSP register
  markReg dup ['A] RSP >body !  ['A] RSP, >body !  ;

: get-tos	\ -- n ; get register used as TOS
  ['A] TOS >body @ $FF and  ;
: get-psp	\ -- n ; get register used as PSP
  ['A] PSP >body @ $FF and  ;
: get-rsp	\ -- n ; get register used as RSP
  ['A] RSP >body @ $FF and  ;

only forth  also c-c definitions

decimal

((
: ;CODE		\ --
\ *G A defining word used in the form:
\ *C 	: <namex>  CREATE ... ;CODE ... END-CODE
\ *P Stops compilation, and enables the assembler. This word is used
\ ** with *\fo{CREATE} to produce defining words whose run-time portion is
\ ** written in code, in the same way that *\fo{CREATE ... DOES>}
\ ** is used to create high level defining words .
\ *P The data structure is defined between *\fo{CREATE} and *\fo{;CODE}
\ ** and the run-time action is defined between *\fo{;CODE} and
\ ** *\fo{END-CODE}. The current value of the data stack pointer
\ ** is saved by *\fo{;CODE} for later use by *\fo{END-CODE} for
\ ** error checking. When *\fo{<namex>} executes the address of
\ ** the data area will be found on the processor stack, from which it
\ ** must be removed.

: ASMCODE	\ --
\ *G Starts a section of assembler code and turns on the assembler, but
\ ** without generating a dictionary header. This action is
\ ** particularly useful for generating the start-up code. Examples of
\ ** this can be found in *\i{CODEARM.FTH}.
\ *C   ASMCODE ... END-CODE

: CODE		\ --
\ *G A defining word used in the form:
\ *C  CODE <name> ... END-CODE
\ *P Creates a dictionary entry for <name> to be defined by a following
\ ** sequence of assembly language words. Words thus defined are called
\ ** code definitions. *\fo{CODE} stores the current data stack pointer for
\ ** later error checking by *\fo{END-CODE}.

: END-CODE	\ --
\ *G Terminates a code definition and checks the data stack pointer
\ ** against the value stored when *\fo{;CODE} or *\fo{CODE} is
\ ** executed. The assembler is disabled. See: *\fo{CODE} *\fo{;CODE}.

: IS-ACTION-OF	\ addr --
\ *G Used to tell the cross-compiler that the given address is to be
\ ** used as the run time action of the word whose name follows.
\ ** Usually found in code definitions, but can also be used for
\ ** high-level definitions. For example:
\ *E 	ASMCODE
\ ** 	HERE IS-ACTION-OF CONSTANT
\ ** 	  ...
\ ** 	END-CODE
\ ** 	ASSEMBLER
\ ** 	HERE IS-ACTION-OF <<high-level-definer>>
\ ** 	B DODOES
\ ** 	END-CODE
\ ** 	] ... EXIT [

: PROC		\ -- ; PROC <label-name>
\ *G Starts a section of assembler code and turns on the assembler,
\ ** defining a label. This action is particularly useful for
\ ** generating interrupts or shared subroutines. Examples of
\ ** this can be found in *\i{Cortex/CodeCortex.fth}.
))

also cc/i
: !call		\ dest ^ins --
\ *G Patch a branch opcode at *\i{^ins} to branch to target
\ ** address *\i{dest}. The opcode portion is not changed, so
\ ** that this word works with both B and BL. Note that this
\ ** word may be redefined in some target code for ARM7 and ARM9
\ ** devices.
  Thumb? if
    swap fix_T2branch
  else
    [+assembler] fix_r24 [previous]
  endif
;
previous definitions


\ ======
\ *> ###
\ ======
